--
-- PostgreSQL database dump
--

-- Dumped from database version 16.1
-- Dumped by pg_dump version 16.1

SET statement_timeout = 0;
SET lock_timeout = 0;
SET idle_in_transaction_session_timeout = 0;
SET client_encoding = 'UTF8';
SET standard_conforming_strings = on;
SELECT pg_catalog.set_config('search_path', '', false);
SET check_function_bodies = false;
SET xmloption = content;
SET client_min_messages = warning;
SET row_security = off;

--
-- Name: bloom_lims; Type: SCHEMA; Schema: -; Owner: daylily
--

CREATE SCHEMA bloom_lims;


ALTER SCHEMA bloom_lims OWNER TO daylily;

--
-- Name: pgcrypto; Type: EXTENSION; Schema: -; Owner: -
--

CREATE EXTENSION IF NOT EXISTS pgcrypto WITH SCHEMA public;


--
-- Name: EXTENSION pgcrypto; Type: COMMENT; Schema: -; Owner: 
--

COMMENT ON EXTENSION pgcrypto IS 'cryptographic functions';


--
-- Name: create_audit_triggers_for_tables(text[]); Type: FUNCTION; Schema: public; Owner: daylily
--

CREATE FUNCTION public.create_audit_triggers_for_tables(tables text[]) RETURNS void
    LANGUAGE plpgsql
    AS $$
DECLARE
    table_name TEXT;
BEGIN
    FOREACH table_name IN ARRAY tables LOOP
        -- Construct the CREATE TRIGGER statement for UPDATE
        EXECUTE format(
            'CREATE TRIGGER %I_audit_trigger_after_update AFTER UPDATE ON %I FOR EACH ROW EXECUTE FUNCTION record_update();',
            table_name, table_name
        );

        -- Construct the CREATE TRIGGER statement for DELETE
        EXECUTE format(
            'CREATE TRIGGER %I_audit_trigger_after_delete AFTER DELETE ON %I FOR EACH ROW EXECUTE FUNCTION record_delete();',
            table_name, table_name
        );
    END LOOP;
END;
$$;


ALTER FUNCTION public.create_audit_triggers_for_tables(tables text[]) OWNER TO daylily;

--
-- Name: create_insert_audit_triggers_for_tables(text[]); Type: FUNCTION; Schema: public; Owner: daylily
--

CREATE FUNCTION public.create_insert_audit_triggers_for_tables(tables text[]) RETURNS void
    LANGUAGE plpgsql
    AS $$
DECLARE
    table_name TEXT;
BEGIN
    FOREACH table_name IN ARRAY tables LOOP
        -- Construct the CREATE TRIGGER statement for INSERT
        EXECUTE format(
            'CREATE TRIGGER %I_audit_trigger_after_insert AFTER INSERT ON %I FOR EACH ROW EXECUTE FUNCTION record_insert();',
            table_name, table_name
        );
    END LOOP;
END;
$$;


ALTER FUNCTION public.create_insert_audit_triggers_for_tables(tables text[]) OWNER TO daylily;

--
-- Name: record_delete(); Type: FUNCTION; Schema: public; Owner: daylily
--

CREATE FUNCTION public.record_delete() RETURNS trigger
    LANGUAGE plpgsql
    AS $$
DECLARE
    app_username TEXT;
BEGIN
    BEGIN
        app_username := current_setting('session.current_username', true);
    EXCEPTION WHEN OTHERS THEN
        app_username := current_user;
    END;
    
    INSERT INTO audit_log (rel_table_name, rel_table_uuid_fk, rel_table_euid_fk, changed_by, operation_type)
    VALUES (TG_TABLE_NAME, OLD.uuid, OLD.euid, app_username, 'DELETE');
    RETURN NULL;
END;
$$;


ALTER FUNCTION public.record_delete() OWNER TO daylily;

--
-- Name: record_insert(); Type: FUNCTION; Schema: public; Owner: daylily
--

CREATE FUNCTION public.record_insert() RETURNS trigger
    LANGUAGE plpgsql
    AS $$
DECLARE
    app_username TEXT;
BEGIN
    BEGIN
        app_username := current_setting('session.current_username', true);
    EXCEPTION WHEN OTHERS THEN
        app_username := current_user;
    END;

    -- Insert record into audit_log
    INSERT INTO audit_log (rel_table_name, rel_table_uuid_fk, rel_table_euid_fk, changed_by, operation_type)
    VALUES (TG_TABLE_NAME, NEW.uuid, NEW.euid, app_username, 'INSERT');

    RETURN NEW;
END;
$$;


ALTER FUNCTION public.record_insert() OWNER TO daylily;

--
-- Name: record_update(); Type: FUNCTION; Schema: public; Owner: daylily
--

CREATE FUNCTION public.record_update() RETURNS trigger
    LANGUAGE plpgsql
    AS $_$
DECLARE
    r RECORD;
    column_name TEXT;
    old_value TEXT;
    new_value TEXT;
    app_username TEXT;
BEGIN
    BEGIN
        app_username := current_setting('session.current_username', true);
    EXCEPTION WHEN OTHERS THEN
        app_username := current_user;
    END;

    FOR r IN SELECT * FROM json_each_text(row_to_json(NEW)) LOOP
        column_name := r.key;
        new_value := r.value;
        EXECUTE format('SELECT ($1).%I', column_name) USING OLD INTO old_value;

        IF old_value IS DISTINCT FROM new_value THEN
            INSERT INTO audit_log (rel_table_name, column_name, old_value, new_value, changed_by, rel_table_uuid_fk, rel_table_euid_fk, operation_type)
            VALUES (TG_TABLE_NAME, column_name, old_value, new_value, app_username, NEW.uuid, NEW.euid, TG_OP);
        END IF;
    END LOOP;
    
    RETURN NEW;
END;
$_$;


ALTER FUNCTION public.record_update() OWNER TO daylily;

--
-- Name: set_generic_instance_euid(); Type: FUNCTION; Schema: public; Owner: daylily
--

CREATE FUNCTION public.set_generic_instance_euid() RETURNS trigger
    LANGUAGE plpgsql
    AS $$
DECLARE
    prefix TEXT;
    sequence_name TEXT;
BEGIN
    -- Fetch the instance_prefix from generic_template
    SELECT instance_prefix INTO prefix FROM generic_template WHERE uuid = NEW.template_uuid;
    -- Set the euid of the new generic_instance
    sequence_name := CASE
            WHEN prefix = 'CX' THEN nextval('cx_instance_seq')
            WHEN prefix = 'MX' THEN nextval('mx_instance_seq')
            WHEN prefix = 'DX' THEN nextval('dx_instance_seq')
            WHEN prefix = 'AX' THEN nextval('ax_instance_seq')
            WHEN prefix = 'WX' THEN nextval('wx_instance_seq')
            WHEN prefix = 'WSX' THEN nextval('wsx_instance_seq')
            WHEN prefix = 'TRX' THEN nextval('trx_instance_seq')
            WHEN prefix = 'EX' THEN nextval('ex_instance_seq')
            WHEN prefix = 'GX' THEN nextval('gx_instance_seq')
            WHEN prefix = 'CWX' THEN nextval('cwx_instance_seq')
            WHEN prefix = 'AY' THEN nextval('ay_instance_seq')
            WHEN prefix = 'XX' THEN nextval('xx_instance_seq')
            WHEN prefix = 'QX' THEN nextval('wsq_instance_seq')
            WHEN prefix = 'MRX' THEN nextval('mrxq_instance_seq')
            WHEN prefix = 'MCX' THEN nextval('mcxq_instance_seq')
            WHEN prefix = 'FX' THEN nextval('fx_instance_seq')
            WHEN prefix = 'FI' THEN nextval('fi_instance_seq')
            WHEN prefix = 'FS' THEN nextval('fs_instance_seq')
            WHEN prefix = 'EV' THEN nextval('ev_instance_seq')

            -- Add more cases for other prefixes
            ELSE nextval('generic_instance_seq') -- Default sequence
        END;

    NEW.euid := prefix || sequence_name;
    RETURN NEW;
END;
$$;


ALTER FUNCTION public.set_generic_instance_euid() OWNER TO daylily;

--
-- Name: soft_delete_row(); Type: FUNCTION; Schema: public; Owner: daylily
--

CREATE FUNCTION public.soft_delete_row() RETURNS trigger
    LANGUAGE plpgsql
    AS $_$
DECLARE
    jsonData JSON;
    app_username TEXT;
BEGIN
    BEGIN
        app_username := current_setting('session.current_username', true);
    EXCEPTION WHEN OTHERS THEN
        app_username := current_user;
    END;
    -- Check if the 'is_deleted' column exists in the table
    IF EXISTS(SELECT 1 FROM information_schema.columns 
              WHERE table_name = TG_TABLE_NAME 
              AND column_name = 'is_deleted') THEN
        -- Set 'is_deleted' to TRUE for the row identified by UUID
        EXECUTE format('UPDATE %I SET is_deleted = TRUE WHERE uuid = $1', TG_TABLE_NAME) USING OLD.uuid;

        jsonData := row_to_json(OLD); -- Convert the OLD record to JSON

        INSERT INTO audit_log (rel_table_name, rel_table_uuid_fk, rel_table_euid_fk, changed_by, operation_type, deleted_record_json)
        VALUES (TG_TABLE_NAME, OLD.uuid, OLD.euid, app_username, 'DELETE', jsonData);
    
        RETURN NULL; -- Cancel the delete operation
    ELSE
        -- If 'is_deleted' column doesn't exist, continue with the delete
        RETURN OLD;
    END IF;
END;
$_$;


ALTER FUNCTION public.soft_delete_row() OWNER TO daylily;

--
-- Name: update_modified_dt(); Type: FUNCTION; Schema: public; Owner: daylily
--

CREATE FUNCTION public.update_modified_dt() RETURNS trigger
    LANGUAGE plpgsql
    AS $$
BEGIN
    NEW.modified_dt = CURRENT_TIMESTAMP;
    RETURN NEW;
END;
$$;


ALTER FUNCTION public.update_modified_dt() OWNER TO daylily;

SET default_tablespace = '';

SET default_table_access_method = heap;

--
-- Name: audit_log; Type: TABLE; Schema: public; Owner: daylily
--

CREATE TABLE public.audit_log (
    uuid uuid DEFAULT gen_random_uuid() NOT NULL,
    rel_table_name text NOT NULL,
    column_name text,
    rel_table_uuid_fk uuid NOT NULL,
    rel_table_euid_fk text NOT NULL,
    old_value text,
    new_value text,
    changed_by text,
    changed_at timestamp with time zone DEFAULT CURRENT_TIMESTAMP NOT NULL,
    operation_type text,
    json_addl jsonb,
    super_type text,
    deleted_record_json jsonb,
    is_deleted boolean DEFAULT false NOT NULL,
    is_singleton boolean DEFAULT false NOT NULL,
    CONSTRAINT audit_log_operation_type_check CHECK ((operation_type = ANY (ARRAY['INSERT'::text, 'UPDATE'::text, 'DELETE'::text])))
);


ALTER TABLE public.audit_log OWNER TO daylily;

--
-- Name: audit_log_seq; Type: SEQUENCE; Schema: public; Owner: daylily
--

CREATE SEQUENCE public.audit_log_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


ALTER SEQUENCE public.audit_log_seq OWNER TO daylily;

--
-- Name: ax_instance_seq; Type: SEQUENCE; Schema: public; Owner: daylily
--

CREATE SEQUENCE public.ax_instance_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


ALTER SEQUENCE public.ax_instance_seq OWNER TO daylily;

--
-- Name: ay_instance_seq; Type: SEQUENCE; Schema: public; Owner: daylily
--

CREATE SEQUENCE public.ay_instance_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


ALTER SEQUENCE public.ay_instance_seq OWNER TO daylily;

--
-- Name: cwx_instance_seq; Type: SEQUENCE; Schema: public; Owner: daylily
--

CREATE SEQUENCE public.cwx_instance_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


ALTER SEQUENCE public.cwx_instance_seq OWNER TO daylily;

--
-- Name: cx_instance_seq; Type: SEQUENCE; Schema: public; Owner: daylily
--

CREATE SEQUENCE public.cx_instance_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


ALTER SEQUENCE public.cx_instance_seq OWNER TO daylily;

--
-- Name: dx_instance_seq; Type: SEQUENCE; Schema: public; Owner: daylily
--

CREATE SEQUENCE public.dx_instance_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


ALTER SEQUENCE public.dx_instance_seq OWNER TO daylily;

--
-- Name: ev_instance_seq; Type: SEQUENCE; Schema: public; Owner: daylily
--

CREATE SEQUENCE public.ev_instance_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


ALTER SEQUENCE public.ev_instance_seq OWNER TO daylily;

--
-- Name: ex_instance_seq; Type: SEQUENCE; Schema: public; Owner: daylily
--

CREATE SEQUENCE public.ex_instance_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


ALTER SEQUENCE public.ex_instance_seq OWNER TO daylily;

--
-- Name: fi_instance_seq; Type: SEQUENCE; Schema: public; Owner: daylily
--

CREATE SEQUENCE public.fi_instance_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


ALTER SEQUENCE public.fi_instance_seq OWNER TO daylily;

--
-- Name: fs_instance_seq; Type: SEQUENCE; Schema: public; Owner: daylily
--

CREATE SEQUENCE public.fs_instance_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


ALTER SEQUENCE public.fs_instance_seq OWNER TO daylily;

--
-- Name: fx_instance_seq; Type: SEQUENCE; Schema: public; Owner: daylily
--

CREATE SEQUENCE public.fx_instance_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


ALTER SEQUENCE public.fx_instance_seq OWNER TO daylily;

--
-- Name: generic_instance; Type: TABLE; Schema: public; Owner: daylily
--

CREATE TABLE public.generic_instance (
    uuid uuid DEFAULT gen_random_uuid() NOT NULL,
    euid text,
    name text NOT NULL,
    json_addl jsonb NOT NULL,
    btype text NOT NULL,
    polymorphic_discriminator text NOT NULL,
    super_type text NOT NULL,
    b_sub_type text NOT NULL,
    created_dt timestamp with time zone DEFAULT CURRENT_TIMESTAMP NOT NULL,
    version text NOT NULL,
    bstatus text NOT NULL,
    is_deleted boolean DEFAULT false NOT NULL,
    template_uuid uuid NOT NULL,
    modified_dt timestamp with time zone DEFAULT CURRENT_TIMESTAMP,
    is_singleton boolean DEFAULT false NOT NULL
);


ALTER TABLE public.generic_instance OWNER TO daylily;

--
-- Name: generic_instance_lineage_seq; Type: SEQUENCE; Schema: public; Owner: daylily
--

CREATE SEQUENCE public.generic_instance_lineage_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


ALTER SEQUENCE public.generic_instance_lineage_seq OWNER TO daylily;

--
-- Name: generic_instance_lineage; Type: TABLE; Schema: public; Owner: daylily
--

CREATE TABLE public.generic_instance_lineage (
    uuid uuid DEFAULT gen_random_uuid() NOT NULL,
    euid text DEFAULT ('GL'::text || nextval('public.generic_instance_lineage_seq'::regclass)) NOT NULL,
    parent_instance_uuid uuid NOT NULL,
    parent_type text NOT NULL,
    child_type text NOT NULL,
    polymorphic_discriminator text NOT NULL,
    child_instance_uuid uuid NOT NULL,
    created_dt timestamp with time zone DEFAULT CURRENT_TIMESTAMP NOT NULL,
    name text NOT NULL,
    super_type text NOT NULL,
    btype text,
    b_sub_type text,
    json_addl jsonb,
    bstatus text,
    version text,
    is_deleted boolean DEFAULT false NOT NULL,
    modified_dt timestamp with time zone DEFAULT CURRENT_TIMESTAMP,
    is_singleton boolean DEFAULT false NOT NULL,
    relationship_type text
);


ALTER TABLE public.generic_instance_lineage OWNER TO daylily;

--
-- Name: generic_instance_seq; Type: SEQUENCE; Schema: public; Owner: daylily
--

CREATE SEQUENCE public.generic_instance_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


ALTER SEQUENCE public.generic_instance_seq OWNER TO daylily;

--
-- Name: generic_template_seq; Type: SEQUENCE; Schema: public; Owner: daylily
--

CREATE SEQUENCE public.generic_template_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


ALTER SEQUENCE public.generic_template_seq OWNER TO daylily;

--
-- Name: generic_template; Type: TABLE; Schema: public; Owner: daylily
--

CREATE TABLE public.generic_template (
    uuid uuid DEFAULT gen_random_uuid() NOT NULL,
    euid text DEFAULT ('GT'::text || nextval('public.generic_template_seq'::regclass)) NOT NULL,
    name text NOT NULL,
    instance_prefix text NOT NULL,
    polymorphic_discriminator text NOT NULL,
    super_type text NOT NULL,
    btype text NOT NULL,
    b_sub_type text NOT NULL,
    created_dt timestamp with time zone DEFAULT CURRENT_TIMESTAMP NOT NULL,
    version text NOT NULL,
    json_addl jsonb NOT NULL,
    json_addl_schema jsonb,
    bstatus text NOT NULL,
    is_deleted boolean DEFAULT false NOT NULL,
    modified_dt timestamp with time zone DEFAULT CURRENT_TIMESTAMP,
    is_singleton boolean DEFAULT true NOT NULL
);


ALTER TABLE public.generic_template OWNER TO daylily;

--
-- Name: gx_instance_seq; Type: SEQUENCE; Schema: public; Owner: daylily
--

CREATE SEQUENCE public.gx_instance_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


ALTER SEQUENCE public.gx_instance_seq OWNER TO daylily;

--
-- Name: mcxq_instance_seq; Type: SEQUENCE; Schema: public; Owner: daylily
--

CREATE SEQUENCE public.mcxq_instance_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


ALTER SEQUENCE public.mcxq_instance_seq OWNER TO daylily;

--
-- Name: mrxq_instance_seq; Type: SEQUENCE; Schema: public; Owner: daylily
--

CREATE SEQUENCE public.mrxq_instance_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


ALTER SEQUENCE public.mrxq_instance_seq OWNER TO daylily;

--
-- Name: mx_instance_seq; Type: SEQUENCE; Schema: public; Owner: daylily
--

CREATE SEQUENCE public.mx_instance_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


ALTER SEQUENCE public.mx_instance_seq OWNER TO daylily;

--
-- Name: trx_instance_seq; Type: SEQUENCE; Schema: public; Owner: daylily
--

CREATE SEQUENCE public.trx_instance_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


ALTER SEQUENCE public.trx_instance_seq OWNER TO daylily;

--
-- Name: wsq_instance_seq; Type: SEQUENCE; Schema: public; Owner: daylily
--

CREATE SEQUENCE public.wsq_instance_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


ALTER SEQUENCE public.wsq_instance_seq OWNER TO daylily;

--
-- Name: wsx_instance_seq; Type: SEQUENCE; Schema: public; Owner: daylily
--

CREATE SEQUENCE public.wsx_instance_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


ALTER SEQUENCE public.wsx_instance_seq OWNER TO daylily;

--
-- Name: wx_instance_seq; Type: SEQUENCE; Schema: public; Owner: daylily
--

CREATE SEQUENCE public.wx_instance_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


ALTER SEQUENCE public.wx_instance_seq OWNER TO daylily;

--
-- Name: xx_instance_seq; Type: SEQUENCE; Schema: public; Owner: daylily
--

CREATE SEQUENCE public.xx_instance_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


ALTER SEQUENCE public.xx_instance_seq OWNER TO daylily;

--
-- Data for Name: audit_log; Type: TABLE DATA; Schema: public; Owner: daylily
--

COPY public.audit_log (uuid, rel_table_name, column_name, rel_table_uuid_fk, rel_table_euid_fk, old_value, new_value, changed_by, changed_at, operation_type, json_addl, super_type, deleted_record_json, is_deleted, is_singleton) FROM stdin;
bce11bfd-9a6f-4c15-9a1b-348ab37c207a	generic_template	\N	4a35bc93-97d4-43e5-b228-085f24c16807	GT1	\N	\N	bloom_db_init	2025-07-03 05:56:45.597163-07	INSERT	\N	\N	\N	f	f
0febfc94-8789-4f5d-8f7f-a75f2b8f1cbe	generic_template	\N	f231825e-55c9-4b14-9ea1-98df6b41172b	GT2	\N	\N	bloom_db_init	2025-07-03 05:56:46.067501-07	INSERT	\N	\N	\N	f	f
30a3a772-2c8e-43ca-b0ab-db724111f35c	generic_template	\N	07d797c4-599d-4b25-af0b-36ef32b875d7	GT3	\N	\N	bloom_db_init	2025-07-03 05:56:46.532999-07	INSERT	\N	\N	\N	f	f
c1f882ec-e311-40ef-9513-f5986342ecc6	generic_template	\N	08fac2ba-c5e1-4393-9926-e4c3d67328c5	GT4	\N	\N	bloom_db_init	2025-07-03 05:56:46.542929-07	INSERT	\N	\N	\N	f	f
89bd4019-24ea-4c76-a57e-c74443b42550	generic_template	\N	a523522b-428e-4175-9e8f-00aa181fb7f5	GT5	\N	\N	bloom_db_init	2025-07-03 05:56:46.552081-07	INSERT	\N	\N	\N	f	f
9822d836-f8c6-443f-8d64-4a5dcfee6bd4	generic_template	\N	58348944-3c6d-418a-844a-1fd146c8f2e1	GT6	\N	\N	bloom_db_init	2025-07-03 05:56:47.012849-07	INSERT	\N	\N	\N	f	f
3cc01bf7-d3f1-4c0a-973a-4bb44c9ab6d0	generic_template	\N	48960c63-2668-40d4-b442-7121e3ba9054	GT7	\N	\N	bloom_db_init	2025-07-03 05:56:47.022972-07	INSERT	\N	\N	\N	f	f
012f799f-036f-4f7a-9fc8-583275d071da	generic_template	\N	37a4e666-4b84-466d-8498-0baefd7a3edd	GT8	\N	\N	bloom_db_init	2025-07-03 05:56:47.483943-07	INSERT	\N	\N	\N	f	f
ecd3a4d7-5c86-4748-a744-130e96b4c2af	generic_template	\N	09f033cd-f0df-462b-a151-43d92049b8b2	GT9	\N	\N	bloom_db_init	2025-07-03 05:56:47.492776-07	INSERT	\N	\N	\N	f	f
ade926b9-fa55-447a-bdea-cd864f5e974a	generic_template	\N	3412d0e5-3fd2-40c1-a84b-ceb1ddb90e3e	GT10	\N	\N	bloom_db_init	2025-07-03 05:56:47.501987-07	INSERT	\N	\N	\N	f	f
f389f821-f145-4a89-abda-2518f6049913	generic_template	\N	bfab0270-6453-43cb-a2e8-9d3abd12eca8	GT11	\N	\N	bloom_db_init	2025-07-03 05:56:47.510056-07	INSERT	\N	\N	\N	f	f
8375b7ec-445e-475a-a888-56c5a6a84352	generic_template	\N	574c583b-b1e3-4142-9922-00529d7126e8	GT12	\N	\N	bloom_db_init	2025-07-03 05:56:47.963962-07	INSERT	\N	\N	\N	f	f
382b362c-a808-431e-9b2f-dcc5bf30f222	generic_template	\N	1a25617a-4bc5-471c-ad68-08087dfe2488	GT13	\N	\N	bloom_db_init	2025-07-03 05:56:47.975234-07	INSERT	\N	\N	\N	f	f
28c2834e-3edb-4294-8567-409fc10f6501	generic_template	\N	fe026be9-ef1e-4db0-8f53-4f254e4c212f	GT14	\N	\N	bloom_db_init	2025-07-03 05:56:47.982632-07	INSERT	\N	\N	\N	f	f
3c2325a4-c636-41a7-987c-03f34cfa3e81	generic_template	\N	bad547f4-644b-4456-9c0b-1d26d160b9a7	GT15	\N	\N	bloom_db_init	2025-07-03 05:56:48.43328-07	INSERT	\N	\N	\N	f	f
c0f800cd-424b-4153-99c1-56d5044fde4f	generic_template	\N	4439e2c0-0ccf-4acd-8454-1cf2dc9245f3	GT16	\N	\N	bloom_db_init	2025-07-03 05:56:48.901482-07	INSERT	\N	\N	\N	f	f
b73d9f2c-2092-4972-80f8-d25a827d572d	generic_template	\N	93e3736d-0342-4b13-bcb1-43f6f0ad046d	GT17	\N	\N	bloom_db_init	2025-07-03 05:56:49.356324-07	INSERT	\N	\N	\N	f	f
16378824-1634-4c3f-aba8-5681fbf4df18	generic_template	\N	f204a8a7-c93f-4647-8a83-60cd13730368	GT18	\N	\N	bloom_db_init	2025-07-03 05:56:49.845446-07	INSERT	\N	\N	\N	f	f
2b8fee28-079a-4344-a120-b926bd7bb76f	generic_template	\N	049706cb-9b8f-467d-954f-9642c7abeb77	GT19	\N	\N	bloom_db_init	2025-07-03 05:56:50.294806-07	INSERT	\N	\N	\N	f	f
60963e4e-6b2c-4ed8-a3c1-fcd10fd59efe	generic_template	\N	5b308530-e7e5-458e-8cf2-1dd82bbdf73b	GT20	\N	\N	bloom_db_init	2025-07-03 05:56:50.304285-07	INSERT	\N	\N	\N	f	f
7e30fa1a-0c28-420f-a9c1-8290f2fdb74d	generic_template	\N	589df01e-8e35-4949-8af2-873ba2772644	GT21	\N	\N	bloom_db_init	2025-07-03 05:56:50.771645-07	INSERT	\N	\N	\N	f	f
843478c5-706e-4a8f-b26c-ddecb3527a21	generic_template	\N	ef501631-cae6-4e0e-b901-c6fc61245b56	GT22	\N	\N	bloom_db_init	2025-07-03 05:56:51.235388-07	INSERT	\N	\N	\N	f	f
08bfa869-e49e-4f6d-b594-8c28c50c5472	generic_template	\N	b08268ac-cb7a-4e88-baae-ed61c441ea37	GT23	\N	\N	bloom_db_init	2025-07-03 05:56:51.245386-07	INSERT	\N	\N	\N	f	f
68342c01-215d-4175-9627-54fc21794ee6	generic_template	\N	98df25d3-bf76-4ae0-a3cf-295bb667652f	GT24	\N	\N	bloom_db_init	2025-07-03 05:56:51.702162-07	INSERT	\N	\N	\N	f	f
cf410914-aef1-4e51-bc25-c700cd3a84e0	generic_template	\N	cae63b72-f330-441c-8200-30509666adc2	GT25	\N	\N	bloom_db_init	2025-07-03 05:56:51.71187-07	INSERT	\N	\N	\N	f	f
143f1964-084a-4c58-a3a2-6874024c1660	generic_template	\N	65bccbd3-92a2-4f05-a6f9-9cbf95af621d	GT26	\N	\N	bloom_db_init	2025-07-03 05:56:51.720854-07	INSERT	\N	\N	\N	f	f
e9a9c33c-d7e9-4be1-97a0-46adfe7c2f44	generic_template	\N	7463cce0-e0a2-401c-a36a-50823db2f8cf	GT27	\N	\N	bloom_db_init	2025-07-03 05:56:51.729001-07	INSERT	\N	\N	\N	f	f
0a467a29-c307-43ab-a130-6b7599ca52ca	generic_template	\N	48bf27c7-068c-4625-bce5-1826388f36fb	GT28	\N	\N	bloom_db_init	2025-07-03 05:56:51.736163-07	INSERT	\N	\N	\N	f	f
1d5dcc21-b1fc-4148-b9ed-185163fb1876	generic_template	\N	77d0a516-d631-4452-9664-b46b1fb760f9	GT29	\N	\N	bloom_db_init	2025-07-03 05:56:52.188322-07	INSERT	\N	\N	\N	f	f
157d508b-94d6-4414-8821-605eed78b414	generic_template	\N	a963a0b5-9398-493c-8a4f-7a0f6cb58a0a	GT30	\N	\N	bloom_db_init	2025-07-03 05:56:52.641013-07	INSERT	\N	\N	\N	f	f
5fedf907-e4b9-4808-992a-86a5fa9d3ff8	generic_template	\N	a9a238f2-2fb8-41c8-b64e-5f0f2f51e26d	GT31	\N	\N	bloom_db_init	2025-07-03 05:56:53.087159-07	INSERT	\N	\N	\N	f	f
5806af51-ca50-452d-9fa2-e69e4fd1c831	generic_template	\N	595b2ed6-7759-4dd9-bce7-e755190406e9	GT32	\N	\N	bloom_db_init	2025-07-03 05:56:53.556901-07	INSERT	\N	\N	\N	f	f
bcb0e4fa-edb4-43c4-b6cb-33252d9d0094	generic_template	\N	f9d50b10-de39-4ad4-ade4-767adbd86ae8	GT33	\N	\N	bloom_db_init	2025-07-03 05:56:53.567628-07	INSERT	\N	\N	\N	f	f
9face793-faca-4651-886b-e58eec10bf8f	generic_template	\N	72f7584f-a95f-47ca-b02d-10e9f734bfef	GT34	\N	\N	bloom_db_init	2025-07-03 05:56:54.021859-07	INSERT	\N	\N	\N	f	f
caa122f4-eee6-4a86-a4f3-168060a21a81	generic_template	\N	2810a6e6-bf77-40f3-aa57-feeecaf7d589	GT35	\N	\N	bloom_db_init	2025-07-03 05:56:54.03264-07	INSERT	\N	\N	\N	f	f
ed25de8b-1a55-46b2-81cb-3dc45faa2356	generic_template	\N	c28e1cd1-224c-429c-8397-5813b2b9d50e	GT36	\N	\N	bloom_db_init	2025-07-03 05:56:54.040451-07	INSERT	\N	\N	\N	f	f
dfb0a67e-55d7-46e1-9a20-ed29feefc961	generic_template	\N	2e9cce06-c2e4-4bc6-ab28-9f1fb91072cd	GT37	\N	\N	bloom_db_init	2025-07-03 05:56:54.048666-07	INSERT	\N	\N	\N	f	f
9d3c4e97-4cb4-46d4-9864-9c18131834c3	generic_template	\N	2855fdeb-00d2-46d8-823d-8dd13fb1d82f	GT38	\N	\N	bloom_db_init	2025-07-03 05:56:54.056092-07	INSERT	\N	\N	\N	f	f
a80429c9-8d0f-437d-8e78-7918d898f326	generic_template	\N	094d3f62-c3d3-4501-a801-9ffab36ec973	GT39	\N	\N	bloom_db_init	2025-07-03 05:56:54.51131-07	INSERT	\N	\N	\N	f	f
b4fbedf5-a658-44c5-8c37-88cc15ffab50	generic_template	\N	558b3f42-d3db-46dc-bc44-dba056406fe5	GT40	\N	\N	bloom_db_init	2025-07-03 05:56:54.522819-07	INSERT	\N	\N	\N	f	f
91ab510a-3c02-42c9-8d5d-7678de0623f9	generic_template	\N	a54b1949-0198-401f-81ef-6071793ccfe9	GT41	\N	\N	bloom_db_init	2025-07-03 05:56:54.982533-07	INSERT	\N	\N	\N	f	f
66bb6f73-9d09-4d34-ad56-2638ff5c0814	generic_template	\N	e76cf793-06c2-4109-8df8-6e0e3517ddd6	GT42	\N	\N	bloom_db_init	2025-07-03 05:56:54.992571-07	INSERT	\N	\N	\N	f	f
7b182a21-5ec0-424b-a94c-d6b27f055a0e	generic_template	\N	2959b93a-fd1b-4d2c-a5ca-b6fd917b4cbc	GT43	\N	\N	bloom_db_init	2025-07-03 05:56:55.446608-07	INSERT	\N	\N	\N	f	f
375ea3cf-cae7-427a-9e34-2a85d1d4747e	generic_template	\N	4bb99ae7-dc05-460e-95c5-6fec42076080	GT44	\N	\N	bloom_db_init	2025-07-03 05:56:55.898352-07	INSERT	\N	\N	\N	f	f
cf403f2c-2c4a-425a-8579-f1cdcefa1866	generic_template	\N	6bae2028-8873-4717-9e10-821a62bcd9f0	GT45	\N	\N	bloom_db_init	2025-07-03 05:56:56.347558-07	INSERT	\N	\N	\N	f	f
771b1899-538a-4ffb-95e2-484452769549	generic_template	\N	36f86ab4-1ac5-49a6-8dc3-0285b1185fe3	GT46	\N	\N	bloom_db_init	2025-07-03 05:56:56.357901-07	INSERT	\N	\N	\N	f	f
97118fe2-dd2f-4015-81de-fe1031640a21	generic_template	\N	88594a47-129c-4db5-886a-aca0f62ba7d2	GT47	\N	\N	bloom_db_init	2025-07-03 05:56:56.812244-07	INSERT	\N	\N	\N	f	f
f55c9bce-bd77-4094-9ae0-553f50e21a4a	generic_template	\N	d5ee3d31-ad0b-4780-a97d-b0c78bc74ac3	GT48	\N	\N	bloom_db_init	2025-07-03 05:56:56.82144-07	INSERT	\N	\N	\N	f	f
708bb7d2-1ba6-4bba-a67a-5ad3788f3b25	generic_template	\N	4c7ddd40-8031-4060-87e4-1755fea7ba18	GT49	\N	\N	bloom_db_init	2025-07-03 05:56:57.280881-07	INSERT	\N	\N	\N	f	f
9b03baec-346d-4acb-bf5b-147a82f8e43c	generic_template	\N	c5ef9747-3879-4e3e-bab9-56164b1dda95	GT50	\N	\N	bloom_db_init	2025-07-03 05:56:57.734023-07	INSERT	\N	\N	\N	f	f
4a2cef3f-3b30-4946-94f3-aa3ee066bfe5	generic_template	\N	d1533fbe-9bf0-46f2-abff-3c74cc80c744	GT51	\N	\N	bloom_db_init	2025-07-03 05:56:57.744961-07	INSERT	\N	\N	\N	f	f
0a7846c7-3eea-4e87-ade2-b5f036a15f57	generic_template	\N	ceca7fef-9b77-433d-9420-8d8a15cf1147	GT52	\N	\N	bloom_db_init	2025-07-03 05:56:57.752706-07	INSERT	\N	\N	\N	f	f
fb3d4c26-6618-4126-81ec-5eb57b0d9f2c	generic_template	\N	7651bf71-c917-4c36-bd28-ded3fa87e1c2	GT53	\N	\N	bloom_db_init	2025-07-03 05:56:57.760583-07	INSERT	\N	\N	\N	f	f
77de547d-c03c-45e3-b1e2-6a451ba6b4d6	generic_template	\N	507a3395-14f5-4e21-bf68-66df5d242281	GT54	\N	\N	bloom_db_init	2025-07-03 05:56:58.215895-07	INSERT	\N	\N	\N	f	f
100122fb-5201-4f26-ad01-029e10e75be8	generic_template	\N	f974507f-5cd4-48ce-82c2-696aa8770434	GT55	\N	\N	bloom_db_init	2025-07-03 05:56:58.667599-07	INSERT	\N	\N	\N	f	f
a009883a-ec07-4675-8319-d56cc84a785f	generic_template	\N	50d08295-a50b-422e-9c2a-6c1791c08e8f	GT56	\N	\N	bloom_db_init	2025-07-03 05:56:58.67764-07	INSERT	\N	\N	\N	f	f
7cddc23d-5f60-4171-aba3-24aa5ce5d267	generic_template	\N	70d42dd6-b1f4-4293-ba62-97c902f6ffb0	GT57	\N	\N	bloom_db_init	2025-07-03 05:56:58.684637-07	INSERT	\N	\N	\N	f	f
f15b0500-593b-45a8-87cc-b640dc89df8d	generic_template	\N	4765a344-581e-4bc6-9c70-cca99e85637d	GT58	\N	\N	bloom_db_init	2025-07-03 05:56:58.691421-07	INSERT	\N	\N	\N	f	f
f527dad6-e5e1-43fa-aecf-9eed29701391	generic_template	\N	ba1686fd-a5c7-44d1-82c4-4fae8ebcb888	GT59	\N	\N	bloom_db_init	2025-07-03 05:56:59.14715-07	INSERT	\N	\N	\N	f	f
4028f89f-0577-4b45-87f2-6cc467874953	generic_template	\N	37a16678-0fa3-401e-8467-f5842fa37866	GT60	\N	\N	bloom_db_init	2025-07-03 05:56:59.607026-07	INSERT	\N	\N	\N	f	f
6377cd44-4b43-47aa-bc90-1e80bcb91fbe	generic_template	\N	95078f21-b1f0-46c6-99cf-e198717cca5e	GT61	\N	\N	bloom_db_init	2025-07-03 05:56:59.616012-07	INSERT	\N	\N	\N	f	f
f14dca3e-3dbb-4bad-af07-3045d0cda57d	generic_template	\N	1eded08b-0b24-4da3-9259-f4cbcca1b4a4	GT62	\N	\N	bloom_db_init	2025-07-03 05:56:59.623449-07	INSERT	\N	\N	\N	f	f
c229d19e-70c4-45a0-9250-09fefc5fa31e	generic_template	\N	ab7ccba5-c577-4dad-b713-59623d41b45a	GT63	\N	\N	bloom_db_init	2025-07-03 05:57:00.08252-07	INSERT	\N	\N	\N	f	f
d6f1187d-8ec0-48e6-ab49-453624e7d33a	generic_template	\N	0b2a2111-b3a7-415a-b61b-e9fc0051c56b	GT64	\N	\N	bloom_db_init	2025-07-03 05:57:00.091674-07	INSERT	\N	\N	\N	f	f
9adf628a-f3e0-4f2c-a44c-030c3a41e2bd	generic_template	\N	6b0e3b65-2244-4963-b36c-e2a9582344b2	GT65	\N	\N	bloom_db_init	2025-07-03 05:57:00.098763-07	INSERT	\N	\N	\N	f	f
57d9ffd4-a4a0-4f82-bda0-b4503ec6cb5d	generic_template	\N	f43b7ac4-b6c7-4159-9ede-58f9dabc5414	GT66	\N	\N	bloom_db_init	2025-07-03 05:57:00.560583-07	INSERT	\N	\N	\N	f	f
58f9ac1b-0991-454d-b544-6648f663160a	generic_template	\N	bb3a55ef-c1d8-4b21-8126-e6d2f8b71cc7	GT67	\N	\N	bloom_db_init	2025-07-03 05:57:00.571084-07	INSERT	\N	\N	\N	f	f
98226f31-7058-4816-809e-0e8c46b70051	generic_template	\N	f12aaba1-dba0-4228-8f42-bc8404095117	GT68	\N	\N	bloom_db_init	2025-07-03 05:57:00.578556-07	INSERT	\N	\N	\N	f	f
5760dbec-e17c-4259-8480-5b861c84405f	generic_template	\N	2b1a67ed-2156-484d-aeb7-04566a9307db	GT69	\N	\N	bloom_db_init	2025-07-03 05:57:01.034457-07	INSERT	\N	\N	\N	f	f
ffe402e5-82b1-4fcd-9c53-fe23c6ae2d09	generic_template	\N	e14dcbc9-15a4-4df5-a246-545a24e0fdec	GT70	\N	\N	bloom_db_init	2025-07-03 05:57:01.044305-07	INSERT	\N	\N	\N	f	f
312c3ce0-de45-44b1-8ddd-b23e46a77d82	generic_template	\N	1167bc43-6f91-475b-bd87-810fe80ff930	GT71	\N	\N	bloom_db_init	2025-07-03 05:57:01.492497-07	INSERT	\N	\N	\N	f	f
3c85003a-dbce-4a1e-a03a-22fdf47b9904	generic_template	\N	85684d6e-da67-48c5-b565-3891c44d2859	GT72	\N	\N	bloom_db_init	2025-07-03 05:57:01.502472-07	INSERT	\N	\N	\N	f	f
9a635fbd-55e2-4d54-b45a-63d494446671	generic_template	\N	42b5243c-25db-4bdb-862b-82973a4a162d	GT73	\N	\N	bloom_db_init	2025-07-03 05:57:01.955911-07	INSERT	\N	\N	\N	f	f
bb1b9f26-0dc8-477b-a033-d0678e739fe8	generic_template	\N	e099d6f0-db3b-4613-b7a6-496537eef95a	GT74	\N	\N	bloom_db_init	2025-07-03 05:57:02.395505-07	INSERT	\N	\N	\N	f	f
dac3722e-6242-4c95-8e92-8366a49781de	generic_template	\N	7514dcbb-cca8-4793-8b61-6aa4ce3312b9	GT75	\N	\N	bloom_db_init	2025-07-03 05:57:02.403676-07	INSERT	\N	\N	\N	f	f
065b676c-bdcc-4a90-aff9-5f905f330b59	generic_template	\N	c91acde4-57c5-4dcf-9704-cbb0fb297475	GT76	\N	\N	bloom_db_init	2025-07-03 05:57:02.41276-07	INSERT	\N	\N	\N	f	f
88ac8410-7a86-490e-ae2c-5f66ca57911a	generic_template	\N	1c57e0bf-8b44-426b-8f9e-e96aff46a341	GT77	\N	\N	bloom_db_init	2025-07-03 05:57:02.835006-07	INSERT	\N	\N	\N	f	f
af70d5ef-431c-4fc2-99ef-be144bf77067	generic_template	\N	9a500c83-8b62-48e1-a0b0-cddd7c498e29	GT78	\N	\N	bloom_db_init	2025-07-03 05:57:03.249325-07	INSERT	\N	\N	\N	f	f
a923b9ec-6655-4d84-8ba6-6fb6d818ecc5	generic_template	\N	64646531-f2fd-463a-9790-a4587d83f2a5	GT79	\N	\N	bloom_db_init	2025-07-03 05:57:03.668781-07	INSERT	\N	\N	\N	f	f
f1944110-29af-42ab-8843-dfea5258ffdf	generic_template	\N	b1dfed48-902d-4761-8e91-3588388a3247	GT80	\N	\N	bloom_db_init	2025-07-03 05:57:04.1234-07	INSERT	\N	\N	\N	f	f
795514b4-2ebb-4e49-bc8e-0bfd5ad0cc5e	generic_template	\N	b7102e95-8645-4f7f-ad5c-269b41b356ac	GT81	\N	\N	bloom_db_init	2025-07-03 05:57:04.535922-07	INSERT	\N	\N	\N	f	f
168c0a1e-6597-4f4f-ad5f-66160c908329	generic_template	\N	788e3a58-f719-4bfe-a013-2d3e7d33fc1d	GT82	\N	\N	bloom_db_init	2025-07-03 05:57:04.952392-07	INSERT	\N	\N	\N	f	f
fd4be1cb-83f9-463d-a205-609f814e46c4	generic_template	\N	544eb81d-8a02-4606-87b5-4eff70d7d1e1	GT83	\N	\N	bloom_db_init	2025-07-03 05:57:05.368267-07	INSERT	\N	\N	\N	f	f
4a6c44a0-7c77-43df-9ba2-15f9b962ec02	generic_template	\N	4f62b8c2-2d9c-49fe-a609-086f412026be	GT84	\N	\N	bloom_db_init	2025-07-03 05:57:05.787722-07	INSERT	\N	\N	\N	f	f
0acc020d-c6dc-496e-9084-c63814ebc636	generic_template	\N	19e90d71-6ca5-4337-9611-dcaefae94281	GT85	\N	\N	bloom_db_init	2025-07-03 05:57:05.795953-07	INSERT	\N	\N	\N	f	f
45e421dc-6ebd-41ae-b07e-e174a923a3ff	generic_template	\N	990367b0-c995-4ac3-9d8a-527eb85151b8	GT86	\N	\N	bloom_db_init	2025-07-03 05:57:05.802888-07	INSERT	\N	\N	\N	f	f
233974e6-2992-4f69-8189-93ba57ad0769	generic_template	\N	dd00cc24-18ce-4b4c-a6a1-c7ac503f2ec1	GT87	\N	\N	bloom_db_init	2025-07-03 05:57:05.809117-07	INSERT	\N	\N	\N	f	f
f77be09c-ffc4-4f34-a8e8-8292ed5fbbc3	generic_template	\N	285d1278-67df-4d62-b2de-67d49bca1572	GT88	\N	\N	bloom_db_init	2025-07-03 05:57:06.256703-07	INSERT	\N	\N	\N	f	f
f0a55b47-9ab5-4a60-a79c-428e06dc2648	generic_template	\N	9a550109-ad1e-4afd-b127-3ee9378a542f	GT89	\N	\N	bloom_db_init	2025-07-03 05:57:06.265499-07	INSERT	\N	\N	\N	f	f
ed20c7e8-3453-42ee-934d-b5c5e8548cfc	generic_template	\N	b7eaddd6-0201-4579-b422-95af5dbb0805	GT90	\N	\N	bloom_db_init	2025-07-03 05:57:06.697285-07	INSERT	\N	\N	\N	f	f
63e330da-3036-476d-9059-d34f56c31fa9	generic_template	\N	3862d63d-1127-4727-af80-064ff9c105e2	GT91	\N	\N	bloom_db_init	2025-07-03 05:57:07.119389-07	INSERT	\N	\N	\N	f	f
6af2cc54-a69a-4bd8-99a9-bc35f7da894d	generic_template	\N	34966c45-2cd1-485a-b2db-cfb932024bed	GT92	\N	\N	bloom_db_init	2025-07-03 05:57:07.541107-07	INSERT	\N	\N	\N	f	f
014806c4-ec1b-4695-941d-b15c3d0b51dc	generic_template	\N	4f614c77-6499-4b50-bb8b-3bde0c5cd5be	GT93	\N	\N	bloom_db_init	2025-07-03 05:57:07.550559-07	INSERT	\N	\N	\N	f	f
3ef9ce76-3320-4256-b3ea-cc333067bbf2	generic_template	\N	c96906f6-fa81-4c5f-90dc-25a5a099b745	GT94	\N	\N	bloom_db_init	2025-07-03 05:57:07.557425-07	INSERT	\N	\N	\N	f	f
da8f463e-f4a1-49fa-b341-95a8d82b50fd	generic_template	\N	6c7c2ad6-3774-42c5-a226-c09b52e011b0	GT95	\N	\N	bloom_db_init	2025-07-03 05:57:07.563973-07	INSERT	\N	\N	\N	f	f
ab7dc945-1037-4dae-b719-6b09e74f14c8	generic_template	\N	3c8306f0-1d56-4dae-b063-76cf78fd96dd	GT96	\N	\N	bloom_db_init	2025-07-03 05:57:07.985297-07	INSERT	\N	\N	\N	f	f
873bcbcf-fa1a-4f4b-a4da-c227bc2f030a	generic_template	\N	6f7b03b6-fc05-48be-9372-a2480ce99349	GT97	\N	\N	bloom_db_init	2025-07-03 05:57:07.993607-07	INSERT	\N	\N	\N	f	f
5e35bc43-1709-49b0-a6a7-ca21604e94b0	generic_template	\N	806a641e-fc42-4c92-ab2f-25b248c82db4	GT98	\N	\N	bloom_db_init	2025-07-03 05:57:08.000826-07	INSERT	\N	\N	\N	f	f
839b3060-ff11-4456-9d87-2abb1e90f0a3	generic_template	\N	575cfbc5-68d3-40a0-a752-15638c82b871	GT99	\N	\N	bloom_db_init	2025-07-03 05:57:08.007888-07	INSERT	\N	\N	\N	f	f
440a9a91-a4c8-4c0b-b540-f6be20b0d164	generic_template	\N	34f5d0b7-534e-45dc-9e95-89b082c9f496	GT100	\N	\N	bloom_db_init	2025-07-03 05:57:08.426236-07	INSERT	\N	\N	\N	f	f
7991ddda-6176-48ef-a8a5-b010ee417120	generic_template	\N	ba6f37f3-857e-430f-b4ea-46be96b43945	GT101	\N	\N	bloom_db_init	2025-07-03 05:57:08.844981-07	INSERT	\N	\N	\N	f	f
ca211aab-fce1-458f-8e86-dbff7db5ca2f	generic_template	\N	ce71f5cc-14fb-4bdc-8798-b726f5c74966	GT102	\N	\N	bloom_db_init	2025-07-03 05:57:08.854493-07	INSERT	\N	\N	\N	f	f
c634cdc7-9f97-49b6-ba7a-552b67d927c4	generic_template	\N	c5674a0e-552c-4310-aa7b-92d147f1e22c	GT103	\N	\N	bloom_db_init	2025-07-03 05:57:08.861101-07	INSERT	\N	\N	\N	f	f
be146f16-c0ab-4101-9822-736fada4ad97	generic_template	\N	c719e675-9b0f-4225-b133-86ad974014d2	GT104	\N	\N	bloom_db_init	2025-07-03 05:57:08.867425-07	INSERT	\N	\N	\N	f	f
3c6b6dec-6492-48ae-a549-a2883e96127d	generic_template	\N	e498236a-d679-4005-8c6f-dcef0cdc44af	GT105	\N	\N	bloom_db_init	2025-07-03 05:57:08.874083-07	INSERT	\N	\N	\N	f	f
ec4beda7-21a8-4655-9d64-a26ff07dfa77	generic_template	\N	6e815277-8197-46dc-8cc3-c0741db7cb31	GT106	\N	\N	bloom_db_init	2025-07-03 05:57:09.302089-07	INSERT	\N	\N	\N	f	f
02b91154-6ab1-455e-83c3-747c3b6ebc6b	generic_template	\N	ed324835-92b7-4d82-b255-b5909b887a57	GT107	\N	\N	bloom_db_init	2025-07-03 05:57:09.725266-07	INSERT	\N	\N	\N	f	f
415a2a25-b0b4-47bd-8a87-39a295e6340a	generic_template	\N	538ca2de-0f1c-4e0b-ac0f-3aafbd176d9e	GT108	\N	\N	bloom_db_init	2025-07-03 05:57:09.735431-07	INSERT	\N	\N	\N	f	f
7d3648a0-60e5-4f9b-846f-f1c0d3cd862c	generic_template	\N	d9a37828-c1d8-44f0-979b-e293cb521c30	GT109	\N	\N	bloom_db_init	2025-07-03 05:57:10.154217-07	INSERT	\N	\N	\N	f	f
69f0caa0-5024-429a-b87a-38210d2ef38f	generic_template	\N	1251878d-5c20-490d-b895-0349b9ae827b	GT110	\N	\N	bloom_db_init	2025-07-03 05:57:10.162593-07	INSERT	\N	\N	\N	f	f
d12d43e4-8512-43d3-8be9-1011a108cb9a	generic_template	\N	a3e7194e-e1f3-47ab-9c88-8ffd6d8cbfd0	GT111	\N	\N	bloom_db_init	2025-07-03 05:57:10.16952-07	INSERT	\N	\N	\N	f	f
acdd4a0e-cd9a-48e4-8989-81e4519536c0	generic_template	\N	02ca20b8-6864-4bf7-bae8-cbf445b3324e	GT112	\N	\N	bloom_db_init	2025-07-03 05:57:10.17588-07	INSERT	\N	\N	\N	f	f
53deebe5-8fdd-4f68-90e0-0f6e55af1ec9	generic_template	\N	13028ffb-87ef-4e09-ba4e-5941405f90b9	GT113	\N	\N	bloom_db_init	2025-07-03 05:57:10.603883-07	INSERT	\N	\N	\N	f	f
4af0a7de-1c67-483a-a6c2-c14d0106a41a	generic_template	\N	ce435b29-2c59-4e13-96c6-00e1b8c50b37	GT114	\N	\N	bloom_db_init	2025-07-03 05:57:11.0287-07	INSERT	\N	\N	\N	f	f
4b303823-0c3a-4557-b0f7-99274326acfc	generic_template	\N	7f795301-5b20-47bf-9656-9073ca292af6	GT115	\N	\N	bloom_db_init	2025-07-03 05:57:11.03695-07	INSERT	\N	\N	\N	f	f
e2b8e7f7-d0bb-471c-938e-f3a1c0139f09	generic_instance	\N	a95a3806-06d0-4ce3-bf5b-042fb400562b	FS1	\N	\N	john@daylilyinformatics.com	2025-07-03 05:57:59.406061-07	INSERT	\N	\N	\N	f	f
9542ed09-5971-433d-a242-81be3dc91c3b	generic_instance	\N	01045846-479f-43f7-8d20-e84e3bec3420	FI7003	\N	\N	john@daylilyinformatics.com	2025-07-03 05:57:59.595243-07	INSERT	\N	\N	\N	f	f
0d8f86b1-d6b6-4065-a946-5ef712d650ea	generic_instance	json_addl	01045846-479f-43f7-8d20-e84e3bec3420	FI7003	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T125759", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T125759", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 05:57:59.604217-07	UPDATE	\N	\N	\N	f	f
ca86c41a-8e4d-414a-ac90-9b4968b3d6e0	generic_instance	created_dt	01045846-479f-43f7-8d20-e84e3bec3420	FI7003	2025-07-03 05:57:59.595243-07	2025-07-03T05:57:59.595243-07:00	john@daylilyinformatics.com	2025-07-03 05:57:59.604217-07	UPDATE	\N	\N	\N	f	f
32decfed-24ab-448f-9a5e-be99ca0e6f4a	generic_instance	modified_dt	01045846-479f-43f7-8d20-e84e3bec3420	FI7003	2025-07-03 05:57:59.595243-07	2025-07-03T05:57:59.604217-07:00	john@daylilyinformatics.com	2025-07-03 05:57:59.604217-07	UPDATE	\N	\N	\N	f	f
093956be-743b-4e42-b23d-990082ef09a2	generic_instance	json_addl	01045846-479f-43f7-8d20-e84e3bec3420	FI7003	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T125759", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7003.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7003.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T125759", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 05:57:59.944771-07	UPDATE	\N	\N	\N	f	f
3f416911-046d-46a0-ae0d-1cfaa4e34c74	generic_instance	created_dt	01045846-479f-43f7-8d20-e84e3bec3420	FI7003	2025-07-03 05:57:59.595243-07	2025-07-03T05:57:59.595243-07:00	john@daylilyinformatics.com	2025-07-03 05:57:59.944771-07	UPDATE	\N	\N	\N	f	f
f5e906fc-4ac3-45fa-86e6-f7c6430e5e28	generic_instance	modified_dt	01045846-479f-43f7-8d20-e84e3bec3420	FI7003	2025-07-03 05:57:59.604217-07	2025-07-03T05:57:59.944771-07:00	john@daylilyinformatics.com	2025-07-03 05:57:59.944771-07	UPDATE	\N	\N	\N	f	f
f0cb4e03-caee-49ff-912b-035a33d912a5	generic_instance_lineage	\N	54f3e9c7-4f9f-48c7-8b3a-67ecfc60d506	GL1	\N	\N	john@daylilyinformatics.com	2025-07-03 05:58:00.946314-07	INSERT	\N	\N	\N	f	f
f8d61306-28eb-4e13-91bb-9465415fe0cd	generic_instance	\N	c76338b6-6ba8-4dc2-9b60-eb20a9e253a6	FI7004	\N	\N	john@daylilyinformatics.com	2025-07-03 05:58:00.957798-07	INSERT	\N	\N	\N	f	f
83a780a7-aeb2-4457-a58a-7ee8344d71a5	generic_instance	json_addl	c76338b6-6ba8-4dc2-9b60-eb20a9e253a6	FI7004	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T125759", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T125759", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 05:58:01.15712-07	UPDATE	\N	\N	\N	f	f
e750f41b-615d-47da-ad59-9626fe626e43	generic_instance	created_dt	c76338b6-6ba8-4dc2-9b60-eb20a9e253a6	FI7004	2025-07-03 05:58:00.957798-07	2025-07-03T05:58:00.957798-07:00	john@daylilyinformatics.com	2025-07-03 05:58:01.15712-07	UPDATE	\N	\N	\N	f	f
1ee8a1d9-f6a7-4cc8-a5d7-a244fd91575c	generic_instance	modified_dt	c76338b6-6ba8-4dc2-9b60-eb20a9e253a6	FI7004	2025-07-03 05:58:00.957798-07	2025-07-03T05:58:01.15712-07:00	john@daylilyinformatics.com	2025-07-03 05:58:01.15712-07	UPDATE	\N	\N	\N	f	f
04c779ce-6323-4030-873f-778ac8c78a73	generic_instance	json_addl	c76338b6-6ba8-4dc2-9b60-eb20a9e253a6	FI7004	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T125759", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7004.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7004.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T125759", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 05:58:01.279603-07	UPDATE	\N	\N	\N	f	f
56414176-2494-443b-abeb-274457c1f899	generic_instance	created_dt	c76338b6-6ba8-4dc2-9b60-eb20a9e253a6	FI7004	2025-07-03 05:58:00.957798-07	2025-07-03T05:58:00.957798-07:00	john@daylilyinformatics.com	2025-07-03 05:58:01.279603-07	UPDATE	\N	\N	\N	f	f
0a623c70-f647-4487-8ee9-b3c6b2edb395	generic_instance	modified_dt	c76338b6-6ba8-4dc2-9b60-eb20a9e253a6	FI7004	2025-07-03 05:58:01.15712-07	2025-07-03T05:58:01.279603-07:00	john@daylilyinformatics.com	2025-07-03 05:58:01.279603-07	UPDATE	\N	\N	\N	f	f
c9b5f92c-8f99-4bc9-b469-b2bf51abb9cf	generic_instance_lineage	\N	96815b0b-1120-4db7-a199-f709a5961913	GL2	\N	\N	john@daylilyinformatics.com	2025-07-03 05:58:02.248251-07	INSERT	\N	\N	\N	f	f
c94295af-1220-4dd4-ba8a-a45b35747f6d	generic_instance	\N	6dbd1b9c-394e-43ad-9935-dce8db433908	FI7005	\N	\N	john@daylilyinformatics.com	2025-07-03 05:58:02.258533-07	INSERT	\N	\N	\N	f	f
49a841ec-f7a4-4f7c-88c8-96f400c1e148	generic_instance	json_addl	6dbd1b9c-394e-43ad-9935-dce8db433908	FI7005	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T125759", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T125759", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 05:58:02.283521-07	UPDATE	\N	\N	\N	f	f
85b39b14-5b83-4ca6-9fd1-d8f846e95ecb	generic_instance	created_dt	6dbd1b9c-394e-43ad-9935-dce8db433908	FI7005	2025-07-03 05:58:02.258533-07	2025-07-03T05:58:02.258533-07:00	john@daylilyinformatics.com	2025-07-03 05:58:02.283521-07	UPDATE	\N	\N	\N	f	f
5cd86dd4-af1d-44c3-a8c7-b0c03bcc16df	generic_instance	modified_dt	6dbd1b9c-394e-43ad-9935-dce8db433908	FI7005	2025-07-03 05:58:02.258533-07	2025-07-03T05:58:02.283521-07:00	john@daylilyinformatics.com	2025-07-03 05:58:02.283521-07	UPDATE	\N	\N	\N	f	f
3c946218-083f-40e6-9a11-556e2dc46389	generic_instance	json_addl	6dbd1b9c-394e-43ad-9935-dce8db433908	FI7005	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T125759", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7005.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7005.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T125759", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 05:58:02.539249-07	UPDATE	\N	\N	\N	f	f
3ffc8533-a1c9-4d73-84ca-4238a2305ddc	generic_instance	created_dt	6dbd1b9c-394e-43ad-9935-dce8db433908	FI7005	2025-07-03 05:58:02.258533-07	2025-07-03T05:58:02.258533-07:00	john@daylilyinformatics.com	2025-07-03 05:58:02.539249-07	UPDATE	\N	\N	\N	f	f
d7492783-18e9-4a02-b55b-0d025db4db82	generic_instance	modified_dt	6dbd1b9c-394e-43ad-9935-dce8db433908	FI7005	2025-07-03 05:58:02.283521-07	2025-07-03T05:58:02.539249-07:00	john@daylilyinformatics.com	2025-07-03 05:58:02.539249-07	UPDATE	\N	\N	\N	f	f
9fe9e55f-1553-4298-8238-7634558411d1	generic_instance_lineage	\N	8796fe96-65e1-4b68-aa6a-fc0a978ecbe8	GL3	\N	\N	john@daylilyinformatics.com	2025-07-03 05:58:03.384589-07	INSERT	\N	\N	\N	f	f
fc953663-cfb7-4715-9aed-bc6fdab4013e	generic_instance	\N	2daf2aa2-3633-4fd6-a8a4-784977febd55	FI7006	\N	\N	john@daylilyinformatics.com	2025-07-03 05:58:03.395505-07	INSERT	\N	\N	\N	f	f
13f8916f-2e8f-43ee-b39b-1b00f9d66475	generic_instance	json_addl	2daf2aa2-3633-4fd6-a8a4-784977febd55	FI7006	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T125759", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T125759", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 05:58:03.576243-07	UPDATE	\N	\N	\N	f	f
c6267f8b-3307-4263-98c9-da88f369ee38	generic_instance	created_dt	2daf2aa2-3633-4fd6-a8a4-784977febd55	FI7006	2025-07-03 05:58:03.395505-07	2025-07-03T05:58:03.395505-07:00	john@daylilyinformatics.com	2025-07-03 05:58:03.576243-07	UPDATE	\N	\N	\N	f	f
12572dd7-0251-4629-92bd-417443f08eb2	generic_instance	modified_dt	2daf2aa2-3633-4fd6-a8a4-784977febd55	FI7006	2025-07-03 05:58:03.395505-07	2025-07-03T05:58:03.576243-07:00	john@daylilyinformatics.com	2025-07-03 05:58:03.576243-07	UPDATE	\N	\N	\N	f	f
3df87d0d-74d1-4d23-b6f2-3a9521aa32d8	generic_instance	json_addl	2daf2aa2-3633-4fd6-a8a4-784977febd55	FI7006	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T125759", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7006.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7006.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T125759", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 05:58:03.705654-07	UPDATE	\N	\N	\N	f	f
755021d6-a421-45b5-b1a8-2560b49da832	generic_instance	created_dt	2daf2aa2-3633-4fd6-a8a4-784977febd55	FI7006	2025-07-03 05:58:03.395505-07	2025-07-03T05:58:03.395505-07:00	john@daylilyinformatics.com	2025-07-03 05:58:03.705654-07	UPDATE	\N	\N	\N	f	f
92f30b6a-05b4-44de-9f20-702a6a749812	generic_instance	modified_dt	2daf2aa2-3633-4fd6-a8a4-784977febd55	FI7006	2025-07-03 05:58:03.576243-07	2025-07-03T05:58:03.705654-07:00	john@daylilyinformatics.com	2025-07-03 05:58:03.705654-07	UPDATE	\N	\N	\N	f	f
7b6c73e5-f77d-49c1-a077-4fafaacdd40d	generic_instance_lineage	\N	24ef7ef4-c0e6-485f-a117-52b457642f9f	GL4	\N	\N	john@daylilyinformatics.com	2025-07-03 05:58:04.619145-07	INSERT	\N	\N	\N	f	f
67101870-d812-4fa9-bbed-2424916cf336	generic_instance	\N	f7c652b1-51bc-483b-a485-8d70c3630310	FS2	\N	\N	john@daylilyinformatics.com	2025-07-03 06:00:25.841409-07	INSERT	\N	\N	\N	f	f
246a784e-fdef-4ac3-9745-d3ccfd02b7ea	generic_instance	\N	c8552b03-f549-47c5-962f-bb123d63fe02	FI7007	\N	\N	john@daylilyinformatics.com	2025-07-03 06:00:25.954212-07	INSERT	\N	\N	\N	f	f
f3ec4689-356b-4ddd-81f6-2167ad2099ed	generic_instance	json_addl	c8552b03-f549-47c5-962f-bb123d63fe02	FI7007	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T130014", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T130014", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:00:25.966564-07	UPDATE	\N	\N	\N	f	f
e2f1a557-3859-4835-8011-ff03d5957798	generic_instance	created_dt	c8552b03-f549-47c5-962f-bb123d63fe02	FI7007	2025-07-03 06:00:25.954212-07	2025-07-03T06:00:25.954212-07:00	john@daylilyinformatics.com	2025-07-03 06:00:25.966564-07	UPDATE	\N	\N	\N	f	f
cc6726ec-4666-4f8f-872b-35b384c6b7ce	generic_instance	modified_dt	c8552b03-f549-47c5-962f-bb123d63fe02	FI7007	2025-07-03 06:00:25.954212-07	2025-07-03T06:00:25.966564-07:00	john@daylilyinformatics.com	2025-07-03 06:00:25.966564-07	UPDATE	\N	\N	\N	f	f
9d61f950-e942-4e2e-a3b1-5699090cd3a6	generic_instance	json_addl	c8552b03-f549-47c5-962f-bb123d63fe02	FI7007	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T130014", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "backup_2025-07-03_00-06-33.sql", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "sql", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7007.sql", "current_s3_uri": "s3://daylily-dewey-0/0/FI7007.sql", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T130014", "original_file_md5": "", "original_file_name": "backup_2025-07-03_00-06-33.sql", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "sql", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1012876, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:00:26.245305-07	UPDATE	\N	\N	\N	f	f
49ffdc1b-db04-4b46-8aca-e420919ef985	generic_instance	created_dt	c8552b03-f549-47c5-962f-bb123d63fe02	FI7007	2025-07-03 06:00:25.954212-07	2025-07-03T06:00:25.954212-07:00	john@daylilyinformatics.com	2025-07-03 06:00:26.245305-07	UPDATE	\N	\N	\N	f	f
fda71efb-d909-4ab3-96ef-df2024f02be7	generic_instance	modified_dt	c8552b03-f549-47c5-962f-bb123d63fe02	FI7007	2025-07-03 06:00:25.966564-07	2025-07-03T06:00:26.245305-07:00	john@daylilyinformatics.com	2025-07-03 06:00:26.245305-07	UPDATE	\N	\N	\N	f	f
0366e511-d71b-4216-baab-239283c90263	generic_instance_lineage	\N	54ee535d-5c73-49b5-a784-5365b0f0e7fa	GL5	\N	\N	john@daylilyinformatics.com	2025-07-03 06:00:27.011789-07	INSERT	\N	\N	\N	f	f
68db883b-7263-4e2e-8765-b4044ca1c71d	generic_instance	\N	0d2b2a11-5fa7-407f-9053-c2933c838bdf	FS3	\N	\N	john@daylilyinformatics.com	2025-07-03 06:05:38.782513-07	INSERT	\N	\N	\N	f	f
db662951-e187-4443-8c75-6569f302d87a	generic_instance	\N	c9f38309-e10a-4fae-a3d0-f1f8488f7eae	FI7008	\N	\N	john@daylilyinformatics.com	2025-07-03 06:05:38.980249-07	INSERT	\N	\N	\N	f	f
40927be1-e92c-4cb3-b047-d4ce1ce5b6bc	generic_instance	json_addl	c9f38309-e10a-4fae-a3d0-f1f8488f7eae	FI7008	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T130538", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T130538", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:05:38.996341-07	UPDATE	\N	\N	\N	f	f
a8ccb528-434e-47c6-baeb-78418f9325bd	generic_instance	created_dt	c9f38309-e10a-4fae-a3d0-f1f8488f7eae	FI7008	2025-07-03 06:05:38.980249-07	2025-07-03T06:05:38.980249-07:00	john@daylilyinformatics.com	2025-07-03 06:05:38.996341-07	UPDATE	\N	\N	\N	f	f
d7885fa3-96dc-4c05-bfef-076dc17202e1	generic_instance_lineage	\N	e087e2fe-5053-4cc2-9058-62ff7149a22e	GL13	\N	\N	john@daylilyinformatics.com	2025-07-03 06:41:49.286757-07	INSERT	\N	\N	\N	f	f
b3d66cd1-8985-47c4-b802-a85cfeb33e10	generic_instance	modified_dt	c9f38309-e10a-4fae-a3d0-f1f8488f7eae	FI7008	2025-07-03 06:05:38.980249-07	2025-07-03T06:05:38.996341-07:00	john@daylilyinformatics.com	2025-07-03 06:05:38.996341-07	UPDATE	\N	\N	\N	f	f
0394d394-859b-4431-9288-b328eaac7501	generic_instance	json_addl	c9f38309-e10a-4fae-a3d0-f1f8488f7eae	FI7008	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T130538", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7008.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7008.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T130538", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:05:39.314807-07	UPDATE	\N	\N	\N	f	f
0e1812de-60ab-4b54-a074-07c8cdcc6313	generic_instance	created_dt	c9f38309-e10a-4fae-a3d0-f1f8488f7eae	FI7008	2025-07-03 06:05:38.980249-07	2025-07-03T06:05:38.980249-07:00	john@daylilyinformatics.com	2025-07-03 06:05:39.314807-07	UPDATE	\N	\N	\N	f	f
de52dab9-4dbe-487e-865e-30aa4f8601b1	generic_instance	modified_dt	c9f38309-e10a-4fae-a3d0-f1f8488f7eae	FI7008	2025-07-03 06:05:38.996341-07	2025-07-03T06:05:39.314807-07:00	john@daylilyinformatics.com	2025-07-03 06:05:39.314807-07	UPDATE	\N	\N	\N	f	f
19c2e91d-68ba-407f-ac33-fd254712552f	generic_instance_lineage	\N	056be3d1-ec54-4164-8389-037e32201f10	GL6	\N	\N	john@daylilyinformatics.com	2025-07-03 06:05:40.421955-07	INSERT	\N	\N	\N	f	f
c25a6233-7e4a-4c35-b823-5706a0673916	generic_instance	\N	c98ef464-2308-47c8-aaa7-cd1cee22edd1	FI7009	\N	\N	john@daylilyinformatics.com	2025-07-03 06:05:40.43767-07	INSERT	\N	\N	\N	f	f
15e20baf-63fd-4392-bc34-41907f9d2c33	generic_instance	json_addl	c98ef464-2308-47c8-aaa7-cd1cee22edd1	FI7009	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T130538", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T130538", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:05:40.646331-07	UPDATE	\N	\N	\N	f	f
f7af1153-6c93-4206-a1e8-d86529e38c8a	generic_instance	created_dt	c98ef464-2308-47c8-aaa7-cd1cee22edd1	FI7009	2025-07-03 06:05:40.43767-07	2025-07-03T06:05:40.43767-07:00	john@daylilyinformatics.com	2025-07-03 06:05:40.646331-07	UPDATE	\N	\N	\N	f	f
6ace3e79-19ca-4fa0-a8b3-8d8af2da48fc	generic_instance	modified_dt	c98ef464-2308-47c8-aaa7-cd1cee22edd1	FI7009	2025-07-03 06:05:40.43767-07	2025-07-03T06:05:40.646331-07:00	john@daylilyinformatics.com	2025-07-03 06:05:40.646331-07	UPDATE	\N	\N	\N	f	f
231f3acd-9a09-4b5b-9bd0-63242e56f43a	generic_instance	json_addl	c98ef464-2308-47c8-aaa7-cd1cee22edd1	FI7009	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T130538", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7009.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7009.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T130538", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:05:40.788879-07	UPDATE	\N	\N	\N	f	f
f9707876-e983-413a-a7b5-1cb9a244a965	generic_instance	created_dt	c98ef464-2308-47c8-aaa7-cd1cee22edd1	FI7009	2025-07-03 06:05:40.43767-07	2025-07-03T06:05:40.43767-07:00	john@daylilyinformatics.com	2025-07-03 06:05:40.788879-07	UPDATE	\N	\N	\N	f	f
83f048cb-33de-4a45-bd0b-d74e7430f519	generic_instance	modified_dt	c98ef464-2308-47c8-aaa7-cd1cee22edd1	FI7009	2025-07-03 06:05:40.646331-07	2025-07-03T06:05:40.788879-07:00	john@daylilyinformatics.com	2025-07-03 06:05:40.788879-07	UPDATE	\N	\N	\N	f	f
27883b18-84d9-44a3-93af-d6a76ecfaceb	generic_instance_lineage	\N	db96380a-d111-4a49-b605-fd48025f640f	GL7	\N	\N	john@daylilyinformatics.com	2025-07-03 06:05:41.67887-07	INSERT	\N	\N	\N	f	f
aad25ea4-2458-4b8a-b805-c07e002e33f4	generic_instance	\N	25cb2b88-6898-400b-9282-3285aa3cd2e6	FI7010	\N	\N	john@daylilyinformatics.com	2025-07-03 06:05:41.701199-07	INSERT	\N	\N	\N	f	f
e1b836a2-9048-46cb-8719-c20c8af07822	generic_instance	json_addl	25cb2b88-6898-400b-9282-3285aa3cd2e6	FI7010	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T130538", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T130538", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:05:41.724186-07	UPDATE	\N	\N	\N	f	f
a34333e7-958c-4220-8f97-808dc7b9859c	generic_instance	created_dt	25cb2b88-6898-400b-9282-3285aa3cd2e6	FI7010	2025-07-03 06:05:41.701199-07	2025-07-03T06:05:41.701199-07:00	john@daylilyinformatics.com	2025-07-03 06:05:41.724186-07	UPDATE	\N	\N	\N	f	f
4b9fc5e3-96b6-4852-a03b-8ed82e62f7e9	generic_instance	modified_dt	25cb2b88-6898-400b-9282-3285aa3cd2e6	FI7010	2025-07-03 06:05:41.701199-07	2025-07-03T06:05:41.724186-07:00	john@daylilyinformatics.com	2025-07-03 06:05:41.724186-07	UPDATE	\N	\N	\N	f	f
42ef7328-0106-4e2e-852f-0fce47cc78b6	generic_instance	json_addl	25cb2b88-6898-400b-9282-3285aa3cd2e6	FI7010	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T130538", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7010.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7010.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T130538", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:05:41.939746-07	UPDATE	\N	\N	\N	f	f
4d5bc7d5-f3be-4c6f-83bf-c4b39c2cc899	generic_instance	created_dt	25cb2b88-6898-400b-9282-3285aa3cd2e6	FI7010	2025-07-03 06:05:41.701199-07	2025-07-03T06:05:41.701199-07:00	john@daylilyinformatics.com	2025-07-03 06:05:41.939746-07	UPDATE	\N	\N	\N	f	f
47bf7797-aed8-4584-97b1-1e1aab7a16db	generic_instance	modified_dt	25cb2b88-6898-400b-9282-3285aa3cd2e6	FI7010	2025-07-03 06:05:41.724186-07	2025-07-03T06:05:41.939746-07:00	john@daylilyinformatics.com	2025-07-03 06:05:41.939746-07	UPDATE	\N	\N	\N	f	f
dabf1824-8479-4328-a924-fa37d65f2b20	generic_instance_lineage	\N	0c13ca5b-8813-42af-b226-7b3d2c6ada7c	GL8	\N	\N	john@daylilyinformatics.com	2025-07-03 06:05:42.641494-07	INSERT	\N	\N	\N	f	f
c29dfab4-4681-43c4-ba38-e10a944f75f9	generic_instance	\N	6f5b9dcf-45e5-43cc-b340-ed644031f886	FI7011	\N	\N	john@daylilyinformatics.com	2025-07-03 06:05:42.664112-07	INSERT	\N	\N	\N	f	f
77e7e3df-de63-4647-8bb6-cc097255a50a	generic_instance	json_addl	6f5b9dcf-45e5-43cc-b340-ed644031f886	FI7011	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T130538", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T130538", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:05:42.86122-07	UPDATE	\N	\N	\N	f	f
4dfdd4e0-20ac-41df-a280-cd4fcc1829dd	generic_instance	created_dt	6f5b9dcf-45e5-43cc-b340-ed644031f886	FI7011	2025-07-03 06:05:42.664112-07	2025-07-03T06:05:42.664112-07:00	john@daylilyinformatics.com	2025-07-03 06:05:42.86122-07	UPDATE	\N	\N	\N	f	f
40abdc2d-525d-40a4-bab4-b4837c9a4cc6	generic_instance	modified_dt	6f5b9dcf-45e5-43cc-b340-ed644031f886	FI7011	2025-07-03 06:05:42.664112-07	2025-07-03T06:05:42.86122-07:00	john@daylilyinformatics.com	2025-07-03 06:05:42.86122-07	UPDATE	\N	\N	\N	f	f
4fe6aff1-5dcf-4e37-8020-85406e348e59	generic_instance	json_addl	6f5b9dcf-45e5-43cc-b340-ed644031f886	FI7011	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T130538", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7011.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7011.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T130538", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:05:43.012234-07	UPDATE	\N	\N	\N	f	f
30242e16-f475-46d5-9ab3-2a36e958ec1e	generic_instance	created_dt	6f5b9dcf-45e5-43cc-b340-ed644031f886	FI7011	2025-07-03 06:05:42.664112-07	2025-07-03T06:05:42.664112-07:00	john@daylilyinformatics.com	2025-07-03 06:05:43.012234-07	UPDATE	\N	\N	\N	f	f
7636f870-efcf-4e31-8e5f-ac4735c575d4	generic_instance	modified_dt	6f5b9dcf-45e5-43cc-b340-ed644031f886	FI7011	2025-07-03 06:05:42.86122-07	2025-07-03T06:05:43.012234-07:00	john@daylilyinformatics.com	2025-07-03 06:05:43.012234-07	UPDATE	\N	\N	\N	f	f
b35e2d87-df10-4316-8f0c-d4c2cf92e902	generic_instance_lineage	\N	21d57f42-7ff7-45ef-b75d-182191c6a5d8	GL9	\N	\N	john@daylilyinformatics.com	2025-07-03 06:05:43.745629-07	INSERT	\N	\N	\N	f	f
9e0e5010-3596-44d7-bbdf-ee017b7209c0	generic_instance	json_addl	0d2b2a11-5fa7-407f-9053-c2933c838bdf	FS3	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T130538", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T130538", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "end_operator": {"email": "john@daylilyinformatics.com", "dag_fnv2": ""}, "creating_user": "", "end_timestamp": "2025-07-03 09:06:37 EDT-0400", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "0", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	john@daylilyinformatics.com	2025-07-03 06:06:37.18394-07	UPDATE	\N	\N	\N	f	f
e3b3c5f8-df8d-48cd-955a-e8742a4865c5	generic_instance	created_dt	0d2b2a11-5fa7-407f-9053-c2933c838bdf	FS3	2025-07-03 06:05:38.782513-07	2025-07-03T06:05:38.782513-07:00	john@daylilyinformatics.com	2025-07-03 06:06:37.18394-07	UPDATE	\N	\N	\N	f	f
e424a7d5-d13d-475a-8070-105de6b68c91	generic_instance	bstatus	0d2b2a11-5fa7-407f-9053-c2933c838bdf	FS3	ready	complete	john@daylilyinformatics.com	2025-07-03 06:06:37.18394-07	UPDATE	\N	\N	\N	f	f
4e60640c-a156-4fe9-af0a-9fb027e42b98	generic_instance	modified_dt	0d2b2a11-5fa7-407f-9053-c2933c838bdf	FS3	2025-07-03 06:05:38.782513-07	2025-07-03T06:06:37.18394-07:00	john@daylilyinformatics.com	2025-07-03 06:06:37.18394-07	UPDATE	\N	\N	\N	f	f
ba919bd7-e8f7-4241-8849-3792bcc35d2e	generic_instance	json_addl	0d2b2a11-5fa7-407f-9053-c2933c838bdf	FS3	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T130538", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "end_operator": {"email": "john@daylilyinformatics.com", "dag_fnv2": ""}, "creating_user": "", "end_timestamp": "2025-07-03 09:06:37 EDT-0400", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "0", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T130538", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "end_operator": {"email": "john@daylilyinformatics.com", "dag_fnv2": ""}, "creating_user": "", "end_timestamp": "2025-07-03 09:06:37 EDT-0400", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [{"email": "john@daylilyinformatics.com", "dag_fnv2": ""}], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "0", "max_executions": "-1", "action_executed": "1", "executed_datetime": ["2025-07-03 09:06:37 EDT-0400"], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	john@daylilyinformatics.com	2025-07-03 06:06:37.203894-07	UPDATE	\N	\N	\N	f	f
94c04f87-7732-4d41-965a-1e25a85d3880	generic_instance	created_dt	0d2b2a11-5fa7-407f-9053-c2933c838bdf	FS3	2025-07-03 06:05:38.782513-07	2025-07-03T06:05:38.782513-07:00	john@daylilyinformatics.com	2025-07-03 06:06:37.203894-07	UPDATE	\N	\N	\N	f	f
620d8cc4-05ea-4b6f-be5e-50dc4f7db525	generic_instance	modified_dt	0d2b2a11-5fa7-407f-9053-c2933c838bdf	FS3	2025-07-03 06:06:37.18394-07	2025-07-03T06:06:37.203894-07:00	john@daylilyinformatics.com	2025-07-03 06:06:37.203894-07	UPDATE	\N	\N	\N	f	f
71ec5179-30a0-4114-a8bc-7554ea2d56d9	generic_instance	\N	8c274106-80d4-4ca6-b317-2a1ad9cc1f7c	FX1	\N	\N	john@daylilyinformatics.com	2025-07-03 06:13:20.882236-07	INSERT	\N	\N	\N	f	f
8766a521-578f-488c-aa6d-8f7b941c4ba1	generic_instance_lineage	\N	40d8c3fb-7df1-484a-872a-7d6847061560	GL10	\N	\N	john@daylilyinformatics.com	2025-07-03 06:13:20.915846-07	INSERT	\N	\N	\N	f	f
9aa0cde7-6458-4267-a4d3-f4915cffde36	generic_instance_lineage	\N	afaa336c-a28e-41ee-a3db-820af546ebcf	GL11	\N	\N	john@daylilyinformatics.com	2025-07-03 06:13:20.915846-07	INSERT	\N	\N	\N	f	f
de95ad6b-c553-40ea-9471-4f187aa2e616	generic_instance	json_addl	f7c652b1-51bc-483b-a485-8d70c3630310	FS2	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T130014", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T130014", "tags": [], "comments": "", "duration": 1.0, "lab_code": "", "ref_type": "presigned_url", "description": "File set created by Dewey file manager", "creating_user": "john@daylilyinformatics.com", "rclone_config": {"host": "0.0.0.0", "port": 8080, "user": "user", "bucket": "daylily-dewey-0", "passwd": "passwd"}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	john@daylilyinformatics.com	2025-07-03 06:13:20.875767-07	UPDATE	\N	\N	\N	f	f
bd0fc054-6422-4170-88cd-2961a4aeeaf6	generic_instance	created_dt	f7c652b1-51bc-483b-a485-8d70c3630310	FS2	2025-07-03 06:00:25.841409-07	2025-07-03T06:00:25.841409-07:00	john@daylilyinformatics.com	2025-07-03 06:13:20.875767-07	UPDATE	\N	\N	\N	f	f
faaf3b89-1fbb-474d-8049-0e6723b8550e	generic_instance	modified_dt	f7c652b1-51bc-483b-a485-8d70c3630310	FS2	2025-07-03 06:00:25.841409-07	2025-07-03T06:13:20.875767-07:00	john@daylilyinformatics.com	2025-07-03 06:13:20.875767-07	UPDATE	\N	\N	\N	f	f
7f130532-102b-4fe7-b8f2-8c2c20a335c5	generic_instance	\N	d7e5884d-3ffe-451a-8fbd-99716aff9bb9	FS4	\N	\N	john@daylilyinformatics.com	2025-07-03 06:40:40.947157-07	INSERT	\N	\N	\N	f	f
8883021e-1634-4ed2-b25b-cc77392616c8	generic_instance	\N	44f71e98-963f-4ee5-90c1-d4e544a735da	FI7012	\N	\N	john@daylilyinformatics.com	2025-07-03 06:40:41.075897-07	INSERT	\N	\N	\N	f	f
95095cb4-3fde-4d7c-8791-6e75b58fe52f	generic_instance	json_addl	44f71e98-963f-4ee5-90c1-d4e544a735da	FI7012	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134027", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134027", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:40:41.09405-07	UPDATE	\N	\N	\N	f	f
e6d9940b-1a03-4349-92af-22d699ca8f9f	generic_instance	created_dt	44f71e98-963f-4ee5-90c1-d4e544a735da	FI7012	2025-07-03 06:40:41.075897-07	2025-07-03T06:40:41.075897-07:00	john@daylilyinformatics.com	2025-07-03 06:40:41.09405-07	UPDATE	\N	\N	\N	f	f
f6cb73ed-b94a-4996-ba8d-79163674e9d5	generic_instance	modified_dt	44f71e98-963f-4ee5-90c1-d4e544a735da	FI7012	2025-07-03 06:40:41.075897-07	2025-07-03T06:40:41.09405-07:00	john@daylilyinformatics.com	2025-07-03 06:40:41.09405-07	UPDATE	\N	\N	\N	f	f
bb13d54d-c28f-4f7f-99b6-01b51021f49c	generic_instance	json_addl	44f71e98-963f-4ee5-90c1-d4e544a735da	FI7012	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134027", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "backup_2025-07-03_00-06-33.sql", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "sql", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7012.sql", "current_s3_uri": "s3://daylily-dewey-0/0/FI7012.sql", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134027", "original_file_md5": "", "original_file_name": "backup_2025-07-03_00-06-33.sql", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "sql", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1012876, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:40:41.379167-07	UPDATE	\N	\N	\N	f	f
bfd4404d-30ad-4e2c-8585-aee49c118ae3	generic_instance	created_dt	44f71e98-963f-4ee5-90c1-d4e544a735da	FI7012	2025-07-03 06:40:41.075897-07	2025-07-03T06:40:41.075897-07:00	john@daylilyinformatics.com	2025-07-03 06:40:41.379167-07	UPDATE	\N	\N	\N	f	f
e931263f-b1df-47fe-89c4-dafbd6c7ebf6	generic_instance	modified_dt	44f71e98-963f-4ee5-90c1-d4e544a735da	FI7012	2025-07-03 06:40:41.09405-07	2025-07-03T06:40:41.379167-07:00	john@daylilyinformatics.com	2025-07-03 06:40:41.379167-07	UPDATE	\N	\N	\N	f	f
492eaf11-d1c0-4bc2-b937-86f90c0aeb1d	generic_instance_lineage	\N	ffe8488a-73cd-44bb-a54c-1a32202d24c7	GL12	\N	\N	john@daylilyinformatics.com	2025-07-03 06:40:42.323173-07	INSERT	\N	\N	\N	f	f
3673d5bb-100c-4b1d-b4fd-8b5946f013a1	generic_instance	\N	c54a4608-c3e9-4453-83b3-302ca39467f3	FS5	\N	\N	john@daylilyinformatics.com	2025-07-03 06:41:43.824381-07	INSERT	\N	\N	\N	f	f
ca1821d3-e056-4702-bbb4-a83ca2af14e3	generic_instance	\N	5aa56483-8480-4d03-a3f2-36506ae89d6c	FI7013	\N	\N	john@daylilyinformatics.com	2025-07-03 06:41:43.893844-07	INSERT	\N	\N	\N	f	f
84f2fcfc-c211-4c77-98b6-8f9b41db2683	generic_instance	json_addl	5aa56483-8480-4d03-a3f2-36506ae89d6c	FI7013	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134109", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134109", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:41:43.903867-07	UPDATE	\N	\N	\N	f	f
534764ee-261f-4c2e-bc6f-a3d9bb7b2c86	generic_instance	created_dt	5aa56483-8480-4d03-a3f2-36506ae89d6c	FI7013	2025-07-03 06:41:43.893844-07	2025-07-03T06:41:43.893844-07:00	john@daylilyinformatics.com	2025-07-03 06:41:43.903867-07	UPDATE	\N	\N	\N	f	f
9eb12eeb-adcf-4bb2-8323-8ab5f5f03f56	generic_instance	modified_dt	5aa56483-8480-4d03-a3f2-36506ae89d6c	FI7013	2025-07-03 06:41:43.893844-07	2025-07-03T06:41:43.903867-07:00	john@daylilyinformatics.com	2025-07-03 06:41:43.903867-07	UPDATE	\N	\N	\N	f	f
bb8d058a-0ab2-43f9-86e4-7d59fa4a00d6	generic_instance	json_addl	5aa56483-8480-4d03-a3f2-36506ae89d6c	FI7013	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134109", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "NA12878/HG001_GRCh38_1_22_v4.2.1_benchmark_hifiasm_v11_phasetransfer.PS_FIX.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7013.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7013.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134109", "original_file_md5": "", "original_file_name": "HG001_GRCh38_1_22_v4.2.1_benchmark_hifiasm_v11_phasetransfer.PS_FIX.vcf.gz", "original_file_path": "NA12878", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 111830815, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:41:44.167933-07	UPDATE	\N	\N	\N	f	f
1967a9e6-b3a5-48ec-923d-6b5fd7f23b82	generic_instance	created_dt	5aa56483-8480-4d03-a3f2-36506ae89d6c	FI7013	2025-07-03 06:41:43.893844-07	2025-07-03T06:41:43.893844-07:00	john@daylilyinformatics.com	2025-07-03 06:41:44.167933-07	UPDATE	\N	\N	\N	f	f
c5d2dbdc-e9f3-46d2-a643-de4462aeafdc	generic_instance	modified_dt	5aa56483-8480-4d03-a3f2-36506ae89d6c	FI7013	2025-07-03 06:41:43.903867-07	2025-07-03T06:41:44.167933-07:00	john@daylilyinformatics.com	2025-07-03 06:41:44.167933-07	UPDATE	\N	\N	\N	f	f
572e066f-f373-40fc-883a-5fc34637a912	generic_instance	\N	9312ed99-a2e8-45a4-9f4d-802f27c08ea5	FI7014	\N	\N	john@daylilyinformatics.com	2025-07-03 06:41:49.285358-07	INSERT	\N	\N	\N	f	f
02de4dd9-146a-490d-8b08-a3ccaebdbbbc	generic_instance	json_addl	9312ed99-a2e8-45a4-9f4d-802f27c08ea5	FI7014	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134109", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134109", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:41:49.300875-07	UPDATE	\N	\N	\N	f	f
81a13aae-e518-4048-b65b-b3a244eaa222	generic_instance	created_dt	9312ed99-a2e8-45a4-9f4d-802f27c08ea5	FI7014	2025-07-03 06:41:49.285358-07	2025-07-03T06:41:49.285358-07:00	john@daylilyinformatics.com	2025-07-03 06:41:49.300875-07	UPDATE	\N	\N	\N	f	f
2b17221b-06bc-4b97-89ed-c8249dfd0764	generic_instance	modified_dt	9312ed99-a2e8-45a4-9f4d-802f27c08ea5	FI7014	2025-07-03 06:41:49.285358-07	2025-07-03T06:41:49.300875-07:00	john@daylilyinformatics.com	2025-07-03 06:41:49.300875-07	UPDATE	\N	\N	\N	f	f
cd6fd8fc-5e52-4a31-800e-8ad2ef8a7fee	generic_instance	json_addl	9312ed99-a2e8-45a4-9f4d-802f27c08ea5	FI7014	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134109", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "NA12878/Lee_washington_0250E_19948.pdf", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "pdf", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7014.pdf", "current_s3_uri": "s3://daylily-dewey-0/0/FI7014.pdf", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134109", "original_file_md5": "", "original_file_name": "Lee_washington_0250E_19948.pdf", "original_file_path": "NA12878", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "pdf", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 11203779, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:41:49.405737-07	UPDATE	\N	\N	\N	f	f
aa4d3f63-ba26-42b4-9d1a-ef4c707d7e08	generic_instance	created_dt	9312ed99-a2e8-45a4-9f4d-802f27c08ea5	FI7014	2025-07-03 06:41:49.285358-07	2025-07-03T06:41:49.285358-07:00	john@daylilyinformatics.com	2025-07-03 06:41:49.405737-07	UPDATE	\N	\N	\N	f	f
a99fae26-08a6-4e67-aabb-05111200b72f	generic_instance	modified_dt	9312ed99-a2e8-45a4-9f4d-802f27c08ea5	FI7014	2025-07-03 06:41:49.300875-07	2025-07-03T06:41:49.405737-07:00	john@daylilyinformatics.com	2025-07-03 06:41:49.405737-07	UPDATE	\N	\N	\N	f	f
b2d9194e-ddda-42ef-b6c6-86043da91a6c	generic_instance	\N	ec941c78-3c25-4399-bb69-f266930774bb	CX8	\N	\N	daylily	2025-07-03 08:23:08.671329-07	INSERT	\N	\N	\N	f	f
6f730001-0138-4436-bacb-67b0960932b1	generic_instance	\N	a778a7e1-7154-40f8-90c2-8c264b021681	CWX85	\N	\N	daylily	2025-07-03 08:23:08.707623-07	INSERT	\N	\N	\N	f	f
81286f94-2b16-4950-8e82-d70a10a1710b	generic_instance	json_addl	a778a7e1-7154-40f8-90c2-8c264b021681	CWX85	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:08.7125-07	UPDATE	\N	\N	\N	f	f
d93d547b-6b8d-48f2-acf0-01dd1b749a48	generic_instance	created_dt	a778a7e1-7154-40f8-90c2-8c264b021681	CWX85	2025-07-03 08:23:08.707623-07	2025-07-03T08:23:08.707623-07:00	daylily	2025-07-03 08:23:08.7125-07	UPDATE	\N	\N	\N	f	f
2fe8dac8-ba34-4016-8776-9657bcac4ab7	generic_instance	modified_dt	a778a7e1-7154-40f8-90c2-8c264b021681	CWX85	2025-07-03 08:23:08.707623-07	2025-07-03T08:23:08.7125-07:00	daylily	2025-07-03 08:23:08.7125-07	UPDATE	\N	\N	\N	f	f
4d387983-a68b-48a9-886c-aae822ef4956	generic_instance_lineage	\N	bba10ec4-9246-4469-998b-88a4bd0a2163	GL303	\N	\N	daylily	2025-07-03 08:23:08.715457-07	INSERT	\N	\N	\N	f	f
0cdd4551-2a9d-4983-bac5-e93c147aa6f6	generic_instance	\N	ccf9f662-857d-49b6-8a43-8f9097e466f6	CWX86	\N	\N	daylily	2025-07-03 08:23:08.715457-07	INSERT	\N	\N	\N	f	f
07e02313-14d1-4c9e-b674-cf50592148a3	generic_instance	json_addl	ccf9f662-857d-49b6-8a43-8f9097e466f6	CWX86	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:08.721885-07	UPDATE	\N	\N	\N	f	f
9626abd6-98d9-4e1a-8140-0d28da5b9584	generic_instance	created_dt	ccf9f662-857d-49b6-8a43-8f9097e466f6	CWX86	2025-07-03 08:23:08.715457-07	2025-07-03T08:23:08.715457-07:00	daylily	2025-07-03 08:23:08.721885-07	UPDATE	\N	\N	\N	f	f
4ae97121-e044-43b3-8fec-b83cf7f27c05	generic_instance	modified_dt	ccf9f662-857d-49b6-8a43-8f9097e466f6	CWX86	2025-07-03 08:23:08.715457-07	2025-07-03T08:23:08.721885-07:00	daylily	2025-07-03 08:23:08.721885-07	UPDATE	\N	\N	\N	f	f
4731a241-7d56-493d-9fe4-d29d90307dc3	generic_instance_lineage	\N	7048bc4b-058d-4fe4-822f-f2b6d5368a5c	GL304	\N	\N	daylily	2025-07-03 08:23:08.723816-07	INSERT	\N	\N	\N	f	f
cc0c403d-e0b5-40f3-8e11-5fb91cad0f50	generic_instance	\N	7bffe3df-228e-49d0-b3d3-ea7566c97b58	CWX87	\N	\N	daylily	2025-07-03 08:23:08.723816-07	INSERT	\N	\N	\N	f	f
70a59970-0a3b-41c9-90e5-e4a2389c81ae	generic_instance	json_addl	d6dc35b7-cf4c-4e70-a692-c23346f61bcd	CX12	{"cogs": {"cost": "1.50", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "10ml tube generic", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "Generic 10ml tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "1.50", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "giab NA24385", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "Generic 10ml tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.494103-07	UPDATE	\N	\N	\N	f	f
357576af-985c-4d3b-b65b-b66db23615e9	generic_instance	created_dt	d6dc35b7-cf4c-4e70-a692-c23346f61bcd	CX12	2025-07-03 08:23:17.484168-07	2025-07-03T08:23:17.484168-07:00	daylily	2025-07-03 08:23:17.494103-07	UPDATE	\N	\N	\N	f	f
b12964b1-7ce1-43d8-9d65-0199fcf56a1e	generic_instance_lineage	\N	ad18c92a-a1c6-47dd-982e-0e5f5cb9ece8	GL14	\N	\N	john@daylilyinformatics.com	2025-07-03 06:41:50.40051-07	INSERT	\N	\N	\N	f	f
8152c5dd-48a5-461d-97ee-22ab492c7a5c	generic_instance	\N	d77e413a-c96e-4048-b4a7-a5ca4c00828a	FS6	\N	\N	john@daylilyinformatics.com	2025-07-03 06:42:12.436379-07	INSERT	\N	\N	\N	f	f
131e0432-737a-41a5-a4e7-07f665599286	generic_instance	\N	675337ff-ab88-4ba5-bce0-a80e5446111e	FI7015	\N	\N	john@daylilyinformatics.com	2025-07-03 06:42:12.453434-07	INSERT	\N	\N	\N	f	f
5a1e3545-dd59-48b0-b4b4-367c05f24b01	generic_instance	json_addl	675337ff-ab88-4ba5-bce0-a80e5446111e	FI7015	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T134212", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T134212", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:42:12.458295-07	UPDATE	\N	\N	\N	f	f
36046d8c-23ec-403b-8d1a-944ef27638f7	generic_instance	created_dt	675337ff-ab88-4ba5-bce0-a80e5446111e	FI7015	2025-07-03 06:42:12.453434-07	2025-07-03T06:42:12.453434-07:00	john@daylilyinformatics.com	2025-07-03 06:42:12.458295-07	UPDATE	\N	\N	\N	f	f
4bafc420-49fc-4ba9-8c56-1e9009a8d2fb	generic_instance	modified_dt	675337ff-ab88-4ba5-bce0-a80e5446111e	FI7015	2025-07-03 06:42:12.453434-07	2025-07-03T06:42:12.458295-07:00	john@daylilyinformatics.com	2025-07-03 06:42:12.458295-07	UPDATE	\N	\N	\N	f	f
4792c819-82a4-4ed2-897b-5ed8f6a6b003	generic_instance	json_addl	675337ff-ab88-4ba5-bce0-a80e5446111e	FI7015	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T134212", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7015.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7015.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T134212", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:42:12.710245-07	UPDATE	\N	\N	\N	f	f
61617ad9-090a-4a02-9ed3-7b1c69c6160f	generic_instance	created_dt	675337ff-ab88-4ba5-bce0-a80e5446111e	FI7015	2025-07-03 06:42:12.453434-07	2025-07-03T06:42:12.453434-07:00	john@daylilyinformatics.com	2025-07-03 06:42:12.710245-07	UPDATE	\N	\N	\N	f	f
e97b00a9-573f-4d65-b3bb-5184520a580c	generic_instance	modified_dt	675337ff-ab88-4ba5-bce0-a80e5446111e	FI7015	2025-07-03 06:42:12.458295-07	2025-07-03T06:42:12.710245-07:00	john@daylilyinformatics.com	2025-07-03 06:42:12.710245-07	UPDATE	\N	\N	\N	f	f
3db10a62-9f07-44a2-9ffc-5d0c2b2a5079	generic_instance_lineage	\N	cf6b66d1-4204-48e0-89b8-a1b3d93e056f	GL15	\N	\N	john@daylilyinformatics.com	2025-07-03 06:42:13.754991-07	INSERT	\N	\N	\N	f	f
76e5c51b-a3f8-43f7-9aa1-cd364cb64fd8	generic_instance	\N	199bb1af-804b-411a-b0ba-f270dc958bb5	FI7016	\N	\N	john@daylilyinformatics.com	2025-07-03 06:42:13.769262-07	INSERT	\N	\N	\N	f	f
8a4a5b35-f7ce-4475-9565-5181b338dcf9	generic_instance	json_addl	199bb1af-804b-411a-b0ba-f270dc958bb5	FI7016	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T134212", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T134212", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:42:13.964522-07	UPDATE	\N	\N	\N	f	f
8f6dd4ba-0fbf-443e-a47a-53d9c39cab65	generic_instance	created_dt	199bb1af-804b-411a-b0ba-f270dc958bb5	FI7016	2025-07-03 06:42:13.769262-07	2025-07-03T06:42:13.769262-07:00	john@daylilyinformatics.com	2025-07-03 06:42:13.964522-07	UPDATE	\N	\N	\N	f	f
4642e929-00e2-4fae-85f5-cd3abdedc5b1	generic_instance	modified_dt	199bb1af-804b-411a-b0ba-f270dc958bb5	FI7016	2025-07-03 06:42:13.769262-07	2025-07-03T06:42:13.964522-07:00	john@daylilyinformatics.com	2025-07-03 06:42:13.964522-07	UPDATE	\N	\N	\N	f	f
478167e6-6014-4ae5-973d-e9775ff52095	generic_instance	json_addl	199bb1af-804b-411a-b0ba-f270dc958bb5	FI7016	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T134212", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7016.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7016.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T134212", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:42:14.076936-07	UPDATE	\N	\N	\N	f	f
c6e9b8d5-679a-43a1-a5f1-7e72a353dc16	generic_instance	created_dt	199bb1af-804b-411a-b0ba-f270dc958bb5	FI7016	2025-07-03 06:42:13.769262-07	2025-07-03T06:42:13.769262-07:00	john@daylilyinformatics.com	2025-07-03 06:42:14.076936-07	UPDATE	\N	\N	\N	f	f
fe2e079f-2ad4-4955-81ba-5d839a10c8ac	generic_instance	modified_dt	199bb1af-804b-411a-b0ba-f270dc958bb5	FI7016	2025-07-03 06:42:13.964522-07	2025-07-03T06:42:14.076936-07:00	john@daylilyinformatics.com	2025-07-03 06:42:14.076936-07	UPDATE	\N	\N	\N	f	f
34d72f85-1b54-4f59-94e8-72cdeda88e92	generic_instance_lineage	\N	ca812806-312b-4c91-9783-4f3b844afe27	GL16	\N	\N	john@daylilyinformatics.com	2025-07-03 06:42:14.974898-07	INSERT	\N	\N	\N	f	f
eda1ead2-25f2-472e-91dc-8c400dfbfe43	generic_instance	\N	b5eca29d-be2c-4c68-8f21-61819fe90f20	FI7017	\N	\N	john@daylilyinformatics.com	2025-07-03 06:42:14.980617-07	INSERT	\N	\N	\N	f	f
3c42b2ca-b5ac-47c4-ba05-ee296259f7f6	generic_instance	json_addl	b5eca29d-be2c-4c68-8f21-61819fe90f20	FI7017	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T134212", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T134212", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:42:14.989124-07	UPDATE	\N	\N	\N	f	f
f6a58545-4b57-4c8e-94d0-b7c6c67cd7a4	generic_instance	created_dt	b5eca29d-be2c-4c68-8f21-61819fe90f20	FI7017	2025-07-03 06:42:14.980617-07	2025-07-03T06:42:14.980617-07:00	john@daylilyinformatics.com	2025-07-03 06:42:14.989124-07	UPDATE	\N	\N	\N	f	f
d6810100-5e37-412e-aab8-a297d6067260	generic_instance	modified_dt	b5eca29d-be2c-4c68-8f21-61819fe90f20	FI7017	2025-07-03 06:42:14.980617-07	2025-07-03T06:42:14.989124-07:00	john@daylilyinformatics.com	2025-07-03 06:42:14.989124-07	UPDATE	\N	\N	\N	f	f
46728d40-6800-4684-bb32-4b8ed825ccf3	generic_instance	json_addl	b5eca29d-be2c-4c68-8f21-61819fe90f20	FI7017	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T134212", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7017.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7017.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T134212", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:42:15.2375-07	UPDATE	\N	\N	\N	f	f
54a9b0b6-b147-4b00-9aee-d79cf2f49c2b	generic_instance	created_dt	b5eca29d-be2c-4c68-8f21-61819fe90f20	FI7017	2025-07-03 06:42:14.980617-07	2025-07-03T06:42:14.980617-07:00	john@daylilyinformatics.com	2025-07-03 06:42:15.2375-07	UPDATE	\N	\N	\N	f	f
8ccb289f-501d-4fd7-994b-1b08aeb038ac	generic_instance	modified_dt	b5eca29d-be2c-4c68-8f21-61819fe90f20	FI7017	2025-07-03 06:42:14.989124-07	2025-07-03T06:42:15.2375-07:00	john@daylilyinformatics.com	2025-07-03 06:42:15.2375-07	UPDATE	\N	\N	\N	f	f
a9cfefc0-a736-4116-b12e-8e09f949323c	generic_instance_lineage	\N	321da091-bd88-482d-81be-9b7b69c4a70d	GL17	\N	\N	john@daylilyinformatics.com	2025-07-03 06:42:15.911874-07	INSERT	\N	\N	\N	f	f
c94e5b15-2cd0-4393-8d3c-9ba23a798ef4	generic_instance	\N	7ee95105-e62b-4f4a-aadf-e76bc8bf12d1	FI7018	\N	\N	john@daylilyinformatics.com	2025-07-03 06:42:15.917897-07	INSERT	\N	\N	\N	f	f
794a6972-ccde-4894-b33b-830ee8867488	generic_instance	json_addl	7ee95105-e62b-4f4a-aadf-e76bc8bf12d1	FI7018	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T134212", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T134212", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:42:16.096322-07	UPDATE	\N	\N	\N	f	f
f2836dfd-2de6-4759-b383-8f93579c637c	generic_instance	created_dt	7ee95105-e62b-4f4a-aadf-e76bc8bf12d1	FI7018	2025-07-03 06:42:15.917897-07	2025-07-03T06:42:15.917897-07:00	john@daylilyinformatics.com	2025-07-03 06:42:16.096322-07	UPDATE	\N	\N	\N	f	f
a008f335-a0b0-414a-84c4-c2ad702c7c89	generic_instance	modified_dt	7ee95105-e62b-4f4a-aadf-e76bc8bf12d1	FI7018	2025-07-03 06:42:15.917897-07	2025-07-03T06:42:16.096322-07:00	john@daylilyinformatics.com	2025-07-03 06:42:16.096322-07	UPDATE	\N	\N	\N	f	f
f5157d4c-9109-4328-a2a5-a7054d2624c5	generic_instance	json_addl	7ee95105-e62b-4f4a-aadf-e76bc8bf12d1	FI7018	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T134212", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7018.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7018.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T134212", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:42:16.202901-07	UPDATE	\N	\N	\N	f	f
98c4beab-2878-4c03-a816-40927f00aa19	generic_instance	created_dt	7ee95105-e62b-4f4a-aadf-e76bc8bf12d1	FI7018	2025-07-03 06:42:15.917897-07	2025-07-03T06:42:15.917897-07:00	john@daylilyinformatics.com	2025-07-03 06:42:16.202901-07	UPDATE	\N	\N	\N	f	f
515fe6a8-6829-4a5a-af30-91bf66d74537	generic_instance	modified_dt	7ee95105-e62b-4f4a-aadf-e76bc8bf12d1	FI7018	2025-07-03 06:42:16.096322-07	2025-07-03T06:42:16.202901-07:00	john@daylilyinformatics.com	2025-07-03 06:42:16.202901-07	UPDATE	\N	\N	\N	f	f
2fae16b2-4414-4651-8f61-7fcc0230d5a5	generic_instance_lineage	\N	6c6fd1b8-10b5-4ffb-820e-e26fe831b976	GL18	\N	\N	john@daylilyinformatics.com	2025-07-03 06:42:16.846283-07	INSERT	\N	\N	\N	f	f
79753053-3026-4b54-ac39-2e92577ce78f	generic_instance	\N	d7eb218e-e19e-402c-b405-ba307911e0ff	FX2	\N	\N	john@daylilyinformatics.com	2025-07-03 06:42:33.390481-07	INSERT	\N	\N	\N	f	f
f825d92e-df7e-445a-b1b7-75aeeb668692	generic_instance_lineage	\N	540527c1-2db0-4e63-8243-f4df42e12b65	GL19	\N	\N	john@daylilyinformatics.com	2025-07-03 06:42:33.412063-07	INSERT	\N	\N	\N	f	f
bf0f7021-2fa7-44ec-8daf-e6c6baadfe62	generic_instance_lineage	\N	e7bf76f4-992f-4dca-88b2-dc064be5140f	GL20	\N	\N	john@daylilyinformatics.com	2025-07-03 06:42:33.412063-07	INSERT	\N	\N	\N	f	f
75e5c268-908c-451c-8ab3-0dc47bec4755	generic_instance	\N	6531f3d5-1217-4d66-93f7-7e40c65d72bf	FX3	\N	\N	john@daylilyinformatics.com	2025-07-03 06:42:33.420047-07	INSERT	\N	\N	\N	f	f
af19e872-fd96-4994-8aa7-446445a44840	generic_instance_lineage	\N	2d8a3441-2f76-4196-b853-d102dded27ec	GL21	\N	\N	john@daylilyinformatics.com	2025-07-03 06:42:33.460569-07	INSERT	\N	\N	\N	f	f
5b368f29-4c34-4560-a3a8-3cabae05b2b6	generic_instance_lineage	\N	f5e8b463-601f-46eb-8004-1956858c2087	GL22	\N	\N	john@daylilyinformatics.com	2025-07-03 06:42:33.460569-07	INSERT	\N	\N	\N	f	f
17fc596f-fa69-4415-9a73-eb9ec3e7e802	generic_instance	\N	b0c04ae3-a3bd-4721-9834-9cb6ae4f3526	FX4	\N	\N	john@daylilyinformatics.com	2025-07-03 06:42:33.469089-07	INSERT	\N	\N	\N	f	f
f55a2429-68d4-47ab-81cd-cab70ca6db98	generic_instance_lineage	\N	a8a79f41-7e7d-4177-842f-12b328b26065	GL23	\N	\N	john@daylilyinformatics.com	2025-07-03 06:42:33.514661-07	INSERT	\N	\N	\N	f	f
bec5dd53-3c6b-4c2c-aa67-6c5a3de53411	generic_instance_lineage	\N	d24ce294-8388-4b74-83ed-7fc017912513	GL24	\N	\N	john@daylilyinformatics.com	2025-07-03 06:42:33.514661-07	INSERT	\N	\N	\N	f	f
d3a104e0-e1a8-4419-82a1-b3f39b38cbc0	generic_instance	\N	9728d18d-696b-47f7-a548-33e93b36b0f0	FX5	\N	\N	john@daylilyinformatics.com	2025-07-03 06:42:33.521204-07	INSERT	\N	\N	\N	f	f
b911caa0-9c7f-4578-8c5a-573842b9b854	generic_instance_lineage	\N	1643f84d-a93c-43a6-bfea-90a8578428e3	GL25	\N	\N	john@daylilyinformatics.com	2025-07-03 06:42:33.566659-07	INSERT	\N	\N	\N	f	f
1c4f421c-9e9b-415e-b727-f69a938c4112	generic_instance_lineage	\N	8deb54e7-2959-42b4-80d6-feb2ec638056	GL26	\N	\N	john@daylilyinformatics.com	2025-07-03 06:42:33.566659-07	INSERT	\N	\N	\N	f	f
b22ecaaf-e7f5-4aff-8e4b-87777fcf7967	generic_instance	json_addl	0d2b2a11-5fa7-407f-9053-c2933c838bdf	FS3	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T130538", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "end_operator": {"email": "john@daylilyinformatics.com", "dag_fnv2": ""}, "creating_user": "", "end_timestamp": "2025-07-03 09:06:37 EDT-0400", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [{"email": "john@daylilyinformatics.com", "dag_fnv2": ""}], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "0", "max_executions": "-1", "action_executed": "1", "executed_datetime": ["2025-07-03 09:06:37 EDT-0400"], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T130538", "tags": [], "comments": "", "duration": 1.0, "lab_code": "", "ref_type": "presigned_url", "description": "Bulk upload", "end_operator": {"email": "john@daylilyinformatics.com", "dag_fnv2": ""}, "creating_user": "john@daylilyinformatics.com", "end_timestamp": "2025-07-03 09:06:37 EDT-0400", "rclone_config": {"host": "0.0.0.0", "port": 8080, "user": "user", "bucket": "daylily-dewey-0", "passwd": "passwd"}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [{"email": "john@daylilyinformatics.com", "dag_fnv2": ""}], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "0", "max_executions": "-1", "action_executed": "1", "executed_datetime": ["2025-07-03 09:06:37 EDT-0400"], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	john@daylilyinformatics.com	2025-07-03 06:42:33.38115-07	UPDATE	\N	\N	\N	f	f
6f65d595-faa3-4993-9370-4da3aeddf227	generic_instance	created_dt	0d2b2a11-5fa7-407f-9053-c2933c838bdf	FS3	2025-07-03 06:05:38.782513-07	2025-07-03T06:05:38.782513-07:00	john@daylilyinformatics.com	2025-07-03 06:42:33.38115-07	UPDATE	\N	\N	\N	f	f
b927a9d2-ba4a-4fcb-865e-a8e554f9067e	generic_instance	modified_dt	0d2b2a11-5fa7-407f-9053-c2933c838bdf	FS3	2025-07-03 06:06:37.203894-07	2025-07-03T06:42:33.38115-07:00	john@daylilyinformatics.com	2025-07-03 06:42:33.38115-07	UPDATE	\N	\N	\N	f	f
718457cb-76d4-4153-9726-15a3fb120285	generic_instance	\N	ce4a70c5-bcb6-4305-9736-b14390e6c4a2	FS7	\N	\N	john@daylilyinformatics.com	2025-07-03 06:43:21.789558-07	INSERT	\N	\N	\N	f	f
fd54f1b3-a6d8-4328-a2ed-3e9cd7796a7a	generic_instance	\N	0098cfb3-be50-4e6d-8633-fef7cc11846c	FI7019	\N	\N	john@daylilyinformatics.com	2025-07-03 06:43:21.841829-07	INSERT	\N	\N	\N	f	f
0c57bd14-3d78-4090-9b62-0cfc5c7e723b	generic_instance	json_addl	0098cfb3-be50-4e6d-8633-fef7cc11846c	FI7019	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Remote", "upload_group_key": "20250703T134315", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Remote", "upload_group_key": "20250703T134315", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:43:21.850084-07	UPDATE	\N	\N	\N	f	f
4eb2100e-b59d-441c-8e9c-a520c2e89c29	generic_instance	created_dt	0098cfb3-be50-4e6d-8633-fef7cc11846c	FI7019	2025-07-03 06:43:21.841829-07	2025-07-03T06:43:21.841829-07:00	john@daylilyinformatics.com	2025-07-03 06:43:21.850084-07	UPDATE	\N	\N	\N	f	f
fb6bd520-cb46-4241-a680-4a105b85f2e8	generic_instance	modified_dt	0098cfb3-be50-4e6d-8633-fef7cc11846c	FI7019	2025-07-03 06:43:21.841829-07	2025-07-03T06:43:21.850084-07:00	john@daylilyinformatics.com	2025-07-03 06:43:21.850084-07	UPDATE	\N	\N	\N	f	f
7343255e-a33a-4bde-ac11-61012f86a38a	generic_instance	json_addl	0098cfb3-be50-4e6d-8633-fef7cc11846c	FI7019	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Remote", "upload_group_key": "20250703T134315", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "Rare_Cancer_Research_Foundation_eval.lic", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "lic", "patient_id": "", "remote_url": "https://daylily-service-omics-analysis-us-west-2.s3.us-west-2.amazonaws.com/data/cached_envs/Rare_Cancer_Research_Foundation_eval.lic?response-content-disposition=inline&X-Amz-Content-Sha256=UNSIGNED-PAYLOAD&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEA4aCXVzLXdlc3QtMiJIMEYCIQCZXoyLH34gj8x7%2B3rf73TQmLO4DbXXuM8tknhpdj%2Bw3AIhAIrgR1QwhzUkSqhzS%2FlzNlkHH2LR0BeExhlYOOlC0E2aKrkDCBcQABoMNjcwNDg0MDUwNzM4IgwjfPjz6GvCEvqnDjEqlgPZc0VsvDt6%2BcJJMHJ23ouM15EJ1mTLlwiLz4J4nNIbTxJOOFRTu6qBJWmw2YGwoBZM4v5iMyBSMI4L4NYQy4RTEsCljp8S6EOzBhjfU604eTRuPT3qyzgyrD1PkBLCQl2KdRLOdz%2BFdWnDEvl9sqQoityKo%2BOJE7PTlmXcHTcQnPRh2Ns2b%2Br83xJrk%2BrAOBi4U4u9lNDFJztlr32cUCsbPE8GvnGZ%2FVcdJej%2B7E7LRoE82bzm3QGrkHBzls2OKOA1vqCQ59THyRi4Lyj7rRG%2B7z13moLLDUny1ivZSxjaNgAQUbE1lVEkpnUZEl7DLHlWlDPNVcW%2FPSd5kEc6VhYOWajuGwCX6HUYnixTu9VhbYsCk4jyTbAAkm2tBMq7znJeFv%2B4IZWG1Tqp5Tf%2B68YFX57h0u13rL7G1qSsWlU5s4k5fnmQHXnPm%2BNreE6mhBlgbrvlXUpm%2Bq1xL%2Bee3Jw6lx%2BCtjgude6iG%2FqiYGjWpNgsQH%2ByrvsJtPhfT7MthZQnBjbJnCmjY7viuMpF6btOSGsQ4RuUMIS6mMMGOt0Cit1Z8coQMzmtNe8V7viWPNQU1JTLgBAaIkHmCsqtFjS3VxpRsYTt%2FgM1GbCysKXYyOs5oVAfM9g%2F3BB2H5O824KPPEtU4NBybVXIvdAP%2B02ZVtn5aAaIReO%2BuJVhG0rhXp1ZWsbc6WP0rUhLKhZ3SN%2FnAGEcsTqH9H%2F69wQYhC95gFix%2BapN2H7Hc%2BQZjD2mHG%2FWa6de4I4b7S32pAFsshBg%2Bpaa9KnAZ3MydygxMU5MwO4kXsEWvaUNYnbLNlniQ9a1CT02KtRd6AkmI9ReLjANDsL665JzJkL5gI5YE%2Fr9menC%2FfL4Qegz1a7UNN3x22VhQxlycDVe3iJC%2Bz56ASG4Y%2FNp9P3hyFnZeQ9hvxnSmZEsU%2BHf%2BpUsBes0D8p2ltqLyQWOH02RfkA2ry73VjeYZaFaTHRVUQCAJipykllOXUbBTculx%2BuiP9tjDfTxnDaI5x4JMueXdAayZg%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAZYG7WWMZFG4S7ANK%2F20250703%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20250703T134309Z&X-Amz-Expires=18000&X-Amz-SignedHeaders=host&X-Amz-Signature=8a0bd2fd61a1c4f4c7519ba7ddcae103f73614e41c27e88c6615e23cbed7d3c0", "clinician_id": "", "external_ids": [], "original_url": "https://daylily-service-omics-analysis-us-west-2.s3.us-west-2.amazonaws.com/data/cached_envs/Rare_Cancer_Research_Foundation_eval.lic?response-content-disposition=inline&X-Amz-Content-Sha256=UNSIGNED-PAYLOAD&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEA4aCXVzLXdlc3QtMiJIMEYCIQCZXoyLH34gj8x7%2B3rf73TQmLO4DbXXuM8tknhpdj%2Bw3AIhAIrgR1QwhzUkSqhzS%2FlzNlkHH2LR0BeExhlYOOlC0E2aKrkDCBcQABoMNjcwNDg0MDUwNzM4IgwjfPjz6GvCEvqnDjEqlgPZc0VsvDt6%2BcJJMHJ23ouM15EJ1mTLlwiLz4J4nNIbTxJOOFRTu6qBJWmw2YGwoBZM4v5iMyBSMI4L4NYQy4RTEsCljp8S6EOzBhjfU604eTRuPT3qyzgyrD1PkBLCQl2KdRLOdz%2BFdWnDEvl9sqQoityKo%2BOJE7PTlmXcHTcQnPRh2Ns2b%2Br83xJrk%2BrAOBi4U4u9lNDFJztlr32cUCsbPE8GvnGZ%2FVcdJej%2B7E7LRoE82bzm3QGrkHBzls2OKOA1vqCQ59THyRi4Lyj7rRG%2B7z13moLLDUny1ivZSxjaNgAQUbE1lVEkpnUZEl7DLHlWlDPNVcW%2FPSd5kEc6VhYOWajuGwCX6HUYnixTu9VhbYsCk4jyTbAAkm2tBMq7znJeFv%2B4IZWG1Tqp5Tf%2B68YFX57h0u13rL7G1qSsWlU5s4k5fnmQHXnPm%2BNreE6mhBlgbrvlXUpm%2Bq1xL%2Bee3Jw6lx%2BCtjgude6iG%2FqiYGjWpNgsQH%2ByrvsJtPhfT7MthZQnBjbJnCmjY7viuMpF6btOSGsQ4RuUMIS6mMMGOt0Cit1Z8coQMzmtNe8V7viWPNQU1JTLgBAaIkHmCsqtFjS3VxpRsYTt%2FgM1GbCysKXYyOs5oVAfM9g%2F3BB2H5O824KPPEtU4NBybVXIvdAP%2B02ZVtn5aAaIReO%2BuJVhG0rhXp1ZWsbc6WP0rUhLKhZ3SN%2FnAGEcsTqH9H%2F69wQYhC95gFix%2BapN2H7Hc%2BQZjD2mHG%2FWa6de4I4b7S32pAFsshBg%2Bpaa9KnAZ3MydygxMU5MwO4kXsEWvaUNYnbLNlniQ9a1CT02KtRd6AkmI9ReLjANDsL665JzJkL5gI5YE%2Fr9menC%2FfL4Qegz1a7UNN3x22VhQxlycDVe3iJC%2Bz56ASG4Y%2FNp9P3hyFnZeQ9hvxnSmZEsU%2BHf%2BpUsBes0D8p2ltqLyQWOH02RfkA2ry73VjeYZaFaTHRVUQCAJipykllOXUbBTculx%2BuiP9tjDfTxnDaI5x4JMueXdAayZg%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAZYG7WWMZFG4S7ANK%2F20250703%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20250703T134309Z&X-Amz-Expires=18000&X-Amz-SignedHeaders=host&X-Amz-Signature=8a0bd2fd61a1c4f4c7519ba7ddcae103f73614e41c27e88c6615e23cbed7d3c0", "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "https://daylily-service-omics-analysis-us-west-2.s3.us-west-2.amazonaws.com/data/cached_envs/Rare_Cancer_Research_Foundation_eval.lic?response-content-disposition=inline&X-Amz-Content-Sha256=UNSIGNED-PAYLOAD&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEA4aCXVzLXdlc3QtMiJIMEYCIQCZXoyLH34gj8x7%2B3rf73TQmLO4DbXXuM8tknhpdj%2Bw3AIhAIrgR1QwhzUkSqhzS%2FlzNlkHH2LR0BeExhlYOOlC0E2aKrkDCBcQABoMNjcwNDg0MDUwNzM4IgwjfPjz6GvCEvqnDjEqlgPZc0VsvDt6%2BcJJMHJ23ouM15EJ1mTLlwiLz4J4nNIbTxJOOFRTu6qBJWmw2YGwoBZM4v5iMyBSMI4L4NYQy4RTEsCljp8S6EOzBhjfU604eTRuPT3qyzgyrD1PkBLCQl2KdRLOdz%2BFdWnDEvl9sqQoityKo%2BOJE7PTlmXcHTcQnPRh2Ns2b%2Br83xJrk%2BrAOBi4U4u9lNDFJztlr32cUCsbPE8GvnGZ%2FVcdJej%2B7E7LRoE82bzm3QGrkHBzls2OKOA1vqCQ59THyRi4Lyj7rRG%2B7z13moLLDUny1ivZSxjaNgAQUbE1lVEkpnUZEl7DLHlWlDPNVcW%2FPSd5kEc6VhYOWajuGwCX6HUYnixTu9VhbYsCk4jyTbAAkm2tBMq7znJeFv%2B4IZWG1Tqp5Tf%2B68YFX57h0u13rL7G1qSsWlU5s4k5fnmQHXnPm%2BNreE6mhBlgbrvlXUpm%2Bq1xL%2Bee3Jw6lx%2BCtjgude6iG%2FqiYGjWpNgsQH%2ByrvsJtPhfT7MthZQnBjbJnCmjY7viuMpF6btOSGsQ4RuUMIS6mMMGOt0Cit1Z8coQMzmtNe8V7viWPNQU1JTLgBAaIkHmCsqtFjS3VxpRsYTt%2FgM1GbCysKXYyOs5oVAfM9g%2F3BB2H5O824KPPEtU4NBybVXIvdAP%2B02ZVtn5aAaIReO%2BuJVhG0rhXp1ZWsbc6WP0rUhLKhZ3SN%2FnAGEcsTqH9H%2F69wQYhC95gFix%2BapN2H7Hc%2BQZjD2mHG%2FWa6de4I4b7S32pAFsshBg%2Bpaa9KnAZ3MydygxMU5MwO4kXsEWvaUNYnbLNlniQ9a1CT02KtRd6AkmI9ReLjANDsL665JzJkL5gI5YE%2Fr9menC%2FfL4Qegz1a7UNN3x22VhQxlycDVe3iJC%2Bz56ASG4Y%2FNp9P3hyFnZeQ9hvxnSmZEsU%2BHf%2BpUsBes0D8p2ltqLyQWOH02RfkA2ry73VjeYZaFaTHRVUQCAJipykllOXUbBTculx%2BuiP9tjDfTxnDaI5x4JMueXdAayZg%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAZYG7WWMZFG4S7ANK%2F20250703%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20250703T134309Z&X-Amz-Expires=18000&X-Amz-SignedHeaders=host&X-Amz-Signature=8a0bd2fd61a1c4f4c7519ba7ddcae103f73614e41c27e88c6615e23cbed7d3c0", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "remote", "upload_group_key": "20250703T134315", "original_file_md5": "", "original_file_name": "Rare_Cancer_Research_Foundation_eval.lic", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "lic", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:43:22.101299-07	UPDATE	\N	\N	\N	f	f
8047b274-c043-43bd-a94c-1475524d43df	generic_instance	created_dt	0098cfb3-be50-4e6d-8633-fef7cc11846c	FI7019	2025-07-03 06:43:21.841829-07	2025-07-03T06:43:21.841829-07:00	john@daylilyinformatics.com	2025-07-03 06:43:22.101299-07	UPDATE	\N	\N	\N	f	f
dac41e26-fe41-41ab-8c1d-8f3fd7548dd2	generic_instance	modified_dt	0098cfb3-be50-4e6d-8633-fef7cc11846c	FI7019	2025-07-03 06:43:21.850084-07	2025-07-03T06:43:22.101299-07:00	john@daylilyinformatics.com	2025-07-03 06:43:22.101299-07	UPDATE	\N	\N	\N	f	f
650a87a7-d3ae-41bb-b320-bb17dd5e4e1d	generic_instance_lineage	\N	0e5eb005-a07a-46ec-a552-71c5ecce7dbf	GL27	\N	\N	john@daylilyinformatics.com	2025-07-03 06:43:22.491105-07	INSERT	\N	\N	\N	f	f
c8d83a70-b24c-4e8d-9c3c-ee5b0680af64	generic_instance	\N	b444546e-7e18-4775-a0a5-6a48aaebd985	FS8	\N	\N	john@daylilyinformatics.com	2025-07-03 06:43:34.904833-07	INSERT	\N	\N	\N	f	f
d3624bd5-5596-41d5-aefe-07284252f61c	generic_instance	\N	4f291506-a724-4288-b37b-8a748c6eb737	FI7020	\N	\N	john@daylilyinformatics.com	2025-07-03 06:43:34.954475-07	INSERT	\N	\N	\N	f	f
c22769b1-2125-454c-84c3-17ba87046404	generic_instance	json_addl	4f291506-a724-4288-b37b-8a748c6eb737	FI7020	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134326", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134326", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:43:34.963133-07	UPDATE	\N	\N	\N	f	f
99837ca4-7f70-4f71-b283-828c82711acb	generic_instance	created_dt	4f291506-a724-4288-b37b-8a748c6eb737	FI7020	2025-07-03 06:43:34.954475-07	2025-07-03T06:43:34.954475-07:00	john@daylilyinformatics.com	2025-07-03 06:43:34.963133-07	UPDATE	\N	\N	\N	f	f
18f82aba-b95b-434f-b293-8f704ffe46bf	generic_instance	modified_dt	4f291506-a724-4288-b37b-8a748c6eb737	FI7020	2025-07-03 06:43:34.954475-07	2025-07-03T06:43:34.963133-07:00	john@daylilyinformatics.com	2025-07-03 06:43:34.963133-07	UPDATE	\N	\N	\N	f	f
ecb2543a-2120-4c6b-8455-c3da62778846	generic_instance	json_addl	4f291506-a724-4288-b37b-8a748c6eb737	FI7020	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134326", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "Rare_Cancer_Research_Foundation_eval.lic", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "lic", "patient_id": "", "clinician_id": "", "external_ids": [], "original_url": "https://daylily-service-omics-analysis-us-west-2.s3.us-west-2.amazonaws.com/data/cached_envs/Rare_Cancer_Research_Foundation_eval.lic?response-content-disposition=inline&X-Amz-Content-Sha256=UNSIGNED-PAYLOAD&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEA4aCXVzLXdlc3QtMiJIMEYCIQCZXoyLH34gj8x7%2B3rf73TQmLO4DbXXuM8tknhpdj%2Bw3AIhAIrgR1QwhzUkSqhzS%2FlzNlkHH2LR0BeExhlYOOlC0E2aKrkDCBcQABoMNjcwNDg0MDUwNzM4IgwjfPjz6GvCEvqnDjEqlgPZc0VsvDt6%2BcJJMHJ23ouM15EJ1mTLlwiLz4J4nNIbTxJOOFRTu6qBJWmw2YGwoBZM4v5iMyBSMI4L4NYQy4RTEsCljp8S6EOzBhjfU604eTRuPT3qyzgyrD1PkBLCQl2KdRLOdz%2BFdWnDEvl9sqQoityKo%2BOJE7PTlmXcHTcQnPRh2Ns2b%2Br83xJrk%2BrAOBi4U4u9lNDFJztlr32cUCsbPE8GvnGZ%2FVcdJej%2B7E7LRoE82bzm3QGrkHBzls2OKOA1vqCQ59THyRi4Lyj7rRG%2B7z13moLLDUny1ivZSxjaNgAQUbE1lVEkpnUZEl7DLHlWlDPNVcW%2FPSd5kEc6VhYOWajuGwCX6HUYnixTu9VhbYsCk4jyTbAAkm2tBMq7znJeFv%2B4IZWG1Tqp5Tf%2B68YFX57h0u13rL7G1qSsWlU5s4k5fnmQHXnPm%2BNreE6mhBlgbrvlXUpm%2Bq1xL%2Bee3Jw6lx%2BCtjgude6iG%2FqiYGjWpNgsQH%2ByrvsJtPhfT7MthZQnBjbJnCmjY7viuMpF6btOSGsQ4RuUMIS6mMMGOt0Cit1Z8coQMzmtNe8V7viWPNQU1JTLgBAaIkHmCsqtFjS3VxpRsYTt%2FgM1GbCysKXYyOs5oVAfM9g%2F3BB2H5O824KPPEtU4NBybVXIvdAP%2B02ZVtn5aAaIReO%2BuJVhG0rhXp1ZWsbc6WP0rUhLKhZ3SN%2FnAGEcsTqH9H%2F69wQYhC95gFix%2BapN2H7Hc%2BQZjD2mHG%2FWa6de4I4b7S32pAFsshBg%2Bpaa9KnAZ3MydygxMU5MwO4kXsEWvaUNYnbLNlniQ9a1CT02KtRd6AkmI9ReLjANDsL665JzJkL5gI5YE%2Fr9menC%2FfL4Qegz1a7UNN3x22VhQxlycDVe3iJC%2Bz56ASG4Y%2FNp9P3hyFnZeQ9hvxnSmZEsU%2BHf%2BpUsBes0D8p2ltqLyQWOH02RfkA2ry73VjeYZaFaTHRVUQCAJipykllOXUbBTculx%2BuiP9tjDfTxnDaI5x4JMueXdAayZg%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAZYG7WWMZFG4S7ANK%2F20250703%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20250703T134309Z&X-Amz-Expires=18000&X-Amz-SignedHeaders=host&X-Amz-Signature=8a0bd2fd61a1c4f4c7519ba7ddcae103f73614e41c27e88c6615e23cbed7d3c0", "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7020.lic", "current_s3_uri": "s3://daylily-dewey-0/0/FI7020.lic", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134326", "original_file_md5": "", "original_file_name": "Rare_Cancer_Research_Foundation_eval.lic", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "lic", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 4732, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:43:35.204833-07	UPDATE	\N	\N	\N	f	f
1bc1bb9b-e038-4474-ae39-b9f3f861e01d	generic_instance	created_dt	4f291506-a724-4288-b37b-8a748c6eb737	FI7020	2025-07-03 06:43:34.954475-07	2025-07-03T06:43:34.954475-07:00	john@daylilyinformatics.com	2025-07-03 06:43:35.204833-07	UPDATE	\N	\N	\N	f	f
f36b5ec8-f0b2-4e8d-82b7-8417e8fc9c25	generic_instance	modified_dt	4f291506-a724-4288-b37b-8a748c6eb737	FI7020	2025-07-03 06:43:34.963133-07	2025-07-03T06:43:35.204833-07:00	john@daylilyinformatics.com	2025-07-03 06:43:35.204833-07	UPDATE	\N	\N	\N	f	f
ddd1b4e3-4724-4bb5-abbf-33c03d282160	generic_instance_lineage	\N	de008fbc-7243-4bdb-a8fa-f6761bb05959	GL28	\N	\N	john@daylilyinformatics.com	2025-07-03 06:43:35.9415-07	INSERT	\N	\N	\N	f	f
d28bc03a-4f18-43f6-8c9d-ad17eed8750c	generic_instance	\N	2321b538-2fee-484b-8d18-7b6df27371ea	FS9	\N	\N	john@daylilyinformatics.com	2025-07-03 06:44:50.554495-07	INSERT	\N	\N	\N	f	f
37b64b73-4582-464d-b68b-2e130f7d78ee	generic_instance	\N	bee9e98b-3b9e-4fe1-8729-1b8fd38bac40	FI7021	\N	\N	john@daylilyinformatics.com	2025-07-03 06:44:50.612932-07	INSERT	\N	\N	\N	f	f
b8f10784-47a1-4721-b9d7-7a08fbe37224	generic_instance	json_addl	bee9e98b-3b9e-4fe1-8729-1b8fd38bac40	FI7021	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134442", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134442", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:44:50.622083-07	UPDATE	\N	\N	\N	f	f
89e35bd2-d5bc-4e6a-acbd-868bd2182e2a	generic_instance	created_dt	bee9e98b-3b9e-4fe1-8729-1b8fd38bac40	FI7021	2025-07-03 06:44:50.612932-07	2025-07-03T06:44:50.612932-07:00	john@daylilyinformatics.com	2025-07-03 06:44:50.622083-07	UPDATE	\N	\N	\N	f	f
ae8ce9e2-6596-4557-9076-ed24b13daf0d	generic_instance	modified_dt	bee9e98b-3b9e-4fe1-8729-1b8fd38bac40	FI7021	2025-07-03 06:44:50.612932-07	2025-07-03T06:44:50.622083-07:00	john@daylilyinformatics.com	2025-07-03 06:44:50.622083-07	UPDATE	\N	\N	\N	f	f
7a7b1480-0c9a-494c-bcc5-7213ca8e2a2a	generic_instance	json_addl	bee9e98b-3b9e-4fe1-8729-1b8fd38bac40	FI7021	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134442", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "benchmarks_summary.tsv.old", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "old", "patient_id": "", "clinician_id": "", "external_ids": [], "original_url": "https://daylily-service-omics-analysis-us-west-2.s3.us-west-2.amazonaws.com/FSxLustre20250123T214325Z/analysis_results/ubuntu/hg38/remaining/daylily/benchmarks_summary.tsv.old?response-content-disposition=inline&X-Amz-Content-Sha256=UNSIGNED-PAYLOAD&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEA4aCXVzLXdlc3QtMiJIMEYCIQCZXoyLH34gj8x7%2B3rf73TQmLO4DbXXuM8tknhpdj%2Bw3AIhAIrgR1QwhzUkSqhzS%2FlzNlkHH2LR0BeExhlYOOlC0E2aKrkDCBcQABoMNjcwNDg0MDUwNzM4IgwjfPjz6GvCEvqnDjEqlgPZc0VsvDt6%2BcJJMHJ23ouM15EJ1mTLlwiLz4J4nNIbTxJOOFRTu6qBJWmw2YGwoBZM4v5iMyBSMI4L4NYQy4RTEsCljp8S6EOzBhjfU604eTRuPT3qyzgyrD1PkBLCQl2KdRLOdz%2BFdWnDEvl9sqQoityKo%2BOJE7PTlmXcHTcQnPRh2Ns2b%2Br83xJrk%2BrAOBi4U4u9lNDFJztlr32cUCsbPE8GvnGZ%2FVcdJej%2B7E7LRoE82bzm3QGrkHBzls2OKOA1vqCQ59THyRi4Lyj7rRG%2B7z13moLLDUny1ivZSxjaNgAQUbE1lVEkpnUZEl7DLHlWlDPNVcW%2FPSd5kEc6VhYOWajuGwCX6HUYnixTu9VhbYsCk4jyTbAAkm2tBMq7znJeFv%2B4IZWG1Tqp5Tf%2B68YFX57h0u13rL7G1qSsWlU5s4k5fnmQHXnPm%2BNreE6mhBlgbrvlXUpm%2Bq1xL%2Bee3Jw6lx%2BCtjgude6iG%2FqiYGjWpNgsQH%2ByrvsJtPhfT7MthZQnBjbJnCmjY7viuMpF6btOSGsQ4RuUMIS6mMMGOt0Cit1Z8coQMzmtNe8V7viWPNQU1JTLgBAaIkHmCsqtFjS3VxpRsYTt%2FgM1GbCysKXYyOs5oVAfM9g%2F3BB2H5O824KPPEtU4NBybVXIvdAP%2B02ZVtn5aAaIReO%2BuJVhG0rhXp1ZWsbc6WP0rUhLKhZ3SN%2FnAGEcsTqH9H%2F69wQYhC95gFix%2BapN2H7Hc%2BQZjD2mHG%2FWa6de4I4b7S32pAFsshBg%2Bpaa9KnAZ3MydygxMU5MwO4kXsEWvaUNYnbLNlniQ9a1CT02KtRd6AkmI9ReLjANDsL665JzJkL5gI5YE%2Fr9menC%2FfL4Qegz1a7UNN3x22VhQxlycDVe3iJC%2Bz56ASG4Y%2FNp9P3hyFnZeQ9hvxnSmZEsU%2BHf%2BpUsBes0D8p2ltqLyQWOH02RfkA2ry73VjeYZaFaTHRVUQCAJipykllOXUbBTculx%2BuiP9tjDfTxnDaI5x4JMueXdAayZg%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAZYG7WWMZFG4S7ANK%2F20250703%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20250703T134434Z&X-Amz-Expires=18000&X-Amz-SignedHeaders=host&X-Amz-Signature=ae73c785dbdc400295532cbf30b18e254b86e4ae4d015a2c61fb2712337fb1b7", "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7021.old", "current_s3_uri": "s3://daylily-dewey-0/0/FI7021.old", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134442", "original_file_md5": "", "original_file_name": "benchmarks_summary.tsv.old", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "old", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 179, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:44:50.874923-07	UPDATE	\N	\N	\N	f	f
0ccfdecd-e473-46d5-8421-25bdd37e54d7	generic_instance	created_dt	bee9e98b-3b9e-4fe1-8729-1b8fd38bac40	FI7021	2025-07-03 06:44:50.612932-07	2025-07-03T06:44:50.612932-07:00	john@daylilyinformatics.com	2025-07-03 06:44:50.874923-07	UPDATE	\N	\N	\N	f	f
f6f11cfd-b79b-4d60-8fc6-9d0bffa579e8	generic_instance	modified_dt	bee9e98b-3b9e-4fe1-8729-1b8fd38bac40	FI7021	2025-07-03 06:44:50.622083-07	2025-07-03T06:44:50.874923-07:00	john@daylilyinformatics.com	2025-07-03 06:44:50.874923-07	UPDATE	\N	\N	\N	f	f
95ca0694-a705-42bd-8841-0fa7bee3c4ba	generic_instance_lineage	\N	8bf7dc19-aa7f-420c-a2a6-2348ce9a21fa	GL29	\N	\N	john@daylilyinformatics.com	2025-07-03 06:44:51.517314-07	INSERT	\N	\N	\N	f	f
741e71d6-918e-467e-8bf2-5780532f4fe8	generic_instance	json_addl	bee9e98b-3b9e-4fe1-8729-1b8fd38bac40	FI7021	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "benchmarks_summary.tsv.old", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "old", "patient_id": "", "clinician_id": "", "external_ids": [], "original_url": "https://daylily-service-omics-analysis-us-west-2.s3.us-west-2.amazonaws.com/FSxLustre20250123T214325Z/analysis_results/ubuntu/hg38/remaining/daylily/benchmarks_summary.tsv.old?response-content-disposition=inline&X-Amz-Content-Sha256=UNSIGNED-PAYLOAD&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEA4aCXVzLXdlc3QtMiJIMEYCIQCZXoyLH34gj8x7%2B3rf73TQmLO4DbXXuM8tknhpdj%2Bw3AIhAIrgR1QwhzUkSqhzS%2FlzNlkHH2LR0BeExhlYOOlC0E2aKrkDCBcQABoMNjcwNDg0MDUwNzM4IgwjfPjz6GvCEvqnDjEqlgPZc0VsvDt6%2BcJJMHJ23ouM15EJ1mTLlwiLz4J4nNIbTxJOOFRTu6qBJWmw2YGwoBZM4v5iMyBSMI4L4NYQy4RTEsCljp8S6EOzBhjfU604eTRuPT3qyzgyrD1PkBLCQl2KdRLOdz%2BFdWnDEvl9sqQoityKo%2BOJE7PTlmXcHTcQnPRh2Ns2b%2Br83xJrk%2BrAOBi4U4u9lNDFJztlr32cUCsbPE8GvnGZ%2FVcdJej%2B7E7LRoE82bzm3QGrkHBzls2OKOA1vqCQ59THyRi4Lyj7rRG%2B7z13moLLDUny1ivZSxjaNgAQUbE1lVEkpnUZEl7DLHlWlDPNVcW%2FPSd5kEc6VhYOWajuGwCX6HUYnixTu9VhbYsCk4jyTbAAkm2tBMq7znJeFv%2B4IZWG1Tqp5Tf%2B68YFX57h0u13rL7G1qSsWlU5s4k5fnmQHXnPm%2BNreE6mhBlgbrvlXUpm%2Bq1xL%2Bee3Jw6lx%2BCtjgude6iG%2FqiYGjWpNgsQH%2ByrvsJtPhfT7MthZQnBjbJnCmjY7viuMpF6btOSGsQ4RuUMIS6mMMGOt0Cit1Z8coQMzmtNe8V7viWPNQU1JTLgBAaIkHmCsqtFjS3VxpRsYTt%2FgM1GbCysKXYyOs5oVAfM9g%2F3BB2H5O824KPPEtU4NBybVXIvdAP%2B02ZVtn5aAaIReO%2BuJVhG0rhXp1ZWsbc6WP0rUhLKhZ3SN%2FnAGEcsTqH9H%2F69wQYhC95gFix%2BapN2H7Hc%2BQZjD2mHG%2FWa6de4I4b7S32pAFsshBg%2Bpaa9KnAZ3MydygxMU5MwO4kXsEWvaUNYnbLNlniQ9a1CT02KtRd6AkmI9ReLjANDsL665JzJkL5gI5YE%2Fr9menC%2FfL4Qegz1a7UNN3x22VhQxlycDVe3iJC%2Bz56ASG4Y%2FNp9P3hyFnZeQ9hvxnSmZEsU%2BHf%2BpUsBes0D8p2ltqLyQWOH02RfkA2ry73VjeYZaFaTHRVUQCAJipykllOXUbBTculx%2BuiP9tjDfTxnDaI5x4JMueXdAayZg%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAZYG7WWMZFG4S7ANK%2F20250703%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20250703T134434Z&X-Amz-Expires=18000&X-Amz-SignedHeaders=host&X-Amz-Signature=ae73c785dbdc400295532cbf30b18e254b86e4ae4d015a2c61fb2712337fb1b7", "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7021.old", "current_s3_uri": "s3://daylily-dewey-0/0/FI7021.old", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134442", "original_file_md5": "", "original_file_name": "benchmarks_summary.tsv.old", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "old", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 179, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "benchmarks_summary.tsv.old", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "old", "patient_id": "", "clinician_id": "", "external_ids": [], "original_url": "https://daylily-service-omics-analysis-us-west-2.s3.us-west-2.amazonaws.com/FSxLustre20250123T214325Z/analysis_results/ubuntu/hg38/remaining/daylily/benchmarks_summary.tsv.old?response-content-disposition=inline&X-Amz-Content-Sha256=UNSIGNED-PAYLOAD&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEA4aCXVzLXdlc3QtMiJIMEYCIQCZXoyLH34gj8x7%2B3rf73TQmLO4DbXXuM8tknhpdj%2Bw3AIhAIrgR1QwhzUkSqhzS%2FlzNlkHH2LR0BeExhlYOOlC0E2aKrkDCBcQABoMNjcwNDg0MDUwNzM4IgwjfPjz6GvCEvqnDjEqlgPZc0VsvDt6%2BcJJMHJ23ouM15EJ1mTLlwiLz4J4nNIbTxJOOFRTu6qBJWmw2YGwoBZM4v5iMyBSMI4L4NYQy4RTEsCljp8S6EOzBhjfU604eTRuPT3qyzgyrD1PkBLCQl2KdRLOdz%2BFdWnDEvl9sqQoityKo%2BOJE7PTlmXcHTcQnPRh2Ns2b%2Br83xJrk%2BrAOBi4U4u9lNDFJztlr32cUCsbPE8GvnGZ%2FVcdJej%2B7E7LRoE82bzm3QGrkHBzls2OKOA1vqCQ59THyRi4Lyj7rRG%2B7z13moLLDUny1ivZSxjaNgAQUbE1lVEkpnUZEl7DLHlWlDPNVcW%2FPSd5kEc6VhYOWajuGwCX6HUYnixTu9VhbYsCk4jyTbAAkm2tBMq7znJeFv%2B4IZWG1Tqp5Tf%2B68YFX57h0u13rL7G1qSsWlU5s4k5fnmQHXnPm%2BNreE6mhBlgbrvlXUpm%2Bq1xL%2Bee3Jw6lx%2BCtjgude6iG%2FqiYGjWpNgsQH%2ByrvsJtPhfT7MthZQnBjbJnCmjY7viuMpF6btOSGsQ4RuUMIS6mMMGOt0Cit1Z8coQMzmtNe8V7viWPNQU1JTLgBAaIkHmCsqtFjS3VxpRsYTt%2FgM1GbCysKXYyOs5oVAfM9g%2F3BB2H5O824KPPEtU4NBybVXIvdAP%2B02ZVtn5aAaIReO%2BuJVhG0rhXp1ZWsbc6WP0rUhLKhZ3SN%2FnAGEcsTqH9H%2F69wQYhC95gFix%2BapN2H7Hc%2BQZjD2mHG%2FWa6de4I4b7S32pAFsshBg%2Bpaa9KnAZ3MydygxMU5MwO4kXsEWvaUNYnbLNlniQ9a1CT02KtRd6AkmI9ReLjANDsL665JzJkL5gI5YE%2Fr9menC%2FfL4Qegz1a7UNN3x22VhQxlycDVe3iJC%2Bz56ASG4Y%2FNp9P3hyFnZeQ9hvxnSmZEsU%2BHf%2BpUsBes0D8p2ltqLyQWOH02RfkA2ry73VjeYZaFaTHRVUQCAJipykllOXUbBTculx%2BuiP9tjDfTxnDaI5x4JMueXdAayZg%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAZYG7WWMZFG4S7ANK%2F20250703%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20250703T134434Z&X-Amz-Expires=18000&X-Amz-SignedHeaders=host&X-Amz-Signature=ae73c785dbdc400295532cbf30b18e254b86e4ae4d015a2c61fb2712337fb1b7", "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7021.old", "current_s3_uri": "s3://daylily-dewey-0/0/FI7021.old", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134442", "original_file_md5": "", "original_file_name": "benchmarks_summary.tsv.old", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "old", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 179, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [{"email": "john@daylilyinformatics.com", "dag_fnv2": ""}], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "1", "executed_datetime": ["2025-07-03 09:45:06 EDT-0400"], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:45:06.36426-07	UPDATE	\N	\N	\N	f	f
65dd45de-d8d7-4bfe-a34a-509e3eb5eebf	generic_instance_lineage	\N	5e1c9097-7b4d-4b95-bff7-048394f6b772	GL45	\N	\N	john@daylilyinformatics.com	2025-07-03 07:33:46.124552-07	INSERT	\N	\N	\N	f	f
e910bea3-9000-40f4-835c-1a1fa2dc7dc4	generic_instance	created_dt	bee9e98b-3b9e-4fe1-8729-1b8fd38bac40	FI7021	2025-07-03 06:44:50.612932-07	2025-07-03T06:44:50.612932-07:00	john@daylilyinformatics.com	2025-07-03 06:45:06.36426-07	UPDATE	\N	\N	\N	f	f
f24c7e56-dbcb-4fb6-9bde-3256f4878418	generic_instance	modified_dt	bee9e98b-3b9e-4fe1-8729-1b8fd38bac40	FI7021	2025-07-03 06:44:50.874923-07	2025-07-03T06:45:06.36426-07:00	john@daylilyinformatics.com	2025-07-03 06:45:06.36426-07	UPDATE	\N	\N	\N	f	f
2a8e5084-4019-4a9e-bced-d77541841347	generic_instance	\N	88d739e7-9eb0-4083-af3b-265a8d380792	FS10	\N	\N	john@daylilyinformatics.com	2025-07-03 06:46:19.410805-07	INSERT	\N	\N	\N	f	f
a589383c-a334-467d-af02-48237488539b	generic_instance	\N	d512ba32-ddef-4c26-acae-5dda9273bab1	FI7022	\N	\N	john@daylilyinformatics.com	2025-07-03 06:46:19.522793-07	INSERT	\N	\N	\N	f	f
6a6e7553-bb84-4f1b-ba9c-f26da182d43d	generic_instance	json_addl	d512ba32-ddef-4c26-acae-5dda9273bab1	FI7022	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134607", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134607", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:46:19.541211-07	UPDATE	\N	\N	\N	f	f
1e1bd0fe-d60b-4768-8987-58c931f863fb	generic_instance	created_dt	d512ba32-ddef-4c26-acae-5dda9273bab1	FI7022	2025-07-03 06:46:19.522793-07	2025-07-03T06:46:19.522793-07:00	john@daylilyinformatics.com	2025-07-03 06:46:19.541211-07	UPDATE	\N	\N	\N	f	f
61345730-6b6a-41bc-b653-2fa6c821fccc	generic_instance	modified_dt	d512ba32-ddef-4c26-acae-5dda9273bab1	FI7022	2025-07-03 06:46:19.522793-07	2025-07-03T06:46:19.541211-07:00	john@daylilyinformatics.com	2025-07-03 06:46:19.541211-07	UPDATE	\N	\N	\N	f	f
63a3a6a7-cf3f-41eb-8173-bdecaf5713a1	generic_instance	json_addl	d512ba32-ddef-4c26-acae-5dda9273bab1	FI7022	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134607", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "benchmarks_summary.tsv", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "tsv", "patient_id": "", "clinician_id": "", "external_ids": [], "original_url": "https://daylily-service-omics-analysis-us-west-2.s3.us-west-2.amazonaws.com/FSxLustre20250123T214325Z/analysis_results/ubuntu/hg38/remaining/daylily/benchmarks_summary.tsv?response-content-disposition=inline&X-Amz-Content-Sha256=UNSIGNED-PAYLOAD&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEA4aCXVzLXdlc3QtMiJIMEYCIQCZXoyLH34gj8x7%2B3rf73TQmLO4DbXXuM8tknhpdj%2Bw3AIhAIrgR1QwhzUkSqhzS%2FlzNlkHH2LR0BeExhlYOOlC0E2aKrkDCBcQABoMNjcwNDg0MDUwNzM4IgwjfPjz6GvCEvqnDjEqlgPZc0VsvDt6%2BcJJMHJ23ouM15EJ1mTLlwiLz4J4nNIbTxJOOFRTu6qBJWmw2YGwoBZM4v5iMyBSMI4L4NYQy4RTEsCljp8S6EOzBhjfU604eTRuPT3qyzgyrD1PkBLCQl2KdRLOdz%2BFdWnDEvl9sqQoityKo%2BOJE7PTlmXcHTcQnPRh2Ns2b%2Br83xJrk%2BrAOBi4U4u9lNDFJztlr32cUCsbPE8GvnGZ%2FVcdJej%2B7E7LRoE82bzm3QGrkHBzls2OKOA1vqCQ59THyRi4Lyj7rRG%2B7z13moLLDUny1ivZSxjaNgAQUbE1lVEkpnUZEl7DLHlWlDPNVcW%2FPSd5kEc6VhYOWajuGwCX6HUYnixTu9VhbYsCk4jyTbAAkm2tBMq7znJeFv%2B4IZWG1Tqp5Tf%2B68YFX57h0u13rL7G1qSsWlU5s4k5fnmQHXnPm%2BNreE6mhBlgbrvlXUpm%2Bq1xL%2Bee3Jw6lx%2BCtjgude6iG%2FqiYGjWpNgsQH%2ByrvsJtPhfT7MthZQnBjbJnCmjY7viuMpF6btOSGsQ4RuUMIS6mMMGOt0Cit1Z8coQMzmtNe8V7viWPNQU1JTLgBAaIkHmCsqtFjS3VxpRsYTt%2FgM1GbCysKXYyOs5oVAfM9g%2F3BB2H5O824KPPEtU4NBybVXIvdAP%2B02ZVtn5aAaIReO%2BuJVhG0rhXp1ZWsbc6WP0rUhLKhZ3SN%2FnAGEcsTqH9H%2F69wQYhC95gFix%2BapN2H7Hc%2BQZjD2mHG%2FWa6de4I4b7S32pAFsshBg%2Bpaa9KnAZ3MydygxMU5MwO4kXsEWvaUNYnbLNlniQ9a1CT02KtRd6AkmI9ReLjANDsL665JzJkL5gI5YE%2Fr9menC%2FfL4Qegz1a7UNN3x22VhQxlycDVe3iJC%2Bz56ASG4Y%2FNp9P3hyFnZeQ9hvxnSmZEsU%2BHf%2BpUsBes0D8p2ltqLyQWOH02RfkA2ry73VjeYZaFaTHRVUQCAJipykllOXUbBTculx%2BuiP9tjDfTxnDaI5x4JMueXdAayZg%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAZYG7WWMZFG4S7ANK%2F20250703%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20250703T134602Z&X-Amz-Expires=18000&X-Amz-SignedHeaders=host&X-Amz-Signature=e93a920e5c7fba5ed1f0b926fde5ba9498ec24e4ac83995dd935dad70a49f04e", "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7022.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7022.tsv", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134607", "original_file_md5": "", "original_file_name": "benchmarks_summary.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 742279, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 06:46:19.801673-07	UPDATE	\N	\N	\N	f	f
ba8caaa1-3f3c-4516-9dd3-3a0b935214b2	generic_instance	created_dt	d512ba32-ddef-4c26-acae-5dda9273bab1	FI7022	2025-07-03 06:46:19.522793-07	2025-07-03T06:46:19.522793-07:00	john@daylilyinformatics.com	2025-07-03 06:46:19.801673-07	UPDATE	\N	\N	\N	f	f
02a94067-47c2-4e09-958e-0e74a439b836	generic_instance	modified_dt	d512ba32-ddef-4c26-acae-5dda9273bab1	FI7022	2025-07-03 06:46:19.541211-07	2025-07-03T06:46:19.801673-07:00	john@daylilyinformatics.com	2025-07-03 06:46:19.801673-07	UPDATE	\N	\N	\N	f	f
bf0badad-0e56-4aa7-b2f1-2bdebb6b2af4	generic_instance_lineage	\N	c5c3501a-a482-4872-8cd5-80135dc2a955	GL30	\N	\N	john@daylilyinformatics.com	2025-07-03 06:46:20.996889-07	INSERT	\N	\N	\N	f	f
85034eb0-14d2-4584-935b-f1bc88a6095b	generic_instance	\N	7e52fd3c-eedc-4e34-8caf-de2f5ad475fa	FS11	\N	\N	john@daylilyinformatics.com	2025-07-03 07:26:30.457615-07	INSERT	\N	\N	\N	f	f
774ec91d-efaf-4db9-9cb9-58de44274677	generic_instance	\N	f835a3a0-a807-41d2-9f29-68d149957d8d	FI7023	\N	\N	john@daylilyinformatics.com	2025-07-03 07:26:30.810766-07	INSERT	\N	\N	\N	f	f
795a119e-b4a2-4174-8d8d-a109bd9ac29c	generic_instance	json_addl	f835a3a0-a807-41d2-9f29-68d149957d8d	FI7023	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142630", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142630", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:26:30.821645-07	UPDATE	\N	\N	\N	f	f
7992320c-47c9-45ef-a391-2dd6eba4e9a8	generic_instance	created_dt	f835a3a0-a807-41d2-9f29-68d149957d8d	FI7023	2025-07-03 07:26:30.810766-07	2025-07-03T07:26:30.810766-07:00	john@daylilyinformatics.com	2025-07-03 07:26:30.821645-07	UPDATE	\N	\N	\N	f	f
9aca6c66-6db4-4d8d-861d-5ee5656c6551	generic_instance	modified_dt	f835a3a0-a807-41d2-9f29-68d149957d8d	FI7023	2025-07-03 07:26:30.810766-07	2025-07-03T07:26:30.821645-07:00	john@daylilyinformatics.com	2025-07-03 07:26:30.821645-07	UPDATE	\N	\N	\N	f	f
b09933ec-32d7-43d7-b709-e9893db62bd1	generic_instance	json_addl	f835a3a0-a807-41d2-9f29-68d149957d8d	FI7023	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142630", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7023.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7023.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142630", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:26:31.242334-07	UPDATE	\N	\N	\N	f	f
7ca7c1f3-38cc-4c10-bc48-23bc89812c96	generic_instance	created_dt	f835a3a0-a807-41d2-9f29-68d149957d8d	FI7023	2025-07-03 07:26:30.810766-07	2025-07-03T07:26:30.810766-07:00	john@daylilyinformatics.com	2025-07-03 07:26:31.242334-07	UPDATE	\N	\N	\N	f	f
a97b1fc8-0a26-480d-8299-9cc00f89be66	generic_instance	modified_dt	f835a3a0-a807-41d2-9f29-68d149957d8d	FI7023	2025-07-03 07:26:30.821645-07	2025-07-03T07:26:31.242334-07:00	john@daylilyinformatics.com	2025-07-03 07:26:31.242334-07	UPDATE	\N	\N	\N	f	f
fb19d981-1671-4221-96a0-8f49cfbdb61f	generic_instance_lineage	\N	00cf7c28-8134-4786-a02f-e5acd0185660	GL31	\N	\N	john@daylilyinformatics.com	2025-07-03 07:26:34.211728-07	INSERT	\N	\N	\N	f	f
7fab408e-74d0-47b9-ad47-9199049b04a2	generic_instance	\N	6ad2a1ab-3b91-4c5b-b4e6-69d087c868f0	FI7024	\N	\N	john@daylilyinformatics.com	2025-07-03 07:26:34.225501-07	INSERT	\N	\N	\N	f	f
496f2ef7-d8b4-4352-994f-8a639c48e4f3	generic_instance	json_addl	6ad2a1ab-3b91-4c5b-b4e6-69d087c868f0	FI7024	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142630", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142630", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:26:34.458407-07	UPDATE	\N	\N	\N	f	f
ee8f38f6-e145-4a49-a583-0bbdd92698f8	generic_instance	created_dt	6ad2a1ab-3b91-4c5b-b4e6-69d087c868f0	FI7024	2025-07-03 07:26:34.225501-07	2025-07-03T07:26:34.225501-07:00	john@daylilyinformatics.com	2025-07-03 07:26:34.458407-07	UPDATE	\N	\N	\N	f	f
dd038221-600f-4080-aa49-813b66a9631d	generic_instance	modified_dt	6ad2a1ab-3b91-4c5b-b4e6-69d087c868f0	FI7024	2025-07-03 07:26:34.225501-07	2025-07-03T07:26:34.458407-07:00	john@daylilyinformatics.com	2025-07-03 07:26:34.458407-07	UPDATE	\N	\N	\N	f	f
f15bf006-afb3-497b-8dd0-ea307b423931	generic_instance	json_addl	6ad2a1ab-3b91-4c5b-b4e6-69d087c868f0	FI7024	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142630", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7024.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7024.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142630", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:26:34.615521-07	UPDATE	\N	\N	\N	f	f
dfeb4dc0-8395-43fd-94bf-94730d27a363	generic_instance	created_dt	6ad2a1ab-3b91-4c5b-b4e6-69d087c868f0	FI7024	2025-07-03 07:26:34.225501-07	2025-07-03T07:26:34.225501-07:00	john@daylilyinformatics.com	2025-07-03 07:26:34.615521-07	UPDATE	\N	\N	\N	f	f
a8bcbd41-6093-4a74-9e1a-8578c586f9a9	generic_instance	modified_dt	6ad2a1ab-3b91-4c5b-b4e6-69d087c868f0	FI7024	2025-07-03 07:26:34.458407-07	2025-07-03T07:26:34.615521-07:00	john@daylilyinformatics.com	2025-07-03 07:26:34.615521-07	UPDATE	\N	\N	\N	f	f
a7bf1ed9-0d0c-4df9-b085-604e89285c5c	generic_instance_lineage	\N	4ef08fac-4436-4aa3-9afa-aba2b643d417	GL32	\N	\N	john@daylilyinformatics.com	2025-07-03 07:26:35.660855-07	INSERT	\N	\N	\N	f	f
5f50d381-1d57-4d3a-9ae4-cdde97242f08	generic_instance	\N	7b446c7d-69b4-45f5-9320-dff366d18cce	FI7025	\N	\N	john@daylilyinformatics.com	2025-07-03 07:26:35.675084-07	INSERT	\N	\N	\N	f	f
942cbf83-4534-4df7-9f37-8bfd447cb229	generic_instance	json_addl	7b446c7d-69b4-45f5-9320-dff366d18cce	FI7025	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142630", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142630", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:26:35.701109-07	UPDATE	\N	\N	\N	f	f
23161a34-b3bb-4abc-a8a1-50a16b705c2b	generic_instance	created_dt	7b446c7d-69b4-45f5-9320-dff366d18cce	FI7025	2025-07-03 07:26:35.675084-07	2025-07-03T07:26:35.675084-07:00	john@daylilyinformatics.com	2025-07-03 07:26:35.701109-07	UPDATE	\N	\N	\N	f	f
4030f433-9f38-43e6-8987-a539acaaa69e	generic_instance	modified_dt	7b446c7d-69b4-45f5-9320-dff366d18cce	FI7025	2025-07-03 07:26:35.675084-07	2025-07-03T07:26:35.701109-07:00	john@daylilyinformatics.com	2025-07-03 07:26:35.701109-07	UPDATE	\N	\N	\N	f	f
2c2a7671-c8c9-4c6d-8957-f7935a04e3f2	generic_instance	json_addl	7b446c7d-69b4-45f5-9320-dff366d18cce	FI7025	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142630", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7025.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7025.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142630", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:26:35.966492-07	UPDATE	\N	\N	\N	f	f
6daf3bb6-cb68-4db6-8bbf-ec5bdac9c8bd	generic_instance	created_dt	7b446c7d-69b4-45f5-9320-dff366d18cce	FI7025	2025-07-03 07:26:35.675084-07	2025-07-03T07:26:35.675084-07:00	john@daylilyinformatics.com	2025-07-03 07:26:35.966492-07	UPDATE	\N	\N	\N	f	f
44cd2613-122e-43c3-9657-e46b2d38b244	generic_instance	modified_dt	7b446c7d-69b4-45f5-9320-dff366d18cce	FI7025	2025-07-03 07:26:35.701109-07	2025-07-03T07:26:35.966492-07:00	john@daylilyinformatics.com	2025-07-03 07:26:35.966492-07	UPDATE	\N	\N	\N	f	f
b4fba297-ad51-4933-8e3d-1b153814e442	generic_instance_lineage	\N	e2c2ce43-cead-4515-b697-5be784cfcf97	GL33	\N	\N	john@daylilyinformatics.com	2025-07-03 07:26:36.904618-07	INSERT	\N	\N	\N	f	f
cda9853d-c6b1-4700-be05-665f30253153	generic_instance	\N	11568060-9f81-417e-baf9-9e7e56b3a759	FI7026	\N	\N	john@daylilyinformatics.com	2025-07-03 07:26:36.917375-07	INSERT	\N	\N	\N	f	f
dfb5a127-b9c1-43e7-a4d3-db44f6989670	generic_instance	json_addl	11568060-9f81-417e-baf9-9e7e56b3a759	FI7026	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142630", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142630", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:26:37.117862-07	UPDATE	\N	\N	\N	f	f
ebd043c7-1aa1-46e7-a766-f236ca432326	generic_instance	created_dt	11568060-9f81-417e-baf9-9e7e56b3a759	FI7026	2025-07-03 07:26:36.917375-07	2025-07-03T07:26:36.917375-07:00	john@daylilyinformatics.com	2025-07-03 07:26:37.117862-07	UPDATE	\N	\N	\N	f	f
a6049b57-954b-4081-8850-bdc25abf44b4	generic_instance	modified_dt	11568060-9f81-417e-baf9-9e7e56b3a759	FI7026	2025-07-03 07:26:36.917375-07	2025-07-03T07:26:37.117862-07:00	john@daylilyinformatics.com	2025-07-03 07:26:37.117862-07	UPDATE	\N	\N	\N	f	f
d17d05f4-9838-4284-ae7d-421bbc71b086	generic_instance	json_addl	11568060-9f81-417e-baf9-9e7e56b3a759	FI7026	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142630", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7026.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7026.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142630", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:26:37.251557-07	UPDATE	\N	\N	\N	f	f
b85a0350-515d-400a-a713-44eed5361b27	generic_instance	created_dt	11568060-9f81-417e-baf9-9e7e56b3a759	FI7026	2025-07-03 07:26:36.917375-07	2025-07-03T07:26:36.917375-07:00	john@daylilyinformatics.com	2025-07-03 07:26:37.251557-07	UPDATE	\N	\N	\N	f	f
31c68d50-fef1-4c93-a2e1-0125f109dad0	generic_instance	modified_dt	11568060-9f81-417e-baf9-9e7e56b3a759	FI7026	2025-07-03 07:26:37.117862-07	2025-07-03T07:26:37.251557-07:00	john@daylilyinformatics.com	2025-07-03 07:26:37.251557-07	UPDATE	\N	\N	\N	f	f
4fea2a38-f501-43a0-b27f-ce01e55ec3e3	generic_instance_lineage	\N	8e3d71b8-d06e-40bd-a992-15f32dc14778	GL34	\N	\N	john@daylilyinformatics.com	2025-07-03 07:26:38.31003-07	INSERT	\N	\N	\N	f	f
6d9bad3d-e47a-4e75-93d4-9fed02b95d4a	generic_instance	\N	1aab709b-8c59-4117-9bcc-48c2bf549a79	FS12	\N	\N	john@daylilyinformatics.com	2025-07-03 07:27:20.521229-07	INSERT	\N	\N	\N	f	f
241c047e-1b7c-4b90-bdc7-e5b63351d118	generic_instance	\N	0886f894-a724-4452-8007-54b3598c84d6	FI7027	\N	\N	john@daylilyinformatics.com	2025-07-03 07:27:20.555143-07	INSERT	\N	\N	\N	f	f
c38907c4-3511-45e3-8138-b66a34c4199a	generic_instance	json_addl	0886f894-a724-4452-8007-54b3598c84d6	FI7027	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142720", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142720", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:27:20.566489-07	UPDATE	\N	\N	\N	f	f
716e522f-5fa8-43e7-b768-936945131582	generic_instance	created_dt	0886f894-a724-4452-8007-54b3598c84d6	FI7027	2025-07-03 07:27:20.555143-07	2025-07-03T07:27:20.555143-07:00	john@daylilyinformatics.com	2025-07-03 07:27:20.566489-07	UPDATE	\N	\N	\N	f	f
77d3276e-b903-469e-ba57-b4ab4cd41133	generic_instance	modified_dt	0886f894-a724-4452-8007-54b3598c84d6	FI7027	2025-07-03 07:27:20.555143-07	2025-07-03T07:27:20.566489-07:00	john@daylilyinformatics.com	2025-07-03 07:27:20.566489-07	UPDATE	\N	\N	\N	f	f
65737ddf-8e39-4f84-8396-bf730f498597	generic_instance	json_addl	0886f894-a724-4452-8007-54b3598c84d6	FI7027	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142720", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7027.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7027.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142720", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:27:20.854339-07	UPDATE	\N	\N	\N	f	f
8bee2d35-c182-472d-bf9b-f8c60b9111f1	generic_instance	created_dt	0886f894-a724-4452-8007-54b3598c84d6	FI7027	2025-07-03 07:27:20.555143-07	2025-07-03T07:27:20.555143-07:00	john@daylilyinformatics.com	2025-07-03 07:27:20.854339-07	UPDATE	\N	\N	\N	f	f
4a7177db-7c17-40f4-a4d7-2721649d2fcf	generic_instance	modified_dt	0886f894-a724-4452-8007-54b3598c84d6	FI7027	2025-07-03 07:27:20.566489-07	2025-07-03T07:27:20.854339-07:00	john@daylilyinformatics.com	2025-07-03 07:27:20.854339-07	UPDATE	\N	\N	\N	f	f
4ea570ab-d53c-49e6-954e-ea18fb31d7ca	generic_instance_lineage	\N	64f3ecb8-9358-43fd-9550-e13eeafb7d63	GL35	\N	\N	john@daylilyinformatics.com	2025-07-03 07:27:21.772941-07	INSERT	\N	\N	\N	f	f
6f76ad87-238c-4de5-9c23-aa5a53e89c90	generic_instance	\N	051fd024-70cd-4874-ba13-088a4becc209	FI7028	\N	\N	john@daylilyinformatics.com	2025-07-03 07:27:21.787463-07	INSERT	\N	\N	\N	f	f
720799c5-48ae-4b70-8a11-7b016a2f619a	generic_instance	json_addl	051fd024-70cd-4874-ba13-088a4becc209	FI7028	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142720", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142720", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:27:21.978522-07	UPDATE	\N	\N	\N	f	f
4eb7984f-fa19-4122-bc6f-e0edf3f3bec9	generic_instance	created_dt	051fd024-70cd-4874-ba13-088a4becc209	FI7028	2025-07-03 07:27:21.787463-07	2025-07-03T07:27:21.787463-07:00	john@daylilyinformatics.com	2025-07-03 07:27:21.978522-07	UPDATE	\N	\N	\N	f	f
9dbdb466-6991-49b1-8bd6-5a18c28b5b1e	generic_instance	modified_dt	051fd024-70cd-4874-ba13-088a4becc209	FI7028	2025-07-03 07:27:21.787463-07	2025-07-03T07:27:21.978522-07:00	john@daylilyinformatics.com	2025-07-03 07:27:21.978522-07	UPDATE	\N	\N	\N	f	f
dd51ee08-4c3c-480f-978a-c5d82c40c6ce	generic_instance	json_addl	051fd024-70cd-4874-ba13-088a4becc209	FI7028	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142720", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7028.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7028.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142720", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:27:22.104097-07	UPDATE	\N	\N	\N	f	f
41d7d2f3-abd5-48bd-839f-9f2446438995	generic_instance	created_dt	051fd024-70cd-4874-ba13-088a4becc209	FI7028	2025-07-03 07:27:21.787463-07	2025-07-03T07:27:21.787463-07:00	john@daylilyinformatics.com	2025-07-03 07:27:22.104097-07	UPDATE	\N	\N	\N	f	f
d97f15de-092a-4392-81cc-f0ffef5752e0	generic_instance	modified_dt	051fd024-70cd-4874-ba13-088a4becc209	FI7028	2025-07-03 07:27:21.978522-07	2025-07-03T07:27:22.104097-07:00	john@daylilyinformatics.com	2025-07-03 07:27:22.104097-07	UPDATE	\N	\N	\N	f	f
6d3dfd8a-b40b-4532-b8e3-2f4d62b1ea55	generic_instance_lineage	\N	8bc65425-a974-4a61-9bf8-639aaa6d8fcd	GL36	\N	\N	john@daylilyinformatics.com	2025-07-03 07:27:23.158652-07	INSERT	\N	\N	\N	f	f
43270865-f356-4214-b733-d8dc684f8348	generic_instance	\N	5608248c-838c-458c-bbba-a6a0ba464775	FI7029	\N	\N	john@daylilyinformatics.com	2025-07-03 07:27:23.20872-07	INSERT	\N	\N	\N	f	f
6edb53b5-3112-4848-9c25-33c34b1c84a0	generic_instance	json_addl	5608248c-838c-458c-bbba-a6a0ba464775	FI7029	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142720", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142720", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:27:23.253827-07	UPDATE	\N	\N	\N	f	f
b73cf461-7a6a-4a17-82fb-7d9ebc453ba2	generic_instance	created_dt	5608248c-838c-458c-bbba-a6a0ba464775	FI7029	2025-07-03 07:27:23.20872-07	2025-07-03T07:27:23.20872-07:00	john@daylilyinformatics.com	2025-07-03 07:27:23.253827-07	UPDATE	\N	\N	\N	f	f
df22bfda-bbd3-4e50-b4fb-334f9624866a	generic_instance	modified_dt	5608248c-838c-458c-bbba-a6a0ba464775	FI7029	2025-07-03 07:27:23.20872-07	2025-07-03T07:27:23.253827-07:00	john@daylilyinformatics.com	2025-07-03 07:27:23.253827-07	UPDATE	\N	\N	\N	f	f
4228ed97-a1ca-45a9-88ca-7000a3b0512d	generic_instance	json_addl	5608248c-838c-458c-bbba-a6a0ba464775	FI7029	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142720", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7029.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7029.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142720", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:27:23.50468-07	UPDATE	\N	\N	\N	f	f
2bea12c9-1e75-47c0-b44f-e2c099d71726	generic_instance	created_dt	5608248c-838c-458c-bbba-a6a0ba464775	FI7029	2025-07-03 07:27:23.20872-07	2025-07-03T07:27:23.20872-07:00	john@daylilyinformatics.com	2025-07-03 07:27:23.50468-07	UPDATE	\N	\N	\N	f	f
745eb7ea-91b2-4726-a5db-eee8b2def5e1	generic_instance	modified_dt	5608248c-838c-458c-bbba-a6a0ba464775	FI7029	2025-07-03 07:27:23.253827-07	2025-07-03T07:27:23.50468-07:00	john@daylilyinformatics.com	2025-07-03 07:27:23.50468-07	UPDATE	\N	\N	\N	f	f
b4118ea9-eb77-436e-ac09-6d63a96ee698	generic_instance_lineage	\N	47dde6fa-2e0b-43a5-bc61-0e8b2ecb48b7	GL37	\N	\N	john@daylilyinformatics.com	2025-07-03 07:27:24.445577-07	INSERT	\N	\N	\N	f	f
d1be89ee-1382-45e2-9628-4f87029da558	generic_instance	\N	ec92091a-246e-472e-914a-66e628bd748b	FI7030	\N	\N	john@daylilyinformatics.com	2025-07-03 07:27:24.475065-07	INSERT	\N	\N	\N	f	f
21aafc73-7050-4d39-b1c8-9002558f4d9e	generic_instance	json_addl	ec92091a-246e-472e-914a-66e628bd748b	FI7030	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142720", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142720", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:27:24.670993-07	UPDATE	\N	\N	\N	f	f
68e3e971-7b68-4369-a75a-a66edc3e9274	generic_instance	created_dt	ec92091a-246e-472e-914a-66e628bd748b	FI7030	2025-07-03 07:27:24.475065-07	2025-07-03T07:27:24.475065-07:00	john@daylilyinformatics.com	2025-07-03 07:27:24.670993-07	UPDATE	\N	\N	\N	f	f
c11b41df-8a72-4276-bccb-d48cc7742308	generic_instance	modified_dt	ec92091a-246e-472e-914a-66e628bd748b	FI7030	2025-07-03 07:27:24.475065-07	2025-07-03T07:27:24.670993-07:00	john@daylilyinformatics.com	2025-07-03 07:27:24.670993-07	UPDATE	\N	\N	\N	f	f
2b58bed7-5ed9-4440-a71d-9e98ec70c6c5	generic_instance	json_addl	ec92091a-246e-472e-914a-66e628bd748b	FI7030	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142720", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7030.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7030.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142720", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:27:24.798792-07	UPDATE	\N	\N	\N	f	f
71ffcf96-577f-4d95-91a4-343ca5c173cd	generic_instance	created_dt	ec92091a-246e-472e-914a-66e628bd748b	FI7030	2025-07-03 07:27:24.475065-07	2025-07-03T07:27:24.475065-07:00	john@daylilyinformatics.com	2025-07-03 07:27:24.798792-07	UPDATE	\N	\N	\N	f	f
1585b74f-419b-495c-a0a1-512419dd6e57	generic_instance	modified_dt	ec92091a-246e-472e-914a-66e628bd748b	FI7030	2025-07-03 07:27:24.670993-07	2025-07-03T07:27:24.798792-07:00	john@daylilyinformatics.com	2025-07-03 07:27:24.798792-07	UPDATE	\N	\N	\N	f	f
e457ffb9-0736-4971-96e4-d05ae2df5506	generic_instance_lineage	\N	c33c0266-22e6-4a82-a537-78a94d4d274f	GL38	\N	\N	john@daylilyinformatics.com	2025-07-03 07:27:25.851199-07	INSERT	\N	\N	\N	f	f
34f70073-564a-46dc-8c7a-753dc797f204	generic_instance	\N	c47bda8d-6c52-4cf8-83d6-213aaa276e3f	FS13	\N	\N	john@daylilyinformatics.com	2025-07-03 07:30:38.949974-07	INSERT	\N	\N	\N	f	f
aa1ac668-b629-454a-acb1-1af592076f17	generic_instance	\N	96c5276b-4501-460e-80a1-c6506a98e020	FI7031	\N	\N	john@daylilyinformatics.com	2025-07-03 07:30:39.072113-07	INSERT	\N	\N	\N	f	f
57fcc79d-5b79-4354-8285-269a672a71ba	generic_instance	json_addl	96c5276b-4501-460e-80a1-c6506a98e020	FI7031	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T143038", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T143038", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:30:39.077004-07	UPDATE	\N	\N	\N	f	f
1777d85b-69bc-48dd-b0c1-ff648602fa8b	generic_instance	created_dt	96c5276b-4501-460e-80a1-c6506a98e020	FI7031	2025-07-03 07:30:39.072113-07	2025-07-03T07:30:39.072113-07:00	john@daylilyinformatics.com	2025-07-03 07:30:39.077004-07	UPDATE	\N	\N	\N	f	f
91f311e0-a9b1-4676-8e3c-e16a01f46fcf	generic_instance	modified_dt	96c5276b-4501-460e-80a1-c6506a98e020	FI7031	2025-07-03 07:30:39.072113-07	2025-07-03T07:30:39.077004-07:00	john@daylilyinformatics.com	2025-07-03 07:30:39.077004-07	UPDATE	\N	\N	\N	f	f
31746e65-2986-4de1-bf60-51b73d7f726f	generic_instance_lineage	\N	82fa7bdb-2695-41d2-b8e5-46378f282371	GL49	\N	\N	john@daylilyinformatics.com	2025-07-03 07:34:39.315097-07	INSERT	\N	\N	\N	f	f
424d1535-5c01-4e29-a16c-608c67a72a5f	generic_instance	json_addl	96c5276b-4501-460e-80a1-c6506a98e020	FI7031	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T143038", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7031.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7031.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T143038", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:30:39.333909-07	UPDATE	\N	\N	\N	f	f
f7b81828-1f13-4f96-85dc-089a0a26389a	generic_instance	created_dt	96c5276b-4501-460e-80a1-c6506a98e020	FI7031	2025-07-03 07:30:39.072113-07	2025-07-03T07:30:39.072113-07:00	john@daylilyinformatics.com	2025-07-03 07:30:39.333909-07	UPDATE	\N	\N	\N	f	f
59154f4b-1960-48fb-a360-f4a215bfe591	generic_instance	modified_dt	96c5276b-4501-460e-80a1-c6506a98e020	FI7031	2025-07-03 07:30:39.077004-07	2025-07-03T07:30:39.333909-07:00	john@daylilyinformatics.com	2025-07-03 07:30:39.333909-07	UPDATE	\N	\N	\N	f	f
57fc4d3c-181d-48be-9014-f4a5c75cf36c	generic_instance_lineage	\N	bda4f9c3-1d9d-4481-b0dc-f3e61b5377aa	GL39	\N	\N	john@daylilyinformatics.com	2025-07-03 07:30:40.065119-07	INSERT	\N	\N	\N	f	f
4664a8fd-e10e-459f-ba2b-a38a912e6ab5	generic_instance	\N	c64b5702-5aea-4bf0-856b-ad9d6a0035c4	FI7032	\N	\N	john@daylilyinformatics.com	2025-07-03 07:30:40.072812-07	INSERT	\N	\N	\N	f	f
058a8128-d7cb-4128-a38d-5487d14987f8	generic_instance	json_addl	c64b5702-5aea-4bf0-856b-ad9d6a0035c4	FI7032	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T143038", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T143038", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:30:40.247955-07	UPDATE	\N	\N	\N	f	f
5209c928-456e-47e5-9bc8-f0eef983983f	generic_instance	created_dt	c64b5702-5aea-4bf0-856b-ad9d6a0035c4	FI7032	2025-07-03 07:30:40.072812-07	2025-07-03T07:30:40.072812-07:00	john@daylilyinformatics.com	2025-07-03 07:30:40.247955-07	UPDATE	\N	\N	\N	f	f
33e0a1ad-bb77-46fd-b31d-8f111faf7b78	generic_instance	modified_dt	c64b5702-5aea-4bf0-856b-ad9d6a0035c4	FI7032	2025-07-03 07:30:40.072812-07	2025-07-03T07:30:40.247955-07:00	john@daylilyinformatics.com	2025-07-03 07:30:40.247955-07	UPDATE	\N	\N	\N	f	f
cd585a3e-e367-49fa-97de-ecbb992642b4	generic_instance	json_addl	c64b5702-5aea-4bf0-856b-ad9d6a0035c4	FI7032	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T143038", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7032.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7032.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T143038", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:30:40.362347-07	UPDATE	\N	\N	\N	f	f
d96afac4-ea4e-49ae-b220-c742a609809f	generic_instance	created_dt	c64b5702-5aea-4bf0-856b-ad9d6a0035c4	FI7032	2025-07-03 07:30:40.072812-07	2025-07-03T07:30:40.072812-07:00	john@daylilyinformatics.com	2025-07-03 07:30:40.362347-07	UPDATE	\N	\N	\N	f	f
d3456219-2831-401e-9e5e-928fbf7d2588	generic_instance	modified_dt	c64b5702-5aea-4bf0-856b-ad9d6a0035c4	FI7032	2025-07-03 07:30:40.247955-07	2025-07-03T07:30:40.362347-07:00	john@daylilyinformatics.com	2025-07-03 07:30:40.362347-07	UPDATE	\N	\N	\N	f	f
0953c4e0-50f5-4f83-9633-185e3a420ca9	generic_instance_lineage	\N	b82352e9-606f-4a0e-bc02-6b29dd67745d	GL40	\N	\N	john@daylilyinformatics.com	2025-07-03 07:30:41.005943-07	INSERT	\N	\N	\N	f	f
b71f0533-b79b-4bbf-898c-83da26cea804	generic_instance	\N	44188f89-cb29-4da6-90c3-592a3e96ce82	FI7033	\N	\N	john@daylilyinformatics.com	2025-07-03 07:30:41.012048-07	INSERT	\N	\N	\N	f	f
06a7b88e-300e-4693-9ad8-82e8158624ee	generic_instance	json_addl	44188f89-cb29-4da6-90c3-592a3e96ce82	FI7033	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T143038", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T143038", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:30:41.022514-07	UPDATE	\N	\N	\N	f	f
4555be28-d2f9-4549-ab5d-c946815ce1e1	generic_instance	created_dt	44188f89-cb29-4da6-90c3-592a3e96ce82	FI7033	2025-07-03 07:30:41.012048-07	2025-07-03T07:30:41.012048-07:00	john@daylilyinformatics.com	2025-07-03 07:30:41.022514-07	UPDATE	\N	\N	\N	f	f
cd719b15-9e5f-4a42-accb-a105af18c63b	generic_instance	modified_dt	44188f89-cb29-4da6-90c3-592a3e96ce82	FI7033	2025-07-03 07:30:41.012048-07	2025-07-03T07:30:41.022514-07:00	john@daylilyinformatics.com	2025-07-03 07:30:41.022514-07	UPDATE	\N	\N	\N	f	f
68fd26c6-a3e0-454f-ba80-c526c9ec0caa	generic_instance	json_addl	44188f89-cb29-4da6-90c3-592a3e96ce82	FI7033	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T143038", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7033.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7033.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T143038", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:30:41.242377-07	UPDATE	\N	\N	\N	f	f
bb31d5c7-251b-431e-86a0-182f6ba60367	generic_instance	created_dt	44188f89-cb29-4da6-90c3-592a3e96ce82	FI7033	2025-07-03 07:30:41.012048-07	2025-07-03T07:30:41.012048-07:00	john@daylilyinformatics.com	2025-07-03 07:30:41.242377-07	UPDATE	\N	\N	\N	f	f
b9025220-ba01-49b7-b9a5-328fdf8afe64	generic_instance	modified_dt	44188f89-cb29-4da6-90c3-592a3e96ce82	FI7033	2025-07-03 07:30:41.022514-07	2025-07-03T07:30:41.242377-07:00	john@daylilyinformatics.com	2025-07-03 07:30:41.242377-07	UPDATE	\N	\N	\N	f	f
1ef49269-ae39-4572-b428-1f3b1e77117a	generic_instance_lineage	\N	105321bc-e952-45e3-b01d-490290c7ee0c	GL41	\N	\N	john@daylilyinformatics.com	2025-07-03 07:30:41.948242-07	INSERT	\N	\N	\N	f	f
b4a9e0de-d15d-4603-be4a-b37499a68a42	generic_instance	\N	fd60bad1-ce79-4e33-a561-8db568b1514c	FI7034	\N	\N	john@daylilyinformatics.com	2025-07-03 07:30:41.961138-07	INSERT	\N	\N	\N	f	f
771449f2-c45c-4001-b54a-bb66e9534ee3	generic_instance	json_addl	fd60bad1-ce79-4e33-a561-8db568b1514c	FI7034	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T143038", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T143038", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:30:42.141749-07	UPDATE	\N	\N	\N	f	f
22bc48ca-da61-418a-a9d1-96b8b6a341fd	generic_instance	created_dt	fd60bad1-ce79-4e33-a561-8db568b1514c	FI7034	2025-07-03 07:30:41.961138-07	2025-07-03T07:30:41.961138-07:00	john@daylilyinformatics.com	2025-07-03 07:30:42.141749-07	UPDATE	\N	\N	\N	f	f
790e9fa0-4371-4215-9cb6-7807a1502b1b	generic_instance	modified_dt	fd60bad1-ce79-4e33-a561-8db568b1514c	FI7034	2025-07-03 07:30:41.961138-07	2025-07-03T07:30:42.141749-07:00	john@daylilyinformatics.com	2025-07-03 07:30:42.141749-07	UPDATE	\N	\N	\N	f	f
fe1d7056-a653-4903-a2a4-62f0e520db97	generic_instance	json_addl	fd60bad1-ce79-4e33-a561-8db568b1514c	FI7034	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T143038", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7034.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7034.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T143038", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:30:42.258511-07	UPDATE	\N	\N	\N	f	f
f4e3ef0e-d0b0-4d71-b156-868ff4151592	generic_instance	created_dt	fd60bad1-ce79-4e33-a561-8db568b1514c	FI7034	2025-07-03 07:30:41.961138-07	2025-07-03T07:30:41.961138-07:00	john@daylilyinformatics.com	2025-07-03 07:30:42.258511-07	UPDATE	\N	\N	\N	f	f
cd12c329-2a4c-4af2-a450-8a7362d205a2	generic_instance	modified_dt	fd60bad1-ce79-4e33-a561-8db568b1514c	FI7034	2025-07-03 07:30:42.141749-07	2025-07-03T07:30:42.258511-07:00	john@daylilyinformatics.com	2025-07-03 07:30:42.258511-07	UPDATE	\N	\N	\N	f	f
2e5646e1-d00b-4d77-9c97-8b36b4d36924	generic_instance_lineage	\N	2c4895d4-8647-488d-82e2-2714097867d3	GL42	\N	\N	john@daylilyinformatics.com	2025-07-03 07:30:42.982424-07	INSERT	\N	\N	\N	f	f
053615bd-6a54-4406-a03e-bbd7073d18dd	generic_instance	\N	99921700-5afa-47c0-aa3b-f6a520dc526b	FS14	\N	\N	john@daylilyinformatics.com	2025-07-03 07:32:27.331864-07	INSERT	\N	\N	\N	f	f
d004196c-0a92-4649-b8eb-b9a42c55c2bb	generic_instance	\N	ef3752a0-fb6c-4294-ac4e-e0f65976b68a	FI7035	\N	\N	john@daylilyinformatics.com	2025-07-03 07:32:27.440193-07	INSERT	\N	\N	\N	f	f
ebb0fd85-342b-4e60-ad18-88679ec30f70	generic_instance	json_addl	ef3752a0-fb6c-4294-ac4e-e0f65976b68a	FI7035	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143216", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143216", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:32:27.457722-07	UPDATE	\N	\N	\N	f	f
170084e9-df31-4d11-8ada-734583c9a99b	generic_instance	created_dt	ef3752a0-fb6c-4294-ac4e-e0f65976b68a	FI7035	2025-07-03 07:32:27.440193-07	2025-07-03T07:32:27.440193-07:00	john@daylilyinformatics.com	2025-07-03 07:32:27.457722-07	UPDATE	\N	\N	\N	f	f
b104ea1c-2338-4f5d-a297-b8532d8b9b87	generic_instance	modified_dt	ef3752a0-fb6c-4294-ac4e-e0f65976b68a	FI7035	2025-07-03 07:32:27.440193-07	2025-07-03T07:32:27.457722-07:00	john@daylilyinformatics.com	2025-07-03 07:32:27.457722-07	UPDATE	\N	\N	\N	f	f
0d60c44e-4045-466a-bfd7-5b4a220d9860	generic_instance	json_addl	ef3752a0-fb6c-4294-ac4e-e0f65976b68a	FI7035	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143216", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "backup_2025-07-03_00-06-33.sql", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "sql", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7035.sql", "current_s3_uri": "s3://daylily-dewey-0/0/FI7035.sql", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143216", "original_file_md5": "", "original_file_name": "backup_2025-07-03_00-06-33.sql", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "sql", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1012876, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:32:27.785525-07	UPDATE	\N	\N	\N	f	f
4ca84b12-2d8c-42fe-a11a-4a9a1e3527e1	generic_instance	created_dt	ef3752a0-fb6c-4294-ac4e-e0f65976b68a	FI7035	2025-07-03 07:32:27.440193-07	2025-07-03T07:32:27.440193-07:00	john@daylilyinformatics.com	2025-07-03 07:32:27.785525-07	UPDATE	\N	\N	\N	f	f
89ef08f9-8b88-48fa-a646-33f285e46ac7	generic_instance	modified_dt	ef3752a0-fb6c-4294-ac4e-e0f65976b68a	FI7035	2025-07-03 07:32:27.457722-07	2025-07-03T07:32:27.785525-07:00	john@daylilyinformatics.com	2025-07-03 07:32:27.785525-07	UPDATE	\N	\N	\N	f	f
7422e4ed-bf9e-4d63-93fd-d70bf8d52838	generic_instance_lineage	\N	735fad8d-eac4-4bdd-8552-34ed0007dd55	GL43	\N	\N	john@daylilyinformatics.com	2025-07-03 07:32:28.834993-07	INSERT	\N	\N	\N	f	f
23bdbccf-1a78-4dd9-b462-cf69308b9c56	generic_instance	\N	def0e204-1561-4031-a477-de7549381c84	FS15	\N	\N	john@daylilyinformatics.com	2025-07-03 07:33:39.079499-07	INSERT	\N	\N	\N	f	f
f4d03a19-4f24-4864-ab9b-7298315222f4	generic_instance	\N	8c44a660-381a-4fd7-8bb3-6093c23991f0	FI7036	\N	\N	john@daylilyinformatics.com	2025-07-03 07:33:39.200006-07	INSERT	\N	\N	\N	f	f
b368622a-f687-46f6-8f75-95d4dbd299e4	generic_instance	json_addl	8c44a660-381a-4fd7-8bb3-6093c23991f0	FI7036	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "Clinical", "category": "Diagnosis", "comments": "a comment", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "", "file_tags": "sometag", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "fastqs", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "Clinical", "category": "Diagnosis", "comments": "a comment", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "", "file_tags": "sometag", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "fastqs", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:33:39.222496-07	UPDATE	\N	\N	\N	f	f
051b4b97-29af-4cf1-9fb8-51c5dc5c7ac8	generic_instance	created_dt	8c44a660-381a-4fd7-8bb3-6093c23991f0	FI7036	2025-07-03 07:33:39.200006-07	2025-07-03T07:33:39.200006-07:00	john@daylilyinformatics.com	2025-07-03 07:33:39.222496-07	UPDATE	\N	\N	\N	f	f
1203fc8c-807f-4065-a442-dc12f2c00499	generic_instance_lineage	\N	b6ee4da9-b42c-435f-95d7-0ef084d6888c	GL44	\N	\N	john@daylilyinformatics.com	2025-07-03 07:33:44.754931-07	INSERT	\N	\N	\N	f	f
b8c48495-70a5-49bc-bfc1-80fcc92f52cf	generic_instance	modified_dt	8c44a660-381a-4fd7-8bb3-6093c23991f0	FI7036	2025-07-03 07:33:39.200006-07	2025-07-03T07:33:39.222496-07:00	john@daylilyinformatics.com	2025-07-03 07:33:39.222496-07	UPDATE	\N	\N	\N	f	f
a8d3a0af-934a-4311-9091-7cf4733e7283	generic_instance	json_addl	8c44a660-381a-4fd7-8bb3-6093c23991f0	FI7036	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "Clinical", "category": "Diagnosis", "comments": "a comment", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "", "file_tags": "sometag", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "fastqs", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "NA12878/HG001_GRCh38_1_22_v4.2.1_benchmark_hifiasm_v11_phasetransfer.PS_FIX.vcf.gz", "purpose": "Clinical", "category": "Diagnosis", "comments": "a comment", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "", "file_tags": "sometag", "file_type": "gz", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7036.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7036.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "fastqs", "original_file_md5": "", "original_file_name": "HG001_GRCh38_1_22_v4.2.1_benchmark_hifiasm_v11_phasetransfer.PS_FIX.vcf.gz", "original_file_path": "NA12878", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 111830815, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:33:39.500051-07	UPDATE	\N	\N	\N	f	f
91a2b57f-3354-4cc1-a0fe-850d39b1a640	generic_instance	created_dt	8c44a660-381a-4fd7-8bb3-6093c23991f0	FI7036	2025-07-03 07:33:39.200006-07	2025-07-03T07:33:39.200006-07:00	john@daylilyinformatics.com	2025-07-03 07:33:39.500051-07	UPDATE	\N	\N	\N	f	f
95a39f19-9b04-4519-bf6f-567500f00d2c	generic_instance	modified_dt	8c44a660-381a-4fd7-8bb3-6093c23991f0	FI7036	2025-07-03 07:33:39.222496-07	2025-07-03T07:33:39.500051-07:00	john@daylilyinformatics.com	2025-07-03 07:33:39.500051-07	UPDATE	\N	\N	\N	f	f
aeef1aff-7a54-44dd-b04d-308b00192e84	generic_instance	\N	76d325cb-204a-490a-817f-ba7917e86e20	FI7037	\N	\N	john@daylilyinformatics.com	2025-07-03 07:33:44.749755-07	INSERT	\N	\N	\N	f	f
7c482dc1-40f0-4113-99ce-122fbf01cb08	generic_instance	json_addl	76d325cb-204a-490a-817f-ba7917e86e20	FI7037	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "Clinical", "category": "Diagnosis", "comments": "a comment", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "", "file_tags": "sometag", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "fastqs", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "Clinical", "category": "Diagnosis", "comments": "a comment", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "", "file_tags": "sometag", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "fastqs", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:33:44.791701-07	UPDATE	\N	\N	\N	f	f
d8b69407-afb2-42fd-b45f-b11070d75039	generic_instance	created_dt	76d325cb-204a-490a-817f-ba7917e86e20	FI7037	2025-07-03 07:33:44.749755-07	2025-07-03T07:33:44.749755-07:00	john@daylilyinformatics.com	2025-07-03 07:33:44.791701-07	UPDATE	\N	\N	\N	f	f
2ff04998-35a1-4fcd-b4e8-795b90b4e0dd	generic_instance	modified_dt	76d325cb-204a-490a-817f-ba7917e86e20	FI7037	2025-07-03 07:33:44.749755-07	2025-07-03T07:33:44.791701-07:00	john@daylilyinformatics.com	2025-07-03 07:33:44.791701-07	UPDATE	\N	\N	\N	f	f
e905243f-f1e0-4fc0-b9f7-a3a209586f43	generic_instance	json_addl	76d325cb-204a-490a-817f-ba7917e86e20	FI7037	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "Clinical", "category": "Diagnosis", "comments": "a comment", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "", "file_tags": "sometag", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "fastqs", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "NA12878/Lee_washington_0250E_19948.pdf", "purpose": "Clinical", "category": "Diagnosis", "comments": "a comment", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "", "file_tags": "sometag", "file_type": "pdf", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7037.pdf", "current_s3_uri": "s3://daylily-dewey-0/0/FI7037.pdf", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "fastqs", "original_file_md5": "", "original_file_name": "Lee_washington_0250E_19948.pdf", "original_file_path": "NA12878", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "pdf", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 11203779, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:33:44.943376-07	UPDATE	\N	\N	\N	f	f
3d4388c7-f4c7-4205-9c82-98465e29a40b	generic_instance	created_dt	76d325cb-204a-490a-817f-ba7917e86e20	FI7037	2025-07-03 07:33:44.749755-07	2025-07-03T07:33:44.749755-07:00	john@daylilyinformatics.com	2025-07-03 07:33:44.943376-07	UPDATE	\N	\N	\N	f	f
91eed1c9-72e9-498b-99f5-cd4044f146ff	generic_instance	modified_dt	76d325cb-204a-490a-817f-ba7917e86e20	FI7037	2025-07-03 07:33:44.791701-07	2025-07-03T07:33:44.943376-07:00	john@daylilyinformatics.com	2025-07-03 07:33:44.943376-07	UPDATE	\N	\N	\N	f	f
7eee4027-a0e6-4ebb-aaff-2edeeec5f211	generic_instance	\N	9ea2e910-4f31-4ed2-b6c4-449c753c3692	FS16	\N	\N	john@daylilyinformatics.com	2025-07-03 07:34:23.540595-07	INSERT	\N	\N	\N	f	f
9e8afb64-b5f3-4759-bc41-651cdc8deb73	generic_instance	\N	2158f0a3-f8e1-4bc9-84f7-4620f6fc8f28	FI7038	\N	\N	john@daylilyinformatics.com	2025-07-03 07:34:23.593818-07	INSERT	\N	\N	\N	f	f
8b49efe6-26a6-48ca-9557-3e252ddb5157	generic_instance	json_addl	2158f0a3-f8e1-4bc9-84f7-4620f6fc8f28	FI7038	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Remote", "upload_group_key": "20250703T143413", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Remote", "upload_group_key": "20250703T143413", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:34:23.602741-07	UPDATE	\N	\N	\N	f	f
5f038ddb-6bfe-4387-ad8e-b2cd6cefbffa	generic_instance	created_dt	2158f0a3-f8e1-4bc9-84f7-4620f6fc8f28	FI7038	2025-07-03 07:34:23.593818-07	2025-07-03T07:34:23.593818-07:00	john@daylilyinformatics.com	2025-07-03 07:34:23.602741-07	UPDATE	\N	\N	\N	f	f
cc87503b-276f-447e-bdda-e24c18fc7cff	generic_instance	modified_dt	2158f0a3-f8e1-4bc9-84f7-4620f6fc8f28	FI7038	2025-07-03 07:34:23.593818-07	2025-07-03T07:34:23.602741-07:00	john@daylilyinformatics.com	2025-07-03 07:34:23.602741-07	UPDATE	\N	\N	\N	f	f
8e653d59-96ef-49f0-b4dd-07fb726300c3	generic_instance	json_addl	2158f0a3-f8e1-4bc9-84f7-4620f6fc8f28	FI7038	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Remote", "upload_group_key": "20250703T143413", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "day_cmd.log", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "log", "patient_id": "", "remote_url": "https://daylily-service-omics-analysis-us-west-2.s3.us-west-2.amazonaws.com/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log?response-content-disposition=inline&X-Amz-Content-Sha256=UNSIGNED-PAYLOAD&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEA8aCXVzLXdlc3QtMiJHMEUCIEwkFC7typ%2B7k712by339aiZJx867GzZv1rwLjaMWjy9AiEAyFFAGSuNMLiCibMYT9IuQiYJjXD59ZexbYTehsL2hXoquQMIGBAAGgw2NzA0ODQwNTA3MzgiDOLLh4KTH%2Bja%2FjEo0CqWA32BzYyhHkRrNyY1g6iBHMD91fBPNWdcSCjPQ3hFJ9USdNJZJG9R7%2BuWbMT0EQgqOu%2FUTCKIX7JLD8tJJHb1bx90R7BfXWfXKbrg94tOrItTV11IAv5uFUeVa24iDH%2BLCoStSBxskYQJl7MosBPuA0BXnrGsSTi545yLxxZnoUfHNRqltwIQRNut3XxdoDTpLlkirFKMGkIeyY8wJLQZfGoXj9bcau%2BGPw3Ikb20%2FdA79erq9GcJaANTefB3JHU%2FK2c%2B8zZLVnSB1V3%2BtR%2FTWW6thVzY8IFeXGMd9jvKnt9bX4CciVWPyQeBvKVniiLYqwFAdBGbLYmTr72bCfP%2FxA6JAsHooKLkX6bNgGNYbbT38%2BNGS4PDV5tZ7Ud8H13vKDvwn7Bux6JRqbf9erjLK7UJ9KmsLFPGdCvGMWO0c3STI3sw1vJgRQ6DAwwM6kKq4ui3k3yBHMGfF%2FW8MSdLB93B4W3qeKTriP5f8UX8Kw2LmNzLQnat7mPAmCunY3NHOpwaKMiES5XVhXcHGlKFYTz9V6spE3cwhLqYwwY63gJO%2FcyKZeiz3yHTivM0FJNm2TMuIhLmUm9KHl%2BEDGDP9x4XfVnbWza1M%2Fhjhkp7OQVZK0WWMsRmgzSKN%2F%2B0AYGf%2ByswpK6fUKJbeFeDDIxOTh3d8TWJTXj3wDTKuQddBNqLVw6XUuOmVwFBB5%2FLpVvywLDb3iVcNqOmlFJcR2jBR53n0EpZbfIW9j28xVlI965f%2BhcXMhK0D6yvNody3PegC9lDNi7bRERsj72QGCC7Na3HNliMURha2TZoo24SHLa1YLcsyu5ojw%2FIN%2BggWTTMg%2FE%2BWsP9eeEIxNLDj5G7nXkCEdjbfuKPoWUdebP15z9Dknyl7W1sDN2hai5B%2Bk%2B6hTaZziZxr%2FI%2FU2EHgwWkj2ya5qArGlpfK%2BAhlH5%2BVA9I3P3pv7ckjp2J6f2TUErgI3maueoSiTlbQqpEeSVNGu%2BDODQx5799eMGf0f448b29I%2BXKUrDlH0Yoj1RmEA%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAZYG7WWMZCCDE37TC%2F20250703%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20250703T143406Z&X-Amz-Expires=21600&X-Amz-SignedHeaders=host&X-Amz-Signature=3d7e773fe4a25d0a6dd36aecd9176b13b1717033a69610cfb6815ed156a1b93c", "clinician_id": "", "external_ids": [], "original_url": "https://daylily-service-omics-analysis-us-west-2.s3.us-west-2.amazonaws.com/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log?response-content-disposition=inline&X-Amz-Content-Sha256=UNSIGNED-PAYLOAD&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEA8aCXVzLXdlc3QtMiJHMEUCIEwkFC7typ%2B7k712by339aiZJx867GzZv1rwLjaMWjy9AiEAyFFAGSuNMLiCibMYT9IuQiYJjXD59ZexbYTehsL2hXoquQMIGBAAGgw2NzA0ODQwNTA3MzgiDOLLh4KTH%2Bja%2FjEo0CqWA32BzYyhHkRrNyY1g6iBHMD91fBPNWdcSCjPQ3hFJ9USdNJZJG9R7%2BuWbMT0EQgqOu%2FUTCKIX7JLD8tJJHb1bx90R7BfXWfXKbrg94tOrItTV11IAv5uFUeVa24iDH%2BLCoStSBxskYQJl7MosBPuA0BXnrGsSTi545yLxxZnoUfHNRqltwIQRNut3XxdoDTpLlkirFKMGkIeyY8wJLQZfGoXj9bcau%2BGPw3Ikb20%2FdA79erq9GcJaANTefB3JHU%2FK2c%2B8zZLVnSB1V3%2BtR%2FTWW6thVzY8IFeXGMd9jvKnt9bX4CciVWPyQeBvKVniiLYqwFAdBGbLYmTr72bCfP%2FxA6JAsHooKLkX6bNgGNYbbT38%2BNGS4PDV5tZ7Ud8H13vKDvwn7Bux6JRqbf9erjLK7UJ9KmsLFPGdCvGMWO0c3STI3sw1vJgRQ6DAwwM6kKq4ui3k3yBHMGfF%2FW8MSdLB93B4W3qeKTriP5f8UX8Kw2LmNzLQnat7mPAmCunY3NHOpwaKMiES5XVhXcHGlKFYTz9V6spE3cwhLqYwwY63gJO%2FcyKZeiz3yHTivM0FJNm2TMuIhLmUm9KHl%2BEDGDP9x4XfVnbWza1M%2Fhjhkp7OQVZK0WWMsRmgzSKN%2F%2B0AYGf%2ByswpK6fUKJbeFeDDIxOTh3d8TWJTXj3wDTKuQddBNqLVw6XUuOmVwFBB5%2FLpVvywLDb3iVcNqOmlFJcR2jBR53n0EpZbfIW9j28xVlI965f%2BhcXMhK0D6yvNody3PegC9lDNi7bRERsj72QGCC7Na3HNliMURha2TZoo24SHLa1YLcsyu5ojw%2FIN%2BggWTTMg%2FE%2BWsP9eeEIxNLDj5G7nXkCEdjbfuKPoWUdebP15z9Dknyl7W1sDN2hai5B%2Bk%2B6hTaZziZxr%2FI%2FU2EHgwWkj2ya5qArGlpfK%2BAhlH5%2BVA9I3P3pv7ckjp2J6f2TUErgI3maueoSiTlbQqpEeSVNGu%2BDODQx5799eMGf0f448b29I%2BXKUrDlH0Yoj1RmEA%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAZYG7WWMZCCDE37TC%2F20250703%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20250703T143406Z&X-Amz-Expires=21600&X-Amz-SignedHeaders=host&X-Amz-Signature=3d7e773fe4a25d0a6dd36aecd9176b13b1717033a69610cfb6815ed156a1b93c", "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "https://daylily-service-omics-analysis-us-west-2.s3.us-west-2.amazonaws.com/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log?response-content-disposition=inline&X-Amz-Content-Sha256=UNSIGNED-PAYLOAD&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEA8aCXVzLXdlc3QtMiJHMEUCIEwkFC7typ%2B7k712by339aiZJx867GzZv1rwLjaMWjy9AiEAyFFAGSuNMLiCibMYT9IuQiYJjXD59ZexbYTehsL2hXoquQMIGBAAGgw2NzA0ODQwNTA3MzgiDOLLh4KTH%2Bja%2FjEo0CqWA32BzYyhHkRrNyY1g6iBHMD91fBPNWdcSCjPQ3hFJ9USdNJZJG9R7%2BuWbMT0EQgqOu%2FUTCKIX7JLD8tJJHb1bx90R7BfXWfXKbrg94tOrItTV11IAv5uFUeVa24iDH%2BLCoStSBxskYQJl7MosBPuA0BXnrGsSTi545yLxxZnoUfHNRqltwIQRNut3XxdoDTpLlkirFKMGkIeyY8wJLQZfGoXj9bcau%2BGPw3Ikb20%2FdA79erq9GcJaANTefB3JHU%2FK2c%2B8zZLVnSB1V3%2BtR%2FTWW6thVzY8IFeXGMd9jvKnt9bX4CciVWPyQeBvKVniiLYqwFAdBGbLYmTr72bCfP%2FxA6JAsHooKLkX6bNgGNYbbT38%2BNGS4PDV5tZ7Ud8H13vKDvwn7Bux6JRqbf9erjLK7UJ9KmsLFPGdCvGMWO0c3STI3sw1vJgRQ6DAwwM6kKq4ui3k3yBHMGfF%2FW8MSdLB93B4W3qeKTriP5f8UX8Kw2LmNzLQnat7mPAmCunY3NHOpwaKMiES5XVhXcHGlKFYTz9V6spE3cwhLqYwwY63gJO%2FcyKZeiz3yHTivM0FJNm2TMuIhLmUm9KHl%2BEDGDP9x4XfVnbWza1M%2Fhjhkp7OQVZK0WWMsRmgzSKN%2F%2B0AYGf%2ByswpK6fUKJbeFeDDIxOTh3d8TWJTXj3wDTKuQddBNqLVw6XUuOmVwFBB5%2FLpVvywLDb3iVcNqOmlFJcR2jBR53n0EpZbfIW9j28xVlI965f%2BhcXMhK0D6yvNody3PegC9lDNi7bRERsj72QGCC7Na3HNliMURha2TZoo24SHLa1YLcsyu5ojw%2FIN%2BggWTTMg%2FE%2BWsP9eeEIxNLDj5G7nXkCEdjbfuKPoWUdebP15z9Dknyl7W1sDN2hai5B%2Bk%2B6hTaZziZxr%2FI%2FU2EHgwWkj2ya5qArGlpfK%2BAhlH5%2BVA9I3P3pv7ckjp2J6f2TUErgI3maueoSiTlbQqpEeSVNGu%2BDODQx5799eMGf0f448b29I%2BXKUrDlH0Yoj1RmEA%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAZYG7WWMZCCDE37TC%2F20250703%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20250703T143406Z&X-Amz-Expires=21600&X-Amz-SignedHeaders=host&X-Amz-Signature=3d7e773fe4a25d0a6dd36aecd9176b13b1717033a69610cfb6815ed156a1b93c", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "remote", "upload_group_key": "20250703T143413", "original_file_md5": "", "original_file_name": "day_cmd.log", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "log", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:34:23.863877-07	UPDATE	\N	\N	\N	f	f
84358e38-c622-4410-9a9c-251c1c96d101	generic_instance	created_dt	2158f0a3-f8e1-4bc9-84f7-4620f6fc8f28	FI7038	2025-07-03 07:34:23.593818-07	2025-07-03T07:34:23.593818-07:00	john@daylilyinformatics.com	2025-07-03 07:34:23.863877-07	UPDATE	\N	\N	\N	f	f
418e0f30-9b76-4bd1-b482-4f9b94b84d1e	generic_instance	modified_dt	2158f0a3-f8e1-4bc9-84f7-4620f6fc8f28	FI7038	2025-07-03 07:34:23.602741-07	2025-07-03T07:34:23.863877-07:00	john@daylilyinformatics.com	2025-07-03 07:34:23.863877-07	UPDATE	\N	\N	\N	f	f
e8d387f8-ad33-48e9-a651-2bad74599572	generic_instance_lineage	\N	a37f01e7-1f5e-41b2-82ce-99c6ab0891f9	GL46	\N	\N	john@daylilyinformatics.com	2025-07-03 07:34:24.204351-07	INSERT	\N	\N	\N	f	f
31f861fa-7e56-4f48-8c2a-700a56553089	generic_instance	\N	cf8a5e09-2051-44be-9121-e9a446b85e5d	FI7039	\N	\N	john@daylilyinformatics.com	2025-07-03 07:34:24.203098-07	INSERT	\N	\N	\N	f	f
1d6cd410-16cb-4529-8cc1-e1e0813313e7	generic_instance	json_addl	cf8a5e09-2051-44be-9121-e9a446b85e5d	FI7039	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Remote", "upload_group_key": "20250703T143413", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Remote", "upload_group_key": "20250703T143413", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:34:24.214508-07	UPDATE	\N	\N	\N	f	f
aa696a2f-7811-415e-9d59-2fd1afb1efca	generic_instance	created_dt	cf8a5e09-2051-44be-9121-e9a446b85e5d	FI7039	2025-07-03 07:34:24.203098-07	2025-07-03T07:34:24.203098-07:00	john@daylilyinformatics.com	2025-07-03 07:34:24.214508-07	UPDATE	\N	\N	\N	f	f
190b3b3a-b4d0-4899-88f5-58ce19a7a39c	generic_instance	modified_dt	cf8a5e09-2051-44be-9121-e9a446b85e5d	FI7039	2025-07-03 07:34:24.203098-07	2025-07-03T07:34:24.214508-07:00	john@daylilyinformatics.com	2025-07-03 07:34:24.214508-07	UPDATE	\N	\N	\N	f	f
047aaa52-1199-4dfc-8719-ea44d1d69c86	generic_instance	json_addl	cf8a5e09-2051-44be-9121-e9a446b85e5d	FI7039	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Remote", "upload_group_key": "20250703T143413", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "day_cmd.log", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "log", "patient_id": "", "remote_url": "https://daylily-service-omics-analysis-us-west-2.s3.us-west-2.amazonaws.com/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log?response-content-disposition=inline&X-Amz-Content-Sha256=UNSIGNED-PAYLOAD&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEA8aCXVzLXdlc3QtMiJHMEUCIEwkFC7typ%2B7k712by339aiZJx867GzZv1rwLjaMWjy9AiEAyFFAGSuNMLiCibMYT9IuQiYJjXD59ZexbYTehsL2hXoquQMIGBAAGgw2NzA0ODQwNTA3MzgiDOLLh4KTH%2Bja%2FjEo0CqWA32BzYyhHkRrNyY1g6iBHMD91fBPNWdcSCjPQ3hFJ9USdNJZJG9R7%2BuWbMT0EQgqOu%2FUTCKIX7JLD8tJJHb1bx90R7BfXWfXKbrg94tOrItTV11IAv5uFUeVa24iDH%2BLCoStSBxskYQJl7MosBPuA0BXnrGsSTi545yLxxZnoUfHNRqltwIQRNut3XxdoDTpLlkirFKMGkIeyY8wJLQZfGoXj9bcau%2BGPw3Ikb20%2FdA79erq9GcJaANTefB3JHU%2FK2c%2B8zZLVnSB1V3%2BtR%2FTWW6thVzY8IFeXGMd9jvKnt9bX4CciVWPyQeBvKVniiLYqwFAdBGbLYmTr72bCfP%2FxA6JAsHooKLkX6bNgGNYbbT38%2BNGS4PDV5tZ7Ud8H13vKDvwn7Bux6JRqbf9erjLK7UJ9KmsLFPGdCvGMWO0c3STI3sw1vJgRQ6DAwwM6kKq4ui3k3yBHMGfF%2FW8MSdLB93B4W3qeKTriP5f8UX8Kw2LmNzLQnat7mPAmCunY3NHOpwaKMiES5XVhXcHGlKFYTz9V6spE3cwhLqYwwY63gJO%2FcyKZeiz3yHTivM0FJNm2TMuIhLmUm9KHl%2BEDGDP9x4XfVnbWza1M%2Fhjhkp7OQVZK0WWMsRmgzSKN%2F%2B0AYGf%2ByswpK6fUKJbeFeDDIxOTh3d8TWJTXj3wDTKuQddBNqLVw6XUuOmVwFBB5%2FLpVvywLDb3iVcNqOmlFJcR2jBR53n0EpZbfIW9j28xVlI965f%2BhcXMhK0D6yvNody3PegC9lDNi7bRERsj72QGCC7Na3HNliMURha2TZoo24SHLa1YLcsyu5ojw%2FIN%2BggWTTMg%2FE%2BWsP9eeEIxNLDj5G7nXkCEdjbfuKPoWUdebP15z9Dknyl7W1sDN2hai5B%2Bk%2B6hTaZziZxr%2FI%2FU2EHgwWkj2ya5qArGlpfK%2BAhlH5%2BVA9I3P3pv7ckjp2J6f2TUErgI3maueoSiTlbQqpEeSVNGu%2BDODQx5799eMGf0f448b29I%2BXKUrDlH0Yoj1RmEA%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAZYG7WWMZCCDE37TC%2F20250703%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20250703T143406Z&X-Amz-Expires=21600&X-Amz-SignedHeaders=host&X-Amz-Signature=3d7e773fe4a25d0a6dd36aecd9176b13b1717033a69610cfb6815ed156a1b93c", "clinician_id": "", "external_ids": [], "original_url": "https://daylily-service-omics-analysis-us-west-2.s3.us-west-2.amazonaws.com/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log?response-content-disposition=inline&X-Amz-Content-Sha256=UNSIGNED-PAYLOAD&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEA8aCXVzLXdlc3QtMiJHMEUCIEwkFC7typ%2B7k712by339aiZJx867GzZv1rwLjaMWjy9AiEAyFFAGSuNMLiCibMYT9IuQiYJjXD59ZexbYTehsL2hXoquQMIGBAAGgw2NzA0ODQwNTA3MzgiDOLLh4KTH%2Bja%2FjEo0CqWA32BzYyhHkRrNyY1g6iBHMD91fBPNWdcSCjPQ3hFJ9USdNJZJG9R7%2BuWbMT0EQgqOu%2FUTCKIX7JLD8tJJHb1bx90R7BfXWfXKbrg94tOrItTV11IAv5uFUeVa24iDH%2BLCoStSBxskYQJl7MosBPuA0BXnrGsSTi545yLxxZnoUfHNRqltwIQRNut3XxdoDTpLlkirFKMGkIeyY8wJLQZfGoXj9bcau%2BGPw3Ikb20%2FdA79erq9GcJaANTefB3JHU%2FK2c%2B8zZLVnSB1V3%2BtR%2FTWW6thVzY8IFeXGMd9jvKnt9bX4CciVWPyQeBvKVniiLYqwFAdBGbLYmTr72bCfP%2FxA6JAsHooKLkX6bNgGNYbbT38%2BNGS4PDV5tZ7Ud8H13vKDvwn7Bux6JRqbf9erjLK7UJ9KmsLFPGdCvGMWO0c3STI3sw1vJgRQ6DAwwM6kKq4ui3k3yBHMGfF%2FW8MSdLB93B4W3qeKTriP5f8UX8Kw2LmNzLQnat7mPAmCunY3NHOpwaKMiES5XVhXcHGlKFYTz9V6spE3cwhLqYwwY63gJO%2FcyKZeiz3yHTivM0FJNm2TMuIhLmUm9KHl%2BEDGDP9x4XfVnbWza1M%2Fhjhkp7OQVZK0WWMsRmgzSKN%2F%2B0AYGf%2ByswpK6fUKJbeFeDDIxOTh3d8TWJTXj3wDTKuQddBNqLVw6XUuOmVwFBB5%2FLpVvywLDb3iVcNqOmlFJcR2jBR53n0EpZbfIW9j28xVlI965f%2BhcXMhK0D6yvNody3PegC9lDNi7bRERsj72QGCC7Na3HNliMURha2TZoo24SHLa1YLcsyu5ojw%2FIN%2BggWTTMg%2FE%2BWsP9eeEIxNLDj5G7nXkCEdjbfuKPoWUdebP15z9Dknyl7W1sDN2hai5B%2Bk%2B6hTaZziZxr%2FI%2FU2EHgwWkj2ya5qArGlpfK%2BAhlH5%2BVA9I3P3pv7ckjp2J6f2TUErgI3maueoSiTlbQqpEeSVNGu%2BDODQx5799eMGf0f448b29I%2BXKUrDlH0Yoj1RmEA%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAZYG7WWMZCCDE37TC%2F20250703%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20250703T143406Z&X-Amz-Expires=21600&X-Amz-SignedHeaders=host&X-Amz-Signature=3d7e773fe4a25d0a6dd36aecd9176b13b1717033a69610cfb6815ed156a1b93c", "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "https://daylily-service-omics-analysis-us-west-2.s3.us-west-2.amazonaws.com/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log?response-content-disposition=inline&X-Amz-Content-Sha256=UNSIGNED-PAYLOAD&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEA8aCXVzLXdlc3QtMiJHMEUCIEwkFC7typ%2B7k712by339aiZJx867GzZv1rwLjaMWjy9AiEAyFFAGSuNMLiCibMYT9IuQiYJjXD59ZexbYTehsL2hXoquQMIGBAAGgw2NzA0ODQwNTA3MzgiDOLLh4KTH%2Bja%2FjEo0CqWA32BzYyhHkRrNyY1g6iBHMD91fBPNWdcSCjPQ3hFJ9USdNJZJG9R7%2BuWbMT0EQgqOu%2FUTCKIX7JLD8tJJHb1bx90R7BfXWfXKbrg94tOrItTV11IAv5uFUeVa24iDH%2BLCoStSBxskYQJl7MosBPuA0BXnrGsSTi545yLxxZnoUfHNRqltwIQRNut3XxdoDTpLlkirFKMGkIeyY8wJLQZfGoXj9bcau%2BGPw3Ikb20%2FdA79erq9GcJaANTefB3JHU%2FK2c%2B8zZLVnSB1V3%2BtR%2FTWW6thVzY8IFeXGMd9jvKnt9bX4CciVWPyQeBvKVniiLYqwFAdBGbLYmTr72bCfP%2FxA6JAsHooKLkX6bNgGNYbbT38%2BNGS4PDV5tZ7Ud8H13vKDvwn7Bux6JRqbf9erjLK7UJ9KmsLFPGdCvGMWO0c3STI3sw1vJgRQ6DAwwM6kKq4ui3k3yBHMGfF%2FW8MSdLB93B4W3qeKTriP5f8UX8Kw2LmNzLQnat7mPAmCunY3NHOpwaKMiES5XVhXcHGlKFYTz9V6spE3cwhLqYwwY63gJO%2FcyKZeiz3yHTivM0FJNm2TMuIhLmUm9KHl%2BEDGDP9x4XfVnbWza1M%2Fhjhkp7OQVZK0WWMsRmgzSKN%2F%2B0AYGf%2ByswpK6fUKJbeFeDDIxOTh3d8TWJTXj3wDTKuQddBNqLVw6XUuOmVwFBB5%2FLpVvywLDb3iVcNqOmlFJcR2jBR53n0EpZbfIW9j28xVlI965f%2BhcXMhK0D6yvNody3PegC9lDNi7bRERsj72QGCC7Na3HNliMURha2TZoo24SHLa1YLcsyu5ojw%2FIN%2BggWTTMg%2FE%2BWsP9eeEIxNLDj5G7nXkCEdjbfuKPoWUdebP15z9Dknyl7W1sDN2hai5B%2Bk%2B6hTaZziZxr%2FI%2FU2EHgwWkj2ya5qArGlpfK%2BAhlH5%2BVA9I3P3pv7ckjp2J6f2TUErgI3maueoSiTlbQqpEeSVNGu%2BDODQx5799eMGf0f448b29I%2BXKUrDlH0Yoj1RmEA%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAZYG7WWMZCCDE37TC%2F20250703%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20250703T143406Z&X-Amz-Expires=21600&X-Amz-SignedHeaders=host&X-Amz-Signature=3d7e773fe4a25d0a6dd36aecd9176b13b1717033a69610cfb6815ed156a1b93c", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "remote", "upload_group_key": "20250703T143413", "original_file_md5": "", "original_file_name": "day_cmd.log", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "log", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:34:24.279969-07	UPDATE	\N	\N	\N	f	f
e3b2f2fa-d60f-4f37-a67b-665038afea8b	generic_instance	created_dt	cf8a5e09-2051-44be-9121-e9a446b85e5d	FI7039	2025-07-03 07:34:24.203098-07	2025-07-03T07:34:24.203098-07:00	john@daylilyinformatics.com	2025-07-03 07:34:24.279969-07	UPDATE	\N	\N	\N	f	f
0e685909-b128-42e3-9276-bf7455463150	generic_instance	modified_dt	cf8a5e09-2051-44be-9121-e9a446b85e5d	FI7039	2025-07-03 07:34:24.214508-07	2025-07-03T07:34:24.279969-07:00	john@daylilyinformatics.com	2025-07-03 07:34:24.279969-07	UPDATE	\N	\N	\N	f	f
248aa503-a0f5-484c-a80f-8606666f9b7d	generic_instance_lineage	\N	f0e1b391-4ccf-4109-953b-41720336197d	GL47	\N	\N	john@daylilyinformatics.com	2025-07-03 07:34:24.520501-07	INSERT	\N	\N	\N	f	f
1fc05b36-e294-4177-be70-3d1a4c6074c1	generic_instance	\N	382785ea-95f6-45b9-b63a-71ba7c409c3d	FS17	\N	\N	john@daylilyinformatics.com	2025-07-03 07:34:37.863248-07	INSERT	\N	\N	\N	f	f
1926a3e2-41b1-42da-af6a-50bbb6bbd925	generic_instance	\N	47f023f3-25bd-4d38-bbf1-c9465815efe0	FI7040	\N	\N	john@daylilyinformatics.com	2025-07-03 07:34:37.902654-07	INSERT	\N	\N	\N	f	f
bcd6b220-bcdf-4770-88e9-5b7de819c006	generic_instance	json_addl	47f023f3-25bd-4d38-bbf1-c9465815efe0	FI7040	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143426", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143426", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:34:37.910349-07	UPDATE	\N	\N	\N	f	f
374f259b-2f50-4d47-bd6d-c522143d1adb	generic_instance	created_dt	47f023f3-25bd-4d38-bbf1-c9465815efe0	FI7040	2025-07-03 07:34:37.902654-07	2025-07-03T07:34:37.902654-07:00	john@daylilyinformatics.com	2025-07-03 07:34:37.910349-07	UPDATE	\N	\N	\N	f	f
9e83f51e-7dae-4279-999d-c4a9efe33203	generic_instance	modified_dt	47f023f3-25bd-4d38-bbf1-c9465815efe0	FI7040	2025-07-03 07:34:37.902654-07	2025-07-03T07:34:37.910349-07:00	john@daylilyinformatics.com	2025-07-03 07:34:37.910349-07	UPDATE	\N	\N	\N	f	f
60952a2e-3495-4420-851f-0a216b300a9b	generic_instance	json_addl	47f023f3-25bd-4d38-bbf1-c9465815efe0	FI7040	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143426", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "day_cmd.log", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "log", "patient_id": "", "clinician_id": "", "external_ids": [], "original_url": "https://daylily-service-omics-analysis-us-west-2.s3.us-west-2.amazonaws.com/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log?response-content-disposition=inline&X-Amz-Content-Sha256=UNSIGNED-PAYLOAD&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEA8aCXVzLXdlc3QtMiJHMEUCIEwkFC7typ%2B7k712by339aiZJx867GzZv1rwLjaMWjy9AiEAyFFAGSuNMLiCibMYT9IuQiYJjXD59ZexbYTehsL2hXoquQMIGBAAGgw2NzA0ODQwNTA3MzgiDOLLh4KTH%2Bja%2FjEo0CqWA32BzYyhHkRrNyY1g6iBHMD91fBPNWdcSCjPQ3hFJ9USdNJZJG9R7%2BuWbMT0EQgqOu%2FUTCKIX7JLD8tJJHb1bx90R7BfXWfXKbrg94tOrItTV11IAv5uFUeVa24iDH%2BLCoStSBxskYQJl7MosBPuA0BXnrGsSTi545yLxxZnoUfHNRqltwIQRNut3XxdoDTpLlkirFKMGkIeyY8wJLQZfGoXj9bcau%2BGPw3Ikb20%2FdA79erq9GcJaANTefB3JHU%2FK2c%2B8zZLVnSB1V3%2BtR%2FTWW6thVzY8IFeXGMd9jvKnt9bX4CciVWPyQeBvKVniiLYqwFAdBGbLYmTr72bCfP%2FxA6JAsHooKLkX6bNgGNYbbT38%2BNGS4PDV5tZ7Ud8H13vKDvwn7Bux6JRqbf9erjLK7UJ9KmsLFPGdCvGMWO0c3STI3sw1vJgRQ6DAwwM6kKq4ui3k3yBHMGfF%2FW8MSdLB93B4W3qeKTriP5f8UX8Kw2LmNzLQnat7mPAmCunY3NHOpwaKMiES5XVhXcHGlKFYTz9V6spE3cwhLqYwwY63gJO%2FcyKZeiz3yHTivM0FJNm2TMuIhLmUm9KHl%2BEDGDP9x4XfVnbWza1M%2Fhjhkp7OQVZK0WWMsRmgzSKN%2F%2B0AYGf%2ByswpK6fUKJbeFeDDIxOTh3d8TWJTXj3wDTKuQddBNqLVw6XUuOmVwFBB5%2FLpVvywLDb3iVcNqOmlFJcR2jBR53n0EpZbfIW9j28xVlI965f%2BhcXMhK0D6yvNody3PegC9lDNi7bRERsj72QGCC7Na3HNliMURha2TZoo24SHLa1YLcsyu5ojw%2FIN%2BggWTTMg%2FE%2BWsP9eeEIxNLDj5G7nXkCEdjbfuKPoWUdebP15z9Dknyl7W1sDN2hai5B%2Bk%2B6hTaZziZxr%2FI%2FU2EHgwWkj2ya5qArGlpfK%2BAhlH5%2BVA9I3P3pv7ckjp2J6f2TUErgI3maueoSiTlbQqpEeSVNGu%2BDODQx5799eMGf0f448b29I%2BXKUrDlH0Yoj1RmEA%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAZYG7WWMZCCDE37TC%2F20250703%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20250703T143406Z&X-Amz-Expires=21600&X-Amz-SignedHeaders=host&X-Amz-Signature=3d7e773fe4a25d0a6dd36aecd9176b13b1717033a69610cfb6815ed156a1b93c", "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7040.log", "current_s3_uri": "s3://daylily-dewey-0/0/FI7040.log", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143426", "original_file_md5": "", "original_file_name": "day_cmd.log", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "log", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 3567, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:34:38.160581-07	UPDATE	\N	\N	\N	f	f
b3952a94-b9f1-4dac-a914-afae26f67c6e	generic_instance	created_dt	47f023f3-25bd-4d38-bbf1-c9465815efe0	FI7040	2025-07-03 07:34:37.902654-07	2025-07-03T07:34:37.902654-07:00	john@daylilyinformatics.com	2025-07-03 07:34:38.160581-07	UPDATE	\N	\N	\N	f	f
8336c3c6-bff3-442a-b9fa-6365dc19907c	generic_instance	modified_dt	47f023f3-25bd-4d38-bbf1-c9465815efe0	FI7040	2025-07-03 07:34:37.910349-07	2025-07-03T07:34:38.160581-07:00	john@daylilyinformatics.com	2025-07-03 07:34:38.160581-07	UPDATE	\N	\N	\N	f	f
0c2c8582-360c-43f2-8cd6-b9bbaca83fe2	generic_instance_lineage	\N	5e646d28-0b01-4fc8-95d4-837b4887af5c	GL48	\N	\N	john@daylilyinformatics.com	2025-07-03 07:34:38.766287-07	INSERT	\N	\N	\N	f	f
fa882671-be34-4d66-827b-0d7bea98e7aa	generic_instance	\N	471467e2-7d73-4ec1-957c-980027da7ccc	FI7041	\N	\N	john@daylilyinformatics.com	2025-07-03 07:34:38.765076-07	INSERT	\N	\N	\N	f	f
5c6d85a4-b369-4a15-a25e-773f9fabdd50	generic_instance	json_addl	471467e2-7d73-4ec1-957c-980027da7ccc	FI7041	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143426", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143426", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:34:38.776702-07	UPDATE	\N	\N	\N	f	f
538d2e8c-6442-409e-9dc2-4972f6079552	generic_instance	created_dt	471467e2-7d73-4ec1-957c-980027da7ccc	FI7041	2025-07-03 07:34:38.765076-07	2025-07-03T07:34:38.765076-07:00	john@daylilyinformatics.com	2025-07-03 07:34:38.776702-07	UPDATE	\N	\N	\N	f	f
a0abfb8a-336a-4af3-b2db-afb84b3a1b05	generic_instance	modified_dt	471467e2-7d73-4ec1-957c-980027da7ccc	FI7041	2025-07-03 07:34:38.765076-07	2025-07-03T07:34:38.776702-07:00	john@daylilyinformatics.com	2025-07-03 07:34:38.776702-07	UPDATE	\N	\N	\N	f	f
f0e26644-e124-4117-8438-b1b803ec94e1	generic_instance	json_addl	471467e2-7d73-4ec1-957c-980027da7ccc	FI7041	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143426", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "day_cmd.log", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "log", "patient_id": "", "clinician_id": "", "external_ids": [], "original_url": "https://daylily-service-omics-analysis-us-west-2.s3.us-west-2.amazonaws.com/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log?response-content-disposition=inline&X-Amz-Content-Sha256=UNSIGNED-PAYLOAD&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEA8aCXVzLXdlc3QtMiJHMEUCIEwkFC7typ%2B7k712by339aiZJx867GzZv1rwLjaMWjy9AiEAyFFAGSuNMLiCibMYT9IuQiYJjXD59ZexbYTehsL2hXoquQMIGBAAGgw2NzA0ODQwNTA3MzgiDOLLh4KTH%2Bja%2FjEo0CqWA32BzYyhHkRrNyY1g6iBHMD91fBPNWdcSCjPQ3hFJ9USdNJZJG9R7%2BuWbMT0EQgqOu%2FUTCKIX7JLD8tJJHb1bx90R7BfXWfXKbrg94tOrItTV11IAv5uFUeVa24iDH%2BLCoStSBxskYQJl7MosBPuA0BXnrGsSTi545yLxxZnoUfHNRqltwIQRNut3XxdoDTpLlkirFKMGkIeyY8wJLQZfGoXj9bcau%2BGPw3Ikb20%2FdA79erq9GcJaANTefB3JHU%2FK2c%2B8zZLVnSB1V3%2BtR%2FTWW6thVzY8IFeXGMd9jvKnt9bX4CciVWPyQeBvKVniiLYqwFAdBGbLYmTr72bCfP%2FxA6JAsHooKLkX6bNgGNYbbT38%2BNGS4PDV5tZ7Ud8H13vKDvwn7Bux6JRqbf9erjLK7UJ9KmsLFPGdCvGMWO0c3STI3sw1vJgRQ6DAwwM6kKq4ui3k3yBHMGfF%2FW8MSdLB93B4W3qeKTriP5f8UX8Kw2LmNzLQnat7mPAmCunY3NHOpwaKMiES5XVhXcHGlKFYTz9V6spE3cwhLqYwwY63gJO%2FcyKZeiz3yHTivM0FJNm2TMuIhLmUm9KHl%2BEDGDP9x4XfVnbWza1M%2Fhjhkp7OQVZK0WWMsRmgzSKN%2F%2B0AYGf%2ByswpK6fUKJbeFeDDIxOTh3d8TWJTXj3wDTKuQddBNqLVw6XUuOmVwFBB5%2FLpVvywLDb3iVcNqOmlFJcR2jBR53n0EpZbfIW9j28xVlI965f%2BhcXMhK0D6yvNody3PegC9lDNi7bRERsj72QGCC7Na3HNliMURha2TZoo24SHLa1YLcsyu5ojw%2FIN%2BggWTTMg%2FE%2BWsP9eeEIxNLDj5G7nXkCEdjbfuKPoWUdebP15z9Dknyl7W1sDN2hai5B%2Bk%2B6hTaZziZxr%2FI%2FU2EHgwWkj2ya5qArGlpfK%2BAhlH5%2BVA9I3P3pv7ckjp2J6f2TUErgI3maueoSiTlbQqpEeSVNGu%2BDODQx5799eMGf0f448b29I%2BXKUrDlH0Yoj1RmEA%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAZYG7WWMZCCDE37TC%2F20250703%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20250703T143406Z&X-Amz-Expires=21600&X-Amz-SignedHeaders=host&X-Amz-Signature=3d7e773fe4a25d0a6dd36aecd9176b13b1717033a69610cfb6815ed156a1b93c", "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7041.log", "current_s3_uri": "s3://daylily-dewey-0/0/FI7041.log", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143426", "original_file_md5": "", "original_file_name": "day_cmd.log", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "log", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 3567, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:34:38.844921-07	UPDATE	\N	\N	\N	f	f
5f90f339-5d3b-46ab-9753-54bf9e94db56	generic_instance	created_dt	471467e2-7d73-4ec1-957c-980027da7ccc	FI7041	2025-07-03 07:34:38.765076-07	2025-07-03T07:34:38.765076-07:00	john@daylilyinformatics.com	2025-07-03 07:34:38.844921-07	UPDATE	\N	\N	\N	f	f
2b1de284-ff75-46ad-aa36-d4921df0f9b8	generic_instance	modified_dt	471467e2-7d73-4ec1-957c-980027da7ccc	FI7041	2025-07-03 07:34:38.776702-07	2025-07-03T07:34:38.844921-07:00	john@daylilyinformatics.com	2025-07-03 07:34:38.844921-07	UPDATE	\N	\N	\N	f	f
2bd0293a-cd84-4e61-b350-307a78cd92b6	generic_instance	\N	42090aeb-3afd-498a-b55f-4af3689f2379	FI7042	\N	\N	john@daylilyinformatics.com	2025-07-03 07:34:39.314264-07	INSERT	\N	\N	\N	f	f
d9d5a41f-8ec4-4556-87f4-8fc9f885f04d	generic_instance	json_addl	42090aeb-3afd-498a-b55f-4af3689f2379	FI7042	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143426", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143426", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:34:39.324573-07	UPDATE	\N	\N	\N	f	f
1ff14d84-9cd4-4e77-ad1e-ecbb3babfa32	generic_instance	created_dt	42090aeb-3afd-498a-b55f-4af3689f2379	FI7042	2025-07-03 07:34:39.314264-07	2025-07-03T07:34:39.314264-07:00	john@daylilyinformatics.com	2025-07-03 07:34:39.324573-07	UPDATE	\N	\N	\N	f	f
2095365e-257b-4b4d-a303-cf1629370ce2	generic_instance	modified_dt	42090aeb-3afd-498a-b55f-4af3689f2379	FI7042	2025-07-03 07:34:39.314264-07	2025-07-03T07:34:39.324573-07:00	john@daylilyinformatics.com	2025-07-03 07:34:39.324573-07	UPDATE	\N	\N	\N	f	f
3f9e19e0-728e-4a73-bf8c-cb97419bdeca	generic_instance	json_addl	42090aeb-3afd-498a-b55f-4af3689f2379	FI7042	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143426", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "day_cmd.log", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "log", "patient_id": "", "clinician_id": "", "external_ids": [], "original_url": "https://daylily-service-omics-analysis-us-west-2.s3.us-west-2.amazonaws.com/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log?response-content-disposition=inline&X-Amz-Content-Sha256=UNSIGNED-PAYLOAD&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEA8aCXVzLXdlc3QtMiJHMEUCIEwkFC7typ%2B7k712by339aiZJx867GzZv1rwLjaMWjy9AiEAyFFAGSuNMLiCibMYT9IuQiYJjXD59ZexbYTehsL2hXoquQMIGBAAGgw2NzA0ODQwNTA3MzgiDOLLh4KTH%2Bja%2FjEo0CqWA32BzYyhHkRrNyY1g6iBHMD91fBPNWdcSCjPQ3hFJ9USdNJZJG9R7%2BuWbMT0EQgqOu%2FUTCKIX7JLD8tJJHb1bx90R7BfXWfXKbrg94tOrItTV11IAv5uFUeVa24iDH%2BLCoStSBxskYQJl7MosBPuA0BXnrGsSTi545yLxxZnoUfHNRqltwIQRNut3XxdoDTpLlkirFKMGkIeyY8wJLQZfGoXj9bcau%2BGPw3Ikb20%2FdA79erq9GcJaANTefB3JHU%2FK2c%2B8zZLVnSB1V3%2BtR%2FTWW6thVzY8IFeXGMd9jvKnt9bX4CciVWPyQeBvKVniiLYqwFAdBGbLYmTr72bCfP%2FxA6JAsHooKLkX6bNgGNYbbT38%2BNGS4PDV5tZ7Ud8H13vKDvwn7Bux6JRqbf9erjLK7UJ9KmsLFPGdCvGMWO0c3STI3sw1vJgRQ6DAwwM6kKq4ui3k3yBHMGfF%2FW8MSdLB93B4W3qeKTriP5f8UX8Kw2LmNzLQnat7mPAmCunY3NHOpwaKMiES5XVhXcHGlKFYTz9V6spE3cwhLqYwwY63gJO%2FcyKZeiz3yHTivM0FJNm2TMuIhLmUm9KHl%2BEDGDP9x4XfVnbWza1M%2Fhjhkp7OQVZK0WWMsRmgzSKN%2F%2B0AYGf%2ByswpK6fUKJbeFeDDIxOTh3d8TWJTXj3wDTKuQddBNqLVw6XUuOmVwFBB5%2FLpVvywLDb3iVcNqOmlFJcR2jBR53n0EpZbfIW9j28xVlI965f%2BhcXMhK0D6yvNody3PegC9lDNi7bRERsj72QGCC7Na3HNliMURha2TZoo24SHLa1YLcsyu5ojw%2FIN%2BggWTTMg%2FE%2BWsP9eeEIxNLDj5G7nXkCEdjbfuKPoWUdebP15z9Dknyl7W1sDN2hai5B%2Bk%2B6hTaZziZxr%2FI%2FU2EHgwWkj2ya5qArGlpfK%2BAhlH5%2BVA9I3P3pv7ckjp2J6f2TUErgI3maueoSiTlbQqpEeSVNGu%2BDODQx5799eMGf0f448b29I%2BXKUrDlH0Yoj1RmEA%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAZYG7WWMZCCDE37TC%2F20250703%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20250703T143406Z&X-Amz-Expires=21600&X-Amz-SignedHeaders=host&X-Amz-Signature=3d7e773fe4a25d0a6dd36aecd9176b13b1717033a69610cfb6815ed156a1b93c", "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7042.log", "current_s3_uri": "s3://daylily-dewey-0/0/FI7042.log", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143426", "original_file_md5": "", "original_file_name": "day_cmd.log", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "log", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 3567, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:34:39.393563-07	UPDATE	\N	\N	\N	f	f
ffa6eb89-02b6-4e5a-bbec-e6d7c0cbc001	generic_instance	created_dt	42090aeb-3afd-498a-b55f-4af3689f2379	FI7042	2025-07-03 07:34:39.314264-07	2025-07-03T07:34:39.314264-07:00	john@daylilyinformatics.com	2025-07-03 07:34:39.393563-07	UPDATE	\N	\N	\N	f	f
8864d526-41f1-44dc-8e33-c2b143361503	generic_instance	modified_dt	42090aeb-3afd-498a-b55f-4af3689f2379	FI7042	2025-07-03 07:34:39.324573-07	2025-07-03T07:34:39.393563-07:00	john@daylilyinformatics.com	2025-07-03 07:34:39.393563-07	UPDATE	\N	\N	\N	f	f
bdd20151-4ec3-484f-98de-8cada323306c	generic_instance	json_addl	7bffe3df-228e-49d0-b3d3-ea7566c97b58	CWX87	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:08.729768-07	UPDATE	\N	\N	\N	f	f
9fa053fe-2fa1-47f9-9e51-f0f00504db4d	generic_instance	created_dt	7bffe3df-228e-49d0-b3d3-ea7566c97b58	CWX87	2025-07-03 08:23:08.723816-07	2025-07-03T08:23:08.723816-07:00	daylily	2025-07-03 08:23:08.729768-07	UPDATE	\N	\N	\N	f	f
5effd6ce-67e6-45dc-9f14-db3caccb4549	generic_instance	modified_dt	7bffe3df-228e-49d0-b3d3-ea7566c97b58	CWX87	2025-07-03 08:23:08.723816-07	2025-07-03T08:23:08.729768-07:00	daylily	2025-07-03 08:23:08.729768-07	UPDATE	\N	\N	\N	f	f
62d5bb21-ed4e-4d78-be6a-dbd74d46f78e	generic_instance_lineage	\N	c3a38f19-4c21-4950-a3b2-008298f34e12	GL305	\N	\N	daylily	2025-07-03 08:23:08.735751-07	INSERT	\N	\N	\N	f	f
9e316268-0c3d-420a-86e0-fb23a62210fd	generic_instance	\N	29ca4ea9-a532-40b8-90a2-bf62f212272c	CWX88	\N	\N	daylily	2025-07-03 08:23:08.735751-07	INSERT	\N	\N	\N	f	f
cd682356-be60-4e6d-9c5b-b833317b544a	generic_instance	json_addl	29ca4ea9-a532-40b8-90a2-bf62f212272c	CWX88	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:08.741924-07	UPDATE	\N	\N	\N	f	f
97ac1840-da11-439a-9ba5-018624360ba5	generic_instance	created_dt	29ca4ea9-a532-40b8-90a2-bf62f212272c	CWX88	2025-07-03 08:23:08.735751-07	2025-07-03T08:23:08.735751-07:00	daylily	2025-07-03 08:23:08.741924-07	UPDATE	\N	\N	\N	f	f
95bacb23-6755-4add-a820-5ed7294fb3a5	generic_instance	modified_dt	29ca4ea9-a532-40b8-90a2-bf62f212272c	CWX88	2025-07-03 08:23:08.735751-07	2025-07-03T08:23:08.741924-07:00	daylily	2025-07-03 08:23:08.741924-07	UPDATE	\N	\N	\N	f	f
5ad8a112-2ad0-41b4-90fa-93bcda7aa25c	generic_instance_lineage	\N	c6a27932-f97c-41d6-8bce-8a647a177d63	GL306	\N	\N	daylily	2025-07-03 08:23:08.743657-07	INSERT	\N	\N	\N	f	f
58b224d7-3066-4ce8-8caa-509e740d6a7b	generic_instance	\N	3b804416-77a1-4840-a1cb-ddce018c129b	CWX89	\N	\N	daylily	2025-07-03 08:23:08.743657-07	INSERT	\N	\N	\N	f	f
b0f49cc0-0dc3-4915-b4ce-3850e3b45a0d	generic_instance	modified_dt	d6dc35b7-cf4c-4e70-a692-c23346f61bcd	CX12	2025-07-03 08:23:17.484168-07	2025-07-03T08:23:17.494103-07:00	daylily	2025-07-03 08:23:17.494103-07	UPDATE	\N	\N	\N	f	f
123a9b3d-4ad2-419c-bab3-1a538ebf2153	generic_instance_lineage	\N	c4bb2d2b-0eb9-4e7a-b314-4a4ca8af24c1	GL527	\N	\N	daylily	2025-07-03 08:23:17.494103-07	INSERT	\N	\N	\N	f	f
bd41c926-bca5-4775-94d9-1c76e5fcfa70	generic_instance	\N	9c6dc920-ebc9-44ca-bf88-13f9ca571726	CX13	\N	\N	daylily	2025-07-03 08:23:17.500261-07	INSERT	\N	\N	\N	f	f
f1b26a73-f865-400f-880f-175b9a5f6a1a	generic_instance	\N	cab1f6d5-eb37-459c-8fc4-4e3a8e81855f	MCX5	\N	\N	daylily	2025-07-03 08:23:17.502951-07	INSERT	\N	\N	\N	f	f
37021e5b-ce4f-4932-a1b8-81f6eca2306b	generic_instance_lineage	\N	17fe66cc-a1bb-4cad-846c-f3d85b03cdfb	GL50	\N	\N	john@daylilyinformatics.com	2025-07-03 07:34:39.874607-07	INSERT	\N	\N	\N	f	f
9d9ba608-5e36-40ba-a52f-a248d97144a6	generic_instance	\N	7ddba52c-036c-420d-b112-ba9419d50c1d	FS18	\N	\N	john@daylilyinformatics.com	2025-07-03 07:35:07.907667-07	INSERT	\N	\N	\N	f	f
f8d1ecc5-16e5-4beb-b2a9-4e6dc26cbef4	generic_instance	\N	5f8a84d3-a799-47fa-8e6e-e95496c7d320	FI7043	\N	\N	john@daylilyinformatics.com	2025-07-03 07:35:08.22041-07	INSERT	\N	\N	\N	f	f
d9ecba3c-6964-4d0f-8945-6a174c8915f6	generic_instance	json_addl	5f8a84d3-a799-47fa-8e6e-e95496c7d320	FI7043	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Remote", "upload_group_key": "20250703T143457", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Remote", "upload_group_key": "20250703T143457", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:35:08.238029-07	UPDATE	\N	\N	\N	f	f
a74b9c84-e8f3-4dea-871b-cc20ec2ab781	generic_instance	created_dt	5f8a84d3-a799-47fa-8e6e-e95496c7d320	FI7043	2025-07-03 07:35:08.22041-07	2025-07-03T07:35:08.22041-07:00	john@daylilyinformatics.com	2025-07-03 07:35:08.238029-07	UPDATE	\N	\N	\N	f	f
e76b6cf7-fd80-4c5b-814d-e92575bb633f	generic_instance	modified_dt	5f8a84d3-a799-47fa-8e6e-e95496c7d320	FI7043	2025-07-03 07:35:08.22041-07	2025-07-03T07:35:08.238029-07:00	john@daylilyinformatics.com	2025-07-03 07:35:08.238029-07	UPDATE	\N	\N	\N	f	f
745e8161-6ecd-4c09-adfe-167efcb6fca9	generic_instance	json_addl	5f8a84d3-a799-47fa-8e6e-e95496c7d320	FI7043	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Remote", "upload_group_key": "20250703T143457", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "day_cmd.log", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "log", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "remote_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "current_s3_key": "FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "current_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "remote", "upload_group_key": "20250703T143457", "original_file_md5": "", "original_file_name": "day_cmd.log", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "log", "current_s3_bucket_name": "daylily-service-omics-analysis-us-west-2", "original_file_data_type": "s3uri", "original_file_size_bytes": null, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:35:08.51139-07	UPDATE	\N	\N	\N	f	f
f6816b19-7a94-40f8-862a-1ff409bdd5bc	generic_instance	created_dt	5f8a84d3-a799-47fa-8e6e-e95496c7d320	FI7043	2025-07-03 07:35:08.22041-07	2025-07-03T07:35:08.22041-07:00	john@daylilyinformatics.com	2025-07-03 07:35:08.51139-07	UPDATE	\N	\N	\N	f	f
aab441b3-f90d-424d-b46e-5f5dd3e0df43	generic_instance	modified_dt	5f8a84d3-a799-47fa-8e6e-e95496c7d320	FI7043	2025-07-03 07:35:08.238029-07	2025-07-03T07:35:08.51139-07:00	john@daylilyinformatics.com	2025-07-03 07:35:08.51139-07	UPDATE	\N	\N	\N	f	f
09d7cfae-9aea-435c-9000-78edc56241fe	generic_instance_lineage	\N	daeb0b62-5fab-4e71-a698-f4460e9ee66f	GL51	\N	\N	john@daylilyinformatics.com	2025-07-03 07:35:09.126574-07	INSERT	\N	\N	\N	f	f
ce3d3d50-cff6-44b6-a9df-f7452c648a7d	generic_instance	\N	afd98d90-d88b-4d7f-8384-8af9b64f4550	FI7044	\N	\N	john@daylilyinformatics.com	2025-07-03 07:35:09.124723-07	INSERT	\N	\N	\N	f	f
985bf796-0715-4ef0-ac56-5ac4eb2282aa	generic_instance	json_addl	afd98d90-d88b-4d7f-8384-8af9b64f4550	FI7044	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Remote", "upload_group_key": "20250703T143457", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Remote", "upload_group_key": "20250703T143457", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:35:09.267574-07	UPDATE	\N	\N	\N	f	f
36a0f138-a080-4f3e-bb82-07da9b6b0c52	generic_instance	created_dt	afd98d90-d88b-4d7f-8384-8af9b64f4550	FI7044	2025-07-03 07:35:09.124723-07	2025-07-03T07:35:09.124723-07:00	john@daylilyinformatics.com	2025-07-03 07:35:09.267574-07	UPDATE	\N	\N	\N	f	f
d8ef6a2e-8876-46e7-a82b-fc736510b43c	generic_instance	modified_dt	afd98d90-d88b-4d7f-8384-8af9b64f4550	FI7044	2025-07-03 07:35:09.124723-07	2025-07-03T07:35:09.267574-07:00	john@daylilyinformatics.com	2025-07-03 07:35:09.267574-07	UPDATE	\N	\N	\N	f	f
05f6a238-b677-4e47-83df-d84cb5945e96	generic_instance	json_addl	afd98d90-d88b-4d7f-8384-8af9b64f4550	FI7044	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Remote", "upload_group_key": "20250703T143457", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "day_cmd.log", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "log", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "remote_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "current_s3_key": "FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "current_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "remote", "upload_group_key": "20250703T143457", "original_file_md5": "", "original_file_name": "day_cmd.log", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "log", "current_s3_bucket_name": "daylily-service-omics-analysis-us-west-2", "original_file_data_type": "s3uri", "original_file_size_bytes": null, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:35:09.391963-07	UPDATE	\N	\N	\N	f	f
dc18f765-5a0b-4385-9ce8-a1244a232754	generic_instance	created_dt	afd98d90-d88b-4d7f-8384-8af9b64f4550	FI7044	2025-07-03 07:35:09.124723-07	2025-07-03T07:35:09.124723-07:00	john@daylilyinformatics.com	2025-07-03 07:35:09.391963-07	UPDATE	\N	\N	\N	f	f
f8bab748-ee30-4cfe-ade7-bab190a103b7	generic_instance	modified_dt	afd98d90-d88b-4d7f-8384-8af9b64f4550	FI7044	2025-07-03 07:35:09.267574-07	2025-07-03T07:35:09.391963-07:00	john@daylilyinformatics.com	2025-07-03 07:35:09.391963-07	UPDATE	\N	\N	\N	f	f
83574958-e18b-4b1e-a508-5cf4bc4bc03d	generic_instance_lineage	\N	8c70110d-9be7-44f3-b54c-7ee15ea3b0ef	GL52	\N	\N	john@daylilyinformatics.com	2025-07-03 07:35:09.825727-07	INSERT	\N	\N	\N	f	f
6643f9c8-a394-413a-9e99-007dc43fbf03	generic_instance	\N	cd291654-0794-44d6-ad84-b147d8b5f61b	FI7045	\N	\N	john@daylilyinformatics.com	2025-07-03 07:35:09.824113-07	INSERT	\N	\N	\N	f	f
27016ee4-0783-4596-81a4-ec60c3c22276	generic_instance	json_addl	cd291654-0794-44d6-ad84-b147d8b5f61b	FI7045	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Remote", "upload_group_key": "20250703T143457", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Remote", "upload_group_key": "20250703T143457", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:35:09.943883-07	UPDATE	\N	\N	\N	f	f
208cd2d6-413b-433c-9d99-c21855333520	generic_instance	created_dt	cd291654-0794-44d6-ad84-b147d8b5f61b	FI7045	2025-07-03 07:35:09.824113-07	2025-07-03T07:35:09.824113-07:00	john@daylilyinformatics.com	2025-07-03 07:35:09.943883-07	UPDATE	\N	\N	\N	f	f
2b2d46cf-d8d3-442d-b886-d77d8ef5e217	generic_instance	modified_dt	cd291654-0794-44d6-ad84-b147d8b5f61b	FI7045	2025-07-03 07:35:09.824113-07	2025-07-03T07:35:09.943883-07:00	john@daylilyinformatics.com	2025-07-03 07:35:09.943883-07	UPDATE	\N	\N	\N	f	f
0427de7a-cb72-46b3-a83d-ff8fea21817d	generic_instance	json_addl	cd291654-0794-44d6-ad84-b147d8b5f61b	FI7045	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Remote", "upload_group_key": "20250703T143457", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "day_cmd.log", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "log", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "remote_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "current_s3_key": "FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "current_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "remote", "upload_group_key": "20250703T143457", "original_file_md5": "", "original_file_name": "day_cmd.log", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "log", "current_s3_bucket_name": "daylily-service-omics-analysis-us-west-2", "original_file_data_type": "s3uri", "original_file_size_bytes": null, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:35:10.060044-07	UPDATE	\N	\N	\N	f	f
55d42393-7ed8-4a27-8b80-4b7ae3dcca2b	generic_instance	created_dt	cd291654-0794-44d6-ad84-b147d8b5f61b	FI7045	2025-07-03 07:35:09.824113-07	2025-07-03T07:35:09.824113-07:00	john@daylilyinformatics.com	2025-07-03 07:35:10.060044-07	UPDATE	\N	\N	\N	f	f
f1511fa6-f3b3-48a0-9323-6632cb9dc244	generic_instance	modified_dt	cd291654-0794-44d6-ad84-b147d8b5f61b	FI7045	2025-07-03 07:35:09.943883-07	2025-07-03T07:35:10.060044-07:00	john@daylilyinformatics.com	2025-07-03 07:35:10.060044-07	UPDATE	\N	\N	\N	f	f
26bab69c-1cfb-4184-8cb7-39a77d349eef	generic_instance_lineage	\N	7d6bdab9-e67c-4c0c-954e-29e82c47f92b	GL53	\N	\N	john@daylilyinformatics.com	2025-07-03 07:35:10.436307-07	INSERT	\N	\N	\N	f	f
6c9fcda2-ad3e-4892-a0c2-5d060e58e155	generic_instance	\N	d8f83592-c673-4ac6-aef5-b79834139010	FS19	\N	\N	john@daylilyinformatics.com	2025-07-03 07:35:23.676353-07	INSERT	\N	\N	\N	f	f
d0ff1561-e4a6-49e1-8ce3-d3c6454a485e	generic_instance	\N	2975d2a7-516a-4e2d-8149-1835e51641fa	FI7046	\N	\N	john@daylilyinformatics.com	2025-07-03 07:35:24.127515-07	INSERT	\N	\N	\N	f	f
a1fa4116-d483-445b-982a-a9f21f180cd4	generic_instance	json_addl	2975d2a7-516a-4e2d-8149-1835e51641fa	FI7046	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143512", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143512", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:35:24.149113-07	UPDATE	\N	\N	\N	f	f
de8b328a-bd46-4010-91d5-fa081c76128c	generic_instance	created_dt	2975d2a7-516a-4e2d-8149-1835e51641fa	FI7046	2025-07-03 07:35:24.127515-07	2025-07-03T07:35:24.127515-07:00	john@daylilyinformatics.com	2025-07-03 07:35:24.149113-07	UPDATE	\N	\N	\N	f	f
515691e7-d47d-4343-a0b7-5c547eae4a96	generic_instance	modified_dt	2975d2a7-516a-4e2d-8149-1835e51641fa	FI7046	2025-07-03 07:35:24.127515-07	2025-07-03T07:35:24.149113-07:00	john@daylilyinformatics.com	2025-07-03 07:35:24.149113-07	UPDATE	\N	\N	\N	f	f
e088bcbf-f66a-4613-a1d0-e8394f020605	generic_instance	json_addl	2975d2a7-516a-4e2d-8149-1835e51641fa	FI7046	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143512", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "day_cmd.log", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "log", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7046.log", "current_s3_uri": "s3://daylily-dewey-0/0/FI7046.log", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143512", "original_file_md5": "", "original_file_name": "day_cmd.log", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "log", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 3567, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:35:24.418254-07	UPDATE	\N	\N	\N	f	f
c1373865-4999-46ab-9258-a5eeda4d0b24	generic_instance	created_dt	2975d2a7-516a-4e2d-8149-1835e51641fa	FI7046	2025-07-03 07:35:24.127515-07	2025-07-03T07:35:24.127515-07:00	john@daylilyinformatics.com	2025-07-03 07:35:24.418254-07	UPDATE	\N	\N	\N	f	f
1612cab9-2d5f-475f-beed-64baef719e56	generic_instance	modified_dt	2975d2a7-516a-4e2d-8149-1835e51641fa	FI7046	2025-07-03 07:35:24.149113-07	2025-07-03T07:35:24.418254-07:00	john@daylilyinformatics.com	2025-07-03 07:35:24.418254-07	UPDATE	\N	\N	\N	f	f
7415fbcc-eb2f-4545-a644-e326aa5142ca	generic_instance_lineage	\N	3ce6bc0d-f5fa-4db4-b5f9-e3c2372ad3fc	GL54	\N	\N	john@daylilyinformatics.com	2025-07-03 07:35:25.385411-07	INSERT	\N	\N	\N	f	f
698748db-56eb-4f9f-860e-bf861cee4cbb	generic_instance	\N	6c35ea5e-3cf4-45dd-9cc7-b1e1dc8f78a1	FI7047	\N	\N	john@daylilyinformatics.com	2025-07-03 07:35:25.377865-07	INSERT	\N	\N	\N	f	f
afaadaee-826a-432a-b9ca-123dab8a03b3	generic_instance	json_addl	6c35ea5e-3cf4-45dd-9cc7-b1e1dc8f78a1	FI7047	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143512", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143512", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:35:25.525869-07	UPDATE	\N	\N	\N	f	f
8ab58eca-1af8-45d1-8ae1-18e854ab7631	generic_instance	created_dt	6c35ea5e-3cf4-45dd-9cc7-b1e1dc8f78a1	FI7047	2025-07-03 07:35:25.377865-07	2025-07-03T07:35:25.377865-07:00	john@daylilyinformatics.com	2025-07-03 07:35:25.525869-07	UPDATE	\N	\N	\N	f	f
a356c8f7-9a86-49a6-9f6b-2879d14a62c9	generic_instance	modified_dt	6c35ea5e-3cf4-45dd-9cc7-b1e1dc8f78a1	FI7047	2025-07-03 07:35:25.377865-07	2025-07-03T07:35:25.525869-07:00	john@daylilyinformatics.com	2025-07-03 07:35:25.525869-07	UPDATE	\N	\N	\N	f	f
32bc8a76-61c4-401a-b3e7-cac08e63b00b	generic_instance	json_addl	6c35ea5e-3cf4-45dd-9cc7-b1e1dc8f78a1	FI7047	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143512", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "day_cmd.log", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "log", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7047.log", "current_s3_uri": "s3://daylily-dewey-0/0/FI7047.log", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143512", "original_file_md5": "", "original_file_name": "day_cmd.log", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "log", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 3567, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:35:25.612131-07	UPDATE	\N	\N	\N	f	f
b91b0d94-bc0a-4b32-b85a-4558edbead67	generic_instance	created_dt	6c35ea5e-3cf4-45dd-9cc7-b1e1dc8f78a1	FI7047	2025-07-03 07:35:25.377865-07	2025-07-03T07:35:25.377865-07:00	john@daylilyinformatics.com	2025-07-03 07:35:25.612131-07	UPDATE	\N	\N	\N	f	f
f5a39134-f160-49d1-aac0-421aef104993	generic_instance_lineage	\N	70dfbfc9-7337-4786-9fcc-908811003123	GL55	\N	\N	john@daylilyinformatics.com	2025-07-03 07:35:26.394088-07	INSERT	\N	\N	\N	f	f
cb7c7ef2-487d-444a-b976-b318037629a7	generic_instance	modified_dt	6c35ea5e-3cf4-45dd-9cc7-b1e1dc8f78a1	FI7047	2025-07-03 07:35:25.525869-07	2025-07-03T07:35:25.612131-07:00	john@daylilyinformatics.com	2025-07-03 07:35:25.612131-07	UPDATE	\N	\N	\N	f	f
cbcf33f0-e667-4496-b72a-3c15821cc37c	generic_instance	\N	f8e647ac-e9b7-4af5-ba52-ddac34812049	FI7048	\N	\N	john@daylilyinformatics.com	2025-07-03 07:35:26.392421-07	INSERT	\N	\N	\N	f	f
603c4bb4-f270-49fc-9b6e-c2f1ed96e1db	generic_instance	json_addl	f8e647ac-e9b7-4af5-ba52-ddac34812049	FI7048	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143512", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143512", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:35:26.527634-07	UPDATE	\N	\N	\N	f	f
37776457-eb6e-46c4-8bc6-a3a761921ce4	generic_instance	created_dt	f8e647ac-e9b7-4af5-ba52-ddac34812049	FI7048	2025-07-03 07:35:26.392421-07	2025-07-03T07:35:26.392421-07:00	john@daylilyinformatics.com	2025-07-03 07:35:26.527634-07	UPDATE	\N	\N	\N	f	f
d6d702ba-307d-45e3-8079-f476701670a1	generic_instance	modified_dt	f8e647ac-e9b7-4af5-ba52-ddac34812049	FI7048	2025-07-03 07:35:26.392421-07	2025-07-03T07:35:26.527634-07:00	john@daylilyinformatics.com	2025-07-03 07:35:26.527634-07	UPDATE	\N	\N	\N	f	f
5a75814f-7bdf-4648-b0b3-fb798db58604	generic_instance	json_addl	f8e647ac-e9b7-4af5-ba52-ddac34812049	FI7048	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143512", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "day_cmd.log", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "log", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7048.log", "current_s3_uri": "s3://daylily-dewey-0/0/FI7048.log", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143512", "original_file_md5": "", "original_file_name": "day_cmd.log", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "log", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 3567, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:35:26.629641-07	UPDATE	\N	\N	\N	f	f
c2b8a19b-46db-4107-b37b-35c8164c23b0	generic_instance	created_dt	f8e647ac-e9b7-4af5-ba52-ddac34812049	FI7048	2025-07-03 07:35:26.392421-07	2025-07-03T07:35:26.392421-07:00	john@daylilyinformatics.com	2025-07-03 07:35:26.629641-07	UPDATE	\N	\N	\N	f	f
6a857307-55b1-4626-9fd7-438ecebec43f	generic_instance	modified_dt	f8e647ac-e9b7-4af5-ba52-ddac34812049	FI7048	2025-07-03 07:35:26.527634-07	2025-07-03T07:35:26.629641-07:00	john@daylilyinformatics.com	2025-07-03 07:35:26.629641-07	UPDATE	\N	\N	\N	f	f
2c7f61b0-efbe-496e-b31c-b2bec0c51f49	generic_instance	json_addl	3b804416-77a1-4840-a1cb-ddce018c129b	CWX89	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:08.749126-07	UPDATE	\N	\N	\N	f	f
c5d8cbd6-83ea-4d63-a610-dffc66b55b01	generic_instance	created_dt	3b804416-77a1-4840-a1cb-ddce018c129b	CWX89	2025-07-03 08:23:08.743657-07	2025-07-03T08:23:08.743657-07:00	daylily	2025-07-03 08:23:08.749126-07	UPDATE	\N	\N	\N	f	f
6b4f8b81-b649-4bc1-8388-18f92812af13	generic_instance	modified_dt	3b804416-77a1-4840-a1cb-ddce018c129b	CWX89	2025-07-03 08:23:08.743657-07	2025-07-03T08:23:08.749126-07:00	daylily	2025-07-03 08:23:08.749126-07	UPDATE	\N	\N	\N	f	f
d37ad23a-9d82-41f2-9d4c-47808d2597a0	generic_instance_lineage	\N	b72dec32-16c0-4ddb-ae94-2743c0db5d82	GL307	\N	\N	daylily	2025-07-03 08:23:08.750583-07	INSERT	\N	\N	\N	f	f
dc987fca-e493-4f02-91cc-129244125713	generic_instance	\N	1bc05522-5e1d-4faa-8aaa-5b4a38ebc55a	CWX90	\N	\N	daylily	2025-07-03 08:23:08.750583-07	INSERT	\N	\N	\N	f	f
6a92d5d8-7f47-4e6a-b80e-5bc0b164d554	generic_instance	json_addl	1bc05522-5e1d-4faa-8aaa-5b4a38ebc55a	CWX90	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:08.755784-07	UPDATE	\N	\N	\N	f	f
86af4fb6-a8de-4066-a993-becb8d526a74	generic_instance	created_dt	1bc05522-5e1d-4faa-8aaa-5b4a38ebc55a	CWX90	2025-07-03 08:23:08.750583-07	2025-07-03T08:23:08.750583-07:00	daylily	2025-07-03 08:23:08.755784-07	UPDATE	\N	\N	\N	f	f
63a77932-f8ea-4fe5-bf36-2579081a8595	generic_instance	modified_dt	1bc05522-5e1d-4faa-8aaa-5b4a38ebc55a	CWX90	2025-07-03 08:23:08.750583-07	2025-07-03T08:23:08.755784-07:00	daylily	2025-07-03 08:23:08.755784-07	UPDATE	\N	\N	\N	f	f
963d6bd8-f821-4277-abca-623462be0291	generic_instance_lineage	\N	ce9cd592-fd05-4d59-ab69-cb1fdbf569ff	GL308	\N	\N	daylily	2025-07-03 08:23:08.757206-07	INSERT	\N	\N	\N	f	f
2bcddd34-9957-47f8-a193-76ad186d16ab	generic_instance	\N	8ce6ad46-136d-48a8-ae5a-0d860b66c006	CWX91	\N	\N	daylily	2025-07-03 08:23:08.757206-07	INSERT	\N	\N	\N	f	f
5d9f7f3a-f802-453b-a47e-846e3185e70a	generic_instance	json_addl	9c6dc920-ebc9-44ca-bf88-13f9ca571726	CX13	{"cogs": {"cost": "4.70", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "10ml glass tube", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "10ml glass tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "4.70", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "s ynthetic-cfdna", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "10ml glass tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.505305-07	UPDATE	\N	\N	\N	f	f
3e716f85-c742-4409-8109-ef42f5f699b7	generic_instance	created_dt	9c6dc920-ebc9-44ca-bf88-13f9ca571726	CX13	2025-07-03 08:23:17.500261-07	2025-07-03T08:23:17.500261-07:00	daylily	2025-07-03 08:23:17.505305-07	UPDATE	\N	\N	\N	f	f
0f6677e0-c597-437f-b0f4-105767097cbd	generic_instance	modified_dt	9c6dc920-ebc9-44ca-bf88-13f9ca571726	CX13	2025-07-03 08:23:17.500261-07	2025-07-03T08:23:17.505305-07:00	daylily	2025-07-03 08:23:17.505305-07	UPDATE	\N	\N	\N	f	f
43299d18-7732-4209-99a8-663b81c33145	generic_instance_lineage	\N	41268426-8f93-40b8-a880-44396be641ad	GL56	\N	\N	john@daylilyinformatics.com	2025-07-03 07:35:27.2801-07	INSERT	\N	\N	\N	f	f
1b424a69-26a1-41f6-a939-9c6e0e92dddf	generic_instance	\N	d3138620-73a5-40a2-b862-ca9cff5395a9	FX6	\N	\N	john@daylilyinformatics.com	2025-07-03 07:35:57.329457-07	INSERT	\N	\N	\N	f	f
b2daf098-07c8-4845-9ef9-89949cb99386	generic_instance_lineage	\N	af9d13bc-f3c6-4942-b097-afca275ee0dd	GL57	\N	\N	john@daylilyinformatics.com	2025-07-03 07:35:57.37592-07	INSERT	\N	\N	\N	f	f
a2d7885d-cf7d-4aa4-9cda-83bf73aeed1c	generic_instance_lineage	\N	ff55ed19-d77e-42e0-92fb-b1367a3f3e83	GL58	\N	\N	john@daylilyinformatics.com	2025-07-03 07:35:57.37592-07	INSERT	\N	\N	\N	f	f
ece342df-21f8-4f02-8c6a-40670da0498c	generic_instance	json_addl	d7e5884d-3ffe-451a-8fbd-99716aff9bb9	FS4	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T134027", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T134027", "tags": [], "comments": "", "duration": 1.0, "lab_code": "", "ref_type": "presigned_url", "description": "File set created by Dewey file manager", "creating_user": "john@daylilyinformatics.com", "rclone_config": {"host": "0.0.0.0", "port": 8080, "user": "user", "bucket": "daylily-dewey-0", "passwd": "passwd"}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	john@daylilyinformatics.com	2025-07-03 07:35:57.312669-07	UPDATE	\N	\N	\N	f	f
1301bee1-084c-4b25-b99d-9dd65cd05967	generic_instance	created_dt	d7e5884d-3ffe-451a-8fbd-99716aff9bb9	FS4	2025-07-03 06:40:40.947157-07	2025-07-03T06:40:40.947157-07:00	john@daylilyinformatics.com	2025-07-03 07:35:57.312669-07	UPDATE	\N	\N	\N	f	f
ed7953e0-d454-4a91-9b01-bc0336ff2928	generic_instance	modified_dt	d7e5884d-3ffe-451a-8fbd-99716aff9bb9	FS4	2025-07-03 06:40:40.947157-07	2025-07-03T07:35:57.312669-07:00	john@daylilyinformatics.com	2025-07-03 07:35:57.312669-07	UPDATE	\N	\N	\N	f	f
3b41615f-368b-42c6-87df-216e74bb8696	generic_instance	\N	d5b8c535-499d-4586-ae3b-4bf5a9871387	FS20	\N	\N	john@daylilyinformatics.com	2025-07-03 07:37:06.886121-07	INSERT	\N	\N	\N	f	f
7f47b91c-a28e-4911-a071-c70bb9c83244	generic_instance	\N	82e5c3f3-22bc-4207-b3a5-7b29bd0ae741	FI7049	\N	\N	john@daylilyinformatics.com	2025-07-03 07:37:07.256559-07	INSERT	\N	\N	\N	f	f
e06be433-33f5-4f5a-8186-ae7a07bdac1c	generic_instance	json_addl	82e5c3f3-22bc-4207-b3a5-7b29bd0ae741	FI7049	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Remote", "upload_group_key": "20250703T143658", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Remote", "upload_group_key": "20250703T143658", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:37:07.281197-07	UPDATE	\N	\N	\N	f	f
dfb5a51f-14be-4ae8-a216-38418979a143	generic_instance	created_dt	82e5c3f3-22bc-4207-b3a5-7b29bd0ae741	FI7049	2025-07-03 07:37:07.256559-07	2025-07-03T07:37:07.256559-07:00	john@daylilyinformatics.com	2025-07-03 07:37:07.281197-07	UPDATE	\N	\N	\N	f	f
70114928-c8c8-45b0-ab56-1be9533722d5	generic_instance	modified_dt	82e5c3f3-22bc-4207-b3a5-7b29bd0ae741	FI7049	2025-07-03 07:37:07.256559-07	2025-07-03T07:37:07.281197-07:00	john@daylilyinformatics.com	2025-07-03 07:37:07.281197-07	UPDATE	\N	\N	\N	f	f
47da29cc-35bf-410c-ace2-f41d0d1cd1c8	generic_instance	json_addl	82e5c3f3-22bc-4207-b3a5-7b29bd0ae741	FI7049	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Remote", "upload_group_key": "20250703T143658", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "day_cmd.log", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "log", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "remote_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "current_s3_key": "FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "current_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "remote", "upload_group_key": "20250703T143658", "original_file_md5": "", "original_file_name": "day_cmd.log", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "log", "current_s3_bucket_name": "daylily-service-omics-analysis-us-west-2", "original_file_data_type": "s3uri", "original_file_size_bytes": null, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:37:07.618985-07	UPDATE	\N	\N	\N	f	f
442e0f35-55b4-4abb-87bd-389f4cf4ecc9	generic_instance	created_dt	82e5c3f3-22bc-4207-b3a5-7b29bd0ae741	FI7049	2025-07-03 07:37:07.256559-07	2025-07-03T07:37:07.256559-07:00	john@daylilyinformatics.com	2025-07-03 07:37:07.618985-07	UPDATE	\N	\N	\N	f	f
8a862396-3c7c-4ad9-9f13-5df0ea55e5cf	generic_instance	modified_dt	82e5c3f3-22bc-4207-b3a5-7b29bd0ae741	FI7049	2025-07-03 07:37:07.281197-07	2025-07-03T07:37:07.618985-07:00	john@daylilyinformatics.com	2025-07-03 07:37:07.618985-07	UPDATE	\N	\N	\N	f	f
9e09ac5a-2109-4e67-b4bf-d049194dccab	generic_instance_lineage	\N	a35749a3-0485-4a30-80b8-6676da5d722e	GL59	\N	\N	john@daylilyinformatics.com	2025-07-03 07:37:08.404927-07	INSERT	\N	\N	\N	f	f
8ab73ba1-a937-4a58-bd53-aec537ba9433	generic_instance	\N	11d26eb0-da43-4f29-b0c9-c234ca621a60	FX7	\N	\N	john@daylilyinformatics.com	2025-07-03 07:37:32.782376-07	INSERT	\N	\N	\N	f	f
08008ece-c53d-45b3-aca5-907cde60537d	generic_instance_lineage	\N	08be9c2a-f237-4a44-b4b7-f630ecf3f8b8	GL60	\N	\N	john@daylilyinformatics.com	2025-07-03 07:37:32.832686-07	INSERT	\N	\N	\N	f	f
dd79ae47-a789-40db-a100-93299419f82d	generic_instance_lineage	\N	fbb6d5e9-1e89-412f-b073-ad4485f40718	GL61	\N	\N	john@daylilyinformatics.com	2025-07-03 07:37:32.832686-07	INSERT	\N	\N	\N	f	f
87698b58-c5e8-4acc-9183-0411c2c5c065	generic_instance	json_addl	d5b8c535-499d-4586-ae3b-4bf5a9871387	FS20	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T143658", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T143658", "tags": [], "comments": "", "duration": 1.0, "lab_code": "", "ref_type": "presigned_url", "description": "File set created by Dewey file manager", "creating_user": "john@daylilyinformatics.com", "rclone_config": {"host": "0.0.0.0", "port": 8080, "user": "user", "bucket": "daylily-dewey-0", "passwd": "passwd"}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	john@daylilyinformatics.com	2025-07-03 07:37:32.770575-07	UPDATE	\N	\N	\N	f	f
8290feb1-32a0-4a3b-b2ba-e8800d7682db	generic_instance	created_dt	d5b8c535-499d-4586-ae3b-4bf5a9871387	FS20	2025-07-03 07:37:06.886121-07	2025-07-03T07:37:06.886121-07:00	john@daylilyinformatics.com	2025-07-03 07:37:32.770575-07	UPDATE	\N	\N	\N	f	f
25e33202-1caf-4e57-a194-7a896352a800	generic_instance	modified_dt	d5b8c535-499d-4586-ae3b-4bf5a9871387	FS20	2025-07-03 07:37:06.886121-07	2025-07-03T07:37:32.770575-07:00	john@daylilyinformatics.com	2025-07-03 07:37:32.770575-07	UPDATE	\N	\N	\N	f	f
c40c9fa1-b404-4174-ae8c-edd6916a9e3b	generic_instance	\N	5834526f-b5d3-4f42-8384-9c26e5c74bc8	FS21	\N	\N	john@daylilyinformatics.com	2025-07-03 07:39:23.027872-07	INSERT	\N	\N	\N	f	f
1e0abf54-795a-4f4b-b1ee-c390e3275215	generic_instance	\N	dde25f19-7032-40ff-a0e6-8ae39183e84b	FI7050	\N	\N	john@daylilyinformatics.com	2025-07-03 07:39:23.277961-07	INSERT	\N	\N	\N	f	f
95de267d-a5a3-4025-b5ec-a401c489897e	generic_instance	json_addl	dde25f19-7032-40ff-a0e6-8ae39183e84b	FI7050	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Remote", "upload_group_key": "20250703T143915", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Remote", "upload_group_key": "20250703T143915", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:39:23.288139-07	UPDATE	\N	\N	\N	f	f
53a9d1e4-edd8-484e-a176-eb30cc085695	generic_instance	created_dt	dde25f19-7032-40ff-a0e6-8ae39183e84b	FI7050	2025-07-03 07:39:23.277961-07	2025-07-03T07:39:23.277961-07:00	john@daylilyinformatics.com	2025-07-03 07:39:23.288139-07	UPDATE	\N	\N	\N	f	f
46282f0f-f7c2-4395-b886-a2068f8b00b4	generic_instance	modified_dt	dde25f19-7032-40ff-a0e6-8ae39183e84b	FI7050	2025-07-03 07:39:23.277961-07	2025-07-03T07:39:23.288139-07:00	john@daylilyinformatics.com	2025-07-03 07:39:23.288139-07	UPDATE	\N	\N	\N	f	f
adb6fd9f-c7c7-400a-b906-29fa8ba2bd3e	generic_instance	json_addl	dde25f19-7032-40ff-a0e6-8ae39183e84b	FI7050	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Remote", "upload_group_key": "20250703T143915", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "main.py", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "py", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "remote_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/main.py", "current_s3_key": "FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/main.py", "current_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/main.py", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/main.py", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "remote", "upload_group_key": "20250703T143915", "original_file_md5": "", "original_file_name": "main.py", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "py", "current_s3_bucket_name": "daylily-service-omics-analysis-us-west-2", "original_file_data_type": "s3uri", "original_file_size_bytes": null, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:39:23.525991-07	UPDATE	\N	\N	\N	f	f
7575c2b3-54e2-413f-94b5-f40447c48fea	generic_instance	created_dt	dde25f19-7032-40ff-a0e6-8ae39183e84b	FI7050	2025-07-03 07:39:23.277961-07	2025-07-03T07:39:23.277961-07:00	john@daylilyinformatics.com	2025-07-03 07:39:23.525991-07	UPDATE	\N	\N	\N	f	f
f4608d31-3b2e-4924-99bc-274e3b0db755	generic_instance	modified_dt	dde25f19-7032-40ff-a0e6-8ae39183e84b	FI7050	2025-07-03 07:39:23.288139-07	2025-07-03T07:39:23.525991-07:00	john@daylilyinformatics.com	2025-07-03 07:39:23.525991-07	UPDATE	\N	\N	\N	f	f
6fe85fa8-4d0c-4280-9666-df9efcf24609	generic_instance	json_addl	8ce6ad46-136d-48a8-ae5a-0d860b66c006	CWX91	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:08.762245-07	UPDATE	\N	\N	\N	f	f
4267637c-b48c-4fea-80d5-98584dce055a	generic_instance	created_dt	8ce6ad46-136d-48a8-ae5a-0d860b66c006	CWX91	2025-07-03 08:23:08.757206-07	2025-07-03T08:23:08.757206-07:00	daylily	2025-07-03 08:23:08.762245-07	UPDATE	\N	\N	\N	f	f
ea86574a-f3ce-43ce-9f51-09ba56d4242a	generic_instance	modified_dt	8ce6ad46-136d-48a8-ae5a-0d860b66c006	CWX91	2025-07-03 08:23:08.757206-07	2025-07-03T08:23:08.762245-07:00	daylily	2025-07-03 08:23:08.762245-07	UPDATE	\N	\N	\N	f	f
01c881ed-c960-43bd-8745-aa997e201f6d	generic_instance_lineage	\N	ca6d1ddb-c0ce-42b6-9f23-eac0863bfc7b	GL309	\N	\N	daylily	2025-07-03 08:23:08.763819-07	INSERT	\N	\N	\N	f	f
5adb0bb0-2c11-4578-baf3-7e4c8c703d46	generic_instance	\N	cb6e1bb2-13ab-4245-b0bb-398bfdae6ac6	CWX92	\N	\N	daylily	2025-07-03 08:23:08.763819-07	INSERT	\N	\N	\N	f	f
e92440df-9425-4f3f-89cc-871b05bea511	generic_instance	json_addl	cb6e1bb2-13ab-4245-b0bb-398bfdae6ac6	CWX92	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:08.769863-07	UPDATE	\N	\N	\N	f	f
49fc84a3-91ae-4393-8776-82d0321dce9f	generic_instance	created_dt	cb6e1bb2-13ab-4245-b0bb-398bfdae6ac6	CWX92	2025-07-03 08:23:08.763819-07	2025-07-03T08:23:08.763819-07:00	daylily	2025-07-03 08:23:08.769863-07	UPDATE	\N	\N	\N	f	f
1d5caa92-aefb-4188-b860-88e553a12485	generic_instance	modified_dt	cb6e1bb2-13ab-4245-b0bb-398bfdae6ac6	CWX92	2025-07-03 08:23:08.763819-07	2025-07-03T08:23:08.769863-07:00	daylily	2025-07-03 08:23:08.769863-07	UPDATE	\N	\N	\N	f	f
fff5108f-2449-44b9-a5aa-3852ff90a248	generic_instance_lineage	\N	64132bfa-d1cb-4d53-8820-2a3d526b34ce	GL310	\N	\N	daylily	2025-07-03 08:23:08.771673-07	INSERT	\N	\N	\N	f	f
ebe734ae-7518-4ab6-8c76-0a5261f5f47e	generic_instance	\N	da03904f-c644-4f7b-a603-11082db6655d	CWX93	\N	\N	daylily	2025-07-03 08:23:08.771673-07	INSERT	\N	\N	\N	f	f
62dff6d1-1072-4f05-b953-8b79190ff871	generic_instance	json_addl	da03904f-c644-4f7b-a603-11082db6655d	CWX93	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:08.777739-07	UPDATE	\N	\N	\N	f	f
88ebdfb7-95f7-4b4b-a93f-1a5f48d4e8f0	generic_instance	created_dt	da03904f-c644-4f7b-a603-11082db6655d	CWX93	2025-07-03 08:23:08.771673-07	2025-07-03T08:23:08.771673-07:00	daylily	2025-07-03 08:23:08.777739-07	UPDATE	\N	\N	\N	f	f
a5c72aad-e2d4-4da2-a308-faff22618395	generic_instance	modified_dt	da03904f-c644-4f7b-a603-11082db6655d	CWX93	2025-07-03 08:23:08.771673-07	2025-07-03T08:23:08.777739-07:00	daylily	2025-07-03 08:23:08.777739-07	UPDATE	\N	\N	\N	f	f
72c39331-d324-4ad2-8137-1ee089952f44	generic_instance_lineage	\N	074be2e1-7e8f-45c2-aac1-b0ecadfb866a	GL311	\N	\N	daylily	2025-07-03 08:23:08.779117-07	INSERT	\N	\N	\N	f	f
aca84a69-1125-4328-8537-5be936216839	generic_instance	\N	bc77f3ee-c50b-43eb-8f41-d9fec67a5da5	CWX94	\N	\N	daylily	2025-07-03 08:23:08.779117-07	INSERT	\N	\N	\N	f	f
141c0b5d-b4ba-4ab4-886c-eaa264f7a131	generic_instance_lineage	\N	3a410823-ee1b-4542-bd9a-d41c070c02a2	GL528	\N	\N	daylily	2025-07-03 08:23:17.505305-07	INSERT	\N	\N	\N	f	f
9886445e-9673-4905-8859-2473f60c9085	generic_instance	\N	fe1b21ed-3957-4df1-b13c-fcaccc035a83	CX14	\N	\N	daylily	2025-07-03 08:23:17.509268-07	INSERT	\N	\N	\N	f	f
52bb0e92-ded8-4d46-92bc-0b479723b871	generic_instance_lineage	\N	d82bcd53-e48a-4fbc-add8-11809e0c36a0	GL62	\N	\N	john@daylilyinformatics.com	2025-07-03 07:39:24.078427-07	INSERT	\N	\N	\N	f	f
36919be4-a1cd-4539-ac08-7bd9da629f47	generic_instance	\N	21aad253-3d80-4b11-9220-290cf6ff8a34	FX8	\N	\N	john@daylilyinformatics.com	2025-07-03 07:39:39.332504-07	INSERT	\N	\N	\N	f	f
99b2725c-efca-4d22-aa81-dc0c9881ce75	generic_instance_lineage	\N	ff98a429-d339-47c9-815a-cab619e20fcc	GL63	\N	\N	john@daylilyinformatics.com	2025-07-03 07:39:39.360451-07	INSERT	\N	\N	\N	f	f
73298243-2c44-46d6-9437-eccaaae7772c	generic_instance_lineage	\N	3177f944-91c2-4db1-a3e4-91f333ce65fd	GL64	\N	\N	john@daylilyinformatics.com	2025-07-03 07:39:39.360451-07	INSERT	\N	\N	\N	f	f
e183c420-03a0-4c41-a1e3-852b5942f1d5	generic_instance	json_addl	5834526f-b5d3-4f42-8384-9c26e5c74bc8	FS21	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T143915", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T143915", "tags": [], "comments": "", "duration": 1.0, "lab_code": "", "ref_type": "presigned_url", "description": "File set created by Dewey file manager", "creating_user": "john@daylilyinformatics.com", "rclone_config": {"host": "0.0.0.0", "port": 8080, "user": "user", "bucket": "daylily-dewey-0", "passwd": "passwd"}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	john@daylilyinformatics.com	2025-07-03 07:39:39.32352-07	UPDATE	\N	\N	\N	f	f
8884562c-a9a3-4339-ad64-697e09144191	generic_instance	created_dt	5834526f-b5d3-4f42-8384-9c26e5c74bc8	FS21	2025-07-03 07:39:23.027872-07	2025-07-03T07:39:23.027872-07:00	john@daylilyinformatics.com	2025-07-03 07:39:39.32352-07	UPDATE	\N	\N	\N	f	f
29436d08-1526-4913-82e9-2de56f355af5	generic_instance	modified_dt	5834526f-b5d3-4f42-8384-9c26e5c74bc8	FS21	2025-07-03 07:39:23.027872-07	2025-07-03T07:39:39.32352-07:00	john@daylilyinformatics.com	2025-07-03 07:39:39.32352-07	UPDATE	\N	\N	\N	f	f
6b7e7306-5f4c-4f95-a910-8a0844a90e29	generic_instance	\N	2ff79f5f-1663-4dce-bbf3-13e0cc787c2d	FS22	\N	\N	john@daylilyinformatics.com	2025-07-03 07:40:54.268794-07	INSERT	\N	\N	\N	f	f
9fc0b06c-da01-4e21-a42d-0d306572653b	generic_instance	\N	1bccdecd-c0cf-41d5-a0b5-13be26cdb88c	FI7051	\N	\N	john@daylilyinformatics.com	2025-07-03 07:40:54.504759-07	INSERT	\N	\N	\N	f	f
cbb66286-bc3b-4768-a857-338092c59e09	generic_instance	json_addl	1bccdecd-c0cf-41d5-a0b5-13be26cdb88c	FI7051	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Remote", "upload_group_key": "20250703T144046", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Remote", "upload_group_key": "20250703T144046", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:40:54.513586-07	UPDATE	\N	\N	\N	f	f
59513bc0-df8d-495e-8772-1a711c5db471	generic_instance	created_dt	1bccdecd-c0cf-41d5-a0b5-13be26cdb88c	FI7051	2025-07-03 07:40:54.504759-07	2025-07-03T07:40:54.504759-07:00	john@daylilyinformatics.com	2025-07-03 07:40:54.513586-07	UPDATE	\N	\N	\N	f	f
ac2a0c62-3eb2-4f72-b612-09fdbc48935e	generic_instance	modified_dt	1bccdecd-c0cf-41d5-a0b5-13be26cdb88c	FI7051	2025-07-03 07:40:54.504759-07	2025-07-03T07:40:54.513586-07:00	john@daylilyinformatics.com	2025-07-03 07:40:54.513586-07	UPDATE	\N	\N	\N	f	f
8df13f88-bb47-4ff7-a320-23d655d57ba9	generic_instance	json_addl	1bccdecd-c0cf-41d5-a0b5-13be26cdb88c	FI7051	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Remote", "upload_group_key": "20250703T144046", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "rule_dag.png", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "png", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "remote_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/rule_dag.png", "current_s3_key": "FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/rule_dag.png", "current_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/rule_dag.png", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/rule_dag.png", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "remote", "upload_group_key": "20250703T144046", "original_file_md5": "", "original_file_name": "rule_dag.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-service-omics-analysis-us-west-2", "original_file_data_type": "s3uri", "original_file_size_bytes": null, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:40:54.771021-07	UPDATE	\N	\N	\N	f	f
7fa60029-acc9-4058-a708-209b5bd6d8e2	generic_instance	created_dt	1bccdecd-c0cf-41d5-a0b5-13be26cdb88c	FI7051	2025-07-03 07:40:54.504759-07	2025-07-03T07:40:54.504759-07:00	john@daylilyinformatics.com	2025-07-03 07:40:54.771021-07	UPDATE	\N	\N	\N	f	f
c078cb44-36c6-452b-a123-4eaecfb3378b	generic_instance	modified_dt	1bccdecd-c0cf-41d5-a0b5-13be26cdb88c	FI7051	2025-07-03 07:40:54.513586-07	2025-07-03T07:40:54.771021-07:00	john@daylilyinformatics.com	2025-07-03 07:40:54.771021-07	UPDATE	\N	\N	\N	f	f
1fdcdbd7-59f7-44f8-b756-169c665fd9b4	generic_instance_lineage	\N	cfbf10c4-6988-40b1-8360-c61d9fc764b2	GL65	\N	\N	john@daylilyinformatics.com	2025-07-03 07:40:55.423969-07	INSERT	\N	\N	\N	f	f
3bd2c301-cfbf-4d71-82ee-3689d4ca2844	generic_instance	\N	6182f00d-f0e3-4b74-888f-54f693a6f8fe	FX9	\N	\N	john@daylilyinformatics.com	2025-07-03 07:41:07.378265-07	INSERT	\N	\N	\N	f	f
261e81ec-8823-40b1-8716-1ca0d202822d	generic_instance_lineage	\N	808f1552-93c1-40ee-bd40-b31602d8edae	GL66	\N	\N	john@daylilyinformatics.com	2025-07-03 07:41:07.398316-07	INSERT	\N	\N	\N	f	f
e0628225-9d26-4c7e-87c3-d09d9121bf16	generic_instance_lineage	\N	390be49f-6b09-4027-8804-ca78852fd601	GL67	\N	\N	john@daylilyinformatics.com	2025-07-03 07:41:07.398316-07	INSERT	\N	\N	\N	f	f
29a942fe-d835-4bc0-91cd-8a9101e04ef0	generic_instance	json_addl	2ff79f5f-1663-4dce-bbf3-13e0cc787c2d	FS22	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144046", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144046", "tags": [], "comments": "", "duration": 1.0, "lab_code": "", "ref_type": "presigned_url", "description": "File set created by Dewey file manager", "creating_user": "john@daylilyinformatics.com", "rclone_config": {"host": "0.0.0.0", "port": 8080, "user": "user", "bucket": "daylily-dewey-0", "passwd": "passwd"}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	john@daylilyinformatics.com	2025-07-03 07:41:07.372769-07	UPDATE	\N	\N	\N	f	f
f4bbd65d-def9-4b9e-bab2-192167b14b5c	generic_instance	created_dt	2ff79f5f-1663-4dce-bbf3-13e0cc787c2d	FS22	2025-07-03 07:40:54.268794-07	2025-07-03T07:40:54.268794-07:00	john@daylilyinformatics.com	2025-07-03 07:41:07.372769-07	UPDATE	\N	\N	\N	f	f
64b93479-3a6b-4a5f-b05c-5a8493c3dfc3	generic_instance	modified_dt	2ff79f5f-1663-4dce-bbf3-13e0cc787c2d	FS22	2025-07-03 07:40:54.268794-07	2025-07-03T07:41:07.372769-07:00	john@daylilyinformatics.com	2025-07-03 07:41:07.372769-07	UPDATE	\N	\N	\N	f	f
cecfb192-ba08-4ac5-89da-9c7060684f8e	generic_instance	\N	cd99a9ea-224d-4cda-b8a2-2027e4be1767	FS23	\N	\N	john@daylilyinformatics.com	2025-07-03 07:43:08.238971-07	INSERT	\N	\N	\N	f	f
d25b5543-19a6-409c-b84a-66f0ee05a194	generic_instance	\N	751326b8-0128-4f6d-8f82-05bf70cbd6c8	FI7052	\N	\N	john@daylilyinformatics.com	2025-07-03 07:43:08.26847-07	INSERT	\N	\N	\N	f	f
7de38aa8-a160-402d-9a2b-7a12cd0bdb88	generic_instance	json_addl	751326b8-0128-4f6d-8f82-05bf70cbd6c8	FI7052	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T144308", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T144308", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:43:08.272789-07	UPDATE	\N	\N	\N	f	f
2a7d2fbc-3495-4113-a4e6-0872b0f223e0	generic_instance	created_dt	751326b8-0128-4f6d-8f82-05bf70cbd6c8	FI7052	2025-07-03 07:43:08.26847-07	2025-07-03T07:43:08.26847-07:00	john@daylilyinformatics.com	2025-07-03 07:43:08.272789-07	UPDATE	\N	\N	\N	f	f
8705f523-c4a4-4a03-a865-7e7a5145150c	generic_instance	modified_dt	751326b8-0128-4f6d-8f82-05bf70cbd6c8	FI7052	2025-07-03 07:43:08.26847-07	2025-07-03T07:43:08.272789-07:00	john@daylilyinformatics.com	2025-07-03 07:43:08.272789-07	UPDATE	\N	\N	\N	f	f
0a430a6b-2a97-4cd8-8ee9-6b2b4d80d467	generic_instance	json_addl	751326b8-0128-4f6d-8f82-05bf70cbd6c8	FI7052	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T144308", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7052.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7052.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T144308", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:43:08.517518-07	UPDATE	\N	\N	\N	f	f
ddb2fe59-46ba-40e3-8cbd-1d0e357a51d7	generic_instance	created_dt	751326b8-0128-4f6d-8f82-05bf70cbd6c8	FI7052	2025-07-03 07:43:08.26847-07	2025-07-03T07:43:08.26847-07:00	john@daylilyinformatics.com	2025-07-03 07:43:08.517518-07	UPDATE	\N	\N	\N	f	f
976e1aa4-e03c-49d0-bc41-a10fb77eea30	generic_instance	modified_dt	751326b8-0128-4f6d-8f82-05bf70cbd6c8	FI7052	2025-07-03 07:43:08.272789-07	2025-07-03T07:43:08.517518-07:00	john@daylilyinformatics.com	2025-07-03 07:43:08.517518-07	UPDATE	\N	\N	\N	f	f
71bc1ac3-8395-4709-8a8c-190acf9d88dc	generic_instance_lineage	\N	7c6aa1df-232f-48f8-b5c5-a58cce4b1b6e	GL68	\N	\N	john@daylilyinformatics.com	2025-07-03 07:43:09.55241-07	INSERT	\N	\N	\N	f	f
e5d27f2c-f2c5-4831-9cc5-68995fc3cf97	generic_instance	\N	42c52d3e-cb20-4c08-87fe-c36891087f40	FI7053	\N	\N	john@daylilyinformatics.com	2025-07-03 07:43:09.559286-07	INSERT	\N	\N	\N	f	f
77bf1e02-2fe8-4598-a467-f430c2119629	generic_instance	json_addl	42c52d3e-cb20-4c08-87fe-c36891087f40	FI7053	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T144308", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T144308", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:43:09.73722-07	UPDATE	\N	\N	\N	f	f
6858c517-4842-4ce9-89de-8f86f7d01c51	generic_instance	created_dt	42c52d3e-cb20-4c08-87fe-c36891087f40	FI7053	2025-07-03 07:43:09.559286-07	2025-07-03T07:43:09.559286-07:00	john@daylilyinformatics.com	2025-07-03 07:43:09.73722-07	UPDATE	\N	\N	\N	f	f
ec94a82d-fb6e-4f41-8e45-3eda2dd3e816	generic_instance	modified_dt	42c52d3e-cb20-4c08-87fe-c36891087f40	FI7053	2025-07-03 07:43:09.559286-07	2025-07-03T07:43:09.73722-07:00	john@daylilyinformatics.com	2025-07-03 07:43:09.73722-07	UPDATE	\N	\N	\N	f	f
3803565a-a364-4be4-acc8-359bebfa4a8b	generic_instance	json_addl	42c52d3e-cb20-4c08-87fe-c36891087f40	FI7053	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T144308", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7053.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7053.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T144308", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:43:09.844984-07	UPDATE	\N	\N	\N	f	f
7733c2bb-2496-4f5e-afb7-5059a16e8782	generic_instance	created_dt	42c52d3e-cb20-4c08-87fe-c36891087f40	FI7053	2025-07-03 07:43:09.559286-07	2025-07-03T07:43:09.559286-07:00	john@daylilyinformatics.com	2025-07-03 07:43:09.844984-07	UPDATE	\N	\N	\N	f	f
88a8a62d-1cc4-4533-9972-2f9893970400	generic_instance	modified_dt	42c52d3e-cb20-4c08-87fe-c36891087f40	FI7053	2025-07-03 07:43:09.73722-07	2025-07-03T07:43:09.844984-07:00	john@daylilyinformatics.com	2025-07-03 07:43:09.844984-07	UPDATE	\N	\N	\N	f	f
e29dcebe-3df6-4e9c-8af3-dbbbcbc9a503	generic_instance_lineage	\N	e3e7fe3d-2472-4cbe-8837-3abbc8c5124c	GL69	\N	\N	john@daylilyinformatics.com	2025-07-03 07:43:10.443934-07	INSERT	\N	\N	\N	f	f
38eadb3f-7529-42ed-917f-c48f2da0468b	generic_instance	\N	ade4c01a-0859-4437-b6b1-63a99088aaae	FI7054	\N	\N	john@daylilyinformatics.com	2025-07-03 07:43:10.449991-07	INSERT	\N	\N	\N	f	f
d29c5d86-a646-49b6-bfe2-7bf4a25fb2a0	generic_instance	json_addl	ade4c01a-0859-4437-b6b1-63a99088aaae	FI7054	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T144308", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T144308", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:43:10.46002-07	UPDATE	\N	\N	\N	f	f
6510279a-90d6-496d-9d8e-3620867127bf	generic_instance	created_dt	ade4c01a-0859-4437-b6b1-63a99088aaae	FI7054	2025-07-03 07:43:10.449991-07	2025-07-03T07:43:10.449991-07:00	john@daylilyinformatics.com	2025-07-03 07:43:10.46002-07	UPDATE	\N	\N	\N	f	f
6f02525a-ab9a-42be-8cf3-bd974dd4fe41	generic_instance	modified_dt	ade4c01a-0859-4437-b6b1-63a99088aaae	FI7054	2025-07-03 07:43:10.449991-07	2025-07-03T07:43:10.46002-07:00	john@daylilyinformatics.com	2025-07-03 07:43:10.46002-07	UPDATE	\N	\N	\N	f	f
17df379c-cf1c-4a70-8555-1f506e80be6b	generic_instance	json_addl	ade4c01a-0859-4437-b6b1-63a99088aaae	FI7054	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T144308", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7054.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7054.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T144308", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:43:10.679092-07	UPDATE	\N	\N	\N	f	f
67eda34f-7b7d-47d0-869b-5baeaa8b07ae	generic_instance	created_dt	ade4c01a-0859-4437-b6b1-63a99088aaae	FI7054	2025-07-03 07:43:10.449991-07	2025-07-03T07:43:10.449991-07:00	john@daylilyinformatics.com	2025-07-03 07:43:10.679092-07	UPDATE	\N	\N	\N	f	f
2d3b2a87-6688-4e83-a9cb-f719fb164947	generic_instance	modified_dt	ade4c01a-0859-4437-b6b1-63a99088aaae	FI7054	2025-07-03 07:43:10.46002-07	2025-07-03T07:43:10.679092-07:00	john@daylilyinformatics.com	2025-07-03 07:43:10.679092-07	UPDATE	\N	\N	\N	f	f
5735d3ff-1873-447e-a440-66d8b5b481ac	generic_instance_lineage	\N	1b86b600-46b4-4f13-906b-7c4f124ee281	GL70	\N	\N	john@daylilyinformatics.com	2025-07-03 07:43:11.454347-07	INSERT	\N	\N	\N	f	f
91815dcb-dc9a-44e5-a94e-d8e587fa45d5	generic_instance	\N	50d62952-ee77-48c2-adf1-a4fd30234dd6	FI7055	\N	\N	john@daylilyinformatics.com	2025-07-03 07:43:11.46863-07	INSERT	\N	\N	\N	f	f
3c5b1ddf-eb14-496c-ae6e-62a918bef825	generic_instance	json_addl	50d62952-ee77-48c2-adf1-a4fd30234dd6	FI7055	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T144308", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T144308", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:43:11.656813-07	UPDATE	\N	\N	\N	f	f
4ebbbc4d-cfda-449d-889b-b31643719782	generic_instance	created_dt	50d62952-ee77-48c2-adf1-a4fd30234dd6	FI7055	2025-07-03 07:43:11.46863-07	2025-07-03T07:43:11.46863-07:00	john@daylilyinformatics.com	2025-07-03 07:43:11.656813-07	UPDATE	\N	\N	\N	f	f
e6852a9f-060d-4058-8383-033742f75af8	generic_instance	modified_dt	50d62952-ee77-48c2-adf1-a4fd30234dd6	FI7055	2025-07-03 07:43:11.46863-07	2025-07-03T07:43:11.656813-07:00	john@daylilyinformatics.com	2025-07-03 07:43:11.656813-07	UPDATE	\N	\N	\N	f	f
9a097c1d-3455-4455-8198-8e433919e5bb	generic_instance	json_addl	50d62952-ee77-48c2-adf1-a4fd30234dd6	FI7055	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T144308", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7055.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7055.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T144308", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:43:11.77149-07	UPDATE	\N	\N	\N	f	f
dad375b0-42df-4e2f-80f0-ee6ad904b2c6	generic_instance	created_dt	50d62952-ee77-48c2-adf1-a4fd30234dd6	FI7055	2025-07-03 07:43:11.46863-07	2025-07-03T07:43:11.46863-07:00	john@daylilyinformatics.com	2025-07-03 07:43:11.77149-07	UPDATE	\N	\N	\N	f	f
d3a6ba32-9618-4fe4-8026-b7561c7869c7	generic_instance	modified_dt	50d62952-ee77-48c2-adf1-a4fd30234dd6	FI7055	2025-07-03 07:43:11.656813-07	2025-07-03T07:43:11.77149-07:00	john@daylilyinformatics.com	2025-07-03 07:43:11.77149-07	UPDATE	\N	\N	\N	f	f
75631738-9d3e-4646-a18b-814ee64549af	generic_instance_lineage	\N	ae6cb85a-3060-4047-b619-78f7a7e101ce	GL71	\N	\N	john@daylilyinformatics.com	2025-07-03 07:43:12.588778-07	INSERT	\N	\N	\N	f	f
63d92e84-4845-4954-910a-c168c5b5762e	generic_instance	\N	81d233d2-dffb-4b82-9cea-61c8009b5f2b	FS24	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:17.728972-07	INSERT	\N	\N	\N	f	f
34c9db69-4333-427d-9829-eb5dffd1b9a6	generic_instance	\N	f3300273-7c20-42a6-aa2d-101f1f721a6d	FI7056	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:17.796677-07	INSERT	\N	\N	\N	f	f
afb4cd18-a6dd-4337-a53d-dc9525a1d1a1	generic_instance	json_addl	f3300273-7c20-42a6-aa2d-101f1f721a6d	FI7056	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:17.806281-07	UPDATE	\N	\N	\N	f	f
7d60883e-94ba-4d95-9003-723f21cd2235	generic_instance	created_dt	f3300273-7c20-42a6-aa2d-101f1f721a6d	FI7056	2025-07-03 07:46:17.796677-07	2025-07-03T07:46:17.796677-07:00	john@daylilyinformatics.com	2025-07-03 07:46:17.806281-07	UPDATE	\N	\N	\N	f	f
7007d3e9-37a8-4929-9d02-8e4b125e2e42	generic_instance	modified_dt	f3300273-7c20-42a6-aa2d-101f1f721a6d	FI7056	2025-07-03 07:46:17.796677-07	2025-07-03T07:46:17.806281-07:00	john@daylilyinformatics.com	2025-07-03 07:46:17.806281-07	UPDATE	\N	\N	\N	f	f
35451b6d-2170-412e-8d12-33aef21d971a	generic_instance	json_addl	f3300273-7c20-42a6-aa2d-101f1f721a6d	FI7056	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp2s1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7056.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7056.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp2s1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 809418, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:18.064288-07	UPDATE	\N	\N	\N	f	f
3a84f28c-296e-4997-9156-ffe6421a4512	generic_instance	created_dt	f3300273-7c20-42a6-aa2d-101f1f721a6d	FI7056	2025-07-03 07:46:17.796677-07	2025-07-03T07:46:17.796677-07:00	john@daylilyinformatics.com	2025-07-03 07:46:18.064288-07	UPDATE	\N	\N	\N	f	f
d3c197ab-e2c8-4112-803d-1ed1974bafc7	generic_instance	modified_dt	f3300273-7c20-42a6-aa2d-101f1f721a6d	FI7056	2025-07-03 07:46:17.806281-07	2025-07-03T07:46:18.064288-07:00	john@daylilyinformatics.com	2025-07-03 07:46:18.064288-07	UPDATE	\N	\N	\N	f	f
c7bab10a-579e-421c-996c-82e5b22dee17	generic_instance_lineage	\N	af276406-fc93-4896-b3e0-8227c6058a29	GL72	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:18.805557-07	INSERT	\N	\N	\N	f	f
a9d67fe8-7e8f-4fe2-9619-e962c8b5d14c	generic_instance	\N	5b761ebc-9320-4d44-9d95-c80a289eb8ae	FI7057	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:18.804449-07	INSERT	\N	\N	\N	f	f
1b536ce4-9c29-487e-bbf5-6fa7055c5167	generic_instance	json_addl	5b761ebc-9320-4d44-9d95-c80a289eb8ae	FI7057	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:18.816712-07	UPDATE	\N	\N	\N	f	f
9ec280c1-1ad9-4bb8-8dca-2da3dd1f6166	generic_instance	created_dt	5b761ebc-9320-4d44-9d95-c80a289eb8ae	FI7057	2025-07-03 07:46:18.804449-07	2025-07-03T07:46:18.804449-07:00	john@daylilyinformatics.com	2025-07-03 07:46:18.816712-07	UPDATE	\N	\N	\N	f	f
9138978c-be50-47c6-a08c-8dbbeb8f2af4	generic_instance	modified_dt	5b761ebc-9320-4d44-9d95-c80a289eb8ae	FI7057	2025-07-03 07:46:18.804449-07	2025-07-03T07:46:18.816712-07:00	john@daylilyinformatics.com	2025-07-03 07:46:18.816712-07	UPDATE	\N	\N	\N	f	f
3a4d5704-1689-4bf5-9586-65d0453ca70c	generic_instance	\N	7fcfa70d-2016-45c2-a092-4f42ebed1393	MCX6	\N	\N	daylily	2025-07-03 08:23:17.511787-07	INSERT	\N	\N	\N	f	f
f12df035-25ef-4aaa-aca4-a3841d04cc5b	generic_instance	json_addl	5b761ebc-9320-4d44-9d95-c80a289eb8ae	FI7057	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp2f1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7057.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7057.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp2f1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 980260, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:18.918822-07	UPDATE	\N	\N	\N	f	f
269f15af-0c42-4027-ac0c-1fe5df9753d9	generic_instance	created_dt	5b761ebc-9320-4d44-9d95-c80a289eb8ae	FI7057	2025-07-03 07:46:18.804449-07	2025-07-03T07:46:18.804449-07:00	john@daylilyinformatics.com	2025-07-03 07:46:18.918822-07	UPDATE	\N	\N	\N	f	f
3427a2d5-f491-4ffd-93a4-887fcffccb63	generic_instance	modified_dt	5b761ebc-9320-4d44-9d95-c80a289eb8ae	FI7057	2025-07-03 07:46:18.816712-07	2025-07-03T07:46:18.918822-07:00	john@daylilyinformatics.com	2025-07-03 07:46:18.918822-07	UPDATE	\N	\N	\N	f	f
ff42845b-1d2d-455c-93f5-48be6acf03d3	generic_instance	\N	c260d8ec-016f-445e-977a-fb012e00481b	FI7058	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:19.561462-07	INSERT	\N	\N	\N	f	f
b9580887-0d18-4cdf-9fd9-fcfa72165934	generic_instance	json_addl	c260d8ec-016f-445e-977a-fb012e00481b	FI7058	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:19.57101-07	UPDATE	\N	\N	\N	f	f
01b85aee-f043-418a-a5a3-45f707cbbb73	generic_instance	created_dt	c260d8ec-016f-445e-977a-fb012e00481b	FI7058	2025-07-03 07:46:19.561462-07	2025-07-03T07:46:19.561462-07:00	john@daylilyinformatics.com	2025-07-03 07:46:19.57101-07	UPDATE	\N	\N	\N	f	f
8ed27598-80f7-407e-81cb-a4af5426f66c	generic_instance	modified_dt	c260d8ec-016f-445e-977a-fb012e00481b	FI7058	2025-07-03 07:46:19.561462-07	2025-07-03T07:46:19.57101-07:00	john@daylilyinformatics.com	2025-07-03 07:46:19.57101-07	UPDATE	\N	\N	\N	f	f
601b0f71-f5e0-490a-998d-70202bb9be9c	generic_instance	json_addl	c260d8ec-016f-445e-977a-fb012e00481b	FI7058	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/tbxas1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7058.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7058.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "tbxas1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1085371, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:19.679578-07	UPDATE	\N	\N	\N	f	f
35ce48a6-568b-4a29-9d6d-2d6add49e78b	generic_instance	created_dt	c260d8ec-016f-445e-977a-fb012e00481b	FI7058	2025-07-03 07:46:19.561462-07	2025-07-03T07:46:19.561462-07:00	john@daylilyinformatics.com	2025-07-03 07:46:19.679578-07	UPDATE	\N	\N	\N	f	f
c48c7a7e-4de2-48ea-8373-00592e8e31b6	generic_instance	modified_dt	c260d8ec-016f-445e-977a-fb012e00481b	FI7058	2025-07-03 07:46:19.57101-07	2025-07-03T07:46:19.679578-07:00	john@daylilyinformatics.com	2025-07-03 07:46:19.679578-07	UPDATE	\N	\N	\N	f	f
13f7856d-d459-4c22-9804-e015710667fe	generic_instance	\N	04b20693-2770-4741-ab5e-af1f0f85f28e	FI7059	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:20.290547-07	INSERT	\N	\N	\N	f	f
705ce4b0-64c9-43eb-bb5d-aa451a90bef2	generic_instance	json_addl	04b20693-2770-4741-ab5e-af1f0f85f28e	FI7059	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:20.299641-07	UPDATE	\N	\N	\N	f	f
16dfdfe7-66d0-4a94-96c2-2df036bce960	generic_instance	created_dt	04b20693-2770-4741-ab5e-af1f0f85f28e	FI7059	2025-07-03 07:46:20.290547-07	2025-07-03T07:46:20.290547-07:00	john@daylilyinformatics.com	2025-07-03 07:46:20.299641-07	UPDATE	\N	\N	\N	f	f
08591861-dd68-41f6-a606-a86047257c21	generic_instance	modified_dt	04b20693-2770-4741-ab5e-af1f0f85f28e	FI7059	2025-07-03 07:46:20.290547-07	2025-07-03T07:46:20.299641-07:00	john@daylilyinformatics.com	2025-07-03 07:46:20.299641-07	UPDATE	\N	\N	\N	f	f
56428779-c6bb-4923-945d-3259ea48156b	generic_instance	json_addl	04b20693-2770-4741-ab5e-af1f0f85f28e	FI7059	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/ugt2b17.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7059.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7059.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "ugt2b17.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 987018, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:20.409103-07	UPDATE	\N	\N	\N	f	f
deb8498a-4f29-4526-95a7-610a6d21b460	generic_instance	created_dt	04b20693-2770-4741-ab5e-af1f0f85f28e	FI7059	2025-07-03 07:46:20.290547-07	2025-07-03T07:46:20.290547-07:00	john@daylilyinformatics.com	2025-07-03 07:46:20.409103-07	UPDATE	\N	\N	\N	f	f
86bf1638-5d73-4114-95c9-4b288f82decf	generic_instance	modified_dt	04b20693-2770-4741-ab5e-af1f0f85f28e	FI7059	2025-07-03 07:46:20.299641-07	2025-07-03T07:46:20.409103-07:00	john@daylilyinformatics.com	2025-07-03 07:46:20.409103-07	UPDATE	\N	\N	\N	f	f
7da41925-159a-4b7b-be69-43f1b48cba60	generic_instance	\N	dc3b6ed6-0ec6-4cde-98e4-5ae861b129fe	FI7060	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:21.05117-07	INSERT	\N	\N	\N	f	f
1e3f25b4-8a60-4000-b7ae-8d96b019b0c3	generic_instance	json_addl	dc3b6ed6-0ec6-4cde-98e4-5ae861b129fe	FI7060	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:21.060725-07	UPDATE	\N	\N	\N	f	f
d44caeb9-e23a-4b46-bb41-25e2c25d0146	generic_instance	created_dt	dc3b6ed6-0ec6-4cde-98e4-5ae861b129fe	FI7060	2025-07-03 07:46:21.05117-07	2025-07-03T07:46:21.05117-07:00	john@daylilyinformatics.com	2025-07-03 07:46:21.060725-07	UPDATE	\N	\N	\N	f	f
fe8ede71-9d8e-416f-a190-f05ccd72a262	generic_instance	modified_dt	dc3b6ed6-0ec6-4cde-98e4-5ae861b129fe	FI7060	2025-07-03 07:46:21.05117-07	2025-07-03T07:46:21.060725-07:00	john@daylilyinformatics.com	2025-07-03 07:46:21.060725-07	UPDATE	\N	\N	\N	f	f
8e7857f9-1c5a-47f1-aa4b-208b20bbf0e0	generic_instance	json_addl	dc3b6ed6-0ec6-4cde-98e4-5ae861b129fe	FI7060	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/por.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7060.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7060.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "por.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 837876, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:21.168782-07	UPDATE	\N	\N	\N	f	f
6a297fb8-9c71-4bb5-8a01-3e1621ada760	generic_instance	created_dt	dc3b6ed6-0ec6-4cde-98e4-5ae861b129fe	FI7060	2025-07-03 07:46:21.05117-07	2025-07-03T07:46:21.05117-07:00	john@daylilyinformatics.com	2025-07-03 07:46:21.168782-07	UPDATE	\N	\N	\N	f	f
a94cbbe6-4066-4c93-bb3b-35fa9f300b42	generic_instance	modified_dt	dc3b6ed6-0ec6-4cde-98e4-5ae861b129fe	FI7060	2025-07-03 07:46:21.060725-07	2025-07-03T07:46:21.168782-07:00	john@daylilyinformatics.com	2025-07-03 07:46:21.168782-07	UPDATE	\N	\N	\N	f	f
f502fcf3-b242-464e-b0a4-b001cd8c5fde	generic_instance	\N	3b9591c8-f892-4b39-ba2f-95894da01bd1	FI7061	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:21.79047-07	INSERT	\N	\N	\N	f	f
e8ac784e-545b-4532-be5b-2b6d21f2fb76	generic_instance	json_addl	3b9591c8-f892-4b39-ba2f-95894da01bd1	FI7061	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:21.801093-07	UPDATE	\N	\N	\N	f	f
2b237335-f8e1-4f9d-8d8c-d3f19710c08b	generic_instance	created_dt	3b9591c8-f892-4b39-ba2f-95894da01bd1	FI7061	2025-07-03 07:46:21.79047-07	2025-07-03T07:46:21.79047-07:00	john@daylilyinformatics.com	2025-07-03 07:46:21.801093-07	UPDATE	\N	\N	\N	f	f
7c78ce2e-8b6f-4d7e-a24e-285b0b107271	generic_instance	modified_dt	3b9591c8-f892-4b39-ba2f-95894da01bd1	FI7061	2025-07-03 07:46:21.79047-07	2025-07-03T07:46:21.801093-07:00	john@daylilyinformatics.com	2025-07-03 07:46:21.801093-07	UPDATE	\N	\N	\N	f	f
e2f825e7-8e17-4a85-a3b2-ad9810271c71	generic_instance	json_addl	3b9591c8-f892-4b39-ba2f-95894da01bd1	FI7061	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp2c8.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7061.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7061.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp2c8.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 947576, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:21.90558-07	UPDATE	\N	\N	\N	f	f
f6b565cc-03ea-48dc-b5c8-7eee47e0a77d	generic_instance	created_dt	3b9591c8-f892-4b39-ba2f-95894da01bd1	FI7061	2025-07-03 07:46:21.79047-07	2025-07-03T07:46:21.79047-07:00	john@daylilyinformatics.com	2025-07-03 07:46:21.90558-07	UPDATE	\N	\N	\N	f	f
b356ae16-5619-497a-928d-76061379ed17	generic_instance	modified_dt	3b9591c8-f892-4b39-ba2f-95894da01bd1	FI7061	2025-07-03 07:46:21.801093-07	2025-07-03T07:46:21.90558-07:00	john@daylilyinformatics.com	2025-07-03 07:46:21.90558-07	UPDATE	\N	\N	\N	f	f
954d93a0-2f8b-428e-b7c7-34a8757812bd	generic_instance	\N	2eb9d931-5a00-44ea-99d0-b50cc6d630f1	FI7062	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:22.523954-07	INSERT	\N	\N	\N	f	f
884eb051-f7e0-4f15-b0b3-2436ed2ed8d3	generic_instance	json_addl	2eb9d931-5a00-44ea-99d0-b50cc6d630f1	FI7062	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:22.532576-07	UPDATE	\N	\N	\N	f	f
f5003a7e-569a-402c-b309-75eb3f7a4644	generic_instance	created_dt	2eb9d931-5a00-44ea-99d0-b50cc6d630f1	FI7062	2025-07-03 07:46:22.523954-07	2025-07-03T07:46:22.523954-07:00	john@daylilyinformatics.com	2025-07-03 07:46:22.532576-07	UPDATE	\N	\N	\N	f	f
c21051c2-8d21-4f3d-a863-46744f92937b	generic_instance	modified_dt	2eb9d931-5a00-44ea-99d0-b50cc6d630f1	FI7062	2025-07-03 07:46:22.523954-07	2025-07-03T07:46:22.532576-07:00	john@daylilyinformatics.com	2025-07-03 07:46:22.532576-07	UPDATE	\N	\N	\N	f	f
283374f7-8432-44cc-b23a-6594e198c07f	generic_instance	json_addl	2eb9d931-5a00-44ea-99d0-b50cc6d630f1	FI7062	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp2b6.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7062.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7062.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp2b6.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 992530, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:22.641055-07	UPDATE	\N	\N	\N	f	f
35f64e02-a994-4564-870b-9cbff4f13714	generic_instance	created_dt	2eb9d931-5a00-44ea-99d0-b50cc6d630f1	FI7062	2025-07-03 07:46:22.523954-07	2025-07-03T07:46:22.523954-07:00	john@daylilyinformatics.com	2025-07-03 07:46:22.641055-07	UPDATE	\N	\N	\N	f	f
ce515261-3386-4943-9b4f-e74d5283a857	generic_instance	modified_dt	2eb9d931-5a00-44ea-99d0-b50cc6d630f1	FI7062	2025-07-03 07:46:22.532576-07	2025-07-03T07:46:22.641055-07:00	john@daylilyinformatics.com	2025-07-03 07:46:22.641055-07	UPDATE	\N	\N	\N	f	f
090a3149-7b91-46d3-8bd8-fb3cf5a8cb6b	generic_instance	\N	4d7880f2-f6ba-498d-9d36-a065b7c26a62	FI7063	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:23.119651-07	INSERT	\N	\N	\N	f	f
9ffe5365-82f1-40ce-b14d-f75ed1c62c5f	generic_instance	json_addl	4d7880f2-f6ba-498d-9d36-a065b7c26a62	FI7063	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:23.130055-07	UPDATE	\N	\N	\N	f	f
78aa3e5c-f2a4-43a7-8adc-3151f13203f6	generic_instance_lineage	\N	72874abc-291e-4e36-89d2-13aa068cd7fb	GL73	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:19.562247-07	INSERT	\N	\N	\N	f	f
862753f9-7f4a-432b-ab3a-eaf89e15e43e	generic_instance_lineage	\N	f1409a26-5152-433b-a0dd-7a664bd9be6e	GL74	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:20.291238-07	INSERT	\N	\N	\N	f	f
5c56164e-76d5-4ce1-b35c-e45272c48cc4	generic_instance_lineage	\N	c2c9e2e4-8ed4-4737-85d4-fd75871c7f9c	GL75	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:21.051861-07	INSERT	\N	\N	\N	f	f
5259c318-5fdb-4d49-be31-928d9b5bc784	generic_instance_lineage	\N	efebb8ad-8a01-4e0e-afff-447eb897daed	GL76	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:21.79117-07	INSERT	\N	\N	\N	f	f
a06d18ca-f4c5-4425-b6b9-daba60df718a	generic_instance_lineage	\N	bb0894f4-e56c-4a77-a9cd-9881a32ad0c6	GL77	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:22.524617-07	INSERT	\N	\N	\N	f	f
3239d5f8-b33a-4baa-85e9-7b79bd4de51c	generic_instance_lineage	\N	87253686-7e6c-4c31-97af-b76cc05839e9	GL78	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:23.120485-07	INSERT	\N	\N	\N	f	f
856d7c7d-2516-4830-b138-8ee6866ffbea	generic_instance_lineage	\N	fea94b55-19ff-41ec-8dfc-c3c4fe2b28e1	GL79	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:23.857466-07	INSERT	\N	\N	\N	f	f
34556293-bac4-47a1-a963-c283677597be	generic_instance_lineage	\N	89861523-d00e-450e-9812-dce1eac3a79e	GL80	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:24.636402-07	INSERT	\N	\N	\N	f	f
f7dc1cc4-3c2a-4004-9975-7fe6beb742cc	generic_instance_lineage	\N	0f10a224-e406-4a00-a3c7-bc402d35fc65	GL81	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:25.359055-07	INSERT	\N	\N	\N	f	f
e92c9922-164a-43a0-90be-3a2b4bf7bc0e	generic_instance_lineage	\N	8bd2a248-5b53-47b5-ba43-d357e6bc7912	GL82	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:26.081292-07	INSERT	\N	\N	\N	f	f
879e91ec-f6db-471b-ab4f-78c3845f5431	generic_instance_lineage	\N	c7533e6c-8ade-493f-b39f-af4f02fb5023	GL83	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:26.847362-07	INSERT	\N	\N	\N	f	f
1c11a965-8957-4ea5-a161-577deb8d5565	generic_instance_lineage	\N	444ce4d6-7aa1-479c-8f4d-58c0458182fd	GL84	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:27.532904-07	INSERT	\N	\N	\N	f	f
653ad9f9-4a03-4ff2-87c3-ee535f6f0aba	generic_instance_lineage	\N	70d1e8e6-6b6c-4a63-9f2e-4dc40144a27a	GL85	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:28.296882-07	INSERT	\N	\N	\N	f	f
60da67a0-d27f-437f-9077-42a3599e0fdb	generic_instance_lineage	\N	1b99eca0-0f78-4139-80a1-9eb7ed70c6ad	GL86	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:29.045366-07	INSERT	\N	\N	\N	f	f
34aa2749-475a-485e-bd2f-cc71ae80f408	generic_instance_lineage	\N	acfe0859-e3b1-4ab1-8d62-76e44bdbf393	GL87	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:29.750202-07	INSERT	\N	\N	\N	f	f
68fda457-66f1-4d01-9487-3c9327cbaec5	generic_instance_lineage	\N	be8debd8-6d59-4539-a841-b09fcfce1cee	GL88	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:30.504104-07	INSERT	\N	\N	\N	f	f
3c82528b-9b2f-4c3b-8ba8-4bd5d8b51b32	generic_instance_lineage	\N	17757fe8-5d54-4261-83b0-2e79cc37403d	GL89	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:31.173946-07	INSERT	\N	\N	\N	f	f
bfc0a72b-0635-444d-9362-a6bbf07252b0	generic_instance_lineage	\N	4b9a07d8-6a0c-40f0-9018-5a56f314372a	GL90	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:31.760985-07	INSERT	\N	\N	\N	f	f
dcb5f509-c270-4c00-b3f5-094536bf50a8	generic_instance_lineage	\N	b3c10c39-4c5b-4bca-a819-9aa56d472845	GL91	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:32.282593-07	INSERT	\N	\N	\N	f	f
0d80c262-7b5d-40c8-91b4-d516ffd1e1bb	generic_instance_lineage	\N	9040ac20-4574-440d-aad4-75c340141fed	GL92	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:33.089104-07	INSERT	\N	\N	\N	f	f
5c946213-0dc9-4557-81ed-fd33da8364ad	generic_instance_lineage	\N	4732949c-99fb-4e5d-ae96-8ed0e3da8bd8	GL93	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:33.791329-07	INSERT	\N	\N	\N	f	f
2d503742-3b18-4cff-8d1b-5990629d37ca	generic_instance_lineage	\N	f2a2e709-b723-4798-b2c1-80e4e7aea581	GL94	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:34.387019-07	INSERT	\N	\N	\N	f	f
aa138a57-3930-44ee-abb3-90cb279cdc1e	generic_instance_lineage	\N	89caf14b-7c28-4103-a377-83fd4143a672	GL95	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:34.887954-07	INSERT	\N	\N	\N	f	f
89241b6f-cdb1-432a-b9a3-57d6cbae7a69	generic_instance_lineage	\N	66ec8afd-6983-493f-b354-fad5336a0812	GL96	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:35.526478-07	INSERT	\N	\N	\N	f	f
890682d6-3e4e-4d86-b01d-455062b7f390	generic_instance_lineage	\N	82122e94-57f1-4c55-b3fa-0040228c1ef0	GL97	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:36.091874-07	INSERT	\N	\N	\N	f	f
dd539266-94a4-48c1-9418-6c270a6f4161	generic_instance_lineage	\N	ded88850-1761-431c-80b7-bec5d2bd0a84	GL98	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:36.927029-07	INSERT	\N	\N	\N	f	f
7094a82c-d1ed-4b93-a31d-544222cee091	generic_instance_lineage	\N	3f2a57be-c5c3-4cda-b0b8-ebb3a0e78e0a	GL99	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:37.522901-07	INSERT	\N	\N	\N	f	f
6ad55a45-4962-4c35-93ae-ecab2dd46fb3	generic_instance_lineage	\N	b892a9eb-82cb-4adc-bfc7-ba08c1b4b394	GL100	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:38.146625-07	INSERT	\N	\N	\N	f	f
d60fc05a-7541-4a0a-825e-3df9fe52aa9c	generic_instance_lineage	\N	60162a63-a1bd-41e9-be90-3ac38f7dfda2	GL101	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:38.730951-07	INSERT	\N	\N	\N	f	f
e1c46c4e-16da-44f8-bf27-cbcf0889ebf5	generic_instance_lineage	\N	ae6f130a-4d48-4508-a244-b89d0d6b6598	GL102	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:39.29554-07	INSERT	\N	\N	\N	f	f
39877643-6f27-4511-8fbf-d5c6f125e5b8	generic_instance_lineage	\N	f64146e4-2eb2-4471-9b13-f2eae183b4f9	GL103	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:39.834112-07	INSERT	\N	\N	\N	f	f
ac509a4a-1739-4f90-b357-698cec448429	generic_instance_lineage	\N	42f05e96-4756-4763-8b6f-4269b34ac2bf	GL104	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:40.527301-07	INSERT	\N	\N	\N	f	f
c5f66742-f29b-4a52-826e-c7f16afdd59d	generic_instance_lineage	\N	f0b37d08-2344-4cb4-8726-d018de18327d	GL105	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:41.380446-07	INSERT	\N	\N	\N	f	f
faa552b2-cfc6-484c-a7d3-63c9a77cbdf9	generic_instance_lineage	\N	ab7edb2b-4f52-475e-b151-a597117af5ea	GL106	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:42.277334-07	INSERT	\N	\N	\N	f	f
00fc80b0-e57a-455b-9f8c-3795d77ecb2d	generic_instance_lineage	\N	195eee4f-bda2-4dfb-9830-adbad8b15f4a	GL107	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:43.173901-07	INSERT	\N	\N	\N	f	f
862501f1-25c5-4553-a1a0-cf171a9171a7	generic_instance_lineage	\N	af171de5-ad67-44f6-aa97-0f5ef910079a	GL108	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:44.031159-07	INSERT	\N	\N	\N	f	f
df0302e1-eb65-4c8f-8958-e4f94326bd88	generic_instance_lineage	\N	f550ad34-3355-4709-8674-2bf3bdc551a2	GL109	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:44.855903-07	INSERT	\N	\N	\N	f	f
f5efc4c2-b269-44b3-ac2a-925fdcf73e7d	generic_instance_lineage	\N	7bf69843-e848-4997-a763-9a10121c84b2	GL110	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:45.725978-07	INSERT	\N	\N	\N	f	f
e8f1ed0e-8da4-4bac-a684-14daa2abdf96	generic_instance_lineage	\N	3250fc4a-f90c-40b2-b636-33b7ca342d0d	GL111	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:46.598144-07	INSERT	\N	\N	\N	f	f
ca47dca3-e370-417b-97c2-f81309cd3d0c	generic_instance_lineage	\N	1ab9b618-bc2b-4c0f-a4f0-e027e7177231	GL112	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:47.516113-07	INSERT	\N	\N	\N	f	f
14da0627-957c-48dc-9b64-d40e1715a738	generic_instance_lineage	\N	8c920f6d-9a0a-47d5-9cc4-b7d8a71662ea	GL113	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:48.338904-07	INSERT	\N	\N	\N	f	f
8b0db163-ab16-4f53-b01a-bb218fb9a733	generic_instance_lineage	\N	fc3e8ad2-022c-4996-9404-56452b372fe5	GL114	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:49.187999-07	INSERT	\N	\N	\N	f	f
03916f4f-2a21-4deb-943d-23aa628f58cc	generic_instance_lineage	\N	a6f1a6b1-3a2d-4aa8-9385-193fb4f1b726	GL115	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:50.044671-07	INSERT	\N	\N	\N	f	f
166cdd48-e109-4c76-a569-0b9ec3b6b621	generic_instance_lineage	\N	c862aeba-ff72-4880-bb87-1f8f7df2721a	GL116	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:50.870117-07	INSERT	\N	\N	\N	f	f
af851f1c-df64-47fa-af73-72002e47e9f8	generic_instance_lineage	\N	5513d62f-f16f-446d-9b48-0e09ffc2761c	GL117	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:51.716479-07	INSERT	\N	\N	\N	f	f
97a1d5e2-b82b-473d-ad66-c01d3ea9b6a6	generic_instance_lineage	\N	8fa72706-c3b9-4c81-a9a7-c4764ba8d24f	GL118	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:52.478357-07	INSERT	\N	\N	\N	f	f
5a6405f6-d41d-4cae-93ed-dd8eb64080e1	generic_instance_lineage	\N	480f5b69-6f16-41c5-b6c7-e59d3e71b7f4	GL119	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:53.083873-07	INSERT	\N	\N	\N	f	f
2be76096-7c3c-4187-9687-7daff39672db	generic_instance_lineage	\N	a5ea33ef-0dba-4dab-a402-65c29efe4d4a	GL120	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:53.726108-07	INSERT	\N	\N	\N	f	f
a1bbc7ef-09c6-4180-9331-93e7c478363a	generic_instance_lineage	\N	98bbaad3-4087-4909-aac4-888bd138c712	GL121	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:55.436303-07	INSERT	\N	\N	\N	f	f
bb3108b4-f42b-42fd-b92c-77377caff2ea	generic_instance_lineage	\N	1088fa50-c5bd-4e91-b106-977c13134093	GL122	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:56.365609-07	INSERT	\N	\N	\N	f	f
335a5110-9393-49a1-9a04-745e5b6c596a	generic_instance_lineage	\N	74beaaff-cde2-41ba-8f07-4ab53a6ca1a3	GL123	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:57.21458-07	INSERT	\N	\N	\N	f	f
8f291d39-eb62-40a3-9a24-310f93a4056d	generic_instance_lineage	\N	eb262066-36a2-47f0-8401-dfcc895d3057	GL571	\N	\N	daylily	2025-07-03 08:23:17.848523-07	INSERT	\N	\N	\N	f	f
2aa90d12-d044-4101-9cb3-539840dd2830	generic_instance	created_dt	4d7880f2-f6ba-498d-9d36-a065b7c26a62	FI7063	2025-07-03 07:46:23.119651-07	2025-07-03T07:46:23.119651-07:00	john@daylilyinformatics.com	2025-07-03 07:46:23.130055-07	UPDATE	\N	\N	\N	f	f
39b9d034-67d1-40ec-a66e-397ec8d21dcf	generic_instance	modified_dt	4d7880f2-f6ba-498d-9d36-a065b7c26a62	FI7063	2025-07-03 07:46:23.119651-07	2025-07-03T07:46:23.130055-07:00	john@daylilyinformatics.com	2025-07-03 07:46:23.130055-07	UPDATE	\N	\N	\N	f	f
bb55b194-0a02-4210-9609-f250f5224276	generic_instance	json_addl	4d7880f2-f6ba-498d-9d36-a065b7c26a62	FI7063	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/ugt2b7.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7063.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7063.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "ugt2b7.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1288708, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:23.201282-07	UPDATE	\N	\N	\N	f	f
3f9763cb-19a6-4ab7-a55c-39f19b05b080	generic_instance	created_dt	4d7880f2-f6ba-498d-9d36-a065b7c26a62	FI7063	2025-07-03 07:46:23.119651-07	2025-07-03T07:46:23.119651-07:00	john@daylilyinformatics.com	2025-07-03 07:46:23.201282-07	UPDATE	\N	\N	\N	f	f
27fdd60f-b537-4139-8439-d521f9cf682e	generic_instance	modified_dt	4d7880f2-f6ba-498d-9d36-a065b7c26a62	FI7063	2025-07-03 07:46:23.130055-07	2025-07-03T07:46:23.201282-07:00	john@daylilyinformatics.com	2025-07-03 07:46:23.201282-07	UPDATE	\N	\N	\N	f	f
573c3a16-f2ed-4947-97fc-e67cde2e6a19	generic_instance	\N	6033d3d2-b9bb-4ec7-8afd-a90254f310aa	FI7064	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:23.856566-07	INSERT	\N	\N	\N	f	f
9e648636-371d-41d0-bcae-d3666c953c82	generic_instance	json_addl	6033d3d2-b9bb-4ec7-8afd-a90254f310aa	FI7064	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:23.869024-07	UPDATE	\N	\N	\N	f	f
78f5f581-3ec1-4b1f-9a1e-af7513e5416c	generic_instance	created_dt	6033d3d2-b9bb-4ec7-8afd-a90254f310aa	FI7064	2025-07-03 07:46:23.856566-07	2025-07-03T07:46:23.856566-07:00	john@daylilyinformatics.com	2025-07-03 07:46:23.869024-07	UPDATE	\N	\N	\N	f	f
0cb9d99d-3dda-4191-b5dd-de4bd0fb5fe9	generic_instance	modified_dt	6033d3d2-b9bb-4ec7-8afd-a90254f310aa	FI7064	2025-07-03 07:46:23.856566-07	2025-07-03T07:46:23.869024-07:00	john@daylilyinformatics.com	2025-07-03 07:46:23.869024-07	UPDATE	\N	\N	\N	f	f
c2ca4efe-cf56-4b15-98bf-24098e117767	generic_instance	json_addl	6033d3d2-b9bb-4ec7-8afd-a90254f310aa	FI7064	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp1a1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7064.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7064.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp1a1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 419961, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:24.018626-07	UPDATE	\N	\N	\N	f	f
de3f3d93-cfe9-45e5-9311-7f81950e65b3	generic_instance	created_dt	6033d3d2-b9bb-4ec7-8afd-a90254f310aa	FI7064	2025-07-03 07:46:23.856566-07	2025-07-03T07:46:23.856566-07:00	john@daylilyinformatics.com	2025-07-03 07:46:24.018626-07	UPDATE	\N	\N	\N	f	f
01b6785c-5c93-43c5-b64e-85866ab1d11d	generic_instance	modified_dt	6033d3d2-b9bb-4ec7-8afd-a90254f310aa	FI7064	2025-07-03 07:46:23.869024-07	2025-07-03T07:46:24.018626-07:00	john@daylilyinformatics.com	2025-07-03 07:46:24.018626-07	UPDATE	\N	\N	\N	f	f
715a0cdc-ff39-49a5-bc28-1c6033c2151d	generic_instance	\N	7340a717-a09a-4c3c-9cd6-0a5e52a77978	FI7065	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:24.63568-07	INSERT	\N	\N	\N	f	f
da25cb11-508f-4e0f-8788-632b54fa8cbb	generic_instance	json_addl	7340a717-a09a-4c3c-9cd6-0a5e52a77978	FI7065	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:24.644827-07	UPDATE	\N	\N	\N	f	f
aaad5633-eece-4289-9c6f-ce996cdee459	generic_instance	created_dt	7340a717-a09a-4c3c-9cd6-0a5e52a77978	FI7065	2025-07-03 07:46:24.63568-07	2025-07-03T07:46:24.63568-07:00	john@daylilyinformatics.com	2025-07-03 07:46:24.644827-07	UPDATE	\N	\N	\N	f	f
31b78f39-8414-47c1-b42b-1ff1068d5f8c	generic_instance	modified_dt	7340a717-a09a-4c3c-9cd6-0a5e52a77978	FI7065	2025-07-03 07:46:24.63568-07	2025-07-03T07:46:24.644827-07:00	john@daylilyinformatics.com	2025-07-03 07:46:24.644827-07	UPDATE	\N	\N	\N	f	f
029ae85f-795f-4251-84b0-13fba47e8563	generic_instance	json_addl	7340a717-a09a-4c3c-9cd6-0a5e52a77978	FI7065	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp2r1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7065.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7065.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp2r1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 473765, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:24.753214-07	UPDATE	\N	\N	\N	f	f
08b4e17c-b64c-4745-8dbd-9f0cbba4df60	generic_instance	created_dt	7340a717-a09a-4c3c-9cd6-0a5e52a77978	FI7065	2025-07-03 07:46:24.63568-07	2025-07-03T07:46:24.63568-07:00	john@daylilyinformatics.com	2025-07-03 07:46:24.753214-07	UPDATE	\N	\N	\N	f	f
383fcc48-1fff-44be-8ec5-49a9569dab23	generic_instance	modified_dt	7340a717-a09a-4c3c-9cd6-0a5e52a77978	FI7065	2025-07-03 07:46:24.644827-07	2025-07-03T07:46:24.753214-07:00	john@daylilyinformatics.com	2025-07-03 07:46:24.753214-07	UPDATE	\N	\N	\N	f	f
2f5ec35f-9ebe-465f-b01d-d7cdf27a9a2c	generic_instance	\N	611dcea7-1da6-46e9-b035-b9658cef5ab2	FI7066	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:25.35827-07	INSERT	\N	\N	\N	f	f
409ff13d-9dd5-4889-a475-f03bcc141f58	generic_instance	json_addl	611dcea7-1da6-46e9-b035-b9658cef5ab2	FI7066	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:25.367821-07	UPDATE	\N	\N	\N	f	f
4bd4bf46-8b4d-44d3-844e-35877e56bd29	generic_instance	created_dt	611dcea7-1da6-46e9-b035-b9658cef5ab2	FI7066	2025-07-03 07:46:25.35827-07	2025-07-03T07:46:25.35827-07:00	john@daylilyinformatics.com	2025-07-03 07:46:25.367821-07	UPDATE	\N	\N	\N	f	f
a6df1b77-22fb-4f04-8bf9-0328244bd5c7	generic_instance	modified_dt	611dcea7-1da6-46e9-b035-b9658cef5ab2	FI7066	2025-07-03 07:46:25.35827-07	2025-07-03T07:46:25.367821-07:00	john@daylilyinformatics.com	2025-07-03 07:46:25.367821-07	UPDATE	\N	\N	\N	f	f
1527f2a5-23c7-4455-a561-5291779427c5	generic_instance	json_addl	611dcea7-1da6-46e9-b035-b9658cef5ab2	FI7066	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/ugt2b15.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7066.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7066.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "ugt2b15.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1082490, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:25.476017-07	UPDATE	\N	\N	\N	f	f
b788fc07-2013-4245-ab73-eeb0fe4bd159	generic_instance	created_dt	611dcea7-1da6-46e9-b035-b9658cef5ab2	FI7066	2025-07-03 07:46:25.35827-07	2025-07-03T07:46:25.35827-07:00	john@daylilyinformatics.com	2025-07-03 07:46:25.476017-07	UPDATE	\N	\N	\N	f	f
0381a472-d2e5-4bb3-93db-1a39841c3f19	generic_instance	modified_dt	611dcea7-1da6-46e9-b035-b9658cef5ab2	FI7066	2025-07-03 07:46:25.367821-07	2025-07-03T07:46:25.476017-07:00	john@daylilyinformatics.com	2025-07-03 07:46:25.476017-07	UPDATE	\N	\N	\N	f	f
0e670ffc-6a0e-4cab-8d3c-d28ad1759398	generic_instance	\N	b007b95e-b1a6-42c6-bc19-24e1e5f779b9	FI7067	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:26.080622-07	INSERT	\N	\N	\N	f	f
fd129091-3d68-4f77-9e03-b3724d1b3dbd	generic_instance	json_addl	b007b95e-b1a6-42c6-bc19-24e1e5f779b9	FI7067	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:26.089371-07	UPDATE	\N	\N	\N	f	f
c1d2c813-8ff0-4fc6-be3a-c64de290a75d	generic_instance	created_dt	b007b95e-b1a6-42c6-bc19-24e1e5f779b9	FI7067	2025-07-03 07:46:26.080622-07	2025-07-03T07:46:26.080622-07:00	john@daylilyinformatics.com	2025-07-03 07:46:26.089371-07	UPDATE	\N	\N	\N	f	f
9b6a78d8-c59d-4ddd-a7dc-af1d52b1cc3d	generic_instance	modified_dt	b007b95e-b1a6-42c6-bc19-24e1e5f779b9	FI7067	2025-07-03 07:46:26.080622-07	2025-07-03T07:46:26.089371-07:00	john@daylilyinformatics.com	2025-07-03 07:46:26.089371-07	UPDATE	\N	\N	\N	f	f
642c6f3a-86e1-4431-b698-8f8cfc693e3a	generic_instance	json_addl	b007b95e-b1a6-42c6-bc19-24e1e5f779b9	FI7067	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp2j2.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7067.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7067.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp2j2.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 503177, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:26.192291-07	UPDATE	\N	\N	\N	f	f
77d26379-10b6-4b4c-ba71-4909d9c06342	generic_instance	created_dt	b007b95e-b1a6-42c6-bc19-24e1e5f779b9	FI7067	2025-07-03 07:46:26.080622-07	2025-07-03T07:46:26.080622-07:00	john@daylilyinformatics.com	2025-07-03 07:46:26.192291-07	UPDATE	\N	\N	\N	f	f
2cf637ec-2555-42b2-a073-257027268850	generic_instance	modified_dt	b007b95e-b1a6-42c6-bc19-24e1e5f779b9	FI7067	2025-07-03 07:46:26.089371-07	2025-07-03T07:46:26.192291-07:00	john@daylilyinformatics.com	2025-07-03 07:46:26.192291-07	UPDATE	\N	\N	\N	f	f
455326b5-42ca-470d-abf3-9d1519f73475	generic_instance	\N	2763560b-cabd-424f-a685-df92ca63080c	FI7068	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:26.846633-07	INSERT	\N	\N	\N	f	f
bf4b3ca3-f241-4866-9fe5-d29ebac98622	generic_instance	json_addl	2763560b-cabd-424f-a685-df92ca63080c	FI7068	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:26.855492-07	UPDATE	\N	\N	\N	f	f
9b72034f-3f24-453f-a823-a63d01364550	generic_instance	created_dt	2763560b-cabd-424f-a685-df92ca63080c	FI7068	2025-07-03 07:46:26.846633-07	2025-07-03T07:46:26.846633-07:00	john@daylilyinformatics.com	2025-07-03 07:46:26.855492-07	UPDATE	\N	\N	\N	f	f
88116c58-a464-4a1a-9e22-fed049624b2f	generic_instance	modified_dt	2763560b-cabd-424f-a685-df92ca63080c	FI7068	2025-07-03 07:46:26.846633-07	2025-07-03T07:46:26.855492-07:00	john@daylilyinformatics.com	2025-07-03 07:46:26.855492-07	UPDATE	\N	\N	\N	f	f
dc4a0b5c-3596-468a-b59f-147624450539	generic_instance	json_addl	2763560b-cabd-424f-a685-df92ca63080c	FI7068	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp2e1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7068.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7068.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp2e1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 899848, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:26.958402-07	UPDATE	\N	\N	\N	f	f
f298721a-879a-4336-9cbf-8f51ca136c55	generic_instance	created_dt	2763560b-cabd-424f-a685-df92ca63080c	FI7068	2025-07-03 07:46:26.846633-07	2025-07-03T07:46:26.846633-07:00	john@daylilyinformatics.com	2025-07-03 07:46:26.958402-07	UPDATE	\N	\N	\N	f	f
3588b31d-f1b3-4efd-8544-253aba2ce932	generic_instance	json_addl	9a54fbb8-c9e1-4d2e-8283-db8e5d7b84fc	FI7074	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/slco1b1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7074.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7074.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "slco1b1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1325694, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:31.281848-07	UPDATE	\N	\N	\N	f	f
aa919a9c-f84d-417b-9cc3-214a5d88eeae	generic_instance	modified_dt	2763560b-cabd-424f-a685-df92ca63080c	FI7068	2025-07-03 07:46:26.855492-07	2025-07-03T07:46:26.958402-07:00	john@daylilyinformatics.com	2025-07-03 07:46:26.958402-07	UPDATE	\N	\N	\N	f	f
ca74da86-f85e-4f59-a010-e499f342d3ef	generic_instance	\N	cd4284b9-8e2c-4ab9-94e5-2ed1630e323e	FI7069	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:27.532219-07	INSERT	\N	\N	\N	f	f
7af0ab28-75f2-4c41-9c5e-476ed0c4ccd1	generic_instance	json_addl	cd4284b9-8e2c-4ab9-94e5-2ed1630e323e	FI7069	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:27.541158-07	UPDATE	\N	\N	\N	f	f
7e6d7170-918d-49ec-95bb-9d237a08f91e	generic_instance	created_dt	cd4284b9-8e2c-4ab9-94e5-2ed1630e323e	FI7069	2025-07-03 07:46:27.532219-07	2025-07-03T07:46:27.532219-07:00	john@daylilyinformatics.com	2025-07-03 07:46:27.541158-07	UPDATE	\N	\N	\N	f	f
afe3d30d-114c-449c-ae36-72c4a66fca51	generic_instance	modified_dt	cd4284b9-8e2c-4ab9-94e5-2ed1630e323e	FI7069	2025-07-03 07:46:27.532219-07	2025-07-03T07:46:27.541158-07:00	john@daylilyinformatics.com	2025-07-03 07:46:27.541158-07	UPDATE	\N	\N	\N	f	f
b9e8e7cf-78f9-4add-aa04-5033660f4fed	generic_instance	json_addl	cd4284b9-8e2c-4ab9-94e5-2ed1630e323e	FI7069	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/gstp1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7069.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7069.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "gstp1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 685918, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:27.683211-07	UPDATE	\N	\N	\N	f	f
ae0377a7-17fd-47d4-828c-e10c3ef7aebf	generic_instance	created_dt	cd4284b9-8e2c-4ab9-94e5-2ed1630e323e	FI7069	2025-07-03 07:46:27.532219-07	2025-07-03T07:46:27.532219-07:00	john@daylilyinformatics.com	2025-07-03 07:46:27.683211-07	UPDATE	\N	\N	\N	f	f
12cdd4b0-8a12-439c-bb94-3b6dd412fe4e	generic_instance	modified_dt	cd4284b9-8e2c-4ab9-94e5-2ed1630e323e	FI7069	2025-07-03 07:46:27.541158-07	2025-07-03T07:46:27.683211-07:00	john@daylilyinformatics.com	2025-07-03 07:46:27.683211-07	UPDATE	\N	\N	\N	f	f
75bce288-aae1-4d48-b427-cbc72a688a49	generic_instance	\N	81948d72-dcd4-4211-819c-96e65e693d0e	FI7070	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:28.295525-07	INSERT	\N	\N	\N	f	f
d648359b-6f04-46c7-984d-079a8501cc7c	generic_instance	json_addl	81948d72-dcd4-4211-819c-96e65e693d0e	FI7070	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:28.307288-07	UPDATE	\N	\N	\N	f	f
d555cc6b-f258-40da-8072-0c67fce73a81	generic_instance	created_dt	81948d72-dcd4-4211-819c-96e65e693d0e	FI7070	2025-07-03 07:46:28.295525-07	2025-07-03T07:46:28.295525-07:00	john@daylilyinformatics.com	2025-07-03 07:46:28.307288-07	UPDATE	\N	\N	\N	f	f
ec766105-ce57-4d02-aec6-15d5c36ba94b	generic_instance	modified_dt	81948d72-dcd4-4211-819c-96e65e693d0e	FI7070	2025-07-03 07:46:28.295525-07	2025-07-03T07:46:28.307288-07:00	john@daylilyinformatics.com	2025-07-03 07:46:28.307288-07	UPDATE	\N	\N	\N	f	f
f4e0ba77-f7f7-4cc5-a327-cce5d5e9a67f	generic_instance	json_addl	81948d72-dcd4-4211-819c-96e65e693d0e	FI7070	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/nat2.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7070.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7070.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "nat2.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1160607, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:28.417356-07	UPDATE	\N	\N	\N	f	f
380d5050-4444-4585-ad1e-498ab582ba2d	generic_instance	created_dt	81948d72-dcd4-4211-819c-96e65e693d0e	FI7070	2025-07-03 07:46:28.295525-07	2025-07-03T07:46:28.295525-07:00	john@daylilyinformatics.com	2025-07-03 07:46:28.417356-07	UPDATE	\N	\N	\N	f	f
60f03df6-f598-40bc-a101-dad7f908dbdc	generic_instance	modified_dt	81948d72-dcd4-4211-819c-96e65e693d0e	FI7070	2025-07-03 07:46:28.307288-07	2025-07-03T07:46:28.417356-07:00	john@daylilyinformatics.com	2025-07-03 07:46:28.417356-07	UPDATE	\N	\N	\N	f	f
a7183ff8-c1f2-437b-8181-1f7c069b4b8b	generic_instance	\N	77115b00-ac73-4144-80b9-a0bd9b125164	FI7071	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:29.044606-07	INSERT	\N	\N	\N	f	f
b6d52c70-4229-4c8e-b21a-ddfbc8eba646	generic_instance	json_addl	77115b00-ac73-4144-80b9-a0bd9b125164	FI7071	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:29.05456-07	UPDATE	\N	\N	\N	f	f
6cc5b0d0-6d47-46b2-83b6-01742170e9ad	generic_instance	created_dt	77115b00-ac73-4144-80b9-a0bd9b125164	FI7071	2025-07-03 07:46:29.044606-07	2025-07-03T07:46:29.044606-07:00	john@daylilyinformatics.com	2025-07-03 07:46:29.05456-07	UPDATE	\N	\N	\N	f	f
3c8d3861-e086-4f65-bec5-db95b69eed8b	generic_instance	modified_dt	77115b00-ac73-4144-80b9-a0bd9b125164	FI7071	2025-07-03 07:46:29.044606-07	2025-07-03T07:46:29.05456-07:00	john@daylilyinformatics.com	2025-07-03 07:46:29.05456-07	UPDATE	\N	\N	\N	f	f
69c60c9c-93ec-48bd-84a1-37715872e4ad	generic_instance	json_addl	77115b00-ac73-4144-80b9-a0bd9b125164	FI7071	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp2a13.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7071.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7071.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp2a13.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 987555, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:29.159846-07	UPDATE	\N	\N	\N	f	f
1a952e9e-4add-482a-a2eb-91edd3d87523	generic_instance	created_dt	77115b00-ac73-4144-80b9-a0bd9b125164	FI7071	2025-07-03 07:46:29.044606-07	2025-07-03T07:46:29.044606-07:00	john@daylilyinformatics.com	2025-07-03 07:46:29.159846-07	UPDATE	\N	\N	\N	f	f
bf7e2d7d-85d7-4172-914e-d06166de9eb8	generic_instance	modified_dt	77115b00-ac73-4144-80b9-a0bd9b125164	FI7071	2025-07-03 07:46:29.05456-07	2025-07-03T07:46:29.159846-07:00	john@daylilyinformatics.com	2025-07-03 07:46:29.159846-07	UPDATE	\N	\N	\N	f	f
066e6491-d3aa-453e-adf7-7f1ab57458f2	generic_instance	\N	ee740bac-1bae-443e-8f54-c4d4d7bf371f	FI7072	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:29.749324-07	INSERT	\N	\N	\N	f	f
bbff1cbd-dc32-490e-a458-ab990691f16f	generic_instance	json_addl	ee740bac-1bae-443e-8f54-c4d4d7bf371f	FI7072	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:29.761403-07	UPDATE	\N	\N	\N	f	f
60841149-6637-4fa3-bd11-e40e98029578	generic_instance	created_dt	ee740bac-1bae-443e-8f54-c4d4d7bf371f	FI7072	2025-07-03 07:46:29.749324-07	2025-07-03T07:46:29.749324-07:00	john@daylilyinformatics.com	2025-07-03 07:46:29.761403-07	UPDATE	\N	\N	\N	f	f
97f77a7f-f0cb-4b83-900d-ec527463a2d3	generic_instance	modified_dt	ee740bac-1bae-443e-8f54-c4d4d7bf371f	FI7072	2025-07-03 07:46:29.749324-07	2025-07-03T07:46:29.761403-07:00	john@daylilyinformatics.com	2025-07-03 07:46:29.761403-07	UPDATE	\N	\N	\N	f	f
39a75736-ea88-458a-a5dd-44c86c111af7	generic_instance	json_addl	ee740bac-1bae-443e-8f54-c4d4d7bf371f	FI7072	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp2a6.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7072.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7072.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp2a6.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 882980, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:29.870826-07	UPDATE	\N	\N	\N	f	f
d55b8753-da49-4527-aa88-c8060082b452	generic_instance	created_dt	ee740bac-1bae-443e-8f54-c4d4d7bf371f	FI7072	2025-07-03 07:46:29.749324-07	2025-07-03T07:46:29.749324-07:00	john@daylilyinformatics.com	2025-07-03 07:46:29.870826-07	UPDATE	\N	\N	\N	f	f
588f596f-beb5-4efb-b0a2-0e10dde4bb7d	generic_instance	modified_dt	ee740bac-1bae-443e-8f54-c4d4d7bf371f	FI7072	2025-07-03 07:46:29.761403-07	2025-07-03T07:46:29.870826-07:00	john@daylilyinformatics.com	2025-07-03 07:46:29.870826-07	UPDATE	\N	\N	\N	f	f
511cacbf-cc3d-4761-90c3-c3214975e486	generic_instance	\N	cd142cc8-8540-40d1-8b89-0320a209d5f1	FI7073	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:30.503351-07	INSERT	\N	\N	\N	f	f
b6b7f4d5-b022-44aa-b4bd-60bfe4fa4fc0	generic_instance	json_addl	cd142cc8-8540-40d1-8b89-0320a209d5f1	FI7073	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:30.512606-07	UPDATE	\N	\N	\N	f	f
0e00963b-0703-4c9f-ace4-ba1a87a68819	generic_instance	created_dt	cd142cc8-8540-40d1-8b89-0320a209d5f1	FI7073	2025-07-03 07:46:30.503351-07	2025-07-03T07:46:30.503351-07:00	john@daylilyinformatics.com	2025-07-03 07:46:30.512606-07	UPDATE	\N	\N	\N	f	f
c934e263-ab87-4def-b841-0cc60dfb841b	generic_instance	modified_dt	cd142cc8-8540-40d1-8b89-0320a209d5f1	FI7073	2025-07-03 07:46:30.503351-07	2025-07-03T07:46:30.512606-07:00	john@daylilyinformatics.com	2025-07-03 07:46:30.512606-07	UPDATE	\N	\N	\N	f	f
3b12b0b2-aa7b-4526-9935-8570a4d185ba	generic_instance	json_addl	cd142cc8-8540-40d1-8b89-0320a209d5f1	FI7073	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/slc22a2.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7073.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7073.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "slc22a2.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1197684, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:30.617386-07	UPDATE	\N	\N	\N	f	f
87c46478-6458-4876-b17f-26703fd05f4a	generic_instance	created_dt	cd142cc8-8540-40d1-8b89-0320a209d5f1	FI7073	2025-07-03 07:46:30.503351-07	2025-07-03T07:46:30.503351-07:00	john@daylilyinformatics.com	2025-07-03 07:46:30.617386-07	UPDATE	\N	\N	\N	f	f
bbd8564c-65a4-4845-bd9f-32ac6f5a77eb	generic_instance	modified_dt	cd142cc8-8540-40d1-8b89-0320a209d5f1	FI7073	2025-07-03 07:46:30.512606-07	2025-07-03T07:46:30.617386-07:00	john@daylilyinformatics.com	2025-07-03 07:46:30.617386-07	UPDATE	\N	\N	\N	f	f
6eaeb5aa-debc-4ee2-ac97-889f9c17e56f	generic_instance	\N	9a54fbb8-c9e1-4d2e-8283-db8e5d7b84fc	FI7074	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:31.173251-07	INSERT	\N	\N	\N	f	f
e34ca41c-70d4-43d6-8bed-254149da3b35	generic_instance	json_addl	9a54fbb8-c9e1-4d2e-8283-db8e5d7b84fc	FI7074	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:31.181935-07	UPDATE	\N	\N	\N	f	f
37f1042d-4434-4a52-83c9-565462b93410	generic_instance	created_dt	9a54fbb8-c9e1-4d2e-8283-db8e5d7b84fc	FI7074	2025-07-03 07:46:31.173251-07	2025-07-03T07:46:31.173251-07:00	john@daylilyinformatics.com	2025-07-03 07:46:31.181935-07	UPDATE	\N	\N	\N	f	f
5a11f7c5-596e-4e83-a8d7-3b6ebe557eef	generic_instance	modified_dt	9a54fbb8-c9e1-4d2e-8283-db8e5d7b84fc	FI7074	2025-07-03 07:46:31.173251-07	2025-07-03T07:46:31.181935-07:00	john@daylilyinformatics.com	2025-07-03 07:46:31.181935-07	UPDATE	\N	\N	\N	f	f
4b656482-59bf-4d5e-a24c-861523222591	generic_instance	created_dt	9a54fbb8-c9e1-4d2e-8283-db8e5d7b84fc	FI7074	2025-07-03 07:46:31.173251-07	2025-07-03T07:46:31.173251-07:00	john@daylilyinformatics.com	2025-07-03 07:46:31.281848-07	UPDATE	\N	\N	\N	f	f
a8c25f06-dafd-4162-ad8c-11214cd6e69e	generic_instance	modified_dt	9a54fbb8-c9e1-4d2e-8283-db8e5d7b84fc	FI7074	2025-07-03 07:46:31.181935-07	2025-07-03T07:46:31.281848-07:00	john@daylilyinformatics.com	2025-07-03 07:46:31.281848-07	UPDATE	\N	\N	\N	f	f
1b693651-4d69-4883-9fba-845711140ae4	generic_instance	\N	16c52e6c-5995-449c-bd88-bd1500aab2c2	FI7075	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:31.760213-07	INSERT	\N	\N	\N	f	f
3a3b1f93-c6f8-47ba-abe6-fc4cadd50cd4	generic_instance	json_addl	16c52e6c-5995-449c-bd88-bd1500aab2c2	FI7075	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:31.770257-07	UPDATE	\N	\N	\N	f	f
e3894c4a-7b22-42de-bdda-bb3311cc468f	generic_instance	created_dt	16c52e6c-5995-449c-bd88-bd1500aab2c2	FI7075	2025-07-03 07:46:31.760213-07	2025-07-03T07:46:31.760213-07:00	john@daylilyinformatics.com	2025-07-03 07:46:31.770257-07	UPDATE	\N	\N	\N	f	f
83ec34fe-2a58-44db-b9d1-d975c170e606	generic_instance	modified_dt	16c52e6c-5995-449c-bd88-bd1500aab2c2	FI7075	2025-07-03 07:46:31.760213-07	2025-07-03T07:46:31.770257-07:00	john@daylilyinformatics.com	2025-07-03 07:46:31.770257-07	UPDATE	\N	\N	\N	f	f
2c4a2046-08c5-465e-8eb1-306523a2c9a4	generic_instance	json_addl	16c52e6c-5995-449c-bd88-bd1500aab2c2	FI7075	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/ifnl3.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7075.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7075.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "ifnl3.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 620168, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:31.843366-07	UPDATE	\N	\N	\N	f	f
d1cb07ec-8533-4007-a9f0-aefe0f37c934	generic_instance	created_dt	16c52e6c-5995-449c-bd88-bd1500aab2c2	FI7075	2025-07-03 07:46:31.760213-07	2025-07-03T07:46:31.760213-07:00	john@daylilyinformatics.com	2025-07-03 07:46:31.843366-07	UPDATE	\N	\N	\N	f	f
437cae3f-787c-4e3a-99ac-1bcfe098e976	generic_instance	modified_dt	16c52e6c-5995-449c-bd88-bd1500aab2c2	FI7075	2025-07-03 07:46:31.770257-07	2025-07-03T07:46:31.843366-07:00	john@daylilyinformatics.com	2025-07-03 07:46:31.843366-07	UPDATE	\N	\N	\N	f	f
19095956-7ac7-4f4a-a073-52ea28f7d397	generic_instance	\N	ac534fb1-c4b3-4163-a549-9d6c18fad16f	FI7076	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:32.281743-07	INSERT	\N	\N	\N	f	f
96414a17-f9b9-4e9f-b508-dc3ee8bb7a40	generic_instance	json_addl	ac534fb1-c4b3-4163-a549-9d6c18fad16f	FI7076	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:32.292406-07	UPDATE	\N	\N	\N	f	f
4923d3e7-a3e2-4cf0-8344-5965937b6230	generic_instance	created_dt	ac534fb1-c4b3-4163-a549-9d6c18fad16f	FI7076	2025-07-03 07:46:32.281743-07	2025-07-03T07:46:32.281743-07:00	john@daylilyinformatics.com	2025-07-03 07:46:32.292406-07	UPDATE	\N	\N	\N	f	f
9d69e255-aa7f-42f5-aed0-8c6d77d1499a	generic_instance	modified_dt	ac534fb1-c4b3-4163-a549-9d6c18fad16f	FI7076	2025-07-03 07:46:32.281743-07	2025-07-03T07:46:32.292406-07:00	john@daylilyinformatics.com	2025-07-03 07:46:32.292406-07	UPDATE	\N	\N	\N	f	f
963929f2-966a-494d-b681-ea1951528636	generic_instance	json_addl	ac534fb1-c4b3-4163-a549-9d6c18fad16f	FI7076	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/slco1b3.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7076.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7076.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "slco1b3.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1703270, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:32.362033-07	UPDATE	\N	\N	\N	f	f
6cc0d880-d041-4b96-96ab-c7be0bf1963b	generic_instance	created_dt	ac534fb1-c4b3-4163-a549-9d6c18fad16f	FI7076	2025-07-03 07:46:32.281743-07	2025-07-03T07:46:32.281743-07:00	john@daylilyinformatics.com	2025-07-03 07:46:32.362033-07	UPDATE	\N	\N	\N	f	f
fdf45a7c-2b23-4cec-800b-87b9f641c86e	generic_instance	modified_dt	ac534fb1-c4b3-4163-a549-9d6c18fad16f	FI7076	2025-07-03 07:46:32.292406-07	2025-07-03T07:46:32.362033-07:00	john@daylilyinformatics.com	2025-07-03 07:46:32.362033-07	UPDATE	\N	\N	\N	f	f
6ec8b267-e4ff-4deb-90fc-c19db73eb682	generic_instance	\N	9ab24170-9de3-4cf6-929e-9389976f8a03	FI7077	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:33.088279-07	INSERT	\N	\N	\N	f	f
3dd4d3c2-0163-46b5-90b4-b9e85f22974b	generic_instance	json_addl	9ab24170-9de3-4cf6-929e-9389976f8a03	FI7077	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:33.097063-07	UPDATE	\N	\N	\N	f	f
e3cac74c-d864-46e8-9700-16f5f32d8516	generic_instance	created_dt	9ab24170-9de3-4cf6-929e-9389976f8a03	FI7077	2025-07-03 07:46:33.088279-07	2025-07-03T07:46:33.088279-07:00	john@daylilyinformatics.com	2025-07-03 07:46:33.097063-07	UPDATE	\N	\N	\N	f	f
07532855-c25b-4be9-99a9-f294c98bbd51	generic_instance	modified_dt	9ab24170-9de3-4cf6-929e-9389976f8a03	FI7077	2025-07-03 07:46:33.088279-07	2025-07-03T07:46:33.097063-07:00	john@daylilyinformatics.com	2025-07-03 07:46:33.097063-07	UPDATE	\N	\N	\N	f	f
a224c2f4-45d2-4827-9aeb-a9ea9e797162	generic_instance	json_addl	9ab24170-9de3-4cf6-929e-9389976f8a03	FI7077	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/ugt1a4.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7077.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7077.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "ugt1a4.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1012518, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:33.226272-07	UPDATE	\N	\N	\N	f	f
d2781428-6132-4946-8f04-bbd80208342e	generic_instance	created_dt	9ab24170-9de3-4cf6-929e-9389976f8a03	FI7077	2025-07-03 07:46:33.088279-07	2025-07-03T07:46:33.088279-07:00	john@daylilyinformatics.com	2025-07-03 07:46:33.226272-07	UPDATE	\N	\N	\N	f	f
3166ba62-424c-4b24-ad3f-015bba4499d3	generic_instance	modified_dt	9ab24170-9de3-4cf6-929e-9389976f8a03	FI7077	2025-07-03 07:46:33.097063-07	2025-07-03T07:46:33.226272-07:00	john@daylilyinformatics.com	2025-07-03 07:46:33.226272-07	UPDATE	\N	\N	\N	f	f
bfbd4319-6624-4157-97f1-2bedb7cefe50	generic_instance	\N	5e52c4b2-b976-433e-ba92-2aabefb9585e	FI7078	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:33.790585-07	INSERT	\N	\N	\N	f	f
72cfbe8e-3d4d-4895-92d3-57e6d57d7485	generic_instance	json_addl	5e52c4b2-b976-433e-ba92-2aabefb9585e	FI7078	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:33.800889-07	UPDATE	\N	\N	\N	f	f
7f1ad704-2e2b-4fb8-8a93-16a3ffdb843a	generic_instance	created_dt	5e52c4b2-b976-433e-ba92-2aabefb9585e	FI7078	2025-07-03 07:46:33.790585-07	2025-07-03T07:46:33.790585-07:00	john@daylilyinformatics.com	2025-07-03 07:46:33.800889-07	UPDATE	\N	\N	\N	f	f
13a48e37-9f39-42da-ab6b-65bf31104b44	generic_instance	modified_dt	5e52c4b2-b976-433e-ba92-2aabefb9585e	FI7078	2025-07-03 07:46:33.790585-07	2025-07-03T07:46:33.800889-07:00	john@daylilyinformatics.com	2025-07-03 07:46:33.800889-07	UPDATE	\N	\N	\N	f	f
b5168fd1-4b0a-4c83-9cb2-d9e0eb5e5ff8	generic_instance	json_addl	5e52c4b2-b976-433e-ba92-2aabefb9585e	FI7078	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cftr.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7078.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7078.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cftr.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 738060, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:33.899393-07	UPDATE	\N	\N	\N	f	f
9d53dfad-55c1-4f2c-ab71-7400b57ec719	generic_instance	created_dt	5e52c4b2-b976-433e-ba92-2aabefb9585e	FI7078	2025-07-03 07:46:33.790585-07	2025-07-03T07:46:33.790585-07:00	john@daylilyinformatics.com	2025-07-03 07:46:33.899393-07	UPDATE	\N	\N	\N	f	f
e24801d4-f2a1-4b66-b2e8-fd66e839a4b9	generic_instance	modified_dt	5e52c4b2-b976-433e-ba92-2aabefb9585e	FI7078	2025-07-03 07:46:33.800889-07	2025-07-03T07:46:33.899393-07:00	john@daylilyinformatics.com	2025-07-03 07:46:33.899393-07	UPDATE	\N	\N	\N	f	f
420ddff7-cdad-4583-9c03-9e7f8cf051cb	generic_instance	\N	e391ed08-8608-44ff-abc0-16fe6c7baf3d	FI7079	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:34.386032-07	INSERT	\N	\N	\N	f	f
4437aa89-99a7-441f-8448-eaabc979a17a	generic_instance	json_addl	e391ed08-8608-44ff-abc0-16fe6c7baf3d	FI7079	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:34.397587-07	UPDATE	\N	\N	\N	f	f
e8f56ab2-9a57-4f48-8bd9-dca896595d13	generic_instance	created_dt	e391ed08-8608-44ff-abc0-16fe6c7baf3d	FI7079	2025-07-03 07:46:34.386032-07	2025-07-03T07:46:34.386032-07:00	john@daylilyinformatics.com	2025-07-03 07:46:34.397587-07	UPDATE	\N	\N	\N	f	f
d1415281-22db-4be1-8853-175a544455d2	generic_instance	modified_dt	e391ed08-8608-44ff-abc0-16fe6c7baf3d	FI7079	2025-07-03 07:46:34.386032-07	2025-07-03T07:46:34.397587-07:00	john@daylilyinformatics.com	2025-07-03 07:46:34.397587-07	UPDATE	\N	\N	\N	f	f
f7b6abb2-3e65-4c29-8617-c818e511dd37	generic_instance	json_addl	e391ed08-8608-44ff-abc0-16fe6c7baf3d	FI7079	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp3a4.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7079.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7079.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp3a4.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 457772, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:34.468559-07	UPDATE	\N	\N	\N	f	f
026e31db-3408-4540-ab48-8bf325cad1c2	generic_instance	created_dt	e391ed08-8608-44ff-abc0-16fe6c7baf3d	FI7079	2025-07-03 07:46:34.386032-07	2025-07-03T07:46:34.386032-07:00	john@daylilyinformatics.com	2025-07-03 07:46:34.468559-07	UPDATE	\N	\N	\N	f	f
6521beab-0a62-4d74-a448-23db9cbf5468	generic_instance	modified_dt	e391ed08-8608-44ff-abc0-16fe6c7baf3d	FI7079	2025-07-03 07:46:34.397587-07	2025-07-03T07:46:34.468559-07:00	john@daylilyinformatics.com	2025-07-03 07:46:34.468559-07	UPDATE	\N	\N	\N	f	f
e199b186-c9c7-4ec8-a312-7c328adca540	generic_instance	\N	bbf86074-7a49-44d5-a85a-36cd8bbf2da6	FI7080	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:34.887222-07	INSERT	\N	\N	\N	f	f
d6c4b94c-25fb-4fe5-b63a-907d3078c766	generic_instance	json_addl	bbf86074-7a49-44d5-a85a-36cd8bbf2da6	FI7080	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:34.897471-07	UPDATE	\N	\N	\N	f	f
916f2bc6-bbb7-4dc0-a858-3359d68cff39	generic_instance	created_dt	bbf86074-7a49-44d5-a85a-36cd8bbf2da6	FI7080	2025-07-03 07:46:34.887222-07	2025-07-03T07:46:34.887222-07:00	john@daylilyinformatics.com	2025-07-03 07:46:34.897471-07	UPDATE	\N	\N	\N	f	f
168e9a08-6228-4698-8c80-fbfa9af4d300	generic_instance	modified_dt	bbf86074-7a49-44d5-a85a-36cd8bbf2da6	FI7080	2025-07-03 07:46:34.887222-07	2025-07-03T07:46:34.897471-07:00	john@daylilyinformatics.com	2025-07-03 07:46:34.897471-07	UPDATE	\N	\N	\N	f	f
50e48a99-de36-4883-8710-d83332879e67	generic_instance	json_addl	bbf86074-7a49-44d5-a85a-36cd8bbf2da6	FI7080	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/vkorc1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7080.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7080.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "vkorc1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 398598, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:34.971248-07	UPDATE	\N	\N	\N	f	f
42736dcd-00b9-4ce7-841b-858e423bc35d	generic_instance	created_dt	bbf86074-7a49-44d5-a85a-36cd8bbf2da6	FI7080	2025-07-03 07:46:34.887222-07	2025-07-03T07:46:34.887222-07:00	john@daylilyinformatics.com	2025-07-03 07:46:34.971248-07	UPDATE	\N	\N	\N	f	f
2eea6dd9-4988-41a0-9f4f-0d206d0bd15d	generic_instance	modified_dt	bbf86074-7a49-44d5-a85a-36cd8bbf2da6	FI7080	2025-07-03 07:46:34.897471-07	2025-07-03T07:46:34.971248-07:00	john@daylilyinformatics.com	2025-07-03 07:46:34.971248-07	UPDATE	\N	\N	\N	f	f
bc322c39-0275-41cc-9ba0-78bbaf29822f	generic_instance	\N	6968dcc3-4411-486e-a424-89dc7a8a9875	FI7081	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:35.525532-07	INSERT	\N	\N	\N	f	f
ebb34a25-fab9-4272-8826-3ce38d836256	generic_instance	json_addl	6968dcc3-4411-486e-a424-89dc7a8a9875	FI7081	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:35.536557-07	UPDATE	\N	\N	\N	f	f
b506f5e8-3f3b-4a41-a12f-34cffe83a929	generic_instance	created_dt	6968dcc3-4411-486e-a424-89dc7a8a9875	FI7081	2025-07-03 07:46:35.525532-07	2025-07-03T07:46:35.525532-07:00	john@daylilyinformatics.com	2025-07-03 07:46:35.536557-07	UPDATE	\N	\N	\N	f	f
ed7df46a-d631-4440-8166-a5d4d02ca974	generic_instance	modified_dt	6968dcc3-4411-486e-a424-89dc7a8a9875	FI7081	2025-07-03 07:46:35.525532-07	2025-07-03T07:46:35.536557-07:00	john@daylilyinformatics.com	2025-07-03 07:46:35.536557-07	UPDATE	\N	\N	\N	f	f
61da4280-c928-4e8a-beba-75d9fba640a6	generic_instance	json_addl	6968dcc3-4411-486e-a424-89dc7a8a9875	FI7081	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/gstm1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7081.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7081.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "gstm1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 644122, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:35.615832-07	UPDATE	\N	\N	\N	f	f
afd013d0-2cef-4915-a3d3-f457daf60c60	generic_instance	created_dt	6968dcc3-4411-486e-a424-89dc7a8a9875	FI7081	2025-07-03 07:46:35.525532-07	2025-07-03T07:46:35.525532-07:00	john@daylilyinformatics.com	2025-07-03 07:46:35.615832-07	UPDATE	\N	\N	\N	f	f
ea7caf42-b0c7-4210-81e4-ffe0574e78ef	generic_instance	modified_dt	6968dcc3-4411-486e-a424-89dc7a8a9875	FI7081	2025-07-03 07:46:35.536557-07	2025-07-03T07:46:35.615832-07:00	john@daylilyinformatics.com	2025-07-03 07:46:35.615832-07	UPDATE	\N	\N	\N	f	f
1267659f-8904-41a1-b93b-83b84ad9534b	generic_instance	\N	f0f17b46-d607-4117-b350-68dd1beb7c59	FI7082	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:36.091124-07	INSERT	\N	\N	\N	f	f
782fc413-5a46-4af4-8a71-710aa513c58f	generic_instance	json_addl	f0f17b46-d607-4117-b350-68dd1beb7c59	FI7082	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:36.101101-07	UPDATE	\N	\N	\N	f	f
69fca1b8-6888-45e9-854a-9213590bd3c7	generic_instance	created_dt	f0f17b46-d607-4117-b350-68dd1beb7c59	FI7082	2025-07-03 07:46:36.091124-07	2025-07-03T07:46:36.091124-07:00	john@daylilyinformatics.com	2025-07-03 07:46:36.101101-07	UPDATE	\N	\N	\N	f	f
6ed6cb22-b5e9-4143-924b-49baf773e2a1	generic_instance	modified_dt	f0f17b46-d607-4117-b350-68dd1beb7c59	FI7082	2025-07-03 07:46:36.091124-07	2025-07-03T07:46:36.101101-07:00	john@daylilyinformatics.com	2025-07-03 07:46:36.101101-07	UPDATE	\N	\N	\N	f	f
b68e526d-e802-4041-b2a5-5d51ade8fdfe	generic_instance	json_addl	f0f17b46-d607-4117-b350-68dd1beb7c59	FI7082	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/dpyd.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7082.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7082.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "dpyd.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 2534931, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:36.225606-07	UPDATE	\N	\N	\N	f	f
3f3386c6-87f5-4bd4-bcb5-a27d344357ef	generic_instance	created_dt	f0f17b46-d607-4117-b350-68dd1beb7c59	FI7082	2025-07-03 07:46:36.091124-07	2025-07-03T07:46:36.091124-07:00	john@daylilyinformatics.com	2025-07-03 07:46:36.225606-07	UPDATE	\N	\N	\N	f	f
407a9197-69f5-45b0-a070-1a3a9aedd8c7	generic_instance	modified_dt	f0f17b46-d607-4117-b350-68dd1beb7c59	FI7082	2025-07-03 07:46:36.101101-07	2025-07-03T07:46:36.225606-07:00	john@daylilyinformatics.com	2025-07-03 07:46:36.225606-07	UPDATE	\N	\N	\N	f	f
2c069a47-5799-4bc0-9f63-2894a35eb009	generic_instance	\N	628aaab0-bc64-42ae-9199-121c2cfae0be	FI7083	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:36.926256-07	INSERT	\N	\N	\N	f	f
befe0a86-3416-45b9-a348-c4bf78ca2b3f	generic_instance	json_addl	628aaab0-bc64-42ae-9199-121c2cfae0be	FI7083	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:36.935703-07	UPDATE	\N	\N	\N	f	f
01cf34ce-31b8-4a75-9675-b8208fa6d8ae	generic_instance	created_dt	628aaab0-bc64-42ae-9199-121c2cfae0be	FI7083	2025-07-03 07:46:36.926256-07	2025-07-03T07:46:36.926256-07:00	john@daylilyinformatics.com	2025-07-03 07:46:36.935703-07	UPDATE	\N	\N	\N	f	f
e2ea8787-811e-43f6-9f62-651aaa4e04b5	generic_instance	modified_dt	628aaab0-bc64-42ae-9199-121c2cfae0be	FI7083	2025-07-03 07:46:36.926256-07	2025-07-03T07:46:36.935703-07:00	john@daylilyinformatics.com	2025-07-03 07:46:36.935703-07	UPDATE	\N	\N	\N	f	f
2b351e75-ad48-4241-a45e-4c146c989fa7	generic_instance	json_addl	628aaab0-bc64-42ae-9199-121c2cfae0be	FI7083	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp1b1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7083.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7083.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp1b1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 935399, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:37.02709-07	UPDATE	\N	\N	\N	f	f
123c9d55-d8b9-4ef9-bd83-416da45fc2ba	generic_instance	created_dt	628aaab0-bc64-42ae-9199-121c2cfae0be	FI7083	2025-07-03 07:46:36.926256-07	2025-07-03T07:46:36.926256-07:00	john@daylilyinformatics.com	2025-07-03 07:46:37.02709-07	UPDATE	\N	\N	\N	f	f
af88922e-9b04-4daa-8586-c38179862783	generic_instance	modified_dt	628aaab0-bc64-42ae-9199-121c2cfae0be	FI7083	2025-07-03 07:46:36.935703-07	2025-07-03T07:46:37.02709-07:00	john@daylilyinformatics.com	2025-07-03 07:46:37.02709-07	UPDATE	\N	\N	\N	f	f
4fe01ccd-3056-439d-b2fb-90b6f8904ed6	generic_instance	\N	d730ed0a-3439-475d-a55e-031e0f4fac5c	FI7084	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:37.522065-07	INSERT	\N	\N	\N	f	f
139b6b7c-b148-4177-8ff6-12a0370b83c9	generic_instance	json_addl	d730ed0a-3439-475d-a55e-031e0f4fac5c	FI7084	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:37.532397-07	UPDATE	\N	\N	\N	f	f
4f75290f-4072-4745-a4db-c73af4324d87	generic_instance	created_dt	d730ed0a-3439-475d-a55e-031e0f4fac5c	FI7084	2025-07-03 07:46:37.522065-07	2025-07-03T07:46:37.522065-07:00	john@daylilyinformatics.com	2025-07-03 07:46:37.532397-07	UPDATE	\N	\N	\N	f	f
eda9369f-77a5-48fb-b469-906ff45fbf02	generic_instance	modified_dt	d730ed0a-3439-475d-a55e-031e0f4fac5c	FI7084	2025-07-03 07:46:37.522065-07	2025-07-03T07:46:37.532397-07:00	john@daylilyinformatics.com	2025-07-03 07:46:37.532397-07	UPDATE	\N	\N	\N	f	f
033cc3c3-9d9f-4b7c-bfdd-00717acc765f	generic_instance	json_addl	d730ed0a-3439-475d-a55e-031e0f4fac5c	FI7084	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp1a2.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7084.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7084.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp1a2.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 445280, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:37.679385-07	UPDATE	\N	\N	\N	f	f
6d8613be-c589-4dad-a10f-4fe12568db29	generic_instance	created_dt	d730ed0a-3439-475d-a55e-031e0f4fac5c	FI7084	2025-07-03 07:46:37.522065-07	2025-07-03T07:46:37.522065-07:00	john@daylilyinformatics.com	2025-07-03 07:46:37.679385-07	UPDATE	\N	\N	\N	f	f
6f2deef7-f140-4393-b251-0a08fd9ecb25	generic_instance	modified_dt	d730ed0a-3439-475d-a55e-031e0f4fac5c	FI7084	2025-07-03 07:46:37.532397-07	2025-07-03T07:46:37.679385-07:00	john@daylilyinformatics.com	2025-07-03 07:46:37.679385-07	UPDATE	\N	\N	\N	f	f
72d18361-7e53-4bba-aa5a-fcc9b09dd218	generic_instance	\N	b1f51d3d-d2a3-48c0-9487-bc386b4a658d	FI7085	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:38.14582-07	INSERT	\N	\N	\N	f	f
a5cc3fe6-5c60-4c02-81c5-2e073e0f58c8	generic_instance	json_addl	b1f51d3d-d2a3-48c0-9487-bc386b4a658d	FI7085	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:38.155428-07	UPDATE	\N	\N	\N	f	f
b174a026-91f1-4961-88ee-af2e16ce4e3a	generic_instance	created_dt	b1f51d3d-d2a3-48c0-9487-bc386b4a658d	FI7085	2025-07-03 07:46:38.14582-07	2025-07-03T07:46:38.14582-07:00	john@daylilyinformatics.com	2025-07-03 07:46:38.155428-07	UPDATE	\N	\N	\N	f	f
c4007f63-757a-4e94-b78a-c93e0c14ae39	generic_instance	modified_dt	b1f51d3d-d2a3-48c0-9487-bc386b4a658d	FI7085	2025-07-03 07:46:38.14582-07	2025-07-03T07:46:38.155428-07:00	john@daylilyinformatics.com	2025-07-03 07:46:38.155428-07	UPDATE	\N	\N	\N	f	f
f865bf4c-655f-405a-b7aa-e1d6e0141cdf	generic_instance	json_addl	b1f51d3d-d2a3-48c0-9487-bc386b4a658d	FI7085	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp2c9.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7085.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7085.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp2c9.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 938026, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:38.232699-07	UPDATE	\N	\N	\N	f	f
40225720-9091-4e95-a369-f0e800a04595	generic_instance	created_dt	b1f51d3d-d2a3-48c0-9487-bc386b4a658d	FI7085	2025-07-03 07:46:38.14582-07	2025-07-03T07:46:38.14582-07:00	john@daylilyinformatics.com	2025-07-03 07:46:38.232699-07	UPDATE	\N	\N	\N	f	f
423202f1-1d78-472b-98d8-8c5376852c0a	generic_instance	modified_dt	b1f51d3d-d2a3-48c0-9487-bc386b4a658d	FI7085	2025-07-03 07:46:38.155428-07	2025-07-03T07:46:38.232699-07:00	john@daylilyinformatics.com	2025-07-03 07:46:38.232699-07	UPDATE	\N	\N	\N	f	f
82c56ab0-f985-4ffa-8be1-be8126a336b4	generic_instance	\N	652238e6-836f-4e53-ad34-062159021e26	FI7086	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:38.730285-07	INSERT	\N	\N	\N	f	f
1f939756-95a9-479f-a791-f4db7494c994	generic_instance	json_addl	652238e6-836f-4e53-ad34-062159021e26	FI7086	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:38.740084-07	UPDATE	\N	\N	\N	f	f
e29f78a1-4d96-43ac-aef8-8f1af6230058	generic_instance	created_dt	652238e6-836f-4e53-ad34-062159021e26	FI7086	2025-07-03 07:46:38.730285-07	2025-07-03T07:46:38.730285-07:00	john@daylilyinformatics.com	2025-07-03 07:46:38.740084-07	UPDATE	\N	\N	\N	f	f
6275d07d-7312-47d7-b7b4-5e1cfbaa5522	generic_instance	modified_dt	652238e6-836f-4e53-ad34-062159021e26	FI7086	2025-07-03 07:46:38.730285-07	2025-07-03T07:46:38.740084-07:00	john@daylilyinformatics.com	2025-07-03 07:46:38.740084-07	UPDATE	\N	\N	\N	f	f
fe35281e-888e-44fb-bb47-6e3fbce8916d	generic_instance	json_addl	652238e6-836f-4e53-ad34-062159021e26	FI7086	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/ugt1a1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7086.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7086.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "ugt1a1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 922059, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:38.814929-07	UPDATE	\N	\N	\N	f	f
5602c4a3-965e-4b4e-8d63-3c0c945198c5	generic_instance	created_dt	652238e6-836f-4e53-ad34-062159021e26	FI7086	2025-07-03 07:46:38.730285-07	2025-07-03T07:46:38.730285-07:00	john@daylilyinformatics.com	2025-07-03 07:46:38.814929-07	UPDATE	\N	\N	\N	f	f
2dc0c411-4137-40cd-a4f9-17fd6a1ad203	generic_instance	modified_dt	652238e6-836f-4e53-ad34-062159021e26	FI7086	2025-07-03 07:46:38.740084-07	2025-07-03T07:46:38.814929-07:00	john@daylilyinformatics.com	2025-07-03 07:46:38.814929-07	UPDATE	\N	\N	\N	f	f
b17539af-bed6-4246-a9db-61a6e14b4845	generic_instance	\N	a99ca08f-34f7-4f31-b30f-d765444bc221	FI7087	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:39.294807-07	INSERT	\N	\N	\N	f	f
b5088640-5af6-491e-9058-1cc731d1c517	generic_instance	json_addl	a99ca08f-34f7-4f31-b30f-d765444bc221	FI7087	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:39.305357-07	UPDATE	\N	\N	\N	f	f
29cd0419-a9dd-43c5-be5b-e40f6026f94d	generic_instance	created_dt	a99ca08f-34f7-4f31-b30f-d765444bc221	FI7087	2025-07-03 07:46:39.294807-07	2025-07-03T07:46:39.294807-07:00	john@daylilyinformatics.com	2025-07-03 07:46:39.305357-07	UPDATE	\N	\N	\N	f	f
c30b1ba7-7122-48ff-a155-7abafe486d59	generic_instance	modified_dt	a99ca08f-34f7-4f31-b30f-d765444bc221	FI7087	2025-07-03 07:46:39.294807-07	2025-07-03T07:46:39.305357-07:00	john@daylilyinformatics.com	2025-07-03 07:46:39.305357-07	UPDATE	\N	\N	\N	f	f
ceca606d-9a98-4334-80f7-8a49116d04bc	generic_instance	json_addl	a99ca08f-34f7-4f31-b30f-d765444bc221	FI7087	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/gstt1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7087.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7087.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "gstt1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 553675, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:39.391185-07	UPDATE	\N	\N	\N	f	f
1c87e9f6-7ff1-47a4-876a-d0c39fa297d4	generic_instance	created_dt	a99ca08f-34f7-4f31-b30f-d765444bc221	FI7087	2025-07-03 07:46:39.294807-07	2025-07-03T07:46:39.294807-07:00	john@daylilyinformatics.com	2025-07-03 07:46:39.391185-07	UPDATE	\N	\N	\N	f	f
5b6dd56f-e3b7-4f6d-87dd-ab6e72ca4304	generic_instance	modified_dt	a99ca08f-34f7-4f31-b30f-d765444bc221	FI7087	2025-07-03 07:46:39.305357-07	2025-07-03T07:46:39.391185-07:00	john@daylilyinformatics.com	2025-07-03 07:46:39.391185-07	UPDATE	\N	\N	\N	f	f
12cc8113-96e4-4c1c-921e-b54c8700b8b8	generic_instance	\N	6eeeffdd-a461-4bb2-ace0-e0f74637b47d	FI7088	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:39.833297-07	INSERT	\N	\N	\N	f	f
138dd0a3-10ef-44db-81a7-ccf685a8ee3a	generic_instance	json_addl	6eeeffdd-a461-4bb2-ace0-e0f74637b47d	FI7088	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:39.843973-07	UPDATE	\N	\N	\N	f	f
d4f28a20-480a-476f-857d-628da22783dc	generic_instance	created_dt	6eeeffdd-a461-4bb2-ace0-e0f74637b47d	FI7088	2025-07-03 07:46:39.833297-07	2025-07-03T07:46:39.833297-07:00	john@daylilyinformatics.com	2025-07-03 07:46:39.843973-07	UPDATE	\N	\N	\N	f	f
fb1651f8-76c8-4b10-9905-dbe502851b30	generic_instance	modified_dt	6eeeffdd-a461-4bb2-ace0-e0f74637b47d	FI7088	2025-07-03 07:46:39.833297-07	2025-07-03T07:46:39.843973-07:00	john@daylilyinformatics.com	2025-07-03 07:46:39.843973-07	UPDATE	\N	\N	\N	f	f
d2045495-24a8-471a-a96f-73825029631d	generic_instance	json_addl	6eeeffdd-a461-4bb2-ace0-e0f74637b47d	FI7088	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp2d6.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7088.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7088.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp2d6.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 682704, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:39.915264-07	UPDATE	\N	\N	\N	f	f
e57e33e4-2227-4d16-8065-97078c20073d	generic_instance	created_dt	6eeeffdd-a461-4bb2-ace0-e0f74637b47d	FI7088	2025-07-03 07:46:39.833297-07	2025-07-03T07:46:39.833297-07:00	john@daylilyinformatics.com	2025-07-03 07:46:39.915264-07	UPDATE	\N	\N	\N	f	f
1d3d3c54-e1f4-44d5-9bb5-028932cf5cb2	generic_instance	modified_dt	6eeeffdd-a461-4bb2-ace0-e0f74637b47d	FI7088	2025-07-03 07:46:39.843973-07	2025-07-03T07:46:39.915264-07:00	john@daylilyinformatics.com	2025-07-03 07:46:39.915264-07	UPDATE	\N	\N	\N	f	f
26b97604-30c3-4a19-b98b-aea8b171e004	generic_instance	\N	1f8fc77a-8554-4716-8f6f-bd177139f82f	FI7089	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:40.525269-07	INSERT	\N	\N	\N	f	f
3f1fe61a-61aa-4c8c-91d7-5655a0658cc4	generic_instance	json_addl	1f8fc77a-8554-4716-8f6f-bd177139f82f	FI7089	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:40.547039-07	UPDATE	\N	\N	\N	f	f
dcf2b33f-d8a1-46ee-84c7-817b0659b198	generic_instance	created_dt	1f8fc77a-8554-4716-8f6f-bd177139f82f	FI7089	2025-07-03 07:46:40.525269-07	2025-07-03T07:46:40.525269-07:00	john@daylilyinformatics.com	2025-07-03 07:46:40.547039-07	UPDATE	\N	\N	\N	f	f
114fbe66-797a-44c5-901c-cde26ca63e05	generic_instance	modified_dt	1f8fc77a-8554-4716-8f6f-bd177139f82f	FI7089	2025-07-03 07:46:40.525269-07	2025-07-03T07:46:40.547039-07:00	john@daylilyinformatics.com	2025-07-03 07:46:40.547039-07	UPDATE	\N	\N	\N	f	f
c9b5b823-2171-4d75-a881-be938b070b31	generic_instance	json_addl	1f8fc77a-8554-4716-8f6f-bd177139f82f	FI7089	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp4b1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7089.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7089.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp4b1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 734217, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:40.630664-07	UPDATE	\N	\N	\N	f	f
c4cdb620-87f2-4ad7-a092-b1cc980919e8	generic_instance	created_dt	1f8fc77a-8554-4716-8f6f-bd177139f82f	FI7089	2025-07-03 07:46:40.525269-07	2025-07-03T07:46:40.525269-07:00	john@daylilyinformatics.com	2025-07-03 07:46:40.630664-07	UPDATE	\N	\N	\N	f	f
3091c54c-923b-4f47-9382-cbafd47b0d8f	generic_instance	modified_dt	1f8fc77a-8554-4716-8f6f-bd177139f82f	FI7089	2025-07-03 07:46:40.547039-07	2025-07-03T07:46:40.630664-07:00	john@daylilyinformatics.com	2025-07-03 07:46:40.630664-07	UPDATE	\N	\N	\N	f	f
6407af33-5e8f-46fd-a538-643d54e7aa4d	generic_instance	\N	6a8be340-a613-4e02-a89f-e1f09be2be4a	FI7090	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:41.378868-07	INSERT	\N	\N	\N	f	f
e2846cf5-69d0-460d-8c22-b42f144c5bc4	generic_instance	json_addl	6a8be340-a613-4e02-a89f-e1f09be2be4a	FI7090	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:41.401322-07	UPDATE	\N	\N	\N	f	f
7533917e-0730-400b-b52f-7a1f38745845	generic_instance	created_dt	6a8be340-a613-4e02-a89f-e1f09be2be4a	FI7090	2025-07-03 07:46:41.378868-07	2025-07-03T07:46:41.378868-07:00	john@daylilyinformatics.com	2025-07-03 07:46:41.401322-07	UPDATE	\N	\N	\N	f	f
03fa887d-08ca-4b0d-a5ca-db33c39c81ee	generic_instance	modified_dt	6a8be340-a613-4e02-a89f-e1f09be2be4a	FI7090	2025-07-03 07:46:41.378868-07	2025-07-03T07:46:41.401322-07:00	john@daylilyinformatics.com	2025-07-03 07:46:41.401322-07	UPDATE	\N	\N	\N	f	f
846e57d0-e5fb-4d4d-92aa-020cb1a5f6d9	generic_instance	json_addl	6a8be340-a613-4e02-a89f-e1f09be2be4a	FI7090	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/ryr1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7090.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7090.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "ryr1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1094999, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:41.512513-07	UPDATE	\N	\N	\N	f	f
59efc944-79ae-4b92-9b0d-bd49bab2b5d6	generic_instance	created_dt	6a8be340-a613-4e02-a89f-e1f09be2be4a	FI7090	2025-07-03 07:46:41.378868-07	2025-07-03T07:46:41.378868-07:00	john@daylilyinformatics.com	2025-07-03 07:46:41.512513-07	UPDATE	\N	\N	\N	f	f
2a0076be-dd5a-4890-9c48-1c335ecf3976	generic_instance	modified_dt	6a8be340-a613-4e02-a89f-e1f09be2be4a	FI7090	2025-07-03 07:46:41.401322-07	2025-07-03T07:46:41.512513-07:00	john@daylilyinformatics.com	2025-07-03 07:46:41.512513-07	UPDATE	\N	\N	\N	f	f
38472e51-c6bc-4822-9702-84e0dbbc2796	generic_instance	\N	378516a2-6482-48f1-9e85-e33026c12b21	FI7091	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:42.275354-07	INSERT	\N	\N	\N	f	f
b04b6141-d31e-458e-b6ce-62e5dec15f35	generic_instance	json_addl	378516a2-6482-48f1-9e85-e33026c12b21	FI7091	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:42.299583-07	UPDATE	\N	\N	\N	f	f
f074d056-69c6-4fe5-8d81-1408d80d9f57	generic_instance	created_dt	378516a2-6482-48f1-9e85-e33026c12b21	FI7091	2025-07-03 07:46:42.275354-07	2025-07-03T07:46:42.275354-07:00	john@daylilyinformatics.com	2025-07-03 07:46:42.299583-07	UPDATE	\N	\N	\N	f	f
ee199478-c636-443f-9c0e-3da5009def89	generic_instance	modified_dt	378516a2-6482-48f1-9e85-e33026c12b21	FI7091	2025-07-03 07:46:42.275354-07	2025-07-03T07:46:42.299583-07:00	john@daylilyinformatics.com	2025-07-03 07:46:42.299583-07	UPDATE	\N	\N	\N	f	f
a30ab647-c4f7-49ea-a241-81ad43ef4a87	generic_instance	json_addl	378516a2-6482-48f1-9e85-e33026c12b21	FI7091	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp2c19.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7091.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7091.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp2c19.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 925411, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:42.418971-07	UPDATE	\N	\N	\N	f	f
e062395e-a48d-4ae7-82ed-1b24904f605a	generic_instance	created_dt	378516a2-6482-48f1-9e85-e33026c12b21	FI7091	2025-07-03 07:46:42.275354-07	2025-07-03T07:46:42.275354-07:00	john@daylilyinformatics.com	2025-07-03 07:46:42.418971-07	UPDATE	\N	\N	\N	f	f
10e4da19-5223-48ea-96ea-959d61e24b5c	generic_instance	modified_dt	378516a2-6482-48f1-9e85-e33026c12b21	FI7091	2025-07-03 07:46:42.299583-07	2025-07-03T07:46:42.418971-07:00	john@daylilyinformatics.com	2025-07-03 07:46:42.418971-07	UPDATE	\N	\N	\N	f	f
9ff3c209-75d0-4e30-90c5-8c36fdaed51b	generic_instance	\N	50f413f9-6000-4723-896a-926c67f39d59	FI7092	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:43.172535-07	INSERT	\N	\N	\N	f	f
594dee4b-0747-41ac-a79b-ed5a866c6e87	generic_instance	json_addl	50f413f9-6000-4723-896a-926c67f39d59	FI7092	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:43.198082-07	UPDATE	\N	\N	\N	f	f
35a39178-d93b-4d47-aa48-01e711d203b5	generic_instance	created_dt	50f413f9-6000-4723-896a-926c67f39d59	FI7092	2025-07-03 07:46:43.172535-07	2025-07-03T07:46:43.172535-07:00	john@daylilyinformatics.com	2025-07-03 07:46:43.198082-07	UPDATE	\N	\N	\N	f	f
d8a6e968-5081-4367-a022-1967ba446ab7	generic_instance	modified_dt	50f413f9-6000-4723-896a-926c67f39d59	FI7092	2025-07-03 07:46:43.172535-07	2025-07-03T07:46:43.198082-07:00	john@daylilyinformatics.com	2025-07-03 07:46:43.198082-07	UPDATE	\N	\N	\N	f	f
f1eb1a69-113e-4e26-b4de-f4aca6084fb0	generic_instance	json_addl	50f413f9-6000-4723-896a-926c67f39d59	FI7092	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/sult1a1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7092.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7092.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "sult1a1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 468288, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:43.309086-07	UPDATE	\N	\N	\N	f	f
c8f4c5e6-033d-4d17-9c2b-499330553c17	generic_instance	created_dt	50f413f9-6000-4723-896a-926c67f39d59	FI7092	2025-07-03 07:46:43.172535-07	2025-07-03T07:46:43.172535-07:00	john@daylilyinformatics.com	2025-07-03 07:46:43.309086-07	UPDATE	\N	\N	\N	f	f
433388d7-b99d-4829-a6aa-c00086f869d2	generic_instance	modified_dt	50f413f9-6000-4723-896a-926c67f39d59	FI7092	2025-07-03 07:46:43.198082-07	2025-07-03T07:46:43.309086-07:00	john@daylilyinformatics.com	2025-07-03 07:46:43.309086-07	UPDATE	\N	\N	\N	f	f
af8dd4fa-f14a-48dd-a7e5-0a8f20bd0db1	generic_instance	\N	5dae0e31-2c76-4322-bdfa-2f2d2f239bb2	FI7093	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:44.0301-07	INSERT	\N	\N	\N	f	f
c56ff3f9-9e3e-4dfb-be90-e7aca4dcd958	generic_instance	json_addl	5dae0e31-2c76-4322-bdfa-2f2d2f239bb2	FI7093	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:44.052619-07	UPDATE	\N	\N	\N	f	f
86418f20-650e-4c5a-8dc8-1076d9763bfc	generic_instance	created_dt	5dae0e31-2c76-4322-bdfa-2f2d2f239bb2	FI7093	2025-07-03 07:46:44.0301-07	2025-07-03T07:46:44.0301-07:00	john@daylilyinformatics.com	2025-07-03 07:46:44.052619-07	UPDATE	\N	\N	\N	f	f
a07ab713-f8a2-45aa-9055-014f737f1505	generic_instance	modified_dt	5dae0e31-2c76-4322-bdfa-2f2d2f239bb2	FI7093	2025-07-03 07:46:44.0301-07	2025-07-03T07:46:44.052619-07:00	john@daylilyinformatics.com	2025-07-03 07:46:44.052619-07	UPDATE	\N	\N	\N	f	f
013b1226-0d0c-4433-a6ed-0bb9cf36583b	generic_instance	json_addl	5dae0e31-2c76-4322-bdfa-2f2d2f239bb2	FI7093	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/tpmt.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7093.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7093.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "tpmt.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 824178, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:44.229127-07	UPDATE	\N	\N	\N	f	f
22ccfb0b-a156-474f-accc-48f69ef0d82e	generic_instance	created_dt	5dae0e31-2c76-4322-bdfa-2f2d2f239bb2	FI7093	2025-07-03 07:46:44.0301-07	2025-07-03T07:46:44.0301-07:00	john@daylilyinformatics.com	2025-07-03 07:46:44.229127-07	UPDATE	\N	\N	\N	f	f
0bfb1046-b0be-4e9f-b11b-32bd358bfeb1	generic_instance	modified_dt	5dae0e31-2c76-4322-bdfa-2f2d2f239bb2	FI7093	2025-07-03 07:46:44.052619-07	2025-07-03T07:46:44.229127-07:00	john@daylilyinformatics.com	2025-07-03 07:46:44.229127-07	UPDATE	\N	\N	\N	f	f
e931263d-62b0-4e43-ba1a-e3e041ee717f	generic_instance	\N	7828fb01-12d3-4b62-8452-7c919cea3ab9	FI7094	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:44.854199-07	INSERT	\N	\N	\N	f	f
e33c2d9b-825e-41ed-aca3-721fa9d2908b	generic_instance	json_addl	7828fb01-12d3-4b62-8452-7c919cea3ab9	FI7094	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:44.877452-07	UPDATE	\N	\N	\N	f	f
891daae8-908f-42f6-bd23-ae16808a2bd2	generic_instance	created_dt	7828fb01-12d3-4b62-8452-7c919cea3ab9	FI7094	2025-07-03 07:46:44.854199-07	2025-07-03T07:46:44.854199-07:00	john@daylilyinformatics.com	2025-07-03 07:46:44.877452-07	UPDATE	\N	\N	\N	f	f
34f707c1-3178-4f29-b6cf-c4b8ca905aa8	generic_instance	modified_dt	7828fb01-12d3-4b62-8452-7c919cea3ab9	FI7094	2025-07-03 07:46:44.854199-07	2025-07-03T07:46:44.877452-07:00	john@daylilyinformatics.com	2025-07-03 07:46:44.877452-07	UPDATE	\N	\N	\N	f	f
bf22e2a2-7d96-4cfc-8dfe-13f1c36a4a30	generic_instance	json_addl	7828fb01-12d3-4b62-8452-7c919cea3ab9	FI7094	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp3a7.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7094.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7094.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp3a7.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 417287, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:44.985226-07	UPDATE	\N	\N	\N	f	f
1af6eb19-e629-465f-8714-a67d86141748	generic_instance	created_dt	7828fb01-12d3-4b62-8452-7c919cea3ab9	FI7094	2025-07-03 07:46:44.854199-07	2025-07-03T07:46:44.854199-07:00	john@daylilyinformatics.com	2025-07-03 07:46:44.985226-07	UPDATE	\N	\N	\N	f	f
46183ae8-c267-4036-b114-b7cca591d140	generic_instance	modified_dt	7828fb01-12d3-4b62-8452-7c919cea3ab9	FI7094	2025-07-03 07:46:44.877452-07	2025-07-03T07:46:44.985226-07:00	john@daylilyinformatics.com	2025-07-03 07:46:44.985226-07	UPDATE	\N	\N	\N	f	f
83d8ef49-3ae3-4355-ac5c-98e61f872edd	generic_instance	\N	e010952a-8da4-4573-9b04-112e6178d9f8	FI7095	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:45.724499-07	INSERT	\N	\N	\N	f	f
cb3cf350-fdfc-4b5d-ae4c-b5b17d367c02	generic_instance	json_addl	e010952a-8da4-4573-9b04-112e6178d9f8	FI7095	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:45.759926-07	UPDATE	\N	\N	\N	f	f
daa0cdd3-5e37-4d81-9618-0fffb072d599	generic_instance	created_dt	e010952a-8da4-4573-9b04-112e6178d9f8	FI7095	2025-07-03 07:46:45.724499-07	2025-07-03T07:46:45.724499-07:00	john@daylilyinformatics.com	2025-07-03 07:46:45.759926-07	UPDATE	\N	\N	\N	f	f
ecc990a0-00f0-4660-99a3-d1c657d88a7a	generic_instance	modified_dt	e010952a-8da4-4573-9b04-112e6178d9f8	FI7095	2025-07-03 07:46:45.724499-07	2025-07-03T07:46:45.759926-07:00	john@daylilyinformatics.com	2025-07-03 07:46:45.759926-07	UPDATE	\N	\N	\N	f	f
d398a7af-40a1-4e75-9269-dfb4ff4e3d3e	generic_instance	json_addl	e010952a-8da4-4573-9b04-112e6178d9f8	FI7095	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp2w1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7095.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7095.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp2w1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1084808, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:45.883195-07	UPDATE	\N	\N	\N	f	f
78e14468-00ea-4bb5-9bc6-1886d27c2c42	generic_instance	created_dt	e010952a-8da4-4573-9b04-112e6178d9f8	FI7095	2025-07-03 07:46:45.724499-07	2025-07-03T07:46:45.724499-07:00	john@daylilyinformatics.com	2025-07-03 07:46:45.883195-07	UPDATE	\N	\N	\N	f	f
822bdc39-1df9-4814-ac41-58b40d7e6d3f	generic_instance	modified_dt	e010952a-8da4-4573-9b04-112e6178d9f8	FI7095	2025-07-03 07:46:45.759926-07	2025-07-03T07:46:45.883195-07:00	john@daylilyinformatics.com	2025-07-03 07:46:45.883195-07	UPDATE	\N	\N	\N	f	f
23a41364-eeb8-4f6b-a4d6-2b7a0b5ff232	generic_instance	\N	e2bd006f-4d32-4a15-b87c-68a3eb36b37e	FI7096	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:46.597019-07	INSERT	\N	\N	\N	f	f
59f4f69b-5bfa-4abe-826e-b4e36d90ad08	generic_instance	json_addl	e2bd006f-4d32-4a15-b87c-68a3eb36b37e	FI7096	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:46.610914-07	UPDATE	\N	\N	\N	f	f
83bef2da-4c19-4cb0-a92c-4a9c7bcab0f7	generic_instance	created_dt	e2bd006f-4d32-4a15-b87c-68a3eb36b37e	FI7096	2025-07-03 07:46:46.597019-07	2025-07-03T07:46:46.597019-07:00	john@daylilyinformatics.com	2025-07-03 07:46:46.610914-07	UPDATE	\N	\N	\N	f	f
3d5d4545-f2fb-4fa1-9d92-89779b2d07f7	generic_instance	modified_dt	e2bd006f-4d32-4a15-b87c-68a3eb36b37e	FI7096	2025-07-03 07:46:46.597019-07	2025-07-03T07:46:46.610914-07:00	john@daylilyinformatics.com	2025-07-03 07:46:46.610914-07	UPDATE	\N	\N	\N	f	f
7f81906e-0726-41a4-aefa-d4fbb262fd9f	generic_instance	json_addl	e2bd006f-4d32-4a15-b87c-68a3eb36b37e	FI7096	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp4f2.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7096.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7096.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp4f2.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1339117, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:46.741659-07	UPDATE	\N	\N	\N	f	f
603bccdb-0726-476a-8057-2948580c6644	generic_instance	created_dt	e2bd006f-4d32-4a15-b87c-68a3eb36b37e	FI7096	2025-07-03 07:46:46.597019-07	2025-07-03T07:46:46.597019-07:00	john@daylilyinformatics.com	2025-07-03 07:46:46.741659-07	UPDATE	\N	\N	\N	f	f
18970502-198a-4d63-bab6-73053a77dfcf	generic_instance	modified_dt	e2bd006f-4d32-4a15-b87c-68a3eb36b37e	FI7096	2025-07-03 07:46:46.610914-07	2025-07-03T07:46:46.741659-07:00	john@daylilyinformatics.com	2025-07-03 07:46:46.741659-07	UPDATE	\N	\N	\N	f	f
2bc18f6e-0d84-4d45-830a-dad3363c0e41	generic_instance	\N	ddbd7604-9140-472d-87fe-44279e0463b4	FI7097	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:47.514419-07	INSERT	\N	\N	\N	f	f
7f69b601-7821-4e31-85a9-be3ae961cf25	generic_instance	json_addl	ddbd7604-9140-472d-87fe-44279e0463b4	FI7097	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:47.545057-07	UPDATE	\N	\N	\N	f	f
50798f87-f23d-49cd-86c6-bd1aaf28e255	generic_instance	created_dt	ddbd7604-9140-472d-87fe-44279e0463b4	FI7097	2025-07-03 07:46:47.514419-07	2025-07-03T07:46:47.514419-07:00	john@daylilyinformatics.com	2025-07-03 07:46:47.545057-07	UPDATE	\N	\N	\N	f	f
b20b6e8f-1a91-4295-b8ef-7f816dc16c21	generic_instance	modified_dt	ddbd7604-9140-472d-87fe-44279e0463b4	FI7097	2025-07-03 07:46:47.514419-07	2025-07-03T07:46:47.545057-07:00	john@daylilyinformatics.com	2025-07-03 07:46:47.545057-07	UPDATE	\N	\N	\N	f	f
950022a7-461b-4e29-9665-7092ff09274e	generic_instance	json_addl	ddbd7604-9140-472d-87fe-44279e0463b4	FI7097	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp26a1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7097.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7097.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp26a1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 611814, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:47.670338-07	UPDATE	\N	\N	\N	f	f
22e7e4c2-741f-43c8-8995-693fa98fb13d	generic_instance	created_dt	ddbd7604-9140-472d-87fe-44279e0463b4	FI7097	2025-07-03 07:46:47.514419-07	2025-07-03T07:46:47.514419-07:00	john@daylilyinformatics.com	2025-07-03 07:46:47.670338-07	UPDATE	\N	\N	\N	f	f
9ed40c8c-0674-4608-b3ac-6f48cc7734d6	generic_instance	modified_dt	ddbd7604-9140-472d-87fe-44279e0463b4	FI7097	2025-07-03 07:46:47.545057-07	2025-07-03T07:46:47.670338-07:00	john@daylilyinformatics.com	2025-07-03 07:46:47.670338-07	UPDATE	\N	\N	\N	f	f
a7bb45cb-f49d-4d28-8d5b-d672e35f8894	generic_instance	\N	16a63367-c957-46fb-aee4-f026757bd059	FI7098	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:48.337838-07	INSERT	\N	\N	\N	f	f
5ec9ad34-b55d-4b20-88f3-d2fd8a7a5469	generic_instance	json_addl	16a63367-c957-46fb-aee4-f026757bd059	FI7098	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:48.355359-07	UPDATE	\N	\N	\N	f	f
83f315b2-b310-411b-a86f-78f3e9d2318b	generic_instance	created_dt	16a63367-c957-46fb-aee4-f026757bd059	FI7098	2025-07-03 07:46:48.337838-07	2025-07-03T07:46:48.337838-07:00	john@daylilyinformatics.com	2025-07-03 07:46:48.355359-07	UPDATE	\N	\N	\N	f	f
b4bbc7f7-d810-451d-b2a6-d6f1f37cf815	generic_instance	modified_dt	16a63367-c957-46fb-aee4-f026757bd059	FI7098	2025-07-03 07:46:48.337838-07	2025-07-03T07:46:48.355359-07:00	john@daylilyinformatics.com	2025-07-03 07:46:48.355359-07	UPDATE	\N	\N	\N	f	f
ea08bc8e-42e2-410d-9b9d-a87c230f02a8	generic_instance	json_addl	16a63367-c957-46fb-aee4-f026757bd059	FI7098	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/nudt15.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7098.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7098.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "nudt15.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 851608, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:48.476434-07	UPDATE	\N	\N	\N	f	f
32d54581-09ff-4aab-a02b-8b378bbcb1ad	generic_instance	created_dt	16a63367-c957-46fb-aee4-f026757bd059	FI7098	2025-07-03 07:46:48.337838-07	2025-07-03T07:46:48.337838-07:00	john@daylilyinformatics.com	2025-07-03 07:46:48.476434-07	UPDATE	\N	\N	\N	f	f
12122817-a0a2-4f97-81f9-b37289a9096b	generic_instance	modified_dt	16a63367-c957-46fb-aee4-f026757bd059	FI7098	2025-07-03 07:46:48.355359-07	2025-07-03T07:46:48.476434-07:00	john@daylilyinformatics.com	2025-07-03 07:46:48.476434-07	UPDATE	\N	\N	\N	f	f
c98df5e1-6b5b-47bc-905c-5d708dbeb30c	generic_instance	\N	205773c7-233a-423b-9b6f-704a210048fc	FI7099	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:49.184805-07	INSERT	\N	\N	\N	f	f
fea7f688-c155-492b-8d8a-17dcb9398cda	generic_instance	json_addl	205773c7-233a-423b-9b6f-704a210048fc	FI7099	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:49.225869-07	UPDATE	\N	\N	\N	f	f
c24d5e71-9fd7-4062-a421-848ddf4c2fa3	generic_instance	created_dt	205773c7-233a-423b-9b6f-704a210048fc	FI7099	2025-07-03 07:46:49.184805-07	2025-07-03T07:46:49.184805-07:00	john@daylilyinformatics.com	2025-07-03 07:46:49.225869-07	UPDATE	\N	\N	\N	f	f
3fc59327-1615-40e2-9c3d-a10530b5910d	generic_instance	modified_dt	205773c7-233a-423b-9b6f-704a210048fc	FI7099	2025-07-03 07:46:49.184805-07	2025-07-03T07:46:49.225869-07:00	john@daylilyinformatics.com	2025-07-03 07:46:49.225869-07	UPDATE	\N	\N	\N	f	f
e5e6f0b3-c878-4492-8cc8-068cb0ef439a	generic_instance	json_addl	205773c7-233a-423b-9b6f-704a210048fc	FI7099	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp19a1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7099.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7099.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp19a1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1103392, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:49.351009-07	UPDATE	\N	\N	\N	f	f
90be2a86-f041-4ee3-bc88-b3dea3324775	generic_instance	created_dt	205773c7-233a-423b-9b6f-704a210048fc	FI7099	2025-07-03 07:46:49.184805-07	2025-07-03T07:46:49.184805-07:00	john@daylilyinformatics.com	2025-07-03 07:46:49.351009-07	UPDATE	\N	\N	\N	f	f
59ca8ffa-7cd8-4e08-a5e7-cb5c9844f5d0	generic_instance	modified_dt	205773c7-233a-423b-9b6f-704a210048fc	FI7099	2025-07-03 07:46:49.225869-07	2025-07-03T07:46:49.351009-07:00	john@daylilyinformatics.com	2025-07-03 07:46:49.351009-07	UPDATE	\N	\N	\N	f	f
d3f66785-42ae-4415-abd9-d89e3c43858e	generic_instance	\N	8b323316-e042-4abb-9e7c-948747452c5e	FI7100	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:50.043178-07	INSERT	\N	\N	\N	f	f
5e303aee-1b25-43d9-8db8-5b44411e5b0b	generic_instance	json_addl	8b323316-e042-4abb-9e7c-948747452c5e	FI7100	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:50.069138-07	UPDATE	\N	\N	\N	f	f
abee51d7-ef53-4c65-a3ce-8fbdbe3bb497	generic_instance	created_dt	8b323316-e042-4abb-9e7c-948747452c5e	FI7100	2025-07-03 07:46:50.043178-07	2025-07-03T07:46:50.043178-07:00	john@daylilyinformatics.com	2025-07-03 07:46:50.069138-07	UPDATE	\N	\N	\N	f	f
e0ce0a8b-5111-4925-b4cd-d918dc0d1985	generic_instance	modified_dt	8b323316-e042-4abb-9e7c-948747452c5e	FI7100	2025-07-03 07:46:50.043178-07	2025-07-03T07:46:50.069138-07:00	john@daylilyinformatics.com	2025-07-03 07:46:50.069138-07	UPDATE	\N	\N	\N	f	f
6eb7b7f4-6f3a-4837-8c47-cb8dec1e2d2c	generic_instance	json_addl	8b323316-e042-4abb-9e7c-948747452c5e	FI7100	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/g6pd.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7100.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7100.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "g6pd.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 253692, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:50.195309-07	UPDATE	\N	\N	\N	f	f
61650c87-ccfe-4798-9ad5-50e3f4487901	generic_instance	created_dt	8b323316-e042-4abb-9e7c-948747452c5e	FI7100	2025-07-03 07:46:50.043178-07	2025-07-03T07:46:50.043178-07:00	john@daylilyinformatics.com	2025-07-03 07:46:50.195309-07	UPDATE	\N	\N	\N	f	f
59290dff-3c37-48d8-8492-092e332c3dbd	generic_instance	modified_dt	8b323316-e042-4abb-9e7c-948747452c5e	FI7100	2025-07-03 07:46:50.069138-07	2025-07-03T07:46:50.195309-07:00	john@daylilyinformatics.com	2025-07-03 07:46:50.195309-07	UPDATE	\N	\N	\N	f	f
61b7701b-d33e-473b-89d1-400409003f7d	generic_instance	\N	51f409ba-37e0-46f3-954f-eeb8b60c8886	FI7101	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:50.867942-07	INSERT	\N	\N	\N	f	f
014b5628-1267-43ce-a288-34f102309038	generic_instance	json_addl	51f409ba-37e0-46f3-954f-eeb8b60c8886	FI7101	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:50.898746-07	UPDATE	\N	\N	\N	f	f
9764cc3a-3e1f-4816-8406-57a8e52816ff	generic_instance	created_dt	51f409ba-37e0-46f3-954f-eeb8b60c8886	FI7101	2025-07-03 07:46:50.867942-07	2025-07-03T07:46:50.867942-07:00	john@daylilyinformatics.com	2025-07-03 07:46:50.898746-07	UPDATE	\N	\N	\N	f	f
55378ab2-7268-42c9-a0d6-7d4ce982ce81	generic_instance	modified_dt	51f409ba-37e0-46f3-954f-eeb8b60c8886	FI7101	2025-07-03 07:46:50.867942-07	2025-07-03T07:46:50.898746-07:00	john@daylilyinformatics.com	2025-07-03 07:46:50.898746-07	UPDATE	\N	\N	\N	f	f
d2f42543-1c32-42f4-933e-cbb42ab1fc59	generic_instance	json_addl	51f409ba-37e0-46f3-954f-eeb8b60c8886	FI7101	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp3a43.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7101.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7101.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp3a43.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 534281, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:51.025392-07	UPDATE	\N	\N	\N	f	f
1ddb09f8-69ec-49d8-b790-40ce91c6533a	generic_instance	created_dt	51f409ba-37e0-46f3-954f-eeb8b60c8886	FI7101	2025-07-03 07:46:50.867942-07	2025-07-03T07:46:50.867942-07:00	john@daylilyinformatics.com	2025-07-03 07:46:51.025392-07	UPDATE	\N	\N	\N	f	f
16ca9ddb-7823-41e6-8745-0a8f86a609c2	generic_instance	modified_dt	51f409ba-37e0-46f3-954f-eeb8b60c8886	FI7101	2025-07-03 07:46:50.898746-07	2025-07-03T07:46:51.025392-07:00	john@daylilyinformatics.com	2025-07-03 07:46:51.025392-07	UPDATE	\N	\N	\N	f	f
6e9d0a3b-0067-4999-9a72-bc04ae2cb98a	generic_instance	\N	6be7b5f3-b9a6-436a-8c5a-d011d47eee2d	FI7102	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:51.714891-07	INSERT	\N	\N	\N	f	f
bd896700-f06f-4a65-9896-5129fbb3e4c8	generic_instance	json_addl	6be7b5f3-b9a6-436a-8c5a-d011d47eee2d	FI7102	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:51.740635-07	UPDATE	\N	\N	\N	f	f
4b407583-8d60-405e-9f43-38369ba05fb6	generic_instance	created_dt	6be7b5f3-b9a6-436a-8c5a-d011d47eee2d	FI7102	2025-07-03 07:46:51.714891-07	2025-07-03T07:46:51.714891-07:00	john@daylilyinformatics.com	2025-07-03 07:46:51.740635-07	UPDATE	\N	\N	\N	f	f
69c50e55-11e3-4fc4-999d-25f4b0c3b206	generic_instance	modified_dt	6be7b5f3-b9a6-436a-8c5a-d011d47eee2d	FI7102	2025-07-03 07:46:51.714891-07	2025-07-03T07:46:51.740635-07:00	john@daylilyinformatics.com	2025-07-03 07:46:51.740635-07	UPDATE	\N	\N	\N	f	f
567cc7aa-30e7-4ae2-9a84-0425d47f73bf	generic_instance	json_addl	6be7b5f3-b9a6-436a-8c5a-d011d47eee2d	FI7102	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/slco2b1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7102.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7102.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "slco2b1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 696892, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:51.879088-07	UPDATE	\N	\N	\N	f	f
85612af8-d022-427e-a8c1-b1a12199de2c	generic_instance	created_dt	6be7b5f3-b9a6-436a-8c5a-d011d47eee2d	FI7102	2025-07-03 07:46:51.714891-07	2025-07-03T07:46:51.714891-07:00	john@daylilyinformatics.com	2025-07-03 07:46:51.879088-07	UPDATE	\N	\N	\N	f	f
86d4985a-a0b3-449b-87af-5cac7523fbee	generic_instance	modified_dt	6be7b5f3-b9a6-436a-8c5a-d011d47eee2d	FI7102	2025-07-03 07:46:51.740635-07	2025-07-03T07:46:51.879088-07:00	john@daylilyinformatics.com	2025-07-03 07:46:51.879088-07	UPDATE	\N	\N	\N	f	f
4b7d5a67-611c-4b5b-8bd2-8f603f4a5c67	generic_instance	\N	9a2d33f4-0c38-4b92-b85c-e3650dbf67d5	FI7103	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:52.475961-07	INSERT	\N	\N	\N	f	f
c1830ccb-742a-4e80-ae2f-3d049d2ced1d	generic_instance	json_addl	9a2d33f4-0c38-4b92-b85c-e3650dbf67d5	FI7103	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:52.498925-07	UPDATE	\N	\N	\N	f	f
21f0a172-829f-48ef-8237-91d8755811d8	generic_instance	created_dt	9a2d33f4-0c38-4b92-b85c-e3650dbf67d5	FI7103	2025-07-03 07:46:52.475961-07	2025-07-03T07:46:52.475961-07:00	john@daylilyinformatics.com	2025-07-03 07:46:52.498925-07	UPDATE	\N	\N	\N	f	f
4d1433d7-475c-45e3-9bcd-cfcc56b0a5c7	generic_instance	modified_dt	9a2d33f4-0c38-4b92-b85c-e3650dbf67d5	FI7103	2025-07-03 07:46:52.475961-07	2025-07-03T07:46:52.498925-07:00	john@daylilyinformatics.com	2025-07-03 07:46:52.498925-07	UPDATE	\N	\N	\N	f	f
d0614c0b-23d9-4a73-bc88-25f3a86209d3	generic_instance	json_addl	9a2d33f4-0c38-4b92-b85c-e3650dbf67d5	FI7103	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cacna1s.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7103.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7103.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cacna1s.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1004301, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:52.59389-07	UPDATE	\N	\N	\N	f	f
8f16e376-b07e-4596-9aa0-97bd088b0111	generic_instance	created_dt	9a2d33f4-0c38-4b92-b85c-e3650dbf67d5	FI7103	2025-07-03 07:46:52.475961-07	2025-07-03T07:46:52.475961-07:00	john@daylilyinformatics.com	2025-07-03 07:46:52.59389-07	UPDATE	\N	\N	\N	f	f
cd6e0b61-4f00-428a-92d3-6202d97b3b83	generic_instance	modified_dt	9a2d33f4-0c38-4b92-b85c-e3650dbf67d5	FI7103	2025-07-03 07:46:52.498925-07	2025-07-03T07:46:52.59389-07:00	john@daylilyinformatics.com	2025-07-03 07:46:52.59389-07	UPDATE	\N	\N	\N	f	f
c53dca08-1470-4e55-ac06-76dbe25f7fbe	generic_instance	\N	7d5dcd5e-7d18-4094-aab9-64c5da8698ec	FI7104	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:53.082013-07	INSERT	\N	\N	\N	f	f
04225706-f1bb-45e8-ad3c-bf94c30538a6	generic_instance	json_addl	7d5dcd5e-7d18-4094-aab9-64c5da8698ec	FI7104	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:53.111501-07	UPDATE	\N	\N	\N	f	f
011d00b2-d6ae-482f-8bfd-6194cdee123f	generic_instance	created_dt	7d5dcd5e-7d18-4094-aab9-64c5da8698ec	FI7104	2025-07-03 07:46:53.082013-07	2025-07-03T07:46:53.082013-07:00	john@daylilyinformatics.com	2025-07-03 07:46:53.111501-07	UPDATE	\N	\N	\N	f	f
18046700-49b8-440c-95bd-323dc98779d6	generic_instance	modified_dt	7d5dcd5e-7d18-4094-aab9-64c5da8698ec	FI7104	2025-07-03 07:46:53.082013-07	2025-07-03T07:46:53.111501-07:00	john@daylilyinformatics.com	2025-07-03 07:46:53.111501-07	UPDATE	\N	\N	\N	f	f
72d37573-967f-4bea-8365-e9184ce162e7	generic_instance	json_addl	7d5dcd5e-7d18-4094-aab9-64c5da8698ec	FI7104	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/slc15a2.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7104.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7104.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "slc15a2.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 793207, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:53.194867-07	UPDATE	\N	\N	\N	f	f
2df63384-e23f-406d-98a7-674d540d6467	generic_instance	created_dt	7d5dcd5e-7d18-4094-aab9-64c5da8698ec	FI7104	2025-07-03 07:46:53.082013-07	2025-07-03T07:46:53.082013-07:00	john@daylilyinformatics.com	2025-07-03 07:46:53.194867-07	UPDATE	\N	\N	\N	f	f
e8e84cf2-a058-4aab-b9a3-db395e33a34c	generic_instance	modified_dt	7d5dcd5e-7d18-4094-aab9-64c5da8698ec	FI7104	2025-07-03 07:46:53.111501-07	2025-07-03T07:46:53.194867-07:00	john@daylilyinformatics.com	2025-07-03 07:46:53.194867-07	UPDATE	\N	\N	\N	f	f
0cdb2b44-cf1f-44ed-84ac-18f5320883f1	generic_instance	\N	bb3c68a8-5a54-4a62-b9e3-d83cfeacf4f6	FI7105	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:53.724439-07	INSERT	\N	\N	\N	f	f
5f354bdc-6cf4-4be8-9265-4513653154d6	generic_instance	json_addl	bb3c68a8-5a54-4a62-b9e3-d83cfeacf4f6	FI7105	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:53.752035-07	UPDATE	\N	\N	\N	f	f
fcf54972-4635-4d6e-8a7f-e7cba718fb6c	generic_instance	created_dt	bb3c68a8-5a54-4a62-b9e3-d83cfeacf4f6	FI7105	2025-07-03 07:46:53.724439-07	2025-07-03T07:46:53.724439-07:00	john@daylilyinformatics.com	2025-07-03 07:46:53.752035-07	UPDATE	\N	\N	\N	f	f
63783d64-ec9b-4bc8-9ae3-12ee387176d3	generic_instance	modified_dt	bb3c68a8-5a54-4a62-b9e3-d83cfeacf4f6	FI7105	2025-07-03 07:46:53.724439-07	2025-07-03T07:46:53.752035-07:00	john@daylilyinformatics.com	2025-07-03 07:46:53.752035-07	UPDATE	\N	\N	\N	f	f
fb0e7975-116c-467e-9348-f1824eaad1d0	generic_instance	json_addl	bb3c68a8-5a54-4a62-b9e3-d83cfeacf4f6	FI7105	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/tpmt.vcf", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "vcf", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7105.vcf", "current_s3_uri": "s3://daylily-dewey-0/0/FI7105.vcf", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "tpmt.vcf", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "vcf", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 28012779, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:53.840522-07	UPDATE	\N	\N	\N	f	f
f8b2ec98-dd9e-4611-9c9b-c2aab139b68c	generic_instance	created_dt	bb3c68a8-5a54-4a62-b9e3-d83cfeacf4f6	FI7105	2025-07-03 07:46:53.724439-07	2025-07-03T07:46:53.724439-07:00	john@daylilyinformatics.com	2025-07-03 07:46:53.840522-07	UPDATE	\N	\N	\N	f	f
aa8456ab-7964-4f93-9c73-5636627f6e1f	generic_instance	modified_dt	bb3c68a8-5a54-4a62-b9e3-d83cfeacf4f6	FI7105	2025-07-03 07:46:53.752035-07	2025-07-03T07:46:53.840522-07:00	john@daylilyinformatics.com	2025-07-03 07:46:53.840522-07	UPDATE	\N	\N	\N	f	f
ba6ce5ac-f41d-497c-bcf2-b6ca032c0416	generic_instance	\N	b32d8a96-efe1-42c3-abe3-7a3241e0c102	FI7106	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:55.434585-07	INSERT	\N	\N	\N	f	f
908a589b-70a9-49d3-a782-82f14f957719	generic_instance	json_addl	b32d8a96-efe1-42c3-abe3-7a3241e0c102	FI7106	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:55.451585-07	UPDATE	\N	\N	\N	f	f
e1efffc3-c974-420d-b559-0328148608ca	generic_instance	created_dt	b32d8a96-efe1-42c3-abe3-7a3241e0c102	FI7106	2025-07-03 07:46:55.434585-07	2025-07-03T07:46:55.434585-07:00	john@daylilyinformatics.com	2025-07-03 07:46:55.451585-07	UPDATE	\N	\N	\N	f	f
ff80865e-f77a-48c6-aab4-a2cd799dae97	generic_instance	modified_dt	b32d8a96-efe1-42c3-abe3-7a3241e0c102	FI7106	2025-07-03 07:46:55.434585-07	2025-07-03T07:46:55.451585-07:00	john@daylilyinformatics.com	2025-07-03 07:46:55.451585-07	UPDATE	\N	\N	\N	f	f
92ee746a-95aa-481c-a624-be98a58ab756	generic_instance	json_addl	b32d8a96-efe1-42c3-abe3-7a3241e0c102	FI7106	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp3a5.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7106.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7106.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp3a5.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 441422, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:55.698071-07	UPDATE	\N	\N	\N	f	f
190b8758-9168-414c-84df-4c50ce416277	generic_instance	created_dt	b32d8a96-efe1-42c3-abe3-7a3241e0c102	FI7106	2025-07-03 07:46:55.434585-07	2025-07-03T07:46:55.434585-07:00	john@daylilyinformatics.com	2025-07-03 07:46:55.698071-07	UPDATE	\N	\N	\N	f	f
fbfabb60-f1a8-4b81-b76a-7d2b136934eb	generic_instance	modified_dt	b32d8a96-efe1-42c3-abe3-7a3241e0c102	FI7106	2025-07-03 07:46:55.451585-07	2025-07-03T07:46:55.698071-07:00	john@daylilyinformatics.com	2025-07-03 07:46:55.698071-07	UPDATE	\N	\N	\N	f	f
773bdd5c-cf00-49f6-b56f-3e594cf8ad69	generic_instance	\N	051aa336-0ce7-4f70-bc74-9efbc624fce9	FI7107	\N	\N	john@daylilyinformatics.com	2025-07-03 07:46:56.362055-07	INSERT	\N	\N	\N	f	f
b1e1134b-39dc-4ffe-b1b1-fa3102877be4	generic_instance	json_addl	051aa336-0ce7-4f70-bc74-9efbc624fce9	FI7107	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:56.396109-07	UPDATE	\N	\N	\N	f	f
2256acd5-14a0-4756-b818-bc1bbfdfd80a	generic_instance	created_dt	051aa336-0ce7-4f70-bc74-9efbc624fce9	FI7107	2025-07-03 07:46:56.362055-07	2025-07-03T07:46:56.362055-07:00	john@daylilyinformatics.com	2025-07-03 07:46:56.396109-07	UPDATE	\N	\N	\N	f	f
17d122de-7c96-4c2a-b74c-0c255bd19dcd	generic_instance	modified_dt	051aa336-0ce7-4f70-bc74-9efbc624fce9	FI7107	2025-07-03 07:46:56.362055-07	2025-07-03T07:46:56.396109-07:00	john@daylilyinformatics.com	2025-07-03 07:46:56.396109-07	UPDATE	\N	\N	\N	f	f
5fb7bbb8-41c1-4b2d-9451-17f9acdc8458	generic_instance	json_addl	051aa336-0ce7-4f70-bc74-9efbc624fce9	FI7107	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/nat1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7107.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7107.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "nat1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1078162, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:46:56.518702-07	UPDATE	\N	\N	\N	f	f
93b44c0b-f4e4-44ba-8e3e-88f9083930c7	generic_instance	created_dt	051aa336-0ce7-4f70-bc74-9efbc624fce9	FI7107	2025-07-03 07:46:56.362055-07	2025-07-03T07:46:56.362055-07:00	john@daylilyinformatics.com	2025-07-03 07:46:56.518702-07	UPDATE	\N	\N	\N	f	f
0b25c624-9235-45f3-b600-b9069580545f	generic_instance	modified_dt	051aa336-0ce7-4f70-bc74-9efbc624fce9	FI7107	2025-07-03 07:46:56.396109-07	2025-07-03T07:46:56.518702-07:00	john@daylilyinformatics.com	2025-07-03 07:46:56.518702-07	UPDATE	\N	\N	\N	f	f
36f2bb57-32a8-4da0-9b1a-c0a2ac632510	generic_instance	\N	245c6374-146a-4bf6-898d-e3a0ce08f78d	FS25	\N	\N	john@daylilyinformatics.com	2025-07-03 07:49:08.725504-07	INSERT	\N	\N	\N	f	f
4927303f-b3bb-40a1-b9a3-a020758db610	generic_instance	\N	82961788-e3f9-441f-96f3-de01e051f122	FI7108	\N	\N	john@daylilyinformatics.com	2025-07-03 07:49:08.773823-07	INSERT	\N	\N	\N	f	f
38cc7c80-55f8-4406-abf3-5696eb90f0aa	generic_instance	json_addl	82961788-e3f9-441f-96f3-de01e051f122	FI7108	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T144901", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T144901", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:49:08.782093-07	UPDATE	\N	\N	\N	f	f
0cfa1598-c646-4d89-b150-2727ac8e6d66	generic_instance	created_dt	82961788-e3f9-441f-96f3-de01e051f122	FI7108	2025-07-03 07:49:08.773823-07	2025-07-03T07:49:08.773823-07:00	john@daylilyinformatics.com	2025-07-03 07:49:08.782093-07	UPDATE	\N	\N	\N	f	f
65c844c4-e9d9-4c98-bfcc-698a94b421bd	generic_instance	modified_dt	82961788-e3f9-441f-96f3-de01e051f122	FI7108	2025-07-03 07:49:08.773823-07	2025-07-03T07:49:08.782093-07:00	john@daylilyinformatics.com	2025-07-03 07:49:08.782093-07	UPDATE	\N	\N	\N	f	f
ce913740-d4c3-4380-afc2-30914feba7f1	generic_instance	json_addl	82961788-e3f9-441f-96f3-de01e051f122	FI7108	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "na", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T144901", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "tbxas1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7108.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7108.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T144901", "original_file_md5": "", "original_file_name": "tbxas1.vcf.gz", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1085371, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:49:09.033598-07	UPDATE	\N	\N	\N	f	f
73480eaa-b7bb-4dea-8a1d-7d9bc8dbb436	generic_instance	created_dt	82961788-e3f9-441f-96f3-de01e051f122	FI7108	2025-07-03 07:49:08.773823-07	2025-07-03T07:49:08.773823-07:00	john@daylilyinformatics.com	2025-07-03 07:49:09.033598-07	UPDATE	\N	\N	\N	f	f
4e4a643e-ac73-49b7-b503-db6a31eefb5f	generic_instance	modified_dt	82961788-e3f9-441f-96f3-de01e051f122	FI7108	2025-07-03 07:49:08.782093-07	2025-07-03T07:49:09.033598-07:00	john@daylilyinformatics.com	2025-07-03 07:49:09.033598-07	UPDATE	\N	\N	\N	f	f
1859a7f6-c2cb-4772-8897-c07d558a0ace	generic_instance	json_addl	bc77f3ee-c50b-43eb-8f41-d9fec67a5da5	CWX94	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:08.783511-07	UPDATE	\N	\N	\N	f	f
1921df5e-a5ed-40d7-b452-d0cbd93a14d2	generic_instance	created_dt	bc77f3ee-c50b-43eb-8f41-d9fec67a5da5	CWX94	2025-07-03 08:23:08.779117-07	2025-07-03T08:23:08.779117-07:00	daylily	2025-07-03 08:23:08.783511-07	UPDATE	\N	\N	\N	f	f
cbc5bd42-9497-4561-b327-4f224885e99e	generic_instance	modified_dt	bc77f3ee-c50b-43eb-8f41-d9fec67a5da5	CWX94	2025-07-03 08:23:08.779117-07	2025-07-03T08:23:08.783511-07:00	daylily	2025-07-03 08:23:08.783511-07	UPDATE	\N	\N	\N	f	f
b688e533-69c9-44b5-afc1-c6697cb890ed	generic_instance_lineage	\N	c5c8cfa0-fe0b-480b-ab03-e5921141b8e1	GL312	\N	\N	daylily	2025-07-03 08:23:08.78484-07	INSERT	\N	\N	\N	f	f
8f2f5708-fdfa-442c-8579-9810189fb105	generic_instance	\N	780f794b-7f49-4805-801b-912cabfd265e	CWX95	\N	\N	daylily	2025-07-03 08:23:08.78484-07	INSERT	\N	\N	\N	f	f
88b61fcc-d96b-414b-8796-aa2819fe33a8	generic_instance	json_addl	780f794b-7f49-4805-801b-912cabfd265e	CWX95	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:08.789232-07	UPDATE	\N	\N	\N	f	f
1ccd0316-2ddd-43c1-a169-38ce0926fd09	generic_instance	created_dt	780f794b-7f49-4805-801b-912cabfd265e	CWX95	2025-07-03 08:23:08.78484-07	2025-07-03T08:23:08.78484-07:00	daylily	2025-07-03 08:23:08.789232-07	UPDATE	\N	\N	\N	f	f
807c1b9a-0ec3-41e5-8bdb-863472f9ac1c	generic_instance	modified_dt	780f794b-7f49-4805-801b-912cabfd265e	CWX95	2025-07-03 08:23:08.78484-07	2025-07-03T08:23:08.789232-07:00	daylily	2025-07-03 08:23:08.789232-07	UPDATE	\N	\N	\N	f	f
ac7b50d6-061b-4a5c-a787-e9aa36f8090d	generic_instance_lineage	\N	62afa9ab-7bbd-451f-b344-504dc4fb7311	GL313	\N	\N	daylily	2025-07-03 08:23:08.790553-07	INSERT	\N	\N	\N	f	f
4b5686dd-4bf9-4dc7-9621-bc1cadc78e63	generic_instance	\N	06fc3873-3a8c-404d-a4af-b8c7116d3213	CWX96	\N	\N	daylily	2025-07-03 08:23:08.790553-07	INSERT	\N	\N	\N	f	f
81ce7092-ac8a-4f4f-8a8d-78675482deeb	generic_instance	json_addl	06fc3873-3a8c-404d-a4af-b8c7116d3213	CWX96	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:08.796577-07	UPDATE	\N	\N	\N	f	f
6692acdc-c35b-4f1b-a7a5-d1f3d9413efe	generic_instance	created_dt	06fc3873-3a8c-404d-a4af-b8c7116d3213	CWX96	2025-07-03 08:23:08.790553-07	2025-07-03T08:23:08.790553-07:00	daylily	2025-07-03 08:23:08.796577-07	UPDATE	\N	\N	\N	f	f
faf33153-77d2-42df-a2fc-b1b7fd9f16da	generic_instance	modified_dt	06fc3873-3a8c-404d-a4af-b8c7116d3213	CWX96	2025-07-03 08:23:08.790553-07	2025-07-03T08:23:08.796577-07:00	daylily	2025-07-03 08:23:08.796577-07	UPDATE	\N	\N	\N	f	f
5a871bf9-70ba-4619-b995-c8a8e109094f	generic_instance_lineage	\N	7aaf3aba-b740-494f-ae80-8d4d6a0539db	GL314	\N	\N	daylily	2025-07-03 08:23:08.799621-07	INSERT	\N	\N	\N	f	f
cfb26e1d-c317-421d-8c4f-1755afac993a	generic_instance	\N	85fa4716-fb10-4279-9999-ebd9c08f6b20	CWX97	\N	\N	daylily	2025-07-03 08:23:08.799621-07	INSERT	\N	\N	\N	f	f
b59572d2-c807-469c-b46c-dbe30f3af624	generic_instance	\N	53dfbf91-c236-4d6e-b31c-a9bf4bc369a2	MRX211	\N	\N	daylily	2025-07-03 08:23:17.848523-07	INSERT	\N	\N	\N	f	f
e2eaddfa-be1e-4368-8d26-7f4cdda4ba8b	generic_instance_lineage	\N	b8044b01-4e2a-4b8c-953e-95d9546307a4	GL572	\N	\N	daylily	2025-07-03 08:23:17.855048-07	INSERT	\N	\N	\N	f	f
f970d543-78d6-401b-829d-be3e1b00c2a5	generic_instance	\N	3666e200-4c24-4efc-a76b-d72a6faa5e65	MRX212	\N	\N	daylily	2025-07-03 08:23:17.855048-07	INSERT	\N	\N	\N	f	f
c2ac9a20-84f5-4d61-9d19-b69e7b9ba6e2	generic_instance_lineage	\N	9560ac70-bc41-4df7-8137-2c05a37394ae	GL573	\N	\N	daylily	2025-07-03 08:23:17.867487-07	INSERT	\N	\N	\N	f	f
086dde28-205e-486d-a0fb-d682def7d9c8	generic_instance	\N	c282db24-d774-48c1-8634-eb60d7b861fc	MRX213	\N	\N	daylily	2025-07-03 08:23:17.867487-07	INSERT	\N	\N	\N	f	f
a988bbb7-008b-452b-b5db-6b4d3a21861d	generic_instance_lineage	\N	e5f9d28a-edb7-4156-abc9-43bbf215f691	GL124	\N	\N	john@daylilyinformatics.com	2025-07-03 07:49:09.90103-07	INSERT	\N	\N	\N	f	f
067b5086-f2ef-40d3-8353-3eda6c00a3c8	generic_instance	\N	eeee8645-360f-4290-a6f7-a7275d34db61	FS26	\N	\N	john@daylilyinformatics.com	2025-07-03 07:50:08.537193-07	INSERT	\N	\N	\N	f	f
99b0ed7d-e10d-4336-8eb2-8ceed3ec3c9e	generic_instance	\N	a2bc4b6d-42e0-4993-91ea-cb3b84c406ba	FI7109	\N	\N	john@daylilyinformatics.com	2025-07-03 07:50:08.569214-07	INSERT	\N	\N	\N	f	f
438585a6-8215-4ce0-96d8-c35f35c714c5	generic_instance	json_addl	a2bc4b6d-42e0-4993-91ea-cb3b84c406ba	FI7109	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T145008", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T145008", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:50:08.574025-07	UPDATE	\N	\N	\N	f	f
f76991bc-63f2-4ee7-87c5-0f168deacfb4	generic_instance	created_dt	a2bc4b6d-42e0-4993-91ea-cb3b84c406ba	FI7109	2025-07-03 07:50:08.569214-07	2025-07-03T07:50:08.569214-07:00	john@daylilyinformatics.com	2025-07-03 07:50:08.574025-07	UPDATE	\N	\N	\N	f	f
2be7bdc5-474d-4ecc-a2b0-62316b184f35	generic_instance	modified_dt	a2bc4b6d-42e0-4993-91ea-cb3b84c406ba	FI7109	2025-07-03 07:50:08.569214-07	2025-07-03T07:50:08.574025-07:00	john@daylilyinformatics.com	2025-07-03 07:50:08.574025-07	UPDATE	\N	\N	\N	f	f
8f7eddb2-356d-4ccc-a2cd-f6894c8475d1	generic_instance	json_addl	a2bc4b6d-42e0-4993-91ea-cb3b84c406ba	FI7109	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T145008", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7109.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7109.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T145008", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:50:08.837143-07	UPDATE	\N	\N	\N	f	f
9f42315b-bcc0-443f-ab74-ef83b743d46e	generic_instance	created_dt	a2bc4b6d-42e0-4993-91ea-cb3b84c406ba	FI7109	2025-07-03 07:50:08.569214-07	2025-07-03T07:50:08.569214-07:00	john@daylilyinformatics.com	2025-07-03 07:50:08.837143-07	UPDATE	\N	\N	\N	f	f
841f7bf6-3299-46e6-bea3-ad086472cbe0	generic_instance	modified_dt	a2bc4b6d-42e0-4993-91ea-cb3b84c406ba	FI7109	2025-07-03 07:50:08.574025-07	2025-07-03T07:50:08.837143-07:00	john@daylilyinformatics.com	2025-07-03 07:50:08.837143-07	UPDATE	\N	\N	\N	f	f
ab31fa8a-ce90-4054-8791-20b065261433	generic_instance_lineage	\N	99745052-4f67-459f-a147-92668662db3b	GL125	\N	\N	john@daylilyinformatics.com	2025-07-03 07:50:09.765023-07	INSERT	\N	\N	\N	f	f
265acffa-f46a-4a4e-9943-969b1f293dcc	generic_instance	\N	9379619e-93d1-4e5e-a313-9576af3ffaab	FI7110	\N	\N	john@daylilyinformatics.com	2025-07-03 07:50:09.773183-07	INSERT	\N	\N	\N	f	f
c3047247-574b-429b-a9fd-a014c9b37f1e	generic_instance	json_addl	9379619e-93d1-4e5e-a313-9576af3ffaab	FI7110	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T145008", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T145008", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:50:09.976718-07	UPDATE	\N	\N	\N	f	f
873f54ca-e4c0-4649-9793-b0ebf5a0eda7	generic_instance	created_dt	9379619e-93d1-4e5e-a313-9576af3ffaab	FI7110	2025-07-03 07:50:09.773183-07	2025-07-03T07:50:09.773183-07:00	john@daylilyinformatics.com	2025-07-03 07:50:09.976718-07	UPDATE	\N	\N	\N	f	f
5f759585-ee20-444a-825c-176999cec15d	generic_instance	modified_dt	9379619e-93d1-4e5e-a313-9576af3ffaab	FI7110	2025-07-03 07:50:09.773183-07	2025-07-03T07:50:09.976718-07:00	john@daylilyinformatics.com	2025-07-03 07:50:09.976718-07	UPDATE	\N	\N	\N	f	f
838f5b9a-2f6b-4283-8ff3-e133ef6072f3	generic_instance	json_addl	9379619e-93d1-4e5e-a313-9576af3ffaab	FI7110	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T145008", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7110.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7110.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T145008", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:50:10.082248-07	UPDATE	\N	\N	\N	f	f
d12cf4df-bfaf-4f0d-9945-fbf7035a16de	generic_instance	created_dt	9379619e-93d1-4e5e-a313-9576af3ffaab	FI7110	2025-07-03 07:50:09.773183-07	2025-07-03T07:50:09.773183-07:00	john@daylilyinformatics.com	2025-07-03 07:50:10.082248-07	UPDATE	\N	\N	\N	f	f
485e157e-d21e-4945-9164-0c43c598a766	generic_instance	modified_dt	9379619e-93d1-4e5e-a313-9576af3ffaab	FI7110	2025-07-03 07:50:09.976718-07	2025-07-03T07:50:10.082248-07:00	john@daylilyinformatics.com	2025-07-03 07:50:10.082248-07	UPDATE	\N	\N	\N	f	f
752f90a7-8567-49f5-88d3-be49a34aa750	generic_instance_lineage	\N	8bf2431f-d5cf-491f-bf01-d728c249643e	GL126	\N	\N	john@daylilyinformatics.com	2025-07-03 07:50:10.707766-07	INSERT	\N	\N	\N	f	f
a14c7162-3acd-4d5f-b003-694b44330023	generic_instance	\N	c01298e9-f706-48dd-b71e-7666927fccf3	FI7111	\N	\N	john@daylilyinformatics.com	2025-07-03 07:50:10.713804-07	INSERT	\N	\N	\N	f	f
d8e6cc97-d71f-4e56-b4e2-e9c89e372083	generic_instance	json_addl	c01298e9-f706-48dd-b71e-7666927fccf3	FI7111	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T145008", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T145008", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:50:10.722424-07	UPDATE	\N	\N	\N	f	f
0c88eed1-2496-443b-b513-c3a0c92ef19e	generic_instance	created_dt	c01298e9-f706-48dd-b71e-7666927fccf3	FI7111	2025-07-03 07:50:10.713804-07	2025-07-03T07:50:10.713804-07:00	john@daylilyinformatics.com	2025-07-03 07:50:10.722424-07	UPDATE	\N	\N	\N	f	f
e9b78f70-04aa-40ef-b728-4555b50e2a47	generic_instance	modified_dt	c01298e9-f706-48dd-b71e-7666927fccf3	FI7111	2025-07-03 07:50:10.713804-07	2025-07-03T07:50:10.722424-07:00	john@daylilyinformatics.com	2025-07-03 07:50:10.722424-07	UPDATE	\N	\N	\N	f	f
2c1202cc-3691-49f9-a7c9-44777cb048f8	generic_instance	json_addl	c01298e9-f706-48dd-b71e-7666927fccf3	FI7111	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T145008", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7111.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7111.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T145008", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:50:10.941052-07	UPDATE	\N	\N	\N	f	f
b4d24507-9b2c-43f7-b14d-6044d02e07ee	generic_instance	created_dt	c01298e9-f706-48dd-b71e-7666927fccf3	FI7111	2025-07-03 07:50:10.713804-07	2025-07-03T07:50:10.713804-07:00	john@daylilyinformatics.com	2025-07-03 07:50:10.941052-07	UPDATE	\N	\N	\N	f	f
823101f3-f109-49a8-a92d-48016de7951c	generic_instance	modified_dt	c01298e9-f706-48dd-b71e-7666927fccf3	FI7111	2025-07-03 07:50:10.722424-07	2025-07-03T07:50:10.941052-07:00	john@daylilyinformatics.com	2025-07-03 07:50:10.941052-07	UPDATE	\N	\N	\N	f	f
2bbc58a8-0274-4e0c-8149-3664653c2fdf	generic_instance_lineage	\N	f8fc87a0-3ca8-4988-93d6-3b3611c8376c	GL127	\N	\N	john@daylilyinformatics.com	2025-07-03 07:50:11.636939-07	INSERT	\N	\N	\N	f	f
a7a8976e-1000-420e-8bdb-6d21e618f6a3	generic_instance	\N	9e89f061-44e0-4163-95f7-03de5e7d595b	FI7112	\N	\N	john@daylilyinformatics.com	2025-07-03 07:50:11.650525-07	INSERT	\N	\N	\N	f	f
676d8e14-8317-4273-98c1-d6047170d76a	generic_instance	json_addl	9e89f061-44e0-4163-95f7-03de5e7d595b	FI7112	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T145008", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T145008", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:50:11.845317-07	UPDATE	\N	\N	\N	f	f
64220e7b-fe76-4db4-8243-3fe66b7c4235	generic_instance	created_dt	9e89f061-44e0-4163-95f7-03de5e7d595b	FI7112	2025-07-03 07:50:11.650525-07	2025-07-03T07:50:11.650525-07:00	john@daylilyinformatics.com	2025-07-03 07:50:11.845317-07	UPDATE	\N	\N	\N	f	f
63f72b68-f223-48cb-b6f5-0e01eca10e8b	generic_instance	modified_dt	9e89f061-44e0-4163-95f7-03de5e7d595b	FI7112	2025-07-03 07:50:11.650525-07	2025-07-03T07:50:11.845317-07:00	john@daylilyinformatics.com	2025-07-03 07:50:11.845317-07	UPDATE	\N	\N	\N	f	f
988e3ed4-4c0e-4611-88b5-661deb700cde	generic_instance	json_addl	9e89f061-44e0-4163-95f7-03de5e7d595b	FI7112	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "example_file", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T145008", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "", "current_s3_bucket_name": "daylily-dewey-0", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7112.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7112.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T145008", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	john@daylilyinformatics.com	2025-07-03 07:50:11.952358-07	UPDATE	\N	\N	\N	f	f
181d3470-4bb8-44d5-9c62-3173c7b748fc	generic_instance	created_dt	9e89f061-44e0-4163-95f7-03de5e7d595b	FI7112	2025-07-03 07:50:11.650525-07	2025-07-03T07:50:11.650525-07:00	john@daylilyinformatics.com	2025-07-03 07:50:11.952358-07	UPDATE	\N	\N	\N	f	f
646a4e80-eb9c-432d-b84f-b69a9853468d	generic_instance	modified_dt	9e89f061-44e0-4163-95f7-03de5e7d595b	FI7112	2025-07-03 07:50:11.845317-07	2025-07-03T07:50:11.952358-07:00	john@daylilyinformatics.com	2025-07-03 07:50:11.952358-07	UPDATE	\N	\N	\N	f	f
cbf64c48-874d-48da-89a2-9b72aeb21ef5	generic_instance_lineage	\N	e7149242-57d5-4d9f-9211-d25bccb90e35	GL128	\N	\N	john@daylilyinformatics.com	2025-07-03 07:50:12.86093-07	INSERT	\N	\N	\N	f	f
09b78971-c507-4395-a1a3-811f958f8c35	generic_instance	\N	3e50c2f7-5e0b-4ae9-b721-cc81d94c22de	CX1	\N	\N	daylily	2025-07-03 08:21:02.222701-07	INSERT	\N	\N	\N	f	f
900f2f0a-5c3f-4f74-b657-0328fb05837c	generic_instance	\N	dacc9c78-e115-4376-8a8c-9869504d64a7	CWX1	\N	\N	daylily	2025-07-03 08:21:02.273663-07	INSERT	\N	\N	\N	f	f
a10a76b3-8328-4141-8485-851c21f61305	generic_instance_lineage	\N	17d06cc1-5d17-477f-99e1-b734663d9520	GL179	\N	\N	daylily	2025-07-03 08:21:02.660099-07	INSERT	\N	\N	\N	f	f
f8818f60-8f13-411c-8277-be96bcac13d3	generic_instance	\N	a0017668-1487-4923-bf46-02ad0e4adae5	MRX24	\N	\N	daylily	2025-07-03 08:21:02.660099-07	INSERT	\N	\N	\N	f	f
6f671af7-1e16-4716-b0fa-3734e650e413	generic_instance_lineage	\N	94c423d6-8770-44ec-b075-685b53f7ee45	GL180	\N	\N	daylily	2025-07-03 08:21:02.664624-07	INSERT	\N	\N	\N	f	f
3c16e279-588c-4867-a68e-cd7fbd45f2b2	generic_instance	\N	0e8b6bac-2a70-473a-a382-8230c530bc86	MRX25	\N	\N	daylily	2025-07-03 08:21:02.664624-07	INSERT	\N	\N	\N	f	f
ae89a49c-8390-4c6e-bbe4-4ab66f2f1f95	generic_instance_lineage	\N	7f6140fa-6011-44e2-ba73-95d1654bad7e	GL181	\N	\N	daylily	2025-07-03 08:21:02.668815-07	INSERT	\N	\N	\N	f	f
c297f3a5-d122-4e63-9fb5-baed2a1030a6	generic_instance	\N	b3c512e2-3442-47bf-a7ae-228de3b99ea8	MRX26	\N	\N	daylily	2025-07-03 08:21:02.668815-07	INSERT	\N	\N	\N	f	f
02c55176-7620-43cc-9ebf-e210836f8d53	generic_instance_lineage	\N	69b4f8ff-fbbe-4a23-938e-27a8f837c23d	GL182	\N	\N	daylily	2025-07-03 08:21:02.672962-07	INSERT	\N	\N	\N	f	f
6a9613a7-ba3c-4309-99b2-8bd8944de667	generic_instance	\N	bf8c8f2e-8303-4213-a40f-25a7cce691d0	MRX27	\N	\N	daylily	2025-07-03 08:21:02.672962-07	INSERT	\N	\N	\N	f	f
336bedaa-e804-4284-8997-5e79d438f9f7	generic_instance_lineage	\N	a3f7dc99-e0bb-458d-a8d6-a1fb2260d74e	GL183	\N	\N	daylily	2025-07-03 08:21:02.679104-07	INSERT	\N	\N	\N	f	f
651e8eb3-ba54-47b7-9405-416117fddd3d	generic_instance	\N	848b30bc-7457-4760-8ed6-4733e1c9bceb	MRX28	\N	\N	daylily	2025-07-03 08:21:02.679104-07	INSERT	\N	\N	\N	f	f
476bd0fb-d86c-4a48-8629-7ce3037ee42e	generic_instance	json_addl	dacc9c78-e115-4376-8a8c-9869504d64a7	CWX1	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.279459-07	UPDATE	\N	\N	\N	f	f
a83647f3-1afd-4e84-bf74-77822633d664	generic_instance	created_dt	dacc9c78-e115-4376-8a8c-9869504d64a7	CWX1	2025-07-03 08:21:02.273663-07	2025-07-03T08:21:02.273663-07:00	daylily	2025-07-03 08:21:02.279459-07	UPDATE	\N	\N	\N	f	f
7b8d62ed-290e-4416-866e-0859b066316b	generic_instance	modified_dt	dacc9c78-e115-4376-8a8c-9869504d64a7	CWX1	2025-07-03 08:21:02.273663-07	2025-07-03T08:21:02.279459-07:00	daylily	2025-07-03 08:21:02.279459-07	UPDATE	\N	\N	\N	f	f
94fa1c17-a300-43d1-a8df-969080702e19	generic_instance_lineage	\N	9dc4ad96-63ce-4ec1-942f-be5d6df671e7	GL129	\N	\N	daylily	2025-07-03 08:21:02.293073-07	INSERT	\N	\N	\N	f	f
e15998b1-9be1-44a3-b23b-374bd256a29a	generic_instance	\N	438c11ad-e609-4041-aa3e-75e8f719cef9	CWX2	\N	\N	daylily	2025-07-03 08:21:02.293073-07	INSERT	\N	\N	\N	f	f
9ac5837b-4a72-49ff-a9ab-43dee3db798a	generic_instance	json_addl	438c11ad-e609-4041-aa3e-75e8f719cef9	CWX2	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.305077-07	UPDATE	\N	\N	\N	f	f
dd168539-65ae-4473-ab58-94beebffe788	generic_instance	created_dt	438c11ad-e609-4041-aa3e-75e8f719cef9	CWX2	2025-07-03 08:21:02.293073-07	2025-07-03T08:21:02.293073-07:00	daylily	2025-07-03 08:21:02.305077-07	UPDATE	\N	\N	\N	f	f
c34d0316-9183-49df-a636-102b3567ca15	generic_instance	modified_dt	438c11ad-e609-4041-aa3e-75e8f719cef9	CWX2	2025-07-03 08:21:02.293073-07	2025-07-03T08:21:02.305077-07:00	daylily	2025-07-03 08:21:02.305077-07	UPDATE	\N	\N	\N	f	f
ab4c7724-a800-4e2a-9215-85a384e11e6b	generic_instance_lineage	\N	e20e6d8a-b365-48ba-bac5-69eea0985427	GL130	\N	\N	daylily	2025-07-03 08:21:02.30858-07	INSERT	\N	\N	\N	f	f
1eb518b7-b9bc-4328-9ca8-2aa6a51b32bd	generic_instance	\N	f4b245da-a7f5-4a87-b23f-e24f173695d1	CWX3	\N	\N	daylily	2025-07-03 08:21:02.30858-07	INSERT	\N	\N	\N	f	f
071435fd-581b-4960-b5f9-c10592ed9614	generic_instance	json_addl	f4b245da-a7f5-4a87-b23f-e24f173695d1	CWX3	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.316622-07	UPDATE	\N	\N	\N	f	f
b00f412b-78d0-4827-884c-cb99a183a0a7	generic_instance	created_dt	f4b245da-a7f5-4a87-b23f-e24f173695d1	CWX3	2025-07-03 08:21:02.30858-07	2025-07-03T08:21:02.30858-07:00	daylily	2025-07-03 08:21:02.316622-07	UPDATE	\N	\N	\N	f	f
c0022f8c-cba8-4ad7-a509-7405ddc227e3	generic_instance	modified_dt	f4b245da-a7f5-4a87-b23f-e24f173695d1	CWX3	2025-07-03 08:21:02.30858-07	2025-07-03T08:21:02.316622-07:00	daylily	2025-07-03 08:21:02.316622-07	UPDATE	\N	\N	\N	f	f
167c3834-935d-4a0a-9eeb-55af1eb68f1e	generic_instance_lineage	\N	15cfe66c-4119-467b-8176-23876dc57dcf	GL131	\N	\N	daylily	2025-07-03 08:21:02.318139-07	INSERT	\N	\N	\N	f	f
f27b9514-7b36-4e24-b712-1f2473a2e495	generic_instance	\N	9abd0e47-97f9-4d8e-ac09-cec264f3ee0d	CWX4	\N	\N	daylily	2025-07-03 08:21:02.318139-07	INSERT	\N	\N	\N	f	f
920a7375-4a84-4f5f-bb28-50223af00f31	generic_instance_lineage	\N	5c1b0cf0-f7e1-4556-885d-c258129f2d93	GL184	\N	\N	daylily	2025-07-03 08:21:02.685324-07	INSERT	\N	\N	\N	f	f
e46d0065-d08e-4614-8b1e-0795a0181429	generic_instance	\N	2fc73838-fb40-4247-9446-971fd96cff46	CX2	\N	\N	daylily	2025-07-03 08:21:02.688189-07	INSERT	\N	\N	\N	f	f
3dcc70ba-e4fc-4688-ae39-64a87c0dac6d	generic_instance	\N	b98e9974-de3d-498d-8371-53ec167ca66f	CWX29	\N	\N	daylily	2025-07-03 08:21:02.692069-07	INSERT	\N	\N	\N	f	f
b381ad7c-790b-46f5-8a30-7e6d4aa11ed0	generic_instance_lineage	\N	e3acbae5-b7cd-41ac-aebf-25d0f233795a	GL353	\N	\N	daylily	2025-07-03 08:23:08.998181-07	INSERT	\N	\N	\N	f	f
36d17c77-1767-4a2d-bc45-a8e08e12135b	generic_instance	\N	859f2a12-8212-4697-893b-19aee54a3c15	MRX108	\N	\N	daylily	2025-07-03 08:23:08.998181-07	INSERT	\N	\N	\N	f	f
e61cfbbf-7156-4397-986f-0bec47924745	generic_instance_lineage	\N	faf2bc63-372f-4d73-ae5b-b7cb93fe8190	GL354	\N	\N	daylily	2025-07-03 08:23:09.002359-07	INSERT	\N	\N	\N	f	f
caab12ac-19ca-4bae-8029-9cce0f376a34	generic_instance	\N	9fe5d8ab-c3a9-4850-b662-07d1349c1d97	MRX109	\N	\N	daylily	2025-07-03 08:23:09.002359-07	INSERT	\N	\N	\N	f	f
20bf0704-1bd9-4ea8-852e-83d7475af890	generic_instance	json_addl	9abd0e47-97f9-4d8e-ac09-cec264f3ee0d	CWX4	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.324744-07	UPDATE	\N	\N	\N	f	f
c03f6a9e-4b6b-43be-85e5-4a0de5e4915c	generic_instance	created_dt	9abd0e47-97f9-4d8e-ac09-cec264f3ee0d	CWX4	2025-07-03 08:21:02.318139-07	2025-07-03T08:21:02.318139-07:00	daylily	2025-07-03 08:21:02.324744-07	UPDATE	\N	\N	\N	f	f
0d70014e-c4b0-4f9b-be9a-8b4929054aca	generic_instance	modified_dt	9abd0e47-97f9-4d8e-ac09-cec264f3ee0d	CWX4	2025-07-03 08:21:02.318139-07	2025-07-03T08:21:02.324744-07:00	daylily	2025-07-03 08:21:02.324744-07	UPDATE	\N	\N	\N	f	f
4d4eb3a7-6dea-449c-a8af-881de6d487d0	generic_instance_lineage	\N	92e7fcc0-f025-4518-a064-87e927e507d5	GL132	\N	\N	daylily	2025-07-03 08:21:02.326916-07	INSERT	\N	\N	\N	f	f
3cb3f7b1-2309-4687-822b-81e42580be76	generic_instance	\N	a339e966-d124-4730-bf7b-f27ca87c4e8d	CWX5	\N	\N	daylily	2025-07-03 08:21:02.326916-07	INSERT	\N	\N	\N	f	f
caaf7666-ade4-47f7-a0fa-3573184615df	generic_instance	json_addl	a339e966-d124-4730-bf7b-f27ca87c4e8d	CWX5	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.333175-07	UPDATE	\N	\N	\N	f	f
b9f7b55d-e095-43dd-bd7d-1a272aece8d2	generic_instance	created_dt	a339e966-d124-4730-bf7b-f27ca87c4e8d	CWX5	2025-07-03 08:21:02.326916-07	2025-07-03T08:21:02.326916-07:00	daylily	2025-07-03 08:21:02.333175-07	UPDATE	\N	\N	\N	f	f
f38e8861-9fe9-4198-baf6-e16cb9c7f0ed	generic_instance	modified_dt	a339e966-d124-4730-bf7b-f27ca87c4e8d	CWX5	2025-07-03 08:21:02.326916-07	2025-07-03T08:21:02.333175-07:00	daylily	2025-07-03 08:21:02.333175-07	UPDATE	\N	\N	\N	f	f
7d930bc6-35c6-4386-be03-a252f858b917	generic_instance_lineage	\N	4cf3557d-3cb8-4572-9720-a3a02f6ff8e2	GL133	\N	\N	daylily	2025-07-03 08:21:02.335164-07	INSERT	\N	\N	\N	f	f
bb112689-5560-4e89-825e-f2b068a1613a	generic_instance	\N	486a95ec-28d0-4b5a-b33e-0de10f65806c	CWX6	\N	\N	daylily	2025-07-03 08:21:02.335164-07	INSERT	\N	\N	\N	f	f
b8a2c677-cf80-4869-9c67-73c67c259f0d	generic_instance	json_addl	486a95ec-28d0-4b5a-b33e-0de10f65806c	CWX6	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.341516-07	UPDATE	\N	\N	\N	f	f
5c9ee7b3-0f45-4323-99a5-67824b8b46df	generic_instance	created_dt	486a95ec-28d0-4b5a-b33e-0de10f65806c	CWX6	2025-07-03 08:21:02.335164-07	2025-07-03T08:21:02.335164-07:00	daylily	2025-07-03 08:21:02.341516-07	UPDATE	\N	\N	\N	f	f
169a0c31-55f8-4f7c-8236-390412ab3a1c	generic_instance	modified_dt	486a95ec-28d0-4b5a-b33e-0de10f65806c	CWX6	2025-07-03 08:21:02.335164-07	2025-07-03T08:21:02.341516-07:00	daylily	2025-07-03 08:21:02.341516-07	UPDATE	\N	\N	\N	f	f
acdd0656-3e41-44e7-8831-279cf645d60d	generic_instance_lineage	\N	33f91995-d65d-4c81-bff9-9b37dc831d72	GL134	\N	\N	daylily	2025-07-03 08:21:02.343198-07	INSERT	\N	\N	\N	f	f
a8609c34-c09e-4d43-b985-1e146e7fc937	generic_instance	\N	dee275bb-a13a-455a-879c-29bcddda4116	CWX7	\N	\N	daylily	2025-07-03 08:21:02.343198-07	INSERT	\N	\N	\N	f	f
86b38a6d-65cf-40f0-8908-d07b6212d46d	generic_instance_lineage	\N	490193a3-525a-4aa2-9895-913a08e17ff6	GL235	\N	\N	daylily	2025-07-03 08:21:02.953494-07	INSERT	\N	\N	\N	f	f
791308bb-95e3-4a07-8795-67f006844182	generic_instance	\N	d5438a6a-5b90-40be-a73d-c418e343a0d5	MRX52	\N	\N	daylily	2025-07-03 08:21:02.953494-07	INSERT	\N	\N	\N	f	f
f0bda722-8e57-4a47-93c5-4b09774df285	generic_instance_lineage	\N	a6626ad6-cbe4-4085-9a33-bc601df27938	GL236	\N	\N	daylily	2025-07-03 08:21:02.957776-07	INSERT	\N	\N	\N	f	f
0e9d95fd-85b0-4b3a-a808-a8ce2cac2349	generic_instance	\N	f1bd3edd-8879-473c-a1ad-d70687254885	MRX53	\N	\N	daylily	2025-07-03 08:21:02.957776-07	INSERT	\N	\N	\N	f	f
2af0b2e1-b1c7-4449-978e-368bd707d084	generic_instance_lineage	\N	cfe1b78e-ffad-4337-be90-af1c9d06c9e7	GL237	\N	\N	daylily	2025-07-03 08:21:02.962054-07	INSERT	\N	\N	\N	f	f
0bf2379a-a937-416e-9d86-9e1c8a5d3ced	generic_instance	\N	a067d8df-fe65-4937-a679-ce28eac8f132	MRX54	\N	\N	daylily	2025-07-03 08:21:02.962054-07	INSERT	\N	\N	\N	f	f
a34f68c1-dcc5-4bf6-a102-ab74d4e0afd8	generic_instance_lineage	\N	9357d737-db54-47c5-baeb-8e0ab20dacf5	GL238	\N	\N	daylily	2025-07-03 08:21:02.966483-07	INSERT	\N	\N	\N	f	f
1b50a475-4b19-499c-aa97-6795d6ab8f2b	generic_instance	json_addl	dee275bb-a13a-455a-879c-29bcddda4116	CWX7	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.349401-07	UPDATE	\N	\N	\N	f	f
c30a84c2-3fc7-4935-ab2d-164f0c25621c	generic_instance	created_dt	dee275bb-a13a-455a-879c-29bcddda4116	CWX7	2025-07-03 08:21:02.343198-07	2025-07-03T08:21:02.343198-07:00	daylily	2025-07-03 08:21:02.349401-07	UPDATE	\N	\N	\N	f	f
5438f8c2-85cb-4000-af65-5edec5595a9d	generic_instance	modified_dt	dee275bb-a13a-455a-879c-29bcddda4116	CWX7	2025-07-03 08:21:02.343198-07	2025-07-03T08:21:02.349401-07:00	daylily	2025-07-03 08:21:02.349401-07	UPDATE	\N	\N	\N	f	f
415af3a8-1540-4060-9419-f70c0eca8121	generic_instance_lineage	\N	952499b9-8efd-4df2-a63c-96321b52cfbd	GL135	\N	\N	daylily	2025-07-03 08:21:02.351211-07	INSERT	\N	\N	\N	f	f
46339ce7-feb6-4401-ac04-277411b86d51	generic_instance	\N	b79a21a8-b8d0-4b24-8b54-01005554bf43	CWX8	\N	\N	daylily	2025-07-03 08:21:02.351211-07	INSERT	\N	\N	\N	f	f
ff938f26-9dfa-4a3b-9f10-ee47fb665390	generic_instance	json_addl	b79a21a8-b8d0-4b24-8b54-01005554bf43	CWX8	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.356925-07	UPDATE	\N	\N	\N	f	f
f48fa60d-9b56-4c11-933c-0b0d537d593e	generic_instance	created_dt	b79a21a8-b8d0-4b24-8b54-01005554bf43	CWX8	2025-07-03 08:21:02.351211-07	2025-07-03T08:21:02.351211-07:00	daylily	2025-07-03 08:21:02.356925-07	UPDATE	\N	\N	\N	f	f
57e84aef-4770-4a5b-8e16-0ec60453c330	generic_instance	modified_dt	b79a21a8-b8d0-4b24-8b54-01005554bf43	CWX8	2025-07-03 08:21:02.351211-07	2025-07-03T08:21:02.356925-07:00	daylily	2025-07-03 08:21:02.356925-07	UPDATE	\N	\N	\N	f	f
8325aa47-0ab8-49d1-a6bb-7fb5e087a18c	generic_instance_lineage	\N	0a698f74-6ffd-470a-a995-96c5fbcbce67	GL136	\N	\N	daylily	2025-07-03 08:21:02.359286-07	INSERT	\N	\N	\N	f	f
94a404f8-306c-4750-aace-043714e9b66c	generic_instance	\N	ad8e242c-7636-42b1-9284-401ea589e648	CWX9	\N	\N	daylily	2025-07-03 08:21:02.359286-07	INSERT	\N	\N	\N	f	f
5b388aff-4c5e-4c32-a3d2-0ccf1d073552	generic_instance	json_addl	ad8e242c-7636-42b1-9284-401ea589e648	CWX9	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.36539-07	UPDATE	\N	\N	\N	f	f
7acbd658-5df0-4135-86eb-23d8fddb699b	generic_instance	created_dt	ad8e242c-7636-42b1-9284-401ea589e648	CWX9	2025-07-03 08:21:02.359286-07	2025-07-03T08:21:02.359286-07:00	daylily	2025-07-03 08:21:02.36539-07	UPDATE	\N	\N	\N	f	f
e48ab09e-4cd7-4f29-b3ec-bb6061a677d4	generic_instance	modified_dt	ad8e242c-7636-42b1-9284-401ea589e648	CWX9	2025-07-03 08:21:02.359286-07	2025-07-03T08:21:02.36539-07:00	daylily	2025-07-03 08:21:02.36539-07	UPDATE	\N	\N	\N	f	f
5d0633d6-13fe-4cd1-86b3-ec1afdb1539e	generic_instance_lineage	\N	3e26d0cc-9cf4-41a0-ae0e-2035e6f13df8	GL137	\N	\N	daylily	2025-07-03 08:21:02.366841-07	INSERT	\N	\N	\N	f	f
44b7eea5-1569-4d1f-bca8-f0a9f8d837c0	generic_instance	\N	84bbdac6-5ab1-480c-8715-8dd7271b19bf	CWX10	\N	\N	daylily	2025-07-03 08:21:02.366841-07	INSERT	\N	\N	\N	f	f
6d01f215-c9c2-4b10-89ce-2651546cf298	generic_instance	\N	8070ee33-11a2-441b-8f99-11df1436e5af	MRX55	\N	\N	daylily	2025-07-03 08:21:02.966483-07	INSERT	\N	\N	\N	f	f
9d7f92ed-bb50-4e99-8be0-632b0e35f315	generic_instance_lineage	\N	2d826236-66eb-4747-b86e-4269cef09d35	GL239	\N	\N	daylily	2025-07-03 08:21:02.978671-07	INSERT	\N	\N	\N	f	f
870c4cd8-8452-4bd8-83f5-08110d844a61	generic_instance	\N	a2e7703b-a781-4d45-9775-6e3812f6f5c7	MRX56	\N	\N	daylily	2025-07-03 08:21:02.978671-07	INSERT	\N	\N	\N	f	f
436448e2-73a6-4366-9f7c-9c774a74657f	generic_instance_lineage	\N	bc3fe620-0fa6-42f9-b3d7-689d16400cfe	GL240	\N	\N	daylily	2025-07-03 08:21:02.98473-07	INSERT	\N	\N	\N	f	f
952ccbce-cb32-4d11-9497-11d0ff579683	generic_instance_lineage	\N	8813022a-cc82-4d68-b5da-a61f008d586a	GL355	\N	\N	daylily	2025-07-03 08:23:09.006676-07	INSERT	\N	\N	\N	f	f
b2ac768f-9859-469d-955a-1f0936a6d71e	generic_instance	\N	901f56f9-59e7-4b78-80b9-6d6f62d6aed7	MRX110	\N	\N	daylily	2025-07-03 08:23:09.006676-07	INSERT	\N	\N	\N	f	f
2f5ea497-4e33-47ae-8b84-bbd60cc13d55	generic_instance_lineage	\N	a2a67686-4243-495c-ac14-47921d1125bb	GL356	\N	\N	daylily	2025-07-03 08:23:09.011055-07	INSERT	\N	\N	\N	f	f
ccfc5023-6ba8-414a-be67-28be38cd3a05	generic_instance	json_addl	84bbdac6-5ab1-480c-8715-8dd7271b19bf	CWX10	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.372549-07	UPDATE	\N	\N	\N	f	f
210f2a67-47ba-4773-9f5f-fd834bd7fd4c	generic_instance	created_dt	84bbdac6-5ab1-480c-8715-8dd7271b19bf	CWX10	2025-07-03 08:21:02.366841-07	2025-07-03T08:21:02.366841-07:00	daylily	2025-07-03 08:21:02.372549-07	UPDATE	\N	\N	\N	f	f
df9f3d25-29e1-422f-9a46-76c68a91e0ba	generic_instance	modified_dt	84bbdac6-5ab1-480c-8715-8dd7271b19bf	CWX10	2025-07-03 08:21:02.366841-07	2025-07-03T08:21:02.372549-07:00	daylily	2025-07-03 08:21:02.372549-07	UPDATE	\N	\N	\N	f	f
3e1091ab-a353-4c7f-84b2-851b1c5e2126	generic_instance_lineage	\N	62156688-56dd-4c18-996a-1cf187c3f186	GL138	\N	\N	daylily	2025-07-03 08:21:02.374692-07	INSERT	\N	\N	\N	f	f
70a5c6bd-902e-4f48-a769-ceb181d0d7db	generic_instance	\N	a1e066c9-b6ad-4644-a288-029b6337c599	CWX11	\N	\N	daylily	2025-07-03 08:21:02.374692-07	INSERT	\N	\N	\N	f	f
be63033d-12ed-4e4e-8bbf-87ea35f41d44	generic_instance	json_addl	a1e066c9-b6ad-4644-a288-029b6337c599	CWX11	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.379966-07	UPDATE	\N	\N	\N	f	f
640f1eb8-0f58-4696-a27e-d468141b8af7	generic_instance	created_dt	a1e066c9-b6ad-4644-a288-029b6337c599	CWX11	2025-07-03 08:21:02.374692-07	2025-07-03T08:21:02.374692-07:00	daylily	2025-07-03 08:21:02.379966-07	UPDATE	\N	\N	\N	f	f
e88e45ef-ba25-441e-b04f-f4a0fb3f3bde	generic_instance	modified_dt	a1e066c9-b6ad-4644-a288-029b6337c599	CWX11	2025-07-03 08:21:02.374692-07	2025-07-03T08:21:02.379966-07:00	daylily	2025-07-03 08:21:02.379966-07	UPDATE	\N	\N	\N	f	f
5908d3c9-7c87-4aec-9f6f-2b78673f0f3a	generic_instance_lineage	\N	7c4c38ac-e94b-4152-9f9d-b55dec1291a1	GL139	\N	\N	daylily	2025-07-03 08:21:02.381424-07	INSERT	\N	\N	\N	f	f
ed1a2726-7ed6-4d13-986c-673d987fe85d	generic_instance	\N	4b07afc5-df46-4a85-8d24-2e74bb943460	CWX12	\N	\N	daylily	2025-07-03 08:21:02.381424-07	INSERT	\N	\N	\N	f	f
7a9e2458-93b8-42d1-876f-c72aeaead5ac	generic_instance	json_addl	4b07afc5-df46-4a85-8d24-2e74bb943460	CWX12	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.386963-07	UPDATE	\N	\N	\N	f	f
f42623eb-2851-4d8d-829d-ef63235c519d	generic_instance	created_dt	4b07afc5-df46-4a85-8d24-2e74bb943460	CWX12	2025-07-03 08:21:02.381424-07	2025-07-03T08:21:02.381424-07:00	daylily	2025-07-03 08:21:02.386963-07	UPDATE	\N	\N	\N	f	f
1db888f0-9c79-41c9-ad59-c167b929bcf8	generic_instance	modified_dt	4b07afc5-df46-4a85-8d24-2e74bb943460	CWX12	2025-07-03 08:21:02.381424-07	2025-07-03T08:21:02.386963-07:00	daylily	2025-07-03 08:21:02.386963-07	UPDATE	\N	\N	\N	f	f
c4364587-004b-4478-b7a4-32d08221bab4	generic_instance_lineage	\N	b496e277-846f-48d0-9473-a24c303cf372	GL140	\N	\N	daylily	2025-07-03 08:21:02.388919-07	INSERT	\N	\N	\N	f	f
12057877-e5b9-41da-ac98-d6aff2420177	generic_instance	\N	01c9d691-4dd1-4a31-88b8-328c9e47a8ec	CWX13	\N	\N	daylily	2025-07-03 08:21:02.388919-07	INSERT	\N	\N	\N	f	f
dc3be99d-73fa-43b0-86e2-b83849214f31	generic_instance	\N	d4b6618e-692a-4952-b7f5-164c2c7e45a2	CX3	\N	\N	daylily	2025-07-03 08:21:02.98976-07	INSERT	\N	\N	\N	f	f
7a0ba692-01de-4cb5-8326-50c758559c37	generic_instance	\N	edd473cb-2593-4ef4-a328-bdbcbae49d87	MCX1	\N	\N	daylily	2025-07-03 08:21:03.013724-07	INSERT	\N	\N	\N	f	f
11231fa0-a0a6-4e11-812e-7d61dda4f69e	generic_instance	\N	c64e172a-31c7-41d4-83e5-a592747c1107	MRX111	\N	\N	daylily	2025-07-03 08:23:09.011055-07	INSERT	\N	\N	\N	f	f
10e2e1d7-bffd-4553-b839-844fdb004c5d	generic_instance_lineage	\N	bd06de39-b100-4b10-8389-caf6c681f7a2	GL357	\N	\N	daylily	2025-07-03 08:23:09.01544-07	INSERT	\N	\N	\N	f	f
564175d8-3bf5-4fd3-886c-7390461a45d9	generic_instance	\N	173f906f-f360-4aa8-bf05-20fed68593d5	MRX112	\N	\N	daylily	2025-07-03 08:23:09.01544-07	INSERT	\N	\N	\N	f	f
2b16f163-69a7-4dd8-8e7b-a761c373522d	generic_instance_lineage	\N	e6161a8c-e2b8-4c00-b382-3dac9c037ec1	GL358	\N	\N	daylily	2025-07-03 08:23:09.019921-07	INSERT	\N	\N	\N	f	f
5e2582f9-75c8-42f8-bd5f-9059cb207f48	generic_instance	\N	36a78f96-96bc-4162-9372-16d115b2e6c3	CX9	\N	\N	daylily	2025-07-03 08:23:09.022787-07	INSERT	\N	\N	\N	f	f
266c3f99-d954-4fff-a2b7-fea18af6c8f6	generic_instance	json_addl	01c9d691-4dd1-4a31-88b8-328c9e47a8ec	CWX13	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.394171-07	UPDATE	\N	\N	\N	f	f
fc3da957-ae93-49fa-ae90-57f8445f1470	generic_instance	created_dt	01c9d691-4dd1-4a31-88b8-328c9e47a8ec	CWX13	2025-07-03 08:21:02.388919-07	2025-07-03T08:21:02.388919-07:00	daylily	2025-07-03 08:21:02.394171-07	UPDATE	\N	\N	\N	f	f
8eb36b2f-5bfc-484d-8c1c-b447b047bdf3	generic_instance	modified_dt	01c9d691-4dd1-4a31-88b8-328c9e47a8ec	CWX13	2025-07-03 08:21:02.388919-07	2025-07-03T08:21:02.394171-07:00	daylily	2025-07-03 08:21:02.394171-07	UPDATE	\N	\N	\N	f	f
5c3ab605-a865-4830-8172-e22d0bc5549a	generic_instance_lineage	\N	8ee8f584-a084-4e05-9d27-c472da7df807	GL141	\N	\N	daylily	2025-07-03 08:21:02.398253-07	INSERT	\N	\N	\N	f	f
18ba1e6f-1598-45b3-b4b7-acae2d81893e	generic_instance	\N	e3a263a5-ad0e-42e5-90f7-3eea9eb68be0	CWX14	\N	\N	daylily	2025-07-03 08:21:02.398253-07	INSERT	\N	\N	\N	f	f
77db5f73-bcd3-436e-ae2c-760a9231fd16	generic_instance	json_addl	e3a263a5-ad0e-42e5-90f7-3eea9eb68be0	CWX14	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.405579-07	UPDATE	\N	\N	\N	f	f
e9cf062e-b863-4256-a725-dda55502e0fa	generic_instance	created_dt	e3a263a5-ad0e-42e5-90f7-3eea9eb68be0	CWX14	2025-07-03 08:21:02.398253-07	2025-07-03T08:21:02.398253-07:00	daylily	2025-07-03 08:21:02.405579-07	UPDATE	\N	\N	\N	f	f
1633e3f8-2f43-4918-b441-9108cce39aec	generic_instance	modified_dt	e3a263a5-ad0e-42e5-90f7-3eea9eb68be0	CWX14	2025-07-03 08:21:02.398253-07	2025-07-03T08:21:02.405579-07:00	daylily	2025-07-03 08:21:02.405579-07	UPDATE	\N	\N	\N	f	f
f79f5b66-9acc-4f19-8c61-2df82a04ef20	generic_instance_lineage	\N	3735d12b-7137-4e39-b871-89926008d49f	GL142	\N	\N	daylily	2025-07-03 08:21:02.407525-07	INSERT	\N	\N	\N	f	f
3cfa8a2c-fba6-4f06-b866-4774e0f17d2a	generic_instance	\N	005041ba-80b0-494a-9982-4f02023c76b9	CWX15	\N	\N	daylily	2025-07-03 08:21:02.407525-07	INSERT	\N	\N	\N	f	f
7175c2e9-d793-41a6-bc21-f8735a025598	generic_instance	json_addl	005041ba-80b0-494a-9982-4f02023c76b9	CWX15	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.412588-07	UPDATE	\N	\N	\N	f	f
9b960874-608f-4c9a-9a05-29a3937871bb	generic_instance	created_dt	005041ba-80b0-494a-9982-4f02023c76b9	CWX15	2025-07-03 08:21:02.407525-07	2025-07-03T08:21:02.407525-07:00	daylily	2025-07-03 08:21:02.412588-07	UPDATE	\N	\N	\N	f	f
ee1ffab8-f404-4888-a42c-4042e1a308d7	generic_instance	modified_dt	005041ba-80b0-494a-9982-4f02023c76b9	CWX15	2025-07-03 08:21:02.407525-07	2025-07-03T08:21:02.412588-07:00	daylily	2025-07-03 08:21:02.412588-07	UPDATE	\N	\N	\N	f	f
58bd8e9e-dc9a-4781-aa95-fb3a305d97de	generic_instance_lineage	\N	f1438358-2b8f-4b56-9f5b-88098a9cd8dd	GL143	\N	\N	daylily	2025-07-03 08:21:02.413998-07	INSERT	\N	\N	\N	f	f
cfa2fa7d-115b-43fb-8ba8-73520f345349	generic_instance	\N	d0e326c1-0ca3-40e7-a7a3-58add70136f4	CWX16	\N	\N	daylily	2025-07-03 08:21:02.413998-07	INSERT	\N	\N	\N	f	f
e0c06171-bcab-46dc-9ee8-57ef58805da4	generic_instance	modified_dt	a20f9cb2-4002-4a69-bff6-d52924a67a7b	CWX57	2025-07-03 08:21:03.049675-07	2025-07-03T08:21:03.053177-07:00	daylily	2025-07-03 08:21:03.053177-07	UPDATE	\N	\N	\N	f	f
80225ca7-3428-4a64-b9f1-8b0a7a6fb701	generic_instance_lineage	\N	6c40ddce-4044-4bf5-90fe-a7f8927f532c	GL244	\N	\N	daylily	2025-07-03 08:21:03.054349-07	INSERT	\N	\N	\N	f	f
6d6c7751-c095-4d08-9afa-6a2db409781a	generic_instance	\N	1a35323d-f111-442b-958f-716b90b5d1f2	CWX58	\N	\N	daylily	2025-07-03 08:21:03.054349-07	INSERT	\N	\N	\N	f	f
64b69c4b-682a-427d-b293-0e2d315b5a43	generic_instance	\N	b806e72c-8000-4615-b23f-b63a79d2da4d	CWX113	\N	\N	daylily	2025-07-03 08:23:09.027734-07	INSERT	\N	\N	\N	f	f
649f3610-d1fb-4f06-b264-90d51ae4ee0e	generic_instance_lineage	\N	5af314ed-67f6-4618-9ffa-68a2f25fd586	GL574	\N	\N	daylily	2025-07-03 08:23:17.873982-07	INSERT	\N	\N	\N	f	f
a867c611-33f5-45f3-b3c9-ca7bb38d9f89	generic_instance	\N	5a55869e-cf2a-470b-b132-b2e375b2872c	MRX214	\N	\N	daylily	2025-07-03 08:23:17.873982-07	INSERT	\N	\N	\N	f	f
90273166-1a93-49e3-83fd-1bd2ac43a496	generic_instance	json_addl	d0e326c1-0ca3-40e7-a7a3-58add70136f4	CWX16	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.419312-07	UPDATE	\N	\N	\N	f	f
c1432051-6919-4223-ad5b-0013d2937c9d	generic_instance	created_dt	d0e326c1-0ca3-40e7-a7a3-58add70136f4	CWX16	2025-07-03 08:21:02.413998-07	2025-07-03T08:21:02.413998-07:00	daylily	2025-07-03 08:21:02.419312-07	UPDATE	\N	\N	\N	f	f
d281e720-940e-470d-860a-ca4f9b5bb034	generic_instance	modified_dt	d0e326c1-0ca3-40e7-a7a3-58add70136f4	CWX16	2025-07-03 08:21:02.413998-07	2025-07-03T08:21:02.419312-07:00	daylily	2025-07-03 08:21:02.419312-07	UPDATE	\N	\N	\N	f	f
ebc08ce6-a383-416b-8ffc-99edc324fd59	generic_instance_lineage	\N	203d92b9-a022-4355-8833-7446c5c90db4	GL144	\N	\N	daylily	2025-07-03 08:21:02.421024-07	INSERT	\N	\N	\N	f	f
45a90a11-f1ec-4a7f-ab1b-4f83949adb7d	generic_instance	\N	6413aa15-e457-4196-a660-67d30c22a768	CWX17	\N	\N	daylily	2025-07-03 08:21:02.421024-07	INSERT	\N	\N	\N	f	f
bbaf1f1c-a1e1-42bb-b651-8a5a7644d147	generic_instance	json_addl	6413aa15-e457-4196-a660-67d30c22a768	CWX17	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.426035-07	UPDATE	\N	\N	\N	f	f
b489a61b-6891-4c74-818e-e695e713726a	generic_instance	created_dt	6413aa15-e457-4196-a660-67d30c22a768	CWX17	2025-07-03 08:21:02.421024-07	2025-07-03T08:21:02.421024-07:00	daylily	2025-07-03 08:21:02.426035-07	UPDATE	\N	\N	\N	f	f
b4bde358-4bae-4e75-a8aa-80498e99a75b	generic_instance	modified_dt	6413aa15-e457-4196-a660-67d30c22a768	CWX17	2025-07-03 08:21:02.421024-07	2025-07-03T08:21:02.426035-07:00	daylily	2025-07-03 08:21:02.426035-07	UPDATE	\N	\N	\N	f	f
bc9891ab-6475-40a6-93af-f6c487805d8d	generic_instance_lineage	\N	0ec8e7dd-8845-47f7-9cd9-efaf356cad19	GL145	\N	\N	daylily	2025-07-03 08:21:02.427792-07	INSERT	\N	\N	\N	f	f
3334b87f-0a9b-468a-a283-4f82f165489e	generic_instance	\N	aabb851f-9f1b-44e8-86e2-2239eb90be7b	CWX18	\N	\N	daylily	2025-07-03 08:21:02.427792-07	INSERT	\N	\N	\N	f	f
5e775aff-0192-426e-9151-b488b6ccda8b	generic_instance	json_addl	aabb851f-9f1b-44e8-86e2-2239eb90be7b	CWX18	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.433789-07	UPDATE	\N	\N	\N	f	f
8a17aa84-182c-4061-9dda-23f9fe0b79b3	generic_instance	created_dt	aabb851f-9f1b-44e8-86e2-2239eb90be7b	CWX18	2025-07-03 08:21:02.427792-07	2025-07-03T08:21:02.427792-07:00	daylily	2025-07-03 08:21:02.433789-07	UPDATE	\N	\N	\N	f	f
2d1361bf-893d-4e03-a874-8bdca70c0f4e	generic_instance	modified_dt	aabb851f-9f1b-44e8-86e2-2239eb90be7b	CWX18	2025-07-03 08:21:02.427792-07	2025-07-03T08:21:02.433789-07:00	daylily	2025-07-03 08:21:02.433789-07	UPDATE	\N	\N	\N	f	f
4997f887-0b34-4b79-850a-37b71f6b9664	generic_instance_lineage	\N	b3b2f32a-fdbf-46d1-bd83-a426b3b8498b	GL146	\N	\N	daylily	2025-07-03 08:21:02.435195-07	INSERT	\N	\N	\N	f	f
be5379f2-c76b-4644-b034-04ef1e6a601a	generic_instance	\N	47df2675-749e-4933-b64a-98e9f92f9891	CWX19	\N	\N	daylily	2025-07-03 08:21:02.435195-07	INSERT	\N	\N	\N	f	f
7da9c49a-ad0a-47fd-b33d-1e0496139082	generic_instance	\N	6fd97ea6-93e2-4184-a45d-46f3431ca109	MRX61	\N	\N	daylily	2025-07-03 08:21:03.25557-07	INSERT	\N	\N	\N	f	f
f19ed702-b2de-4a98-b6dc-a0ffd8b13378	generic_instance_lineage	\N	2c105010-4a1a-473d-85a2-3e37982e6505	GL276	\N	\N	daylily	2025-07-03 08:21:03.260864-07	INSERT	\N	\N	\N	f	f
67b12bf8-6ee8-4a6f-9c29-06995a41b24c	generic_instance	\N	44ef5e8f-8ff4-402e-9ad6-300221cc3de9	MRX62	\N	\N	daylily	2025-07-03 08:21:03.260864-07	INSERT	\N	\N	\N	f	f
38154475-866d-41c6-91ef-f1786013b141	generic_instance_lineage	\N	22ec3501-74c9-4013-a586-e6a6b5abe88b	GL277	\N	\N	daylily	2025-07-03 08:21:03.266682-07	INSERT	\N	\N	\N	f	f
7c32b669-e153-4b46-95d5-65c39773f705	generic_instance	\N	70dc8c52-c634-4743-95a3-f7bc39a9cc12	MRX63	\N	\N	daylily	2025-07-03 08:21:03.266682-07	INSERT	\N	\N	\N	f	f
c77cdd13-fbe8-4b95-a5b0-0629621b17f6	generic_instance_lineage	\N	c6419250-2374-421e-bd13-c13f08c20792	GL278	\N	\N	daylily	2025-07-03 08:21:03.271948-07	INSERT	\N	\N	\N	f	f
a07ac2b9-5bf5-427c-960c-78c6c7da625d	generic_instance	\N	6d743c90-ecc6-4b3b-bcab-30278ece70c4	MRX64	\N	\N	daylily	2025-07-03 08:21:03.271948-07	INSERT	\N	\N	\N	f	f
da012051-3c16-4d5c-8f84-66e92d4fbe86	generic_instance	json_addl	47df2675-749e-4933-b64a-98e9f92f9891	CWX19	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.447507-07	UPDATE	\N	\N	\N	f	f
95af0a32-3237-4973-bfa1-824192f74213	generic_instance	created_dt	47df2675-749e-4933-b64a-98e9f92f9891	CWX19	2025-07-03 08:21:02.435195-07	2025-07-03T08:21:02.435195-07:00	daylily	2025-07-03 08:21:02.447507-07	UPDATE	\N	\N	\N	f	f
87e575d7-2d04-4bba-bd8b-9bb227941001	generic_instance	modified_dt	47df2675-749e-4933-b64a-98e9f92f9891	CWX19	2025-07-03 08:21:02.435195-07	2025-07-03T08:21:02.447507-07:00	daylily	2025-07-03 08:21:02.447507-07	UPDATE	\N	\N	\N	f	f
91508ff7-d315-446e-9086-f381baa945f7	generic_instance_lineage	\N	38e5e199-6846-415d-97f6-85e795b8daf5	GL147	\N	\N	daylily	2025-07-03 08:21:02.465875-07	INSERT	\N	\N	\N	f	f
6dbc7d57-9c18-4f5a-8d0a-6dc0ed5404bf	generic_instance	\N	1d487141-743b-4f01-ac33-bbdd81756836	CWX20	\N	\N	daylily	2025-07-03 08:21:02.465875-07	INSERT	\N	\N	\N	f	f
f35a76c3-0bce-4ae5-944d-0dc47ca27270	generic_instance	json_addl	1d487141-743b-4f01-ac33-bbdd81756836	CWX20	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.50052-07	UPDATE	\N	\N	\N	f	f
71fba059-6141-442f-8f51-1269f253222a	generic_instance	created_dt	1d487141-743b-4f01-ac33-bbdd81756836	CWX20	2025-07-03 08:21:02.465875-07	2025-07-03T08:21:02.465875-07:00	daylily	2025-07-03 08:21:02.50052-07	UPDATE	\N	\N	\N	f	f
53d4783f-db53-4649-8ee9-8d46ecb6ba1f	generic_instance	modified_dt	1d487141-743b-4f01-ac33-bbdd81756836	CWX20	2025-07-03 08:21:02.465875-07	2025-07-03T08:21:02.50052-07:00	daylily	2025-07-03 08:21:02.50052-07	UPDATE	\N	\N	\N	f	f
f44972bd-c500-4a8b-a90d-1bd791e2499e	generic_instance_lineage	\N	0f7d966a-51dd-4977-b50f-11513f15cdbb	GL148	\N	\N	daylily	2025-07-03 08:21:02.50501-07	INSERT	\N	\N	\N	f	f
a5a3db9e-ac20-486f-a3d6-79f557808ed1	generic_instance	\N	d2e4093e-86f0-4c14-b4de-2e90a1f635a5	CWX21	\N	\N	daylily	2025-07-03 08:21:02.50501-07	INSERT	\N	\N	\N	f	f
c5a4bee4-ff6f-4701-8908-38ae97afee7e	generic_instance	json_addl	d2e4093e-86f0-4c14-b4de-2e90a1f635a5	CWX21	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.513494-07	UPDATE	\N	\N	\N	f	f
8dd2f2cd-90a8-4feb-928e-8481c5ed33a6	generic_instance	created_dt	d2e4093e-86f0-4c14-b4de-2e90a1f635a5	CWX21	2025-07-03 08:21:02.50501-07	2025-07-03T08:21:02.50501-07:00	daylily	2025-07-03 08:21:02.513494-07	UPDATE	\N	\N	\N	f	f
b859a0dd-dd03-4485-9434-716af6e5f077	generic_instance	modified_dt	d2e4093e-86f0-4c14-b4de-2e90a1f635a5	CWX21	2025-07-03 08:21:02.50501-07	2025-07-03T08:21:02.513494-07:00	daylily	2025-07-03 08:21:02.513494-07	UPDATE	\N	\N	\N	f	f
3c046e78-0cc8-4576-ade4-58a2f5fb425a	generic_instance_lineage	\N	226492e2-9fbe-4157-829c-b326a3dcfff9	GL149	\N	\N	daylily	2025-07-03 08:21:02.514912-07	INSERT	\N	\N	\N	f	f
835fab9f-de4a-44e7-a0aa-4c17478f0515	generic_instance	\N	2ec41575-3c9e-41db-a210-0aa45aa7c05e	CWX22	\N	\N	daylily	2025-07-03 08:21:02.514912-07	INSERT	\N	\N	\N	f	f
f641b1df-edd4-42ad-8fd9-e5d7c1f55ea1	generic_instance_lineage	\N	ca8920ae-76fa-4082-9d4f-b4f84d940480	GL279	\N	\N	daylily	2025-07-03 08:21:03.277032-07	INSERT	\N	\N	\N	f	f
392c3a18-e515-4164-9d2c-526feb997a00	generic_instance	\N	800a223f-1a21-493f-a97c-5462854e77dd	MRX65	\N	\N	daylily	2025-07-03 08:21:03.277032-07	INSERT	\N	\N	\N	f	f
0fc694aa-f22f-49f3-a22b-b05f4f0cff68	generic_instance_lineage	\N	37f67b45-c251-4ca9-bca5-df5c948b6b71	GL280	\N	\N	daylily	2025-07-03 08:21:03.282022-07	INSERT	\N	\N	\N	f	f
fdd88710-3220-411f-a442-b82f9da4d336	generic_instance	\N	bb6bc18e-109f-4955-a46a-b909b4878979	MRX66	\N	\N	daylily	2025-07-03 08:21:03.282022-07	INSERT	\N	\N	\N	f	f
c5053f32-dd4d-4f1e-ad60-20dc647a9548	generic_instance_lineage	\N	ea70c007-16b1-4589-bbf7-5908e16e4046	GL281	\N	\N	daylily	2025-07-03 08:21:03.287395-07	INSERT	\N	\N	\N	f	f
a671f8fe-048d-437b-9eb9-f5a483e503ac	generic_instance	\N	92d63b59-0d46-4dcc-aadf-ee1ed83a3979	MRX67	\N	\N	daylily	2025-07-03 08:21:03.287395-07	INSERT	\N	\N	\N	f	f
02827d52-035b-49a3-a56a-e4197a3e0fd1	generic_instance_lineage	\N	32fa3742-7427-41e3-aa23-a27a43e065d6	GL282	\N	\N	daylily	2025-07-03 08:21:03.293428-07	INSERT	\N	\N	\N	f	f
2c19452b-4df0-4a63-a40c-e40414d1c3cd	generic_instance	json_addl	2ec41575-3c9e-41db-a210-0aa45aa7c05e	CWX22	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.519435-07	UPDATE	\N	\N	\N	f	f
f8997221-1206-45c8-8018-d2de04b1cb0f	generic_instance	created_dt	2ec41575-3c9e-41db-a210-0aa45aa7c05e	CWX22	2025-07-03 08:21:02.514912-07	2025-07-03T08:21:02.514912-07:00	daylily	2025-07-03 08:21:02.519435-07	UPDATE	\N	\N	\N	f	f
0cc1a039-87cd-48d0-bfa6-43c598acec46	generic_instance	modified_dt	2ec41575-3c9e-41db-a210-0aa45aa7c05e	CWX22	2025-07-03 08:21:02.514912-07	2025-07-03T08:21:02.519435-07:00	daylily	2025-07-03 08:21:02.519435-07	UPDATE	\N	\N	\N	f	f
2ffc3b55-616e-4b6b-888b-5ca8dd1e67f2	generic_instance_lineage	\N	1f2797a2-3440-4e7a-85fa-82a15f6cca11	GL150	\N	\N	daylily	2025-07-03 08:21:02.520696-07	INSERT	\N	\N	\N	f	f
c57ce239-9dc7-4f66-b3b4-758033a1c580	generic_instance	\N	b5b21e36-376a-4491-b698-452597c517e8	CWX23	\N	\N	daylily	2025-07-03 08:21:02.520696-07	INSERT	\N	\N	\N	f	f
d48db8f8-f66e-4d75-850e-afabd2abf868	generic_instance	json_addl	b5b21e36-376a-4491-b698-452597c517e8	CWX23	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.524632-07	UPDATE	\N	\N	\N	f	f
b2477378-cf7e-46c5-a1ce-a13f5d4bef56	generic_instance	created_dt	b5b21e36-376a-4491-b698-452597c517e8	CWX23	2025-07-03 08:21:02.520696-07	2025-07-03T08:21:02.520696-07:00	daylily	2025-07-03 08:21:02.524632-07	UPDATE	\N	\N	\N	f	f
18d748d7-6906-482c-98cd-5794eb9ef96b	generic_instance	modified_dt	b5b21e36-376a-4491-b698-452597c517e8	CWX23	2025-07-03 08:21:02.520696-07	2025-07-03T08:21:02.524632-07:00	daylily	2025-07-03 08:21:02.524632-07	UPDATE	\N	\N	\N	f	f
1dad5f26-fd05-43c1-9643-cb748ea164b3	generic_instance_lineage	\N	082a80b0-d445-4f92-af95-269f521cf314	GL151	\N	\N	daylily	2025-07-03 08:21:02.525924-07	INSERT	\N	\N	\N	f	f
b853d0e3-74e6-4df8-9a34-0873deff888e	generic_instance	\N	23f00e77-1469-4ea9-a978-11a2da61dd29	CWX24	\N	\N	daylily	2025-07-03 08:21:02.525924-07	INSERT	\N	\N	\N	f	f
b2cecb66-9dab-43c4-8937-853eb7ffc35f	generic_instance	json_addl	23f00e77-1469-4ea9-a978-11a2da61dd29	CWX24	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.530253-07	UPDATE	\N	\N	\N	f	f
2f0c4d99-9624-4b0e-95d7-003b803bb4d8	generic_instance	created_dt	23f00e77-1469-4ea9-a978-11a2da61dd29	CWX24	2025-07-03 08:21:02.525924-07	2025-07-03T08:21:02.525924-07:00	daylily	2025-07-03 08:21:02.530253-07	UPDATE	\N	\N	\N	f	f
8b8af68e-9087-48fe-a4b8-04ba42adf6a3	generic_instance	modified_dt	23f00e77-1469-4ea9-a978-11a2da61dd29	CWX24	2025-07-03 08:21:02.525924-07	2025-07-03T08:21:02.530253-07:00	daylily	2025-07-03 08:21:02.530253-07	UPDATE	\N	\N	\N	f	f
9df6b2cb-af1b-4276-aa0f-02ffbc58adcb	generic_instance_lineage	\N	4e5e2d5d-b9a4-4093-a46e-dd0a811cd911	GL152	\N	\N	daylily	2025-07-03 08:21:02.531634-07	INSERT	\N	\N	\N	f	f
0b0c9667-0ad5-40df-891d-9f07da162b76	generic_instance	\N	e4e0fd64-c65d-4cc7-8fa8-1a052ae78bcf	CWX25	\N	\N	daylily	2025-07-03 08:21:02.531634-07	INSERT	\N	\N	\N	f	f
9547b92c-a512-43a1-952b-021b2828d134	generic_instance	\N	ed22c26f-07be-4ee0-8510-9575fec8e997	MRX68	\N	\N	daylily	2025-07-03 08:21:03.293428-07	INSERT	\N	\N	\N	f	f
d2a9f6ed-412d-434f-ae14-f5e2dcfa56da	generic_instance_lineage	\N	25c57886-f486-4aa4-9851-ef4dc99b7d5c	GL283	\N	\N	daylily	2025-07-03 08:21:03.299631-07	INSERT	\N	\N	\N	f	f
a913eed2-ebd7-456f-ad0d-71ed7a624dae	generic_instance	\N	2e08566f-8372-48e0-a356-c2aef6cd92d1	MRX69	\N	\N	daylily	2025-07-03 08:21:03.299631-07	INSERT	\N	\N	\N	f	f
422a51b5-61a2-4bea-805b-edbf98cfb29e	generic_instance_lineage	\N	234f1d94-1d08-431b-b9c7-397b8728fe41	GL284	\N	\N	daylily	2025-07-03 08:21:03.305291-07	INSERT	\N	\N	\N	f	f
7417bc2e-da83-4d20-8b84-f7edca976cf3	generic_instance	\N	e1841886-6ac7-4de7-8c4e-1bb17dd37908	MRX70	\N	\N	daylily	2025-07-03 08:21:03.305291-07	INSERT	\N	\N	\N	f	f
35944138-83a6-400b-b4bd-c949b41b6f3c	generic_instance_lineage	\N	3fc3476c-c2e6-49b5-8b56-b8979e449b7b	GL285	\N	\N	daylily	2025-07-03 08:21:03.311046-07	INSERT	\N	\N	\N	f	f
1b96ea76-9791-4c31-b934-fbe09e8bc8aa	generic_instance	\N	071b553d-fd1a-432b-a258-99de30fb97d9	MRX71	\N	\N	daylily	2025-07-03 08:21:03.311046-07	INSERT	\N	\N	\N	f	f
4f42dd23-f7b8-41ad-becc-4cfeef4ba1e6	generic_instance	json_addl	e4e0fd64-c65d-4cc7-8fa8-1a052ae78bcf	CWX25	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.535926-07	UPDATE	\N	\N	\N	f	f
4f32d4c8-8246-446a-95d9-1111d61f91fe	generic_instance	created_dt	e4e0fd64-c65d-4cc7-8fa8-1a052ae78bcf	CWX25	2025-07-03 08:21:02.531634-07	2025-07-03T08:21:02.531634-07:00	daylily	2025-07-03 08:21:02.535926-07	UPDATE	\N	\N	\N	f	f
b2be3e33-106b-4132-b07e-a411e99e56e4	generic_instance	modified_dt	e4e0fd64-c65d-4cc7-8fa8-1a052ae78bcf	CWX25	2025-07-03 08:21:02.531634-07	2025-07-03T08:21:02.535926-07:00	daylily	2025-07-03 08:21:02.535926-07	UPDATE	\N	\N	\N	f	f
fee29ff6-01fb-4f3f-b876-ef6c63890d42	generic_instance_lineage	\N	800ca384-0363-473f-84b6-bec48c07906d	GL153	\N	\N	daylily	2025-07-03 08:21:02.537114-07	INSERT	\N	\N	\N	f	f
dd3b9f73-f50b-40e4-8964-d090a9c9c33a	generic_instance	\N	9ad9dea4-053c-4429-83e2-de60c13f4f67	CWX26	\N	\N	daylily	2025-07-03 08:21:02.537114-07	INSERT	\N	\N	\N	f	f
491a409b-31dc-48a2-9fc0-13385a62ea63	generic_instance	json_addl	9ad9dea4-053c-4429-83e2-de60c13f4f67	CWX26	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.541082-07	UPDATE	\N	\N	\N	f	f
d57dd978-12ae-4cbd-a062-9cbe1ec07bf1	generic_instance	created_dt	9ad9dea4-053c-4429-83e2-de60c13f4f67	CWX26	2025-07-03 08:21:02.537114-07	2025-07-03T08:21:02.537114-07:00	daylily	2025-07-03 08:21:02.541082-07	UPDATE	\N	\N	\N	f	f
1537cbcf-7fb6-4494-8148-9a93c4a0b852	generic_instance	modified_dt	9ad9dea4-053c-4429-83e2-de60c13f4f67	CWX26	2025-07-03 08:21:02.537114-07	2025-07-03T08:21:02.541082-07:00	daylily	2025-07-03 08:21:02.541082-07	UPDATE	\N	\N	\N	f	f
bca3b8b4-4988-4723-8c27-96963d3a0cb0	generic_instance_lineage	\N	57a46487-2d86-4390-9ec1-e291ca334663	GL154	\N	\N	daylily	2025-07-03 08:21:02.542464-07	INSERT	\N	\N	\N	f	f
6f9d969d-6f0f-47cd-93cd-a1ab2f914e40	generic_instance	\N	d13c47a9-1614-4d77-a15d-1872be28308e	CWX27	\N	\N	daylily	2025-07-03 08:21:02.542464-07	INSERT	\N	\N	\N	f	f
37657be7-bd91-48ba-af93-5cdefd2c795c	generic_instance	json_addl	d13c47a9-1614-4d77-a15d-1872be28308e	CWX27	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.546637-07	UPDATE	\N	\N	\N	f	f
54d6d190-0ea4-429f-bca7-7279a9f488e2	generic_instance	created_dt	d13c47a9-1614-4d77-a15d-1872be28308e	CWX27	2025-07-03 08:21:02.542464-07	2025-07-03T08:21:02.542464-07:00	daylily	2025-07-03 08:21:02.546637-07	UPDATE	\N	\N	\N	f	f
3a1d322b-6555-44f5-881c-256b92b36d7e	generic_instance	modified_dt	d13c47a9-1614-4d77-a15d-1872be28308e	CWX27	2025-07-03 08:21:02.542464-07	2025-07-03T08:21:02.546637-07:00	daylily	2025-07-03 08:21:02.546637-07	UPDATE	\N	\N	\N	f	f
39ee553a-bdfb-41de-838e-1a17fdcf62a5	generic_instance_lineage	\N	2b8f67e1-00ee-437f-be89-571705f8eabc	GL155	\N	\N	daylily	2025-07-03 08:21:02.547786-07	INSERT	\N	\N	\N	f	f
63cf9a45-1852-46ff-9f07-5bfc3699b684	generic_instance	\N	248543ec-5f5f-4421-879c-11e5899229fa	CWX28	\N	\N	daylily	2025-07-03 08:21:02.547786-07	INSERT	\N	\N	\N	f	f
450a36ca-9e1b-4276-a23c-1ef77360e326	generic_instance_lineage	\N	001008c3-0745-41c9-a20d-633615ed3a67	GL286	\N	\N	daylily	2025-07-03 08:21:03.316643-07	INSERT	\N	\N	\N	f	f
facf95ad-6df1-4b7f-911e-b991d286fcb2	generic_instance	\N	6be732d5-ba40-4cd2-95a3-ebbb71ea380b	MRX72	\N	\N	daylily	2025-07-03 08:21:03.316643-07	INSERT	\N	\N	\N	f	f
56206ab5-fc00-423a-a278-a008eb5708e1	generic_instance_lineage	\N	9b260b00-055f-43e1-aded-a45bbc02af39	GL287	\N	\N	daylily	2025-07-03 08:21:03.3223-07	INSERT	\N	\N	\N	f	f
0b54aace-703a-4df3-b697-c6a57b7f3da6	generic_instance	\N	02d4fd7d-df49-48a7-8b80-873d463d7037	MRX73	\N	\N	daylily	2025-07-03 08:21:03.3223-07	INSERT	\N	\N	\N	f	f
9485e29c-7e3f-4b9f-ad63-d5e5410c5987	generic_instance_lineage	\N	d989078d-4d91-407c-bd1a-cabafb5936f3	GL288	\N	\N	daylily	2025-07-03 08:21:03.328263-07	INSERT	\N	\N	\N	f	f
60ab3fba-91e1-41a9-a016-abc6af26c9b8	generic_instance	\N	3f2989ce-c0d1-4b52-9140-b9b19d170768	MRX74	\N	\N	daylily	2025-07-03 08:21:03.328263-07	INSERT	\N	\N	\N	f	f
9fd7c23e-ba49-4e6e-97b4-f696e254371d	generic_instance_lineage	\N	0922e1ea-73cd-43ae-95bf-66960355d3ba	GL289	\N	\N	daylily	2025-07-03 08:21:03.334026-07	INSERT	\N	\N	\N	f	f
930e2632-a3ac-4793-b6b8-f21008e47860	generic_instance	json_addl	248543ec-5f5f-4421-879c-11e5899229fa	CWX28	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.55174-07	UPDATE	\N	\N	\N	f	f
bd92f7a0-5a9c-4352-aabd-d8a342006d20	generic_instance	created_dt	248543ec-5f5f-4421-879c-11e5899229fa	CWX28	2025-07-03 08:21:02.547786-07	2025-07-03T08:21:02.547786-07:00	daylily	2025-07-03 08:21:02.55174-07	UPDATE	\N	\N	\N	f	f
c02203e4-c091-47b4-be40-a0a8e28e75e4	generic_instance	modified_dt	248543ec-5f5f-4421-879c-11e5899229fa	CWX28	2025-07-03 08:21:02.547786-07	2025-07-03T08:21:02.55174-07:00	daylily	2025-07-03 08:21:02.55174-07	UPDATE	\N	\N	\N	f	f
1a7eedf8-10e5-4167-8bb4-5f7de6bc4eba	generic_instance_lineage	\N	2bac7721-51a7-415d-95e6-f1fb15adef8e	GL156	\N	\N	daylily	2025-07-03 08:21:02.553092-07	INSERT	\N	\N	\N	f	f
7155c8f8-0c17-4fb1-95c4-429ac2f99df9	generic_instance	\N	73f06c89-2ce5-4e44-a22e-9441cd06dc0d	MRX1	\N	\N	daylily	2025-07-03 08:21:02.554872-07	INSERT	\N	\N	\N	f	f
781fa5c7-c4e5-4d83-8f4e-112ae635fca0	generic_instance_lineage	\N	492d980e-25a8-46c3-9b87-75d557ba1cac	GL157	\N	\N	daylily	2025-07-03 08:21:02.557162-07	INSERT	\N	\N	\N	f	f
29a56c4d-d385-403e-8511-d44935f6f1c9	generic_instance	\N	fa310d7e-c727-402a-9c6c-47eafd01096c	MRX2	\N	\N	daylily	2025-07-03 08:21:02.557162-07	INSERT	\N	\N	\N	f	f
8e61cd9b-82e1-4347-b6a0-a7a6e2cfb0b0	generic_instance_lineage	\N	d09f5f2e-1cce-425a-b20d-3e971b0575fd	GL158	\N	\N	daylily	2025-07-03 08:21:02.562107-07	INSERT	\N	\N	\N	f	f
e2ef48f0-dcdb-4227-a230-e822842665ec	generic_instance	\N	155d76ca-0ef6-4610-8991-3caf5eebcaf9	MRX3	\N	\N	daylily	2025-07-03 08:21:02.562107-07	INSERT	\N	\N	\N	f	f
62f06a8f-4962-4ab5-8104-d4024c2bf1e5	generic_instance_lineage	\N	33d3182a-a8d7-439a-a752-6ba1d8219778	GL159	\N	\N	daylily	2025-07-03 08:21:02.566978-07	INSERT	\N	\N	\N	f	f
4de2efdc-f705-434a-b524-e7b48d5b62f8	generic_instance	\N	7f61478d-3e5c-47b1-a942-508cfe92ce65	MRX4	\N	\N	daylily	2025-07-03 08:21:02.566978-07	INSERT	\N	\N	\N	f	f
04a4f6b3-c9dc-4657-9377-3a4043f426c8	generic_instance_lineage	\N	926a9448-460d-4f6b-a4f9-e26815314305	GL160	\N	\N	daylily	2025-07-03 08:21:02.571207-07	INSERT	\N	\N	\N	f	f
cd4b5762-bdd2-4751-89e6-68da44bf2320	generic_instance	\N	d7eb32c1-6513-4a8e-b241-0bba13be7302	MRX5	\N	\N	daylily	2025-07-03 08:21:02.571207-07	INSERT	\N	\N	\N	f	f
d8f12cb9-9662-4a3f-9500-7c4bacbd62e9	generic_instance_lineage	\N	09b81e88-6598-4ce5-81a2-206f8e0a9a46	GL161	\N	\N	daylily	2025-07-03 08:21:02.575611-07	INSERT	\N	\N	\N	f	f
fdbc87e2-afd3-40d2-abef-e0d7f491a836	generic_instance	\N	1122f95f-07f3-4153-b7b9-eb3710b3370e	MRX6	\N	\N	daylily	2025-07-03 08:21:02.575611-07	INSERT	\N	\N	\N	f	f
d39a533b-dc38-40ad-8ae4-00bd3523d82a	generic_instance_lineage	\N	94fa20cd-f1a2-441b-84b1-f8e71b11144f	GL162	\N	\N	daylily	2025-07-03 08:21:02.579544-07	INSERT	\N	\N	\N	f	f
bbdc7324-bcdc-4300-b8a5-d36d786c2783	generic_instance	\N	ee6c92d8-88fe-4743-af74-1ea3834c8407	MRX7	\N	\N	daylily	2025-07-03 08:21:02.579544-07	INSERT	\N	\N	\N	f	f
aac5b121-553c-40e2-84d9-4bb913ba8626	generic_instance_lineage	\N	b76e6d69-7def-42ad-98a1-1c1500039235	GL163	\N	\N	daylily	2025-07-03 08:21:02.584966-07	INSERT	\N	\N	\N	f	f
202f6534-139d-48d1-a2e3-8e2f6a1b030c	generic_instance	\N	20150c7c-2e35-4810-8f6c-0e7779530fac	MRX8	\N	\N	daylily	2025-07-03 08:21:02.584966-07	INSERT	\N	\N	\N	f	f
271b8b34-3f68-4a55-86c4-e5c898878bae	generic_instance_lineage	\N	4e88924c-588b-47d8-9221-72432cbbec27	GL164	\N	\N	daylily	2025-07-03 08:21:02.591786-07	INSERT	\N	\N	\N	f	f
ae3e3385-10cd-4485-a304-028ff18184a7	generic_instance	\N	25cad9ef-1267-48b7-9a19-239743a8caea	MRX9	\N	\N	daylily	2025-07-03 08:21:02.591786-07	INSERT	\N	\N	\N	f	f
065e94d5-8699-4402-9122-402ca936e049	generic_instance_lineage	\N	abbdef57-8708-42b4-9dcc-ee3591a5ce2e	GL165	\N	\N	daylily	2025-07-03 08:21:02.596373-07	INSERT	\N	\N	\N	f	f
5a3250e6-1b2d-4cc0-9906-66bbf2bb68bd	generic_instance	\N	e04744fc-bc25-486c-a066-89db9af94e92	MRX10	\N	\N	daylily	2025-07-03 08:21:02.596373-07	INSERT	\N	\N	\N	f	f
947ece3e-9cc6-4c87-8a4e-1bd57308453c	generic_instance_lineage	\N	815f0362-3cc3-4d9d-a0e8-a3e79c963c64	GL166	\N	\N	daylily	2025-07-03 08:21:02.600841-07	INSERT	\N	\N	\N	f	f
7a436de6-75ef-48a4-bd3a-62f35bee4d84	generic_instance	\N	b9dbd159-998b-4db7-98ba-fb4f220cd04b	MRX11	\N	\N	daylily	2025-07-03 08:21:02.600841-07	INSERT	\N	\N	\N	f	f
e905faae-ca6c-4493-87e0-65522d8ded7c	generic_instance_lineage	\N	773ed0fe-4337-4821-b638-76ce2a6d994e	GL167	\N	\N	daylily	2025-07-03 08:21:02.605545-07	INSERT	\N	\N	\N	f	f
4da040b2-ee74-4568-84a9-4d176a320e60	generic_instance	\N	eea51634-39be-405d-b1e1-f2b229f006d2	MRX12	\N	\N	daylily	2025-07-03 08:21:02.605545-07	INSERT	\N	\N	\N	f	f
2ca6bd26-09a5-48e5-ac3a-ba28522e4a74	generic_instance_lineage	\N	cf9b3d21-6aee-456b-b30b-daa1c8c0ab84	GL168	\N	\N	daylily	2025-07-03 08:21:02.610022-07	INSERT	\N	\N	\N	f	f
7fc5860c-49a6-4ddd-9376-5115453ed456	generic_instance	\N	0e1fd8ab-89b9-4237-8f0d-86d844c9c3d9	MRX13	\N	\N	daylily	2025-07-03 08:21:02.610022-07	INSERT	\N	\N	\N	f	f
f6c6a587-e2e6-4774-9f96-2f515a963c84	generic_instance_lineage	\N	947c868a-17cf-433a-8327-a29f7629f8d3	GL169	\N	\N	daylily	2025-07-03 08:21:02.614709-07	INSERT	\N	\N	\N	f	f
0f811fbe-7ed4-4104-b6ab-154a13368fb8	generic_instance	\N	6e9a2e5a-f406-43d6-b310-2357457c94db	MRX14	\N	\N	daylily	2025-07-03 08:21:02.614709-07	INSERT	\N	\N	\N	f	f
793f693b-02ef-459d-bf3d-d6cbe0446802	generic_instance_lineage	\N	5052c3dc-b03b-463c-bec8-6e5bcdacc2f4	GL170	\N	\N	daylily	2025-07-03 08:21:02.619075-07	INSERT	\N	\N	\N	f	f
38211a0c-5f1b-4fac-bc83-a64f319065be	generic_instance	\N	3c3c3a45-565a-4e97-9854-2702e0725454	MRX15	\N	\N	daylily	2025-07-03 08:21:02.619075-07	INSERT	\N	\N	\N	f	f
c6284c06-2bed-4585-812a-60c7786411c7	generic_instance_lineage	\N	ff72f753-ba2a-480e-ba40-825a5b2313b1	GL171	\N	\N	daylily	2025-07-03 08:21:02.623871-07	INSERT	\N	\N	\N	f	f
d9e08571-03c3-4013-a00b-b469ba506a1f	generic_instance	\N	3fddfa66-2e48-47d6-8b04-a54d3c08aa2f	MRX16	\N	\N	daylily	2025-07-03 08:21:02.623871-07	INSERT	\N	\N	\N	f	f
5dc6d747-196d-4088-a0d9-97b72f72ab18	generic_instance_lineage	\N	b23eed0b-8300-422c-a96e-37e46b7ff159	GL172	\N	\N	daylily	2025-07-03 08:21:02.628909-07	INSERT	\N	\N	\N	f	f
8713d551-8f88-4fda-b4d9-c79e77b4e1f6	generic_instance	\N	94af7ed1-6a16-463c-9cce-442d3344703f	MRX17	\N	\N	daylily	2025-07-03 08:21:02.628909-07	INSERT	\N	\N	\N	f	f
f66984a9-6f90-4be0-b53a-d8e43cb23d9f	generic_instance_lineage	\N	8ff2b226-0fb7-4f02-8478-a0f269d8a809	GL173	\N	\N	daylily	2025-07-03 08:21:02.633227-07	INSERT	\N	\N	\N	f	f
dcdcb1cc-d6a2-4e2b-8ef4-ac133e75e933	generic_instance	\N	d9adf020-6a54-47b1-ae11-60769eb81e1d	MRX18	\N	\N	daylily	2025-07-03 08:21:02.633227-07	INSERT	\N	\N	\N	f	f
9f95f3f5-66c2-49cc-babd-39185ca2b14d	generic_instance_lineage	\N	8e1bc899-9b88-4ba3-a783-e243843f01f4	GL174	\N	\N	daylily	2025-07-03 08:21:02.637441-07	INSERT	\N	\N	\N	f	f
d5ebdd3c-d47b-400c-ad3d-a8cb2d13303d	generic_instance	\N	f3dfc3a9-b905-4286-9e05-ba68f5762d52	MRX19	\N	\N	daylily	2025-07-03 08:21:02.637441-07	INSERT	\N	\N	\N	f	f
53a0b326-0fc6-4fc0-b3f8-3060e8eab0df	generic_instance_lineage	\N	f41f629e-60f5-467d-8af0-5beffba4a517	GL175	\N	\N	daylily	2025-07-03 08:21:02.642032-07	INSERT	\N	\N	\N	f	f
2daebd30-be50-4948-a249-fb16c4592573	generic_instance	\N	130e9a38-6071-44ea-b12c-0ac555011c7d	MRX20	\N	\N	daylily	2025-07-03 08:21:02.642032-07	INSERT	\N	\N	\N	f	f
0c4b8563-541e-4814-b976-86313d63cd98	generic_instance_lineage	\N	b723df62-3595-4170-965e-801bf4ff97a9	GL176	\N	\N	daylily	2025-07-03 08:21:02.646617-07	INSERT	\N	\N	\N	f	f
97b9f534-9d77-44b5-9e1b-a5b629f31a2b	generic_instance	\N	add01ba4-438a-48a5-aa94-62b29f76862b	MRX21	\N	\N	daylily	2025-07-03 08:21:02.646617-07	INSERT	\N	\N	\N	f	f
633928dd-269d-499a-85a4-e7dc02b71779	generic_instance_lineage	\N	ec9e080d-ef8a-4fe8-af23-2805090191bb	GL177	\N	\N	daylily	2025-07-03 08:21:02.650909-07	INSERT	\N	\N	\N	f	f
47fd3408-d437-4a4e-a6c1-a0ccba7b001c	generic_instance	\N	ce5401cf-dae7-4b77-988d-a7200152e343	MRX22	\N	\N	daylily	2025-07-03 08:21:02.650909-07	INSERT	\N	\N	\N	f	f
6bafc562-e8b0-411b-b59f-972bfa54f39a	generic_instance_lineage	\N	8dd330c3-9d9a-4849-a349-a69079627dba	GL178	\N	\N	daylily	2025-07-03 08:21:02.655356-07	INSERT	\N	\N	\N	f	f
622baa24-aa85-4b87-aec7-a92e45c13bfd	generic_instance	\N	7d46d938-2e29-4a0e-b9ce-c132e7db46e6	MRX23	\N	\N	daylily	2025-07-03 08:21:02.655356-07	INSERT	\N	\N	\N	f	f
7d8e23f9-4908-4ca0-843b-16e95f7e6d54	generic_instance	json_addl	b98e9974-de3d-498d-8371-53ec167ca66f	CWX29	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.694601-07	UPDATE	\N	\N	\N	f	f
8b52bd41-bf48-499d-9ac8-e4013a1937e4	generic_instance	created_dt	b98e9974-de3d-498d-8371-53ec167ca66f	CWX29	2025-07-03 08:21:02.692069-07	2025-07-03T08:21:02.692069-07:00	daylily	2025-07-03 08:21:02.694601-07	UPDATE	\N	\N	\N	f	f
6a478f76-4c33-4cff-a2e2-af37d02db979	generic_instance	modified_dt	b98e9974-de3d-498d-8371-53ec167ca66f	CWX29	2025-07-03 08:21:02.692069-07	2025-07-03T08:21:02.694601-07:00	daylily	2025-07-03 08:21:02.694601-07	UPDATE	\N	\N	\N	f	f
d49cc266-53eb-4ec7-93be-33b9dd1e6bee	generic_instance_lineage	\N	44ad6c97-bef2-47da-8ed3-7319f5e5a742	GL185	\N	\N	daylily	2025-07-03 08:21:02.69569-07	INSERT	\N	\N	\N	f	f
4a2d18ab-b483-4f3d-a691-2fe8f9f1a810	generic_instance	\N	a68a2de5-45b2-4db5-873f-eb0708d9b999	CWX30	\N	\N	daylily	2025-07-03 08:21:02.69569-07	INSERT	\N	\N	\N	f	f
b3c4487a-fcf3-4756-9554-022d57742ab3	generic_instance	json_addl	a68a2de5-45b2-4db5-873f-eb0708d9b999	CWX30	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.700045-07	UPDATE	\N	\N	\N	f	f
3528ea85-71b8-48dd-b57a-61e39fa7404c	generic_instance	created_dt	a68a2de5-45b2-4db5-873f-eb0708d9b999	CWX30	2025-07-03 08:21:02.69569-07	2025-07-03T08:21:02.69569-07:00	daylily	2025-07-03 08:21:02.700045-07	UPDATE	\N	\N	\N	f	f
378d9d68-eee8-4e9e-a34c-2d3cbbfd864a	generic_instance	modified_dt	a68a2de5-45b2-4db5-873f-eb0708d9b999	CWX30	2025-07-03 08:21:02.69569-07	2025-07-03T08:21:02.700045-07:00	daylily	2025-07-03 08:21:02.700045-07	UPDATE	\N	\N	\N	f	f
1ee7d33f-752f-4804-8673-1eb1222f771b	generic_instance_lineage	\N	8a363d88-87bf-411b-9738-90b7e3aac55b	GL186	\N	\N	daylily	2025-07-03 08:21:02.701299-07	INSERT	\N	\N	\N	f	f
dccad69f-0e0b-4c96-9a32-4fe98f0fdf22	generic_instance	\N	5653c452-7c65-4fbf-a63a-555795d12dfd	CWX31	\N	\N	daylily	2025-07-03 08:21:02.701299-07	INSERT	\N	\N	\N	f	f
35566a34-ac75-42c0-b10c-b9f6f9df37a3	generic_instance	json_addl	5653c452-7c65-4fbf-a63a-555795d12dfd	CWX31	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.705287-07	UPDATE	\N	\N	\N	f	f
f8ad025b-f4da-401d-b5bc-46e6124e30c4	generic_instance	created_dt	5653c452-7c65-4fbf-a63a-555795d12dfd	CWX31	2025-07-03 08:21:02.701299-07	2025-07-03T08:21:02.701299-07:00	daylily	2025-07-03 08:21:02.705287-07	UPDATE	\N	\N	\N	f	f
3da3bd90-c405-4c2b-bd5b-fa30a8a00b65	generic_instance	modified_dt	5653c452-7c65-4fbf-a63a-555795d12dfd	CWX31	2025-07-03 08:21:02.701299-07	2025-07-03T08:21:02.705287-07:00	daylily	2025-07-03 08:21:02.705287-07	UPDATE	\N	\N	\N	f	f
8cb1f35c-85de-4280-a501-aa214fbdacb6	generic_instance_lineage	\N	6154ccf3-52a5-4582-a1df-5d877ae9ac48	GL187	\N	\N	daylily	2025-07-03 08:21:02.706361-07	INSERT	\N	\N	\N	f	f
c80351e6-f884-446a-9d0a-626ff4eb3a21	generic_instance	\N	6ebf400e-84b9-48d3-9d87-05463e046efb	CWX32	\N	\N	daylily	2025-07-03 08:21:02.706361-07	INSERT	\N	\N	\N	f	f
c21935e8-1ac9-4267-aefb-a44e4ead91ac	generic_instance	\N	8bff6ef8-6bc4-4cc5-bf48-8acc081d6f99	MRX75	\N	\N	daylily	2025-07-03 08:21:03.334026-07	INSERT	\N	\N	\N	f	f
21431d47-28bb-42fc-833e-0571d2f9d50f	generic_instance_lineage	\N	0fc74178-ff14-4450-88c2-560425d22852	GL290	\N	\N	daylily	2025-07-03 08:21:03.341416-07	INSERT	\N	\N	\N	f	f
6e2c44ca-e3f1-49d0-8a02-d0909e2fad79	generic_instance	\N	86c7fbcf-32e3-4270-8884-b1617558a29e	MRX76	\N	\N	daylily	2025-07-03 08:21:03.341416-07	INSERT	\N	\N	\N	f	f
3875d7e2-5682-4fe2-a0cf-e0079c7cd829	generic_instance_lineage	\N	da673b37-23d5-4b53-9774-c2a00fd4f313	GL291	\N	\N	daylily	2025-07-03 08:21:03.34811-07	INSERT	\N	\N	\N	f	f
774068f3-fbfd-4ac4-a5d9-225b36d65aed	generic_instance	\N	e258c3a2-c387-44c9-93c7-60868c8046c8	MRX77	\N	\N	daylily	2025-07-03 08:21:03.34811-07	INSERT	\N	\N	\N	f	f
ec38a0a4-0f21-4f04-88b1-59f1bdc493df	generic_instance_lineage	\N	102d4f4b-3df3-43f4-8be9-a208b12463b8	GL292	\N	\N	daylily	2025-07-03 08:21:03.353852-07	INSERT	\N	\N	\N	f	f
594605c7-0498-4e45-b6e1-242195dd1099	generic_instance	\N	82dac929-c652-47f9-ad44-7a7ac8fade75	MRX78	\N	\N	daylily	2025-07-03 08:21:03.353852-07	INSERT	\N	\N	\N	f	f
ceb3559c-b68c-43ec-8e49-35b31ff0c994	generic_instance	json_addl	6ebf400e-84b9-48d3-9d87-05463e046efb	CWX32	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.710076-07	UPDATE	\N	\N	\N	f	f
11678e70-06e8-4be4-bdc3-9cd574a1732c	generic_instance	created_dt	6ebf400e-84b9-48d3-9d87-05463e046efb	CWX32	2025-07-03 08:21:02.706361-07	2025-07-03T08:21:02.706361-07:00	daylily	2025-07-03 08:21:02.710076-07	UPDATE	\N	\N	\N	f	f
539973df-110f-4790-b74c-01163be037ab	generic_instance	modified_dt	6ebf400e-84b9-48d3-9d87-05463e046efb	CWX32	2025-07-03 08:21:02.706361-07	2025-07-03T08:21:02.710076-07:00	daylily	2025-07-03 08:21:02.710076-07	UPDATE	\N	\N	\N	f	f
193d86d2-c24e-43f3-b012-83afe33fb974	generic_instance_lineage	\N	e27c1763-c034-47a7-823e-c177b1630ceb	GL188	\N	\N	daylily	2025-07-03 08:21:02.711446-07	INSERT	\N	\N	\N	f	f
cca748a7-5b2e-4baa-98db-8024b6965488	generic_instance	\N	490088ed-6d23-4654-9f8e-032e607177cf	CWX33	\N	\N	daylily	2025-07-03 08:21:02.711446-07	INSERT	\N	\N	\N	f	f
470f5c20-3332-4790-bde1-d74197ff098a	generic_instance	json_addl	490088ed-6d23-4654-9f8e-032e607177cf	CWX33	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.715244-07	UPDATE	\N	\N	\N	f	f
70c74d0c-673c-4e8c-a18f-4bf4274730ca	generic_instance	created_dt	490088ed-6d23-4654-9f8e-032e607177cf	CWX33	2025-07-03 08:21:02.711446-07	2025-07-03T08:21:02.711446-07:00	daylily	2025-07-03 08:21:02.715244-07	UPDATE	\N	\N	\N	f	f
c98c96ca-07f3-4464-8979-7e3a1f383354	generic_instance	modified_dt	490088ed-6d23-4654-9f8e-032e607177cf	CWX33	2025-07-03 08:21:02.711446-07	2025-07-03T08:21:02.715244-07:00	daylily	2025-07-03 08:21:02.715244-07	UPDATE	\N	\N	\N	f	f
048aa818-5f03-41dd-8328-97d099f2b42d	generic_instance_lineage	\N	a9b23622-9e25-4fde-997c-6e5ae8605847	GL189	\N	\N	daylily	2025-07-03 08:21:02.716431-07	INSERT	\N	\N	\N	f	f
9496aadf-5422-4993-9bde-5eed5a5a8e05	generic_instance	\N	f8b71f68-c515-4c50-9d0e-c4dff9409c13	CWX34	\N	\N	daylily	2025-07-03 08:21:02.716431-07	INSERT	\N	\N	\N	f	f
16ef11e3-7ea2-4ac3-8ab1-c1469252a1ab	generic_instance	json_addl	f8b71f68-c515-4c50-9d0e-c4dff9409c13	CWX34	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.720483-07	UPDATE	\N	\N	\N	f	f
57610570-cff7-4943-9f35-7d3a4c70446b	generic_instance	created_dt	f8b71f68-c515-4c50-9d0e-c4dff9409c13	CWX34	2025-07-03 08:21:02.716431-07	2025-07-03T08:21:02.716431-07:00	daylily	2025-07-03 08:21:02.720483-07	UPDATE	\N	\N	\N	f	f
2ade7e73-5cbd-42d1-8a9f-f170e17903ab	generic_instance	modified_dt	f8b71f68-c515-4c50-9d0e-c4dff9409c13	CWX34	2025-07-03 08:21:02.716431-07	2025-07-03T08:21:02.720483-07:00	daylily	2025-07-03 08:21:02.720483-07	UPDATE	\N	\N	\N	f	f
2ec71996-1754-49b4-ad22-900e30de049d	generic_instance_lineage	\N	2d05e9db-7e43-48cc-a065-08054f89147e	GL190	\N	\N	daylily	2025-07-03 08:21:02.721585-07	INSERT	\N	\N	\N	f	f
7deab7a3-4b11-4ede-bd0d-0e6aa76ec9ca	generic_instance	\N	a0eef834-450f-4780-8b5c-7c1db8b7ea67	CWX35	\N	\N	daylily	2025-07-03 08:21:02.721585-07	INSERT	\N	\N	\N	f	f
b8f93001-e7da-44c9-9fe0-0bc0191df0a2	generic_instance_lineage	\N	a09a0eb6-3eba-414c-9c5a-afb50eebd4cc	GL293	\N	\N	daylily	2025-07-03 08:21:03.359328-07	INSERT	\N	\N	\N	f	f
c0a4aaea-3de0-4ccb-87bd-62ead2038f1e	generic_instance	\N	ba650613-ec31-45d0-8935-5b94a1dc3079	MRX79	\N	\N	daylily	2025-07-03 08:21:03.359328-07	INSERT	\N	\N	\N	f	f
21876579-c750-4f83-9690-7ec920e33bb1	generic_instance_lineage	\N	6fff4d6b-2843-4149-ad1b-6fde2cf5ee2b	GL294	\N	\N	daylily	2025-07-03 08:21:03.365447-07	INSERT	\N	\N	\N	f	f
c349afdc-f8fe-4f18-aad6-e5260f275e49	generic_instance	\N	4d800139-262b-4779-ba70-ab7a011c53cb	MRX80	\N	\N	daylily	2025-07-03 08:21:03.365447-07	INSERT	\N	\N	\N	f	f
941da251-270f-45b5-a5e3-c7ec1dd8d7bc	generic_instance_lineage	\N	79a4a65e-34f6-45e2-8c2a-cec28ca4259c	GL295	\N	\N	daylily	2025-07-03 08:21:03.371137-07	INSERT	\N	\N	\N	f	f
6911eb6e-ac03-4773-a51e-104f9b9e1492	generic_instance	\N	ca7c8e06-c44f-414b-a3f4-7f3abdc7d153	MRX81	\N	\N	daylily	2025-07-03 08:21:03.371137-07	INSERT	\N	\N	\N	f	f
1672a796-45b0-48bd-8d56-7ef78bcdf302	generic_instance_lineage	\N	3e137116-98fc-4c3a-9b8a-e44d70090824	GL296	\N	\N	daylily	2025-07-03 08:21:03.376844-07	INSERT	\N	\N	\N	f	f
8e216554-6bd7-4dc5-9aae-43562a2cef7e	generic_instance	json_addl	a0eef834-450f-4780-8b5c-7c1db8b7ea67	CWX35	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.725532-07	UPDATE	\N	\N	\N	f	f
acc1499b-f8ec-48df-a9b7-b61acc30c0f2	generic_instance	created_dt	a0eef834-450f-4780-8b5c-7c1db8b7ea67	CWX35	2025-07-03 08:21:02.721585-07	2025-07-03T08:21:02.721585-07:00	daylily	2025-07-03 08:21:02.725532-07	UPDATE	\N	\N	\N	f	f
88e6006e-de4a-49e6-bab5-681e04e01319	generic_instance	modified_dt	a0eef834-450f-4780-8b5c-7c1db8b7ea67	CWX35	2025-07-03 08:21:02.721585-07	2025-07-03T08:21:02.725532-07:00	daylily	2025-07-03 08:21:02.725532-07	UPDATE	\N	\N	\N	f	f
f8389de8-d7b1-45bf-a491-5f2666239f6a	generic_instance_lineage	\N	47c42ef6-61a9-4a61-a897-95eb26e74770	GL191	\N	\N	daylily	2025-07-03 08:21:02.727093-07	INSERT	\N	\N	\N	f	f
07c14818-7eb6-4c5d-811b-d1a023faf93d	generic_instance	\N	64d6d4da-233a-483b-b150-f76b15678000	CWX36	\N	\N	daylily	2025-07-03 08:21:02.727093-07	INSERT	\N	\N	\N	f	f
c535909e-5e16-4723-b0ac-c7bc48a89274	generic_instance	json_addl	64d6d4da-233a-483b-b150-f76b15678000	CWX36	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.731263-07	UPDATE	\N	\N	\N	f	f
0c9e31ef-118d-42ca-9936-ab91bc68aab9	generic_instance	created_dt	64d6d4da-233a-483b-b150-f76b15678000	CWX36	2025-07-03 08:21:02.727093-07	2025-07-03T08:21:02.727093-07:00	daylily	2025-07-03 08:21:02.731263-07	UPDATE	\N	\N	\N	f	f
da8cedf8-18c2-478c-b48a-b59682237de3	generic_instance	modified_dt	64d6d4da-233a-483b-b150-f76b15678000	CWX36	2025-07-03 08:21:02.727093-07	2025-07-03T08:21:02.731263-07:00	daylily	2025-07-03 08:21:02.731263-07	UPDATE	\N	\N	\N	f	f
04b2beff-6a19-4ad0-abee-d8713fe807a0	generic_instance_lineage	\N	b1b45f70-b597-4133-9046-681cc3b0dd8d	GL192	\N	\N	daylily	2025-07-03 08:21:02.732411-07	INSERT	\N	\N	\N	f	f
1d605139-8cbf-4412-8a4a-a3a31c8431ba	generic_instance	\N	02114dfa-d423-42f2-adf7-65becb5f1255	CWX37	\N	\N	daylily	2025-07-03 08:21:02.732411-07	INSERT	\N	\N	\N	f	f
415dbe19-ba2b-4614-a3bf-0d967b230247	generic_instance	json_addl	02114dfa-d423-42f2-adf7-65becb5f1255	CWX37	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.73656-07	UPDATE	\N	\N	\N	f	f
5697790d-cdae-4272-9bdc-96c00aaa76d9	generic_instance	created_dt	02114dfa-d423-42f2-adf7-65becb5f1255	CWX37	2025-07-03 08:21:02.732411-07	2025-07-03T08:21:02.732411-07:00	daylily	2025-07-03 08:21:02.73656-07	UPDATE	\N	\N	\N	f	f
88729829-c6d4-4f11-8b8d-146cb5e0c4a0	generic_instance	modified_dt	02114dfa-d423-42f2-adf7-65becb5f1255	CWX37	2025-07-03 08:21:02.732411-07	2025-07-03T08:21:02.73656-07:00	daylily	2025-07-03 08:21:02.73656-07	UPDATE	\N	\N	\N	f	f
c13305c1-78a0-4da8-82a3-349e4981fe82	generic_instance_lineage	\N	5d40c347-7c01-4135-8d6e-a07001f5b902	GL193	\N	\N	daylily	2025-07-03 08:21:02.738142-07	INSERT	\N	\N	\N	f	f
ea39bcae-aa06-49bb-8a26-054ff2041808	generic_instance	\N	ee64ced1-19ac-4f65-844c-7b996b0eda0f	CWX38	\N	\N	daylily	2025-07-03 08:21:02.738142-07	INSERT	\N	\N	\N	f	f
a532ee66-a0d7-411a-8173-f1eef2f88144	generic_instance	\N	c7de2b92-a075-4026-86ca-1f491615948c	MRX82	\N	\N	daylily	2025-07-03 08:21:03.376844-07	INSERT	\N	\N	\N	f	f
d45589b2-3b15-4a1e-8eb3-370fc64e4d63	generic_instance_lineage	\N	39921349-1c43-4fa3-ac6a-92fdbe295072	GL297	\N	\N	daylily	2025-07-03 08:21:03.382514-07	INSERT	\N	\N	\N	f	f
4bfe088f-c1a4-4747-90fe-9de726ea7c42	generic_instance	\N	f3ca1576-6ec8-4e97-835f-dbc25be9d779	MRX83	\N	\N	daylily	2025-07-03 08:21:03.382514-07	INSERT	\N	\N	\N	f	f
4a37ab3a-09ec-408b-af6a-06f629147e29	generic_instance_lineage	\N	51610499-5fa3-46a2-8e63-d240c0fce935	GL298	\N	\N	daylily	2025-07-03 08:21:03.388193-07	INSERT	\N	\N	\N	f	f
c779a691-8810-4fe0-a7d3-3b9653b91044	generic_instance	\N	e526049b-2502-40d3-92d6-00841b98b660	MRX84	\N	\N	daylily	2025-07-03 08:21:03.388193-07	INSERT	\N	\N	\N	f	f
dee0998c-b83f-4b21-8632-ace393d73a2a	generic_instance_lineage	\N	62462a73-6790-4ca9-85bd-371aafe6c537	GL299	\N	\N	daylily	2025-07-03 08:21:03.393933-07	INSERT	\N	\N	\N	f	f
fad0800f-39da-4221-818d-9130d1e72075	generic_instance	\N	6bb43301-6e9a-4fad-9cd4-8c43db0a1159	CX7	\N	\N	daylily	2025-07-03 08:21:03.393933-07	INSERT	\N	\N	\N	f	f
fcb5f0f7-93f6-4e32-a769-05f9fef54ec2	generic_instance	json_addl	ee64ced1-19ac-4f65-844c-7b996b0eda0f	CWX38	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.742113-07	UPDATE	\N	\N	\N	f	f
593eb2af-aba5-41c2-9a68-07c7202f3ae3	generic_instance	created_dt	ee64ced1-19ac-4f65-844c-7b996b0eda0f	CWX38	2025-07-03 08:21:02.738142-07	2025-07-03T08:21:02.738142-07:00	daylily	2025-07-03 08:21:02.742113-07	UPDATE	\N	\N	\N	f	f
cc21c97a-e9ff-42ad-a11d-554759939fac	generic_instance	modified_dt	ee64ced1-19ac-4f65-844c-7b996b0eda0f	CWX38	2025-07-03 08:21:02.738142-07	2025-07-03T08:21:02.742113-07:00	daylily	2025-07-03 08:21:02.742113-07	UPDATE	\N	\N	\N	f	f
313edb0f-3a9a-4b41-83b4-9f68e85b7c0e	generic_instance_lineage	\N	07c15c91-54ca-4638-8729-15abe16442d9	GL194	\N	\N	daylily	2025-07-03 08:21:02.743496-07	INSERT	\N	\N	\N	f	f
face219c-a7b1-42da-82fa-4f6fa4918a9f	generic_instance	\N	7842b714-574f-4cb9-a83a-3eb1e194767a	CWX39	\N	\N	daylily	2025-07-03 08:21:02.743496-07	INSERT	\N	\N	\N	f	f
329598fd-a4c4-4dfa-b066-1a229daac540	generic_instance	json_addl	7842b714-574f-4cb9-a83a-3eb1e194767a	CWX39	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.747516-07	UPDATE	\N	\N	\N	f	f
d1bcda80-05bb-4636-9b0c-baf27846aacd	generic_instance	created_dt	7842b714-574f-4cb9-a83a-3eb1e194767a	CWX39	2025-07-03 08:21:02.743496-07	2025-07-03T08:21:02.743496-07:00	daylily	2025-07-03 08:21:02.747516-07	UPDATE	\N	\N	\N	f	f
b4f8ce38-9bfa-4cbd-8447-8452ad341909	generic_instance	modified_dt	7842b714-574f-4cb9-a83a-3eb1e194767a	CWX39	2025-07-03 08:21:02.743496-07	2025-07-03T08:21:02.747516-07:00	daylily	2025-07-03 08:21:02.747516-07	UPDATE	\N	\N	\N	f	f
b3cbe099-3de4-4fff-b7b3-edf01556af12	generic_instance_lineage	\N	2852f056-f0d6-421c-b344-fcc6f0bd0bff	GL195	\N	\N	daylily	2025-07-03 08:21:02.749097-07	INSERT	\N	\N	\N	f	f
924413ba-118e-42f6-a7da-2a5d52c70467	generic_instance	\N	d4573a76-c4f1-4ed7-9424-90fdadead6a9	CWX40	\N	\N	daylily	2025-07-03 08:21:02.749097-07	INSERT	\N	\N	\N	f	f
1978acb6-3836-47bb-86e3-0ccad8bf5dad	generic_instance	json_addl	d4573a76-c4f1-4ed7-9424-90fdadead6a9	CWX40	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.753202-07	UPDATE	\N	\N	\N	f	f
8e686c9a-ad68-4559-a8e9-9eb97cdf9473	generic_instance	created_dt	d4573a76-c4f1-4ed7-9424-90fdadead6a9	CWX40	2025-07-03 08:21:02.749097-07	2025-07-03T08:21:02.749097-07:00	daylily	2025-07-03 08:21:02.753202-07	UPDATE	\N	\N	\N	f	f
1ac0b396-832b-4ee5-bada-98fa700bb0e3	generic_instance	modified_dt	d4573a76-c4f1-4ed7-9424-90fdadead6a9	CWX40	2025-07-03 08:21:02.749097-07	2025-07-03T08:21:02.753202-07:00	daylily	2025-07-03 08:21:02.753202-07	UPDATE	\N	\N	\N	f	f
f1d2029a-81d5-49ff-83b8-6f4fb41a5b89	generic_instance_lineage	\N	2a18e5a8-fe8f-4144-a6bd-ad89f5db9f53	GL196	\N	\N	daylily	2025-07-03 08:21:02.754263-07	INSERT	\N	\N	\N	f	f
8f80c250-58b7-4105-a549-3cd840fb00a3	generic_instance	\N	f37b4964-1b4a-44e9-b71e-faa7ff7a0bb8	CWX41	\N	\N	daylily	2025-07-03 08:21:02.754263-07	INSERT	\N	\N	\N	f	f
d6343138-68d7-414e-be5e-9c4dcf6ab5e6	generic_instance_lineage	\N	355e4638-e17a-4fe7-87df-ba57989a4c0f	GL409	\N	\N	daylily	2025-07-03 08:23:09.274535-07	INSERT	\N	\N	\N	f	f
09d6da6c-2fd5-4543-a6c0-7be2f0fdb5e1	generic_instance	\N	33903f7e-9f8c-4b25-aae5-d2273f155189	MRX136	\N	\N	daylily	2025-07-03 08:23:09.274535-07	INSERT	\N	\N	\N	f	f
9779239d-a097-459f-ac23-3a70a543fc7e	generic_instance_lineage	\N	7511643a-c11c-4603-9aaa-51dfacd96239	GL410	\N	\N	daylily	2025-07-03 08:23:09.278964-07	INSERT	\N	\N	\N	f	f
b68d8983-8745-4e77-8aed-1510e1b829f5	generic_instance	\N	b94ca2e5-3160-4c4d-ba63-e9b1fd3395e0	MRX137	\N	\N	daylily	2025-07-03 08:23:09.278964-07	INSERT	\N	\N	\N	f	f
9e02526f-08d4-487b-9519-7031e73bdc1d	generic_instance_lineage	\N	9253402c-c2be-41a1-aaa2-5c97ff05605e	GL411	\N	\N	daylily	2025-07-03 08:23:09.283551-07	INSERT	\N	\N	\N	f	f
387f4143-c8bd-4f9d-9032-c51c2835c491	generic_instance	\N	88692020-71a2-4eda-89bc-ebb4cf452fed	MRX138	\N	\N	daylily	2025-07-03 08:23:09.283551-07	INSERT	\N	\N	\N	f	f
03964815-2618-4c6f-b2fd-cfd62aa31622	generic_instance_lineage	\N	94d325b7-2307-4805-87a4-d6feeef522f7	GL412	\N	\N	daylily	2025-07-03 08:23:09.287948-07	INSERT	\N	\N	\N	f	f
dd16d625-34b3-4ca6-82db-e8866f91b1be	generic_instance	json_addl	f37b4964-1b4a-44e9-b71e-faa7ff7a0bb8	CWX41	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.758213-07	UPDATE	\N	\N	\N	f	f
1410952c-4507-45ea-bd85-c4c15460de3f	generic_instance	created_dt	f37b4964-1b4a-44e9-b71e-faa7ff7a0bb8	CWX41	2025-07-03 08:21:02.754263-07	2025-07-03T08:21:02.754263-07:00	daylily	2025-07-03 08:21:02.758213-07	UPDATE	\N	\N	\N	f	f
cb209ec5-abfb-47d3-94b0-90b076ff1e7f	generic_instance	modified_dt	f37b4964-1b4a-44e9-b71e-faa7ff7a0bb8	CWX41	2025-07-03 08:21:02.754263-07	2025-07-03T08:21:02.758213-07:00	daylily	2025-07-03 08:21:02.758213-07	UPDATE	\N	\N	\N	f	f
10704e0f-7383-4685-8f0a-63efc87a5fe8	generic_instance_lineage	\N	6babb41b-8acb-48d1-b062-c9551c09734f	GL197	\N	\N	daylily	2025-07-03 08:21:02.759768-07	INSERT	\N	\N	\N	f	f
708fd4bd-c14c-4f8b-9932-94247433a326	generic_instance	\N	db55beaf-f629-47b6-bf32-d6585fc2f1ca	CWX42	\N	\N	daylily	2025-07-03 08:21:02.759768-07	INSERT	\N	\N	\N	f	f
1acb821e-a17c-4850-b9a7-63720ab77c67	generic_instance	json_addl	db55beaf-f629-47b6-bf32-d6585fc2f1ca	CWX42	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.764508-07	UPDATE	\N	\N	\N	f	f
be39d15f-8f92-40b8-8071-f42cb8fe1409	generic_instance	created_dt	db55beaf-f629-47b6-bf32-d6585fc2f1ca	CWX42	2025-07-03 08:21:02.759768-07	2025-07-03T08:21:02.759768-07:00	daylily	2025-07-03 08:21:02.764508-07	UPDATE	\N	\N	\N	f	f
8a3d92fa-61d2-456c-a96d-dcc931b80c01	generic_instance	modified_dt	db55beaf-f629-47b6-bf32-d6585fc2f1ca	CWX42	2025-07-03 08:21:02.759768-07	2025-07-03T08:21:02.764508-07:00	daylily	2025-07-03 08:21:02.764508-07	UPDATE	\N	\N	\N	f	f
539fbe6c-82da-42f0-ab83-7537242246cd	generic_instance_lineage	\N	a4e2f96e-ba92-40c2-bf66-5b5f1605c8e1	GL198	\N	\N	daylily	2025-07-03 08:21:02.767064-07	INSERT	\N	\N	\N	f	f
5ccd9fca-e2a4-4ab5-b7dd-343b5faf10ff	generic_instance	\N	26349cf2-a95d-4895-a58b-3226f432bfa7	CWX43	\N	\N	daylily	2025-07-03 08:21:02.767064-07	INSERT	\N	\N	\N	f	f
514cc3d9-018a-4d85-b843-915a8cd8ce8d	generic_instance	json_addl	26349cf2-a95d-4895-a58b-3226f432bfa7	CWX43	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.775639-07	UPDATE	\N	\N	\N	f	f
3c594fda-9b02-42dc-9b60-361afc6e0d41	generic_instance	created_dt	26349cf2-a95d-4895-a58b-3226f432bfa7	CWX43	2025-07-03 08:21:02.767064-07	2025-07-03T08:21:02.767064-07:00	daylily	2025-07-03 08:21:02.775639-07	UPDATE	\N	\N	\N	f	f
b98f2066-c7c6-4814-aa8e-c7138d3b1924	generic_instance	modified_dt	26349cf2-a95d-4895-a58b-3226f432bfa7	CWX43	2025-07-03 08:21:02.767064-07	2025-07-03T08:21:02.775639-07:00	daylily	2025-07-03 08:21:02.775639-07	UPDATE	\N	\N	\N	f	f
5af91c5a-0321-44ae-a5bb-40a23b3d591d	generic_instance_lineage	\N	34749d2e-9c24-49a3-bcde-0cd627320294	GL199	\N	\N	daylily	2025-07-03 08:21:02.776932-07	INSERT	\N	\N	\N	f	f
8fa11e9f-5fae-44b7-891b-c3b105b972d8	generic_instance	\N	a237b8a6-236f-4af0-94cc-2ccac3708511	CWX44	\N	\N	daylily	2025-07-03 08:21:02.776932-07	INSERT	\N	\N	\N	f	f
0a9f7ca0-f599-42e9-a730-f0ddf7ef21f5	generic_instance	\N	b26f5000-d3ef-4d72-be05-67228321dee4	MRX139	\N	\N	daylily	2025-07-03 08:23:09.287948-07	INSERT	\N	\N	\N	f	f
846588df-079d-4c5a-ab3d-3cb0cf017ee2	generic_instance_lineage	\N	9513e3a5-1064-45d6-9499-4df4bc43ef72	GL413	\N	\N	daylily	2025-07-03 08:23:09.292413-07	INSERT	\N	\N	\N	f	f
7c4007ce-e0d4-4610-86ae-6fac4bee1569	generic_instance	\N	46df43ab-6fe5-4e67-8f96-0ea193e774c6	MRX140	\N	\N	daylily	2025-07-03 08:23:09.292413-07	INSERT	\N	\N	\N	f	f
b07ed65a-a62e-4822-a075-cbbdc77a5e6f	generic_instance_lineage	\N	1aff911b-21b8-4861-9006-ab254d38e0e9	GL414	\N	\N	daylily	2025-07-03 08:23:09.296939-07	INSERT	\N	\N	\N	f	f
4d840426-d302-45f1-b3ce-17d8ac382f9e	generic_instance_lineage	\N	f77365a9-4599-4ae3-9a9d-b2d387eed316	GL575	\N	\N	daylily	2025-07-03 08:23:17.879853-07	INSERT	\N	\N	\N	f	f
ef1a9ea5-8324-45fd-b05c-a87b90697abc	generic_instance	\N	2d6a5e4e-5f46-48bc-ab25-2f3f5ddb6412	MRX215	\N	\N	daylily	2025-07-03 08:23:17.879853-07	INSERT	\N	\N	\N	f	f
ae22abc6-0195-4e88-a8fe-dc6557774688	generic_instance_lineage	\N	70b82c8f-8355-4f5b-9725-e4584bc36ab2	GL576	\N	\N	daylily	2025-07-03 08:23:17.885713-07	INSERT	\N	\N	\N	f	f
787e0e4f-af8d-48f1-ad69-69c33d6ffc5f	generic_instance	json_addl	a237b8a6-236f-4af0-94cc-2ccac3708511	CWX44	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.782574-07	UPDATE	\N	\N	\N	f	f
94d8015b-5d4a-4684-9f26-8ce466ed03de	generic_instance	created_dt	a237b8a6-236f-4af0-94cc-2ccac3708511	CWX44	2025-07-03 08:21:02.776932-07	2025-07-03T08:21:02.776932-07:00	daylily	2025-07-03 08:21:02.782574-07	UPDATE	\N	\N	\N	f	f
16a7c303-2ca4-4cd1-b9e0-493278b80ef1	generic_instance	modified_dt	a237b8a6-236f-4af0-94cc-2ccac3708511	CWX44	2025-07-03 08:21:02.776932-07	2025-07-03T08:21:02.782574-07:00	daylily	2025-07-03 08:21:02.782574-07	UPDATE	\N	\N	\N	f	f
a533550d-79ae-44cd-bde6-01110d8afe38	generic_instance_lineage	\N	673f53a6-8f9a-458f-9695-698756484132	GL200	\N	\N	daylily	2025-07-03 08:21:02.784106-07	INSERT	\N	\N	\N	f	f
60388baf-5197-4dfa-9906-ddf85ca39774	generic_instance	\N	247ddbff-98e3-4459-b22f-b4e7df6ea651	CWX45	\N	\N	daylily	2025-07-03 08:21:02.784106-07	INSERT	\N	\N	\N	f	f
129bde67-1bb3-4b97-8e6e-9afa634277cf	generic_instance	json_addl	247ddbff-98e3-4459-b22f-b4e7df6ea651	CWX45	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.788755-07	UPDATE	\N	\N	\N	f	f
eebafce5-aa68-4630-8ad6-56dffd5d51d9	generic_instance	created_dt	247ddbff-98e3-4459-b22f-b4e7df6ea651	CWX45	2025-07-03 08:21:02.784106-07	2025-07-03T08:21:02.784106-07:00	daylily	2025-07-03 08:21:02.788755-07	UPDATE	\N	\N	\N	f	f
51fe514b-272c-4e55-87e1-6952f0b1e117	generic_instance	modified_dt	247ddbff-98e3-4459-b22f-b4e7df6ea651	CWX45	2025-07-03 08:21:02.784106-07	2025-07-03T08:21:02.788755-07:00	daylily	2025-07-03 08:21:02.788755-07	UPDATE	\N	\N	\N	f	f
b50e178d-a2aa-4f7f-a28f-e661f2533a8c	generic_instance_lineage	\N	52853f6b-81be-414c-aa56-4fff379d8cf0	GL201	\N	\N	daylily	2025-07-03 08:21:02.790176-07	INSERT	\N	\N	\N	f	f
14330cf5-311c-408a-b655-63ff0b848fb6	generic_instance	\N	e37cde14-ce6c-4cb8-a239-67ebf9db6a5a	CWX46	\N	\N	daylily	2025-07-03 08:21:02.790176-07	INSERT	\N	\N	\N	f	f
34bb5506-419e-44ae-9894-8476701d1e07	generic_instance	json_addl	e37cde14-ce6c-4cb8-a239-67ebf9db6a5a	CWX46	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.794017-07	UPDATE	\N	\N	\N	f	f
40f9d250-96ca-47b6-afdf-9b31bc4b3d23	generic_instance	created_dt	e37cde14-ce6c-4cb8-a239-67ebf9db6a5a	CWX46	2025-07-03 08:21:02.790176-07	2025-07-03T08:21:02.790176-07:00	daylily	2025-07-03 08:21:02.794017-07	UPDATE	\N	\N	\N	f	f
f6309e94-9a77-4d12-96f4-6562074967b9	generic_instance	modified_dt	e37cde14-ce6c-4cb8-a239-67ebf9db6a5a	CWX46	2025-07-03 08:21:02.790176-07	2025-07-03T08:21:02.794017-07:00	daylily	2025-07-03 08:21:02.794017-07	UPDATE	\N	\N	\N	f	f
804200a3-69eb-4f84-bf11-24064076cdf9	generic_instance_lineage	\N	76023953-9a1e-4063-9425-5f905ba3f229	GL202	\N	\N	daylily	2025-07-03 08:21:02.795569-07	INSERT	\N	\N	\N	f	f
cbe162ba-d029-4f6d-a3a7-abeeffc29128	generic_instance	\N	0eaa618c-4ac4-4704-b7b0-95b542240475	CWX47	\N	\N	daylily	2025-07-03 08:21:02.795569-07	INSERT	\N	\N	\N	f	f
b8515150-466d-4d67-a9b5-b90287aec6a8	generic_instance	\N	39ac2d80-8f21-4823-8941-abfe02e40867	CX10	\N	\N	daylily	2025-07-03 08:23:16.838138-07	INSERT	\N	\N	\N	f	f
d8323d01-f1af-4f60-8898-18b8d69248d3	generic_instance	\N	db19c6e5-98cc-4eba-9311-7d9ac31c3cf2	CWX141	\N	\N	daylily	2025-07-03 08:23:16.850501-07	INSERT	\N	\N	\N	f	f
d830be40-2dc2-42d2-bddf-e6065866ebf9	generic_instance	\N	e3be72b1-8052-4d5b-a017-c9db741c7bff	MRX216	\N	\N	daylily	2025-07-03 08:23:17.885713-07	INSERT	\N	\N	\N	f	f
2d46f099-3204-46c0-bd94-c41182d48294	generic_instance_lineage	\N	8632faff-f351-441f-9874-060b94ea3d39	GL577	\N	\N	daylily	2025-07-03 08:23:17.891898-07	INSERT	\N	\N	\N	f	f
45ebcd4a-f9ad-4e86-8061-f9bc16093dbb	generic_instance	\N	c69af3be-aa7a-446d-a0c9-8a3c9d5b06f7	MRX217	\N	\N	daylily	2025-07-03 08:23:17.891898-07	INSERT	\N	\N	\N	f	f
1854a130-13b2-49a8-97e5-ecbe41ed46e7	generic_instance_lineage	\N	21ecb073-d644-4add-bd5e-5ca8503a60ec	GL578	\N	\N	daylily	2025-07-03 08:23:17.900232-07	INSERT	\N	\N	\N	f	f
47fd7dce-fe2c-4aca-b1ae-cf022ddb344c	generic_instance	\N	264f523e-dc9b-4fb7-b188-68d48691adab	MRX218	\N	\N	daylily	2025-07-03 08:23:17.900232-07	INSERT	\N	\N	\N	f	f
581046ce-1902-4601-9a02-29a7072dfecf	generic_instance	json_addl	0eaa618c-4ac4-4704-b7b0-95b542240475	CWX47	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.800059-07	UPDATE	\N	\N	\N	f	f
f010a7b9-f559-444e-97e8-d3f8b84d7c5b	generic_instance	created_dt	0eaa618c-4ac4-4704-b7b0-95b542240475	CWX47	2025-07-03 08:21:02.795569-07	2025-07-03T08:21:02.795569-07:00	daylily	2025-07-03 08:21:02.800059-07	UPDATE	\N	\N	\N	f	f
234b9aa3-ce99-43e2-a9ac-3930ce5cc638	generic_instance	modified_dt	0eaa618c-4ac4-4704-b7b0-95b542240475	CWX47	2025-07-03 08:21:02.795569-07	2025-07-03T08:21:02.800059-07:00	daylily	2025-07-03 08:21:02.800059-07	UPDATE	\N	\N	\N	f	f
cae9fcb3-7750-465b-bcaf-b03207399cff	generic_instance_lineage	\N	de84edea-b8f7-4b00-825b-053219c05727	GL203	\N	\N	daylily	2025-07-03 08:21:02.801109-07	INSERT	\N	\N	\N	f	f
47a73ee5-1b5d-4939-b3b9-fa7ac8730a86	generic_instance	\N	3bd67b36-e968-4875-95ce-1eb8a711d47a	CWX48	\N	\N	daylily	2025-07-03 08:21:02.801109-07	INSERT	\N	\N	\N	f	f
a574769a-d1b0-4595-8cdf-ad9d3cd26ca0	generic_instance	json_addl	3bd67b36-e968-4875-95ce-1eb8a711d47a	CWX48	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.804906-07	UPDATE	\N	\N	\N	f	f
0db106df-ed54-4c49-8848-d353ea622ddb	generic_instance	created_dt	3bd67b36-e968-4875-95ce-1eb8a711d47a	CWX48	2025-07-03 08:21:02.801109-07	2025-07-03T08:21:02.801109-07:00	daylily	2025-07-03 08:21:02.804906-07	UPDATE	\N	\N	\N	f	f
7d036be2-4052-46c9-90f4-40100b92eff9	generic_instance	modified_dt	3bd67b36-e968-4875-95ce-1eb8a711d47a	CWX48	2025-07-03 08:21:02.801109-07	2025-07-03T08:21:02.804906-07:00	daylily	2025-07-03 08:21:02.804906-07	UPDATE	\N	\N	\N	f	f
9570b561-26fc-4bae-bdf8-b6cb419acfdf	generic_instance_lineage	\N	26c05e2c-0945-4aad-b974-992e655c6fde	GL204	\N	\N	daylily	2025-07-03 08:21:02.806355-07	INSERT	\N	\N	\N	f	f
ef9e99db-636f-4632-bea8-935b03247ff6	generic_instance	\N	9104381b-76f1-478f-bf41-baaa240f776c	CWX49	\N	\N	daylily	2025-07-03 08:21:02.806355-07	INSERT	\N	\N	\N	f	f
fba4442c-aea0-4eef-8a5a-6cfbc46802dd	generic_instance	json_addl	9104381b-76f1-478f-bf41-baaa240f776c	CWX49	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.810561-07	UPDATE	\N	\N	\N	f	f
2e69c7ce-167f-4ecc-8a6a-5a9cf6538774	generic_instance	created_dt	9104381b-76f1-478f-bf41-baaa240f776c	CWX49	2025-07-03 08:21:02.806355-07	2025-07-03T08:21:02.806355-07:00	daylily	2025-07-03 08:21:02.810561-07	UPDATE	\N	\N	\N	f	f
d2ff4761-d85e-4b15-bdf9-141dedb53ae0	generic_instance	modified_dt	9104381b-76f1-478f-bf41-baaa240f776c	CWX49	2025-07-03 08:21:02.806355-07	2025-07-03T08:21:02.810561-07:00	daylily	2025-07-03 08:21:02.810561-07	UPDATE	\N	\N	\N	f	f
2d185a47-fbed-4096-bfb2-ab218e9eb16f	generic_instance_lineage	\N	b8d55808-a282-47b1-9f9c-75d2946bf11d	GL205	\N	\N	daylily	2025-07-03 08:21:02.811933-07	INSERT	\N	\N	\N	f	f
781efc44-488b-4da7-a73e-f13db42bfe9b	generic_instance	\N	b80d221a-eb76-431b-bc14-78ef538ad39e	CWX50	\N	\N	daylily	2025-07-03 08:21:02.811933-07	INSERT	\N	\N	\N	f	f
c1e3e45f-0f26-4b1f-b14f-47a458788bb8	generic_instance_lineage	\N	0cef7180-14f3-42cd-a30e-5b4afa6f9143	GL465	\N	\N	daylily	2025-07-03 08:23:17.134831-07	INSERT	\N	\N	\N	f	f
ca5d7057-1ba5-469a-b745-93fe4851c2ac	generic_instance	\N	070fcb47-fb22-499a-afa0-e379811d58c6	MRX164	\N	\N	daylily	2025-07-03 08:23:17.134831-07	INSERT	\N	\N	\N	f	f
eebfac74-42b6-4ff6-9c86-ac3ac1d1cbd2	generic_instance_lineage	\N	c99bbeeb-3988-463d-ac06-71c63e29b0a5	GL466	\N	\N	daylily	2025-07-03 08:23:17.141995-07	INSERT	\N	\N	\N	f	f
3e9ae362-ca10-4124-8dcf-9b2ef56686d8	generic_instance	\N	1ad3b52b-32cc-4ab5-a2b4-deb1c2a81bf2	MRX165	\N	\N	daylily	2025-07-03 08:23:17.141995-07	INSERT	\N	\N	\N	f	f
6bece3e5-c3e9-429e-b3cd-e51eb28ae629	generic_instance_lineage	\N	aef6ce20-c08b-48ea-b26d-c951053fdb4f	GL467	\N	\N	daylily	2025-07-03 08:23:17.147896-07	INSERT	\N	\N	\N	f	f
4275a048-6c98-4c62-8403-010d8a7beb59	generic_instance	\N	44cbde60-02c7-44d2-9032-fe68359e4f02	MRX166	\N	\N	daylily	2025-07-03 08:23:17.147896-07	INSERT	\N	\N	\N	f	f
ad872517-aab8-4e62-aeee-30cc597b2f67	generic_instance_lineage	\N	19cf5f23-9718-496a-a07a-016fe8ee09a6	GL468	\N	\N	daylily	2025-07-03 08:23:17.152976-07	INSERT	\N	\N	\N	f	f
67c49df6-6118-45ec-8c84-f8030b82f88b	generic_instance	json_addl	b80d221a-eb76-431b-bc14-78ef538ad39e	CWX50	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.816008-07	UPDATE	\N	\N	\N	f	f
cf7d501b-0587-4831-85a4-8bb964f89ac0	generic_instance	created_dt	b80d221a-eb76-431b-bc14-78ef538ad39e	CWX50	2025-07-03 08:21:02.811933-07	2025-07-03T08:21:02.811933-07:00	daylily	2025-07-03 08:21:02.816008-07	UPDATE	\N	\N	\N	f	f
949914ca-517e-45f2-9804-b979018b1085	generic_instance	modified_dt	b80d221a-eb76-431b-bc14-78ef538ad39e	CWX50	2025-07-03 08:21:02.811933-07	2025-07-03T08:21:02.816008-07:00	daylily	2025-07-03 08:21:02.816008-07	UPDATE	\N	\N	\N	f	f
62a4d4b7-07cf-426a-a8ee-da554a27bb93	generic_instance_lineage	\N	50f06a3a-15ad-486b-8401-2ba99d5683a1	GL206	\N	\N	daylily	2025-07-03 08:21:02.817355-07	INSERT	\N	\N	\N	f	f
c8f1b4ce-b1d3-4db2-a4e7-74dd7f5596e1	generic_instance	\N	62874067-fe2e-4ba5-a527-2c53ef5166c8	CWX51	\N	\N	daylily	2025-07-03 08:21:02.817355-07	INSERT	\N	\N	\N	f	f
4cda1b92-5f05-4564-b4de-a26d760e8853	generic_instance	json_addl	62874067-fe2e-4ba5-a527-2c53ef5166c8	CWX51	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.821698-07	UPDATE	\N	\N	\N	f	f
525f41c2-6f57-450e-a189-8b809d99dc0b	generic_instance	created_dt	62874067-fe2e-4ba5-a527-2c53ef5166c8	CWX51	2025-07-03 08:21:02.817355-07	2025-07-03T08:21:02.817355-07:00	daylily	2025-07-03 08:21:02.821698-07	UPDATE	\N	\N	\N	f	f
240bec8b-06b0-4670-af2e-b360be78da51	generic_instance	modified_dt	62874067-fe2e-4ba5-a527-2c53ef5166c8	CWX51	2025-07-03 08:21:02.817355-07	2025-07-03T08:21:02.821698-07:00	daylily	2025-07-03 08:21:02.821698-07	UPDATE	\N	\N	\N	f	f
b1648ebc-e4d5-45a0-8458-f909481b2149	generic_instance_lineage	\N	1ab8f294-117c-4879-a40e-924c07298d53	GL207	\N	\N	daylily	2025-07-03 08:21:02.822872-07	INSERT	\N	\N	\N	f	f
9332da83-00a9-4bef-9adb-8d3944b2d46c	generic_instance	\N	ff19a148-c912-4b79-a416-d77eb17f4c8c	CWX52	\N	\N	daylily	2025-07-03 08:21:02.822872-07	INSERT	\N	\N	\N	f	f
3424bd03-cfa5-4178-ae5f-93ce927f4d80	generic_instance	json_addl	ff19a148-c912-4b79-a416-d77eb17f4c8c	CWX52	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.826758-07	UPDATE	\N	\N	\N	f	f
80c349b9-c4fe-4bf5-a0f6-0b379dadd3c7	generic_instance	created_dt	ff19a148-c912-4b79-a416-d77eb17f4c8c	CWX52	2025-07-03 08:21:02.822872-07	2025-07-03T08:21:02.822872-07:00	daylily	2025-07-03 08:21:02.826758-07	UPDATE	\N	\N	\N	f	f
95f67bff-4104-4aea-862b-6341ac4f9d1b	generic_instance	modified_dt	ff19a148-c912-4b79-a416-d77eb17f4c8c	CWX52	2025-07-03 08:21:02.822872-07	2025-07-03T08:21:02.826758-07:00	daylily	2025-07-03 08:21:02.826758-07	UPDATE	\N	\N	\N	f	f
c0c02c89-11d7-4d86-a471-11d6fd9d0539	generic_instance_lineage	\N	73eb904b-2928-4580-8568-9ce82906705b	GL208	\N	\N	daylily	2025-07-03 08:21:02.828239-07	INSERT	\N	\N	\N	f	f
f6d6a687-921f-464d-b62a-2e9d3d7958b3	generic_instance	\N	981a109f-2b0d-464c-ba8c-3ab530503986	CWX53	\N	\N	daylily	2025-07-03 08:21:02.828239-07	INSERT	\N	\N	\N	f	f
385bbfde-efdb-4efd-88aa-9049ca3d765f	generic_instance	\N	0c7c36b7-7848-4205-af9c-ce41e543fd35	MRX167	\N	\N	daylily	2025-07-03 08:23:17.152976-07	INSERT	\N	\N	\N	f	f
e97fc89a-32f1-45e2-80c7-809f68c4c285	generic_instance_lineage	\N	b6822c5d-b88d-40dc-a9b2-1c6753b477b2	GL469	\N	\N	daylily	2025-07-03 08:23:17.157502-07	INSERT	\N	\N	\N	f	f
75de059c-cd04-49d9-8039-d5eeb732853b	generic_instance	\N	35697e67-053e-4fbb-ac45-2223813e0a10	MRX168	\N	\N	daylily	2025-07-03 08:23:17.157502-07	INSERT	\N	\N	\N	f	f
afcd4b23-677f-4c27-b897-39ff44d0551f	generic_instance_lineage	\N	d7c796a9-e84f-47ac-a58a-2308a0b3a9fe	GL470	\N	\N	daylily	2025-07-03 08:23:17.162168-07	INSERT	\N	\N	\N	f	f
8718713d-7680-4334-9bc3-299f282416ab	generic_instance	\N	1852629f-beb7-4f69-98e1-971a34af9928	CX11	\N	\N	daylily	2025-07-03 08:23:17.165033-07	INSERT	\N	\N	\N	f	f
fa066c7b-a663-468d-9e6f-ee120135d751	generic_instance	\N	4b6dd84b-e2c0-413d-aa2a-2a520af200af	CWX169	\N	\N	daylily	2025-07-03 08:23:17.169419-07	INSERT	\N	\N	\N	f	f
87566a57-9c56-4c3e-a45e-57921bca322f	generic_instance_lineage	\N	4109ce5e-ebe2-45ea-8d4c-cbad57f41159	GL579	\N	\N	daylily	2025-07-03 08:23:17.906482-07	INSERT	\N	\N	\N	f	f
92ff590d-d901-40e4-b3fc-65b067450da3	generic_instance	json_addl	981a109f-2b0d-464c-ba8c-3ab530503986	CWX53	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.832557-07	UPDATE	\N	\N	\N	f	f
f8ffe00c-1443-45d1-975b-08b67d718508	generic_instance	created_dt	981a109f-2b0d-464c-ba8c-3ab530503986	CWX53	2025-07-03 08:21:02.828239-07	2025-07-03T08:21:02.828239-07:00	daylily	2025-07-03 08:21:02.832557-07	UPDATE	\N	\N	\N	f	f
91a1772c-7532-4c8a-ab39-61ffd8a76822	generic_instance	modified_dt	981a109f-2b0d-464c-ba8c-3ab530503986	CWX53	2025-07-03 08:21:02.828239-07	2025-07-03T08:21:02.832557-07:00	daylily	2025-07-03 08:21:02.832557-07	UPDATE	\N	\N	\N	f	f
2364274f-f939-42ab-8660-b8aa7c1fc02a	generic_instance_lineage	\N	a833636f-76b0-4ea8-81f2-bcea7cc2c27f	GL209	\N	\N	daylily	2025-07-03 08:21:02.833692-07	INSERT	\N	\N	\N	f	f
f1247c56-1be8-4347-824a-b89f66c2fe57	generic_instance	\N	371656ae-8d28-434f-86a7-21cc03c87d92	CWX54	\N	\N	daylily	2025-07-03 08:21:02.833692-07	INSERT	\N	\N	\N	f	f
95ee4cc7-9015-4254-b4f9-655789d0ded1	generic_instance	json_addl	371656ae-8d28-434f-86a7-21cc03c87d92	CWX54	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.837588-07	UPDATE	\N	\N	\N	f	f
80d194e0-5c80-4a2d-a2c9-5a6b1cf6bc35	generic_instance	created_dt	371656ae-8d28-434f-86a7-21cc03c87d92	CWX54	2025-07-03 08:21:02.833692-07	2025-07-03T08:21:02.833692-07:00	daylily	2025-07-03 08:21:02.837588-07	UPDATE	\N	\N	\N	f	f
20c80ee6-5e48-4d37-beb5-440855cf2302	generic_instance	modified_dt	371656ae-8d28-434f-86a7-21cc03c87d92	CWX54	2025-07-03 08:21:02.833692-07	2025-07-03T08:21:02.837588-07:00	daylily	2025-07-03 08:21:02.837588-07	UPDATE	\N	\N	\N	f	f
0aaf0d17-581c-41e4-ad10-52610925d67a	generic_instance_lineage	\N	2ab06d6c-2d35-45bc-947f-8375fabc287f	GL210	\N	\N	daylily	2025-07-03 08:21:02.838828-07	INSERT	\N	\N	\N	f	f
a867d41d-0620-46c4-983c-7330aa5c7308	generic_instance	\N	7554c1d0-df29-4789-b007-c08a1d9f22e1	CWX55	\N	\N	daylily	2025-07-03 08:21:02.838828-07	INSERT	\N	\N	\N	f	f
ba509b23-22d4-462f-a9e3-0a00677f7a28	generic_instance	json_addl	7554c1d0-df29-4789-b007-c08a1d9f22e1	CWX55	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.842641-07	UPDATE	\N	\N	\N	f	f
a1e72dbd-a86b-4e6c-8f97-e0bd72bc84d9	generic_instance	created_dt	7554c1d0-df29-4789-b007-c08a1d9f22e1	CWX55	2025-07-03 08:21:02.838828-07	2025-07-03T08:21:02.838828-07:00	daylily	2025-07-03 08:21:02.842641-07	UPDATE	\N	\N	\N	f	f
192059ba-bffb-4e13-851b-384fc76d5e8b	generic_instance	modified_dt	7554c1d0-df29-4789-b007-c08a1d9f22e1	CWX55	2025-07-03 08:21:02.838828-07	2025-07-03T08:21:02.842641-07:00	daylily	2025-07-03 08:21:02.842641-07	UPDATE	\N	\N	\N	f	f
ce400814-561d-4a08-8559-50daaf15f43f	generic_instance_lineage	\N	d71443dc-a34a-48f5-9c2e-0b65da03c707	GL211	\N	\N	daylily	2025-07-03 08:21:02.843892-07	INSERT	\N	\N	\N	f	f
e37ea827-602a-4d41-994d-ceba28bc1190	generic_instance	\N	42ed148e-fc5f-4bfc-a277-2bfdc42d1941	CWX56	\N	\N	daylily	2025-07-03 08:21:02.843892-07	INSERT	\N	\N	\N	f	f
817561ac-a173-4f61-87a7-a24bb6f7f821	generic_instance_lineage	\N	97d1bdcb-bcf8-4f31-88de-5e0ba783e762	GL521	\N	\N	daylily	2025-07-03 08:23:17.454094-07	INSERT	\N	\N	\N	f	f
6b127aa3-0c98-40bc-ac98-dc711b1dc517	generic_instance	\N	dfcce43b-c1c2-4169-aca9-4e84cb8f9eb9	MRX192	\N	\N	daylily	2025-07-03 08:23:17.454094-07	INSERT	\N	\N	\N	f	f
26eaa109-776e-452d-ba2b-13adc532908c	generic_instance_lineage	\N	70b892ad-d58c-42c2-8152-e0b156e50d96	GL522	\N	\N	daylily	2025-07-03 08:23:17.459277-07	INSERT	\N	\N	\N	f	f
43b98b69-d7c4-45c0-855b-c6410227ddf4	generic_instance	\N	4380dfd5-bc54-4a8f-9866-4012f2b64ab4	MRX193	\N	\N	daylily	2025-07-03 08:23:17.459277-07	INSERT	\N	\N	\N	f	f
887c4a00-9cb4-4946-a880-a06ca175b7e4	generic_instance_lineage	\N	ddf496e1-7004-4cd9-873d-ca9ac78c5c8e	GL523	\N	\N	daylily	2025-07-03 08:23:17.464105-07	INSERT	\N	\N	\N	f	f
840496e7-1b48-4cfa-8601-7e1a5e6ce58f	generic_instance	\N	7191f250-f552-4d76-9ecd-78256934407d	MRX194	\N	\N	daylily	2025-07-03 08:23:17.464105-07	INSERT	\N	\N	\N	f	f
1c1a4cd1-f654-45c8-bbd2-182e7e792cf6	generic_instance_lineage	\N	ecd66eb6-753d-45ae-ba80-6ba2e90bd86c	GL524	\N	\N	daylily	2025-07-03 08:23:17.471022-07	INSERT	\N	\N	\N	f	f
de6e2043-746b-4aeb-abcf-7069b715c413	generic_instance	json_addl	42ed148e-fc5f-4bfc-a277-2bfdc42d1941	CWX56	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:02.847692-07	UPDATE	\N	\N	\N	f	f
e9db3a84-c148-4147-b1c2-c1b6c5797d42	generic_instance	created_dt	42ed148e-fc5f-4bfc-a277-2bfdc42d1941	CWX56	2025-07-03 08:21:02.843892-07	2025-07-03T08:21:02.843892-07:00	daylily	2025-07-03 08:21:02.847692-07	UPDATE	\N	\N	\N	f	f
995090fb-ff42-4f5f-adff-b2d278f00083	generic_instance	modified_dt	42ed148e-fc5f-4bfc-a277-2bfdc42d1941	CWX56	2025-07-03 08:21:02.843892-07	2025-07-03T08:21:02.847692-07:00	daylily	2025-07-03 08:21:02.847692-07	UPDATE	\N	\N	\N	f	f
27b5af31-f24d-464b-bae9-f10504b1fbea	generic_instance_lineage	\N	20d98b3c-6fb6-49f0-929c-621a3a3e0448	GL212	\N	\N	daylily	2025-07-03 08:21:02.848923-07	INSERT	\N	\N	\N	f	f
68b2dae0-7639-41c8-906f-49743bd54078	generic_instance	\N	de59347a-8518-43d1-8141-7314cb217b54	MRX29	\N	\N	daylily	2025-07-03 08:21:02.851007-07	INSERT	\N	\N	\N	f	f
95b08551-c003-4c48-8fa4-fa56c5bdb448	generic_instance_lineage	\N	52ec52f5-0695-4df6-b0c7-e3f502c343fb	GL213	\N	\N	daylily	2025-07-03 08:21:02.852684-07	INSERT	\N	\N	\N	f	f
88411160-4273-4798-98e2-637d08f9f1eb	generic_instance	\N	70e6974a-4889-42b3-aa50-1dc5162a753a	MRX30	\N	\N	daylily	2025-07-03 08:21:02.852684-07	INSERT	\N	\N	\N	f	f
1c8e1616-cf8a-4d2c-84c5-8b992f8f244d	generic_instance_lineage	\N	c60bfba3-c498-4612-af14-0dd994a6470a	GL214	\N	\N	daylily	2025-07-03 08:21:02.857377-07	INSERT	\N	\N	\N	f	f
5f4bb6c6-8c05-4af3-987e-7a8e80b21c80	generic_instance	\N	6bb5fc9e-205d-4cf1-aaa7-b82e122bb446	MRX31	\N	\N	daylily	2025-07-03 08:21:02.857377-07	INSERT	\N	\N	\N	f	f
22050390-4fb3-484b-bd32-bd4a62621e5c	generic_instance_lineage	\N	0b161150-52f2-4580-9feb-cb6af2f34acf	GL215	\N	\N	daylily	2025-07-03 08:21:02.861318-07	INSERT	\N	\N	\N	f	f
1ebb85cf-9d60-48d8-921b-ea506fbb9da6	generic_instance	\N	d97fc469-cfcf-42f6-b9f4-c694040b88c4	MRX32	\N	\N	daylily	2025-07-03 08:21:02.861318-07	INSERT	\N	\N	\N	f	f
8fb071ec-f24c-45ee-82c5-42f90679deb7	generic_instance_lineage	\N	811021ad-2a1f-45af-8b66-9c09c7e1d0d8	GL216	\N	\N	daylily	2025-07-03 08:21:02.866124-07	INSERT	\N	\N	\N	f	f
50707ffd-f9a9-40ff-b5ef-beb16a2faaa5	generic_instance	\N	f25b65aa-479f-4c10-b039-ecf50ea2340a	MRX33	\N	\N	daylily	2025-07-03 08:21:02.866124-07	INSERT	\N	\N	\N	f	f
ae479b57-55a7-4849-b903-a6de18b08a77	generic_instance_lineage	\N	fa3702d0-b8ea-4bd5-ae46-8dfd9d498456	GL217	\N	\N	daylily	2025-07-03 08:21:02.870652-07	INSERT	\N	\N	\N	f	f
6c93a4c6-d3d3-4f1c-9fe0-4576d9dde786	generic_instance	\N	9fc34612-8c1e-42b4-a30b-c95060c031ab	MRX34	\N	\N	daylily	2025-07-03 08:21:02.870652-07	INSERT	\N	\N	\N	f	f
6a2c81d8-b51f-418a-9e2b-3035ef728500	generic_instance_lineage	\N	582ac088-6f0f-4c81-a954-107b621eb12e	GL218	\N	\N	daylily	2025-07-03 08:21:02.877972-07	INSERT	\N	\N	\N	f	f
3eb41ffe-95b9-4737-8efe-797a38f1b816	generic_instance	\N	c5783a54-74bf-4417-b8d6-cfa029a644ec	MRX35	\N	\N	daylily	2025-07-03 08:21:02.877972-07	INSERT	\N	\N	\N	f	f
df3ecfa1-78a4-4181-8772-f49cd66cb247	generic_instance_lineage	\N	ffe2f273-ef35-4502-bd7b-1c8efd609f16	GL219	\N	\N	daylily	2025-07-03 08:21:02.882499-07	INSERT	\N	\N	\N	f	f
763da4b8-24af-4205-91a0-bd8ac6692691	generic_instance	\N	7af7bee4-1b4e-4357-84a6-13e465094334	MRX36	\N	\N	daylily	2025-07-03 08:21:02.882499-07	INSERT	\N	\N	\N	f	f
02d12fb0-c9cd-481c-bc6b-31fe96ae5920	generic_instance_lineage	\N	fe5dea71-c542-452a-9eeb-c07775028458	GL220	\N	\N	daylily	2025-07-03 08:21:02.886662-07	INSERT	\N	\N	\N	f	f
f329c91c-d8b3-462e-9b74-912e7eaff52f	generic_instance	\N	5932edad-b324-4826-b79a-4ff76745d991	MRX37	\N	\N	daylily	2025-07-03 08:21:02.886662-07	INSERT	\N	\N	\N	f	f
583de0dd-9330-466b-b4e1-43505083a7fa	generic_instance_lineage	\N	d87aaabd-2480-4f6a-b772-a019a906503f	GL221	\N	\N	daylily	2025-07-03 08:21:02.891032-07	INSERT	\N	\N	\N	f	f
2b472242-8c5d-4ed0-a3d5-454b4e86d9c8	generic_instance	\N	da26ed31-1065-4604-8b23-091ce21c1a2f	MRX38	\N	\N	daylily	2025-07-03 08:21:02.891032-07	INSERT	\N	\N	\N	f	f
e019dacb-71b6-4b7d-bbef-a342d567f011	generic_instance_lineage	\N	7e9a729c-47ca-484e-8e76-eab9adef2fdc	GL222	\N	\N	daylily	2025-07-03 08:21:02.895251-07	INSERT	\N	\N	\N	f	f
adda9a01-7e0f-4131-a31d-f83a72a69d76	generic_instance	\N	d6222668-162e-4c2d-b085-2ebbc2e71828	MRX39	\N	\N	daylily	2025-07-03 08:21:02.895251-07	INSERT	\N	\N	\N	f	f
52799ac6-5177-4936-a456-eadaf9415a2b	generic_instance_lineage	\N	46382e63-9abd-4b81-9ef9-27d6a97a295a	GL223	\N	\N	daylily	2025-07-03 08:21:02.900199-07	INSERT	\N	\N	\N	f	f
517f50c3-054a-47e7-bffc-2bb227e0cf81	generic_instance	\N	67777ee6-6878-4749-b64d-57f9eaa72595	MRX40	\N	\N	daylily	2025-07-03 08:21:02.900199-07	INSERT	\N	\N	\N	f	f
013be7f1-33a3-4084-9fb6-fa762f8256a0	generic_instance_lineage	\N	93f9b7b4-f8ce-4cee-9283-732312d1169e	GL224	\N	\N	daylily	2025-07-03 08:21:02.904584-07	INSERT	\N	\N	\N	f	f
0def29bd-5106-47ad-96cf-2ba2e0e21a68	generic_instance	\N	d57dd303-b4c6-4710-b3ef-6de6113f5106	MRX41	\N	\N	daylily	2025-07-03 08:21:02.904584-07	INSERT	\N	\N	\N	f	f
e870c803-2c24-43ee-8c41-03d2f16e25f8	generic_instance_lineage	\N	3eaac0cb-57fd-4f7a-bd2f-3b5b0b694152	GL225	\N	\N	daylily	2025-07-03 08:21:02.909067-07	INSERT	\N	\N	\N	f	f
fafcf92d-79ed-4ae2-82a5-2543aee868b0	generic_instance	\N	d9c14341-79eb-4fc7-b30f-5c19d96ca9c2	MRX42	\N	\N	daylily	2025-07-03 08:21:02.909067-07	INSERT	\N	\N	\N	f	f
52a16865-f43a-4dfa-abae-f0aa5ab8f015	generic_instance_lineage	\N	e512746a-c20f-4fac-b008-c4f72f3676d8	GL226	\N	\N	daylily	2025-07-03 08:21:02.913536-07	INSERT	\N	\N	\N	f	f
525bb954-f1eb-4c0a-906f-9c28d5fe28b8	generic_instance	\N	7352677a-92c8-4835-901a-525861538e8e	MRX43	\N	\N	daylily	2025-07-03 08:21:02.913536-07	INSERT	\N	\N	\N	f	f
7e513f40-a02a-4a8c-9578-0b8c03abe55a	generic_instance_lineage	\N	ef2e2bb9-d7be-4366-9228-bcd8cdbd1f30	GL227	\N	\N	daylily	2025-07-03 08:21:02.918242-07	INSERT	\N	\N	\N	f	f
e0fb063e-e523-4bed-a628-27c606ce229c	generic_instance	\N	a330becf-ac06-416e-957d-ac2852be9ea0	MRX44	\N	\N	daylily	2025-07-03 08:21:02.918242-07	INSERT	\N	\N	\N	f	f
548a3dff-8ef9-44f6-b7d2-2826f8f25e9c	generic_instance_lineage	\N	3c82adc9-90f3-4c77-a4c1-859d21e934db	GL228	\N	\N	daylily	2025-07-03 08:21:02.922626-07	INSERT	\N	\N	\N	f	f
a1191e17-a45d-4321-995d-6a448548f97e	generic_instance	\N	a887fcca-84b8-4a78-8bfb-c400a989c31a	MRX45	\N	\N	daylily	2025-07-03 08:21:02.922626-07	INSERT	\N	\N	\N	f	f
437bdaf2-df38-46d8-98e4-c735a34afe29	generic_instance_lineage	\N	41c488f6-b789-4431-ad59-3e865965fa45	GL229	\N	\N	daylily	2025-07-03 08:21:02.926982-07	INSERT	\N	\N	\N	f	f
131ba860-a8c7-4209-abd7-be8061454b1c	generic_instance	\N	5d5e68be-5a3d-4089-a30b-5d7bbbe4a2a7	MRX46	\N	\N	daylily	2025-07-03 08:21:02.926982-07	INSERT	\N	\N	\N	f	f
2d5be978-36a6-4fcc-b85d-5a0af47dd827	generic_instance_lineage	\N	c1aed6a3-140e-4484-a094-17f4e342418d	GL230	\N	\N	daylily	2025-07-03 08:21:02.931525-07	INSERT	\N	\N	\N	f	f
edbc8f91-fcf3-48f4-bd21-85d42f32029b	generic_instance	\N	ccc52ad7-c722-4a84-b999-5f97f4d53ffb	MRX47	\N	\N	daylily	2025-07-03 08:21:02.931525-07	INSERT	\N	\N	\N	f	f
5c83d59a-d4b8-48d2-9ac9-fc8b1809a31c	generic_instance_lineage	\N	16d98e3a-cb00-4c8a-9457-d2c8b473cf5d	GL231	\N	\N	daylily	2025-07-03 08:21:02.935814-07	INSERT	\N	\N	\N	f	f
5573b5f9-510a-4e96-a884-302548aba927	generic_instance	\N	1e7602c7-c553-4988-8eb4-014fe5e4c3ef	MRX48	\N	\N	daylily	2025-07-03 08:21:02.935814-07	INSERT	\N	\N	\N	f	f
2015d5c6-042f-4571-8ee4-22d0bfbb0038	generic_instance_lineage	\N	05975054-b74d-4f4d-8b2b-46ee393fa5f7	GL232	\N	\N	daylily	2025-07-03 08:21:02.940253-07	INSERT	\N	\N	\N	f	f
a5cad16f-e825-4f5d-a0b1-6778370915d3	generic_instance	\N	db082c31-af48-4101-96af-118560adac8e	MRX49	\N	\N	daylily	2025-07-03 08:21:02.940253-07	INSERT	\N	\N	\N	f	f
e334df23-dc11-4eb8-8a24-51b4b964f5b7	generic_instance_lineage	\N	9bd72e8b-3b17-4340-b9b5-94acef1223b7	GL233	\N	\N	daylily	2025-07-03 08:21:02.944649-07	INSERT	\N	\N	\N	f	f
a03f00aa-d926-4941-b2e7-e77dfcfbf672	generic_instance	\N	77754c02-be5c-43cb-b291-8ebdeefc4c2b	MRX50	\N	\N	daylily	2025-07-03 08:21:02.944649-07	INSERT	\N	\N	\N	f	f
b9d7105c-cd36-44df-9f88-67652503316f	generic_instance_lineage	\N	f1004776-858c-4abb-beae-0ca718aff719	GL234	\N	\N	daylily	2025-07-03 08:21:02.94904-07	INSERT	\N	\N	\N	f	f
6394097c-b877-46b2-a52d-cae3ee233a72	generic_instance	\N	8a472e60-5250-43b5-914b-d1bfe5a47927	MRX51	\N	\N	daylily	2025-07-03 08:21:02.94904-07	INSERT	\N	\N	\N	f	f
0636c1a4-50b3-4a9c-b920-c0cc7571a78f	generic_instance	json_addl	d4b6618e-692a-4952-b7f5-164c2c7e45a2	CX3	{"cogs": {"cost": "1.50", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "10ml tube generic", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "Generic 10ml tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "1.50", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "giab NA24385", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "Generic 10ml tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:03.017866-07	UPDATE	\N	\N	\N	f	f
135565b8-0560-4c86-9c8d-8dd8838455bb	generic_instance	created_dt	d4b6618e-692a-4952-b7f5-164c2c7e45a2	CX3	2025-07-03 08:21:02.98976-07	2025-07-03T08:21:02.98976-07:00	daylily	2025-07-03 08:21:03.017866-07	UPDATE	\N	\N	\N	f	f
e249f6c7-c03b-4288-b924-acb449ed1c79	generic_instance	modified_dt	d4b6618e-692a-4952-b7f5-164c2c7e45a2	CX3	2025-07-03 08:21:02.98976-07	2025-07-03T08:21:03.017866-07:00	daylily	2025-07-03 08:21:03.017866-07	UPDATE	\N	\N	\N	f	f
50412388-9feb-4c6d-bdc9-b10d71874ea9	generic_instance_lineage	\N	65dda973-c896-4996-b9a2-395234b91dc5	GL241	\N	\N	daylily	2025-07-03 08:21:03.017866-07	INSERT	\N	\N	\N	f	f
5a64710d-0d7d-48d0-a313-f5a8c3bd5bd0	generic_instance	\N	14f6d115-81d1-4a24-b8d9-a675f97a5251	CX4	\N	\N	daylily	2025-07-03 08:21:03.024427-07	INSERT	\N	\N	\N	f	f
c341a3da-e5fd-4dfc-a43a-001a9f8fcb6d	generic_instance	\N	26562587-4869-4ae7-82f8-06ce603df7f3	MCX2	\N	\N	daylily	2025-07-03 08:21:03.027568-07	INSERT	\N	\N	\N	f	f
2fa85f04-e517-4112-b5a2-6deeea73f849	generic_instance	json_addl	14f6d115-81d1-4a24-b8d9-a675f97a5251	CX4	{"cogs": {"cost": "4.70", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "10ml glass tube", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "10ml glass tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "4.70", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "s ynthetic-cfdna", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "10ml glass tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:03.030939-07	UPDATE	\N	\N	\N	f	f
dbb96b38-ad39-4a33-8b6f-26cfbe381c49	generic_instance	created_dt	14f6d115-81d1-4a24-b8d9-a675f97a5251	CX4	2025-07-03 08:21:03.024427-07	2025-07-03T08:21:03.024427-07:00	daylily	2025-07-03 08:21:03.030939-07	UPDATE	\N	\N	\N	f	f
efc205d7-c71f-4f70-8e2c-6fff728534ff	generic_instance	modified_dt	14f6d115-81d1-4a24-b8d9-a675f97a5251	CX4	2025-07-03 08:21:03.024427-07	2025-07-03T08:21:03.030939-07:00	daylily	2025-07-03 08:21:03.030939-07	UPDATE	\N	\N	\N	f	f
a2daf9eb-2b68-43eb-85ff-92cd0c02ce81	generic_instance_lineage	\N	9a121f62-bcd3-48be-a7e6-8f9f962b21d2	GL242	\N	\N	daylily	2025-07-03 08:21:03.030939-07	INSERT	\N	\N	\N	f	f
8751eca9-6bfd-4c1f-a1d4-761c30d8f3c3	generic_instance	\N	ed15ff68-5a5a-4bf8-99aa-b4619eb64fd1	CX5	\N	\N	daylily	2025-07-03 08:21:03.035534-07	INSERT	\N	\N	\N	f	f
7ebf5f39-86d6-4270-a280-92da80f1d6be	generic_instance	\N	8571eed1-4c25-4b2d-9f40-2e3bf265e6b3	MCX3	\N	\N	daylily	2025-07-03 08:21:03.038-07	INSERT	\N	\N	\N	f	f
3775b97a-b281-4d54-8a0e-d57c50e7fb5d	generic_instance	json_addl	ed15ff68-5a5a-4bf8-99aa-b4619eb64fd1	CX5	{"cogs": {"cost": "1.70", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "1.5ml eppi tube", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "1.5ml eppi tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "1.70", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "NTC - water", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "1.5ml eppi tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:03.040775-07	UPDATE	\N	\N	\N	f	f
765f0e15-af49-4bb3-8f18-d0ff2f371619	generic_instance	created_dt	ed15ff68-5a5a-4bf8-99aa-b4619eb64fd1	CX5	2025-07-03 08:21:03.035534-07	2025-07-03T08:21:03.035534-07:00	daylily	2025-07-03 08:21:03.040775-07	UPDATE	\N	\N	\N	f	f
a03cbd7a-c174-43cc-b554-42d669ad968a	generic_instance	modified_dt	ed15ff68-5a5a-4bf8-99aa-b4619eb64fd1	CX5	2025-07-03 08:21:03.035534-07	2025-07-03T08:21:03.040775-07:00	daylily	2025-07-03 08:21:03.040775-07	UPDATE	\N	\N	\N	f	f
8048b39c-8752-42bd-b07f-1b08f73da619	generic_instance_lineage	\N	77853896-887c-4d79-bee7-2cd9078883c5	GL243	\N	\N	daylily	2025-07-03 08:21:03.040775-07	INSERT	\N	\N	\N	f	f
65a4a4c8-ec39-41b2-bba8-340c9e88f258	generic_instance	\N	c5f7422b-9ae1-4496-a645-28e6decfdcd9	CX6	\N	\N	daylily	2025-07-03 08:21:03.044814-07	INSERT	\N	\N	\N	f	f
98084311-34a5-473f-a270-ba28422eeedc	generic_instance	\N	a20f9cb2-4002-4a69-bff6-d52924a67a7b	CWX57	\N	\N	daylily	2025-07-03 08:21:03.049675-07	INSERT	\N	\N	\N	f	f
ef9c11f6-ff7d-4fd9-8ed4-e2216f6f06e4	generic_instance	json_addl	a20f9cb2-4002-4a69-bff6-d52924a67a7b	CWX57	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:03.053177-07	UPDATE	\N	\N	\N	f	f
b0323309-f218-4e19-ab98-f4b6bc0fd551	generic_instance	created_dt	a20f9cb2-4002-4a69-bff6-d52924a67a7b	CWX57	2025-07-03 08:21:03.049675-07	2025-07-03T08:21:03.049675-07:00	daylily	2025-07-03 08:21:03.053177-07	UPDATE	\N	\N	\N	f	f
dfa4cd72-d5de-435b-b0e0-ecce385490db	generic_instance	json_addl	1a35323d-f111-442b-958f-716b90b5d1f2	CWX58	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:03.063623-07	UPDATE	\N	\N	\N	f	f
43cdc9df-f52c-4445-9fe8-a15333d325f5	generic_instance	created_dt	1a35323d-f111-442b-958f-716b90b5d1f2	CWX58	2025-07-03 08:21:03.054349-07	2025-07-03T08:21:03.054349-07:00	daylily	2025-07-03 08:21:03.063623-07	UPDATE	\N	\N	\N	f	f
24fd84a1-eca0-4760-a11e-2d1556dd2ea8	generic_instance	modified_dt	1a35323d-f111-442b-958f-716b90b5d1f2	CWX58	2025-07-03 08:21:03.054349-07	2025-07-03T08:21:03.063623-07:00	daylily	2025-07-03 08:21:03.063623-07	UPDATE	\N	\N	\N	f	f
accf16cc-ace3-4e2c-acf6-b9028a945f4d	generic_instance_lineage	\N	9e36ce27-9cda-4b79-9d10-582ff86a73c5	GL245	\N	\N	daylily	2025-07-03 08:21:03.068927-07	INSERT	\N	\N	\N	f	f
8fe26f3f-2828-417a-9a3c-3fbd123ba881	generic_instance	\N	9b76a542-1cdd-4c04-bb17-9fd9766d78cd	CWX59	\N	\N	daylily	2025-07-03 08:21:03.068927-07	INSERT	\N	\N	\N	f	f
46b54296-83f7-41a2-bb2a-e6136928b4c1	generic_instance	json_addl	9b76a542-1cdd-4c04-bb17-9fd9766d78cd	CWX59	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:03.076386-07	UPDATE	\N	\N	\N	f	f
0cc8b991-bbe2-4a00-acfa-081b99c766e7	generic_instance	created_dt	9b76a542-1cdd-4c04-bb17-9fd9766d78cd	CWX59	2025-07-03 08:21:03.068927-07	2025-07-03T08:21:03.068927-07:00	daylily	2025-07-03 08:21:03.076386-07	UPDATE	\N	\N	\N	f	f
3df69771-2347-4a8a-bbf3-3c3864250318	generic_instance	modified_dt	9b76a542-1cdd-4c04-bb17-9fd9766d78cd	CWX59	2025-07-03 08:21:03.068927-07	2025-07-03T08:21:03.076386-07:00	daylily	2025-07-03 08:21:03.076386-07	UPDATE	\N	\N	\N	f	f
2b06271c-54ec-4c05-b8d5-1f3eb94f047c	generic_instance_lineage	\N	6b687a84-7642-4f88-b4d6-78e5c6091d39	GL246	\N	\N	daylily	2025-07-03 08:21:03.0776-07	INSERT	\N	\N	\N	f	f
9be09fac-0166-44f0-b946-89a260f5f750	generic_instance	\N	e025d800-f0c0-4537-b1b1-cf50c0b976af	CWX60	\N	\N	daylily	2025-07-03 08:21:03.0776-07	INSERT	\N	\N	\N	f	f
5dd4673e-4bd8-4443-b974-897e9587d20c	generic_instance	json_addl	e025d800-f0c0-4537-b1b1-cf50c0b976af	CWX60	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:03.081513-07	UPDATE	\N	\N	\N	f	f
de4b633f-cfd0-4f4b-9b34-81b5c59e6bc1	generic_instance	created_dt	e025d800-f0c0-4537-b1b1-cf50c0b976af	CWX60	2025-07-03 08:21:03.0776-07	2025-07-03T08:21:03.0776-07:00	daylily	2025-07-03 08:21:03.081513-07	UPDATE	\N	\N	\N	f	f
8ef139da-0024-4318-ad76-3f105d5503e1	generic_instance	modified_dt	e025d800-f0c0-4537-b1b1-cf50c0b976af	CWX60	2025-07-03 08:21:03.0776-07	2025-07-03T08:21:03.081513-07:00	daylily	2025-07-03 08:21:03.081513-07	UPDATE	\N	\N	\N	f	f
233ac720-d327-43af-88b1-974d2f450d24	generic_instance_lineage	\N	2790e507-5b50-4002-8729-7ca14689e420	GL247	\N	\N	daylily	2025-07-03 08:21:03.082741-07	INSERT	\N	\N	\N	f	f
aa447677-9aa3-4c61-ab3a-514035514e46	generic_instance	\N	9fb53135-baa2-4a38-895f-c36ea22b3b11	CWX61	\N	\N	daylily	2025-07-03 08:21:03.082741-07	INSERT	\N	\N	\N	f	f
040f30a7-fc8f-40ca-a20b-bbe97e8d3b1a	generic_instance	\N	5e7485c4-94f0-4ff4-b03a-fd6f67a262d0	MRX195	\N	\N	daylily	2025-07-03 08:23:17.471022-07	INSERT	\N	\N	\N	f	f
63b14ac2-5fde-43af-b0b1-e8a8994f8c64	generic_instance_lineage	\N	ed977d1c-245a-4c41-8a4d-2db1f1485b13	GL525	\N	\N	daylily	2025-07-03 08:23:17.475801-07	INSERT	\N	\N	\N	f	f
b0146693-f35b-4569-b651-21f2ca00e8ee	generic_instance	\N	275089ba-354d-4b77-a90b-c9eb2be85073	MRX196	\N	\N	daylily	2025-07-03 08:23:17.475801-07	INSERT	\N	\N	\N	f	f
3c7a0820-cf77-4844-999f-5ba053640937	generic_instance_lineage	\N	ae9018b8-f4d5-4e32-a17c-1d40eb514619	GL526	\N	\N	daylily	2025-07-03 08:23:17.480331-07	INSERT	\N	\N	\N	f	f
a8980ad8-2855-49c5-baf8-06d7c32ab2ab	generic_instance	\N	b1f40745-93ea-4e41-8d30-5a900f02aab9	MRX219	\N	\N	daylily	2025-07-03 08:23:17.906482-07	INSERT	\N	\N	\N	f	f
98246f93-7cdb-4d93-ad90-4b12bae9d6cb	generic_instance_lineage	\N	61d1d09f-4fb2-4dde-9156-702072657d31	GL580	\N	\N	daylily	2025-07-03 08:23:17.912796-07	INSERT	\N	\N	\N	f	f
7595bd64-d7f5-4662-9ee5-8d68ff1820c0	generic_instance	\N	1452bef9-03fa-4bf2-9908-b85eb998617a	MRX220	\N	\N	daylily	2025-07-03 08:23:17.912796-07	INSERT	\N	\N	\N	f	f
ec2762f3-76a2-4d37-b6ee-3488b08da777	generic_instance	json_addl	9fb53135-baa2-4a38-895f-c36ea22b3b11	CWX61	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:03.086591-07	UPDATE	\N	\N	\N	f	f
9eb46930-36b5-465d-a9de-66ba397b7170	generic_instance	created_dt	9fb53135-baa2-4a38-895f-c36ea22b3b11	CWX61	2025-07-03 08:21:03.082741-07	2025-07-03T08:21:03.082741-07:00	daylily	2025-07-03 08:21:03.086591-07	UPDATE	\N	\N	\N	f	f
af1fdafb-f48e-4f71-800e-fe7b3d8215b2	generic_instance	modified_dt	9fb53135-baa2-4a38-895f-c36ea22b3b11	CWX61	2025-07-03 08:21:03.082741-07	2025-07-03T08:21:03.086591-07:00	daylily	2025-07-03 08:21:03.086591-07	UPDATE	\N	\N	\N	f	f
4976718c-dd70-4aa3-9c60-0fc932179fb5	generic_instance_lineage	\N	5d59836e-6817-4a46-9b6a-cbe41017d7c2	GL248	\N	\N	daylily	2025-07-03 08:21:03.087697-07	INSERT	\N	\N	\N	f	f
c4f77e02-d390-49b6-86ef-e5d3eaf53c0f	generic_instance	\N	4b2787c8-670e-45a3-93e9-7c32160e7800	CWX62	\N	\N	daylily	2025-07-03 08:21:03.087697-07	INSERT	\N	\N	\N	f	f
bc1ee7ff-ed6a-46ab-a3be-29cd14bcd097	generic_instance	json_addl	4b2787c8-670e-45a3-93e9-7c32160e7800	CWX62	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:03.091585-07	UPDATE	\N	\N	\N	f	f
274eb2b8-b264-4f5e-b3c6-074060d5cd87	generic_instance	created_dt	4b2787c8-670e-45a3-93e9-7c32160e7800	CWX62	2025-07-03 08:21:03.087697-07	2025-07-03T08:21:03.087697-07:00	daylily	2025-07-03 08:21:03.091585-07	UPDATE	\N	\N	\N	f	f
69d14df1-f65a-4a87-9e23-26cdaa31c6fb	generic_instance	modified_dt	4b2787c8-670e-45a3-93e9-7c32160e7800	CWX62	2025-07-03 08:21:03.087697-07	2025-07-03T08:21:03.091585-07:00	daylily	2025-07-03 08:21:03.091585-07	UPDATE	\N	\N	\N	f	f
a2c52fd0-3f6d-4de1-9fd5-07bd1a62d57d	generic_instance_lineage	\N	c488e6e8-5fb6-4cf1-9c6b-df9f32f49895	GL249	\N	\N	daylily	2025-07-03 08:21:03.093038-07	INSERT	\N	\N	\N	f	f
03a5d0af-2140-4851-8baf-d6bef5482e9f	generic_instance	\N	c188bf5e-cf34-431e-9949-ce0743474003	CWX63	\N	\N	daylily	2025-07-03 08:21:03.093038-07	INSERT	\N	\N	\N	f	f
aa1948b3-f74e-46b4-bcc3-89dfbe3c3303	generic_instance	json_addl	c188bf5e-cf34-431e-9949-ce0743474003	CWX63	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:03.097542-07	UPDATE	\N	\N	\N	f	f
7bd4e47e-b26c-46cd-a520-6cebb99f144a	generic_instance	created_dt	c188bf5e-cf34-431e-9949-ce0743474003	CWX63	2025-07-03 08:21:03.093038-07	2025-07-03T08:21:03.093038-07:00	daylily	2025-07-03 08:21:03.097542-07	UPDATE	\N	\N	\N	f	f
76cbe3a7-6f15-4397-9125-5e8c26c3c931	generic_instance	modified_dt	c188bf5e-cf34-431e-9949-ce0743474003	CWX63	2025-07-03 08:21:03.093038-07	2025-07-03T08:21:03.097542-07:00	daylily	2025-07-03 08:21:03.097542-07	UPDATE	\N	\N	\N	f	f
360a1a6b-c3bd-4215-9d04-eef45e4dc561	generic_instance_lineage	\N	fef9e938-8114-4b49-9324-c0c65b6decb3	GL250	\N	\N	daylily	2025-07-03 08:21:03.09916-07	INSERT	\N	\N	\N	f	f
71eb8261-8062-46ef-a825-42bf85f852ea	generic_instance	\N	0c8feadc-bdae-4f20-b475-839e34f912c2	CWX64	\N	\N	daylily	2025-07-03 08:21:03.09916-07	INSERT	\N	\N	\N	f	f
a686c06b-1098-46ab-a0c6-d66d1a2403d1	generic_instance	\N	d6dc35b7-cf4c-4e70-a692-c23346f61bcd	CX12	\N	\N	daylily	2025-07-03 08:23:17.484168-07	INSERT	\N	\N	\N	f	f
9e3d06e8-1756-4977-b774-b33d55100687	generic_instance	\N	f9786af0-c6a2-470c-b39b-8bc18c43c409	MCX4	\N	\N	daylily	2025-07-03 08:23:17.491289-07	INSERT	\N	\N	\N	f	f
1f56de57-5907-42bd-9729-5865b20dbfeb	generic_instance_lineage	\N	ff05c678-a518-47e4-a180-36ee383a75c3	GL581	\N	\N	daylily	2025-07-03 08:23:17.919066-07	INSERT	\N	\N	\N	f	f
8cb1f5a6-28e1-4df7-8e12-31e665aee903	generic_instance	\N	5551791d-203e-4584-b45d-4cce837d92a7	MRX221	\N	\N	daylily	2025-07-03 08:23:17.919066-07	INSERT	\N	\N	\N	f	f
6e766caa-32d2-4fd2-aea2-281c8c86f02b	generic_instance_lineage	\N	9a04149d-0f0b-49a4-aea9-732058930817	GL582	\N	\N	daylily	2025-07-03 08:23:17.925122-07	INSERT	\N	\N	\N	f	f
466fa621-3ead-4940-bec4-fef39cd79f14	generic_instance	\N	a0110d14-da41-463b-86a6-81b15bdd0775	MRX222	\N	\N	daylily	2025-07-03 08:23:17.925122-07	INSERT	\N	\N	\N	f	f
f23abc4b-e545-46e7-9675-baa9c22cab74	generic_instance_lineage	\N	ba879a78-3dee-475f-8ec0-53e22579cfb2	GL583	\N	\N	daylily	2025-07-03 08:23:17.931001-07	INSERT	\N	\N	\N	f	f
276f59c1-23a7-49f0-8807-f43c1d40abed	generic_instance	json_addl	0c8feadc-bdae-4f20-b475-839e34f912c2	CWX64	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:03.103438-07	UPDATE	\N	\N	\N	f	f
fd8e720c-fe42-446a-bfec-63cffb541dd6	generic_instance	created_dt	0c8feadc-bdae-4f20-b475-839e34f912c2	CWX64	2025-07-03 08:21:03.09916-07	2025-07-03T08:21:03.09916-07:00	daylily	2025-07-03 08:21:03.103438-07	UPDATE	\N	\N	\N	f	f
666cbe7e-3f16-487d-9c62-eb532f134822	generic_instance	modified_dt	0c8feadc-bdae-4f20-b475-839e34f912c2	CWX64	2025-07-03 08:21:03.09916-07	2025-07-03T08:21:03.103438-07:00	daylily	2025-07-03 08:21:03.103438-07	UPDATE	\N	\N	\N	f	f
d5077e03-32c4-4b1b-92fd-25eecba354b0	generic_instance_lineage	\N	131c2d42-303a-4ab7-928b-876a6d1fecf8	GL251	\N	\N	daylily	2025-07-03 08:21:03.104822-07	INSERT	\N	\N	\N	f	f
ef49860a-c86f-4390-adee-7b4c368928e1	generic_instance	\N	96d3bf7e-a12d-4167-b3eb-6f30a69c3b24	CWX65	\N	\N	daylily	2025-07-03 08:21:03.104822-07	INSERT	\N	\N	\N	f	f
39035679-66d2-4cc5-8c39-84bbb5309d1b	generic_instance	json_addl	96d3bf7e-a12d-4167-b3eb-6f30a69c3b24	CWX65	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:03.108766-07	UPDATE	\N	\N	\N	f	f
1e2a303b-8054-4a89-867d-0f5463b9ab89	generic_instance	created_dt	96d3bf7e-a12d-4167-b3eb-6f30a69c3b24	CWX65	2025-07-03 08:21:03.104822-07	2025-07-03T08:21:03.104822-07:00	daylily	2025-07-03 08:21:03.108766-07	UPDATE	\N	\N	\N	f	f
bd53040c-d67b-4fab-94ae-656dc2a2b6ff	generic_instance	modified_dt	96d3bf7e-a12d-4167-b3eb-6f30a69c3b24	CWX65	2025-07-03 08:21:03.104822-07	2025-07-03T08:21:03.108766-07:00	daylily	2025-07-03 08:21:03.108766-07	UPDATE	\N	\N	\N	f	f
849fe298-4dee-4240-b455-ba546d3c6fbb	generic_instance_lineage	\N	84d114ba-5d11-408f-bc8d-6c7a702b8d2d	GL252	\N	\N	daylily	2025-07-03 08:21:03.110199-07	INSERT	\N	\N	\N	f	f
36e7fd2e-69e4-4aae-87eb-01de405b46fa	generic_instance	\N	6e0aa5d8-9855-45b9-8da1-1ae98a7dc57c	CWX66	\N	\N	daylily	2025-07-03 08:21:03.110199-07	INSERT	\N	\N	\N	f	f
9d4e3dfb-34b8-4d77-b9a3-a63fc3d5e566	generic_instance	json_addl	6e0aa5d8-9855-45b9-8da1-1ae98a7dc57c	CWX66	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:03.114005-07	UPDATE	\N	\N	\N	f	f
f2a5d826-6854-49d2-8f3e-fb7efe646967	generic_instance	created_dt	6e0aa5d8-9855-45b9-8da1-1ae98a7dc57c	CWX66	2025-07-03 08:21:03.110199-07	2025-07-03T08:21:03.110199-07:00	daylily	2025-07-03 08:21:03.114005-07	UPDATE	\N	\N	\N	f	f
6c1b5f7f-1995-4ef4-9d73-cfadbdf5113f	generic_instance	modified_dt	6e0aa5d8-9855-45b9-8da1-1ae98a7dc57c	CWX66	2025-07-03 08:21:03.110199-07	2025-07-03T08:21:03.114005-07:00	daylily	2025-07-03 08:21:03.114005-07	UPDATE	\N	\N	\N	f	f
793b55ca-942f-4de3-b0f1-6acd64b29f1c	generic_instance_lineage	\N	14bdc49f-93f2-4d8b-adbb-dd480f1d4f5d	GL253	\N	\N	daylily	2025-07-03 08:21:03.115153-07	INSERT	\N	\N	\N	f	f
9ef2fc7e-9bce-4585-9343-18b6541224ba	generic_instance	\N	42c4b4ad-e85c-4b20-9276-a9f98084fd5c	CWX67	\N	\N	daylily	2025-07-03 08:21:03.115153-07	INSERT	\N	\N	\N	f	f
be4a6a86-a2c2-488e-a6ff-1cd75627ef21	generic_instance	\N	c9e2dab1-38f2-4ade-9c19-814d76031f2d	MRX223	\N	\N	daylily	2025-07-03 08:23:17.931001-07	INSERT	\N	\N	\N	f	f
c070b354-4369-476c-9946-8a7cfa4354a3	generic_instance_lineage	\N	a511a619-24ee-41a3-a8e6-ee7a7f1bbaad	GL584	\N	\N	daylily	2025-07-03 08:23:17.936775-07	INSERT	\N	\N	\N	f	f
fb56cfb6-bfd3-4954-8bac-a8b73b64a60d	generic_instance	\N	9999a8ff-1161-4ba5-a8c3-347e452a5f92	MRX224	\N	\N	daylily	2025-07-03 08:23:17.936775-07	INSERT	\N	\N	\N	f	f
deac7090-02f1-4547-8954-05ec61ecdd81	generic_instance_lineage	\N	0401fdfb-3f0e-47f9-8bab-da87d20816bf	GL585	\N	\N	daylily	2025-07-03 08:23:17.942647-07	INSERT	\N	\N	\N	f	f
6dac1033-7b07-469d-ace4-e3ff294ed063	generic_instance	\N	8be1afd7-55da-4457-90b1-f02f5cc77e5c	CX16	\N	\N	daylily	2025-07-03 08:23:17.942647-07	INSERT	\N	\N	\N	f	f
d7f7d740-d223-4968-953d-13017464e229	generic_instance	json_addl	42c4b4ad-e85c-4b20-9276-a9f98084fd5c	CWX67	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:03.119129-07	UPDATE	\N	\N	\N	f	f
c7b2e541-f936-4eb5-8618-3f9cbc1943b7	generic_instance	created_dt	42c4b4ad-e85c-4b20-9276-a9f98084fd5c	CWX67	2025-07-03 08:21:03.115153-07	2025-07-03T08:21:03.115153-07:00	daylily	2025-07-03 08:21:03.119129-07	UPDATE	\N	\N	\N	f	f
dfe92f73-5a3e-4c4f-be4d-598a4899d236	generic_instance	modified_dt	42c4b4ad-e85c-4b20-9276-a9f98084fd5c	CWX67	2025-07-03 08:21:03.115153-07	2025-07-03T08:21:03.119129-07:00	daylily	2025-07-03 08:21:03.119129-07	UPDATE	\N	\N	\N	f	f
af0f9199-6e97-4c0e-a53c-700661499c9a	generic_instance_lineage	\N	fd8a8aec-ee5b-4b81-88da-1f736db065f1	GL254	\N	\N	daylily	2025-07-03 08:21:03.120564-07	INSERT	\N	\N	\N	f	f
d55bb8de-b745-4902-8b12-590e8d617a12	generic_instance	\N	723fff3b-fa28-4af0-82a4-c5228c05a429	CWX68	\N	\N	daylily	2025-07-03 08:21:03.120564-07	INSERT	\N	\N	\N	f	f
d0ef9a52-0af5-49bd-9539-d207bb55242f	generic_instance	json_addl	723fff3b-fa28-4af0-82a4-c5228c05a429	CWX68	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:03.124341-07	UPDATE	\N	\N	\N	f	f
c1587b3d-9e73-46b6-af3e-785f950aaa08	generic_instance	created_dt	723fff3b-fa28-4af0-82a4-c5228c05a429	CWX68	2025-07-03 08:21:03.120564-07	2025-07-03T08:21:03.120564-07:00	daylily	2025-07-03 08:21:03.124341-07	UPDATE	\N	\N	\N	f	f
0cabf796-7bfd-48ec-8936-e7c90b5d8549	generic_instance	modified_dt	723fff3b-fa28-4af0-82a4-c5228c05a429	CWX68	2025-07-03 08:21:03.120564-07	2025-07-03T08:21:03.124341-07:00	daylily	2025-07-03 08:21:03.124341-07	UPDATE	\N	\N	\N	f	f
a559f683-9de1-4343-9d08-8f344c0c515e	generic_instance_lineage	\N	cbbd217c-0824-417f-a034-d06d7d3f50f3	GL255	\N	\N	daylily	2025-07-03 08:21:03.125565-07	INSERT	\N	\N	\N	f	f
5c179241-9f25-4a0a-b9d0-5dfc171d0dba	generic_instance	\N	fd2c7148-44ec-48bf-abe8-38d5fb4c9e93	CWX69	\N	\N	daylily	2025-07-03 08:21:03.125565-07	INSERT	\N	\N	\N	f	f
fa6f4128-e336-4fe4-b076-e2ea2dee981a	generic_instance	json_addl	fd2c7148-44ec-48bf-abe8-38d5fb4c9e93	CWX69	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:03.130106-07	UPDATE	\N	\N	\N	f	f
2d2ef867-cac6-4e4e-a365-cff9cf8d5635	generic_instance	created_dt	fd2c7148-44ec-48bf-abe8-38d5fb4c9e93	CWX69	2025-07-03 08:21:03.125565-07	2025-07-03T08:21:03.125565-07:00	daylily	2025-07-03 08:21:03.130106-07	UPDATE	\N	\N	\N	f	f
282b0b74-a217-49db-91dd-07894a86159f	generic_instance	modified_dt	fd2c7148-44ec-48bf-abe8-38d5fb4c9e93	CWX69	2025-07-03 08:21:03.125565-07	2025-07-03T08:21:03.130106-07:00	daylily	2025-07-03 08:21:03.130106-07	UPDATE	\N	\N	\N	f	f
9aabfbed-2826-4928-bda9-14fa6969b334	generic_instance_lineage	\N	de385663-a99e-4d19-bd08-64567298633a	GL256	\N	\N	daylily	2025-07-03 08:21:03.138604-07	INSERT	\N	\N	\N	f	f
80dfa59c-fd4b-474e-9135-64b741c1d663	generic_instance	\N	2b67f9bd-b749-40d0-8189-51ed2c0da68e	CWX70	\N	\N	daylily	2025-07-03 08:21:03.138604-07	INSERT	\N	\N	\N	f	f
46cf8f83-30e5-4a29-8a0e-8d6b4e29d600	generic_instance	json_addl	2b67f9bd-b749-40d0-8189-51ed2c0da68e	CWX70	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:03.14646-07	UPDATE	\N	\N	\N	f	f
ce10e484-3c66-49c8-be20-bc5dfb043ce2	generic_instance	created_dt	2b67f9bd-b749-40d0-8189-51ed2c0da68e	CWX70	2025-07-03 08:21:03.138604-07	2025-07-03T08:21:03.138604-07:00	daylily	2025-07-03 08:21:03.14646-07	UPDATE	\N	\N	\N	f	f
4b99e682-cdd0-494c-8a5a-1ca345ce0a46	generic_instance	modified_dt	2b67f9bd-b749-40d0-8189-51ed2c0da68e	CWX70	2025-07-03 08:21:03.138604-07	2025-07-03T08:21:03.14646-07:00	daylily	2025-07-03 08:21:03.14646-07	UPDATE	\N	\N	\N	f	f
9fd071ec-91c0-4b74-b03c-2ece4807355e	generic_instance_lineage	\N	8a9c03a2-5b4c-4f87-a61a-f7d6d32bf1d1	GL257	\N	\N	daylily	2025-07-03 08:21:03.147822-07	INSERT	\N	\N	\N	f	f
8fd617e5-f86e-4d3f-b416-44fa17706c0c	generic_instance	\N	467f92c0-58ff-47fa-85a6-e9268901935e	CWX71	\N	\N	daylily	2025-07-03 08:21:03.147822-07	INSERT	\N	\N	\N	f	f
0d1016d2-be57-4b37-8738-96700eb415b0	generic_instance	json_addl	467f92c0-58ff-47fa-85a6-e9268901935e	CWX71	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:03.152282-07	UPDATE	\N	\N	\N	f	f
6df0d845-4de1-4157-b40c-c61d3b565f47	generic_instance	created_dt	467f92c0-58ff-47fa-85a6-e9268901935e	CWX71	2025-07-03 08:21:03.147822-07	2025-07-03T08:21:03.147822-07:00	daylily	2025-07-03 08:21:03.152282-07	UPDATE	\N	\N	\N	f	f
b08fcaf6-6f3e-4695-a645-a9c48b250b44	generic_instance	modified_dt	467f92c0-58ff-47fa-85a6-e9268901935e	CWX71	2025-07-03 08:21:03.147822-07	2025-07-03T08:21:03.152282-07:00	daylily	2025-07-03 08:21:03.152282-07	UPDATE	\N	\N	\N	f	f
236e179b-c7da-4158-a3ca-2793a3ae4907	generic_instance_lineage	\N	4b8562a1-3f1b-48e1-aff6-3425a41b0b8a	GL258	\N	\N	daylily	2025-07-03 08:21:03.153581-07	INSERT	\N	\N	\N	f	f
b905a7fe-daa1-435e-8945-8824ef4d066b	generic_instance	\N	ecc31348-3e43-46dc-b0d9-c81f9ba47938	CWX72	\N	\N	daylily	2025-07-03 08:21:03.153581-07	INSERT	\N	\N	\N	f	f
5825c572-f63d-46d9-af60-565beee55892	generic_instance	json_addl	ecc31348-3e43-46dc-b0d9-c81f9ba47938	CWX72	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:03.157767-07	UPDATE	\N	\N	\N	f	f
98e95f74-ee3f-45c3-8f8c-90c14f07e9c3	generic_instance	created_dt	ecc31348-3e43-46dc-b0d9-c81f9ba47938	CWX72	2025-07-03 08:21:03.153581-07	2025-07-03T08:21:03.153581-07:00	daylily	2025-07-03 08:21:03.157767-07	UPDATE	\N	\N	\N	f	f
6a7b491f-f181-4f0b-8f42-71bb8f920e0e	generic_instance	modified_dt	ecc31348-3e43-46dc-b0d9-c81f9ba47938	CWX72	2025-07-03 08:21:03.153581-07	2025-07-03T08:21:03.157767-07:00	daylily	2025-07-03 08:21:03.157767-07	UPDATE	\N	\N	\N	f	f
b8bd53eb-2dc4-4446-811a-cd595494c3b4	generic_instance_lineage	\N	923970ef-2781-4273-a175-761c91dabee8	GL259	\N	\N	daylily	2025-07-03 08:21:03.15928-07	INSERT	\N	\N	\N	f	f
b6e05f26-bdd8-416a-9040-8b51ba3381f1	generic_instance	\N	f8ac8009-49e0-4a5d-a1d2-6dd107e496e7	CWX73	\N	\N	daylily	2025-07-03 08:21:03.15928-07	INSERT	\N	\N	\N	f	f
26b1c855-e303-430e-9208-5b69938c7911	generic_instance	\N	a6ab103e-ac0d-4bda-b989-6beb34419280	CX17	\N	\N	daylily	2025-07-03 08:24:15.648493-07	INSERT	\N	\N	\N	f	f
c8a8d997-07af-41d4-8938-ecf960b05116	generic_instance	\N	843a9ccc-36ca-4d8b-80c8-a38487e0a12d	CWX225	\N	\N	daylily	2025-07-03 08:24:15.677773-07	INSERT	\N	\N	\N	f	f
5e18e934-0243-4554-9e25-98ca477a0fa7	generic_instance	json_addl	f8ac8009-49e0-4a5d-a1d2-6dd107e496e7	CWX73	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:03.163663-07	UPDATE	\N	\N	\N	f	f
5613c472-f00b-419a-b7da-a3a36c16c579	generic_instance	created_dt	f8ac8009-49e0-4a5d-a1d2-6dd107e496e7	CWX73	2025-07-03 08:21:03.15928-07	2025-07-03T08:21:03.15928-07:00	daylily	2025-07-03 08:21:03.163663-07	UPDATE	\N	\N	\N	f	f
c444a7a4-176a-4f23-9eb9-0d1c2d00d0c5	generic_instance	modified_dt	f8ac8009-49e0-4a5d-a1d2-6dd107e496e7	CWX73	2025-07-03 08:21:03.15928-07	2025-07-03T08:21:03.163663-07:00	daylily	2025-07-03 08:21:03.163663-07	UPDATE	\N	\N	\N	f	f
6278bb59-b0cd-42e0-936d-77c23506f321	generic_instance_lineage	\N	4e09b61d-bcf7-4dd5-9abc-a2663c96e8c7	GL260	\N	\N	daylily	2025-07-03 08:21:03.165817-07	INSERT	\N	\N	\N	f	f
31da1013-0144-4806-a4e0-8108195634a5	generic_instance	\N	c30fd0fb-b696-4632-895b-ac4e0b089522	CWX74	\N	\N	daylily	2025-07-03 08:21:03.165817-07	INSERT	\N	\N	\N	f	f
8c3f62ab-b297-436c-b8a8-dc60d2bec65c	generic_instance	json_addl	c30fd0fb-b696-4632-895b-ac4e0b089522	CWX74	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:03.170429-07	UPDATE	\N	\N	\N	f	f
c2fbf296-1f0f-48ad-bed2-4fe2bd2ecf3d	generic_instance	created_dt	c30fd0fb-b696-4632-895b-ac4e0b089522	CWX74	2025-07-03 08:21:03.165817-07	2025-07-03T08:21:03.165817-07:00	daylily	2025-07-03 08:21:03.170429-07	UPDATE	\N	\N	\N	f	f
04300475-f29c-4586-8d95-88a3b9f883e0	generic_instance	modified_dt	c30fd0fb-b696-4632-895b-ac4e0b089522	CWX74	2025-07-03 08:21:03.165817-07	2025-07-03T08:21:03.170429-07:00	daylily	2025-07-03 08:21:03.170429-07	UPDATE	\N	\N	\N	f	f
7d1b904b-04d8-4cb6-9d8f-1b54c8d988fe	generic_instance_lineage	\N	4e428bc7-aa44-455f-82e9-1b0f48b3bfd5	GL261	\N	\N	daylily	2025-07-03 08:21:03.171921-07	INSERT	\N	\N	\N	f	f
0c4d3baa-80bb-4ff6-bf8b-55bed957fd15	generic_instance	\N	18266d0e-48c1-4a3c-a530-bd7d02a60043	CWX75	\N	\N	daylily	2025-07-03 08:21:03.171921-07	INSERT	\N	\N	\N	f	f
d8db6b06-ef4d-4d2b-bd4f-1ab284a99119	generic_instance	json_addl	18266d0e-48c1-4a3c-a530-bd7d02a60043	CWX75	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:03.175836-07	UPDATE	\N	\N	\N	f	f
081d497e-1151-433c-adbb-c7b71be3bcaf	generic_instance	created_dt	18266d0e-48c1-4a3c-a530-bd7d02a60043	CWX75	2025-07-03 08:21:03.171921-07	2025-07-03T08:21:03.171921-07:00	daylily	2025-07-03 08:21:03.175836-07	UPDATE	\N	\N	\N	f	f
2873681e-9100-46b4-8d4f-fe4a6133c1c9	generic_instance	modified_dt	18266d0e-48c1-4a3c-a530-bd7d02a60043	CWX75	2025-07-03 08:21:03.171921-07	2025-07-03T08:21:03.175836-07:00	daylily	2025-07-03 08:21:03.175836-07	UPDATE	\N	\N	\N	f	f
5e1515da-9448-49f3-98d2-5ce79e9faeeb	generic_instance_lineage	\N	9db54eed-808f-4ce6-8e2c-03ddab45bce1	GL262	\N	\N	daylily	2025-07-03 08:21:03.177099-07	INSERT	\N	\N	\N	f	f
e770b540-09a2-4675-9c45-3e25e44f13d1	generic_instance	\N	ffb4a762-b5dc-4494-b040-eb49d98a3e89	CWX76	\N	\N	daylily	2025-07-03 08:21:03.177099-07	INSERT	\N	\N	\N	f	f
3bfd1ae7-f5d4-421f-a437-d4822499b0ee	generic_instance_lineage	\N	f5cbe271-a522-4ca7-b600-a40d038b5efc	GL639	\N	\N	daylily	2025-07-03 08:24:16.075965-07	INSERT	\N	\N	\N	f	f
ccaa792c-f527-4124-9455-f57f82fdec08	generic_instance	\N	4ef01449-ede9-4847-9074-67d56aa0e462	MRX248	\N	\N	daylily	2025-07-03 08:24:16.075965-07	INSERT	\N	\N	\N	f	f
bbd95c9a-d64e-4510-9215-619786ed890d	generic_instance_lineage	\N	66e1fbee-48f4-402e-9a7e-267de546c2c1	GL640	\N	\N	daylily	2025-07-03 08:24:16.080477-07	INSERT	\N	\N	\N	f	f
7b347141-5748-492d-a7d8-a3320f6caf92	generic_instance	\N	955793c4-ddbf-4bdf-a4e2-51f541d2d828	MRX249	\N	\N	daylily	2025-07-03 08:24:16.080477-07	INSERT	\N	\N	\N	f	f
efd78e86-cd03-4206-a9d9-af147442f018	generic_instance_lineage	\N	61a5e379-dc70-40d1-8540-a84cff8b5596	GL641	\N	\N	daylily	2025-07-03 08:24:16.08627-07	INSERT	\N	\N	\N	f	f
b538336c-a210-47eb-851f-fb2b3496b2a1	generic_instance	\N	275fe9ac-a3f8-4182-bd15-d5982dab919d	MRX250	\N	\N	daylily	2025-07-03 08:24:16.08627-07	INSERT	\N	\N	\N	f	f
ae76c250-92fd-4c1c-901a-392f0b6050db	generic_instance_lineage	\N	f042aa35-b703-49ce-a177-5dec670b35f8	GL642	\N	\N	daylily	2025-07-03 08:24:16.090668-07	INSERT	\N	\N	\N	f	f
6565a861-ab42-48a5-b323-4ba16aa13441	generic_instance	json_addl	ffb4a762-b5dc-4494-b040-eb49d98a3e89	CWX76	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:03.180898-07	UPDATE	\N	\N	\N	f	f
13c6cc16-a01e-4977-b657-93e50aa20169	generic_instance	created_dt	ffb4a762-b5dc-4494-b040-eb49d98a3e89	CWX76	2025-07-03 08:21:03.177099-07	2025-07-03T08:21:03.177099-07:00	daylily	2025-07-03 08:21:03.180898-07	UPDATE	\N	\N	\N	f	f
e94a9305-4cad-4fc4-910c-c49b7c1f9575	generic_instance	modified_dt	ffb4a762-b5dc-4494-b040-eb49d98a3e89	CWX76	2025-07-03 08:21:03.177099-07	2025-07-03T08:21:03.180898-07:00	daylily	2025-07-03 08:21:03.180898-07	UPDATE	\N	\N	\N	f	f
59e86b25-9194-484e-a5b5-56d754ec5382	generic_instance_lineage	\N	e720145e-8133-4106-9e2e-ae0737e22666	GL263	\N	\N	daylily	2025-07-03 08:21:03.182487-07	INSERT	\N	\N	\N	f	f
3afd339b-c6d1-4b6c-b87a-2d97943f7616	generic_instance	\N	edc810e0-165b-4d5d-8e0e-aa000572600d	CWX77	\N	\N	daylily	2025-07-03 08:21:03.182487-07	INSERT	\N	\N	\N	f	f
06c2cfaf-546e-4565-8f16-c5a9b93def15	generic_instance	json_addl	edc810e0-165b-4d5d-8e0e-aa000572600d	CWX77	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:03.186572-07	UPDATE	\N	\N	\N	f	f
3ad58a5c-be5e-4e4d-924a-3699291daf15	generic_instance	created_dt	edc810e0-165b-4d5d-8e0e-aa000572600d	CWX77	2025-07-03 08:21:03.182487-07	2025-07-03T08:21:03.182487-07:00	daylily	2025-07-03 08:21:03.186572-07	UPDATE	\N	\N	\N	f	f
0cf4b0cc-9338-4548-b154-5e1db970a081	generic_instance	modified_dt	edc810e0-165b-4d5d-8e0e-aa000572600d	CWX77	2025-07-03 08:21:03.182487-07	2025-07-03T08:21:03.186572-07:00	daylily	2025-07-03 08:21:03.186572-07	UPDATE	\N	\N	\N	f	f
93d8741a-4a77-43b2-a74b-ebfd59204ee1	generic_instance_lineage	\N	804320c8-a0c9-4c29-89a0-fcb543c1c226	GL264	\N	\N	daylily	2025-07-03 08:21:03.187621-07	INSERT	\N	\N	\N	f	f
66b8b7d2-4702-4489-9993-c61110cb42fd	generic_instance	\N	de6985c5-652b-4542-b1e5-a5ff86fe3d52	CWX78	\N	\N	daylily	2025-07-03 08:21:03.187621-07	INSERT	\N	\N	\N	f	f
c2e19f49-2552-46a6-887d-4f97e3fc6a1f	generic_instance	json_addl	de6985c5-652b-4542-b1e5-a5ff86fe3d52	CWX78	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:03.191325-07	UPDATE	\N	\N	\N	f	f
24687716-1155-4da6-8a90-e65341aa0834	generic_instance	created_dt	de6985c5-652b-4542-b1e5-a5ff86fe3d52	CWX78	2025-07-03 08:21:03.187621-07	2025-07-03T08:21:03.187621-07:00	daylily	2025-07-03 08:21:03.191325-07	UPDATE	\N	\N	\N	f	f
630ad8aa-2aef-4961-9fe7-dcc19d3bf077	generic_instance	modified_dt	de6985c5-652b-4542-b1e5-a5ff86fe3d52	CWX78	2025-07-03 08:21:03.187621-07	2025-07-03T08:21:03.191325-07:00	daylily	2025-07-03 08:21:03.191325-07	UPDATE	\N	\N	\N	f	f
b59e5e23-da16-43db-9b3f-21f13f54f02a	generic_instance_lineage	\N	8b551b71-f40d-4f21-bf52-8bf726d2fd31	GL265	\N	\N	daylily	2025-07-03 08:21:03.192956-07	INSERT	\N	\N	\N	f	f
aef9e94d-c4de-46f1-a5ae-074780952264	generic_instance	\N	3f1c2a4d-7892-4b46-9059-f884aad34e80	CWX79	\N	\N	daylily	2025-07-03 08:21:03.192956-07	INSERT	\N	\N	\N	f	f
5d6e0078-922d-41f7-aae0-df09e84f583c	generic_instance	\N	890ef728-0ffb-4f4d-90bd-29924f4bb340	MRX251	\N	\N	daylily	2025-07-03 08:24:16.090668-07	INSERT	\N	\N	\N	f	f
d3356a2d-e75a-450a-be18-b0e16a5c26dc	generic_instance_lineage	\N	1bfb4d22-5674-48f0-a82c-7e3d344d6ecf	GL643	\N	\N	daylily	2025-07-03 08:24:16.095989-07	INSERT	\N	\N	\N	f	f
33a44098-61a5-4396-8860-4cfcf5dc5c6d	generic_instance	\N	5b8e86dd-588a-4518-9713-3a48aef550a6	MRX252	\N	\N	daylily	2025-07-03 08:24:16.095989-07	INSERT	\N	\N	\N	f	f
6841028e-6825-47c8-870a-0ffc3848c9ef	generic_instance_lineage	\N	81eb69f1-60cc-4b3d-ab64-480dc08cc57a	GL644	\N	\N	daylily	2025-07-03 08:24:16.10054-07	INSERT	\N	\N	\N	f	f
c4532b24-bb19-42fc-bcbb-3ad9e8120a0d	generic_instance	\N	f0bbdd4d-da05-445d-8c05-a8ed4769fe79	CX18	\N	\N	daylily	2025-07-03 08:24:16.103179-07	INSERT	\N	\N	\N	f	f
58f7c2a0-9496-4f17-8985-2979fce466cd	generic_instance	\N	7a7a05e0-d926-4739-8f3c-c9a41c715f1a	CWX253	\N	\N	daylily	2025-07-03 08:24:16.108836-07	INSERT	\N	\N	\N	f	f
5072cfd3-83b0-492f-a78b-569830d6fa71	generic_instance	json_addl	3f1c2a4d-7892-4b46-9059-f884aad34e80	CWX79	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:03.197291-07	UPDATE	\N	\N	\N	f	f
71db19a0-7268-48f9-beb2-a36620a2816c	generic_instance	created_dt	3f1c2a4d-7892-4b46-9059-f884aad34e80	CWX79	2025-07-03 08:21:03.192956-07	2025-07-03T08:21:03.192956-07:00	daylily	2025-07-03 08:21:03.197291-07	UPDATE	\N	\N	\N	f	f
a0662d33-e410-40b8-b092-a8c53f1f4e56	generic_instance	modified_dt	3f1c2a4d-7892-4b46-9059-f884aad34e80	CWX79	2025-07-03 08:21:03.192956-07	2025-07-03T08:21:03.197291-07:00	daylily	2025-07-03 08:21:03.197291-07	UPDATE	\N	\N	\N	f	f
ac8907bb-8cee-48a2-a4fc-7e96d71928b1	generic_instance_lineage	\N	90c0564c-0c11-4afc-873c-d31e5aa13fc5	GL266	\N	\N	daylily	2025-07-03 08:21:03.198703-07	INSERT	\N	\N	\N	f	f
ddf29179-6e4b-4dcf-8378-80b1e9a0389f	generic_instance	\N	2f43dd3a-7c4c-4300-a06a-1af3bbf880be	CWX80	\N	\N	daylily	2025-07-03 08:21:03.198703-07	INSERT	\N	\N	\N	f	f
ee8ee714-438f-4114-a10a-8dedadcd332c	generic_instance	json_addl	2f43dd3a-7c4c-4300-a06a-1af3bbf880be	CWX80	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:03.202601-07	UPDATE	\N	\N	\N	f	f
dde10b09-df08-43d6-adc0-46aa28a9d155	generic_instance	created_dt	2f43dd3a-7c4c-4300-a06a-1af3bbf880be	CWX80	2025-07-03 08:21:03.198703-07	2025-07-03T08:21:03.198703-07:00	daylily	2025-07-03 08:21:03.202601-07	UPDATE	\N	\N	\N	f	f
50f6b2df-6cf7-4933-838f-29c58113d9fa	generic_instance	modified_dt	2f43dd3a-7c4c-4300-a06a-1af3bbf880be	CWX80	2025-07-03 08:21:03.198703-07	2025-07-03T08:21:03.202601-07:00	daylily	2025-07-03 08:21:03.202601-07	UPDATE	\N	\N	\N	f	f
9b49d706-bc72-4542-b669-8f14089e9d67	generic_instance_lineage	\N	0e183d54-ca39-4fab-9cf4-2a43c7f15220	GL267	\N	\N	daylily	2025-07-03 08:21:03.203859-07	INSERT	\N	\N	\N	f	f
4f44d3d5-7b40-45aa-a234-6e31e03b09c0	generic_instance	\N	88dc0886-2c10-464c-9b20-0fe485c4223a	CWX81	\N	\N	daylily	2025-07-03 08:21:03.203859-07	INSERT	\N	\N	\N	f	f
8a41b807-7db5-4331-8b65-b1fe13d891b9	generic_instance	json_addl	88dc0886-2c10-464c-9b20-0fe485c4223a	CWX81	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:03.208327-07	UPDATE	\N	\N	\N	f	f
f813cc3a-4066-4a49-a4e0-f03beff644b0	generic_instance	created_dt	88dc0886-2c10-464c-9b20-0fe485c4223a	CWX81	2025-07-03 08:21:03.203859-07	2025-07-03T08:21:03.203859-07:00	daylily	2025-07-03 08:21:03.208327-07	UPDATE	\N	\N	\N	f	f
87edf1af-ebbf-405c-a03e-93d9da03932d	generic_instance	modified_dt	88dc0886-2c10-464c-9b20-0fe485c4223a	CWX81	2025-07-03 08:21:03.203859-07	2025-07-03T08:21:03.208327-07:00	daylily	2025-07-03 08:21:03.208327-07	UPDATE	\N	\N	\N	f	f
5830c0e8-5b1f-46a5-8fde-4ae86aeaf4be	generic_instance_lineage	\N	881652e7-d59f-447c-9b11-da331084f35a	GL268	\N	\N	daylily	2025-07-03 08:21:03.209531-07	INSERT	\N	\N	\N	f	f
b7dec14e-6b3c-4c6d-8009-aabb7b3dc2cf	generic_instance	\N	479bb3b4-e408-4cc2-ab2f-1d56b69ace6e	CWX82	\N	\N	daylily	2025-07-03 08:21:03.209531-07	INSERT	\N	\N	\N	f	f
da6d3a9b-9470-4390-a5bc-e0c0ae44428f	generic_instance_lineage	\N	2e61ac89-4ec1-454a-9f36-d8f1e7fa6e6d	GL695	\N	\N	daylily	2025-07-03 08:24:16.395568-07	INSERT	\N	\N	\N	f	f
ec08b667-c7cd-40d6-a50f-31c6186985c4	generic_instance	\N	e23c847f-8774-4450-9dc9-ec4f7233a6d8	MRX276	\N	\N	daylily	2025-07-03 08:24:16.395568-07	INSERT	\N	\N	\N	f	f
9429eef9-2280-4e20-bd32-ea50aeee55f8	generic_instance_lineage	\N	7cc98bd0-c920-481f-9d19-2bb5e1afde70	GL696	\N	\N	daylily	2025-07-03 08:24:16.400057-07	INSERT	\N	\N	\N	f	f
e6e35d54-4d51-40ad-b79a-df3fca9f61a1	generic_instance	\N	f5228cdc-84b8-4f13-8171-543a997417ec	MRX277	\N	\N	daylily	2025-07-03 08:24:16.400057-07	INSERT	\N	\N	\N	f	f
62cdead8-df69-42b9-9113-7b295f4a0e18	generic_instance_lineage	\N	7981eae4-84ba-4695-84c8-31e3abd962b3	GL697	\N	\N	daylily	2025-07-03 08:24:16.404841-07	INSERT	\N	\N	\N	f	f
b3f104c4-3c15-4be8-9eac-54852b574bf8	generic_instance	\N	7d0c6c07-ac47-47df-b229-c005d7d1685d	MRX278	\N	\N	daylily	2025-07-03 08:24:16.404841-07	INSERT	\N	\N	\N	f	f
8cfa788c-5784-43bb-9d87-8b801e41319e	generic_instance_lineage	\N	6c427ffb-e96a-464d-a307-d5bb52318f11	GL698	\N	\N	daylily	2025-07-03 08:24:16.410484-07	INSERT	\N	\N	\N	f	f
ef7ec3a0-8ada-49e3-8546-cc4aad858a29	generic_instance	json_addl	479bb3b4-e408-4cc2-ab2f-1d56b69ace6e	CWX82	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:03.213821-07	UPDATE	\N	\N	\N	f	f
6043a356-a920-4586-8500-1037fed6a93e	generic_instance	created_dt	479bb3b4-e408-4cc2-ab2f-1d56b69ace6e	CWX82	2025-07-03 08:21:03.209531-07	2025-07-03T08:21:03.209531-07:00	daylily	2025-07-03 08:21:03.213821-07	UPDATE	\N	\N	\N	f	f
d95f8eb5-541c-40d2-8a22-0dd2dcf0b829	generic_instance	modified_dt	479bb3b4-e408-4cc2-ab2f-1d56b69ace6e	CWX82	2025-07-03 08:21:03.209531-07	2025-07-03T08:21:03.213821-07:00	daylily	2025-07-03 08:21:03.213821-07	UPDATE	\N	\N	\N	f	f
e8845f33-b048-41e4-9cd3-e8f8e1edeca0	generic_instance_lineage	\N	ebf4f643-f9f2-4ff6-b993-ee5cee172801	GL269	\N	\N	daylily	2025-07-03 08:21:03.220289-07	INSERT	\N	\N	\N	f	f
b0f696b6-d31c-4003-a90e-973a3f2603c8	generic_instance	\N	9039cfde-66bb-4188-b1ea-e6075bb37449	CWX83	\N	\N	daylily	2025-07-03 08:21:03.220289-07	INSERT	\N	\N	\N	f	f
9bbd0da1-bd3b-4826-81a8-2092a79805a3	generic_instance	json_addl	9039cfde-66bb-4188-b1ea-e6075bb37449	CWX83	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:03.227158-07	UPDATE	\N	\N	\N	f	f
611d99ab-526b-4579-86d8-243e9d6f3368	generic_instance	created_dt	9039cfde-66bb-4188-b1ea-e6075bb37449	CWX83	2025-07-03 08:21:03.220289-07	2025-07-03T08:21:03.220289-07:00	daylily	2025-07-03 08:21:03.227158-07	UPDATE	\N	\N	\N	f	f
4c46088c-353b-40f7-95fb-d89a383df979	generic_instance	modified_dt	9039cfde-66bb-4188-b1ea-e6075bb37449	CWX83	2025-07-03 08:21:03.220289-07	2025-07-03T08:21:03.227158-07:00	daylily	2025-07-03 08:21:03.227158-07	UPDATE	\N	\N	\N	f	f
e96735f3-98d3-41c7-ae96-567df0944dd0	generic_instance_lineage	\N	c367a08e-49cb-41a6-b5be-8c2032a41839	GL270	\N	\N	daylily	2025-07-03 08:21:03.228423-07	INSERT	\N	\N	\N	f	f
478c7210-0607-4bfa-83f5-3301a48ad24d	generic_instance	\N	9409b6fd-e01b-43b1-ab67-77fe15748f54	CWX84	\N	\N	daylily	2025-07-03 08:21:03.228423-07	INSERT	\N	\N	\N	f	f
f0e98c77-c5de-4c6e-b280-caa03d6a8084	generic_instance	json_addl	9409b6fd-e01b-43b1-ab67-77fe15748f54	CWX84	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:21:03.232536-07	UPDATE	\N	\N	\N	f	f
3cdfa228-074d-42c1-ac66-d768653d6344	generic_instance	created_dt	9409b6fd-e01b-43b1-ab67-77fe15748f54	CWX84	2025-07-03 08:21:03.228423-07	2025-07-03T08:21:03.228423-07:00	daylily	2025-07-03 08:21:03.232536-07	UPDATE	\N	\N	\N	f	f
ccd52828-e552-48d2-8dc5-12d7d49d05ab	generic_instance	modified_dt	9409b6fd-e01b-43b1-ab67-77fe15748f54	CWX84	2025-07-03 08:21:03.228423-07	2025-07-03T08:21:03.232536-07:00	daylily	2025-07-03 08:21:03.232536-07	UPDATE	\N	\N	\N	f	f
8cad4822-f923-46e9-aa6a-1cccad4cee55	generic_instance_lineage	\N	4e977570-3baa-4586-92d2-e0c6c6f178e1	GL271	\N	\N	daylily	2025-07-03 08:21:03.233611-07	INSERT	\N	\N	\N	f	f
93e26289-4b46-4fb7-88f1-b8d13e006978	generic_instance	\N	42207d28-a267-4548-b1af-b7e93f4982d0	MRX57	\N	\N	daylily	2025-07-03 08:21:03.235399-07	INSERT	\N	\N	\N	f	f
15f99cdc-7d6a-4873-84f4-f0f34e0e0778	generic_instance_lineage	\N	8838f3c9-c234-4079-a0ab-ff182c5c4a27	GL272	\N	\N	daylily	2025-07-03 08:21:03.23909-07	INSERT	\N	\N	\N	f	f
a7fa0406-e17c-421b-8f8a-f51ca971409e	generic_instance	\N	9d3d6ea3-0299-4b7b-aa33-f361910a58e3	MRX58	\N	\N	daylily	2025-07-03 08:21:03.23909-07	INSERT	\N	\N	\N	f	f
f82ee561-2866-41df-87b0-e3d8a36b114f	generic_instance_lineage	\N	841c638a-f684-4594-a793-c513c1f6e97b	GL273	\N	\N	daylily	2025-07-03 08:21:03.244636-07	INSERT	\N	\N	\N	f	f
12aa2c64-e46a-4d18-a0ba-cf4c6ba8c159	generic_instance	\N	e70c4c91-bcf8-487a-8cc8-45817cd72396	MRX59	\N	\N	daylily	2025-07-03 08:21:03.244636-07	INSERT	\N	\N	\N	f	f
8c9eb637-5ee5-4071-ad92-41602503ded9	generic_instance_lineage	\N	dcdabb27-0af0-418c-9d52-511677e47cde	GL274	\N	\N	daylily	2025-07-03 08:21:03.250346-07	INSERT	\N	\N	\N	f	f
cb752536-60db-4dca-8ab9-370438a6b1c1	generic_instance	\N	ed1d3d5a-e234-4ee2-8c5f-b88300456d8b	MRX60	\N	\N	daylily	2025-07-03 08:21:03.250346-07	INSERT	\N	\N	\N	f	f
1e44bf78-8f9c-4a9e-9da4-6c7114143f89	generic_instance_lineage	\N	3162a194-c8ff-424b-90b6-f49e372d6576	GL275	\N	\N	daylily	2025-07-03 08:21:03.25557-07	INSERT	\N	\N	\N	f	f
a091cf4b-37ad-4a52-81c0-e5b0b7f28d19	generic_instance	json_addl	85fa4716-fb10-4279-9999-ebd9c08f6b20	CWX97	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:08.806605-07	UPDATE	\N	\N	\N	f	f
be25e922-dd6e-47c4-8843-22025d2e4989	generic_instance	created_dt	85fa4716-fb10-4279-9999-ebd9c08f6b20	CWX97	2025-07-03 08:23:08.799621-07	2025-07-03T08:23:08.799621-07:00	daylily	2025-07-03 08:23:08.806605-07	UPDATE	\N	\N	\N	f	f
c039694f-43f3-4d8d-9fee-0a4f66191977	generic_instance	modified_dt	85fa4716-fb10-4279-9999-ebd9c08f6b20	CWX97	2025-07-03 08:23:08.799621-07	2025-07-03T08:23:08.806605-07:00	daylily	2025-07-03 08:23:08.806605-07	UPDATE	\N	\N	\N	f	f
96eacb98-92a9-4521-828b-2393f5100e6b	generic_instance_lineage	\N	a2a85981-bdb9-412b-b9b4-5878c7969e70	GL315	\N	\N	daylily	2025-07-03 08:23:08.808977-07	INSERT	\N	\N	\N	f	f
9d4cac90-9fd7-466e-8568-697fba0cff8d	generic_instance	\N	b4b7ac21-3798-4e5d-8ed4-735cec860516	CWX98	\N	\N	daylily	2025-07-03 08:23:08.808977-07	INSERT	\N	\N	\N	f	f
4fe79764-9a7d-4036-963f-1335008af218	generic_instance	json_addl	b4b7ac21-3798-4e5d-8ed4-735cec860516	CWX98	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:08.815741-07	UPDATE	\N	\N	\N	f	f
14d4b68c-94eb-46bb-b295-46d9a9a05c98	generic_instance	created_dt	b4b7ac21-3798-4e5d-8ed4-735cec860516	CWX98	2025-07-03 08:23:08.808977-07	2025-07-03T08:23:08.808977-07:00	daylily	2025-07-03 08:23:08.815741-07	UPDATE	\N	\N	\N	f	f
03a7e343-ba68-49f2-abfb-de2f95a3a2d4	generic_instance	modified_dt	b4b7ac21-3798-4e5d-8ed4-735cec860516	CWX98	2025-07-03 08:23:08.808977-07	2025-07-03T08:23:08.815741-07:00	daylily	2025-07-03 08:23:08.815741-07	UPDATE	\N	\N	\N	f	f
817b68c0-e2c0-4c88-b2de-663acabfa8eb	generic_instance_lineage	\N	fcba1c1e-024c-4be0-941a-5937852dc425	GL316	\N	\N	daylily	2025-07-03 08:23:08.818645-07	INSERT	\N	\N	\N	f	f
15f5b534-8c17-4304-898e-2b3f6fbbb654	generic_instance	\N	50a01be1-89ed-4814-93c6-de250293b4e6	CWX99	\N	\N	daylily	2025-07-03 08:23:08.818645-07	INSERT	\N	\N	\N	f	f
dfff1153-223c-48c5-8dcc-cdada9456133	generic_instance	json_addl	50a01be1-89ed-4814-93c6-de250293b4e6	CWX99	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:08.823329-07	UPDATE	\N	\N	\N	f	f
49c4e466-9ca1-4929-b07f-e524365d8236	generic_instance	created_dt	50a01be1-89ed-4814-93c6-de250293b4e6	CWX99	2025-07-03 08:23:08.818645-07	2025-07-03T08:23:08.818645-07:00	daylily	2025-07-03 08:23:08.823329-07	UPDATE	\N	\N	\N	f	f
1123defb-df9c-4c8b-8537-d7cfddeb1977	generic_instance	modified_dt	50a01be1-89ed-4814-93c6-de250293b4e6	CWX99	2025-07-03 08:23:08.818645-07	2025-07-03T08:23:08.823329-07:00	daylily	2025-07-03 08:23:08.823329-07	UPDATE	\N	\N	\N	f	f
a91c3db1-0570-47cc-b969-5f071219e2cf	generic_instance_lineage	\N	222726b1-4659-44c6-b6ac-704278ca2587	GL317	\N	\N	daylily	2025-07-03 08:23:08.824708-07	INSERT	\N	\N	\N	f	f
a597001d-458a-4ae1-a9f7-a032398aaea0	generic_instance	\N	1fd00059-c3de-43e5-9cb4-7d4c273343ad	CWX100	\N	\N	daylily	2025-07-03 08:23:08.824708-07	INSERT	\N	\N	\N	f	f
333c498c-6383-4f5c-ad35-d860e5d9236f	generic_instance	\N	c5a56bf4-3463-40b3-8b68-5e47b32a5cff	MRX279	\N	\N	daylily	2025-07-03 08:24:16.410484-07	INSERT	\N	\N	\N	f	f
fd4f07ab-cba4-46ff-99e0-0782f6e7392a	generic_instance_lineage	\N	662e0651-f894-4aad-a4ad-59700fc4ec95	GL699	\N	\N	daylily	2025-07-03 08:24:16.415536-07	INSERT	\N	\N	\N	f	f
9d8371b1-d531-4e5f-b14f-5e6db73906c6	generic_instance	\N	1b97a184-a3da-4d82-a043-a02a95207c0e	MRX280	\N	\N	daylily	2025-07-03 08:24:16.415536-07	INSERT	\N	\N	\N	f	f
b074dc92-f560-4d9d-af1c-b47ebaa94ad9	generic_instance_lineage	\N	417511eb-04d3-4027-ba31-6ba3da9e772d	GL700	\N	\N	daylily	2025-07-03 08:24:16.421056-07	INSERT	\N	\N	\N	f	f
db473ca7-1641-48c7-a485-c251eb9b0f34	generic_instance	json_addl	1fd00059-c3de-43e5-9cb4-7d4c273343ad	CWX100	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:08.82929-07	UPDATE	\N	\N	\N	f	f
bc7ad15e-8759-4e2c-a220-211b3cb6eb38	generic_instance	created_dt	1fd00059-c3de-43e5-9cb4-7d4c273343ad	CWX100	2025-07-03 08:23:08.824708-07	2025-07-03T08:23:08.824708-07:00	daylily	2025-07-03 08:23:08.82929-07	UPDATE	\N	\N	\N	f	f
324e1473-5032-4a1a-8111-dd0a7b3ff6a9	generic_instance	modified_dt	1fd00059-c3de-43e5-9cb4-7d4c273343ad	CWX100	2025-07-03 08:23:08.824708-07	2025-07-03T08:23:08.82929-07:00	daylily	2025-07-03 08:23:08.82929-07	UPDATE	\N	\N	\N	f	f
692f7038-4deb-4407-85c8-7d9d36f9581d	generic_instance_lineage	\N	a8736dda-29e9-4b62-b8f3-af9869d44f6b	GL318	\N	\N	daylily	2025-07-03 08:23:08.831455-07	INSERT	\N	\N	\N	f	f
68638a0b-737d-45f0-9c22-537ed862f0e9	generic_instance	\N	53c38748-0c95-4d1d-bb0c-83201d2d54db	CWX101	\N	\N	daylily	2025-07-03 08:23:08.831455-07	INSERT	\N	\N	\N	f	f
14fb5998-9f23-47c2-8983-b343350e552c	generic_instance	json_addl	53c38748-0c95-4d1d-bb0c-83201d2d54db	CWX101	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:08.836095-07	UPDATE	\N	\N	\N	f	f
0505679e-3a13-4afa-96e0-7b3979701edf	generic_instance	created_dt	53c38748-0c95-4d1d-bb0c-83201d2d54db	CWX101	2025-07-03 08:23:08.831455-07	2025-07-03T08:23:08.831455-07:00	daylily	2025-07-03 08:23:08.836095-07	UPDATE	\N	\N	\N	f	f
f34ffe67-1c5f-4323-b3d8-bf4d12ae038e	generic_instance	modified_dt	53c38748-0c95-4d1d-bb0c-83201d2d54db	CWX101	2025-07-03 08:23:08.831455-07	2025-07-03T08:23:08.836095-07:00	daylily	2025-07-03 08:23:08.836095-07	UPDATE	\N	\N	\N	f	f
e4a8d40a-e49e-4e35-a77d-513a98f937fb	generic_instance_lineage	\N	d70377da-64d3-4af3-b6c9-080adf9e5495	GL319	\N	\N	daylily	2025-07-03 08:23:08.837365-07	INSERT	\N	\N	\N	f	f
6f80b44f-19d9-4c60-a8d1-92cc4ab8ea9a	generic_instance	\N	56bc2733-1be4-4f33-8c35-2b522209ae67	CWX102	\N	\N	daylily	2025-07-03 08:23:08.837365-07	INSERT	\N	\N	\N	f	f
9d481b78-0b3e-47b5-9887-396ce054a11c	generic_instance	json_addl	56bc2733-1be4-4f33-8c35-2b522209ae67	CWX102	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:08.841445-07	UPDATE	\N	\N	\N	f	f
51ed4783-7c1f-4614-86ab-b61fd7d47060	generic_instance	created_dt	56bc2733-1be4-4f33-8c35-2b522209ae67	CWX102	2025-07-03 08:23:08.837365-07	2025-07-03T08:23:08.837365-07:00	daylily	2025-07-03 08:23:08.841445-07	UPDATE	\N	\N	\N	f	f
2c9ac031-0283-4563-9d47-40e6f05d44d4	generic_instance	modified_dt	56bc2733-1be4-4f33-8c35-2b522209ae67	CWX102	2025-07-03 08:23:08.837365-07	2025-07-03T08:23:08.841445-07:00	daylily	2025-07-03 08:23:08.841445-07	UPDATE	\N	\N	\N	f	f
04dd3e51-d516-4a08-a7a6-6b657c8f16d5	generic_instance_lineage	\N	beb97add-7667-4b29-9254-37ffc0604634	GL320	\N	\N	daylily	2025-07-03 08:23:08.842566-07	INSERT	\N	\N	\N	f	f
9948ea7c-248d-492f-86bb-58b2cf32c1e3	generic_instance	\N	72061fa5-3321-4a96-b64c-275d915fc9ee	CWX103	\N	\N	daylily	2025-07-03 08:23:08.842566-07	INSERT	\N	\N	\N	f	f
ba52c4cf-19e1-4e4f-acb1-96b573fd503a	generic_instance	\N	b9d2526e-ab8a-4ff8-a1b0-947071f5d2d7	CX19	\N	\N	daylily	2025-07-03 08:24:16.425128-07	INSERT	\N	\N	\N	f	f
1dfe64e5-9dca-434e-a205-0104998fabfa	generic_instance	\N	845ebde8-23d0-447c-a05e-538a37a81f97	MCX7	\N	\N	daylily	2025-07-03 08:24:16.432998-07	INSERT	\N	\N	\N	f	f
2a74a0c2-5263-4735-a30b-1b576129439d	generic_instance	json_addl	72061fa5-3321-4a96-b64c-275d915fc9ee	CWX103	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:08.84644-07	UPDATE	\N	\N	\N	f	f
04cb4824-2c8f-45ff-a665-7eadbbb2dd9d	generic_instance	created_dt	72061fa5-3321-4a96-b64c-275d915fc9ee	CWX103	2025-07-03 08:23:08.842566-07	2025-07-03T08:23:08.842566-07:00	daylily	2025-07-03 08:23:08.84644-07	UPDATE	\N	\N	\N	f	f
3db88953-096e-4f95-9e05-0a364035f122	generic_instance	modified_dt	72061fa5-3321-4a96-b64c-275d915fc9ee	CWX103	2025-07-03 08:23:08.842566-07	2025-07-03T08:23:08.84644-07:00	daylily	2025-07-03 08:23:08.84644-07	UPDATE	\N	\N	\N	f	f
bcbe495b-52a7-4aab-8dff-d95a44f1359f	generic_instance_lineage	\N	beac40fc-cbef-451f-8ab7-863bd606db77	GL321	\N	\N	daylily	2025-07-03 08:23:08.847831-07	INSERT	\N	\N	\N	f	f
12e92b61-ea29-4848-ab8d-3bbaf54f490e	generic_instance	\N	e6d570a4-5fcf-4887-b23c-31b36a64854c	CWX104	\N	\N	daylily	2025-07-03 08:23:08.847831-07	INSERT	\N	\N	\N	f	f
a53d791a-e37e-4a93-ab64-c87e7fa59c7a	generic_instance	json_addl	e6d570a4-5fcf-4887-b23c-31b36a64854c	CWX104	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:08.851964-07	UPDATE	\N	\N	\N	f	f
14fc4442-1f06-4160-bdde-55d3f02f26db	generic_instance	created_dt	e6d570a4-5fcf-4887-b23c-31b36a64854c	CWX104	2025-07-03 08:23:08.847831-07	2025-07-03T08:23:08.847831-07:00	daylily	2025-07-03 08:23:08.851964-07	UPDATE	\N	\N	\N	f	f
f7116591-ee84-4a5d-992d-608cfb570771	generic_instance	modified_dt	e6d570a4-5fcf-4887-b23c-31b36a64854c	CWX104	2025-07-03 08:23:08.847831-07	2025-07-03T08:23:08.851964-07:00	daylily	2025-07-03 08:23:08.851964-07	UPDATE	\N	\N	\N	f	f
8c9b6e40-ed4f-4726-ad43-9bd0b541cba2	generic_instance_lineage	\N	10b53fe6-003b-4068-82a5-ca3a0e5de42c	GL322	\N	\N	daylily	2025-07-03 08:23:08.853275-07	INSERT	\N	\N	\N	f	f
049e48fe-f828-4112-880b-b57dd80126ef	generic_instance	\N	92d78938-91b6-4aee-a8b4-a74e51ad8b95	CWX105	\N	\N	daylily	2025-07-03 08:23:08.853275-07	INSERT	\N	\N	\N	f	f
ec8ca7dd-165a-46da-9d3c-ab5dd36efb8c	generic_instance	json_addl	92d78938-91b6-4aee-a8b4-a74e51ad8b95	CWX105	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:08.857328-07	UPDATE	\N	\N	\N	f	f
f501c577-e141-4d97-8398-f16956da9cbc	generic_instance	created_dt	92d78938-91b6-4aee-a8b4-a74e51ad8b95	CWX105	2025-07-03 08:23:08.853275-07	2025-07-03T08:23:08.853275-07:00	daylily	2025-07-03 08:23:08.857328-07	UPDATE	\N	\N	\N	f	f
a0148873-1af4-4837-8a54-3528d9035306	generic_instance	modified_dt	92d78938-91b6-4aee-a8b4-a74e51ad8b95	CWX105	2025-07-03 08:23:08.853275-07	2025-07-03T08:23:08.857328-07:00	daylily	2025-07-03 08:23:08.857328-07	UPDATE	\N	\N	\N	f	f
b6fc9b08-5893-4abe-b5a7-56777f609d03	generic_instance_lineage	\N	4f9a6005-3c45-4eee-9a6b-44808a8402bd	GL323	\N	\N	daylily	2025-07-03 08:23:08.858591-07	INSERT	\N	\N	\N	f	f
560c4907-a670-48ba-b63c-67047c44c5b5	generic_instance	\N	5dff72b8-b5c4-4662-b097-b09c106854c6	CWX106	\N	\N	daylily	2025-07-03 08:23:08.858591-07	INSERT	\N	\N	\N	f	f
84a76cf7-8e8f-451f-84d7-53175cdf19f3	generic_instance	json_addl	5dff72b8-b5c4-4662-b097-b09c106854c6	CWX106	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:08.862637-07	UPDATE	\N	\N	\N	f	f
4918f1de-8109-4068-b419-2f809c767e04	generic_instance	created_dt	5dff72b8-b5c4-4662-b097-b09c106854c6	CWX106	2025-07-03 08:23:08.858591-07	2025-07-03T08:23:08.858591-07:00	daylily	2025-07-03 08:23:08.862637-07	UPDATE	\N	\N	\N	f	f
a681c343-2243-4b44-b144-b6e8c673f2a7	generic_instance	modified_dt	5dff72b8-b5c4-4662-b097-b09c106854c6	CWX106	2025-07-03 08:23:08.858591-07	2025-07-03T08:23:08.862637-07:00	daylily	2025-07-03 08:23:08.862637-07	UPDATE	\N	\N	\N	f	f
de4877a5-e361-4008-840d-98453303348f	generic_instance_lineage	\N	46179fbe-cf4b-4434-a326-59fb46c8ebe8	GL324	\N	\N	daylily	2025-07-03 08:23:08.863842-07	INSERT	\N	\N	\N	f	f
73991a9e-5fac-4acf-9bff-8b582b3ab3e2	generic_instance	\N	974b84e5-0ee2-4545-8ff0-2de08d139c6f	CWX107	\N	\N	daylily	2025-07-03 08:23:08.863842-07	INSERT	\N	\N	\N	f	f
193dddbd-011b-4982-bac0-72a448cd19fb	generic_instance	json_addl	974b84e5-0ee2-4545-8ff0-2de08d139c6f	CWX107	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:08.868073-07	UPDATE	\N	\N	\N	f	f
a599cbd0-4e08-4cd5-8f41-d319bbef5664	generic_instance	created_dt	974b84e5-0ee2-4545-8ff0-2de08d139c6f	CWX107	2025-07-03 08:23:08.863842-07	2025-07-03T08:23:08.863842-07:00	daylily	2025-07-03 08:23:08.868073-07	UPDATE	\N	\N	\N	f	f
a42048c3-0f46-42c6-8276-b8d7bc000f24	generic_instance	modified_dt	974b84e5-0ee2-4545-8ff0-2de08d139c6f	CWX107	2025-07-03 08:23:08.863842-07	2025-07-03T08:23:08.868073-07:00	daylily	2025-07-03 08:23:08.868073-07	UPDATE	\N	\N	\N	f	f
ebcb468d-73f9-4cc7-9f4b-0d04cd199479	generic_instance_lineage	\N	8179c198-0681-42c1-9b15-6234dfe69da3	GL325	\N	\N	daylily	2025-07-03 08:23:08.869272-07	INSERT	\N	\N	\N	f	f
4d309f89-ea64-46da-8239-3814f3dc382b	generic_instance	\N	32986026-4b6a-4a23-a81c-ae11fd925582	CWX108	\N	\N	daylily	2025-07-03 08:23:08.869272-07	INSERT	\N	\N	\N	f	f
c478f1da-8cc6-425d-bf38-164a44e06dc6	generic_instance	json_addl	32986026-4b6a-4a23-a81c-ae11fd925582	CWX108	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:08.872983-07	UPDATE	\N	\N	\N	f	f
464e2609-2ecd-4ba3-8506-9ca7b1f0df97	generic_instance	created_dt	32986026-4b6a-4a23-a81c-ae11fd925582	CWX108	2025-07-03 08:23:08.869272-07	2025-07-03T08:23:08.869272-07:00	daylily	2025-07-03 08:23:08.872983-07	UPDATE	\N	\N	\N	f	f
2d2ae6e9-c4cf-4993-9b20-e2f4d46e5d49	generic_instance	modified_dt	32986026-4b6a-4a23-a81c-ae11fd925582	CWX108	2025-07-03 08:23:08.869272-07	2025-07-03T08:23:08.872983-07:00	daylily	2025-07-03 08:23:08.872983-07	UPDATE	\N	\N	\N	f	f
3d001e46-c418-4390-b121-835b3403001c	generic_instance_lineage	\N	85d8ef1d-1e34-4970-b863-0119887cb537	GL326	\N	\N	daylily	2025-07-03 08:23:08.874087-07	INSERT	\N	\N	\N	f	f
8aa889e6-d84e-48b7-821a-4bb40c5dfe05	generic_instance	\N	8fd06f8b-3195-4a98-882f-49d1c9a4834b	CWX109	\N	\N	daylily	2025-07-03 08:23:08.874087-07	INSERT	\N	\N	\N	f	f
9052f0da-9fe1-4c54-ba98-42bae4b89522	generic_instance	json_addl	8fd06f8b-3195-4a98-882f-49d1c9a4834b	CWX109	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:08.877679-07	UPDATE	\N	\N	\N	f	f
b722660d-3841-4de7-a8c6-98d268ba1a02	generic_instance	created_dt	8fd06f8b-3195-4a98-882f-49d1c9a4834b	CWX109	2025-07-03 08:23:08.874087-07	2025-07-03T08:23:08.874087-07:00	daylily	2025-07-03 08:23:08.877679-07	UPDATE	\N	\N	\N	f	f
cbde20d9-d3a5-42a9-baf2-0ff2a59e43a3	generic_instance	modified_dt	8fd06f8b-3195-4a98-882f-49d1c9a4834b	CWX109	2025-07-03 08:23:08.874087-07	2025-07-03T08:23:08.877679-07:00	daylily	2025-07-03 08:23:08.877679-07	UPDATE	\N	\N	\N	f	f
e2a8d4da-b07c-4380-8c46-5a2ae16f3056	generic_instance_lineage	\N	1ab2da74-9e5d-417d-8628-f23b553a8990	GL327	\N	\N	daylily	2025-07-03 08:23:08.878751-07	INSERT	\N	\N	\N	f	f
2365abb0-361c-4362-97cb-ae2930846b7e	generic_instance	\N	b7122674-3965-4b7f-b5be-c6b1c869e1c5	CWX110	\N	\N	daylily	2025-07-03 08:23:08.878751-07	INSERT	\N	\N	\N	f	f
6fc2859e-6389-4921-8c43-a8d3b4b1a5d2	generic_instance	json_addl	b7122674-3965-4b7f-b5be-c6b1c869e1c5	CWX110	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:08.882174-07	UPDATE	\N	\N	\N	f	f
d27d659f-11be-47f6-bc62-a7e4fbcaa4fa	generic_instance	created_dt	b7122674-3965-4b7f-b5be-c6b1c869e1c5	CWX110	2025-07-03 08:23:08.878751-07	2025-07-03T08:23:08.878751-07:00	daylily	2025-07-03 08:23:08.882174-07	UPDATE	\N	\N	\N	f	f
275e4157-935c-45c2-8c59-7f8beed33623	generic_instance	modified_dt	b7122674-3965-4b7f-b5be-c6b1c869e1c5	CWX110	2025-07-03 08:23:08.878751-07	2025-07-03T08:23:08.882174-07:00	daylily	2025-07-03 08:23:08.882174-07	UPDATE	\N	\N	\N	f	f
6f426774-d57e-4cce-9558-06822eb92ccc	generic_instance_lineage	\N	04464188-a83d-4e7e-9c68-2c94bd16a0ce	GL328	\N	\N	daylily	2025-07-03 08:23:08.883518-07	INSERT	\N	\N	\N	f	f
012b0f07-c4a2-40b1-838f-0549f32de059	generic_instance	\N	de7745e0-b7ec-49e7-8436-e6f0152453b6	CWX111	\N	\N	daylily	2025-07-03 08:23:08.883518-07	INSERT	\N	\N	\N	f	f
3067d6dc-fe24-4f4c-ac2b-85f5537af75c	generic_instance	json_addl	de7745e0-b7ec-49e7-8436-e6f0152453b6	CWX111	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:08.887717-07	UPDATE	\N	\N	\N	f	f
ff4a3321-654f-4ecc-83b3-c97dd58b71e2	generic_instance	created_dt	de7745e0-b7ec-49e7-8436-e6f0152453b6	CWX111	2025-07-03 08:23:08.883518-07	2025-07-03T08:23:08.883518-07:00	daylily	2025-07-03 08:23:08.887717-07	UPDATE	\N	\N	\N	f	f
6293ac2d-d62e-4744-bd7f-c93ec5712ac3	generic_instance	modified_dt	de7745e0-b7ec-49e7-8436-e6f0152453b6	CWX111	2025-07-03 08:23:08.883518-07	2025-07-03T08:23:08.887717-07:00	daylily	2025-07-03 08:23:08.887717-07	UPDATE	\N	\N	\N	f	f
5050060e-4615-4585-963d-b83ca366baac	generic_instance_lineage	\N	f4b4bebf-5690-4124-ae96-e757f67ab71d	GL329	\N	\N	daylily	2025-07-03 08:23:08.8891-07	INSERT	\N	\N	\N	f	f
6bc7ea18-6bfd-4720-a047-39f4b1fef178	generic_instance	\N	78c3c9b5-ea50-4c51-8560-9929af6a202d	CWX112	\N	\N	daylily	2025-07-03 08:23:08.8891-07	INSERT	\N	\N	\N	f	f
fa1a0bca-d362-433f-a2df-680d64af1299	generic_instance	json_addl	78c3c9b5-ea50-4c51-8560-9929af6a202d	CWX112	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:08.893199-07	UPDATE	\N	\N	\N	f	f
6705e2a7-2f9f-4fb0-9537-b490f58d06bd	generic_instance	created_dt	78c3c9b5-ea50-4c51-8560-9929af6a202d	CWX112	2025-07-03 08:23:08.8891-07	2025-07-03T08:23:08.8891-07:00	daylily	2025-07-03 08:23:08.893199-07	UPDATE	\N	\N	\N	f	f
4d677c79-43f5-43c5-b6fc-0bf64edb9dbb	generic_instance	modified_dt	78c3c9b5-ea50-4c51-8560-9929af6a202d	CWX112	2025-07-03 08:23:08.8891-07	2025-07-03T08:23:08.893199-07:00	daylily	2025-07-03 08:23:08.893199-07	UPDATE	\N	\N	\N	f	f
c0262534-d821-43db-aab9-e4de3858862e	generic_instance_lineage	\N	2b96af66-a954-43b7-9879-ad233e56b391	GL330	\N	\N	daylily	2025-07-03 08:23:08.894521-07	INSERT	\N	\N	\N	f	f
6bc8ced9-1e2f-4cf6-a34c-e19e1bc8f388	generic_instance	\N	4366985c-1c27-4e8e-8f17-5f2bb448f0d2	MRX85	\N	\N	daylily	2025-07-03 08:23:08.897574-07	INSERT	\N	\N	\N	f	f
1f3a9235-e8fc-4085-aeaf-fe55a4ac60df	generic_instance_lineage	\N	edeabf86-33b6-44a3-a590-5b2fd61df19f	GL331	\N	\N	daylily	2025-07-03 08:23:08.901124-07	INSERT	\N	\N	\N	f	f
c641ac2b-bd28-47b6-9bd2-216e65fa6fe6	generic_instance	\N	5753d5e2-98b7-44e0-88ce-6ee0096cf3ee	MRX86	\N	\N	daylily	2025-07-03 08:23:08.901124-07	INSERT	\N	\N	\N	f	f
8721b040-d344-4b3f-b590-841a8a628d73	generic_instance_lineage	\N	7d6f7ec7-5baf-4825-bda5-04224bb9460c	GL332	\N	\N	daylily	2025-07-03 08:23:08.907676-07	INSERT	\N	\N	\N	f	f
f102e2e4-a725-45f8-b98d-64aa543b26c2	generic_instance	\N	06518ad4-6d64-4cd9-ad67-13c330248de9	MRX87	\N	\N	daylily	2025-07-03 08:23:08.907676-07	INSERT	\N	\N	\N	f	f
a6f42378-ef28-416c-a18f-c74df2f9fbc0	generic_instance_lineage	\N	ee3b566e-9c9a-4310-ab79-44f89489453c	GL333	\N	\N	daylily	2025-07-03 08:23:08.912175-07	INSERT	\N	\N	\N	f	f
3ea487f2-f817-4845-91d6-59ffe6a28b9f	generic_instance	\N	cb8e67ed-339b-4c03-b725-1a60d7ad9e54	MRX88	\N	\N	daylily	2025-07-03 08:23:08.912175-07	INSERT	\N	\N	\N	f	f
2272e534-0189-4ab3-8b67-9d93fbd5e02c	generic_instance_lineage	\N	8a450bd4-2d67-4314-84fa-9e096cc6149f	GL334	\N	\N	daylily	2025-07-03 08:23:08.916028-07	INSERT	\N	\N	\N	f	f
f4dc8a54-c8c2-457c-8160-b3e330ffcbfa	generic_instance	\N	b2421416-938d-46ed-bcc4-1aca83a2e2ab	MRX89	\N	\N	daylily	2025-07-03 08:23:08.916028-07	INSERT	\N	\N	\N	f	f
64381c48-d402-404b-95f9-d35bc9dad3e5	generic_instance_lineage	\N	a91c15e0-0d49-4e69-91d7-7f386b35eb0a	GL335	\N	\N	daylily	2025-07-03 08:23:08.92017-07	INSERT	\N	\N	\N	f	f
bf269a31-98f5-4904-975c-9415bfac7609	generic_instance	\N	7ea54466-0a39-4279-becc-c379d09d80cc	MRX90	\N	\N	daylily	2025-07-03 08:23:08.92017-07	INSERT	\N	\N	\N	f	f
50c16699-94a4-40ec-be68-fccf2f20debd	generic_instance_lineage	\N	2db7a90c-1a11-4989-95e9-35419eeb39f9	GL336	\N	\N	daylily	2025-07-03 08:23:08.924004-07	INSERT	\N	\N	\N	f	f
3768bb98-5310-4554-8cc4-128b1eba3e71	generic_instance	\N	878e54ab-5d67-4d62-8602-5d9700906f19	MRX91	\N	\N	daylily	2025-07-03 08:23:08.924004-07	INSERT	\N	\N	\N	f	f
9e26cd0a-b1f7-410f-b0a0-072b49bf94a2	generic_instance_lineage	\N	295a6f67-0a29-4b88-96e2-c6362e7c325f	GL337	\N	\N	daylily	2025-07-03 08:23:08.927791-07	INSERT	\N	\N	\N	f	f
8c77609c-5d97-4697-91c4-b75350235f25	generic_instance	\N	ee71e870-facd-47db-a77a-0005c2199964	MRX92	\N	\N	daylily	2025-07-03 08:23:08.927791-07	INSERT	\N	\N	\N	f	f
a68875c5-58dc-42eb-84da-5b2f0ee66f67	generic_instance_lineage	\N	689af8bc-b8ac-46b8-8a96-7a87f191a7e8	GL338	\N	\N	daylily	2025-07-03 08:23:08.93191-07	INSERT	\N	\N	\N	f	f
dfab933d-8fba-4105-8095-18339897176c	generic_instance	\N	64e1cb24-3e88-4abb-807e-4717e3f32b73	MRX93	\N	\N	daylily	2025-07-03 08:23:08.93191-07	INSERT	\N	\N	\N	f	f
bdf3f96c-f7c3-4d27-8095-016a28bb4af2	generic_instance_lineage	\N	fe7a8b20-63eb-4386-a27d-c83949a04533	GL339	\N	\N	daylily	2025-07-03 08:23:08.935993-07	INSERT	\N	\N	\N	f	f
4aa170ea-65c3-4b7c-87e5-1e322a0f21bb	generic_instance	\N	968e67a1-9b12-4960-83a7-912ca28870e3	MRX94	\N	\N	daylily	2025-07-03 08:23:08.935993-07	INSERT	\N	\N	\N	f	f
74fd11c9-bad7-4dcf-880a-6102d7b0774a	generic_instance_lineage	\N	88a734ce-3de2-4eb7-9806-bab64116c701	GL340	\N	\N	daylily	2025-07-03 08:23:08.940342-07	INSERT	\N	\N	\N	f	f
de48da80-c876-4173-a890-72832dcc3fbd	generic_instance	\N	109cb0f0-2f2b-4a27-992d-38184b13d248	MRX95	\N	\N	daylily	2025-07-03 08:23:08.940342-07	INSERT	\N	\N	\N	f	f
41668953-8790-45ce-815a-c9b80bde9b30	generic_instance_lineage	\N	eaf76fd7-118e-4dbf-897c-f4a1b4930935	GL341	\N	\N	daylily	2025-07-03 08:23:08.944735-07	INSERT	\N	\N	\N	f	f
9b2c3f38-9571-43b7-b89c-e23c9c474ef0	generic_instance	\N	df54fb98-5e91-448f-81c4-c19586df73ee	MRX96	\N	\N	daylily	2025-07-03 08:23:08.944735-07	INSERT	\N	\N	\N	f	f
2f006450-ad32-42c8-9983-085f4ec9251d	generic_instance_lineage	\N	c20c35d3-4f9d-46bf-8317-94bd9413f0d6	GL342	\N	\N	daylily	2025-07-03 08:23:08.94886-07	INSERT	\N	\N	\N	f	f
cb787984-2732-4703-9466-65efffeb6a8f	generic_instance	\N	5d354a84-be84-403e-8777-7b9a2520761f	MRX97	\N	\N	daylily	2025-07-03 08:23:08.94886-07	INSERT	\N	\N	\N	f	f
bafcaf22-70c6-41e6-b341-c071200ab3cd	generic_instance_lineage	\N	5fa030fd-745e-4694-a871-66342e421b9d	GL343	\N	\N	daylily	2025-07-03 08:23:08.95349-07	INSERT	\N	\N	\N	f	f
f7db6947-82b8-4036-8c14-9d5c9b438556	generic_instance	\N	94807e9b-79b5-43ee-928c-b1977a3f96f5	MRX98	\N	\N	daylily	2025-07-03 08:23:08.95349-07	INSERT	\N	\N	\N	f	f
77f324f7-ceef-463e-b395-39a0ff1bd736	generic_instance_lineage	\N	9023c24c-9905-4973-beee-b40b7271f8b2	GL344	\N	\N	daylily	2025-07-03 08:23:08.957729-07	INSERT	\N	\N	\N	f	f
1a8cb652-afcb-4e59-ad84-531035bb3d8d	generic_instance	\N	7de1ceef-160a-464e-92ca-2b68522b3eba	MRX99	\N	\N	daylily	2025-07-03 08:23:08.957729-07	INSERT	\N	\N	\N	f	f
e1693df9-d407-45ee-a71f-04efe8fd6bb1	generic_instance_lineage	\N	e36d8a9f-f261-4f80-9459-4e21d1ac8284	GL345	\N	\N	daylily	2025-07-03 08:23:08.962449-07	INSERT	\N	\N	\N	f	f
74a59716-8472-4cfe-ac36-161a8cbfc1a1	generic_instance	\N	733386da-f0db-4671-9ea4-5dda616d2631	MRX100	\N	\N	daylily	2025-07-03 08:23:08.962449-07	INSERT	\N	\N	\N	f	f
5e982e62-c92d-4a16-84c4-f0196aaf660d	generic_instance_lineage	\N	d7373e5b-8f66-444e-a2ce-c353641f3149	GL346	\N	\N	daylily	2025-07-03 08:23:08.967393-07	INSERT	\N	\N	\N	f	f
f4172db9-dabd-4367-9f7e-dbb162cc7e83	generic_instance	\N	64c5fd73-d9fc-4c22-9f2e-1f91998a02c1	MRX101	\N	\N	daylily	2025-07-03 08:23:08.967393-07	INSERT	\N	\N	\N	f	f
c4a5ddfa-c852-4586-844d-e801149bf3fb	generic_instance_lineage	\N	f98baef5-46bd-44a9-85d5-b6a98e09c842	GL347	\N	\N	daylily	2025-07-03 08:23:08.971744-07	INSERT	\N	\N	\N	f	f
7a1794a3-1d0d-4915-abf1-dc450f3d79ec	generic_instance	\N	148a355a-e554-4686-85f1-484bb92e12eb	MRX102	\N	\N	daylily	2025-07-03 08:23:08.971744-07	INSERT	\N	\N	\N	f	f
13f9f30a-e712-40a8-bbe9-432ee210a370	generic_instance_lineage	\N	5127cd1d-6907-499e-b7e1-30808e04528b	GL348	\N	\N	daylily	2025-07-03 08:23:08.976128-07	INSERT	\N	\N	\N	f	f
5001c279-ca40-405e-8da2-6dc36059f922	generic_instance	\N	8259854f-3495-44e0-8850-57466281e5a5	MRX103	\N	\N	daylily	2025-07-03 08:23:08.976128-07	INSERT	\N	\N	\N	f	f
0f860f3f-4739-4781-ba9a-cc3ca1a9d67c	generic_instance_lineage	\N	d649b2f7-f439-4b04-b383-d8927030277c	GL349	\N	\N	daylily	2025-07-03 08:23:08.980265-07	INSERT	\N	\N	\N	f	f
48b3f247-83b9-4df3-8beb-b16f0715170b	generic_instance	\N	a9098742-f68e-4d41-8ad5-58b225c762ec	MRX104	\N	\N	daylily	2025-07-03 08:23:08.980265-07	INSERT	\N	\N	\N	f	f
646e03df-24cd-445b-b8df-63db64f0a064	generic_instance_lineage	\N	2fa16d80-d276-4209-90fd-a7a87771ee22	GL350	\N	\N	daylily	2025-07-03 08:23:08.984738-07	INSERT	\N	\N	\N	f	f
cb8b6206-2440-415d-97a5-55ef00706366	generic_instance	\N	4da41248-557d-40ac-9a52-c0deaea86440	MRX105	\N	\N	daylily	2025-07-03 08:23:08.984738-07	INSERT	\N	\N	\N	f	f
4ed8eb43-f50f-47d5-8aee-fb12b323e276	generic_instance_lineage	\N	05b798ea-0c44-4bd8-8524-25ba31e10209	GL351	\N	\N	daylily	2025-07-03 08:23:08.989051-07	INSERT	\N	\N	\N	f	f
2900969d-101b-4bca-bca5-359e3db1dc2b	generic_instance	\N	c02979bf-e445-4ee7-8505-67bf010865fd	MRX106	\N	\N	daylily	2025-07-03 08:23:08.989051-07	INSERT	\N	\N	\N	f	f
f50890fe-a17b-43c1-8d9d-a923ebe499b4	generic_instance_lineage	\N	354556e6-e1b3-447d-bc8e-073ca8c80e08	GL352	\N	\N	daylily	2025-07-03 08:23:08.993641-07	INSERT	\N	\N	\N	f	f
054946c6-9586-4d67-b260-a2df4fd6f08c	generic_instance	\N	431e9885-70e6-4269-9e55-5a6b6de2a8ea	MRX107	\N	\N	daylily	2025-07-03 08:23:08.993641-07	INSERT	\N	\N	\N	f	f
ab321284-4fb8-47f8-98fe-174f2111873c	generic_instance	json_addl	b806e72c-8000-4615-b23f-b63a79d2da4d	CWX113	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:09.030328-07	UPDATE	\N	\N	\N	f	f
fd4abb79-7eab-411d-952c-468b873e9d4a	generic_instance	created_dt	b806e72c-8000-4615-b23f-b63a79d2da4d	CWX113	2025-07-03 08:23:09.027734-07	2025-07-03T08:23:09.027734-07:00	daylily	2025-07-03 08:23:09.030328-07	UPDATE	\N	\N	\N	f	f
cd994c4f-1714-4e16-b6bc-1d5f3972f205	generic_instance	modified_dt	b806e72c-8000-4615-b23f-b63a79d2da4d	CWX113	2025-07-03 08:23:09.027734-07	2025-07-03T08:23:09.030328-07:00	daylily	2025-07-03 08:23:09.030328-07	UPDATE	\N	\N	\N	f	f
26eaea94-3620-45ff-9bec-d1e3880d221d	generic_instance_lineage	\N	cdb5cec8-fdbd-4b2c-8f26-c0621c14b9a3	GL359	\N	\N	daylily	2025-07-03 08:23:09.031758-07	INSERT	\N	\N	\N	f	f
ac06b7d5-75ec-420f-be8a-f1e7153e177b	generic_instance	\N	701b1431-fa54-4f76-9076-c445285f8ff5	CWX114	\N	\N	daylily	2025-07-03 08:23:09.031758-07	INSERT	\N	\N	\N	f	f
a8c6c4f1-b038-4f43-9a91-fbaafb540af2	generic_instance	json_addl	701b1431-fa54-4f76-9076-c445285f8ff5	CWX114	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:09.036299-07	UPDATE	\N	\N	\N	f	f
f17b4439-d04e-4e5b-a93a-9a3281cd511c	generic_instance	created_dt	701b1431-fa54-4f76-9076-c445285f8ff5	CWX114	2025-07-03 08:23:09.031758-07	2025-07-03T08:23:09.031758-07:00	daylily	2025-07-03 08:23:09.036299-07	UPDATE	\N	\N	\N	f	f
00b3b166-41ae-47cb-80a7-7b40a51567d0	generic_instance	modified_dt	701b1431-fa54-4f76-9076-c445285f8ff5	CWX114	2025-07-03 08:23:09.031758-07	2025-07-03T08:23:09.036299-07:00	daylily	2025-07-03 08:23:09.036299-07	UPDATE	\N	\N	\N	f	f
5fa7e096-fd52-42a1-abe1-6569e73e8a81	generic_instance_lineage	\N	fec6528b-07bf-4b1e-ac36-168b46d1e7e6	GL360	\N	\N	daylily	2025-07-03 08:23:09.037395-07	INSERT	\N	\N	\N	f	f
93372da5-7091-42ff-ab60-e1301e51689b	generic_instance	\N	6a2dc8d6-7470-4baa-b4d8-cd0e20761c7d	CWX115	\N	\N	daylily	2025-07-03 08:23:09.037395-07	INSERT	\N	\N	\N	f	f
b6a5da9a-2c83-4232-a7ac-afdbce7b0f01	generic_instance	json_addl	6a2dc8d6-7470-4baa-b4d8-cd0e20761c7d	CWX115	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:09.041179-07	UPDATE	\N	\N	\N	f	f
1c6a23ed-8986-44f6-a110-746000559172	generic_instance	created_dt	6a2dc8d6-7470-4baa-b4d8-cd0e20761c7d	CWX115	2025-07-03 08:23:09.037395-07	2025-07-03T08:23:09.037395-07:00	daylily	2025-07-03 08:23:09.041179-07	UPDATE	\N	\N	\N	f	f
2c126de6-15a2-4b1a-a7de-971bcaace7e0	generic_instance	modified_dt	6a2dc8d6-7470-4baa-b4d8-cd0e20761c7d	CWX115	2025-07-03 08:23:09.037395-07	2025-07-03T08:23:09.041179-07:00	daylily	2025-07-03 08:23:09.041179-07	UPDATE	\N	\N	\N	f	f
10632dc8-b533-4aa5-b4f5-f2e22b17b1c9	generic_instance_lineage	\N	bd4f4fe4-b242-41f2-b297-7c4e25c79a43	GL361	\N	\N	daylily	2025-07-03 08:23:09.042644-07	INSERT	\N	\N	\N	f	f
71c81f61-f254-4c35-af6d-998cf6999e2c	generic_instance	\N	7f1ef106-8285-4641-bc6c-b2cb1e199946	CWX116	\N	\N	daylily	2025-07-03 08:23:09.042644-07	INSERT	\N	\N	\N	f	f
736cf548-f0f0-4e43-a2fe-a0a603340ab1	generic_instance	json_addl	7f1ef106-8285-4641-bc6c-b2cb1e199946	CWX116	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:09.04668-07	UPDATE	\N	\N	\N	f	f
46a12d22-256a-4066-a61f-614edca87ce4	generic_instance	created_dt	7f1ef106-8285-4641-bc6c-b2cb1e199946	CWX116	2025-07-03 08:23:09.042644-07	2025-07-03T08:23:09.042644-07:00	daylily	2025-07-03 08:23:09.04668-07	UPDATE	\N	\N	\N	f	f
317b3833-2a57-486f-9e8e-b8df9e025e4a	generic_instance	modified_dt	7f1ef106-8285-4641-bc6c-b2cb1e199946	CWX116	2025-07-03 08:23:09.042644-07	2025-07-03T08:23:09.04668-07:00	daylily	2025-07-03 08:23:09.04668-07	UPDATE	\N	\N	\N	f	f
e6d37dba-b391-4494-acc0-167bafd91828	generic_instance_lineage	\N	5967cb29-d6df-4e6e-a318-d6279ff16249	GL362	\N	\N	daylily	2025-07-03 08:23:09.048349-07	INSERT	\N	\N	\N	f	f
7c51d1c3-ce59-438c-bd2c-d2013a41bb4a	generic_instance	\N	00b1f951-e684-487b-91fc-803626c26deb	CWX117	\N	\N	daylily	2025-07-03 08:23:09.048349-07	INSERT	\N	\N	\N	f	f
82265405-2f15-4ac2-93a3-1e231ffcc897	generic_instance	json_addl	00b1f951-e684-487b-91fc-803626c26deb	CWX117	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:09.052106-07	UPDATE	\N	\N	\N	f	f
447c13dd-ac5a-4971-9035-eff4f4313f94	generic_instance	created_dt	00b1f951-e684-487b-91fc-803626c26deb	CWX117	2025-07-03 08:23:09.048349-07	2025-07-03T08:23:09.048349-07:00	daylily	2025-07-03 08:23:09.052106-07	UPDATE	\N	\N	\N	f	f
4f3d89e0-f91c-4bb4-ad87-1fdac94909d2	generic_instance	modified_dt	00b1f951-e684-487b-91fc-803626c26deb	CWX117	2025-07-03 08:23:09.048349-07	2025-07-03T08:23:09.052106-07:00	daylily	2025-07-03 08:23:09.052106-07	UPDATE	\N	\N	\N	f	f
736c05db-ba1f-4b50-b7ea-9f18ffb87b7f	generic_instance_lineage	\N	377a531c-2df4-483d-bff5-e5251a9caa86	GL363	\N	\N	daylily	2025-07-03 08:23:09.053251-07	INSERT	\N	\N	\N	f	f
901d8e46-8971-47f0-ae04-24c4020273ba	generic_instance	\N	a7ce7f5a-f641-42ae-bb64-e32665f3383f	CWX118	\N	\N	daylily	2025-07-03 08:23:09.053251-07	INSERT	\N	\N	\N	f	f
fce3bb67-78f8-4358-b4aa-66ca7b32de40	generic_instance	json_addl	a7ce7f5a-f641-42ae-bb64-e32665f3383f	CWX118	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:09.05703-07	UPDATE	\N	\N	\N	f	f
76972d2f-de2e-4b0f-bc85-609633b34821	generic_instance	created_dt	a7ce7f5a-f641-42ae-bb64-e32665f3383f	CWX118	2025-07-03 08:23:09.053251-07	2025-07-03T08:23:09.053251-07:00	daylily	2025-07-03 08:23:09.05703-07	UPDATE	\N	\N	\N	f	f
e0aef8d7-2d78-4852-bc0c-ec8423857eaf	generic_instance	modified_dt	a7ce7f5a-f641-42ae-bb64-e32665f3383f	CWX118	2025-07-03 08:23:09.053251-07	2025-07-03T08:23:09.05703-07:00	daylily	2025-07-03 08:23:09.05703-07	UPDATE	\N	\N	\N	f	f
8955ee39-9945-4a01-8d85-7d4d274e299d	generic_instance_lineage	\N	b092a54b-5cbf-4063-a27e-9b90a7fdb23d	GL364	\N	\N	daylily	2025-07-03 08:23:09.058174-07	INSERT	\N	\N	\N	f	f
37004ef9-8263-445d-87c4-8dd67442ac1c	generic_instance	\N	04078517-815f-4032-a223-10dc61971326	CWX119	\N	\N	daylily	2025-07-03 08:23:09.058174-07	INSERT	\N	\N	\N	f	f
02cc628a-5efd-4f3e-94c7-4591c8f1e37d	generic_instance	json_addl	04078517-815f-4032-a223-10dc61971326	CWX119	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:09.062041-07	UPDATE	\N	\N	\N	f	f
fd5c23da-36e6-4e82-b96e-84036a4881e2	generic_instance	created_dt	04078517-815f-4032-a223-10dc61971326	CWX119	2025-07-03 08:23:09.058174-07	2025-07-03T08:23:09.058174-07:00	daylily	2025-07-03 08:23:09.062041-07	UPDATE	\N	\N	\N	f	f
fee13281-84e0-432a-9473-ed170105797d	generic_instance	modified_dt	04078517-815f-4032-a223-10dc61971326	CWX119	2025-07-03 08:23:09.058174-07	2025-07-03T08:23:09.062041-07:00	daylily	2025-07-03 08:23:09.062041-07	UPDATE	\N	\N	\N	f	f
339a65e0-7813-40ff-86db-d5ce858aad91	generic_instance_lineage	\N	90baf748-56e6-4cb8-8fe4-05283ab8b489	GL365	\N	\N	daylily	2025-07-03 08:23:09.063341-07	INSERT	\N	\N	\N	f	f
28fc7245-bcc1-46ad-b752-b292778c9eb5	generic_instance	\N	fb489163-90aa-4b36-83d6-71b946fb43f5	CWX120	\N	\N	daylily	2025-07-03 08:23:09.063341-07	INSERT	\N	\N	\N	f	f
6fc3731c-a7fa-4d32-b262-8c517e80cdd3	generic_instance	json_addl	fb489163-90aa-4b36-83d6-71b946fb43f5	CWX120	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:09.067566-07	UPDATE	\N	\N	\N	f	f
9bc4a509-6e47-404d-a081-2f09235180ba	generic_instance	created_dt	fb489163-90aa-4b36-83d6-71b946fb43f5	CWX120	2025-07-03 08:23:09.063341-07	2025-07-03T08:23:09.063341-07:00	daylily	2025-07-03 08:23:09.067566-07	UPDATE	\N	\N	\N	f	f
f39f6e4f-878a-45a3-9ec4-581dbdb46e23	generic_instance	modified_dt	fb489163-90aa-4b36-83d6-71b946fb43f5	CWX120	2025-07-03 08:23:09.063341-07	2025-07-03T08:23:09.067566-07:00	daylily	2025-07-03 08:23:09.067566-07	UPDATE	\N	\N	\N	f	f
6079265a-04b6-4e45-b9ce-eba22eaf4957	generic_instance_lineage	\N	5ab67372-ca46-47ce-a91f-b01a6564a685	GL366	\N	\N	daylily	2025-07-03 08:23:09.068635-07	INSERT	\N	\N	\N	f	f
edd54268-6822-4fc6-b6cd-d68f95c0e0fa	generic_instance	\N	0e06a2a9-31a7-4ab9-9074-9cc9ce191775	CWX121	\N	\N	daylily	2025-07-03 08:23:09.068635-07	INSERT	\N	\N	\N	f	f
3151d309-b6c1-44d3-84c2-ee770739e675	generic_instance	json_addl	0e06a2a9-31a7-4ab9-9074-9cc9ce191775	CWX121	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:09.072453-07	UPDATE	\N	\N	\N	f	f
01f81383-1c5e-4272-8473-ee9405102b39	generic_instance	created_dt	0e06a2a9-31a7-4ab9-9074-9cc9ce191775	CWX121	2025-07-03 08:23:09.068635-07	2025-07-03T08:23:09.068635-07:00	daylily	2025-07-03 08:23:09.072453-07	UPDATE	\N	\N	\N	f	f
eb0d67fb-82c3-42f2-a050-1a19a9dd640c	generic_instance	modified_dt	0e06a2a9-31a7-4ab9-9074-9cc9ce191775	CWX121	2025-07-03 08:23:09.068635-07	2025-07-03T08:23:09.072453-07:00	daylily	2025-07-03 08:23:09.072453-07	UPDATE	\N	\N	\N	f	f
827b126c-9ad9-47e8-b1cd-9d339449cb06	generic_instance_lineage	\N	6c8544e3-8e57-44c8-9d6d-649071cc1eae	GL367	\N	\N	daylily	2025-07-03 08:23:09.073646-07	INSERT	\N	\N	\N	f	f
e4da43ca-5485-4911-9614-8aac53513f4f	generic_instance	\N	f893cb99-8261-4d16-82b7-195a2985abb8	CWX122	\N	\N	daylily	2025-07-03 08:23:09.073646-07	INSERT	\N	\N	\N	f	f
a2df25aa-e6f6-4eff-a709-a259410def7c	generic_instance	json_addl	f893cb99-8261-4d16-82b7-195a2985abb8	CWX122	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:09.07746-07	UPDATE	\N	\N	\N	f	f
245736a5-af91-43cd-83a5-fa8d1f94a100	generic_instance	created_dt	f893cb99-8261-4d16-82b7-195a2985abb8	CWX122	2025-07-03 08:23:09.073646-07	2025-07-03T08:23:09.073646-07:00	daylily	2025-07-03 08:23:09.07746-07	UPDATE	\N	\N	\N	f	f
a8a436d6-eb0e-4371-9178-f673ea3fa3c5	generic_instance	modified_dt	f893cb99-8261-4d16-82b7-195a2985abb8	CWX122	2025-07-03 08:23:09.073646-07	2025-07-03T08:23:09.07746-07:00	daylily	2025-07-03 08:23:09.07746-07	UPDATE	\N	\N	\N	f	f
9e842c62-741e-4200-a650-fdf686585af5	generic_instance_lineage	\N	47d078b3-fbfd-458f-b210-4264f46ba19e	GL368	\N	\N	daylily	2025-07-03 08:23:09.078849-07	INSERT	\N	\N	\N	f	f
f4325604-941d-4f1e-9705-3774f1fa45ed	generic_instance	\N	842e7d48-9605-4321-956e-17b670298506	CWX123	\N	\N	daylily	2025-07-03 08:23:09.078849-07	INSERT	\N	\N	\N	f	f
0272c034-c864-4dc9-903e-1c51ae841698	generic_instance	json_addl	842e7d48-9605-4321-956e-17b670298506	CWX123	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:09.08265-07	UPDATE	\N	\N	\N	f	f
6b73563c-71d9-4769-b247-3d2205520d8e	generic_instance	created_dt	842e7d48-9605-4321-956e-17b670298506	CWX123	2025-07-03 08:23:09.078849-07	2025-07-03T08:23:09.078849-07:00	daylily	2025-07-03 08:23:09.08265-07	UPDATE	\N	\N	\N	f	f
a001f0b8-eeab-44a4-9af7-48dcfb428bf2	generic_instance	modified_dt	842e7d48-9605-4321-956e-17b670298506	CWX123	2025-07-03 08:23:09.078849-07	2025-07-03T08:23:09.08265-07:00	daylily	2025-07-03 08:23:09.08265-07	UPDATE	\N	\N	\N	f	f
b10a195d-30b3-41fe-97c9-8b02ea39da89	generic_instance_lineage	\N	49b6322f-e51e-4130-bb66-bdde2a7c40d9	GL369	\N	\N	daylily	2025-07-03 08:23:09.08386-07	INSERT	\N	\N	\N	f	f
67f5b2ac-b020-4692-936e-725eb9c8c2cd	generic_instance	\N	b1519897-fb8e-4607-898f-3f0575c8a04a	CWX124	\N	\N	daylily	2025-07-03 08:23:09.08386-07	INSERT	\N	\N	\N	f	f
381159a1-33c0-4d94-965f-3d41ae3c3d3e	generic_instance	json_addl	b1519897-fb8e-4607-898f-3f0575c8a04a	CWX124	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:09.087607-07	UPDATE	\N	\N	\N	f	f
c4c26b97-3a01-4626-9e1f-c8b9e7384162	generic_instance	created_dt	b1519897-fb8e-4607-898f-3f0575c8a04a	CWX124	2025-07-03 08:23:09.08386-07	2025-07-03T08:23:09.08386-07:00	daylily	2025-07-03 08:23:09.087607-07	UPDATE	\N	\N	\N	f	f
32b6e254-3a8d-438b-a7b4-f5a0407f3f7c	generic_instance	modified_dt	b1519897-fb8e-4607-898f-3f0575c8a04a	CWX124	2025-07-03 08:23:09.08386-07	2025-07-03T08:23:09.087607-07:00	daylily	2025-07-03 08:23:09.087607-07	UPDATE	\N	\N	\N	f	f
c22ae334-c505-4452-bc74-81f8aa9397a9	generic_instance_lineage	\N	4137bc31-053d-4639-97ea-f15d689dc56b	GL370	\N	\N	daylily	2025-07-03 08:23:09.088818-07	INSERT	\N	\N	\N	f	f
6f7ed26a-9af1-4c20-be97-00e32798adf2	generic_instance	\N	550d0d97-6fda-47b0-8e07-d820e912a145	CWX125	\N	\N	daylily	2025-07-03 08:23:09.088818-07	INSERT	\N	\N	\N	f	f
8e9a0eb4-c906-4955-b24d-3d43e924c4a6	generic_instance	json_addl	550d0d97-6fda-47b0-8e07-d820e912a145	CWX125	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:09.092469-07	UPDATE	\N	\N	\N	f	f
e48c5fdd-cb12-492f-bb52-d351900cc437	generic_instance	created_dt	550d0d97-6fda-47b0-8e07-d820e912a145	CWX125	2025-07-03 08:23:09.088818-07	2025-07-03T08:23:09.088818-07:00	daylily	2025-07-03 08:23:09.092469-07	UPDATE	\N	\N	\N	f	f
923f7b32-ab4f-4618-9661-67be77216a61	generic_instance	modified_dt	550d0d97-6fda-47b0-8e07-d820e912a145	CWX125	2025-07-03 08:23:09.088818-07	2025-07-03T08:23:09.092469-07:00	daylily	2025-07-03 08:23:09.092469-07	UPDATE	\N	\N	\N	f	f
87a0e41b-0150-450f-96c1-c023a97c1e9f	generic_instance_lineage	\N	71348741-eb8b-4c9c-bfcf-4af826116a13	GL371	\N	\N	daylily	2025-07-03 08:23:09.093617-07	INSERT	\N	\N	\N	f	f
113ee299-f6ed-47de-972e-4375e10d68df	generic_instance	\N	bd871a52-972e-4179-a5f8-2aacf0466302	CWX126	\N	\N	daylily	2025-07-03 08:23:09.093617-07	INSERT	\N	\N	\N	f	f
c4813fe7-3e16-4d63-bf72-43d399b94cf8	generic_instance	json_addl	bd871a52-972e-4179-a5f8-2aacf0466302	CWX126	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:09.09752-07	UPDATE	\N	\N	\N	f	f
a5874f52-b604-497d-bf0e-c96fe2e760c4	generic_instance	created_dt	bd871a52-972e-4179-a5f8-2aacf0466302	CWX126	2025-07-03 08:23:09.093617-07	2025-07-03T08:23:09.093617-07:00	daylily	2025-07-03 08:23:09.09752-07	UPDATE	\N	\N	\N	f	f
6cec6e1d-bf88-42ce-9816-96c94371774c	generic_instance	modified_dt	bd871a52-972e-4179-a5f8-2aacf0466302	CWX126	2025-07-03 08:23:09.093617-07	2025-07-03T08:23:09.09752-07:00	daylily	2025-07-03 08:23:09.09752-07	UPDATE	\N	\N	\N	f	f
a132f39a-2416-4840-831b-43ad03ecc266	generic_instance_lineage	\N	2c63e770-24e0-4053-8e3f-8a97aa1d7a79	GL372	\N	\N	daylily	2025-07-03 08:23:09.099273-07	INSERT	\N	\N	\N	f	f
45cba349-ce4e-4828-b0fa-608dd9499cbe	generic_instance	\N	4ba333d5-a753-4026-9358-75385fec301a	CWX127	\N	\N	daylily	2025-07-03 08:23:09.099273-07	INSERT	\N	\N	\N	f	f
9caff9aa-821c-4c83-9f27-a3e1a86bbcaa	generic_instance	json_addl	4ba333d5-a753-4026-9358-75385fec301a	CWX127	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:09.1057-07	UPDATE	\N	\N	\N	f	f
7f092ed3-e971-4785-a487-b87ab6659fff	generic_instance	created_dt	4ba333d5-a753-4026-9358-75385fec301a	CWX127	2025-07-03 08:23:09.099273-07	2025-07-03T08:23:09.099273-07:00	daylily	2025-07-03 08:23:09.1057-07	UPDATE	\N	\N	\N	f	f
5885a01b-35f7-4cff-8372-fe18d694b670	generic_instance	modified_dt	4ba333d5-a753-4026-9358-75385fec301a	CWX127	2025-07-03 08:23:09.099273-07	2025-07-03T08:23:09.1057-07:00	daylily	2025-07-03 08:23:09.1057-07	UPDATE	\N	\N	\N	f	f
c6296cd8-5b62-422b-96ef-5686542fb13c	generic_instance_lineage	\N	656a6e1c-4bf3-40d4-9a3d-f03ef843d3b0	GL373	\N	\N	daylily	2025-07-03 08:23:09.10718-07	INSERT	\N	\N	\N	f	f
a8680bf0-470c-45f1-9f64-1e7314d5acd0	generic_instance	\N	60b56382-b9de-4a9d-892b-a6e1bcae9e89	CWX128	\N	\N	daylily	2025-07-03 08:23:09.10718-07	INSERT	\N	\N	\N	f	f
4535e08b-dad6-4b69-8b3f-096c63419cba	generic_instance	json_addl	60b56382-b9de-4a9d-892b-a6e1bcae9e89	CWX128	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:09.110903-07	UPDATE	\N	\N	\N	f	f
5757c34c-fee3-476f-836f-88a5dd0304b1	generic_instance	created_dt	60b56382-b9de-4a9d-892b-a6e1bcae9e89	CWX128	2025-07-03 08:23:09.10718-07	2025-07-03T08:23:09.10718-07:00	daylily	2025-07-03 08:23:09.110903-07	UPDATE	\N	\N	\N	f	f
6f004b49-7ea2-4e54-a7a6-be36f16f5335	generic_instance	modified_dt	60b56382-b9de-4a9d-892b-a6e1bcae9e89	CWX128	2025-07-03 08:23:09.10718-07	2025-07-03T08:23:09.110903-07:00	daylily	2025-07-03 08:23:09.110903-07	UPDATE	\N	\N	\N	f	f
8fa64f7d-cfed-4217-8f9c-ba5128f7e49f	generic_instance_lineage	\N	cd1cf9f9-3297-4276-895f-f1a678b9f1dc	GL374	\N	\N	daylily	2025-07-03 08:23:09.112412-07	INSERT	\N	\N	\N	f	f
eaba59fc-99d5-4e34-87b1-ac05aae1e4c0	generic_instance	\N	a9314a45-c45f-4522-a0e9-9a99459a6924	CWX129	\N	\N	daylily	2025-07-03 08:23:09.112412-07	INSERT	\N	\N	\N	f	f
324954b8-1597-449b-a58e-2f87bfd29cfb	generic_instance	json_addl	a9314a45-c45f-4522-a0e9-9a99459a6924	CWX129	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:09.11608-07	UPDATE	\N	\N	\N	f	f
d5c40ba1-0b72-4244-95b3-d2593c76dec4	generic_instance	created_dt	a9314a45-c45f-4522-a0e9-9a99459a6924	CWX129	2025-07-03 08:23:09.112412-07	2025-07-03T08:23:09.112412-07:00	daylily	2025-07-03 08:23:09.11608-07	UPDATE	\N	\N	\N	f	f
114de951-2ffb-42b0-a97f-b5d8b6233c6a	generic_instance	modified_dt	a9314a45-c45f-4522-a0e9-9a99459a6924	CWX129	2025-07-03 08:23:09.112412-07	2025-07-03T08:23:09.11608-07:00	daylily	2025-07-03 08:23:09.11608-07	UPDATE	\N	\N	\N	f	f
d762b4c5-4995-4afa-8617-d648b89c8c70	generic_instance_lineage	\N	f2c44965-a71e-44b9-b5ae-1aabb95e4d1e	GL375	\N	\N	daylily	2025-07-03 08:23:09.117223-07	INSERT	\N	\N	\N	f	f
aa2bf7d0-3765-4652-ad2b-d71914990135	generic_instance	\N	43f0aac8-c4da-48d5-ade3-02f63fdd00a5	CWX130	\N	\N	daylily	2025-07-03 08:23:09.117223-07	INSERT	\N	\N	\N	f	f
33fcdf93-9847-4992-8a03-33bb4ee59096	generic_instance	json_addl	43f0aac8-c4da-48d5-ade3-02f63fdd00a5	CWX130	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:09.121175-07	UPDATE	\N	\N	\N	f	f
330dfafd-cee8-40f4-b871-26a518d00259	generic_instance	created_dt	43f0aac8-c4da-48d5-ade3-02f63fdd00a5	CWX130	2025-07-03 08:23:09.117223-07	2025-07-03T08:23:09.117223-07:00	daylily	2025-07-03 08:23:09.121175-07	UPDATE	\N	\N	\N	f	f
b5274f45-d64b-4a35-9343-97e262847058	generic_instance	modified_dt	43f0aac8-c4da-48d5-ade3-02f63fdd00a5	CWX130	2025-07-03 08:23:09.117223-07	2025-07-03T08:23:09.121175-07:00	daylily	2025-07-03 08:23:09.121175-07	UPDATE	\N	\N	\N	f	f
2c9d7dfd-7667-46c2-8ec6-ecdf26e97fdd	generic_instance_lineage	\N	9cc96fc0-394e-4a0b-9dc0-7287b39afbaf	GL376	\N	\N	daylily	2025-07-03 08:23:09.122373-07	INSERT	\N	\N	\N	f	f
61fe6d10-7bae-447b-adae-74c2c487b126	generic_instance	\N	2d325520-3bd4-44b8-93e0-874881087ae6	CWX131	\N	\N	daylily	2025-07-03 08:23:09.122373-07	INSERT	\N	\N	\N	f	f
03ac1d21-1c06-44c5-b6df-f5e11cd9e9a5	generic_instance	json_addl	2d325520-3bd4-44b8-93e0-874881087ae6	CWX131	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:09.126199-07	UPDATE	\N	\N	\N	f	f
c65e65ce-7803-43ea-9819-11ca547e5d6d	generic_instance	created_dt	2d325520-3bd4-44b8-93e0-874881087ae6	CWX131	2025-07-03 08:23:09.122373-07	2025-07-03T08:23:09.122373-07:00	daylily	2025-07-03 08:23:09.126199-07	UPDATE	\N	\N	\N	f	f
6a9df52e-dc80-449c-bb7c-4cb1f2552216	generic_instance	modified_dt	2d325520-3bd4-44b8-93e0-874881087ae6	CWX131	2025-07-03 08:23:09.122373-07	2025-07-03T08:23:09.126199-07:00	daylily	2025-07-03 08:23:09.126199-07	UPDATE	\N	\N	\N	f	f
6a0ba524-cdc8-4cb1-8fe9-3bc1db05396d	generic_instance_lineage	\N	0f9233ce-f7d6-4a54-8a34-70a67560b75f	GL377	\N	\N	daylily	2025-07-03 08:23:09.127525-07	INSERT	\N	\N	\N	f	f
da0a4d4e-594c-4e06-b50f-57e531cf4036	generic_instance	\N	57556f94-c37c-41c8-88db-53e8cee975b6	CWX132	\N	\N	daylily	2025-07-03 08:23:09.127525-07	INSERT	\N	\N	\N	f	f
27429756-42bd-4361-b89b-166b37a4d429	generic_instance	json_addl	57556f94-c37c-41c8-88db-53e8cee975b6	CWX132	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:09.1318-07	UPDATE	\N	\N	\N	f	f
656e237d-cdec-4ff5-9054-f709cb01c8e9	generic_instance	created_dt	57556f94-c37c-41c8-88db-53e8cee975b6	CWX132	2025-07-03 08:23:09.127525-07	2025-07-03T08:23:09.127525-07:00	daylily	2025-07-03 08:23:09.1318-07	UPDATE	\N	\N	\N	f	f
a159f351-7909-4de3-aac9-7f1ab758f15f	generic_instance	modified_dt	57556f94-c37c-41c8-88db-53e8cee975b6	CWX132	2025-07-03 08:23:09.127525-07	2025-07-03T08:23:09.1318-07:00	daylily	2025-07-03 08:23:09.1318-07	UPDATE	\N	\N	\N	f	f
fe05b4f1-7092-4492-a2f8-ce47359779b5	generic_instance_lineage	\N	c89fc23d-30a7-4027-8168-f25bc8e1de41	GL378	\N	\N	daylily	2025-07-03 08:23:09.132921-07	INSERT	\N	\N	\N	f	f
03e96784-ec56-4662-ae6d-a986b888fe9a	generic_instance	\N	820ef252-cff7-4a46-a803-2dc8eee77103	CWX133	\N	\N	daylily	2025-07-03 08:23:09.132921-07	INSERT	\N	\N	\N	f	f
bc78195e-b00b-486e-ab8b-414ebf313696	generic_instance	json_addl	820ef252-cff7-4a46-a803-2dc8eee77103	CWX133	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:09.136654-07	UPDATE	\N	\N	\N	f	f
513d4329-c135-420c-a13c-b861dd116c41	generic_instance	created_dt	820ef252-cff7-4a46-a803-2dc8eee77103	CWX133	2025-07-03 08:23:09.132921-07	2025-07-03T08:23:09.132921-07:00	daylily	2025-07-03 08:23:09.136654-07	UPDATE	\N	\N	\N	f	f
c4223908-0ca5-4df4-a21f-9ca5c8a3ff7d	generic_instance	modified_dt	820ef252-cff7-4a46-a803-2dc8eee77103	CWX133	2025-07-03 08:23:09.132921-07	2025-07-03T08:23:09.136654-07:00	daylily	2025-07-03 08:23:09.136654-07	UPDATE	\N	\N	\N	f	f
caaa50d5-1bbe-4fad-94c0-6b2278d0831d	generic_instance_lineage	\N	965312e8-cc48-4124-a96e-76a6a9e775a3	GL379	\N	\N	daylily	2025-07-03 08:23:09.137843-07	INSERT	\N	\N	\N	f	f
108d0d4c-f0d8-487d-abea-ac3e8b669479	generic_instance	\N	3b447fb4-8438-4485-bac4-1553e554d7d2	CWX134	\N	\N	daylily	2025-07-03 08:23:09.137843-07	INSERT	\N	\N	\N	f	f
817e3f93-d272-4d43-9ae6-1eb88d8a942e	generic_instance	json_addl	3b447fb4-8438-4485-bac4-1553e554d7d2	CWX134	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:09.141751-07	UPDATE	\N	\N	\N	f	f
c3974e20-b2ce-4171-b3d3-76a2036370d6	generic_instance	created_dt	3b447fb4-8438-4485-bac4-1553e554d7d2	CWX134	2025-07-03 08:23:09.137843-07	2025-07-03T08:23:09.137843-07:00	daylily	2025-07-03 08:23:09.141751-07	UPDATE	\N	\N	\N	f	f
888fb70d-4f1a-467a-bf6c-a6f4311457c4	generic_instance	modified_dt	3b447fb4-8438-4485-bac4-1553e554d7d2	CWX134	2025-07-03 08:23:09.137843-07	2025-07-03T08:23:09.141751-07:00	daylily	2025-07-03 08:23:09.141751-07	UPDATE	\N	\N	\N	f	f
bca4012a-e8e0-432d-8f34-a6e3403a38c7	generic_instance_lineage	\N	e0868793-d127-4bfb-a19d-e2fb2741fa3e	GL380	\N	\N	daylily	2025-07-03 08:23:09.142959-07	INSERT	\N	\N	\N	f	f
9581588d-01f6-4464-9b22-2e7492fb80d2	generic_instance	\N	c6928136-7211-428a-aa93-424927c53157	CWX135	\N	\N	daylily	2025-07-03 08:23:09.142959-07	INSERT	\N	\N	\N	f	f
8692d52a-1334-4591-a0d8-6a3d48d45a2b	generic_instance	json_addl	c6928136-7211-428a-aa93-424927c53157	CWX135	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:09.14677-07	UPDATE	\N	\N	\N	f	f
752a04d7-6f04-4e2f-98f9-23ceb249cc17	generic_instance	created_dt	c6928136-7211-428a-aa93-424927c53157	CWX135	2025-07-03 08:23:09.142959-07	2025-07-03T08:23:09.142959-07:00	daylily	2025-07-03 08:23:09.14677-07	UPDATE	\N	\N	\N	f	f
1a0189bd-9fa9-4f74-b031-060e27bace44	generic_instance	modified_dt	c6928136-7211-428a-aa93-424927c53157	CWX135	2025-07-03 08:23:09.142959-07	2025-07-03T08:23:09.14677-07:00	daylily	2025-07-03 08:23:09.14677-07	UPDATE	\N	\N	\N	f	f
fdfca6ea-638e-43d0-8f71-4613fdca2856	generic_instance_lineage	\N	87df48f5-fead-4a10-9166-10a081b27e42	GL381	\N	\N	daylily	2025-07-03 08:23:09.14792-07	INSERT	\N	\N	\N	f	f
079d526a-c030-477b-893b-861671c7c7cf	generic_instance	\N	5633367b-29d0-4a2c-bf49-950a922e81bb	CWX136	\N	\N	daylily	2025-07-03 08:23:09.14792-07	INSERT	\N	\N	\N	f	f
86d6080d-6106-45c4-abc3-81eb5a20becc	generic_instance	json_addl	5633367b-29d0-4a2c-bf49-950a922e81bb	CWX136	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:09.151572-07	UPDATE	\N	\N	\N	f	f
9d394e4e-98b7-4dca-9765-0dfc352440f7	generic_instance	created_dt	5633367b-29d0-4a2c-bf49-950a922e81bb	CWX136	2025-07-03 08:23:09.14792-07	2025-07-03T08:23:09.14792-07:00	daylily	2025-07-03 08:23:09.151572-07	UPDATE	\N	\N	\N	f	f
2bb3ea2a-f3fe-4e71-9223-86bfc11876f4	generic_instance	modified_dt	5633367b-29d0-4a2c-bf49-950a922e81bb	CWX136	2025-07-03 08:23:09.14792-07	2025-07-03T08:23:09.151572-07:00	daylily	2025-07-03 08:23:09.151572-07	UPDATE	\N	\N	\N	f	f
0abeddf6-faca-45c8-bd6d-a37924016f55	generic_instance_lineage	\N	ae944efe-2835-418f-93fc-fec9a20e20f7	GL382	\N	\N	daylily	2025-07-03 08:23:09.152702-07	INSERT	\N	\N	\N	f	f
321ee276-0754-4bd0-b897-9a3db22fa05a	generic_instance	\N	4b42200e-1ca5-4eb8-8947-8cf89ae11946	CWX137	\N	\N	daylily	2025-07-03 08:23:09.152702-07	INSERT	\N	\N	\N	f	f
ba8d5713-7834-4764-a3c6-b13a9311c625	generic_instance	json_addl	4b42200e-1ca5-4eb8-8947-8cf89ae11946	CWX137	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:09.156632-07	UPDATE	\N	\N	\N	f	f
c2f5327d-d413-449c-a79b-be9139df27e9	generic_instance	created_dt	4b42200e-1ca5-4eb8-8947-8cf89ae11946	CWX137	2025-07-03 08:23:09.152702-07	2025-07-03T08:23:09.152702-07:00	daylily	2025-07-03 08:23:09.156632-07	UPDATE	\N	\N	\N	f	f
83d49be1-4f46-4734-8998-d62a1b7ba117	generic_instance	modified_dt	4b42200e-1ca5-4eb8-8947-8cf89ae11946	CWX137	2025-07-03 08:23:09.152702-07	2025-07-03T08:23:09.156632-07:00	daylily	2025-07-03 08:23:09.156632-07	UPDATE	\N	\N	\N	f	f
6aa40f6b-19f5-4310-9a0d-c417e7e9ba77	generic_instance_lineage	\N	39fc1d4e-f58d-4fe7-b0f7-8fdded5395b0	GL383	\N	\N	daylily	2025-07-03 08:23:09.15779-07	INSERT	\N	\N	\N	f	f
87473c68-a81e-47ad-93ee-24685881b68d	generic_instance	\N	fc4f5d57-c0c6-4d38-bfd6-d50850a8f148	CWX138	\N	\N	daylily	2025-07-03 08:23:09.15779-07	INSERT	\N	\N	\N	f	f
2cbdb50b-5171-446f-91c6-bd8a18cb6787	generic_instance	json_addl	fc4f5d57-c0c6-4d38-bfd6-d50850a8f148	CWX138	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:09.161736-07	UPDATE	\N	\N	\N	f	f
d1904ef7-644f-4593-86c3-b9407b9789c5	generic_instance	created_dt	fc4f5d57-c0c6-4d38-bfd6-d50850a8f148	CWX138	2025-07-03 08:23:09.15779-07	2025-07-03T08:23:09.15779-07:00	daylily	2025-07-03 08:23:09.161736-07	UPDATE	\N	\N	\N	f	f
cc73d1f4-c0da-4b38-859d-809dfcb88975	generic_instance	modified_dt	fc4f5d57-c0c6-4d38-bfd6-d50850a8f148	CWX138	2025-07-03 08:23:09.15779-07	2025-07-03T08:23:09.161736-07:00	daylily	2025-07-03 08:23:09.161736-07	UPDATE	\N	\N	\N	f	f
02337cc8-29bd-4d4b-b1fe-4e5aa1cb1acb	generic_instance_lineage	\N	9a66eec3-08fc-44ae-83ae-fc680437decf	GL384	\N	\N	daylily	2025-07-03 08:23:09.163105-07	INSERT	\N	\N	\N	f	f
973a3525-0ebd-45b8-8067-0d5c1d3b8408	generic_instance	\N	cff309ac-096c-4bc0-93dc-5adc7c8fe0fb	CWX139	\N	\N	daylily	2025-07-03 08:23:09.163105-07	INSERT	\N	\N	\N	f	f
63e92445-e65e-4b1e-b897-48d5e34f4695	generic_instance	json_addl	cff309ac-096c-4bc0-93dc-5adc7c8fe0fb	CWX139	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:09.167363-07	UPDATE	\N	\N	\N	f	f
ea41552d-17b7-4a9f-9da2-43b569011d87	generic_instance	created_dt	cff309ac-096c-4bc0-93dc-5adc7c8fe0fb	CWX139	2025-07-03 08:23:09.163105-07	2025-07-03T08:23:09.163105-07:00	daylily	2025-07-03 08:23:09.167363-07	UPDATE	\N	\N	\N	f	f
0d636151-f0d2-40df-badf-32f95c909742	generic_instance	modified_dt	cff309ac-096c-4bc0-93dc-5adc7c8fe0fb	CWX139	2025-07-03 08:23:09.163105-07	2025-07-03T08:23:09.167363-07:00	daylily	2025-07-03 08:23:09.167363-07	UPDATE	\N	\N	\N	f	f
48dcfa82-cdee-44be-8589-42a6406e332d	generic_instance_lineage	\N	9cc4f63d-0981-4b9c-b12c-d0d0c7f54e33	GL385	\N	\N	daylily	2025-07-03 08:23:09.168511-07	INSERT	\N	\N	\N	f	f
60d86aa3-f05b-4535-ba8d-93be57c8e742	generic_instance	\N	ca5acce6-59f8-4465-83e5-3bb593d9bec3	CWX140	\N	\N	daylily	2025-07-03 08:23:09.168511-07	INSERT	\N	\N	\N	f	f
2ab0acbe-a1eb-423e-b177-eadd5960fe15	generic_instance	json_addl	ca5acce6-59f8-4465-83e5-3bb593d9bec3	CWX140	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:09.172406-07	UPDATE	\N	\N	\N	f	f
031c4e33-a458-4dea-aa4d-6b1110706d3b	generic_instance	created_dt	ca5acce6-59f8-4465-83e5-3bb593d9bec3	CWX140	2025-07-03 08:23:09.168511-07	2025-07-03T08:23:09.168511-07:00	daylily	2025-07-03 08:23:09.172406-07	UPDATE	\N	\N	\N	f	f
dc77c9be-5009-4720-b4e5-eceeaf3c8bda	generic_instance	modified_dt	ca5acce6-59f8-4465-83e5-3bb593d9bec3	CWX140	2025-07-03 08:23:09.168511-07	2025-07-03T08:23:09.172406-07:00	daylily	2025-07-03 08:23:09.172406-07	UPDATE	\N	\N	\N	f	f
953c73d3-94c4-4d3c-bb87-dbec819a0025	generic_instance_lineage	\N	b95e04d7-8737-4824-a2a0-eb7896990a5a	GL386	\N	\N	daylily	2025-07-03 08:23:09.173761-07	INSERT	\N	\N	\N	f	f
39f429fa-a4f9-4c4c-ae29-19b5c0979c57	generic_instance	\N	cb05244f-66db-44ed-9642-77f2976dd0dc	MRX113	\N	\N	daylily	2025-07-03 08:23:09.175657-07	INSERT	\N	\N	\N	f	f
f20d230b-06b6-4a87-9be6-1a5ea43b88e4	generic_instance_lineage	\N	703a5aaf-2a74-4cf9-aaff-eeaf6fee40cd	GL387	\N	\N	daylily	2025-07-03 08:23:09.177524-07	INSERT	\N	\N	\N	f	f
cb226618-a186-4f31-8f7d-f9ad1aaaec75	generic_instance	\N	a3c5b2a8-b081-4788-ac6a-f73a3e942278	MRX114	\N	\N	daylily	2025-07-03 08:23:09.177524-07	INSERT	\N	\N	\N	f	f
ae6a227c-c222-4185-b7b3-4ee99bb92100	generic_instance_lineage	\N	b85974ac-48d4-4f95-8d62-4ed766ae5ce7	GL388	\N	\N	daylily	2025-07-03 08:23:09.181932-07	INSERT	\N	\N	\N	f	f
08004e45-8a78-40b4-b3d3-cdaa6fa47a23	generic_instance	\N	ed1337dc-5377-45a7-a724-e41e29c46a91	MRX115	\N	\N	daylily	2025-07-03 08:23:09.181932-07	INSERT	\N	\N	\N	f	f
1491ba9c-ae8b-4421-a737-c8f54ede21ac	generic_instance_lineage	\N	5d0af38f-b391-4396-a686-0025a58a3f8e	GL389	\N	\N	daylily	2025-07-03 08:23:09.186355-07	INSERT	\N	\N	\N	f	f
810b051d-8d49-495d-bcb0-d1ae408494b6	generic_instance	\N	dd1014f2-4cd8-4e15-b4ce-5095401e8759	MRX116	\N	\N	daylily	2025-07-03 08:23:09.186355-07	INSERT	\N	\N	\N	f	f
5cc1f7ab-293e-43a7-9862-26c50279ff2d	generic_instance_lineage	\N	5861cc21-e93e-46cf-bcdb-afc4afc7225b	GL390	\N	\N	daylily	2025-07-03 08:23:09.190707-07	INSERT	\N	\N	\N	f	f
53630830-b3e8-42ea-bf26-e90839f579ce	generic_instance	\N	170ab794-eb07-48e0-b1c0-f87d77d20590	MRX117	\N	\N	daylily	2025-07-03 08:23:09.190707-07	INSERT	\N	\N	\N	f	f
3d38fb4d-a8fc-45cd-a5a3-d5ad4c1d61b0	generic_instance_lineage	\N	35e7d335-59f2-47ef-a757-8bd04a00a2f8	GL391	\N	\N	daylily	2025-07-03 08:23:09.195105-07	INSERT	\N	\N	\N	f	f
e37589cf-db53-4cad-a1e6-cc99679c92fe	generic_instance	\N	7316e0cf-c132-4afb-aef5-8bcf84ab93df	MRX118	\N	\N	daylily	2025-07-03 08:23:09.195105-07	INSERT	\N	\N	\N	f	f
eb5ae79f-b857-451d-a0b2-27be2d7a1357	generic_instance_lineage	\N	8576a290-e0f4-4301-a63b-df8c53ff07bd	GL392	\N	\N	daylily	2025-07-03 08:23:09.199605-07	INSERT	\N	\N	\N	f	f
7f69b386-a6fb-49a2-a2ca-855f67c75e5b	generic_instance	\N	8a6828b1-67d5-4a72-a00e-70795c41dca6	MRX119	\N	\N	daylily	2025-07-03 08:23:09.199605-07	INSERT	\N	\N	\N	f	f
ba33080b-0d8a-4f90-b843-2e2457b67c80	generic_instance_lineage	\N	319d79f6-0a91-4668-b6e6-6e1676cf9bf2	GL393	\N	\N	daylily	2025-07-03 08:23:09.20388-07	INSERT	\N	\N	\N	f	f
ea183a85-0294-4bc1-a504-3618b2972e32	generic_instance	\N	12f78a64-40a3-4753-bef0-11ad97568fa6	MRX120	\N	\N	daylily	2025-07-03 08:23:09.20388-07	INSERT	\N	\N	\N	f	f
9ce0692c-43ef-480e-9a69-3834e037e0e3	generic_instance_lineage	\N	60fb78d3-e9c7-4553-af86-bcd43b4c3e39	GL394	\N	\N	daylily	2025-07-03 08:23:09.208299-07	INSERT	\N	\N	\N	f	f
73e9bd02-ab06-4b3a-b320-75e4dbe0103b	generic_instance	\N	239cf4e6-27d8-4b54-ad21-a528a3c52252	MRX121	\N	\N	daylily	2025-07-03 08:23:09.208299-07	INSERT	\N	\N	\N	f	f
5ed0a308-194f-44f8-96ec-b770ae152903	generic_instance_lineage	\N	ea0421db-1589-4190-9e5b-4191495ba9c3	GL395	\N	\N	daylily	2025-07-03 08:23:09.212744-07	INSERT	\N	\N	\N	f	f
9032621c-7905-4866-b64e-610fe050a565	generic_instance	\N	512ab731-ec11-4e70-ace2-225ccd1c827c	MRX122	\N	\N	daylily	2025-07-03 08:23:09.212744-07	INSERT	\N	\N	\N	f	f
c8609292-6587-4afa-b3b0-77627d1598c3	generic_instance_lineage	\N	52b216e5-610e-402d-b90d-b587a6dd1e20	GL396	\N	\N	daylily	2025-07-03 08:23:09.217046-07	INSERT	\N	\N	\N	f	f
eb907557-857d-4975-aaa5-87061484db66	generic_instance	\N	60bb2ea0-c3cb-442c-8ced-8495f8693f6f	MRX123	\N	\N	daylily	2025-07-03 08:23:09.217046-07	INSERT	\N	\N	\N	f	f
0de604b6-3664-4c6d-a45b-1c259192aabe	generic_instance_lineage	\N	b6fd6f13-169e-436e-91ef-ba8fc1fd6b3a	GL397	\N	\N	daylily	2025-07-03 08:23:09.221322-07	INSERT	\N	\N	\N	f	f
ade294bd-75de-4e51-8214-65e8a7426cde	generic_instance	\N	cb072692-f302-4e41-9830-dc8d5c7d4289	MRX124	\N	\N	daylily	2025-07-03 08:23:09.221322-07	INSERT	\N	\N	\N	f	f
67c53276-24cd-461a-9d93-e0db8c947aad	generic_instance_lineage	\N	b9dc311a-c011-463c-a9f6-e94d8b415f37	GL398	\N	\N	daylily	2025-07-03 08:23:09.225846-07	INSERT	\N	\N	\N	f	f
e1961d06-0d01-46d3-a24f-72a130a35359	generic_instance	\N	db516863-3a58-4e48-bdea-253df93b7eba	MRX125	\N	\N	daylily	2025-07-03 08:23:09.225846-07	INSERT	\N	\N	\N	f	f
e9b4cb0a-a8f9-4a6e-ab9e-bf3850254ff0	generic_instance_lineage	\N	1d89e2d6-5241-46ab-a357-1d0c0e59382b	GL399	\N	\N	daylily	2025-07-03 08:23:09.23021-07	INSERT	\N	\N	\N	f	f
a730bcfc-a0e8-4792-b712-379a4c81380a	generic_instance	\N	a6f37e7d-af69-46c4-a585-db6bfa69c506	MRX126	\N	\N	daylily	2025-07-03 08:23:09.23021-07	INSERT	\N	\N	\N	f	f
a19716c2-0a94-4e86-a5ef-a7ac880b6afa	generic_instance_lineage	\N	420cf78c-cf53-4208-9ca7-b8d1f3e0d774	GL400	\N	\N	daylily	2025-07-03 08:23:09.234677-07	INSERT	\N	\N	\N	f	f
32a69d34-4d94-4720-b9b4-0483292101d9	generic_instance	\N	c75f8621-9c89-4d84-9938-ed98b1430c39	MRX127	\N	\N	daylily	2025-07-03 08:23:09.234677-07	INSERT	\N	\N	\N	f	f
5fb6d133-e8b6-49a7-b283-23776699c102	generic_instance_lineage	\N	fcd6944d-0c86-4a7a-9aa2-db3490ea589b	GL401	\N	\N	daylily	2025-07-03 08:23:09.239284-07	INSERT	\N	\N	\N	f	f
86146e1c-a7ea-40df-8324-afe06e322578	generic_instance	\N	ad6951f6-0d8e-43e5-a965-d1ac3e2d0035	MRX128	\N	\N	daylily	2025-07-03 08:23:09.239284-07	INSERT	\N	\N	\N	f	f
f43aefe8-8bac-4d9e-acec-61e86993b8b1	generic_instance_lineage	\N	7899ebf9-9a37-4935-9f46-c8fdacde5a1f	GL402	\N	\N	daylily	2025-07-03 08:23:09.243663-07	INSERT	\N	\N	\N	f	f
e63d25d5-b9f3-40fc-a60f-e9ea601807cb	generic_instance	\N	558d4ccc-51af-41e5-a221-6a4d6de4e87e	MRX129	\N	\N	daylily	2025-07-03 08:23:09.243663-07	INSERT	\N	\N	\N	f	f
bfb2c307-ac17-43eb-8661-a41ae02c1ef4	generic_instance_lineage	\N	598f3faa-a496-4f66-8e75-e1882d5a2d65	GL403	\N	\N	daylily	2025-07-03 08:23:09.248075-07	INSERT	\N	\N	\N	f	f
da6abbe8-c716-439f-af03-4a37e5187fa4	generic_instance	\N	3e0d0af1-1210-4cab-84af-c3f9e045953f	MRX130	\N	\N	daylily	2025-07-03 08:23:09.248075-07	INSERT	\N	\N	\N	f	f
cdfaf129-5cf9-4040-bede-3b202141c131	generic_instance_lineage	\N	a701bf49-691f-4c59-8b85-2127da8e5c06	GL404	\N	\N	daylily	2025-07-03 08:23:09.252411-07	INSERT	\N	\N	\N	f	f
b4928cb5-e341-48fa-97e3-fc9c7198bb32	generic_instance	\N	eec83a36-273d-4e8d-a349-5c5abc492832	MRX131	\N	\N	daylily	2025-07-03 08:23:09.252411-07	INSERT	\N	\N	\N	f	f
27f31909-7b74-45f9-bd35-ac0406d14195	generic_instance_lineage	\N	3ed88f84-e80e-4785-9d78-b9a314c4097e	GL405	\N	\N	daylily	2025-07-03 08:23:09.256823-07	INSERT	\N	\N	\N	f	f
a052d76f-1531-4d87-9660-087a34b01f40	generic_instance	\N	502f963c-82a7-45c1-be2b-d96b35f3bdf2	MRX132	\N	\N	daylily	2025-07-03 08:23:09.256823-07	INSERT	\N	\N	\N	f	f
9013dd9a-b6d9-4116-94e4-a609188aae22	generic_instance_lineage	\N	bb9ee74f-1635-4228-9ffb-7af827d417ed	GL406	\N	\N	daylily	2025-07-03 08:23:09.261204-07	INSERT	\N	\N	\N	f	f
6c10e658-0155-4a4b-ad22-657a23385988	generic_instance	\N	00b8f24b-52a1-4993-b78a-1f8de01c1df5	MRX133	\N	\N	daylily	2025-07-03 08:23:09.261204-07	INSERT	\N	\N	\N	f	f
2f209f45-a167-4c5b-b518-532eb6accfa6	generic_instance_lineage	\N	077f07fa-ffe9-41ed-81f8-46e6f1d87674	GL407	\N	\N	daylily	2025-07-03 08:23:09.265782-07	INSERT	\N	\N	\N	f	f
f6cea4d9-c56f-4297-801c-a3237f68fa24	generic_instance	\N	5f74720c-3a0b-426f-981d-9c1db1d9f0cd	MRX134	\N	\N	daylily	2025-07-03 08:23:09.265782-07	INSERT	\N	\N	\N	f	f
01c9ec67-5c86-4ef9-9199-d435235b2860	generic_instance_lineage	\N	fb64de69-cc24-44f9-a83b-307030ff461f	GL408	\N	\N	daylily	2025-07-03 08:23:09.269957-07	INSERT	\N	\N	\N	f	f
8d10139b-0430-49da-a94d-578c1bd26aa1	generic_instance	\N	1f7b5d3a-40fc-4878-bcd6-78a9cef9ae5f	MRX135	\N	\N	daylily	2025-07-03 08:23:09.269957-07	INSERT	\N	\N	\N	f	f
dbaa5ec0-9bbb-441f-b040-a45aafd42eae	generic_instance	json_addl	db19c6e5-98cc-4eba-9311-7d9ac31c3cf2	CWX141	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:16.855209-07	UPDATE	\N	\N	\N	f	f
ef076d7e-fc3e-4fe6-90c4-0064ebeb9b18	generic_instance	created_dt	db19c6e5-98cc-4eba-9311-7d9ac31c3cf2	CWX141	2025-07-03 08:23:16.850501-07	2025-07-03T08:23:16.850501-07:00	daylily	2025-07-03 08:23:16.855209-07	UPDATE	\N	\N	\N	f	f
1fe8d77e-8c6f-42c3-9666-74711abdb84a	generic_instance	modified_dt	db19c6e5-98cc-4eba-9311-7d9ac31c3cf2	CWX141	2025-07-03 08:23:16.850501-07	2025-07-03T08:23:16.855209-07:00	daylily	2025-07-03 08:23:16.855209-07	UPDATE	\N	\N	\N	f	f
8abf75ae-9f16-4aa9-9c1a-757b8510ca14	generic_instance_lineage	\N	74601e8a-d8b3-4fd8-ba7b-437d201f9754	GL415	\N	\N	daylily	2025-07-03 08:23:16.858263-07	INSERT	\N	\N	\N	f	f
d894f3cf-473c-4aa7-996c-29fad8d7c0aa	generic_instance	\N	b00c40c2-0afb-46fa-b628-daffa4dcd1c7	CWX142	\N	\N	daylily	2025-07-03 08:23:16.858263-07	INSERT	\N	\N	\N	f	f
886165a4-c58e-4a84-8561-3f3e28d194b6	generic_instance	json_addl	b00c40c2-0afb-46fa-b628-daffa4dcd1c7	CWX142	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:16.863962-07	UPDATE	\N	\N	\N	f	f
90c275ab-4a1a-47a0-8cff-f1f9b0c68970	generic_instance	created_dt	b00c40c2-0afb-46fa-b628-daffa4dcd1c7	CWX142	2025-07-03 08:23:16.858263-07	2025-07-03T08:23:16.858263-07:00	daylily	2025-07-03 08:23:16.863962-07	UPDATE	\N	\N	\N	f	f
3da33f45-3256-4b6a-99fb-ae6c9957f60a	generic_instance	modified_dt	b00c40c2-0afb-46fa-b628-daffa4dcd1c7	CWX142	2025-07-03 08:23:16.858263-07	2025-07-03T08:23:16.863962-07:00	daylily	2025-07-03 08:23:16.863962-07	UPDATE	\N	\N	\N	f	f
c2862b0c-2822-4044-bb65-2e19690f1550	generic_instance_lineage	\N	85e42541-418b-4d4a-a4d1-49d0df5661bf	GL416	\N	\N	daylily	2025-07-03 08:23:16.865432-07	INSERT	\N	\N	\N	f	f
6a5e8e33-999f-4fc4-a57d-e25273af34b0	generic_instance	\N	662cdf0b-bf33-490a-997c-2ac9121bb7b9	CWX143	\N	\N	daylily	2025-07-03 08:23:16.865432-07	INSERT	\N	\N	\N	f	f
bd05ab92-f3c7-4956-9598-534c35878591	generic_instance	json_addl	662cdf0b-bf33-490a-997c-2ac9121bb7b9	CWX143	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:16.871192-07	UPDATE	\N	\N	\N	f	f
37298a9d-bff6-46e3-a490-d7517b8742c2	generic_instance	created_dt	662cdf0b-bf33-490a-997c-2ac9121bb7b9	CWX143	2025-07-03 08:23:16.865432-07	2025-07-03T08:23:16.865432-07:00	daylily	2025-07-03 08:23:16.871192-07	UPDATE	\N	\N	\N	f	f
5cfffd73-03ca-425c-b19a-e626312c52aa	generic_instance	modified_dt	662cdf0b-bf33-490a-997c-2ac9121bb7b9	CWX143	2025-07-03 08:23:16.865432-07	2025-07-03T08:23:16.871192-07:00	daylily	2025-07-03 08:23:16.871192-07	UPDATE	\N	\N	\N	f	f
6e132bed-4e2f-466a-a05c-34b86c00313b	generic_instance_lineage	\N	826b746c-3175-4835-a1dd-1fa25531c5ee	GL417	\N	\N	daylily	2025-07-03 08:23:16.872345-07	INSERT	\N	\N	\N	f	f
f6620c23-4726-4726-9601-39bd43f7980a	generic_instance	\N	b3bb08ee-3efd-406b-ac7f-0959a69c2ff5	CWX144	\N	\N	daylily	2025-07-03 08:23:16.872345-07	INSERT	\N	\N	\N	f	f
2a96849b-ac83-4610-b2b8-bed042cd53a6	generic_instance	json_addl	b3bb08ee-3efd-406b-ac7f-0959a69c2ff5	CWX144	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:16.876773-07	UPDATE	\N	\N	\N	f	f
348b684f-11c2-47f5-8983-a9e550d6a5c0	generic_instance	created_dt	b3bb08ee-3efd-406b-ac7f-0959a69c2ff5	CWX144	2025-07-03 08:23:16.872345-07	2025-07-03T08:23:16.872345-07:00	daylily	2025-07-03 08:23:16.876773-07	UPDATE	\N	\N	\N	f	f
23ca9131-92df-46be-9171-c1486e5b0ac8	generic_instance	modified_dt	b3bb08ee-3efd-406b-ac7f-0959a69c2ff5	CWX144	2025-07-03 08:23:16.872345-07	2025-07-03T08:23:16.876773-07:00	daylily	2025-07-03 08:23:16.876773-07	UPDATE	\N	\N	\N	f	f
6c3f9e48-c4a2-47d0-ba1e-d2319fafd14c	generic_instance_lineage	\N	dd9a2971-6a8d-45e8-b8ee-83c101c8bca7	GL418	\N	\N	daylily	2025-07-03 08:23:16.878044-07	INSERT	\N	\N	\N	f	f
a60c5ed6-3179-4d56-81fd-90ec838e925c	generic_instance	\N	c0ab8a50-6008-48dc-b1ed-35d21741141b	CWX145	\N	\N	daylily	2025-07-03 08:23:16.878044-07	INSERT	\N	\N	\N	f	f
750e7462-9c80-4f23-bbb1-2f62d9fecc0f	generic_instance	json_addl	c0ab8a50-6008-48dc-b1ed-35d21741141b	CWX145	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:16.882158-07	UPDATE	\N	\N	\N	f	f
a464743f-dacf-4cb5-9487-13009847704c	generic_instance	created_dt	c0ab8a50-6008-48dc-b1ed-35d21741141b	CWX145	2025-07-03 08:23:16.878044-07	2025-07-03T08:23:16.878044-07:00	daylily	2025-07-03 08:23:16.882158-07	UPDATE	\N	\N	\N	f	f
655de1bd-3211-4763-86a1-8f6436b25497	generic_instance	modified_dt	c0ab8a50-6008-48dc-b1ed-35d21741141b	CWX145	2025-07-03 08:23:16.878044-07	2025-07-03T08:23:16.882158-07:00	daylily	2025-07-03 08:23:16.882158-07	UPDATE	\N	\N	\N	f	f
2c2d7f57-1085-4dc5-a6ed-39b28bc268c2	generic_instance_lineage	\N	14ce0dba-39dd-41b0-bc3f-72f6ce423a9d	GL419	\N	\N	daylily	2025-07-03 08:23:16.88381-07	INSERT	\N	\N	\N	f	f
771ad7e8-e2b1-442d-8475-da6458ec84a9	generic_instance	\N	44535f96-b261-441e-8059-2f7c5632858d	CWX146	\N	\N	daylily	2025-07-03 08:23:16.88381-07	INSERT	\N	\N	\N	f	f
9774a0c7-f543-44fa-9d3f-f853b60f7b64	generic_instance	json_addl	44535f96-b261-441e-8059-2f7c5632858d	CWX146	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:16.888274-07	UPDATE	\N	\N	\N	f	f
9b9aff8e-910e-4463-ad0b-a23c6e9848c4	generic_instance	created_dt	44535f96-b261-441e-8059-2f7c5632858d	CWX146	2025-07-03 08:23:16.88381-07	2025-07-03T08:23:16.88381-07:00	daylily	2025-07-03 08:23:16.888274-07	UPDATE	\N	\N	\N	f	f
185ead70-6cb6-4a30-a26b-c878612dc597	generic_instance	modified_dt	44535f96-b261-441e-8059-2f7c5632858d	CWX146	2025-07-03 08:23:16.88381-07	2025-07-03T08:23:16.888274-07:00	daylily	2025-07-03 08:23:16.888274-07	UPDATE	\N	\N	\N	f	f
2a63312c-5ad4-400f-b681-ff2b7a0eeabf	generic_instance_lineage	\N	1e49a2c7-894f-4685-90d3-72c574cb8b29	GL420	\N	\N	daylily	2025-07-03 08:23:16.891679-07	INSERT	\N	\N	\N	f	f
514cffd1-015e-49a0-993f-e8c461b46c8d	generic_instance	\N	4d5db19c-2950-4351-8d87-fe365be8e7aa	CWX147	\N	\N	daylily	2025-07-03 08:23:16.891679-07	INSERT	\N	\N	\N	f	f
e07883ac-0567-4ffc-85a6-232bea8b8857	generic_instance	json_addl	4d5db19c-2950-4351-8d87-fe365be8e7aa	CWX147	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:16.897352-07	UPDATE	\N	\N	\N	f	f
ebbc55c5-5fca-4c58-a69a-de7d14b188e2	generic_instance	created_dt	4d5db19c-2950-4351-8d87-fe365be8e7aa	CWX147	2025-07-03 08:23:16.891679-07	2025-07-03T08:23:16.891679-07:00	daylily	2025-07-03 08:23:16.897352-07	UPDATE	\N	\N	\N	f	f
006a5663-6c43-446b-b2db-dd80c67dc21a	generic_instance	modified_dt	4d5db19c-2950-4351-8d87-fe365be8e7aa	CWX147	2025-07-03 08:23:16.891679-07	2025-07-03T08:23:16.897352-07:00	daylily	2025-07-03 08:23:16.897352-07	UPDATE	\N	\N	\N	f	f
139a741a-5f98-4bb2-8a93-3675494bf37d	generic_instance_lineage	\N	b5bf008b-a217-44f2-816f-63881ecf045f	GL421	\N	\N	daylily	2025-07-03 08:23:16.898679-07	INSERT	\N	\N	\N	f	f
3436faf4-b501-4a70-87b0-15f84a3abe25	generic_instance	\N	3aa39100-5672-4c10-a9f9-d0f60b0f437a	CWX148	\N	\N	daylily	2025-07-03 08:23:16.898679-07	INSERT	\N	\N	\N	f	f
f70ea9e7-630d-473e-9ee1-4d38f75032a1	generic_instance	json_addl	3aa39100-5672-4c10-a9f9-d0f60b0f437a	CWX148	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:16.903472-07	UPDATE	\N	\N	\N	f	f
9184182d-4714-42e1-99fc-d5e7fb21fcc1	generic_instance	created_dt	3aa39100-5672-4c10-a9f9-d0f60b0f437a	CWX148	2025-07-03 08:23:16.898679-07	2025-07-03T08:23:16.898679-07:00	daylily	2025-07-03 08:23:16.903472-07	UPDATE	\N	\N	\N	f	f
73cd937f-3f4c-4c67-9253-6cd0f5a2981e	generic_instance	modified_dt	3aa39100-5672-4c10-a9f9-d0f60b0f437a	CWX148	2025-07-03 08:23:16.898679-07	2025-07-03T08:23:16.903472-07:00	daylily	2025-07-03 08:23:16.903472-07	UPDATE	\N	\N	\N	f	f
f7e22947-67e5-4dd9-a6a3-dece20ec0f52	generic_instance_lineage	\N	70fbbb5a-cb5f-49d1-bdc9-38445865e557	GL422	\N	\N	daylily	2025-07-03 08:23:16.904772-07	INSERT	\N	\N	\N	f	f
629c23a4-4849-474c-ad9b-f176ffa32d59	generic_instance	\N	382bee5c-5a83-4842-8385-eec79639cc66	CWX149	\N	\N	daylily	2025-07-03 08:23:16.904772-07	INSERT	\N	\N	\N	f	f
c11da2c0-575a-42b2-974d-a2cf132854ae	generic_instance	json_addl	382bee5c-5a83-4842-8385-eec79639cc66	CWX149	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:16.909154-07	UPDATE	\N	\N	\N	f	f
d3ce1392-8825-4f05-a98f-2006cd5b24a0	generic_instance	created_dt	382bee5c-5a83-4842-8385-eec79639cc66	CWX149	2025-07-03 08:23:16.904772-07	2025-07-03T08:23:16.904772-07:00	daylily	2025-07-03 08:23:16.909154-07	UPDATE	\N	\N	\N	f	f
fe3ff2a2-358d-4422-8b58-35871decddca	generic_instance	modified_dt	382bee5c-5a83-4842-8385-eec79639cc66	CWX149	2025-07-03 08:23:16.904772-07	2025-07-03T08:23:16.909154-07:00	daylily	2025-07-03 08:23:16.909154-07	UPDATE	\N	\N	\N	f	f
c3641de4-6a1f-4d95-a67f-7c526cb23ad6	generic_instance_lineage	\N	1f479497-6bff-486c-b662-149739c24c70	GL423	\N	\N	daylily	2025-07-03 08:23:16.910378-07	INSERT	\N	\N	\N	f	f
087393fd-5e5b-48ba-8a40-6f7167f4d7c9	generic_instance	\N	36808dc6-c5b6-423e-848d-ed03c8a4e4d4	CWX150	\N	\N	daylily	2025-07-03 08:23:16.910378-07	INSERT	\N	\N	\N	f	f
a09beaa5-c684-48bf-a483-1e1ff137f7b7	generic_instance	json_addl	36808dc6-c5b6-423e-848d-ed03c8a4e4d4	CWX150	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:16.914525-07	UPDATE	\N	\N	\N	f	f
dd74e022-a34b-4004-9c09-d202adb76e6c	generic_instance	created_dt	36808dc6-c5b6-423e-848d-ed03c8a4e4d4	CWX150	2025-07-03 08:23:16.910378-07	2025-07-03T08:23:16.910378-07:00	daylily	2025-07-03 08:23:16.914525-07	UPDATE	\N	\N	\N	f	f
dcf4f08c-e689-4d6a-88e9-da5299ef8d79	generic_instance	modified_dt	36808dc6-c5b6-423e-848d-ed03c8a4e4d4	CWX150	2025-07-03 08:23:16.910378-07	2025-07-03T08:23:16.914525-07:00	daylily	2025-07-03 08:23:16.914525-07	UPDATE	\N	\N	\N	f	f
57073295-d586-49e8-8e31-0f6008d002a3	generic_instance_lineage	\N	f82d692f-aa1e-4bea-9cd1-cccf330bc8e6	GL424	\N	\N	daylily	2025-07-03 08:23:16.91574-07	INSERT	\N	\N	\N	f	f
9cf014e1-c522-4cc0-bb34-7f14e401e3d0	generic_instance	\N	304ab3f3-404d-4038-b781-6e028a3aac54	CWX151	\N	\N	daylily	2025-07-03 08:23:16.91574-07	INSERT	\N	\N	\N	f	f
0cb385d4-d644-42eb-9c98-88e3c42c7197	generic_instance	json_addl	304ab3f3-404d-4038-b781-6e028a3aac54	CWX151	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:16.919569-07	UPDATE	\N	\N	\N	f	f
d693bc29-05fd-44e8-b0c7-fe7298bfaa89	generic_instance	created_dt	304ab3f3-404d-4038-b781-6e028a3aac54	CWX151	2025-07-03 08:23:16.91574-07	2025-07-03T08:23:16.91574-07:00	daylily	2025-07-03 08:23:16.919569-07	UPDATE	\N	\N	\N	f	f
45db8462-62fd-459b-a5c9-a046ec6eb666	generic_instance	modified_dt	304ab3f3-404d-4038-b781-6e028a3aac54	CWX151	2025-07-03 08:23:16.91574-07	2025-07-03T08:23:16.919569-07:00	daylily	2025-07-03 08:23:16.919569-07	UPDATE	\N	\N	\N	f	f
df789b43-4740-4711-9814-7f10ad3a3f41	generic_instance_lineage	\N	50d1361b-7633-44a4-a0a6-2ab9a8040d94	GL425	\N	\N	daylily	2025-07-03 08:23:16.9209-07	INSERT	\N	\N	\N	f	f
7dc87838-5e9f-4338-b582-d9fb2f497b1b	generic_instance	\N	bad6ad8a-77c4-4d48-81a1-858893e9a864	CWX152	\N	\N	daylily	2025-07-03 08:23:16.9209-07	INSERT	\N	\N	\N	f	f
c1b8566c-33ac-440a-97f6-269c0934da7c	generic_instance	json_addl	bad6ad8a-77c4-4d48-81a1-858893e9a864	CWX152	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:16.925115-07	UPDATE	\N	\N	\N	f	f
625d26b9-3355-44cf-9045-163467a5c9ad	generic_instance	created_dt	bad6ad8a-77c4-4d48-81a1-858893e9a864	CWX152	2025-07-03 08:23:16.9209-07	2025-07-03T08:23:16.9209-07:00	daylily	2025-07-03 08:23:16.925115-07	UPDATE	\N	\N	\N	f	f
b60315c1-94ba-4879-b012-bbcf44ebdb62	generic_instance	modified_dt	bad6ad8a-77c4-4d48-81a1-858893e9a864	CWX152	2025-07-03 08:23:16.9209-07	2025-07-03T08:23:16.925115-07:00	daylily	2025-07-03 08:23:16.925115-07	UPDATE	\N	\N	\N	f	f
9deefb4b-c44c-4a11-b75b-a345ef34152f	generic_instance_lineage	\N	ee2fbff9-7804-4a34-9ef2-40c364100fda	GL426	\N	\N	daylily	2025-07-03 08:23:16.92652-07	INSERT	\N	\N	\N	f	f
391ae3ce-fd73-4e69-9bb9-b94c28ee3c7c	generic_instance	\N	6a5c1eac-1cbf-4e8f-87dd-63ceaee4a906	CWX153	\N	\N	daylily	2025-07-03 08:23:16.92652-07	INSERT	\N	\N	\N	f	f
fe0e9b26-da30-43f4-91c3-03deca9662d1	generic_instance	json_addl	6a5c1eac-1cbf-4e8f-87dd-63ceaee4a906	CWX153	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:16.931112-07	UPDATE	\N	\N	\N	f	f
3bce2e6b-08e4-4e9a-b566-bf4532a64e48	generic_instance	created_dt	6a5c1eac-1cbf-4e8f-87dd-63ceaee4a906	CWX153	2025-07-03 08:23:16.92652-07	2025-07-03T08:23:16.92652-07:00	daylily	2025-07-03 08:23:16.931112-07	UPDATE	\N	\N	\N	f	f
e9b837ea-364e-4032-a6ec-c0d1e0ece550	generic_instance	modified_dt	6a5c1eac-1cbf-4e8f-87dd-63ceaee4a906	CWX153	2025-07-03 08:23:16.92652-07	2025-07-03T08:23:16.931112-07:00	daylily	2025-07-03 08:23:16.931112-07	UPDATE	\N	\N	\N	f	f
b2bc6f5c-2224-4f75-85fc-6c874a90f67d	generic_instance_lineage	\N	bbb4a931-9249-4bf4-a71f-fa7e511dc1c4	GL427	\N	\N	daylily	2025-07-03 08:23:16.932457-07	INSERT	\N	\N	\N	f	f
6d5f8639-ff43-43aa-b8cd-cc0d0ed4eb0f	generic_instance	\N	c45f4209-c43e-4e0b-a0c6-4d8a5ea4d202	CWX154	\N	\N	daylily	2025-07-03 08:23:16.932457-07	INSERT	\N	\N	\N	f	f
488f56a8-6c04-413d-b620-38a40dab8abc	generic_instance	json_addl	c45f4209-c43e-4e0b-a0c6-4d8a5ea4d202	CWX154	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:16.937418-07	UPDATE	\N	\N	\N	f	f
e1ecb881-f3e4-453b-983b-d75f750392df	generic_instance	created_dt	c45f4209-c43e-4e0b-a0c6-4d8a5ea4d202	CWX154	2025-07-03 08:23:16.932457-07	2025-07-03T08:23:16.932457-07:00	daylily	2025-07-03 08:23:16.937418-07	UPDATE	\N	\N	\N	f	f
463227fa-4a6d-4431-ba65-f8e424af088a	generic_instance	modified_dt	c45f4209-c43e-4e0b-a0c6-4d8a5ea4d202	CWX154	2025-07-03 08:23:16.932457-07	2025-07-03T08:23:16.937418-07:00	daylily	2025-07-03 08:23:16.937418-07	UPDATE	\N	\N	\N	f	f
221b6557-1eda-422f-98be-0f4fb2e13995	generic_instance_lineage	\N	a24fd27a-cd4e-4640-8230-bd6cdc0238b7	GL428	\N	\N	daylily	2025-07-03 08:23:16.938673-07	INSERT	\N	\N	\N	f	f
94953d3c-9cde-47a1-9f5c-48edfd0ea2e1	generic_instance	\N	cfba7c0a-1491-40ab-9d5c-27d8c5ce367e	CWX155	\N	\N	daylily	2025-07-03 08:23:16.938673-07	INSERT	\N	\N	\N	f	f
b351a389-bb3e-4f07-b6a4-7b831c87fbbf	generic_instance	json_addl	cfba7c0a-1491-40ab-9d5c-27d8c5ce367e	CWX155	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:16.942878-07	UPDATE	\N	\N	\N	f	f
3007f99a-5067-496c-81e0-55e059c838db	generic_instance	created_dt	cfba7c0a-1491-40ab-9d5c-27d8c5ce367e	CWX155	2025-07-03 08:23:16.938673-07	2025-07-03T08:23:16.938673-07:00	daylily	2025-07-03 08:23:16.942878-07	UPDATE	\N	\N	\N	f	f
47bfd4e0-7e2a-4047-80b2-3df6fe2094fc	generic_instance	modified_dt	cfba7c0a-1491-40ab-9d5c-27d8c5ce367e	CWX155	2025-07-03 08:23:16.938673-07	2025-07-03T08:23:16.942878-07:00	daylily	2025-07-03 08:23:16.942878-07	UPDATE	\N	\N	\N	f	f
0f8c27d2-4c04-47a4-a9ff-038f3d263644	generic_instance_lineage	\N	e6006c9a-7cf0-4e4e-bc4b-97570b66418c	GL429	\N	\N	daylily	2025-07-03 08:23:16.944108-07	INSERT	\N	\N	\N	f	f
226ec9ef-8ded-41ec-8768-0c43b82cf2a2	generic_instance	\N	92af61ad-da5e-4378-bf20-1d69de145551	CWX156	\N	\N	daylily	2025-07-03 08:23:16.944108-07	INSERT	\N	\N	\N	f	f
4d59dafa-9a56-480d-b7b8-223094166ee6	generic_instance	json_addl	92af61ad-da5e-4378-bf20-1d69de145551	CWX156	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:16.948262-07	UPDATE	\N	\N	\N	f	f
2dd46e5a-f149-47fc-8d19-16809aa9feb6	generic_instance	created_dt	92af61ad-da5e-4378-bf20-1d69de145551	CWX156	2025-07-03 08:23:16.944108-07	2025-07-03T08:23:16.944108-07:00	daylily	2025-07-03 08:23:16.948262-07	UPDATE	\N	\N	\N	f	f
302c026f-aaa4-4c50-8c87-c8b45e563164	generic_instance	modified_dt	92af61ad-da5e-4378-bf20-1d69de145551	CWX156	2025-07-03 08:23:16.944108-07	2025-07-03T08:23:16.948262-07:00	daylily	2025-07-03 08:23:16.948262-07	UPDATE	\N	\N	\N	f	f
0092239b-2909-44bb-9bd6-1b725f23fd94	generic_instance_lineage	\N	33fb8cab-f33a-45f3-915f-2fab93aa6545	GL430	\N	\N	daylily	2025-07-03 08:23:16.94952-07	INSERT	\N	\N	\N	f	f
bbbaa7ee-358f-400f-9193-3e7baffdb82d	generic_instance	\N	8119664c-7b20-4d9d-b788-7ddefbc086e4	CWX157	\N	\N	daylily	2025-07-03 08:23:16.94952-07	INSERT	\N	\N	\N	f	f
1825b55a-d502-4b24-84ce-af7adcf551c7	generic_instance	json_addl	8119664c-7b20-4d9d-b788-7ddefbc086e4	CWX157	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:16.953604-07	UPDATE	\N	\N	\N	f	f
0eb0c179-05da-402d-a958-ac8242272b1a	generic_instance	created_dt	8119664c-7b20-4d9d-b788-7ddefbc086e4	CWX157	2025-07-03 08:23:16.94952-07	2025-07-03T08:23:16.94952-07:00	daylily	2025-07-03 08:23:16.953604-07	UPDATE	\N	\N	\N	f	f
ac841e68-673f-42f8-9a1e-aacab39064e5	generic_instance	modified_dt	8119664c-7b20-4d9d-b788-7ddefbc086e4	CWX157	2025-07-03 08:23:16.94952-07	2025-07-03T08:23:16.953604-07:00	daylily	2025-07-03 08:23:16.953604-07	UPDATE	\N	\N	\N	f	f
0bfe8abf-1674-478f-bef5-b2bce4f7f31c	generic_instance_lineage	\N	64b043b9-29e4-48fb-b4af-03486b298b9e	GL431	\N	\N	daylily	2025-07-03 08:23:16.955038-07	INSERT	\N	\N	\N	f	f
eb4475c4-9321-4aad-b7cf-f8d7bd276631	generic_instance	\N	be75f19a-9204-45e4-ac52-c07157044eb9	CWX158	\N	\N	daylily	2025-07-03 08:23:16.955038-07	INSERT	\N	\N	\N	f	f
5b3c1e90-f2e8-4dad-8c9a-2e9addd94aea	generic_instance	json_addl	be75f19a-9204-45e4-ac52-c07157044eb9	CWX158	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:16.959483-07	UPDATE	\N	\N	\N	f	f
d356a11f-bb3c-4f2a-af51-a3c10b61d773	generic_instance	created_dt	be75f19a-9204-45e4-ac52-c07157044eb9	CWX158	2025-07-03 08:23:16.955038-07	2025-07-03T08:23:16.955038-07:00	daylily	2025-07-03 08:23:16.959483-07	UPDATE	\N	\N	\N	f	f
fad8e76a-d3ec-4e01-b9f3-47e512980282	generic_instance	modified_dt	be75f19a-9204-45e4-ac52-c07157044eb9	CWX158	2025-07-03 08:23:16.955038-07	2025-07-03T08:23:16.959483-07:00	daylily	2025-07-03 08:23:16.959483-07	UPDATE	\N	\N	\N	f	f
1511eb95-0a9f-41c5-b29a-510b231059b6	generic_instance_lineage	\N	73735461-d4d8-4f6a-86bd-8b566d592fd7	GL432	\N	\N	daylily	2025-07-03 08:23:16.96069-07	INSERT	\N	\N	\N	f	f
c834f1ed-ee29-499e-87c6-4df2a025c351	generic_instance	\N	03fd4454-0700-4766-adf4-40cd1f0847a2	CWX159	\N	\N	daylily	2025-07-03 08:23:16.96069-07	INSERT	\N	\N	\N	f	f
e59fcd0e-66f1-4368-b217-d5a2d01c71f5	generic_instance	json_addl	03fd4454-0700-4766-adf4-40cd1f0847a2	CWX159	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:16.966525-07	UPDATE	\N	\N	\N	f	f
583403da-0597-41ce-ae5f-31b1c26a7ea8	generic_instance	created_dt	03fd4454-0700-4766-adf4-40cd1f0847a2	CWX159	2025-07-03 08:23:16.96069-07	2025-07-03T08:23:16.96069-07:00	daylily	2025-07-03 08:23:16.966525-07	UPDATE	\N	\N	\N	f	f
4a84e735-9987-4b3f-8231-d7904aca79e1	generic_instance	modified_dt	03fd4454-0700-4766-adf4-40cd1f0847a2	CWX159	2025-07-03 08:23:16.96069-07	2025-07-03T08:23:16.966525-07:00	daylily	2025-07-03 08:23:16.966525-07	UPDATE	\N	\N	\N	f	f
058a5963-4e2c-43ba-9b11-3f0920dafee2	generic_instance_lineage	\N	27560227-45fc-4f81-a66e-ac717bf72320	GL433	\N	\N	daylily	2025-07-03 08:23:16.968505-07	INSERT	\N	\N	\N	f	f
d1819d1d-4e44-448d-91df-32c4a9be2377	generic_instance	\N	f653b34e-100a-46a3-8ab7-9afaa9cd74d0	CWX160	\N	\N	daylily	2025-07-03 08:23:16.968505-07	INSERT	\N	\N	\N	f	f
adfb2b8e-6b66-4bb0-8e13-3506aabc0054	generic_instance	json_addl	f653b34e-100a-46a3-8ab7-9afaa9cd74d0	CWX160	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:16.972804-07	UPDATE	\N	\N	\N	f	f
941fd2b0-6fc1-47f9-a0bf-801a2ba04f6f	generic_instance	created_dt	f653b34e-100a-46a3-8ab7-9afaa9cd74d0	CWX160	2025-07-03 08:23:16.968505-07	2025-07-03T08:23:16.968505-07:00	daylily	2025-07-03 08:23:16.972804-07	UPDATE	\N	\N	\N	f	f
5a80cc66-6e67-4108-bcf7-e3c7eafb48c0	generic_instance	modified_dt	f653b34e-100a-46a3-8ab7-9afaa9cd74d0	CWX160	2025-07-03 08:23:16.968505-07	2025-07-03T08:23:16.972804-07:00	daylily	2025-07-03 08:23:16.972804-07	UPDATE	\N	\N	\N	f	f
8acb2541-3f2c-4188-a513-a757cfb34621	generic_instance_lineage	\N	e5518eff-c507-4a67-8647-582810906e88	GL434	\N	\N	daylily	2025-07-03 08:23:16.974229-07	INSERT	\N	\N	\N	f	f
cdbc1464-4384-4eeb-a258-886d91e04870	generic_instance	\N	db0c762c-c36b-42de-bc97-6976ba8c53be	CWX161	\N	\N	daylily	2025-07-03 08:23:16.974229-07	INSERT	\N	\N	\N	f	f
7428cc96-5a43-4bfe-a3ec-ce6c128f0223	generic_instance	json_addl	db0c762c-c36b-42de-bc97-6976ba8c53be	CWX161	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:16.980754-07	UPDATE	\N	\N	\N	f	f
b4c392e9-e985-41cd-aad7-7e0fcad9fd04	generic_instance	created_dt	db0c762c-c36b-42de-bc97-6976ba8c53be	CWX161	2025-07-03 08:23:16.974229-07	2025-07-03T08:23:16.974229-07:00	daylily	2025-07-03 08:23:16.980754-07	UPDATE	\N	\N	\N	f	f
8bc80a32-8998-4331-beb1-aa601513608f	generic_instance	modified_dt	db0c762c-c36b-42de-bc97-6976ba8c53be	CWX161	2025-07-03 08:23:16.974229-07	2025-07-03T08:23:16.980754-07:00	daylily	2025-07-03 08:23:16.980754-07	UPDATE	\N	\N	\N	f	f
9c3438cb-3872-4e60-8421-d9b8533ce413	generic_instance_lineage	\N	178aac1e-4d01-4c01-9f6c-c44d25d1fa1f	GL435	\N	\N	daylily	2025-07-03 08:23:16.982042-07	INSERT	\N	\N	\N	f	f
dede83ad-3fe0-4775-a1d8-b21cee556659	generic_instance	\N	01b72489-6786-4b1f-a8b7-25b56770d48f	CWX162	\N	\N	daylily	2025-07-03 08:23:16.982042-07	INSERT	\N	\N	\N	f	f
08e016ce-52fa-4efe-b563-56857703f9fc	generic_instance	json_addl	01b72489-6786-4b1f-a8b7-25b56770d48f	CWX162	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:16.986258-07	UPDATE	\N	\N	\N	f	f
6f782f40-d180-4b5d-b21f-7eea62771f6f	generic_instance	created_dt	01b72489-6786-4b1f-a8b7-25b56770d48f	CWX162	2025-07-03 08:23:16.982042-07	2025-07-03T08:23:16.982042-07:00	daylily	2025-07-03 08:23:16.986258-07	UPDATE	\N	\N	\N	f	f
c430f16b-e113-4eeb-82f3-2708eeb6cc4b	generic_instance	modified_dt	01b72489-6786-4b1f-a8b7-25b56770d48f	CWX162	2025-07-03 08:23:16.982042-07	2025-07-03T08:23:16.986258-07:00	daylily	2025-07-03 08:23:16.986258-07	UPDATE	\N	\N	\N	f	f
664075ca-eba6-498b-81a1-d1f152f634aa	generic_instance_lineage	\N	d5ce552e-a6d9-4d70-b1ce-305f17f7bc13	GL436	\N	\N	daylily	2025-07-03 08:23:16.987527-07	INSERT	\N	\N	\N	f	f
8c5fe289-b7da-4646-bdb5-01fa7778e7d4	generic_instance	\N	70bfdd8e-9926-4052-a785-1555e295832a	CWX163	\N	\N	daylily	2025-07-03 08:23:16.987527-07	INSERT	\N	\N	\N	f	f
5b6243f3-dde7-4d55-80de-7e6aa457859f	generic_instance	json_addl	70bfdd8e-9926-4052-a785-1555e295832a	CWX163	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:16.991348-07	UPDATE	\N	\N	\N	f	f
29ed7d52-f35b-4602-9778-4473b317ff31	generic_instance	created_dt	70bfdd8e-9926-4052-a785-1555e295832a	CWX163	2025-07-03 08:23:16.987527-07	2025-07-03T08:23:16.987527-07:00	daylily	2025-07-03 08:23:16.991348-07	UPDATE	\N	\N	\N	f	f
c90bc8cc-c536-4f08-a609-47113fe0c9cd	generic_instance	modified_dt	70bfdd8e-9926-4052-a785-1555e295832a	CWX163	2025-07-03 08:23:16.987527-07	2025-07-03T08:23:16.991348-07:00	daylily	2025-07-03 08:23:16.991348-07	UPDATE	\N	\N	\N	f	f
14f9d4ec-48b5-4a3c-a918-2dea5c1d9b67	generic_instance_lineage	\N	ef7b9389-0bc2-437d-81ed-dcb394ffbc7f	GL437	\N	\N	daylily	2025-07-03 08:23:16.992526-07	INSERT	\N	\N	\N	f	f
4ae61510-64e8-4009-aafe-17664f9edb5b	generic_instance	\N	ed6bd8d5-9e73-419c-afd6-6399c05e75a3	CWX164	\N	\N	daylily	2025-07-03 08:23:16.992526-07	INSERT	\N	\N	\N	f	f
1755fcb9-b671-4e6a-84fa-96f7a2296423	generic_instance	json_addl	ed6bd8d5-9e73-419c-afd6-6399c05e75a3	CWX164	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:16.996405-07	UPDATE	\N	\N	\N	f	f
1a5637a1-9016-427c-9a1e-27cf1e9e5e16	generic_instance	created_dt	ed6bd8d5-9e73-419c-afd6-6399c05e75a3	CWX164	2025-07-03 08:23:16.992526-07	2025-07-03T08:23:16.992526-07:00	daylily	2025-07-03 08:23:16.996405-07	UPDATE	\N	\N	\N	f	f
3994b4e1-8784-41ef-8302-fa74eef93f51	generic_instance	modified_dt	ed6bd8d5-9e73-419c-afd6-6399c05e75a3	CWX164	2025-07-03 08:23:16.992526-07	2025-07-03T08:23:16.996405-07:00	daylily	2025-07-03 08:23:16.996405-07	UPDATE	\N	\N	\N	f	f
7389d83a-94b8-426d-bdc9-9352e4f30eeb	generic_instance_lineage	\N	25a39b3d-48be-4584-b83a-43e2985ce506	GL438	\N	\N	daylily	2025-07-03 08:23:16.997679-07	INSERT	\N	\N	\N	f	f
8efb741f-748e-4c67-bcdd-fb1ce03260ec	generic_instance	\N	bbb4417b-7369-41c9-91b0-8fb0773fdab6	CWX165	\N	\N	daylily	2025-07-03 08:23:16.997679-07	INSERT	\N	\N	\N	f	f
b3d40e50-ae38-426f-8866-9fda3e9259d5	generic_instance	json_addl	bbb4417b-7369-41c9-91b0-8fb0773fdab6	CWX165	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.00214-07	UPDATE	\N	\N	\N	f	f
34932849-7be9-4cdf-81da-9dce6328492b	generic_instance	created_dt	bbb4417b-7369-41c9-91b0-8fb0773fdab6	CWX165	2025-07-03 08:23:16.997679-07	2025-07-03T08:23:16.997679-07:00	daylily	2025-07-03 08:23:17.00214-07	UPDATE	\N	\N	\N	f	f
44d16eb6-4f20-421b-ac63-52c3996775ea	generic_instance	modified_dt	bbb4417b-7369-41c9-91b0-8fb0773fdab6	CWX165	2025-07-03 08:23:16.997679-07	2025-07-03T08:23:17.00214-07:00	daylily	2025-07-03 08:23:17.00214-07	UPDATE	\N	\N	\N	f	f
41b98ffc-4f4c-445e-830e-0d9a60b51131	generic_instance_lineage	\N	3da3c520-44d8-408d-bce2-574814903a11	GL439	\N	\N	daylily	2025-07-03 08:23:17.003419-07	INSERT	\N	\N	\N	f	f
7a4a7ab5-c68f-490e-b0b5-2f21e704d8a3	generic_instance	\N	cf2a5e5f-bd04-40fa-a03e-f3efecc404b3	CWX166	\N	\N	daylily	2025-07-03 08:23:17.003419-07	INSERT	\N	\N	\N	f	f
976aa9c2-34a2-4b0e-a9f1-a4d6934d45f3	generic_instance	json_addl	cf2a5e5f-bd04-40fa-a03e-f3efecc404b3	CWX166	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.007562-07	UPDATE	\N	\N	\N	f	f
77065832-8404-492b-9b51-fa81adb314f5	generic_instance	created_dt	cf2a5e5f-bd04-40fa-a03e-f3efecc404b3	CWX166	2025-07-03 08:23:17.003419-07	2025-07-03T08:23:17.003419-07:00	daylily	2025-07-03 08:23:17.007562-07	UPDATE	\N	\N	\N	f	f
1cb397fe-5a4c-4dda-ab40-88a0f096d6a7	generic_instance	modified_dt	cf2a5e5f-bd04-40fa-a03e-f3efecc404b3	CWX166	2025-07-03 08:23:17.003419-07	2025-07-03T08:23:17.007562-07:00	daylily	2025-07-03 08:23:17.007562-07	UPDATE	\N	\N	\N	f	f
ab2fb279-cac7-4fd9-8e26-241d00dc7358	generic_instance_lineage	\N	dc00b76c-2219-4851-a0bf-15ca81b49e4e	GL440	\N	\N	daylily	2025-07-03 08:23:17.008874-07	INSERT	\N	\N	\N	f	f
57b3c106-dd74-4f72-9439-20ceeceae5c6	generic_instance	\N	6077eb5c-7bd2-436b-85a4-f1ec9e26cc5c	CWX167	\N	\N	daylily	2025-07-03 08:23:17.008874-07	INSERT	\N	\N	\N	f	f
fe4a39cf-09de-4cab-8436-58d1f279d363	generic_instance	json_addl	6077eb5c-7bd2-436b-85a4-f1ec9e26cc5c	CWX167	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.012753-07	UPDATE	\N	\N	\N	f	f
14dc6252-7bc5-482c-ae07-9a1fcd2a020b	generic_instance	created_dt	6077eb5c-7bd2-436b-85a4-f1ec9e26cc5c	CWX167	2025-07-03 08:23:17.008874-07	2025-07-03T08:23:17.008874-07:00	daylily	2025-07-03 08:23:17.012753-07	UPDATE	\N	\N	\N	f	f
fc3c75f0-a2a5-40fc-8cd4-712830e0a9a4	generic_instance	modified_dt	6077eb5c-7bd2-436b-85a4-f1ec9e26cc5c	CWX167	2025-07-03 08:23:17.008874-07	2025-07-03T08:23:17.012753-07:00	daylily	2025-07-03 08:23:17.012753-07	UPDATE	\N	\N	\N	f	f
2fa1f23d-2051-4db5-8b31-5e358c3e0580	generic_instance_lineage	\N	d89a3d71-6866-47df-81ac-4e80f39eeef5	GL441	\N	\N	daylily	2025-07-03 08:23:17.013944-07	INSERT	\N	\N	\N	f	f
97b0a380-8815-442d-bf62-ab8736a0e23b	generic_instance	\N	4218d37a-128b-45cf-b2ef-1a111112b524	CWX168	\N	\N	daylily	2025-07-03 08:23:17.013944-07	INSERT	\N	\N	\N	f	f
787cf3bc-146f-47c7-8cd9-0945a35b00dd	generic_instance	json_addl	4218d37a-128b-45cf-b2ef-1a111112b524	CWX168	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.017956-07	UPDATE	\N	\N	\N	f	f
b05c4da8-bc05-4042-9c0e-5f6596e31ddf	generic_instance	created_dt	4218d37a-128b-45cf-b2ef-1a111112b524	CWX168	2025-07-03 08:23:17.013944-07	2025-07-03T08:23:17.013944-07:00	daylily	2025-07-03 08:23:17.017956-07	UPDATE	\N	\N	\N	f	f
cbd6c075-a20b-4bdc-979b-ec59b5c9981b	generic_instance	modified_dt	4218d37a-128b-45cf-b2ef-1a111112b524	CWX168	2025-07-03 08:23:17.013944-07	2025-07-03T08:23:17.017956-07:00	daylily	2025-07-03 08:23:17.017956-07	UPDATE	\N	\N	\N	f	f
21a3c4f1-99dd-4d3e-8c19-042324d1a7a9	generic_instance_lineage	\N	871df990-bfd8-4d14-a5e5-2f40cc17e76a	GL442	\N	\N	daylily	2025-07-03 08:23:17.01932-07	INSERT	\N	\N	\N	f	f
2583d183-1668-4492-af9b-607da1d273db	generic_instance	\N	82f7d2fc-6a90-4574-a482-546dc5be34c5	MRX141	\N	\N	daylily	2025-07-03 08:23:17.021151-07	INSERT	\N	\N	\N	f	f
5aee4227-b396-4ae8-8d72-97307e35c159	generic_instance_lineage	\N	ff8de42c-43e0-4b9e-91d5-3dd56d1bef14	GL443	\N	\N	daylily	2025-07-03 08:23:17.023102-07	INSERT	\N	\N	\N	f	f
113988f6-e075-4cd6-b1ab-f5789c1d10c4	generic_instance	\N	c5d41a18-1176-4b9d-bfd7-e222b66b1dfc	MRX142	\N	\N	daylily	2025-07-03 08:23:17.023102-07	INSERT	\N	\N	\N	f	f
1d176c13-d1cc-4ee2-835d-2b879b251749	generic_instance_lineage	\N	e09f0196-336a-4f70-97c5-428580379204	GL444	\N	\N	daylily	2025-07-03 08:23:17.029585-07	INSERT	\N	\N	\N	f	f
cd44c652-ff41-4bac-8964-411c2195291b	generic_instance	\N	ba952458-8cff-4ff5-a9d2-0a6ab5405602	MRX143	\N	\N	daylily	2025-07-03 08:23:17.029585-07	INSERT	\N	\N	\N	f	f
7fef2bb3-4004-4a02-9505-f945aa13ef8b	generic_instance_lineage	\N	c9f99b5f-6046-487a-8b40-55df8e34e46f	GL445	\N	\N	daylily	2025-07-03 08:23:17.034913-07	INSERT	\N	\N	\N	f	f
df5e79ae-4897-45c3-9fd3-5685ac10673e	generic_instance	\N	525493a1-c68d-4d24-a2ee-d51be1599878	MRX144	\N	\N	daylily	2025-07-03 08:23:17.034913-07	INSERT	\N	\N	\N	f	f
0b99a654-bb06-4d03-8950-4dedd5de2ba3	generic_instance_lineage	\N	743efc0f-8bcd-435a-9f46-4f3dc62d85a6	GL446	\N	\N	daylily	2025-07-03 08:23:17.039681-07	INSERT	\N	\N	\N	f	f
121830ad-83ed-4046-bcef-f32624c9a948	generic_instance	\N	ef23526a-8c13-4561-b742-e38af3f43e6d	MRX145	\N	\N	daylily	2025-07-03 08:23:17.039681-07	INSERT	\N	\N	\N	f	f
0d57d6f2-ca23-47b2-8751-7792c67b0eeb	generic_instance_lineage	\N	94ad3b04-a6fa-430a-acba-645ce02212fa	GL447	\N	\N	daylily	2025-07-03 08:23:17.044588-07	INSERT	\N	\N	\N	f	f
d7ec9df5-f78b-4599-9771-057858321898	generic_instance	\N	a6a0f6f1-6a24-475e-919b-c0fbe85a297c	MRX146	\N	\N	daylily	2025-07-03 08:23:17.044588-07	INSERT	\N	\N	\N	f	f
a0acda79-761e-47b4-bcac-b1326ec30326	generic_instance_lineage	\N	93037a80-2592-47d1-9806-8f9b979149f5	GL448	\N	\N	daylily	2025-07-03 08:23:17.049301-07	INSERT	\N	\N	\N	f	f
74c7bb80-7d6d-47de-84b8-ba48e02ead7e	generic_instance	\N	7ad6a870-8313-4efb-ba81-5f4fb0ab06c3	MRX147	\N	\N	daylily	2025-07-03 08:23:17.049301-07	INSERT	\N	\N	\N	f	f
a946a241-7a3f-4816-a7a5-6e81b0df494c	generic_instance_lineage	\N	37476cb1-99c8-423d-a35e-8a806250b1b9	GL449	\N	\N	daylily	2025-07-03 08:23:17.053597-07	INSERT	\N	\N	\N	f	f
b3eecf8d-ed37-4a89-91eb-92232c06eb9e	generic_instance	\N	2fd021db-8cba-4e2b-9550-c0f078ed5ca5	MRX148	\N	\N	daylily	2025-07-03 08:23:17.053597-07	INSERT	\N	\N	\N	f	f
7619fab7-cdc7-4877-be0d-59f9f0c5bac2	generic_instance_lineage	\N	5ad35ea5-0160-4b55-b204-cfee16c12b02	GL450	\N	\N	daylily	2025-07-03 08:23:17.059703-07	INSERT	\N	\N	\N	f	f
c199da53-9000-4218-bf48-d86aebf42e92	generic_instance	\N	7c504afb-63e5-4f3e-9684-d7a685a74f33	MRX149	\N	\N	daylily	2025-07-03 08:23:17.059703-07	INSERT	\N	\N	\N	f	f
36af2908-7f10-482b-b43a-a9969ffd7c80	generic_instance_lineage	\N	c169b2c3-5d29-4c89-9d27-0c73abf43119	GL451	\N	\N	daylily	2025-07-03 08:23:17.067761-07	INSERT	\N	\N	\N	f	f
537a3ea9-2102-4d03-8179-7bd2347c37b0	generic_instance	\N	3c9ca22e-fa5e-4304-916b-d9b14224b0da	MRX150	\N	\N	daylily	2025-07-03 08:23:17.067761-07	INSERT	\N	\N	\N	f	f
0dede0f0-97f2-4b61-b43b-f59d373e9c3d	generic_instance_lineage	\N	9f726eea-7186-421e-b3f4-d28c19b15fc4	GL452	\N	\N	daylily	2025-07-03 08:23:17.072967-07	INSERT	\N	\N	\N	f	f
5018b98a-2e2d-41a6-aa43-019f06ea4eec	generic_instance	\N	47982770-15aa-46c6-bcb2-2dcfe0cf9fde	MRX151	\N	\N	daylily	2025-07-03 08:23:17.072967-07	INSERT	\N	\N	\N	f	f
a7ae7cf3-0986-47e0-832a-772fc569a2db	generic_instance_lineage	\N	edb196ae-ea93-47f3-afe8-77dbb23a98c6	GL453	\N	\N	daylily	2025-07-03 08:23:17.077892-07	INSERT	\N	\N	\N	f	f
989c2e71-08cf-41cf-930b-cbfbb3dd6d5c	generic_instance	\N	864663d1-dd5c-49ea-80c5-bcc1afaf6049	MRX152	\N	\N	daylily	2025-07-03 08:23:17.077892-07	INSERT	\N	\N	\N	f	f
56c21284-ee0a-43af-b5ab-95ba55c6cbb5	generic_instance_lineage	\N	f2dc86cc-069e-43ba-8ee1-84b2417f8a22	GL454	\N	\N	daylily	2025-07-03 08:23:17.082628-07	INSERT	\N	\N	\N	f	f
e4d2a24a-9297-4259-bdd1-b84886230196	generic_instance	\N	da2ddfd1-f611-4c8f-b85b-1c5e8f4d1a3f	MRX153	\N	\N	daylily	2025-07-03 08:23:17.082628-07	INSERT	\N	\N	\N	f	f
2f1a1a40-c4d4-4014-8bdf-d7c336f1d150	generic_instance_lineage	\N	484a0633-3de3-4c63-ab69-983e87374dc0	GL455	\N	\N	daylily	2025-07-03 08:23:17.087322-07	INSERT	\N	\N	\N	f	f
d26ae8eb-9226-4e3d-bda8-865bbd7439b5	generic_instance	\N	7aff050c-94f4-4a86-9bd3-e5ce0b2bcb7b	MRX154	\N	\N	daylily	2025-07-03 08:23:17.087322-07	INSERT	\N	\N	\N	f	f
2a18a5cb-3c98-47f3-8623-ddd7ce453b32	generic_instance_lineage	\N	9c1a6464-47ce-4a78-bc4a-2b812bfad394	GL456	\N	\N	daylily	2025-07-03 08:23:17.091748-07	INSERT	\N	\N	\N	f	f
f26fae81-ad06-4c38-bcb7-d3d8e6a93495	generic_instance	\N	e0a2ec5b-c727-46ff-8cfd-541ded1a37da	MRX155	\N	\N	daylily	2025-07-03 08:23:17.091748-07	INSERT	\N	\N	\N	f	f
9888ca22-901b-45f8-8417-fa5e209a07e7	generic_instance_lineage	\N	ddc86ea0-9241-4da0-bd9d-4147c24a40ef	GL457	\N	\N	daylily	2025-07-03 08:23:17.096479-07	INSERT	\N	\N	\N	f	f
73936927-fdeb-4b78-8532-5a093627d69c	generic_instance	\N	461bb485-e99a-4c28-8499-92794e89166b	MRX156	\N	\N	daylily	2025-07-03 08:23:17.096479-07	INSERT	\N	\N	\N	f	f
ed4cbb5d-ef03-47ef-b52b-889a911e139b	generic_instance_lineage	\N	6179cc59-385a-4a4a-8cc3-3b1c3fd6dd10	GL458	\N	\N	daylily	2025-07-03 08:23:17.101187-07	INSERT	\N	\N	\N	f	f
fdce61e4-644a-4d50-8b57-5bf38cc0b25f	generic_instance	\N	952fb2a0-50f8-426f-baaa-95d86e9af105	MRX157	\N	\N	daylily	2025-07-03 08:23:17.101187-07	INSERT	\N	\N	\N	f	f
aeb6cc6b-93d0-4cb2-940f-e8620b869f2e	generic_instance_lineage	\N	28a09cce-5e74-4bcf-9402-f2737dac2e7c	GL459	\N	\N	daylily	2025-07-03 08:23:17.105736-07	INSERT	\N	\N	\N	f	f
97213b88-944f-4194-95a7-8ecbf05dd5ce	generic_instance	\N	cb217410-9c27-445b-82c2-03e605c21be5	MRX158	\N	\N	daylily	2025-07-03 08:23:17.105736-07	INSERT	\N	\N	\N	f	f
d81e9f02-0e1e-4662-a5da-4c9dc6dbaa66	generic_instance_lineage	\N	a5d3b00d-160b-4960-baab-4b81a68d3210	GL460	\N	\N	daylily	2025-07-03 08:23:17.110472-07	INSERT	\N	\N	\N	f	f
66a0fa01-abf6-4b85-91ea-4cb01e006d4e	generic_instance	\N	c7c571ec-1754-432e-b4e7-a8b18a0dc670	MRX159	\N	\N	daylily	2025-07-03 08:23:17.110472-07	INSERT	\N	\N	\N	f	f
a0f6dd3a-1fb5-4de6-8fcf-33d83f9cf88b	generic_instance_lineage	\N	f01fc837-0b95-4eb1-ae0e-efb1bc4352b4	GL461	\N	\N	daylily	2025-07-03 08:23:17.115343-07	INSERT	\N	\N	\N	f	f
8be3c55f-66d4-4c17-bed0-18098b517f7c	generic_instance	\N	e6c30471-04b1-41ce-a0f9-dea3962bb40f	MRX160	\N	\N	daylily	2025-07-03 08:23:17.115343-07	INSERT	\N	\N	\N	f	f
7ba9994a-d620-46f7-b432-70e3369f40ac	generic_instance_lineage	\N	7c4c632e-22d5-45b2-bce0-1e207ef28dca	GL462	\N	\N	daylily	2025-07-03 08:23:17.120297-07	INSERT	\N	\N	\N	f	f
621f24b1-1479-4791-b1e5-1d130c8e7571	generic_instance	\N	61bd9a3a-9f50-40fa-bcb0-51a811d28a48	MRX161	\N	\N	daylily	2025-07-03 08:23:17.120297-07	INSERT	\N	\N	\N	f	f
f91ab12e-82db-48a8-8358-685c1225f12c	generic_instance_lineage	\N	11d20684-9b3c-4482-b247-2df05f3bb0ce	GL463	\N	\N	daylily	2025-07-03 08:23:17.124922-07	INSERT	\N	\N	\N	f	f
fd0239d9-ecf3-4378-a72a-23e8ac326515	generic_instance	\N	5b503e9f-0587-4d68-a122-eee644adffb3	MRX162	\N	\N	daylily	2025-07-03 08:23:17.124922-07	INSERT	\N	\N	\N	f	f
d5c7d935-061f-40c3-b371-bd7a3d7943df	generic_instance_lineage	\N	e7d5efbd-d094-4870-afbe-75756790f551	GL464	\N	\N	daylily	2025-07-03 08:23:17.129883-07	INSERT	\N	\N	\N	f	f
3f15fab2-e381-4f4c-bd38-5491b360ac13	generic_instance	\N	11bf48ed-61ca-4a6f-ad36-76f4f2eb9f55	MRX163	\N	\N	daylily	2025-07-03 08:23:17.129883-07	INSERT	\N	\N	\N	f	f
68272d04-7702-4ea4-bd55-85bff9679837	generic_instance	json_addl	4b6dd84b-e2c0-413d-aa2a-2a520af200af	CWX169	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.172522-07	UPDATE	\N	\N	\N	f	f
e98db400-7d36-4bb3-aa45-11d89ab31be1	generic_instance	created_dt	4b6dd84b-e2c0-413d-aa2a-2a520af200af	CWX169	2025-07-03 08:23:17.169419-07	2025-07-03T08:23:17.169419-07:00	daylily	2025-07-03 08:23:17.172522-07	UPDATE	\N	\N	\N	f	f
7192b263-c07a-4b3d-a9a8-16827856acce	generic_instance	modified_dt	4b6dd84b-e2c0-413d-aa2a-2a520af200af	CWX169	2025-07-03 08:23:17.169419-07	2025-07-03T08:23:17.172522-07:00	daylily	2025-07-03 08:23:17.172522-07	UPDATE	\N	\N	\N	f	f
33640b74-fd13-42ff-8c55-b0c555afbabc	generic_instance_lineage	\N	3dea1ad0-ff27-43bf-83fe-cb70f0bd154a	GL471	\N	\N	daylily	2025-07-03 08:23:17.174531-07	INSERT	\N	\N	\N	f	f
6325106c-06d0-43c9-ac79-957fe3366d7e	generic_instance	\N	e2728bf3-67bc-4811-bbe4-f0867231f6e0	CWX170	\N	\N	daylily	2025-07-03 08:23:17.174531-07	INSERT	\N	\N	\N	f	f
f2932ea3-9ef9-4049-985f-11089b5556e5	generic_instance	json_addl	e2728bf3-67bc-4811-bbe4-f0867231f6e0	CWX170	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.179427-07	UPDATE	\N	\N	\N	f	f
10bed35c-05b5-45dc-aae7-2adb7e25eda0	generic_instance	created_dt	e2728bf3-67bc-4811-bbe4-f0867231f6e0	CWX170	2025-07-03 08:23:17.174531-07	2025-07-03T08:23:17.174531-07:00	daylily	2025-07-03 08:23:17.179427-07	UPDATE	\N	\N	\N	f	f
1e432878-5e4f-490b-8e7e-9703db3c57b1	generic_instance	modified_dt	e2728bf3-67bc-4811-bbe4-f0867231f6e0	CWX170	2025-07-03 08:23:17.174531-07	2025-07-03T08:23:17.179427-07:00	daylily	2025-07-03 08:23:17.179427-07	UPDATE	\N	\N	\N	f	f
bbbd4732-29b7-4759-ac1a-542f44bf2fbf	generic_instance_lineage	\N	f003b658-5069-4534-840c-5ed3ed5ea149	GL472	\N	\N	daylily	2025-07-03 08:23:17.180834-07	INSERT	\N	\N	\N	f	f
8e1fb8e1-4b77-44f8-a7e0-dbe3736a198b	generic_instance	\N	521b22bc-7d76-41ca-b735-17767c6be7bb	CWX171	\N	\N	daylily	2025-07-03 08:23:17.180834-07	INSERT	\N	\N	\N	f	f
eb512652-5639-43d3-bd57-cd5b0f59ffc9	generic_instance	json_addl	521b22bc-7d76-41ca-b735-17767c6be7bb	CWX171	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.185151-07	UPDATE	\N	\N	\N	f	f
340d6d23-c023-48ae-9b05-df4b76c9eed1	generic_instance	created_dt	521b22bc-7d76-41ca-b735-17767c6be7bb	CWX171	2025-07-03 08:23:17.180834-07	2025-07-03T08:23:17.180834-07:00	daylily	2025-07-03 08:23:17.185151-07	UPDATE	\N	\N	\N	f	f
950918e7-59c4-4983-a6ae-10a0e95bc584	generic_instance	modified_dt	521b22bc-7d76-41ca-b735-17767c6be7bb	CWX171	2025-07-03 08:23:17.180834-07	2025-07-03T08:23:17.185151-07:00	daylily	2025-07-03 08:23:17.185151-07	UPDATE	\N	\N	\N	f	f
85b8a0c6-e9c1-46d1-b539-e489a96ee6f6	generic_instance_lineage	\N	fbcb4f04-2ebd-4692-9fa7-b493a77db7e8	GL473	\N	\N	daylily	2025-07-03 08:23:17.187614-07	INSERT	\N	\N	\N	f	f
7222825d-ccab-4c9a-9251-3e51fb321963	generic_instance	\N	7de76476-6e33-4d96-9a7b-333393b79eb4	CWX172	\N	\N	daylily	2025-07-03 08:23:17.187614-07	INSERT	\N	\N	\N	f	f
6087c68d-6848-4237-a9ad-a81e4cba89a1	generic_instance	json_addl	7de76476-6e33-4d96-9a7b-333393b79eb4	CWX172	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.191641-07	UPDATE	\N	\N	\N	f	f
516d1ad4-77f6-4f0f-b6e0-31f77b3167f8	generic_instance	created_dt	7de76476-6e33-4d96-9a7b-333393b79eb4	CWX172	2025-07-03 08:23:17.187614-07	2025-07-03T08:23:17.187614-07:00	daylily	2025-07-03 08:23:17.191641-07	UPDATE	\N	\N	\N	f	f
d1ff13d5-5567-49d8-9408-d1fc77912655	generic_instance	modified_dt	7de76476-6e33-4d96-9a7b-333393b79eb4	CWX172	2025-07-03 08:23:17.187614-07	2025-07-03T08:23:17.191641-07:00	daylily	2025-07-03 08:23:17.191641-07	UPDATE	\N	\N	\N	f	f
57b4ff07-eec7-49ec-abd3-1cd7286473eb	generic_instance_lineage	\N	436965c6-433f-4bca-b2c7-22f0d14da31b	GL474	\N	\N	daylily	2025-07-03 08:23:17.192961-07	INSERT	\N	\N	\N	f	f
36dbc0b2-c91a-4c68-b8e5-508ed4f4813f	generic_instance	\N	c938574e-08aa-44ce-9f09-460240d55a12	CWX173	\N	\N	daylily	2025-07-03 08:23:17.192961-07	INSERT	\N	\N	\N	f	f
298f3bbb-e307-414e-a47e-8b6e53eaf66e	generic_instance	json_addl	c938574e-08aa-44ce-9f09-460240d55a12	CWX173	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.197314-07	UPDATE	\N	\N	\N	f	f
d3960b72-d9d6-4a5b-af28-c874b82f08c7	generic_instance	created_dt	c938574e-08aa-44ce-9f09-460240d55a12	CWX173	2025-07-03 08:23:17.192961-07	2025-07-03T08:23:17.192961-07:00	daylily	2025-07-03 08:23:17.197314-07	UPDATE	\N	\N	\N	f	f
015d8f76-7949-49f3-befd-2c163ad172d4	generic_instance	modified_dt	c938574e-08aa-44ce-9f09-460240d55a12	CWX173	2025-07-03 08:23:17.192961-07	2025-07-03T08:23:17.197314-07:00	daylily	2025-07-03 08:23:17.197314-07	UPDATE	\N	\N	\N	f	f
86ba667f-c155-4787-8b26-34443c57880f	generic_instance_lineage	\N	101b2aa8-0038-41be-9569-78350b55685d	GL475	\N	\N	daylily	2025-07-03 08:23:17.198868-07	INSERT	\N	\N	\N	f	f
07c99606-b867-4eb3-8787-1b42fb219ce6	generic_instance	\N	c43f8239-2b30-4153-8e0e-1ca8109d4847	CWX174	\N	\N	daylily	2025-07-03 08:23:17.198868-07	INSERT	\N	\N	\N	f	f
aabca3bc-cdc6-4857-be67-a2499404b07d	generic_instance	json_addl	c43f8239-2b30-4153-8e0e-1ca8109d4847	CWX174	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.203494-07	UPDATE	\N	\N	\N	f	f
7e36b396-b8a5-41d6-856e-1cae59d48663	generic_instance	created_dt	c43f8239-2b30-4153-8e0e-1ca8109d4847	CWX174	2025-07-03 08:23:17.198868-07	2025-07-03T08:23:17.198868-07:00	daylily	2025-07-03 08:23:17.203494-07	UPDATE	\N	\N	\N	f	f
cc043ade-8811-4a99-b66c-42066dc4b423	generic_instance	modified_dt	c43f8239-2b30-4153-8e0e-1ca8109d4847	CWX174	2025-07-03 08:23:17.198868-07	2025-07-03T08:23:17.203494-07:00	daylily	2025-07-03 08:23:17.203494-07	UPDATE	\N	\N	\N	f	f
f5534c7c-ba42-4fd5-af2f-72d82d60440f	generic_instance_lineage	\N	8dcc3dcf-7e45-4b0d-9b77-ed993558750d	GL476	\N	\N	daylily	2025-07-03 08:23:17.20482-07	INSERT	\N	\N	\N	f	f
1c0e28bb-26b5-4f95-bfa6-92b273328d05	generic_instance	\N	38d72d79-0c20-43bf-a047-265347aaf013	CWX175	\N	\N	daylily	2025-07-03 08:23:17.20482-07	INSERT	\N	\N	\N	f	f
7d36c260-f55f-4123-8376-e92b20bb5608	generic_instance	json_addl	38d72d79-0c20-43bf-a047-265347aaf013	CWX175	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.209266-07	UPDATE	\N	\N	\N	f	f
4d609ce4-2162-45a9-a4c2-f0d222328e96	generic_instance	created_dt	38d72d79-0c20-43bf-a047-265347aaf013	CWX175	2025-07-03 08:23:17.20482-07	2025-07-03T08:23:17.20482-07:00	daylily	2025-07-03 08:23:17.209266-07	UPDATE	\N	\N	\N	f	f
c1d6e474-6446-47e0-864f-e3b53f31861b	generic_instance	modified_dt	38d72d79-0c20-43bf-a047-265347aaf013	CWX175	2025-07-03 08:23:17.20482-07	2025-07-03T08:23:17.209266-07:00	daylily	2025-07-03 08:23:17.209266-07	UPDATE	\N	\N	\N	f	f
22e7fa99-9d8f-49f2-9e9a-57a9bcfe7bc0	generic_instance_lineage	\N	b1581797-e19f-4c79-b915-9a3b77b87fe7	GL477	\N	\N	daylily	2025-07-03 08:23:17.210421-07	INSERT	\N	\N	\N	f	f
03bb013b-f111-48ee-bb35-00c30ce35478	generic_instance	\N	ad4b62e4-5a8f-4bb6-963b-923da6668cf8	CWX176	\N	\N	daylily	2025-07-03 08:23:17.210421-07	INSERT	\N	\N	\N	f	f
8088d13c-66bf-48f8-87cd-046d3e0f8458	generic_instance	json_addl	ad4b62e4-5a8f-4bb6-963b-923da6668cf8	CWX176	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.214462-07	UPDATE	\N	\N	\N	f	f
e28d47af-1956-41ad-994a-1b7e468f4aeb	generic_instance	created_dt	ad4b62e4-5a8f-4bb6-963b-923da6668cf8	CWX176	2025-07-03 08:23:17.210421-07	2025-07-03T08:23:17.210421-07:00	daylily	2025-07-03 08:23:17.214462-07	UPDATE	\N	\N	\N	f	f
60d3711c-c504-4046-a396-a64b56e81c24	generic_instance	modified_dt	ad4b62e4-5a8f-4bb6-963b-923da6668cf8	CWX176	2025-07-03 08:23:17.210421-07	2025-07-03T08:23:17.214462-07:00	daylily	2025-07-03 08:23:17.214462-07	UPDATE	\N	\N	\N	f	f
c6f704a8-1ae0-4fe3-bb2e-900e6badd33c	generic_instance_lineage	\N	de715d8c-6933-4329-81cc-1f373d667034	GL478	\N	\N	daylily	2025-07-03 08:23:17.21569-07	INSERT	\N	\N	\N	f	f
93c88f13-3829-4ed6-9452-d15a4b8d6c63	generic_instance	\N	45fef90b-f603-46b7-a07f-206b191d251e	CWX177	\N	\N	daylily	2025-07-03 08:23:17.21569-07	INSERT	\N	\N	\N	f	f
73c83cea-3727-44ea-b888-6fca4f708801	generic_instance	json_addl	45fef90b-f603-46b7-a07f-206b191d251e	CWX177	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.219404-07	UPDATE	\N	\N	\N	f	f
1cb5a156-edee-42ba-ac43-fe2f2c587eb4	generic_instance	created_dt	45fef90b-f603-46b7-a07f-206b191d251e	CWX177	2025-07-03 08:23:17.21569-07	2025-07-03T08:23:17.21569-07:00	daylily	2025-07-03 08:23:17.219404-07	UPDATE	\N	\N	\N	f	f
cf3e5ae8-3101-49ae-abfe-41bde5446c73	generic_instance	modified_dt	45fef90b-f603-46b7-a07f-206b191d251e	CWX177	2025-07-03 08:23:17.21569-07	2025-07-03T08:23:17.219404-07:00	daylily	2025-07-03 08:23:17.219404-07	UPDATE	\N	\N	\N	f	f
b8054487-960c-4092-b778-9af36b15dd31	generic_instance_lineage	\N	90a878b6-8fcd-4d61-afa8-21e1768824ae	GL479	\N	\N	daylily	2025-07-03 08:23:17.220545-07	INSERT	\N	\N	\N	f	f
2b774d0f-3f8f-4af0-bbb5-199885972ab5	generic_instance	\N	2d81e507-37d2-4131-b774-958d19466c06	CWX178	\N	\N	daylily	2025-07-03 08:23:17.220545-07	INSERT	\N	\N	\N	f	f
48bff627-6d28-4de9-8733-5bb3e143cd5d	generic_instance	json_addl	2d81e507-37d2-4131-b774-958d19466c06	CWX178	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.224535-07	UPDATE	\N	\N	\N	f	f
667f2134-d8a9-46bd-afe4-f728ff4ba2e8	generic_instance	created_dt	2d81e507-37d2-4131-b774-958d19466c06	CWX178	2025-07-03 08:23:17.220545-07	2025-07-03T08:23:17.220545-07:00	daylily	2025-07-03 08:23:17.224535-07	UPDATE	\N	\N	\N	f	f
0b64f671-949d-4576-8cb0-70c02c617882	generic_instance	modified_dt	2d81e507-37d2-4131-b774-958d19466c06	CWX178	2025-07-03 08:23:17.220545-07	2025-07-03T08:23:17.224535-07:00	daylily	2025-07-03 08:23:17.224535-07	UPDATE	\N	\N	\N	f	f
47164252-d49a-4485-87d4-64e378e4d19d	generic_instance_lineage	\N	92e8b630-74d0-461e-a534-fa5c88614b3f	GL480	\N	\N	daylily	2025-07-03 08:23:17.225911-07	INSERT	\N	\N	\N	f	f
b5a65f30-f9d7-4e57-a7db-c976b72280cb	generic_instance	\N	40c85e8c-792d-49af-930a-91f33e352dab	CWX179	\N	\N	daylily	2025-07-03 08:23:17.225911-07	INSERT	\N	\N	\N	f	f
2eaf7269-52e4-4d73-bf07-b06704efc8c3	generic_instance	json_addl	40c85e8c-792d-49af-930a-91f33e352dab	CWX179	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.230668-07	UPDATE	\N	\N	\N	f	f
9522a9c9-5349-40aa-b7fc-5ab87a047423	generic_instance	created_dt	40c85e8c-792d-49af-930a-91f33e352dab	CWX179	2025-07-03 08:23:17.225911-07	2025-07-03T08:23:17.225911-07:00	daylily	2025-07-03 08:23:17.230668-07	UPDATE	\N	\N	\N	f	f
0e2e123c-ae4e-4346-8608-a7475b612438	generic_instance	modified_dt	40c85e8c-792d-49af-930a-91f33e352dab	CWX179	2025-07-03 08:23:17.225911-07	2025-07-03T08:23:17.230668-07:00	daylily	2025-07-03 08:23:17.230668-07	UPDATE	\N	\N	\N	f	f
4bf0862c-6d9b-4fa5-9a34-f7fa80007a8d	generic_instance_lineage	\N	0832cc59-ca71-4460-b826-828898f42e1e	GL481	\N	\N	daylily	2025-07-03 08:23:17.23526-07	INSERT	\N	\N	\N	f	f
cf32aca5-e646-4636-9edc-600225e196c8	generic_instance	\N	7ed48914-f3a4-462b-baaa-845a9b3b31e6	CWX180	\N	\N	daylily	2025-07-03 08:23:17.23526-07	INSERT	\N	\N	\N	f	f
5b2280cd-a00c-4ea8-8dfa-1e4df28cacd3	generic_instance	json_addl	7ed48914-f3a4-462b-baaa-845a9b3b31e6	CWX180	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.241973-07	UPDATE	\N	\N	\N	f	f
7deabcc1-fcd1-4cbe-81ca-15ba8a8be896	generic_instance	created_dt	7ed48914-f3a4-462b-baaa-845a9b3b31e6	CWX180	2025-07-03 08:23:17.23526-07	2025-07-03T08:23:17.23526-07:00	daylily	2025-07-03 08:23:17.241973-07	UPDATE	\N	\N	\N	f	f
97980d02-602a-4ffb-9089-2a282c395d9b	generic_instance	modified_dt	7ed48914-f3a4-462b-baaa-845a9b3b31e6	CWX180	2025-07-03 08:23:17.23526-07	2025-07-03T08:23:17.241973-07:00	daylily	2025-07-03 08:23:17.241973-07	UPDATE	\N	\N	\N	f	f
b9f57578-dc61-4e1d-92ac-4918a9e9aed9	generic_instance_lineage	\N	56d1a636-c82f-41a3-92b3-afba14b16911	GL482	\N	\N	daylily	2025-07-03 08:23:17.243204-07	INSERT	\N	\N	\N	f	f
23c21c88-d911-4ae7-9721-97211e42ff00	generic_instance	\N	3802ef66-effb-4e9a-bcf9-d1ed8c84beba	CWX181	\N	\N	daylily	2025-07-03 08:23:17.243204-07	INSERT	\N	\N	\N	f	f
e006e7ea-8a1f-4f15-9ad3-72cbfe194ced	generic_instance	json_addl	3802ef66-effb-4e9a-bcf9-d1ed8c84beba	CWX181	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.247337-07	UPDATE	\N	\N	\N	f	f
690bac65-590c-4131-992e-5490d9505236	generic_instance	created_dt	3802ef66-effb-4e9a-bcf9-d1ed8c84beba	CWX181	2025-07-03 08:23:17.243204-07	2025-07-03T08:23:17.243204-07:00	daylily	2025-07-03 08:23:17.247337-07	UPDATE	\N	\N	\N	f	f
0279bb1e-fd0f-4bc6-ae79-1c6f7674173e	generic_instance	modified_dt	3802ef66-effb-4e9a-bcf9-d1ed8c84beba	CWX181	2025-07-03 08:23:17.243204-07	2025-07-03T08:23:17.247337-07:00	daylily	2025-07-03 08:23:17.247337-07	UPDATE	\N	\N	\N	f	f
c7b05aa9-2c7e-4441-9a5f-a417efcdfbca	generic_instance_lineage	\N	b446d6cb-f7f3-49e1-9650-7b6b0519af6d	GL483	\N	\N	daylily	2025-07-03 08:23:17.249395-07	INSERT	\N	\N	\N	f	f
3e8ebd85-ef89-44ae-88a5-32ccfae3455d	generic_instance	\N	a9cc088c-0aec-4d5f-a5ab-fae4fe7f030f	CWX182	\N	\N	daylily	2025-07-03 08:23:17.249395-07	INSERT	\N	\N	\N	f	f
fc3de4ef-517d-4e72-9cbd-fd2a2d2c17a5	generic_instance	json_addl	a9cc088c-0aec-4d5f-a5ab-fae4fe7f030f	CWX182	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.254516-07	UPDATE	\N	\N	\N	f	f
c601a785-d82b-4852-a571-d85a9c310fa6	generic_instance	created_dt	a9cc088c-0aec-4d5f-a5ab-fae4fe7f030f	CWX182	2025-07-03 08:23:17.249395-07	2025-07-03T08:23:17.249395-07:00	daylily	2025-07-03 08:23:17.254516-07	UPDATE	\N	\N	\N	f	f
82fb770a-8e18-4291-be53-19d336583ba3	generic_instance	modified_dt	a9cc088c-0aec-4d5f-a5ab-fae4fe7f030f	CWX182	2025-07-03 08:23:17.249395-07	2025-07-03T08:23:17.254516-07:00	daylily	2025-07-03 08:23:17.254516-07	UPDATE	\N	\N	\N	f	f
16493119-63d5-4dfa-abec-9f22f34fdc44	generic_instance_lineage	\N	711ed583-eefa-4f39-8784-645d104d9648	GL484	\N	\N	daylily	2025-07-03 08:23:17.255731-07	INSERT	\N	\N	\N	f	f
72bbbe33-2ba7-4ee7-bd0f-3f7f2fd89ccd	generic_instance	\N	fc480673-f73e-44cb-b6d4-64df95e6b3d0	CWX183	\N	\N	daylily	2025-07-03 08:23:17.255731-07	INSERT	\N	\N	\N	f	f
5c9da1d6-c5d3-4b00-a6b4-b3a09b932bed	generic_instance	json_addl	fc480673-f73e-44cb-b6d4-64df95e6b3d0	CWX183	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.260046-07	UPDATE	\N	\N	\N	f	f
e44a93fc-3531-48e5-af5b-79f374ac3ab4	generic_instance	created_dt	fc480673-f73e-44cb-b6d4-64df95e6b3d0	CWX183	2025-07-03 08:23:17.255731-07	2025-07-03T08:23:17.255731-07:00	daylily	2025-07-03 08:23:17.260046-07	UPDATE	\N	\N	\N	f	f
f366e819-b3c3-45aa-b1bb-33f41399e053	generic_instance	modified_dt	fc480673-f73e-44cb-b6d4-64df95e6b3d0	CWX183	2025-07-03 08:23:17.255731-07	2025-07-03T08:23:17.260046-07:00	daylily	2025-07-03 08:23:17.260046-07	UPDATE	\N	\N	\N	f	f
2703a18b-9b1f-4ba7-9d81-2b10570dc598	generic_instance_lineage	\N	e86af435-51d6-46ff-aeb2-36e10bbee947	GL485	\N	\N	daylily	2025-07-03 08:23:17.261154-07	INSERT	\N	\N	\N	f	f
4bcef29e-c001-4c3f-bbe8-e12f530d2029	generic_instance	\N	8744e5ca-a013-46fc-99ac-b67f436d0827	CWX184	\N	\N	daylily	2025-07-03 08:23:17.261154-07	INSERT	\N	\N	\N	f	f
1327ade1-01a3-44a9-af94-bb0bd73d6f4a	generic_instance	json_addl	8744e5ca-a013-46fc-99ac-b67f436d0827	CWX184	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.265857-07	UPDATE	\N	\N	\N	f	f
d868397c-0c12-452e-90f8-47615ceb8ed1	generic_instance	created_dt	8744e5ca-a013-46fc-99ac-b67f436d0827	CWX184	2025-07-03 08:23:17.261154-07	2025-07-03T08:23:17.261154-07:00	daylily	2025-07-03 08:23:17.265857-07	UPDATE	\N	\N	\N	f	f
8c1da6b0-2295-4e06-b13b-f52229720ca5	generic_instance	modified_dt	8744e5ca-a013-46fc-99ac-b67f436d0827	CWX184	2025-07-03 08:23:17.261154-07	2025-07-03T08:23:17.265857-07:00	daylily	2025-07-03 08:23:17.265857-07	UPDATE	\N	\N	\N	f	f
9d5e7b84-adf8-421f-b8ad-d700e4610cc2	generic_instance_lineage	\N	9b86fafd-0e91-41b3-9958-2e341240e13d	GL486	\N	\N	daylily	2025-07-03 08:23:17.267378-07	INSERT	\N	\N	\N	f	f
3e032042-106a-4d74-89f7-bc47a171a78d	generic_instance	\N	1cf84a4a-f071-46fa-ae06-d39f5f8aa6df	CWX185	\N	\N	daylily	2025-07-03 08:23:17.267378-07	INSERT	\N	\N	\N	f	f
4bdc0b5f-cf7d-48d8-bf8b-c5ab5bb1f76d	generic_instance	json_addl	1cf84a4a-f071-46fa-ae06-d39f5f8aa6df	CWX185	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.271569-07	UPDATE	\N	\N	\N	f	f
6ad9379a-210e-40c6-9e18-ab9a840ce823	generic_instance	created_dt	1cf84a4a-f071-46fa-ae06-d39f5f8aa6df	CWX185	2025-07-03 08:23:17.267378-07	2025-07-03T08:23:17.267378-07:00	daylily	2025-07-03 08:23:17.271569-07	UPDATE	\N	\N	\N	f	f
4729534f-1a7d-4fc6-bd15-4e6cb454c112	generic_instance	modified_dt	1cf84a4a-f071-46fa-ae06-d39f5f8aa6df	CWX185	2025-07-03 08:23:17.267378-07	2025-07-03T08:23:17.271569-07:00	daylily	2025-07-03 08:23:17.271569-07	UPDATE	\N	\N	\N	f	f
da667d3b-eea1-4e76-a03a-e9805f722aa6	generic_instance_lineage	\N	ce43688c-c273-4f42-ab5a-38e2f0be6274	GL487	\N	\N	daylily	2025-07-03 08:23:17.272936-07	INSERT	\N	\N	\N	f	f
88ca3d2e-5762-4d8e-94f8-4897f4118663	generic_instance	\N	fb64119b-a9bf-4920-ac9f-42025d6486cd	CWX186	\N	\N	daylily	2025-07-03 08:23:17.272936-07	INSERT	\N	\N	\N	f	f
51d20dbb-cfd0-48ab-b27d-59a8f5f87b57	generic_instance	json_addl	fb64119b-a9bf-4920-ac9f-42025d6486cd	CWX186	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.277361-07	UPDATE	\N	\N	\N	f	f
950ba16a-7f8a-4693-9f2c-a2aeea007c5c	generic_instance	created_dt	fb64119b-a9bf-4920-ac9f-42025d6486cd	CWX186	2025-07-03 08:23:17.272936-07	2025-07-03T08:23:17.272936-07:00	daylily	2025-07-03 08:23:17.277361-07	UPDATE	\N	\N	\N	f	f
4cbf0655-bc97-4380-81e2-2fd91ec9db65	generic_instance	modified_dt	fb64119b-a9bf-4920-ac9f-42025d6486cd	CWX186	2025-07-03 08:23:17.272936-07	2025-07-03T08:23:17.277361-07:00	daylily	2025-07-03 08:23:17.277361-07	UPDATE	\N	\N	\N	f	f
8486294a-dd83-49ef-9e60-4e4fb3d89cf7	generic_instance_lineage	\N	eb343497-2a96-4f4e-82dd-f15e3c19ed8a	GL488	\N	\N	daylily	2025-07-03 08:23:17.278636-07	INSERT	\N	\N	\N	f	f
69f11e15-7bfe-4086-b63e-e629530c5d44	generic_instance	\N	647cfc65-654d-4ff4-98b5-737645b3e930	CWX187	\N	\N	daylily	2025-07-03 08:23:17.278636-07	INSERT	\N	\N	\N	f	f
28c29bbe-559b-4f5c-9ce0-353f87b7a1a6	generic_instance	json_addl	647cfc65-654d-4ff4-98b5-737645b3e930	CWX187	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.28277-07	UPDATE	\N	\N	\N	f	f
4a639775-1676-4b82-896e-b16e7edbc97f	generic_instance	created_dt	647cfc65-654d-4ff4-98b5-737645b3e930	CWX187	2025-07-03 08:23:17.278636-07	2025-07-03T08:23:17.278636-07:00	daylily	2025-07-03 08:23:17.28277-07	UPDATE	\N	\N	\N	f	f
480e9bbd-ad0b-475c-8ec9-365b74c14cca	generic_instance	modified_dt	647cfc65-654d-4ff4-98b5-737645b3e930	CWX187	2025-07-03 08:23:17.278636-07	2025-07-03T08:23:17.28277-07:00	daylily	2025-07-03 08:23:17.28277-07	UPDATE	\N	\N	\N	f	f
5667486d-e8aa-4114-80fa-1a45f4cef925	generic_instance_lineage	\N	754681d9-0ece-4b7f-bf4d-e8dc2d256a6d	GL489	\N	\N	daylily	2025-07-03 08:23:17.28404-07	INSERT	\N	\N	\N	f	f
0a289d72-b2db-4ebd-b98b-a00e6007d3fe	generic_instance	\N	73d57912-9cb8-44c9-88a1-34e13b8b65fd	CWX188	\N	\N	daylily	2025-07-03 08:23:17.28404-07	INSERT	\N	\N	\N	f	f
f2b62849-32ad-4a76-bb0e-b9f21bad49d1	generic_instance	json_addl	73d57912-9cb8-44c9-88a1-34e13b8b65fd	CWX188	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.288198-07	UPDATE	\N	\N	\N	f	f
c9b50232-e987-4402-a2b1-9108957e4a53	generic_instance	created_dt	73d57912-9cb8-44c9-88a1-34e13b8b65fd	CWX188	2025-07-03 08:23:17.28404-07	2025-07-03T08:23:17.28404-07:00	daylily	2025-07-03 08:23:17.288198-07	UPDATE	\N	\N	\N	f	f
1917a4e8-8b96-49f9-9941-5717dfe94176	generic_instance	modified_dt	73d57912-9cb8-44c9-88a1-34e13b8b65fd	CWX188	2025-07-03 08:23:17.28404-07	2025-07-03T08:23:17.288198-07:00	daylily	2025-07-03 08:23:17.288198-07	UPDATE	\N	\N	\N	f	f
a9ebe926-0e4d-4df1-b62e-6bedc13e5489	generic_instance_lineage	\N	1a714a26-b5da-442a-94bb-a4a2bb5dc5ea	GL490	\N	\N	daylily	2025-07-03 08:23:17.289374-07	INSERT	\N	\N	\N	f	f
b278d00d-a014-4eed-bb00-7693ef041d01	generic_instance	\N	8a96e5af-aa96-4369-9f19-a004222d92bf	CWX189	\N	\N	daylily	2025-07-03 08:23:17.289374-07	INSERT	\N	\N	\N	f	f
e1a1cb21-5fa3-4b81-9cad-933171418363	generic_instance	json_addl	8a96e5af-aa96-4369-9f19-a004222d92bf	CWX189	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.29369-07	UPDATE	\N	\N	\N	f	f
876b44a1-0052-4828-9cd4-4e4c47852020	generic_instance	created_dt	8a96e5af-aa96-4369-9f19-a004222d92bf	CWX189	2025-07-03 08:23:17.289374-07	2025-07-03T08:23:17.289374-07:00	daylily	2025-07-03 08:23:17.29369-07	UPDATE	\N	\N	\N	f	f
a1180012-0e02-4e1e-b9af-4c040da2eb0d	generic_instance	modified_dt	8a96e5af-aa96-4369-9f19-a004222d92bf	CWX189	2025-07-03 08:23:17.289374-07	2025-07-03T08:23:17.29369-07:00	daylily	2025-07-03 08:23:17.29369-07	UPDATE	\N	\N	\N	f	f
5455aece-d1e7-43a4-8087-c80574db6b56	generic_instance_lineage	\N	3c7a4747-1927-42f2-8229-775b1d5a68ff	GL491	\N	\N	daylily	2025-07-03 08:23:17.29493-07	INSERT	\N	\N	\N	f	f
d4799549-bf57-4198-92af-f1c2dcacc691	generic_instance	\N	dd25a9e1-d8dc-4eea-be62-2d3d6a4a7829	CWX190	\N	\N	daylily	2025-07-03 08:23:17.29493-07	INSERT	\N	\N	\N	f	f
0afffbdf-0b68-4ddd-989f-2e1e4c63a79b	generic_instance	json_addl	dd25a9e1-d8dc-4eea-be62-2d3d6a4a7829	CWX190	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.2998-07	UPDATE	\N	\N	\N	f	f
6cb7985b-b92e-4e1f-a6dc-acd0e9155fc2	generic_instance	created_dt	dd25a9e1-d8dc-4eea-be62-2d3d6a4a7829	CWX190	2025-07-03 08:23:17.29493-07	2025-07-03T08:23:17.29493-07:00	daylily	2025-07-03 08:23:17.2998-07	UPDATE	\N	\N	\N	f	f
e593228d-b63d-46fb-8d77-30a44e891187	generic_instance	modified_dt	dd25a9e1-d8dc-4eea-be62-2d3d6a4a7829	CWX190	2025-07-03 08:23:17.29493-07	2025-07-03T08:23:17.2998-07:00	daylily	2025-07-03 08:23:17.2998-07	UPDATE	\N	\N	\N	f	f
3dfc3f2f-ba32-439f-8626-bae3861108c2	generic_instance_lineage	\N	b7eb8ee6-4cba-41c8-885f-1470368a38b3	GL492	\N	\N	daylily	2025-07-03 08:23:17.301168-07	INSERT	\N	\N	\N	f	f
bca29419-243b-4e07-b4be-1ff36d419201	generic_instance	\N	23e5c523-46c3-4f9b-a0ca-e30781d73d3c	CWX191	\N	\N	daylily	2025-07-03 08:23:17.301168-07	INSERT	\N	\N	\N	f	f
1f50bbf9-692e-4755-aa6a-594da1117b9d	generic_instance	json_addl	23e5c523-46c3-4f9b-a0ca-e30781d73d3c	CWX191	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.305344-07	UPDATE	\N	\N	\N	f	f
c74b93bc-da57-4bd8-a3ca-235fbb371e65	generic_instance	created_dt	23e5c523-46c3-4f9b-a0ca-e30781d73d3c	CWX191	2025-07-03 08:23:17.301168-07	2025-07-03T08:23:17.301168-07:00	daylily	2025-07-03 08:23:17.305344-07	UPDATE	\N	\N	\N	f	f
e2c4fe46-2ed2-4d8c-bd9f-79cdbfebac26	generic_instance	modified_dt	23e5c523-46c3-4f9b-a0ca-e30781d73d3c	CWX191	2025-07-03 08:23:17.301168-07	2025-07-03T08:23:17.305344-07:00	daylily	2025-07-03 08:23:17.305344-07	UPDATE	\N	\N	\N	f	f
38e10231-a606-4bac-b284-b5562d64064f	generic_instance_lineage	\N	fc077a22-296f-4e69-a3a4-6d40439d18ea	GL493	\N	\N	daylily	2025-07-03 08:23:17.306639-07	INSERT	\N	\N	\N	f	f
c64c45b4-c2eb-42e1-a625-606572b689fc	generic_instance	\N	6d930317-d930-41e8-95c6-dccd9add54ff	CWX192	\N	\N	daylily	2025-07-03 08:23:17.306639-07	INSERT	\N	\N	\N	f	f
735845af-9814-4164-b996-98a15df57921	generic_instance	json_addl	6d930317-d930-41e8-95c6-dccd9add54ff	CWX192	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.310939-07	UPDATE	\N	\N	\N	f	f
0f69702b-e801-444f-80d7-9debcd5696c1	generic_instance	created_dt	6d930317-d930-41e8-95c6-dccd9add54ff	CWX192	2025-07-03 08:23:17.306639-07	2025-07-03T08:23:17.306639-07:00	daylily	2025-07-03 08:23:17.310939-07	UPDATE	\N	\N	\N	f	f
fe28812f-ab37-4960-b6df-6c0c48d81f6b	generic_instance	modified_dt	6d930317-d930-41e8-95c6-dccd9add54ff	CWX192	2025-07-03 08:23:17.306639-07	2025-07-03T08:23:17.310939-07:00	daylily	2025-07-03 08:23:17.310939-07	UPDATE	\N	\N	\N	f	f
784cc240-2723-4c91-a14c-f5111515ff94	generic_instance_lineage	\N	d70e0389-0c1b-4a2a-92c4-1a5dc0511299	GL494	\N	\N	daylily	2025-07-03 08:23:17.312181-07	INSERT	\N	\N	\N	f	f
0f39db06-1785-44a5-931a-42f9c1f36367	generic_instance	\N	4ff0926c-88e1-40fd-90ac-dafe861eb746	CWX193	\N	\N	daylily	2025-07-03 08:23:17.312181-07	INSERT	\N	\N	\N	f	f
2080cea8-7919-4e10-a04b-95a18e978f7e	generic_instance	json_addl	4ff0926c-88e1-40fd-90ac-dafe861eb746	CWX193	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.316077-07	UPDATE	\N	\N	\N	f	f
f89ac454-357a-46d1-8c11-f74c2fb4928d	generic_instance	created_dt	4ff0926c-88e1-40fd-90ac-dafe861eb746	CWX193	2025-07-03 08:23:17.312181-07	2025-07-03T08:23:17.312181-07:00	daylily	2025-07-03 08:23:17.316077-07	UPDATE	\N	\N	\N	f	f
2faa3764-8f93-4031-8d01-9402bc8927d0	generic_instance	modified_dt	4ff0926c-88e1-40fd-90ac-dafe861eb746	CWX193	2025-07-03 08:23:17.312181-07	2025-07-03T08:23:17.316077-07:00	daylily	2025-07-03 08:23:17.316077-07	UPDATE	\N	\N	\N	f	f
0d23e9b0-1c88-4593-826b-92bc917496a2	generic_instance_lineage	\N	f1c9dfb8-17d8-43a0-9d0f-7cafa83b3f4c	GL495	\N	\N	daylily	2025-07-03 08:23:17.317516-07	INSERT	\N	\N	\N	f	f
6f72eed4-b8c8-453e-801e-a370f8062d52	generic_instance	\N	a72a0a3c-2e05-4aa3-bb8f-1d24163a15fa	CWX194	\N	\N	daylily	2025-07-03 08:23:17.317516-07	INSERT	\N	\N	\N	f	f
75397d0e-4851-4ca2-91d0-0e556495bb49	generic_instance	json_addl	a72a0a3c-2e05-4aa3-bb8f-1d24163a15fa	CWX194	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.322009-07	UPDATE	\N	\N	\N	f	f
d933851c-77f0-4562-b48a-0dbfbadb92ca	generic_instance	created_dt	a72a0a3c-2e05-4aa3-bb8f-1d24163a15fa	CWX194	2025-07-03 08:23:17.317516-07	2025-07-03T08:23:17.317516-07:00	daylily	2025-07-03 08:23:17.322009-07	UPDATE	\N	\N	\N	f	f
9ebc51d9-62e7-4840-a705-e27aedc4d3af	generic_instance	modified_dt	a72a0a3c-2e05-4aa3-bb8f-1d24163a15fa	CWX194	2025-07-03 08:23:17.317516-07	2025-07-03T08:23:17.322009-07:00	daylily	2025-07-03 08:23:17.322009-07	UPDATE	\N	\N	\N	f	f
0541c36e-056f-410a-abd6-a7b270e9e8ed	generic_instance_lineage	\N	478f219f-c250-4739-aa4f-89be44446078	GL496	\N	\N	daylily	2025-07-03 08:23:17.323224-07	INSERT	\N	\N	\N	f	f
994f83f8-b074-48eb-9439-caf8507d3684	generic_instance	\N	ebd9b442-c35d-4c7f-a9d3-514321045d51	CWX195	\N	\N	daylily	2025-07-03 08:23:17.323224-07	INSERT	\N	\N	\N	f	f
018b3549-8970-46d1-afeb-8c78c54a1a69	generic_instance	json_addl	ebd9b442-c35d-4c7f-a9d3-514321045d51	CWX195	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.327652-07	UPDATE	\N	\N	\N	f	f
73407508-410b-4354-9b10-5ed87acdaaa1	generic_instance	created_dt	ebd9b442-c35d-4c7f-a9d3-514321045d51	CWX195	2025-07-03 08:23:17.323224-07	2025-07-03T08:23:17.323224-07:00	daylily	2025-07-03 08:23:17.327652-07	UPDATE	\N	\N	\N	f	f
3ffdf0da-30a1-4cfa-a072-2dde92716701	generic_instance	modified_dt	ebd9b442-c35d-4c7f-a9d3-514321045d51	CWX195	2025-07-03 08:23:17.323224-07	2025-07-03T08:23:17.327652-07:00	daylily	2025-07-03 08:23:17.327652-07	UPDATE	\N	\N	\N	f	f
0bbd18f0-f28e-4ba9-81af-cf64154e1c72	generic_instance_lineage	\N	2adb0e7f-f3b2-4acc-a2d8-f99eea0cf1f9	GL497	\N	\N	daylily	2025-07-03 08:23:17.328821-07	INSERT	\N	\N	\N	f	f
92d5cb62-d374-4478-8000-7ccc6121c58a	generic_instance	\N	116a5243-bb5e-4e4c-b293-c07586d1d79e	CWX196	\N	\N	daylily	2025-07-03 08:23:17.328821-07	INSERT	\N	\N	\N	f	f
83f5be73-f433-47c4-940b-987dab800e5a	generic_instance	json_addl	116a5243-bb5e-4e4c-b293-c07586d1d79e	CWX196	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.334892-07	UPDATE	\N	\N	\N	f	f
ed264b5b-8f83-446a-b66f-02d6f7f77ad7	generic_instance	created_dt	116a5243-bb5e-4e4c-b293-c07586d1d79e	CWX196	2025-07-03 08:23:17.328821-07	2025-07-03T08:23:17.328821-07:00	daylily	2025-07-03 08:23:17.334892-07	UPDATE	\N	\N	\N	f	f
3f937414-980a-44ca-b6e0-f5305863292f	generic_instance	modified_dt	116a5243-bb5e-4e4c-b293-c07586d1d79e	CWX196	2025-07-03 08:23:17.328821-07	2025-07-03T08:23:17.334892-07:00	daylily	2025-07-03 08:23:17.334892-07	UPDATE	\N	\N	\N	f	f
95619675-eeb1-403a-8038-02a9d00a5b71	generic_instance_lineage	\N	57a29989-1636-4134-8a6d-0dddf3c0745a	GL498	\N	\N	daylily	2025-07-03 08:23:17.336917-07	INSERT	\N	\N	\N	f	f
f46d3b1f-5a15-401d-82b5-e016589e5530	generic_instance	\N	a7fdfb14-fc01-4349-a8ad-8de611a15b1c	MRX169	\N	\N	daylily	2025-07-03 08:23:17.34078-07	INSERT	\N	\N	\N	f	f
ca073c59-2e39-499a-8868-515d2f7daeb8	generic_instance_lineage	\N	d9c50ec5-f17c-4f1d-9bd5-b2598006394b	GL499	\N	\N	daylily	2025-07-03 08:23:17.342965-07	INSERT	\N	\N	\N	f	f
94677142-a8dd-4ea9-8dac-1f055c1f2c3b	generic_instance	\N	8205e9ec-903c-4f91-bc3b-bbdb81479279	MRX170	\N	\N	daylily	2025-07-03 08:23:17.342965-07	INSERT	\N	\N	\N	f	f
fe443192-707e-47c6-bc21-e1578bcd17d5	generic_instance_lineage	\N	208fabab-e2e4-4232-b8cb-99afe8c25b9f	GL500	\N	\N	daylily	2025-07-03 08:23:17.347665-07	INSERT	\N	\N	\N	f	f
4cd0c6a2-2811-464f-8a14-70671f675c2e	generic_instance	\N	4e28eeb7-6056-407f-92d0-53a349a90407	MRX171	\N	\N	daylily	2025-07-03 08:23:17.347665-07	INSERT	\N	\N	\N	f	f
07e5f24f-731d-4261-9a38-9c1a913a2388	generic_instance_lineage	\N	91603704-a959-40ac-9fb5-afe7e62a93e0	GL501	\N	\N	daylily	2025-07-03 08:23:17.354348-07	INSERT	\N	\N	\N	f	f
dd3809d0-f75d-4766-ae32-09f18bf381c9	generic_instance	\N	7a771f0c-8c4e-4184-91f0-be5b15eb4a1d	MRX172	\N	\N	daylily	2025-07-03 08:23:17.354348-07	INSERT	\N	\N	\N	f	f
06f62793-bdc0-4b7a-a5c0-6f777ee1f283	generic_instance_lineage	\N	225dfd3a-ab52-453e-b228-f639f4607292	GL502	\N	\N	daylily	2025-07-03 08:23:17.359712-07	INSERT	\N	\N	\N	f	f
2adbe4ae-63ab-4df0-a1f5-474a3764b367	generic_instance	\N	b348c7b2-83c2-4c6e-ac63-25d711fed70a	MRX173	\N	\N	daylily	2025-07-03 08:23:17.359712-07	INSERT	\N	\N	\N	f	f
262e73d2-5685-41d5-885e-7608b2fce334	generic_instance_lineage	\N	e7d98d83-461d-46e1-b18b-860a4fa58d1a	GL503	\N	\N	daylily	2025-07-03 08:23:17.364937-07	INSERT	\N	\N	\N	f	f
2d2ca564-7c1b-4688-8cbf-61a8e5be0241	generic_instance	\N	ad0b6013-0a58-4c0d-a2a2-acf5cf2fcf0b	MRX174	\N	\N	daylily	2025-07-03 08:23:17.364937-07	INSERT	\N	\N	\N	f	f
6c44203f-4b64-4f92-8915-a8da90d0471b	generic_instance_lineage	\N	dd3a1113-92e0-4264-a566-0ac6be54223d	GL504	\N	\N	daylily	2025-07-03 08:23:17.36972-07	INSERT	\N	\N	\N	f	f
1b4826a5-9cd3-4c6a-a9cd-c8ff847baa66	generic_instance	\N	269038aa-810c-4401-9ae5-b2474d3afe38	MRX175	\N	\N	daylily	2025-07-03 08:23:17.36972-07	INSERT	\N	\N	\N	f	f
09e8b5b7-4eec-4e38-bde5-74c3fc9172a7	generic_instance_lineage	\N	748d5cb4-3405-421e-9e56-fdafcb8228f6	GL505	\N	\N	daylily	2025-07-03 08:23:17.374602-07	INSERT	\N	\N	\N	f	f
3ec3f84f-640c-4179-bf64-774064109b6a	generic_instance	\N	538a6afc-a419-46c2-b3b1-79ac907c1b6f	MRX176	\N	\N	daylily	2025-07-03 08:23:17.374602-07	INSERT	\N	\N	\N	f	f
9f2c71b2-38fb-4a9e-8a48-c737281c69b2	generic_instance_lineage	\N	d26d60f3-7c03-424b-ad7a-d39921a63425	GL506	\N	\N	daylily	2025-07-03 08:23:17.379232-07	INSERT	\N	\N	\N	f	f
dc1c9d7e-dcff-4f01-abb2-af5bfde28de5	generic_instance	\N	376ecd05-c2eb-4b37-b19b-948ef80045b4	MRX177	\N	\N	daylily	2025-07-03 08:23:17.379232-07	INSERT	\N	\N	\N	f	f
5f29e6ad-fed1-4202-8f3e-5eb0bd477181	generic_instance_lineage	\N	c449c4f7-937f-4534-9a18-98ad969e82cf	GL507	\N	\N	daylily	2025-07-03 08:23:17.383898-07	INSERT	\N	\N	\N	f	f
cea5ad0a-2996-4d77-9387-9bb9e915a95a	generic_instance	\N	6f2d6e6f-012e-46fc-82d9-8323bc889a37	MRX178	\N	\N	daylily	2025-07-03 08:23:17.383898-07	INSERT	\N	\N	\N	f	f
408526d5-366d-432e-a5e1-12b60e817496	generic_instance_lineage	\N	7a7224e8-222a-42a7-83cf-c40ee155501c	GL508	\N	\N	daylily	2025-07-03 08:23:17.388733-07	INSERT	\N	\N	\N	f	f
977c566c-0d02-4752-9e76-f3427e3aebac	generic_instance	\N	009eef87-568f-4b42-8aff-4d78a5c238af	MRX179	\N	\N	daylily	2025-07-03 08:23:17.388733-07	INSERT	\N	\N	\N	f	f
bcd9b013-0805-4886-a0bb-dddfd98ae783	generic_instance_lineage	\N	84e4b868-c7d5-46fa-ac1c-9948f2b25016	GL509	\N	\N	daylily	2025-07-03 08:23:17.393467-07	INSERT	\N	\N	\N	f	f
dafdba1b-b663-4acb-83df-43e5760473e5	generic_instance	\N	b2c3158a-af0b-43d2-a6fe-6b005335a909	MRX180	\N	\N	daylily	2025-07-03 08:23:17.393467-07	INSERT	\N	\N	\N	f	f
fa68f0d7-abdc-419e-9e0e-42414e6fda7d	generic_instance_lineage	\N	c9f43a9a-337e-4d13-ac73-8706fcb15a20	GL510	\N	\N	daylily	2025-07-03 08:23:17.398828-07	INSERT	\N	\N	\N	f	f
f5b32332-4bec-4df0-8a85-189da64eb714	generic_instance	\N	7ec4cacc-52bd-4a88-bfcd-3167baeab3a2	MRX181	\N	\N	daylily	2025-07-03 08:23:17.398828-07	INSERT	\N	\N	\N	f	f
396ac278-38af-486c-ab29-6243325eb0c6	generic_instance_lineage	\N	d93f0d46-960f-42f5-874c-dcd85cec0337	GL511	\N	\N	daylily	2025-07-03 08:23:17.404929-07	INSERT	\N	\N	\N	f	f
704159ba-2912-4030-9dc2-3279711fb159	generic_instance	\N	bda1c743-e436-4a9c-a9fc-71f22cda28c1	MRX182	\N	\N	daylily	2025-07-03 08:23:17.404929-07	INSERT	\N	\N	\N	f	f
5ee20aa7-0713-416d-a62b-4b0c8c35957a	generic_instance_lineage	\N	fae8a86e-5587-403c-9648-8f6826596e9b	GL512	\N	\N	daylily	2025-07-03 08:23:17.409505-07	INSERT	\N	\N	\N	f	f
fd32648f-5a5a-4da6-8481-489ec4aac800	generic_instance	\N	c17a9260-b36f-4a62-b076-e13de08d64cb	MRX183	\N	\N	daylily	2025-07-03 08:23:17.409505-07	INSERT	\N	\N	\N	f	f
87818157-f618-4170-91cb-b7b39e5298ee	generic_instance_lineage	\N	d005a268-21e0-47cf-bbc4-8694d07be3b7	GL513	\N	\N	daylily	2025-07-03 08:23:17.416652-07	INSERT	\N	\N	\N	f	f
3aa2129a-a498-4d37-b000-8f910a53372c	generic_instance	\N	498ad847-9c74-4b5a-8725-0a56a3c84f94	MRX184	\N	\N	daylily	2025-07-03 08:23:17.416652-07	INSERT	\N	\N	\N	f	f
cf801e5a-753b-45f7-8c2e-8382ebd7ca9b	generic_instance_lineage	\N	761d0c36-8f03-491e-8f44-0bd503b7cc0e	GL514	\N	\N	daylily	2025-07-03 08:23:17.422874-07	INSERT	\N	\N	\N	f	f
d06c3763-a1e9-4b2c-8d08-1d6a45946e23	generic_instance	\N	9abbbeeb-2a9f-4a05-95f5-070fdc8c3483	MRX185	\N	\N	daylily	2025-07-03 08:23:17.422874-07	INSERT	\N	\N	\N	f	f
46e20899-60d6-44a3-8af9-42c96e5894a7	generic_instance_lineage	\N	2c081129-7e27-4ab7-8d41-2d6e0227837e	GL515	\N	\N	daylily	2025-07-03 08:23:17.427323-07	INSERT	\N	\N	\N	f	f
d6802bb4-9ef8-407e-8169-9bfe0e9811f3	generic_instance	\N	58b3a670-dedc-4045-b70d-d90479915351	MRX186	\N	\N	daylily	2025-07-03 08:23:17.427323-07	INSERT	\N	\N	\N	f	f
bd38cfa0-a513-4af9-9212-75da5e8beaa2	generic_instance_lineage	\N	400b16bf-edce-406b-bc9b-7ffab4cfe2c2	GL516	\N	\N	daylily	2025-07-03 08:23:17.431746-07	INSERT	\N	\N	\N	f	f
eba36536-0eb4-4d79-a0f6-ca15d38d544a	generic_instance	\N	ef8398cc-3507-42a0-b8f5-99c0a9d033da	MRX187	\N	\N	daylily	2025-07-03 08:23:17.431746-07	INSERT	\N	\N	\N	f	f
01c76af1-1fe7-4fe7-a6c4-1e6b05abae07	generic_instance_lineage	\N	ebfaecf1-8a8c-4b7e-b55f-c680c68c5949	GL517	\N	\N	daylily	2025-07-03 08:23:17.436354-07	INSERT	\N	\N	\N	f	f
72a7f3eb-e5d7-4915-a2cd-66997998f436	generic_instance	\N	04853cf5-aee7-4abf-9240-600265068cb2	MRX188	\N	\N	daylily	2025-07-03 08:23:17.436354-07	INSERT	\N	\N	\N	f	f
73838c0b-83c1-41d4-8e3b-fbe9f5996560	generic_instance_lineage	\N	a655b533-585a-4ddf-9581-6512e81ae166	GL518	\N	\N	daylily	2025-07-03 08:23:17.440646-07	INSERT	\N	\N	\N	f	f
5fc18741-2f63-41f6-8210-07f7a92a35fc	generic_instance	\N	9401369a-7357-4e36-b527-2d208017597d	MRX189	\N	\N	daylily	2025-07-03 08:23:17.440646-07	INSERT	\N	\N	\N	f	f
980ec996-68b1-468e-a24a-29cd528373cf	generic_instance_lineage	\N	0b8f0035-8c7a-4c51-8eb4-08d77e18103b	GL519	\N	\N	daylily	2025-07-03 08:23:17.445219-07	INSERT	\N	\N	\N	f	f
07c57151-bb2f-4b08-ab6e-b3c928ee0fdb	generic_instance	\N	e6868a3f-63f7-4302-be5e-70a1bc5ec46e	MRX190	\N	\N	daylily	2025-07-03 08:23:17.445219-07	INSERT	\N	\N	\N	f	f
8d3f8ee6-2c1e-4ac6-810c-a28266f54b46	generic_instance_lineage	\N	09a206d5-d266-4242-8f79-c83cc5366fe7	GL520	\N	\N	daylily	2025-07-03 08:23:17.449624-07	INSERT	\N	\N	\N	f	f
2ebff752-c3ef-46c6-86f0-6303dfaa9955	generic_instance	\N	8269b692-aa3c-45f6-9b49-415646e12639	MRX191	\N	\N	daylily	2025-07-03 08:23:17.449624-07	INSERT	\N	\N	\N	f	f
df065526-ff09-4232-a754-e0d8ded9c9d4	generic_instance	json_addl	fe1b21ed-3957-4df1-b13c-fcaccc035a83	CX14	{"cogs": {"cost": "1.70", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "1.5ml eppi tube", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "1.5ml eppi tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "1.70", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "NTC - water", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "1.5ml eppi tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.514422-07	UPDATE	\N	\N	\N	f	f
a5763c9b-41aa-4467-b9cc-fb99ad816930	generic_instance	created_dt	fe1b21ed-3957-4df1-b13c-fcaccc035a83	CX14	2025-07-03 08:23:17.509268-07	2025-07-03T08:23:17.509268-07:00	daylily	2025-07-03 08:23:17.514422-07	UPDATE	\N	\N	\N	f	f
1497e4d8-4cc7-49de-ba3d-d0e7351c5aa3	generic_instance	modified_dt	fe1b21ed-3957-4df1-b13c-fcaccc035a83	CX14	2025-07-03 08:23:17.509268-07	2025-07-03T08:23:17.514422-07:00	daylily	2025-07-03 08:23:17.514422-07	UPDATE	\N	\N	\N	f	f
adb4ad58-34d3-4682-8bf2-811943587737	generic_instance_lineage	\N	ef0f706f-574e-4646-8556-90f289c2653e	GL529	\N	\N	daylily	2025-07-03 08:23:17.514422-07	INSERT	\N	\N	\N	f	f
bc66a6a8-71a0-4807-8daf-3088a99ebbd5	generic_instance	\N	43b72609-8eab-47a3-9cd2-fc54e4092892	CX15	\N	\N	daylily	2025-07-03 08:23:17.5205-07	INSERT	\N	\N	\N	f	f
0672de88-e835-4f1c-a277-3e3505013415	generic_instance	\N	67da16e4-6e1d-451f-9319-d3b627ef5a3e	CWX197	\N	\N	daylily	2025-07-03 08:23:17.525364-07	INSERT	\N	\N	\N	f	f
1ce3d0cc-bae3-44a6-afdc-c016b64226f7	generic_instance	json_addl	67da16e4-6e1d-451f-9319-d3b627ef5a3e	CWX197	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.530954-07	UPDATE	\N	\N	\N	f	f
1208238f-d91e-4f5f-aba3-1d593772b7a7	generic_instance	created_dt	67da16e4-6e1d-451f-9319-d3b627ef5a3e	CWX197	2025-07-03 08:23:17.525364-07	2025-07-03T08:23:17.525364-07:00	daylily	2025-07-03 08:23:17.530954-07	UPDATE	\N	\N	\N	f	f
1af20bce-0eba-42bf-8a09-d8a97f540c27	generic_instance	modified_dt	67da16e4-6e1d-451f-9319-d3b627ef5a3e	CWX197	2025-07-03 08:23:17.525364-07	2025-07-03T08:23:17.530954-07:00	daylily	2025-07-03 08:23:17.530954-07	UPDATE	\N	\N	\N	f	f
3166edba-7098-4a0b-a495-40b0874f7ba8	generic_instance_lineage	\N	91f5a54b-5d03-4ffe-a643-76c4370b80e9	GL530	\N	\N	daylily	2025-07-03 08:23:17.532606-07	INSERT	\N	\N	\N	f	f
89520005-b0db-432d-a92b-b17297414a60	generic_instance	\N	3ab97a64-e71c-41b1-b799-60b928503b2e	CWX198	\N	\N	daylily	2025-07-03 08:23:17.532606-07	INSERT	\N	\N	\N	f	f
40f40732-ab06-4dd9-a089-c7fb67e4d61d	generic_instance	json_addl	3ab97a64-e71c-41b1-b799-60b928503b2e	CWX198	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.536954-07	UPDATE	\N	\N	\N	f	f
33fa4067-3aac-41b0-ad8b-424abbeb2f16	generic_instance	created_dt	3ab97a64-e71c-41b1-b799-60b928503b2e	CWX198	2025-07-03 08:23:17.532606-07	2025-07-03T08:23:17.532606-07:00	daylily	2025-07-03 08:23:17.536954-07	UPDATE	\N	\N	\N	f	f
cd15f55b-a2b3-4162-9f82-9acba8af79cd	generic_instance	modified_dt	3ab97a64-e71c-41b1-b799-60b928503b2e	CWX198	2025-07-03 08:23:17.532606-07	2025-07-03T08:23:17.536954-07:00	daylily	2025-07-03 08:23:17.536954-07	UPDATE	\N	\N	\N	f	f
0cac239d-06a2-4ac3-bea9-ae30ed581e16	generic_instance_lineage	\N	e07a7bdc-4437-469c-88d1-ab4e8f2e0b7b	GL531	\N	\N	daylily	2025-07-03 08:23:17.538143-07	INSERT	\N	\N	\N	f	f
bedcd392-8447-4477-a289-78d21a120ec4	generic_instance	\N	1773aa0b-fcb9-440c-99ba-e97d08c897d2	CWX199	\N	\N	daylily	2025-07-03 08:23:17.538143-07	INSERT	\N	\N	\N	f	f
3ebbd47f-32fa-401f-8f83-ba8160bf5a62	generic_instance	json_addl	b9d2526e-ab8a-4ff8-a1b0-947071f5d2d7	CX19	{"cogs": {"cost": "1.50", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "10ml tube generic", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "Generic 10ml tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "1.50", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "giab NA24385", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "Generic 10ml tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.436412-07	UPDATE	\N	\N	\N	f	f
a4ee8b40-cf4f-4392-b991-f7b93614beca	generic_instance	json_addl	1773aa0b-fcb9-440c-99ba-e97d08c897d2	CWX199	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.54221-07	UPDATE	\N	\N	\N	f	f
f7375111-2544-4cee-9c94-638bb3ca23ee	generic_instance	created_dt	1773aa0b-fcb9-440c-99ba-e97d08c897d2	CWX199	2025-07-03 08:23:17.538143-07	2025-07-03T08:23:17.538143-07:00	daylily	2025-07-03 08:23:17.54221-07	UPDATE	\N	\N	\N	f	f
414f505a-9c99-4757-af46-d62bdc1ba221	generic_instance	modified_dt	1773aa0b-fcb9-440c-99ba-e97d08c897d2	CWX199	2025-07-03 08:23:17.538143-07	2025-07-03T08:23:17.54221-07:00	daylily	2025-07-03 08:23:17.54221-07	UPDATE	\N	\N	\N	f	f
3294dfa9-b336-4980-aa98-ff48e8ed02c7	generic_instance_lineage	\N	00c14f88-0d17-46a0-b361-30b1db2e82be	GL532	\N	\N	daylily	2025-07-03 08:23:17.543423-07	INSERT	\N	\N	\N	f	f
929ca021-20d7-40ab-8716-93049f49fae7	generic_instance	\N	245e0cf5-ffc9-4b32-bd05-857c40d77e4d	CWX200	\N	\N	daylily	2025-07-03 08:23:17.543423-07	INSERT	\N	\N	\N	f	f
2c47e0bb-fbcd-4dec-84d0-783115aa5dc7	generic_instance	json_addl	245e0cf5-ffc9-4b32-bd05-857c40d77e4d	CWX200	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.547902-07	UPDATE	\N	\N	\N	f	f
7fd6c87b-2f5e-43dd-a215-813240411a83	generic_instance	created_dt	245e0cf5-ffc9-4b32-bd05-857c40d77e4d	CWX200	2025-07-03 08:23:17.543423-07	2025-07-03T08:23:17.543423-07:00	daylily	2025-07-03 08:23:17.547902-07	UPDATE	\N	\N	\N	f	f
fab3b11c-0376-4808-a0c4-e4058135cde8	generic_instance	modified_dt	245e0cf5-ffc9-4b32-bd05-857c40d77e4d	CWX200	2025-07-03 08:23:17.543423-07	2025-07-03T08:23:17.547902-07:00	daylily	2025-07-03 08:23:17.547902-07	UPDATE	\N	\N	\N	f	f
1ff5da11-dc90-4e17-a640-c4cbf2052aa1	generic_instance_lineage	\N	ad6515f0-ae8c-4813-98af-f1ad0a80bd22	GL533	\N	\N	daylily	2025-07-03 08:23:17.549119-07	INSERT	\N	\N	\N	f	f
14bb837b-734c-4f2d-9329-12135403f794	generic_instance	\N	3f3515b1-157e-4537-9270-70187b0b55b2	CWX201	\N	\N	daylily	2025-07-03 08:23:17.549119-07	INSERT	\N	\N	\N	f	f
a08db1a3-d4ac-4701-8de1-9172281b0a68	generic_instance	json_addl	3f3515b1-157e-4537-9270-70187b0b55b2	CWX201	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.553401-07	UPDATE	\N	\N	\N	f	f
6ca363db-fffe-4145-a7a2-2a7226de7f1d	generic_instance	created_dt	3f3515b1-157e-4537-9270-70187b0b55b2	CWX201	2025-07-03 08:23:17.549119-07	2025-07-03T08:23:17.549119-07:00	daylily	2025-07-03 08:23:17.553401-07	UPDATE	\N	\N	\N	f	f
9f40ceec-e90f-4b5e-aefd-193f87996ce4	generic_instance	modified_dt	3f3515b1-157e-4537-9270-70187b0b55b2	CWX201	2025-07-03 08:23:17.549119-07	2025-07-03T08:23:17.553401-07:00	daylily	2025-07-03 08:23:17.553401-07	UPDATE	\N	\N	\N	f	f
3a893061-09a0-4271-ae23-d725c19925ea	generic_instance_lineage	\N	a66eb90a-d406-4b1b-8a15-e8e68faa0a5a	GL534	\N	\N	daylily	2025-07-03 08:23:17.554699-07	INSERT	\N	\N	\N	f	f
7c5b9750-80c1-4483-ba0d-d79b67d268be	generic_instance	\N	6282522a-6923-4b31-8c11-9945e5d04e9c	CWX202	\N	\N	daylily	2025-07-03 08:23:17.554699-07	INSERT	\N	\N	\N	f	f
55814a85-a7c6-461c-bf30-565b8e651690	generic_instance	created_dt	b9d2526e-ab8a-4ff8-a1b0-947071f5d2d7	CX19	2025-07-03 08:24:16.425128-07	2025-07-03T08:24:16.425128-07:00	daylily	2025-07-03 08:24:16.436412-07	UPDATE	\N	\N	\N	f	f
3ef3a535-3562-488e-a4f9-0add2e53bfa4	generic_instance	modified_dt	b9d2526e-ab8a-4ff8-a1b0-947071f5d2d7	CX19	2025-07-03 08:24:16.425128-07	2025-07-03T08:24:16.436412-07:00	daylily	2025-07-03 08:24:16.436412-07	UPDATE	\N	\N	\N	f	f
e495b7bb-f280-4c40-b52c-665d387f936c	generic_instance_lineage	\N	ed5b3cd1-d211-4309-ac26-90ed4d8816b6	GL701	\N	\N	daylily	2025-07-03 08:24:16.436412-07	INSERT	\N	\N	\N	f	f
3fbbadc1-15a7-4ef0-972b-3646c66653ac	generic_instance	\N	0bbf1234-7c4b-4afc-9352-56a600deb076	CX20	\N	\N	daylily	2025-07-03 08:24:16.445109-07	INSERT	\N	\N	\N	f	f
f71130d2-2473-4cb4-84d4-216821aa3249	generic_instance	\N	4f0b197f-9f39-43fd-90bd-3ee2cf2ca0a6	MCX8	\N	\N	daylily	2025-07-03 08:24:16.448073-07	INSERT	\N	\N	\N	f	f
20184436-f419-4353-aa0c-417583ba6843	generic_instance	json_addl	6282522a-6923-4b31-8c11-9945e5d04e9c	CWX202	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.558745-07	UPDATE	\N	\N	\N	f	f
ba1081fe-f218-4aa2-9df6-50409ef23a68	generic_instance	created_dt	6282522a-6923-4b31-8c11-9945e5d04e9c	CWX202	2025-07-03 08:23:17.554699-07	2025-07-03T08:23:17.554699-07:00	daylily	2025-07-03 08:23:17.558745-07	UPDATE	\N	\N	\N	f	f
4ab4369e-7023-44c7-a10e-77fafb7d063c	generic_instance	modified_dt	6282522a-6923-4b31-8c11-9945e5d04e9c	CWX202	2025-07-03 08:23:17.554699-07	2025-07-03T08:23:17.558745-07:00	daylily	2025-07-03 08:23:17.558745-07	UPDATE	\N	\N	\N	f	f
3d91e19a-0e64-4d6d-93f1-1f4e615bd06f	generic_instance_lineage	\N	cbabf5f8-b703-43ee-8dd5-3bc08bbe8cbf	GL535	\N	\N	daylily	2025-07-03 08:23:17.560004-07	INSERT	\N	\N	\N	f	f
cadfb18b-ed18-4922-9fd2-d894c9d0af98	generic_instance	\N	743e4c9a-0b92-48f1-b15e-aa89d9ef6f17	CWX203	\N	\N	daylily	2025-07-03 08:23:17.560004-07	INSERT	\N	\N	\N	f	f
18190ac7-3d07-42cf-872d-1e67fba684c9	generic_instance	json_addl	743e4c9a-0b92-48f1-b15e-aa89d9ef6f17	CWX203	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.564347-07	UPDATE	\N	\N	\N	f	f
50c9a0d2-6712-44f2-8a57-92dfa7e44af7	generic_instance	created_dt	743e4c9a-0b92-48f1-b15e-aa89d9ef6f17	CWX203	2025-07-03 08:23:17.560004-07	2025-07-03T08:23:17.560004-07:00	daylily	2025-07-03 08:23:17.564347-07	UPDATE	\N	\N	\N	f	f
ab175ede-1f50-4883-9ac3-7fa3bb488afe	generic_instance	modified_dt	743e4c9a-0b92-48f1-b15e-aa89d9ef6f17	CWX203	2025-07-03 08:23:17.560004-07	2025-07-03T08:23:17.564347-07:00	daylily	2025-07-03 08:23:17.564347-07	UPDATE	\N	\N	\N	f	f
833e2464-49e4-4209-8803-adeba7aaae9c	generic_instance_lineage	\N	206bdf72-c5b7-4fd1-b83a-523d1d96ac15	GL536	\N	\N	daylily	2025-07-03 08:23:17.566042-07	INSERT	\N	\N	\N	f	f
a985eb9d-3048-4485-89b5-aaa8da4a0099	generic_instance	\N	3fb7f5a8-6c7b-44c6-af3b-875bb8f8fcc9	CWX204	\N	\N	daylily	2025-07-03 08:23:17.566042-07	INSERT	\N	\N	\N	f	f
97ea7407-6259-40c2-85d2-ba505d4e6624	generic_instance	json_addl	3fb7f5a8-6c7b-44c6-af3b-875bb8f8fcc9	CWX204	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.570514-07	UPDATE	\N	\N	\N	f	f
fcb0f065-763c-46db-ac5a-c3486a04e49b	generic_instance	created_dt	3fb7f5a8-6c7b-44c6-af3b-875bb8f8fcc9	CWX204	2025-07-03 08:23:17.566042-07	2025-07-03T08:23:17.566042-07:00	daylily	2025-07-03 08:23:17.570514-07	UPDATE	\N	\N	\N	f	f
cc042bab-9000-4630-9f84-1203cc6ed4bc	generic_instance	modified_dt	3fb7f5a8-6c7b-44c6-af3b-875bb8f8fcc9	CWX204	2025-07-03 08:23:17.566042-07	2025-07-03T08:23:17.570514-07:00	daylily	2025-07-03 08:23:17.570514-07	UPDATE	\N	\N	\N	f	f
3438c518-e722-41b9-9676-576c210ac4c2	generic_instance_lineage	\N	abcd409c-29c1-4b65-98e5-11f2f9263ddd	GL537	\N	\N	daylily	2025-07-03 08:23:17.571878-07	INSERT	\N	\N	\N	f	f
eeaf856b-9285-4bfb-9b72-9cbf4b6e6192	generic_instance	\N	24c55350-7ef4-4463-90ab-46252ab1c1c5	CWX205	\N	\N	daylily	2025-07-03 08:23:17.571878-07	INSERT	\N	\N	\N	f	f
7437e8da-0db6-459e-8112-3a6abbc68a1b	generic_instance	\N	ca939326-d2b1-4b88-80a3-15cb849ba2af	MRX285	\N	\N	daylily	2025-07-03 08:24:16.655929-07	INSERT	\N	\N	\N	f	f
dc8c171d-73e3-4b2e-9d53-e2fa6ca4f0b4	generic_instance_lineage	\N	7916e637-e90f-4e7e-858e-e2c038f113e3	GL736	\N	\N	daylily	2025-07-03 08:24:16.662326-07	INSERT	\N	\N	\N	f	f
cdf76f52-ecfa-4188-93b8-3359627c7d11	generic_instance	\N	88694a6f-07f1-4ef8-a9b8-97467620e235	MRX286	\N	\N	daylily	2025-07-03 08:24:16.662326-07	INSERT	\N	\N	\N	f	f
b45faaa7-4aca-4f00-8d99-11d2a07e9d1c	generic_instance_lineage	\N	2b1f02f9-59c7-4526-a897-b4ba3cf6d12f	GL737	\N	\N	daylily	2025-07-03 08:24:16.668647-07	INSERT	\N	\N	\N	f	f
c676c478-bd3c-4705-8426-e13f69b0a1f1	generic_instance	\N	86c0d971-5f10-4cb5-ae88-c21624503ba6	MRX287	\N	\N	daylily	2025-07-03 08:24:16.668647-07	INSERT	\N	\N	\N	f	f
2ff0976a-2ec0-449f-bc32-05a834ed073d	generic_instance_lineage	\N	c0364bde-d980-4756-b7da-a924defc131a	GL738	\N	\N	daylily	2025-07-03 08:24:16.67413-07	INSERT	\N	\N	\N	f	f
79e3c33a-acbf-4f0b-b262-cc9a2c5718e2	generic_instance	\N	9f1fb3f4-0611-4f6e-a39a-71b3ae2dba25	MRX288	\N	\N	daylily	2025-07-03 08:24:16.67413-07	INSERT	\N	\N	\N	f	f
1e04bb53-44e6-4b32-bf35-e65652ebcee4	generic_instance	json_addl	24c55350-7ef4-4463-90ab-46252ab1c1c5	CWX205	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.576571-07	UPDATE	\N	\N	\N	f	f
ae1c3c1b-2f09-4aad-9753-79baa4313830	generic_instance	created_dt	24c55350-7ef4-4463-90ab-46252ab1c1c5	CWX205	2025-07-03 08:23:17.571878-07	2025-07-03T08:23:17.571878-07:00	daylily	2025-07-03 08:23:17.576571-07	UPDATE	\N	\N	\N	f	f
b9af2457-16c3-430a-b8c2-90e246351f21	generic_instance	modified_dt	24c55350-7ef4-4463-90ab-46252ab1c1c5	CWX205	2025-07-03 08:23:17.571878-07	2025-07-03T08:23:17.576571-07:00	daylily	2025-07-03 08:23:17.576571-07	UPDATE	\N	\N	\N	f	f
f92ee3ab-36f0-4d04-9e9b-7dbe29f2fa6c	generic_instance_lineage	\N	813107eb-d656-4fb7-aab3-8d12a0e8711d	GL538	\N	\N	daylily	2025-07-03 08:23:17.577992-07	INSERT	\N	\N	\N	f	f
239fd4a9-6142-409c-a167-a800bba6c535	generic_instance	\N	8ad330a0-fad3-4b94-be1d-6c2aade5b972	CWX206	\N	\N	daylily	2025-07-03 08:23:17.577992-07	INSERT	\N	\N	\N	f	f
e3b7c050-955f-4395-998b-0c596796d594	generic_instance	json_addl	8ad330a0-fad3-4b94-be1d-6c2aade5b972	CWX206	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.582129-07	UPDATE	\N	\N	\N	f	f
09a5a18c-d07d-4c43-8b5b-c116d40c7be6	generic_instance	created_dt	8ad330a0-fad3-4b94-be1d-6c2aade5b972	CWX206	2025-07-03 08:23:17.577992-07	2025-07-03T08:23:17.577992-07:00	daylily	2025-07-03 08:23:17.582129-07	UPDATE	\N	\N	\N	f	f
614cfeae-501d-4d9d-9372-29530e8a1026	generic_instance	modified_dt	8ad330a0-fad3-4b94-be1d-6c2aade5b972	CWX206	2025-07-03 08:23:17.577992-07	2025-07-03T08:23:17.582129-07:00	daylily	2025-07-03 08:23:17.582129-07	UPDATE	\N	\N	\N	f	f
877bb1e9-2f03-43b5-bdfe-1541a48d2ca8	generic_instance_lineage	\N	40c0072f-2304-438d-ac92-c751dc363d54	GL539	\N	\N	daylily	2025-07-03 08:23:17.583496-07	INSERT	\N	\N	\N	f	f
25b3d1f8-4420-4378-820e-f908ce4af071	generic_instance	\N	801a01d3-8107-4911-9822-69f73fef2dee	CWX207	\N	\N	daylily	2025-07-03 08:23:17.583496-07	INSERT	\N	\N	\N	f	f
06255350-1de6-490f-aae3-797486a7fed1	generic_instance	json_addl	801a01d3-8107-4911-9822-69f73fef2dee	CWX207	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.588081-07	UPDATE	\N	\N	\N	f	f
1dc1208f-3c0f-4838-ab26-1a76dc4fa84e	generic_instance	created_dt	801a01d3-8107-4911-9822-69f73fef2dee	CWX207	2025-07-03 08:23:17.583496-07	2025-07-03T08:23:17.583496-07:00	daylily	2025-07-03 08:23:17.588081-07	UPDATE	\N	\N	\N	f	f
a9b04400-114e-4411-98c6-124159d6da6b	generic_instance	modified_dt	801a01d3-8107-4911-9822-69f73fef2dee	CWX207	2025-07-03 08:23:17.583496-07	2025-07-03T08:23:17.588081-07:00	daylily	2025-07-03 08:23:17.588081-07	UPDATE	\N	\N	\N	f	f
64a7760d-294b-443f-a635-f8337e2b5860	generic_instance_lineage	\N	95530e78-12ef-44c1-80f4-d1edb0c6abf1	GL540	\N	\N	daylily	2025-07-03 08:23:17.589387-07	INSERT	\N	\N	\N	f	f
8c74837b-fb5f-4b4a-b1b6-226505f500e6	generic_instance	\N	f6e70812-dce0-4359-848a-7c86a15b25eb	CWX208	\N	\N	daylily	2025-07-03 08:23:17.589387-07	INSERT	\N	\N	\N	f	f
87d97a03-1575-4ae5-8568-b177ea806066	generic_instance_lineage	\N	157ddf80-29b8-445d-93d1-b2088c505dc8	GL739	\N	\N	daylily	2025-07-03 08:24:16.679546-07	INSERT	\N	\N	\N	f	f
b684ce56-c8db-45f7-bef5-daa65bf5b7d9	generic_instance	\N	e47de7d0-f3d5-4fd5-ae77-6f3d053caf84	MRX289	\N	\N	daylily	2025-07-03 08:24:16.679546-07	INSERT	\N	\N	\N	f	f
89e37801-316b-4021-999f-f791239c3a93	generic_instance_lineage	\N	b931f168-c2da-4392-a535-653e5daa6ffa	GL740	\N	\N	daylily	2025-07-03 08:24:16.684741-07	INSERT	\N	\N	\N	f	f
c6498f24-0254-4d51-89e2-26a99b6ce863	generic_instance	\N	e72c6cc1-db0a-41bd-b271-a7d5b645df0c	MRX290	\N	\N	daylily	2025-07-03 08:24:16.684741-07	INSERT	\N	\N	\N	f	f
22eb8e6e-3992-47b5-a1e8-30cf8564e705	generic_instance_lineage	\N	1d967a38-f06d-47f0-a544-95629e27e900	GL741	\N	\N	daylily	2025-07-03 08:24:16.690129-07	INSERT	\N	\N	\N	f	f
8dfd7bd2-9385-47b8-aa81-0565cb66bed4	generic_instance	\N	49a12695-1781-4daa-80c6-fbb71f9ba0ba	MRX291	\N	\N	daylily	2025-07-03 08:24:16.690129-07	INSERT	\N	\N	\N	f	f
ebc498f4-34d6-4b83-86ba-c0e844919455	generic_instance_lineage	\N	b993f737-d611-4508-a4a0-fe83b78dc322	GL742	\N	\N	daylily	2025-07-03 08:24:16.697069-07	INSERT	\N	\N	\N	f	f
42d70bfc-a843-4a22-95ab-59bdfe37750c	generic_instance	json_addl	f6e70812-dce0-4359-848a-7c86a15b25eb	CWX208	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.593702-07	UPDATE	\N	\N	\N	f	f
d227ea69-0083-41d0-b30e-3ff7b14e95f8	generic_instance	created_dt	f6e70812-dce0-4359-848a-7c86a15b25eb	CWX208	2025-07-03 08:23:17.589387-07	2025-07-03T08:23:17.589387-07:00	daylily	2025-07-03 08:23:17.593702-07	UPDATE	\N	\N	\N	f	f
b7eb900a-8454-402d-82c8-94a67b82067a	generic_instance	modified_dt	f6e70812-dce0-4359-848a-7c86a15b25eb	CWX208	2025-07-03 08:23:17.589387-07	2025-07-03T08:23:17.593702-07:00	daylily	2025-07-03 08:23:17.593702-07	UPDATE	\N	\N	\N	f	f
95ae2933-363b-42d1-b2e8-020e7416eebc	generic_instance_lineage	\N	05f8ddab-694b-462d-92e7-3e4d8f1c88ca	GL541	\N	\N	daylily	2025-07-03 08:23:17.594959-07	INSERT	\N	\N	\N	f	f
7dcfbc7e-7452-48da-97d6-2e857ecc9d5c	generic_instance	\N	226cfe6d-f38c-4275-b630-54bc1a181c4f	CWX209	\N	\N	daylily	2025-07-03 08:23:17.594959-07	INSERT	\N	\N	\N	f	f
ab8083c2-df22-4a8a-a3aa-e8ee84e29539	generic_instance	json_addl	226cfe6d-f38c-4275-b630-54bc1a181c4f	CWX209	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.602179-07	UPDATE	\N	\N	\N	f	f
37c9aa28-60ed-45c6-885a-0ab3851342c3	generic_instance	created_dt	226cfe6d-f38c-4275-b630-54bc1a181c4f	CWX209	2025-07-03 08:23:17.594959-07	2025-07-03T08:23:17.594959-07:00	daylily	2025-07-03 08:23:17.602179-07	UPDATE	\N	\N	\N	f	f
fae7fe2b-6ef5-4f64-ad15-a04703a1b2aa	generic_instance	modified_dt	226cfe6d-f38c-4275-b630-54bc1a181c4f	CWX209	2025-07-03 08:23:17.594959-07	2025-07-03T08:23:17.602179-07:00	daylily	2025-07-03 08:23:17.602179-07	UPDATE	\N	\N	\N	f	f
1866f857-c35f-4ece-9ae7-d47253428f58	generic_instance_lineage	\N	fdfe2b7f-84f2-4096-8e9a-1223df323f2d	GL542	\N	\N	daylily	2025-07-03 08:23:17.605463-07	INSERT	\N	\N	\N	f	f
49a9ba61-705b-4b6d-91e2-27dfb670ec3e	generic_instance	\N	6d5743cc-81b2-46c0-bf95-ec24bf59c944	CWX210	\N	\N	daylily	2025-07-03 08:23:17.605463-07	INSERT	\N	\N	\N	f	f
22d225fb-2515-453f-9742-d083401835ea	generic_instance	json_addl	6d5743cc-81b2-46c0-bf95-ec24bf59c944	CWX210	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.610861-07	UPDATE	\N	\N	\N	f	f
008d716e-e000-48e4-be76-0cc489cdc9f1	generic_instance	created_dt	6d5743cc-81b2-46c0-bf95-ec24bf59c944	CWX210	2025-07-03 08:23:17.605463-07	2025-07-03T08:23:17.605463-07:00	daylily	2025-07-03 08:23:17.610861-07	UPDATE	\N	\N	\N	f	f
a076a395-87c3-4182-948c-261d9014b588	generic_instance	modified_dt	6d5743cc-81b2-46c0-bf95-ec24bf59c944	CWX210	2025-07-03 08:23:17.605463-07	2025-07-03T08:23:17.610861-07:00	daylily	2025-07-03 08:23:17.610861-07	UPDATE	\N	\N	\N	f	f
914f0ace-e874-4c4a-97ce-db85fda21e63	generic_instance_lineage	\N	24a4da19-904a-4b6b-9215-1c3bf022926c	GL543	\N	\N	daylily	2025-07-03 08:23:17.612317-07	INSERT	\N	\N	\N	f	f
832befe4-4431-43d0-8a5a-4ee53afe5b1a	generic_instance	\N	e63af02f-2942-4849-a657-ab339e79a86c	CWX211	\N	\N	daylily	2025-07-03 08:23:17.612317-07	INSERT	\N	\N	\N	f	f
d34cc426-c919-4d9a-adef-de0df1b9cf4e	generic_instance	\N	5117b15a-474f-43b7-b9eb-e367af711f2a	MRX292	\N	\N	daylily	2025-07-03 08:24:16.697069-07	INSERT	\N	\N	\N	f	f
a7abbda0-f190-45d7-939c-ae47d13bc1ba	generic_instance_lineage	\N	80c92374-35f4-4ed1-82ad-ed50eda7a153	GL743	\N	\N	daylily	2025-07-03 08:24:16.705224-07	INSERT	\N	\N	\N	f	f
86ff2a86-03e0-479a-aa38-481c5ef3376c	generic_instance	\N	cb3a3bb1-aefc-4355-aae0-fcf5a8a94013	MRX293	\N	\N	daylily	2025-07-03 08:24:16.705224-07	INSERT	\N	\N	\N	f	f
0f9a4db1-42f8-4456-90af-b45ac93f276b	generic_instance_lineage	\N	a2eda545-dd1a-49f3-aa99-a19375707042	GL744	\N	\N	daylily	2025-07-03 08:24:16.710408-07	INSERT	\N	\N	\N	f	f
ef1c4308-4429-45ec-8781-3e5b35f93d48	generic_instance	\N	51e3ccc5-9470-49f4-987d-b96a897a3fd2	MRX294	\N	\N	daylily	2025-07-03 08:24:16.710408-07	INSERT	\N	\N	\N	f	f
4176b108-44c8-45dd-b2fc-d10248fe4f43	generic_instance_lineage	\N	74a5f05a-5eb8-42c3-aaa2-7bcc9fb6aafd	GL745	\N	\N	daylily	2025-07-03 08:24:16.715575-07	INSERT	\N	\N	\N	f	f
36ed7e25-ff88-47af-8f88-a2d28c8c02d2	generic_instance	\N	dea04455-e99a-4424-ac3b-fc7124de43f8	MRX295	\N	\N	daylily	2025-07-03 08:24:16.715575-07	INSERT	\N	\N	\N	f	f
d64f6673-bb70-4dfa-8f26-a5e643d2e1ad	generic_instance	json_addl	e63af02f-2942-4849-a657-ab339e79a86c	CWX211	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.616695-07	UPDATE	\N	\N	\N	f	f
a3a4abcf-42c5-435f-8574-86594e19e82a	generic_instance	created_dt	e63af02f-2942-4849-a657-ab339e79a86c	CWX211	2025-07-03 08:23:17.612317-07	2025-07-03T08:23:17.612317-07:00	daylily	2025-07-03 08:23:17.616695-07	UPDATE	\N	\N	\N	f	f
42908080-9144-4e46-914c-09ad316f6ca7	generic_instance	modified_dt	e63af02f-2942-4849-a657-ab339e79a86c	CWX211	2025-07-03 08:23:17.612317-07	2025-07-03T08:23:17.616695-07:00	daylily	2025-07-03 08:23:17.616695-07	UPDATE	\N	\N	\N	f	f
db58435d-90d3-4a7f-933f-b37897dbed29	generic_instance_lineage	\N	04b0a9cf-6877-41d3-b1bc-b06cbdc6c386	GL544	\N	\N	daylily	2025-07-03 08:23:17.618607-07	INSERT	\N	\N	\N	f	f
86d0204b-563b-40c1-b8a9-9974ce3287f1	generic_instance	\N	ab47b771-3bc5-4ec6-88ea-14a05ab36fe7	CWX212	\N	\N	daylily	2025-07-03 08:23:17.618607-07	INSERT	\N	\N	\N	f	f
d6f94971-ee4a-43bd-b469-32e67e6c4409	generic_instance	json_addl	ab47b771-3bc5-4ec6-88ea-14a05ab36fe7	CWX212	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.624378-07	UPDATE	\N	\N	\N	f	f
12cb3758-848b-40e5-bf1d-5b5398d50a5c	generic_instance	created_dt	ab47b771-3bc5-4ec6-88ea-14a05ab36fe7	CWX212	2025-07-03 08:23:17.618607-07	2025-07-03T08:23:17.618607-07:00	daylily	2025-07-03 08:23:17.624378-07	UPDATE	\N	\N	\N	f	f
dfcbf6ed-2c0c-4853-99c3-19f75f844efc	generic_instance	modified_dt	ab47b771-3bc5-4ec6-88ea-14a05ab36fe7	CWX212	2025-07-03 08:23:17.618607-07	2025-07-03T08:23:17.624378-07:00	daylily	2025-07-03 08:23:17.624378-07	UPDATE	\N	\N	\N	f	f
cf8e22d8-007b-4670-99ae-255d308a7ea4	generic_instance_lineage	\N	7626f0e3-d4d8-4151-9e78-5f37f2bd2533	GL545	\N	\N	daylily	2025-07-03 08:23:17.625699-07	INSERT	\N	\N	\N	f	f
d38f7b01-44dd-4533-90ab-961d50bc3c2c	generic_instance	\N	32a38823-7229-4403-b7c5-4564b57105f6	CWX213	\N	\N	daylily	2025-07-03 08:23:17.625699-07	INSERT	\N	\N	\N	f	f
7b9f8655-94ee-4b9f-8380-22c8c17d0af4	generic_instance	json_addl	32a38823-7229-4403-b7c5-4564b57105f6	CWX213	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.629956-07	UPDATE	\N	\N	\N	f	f
43f8a786-98bc-4f0b-b437-900b9b0938e7	generic_instance	created_dt	32a38823-7229-4403-b7c5-4564b57105f6	CWX213	2025-07-03 08:23:17.625699-07	2025-07-03T08:23:17.625699-07:00	daylily	2025-07-03 08:23:17.629956-07	UPDATE	\N	\N	\N	f	f
779f28cb-e2b4-4138-b548-2e35bf704185	generic_instance	modified_dt	32a38823-7229-4403-b7c5-4564b57105f6	CWX213	2025-07-03 08:23:17.625699-07	2025-07-03T08:23:17.629956-07:00	daylily	2025-07-03 08:23:17.629956-07	UPDATE	\N	\N	\N	f	f
bd41d8b3-bc97-444a-b28b-5a5c306e44f0	generic_instance_lineage	\N	fcb84219-4735-4483-bf9a-1d6fb3746398	GL546	\N	\N	daylily	2025-07-03 08:23:17.631317-07	INSERT	\N	\N	\N	f	f
2678c987-21d7-4fb6-9d53-303e50cf85de	generic_instance	\N	ed9a04fc-0e00-4fd1-8813-54c833e62db0	CWX214	\N	\N	daylily	2025-07-03 08:23:17.631317-07	INSERT	\N	\N	\N	f	f
62386970-8854-4151-af19-451d53aff805	generic_instance_lineage	\N	87175ff8-4efe-4080-b847-74d4fbbe650d	GL746	\N	\N	daylily	2025-07-03 08:24:16.720517-07	INSERT	\N	\N	\N	f	f
4271e0ef-0717-47b8-bb32-b3ad50b6ca4f	generic_instance	\N	0559ee06-9d4f-4f58-a12e-1f32bdff4633	MRX296	\N	\N	daylily	2025-07-03 08:24:16.720517-07	INSERT	\N	\N	\N	f	f
325e1c7b-c435-40d3-bd6c-d1eb9a272b1c	generic_instance_lineage	\N	ac1ae2ef-839c-4f9c-a3e1-6ee4a006106b	GL747	\N	\N	daylily	2025-07-03 08:24:16.725843-07	INSERT	\N	\N	\N	f	f
51ebf65b-8883-4f91-b919-b7bd8413c8e8	generic_instance	\N	d20de8f6-379c-429c-9a0b-916b7833b1a4	MRX297	\N	\N	daylily	2025-07-03 08:24:16.725843-07	INSERT	\N	\N	\N	f	f
d4c03b43-bfc9-4c38-a012-0df73d7f3f30	generic_instance_lineage	\N	18f6a81b-71be-42e0-8cb5-7050c49b376c	GL748	\N	\N	daylily	2025-07-03 08:24:16.730975-07	INSERT	\N	\N	\N	f	f
8ddc3867-8fa6-4f13-bdb5-8e4f5c546064	generic_instance	\N	bb287c24-62a2-4531-8ca1-909803f845fc	MRX298	\N	\N	daylily	2025-07-03 08:24:16.730975-07	INSERT	\N	\N	\N	f	f
50aa7a08-ad71-4615-86f0-4ddf2608f951	generic_instance_lineage	\N	63dc7384-6002-45ac-9818-63728e7cce47	GL749	\N	\N	daylily	2025-07-03 08:24:16.736556-07	INSERT	\N	\N	\N	f	f
c160393b-ee90-45d7-a6a6-28f08f64883e	generic_instance	json_addl	ed9a04fc-0e00-4fd1-8813-54c833e62db0	CWX214	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.657687-07	UPDATE	\N	\N	\N	f	f
debbcf7b-11f5-42bd-bed3-f59c28c13a22	generic_instance	created_dt	ed9a04fc-0e00-4fd1-8813-54c833e62db0	CWX214	2025-07-03 08:23:17.631317-07	2025-07-03T08:23:17.631317-07:00	daylily	2025-07-03 08:23:17.657687-07	UPDATE	\N	\N	\N	f	f
4238267d-4472-4896-b854-a873a04678a0	generic_instance	modified_dt	ed9a04fc-0e00-4fd1-8813-54c833e62db0	CWX214	2025-07-03 08:23:17.631317-07	2025-07-03T08:23:17.657687-07:00	daylily	2025-07-03 08:23:17.657687-07	UPDATE	\N	\N	\N	f	f
9bfa9efc-a72b-4ff3-9c12-635a24efd363	generic_instance_lineage	\N	ae2cc2a9-2643-4b26-bb1b-048663237d5e	GL547	\N	\N	daylily	2025-07-03 08:23:17.685254-07	INSERT	\N	\N	\N	f	f
322f8a42-2e5e-419c-bc01-62f9779ad092	generic_instance	\N	39bff9ce-0ded-465c-b704-dda606530e7a	CWX215	\N	\N	daylily	2025-07-03 08:23:17.685254-07	INSERT	\N	\N	\N	f	f
83cd476a-896b-45d1-8d75-3093829cea8f	generic_instance	json_addl	39bff9ce-0ded-465c-b704-dda606530e7a	CWX215	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.700558-07	UPDATE	\N	\N	\N	f	f
6b781bab-a0f4-4599-ad71-359664ec1b62	generic_instance	created_dt	39bff9ce-0ded-465c-b704-dda606530e7a	CWX215	2025-07-03 08:23:17.685254-07	2025-07-03T08:23:17.685254-07:00	daylily	2025-07-03 08:23:17.700558-07	UPDATE	\N	\N	\N	f	f
d2786f4f-6c38-4b84-94b3-1300446b3d30	generic_instance	modified_dt	39bff9ce-0ded-465c-b704-dda606530e7a	CWX215	2025-07-03 08:23:17.685254-07	2025-07-03T08:23:17.700558-07:00	daylily	2025-07-03 08:23:17.700558-07	UPDATE	\N	\N	\N	f	f
4f8dcca2-aefd-4469-b1ee-cc2718fa4fbc	generic_instance_lineage	\N	f78e5840-de52-49a1-9f70-67ed6f26f822	GL548	\N	\N	daylily	2025-07-03 08:23:17.703574-07	INSERT	\N	\N	\N	f	f
d8fe367d-29ae-4349-8a17-4bd1f05da867	generic_instance	\N	fa6b2d22-7543-4429-b740-abd29980cabb	CWX216	\N	\N	daylily	2025-07-03 08:23:17.703574-07	INSERT	\N	\N	\N	f	f
04416f57-6d21-4681-8e02-75af3c6ce06e	generic_instance	json_addl	fa6b2d22-7543-4429-b740-abd29980cabb	CWX216	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.708097-07	UPDATE	\N	\N	\N	f	f
c030db16-10a6-400f-87d9-6eddc0597d9d	generic_instance	created_dt	fa6b2d22-7543-4429-b740-abd29980cabb	CWX216	2025-07-03 08:23:17.703574-07	2025-07-03T08:23:17.703574-07:00	daylily	2025-07-03 08:23:17.708097-07	UPDATE	\N	\N	\N	f	f
7af75c64-4f24-4253-b0b1-ab33eb67787f	generic_instance	modified_dt	fa6b2d22-7543-4429-b740-abd29980cabb	CWX216	2025-07-03 08:23:17.703574-07	2025-07-03T08:23:17.708097-07:00	daylily	2025-07-03 08:23:17.708097-07	UPDATE	\N	\N	\N	f	f
c5817715-cd06-48b3-929e-c75ef2b17c58	generic_instance_lineage	\N	86c40630-ff92-4e95-a713-dd50a3d7a8a8	GL549	\N	\N	daylily	2025-07-03 08:23:17.709524-07	INSERT	\N	\N	\N	f	f
c90254e6-b217-4a2b-a20f-5449d07f2eb1	generic_instance	\N	2fe1c959-22e0-457c-b5f1-2e043fe9c6e4	CWX217	\N	\N	daylily	2025-07-03 08:23:17.709524-07	INSERT	\N	\N	\N	f	f
8e666d1c-dc5a-441d-a273-7aef5dacb825	generic_instance	\N	eb4ac051-5a75-4ba3-9725-081583664880	MRX299	\N	\N	daylily	2025-07-03 08:24:16.736556-07	INSERT	\N	\N	\N	f	f
c1f5810e-7d19-4dad-9ec3-acd44bb7c85d	generic_instance_lineage	\N	61b19c14-f268-41dc-ba4f-d56d1793e5c7	GL750	\N	\N	daylily	2025-07-03 08:24:16.741873-07	INSERT	\N	\N	\N	f	f
96b98b97-fe97-4ba3-aba9-a3c40d5df1c2	generic_instance	\N	89e14590-5f15-40df-8f99-d3784fb556be	MRX300	\N	\N	daylily	2025-07-03 08:24:16.741873-07	INSERT	\N	\N	\N	f	f
5a6fa93a-0f56-4bcc-8e05-498084970641	generic_instance_lineage	\N	68c70c14-45bf-4eba-8275-6b146d767096	GL751	\N	\N	daylily	2025-07-03 08:24:16.746901-07	INSERT	\N	\N	\N	f	f
6465660e-0c46-4591-ac86-521cfd363872	generic_instance	\N	7fc49341-3ee8-463f-b10a-5919cd6f2f0a	MRX301	\N	\N	daylily	2025-07-03 08:24:16.746901-07	INSERT	\N	\N	\N	f	f
c640dea3-3211-41a4-ae43-c5a8f5de4c18	generic_instance_lineage	\N	e64808e3-e2bd-4528-9fe1-c2f41c80b1d6	GL752	\N	\N	daylily	2025-07-03 08:24:16.752108-07	INSERT	\N	\N	\N	f	f
e92c6ad6-7dec-4bd9-85b2-c90d429a3c6e	generic_instance	\N	0262fa19-4669-4ca2-94d9-6fd44dfebbcf	MRX302	\N	\N	daylily	2025-07-03 08:24:16.752108-07	INSERT	\N	\N	\N	f	f
c6be4448-b068-4796-904d-75b02210b5a3	generic_instance	json_addl	2fe1c959-22e0-457c-b5f1-2e043fe9c6e4	CWX217	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.713615-07	UPDATE	\N	\N	\N	f	f
ef14af14-c42b-47ce-83d9-8c513a371e42	generic_instance	created_dt	2fe1c959-22e0-457c-b5f1-2e043fe9c6e4	CWX217	2025-07-03 08:23:17.709524-07	2025-07-03T08:23:17.709524-07:00	daylily	2025-07-03 08:23:17.713615-07	UPDATE	\N	\N	\N	f	f
5f9e16c0-1c92-4bc9-ae0c-29065f014572	generic_instance	modified_dt	2fe1c959-22e0-457c-b5f1-2e043fe9c6e4	CWX217	2025-07-03 08:23:17.709524-07	2025-07-03T08:23:17.713615-07:00	daylily	2025-07-03 08:23:17.713615-07	UPDATE	\N	\N	\N	f	f
72d4d554-2f85-4c1e-bd5d-fc4662329991	generic_instance_lineage	\N	2bda3132-9788-40e7-ac3a-58fac4c774b7	GL550	\N	\N	daylily	2025-07-03 08:23:17.714876-07	INSERT	\N	\N	\N	f	f
6b98f3f8-15a2-449a-a36f-6a1b9b795b59	generic_instance	\N	c42216f6-5feb-418b-af3c-792c2e43b666	CWX218	\N	\N	daylily	2025-07-03 08:23:17.714876-07	INSERT	\N	\N	\N	f	f
a0dc6a78-b6e0-4bcc-9267-3929fb0698c9	generic_instance	json_addl	c42216f6-5feb-418b-af3c-792c2e43b666	CWX218	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.719105-07	UPDATE	\N	\N	\N	f	f
a5cfdabf-308d-4fad-bf23-a129f480a491	generic_instance	created_dt	c42216f6-5feb-418b-af3c-792c2e43b666	CWX218	2025-07-03 08:23:17.714876-07	2025-07-03T08:23:17.714876-07:00	daylily	2025-07-03 08:23:17.719105-07	UPDATE	\N	\N	\N	f	f
e9ee28a7-9b58-44ea-8f2e-fcc9b8de6012	generic_instance	modified_dt	c42216f6-5feb-418b-af3c-792c2e43b666	CWX218	2025-07-03 08:23:17.714876-07	2025-07-03T08:23:17.719105-07:00	daylily	2025-07-03 08:23:17.719105-07	UPDATE	\N	\N	\N	f	f
ad064afa-a710-404f-9510-76ab0affedcf	generic_instance_lineage	\N	c1d9a385-213c-45a7-8ef9-0e4d0054f47a	GL551	\N	\N	daylily	2025-07-03 08:23:17.720404-07	INSERT	\N	\N	\N	f	f
820e52ad-2fe9-47e6-945a-3200c5be6dc2	generic_instance	\N	7207e508-99fc-41ef-91ab-fd60d5306f4e	CWX219	\N	\N	daylily	2025-07-03 08:23:17.720404-07	INSERT	\N	\N	\N	f	f
d3c66099-8622-4170-9280-0bafb2fde52a	generic_instance	json_addl	7207e508-99fc-41ef-91ab-fd60d5306f4e	CWX219	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.72516-07	UPDATE	\N	\N	\N	f	f
c591e055-c15b-41cf-b3e4-7750a8009f73	generic_instance	created_dt	7207e508-99fc-41ef-91ab-fd60d5306f4e	CWX219	2025-07-03 08:23:17.720404-07	2025-07-03T08:23:17.720404-07:00	daylily	2025-07-03 08:23:17.72516-07	UPDATE	\N	\N	\N	f	f
16f9bd99-453f-4cdb-9bae-501c13a66da4	generic_instance	modified_dt	7207e508-99fc-41ef-91ab-fd60d5306f4e	CWX219	2025-07-03 08:23:17.720404-07	2025-07-03T08:23:17.72516-07:00	daylily	2025-07-03 08:23:17.72516-07	UPDATE	\N	\N	\N	f	f
27ed8a25-bf45-4a32-bac2-d72c0797c266	generic_instance_lineage	\N	f8f1ced5-70e6-4a61-a9d9-d06d1f4ec0a5	GL552	\N	\N	daylily	2025-07-03 08:23:17.72663-07	INSERT	\N	\N	\N	f	f
dded9f5d-d1a7-46a2-8483-324bc5065fc9	generic_instance	\N	13a17e5f-6d85-417c-9d4a-956b5282800a	CWX220	\N	\N	daylily	2025-07-03 08:23:17.72663-07	INSERT	\N	\N	\N	f	f
78a5f102-7cf8-4fa5-93d1-40bc22688007	generic_instance_lineage	\N	81ebce8b-fd0c-4e36-880a-544ddd442ab0	GL753	\N	\N	daylily	2025-07-03 08:24:16.757112-07	INSERT	\N	\N	\N	f	f
dfa622ee-196f-41ba-b035-9cf03c2a2e89	generic_instance	\N	1e661782-0d7e-478b-834a-cf1b8e5619ad	MRX303	\N	\N	daylily	2025-07-03 08:24:16.757112-07	INSERT	\N	\N	\N	f	f
6ad58e93-604a-496a-bbf9-e65236ed6890	generic_instance_lineage	\N	446912ce-94a9-491f-9925-00e5e85aff7b	GL754	\N	\N	daylily	2025-07-03 08:24:16.762394-07	INSERT	\N	\N	\N	f	f
0796cfe1-b929-4d84-a03e-6ce173643e89	generic_instance	\N	1d281d59-1305-4a9a-b542-7a874fea7e37	MRX304	\N	\N	daylily	2025-07-03 08:24:16.762394-07	INSERT	\N	\N	\N	f	f
0b9d69c9-c894-4bf6-a685-73c909f22783	generic_instance_lineage	\N	0b6fa778-a9ff-4848-a072-6f37f098aa01	GL755	\N	\N	daylily	2025-07-03 08:24:16.767513-07	INSERT	\N	\N	\N	f	f
f42cf791-b64e-43c0-8786-df330eb4ab1d	generic_instance	\N	97b8ade0-e5da-4ab0-97a5-a03f9ce9b362	MRX305	\N	\N	daylily	2025-07-03 08:24:16.767513-07	INSERT	\N	\N	\N	f	f
4bb92e48-7e67-44c7-afbf-b4618b9d050f	generic_instance_lineage	\N	03021265-7b72-4ab7-b3f3-06c3e8bb2432	GL756	\N	\N	daylily	2025-07-03 08:24:16.772618-07	INSERT	\N	\N	\N	f	f
ef7ee9e1-c6bc-42eb-8347-098e880726c4	generic_instance	json_addl	13a17e5f-6d85-417c-9d4a-956b5282800a	CWX220	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.732601-07	UPDATE	\N	\N	\N	f	f
72858816-95ec-4ab5-84b5-2a1fa4c17310	generic_instance	created_dt	13a17e5f-6d85-417c-9d4a-956b5282800a	CWX220	2025-07-03 08:23:17.72663-07	2025-07-03T08:23:17.72663-07:00	daylily	2025-07-03 08:23:17.732601-07	UPDATE	\N	\N	\N	f	f
4cd2996f-5c64-454e-8fe7-57a20d86c094	generic_instance	modified_dt	13a17e5f-6d85-417c-9d4a-956b5282800a	CWX220	2025-07-03 08:23:17.72663-07	2025-07-03T08:23:17.732601-07:00	daylily	2025-07-03 08:23:17.732601-07	UPDATE	\N	\N	\N	f	f
f9e89603-3438-4cd9-acaa-a605996e9df2	generic_instance_lineage	\N	dd69217d-fc02-4b82-bd19-6e38124d5c8f	GL553	\N	\N	daylily	2025-07-03 08:23:17.733993-07	INSERT	\N	\N	\N	f	f
71dbdaf4-06e6-4180-aed6-e72f29bf5f4a	generic_instance	\N	f4382eed-95c5-4d1b-a046-f2b9b277ec34	CWX221	\N	\N	daylily	2025-07-03 08:23:17.733993-07	INSERT	\N	\N	\N	f	f
1a2f4351-14aa-455e-94c4-7147bfb61aaf	generic_instance	json_addl	f4382eed-95c5-4d1b-a046-f2b9b277ec34	CWX221	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.738774-07	UPDATE	\N	\N	\N	f	f
98455e5b-df79-4d4c-9c40-cceac3a6f56c	generic_instance	created_dt	f4382eed-95c5-4d1b-a046-f2b9b277ec34	CWX221	2025-07-03 08:23:17.733993-07	2025-07-03T08:23:17.733993-07:00	daylily	2025-07-03 08:23:17.738774-07	UPDATE	\N	\N	\N	f	f
0ffb1edb-9784-42c9-82d4-844d57dc3909	generic_instance	modified_dt	f4382eed-95c5-4d1b-a046-f2b9b277ec34	CWX221	2025-07-03 08:23:17.733993-07	2025-07-03T08:23:17.738774-07:00	daylily	2025-07-03 08:23:17.738774-07	UPDATE	\N	\N	\N	f	f
8c6028d4-efa0-4eb0-a79e-ef4c7bc17575	generic_instance_lineage	\N	fd45b679-7d8e-4072-ae36-602a54610c84	GL554	\N	\N	daylily	2025-07-03 08:23:17.740122-07	INSERT	\N	\N	\N	f	f
25a65f9b-683f-4749-9c48-c575535a024c	generic_instance	\N	741b104f-826e-443e-8efd-9d6aaf9ec330	CWX222	\N	\N	daylily	2025-07-03 08:23:17.740122-07	INSERT	\N	\N	\N	f	f
1da43367-8a81-4f4f-a1ce-a3195dd60568	generic_instance	json_addl	741b104f-826e-443e-8efd-9d6aaf9ec330	CWX222	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.744649-07	UPDATE	\N	\N	\N	f	f
ec77206e-02eb-41e8-9318-18a98aa1fec0	generic_instance	created_dt	741b104f-826e-443e-8efd-9d6aaf9ec330	CWX222	2025-07-03 08:23:17.740122-07	2025-07-03T08:23:17.740122-07:00	daylily	2025-07-03 08:23:17.744649-07	UPDATE	\N	\N	\N	f	f
7affe974-cc0f-4210-b077-e3c457bf9b9a	generic_instance	modified_dt	741b104f-826e-443e-8efd-9d6aaf9ec330	CWX222	2025-07-03 08:23:17.740122-07	2025-07-03T08:23:17.744649-07:00	daylily	2025-07-03 08:23:17.744649-07	UPDATE	\N	\N	\N	f	f
5faad909-7e0f-4a96-9071-f86a522e834e	generic_instance_lineage	\N	ef95e543-129f-4d06-8bb3-b580c983cc3c	GL555	\N	\N	daylily	2025-07-03 08:23:17.746439-07	INSERT	\N	\N	\N	f	f
060e7f5d-ecbe-45a7-a835-63ba4c8e84ad	generic_instance	\N	340f2341-eb12-4958-9c03-224ed6cc0747	CWX223	\N	\N	daylily	2025-07-03 08:23:17.746439-07	INSERT	\N	\N	\N	f	f
aa0ddd1d-0141-4c74-882e-8dadadbff923	generic_instance	\N	fa1ac264-1427-48af-9781-70e9f462b347	MRX306	\N	\N	daylily	2025-07-03 08:24:16.772618-07	INSERT	\N	\N	\N	f	f
c26e608f-6cde-4f52-b379-000ed44fa75a	generic_instance_lineage	\N	852c51fb-14e4-4bef-a935-2919b6048be1	GL757	\N	\N	daylily	2025-07-03 08:24:16.777579-07	INSERT	\N	\N	\N	f	f
280b2237-fec2-451f-99a7-acc4ec0309d2	generic_instance	\N	ed96bfa1-f338-408f-86b5-48328def257e	MRX307	\N	\N	daylily	2025-07-03 08:24:16.777579-07	INSERT	\N	\N	\N	f	f
442c2682-0bf1-47ab-9272-96d9c25291f8	generic_instance_lineage	\N	d649b760-159d-41fb-9970-82ab7455d1c7	GL758	\N	\N	daylily	2025-07-03 08:24:16.782698-07	INSERT	\N	\N	\N	f	f
0406c0af-9c75-4b48-bf3d-da0e6dd733e9	generic_instance	\N	9cd2fa8f-9130-42fa-b43d-5e9110ceaec6	MRX308	\N	\N	daylily	2025-07-03 08:24:16.782698-07	INSERT	\N	\N	\N	f	f
9c3d0a54-4334-415b-8e55-2f430a89e9d7	generic_instance_lineage	\N	ef5a37e6-2b00-4820-a71c-03d6039f4358	GL759	\N	\N	daylily	2025-07-03 08:24:16.787731-07	INSERT	\N	\N	\N	f	f
d612a684-0b0b-49d7-90c3-77edeb032e9a	generic_instance	\N	88e5e9bc-0bf2-40e7-98fb-cc2576cf330d	CX23	\N	\N	daylily	2025-07-03 08:24:16.787731-07	INSERT	\N	\N	\N	f	f
96f4d917-59cb-48dd-b43d-aadbff421ed3	generic_instance	json_addl	340f2341-eb12-4958-9c03-224ed6cc0747	CWX223	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.751906-07	UPDATE	\N	\N	\N	f	f
a9469d19-657d-4103-aada-ab3242831524	generic_instance	created_dt	340f2341-eb12-4958-9c03-224ed6cc0747	CWX223	2025-07-03 08:23:17.746439-07	2025-07-03T08:23:17.746439-07:00	daylily	2025-07-03 08:23:17.751906-07	UPDATE	\N	\N	\N	f	f
7857502f-2e42-4037-9e53-fe8356baa37e	generic_instance	modified_dt	340f2341-eb12-4958-9c03-224ed6cc0747	CWX223	2025-07-03 08:23:17.746439-07	2025-07-03T08:23:17.751906-07:00	daylily	2025-07-03 08:23:17.751906-07	UPDATE	\N	\N	\N	f	f
e61ab392-fb55-4414-9186-b3f966f08b77	generic_instance_lineage	\N	0e0d51ed-217a-42e4-a32c-0ac91f8375e1	GL556	\N	\N	daylily	2025-07-03 08:23:17.753422-07	INSERT	\N	\N	\N	f	f
cbf69892-8adb-403f-8bc1-4e6bca678116	generic_instance	\N	c6eb89d6-4a49-4da3-a615-6be4294b9235	CWX224	\N	\N	daylily	2025-07-03 08:23:17.753422-07	INSERT	\N	\N	\N	f	f
40ab2fbd-629b-4abc-84f7-49f81d8cba63	generic_instance	json_addl	c6eb89d6-4a49-4da3-a615-6be4294b9235	CWX224	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:23:17.757567-07	UPDATE	\N	\N	\N	f	f
dd0fe09d-5484-4844-921d-993841b1f417	generic_instance	created_dt	c6eb89d6-4a49-4da3-a615-6be4294b9235	CWX224	2025-07-03 08:23:17.753422-07	2025-07-03T08:23:17.753422-07:00	daylily	2025-07-03 08:23:17.757567-07	UPDATE	\N	\N	\N	f	f
94957756-b1ab-4d63-89d2-205b6cf1c681	generic_instance	modified_dt	c6eb89d6-4a49-4da3-a615-6be4294b9235	CWX224	2025-07-03 08:23:17.753422-07	2025-07-03T08:23:17.757567-07:00	daylily	2025-07-03 08:23:17.757567-07	UPDATE	\N	\N	\N	f	f
69799ee0-1dd1-4c16-9dc2-0811bc9dc1b4	generic_instance_lineage	\N	c9960503-0eeb-44bc-9fc8-7aa89a55623b	GL557	\N	\N	daylily	2025-07-03 08:23:17.759189-07	INSERT	\N	\N	\N	f	f
db51224b-ada7-41fa-8f89-157fccadd74c	generic_instance	\N	0d22f68d-dd27-4e92-ba89-b39005cfca81	MRX197	\N	\N	daylily	2025-07-03 08:23:17.761172-07	INSERT	\N	\N	\N	f	f
3c976b00-90fc-4b8b-867f-289ae27a969f	generic_instance_lineage	\N	991a8ecc-ff6c-43fa-bc74-8aaa5af0816b	GL558	\N	\N	daylily	2025-07-03 08:23:17.766127-07	INSERT	\N	\N	\N	f	f
f2667fac-371a-4190-9a16-96e3a808739a	generic_instance	\N	f186463f-a0de-4e66-9a85-60d861f29c20	MRX198	\N	\N	daylily	2025-07-03 08:23:17.766127-07	INSERT	\N	\N	\N	f	f
acaad608-e883-43e7-908b-95e7059f76ff	generic_instance_lineage	\N	204e9535-3393-4a36-83ee-179de1d3e671	GL559	\N	\N	daylily	2025-07-03 08:23:17.772121-07	INSERT	\N	\N	\N	f	f
76335d29-c4c3-4c7d-a243-4f72873c38c0	generic_instance	\N	78f2fa4b-bfcb-46fc-a9ef-2f0bb27926f8	MRX199	\N	\N	daylily	2025-07-03 08:23:17.772121-07	INSERT	\N	\N	\N	f	f
d28cb895-b167-4cd1-9133-d8ad3abe02bf	generic_instance_lineage	\N	a73eaa6e-cb5c-4dab-828a-0c27cdc81801	GL560	\N	\N	daylily	2025-07-03 08:23:17.781168-07	INSERT	\N	\N	\N	f	f
52dfb5f7-84f3-48bb-85a9-48d20788fcc9	generic_instance	\N	906cdad8-145e-4040-b4f2-5ef1f31c907b	MRX200	\N	\N	daylily	2025-07-03 08:23:17.781168-07	INSERT	\N	\N	\N	f	f
5570502d-0221-4fa0-a0f6-48e7d2512702	generic_instance_lineage	\N	37a75b76-cfaf-4790-8424-090bc57d284a	GL561	\N	\N	daylily	2025-07-03 08:23:17.786967-07	INSERT	\N	\N	\N	f	f
b1dd18ff-4cad-48d7-8b63-66859d31058c	generic_instance	\N	dbc67543-2fb8-4e02-949f-1c2146b9f1c3	MRX201	\N	\N	daylily	2025-07-03 08:23:17.786967-07	INSERT	\N	\N	\N	f	f
b234fa0f-7087-481b-bd86-d60f6f777bd5	generic_instance_lineage	\N	36f5a79b-9d80-4c2a-a1f4-aaca37b2bd6a	GL562	\N	\N	daylily	2025-07-03 08:23:17.792875-07	INSERT	\N	\N	\N	f	f
076645e0-6e9b-490b-bc2c-7c640e958639	generic_instance	\N	d9d3755e-bd33-44fe-a4de-a68f8f0e4275	MRX202	\N	\N	daylily	2025-07-03 08:23:17.792875-07	INSERT	\N	\N	\N	f	f
c03a7537-c24b-4223-95b6-6e9de6e81928	generic_instance_lineage	\N	5e3cfb28-fc2c-4fc9-aa77-fdc989624052	GL563	\N	\N	daylily	2025-07-03 08:23:17.799241-07	INSERT	\N	\N	\N	f	f
3ea4c2a6-2701-45fa-b30f-1dd5ad5ed57b	generic_instance	\N	9c4cca2b-65bf-4957-9a49-fefe346bdca8	MRX203	\N	\N	daylily	2025-07-03 08:23:17.799241-07	INSERT	\N	\N	\N	f	f
e734fdd1-27a8-4f10-b100-5435d4367e07	generic_instance_lineage	\N	c485273e-9dcf-422e-b10c-405cd3958b6f	GL564	\N	\N	daylily	2025-07-03 08:23:17.805147-07	INSERT	\N	\N	\N	f	f
680821e5-aee5-4179-8dad-59f3bf03904c	generic_instance	\N	a6902e58-6edc-4679-a335-fffd804cd1b3	MRX204	\N	\N	daylily	2025-07-03 08:23:17.805147-07	INSERT	\N	\N	\N	f	f
9cbf9aa5-add3-4384-b80c-d3dba6063617	generic_instance_lineage	\N	fbd21f8d-51f8-4a37-a07c-deed87faa0e2	GL565	\N	\N	daylily	2025-07-03 08:23:17.811179-07	INSERT	\N	\N	\N	f	f
e2061092-e42b-4ad0-b283-60fcdb6538c2	generic_instance	\N	4396a723-1c2c-490b-b006-94ab6438e2d8	MRX205	\N	\N	daylily	2025-07-03 08:23:17.811179-07	INSERT	\N	\N	\N	f	f
fbd5623e-8536-4499-930e-e57425a01501	generic_instance_lineage	\N	b51ec6b2-0b5e-48f3-8755-de7aabf8f9a0	GL566	\N	\N	daylily	2025-07-03 08:23:17.817124-07	INSERT	\N	\N	\N	f	f
1fc1805b-7850-4dfa-93ea-f9f521f1cc3b	generic_instance	\N	5bb46d8b-d389-4d1f-a86f-2bd41aef6aa1	MRX206	\N	\N	daylily	2025-07-03 08:23:17.817124-07	INSERT	\N	\N	\N	f	f
4fbaf611-6d95-4d34-abc2-9777a001d99c	generic_instance_lineage	\N	07ef2f1c-009c-492d-9d77-72862b15a277	GL567	\N	\N	daylily	2025-07-03 08:23:17.823295-07	INSERT	\N	\N	\N	f	f
0de89661-1612-42a2-99a8-e90019d83e5b	generic_instance	\N	3813f28b-8d60-4dee-9752-1dabc8ba2ee2	MRX207	\N	\N	daylily	2025-07-03 08:23:17.823295-07	INSERT	\N	\N	\N	f	f
34a086a8-d3c2-4067-9943-aa7c48ee1b73	generic_instance_lineage	\N	dcf4e965-78bd-4eb8-a1f4-547fdc7c6d17	GL568	\N	\N	daylily	2025-07-03 08:23:17.829526-07	INSERT	\N	\N	\N	f	f
46ebedf5-bdce-401a-a477-f254cea378d1	generic_instance	\N	95a96de5-0287-4548-a6bc-5427e9da5086	MRX208	\N	\N	daylily	2025-07-03 08:23:17.829526-07	INSERT	\N	\N	\N	f	f
93286b3d-a101-4671-ac0b-f35bbbfd5002	generic_instance_lineage	\N	eec2f323-46ea-4380-8c58-4b1065f85ad5	GL569	\N	\N	daylily	2025-07-03 08:23:17.836604-07	INSERT	\N	\N	\N	f	f
d88516d2-debb-413b-983f-5d5a53f87edd	generic_instance	\N	9aea1a13-4383-4c3d-b24b-aad55d4e8111	MRX209	\N	\N	daylily	2025-07-03 08:23:17.836604-07	INSERT	\N	\N	\N	f	f
a8b0c8c5-8a93-47a3-8e0f-70af022cc2bc	generic_instance_lineage	\N	554e32b1-9710-41fe-b057-84dfe55c8ab0	GL570	\N	\N	daylily	2025-07-03 08:23:17.842567-07	INSERT	\N	\N	\N	f	f
72036011-75c0-419b-86c3-14596c612ceb	generic_instance	\N	604c069f-71dd-4ee8-b4e8-ef6e7bb46e5c	MRX210	\N	\N	daylily	2025-07-03 08:23:17.842567-07	INSERT	\N	\N	\N	f	f
7b6935b4-2de6-4746-a6f9-655bff32c8f2	generic_instance	json_addl	843a9ccc-36ca-4d8b-80c8-a38487e0a12d	CWX225	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:15.683175-07	UPDATE	\N	\N	\N	f	f
309704c9-b765-4abc-8dbc-9706f5311f15	generic_instance	created_dt	843a9ccc-36ca-4d8b-80c8-a38487e0a12d	CWX225	2025-07-03 08:24:15.677773-07	2025-07-03T08:24:15.677773-07:00	daylily	2025-07-03 08:24:15.683175-07	UPDATE	\N	\N	\N	f	f
fd9c00c1-ac11-42a0-9e93-56b7cc56e836	generic_instance	modified_dt	843a9ccc-36ca-4d8b-80c8-a38487e0a12d	CWX225	2025-07-03 08:24:15.677773-07	2025-07-03T08:24:15.683175-07:00	daylily	2025-07-03 08:24:15.683175-07	UPDATE	\N	\N	\N	f	f
5cd8bc40-4f9c-49f2-9dad-e91d4564096e	generic_instance_lineage	\N	5a3319de-98fc-4251-acae-a6406be2d8db	GL589	\N	\N	daylily	2025-07-03 08:24:15.690826-07	INSERT	\N	\N	\N	f	f
5aa20e09-3409-4eaf-bc23-02808bbfff32	generic_instance	\N	37a7fae4-216d-4842-ab8f-5c048d4a2809	CWX226	\N	\N	daylily	2025-07-03 08:24:15.690826-07	INSERT	\N	\N	\N	f	f
79e6a93c-e293-4959-b797-019f6a68f9c0	generic_instance	json_addl	37a7fae4-216d-4842-ab8f-5c048d4a2809	CWX226	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:15.702113-07	UPDATE	\N	\N	\N	f	f
975d1cfd-5d71-4f66-88c3-d73c0ce2e68a	generic_instance	created_dt	37a7fae4-216d-4842-ab8f-5c048d4a2809	CWX226	2025-07-03 08:24:15.690826-07	2025-07-03T08:24:15.690826-07:00	daylily	2025-07-03 08:24:15.702113-07	UPDATE	\N	\N	\N	f	f
dc29aacf-cf70-4628-8d48-7fb4d6715364	generic_instance	modified_dt	37a7fae4-216d-4842-ab8f-5c048d4a2809	CWX226	2025-07-03 08:24:15.690826-07	2025-07-03T08:24:15.702113-07:00	daylily	2025-07-03 08:24:15.702113-07	UPDATE	\N	\N	\N	f	f
209939f1-b442-448f-b34f-969cfdf98b31	generic_instance_lineage	\N	f1be33a3-0cc2-4b05-90fc-4d8a05d08c16	GL590	\N	\N	daylily	2025-07-03 08:24:15.704863-07	INSERT	\N	\N	\N	f	f
ee7d5400-8ca1-4912-b715-a9778d55225f	generic_instance	\N	109675e8-0498-4c1c-bad5-788a996e69a2	CWX227	\N	\N	daylily	2025-07-03 08:24:15.704863-07	INSERT	\N	\N	\N	f	f
57832550-5b96-440e-bdf2-850056148914	generic_instance	json_addl	109675e8-0498-4c1c-bad5-788a996e69a2	CWX227	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:15.710942-07	UPDATE	\N	\N	\N	f	f
6ce0648c-0b76-4900-9391-ca80041fe698	generic_instance	created_dt	109675e8-0498-4c1c-bad5-788a996e69a2	CWX227	2025-07-03 08:24:15.704863-07	2025-07-03T08:24:15.704863-07:00	daylily	2025-07-03 08:24:15.710942-07	UPDATE	\N	\N	\N	f	f
9ba81522-31ae-4e4f-8b45-a9e833843b41	generic_instance	modified_dt	109675e8-0498-4c1c-bad5-788a996e69a2	CWX227	2025-07-03 08:24:15.704863-07	2025-07-03T08:24:15.710942-07:00	daylily	2025-07-03 08:24:15.710942-07	UPDATE	\N	\N	\N	f	f
aedaa61b-52e6-4b0f-a9b3-df71817b3a26	generic_instance_lineage	\N	7a924f25-8fc4-499a-98f6-05a8f45a7c52	GL591	\N	\N	daylily	2025-07-03 08:24:15.712473-07	INSERT	\N	\N	\N	f	f
7ddf23b3-6278-418d-b096-5908a0e94e5f	generic_instance	\N	20febaa0-790b-4079-8124-3d2d9586d49e	CWX228	\N	\N	daylily	2025-07-03 08:24:15.712473-07	INSERT	\N	\N	\N	f	f
28406757-9eb4-4514-bc9c-afb3e3ce27ec	generic_instance	json_addl	20febaa0-790b-4079-8124-3d2d9586d49e	CWX228	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:15.721173-07	UPDATE	\N	\N	\N	f	f
104113f6-a7d9-41af-ac75-efc556a3b9cc	generic_instance	created_dt	20febaa0-790b-4079-8124-3d2d9586d49e	CWX228	2025-07-03 08:24:15.712473-07	2025-07-03T08:24:15.712473-07:00	daylily	2025-07-03 08:24:15.721173-07	UPDATE	\N	\N	\N	f	f
563d9e0e-58e7-49d6-8ada-7cf7322f50cb	generic_instance	modified_dt	20febaa0-790b-4079-8124-3d2d9586d49e	CWX228	2025-07-03 08:24:15.712473-07	2025-07-03T08:24:15.721173-07:00	daylily	2025-07-03 08:24:15.721173-07	UPDATE	\N	\N	\N	f	f
d911d1c7-62c9-46f1-9475-cd9bbf3b0399	generic_instance_lineage	\N	2c9f09cb-5300-4d3c-b98d-662a63a2eb52	GL592	\N	\N	daylily	2025-07-03 08:24:15.723773-07	INSERT	\N	\N	\N	f	f
78eda6b7-3335-468b-8b2f-78fd1378b15f	generic_instance	\N	a0398a18-4ab8-469e-9010-55fda1de05b0	CWX229	\N	\N	daylily	2025-07-03 08:24:15.723773-07	INSERT	\N	\N	\N	f	f
39ee885b-65d4-4610-95d1-4a5c56f4994a	generic_instance	json_addl	a0398a18-4ab8-469e-9010-55fda1de05b0	CWX229	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:15.732297-07	UPDATE	\N	\N	\N	f	f
9893f5e1-0cce-42c8-973b-432cfe85a716	generic_instance	created_dt	a0398a18-4ab8-469e-9010-55fda1de05b0	CWX229	2025-07-03 08:24:15.723773-07	2025-07-03T08:24:15.723773-07:00	daylily	2025-07-03 08:24:15.732297-07	UPDATE	\N	\N	\N	f	f
003530de-cdaf-4090-9ed3-d43bb455e95f	generic_instance	modified_dt	a0398a18-4ab8-469e-9010-55fda1de05b0	CWX229	2025-07-03 08:24:15.723773-07	2025-07-03T08:24:15.732297-07:00	daylily	2025-07-03 08:24:15.732297-07	UPDATE	\N	\N	\N	f	f
a2b83336-8a30-4acf-8674-8ee7a2557b14	generic_instance_lineage	\N	66664f7c-b45c-4109-a12b-442de26397ce	GL593	\N	\N	daylily	2025-07-03 08:24:15.734602-07	INSERT	\N	\N	\N	f	f
9021db33-11d6-4a35-8bd7-c8c29c76cf39	generic_instance	\N	4ba3e341-5121-4322-bc64-7b192e496768	CWX230	\N	\N	daylily	2025-07-03 08:24:15.734602-07	INSERT	\N	\N	\N	f	f
0360cf30-20f0-4682-889e-94d770163468	generic_instance	json_addl	4ba3e341-5121-4322-bc64-7b192e496768	CWX230	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:15.743868-07	UPDATE	\N	\N	\N	f	f
d1dd63af-4829-4aca-b624-6c63cd5be095	generic_instance	created_dt	4ba3e341-5121-4322-bc64-7b192e496768	CWX230	2025-07-03 08:24:15.734602-07	2025-07-03T08:24:15.734602-07:00	daylily	2025-07-03 08:24:15.743868-07	UPDATE	\N	\N	\N	f	f
f55040f6-a8ef-4984-a6ec-260af4e47307	generic_instance	modified_dt	4ba3e341-5121-4322-bc64-7b192e496768	CWX230	2025-07-03 08:24:15.734602-07	2025-07-03T08:24:15.743868-07:00	daylily	2025-07-03 08:24:15.743868-07	UPDATE	\N	\N	\N	f	f
11a1ad7e-4957-4c33-b026-94dd83cd2d6b	generic_instance_lineage	\N	136d6f57-3093-4fd7-af0a-a7bc175b96f6	GL594	\N	\N	daylily	2025-07-03 08:24:15.746269-07	INSERT	\N	\N	\N	f	f
cae5d494-68f1-4dbf-afbc-9f25a3c80ca7	generic_instance	\N	c478dfca-9c72-4f15-93f1-fe3efd6274d1	CWX231	\N	\N	daylily	2025-07-03 08:24:15.746269-07	INSERT	\N	\N	\N	f	f
00893a81-988a-4a3f-a1a0-7ad107ea689f	generic_instance	json_addl	c478dfca-9c72-4f15-93f1-fe3efd6274d1	CWX231	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:15.755799-07	UPDATE	\N	\N	\N	f	f
e7dbe44b-d959-4026-85c4-ab469908c817	generic_instance	created_dt	c478dfca-9c72-4f15-93f1-fe3efd6274d1	CWX231	2025-07-03 08:24:15.746269-07	2025-07-03T08:24:15.746269-07:00	daylily	2025-07-03 08:24:15.755799-07	UPDATE	\N	\N	\N	f	f
3e148fdf-f3a3-41b7-9cb0-9b3d02b85b61	generic_instance	modified_dt	c478dfca-9c72-4f15-93f1-fe3efd6274d1	CWX231	2025-07-03 08:24:15.746269-07	2025-07-03T08:24:15.755799-07:00	daylily	2025-07-03 08:24:15.755799-07	UPDATE	\N	\N	\N	f	f
a24afe9f-4adb-486f-a481-59e2ea35c1ce	generic_instance_lineage	\N	442c993b-ad25-4467-acc7-acd96b0e797f	GL595	\N	\N	daylily	2025-07-03 08:24:15.759711-07	INSERT	\N	\N	\N	f	f
9094c5cf-fd75-4ed7-b87e-710e44f27766	generic_instance	\N	d4fbc620-8b0d-4f35-b5f0-0b2fe2f00571	CWX232	\N	\N	daylily	2025-07-03 08:24:15.759711-07	INSERT	\N	\N	\N	f	f
000f8644-b672-4126-a37a-1d5c85b25776	generic_instance	json_addl	d4fbc620-8b0d-4f35-b5f0-0b2fe2f00571	CWX232	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:15.768378-07	UPDATE	\N	\N	\N	f	f
aa185e66-6fd7-43f1-9510-0b8d1e1026cb	generic_instance	created_dt	d4fbc620-8b0d-4f35-b5f0-0b2fe2f00571	CWX232	2025-07-03 08:24:15.759711-07	2025-07-03T08:24:15.759711-07:00	daylily	2025-07-03 08:24:15.768378-07	UPDATE	\N	\N	\N	f	f
a9ab5c15-fb08-4f23-b6dc-38b9043c064a	generic_instance	modified_dt	d4fbc620-8b0d-4f35-b5f0-0b2fe2f00571	CWX232	2025-07-03 08:24:15.759711-07	2025-07-03T08:24:15.768378-07:00	daylily	2025-07-03 08:24:15.768378-07	UPDATE	\N	\N	\N	f	f
b2b96d21-e9a9-4bda-b302-2555d070ca77	generic_instance_lineage	\N	64995098-fc5c-4cd7-bf45-d693f9e3b958	GL596	\N	\N	daylily	2025-07-03 08:24:15.771546-07	INSERT	\N	\N	\N	f	f
523ecbd9-ca8b-4749-bffd-469587b1eb9e	generic_instance	\N	5c89cd1b-c0d0-4846-896a-ceca2f29fcb6	CWX233	\N	\N	daylily	2025-07-03 08:24:15.771546-07	INSERT	\N	\N	\N	f	f
2b89734b-48f3-4ce3-8964-dcfd4e450f3b	generic_instance	json_addl	5c89cd1b-c0d0-4846-896a-ceca2f29fcb6	CWX233	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:15.777077-07	UPDATE	\N	\N	\N	f	f
c145065b-af23-4138-9160-0a665093803e	generic_instance	created_dt	5c89cd1b-c0d0-4846-896a-ceca2f29fcb6	CWX233	2025-07-03 08:24:15.771546-07	2025-07-03T08:24:15.771546-07:00	daylily	2025-07-03 08:24:15.777077-07	UPDATE	\N	\N	\N	f	f
0ba0c5b4-8951-4599-904a-2e0882fcd802	generic_instance	modified_dt	5c89cd1b-c0d0-4846-896a-ceca2f29fcb6	CWX233	2025-07-03 08:24:15.771546-07	2025-07-03T08:24:15.777077-07:00	daylily	2025-07-03 08:24:15.777077-07	UPDATE	\N	\N	\N	f	f
b73aa0d7-ab16-498f-b819-e3468026cafc	generic_instance_lineage	\N	6a836416-bb10-49b4-b5a5-496236afd94d	GL597	\N	\N	daylily	2025-07-03 08:24:15.778732-07	INSERT	\N	\N	\N	f	f
35d2dc66-12f0-4bb0-8f28-908f41c919f9	generic_instance	\N	5fe3fa73-82ec-4a72-a25d-f4ca3f7575d8	CWX234	\N	\N	daylily	2025-07-03 08:24:15.778732-07	INSERT	\N	\N	\N	f	f
f3fe73b8-dbfa-4718-90e4-16147418f329	generic_instance	json_addl	5fe3fa73-82ec-4a72-a25d-f4ca3f7575d8	CWX234	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:15.784492-07	UPDATE	\N	\N	\N	f	f
9042f2dc-ca10-49d7-9a3b-c52fd7446dbd	generic_instance	created_dt	5fe3fa73-82ec-4a72-a25d-f4ca3f7575d8	CWX234	2025-07-03 08:24:15.778732-07	2025-07-03T08:24:15.778732-07:00	daylily	2025-07-03 08:24:15.784492-07	UPDATE	\N	\N	\N	f	f
4b6dccd0-1afe-4adf-a3a9-39fae1527da1	generic_instance	modified_dt	5fe3fa73-82ec-4a72-a25d-f4ca3f7575d8	CWX234	2025-07-03 08:24:15.778732-07	2025-07-03T08:24:15.784492-07:00	daylily	2025-07-03 08:24:15.784492-07	UPDATE	\N	\N	\N	f	f
f79adfa3-5fb1-44d9-9e94-7fda285bc3b1	generic_instance_lineage	\N	81fdfe0f-4a70-4c28-ac45-885487399eca	GL598	\N	\N	daylily	2025-07-03 08:24:15.786514-07	INSERT	\N	\N	\N	f	f
d526bc4b-e22f-4040-89d5-4d98e751af81	generic_instance	\N	015fbb50-97f2-409e-9b39-e5f548a21402	CWX235	\N	\N	daylily	2025-07-03 08:24:15.786514-07	INSERT	\N	\N	\N	f	f
b784178b-8054-4334-9537-7a6553eb3f8c	generic_instance	json_addl	015fbb50-97f2-409e-9b39-e5f548a21402	CWX235	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:15.79204-07	UPDATE	\N	\N	\N	f	f
a14adafd-e733-499c-b0a1-d4edcddb4442	generic_instance	created_dt	015fbb50-97f2-409e-9b39-e5f548a21402	CWX235	2025-07-03 08:24:15.786514-07	2025-07-03T08:24:15.786514-07:00	daylily	2025-07-03 08:24:15.79204-07	UPDATE	\N	\N	\N	f	f
697d4304-0655-42cb-9d06-dbef00781591	generic_instance	modified_dt	015fbb50-97f2-409e-9b39-e5f548a21402	CWX235	2025-07-03 08:24:15.786514-07	2025-07-03T08:24:15.79204-07:00	daylily	2025-07-03 08:24:15.79204-07	UPDATE	\N	\N	\N	f	f
d95c76ea-11a2-44c3-b3d8-a5f96997d17c	generic_instance_lineage	\N	ace327c9-a487-436d-859b-ccc19cbb11f2	GL599	\N	\N	daylily	2025-07-03 08:24:15.793739-07	INSERT	\N	\N	\N	f	f
bda8829f-7d50-43ee-a808-87f1fe551a5f	generic_instance	\N	427711ca-ab90-45b5-b311-8733db30d9ad	CWX236	\N	\N	daylily	2025-07-03 08:24:15.793739-07	INSERT	\N	\N	\N	f	f
a1fa020f-7416-4771-b0ae-0182db459590	generic_instance	json_addl	427711ca-ab90-45b5-b311-8733db30d9ad	CWX236	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:15.799421-07	UPDATE	\N	\N	\N	f	f
7294e2b8-e74a-4c1e-8c0b-f8f9813bddad	generic_instance	created_dt	427711ca-ab90-45b5-b311-8733db30d9ad	CWX236	2025-07-03 08:24:15.793739-07	2025-07-03T08:24:15.793739-07:00	daylily	2025-07-03 08:24:15.799421-07	UPDATE	\N	\N	\N	f	f
ae11645e-269a-4030-b611-da81bcae712f	generic_instance	modified_dt	427711ca-ab90-45b5-b311-8733db30d9ad	CWX236	2025-07-03 08:24:15.793739-07	2025-07-03T08:24:15.799421-07:00	daylily	2025-07-03 08:24:15.799421-07	UPDATE	\N	\N	\N	f	f
33165b7c-79f2-4ee1-8106-ca7db6d20395	generic_instance_lineage	\N	7eda35a5-c22f-4eea-a5d5-c9ca6422577e	GL600	\N	\N	daylily	2025-07-03 08:24:15.801168-07	INSERT	\N	\N	\N	f	f
226083d8-ef44-4804-a757-187ea7346c88	generic_instance	\N	95d4d6ba-7610-486b-a74c-d1ad8255621a	CWX237	\N	\N	daylily	2025-07-03 08:24:15.801168-07	INSERT	\N	\N	\N	f	f
cdc13e67-752e-4ed3-bcec-5f05fc48b366	generic_instance	json_addl	95d4d6ba-7610-486b-a74c-d1ad8255621a	CWX237	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:15.814182-07	UPDATE	\N	\N	\N	f	f
faeb7646-9cd6-4e65-876b-d964d26c087e	generic_instance	created_dt	95d4d6ba-7610-486b-a74c-d1ad8255621a	CWX237	2025-07-03 08:24:15.801168-07	2025-07-03T08:24:15.801168-07:00	daylily	2025-07-03 08:24:15.814182-07	UPDATE	\N	\N	\N	f	f
4e876273-8835-4a16-bee9-ea532a58dc6c	generic_instance	modified_dt	95d4d6ba-7610-486b-a74c-d1ad8255621a	CWX237	2025-07-03 08:24:15.801168-07	2025-07-03T08:24:15.814182-07:00	daylily	2025-07-03 08:24:15.814182-07	UPDATE	\N	\N	\N	f	f
669189ae-4281-4cfa-94a4-d2b2ed6144cf	generic_instance_lineage	\N	88920470-3dea-40a5-95cd-e345224eecf6	GL601	\N	\N	daylily	2025-07-03 08:24:15.816498-07	INSERT	\N	\N	\N	f	f
d71c2fbd-8efb-4602-870a-138d6704680f	generic_instance	\N	61bfac74-3f5c-474c-8b7a-6fe48898e386	CWX238	\N	\N	daylily	2025-07-03 08:24:15.816498-07	INSERT	\N	\N	\N	f	f
3d343484-9f3f-4005-ac16-6f932653d426	generic_instance	json_addl	61bfac74-3f5c-474c-8b7a-6fe48898e386	CWX238	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:15.82291-07	UPDATE	\N	\N	\N	f	f
065efdc6-e398-4af7-9a85-a98d2afec21c	generic_instance	created_dt	61bfac74-3f5c-474c-8b7a-6fe48898e386	CWX238	2025-07-03 08:24:15.816498-07	2025-07-03T08:24:15.816498-07:00	daylily	2025-07-03 08:24:15.82291-07	UPDATE	\N	\N	\N	f	f
3b0ee5a4-dfe1-46f9-9eec-0f0d97910e55	generic_instance	modified_dt	61bfac74-3f5c-474c-8b7a-6fe48898e386	CWX238	2025-07-03 08:24:15.816498-07	2025-07-03T08:24:15.82291-07:00	daylily	2025-07-03 08:24:15.82291-07	UPDATE	\N	\N	\N	f	f
f43b5346-5e5a-4759-b9e4-6ca7ea1127e0	generic_instance_lineage	\N	c1a191a2-5708-4838-8c42-518388d88bb4	GL602	\N	\N	daylily	2025-07-03 08:24:15.824383-07	INSERT	\N	\N	\N	f	f
796a787f-607b-4436-94b7-b5c610dc2dd4	generic_instance	\N	a1f6ccdc-9ee5-402b-a93d-533b39212bee	CWX239	\N	\N	daylily	2025-07-03 08:24:15.824383-07	INSERT	\N	\N	\N	f	f
ffe46ff4-6cb9-465a-870d-d2a3743aacb1	generic_instance	json_addl	a1f6ccdc-9ee5-402b-a93d-533b39212bee	CWX239	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:15.833736-07	UPDATE	\N	\N	\N	f	f
83f60827-d428-406e-b2fc-35516e035f67	generic_instance	created_dt	a1f6ccdc-9ee5-402b-a93d-533b39212bee	CWX239	2025-07-03 08:24:15.824383-07	2025-07-03T08:24:15.824383-07:00	daylily	2025-07-03 08:24:15.833736-07	UPDATE	\N	\N	\N	f	f
5c2fbfc9-2cc5-4b99-b80d-e455ee24a4ba	generic_instance	modified_dt	a1f6ccdc-9ee5-402b-a93d-533b39212bee	CWX239	2025-07-03 08:24:15.824383-07	2025-07-03T08:24:15.833736-07:00	daylily	2025-07-03 08:24:15.833736-07	UPDATE	\N	\N	\N	f	f
aab3e547-cbae-4227-9637-b68bba0498b7	generic_instance_lineage	\N	f08fc12e-74c2-4b22-ae8b-73110944312f	GL603	\N	\N	daylily	2025-07-03 08:24:15.83564-07	INSERT	\N	\N	\N	f	f
7ce726a9-35d8-45cf-81ac-73c900d1f8b3	generic_instance	\N	4ff5edda-1503-4a1b-aa81-44f0d001dede	CWX240	\N	\N	daylily	2025-07-03 08:24:15.83564-07	INSERT	\N	\N	\N	f	f
ac4e57d5-9529-425c-9aea-f78ff6e5f60f	generic_instance	json_addl	4ff5edda-1503-4a1b-aa81-44f0d001dede	CWX240	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:15.841215-07	UPDATE	\N	\N	\N	f	f
17bc5e60-d4ab-472e-8d66-f308e17597a1	generic_instance	created_dt	4ff5edda-1503-4a1b-aa81-44f0d001dede	CWX240	2025-07-03 08:24:15.83564-07	2025-07-03T08:24:15.83564-07:00	daylily	2025-07-03 08:24:15.841215-07	UPDATE	\N	\N	\N	f	f
a0c53442-3641-4059-a76a-ba3a3e9b8976	generic_instance	modified_dt	4ff5edda-1503-4a1b-aa81-44f0d001dede	CWX240	2025-07-03 08:24:15.83564-07	2025-07-03T08:24:15.841215-07:00	daylily	2025-07-03 08:24:15.841215-07	UPDATE	\N	\N	\N	f	f
efe18451-f722-4fe6-ad7c-029afc730261	generic_instance_lineage	\N	b7fb325c-b99c-4d31-91df-32d531de33ae	GL604	\N	\N	daylily	2025-07-03 08:24:15.84285-07	INSERT	\N	\N	\N	f	f
3ff05530-941d-4509-911e-2f1203cecc7f	generic_instance	\N	c60228c8-099b-47f1-a890-387453eec26b	CWX241	\N	\N	daylily	2025-07-03 08:24:15.84285-07	INSERT	\N	\N	\N	f	f
caf13e35-eb9f-4d33-b299-b3e76cfde1cc	generic_instance	json_addl	c60228c8-099b-47f1-a890-387453eec26b	CWX241	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:15.848337-07	UPDATE	\N	\N	\N	f	f
0d466f1b-9d78-4394-a352-c84e51dc6a8c	generic_instance	created_dt	c60228c8-099b-47f1-a890-387453eec26b	CWX241	2025-07-03 08:24:15.84285-07	2025-07-03T08:24:15.84285-07:00	daylily	2025-07-03 08:24:15.848337-07	UPDATE	\N	\N	\N	f	f
9c3e6e68-9ca1-4375-ae83-b54ed01cb210	generic_instance	modified_dt	c60228c8-099b-47f1-a890-387453eec26b	CWX241	2025-07-03 08:24:15.84285-07	2025-07-03T08:24:15.848337-07:00	daylily	2025-07-03 08:24:15.848337-07	UPDATE	\N	\N	\N	f	f
19bed65e-0736-436a-b51b-6714e025a252	generic_instance_lineage	\N	dc72e345-5707-47fb-9bee-5003ef2dd689	GL605	\N	\N	daylily	2025-07-03 08:24:15.850004-07	INSERT	\N	\N	\N	f	f
e0842548-e567-4e53-90c2-5cbc7fb17368	generic_instance	\N	7833d79c-9c5d-4177-b4ed-b29293438cc4	CWX242	\N	\N	daylily	2025-07-03 08:24:15.850004-07	INSERT	\N	\N	\N	f	f
947026c8-1758-4840-b545-ccba2ae76e6c	generic_instance	json_addl	7833d79c-9c5d-4177-b4ed-b29293438cc4	CWX242	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:15.855556-07	UPDATE	\N	\N	\N	f	f
1e166ac0-0a8d-4645-a83f-fb38238d5fea	generic_instance	created_dt	7833d79c-9c5d-4177-b4ed-b29293438cc4	CWX242	2025-07-03 08:24:15.850004-07	2025-07-03T08:24:15.850004-07:00	daylily	2025-07-03 08:24:15.855556-07	UPDATE	\N	\N	\N	f	f
c4a12f9e-d459-4481-af8d-27aa84b90da5	generic_instance	modified_dt	7833d79c-9c5d-4177-b4ed-b29293438cc4	CWX242	2025-07-03 08:24:15.850004-07	2025-07-03T08:24:15.855556-07:00	daylily	2025-07-03 08:24:15.855556-07	UPDATE	\N	\N	\N	f	f
8c2df5f1-9bec-4db3-8b48-ae9967199e90	generic_instance_lineage	\N	8528102b-a975-4278-bb42-db6feccbfc84	GL606	\N	\N	daylily	2025-07-03 08:24:15.857224-07	INSERT	\N	\N	\N	f	f
60793672-70b9-49f7-b099-e2b2e5678d2b	generic_instance	\N	85859176-cfcb-4ba9-8078-981f2c92c5f7	CWX243	\N	\N	daylily	2025-07-03 08:24:15.857224-07	INSERT	\N	\N	\N	f	f
5bc4c1b3-47bb-4e7c-a96a-350c704ab3b4	generic_instance	json_addl	85859176-cfcb-4ba9-8078-981f2c92c5f7	CWX243	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:15.862723-07	UPDATE	\N	\N	\N	f	f
5c4ab323-4c46-4506-bc4e-258f739ffef9	generic_instance	created_dt	85859176-cfcb-4ba9-8078-981f2c92c5f7	CWX243	2025-07-03 08:24:15.857224-07	2025-07-03T08:24:15.857224-07:00	daylily	2025-07-03 08:24:15.862723-07	UPDATE	\N	\N	\N	f	f
412a9924-4ef7-4f1d-8efa-248494c0a707	generic_instance	modified_dt	85859176-cfcb-4ba9-8078-981f2c92c5f7	CWX243	2025-07-03 08:24:15.857224-07	2025-07-03T08:24:15.862723-07:00	daylily	2025-07-03 08:24:15.862723-07	UPDATE	\N	\N	\N	f	f
d85242c1-2ec6-4055-84d5-98eddd07bee6	generic_instance_lineage	\N	4ba073c0-9fda-4718-ac23-fd5e4bfbc5c4	GL607	\N	\N	daylily	2025-07-03 08:24:15.864295-07	INSERT	\N	\N	\N	f	f
f5ea0ce8-f45f-41f1-bcfa-d8fbbdc9da9a	generic_instance	\N	fca6c1f5-1e60-4b0e-b8dc-de2108eb584c	CWX244	\N	\N	daylily	2025-07-03 08:24:15.864295-07	INSERT	\N	\N	\N	f	f
2f7f74eb-9f9d-4762-a6b4-b8bbc29bd4f5	generic_instance	json_addl	fca6c1f5-1e60-4b0e-b8dc-de2108eb584c	CWX244	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:15.870121-07	UPDATE	\N	\N	\N	f	f
8cb438a9-48a5-4a06-b3e7-bc1ea1931730	generic_instance	created_dt	fca6c1f5-1e60-4b0e-b8dc-de2108eb584c	CWX244	2025-07-03 08:24:15.864295-07	2025-07-03T08:24:15.864295-07:00	daylily	2025-07-03 08:24:15.870121-07	UPDATE	\N	\N	\N	f	f
f467a8d1-65e8-48e3-9ed6-e0165840239f	generic_instance	modified_dt	fca6c1f5-1e60-4b0e-b8dc-de2108eb584c	CWX244	2025-07-03 08:24:15.864295-07	2025-07-03T08:24:15.870121-07:00	daylily	2025-07-03 08:24:15.870121-07	UPDATE	\N	\N	\N	f	f
2c9c3755-b9b8-49cf-98e9-ae43ac63332b	generic_instance_lineage	\N	85903bd7-3177-4463-885e-78e4db0c1929	GL608	\N	\N	daylily	2025-07-03 08:24:15.872125-07	INSERT	\N	\N	\N	f	f
61f6fc97-05de-4fb5-84b5-32228851f372	generic_instance	\N	11d71ea0-89ac-4aad-89f4-68ba3f60525d	CWX245	\N	\N	daylily	2025-07-03 08:24:15.872125-07	INSERT	\N	\N	\N	f	f
2aad8dda-80e3-4c79-84ff-7a2308d1a6da	generic_instance	json_addl	11d71ea0-89ac-4aad-89f4-68ba3f60525d	CWX245	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:15.883029-07	UPDATE	\N	\N	\N	f	f
a0e4252d-4c9f-4659-bb16-df3f89a93c07	generic_instance	created_dt	11d71ea0-89ac-4aad-89f4-68ba3f60525d	CWX245	2025-07-03 08:24:15.872125-07	2025-07-03T08:24:15.872125-07:00	daylily	2025-07-03 08:24:15.883029-07	UPDATE	\N	\N	\N	f	f
72ac7965-8425-4b7a-b51f-43bdf13127d7	generic_instance	modified_dt	11d71ea0-89ac-4aad-89f4-68ba3f60525d	CWX245	2025-07-03 08:24:15.872125-07	2025-07-03T08:24:15.883029-07:00	daylily	2025-07-03 08:24:15.883029-07	UPDATE	\N	\N	\N	f	f
9e877119-1cdc-43a4-b603-579cd34e1d4d	generic_instance_lineage	\N	3adb2680-c576-4c53-a5b9-0ca84a315b80	GL609	\N	\N	daylily	2025-07-03 08:24:15.884775-07	INSERT	\N	\N	\N	f	f
89a60c3f-c311-42b5-a6b2-bbe16d59d0a9	generic_instance	\N	b5c6c063-303c-4bf1-a0b4-d952763bd8fc	CWX246	\N	\N	daylily	2025-07-03 08:24:15.884775-07	INSERT	\N	\N	\N	f	f
761a7a2d-41da-4af8-bdc0-52a84298cfd0	generic_instance	json_addl	b5c6c063-303c-4bf1-a0b4-d952763bd8fc	CWX246	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:15.891088-07	UPDATE	\N	\N	\N	f	f
d0fed34b-4374-48bb-90bb-544ca4e11bf5	generic_instance	created_dt	b5c6c063-303c-4bf1-a0b4-d952763bd8fc	CWX246	2025-07-03 08:24:15.884775-07	2025-07-03T08:24:15.884775-07:00	daylily	2025-07-03 08:24:15.891088-07	UPDATE	\N	\N	\N	f	f
951b7fd5-3fc6-4915-89f9-fdb5a66d22b4	generic_instance	modified_dt	b5c6c063-303c-4bf1-a0b4-d952763bd8fc	CWX246	2025-07-03 08:24:15.884775-07	2025-07-03T08:24:15.891088-07:00	daylily	2025-07-03 08:24:15.891088-07	UPDATE	\N	\N	\N	f	f
2b02688d-2dd5-4569-9538-527a76f60942	generic_instance_lineage	\N	acf08066-ab9c-4a70-8d4c-9ecc63390409	GL610	\N	\N	daylily	2025-07-03 08:24:15.894459-07	INSERT	\N	\N	\N	f	f
8a518ee3-9131-4970-aefd-1e3e984dc73e	generic_instance	\N	d7616aa4-1790-453e-97cd-801233b1bc1e	CWX247	\N	\N	daylily	2025-07-03 08:24:15.894459-07	INSERT	\N	\N	\N	f	f
754345ac-ceee-4e3c-9468-1241c9fe5024	generic_instance	json_addl	d7616aa4-1790-453e-97cd-801233b1bc1e	CWX247	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:15.900439-07	UPDATE	\N	\N	\N	f	f
9dc7a872-e1db-48bc-97c9-0f486427d6b5	generic_instance	created_dt	d7616aa4-1790-453e-97cd-801233b1bc1e	CWX247	2025-07-03 08:24:15.894459-07	2025-07-03T08:24:15.894459-07:00	daylily	2025-07-03 08:24:15.900439-07	UPDATE	\N	\N	\N	f	f
d77ea0af-1143-4551-821a-18d87958f500	generic_instance	modified_dt	d7616aa4-1790-453e-97cd-801233b1bc1e	CWX247	2025-07-03 08:24:15.894459-07	2025-07-03T08:24:15.900439-07:00	daylily	2025-07-03 08:24:15.900439-07	UPDATE	\N	\N	\N	f	f
17eb0fc2-11aa-488b-92bb-bcd09a8e65bb	generic_instance_lineage	\N	a8535f8f-fa7e-473c-a2b4-30b0a1612da5	GL611	\N	\N	daylily	2025-07-03 08:24:15.901925-07	INSERT	\N	\N	\N	f	f
57f9a8c6-98aa-4b03-b17e-749d235c46a4	generic_instance	\N	b9068f68-df90-4cc8-be11-f651be061b2a	CWX248	\N	\N	daylily	2025-07-03 08:24:15.901925-07	INSERT	\N	\N	\N	f	f
9f5bc13b-2306-4a81-afac-7bf54a0779a5	generic_instance	json_addl	b9068f68-df90-4cc8-be11-f651be061b2a	CWX248	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:15.909075-07	UPDATE	\N	\N	\N	f	f
cbb9417c-160c-472e-82c5-ed3504fd5d72	generic_instance	created_dt	b9068f68-df90-4cc8-be11-f651be061b2a	CWX248	2025-07-03 08:24:15.901925-07	2025-07-03T08:24:15.901925-07:00	daylily	2025-07-03 08:24:15.909075-07	UPDATE	\N	\N	\N	f	f
ebe9765d-fdc6-42c6-a4c8-ec523f9434d7	generic_instance	modified_dt	b9068f68-df90-4cc8-be11-f651be061b2a	CWX248	2025-07-03 08:24:15.901925-07	2025-07-03T08:24:15.909075-07:00	daylily	2025-07-03 08:24:15.909075-07	UPDATE	\N	\N	\N	f	f
23b61670-31d2-4d35-be54-df5e0d6f638d	generic_instance_lineage	\N	6b1669e9-21a0-49ba-a9cc-0375fbb1994d	GL612	\N	\N	daylily	2025-07-03 08:24:15.911667-07	INSERT	\N	\N	\N	f	f
f1c96ed3-2b9f-4c73-84ad-3c8b0da375ca	generic_instance	\N	35280b86-54a8-4337-9df0-abcb0cd2c340	CWX249	\N	\N	daylily	2025-07-03 08:24:15.911667-07	INSERT	\N	\N	\N	f	f
fb92665a-5733-491f-8dec-aa16281c2e76	generic_instance	json_addl	35280b86-54a8-4337-9df0-abcb0cd2c340	CWX249	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:15.91939-07	UPDATE	\N	\N	\N	f	f
4a672df5-ba84-44bc-8543-f6d8466a2e49	generic_instance	created_dt	35280b86-54a8-4337-9df0-abcb0cd2c340	CWX249	2025-07-03 08:24:15.911667-07	2025-07-03T08:24:15.911667-07:00	daylily	2025-07-03 08:24:15.91939-07	UPDATE	\N	\N	\N	f	f
be57fe35-2148-40fd-886f-05cba39a9780	generic_instance	modified_dt	35280b86-54a8-4337-9df0-abcb0cd2c340	CWX249	2025-07-03 08:24:15.911667-07	2025-07-03T08:24:15.91939-07:00	daylily	2025-07-03 08:24:15.91939-07	UPDATE	\N	\N	\N	f	f
8866437b-2bce-4d02-8477-2b3d6a4378eb	generic_instance_lineage	\N	d7786da3-1950-47d6-87aa-96f83aa833db	GL613	\N	\N	daylily	2025-07-03 08:24:15.920899-07	INSERT	\N	\N	\N	f	f
85508a46-b832-4575-b5ed-ef48439eb709	generic_instance	\N	3f257485-e9ed-43ba-9245-63fe12620c6e	CWX250	\N	\N	daylily	2025-07-03 08:24:15.920899-07	INSERT	\N	\N	\N	f	f
af089bbc-7d7b-4bfb-8a29-00f947f39085	generic_instance	json_addl	3f257485-e9ed-43ba-9245-63fe12620c6e	CWX250	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:15.92644-07	UPDATE	\N	\N	\N	f	f
5894013a-c6ee-4d7a-b0b5-987584129421	generic_instance	created_dt	3f257485-e9ed-43ba-9245-63fe12620c6e	CWX250	2025-07-03 08:24:15.920899-07	2025-07-03T08:24:15.920899-07:00	daylily	2025-07-03 08:24:15.92644-07	UPDATE	\N	\N	\N	f	f
102e5cef-7424-4fda-af87-e4b601030368	generic_instance	modified_dt	3f257485-e9ed-43ba-9245-63fe12620c6e	CWX250	2025-07-03 08:24:15.920899-07	2025-07-03T08:24:15.92644-07:00	daylily	2025-07-03 08:24:15.92644-07	UPDATE	\N	\N	\N	f	f
b75599b2-81ff-43c2-bd14-7eebb60291fd	generic_instance_lineage	\N	96afcc22-b8db-4421-9597-d4dd2a8892fe	GL614	\N	\N	daylily	2025-07-03 08:24:15.929132-07	INSERT	\N	\N	\N	f	f
37f44a74-f8c9-4f2b-a78b-cf2125b6bf43	generic_instance	\N	c9afb65c-3bce-41aa-94ca-ed1c356a938c	CWX251	\N	\N	daylily	2025-07-03 08:24:15.929132-07	INSERT	\N	\N	\N	f	f
d8cbad26-7f62-425b-a90d-6ff816c44246	generic_instance	json_addl	c9afb65c-3bce-41aa-94ca-ed1c356a938c	CWX251	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:15.935348-07	UPDATE	\N	\N	\N	f	f
e651aee4-85de-4834-a0d6-733467dfd419	generic_instance	created_dt	c9afb65c-3bce-41aa-94ca-ed1c356a938c	CWX251	2025-07-03 08:24:15.929132-07	2025-07-03T08:24:15.929132-07:00	daylily	2025-07-03 08:24:15.935348-07	UPDATE	\N	\N	\N	f	f
e401fcfb-781f-4058-a470-5131793e5835	generic_instance	modified_dt	c9afb65c-3bce-41aa-94ca-ed1c356a938c	CWX251	2025-07-03 08:24:15.929132-07	2025-07-03T08:24:15.935348-07:00	daylily	2025-07-03 08:24:15.935348-07	UPDATE	\N	\N	\N	f	f
a72be874-f827-4572-9baa-c8f3a6b4cc0c	generic_instance_lineage	\N	870719ee-b6b5-4dbb-b2f4-bd30e3e96995	GL615	\N	\N	daylily	2025-07-03 08:24:15.938076-07	INSERT	\N	\N	\N	f	f
a3030dad-7af6-4723-af1d-82e49c322df3	generic_instance	\N	cca985cb-db90-47cb-880a-3749d3611984	CWX252	\N	\N	daylily	2025-07-03 08:24:15.938076-07	INSERT	\N	\N	\N	f	f
8a3e9ad7-f7c2-400b-bbd0-ebe7259fe1f3	generic_instance	json_addl	cca985cb-db90-47cb-880a-3749d3611984	CWX252	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:15.943407-07	UPDATE	\N	\N	\N	f	f
b516b8f1-3c20-4522-9acb-19a80549595c	generic_instance	created_dt	cca985cb-db90-47cb-880a-3749d3611984	CWX252	2025-07-03 08:24:15.938076-07	2025-07-03T08:24:15.938076-07:00	daylily	2025-07-03 08:24:15.943407-07	UPDATE	\N	\N	\N	f	f
0a3e025e-5729-43a3-b1ee-bc94d25ceb5e	generic_instance	modified_dt	cca985cb-db90-47cb-880a-3749d3611984	CWX252	2025-07-03 08:24:15.938076-07	2025-07-03T08:24:15.943407-07:00	daylily	2025-07-03 08:24:15.943407-07	UPDATE	\N	\N	\N	f	f
7fc86ce1-14ff-406b-9e8a-c27e76137d2f	generic_instance_lineage	\N	4342c3b1-5931-4411-a5ae-cca89b42c12b	GL616	\N	\N	daylily	2025-07-03 08:24:15.94856-07	INSERT	\N	\N	\N	f	f
cddee0e3-e6be-4514-9a50-62b523b6e16b	generic_instance	\N	3ed95fad-05dc-4da6-a86e-f9cf24a0215a	MRX225	\N	\N	daylily	2025-07-03 08:24:15.952949-07	INSERT	\N	\N	\N	f	f
b848a768-9598-495c-9590-895a86629438	generic_instance_lineage	\N	e9e3c63e-e74b-4260-8cdd-e5f89b3027c8	GL617	\N	\N	daylily	2025-07-03 08:24:15.956035-07	INSERT	\N	\N	\N	f	f
13caeaea-cdc0-4059-9d09-7cc17079e64d	generic_instance	\N	6b65a1a8-ba49-4cb5-b381-72a0c2ab0f7d	MRX226	\N	\N	daylily	2025-07-03 08:24:15.956035-07	INSERT	\N	\N	\N	f	f
e8bf53a8-1917-48ca-ae47-ce51de5773ed	generic_instance_lineage	\N	703807e9-20eb-4909-9b0f-529685666294	GL618	\N	\N	daylily	2025-07-03 08:24:15.963002-07	INSERT	\N	\N	\N	f	f
6ac192fa-3621-4376-8fee-aa2c102c88c0	generic_instance	\N	467ee04d-f443-49cf-b38d-090e047d1306	MRX227	\N	\N	daylily	2025-07-03 08:24:15.963002-07	INSERT	\N	\N	\N	f	f
dc87f110-fda0-4b54-9ea5-884528baff23	generic_instance_lineage	\N	fd55de73-a8a6-4e98-b1ad-bcf025eb76a6	GL619	\N	\N	daylily	2025-07-03 08:24:15.968746-07	INSERT	\N	\N	\N	f	f
ed5d013c-e90f-4d0e-bf2b-933a6ca4504b	generic_instance	\N	3c31b867-fb64-4aaa-b939-f7e98b8abb92	MRX228	\N	\N	daylily	2025-07-03 08:24:15.968746-07	INSERT	\N	\N	\N	f	f
d2cb6faf-fa20-4077-939e-aaa69cf226fd	generic_instance_lineage	\N	f92f9f48-c82d-4eba-b862-92b4d7c943fb	GL620	\N	\N	daylily	2025-07-03 08:24:15.9742-07	INSERT	\N	\N	\N	f	f
97d713ce-e59b-402b-be39-5109a5a08a5d	generic_instance	\N	94ba58bf-d6b4-4163-bd4d-a69c9e105a68	MRX229	\N	\N	daylily	2025-07-03 08:24:15.9742-07	INSERT	\N	\N	\N	f	f
79eae780-47da-4dfa-9e80-97d53ff47507	generic_instance_lineage	\N	265f4ea7-d290-49fb-a84c-2b862840fbb9	GL621	\N	\N	daylily	2025-07-03 08:24:15.979872-07	INSERT	\N	\N	\N	f	f
b9d17598-9d86-431b-b8c8-1947a695cad9	generic_instance	\N	fb4fba5f-dbd7-42d9-a5a3-7fe0cfca759d	MRX230	\N	\N	daylily	2025-07-03 08:24:15.979872-07	INSERT	\N	\N	\N	f	f
158f0306-7707-4c70-93a0-b1c57591e66f	generic_instance_lineage	\N	f804dfcd-db6b-4d6a-b8c9-3d619404ed21	GL622	\N	\N	daylily	2025-07-03 08:24:15.985312-07	INSERT	\N	\N	\N	f	f
a9d97fcc-18c8-4e15-b973-a2f34ad79c75	generic_instance	\N	c41f43b2-3170-48fd-97c3-304c73f09ad7	MRX231	\N	\N	daylily	2025-07-03 08:24:15.985312-07	INSERT	\N	\N	\N	f	f
7ff770f8-65b0-4d04-843e-22f6c90d6edd	generic_instance_lineage	\N	adbe5a56-c700-42a9-b860-03dddfa86bb2	GL623	\N	\N	daylily	2025-07-03 08:24:15.993143-07	INSERT	\N	\N	\N	f	f
01e5c83b-a545-47b5-bfa0-9e62c944a106	generic_instance	\N	91cccd1e-a078-41ca-bbb3-67985eaaf192	MRX232	\N	\N	daylily	2025-07-03 08:24:15.993143-07	INSERT	\N	\N	\N	f	f
71f5b2da-366b-44b2-ae49-b3f3949cbc77	generic_instance_lineage	\N	63ddbe12-b72a-4a5b-9f84-dfa561b0777c	GL624	\N	\N	daylily	2025-07-03 08:24:15.998517-07	INSERT	\N	\N	\N	f	f
0e9e6f99-1447-4d3f-9f51-8aa36d741dd2	generic_instance	\N	43079fcb-f024-46aa-9320-7c51a5bd6494	MRX233	\N	\N	daylily	2025-07-03 08:24:15.998517-07	INSERT	\N	\N	\N	f	f
894921be-3aca-42c1-9965-44f6312f5c1e	generic_instance_lineage	\N	731da033-4f83-42a5-9c13-160e3fd0234b	GL625	\N	\N	daylily	2025-07-03 08:24:16.005492-07	INSERT	\N	\N	\N	f	f
5de33014-e255-498e-8d9b-843887618cf2	generic_instance	\N	c76c7d77-7a2c-4aaf-8a66-637b61c1f62e	MRX234	\N	\N	daylily	2025-07-03 08:24:16.005492-07	INSERT	\N	\N	\N	f	f
92009d8b-3636-4ec8-882e-1a85c6409da3	generic_instance_lineage	\N	dd716bc7-3f79-459d-a219-cccd102d80eb	GL626	\N	\N	daylily	2025-07-03 08:24:16.010893-07	INSERT	\N	\N	\N	f	f
01d29ada-c59b-4386-937e-5163b6e411dd	generic_instance	\N	d20fb8e0-c8c4-4754-a5b3-4937f66a496c	MRX235	\N	\N	daylily	2025-07-03 08:24:16.010893-07	INSERT	\N	\N	\N	f	f
a76f9a86-00a4-47a9-a477-e0bbdfc677d7	generic_instance_lineage	\N	07d31f16-2540-4795-89ab-83ba3d4316ec	GL627	\N	\N	daylily	2025-07-03 08:24:16.016379-07	INSERT	\N	\N	\N	f	f
835d9a0c-c258-4017-802a-2d449a5959bc	generic_instance	\N	5fff5b28-eb3e-4bbd-b79f-754f8f03eea3	MRX236	\N	\N	daylily	2025-07-03 08:24:16.016379-07	INSERT	\N	\N	\N	f	f
19b0828a-58c3-4adb-a986-69060ba8860b	generic_instance_lineage	\N	5efb0bdd-c4b9-4a92-9652-f35ee49f40a4	GL628	\N	\N	daylily	2025-07-03 08:24:16.021218-07	INSERT	\N	\N	\N	f	f
c4723924-5664-4904-84d1-65ee3921f379	generic_instance	\N	1ddb556f-e2da-4b49-b71a-3201257874a6	MRX237	\N	\N	daylily	2025-07-03 08:24:16.021218-07	INSERT	\N	\N	\N	f	f
31f8e41e-0294-4b50-99c3-6009f7429b98	generic_instance_lineage	\N	0ee22f53-49d0-4e74-ad14-ae16da4384df	GL629	\N	\N	daylily	2025-07-03 08:24:16.025934-07	INSERT	\N	\N	\N	f	f
747c211d-d266-4046-b0b1-ba2d3f0aac69	generic_instance	\N	cbd6c137-3d08-43da-84e9-4f918b6946ce	MRX238	\N	\N	daylily	2025-07-03 08:24:16.025934-07	INSERT	\N	\N	\N	f	f
38c04e01-1957-4512-9349-49d49c3b4422	generic_instance_lineage	\N	8bec5b3b-e1a8-4ce1-a3ee-3f2488d98676	GL630	\N	\N	daylily	2025-07-03 08:24:16.030921-07	INSERT	\N	\N	\N	f	f
ccd4d4b4-06b8-4fd0-82f4-b48cc3f74514	generic_instance	\N	ba9a32dd-9ff0-4bd5-bec2-2be8156a1c60	MRX239	\N	\N	daylily	2025-07-03 08:24:16.030921-07	INSERT	\N	\N	\N	f	f
2e596ea7-ab5c-42ba-8010-8584af5bb0f2	generic_instance_lineage	\N	ee2ef3b9-e776-4631-ba4c-9d95d9ab5e2e	GL631	\N	\N	daylily	2025-07-03 08:24:16.03599-07	INSERT	\N	\N	\N	f	f
254121d0-1ec7-4ba0-ad19-1ac995fc9fc3	generic_instance	\N	653a1b4c-97ca-4469-88ee-08bdeeb09309	MRX240	\N	\N	daylily	2025-07-03 08:24:16.03599-07	INSERT	\N	\N	\N	f	f
da963d93-9e05-487e-95fa-429266cc88ff	generic_instance_lineage	\N	08c934fa-b964-4afb-9bbe-a201e7014d94	GL632	\N	\N	daylily	2025-07-03 08:24:16.040974-07	INSERT	\N	\N	\N	f	f
d9de583b-abbc-4184-ac1f-1966e424e155	generic_instance	\N	045dfa06-cd02-4830-815a-64a0eda2b595	MRX241	\N	\N	daylily	2025-07-03 08:24:16.040974-07	INSERT	\N	\N	\N	f	f
fcd33216-8f01-4f2f-93ab-453fc5dc4b76	generic_instance_lineage	\N	f6cd6d86-310f-4c84-b660-b47c5427993a	GL633	\N	\N	daylily	2025-07-03 08:24:16.045785-07	INSERT	\N	\N	\N	f	f
fc49b0e4-bac5-4ce3-9793-6eef83dd8ce5	generic_instance	\N	e5c9c69f-4bf8-4583-83c3-ab32883c6fd8	MRX242	\N	\N	daylily	2025-07-03 08:24:16.045785-07	INSERT	\N	\N	\N	f	f
fa5aacc8-ba20-4f69-bb41-3aaf9fb0cd78	generic_instance_lineage	\N	32ce890f-5851-4915-a0fd-d6c5ea3925a9	GL634	\N	\N	daylily	2025-07-03 08:24:16.050413-07	INSERT	\N	\N	\N	f	f
2dac19bd-ed2a-4f38-af0e-ece5a19bb523	generic_instance	\N	aa3dceeb-693f-4b52-96fd-fcdacec74279	MRX243	\N	\N	daylily	2025-07-03 08:24:16.050413-07	INSERT	\N	\N	\N	f	f
041a8c8f-34ef-4565-8e3d-b7a109657e39	generic_instance_lineage	\N	e954ff06-ed30-415d-bd18-e3e6e46ae774	GL635	\N	\N	daylily	2025-07-03 08:24:16.057149-07	INSERT	\N	\N	\N	f	f
cb1795aa-4917-401f-be78-4f0f538e1bfc	generic_instance	\N	1806a41a-5c44-4be2-9d25-69589865a6fe	MRX244	\N	\N	daylily	2025-07-03 08:24:16.057149-07	INSERT	\N	\N	\N	f	f
3d0066f5-2383-4453-b59b-298bc0c720e1	generic_instance_lineage	\N	bb35d05f-6d49-49cd-b3b0-b14da7148cd7	GL636	\N	\N	daylily	2025-07-03 08:24:16.061637-07	INSERT	\N	\N	\N	f	f
2cd939e0-33a2-4095-bc33-eb29b0771201	generic_instance	\N	81670de8-fbad-4c5b-8159-65d031907aa0	MRX245	\N	\N	daylily	2025-07-03 08:24:16.061637-07	INSERT	\N	\N	\N	f	f
957adabb-9f58-4017-9afe-d3f6b3f42b1c	generic_instance_lineage	\N	5aed10fb-b550-42c8-a69f-48f87bbe0dda	GL637	\N	\N	daylily	2025-07-03 08:24:16.066932-07	INSERT	\N	\N	\N	f	f
d7713771-e351-487a-aeb5-9ea929283018	generic_instance	\N	8ca2ddc1-bf48-4c18-ae6f-54d1d4bef1e4	MRX246	\N	\N	daylily	2025-07-03 08:24:16.066932-07	INSERT	\N	\N	\N	f	f
15ff6f07-7b56-4dab-848b-9804d2898a63	generic_instance_lineage	\N	07d42fd6-e6fa-4636-be25-6ea6d0465c8f	GL638	\N	\N	daylily	2025-07-03 08:24:16.071479-07	INSERT	\N	\N	\N	f	f
9d6b3693-8c10-4900-8ca8-874e94f10079	generic_instance	\N	e1feb2f2-726f-4aa6-8334-7875ca8a9d76	MRX247	\N	\N	daylily	2025-07-03 08:24:16.071479-07	INSERT	\N	\N	\N	f	f
b01320cb-2503-4547-ab31-04618a439f3a	generic_instance	json_addl	7a7a05e0-d926-4739-8f3c-c9a41c715f1a	CWX253	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.111386-07	UPDATE	\N	\N	\N	f	f
5acc3fbc-576e-446b-82eb-143e245d6885	generic_instance	created_dt	7a7a05e0-d926-4739-8f3c-c9a41c715f1a	CWX253	2025-07-03 08:24:16.108836-07	2025-07-03T08:24:16.108836-07:00	daylily	2025-07-03 08:24:16.111386-07	UPDATE	\N	\N	\N	f	f
b38cb6fe-277e-42e8-966a-0bcd77c9e306	generic_instance	modified_dt	7a7a05e0-d926-4739-8f3c-c9a41c715f1a	CWX253	2025-07-03 08:24:16.108836-07	2025-07-03T08:24:16.111386-07:00	daylily	2025-07-03 08:24:16.111386-07	UPDATE	\N	\N	\N	f	f
e435f924-aa9d-46bf-a192-17c8dbc2db38	generic_instance_lineage	\N	43cd7bd4-c097-406a-be9b-ada4b9091cf3	GL645	\N	\N	daylily	2025-07-03 08:24:16.115443-07	INSERT	\N	\N	\N	f	f
227bc80a-1d4c-46ca-865a-8b2a5b04cdfd	generic_instance	\N	78ea25c1-5b2c-4e77-8781-5fffd25ab2df	CWX254	\N	\N	daylily	2025-07-03 08:24:16.115443-07	INSERT	\N	\N	\N	f	f
30a370f8-e6ed-48bd-bd4b-f5cc29461f0a	generic_instance	json_addl	78ea25c1-5b2c-4e77-8781-5fffd25ab2df	CWX254	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.11985-07	UPDATE	\N	\N	\N	f	f
c5534f21-7f85-42a7-915e-6d44d2915ec1	generic_instance	created_dt	78ea25c1-5b2c-4e77-8781-5fffd25ab2df	CWX254	2025-07-03 08:24:16.115443-07	2025-07-03T08:24:16.115443-07:00	daylily	2025-07-03 08:24:16.11985-07	UPDATE	\N	\N	\N	f	f
efa5376e-df28-4ecb-b279-232d95d6fe8b	generic_instance	modified_dt	78ea25c1-5b2c-4e77-8781-5fffd25ab2df	CWX254	2025-07-03 08:24:16.115443-07	2025-07-03T08:24:16.11985-07:00	daylily	2025-07-03 08:24:16.11985-07	UPDATE	\N	\N	\N	f	f
0221e61d-62cc-4f6f-ae56-04e65547281b	generic_instance_lineage	\N	d67ff291-c680-4499-b6b9-b4de08922b84	GL646	\N	\N	daylily	2025-07-03 08:24:16.121029-07	INSERT	\N	\N	\N	f	f
6d1255d9-6ad6-4992-90b0-60a752adec73	generic_instance	\N	6a69eda7-1e5a-4d4a-8755-94aefc3c74e8	CWX255	\N	\N	daylily	2025-07-03 08:24:16.121029-07	INSERT	\N	\N	\N	f	f
c0f246da-e24e-4d82-9773-ffc09255eed5	generic_instance	json_addl	6a69eda7-1e5a-4d4a-8755-94aefc3c74e8	CWX255	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.124914-07	UPDATE	\N	\N	\N	f	f
2ab0de15-021d-46e9-856b-29bed83d6ad5	generic_instance	created_dt	6a69eda7-1e5a-4d4a-8755-94aefc3c74e8	CWX255	2025-07-03 08:24:16.121029-07	2025-07-03T08:24:16.121029-07:00	daylily	2025-07-03 08:24:16.124914-07	UPDATE	\N	\N	\N	f	f
8e919b75-6b85-4cee-9a57-e99b56fd7284	generic_instance	modified_dt	6a69eda7-1e5a-4d4a-8755-94aefc3c74e8	CWX255	2025-07-03 08:24:16.121029-07	2025-07-03T08:24:16.124914-07:00	daylily	2025-07-03 08:24:16.124914-07	UPDATE	\N	\N	\N	f	f
2c95fc65-0038-48c7-998b-c9e091b5c542	generic_instance_lineage	\N	008a06f2-0160-48c3-a881-b7657e127b0c	GL647	\N	\N	daylily	2025-07-03 08:24:16.126215-07	INSERT	\N	\N	\N	f	f
07c6ed5c-0242-4c76-9208-84e88746d7c6	generic_instance	\N	217b0af9-7c61-4971-8098-d67ce949fc38	CWX256	\N	\N	daylily	2025-07-03 08:24:16.126215-07	INSERT	\N	\N	\N	f	f
f03fd04f-0ffb-4ce7-bb77-32d8e1adb7a3	generic_instance	json_addl	217b0af9-7c61-4971-8098-d67ce949fc38	CWX256	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.129956-07	UPDATE	\N	\N	\N	f	f
5ff235ed-8dcc-44cd-bf3f-1af5827b310b	generic_instance	created_dt	217b0af9-7c61-4971-8098-d67ce949fc38	CWX256	2025-07-03 08:24:16.126215-07	2025-07-03T08:24:16.126215-07:00	daylily	2025-07-03 08:24:16.129956-07	UPDATE	\N	\N	\N	f	f
f3902ca8-2510-4f65-b42c-a7cb7ef9d28f	generic_instance	modified_dt	217b0af9-7c61-4971-8098-d67ce949fc38	CWX256	2025-07-03 08:24:16.126215-07	2025-07-03T08:24:16.129956-07:00	daylily	2025-07-03 08:24:16.129956-07	UPDATE	\N	\N	\N	f	f
a1b404db-dbca-423e-ab6e-568fbac28450	generic_instance_lineage	\N	1d5cc72e-a788-444c-be4a-850c50af8eb7	GL648	\N	\N	daylily	2025-07-03 08:24:16.131433-07	INSERT	\N	\N	\N	f	f
f4215c99-369c-4611-9ef4-ced8a3e43b90	generic_instance	\N	4484b524-4a4b-408d-8153-f19da5a148c1	CWX257	\N	\N	daylily	2025-07-03 08:24:16.131433-07	INSERT	\N	\N	\N	f	f
69e18453-430f-4820-ab8f-6faabcd68cf6	generic_instance	json_addl	4484b524-4a4b-408d-8153-f19da5a148c1	CWX257	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.13607-07	UPDATE	\N	\N	\N	f	f
cbd6020b-da2c-478e-b5e6-13bd2c96e75c	generic_instance	created_dt	4484b524-4a4b-408d-8153-f19da5a148c1	CWX257	2025-07-03 08:24:16.131433-07	2025-07-03T08:24:16.131433-07:00	daylily	2025-07-03 08:24:16.13607-07	UPDATE	\N	\N	\N	f	f
db104ec2-a884-4501-b2de-31c730212cd9	generic_instance	modified_dt	4484b524-4a4b-408d-8153-f19da5a148c1	CWX257	2025-07-03 08:24:16.131433-07	2025-07-03T08:24:16.13607-07:00	daylily	2025-07-03 08:24:16.13607-07	UPDATE	\N	\N	\N	f	f
2242235e-a96c-4ebf-a0ae-a3675b529002	generic_instance_lineage	\N	6de10070-8c76-49bd-8956-1a0634fc1cca	GL649	\N	\N	daylily	2025-07-03 08:24:16.137239-07	INSERT	\N	\N	\N	f	f
2222be89-87cf-4744-991d-e5d15dffc5bc	generic_instance	\N	220d86a6-088f-47a7-a11b-f64519da6a46	CWX258	\N	\N	daylily	2025-07-03 08:24:16.137239-07	INSERT	\N	\N	\N	f	f
b0afb350-d941-4db9-88e9-e56a9c02fffa	generic_instance	json_addl	220d86a6-088f-47a7-a11b-f64519da6a46	CWX258	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.141143-07	UPDATE	\N	\N	\N	f	f
32219a3a-847e-4ca0-ba5a-80ee2f9dc4da	generic_instance	created_dt	220d86a6-088f-47a7-a11b-f64519da6a46	CWX258	2025-07-03 08:24:16.137239-07	2025-07-03T08:24:16.137239-07:00	daylily	2025-07-03 08:24:16.141143-07	UPDATE	\N	\N	\N	f	f
65492d8b-16d0-474a-b471-da9ea4af3be0	generic_instance	modified_dt	220d86a6-088f-47a7-a11b-f64519da6a46	CWX258	2025-07-03 08:24:16.137239-07	2025-07-03T08:24:16.141143-07:00	daylily	2025-07-03 08:24:16.141143-07	UPDATE	\N	\N	\N	f	f
e0ee5479-82f5-43c8-b4a8-f1904bf5b960	generic_instance_lineage	\N	4052a3fd-3659-44d6-9b95-8b211f5300ef	GL650	\N	\N	daylily	2025-07-03 08:24:16.142482-07	INSERT	\N	\N	\N	f	f
6e47c5d9-a6f1-4c10-90bb-ee876ff04e48	generic_instance	\N	c3897a38-49ee-4f23-9461-a55dff9a4413	CWX259	\N	\N	daylily	2025-07-03 08:24:16.142482-07	INSERT	\N	\N	\N	f	f
08d63969-b537-45d9-a938-afd7966d94a2	generic_instance	json_addl	c3897a38-49ee-4f23-9461-a55dff9a4413	CWX259	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.146746-07	UPDATE	\N	\N	\N	f	f
dd1e5332-04b0-4e1e-ad3c-b35f36d838be	generic_instance	created_dt	c3897a38-49ee-4f23-9461-a55dff9a4413	CWX259	2025-07-03 08:24:16.142482-07	2025-07-03T08:24:16.142482-07:00	daylily	2025-07-03 08:24:16.146746-07	UPDATE	\N	\N	\N	f	f
98697483-6abe-4cd2-98a4-6ea5862d909e	generic_instance	modified_dt	c3897a38-49ee-4f23-9461-a55dff9a4413	CWX259	2025-07-03 08:24:16.142482-07	2025-07-03T08:24:16.146746-07:00	daylily	2025-07-03 08:24:16.146746-07	UPDATE	\N	\N	\N	f	f
da637d6e-e170-455d-ad86-756ff2bb2f7a	generic_instance_lineage	\N	82b804fa-0b12-4c36-a305-5f3e3cda81d8	GL651	\N	\N	daylily	2025-07-03 08:24:16.148035-07	INSERT	\N	\N	\N	f	f
b8b980f4-cb62-4f87-bcfd-562ed147ae55	generic_instance	\N	2588b288-14ab-4eab-b9ff-41c51104f8bc	CWX260	\N	\N	daylily	2025-07-03 08:24:16.148035-07	INSERT	\N	\N	\N	f	f
ccd05f54-9de1-4d9c-a342-f6b26e8a0565	generic_instance	json_addl	2588b288-14ab-4eab-b9ff-41c51104f8bc	CWX260	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.152392-07	UPDATE	\N	\N	\N	f	f
e6c3da3a-b4c8-44b9-914d-861b8019fc1a	generic_instance	created_dt	2588b288-14ab-4eab-b9ff-41c51104f8bc	CWX260	2025-07-03 08:24:16.148035-07	2025-07-03T08:24:16.148035-07:00	daylily	2025-07-03 08:24:16.152392-07	UPDATE	\N	\N	\N	f	f
14957de0-c711-4d59-85db-30f910355088	generic_instance	modified_dt	2588b288-14ab-4eab-b9ff-41c51104f8bc	CWX260	2025-07-03 08:24:16.148035-07	2025-07-03T08:24:16.152392-07:00	daylily	2025-07-03 08:24:16.152392-07	UPDATE	\N	\N	\N	f	f
e8fc0509-5e3b-4a53-9d14-f8a958f8da20	generic_instance_lineage	\N	8d3f12ab-72a4-4eed-9bb2-e66f2c4e6e02	GL652	\N	\N	daylily	2025-07-03 08:24:16.153972-07	INSERT	\N	\N	\N	f	f
d453c9a1-4f16-4a95-be1c-152dd3efe0e2	generic_instance	\N	13fb23bc-4066-4e86-bde9-9a56cdc08e57	CWX261	\N	\N	daylily	2025-07-03 08:24:16.153972-07	INSERT	\N	\N	\N	f	f
3bf69208-e5e0-4f46-94cc-48fce8fe6529	generic_instance	json_addl	13fb23bc-4066-4e86-bde9-9a56cdc08e57	CWX261	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.159719-07	UPDATE	\N	\N	\N	f	f
aa278737-03c5-4e7f-9f3b-1102cad112dc	generic_instance	created_dt	13fb23bc-4066-4e86-bde9-9a56cdc08e57	CWX261	2025-07-03 08:24:16.153972-07	2025-07-03T08:24:16.153972-07:00	daylily	2025-07-03 08:24:16.159719-07	UPDATE	\N	\N	\N	f	f
0709e482-fe74-4b8f-a967-70778ffa5c57	generic_instance	modified_dt	13fb23bc-4066-4e86-bde9-9a56cdc08e57	CWX261	2025-07-03 08:24:16.153972-07	2025-07-03T08:24:16.159719-07:00	daylily	2025-07-03 08:24:16.159719-07	UPDATE	\N	\N	\N	f	f
ebccfb1a-10f7-43ac-b4ea-ec7dc3583334	generic_instance_lineage	\N	6002beae-1ea4-4071-a5c5-cadcbdec0af0	GL653	\N	\N	daylily	2025-07-03 08:24:16.160832-07	INSERT	\N	\N	\N	f	f
c82f288a-e94f-4484-8dba-ee2b95cf9e5c	generic_instance	\N	4a40cc94-4c0d-4f56-a5be-7b17905d478f	CWX262	\N	\N	daylily	2025-07-03 08:24:16.160832-07	INSERT	\N	\N	\N	f	f
466eaac5-e631-49b8-80a1-5a06d6872326	generic_instance	json_addl	4a40cc94-4c0d-4f56-a5be-7b17905d478f	CWX262	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.165623-07	UPDATE	\N	\N	\N	f	f
8953bd1f-22a6-4469-8221-ee565d246ebc	generic_instance	created_dt	4a40cc94-4c0d-4f56-a5be-7b17905d478f	CWX262	2025-07-03 08:24:16.160832-07	2025-07-03T08:24:16.160832-07:00	daylily	2025-07-03 08:24:16.165623-07	UPDATE	\N	\N	\N	f	f
befeca14-2bb9-4d69-ae5d-533b83bc29e6	generic_instance	modified_dt	4a40cc94-4c0d-4f56-a5be-7b17905d478f	CWX262	2025-07-03 08:24:16.160832-07	2025-07-03T08:24:16.165623-07:00	daylily	2025-07-03 08:24:16.165623-07	UPDATE	\N	\N	\N	f	f
2b7e1c6a-7500-4dae-ba9d-4de602fc4358	generic_instance_lineage	\N	738644c7-f56a-4454-8db8-7f06be1a0213	GL654	\N	\N	daylily	2025-07-03 08:24:16.16817-07	INSERT	\N	\N	\N	f	f
052fe228-8c80-4e09-ab7b-3e362b97b978	generic_instance	\N	a34c5ab3-1540-4e32-a08c-9f90c1f7894e	CWX263	\N	\N	daylily	2025-07-03 08:24:16.16817-07	INSERT	\N	\N	\N	f	f
7bb51859-8a52-4c84-8889-d9d993e564b4	generic_instance	json_addl	a34c5ab3-1540-4e32-a08c-9f90c1f7894e	CWX263	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.172411-07	UPDATE	\N	\N	\N	f	f
93490ed4-216f-4391-9218-6562c6f1ec7c	generic_instance	created_dt	a34c5ab3-1540-4e32-a08c-9f90c1f7894e	CWX263	2025-07-03 08:24:16.16817-07	2025-07-03T08:24:16.16817-07:00	daylily	2025-07-03 08:24:16.172411-07	UPDATE	\N	\N	\N	f	f
480bd55f-9883-41bc-ad8c-deacf013ea3b	generic_instance	modified_dt	a34c5ab3-1540-4e32-a08c-9f90c1f7894e	CWX263	2025-07-03 08:24:16.16817-07	2025-07-03T08:24:16.172411-07:00	daylily	2025-07-03 08:24:16.172411-07	UPDATE	\N	\N	\N	f	f
30dceb4f-7142-4a72-aaf6-f30d14dc8b18	generic_instance_lineage	\N	a078c524-0cac-44f8-89ae-77f51e0f2ed6	GL655	\N	\N	daylily	2025-07-03 08:24:16.173538-07	INSERT	\N	\N	\N	f	f
744c9b6b-8067-4c09-ae65-461980e3758d	generic_instance	\N	0007e7e3-47ad-48fa-8228-7f0dbe1f4a60	CWX264	\N	\N	daylily	2025-07-03 08:24:16.173538-07	INSERT	\N	\N	\N	f	f
35b6fe4e-c8e5-4301-8b69-abde069b9a03	generic_instance	json_addl	0007e7e3-47ad-48fa-8228-7f0dbe1f4a60	CWX264	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.177886-07	UPDATE	\N	\N	\N	f	f
1c3a6b37-4cc2-41aa-bdd4-d844a47855a3	generic_instance	created_dt	0007e7e3-47ad-48fa-8228-7f0dbe1f4a60	CWX264	2025-07-03 08:24:16.173538-07	2025-07-03T08:24:16.173538-07:00	daylily	2025-07-03 08:24:16.177886-07	UPDATE	\N	\N	\N	f	f
6ac3c888-f5af-4677-86fd-f3ca48f2de47	generic_instance	modified_dt	0007e7e3-47ad-48fa-8228-7f0dbe1f4a60	CWX264	2025-07-03 08:24:16.173538-07	2025-07-03T08:24:16.177886-07:00	daylily	2025-07-03 08:24:16.177886-07	UPDATE	\N	\N	\N	f	f
05d037d7-cc6e-4558-a3a3-b70c7910418b	generic_instance_lineage	\N	72a6dd12-e29b-4456-893d-207104cbad11	GL656	\N	\N	daylily	2025-07-03 08:24:16.179232-07	INSERT	\N	\N	\N	f	f
65d488ea-380c-49ac-a198-42a40352d965	generic_instance	\N	e6a96820-ce9d-4ec4-8f59-fd93f0e11b76	CWX265	\N	\N	daylily	2025-07-03 08:24:16.179232-07	INSERT	\N	\N	\N	f	f
8fecd1f4-ef06-474e-bca0-91edf9fb0779	generic_instance	json_addl	e6a96820-ce9d-4ec4-8f59-fd93f0e11b76	CWX265	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.183037-07	UPDATE	\N	\N	\N	f	f
0e89bae7-03fb-4110-bf83-c114543bee98	generic_instance	created_dt	e6a96820-ce9d-4ec4-8f59-fd93f0e11b76	CWX265	2025-07-03 08:24:16.179232-07	2025-07-03T08:24:16.179232-07:00	daylily	2025-07-03 08:24:16.183037-07	UPDATE	\N	\N	\N	f	f
6f3f747c-995a-4c0e-ac59-466d38117cac	generic_instance	modified_dt	e6a96820-ce9d-4ec4-8f59-fd93f0e11b76	CWX265	2025-07-03 08:24:16.179232-07	2025-07-03T08:24:16.183037-07:00	daylily	2025-07-03 08:24:16.183037-07	UPDATE	\N	\N	\N	f	f
2d3988fc-ee2d-4847-948b-df5233dad9f5	generic_instance_lineage	\N	7ff212fc-f2a2-4b75-b36e-5dc8ee450acb	GL657	\N	\N	daylily	2025-07-03 08:24:16.184262-07	INSERT	\N	\N	\N	f	f
87f1a7a7-809d-48a5-9a20-1c3ce380b8c6	generic_instance	\N	5eab8dce-d0dd-449f-966d-47154aef3cf0	CWX266	\N	\N	daylily	2025-07-03 08:24:16.184262-07	INSERT	\N	\N	\N	f	f
a36f44cd-1565-4824-a1b0-aed7326e0733	generic_instance	json_addl	5eab8dce-d0dd-449f-966d-47154aef3cf0	CWX266	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.188374-07	UPDATE	\N	\N	\N	f	f
7707e712-2494-4f79-b3fb-b304a2c77139	generic_instance	created_dt	5eab8dce-d0dd-449f-966d-47154aef3cf0	CWX266	2025-07-03 08:24:16.184262-07	2025-07-03T08:24:16.184262-07:00	daylily	2025-07-03 08:24:16.188374-07	UPDATE	\N	\N	\N	f	f
57e10165-d491-4a97-a0b2-366e81a81358	generic_instance	modified_dt	5eab8dce-d0dd-449f-966d-47154aef3cf0	CWX266	2025-07-03 08:24:16.184262-07	2025-07-03T08:24:16.188374-07:00	daylily	2025-07-03 08:24:16.188374-07	UPDATE	\N	\N	\N	f	f
b85ba962-f14c-490a-a67a-469c2f90c86e	generic_instance_lineage	\N	989b1900-2d75-4096-a4c6-2ccf62e9d9e2	GL658	\N	\N	daylily	2025-07-03 08:24:16.189469-07	INSERT	\N	\N	\N	f	f
555ff1b8-703a-49ac-bac3-cc11b8b619f9	generic_instance	\N	4e71540e-b707-4978-a4da-25a8ac6dbb48	CWX267	\N	\N	daylily	2025-07-03 08:24:16.189469-07	INSERT	\N	\N	\N	f	f
ee0af028-cfe9-46f2-8804-c4b7c1eb8309	generic_instance	json_addl	4e71540e-b707-4978-a4da-25a8ac6dbb48	CWX267	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.193381-07	UPDATE	\N	\N	\N	f	f
2a2ea106-c1f3-4e16-a2af-224d65fbb2e2	generic_instance	created_dt	4e71540e-b707-4978-a4da-25a8ac6dbb48	CWX267	2025-07-03 08:24:16.189469-07	2025-07-03T08:24:16.189469-07:00	daylily	2025-07-03 08:24:16.193381-07	UPDATE	\N	\N	\N	f	f
009a8f71-3b32-433b-870a-9bee320143d2	generic_instance	modified_dt	4e71540e-b707-4978-a4da-25a8ac6dbb48	CWX267	2025-07-03 08:24:16.189469-07	2025-07-03T08:24:16.193381-07:00	daylily	2025-07-03 08:24:16.193381-07	UPDATE	\N	\N	\N	f	f
4d8a12fc-5c0a-43c0-9812-d33b5b0bc65d	generic_instance_lineage	\N	86dd4863-f3ee-40ab-b399-c8e65d4d65a0	GL659	\N	\N	daylily	2025-07-03 08:24:16.194592-07	INSERT	\N	\N	\N	f	f
72b7d36c-7e43-46c3-93c4-1512b11623b1	generic_instance	\N	7f0ff5ac-55fd-40b8-993c-37fd45de7e70	CWX268	\N	\N	daylily	2025-07-03 08:24:16.194592-07	INSERT	\N	\N	\N	f	f
1bd39710-f1ed-4504-a104-897f5d96cd71	generic_instance	json_addl	7f0ff5ac-55fd-40b8-993c-37fd45de7e70	CWX268	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.199917-07	UPDATE	\N	\N	\N	f	f
fa089b00-b7f7-40bd-a56c-2d5dfa167c34	generic_instance	created_dt	7f0ff5ac-55fd-40b8-993c-37fd45de7e70	CWX268	2025-07-03 08:24:16.194592-07	2025-07-03T08:24:16.194592-07:00	daylily	2025-07-03 08:24:16.199917-07	UPDATE	\N	\N	\N	f	f
d13377f0-02d7-4ccd-96c6-fa328b45531f	generic_instance	modified_dt	7f0ff5ac-55fd-40b8-993c-37fd45de7e70	CWX268	2025-07-03 08:24:16.194592-07	2025-07-03T08:24:16.199917-07:00	daylily	2025-07-03 08:24:16.199917-07	UPDATE	\N	\N	\N	f	f
d995bc74-22ea-4666-8211-27d3b56fde48	generic_instance_lineage	\N	8a7af7d7-ec12-46f0-8775-aa037d309fd8	GL660	\N	\N	daylily	2025-07-03 08:24:16.201071-07	INSERT	\N	\N	\N	f	f
9974a82a-1d14-4eaf-b737-fa462fc69700	generic_instance	\N	5914dd40-bbb0-451f-aabf-17c7970df831	CWX269	\N	\N	daylily	2025-07-03 08:24:16.201071-07	INSERT	\N	\N	\N	f	f
b783fc77-b333-47c1-90f4-7abc699ebcda	generic_instance	json_addl	5914dd40-bbb0-451f-aabf-17c7970df831	CWX269	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.204924-07	UPDATE	\N	\N	\N	f	f
d6615c4c-0fa3-4b79-b19c-b0241f1f9995	generic_instance	created_dt	5914dd40-bbb0-451f-aabf-17c7970df831	CWX269	2025-07-03 08:24:16.201071-07	2025-07-03T08:24:16.201071-07:00	daylily	2025-07-03 08:24:16.204924-07	UPDATE	\N	\N	\N	f	f
c49d0ee4-33e1-4743-b80b-f68ec2b9b4cd	generic_instance	modified_dt	5914dd40-bbb0-451f-aabf-17c7970df831	CWX269	2025-07-03 08:24:16.201071-07	2025-07-03T08:24:16.204924-07:00	daylily	2025-07-03 08:24:16.204924-07	UPDATE	\N	\N	\N	f	f
7107a6db-1455-4045-a2cf-d9f6c295c937	generic_instance_lineage	\N	2c506eda-c6e3-419a-acbe-eb00c2014cc4	GL661	\N	\N	daylily	2025-07-03 08:24:16.20631-07	INSERT	\N	\N	\N	f	f
ce17d512-25de-469b-849e-46fdaeee08cf	generic_instance	\N	d53b5c53-dff3-4c58-be12-ef4497a67766	CWX270	\N	\N	daylily	2025-07-03 08:24:16.20631-07	INSERT	\N	\N	\N	f	f
a0e59503-93b0-4ebb-840a-be173fc04642	generic_instance	json_addl	d53b5c53-dff3-4c58-be12-ef4497a67766	CWX270	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.210561-07	UPDATE	\N	\N	\N	f	f
c0fbdc82-f380-4319-b40e-065f20b87be2	generic_instance	created_dt	d53b5c53-dff3-4c58-be12-ef4497a67766	CWX270	2025-07-03 08:24:16.20631-07	2025-07-03T08:24:16.20631-07:00	daylily	2025-07-03 08:24:16.210561-07	UPDATE	\N	\N	\N	f	f
851bb05b-1ba5-4cae-9e23-c3f438e66d8d	generic_instance	modified_dt	d53b5c53-dff3-4c58-be12-ef4497a67766	CWX270	2025-07-03 08:24:16.20631-07	2025-07-03T08:24:16.210561-07:00	daylily	2025-07-03 08:24:16.210561-07	UPDATE	\N	\N	\N	f	f
34c87580-7884-40ee-9c7e-5c57a924a930	generic_instance_lineage	\N	938c645b-9d7b-40ec-a03f-632a15414104	GL662	\N	\N	daylily	2025-07-03 08:24:16.212024-07	INSERT	\N	\N	\N	f	f
e6cec067-4213-48ff-963c-e442839f66cd	generic_instance	\N	561fbd02-0e42-4c28-b549-540297a81c6a	CWX271	\N	\N	daylily	2025-07-03 08:24:16.212024-07	INSERT	\N	\N	\N	f	f
e72944bb-8820-43cd-85f5-c7f404263875	generic_instance	json_addl	561fbd02-0e42-4c28-b549-540297a81c6a	CWX271	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.216014-07	UPDATE	\N	\N	\N	f	f
231a2ab2-9b8a-43dd-872d-5c7c64f29636	generic_instance	created_dt	561fbd02-0e42-4c28-b549-540297a81c6a	CWX271	2025-07-03 08:24:16.212024-07	2025-07-03T08:24:16.212024-07:00	daylily	2025-07-03 08:24:16.216014-07	UPDATE	\N	\N	\N	f	f
adabd972-6a1a-4ece-8a7a-7834c4377600	generic_instance	modified_dt	561fbd02-0e42-4c28-b549-540297a81c6a	CWX271	2025-07-03 08:24:16.212024-07	2025-07-03T08:24:16.216014-07:00	daylily	2025-07-03 08:24:16.216014-07	UPDATE	\N	\N	\N	f	f
e6f8c063-681c-44b2-a1c5-0d1d4956ffd4	generic_instance_lineage	\N	7f015c91-b548-4b5d-b668-eeee4ea78342	GL663	\N	\N	daylily	2025-07-03 08:24:16.217462-07	INSERT	\N	\N	\N	f	f
0cbb4629-4e9d-4286-ad75-112bb1756827	generic_instance	\N	ddc6efd5-96cb-476a-a565-bb7248a7b9a2	CWX272	\N	\N	daylily	2025-07-03 08:24:16.217462-07	INSERT	\N	\N	\N	f	f
41becdd0-495f-4b56-a3d0-e5d80840c05b	generic_instance	json_addl	ddc6efd5-96cb-476a-a565-bb7248a7b9a2	CWX272	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.221429-07	UPDATE	\N	\N	\N	f	f
29e6b51f-6cac-4467-af2e-732bf2086501	generic_instance	created_dt	ddc6efd5-96cb-476a-a565-bb7248a7b9a2	CWX272	2025-07-03 08:24:16.217462-07	2025-07-03T08:24:16.217462-07:00	daylily	2025-07-03 08:24:16.221429-07	UPDATE	\N	\N	\N	f	f
7b51cffe-2425-4732-ae83-7b9095ca4776	generic_instance	modified_dt	ddc6efd5-96cb-476a-a565-bb7248a7b9a2	CWX272	2025-07-03 08:24:16.217462-07	2025-07-03T08:24:16.221429-07:00	daylily	2025-07-03 08:24:16.221429-07	UPDATE	\N	\N	\N	f	f
6d970692-0dcc-4040-91a3-349647815dfe	generic_instance_lineage	\N	cb7f1fb2-ba5e-4eb8-8ada-c266202ad206	GL664	\N	\N	daylily	2025-07-03 08:24:16.222941-07	INSERT	\N	\N	\N	f	f
0e658a20-ef47-4cb1-bd97-b6dc42174645	generic_instance	\N	f5a92643-727e-42cd-a337-e2fd476dc69d	CWX273	\N	\N	daylily	2025-07-03 08:24:16.222941-07	INSERT	\N	\N	\N	f	f
8354e299-5607-428f-ba75-16aca5ee7c79	generic_instance	json_addl	f5a92643-727e-42cd-a337-e2fd476dc69d	CWX273	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.228929-07	UPDATE	\N	\N	\N	f	f
e1cb8c8d-f1fc-40c9-afc1-295024340456	generic_instance	created_dt	f5a92643-727e-42cd-a337-e2fd476dc69d	CWX273	2025-07-03 08:24:16.222941-07	2025-07-03T08:24:16.222941-07:00	daylily	2025-07-03 08:24:16.228929-07	UPDATE	\N	\N	\N	f	f
7a8a7c1c-6809-4c9b-8501-fef5571e07ff	generic_instance	modified_dt	f5a92643-727e-42cd-a337-e2fd476dc69d	CWX273	2025-07-03 08:24:16.222941-07	2025-07-03T08:24:16.228929-07:00	daylily	2025-07-03 08:24:16.228929-07	UPDATE	\N	\N	\N	f	f
ea6650ca-1893-4ddb-bb51-3f05fc2b5e7a	generic_instance_lineage	\N	68b9b27e-f99d-4291-8e6a-754536172b0a	GL665	\N	\N	daylily	2025-07-03 08:24:16.230475-07	INSERT	\N	\N	\N	f	f
d52d3417-b3c5-4ef8-a013-2fc95a82e8a6	generic_instance	\N	2754a4ac-782d-4991-9e07-ed7cad5dfc2a	CWX274	\N	\N	daylily	2025-07-03 08:24:16.230475-07	INSERT	\N	\N	\N	f	f
8e200b6c-c2ef-4d77-b94f-4d597c8e9b26	generic_instance	json_addl	2754a4ac-782d-4991-9e07-ed7cad5dfc2a	CWX274	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.234608-07	UPDATE	\N	\N	\N	f	f
a0a8a84d-1b66-42bb-b3fd-ead0c1a84edf	generic_instance	created_dt	2754a4ac-782d-4991-9e07-ed7cad5dfc2a	CWX274	2025-07-03 08:24:16.230475-07	2025-07-03T08:24:16.230475-07:00	daylily	2025-07-03 08:24:16.234608-07	UPDATE	\N	\N	\N	f	f
27ad9095-f4f6-440c-9cdc-045846903c87	generic_instance	modified_dt	2754a4ac-782d-4991-9e07-ed7cad5dfc2a	CWX274	2025-07-03 08:24:16.230475-07	2025-07-03T08:24:16.234608-07:00	daylily	2025-07-03 08:24:16.234608-07	UPDATE	\N	\N	\N	f	f
2a180419-aea3-4784-851f-44d950d92dbb	generic_instance_lineage	\N	a65743ed-2090-4b59-b0ce-f99809235d67	GL666	\N	\N	daylily	2025-07-03 08:24:16.235951-07	INSERT	\N	\N	\N	f	f
79e5c3f4-102c-4317-a7dc-9e420ae6d792	generic_instance	\N	74137fae-e9d7-4e0b-b88c-f6e1ef1aeb96	CWX275	\N	\N	daylily	2025-07-03 08:24:16.235951-07	INSERT	\N	\N	\N	f	f
69ec05c3-9349-4733-8787-964fef94dc3f	generic_instance	json_addl	74137fae-e9d7-4e0b-b88c-f6e1ef1aeb96	CWX275	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.240465-07	UPDATE	\N	\N	\N	f	f
92981aad-4883-4ecd-854e-70bc95590696	generic_instance	created_dt	74137fae-e9d7-4e0b-b88c-f6e1ef1aeb96	CWX275	2025-07-03 08:24:16.235951-07	2025-07-03T08:24:16.235951-07:00	daylily	2025-07-03 08:24:16.240465-07	UPDATE	\N	\N	\N	f	f
b892a311-e19e-4bd4-a223-28d07e47e211	generic_instance	modified_dt	74137fae-e9d7-4e0b-b88c-f6e1ef1aeb96	CWX275	2025-07-03 08:24:16.235951-07	2025-07-03T08:24:16.240465-07:00	daylily	2025-07-03 08:24:16.240465-07	UPDATE	\N	\N	\N	f	f
e050a4e3-2a5d-4d69-af83-e74fef0b74d4	generic_instance_lineage	\N	ae2afa26-cc66-4df9-9354-d25769ef3181	GL667	\N	\N	daylily	2025-07-03 08:24:16.241895-07	INSERT	\N	\N	\N	f	f
b755fa1a-dfe3-40c4-92d8-d7873d60ac0d	generic_instance	\N	7b8a94d7-37af-4f1b-a3fc-1cd5399cb3f9	CWX276	\N	\N	daylily	2025-07-03 08:24:16.241895-07	INSERT	\N	\N	\N	f	f
1bc87afe-4248-4933-b1eb-06842e41a7e8	generic_instance	json_addl	7b8a94d7-37af-4f1b-a3fc-1cd5399cb3f9	CWX276	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.246288-07	UPDATE	\N	\N	\N	f	f
1119d578-0d8d-462a-8358-5723bfc3fd10	generic_instance	created_dt	7b8a94d7-37af-4f1b-a3fc-1cd5399cb3f9	CWX276	2025-07-03 08:24:16.241895-07	2025-07-03T08:24:16.241895-07:00	daylily	2025-07-03 08:24:16.246288-07	UPDATE	\N	\N	\N	f	f
b8749ced-eefb-40de-bee3-b136e1213c19	generic_instance	modified_dt	7b8a94d7-37af-4f1b-a3fc-1cd5399cb3f9	CWX276	2025-07-03 08:24:16.241895-07	2025-07-03T08:24:16.246288-07:00	daylily	2025-07-03 08:24:16.246288-07	UPDATE	\N	\N	\N	f	f
94770541-68c2-4282-a8c7-19c9e07d3f71	generic_instance_lineage	\N	ce18cd66-32c6-4ce9-a103-dc52d3425d65	GL668	\N	\N	daylily	2025-07-03 08:24:16.247453-07	INSERT	\N	\N	\N	f	f
470b212b-77de-4efd-aaf8-aa8ac4c93df7	generic_instance	\N	165a43a4-b366-4fc2-a6c1-97f8c9b236dc	CWX277	\N	\N	daylily	2025-07-03 08:24:16.247453-07	INSERT	\N	\N	\N	f	f
8249f080-b30d-49e9-b2a5-4541d66a4d5d	generic_instance	json_addl	165a43a4-b366-4fc2-a6c1-97f8c9b236dc	CWX277	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.251863-07	UPDATE	\N	\N	\N	f	f
f6c55c16-3fb0-4100-a43e-b615dea86f4f	generic_instance	created_dt	165a43a4-b366-4fc2-a6c1-97f8c9b236dc	CWX277	2025-07-03 08:24:16.247453-07	2025-07-03T08:24:16.247453-07:00	daylily	2025-07-03 08:24:16.251863-07	UPDATE	\N	\N	\N	f	f
eed47d3f-ea1a-4cd3-84f1-8ef915c6126d	generic_instance	modified_dt	165a43a4-b366-4fc2-a6c1-97f8c9b236dc	CWX277	2025-07-03 08:24:16.247453-07	2025-07-03T08:24:16.251863-07:00	daylily	2025-07-03 08:24:16.251863-07	UPDATE	\N	\N	\N	f	f
95f0a335-c69c-491c-93e5-4c44c05cb1d2	generic_instance_lineage	\N	2f763498-ce03-48fc-bb96-306795f1bf84	GL669	\N	\N	daylily	2025-07-03 08:24:16.257121-07	INSERT	\N	\N	\N	f	f
66ece850-4fb1-4f4e-a570-d4013c148ae7	generic_instance	\N	b2fc4da9-2a05-49ca-8d5d-b9558cc97c68	CWX278	\N	\N	daylily	2025-07-03 08:24:16.257121-07	INSERT	\N	\N	\N	f	f
73976eb5-7cf8-4047-b1da-0f216d42804c	generic_instance	json_addl	b2fc4da9-2a05-49ca-8d5d-b9558cc97c68	CWX278	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.261195-07	UPDATE	\N	\N	\N	f	f
ec11cbf2-698c-43dc-934c-00560c29e763	generic_instance	created_dt	b2fc4da9-2a05-49ca-8d5d-b9558cc97c68	CWX278	2025-07-03 08:24:16.257121-07	2025-07-03T08:24:16.257121-07:00	daylily	2025-07-03 08:24:16.261195-07	UPDATE	\N	\N	\N	f	f
7f64c547-08fa-4f1c-8c3c-14cdd76037be	generic_instance	modified_dt	b2fc4da9-2a05-49ca-8d5d-b9558cc97c68	CWX278	2025-07-03 08:24:16.257121-07	2025-07-03T08:24:16.261195-07:00	daylily	2025-07-03 08:24:16.261195-07	UPDATE	\N	\N	\N	f	f
c20aa9d2-b09f-4a79-899c-817ea60f2b2c	generic_instance_lineage	\N	52f6dde3-ae7c-4031-8235-52e3b1aea117	GL670	\N	\N	daylily	2025-07-03 08:24:16.262563-07	INSERT	\N	\N	\N	f	f
94a6632a-c447-4ce9-b24d-32ee623b364b	generic_instance	\N	774fcd11-a6a1-4bdc-8b98-7fbc239ad4d7	CWX279	\N	\N	daylily	2025-07-03 08:24:16.262563-07	INSERT	\N	\N	\N	f	f
175fc45a-6612-4527-86b7-8b2020ca4f85	generic_instance	json_addl	774fcd11-a6a1-4bdc-8b98-7fbc239ad4d7	CWX279	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.267222-07	UPDATE	\N	\N	\N	f	f
350be657-5325-40c1-9a8f-606ff6c33759	generic_instance	created_dt	774fcd11-a6a1-4bdc-8b98-7fbc239ad4d7	CWX279	2025-07-03 08:24:16.262563-07	2025-07-03T08:24:16.262563-07:00	daylily	2025-07-03 08:24:16.267222-07	UPDATE	\N	\N	\N	f	f
1c91d94e-c845-41e8-afa2-cf9f903b1fe9	generic_instance	modified_dt	774fcd11-a6a1-4bdc-8b98-7fbc239ad4d7	CWX279	2025-07-03 08:24:16.262563-07	2025-07-03T08:24:16.267222-07:00	daylily	2025-07-03 08:24:16.267222-07	UPDATE	\N	\N	\N	f	f
e1c3618e-ee12-47e4-8c12-fcd5bfc22ee1	generic_instance_lineage	\N	fc77ce96-036c-4f03-a473-a11c8067b14e	GL671	\N	\N	daylily	2025-07-03 08:24:16.268554-07	INSERT	\N	\N	\N	f	f
7a54060b-59c6-4e85-bfc2-8abb0c941da5	generic_instance	\N	91999ed4-21f0-408d-8f7c-a00ad024b9a8	CWX280	\N	\N	daylily	2025-07-03 08:24:16.268554-07	INSERT	\N	\N	\N	f	f
62d8c884-c3b7-4f4d-82e1-c5663af3e4f7	generic_instance	json_addl	91999ed4-21f0-408d-8f7c-a00ad024b9a8	CWX280	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.272866-07	UPDATE	\N	\N	\N	f	f
9b11b575-884d-43e7-9222-8740bacb0558	generic_instance	created_dt	91999ed4-21f0-408d-8f7c-a00ad024b9a8	CWX280	2025-07-03 08:24:16.268554-07	2025-07-03T08:24:16.268554-07:00	daylily	2025-07-03 08:24:16.272866-07	UPDATE	\N	\N	\N	f	f
75677bfb-7a96-4117-8a87-6e25a4416102	generic_instance	modified_dt	91999ed4-21f0-408d-8f7c-a00ad024b9a8	CWX280	2025-07-03 08:24:16.268554-07	2025-07-03T08:24:16.272866-07:00	daylily	2025-07-03 08:24:16.272866-07	UPDATE	\N	\N	\N	f	f
ce8d92b2-26d3-4db5-aaba-82f84b3d17a7	generic_instance_lineage	\N	4afbaba8-6480-45fb-aeed-9c06de5409f2	GL672	\N	\N	daylily	2025-07-03 08:24:16.274066-07	INSERT	\N	\N	\N	f	f
f7c14074-9726-48a3-8ebf-c1ad2141f753	generic_instance	\N	e3b20962-4aaa-4e81-850f-90b4ff6bb72b	MRX253	\N	\N	daylily	2025-07-03 08:24:16.276217-07	INSERT	\N	\N	\N	f	f
9ba0b38c-2a0a-4241-96d1-0fc7f008070e	generic_instance_lineage	\N	87c3ad43-cdba-4c8f-a3e6-0c5dcffa7812	GL673	\N	\N	daylily	2025-07-03 08:24:16.278242-07	INSERT	\N	\N	\N	f	f
65c6f52b-44e9-4580-8960-d2ba55fbd0b6	generic_instance	\N	74ad366b-d25f-4da4-8084-53ea5fa20220	MRX254	\N	\N	daylily	2025-07-03 08:24:16.278242-07	INSERT	\N	\N	\N	f	f
ac1a7aef-c6fd-46d7-a9d1-a0673babbe27	generic_instance_lineage	\N	b2659560-f732-4da9-ace5-539448f993f5	GL674	\N	\N	daylily	2025-07-03 08:24:16.283122-07	INSERT	\N	\N	\N	f	f
261bf5ab-06f5-458e-adc6-9a509becc0b0	generic_instance	\N	0c89558c-3040-41f3-b14f-54281d89a861	MRX255	\N	\N	daylily	2025-07-03 08:24:16.283122-07	INSERT	\N	\N	\N	f	f
18dcd2e2-0602-4386-9892-626eb05b1fb8	generic_instance_lineage	\N	74e2cf41-d124-44c0-998c-d4670f1df903	GL675	\N	\N	daylily	2025-07-03 08:24:16.289351-07	INSERT	\N	\N	\N	f	f
79a8a9fe-c9f6-4acc-8883-d23f1fe8d035	generic_instance	\N	f576931a-14bf-4549-bdad-cfbb1c5ad826	MRX256	\N	\N	daylily	2025-07-03 08:24:16.289351-07	INSERT	\N	\N	\N	f	f
0c61ea9f-a4d8-4787-b12a-8c1a4d255e96	generic_instance_lineage	\N	3f1e9150-2a19-442f-bcd7-14f218946eee	GL676	\N	\N	daylily	2025-07-03 08:24:16.29568-07	INSERT	\N	\N	\N	f	f
b712d597-49e9-4e1b-a6da-b50558362d9e	generic_instance	\N	f5c9a461-3037-48a9-ab53-74984ace3aec	MRX257	\N	\N	daylily	2025-07-03 08:24:16.29568-07	INSERT	\N	\N	\N	f	f
da283f9e-9710-4533-b22d-1c4ddda1142c	generic_instance_lineage	\N	c8a926dc-bfbb-489a-8992-40c4133997ad	GL677	\N	\N	daylily	2025-07-03 08:24:16.301211-07	INSERT	\N	\N	\N	f	f
50bf9001-3a71-4bed-8fd0-7caaaec0eea6	generic_instance	\N	103f84a1-a056-414a-b9a1-07ead9a853e6	MRX258	\N	\N	daylily	2025-07-03 08:24:16.301211-07	INSERT	\N	\N	\N	f	f
5d99cab4-dfff-4f7a-9913-ccea23e0531b	generic_instance_lineage	\N	aeacbbe7-6ba7-4e70-939a-f86a95f641be	GL678	\N	\N	daylily	2025-07-03 08:24:16.307149-07	INSERT	\N	\N	\N	f	f
243dc100-1ac5-49ee-8f42-68aad68670b1	generic_instance	\N	0c26d313-4b07-4a0f-a826-ea2b0242df8b	MRX259	\N	\N	daylily	2025-07-03 08:24:16.307149-07	INSERT	\N	\N	\N	f	f
93b263f2-6b23-41c9-b654-47830a0a417f	generic_instance_lineage	\N	5e087394-2c8b-4e81-ad07-da72b87b5915	GL679	\N	\N	daylily	2025-07-03 08:24:16.311778-07	INSERT	\N	\N	\N	f	f
0308501f-3303-466a-9fef-cdddc8b3ed74	generic_instance	\N	5ef0757b-7e03-409b-87ea-70557be4dac0	MRX260	\N	\N	daylily	2025-07-03 08:24:16.311778-07	INSERT	\N	\N	\N	f	f
13a0ce86-038f-46ea-8c81-6d8290ffe71d	generic_instance_lineage	\N	3a64133b-eb89-4bde-8e3a-9fc9a23f4b05	GL680	\N	\N	daylily	2025-07-03 08:24:16.316737-07	INSERT	\N	\N	\N	f	f
75835c8b-c49d-4f8e-be2e-a92ef1211804	generic_instance	\N	a5bb73df-6539-4173-9687-1763600076bc	MRX261	\N	\N	daylily	2025-07-03 08:24:16.316737-07	INSERT	\N	\N	\N	f	f
00aba006-f613-490e-9459-69d914a39041	generic_instance_lineage	\N	775e4d55-c66f-423d-a000-8cd9eed7099b	GL681	\N	\N	daylily	2025-07-03 08:24:16.321763-07	INSERT	\N	\N	\N	f	f
1845ba60-af63-4dda-9555-dbcc9a8d2544	generic_instance	\N	4830667e-3fe8-4a03-beeb-80786dc41757	MRX262	\N	\N	daylily	2025-07-03 08:24:16.321763-07	INSERT	\N	\N	\N	f	f
18e32558-f9b8-43e7-9cab-b392e5b1b9a4	generic_instance_lineage	\N	d4157848-b040-435a-b13a-92c5006c7d2b	GL682	\N	\N	daylily	2025-07-03 08:24:16.326841-07	INSERT	\N	\N	\N	f	f
ff45a4cc-c25b-4107-8b3e-73ba4fc6a6de	generic_instance	\N	7512b9d2-980e-4174-952b-78facb42ec6d	MRX263	\N	\N	daylily	2025-07-03 08:24:16.326841-07	INSERT	\N	\N	\N	f	f
8e455f11-b8ab-4415-8ab6-f3ae740a2266	generic_instance_lineage	\N	ada1fbec-c3af-44bf-ae9f-3f6e6eecafa1	GL683	\N	\N	daylily	2025-07-03 08:24:16.331496-07	INSERT	\N	\N	\N	f	f
252f106b-2def-4e9a-aa39-81366cc21da9	generic_instance	\N	b0b03902-4397-4f5c-99cf-9bc996f64e45	MRX264	\N	\N	daylily	2025-07-03 08:24:16.331496-07	INSERT	\N	\N	\N	f	f
c2e7f7c0-4c1b-484b-bf4c-16ba7e07368e	generic_instance_lineage	\N	460b6197-2e73-4e6b-bba4-e6d8c8231d5c	GL684	\N	\N	daylily	2025-07-03 08:24:16.336673-07	INSERT	\N	\N	\N	f	f
59290b62-cad7-49be-9566-6ae93c8a789b	generic_instance	\N	90a7b8ca-0912-4f4f-9e1b-35ee43101fe8	MRX265	\N	\N	daylily	2025-07-03 08:24:16.336673-07	INSERT	\N	\N	\N	f	f
93764256-6c15-401e-9e08-138f48695857	generic_instance_lineage	\N	ddb35748-5461-4510-9441-6dd66ce69f9c	GL685	\N	\N	daylily	2025-07-03 08:24:16.342636-07	INSERT	\N	\N	\N	f	f
ca7787a6-c330-47c1-ba71-ca549f9f24ef	generic_instance	\N	ffae08fc-ee93-4871-9d7a-5c45332c4d86	MRX266	\N	\N	daylily	2025-07-03 08:24:16.342636-07	INSERT	\N	\N	\N	f	f
e3c633f5-beac-4465-b097-cb74e2a77c1f	generic_instance_lineage	\N	a4d06364-7a57-41bd-ae5f-cbe0989e7ac6	GL686	\N	\N	daylily	2025-07-03 08:24:16.34712-07	INSERT	\N	\N	\N	f	f
e97b667c-709f-4117-a816-2f19b306781c	generic_instance	\N	7823f81f-b135-4cb1-9937-96ad578a673d	MRX267	\N	\N	daylily	2025-07-03 08:24:16.34712-07	INSERT	\N	\N	\N	f	f
3d514713-95c7-45e8-8f86-c28a59886c7e	generic_instance_lineage	\N	f25eb45f-f21a-471d-ab99-802a4c35dc7d	GL687	\N	\N	daylily	2025-07-03 08:24:16.35246-07	INSERT	\N	\N	\N	f	f
64315a52-faa6-4e02-b218-d5e6ef444a8e	generic_instance	\N	4a67fb94-8ae8-4ca3-bc46-81ed67993cf1	MRX268	\N	\N	daylily	2025-07-03 08:24:16.35246-07	INSERT	\N	\N	\N	f	f
27e5db37-a48e-4240-a30d-1ca8a131e303	generic_instance_lineage	\N	d2b3c985-dbb8-46f7-adb8-22b0ee6bc0b7	GL688	\N	\N	daylily	2025-07-03 08:24:16.357321-07	INSERT	\N	\N	\N	f	f
ff35417e-224d-424b-a1be-3f5927dd9a83	generic_instance	\N	c5f72729-d226-4ab5-b263-52ecffcea927	MRX269	\N	\N	daylily	2025-07-03 08:24:16.357321-07	INSERT	\N	\N	\N	f	f
007dea33-f225-4db9-8356-541d12abd7d6	generic_instance_lineage	\N	de96d141-5742-4e82-a795-eda650bf3640	GL689	\N	\N	daylily	2025-07-03 08:24:16.361823-07	INSERT	\N	\N	\N	f	f
99862c77-2116-4314-bef5-0a8844695efd	generic_instance	\N	debc0fde-6d6a-4e56-899b-cd47f75025bb	MRX270	\N	\N	daylily	2025-07-03 08:24:16.361823-07	INSERT	\N	\N	\N	f	f
80b177ef-2b71-443f-925e-74d8477a4654	generic_instance_lineage	\N	9c5ff9e1-f03d-49e1-b457-255c47315246	GL690	\N	\N	daylily	2025-07-03 08:24:16.369633-07	INSERT	\N	\N	\N	f	f
6e3cbb7d-1f31-4961-b8e8-583e2d83b06f	generic_instance	\N	ebaff42d-7c7c-4d36-b7ce-02db9be33504	MRX271	\N	\N	daylily	2025-07-03 08:24:16.369633-07	INSERT	\N	\N	\N	f	f
22cd673c-c3b9-4127-8e56-71d5d0ffdf7b	generic_instance_lineage	\N	ff8cbce0-c5a2-48aa-9185-f755a67ef106	GL691	\N	\N	daylily	2025-07-03 08:24:16.376565-07	INSERT	\N	\N	\N	f	f
2ad2927d-dfb1-45d4-9870-a7e17444f28c	generic_instance	\N	19af9824-019d-472e-a8b7-e5b098186cd7	MRX272	\N	\N	daylily	2025-07-03 08:24:16.376565-07	INSERT	\N	\N	\N	f	f
e75d2084-4094-4024-8800-fb5bc74f1af3	generic_instance_lineage	\N	db514589-ad0b-43a6-9db4-9235d08dcebf	GL692	\N	\N	daylily	2025-07-03 08:24:16.381663-07	INSERT	\N	\N	\N	f	f
f61ef355-ab97-4084-8c44-9242db784b60	generic_instance	\N	c2af1c10-826b-403b-945b-5068a0fe4998	MRX273	\N	\N	daylily	2025-07-03 08:24:16.381663-07	INSERT	\N	\N	\N	f	f
dfc4ee00-c831-4982-b52f-b5a0748c39fa	generic_instance_lineage	\N	6c29cd28-4078-4524-b84c-3715029e5e39	GL693	\N	\N	daylily	2025-07-03 08:24:16.38591-07	INSERT	\N	\N	\N	f	f
74000ed2-b071-4b58-8b5d-661d8c0ddf27	generic_instance	\N	b9299cec-5b81-48be-b4e8-bcce06bf4d26	MRX274	\N	\N	daylily	2025-07-03 08:24:16.38591-07	INSERT	\N	\N	\N	f	f
8360d8f0-26b7-4cb0-8ae1-2462b07a7452	generic_instance_lineage	\N	87a949a8-d919-4371-8130-186f927a470c	GL694	\N	\N	daylily	2025-07-03 08:24:16.391185-07	INSERT	\N	\N	\N	f	f
255f79a9-5391-4963-b24d-07240fb29d0c	generic_instance	\N	ac122017-f1e1-449f-bdf3-39016e27fad8	MRX275	\N	\N	daylily	2025-07-03 08:24:16.391185-07	INSERT	\N	\N	\N	f	f
132367d7-3e94-4f72-80fb-1f560944c72d	generic_instance	json_addl	0bbf1234-7c4b-4afc-9352-56a600deb076	CX20	{"cogs": {"cost": "4.70", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "10ml glass tube", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "10ml glass tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "4.70", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "s ynthetic-cfdna", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "10ml glass tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.450581-07	UPDATE	\N	\N	\N	f	f
ed1ceb42-7dfb-49d1-b1ad-7f877770972d	generic_instance	created_dt	0bbf1234-7c4b-4afc-9352-56a600deb076	CX20	2025-07-03 08:24:16.445109-07	2025-07-03T08:24:16.445109-07:00	daylily	2025-07-03 08:24:16.450581-07	UPDATE	\N	\N	\N	f	f
67478692-c2f5-47f2-9845-1e566b27a406	generic_instance	modified_dt	0bbf1234-7c4b-4afc-9352-56a600deb076	CX20	2025-07-03 08:24:16.445109-07	2025-07-03T08:24:16.450581-07:00	daylily	2025-07-03 08:24:16.450581-07	UPDATE	\N	\N	\N	f	f
e5c3d14b-b3ce-4d80-be44-96ac8ca1a063	generic_instance_lineage	\N	ad8a63b8-3a0f-4014-8136-dfae38a1650c	GL702	\N	\N	daylily	2025-07-03 08:24:16.450581-07	INSERT	\N	\N	\N	f	f
4e353204-43cd-453c-9f72-1841d53d8d92	generic_instance	\N	f69f3bb2-4931-451c-ba90-a2f283ef8618	CX21	\N	\N	daylily	2025-07-03 08:24:16.454793-07	INSERT	\N	\N	\N	f	f
8411b08d-2140-4c40-8249-0d3b3b05eecd	generic_instance	\N	da2ba524-f58e-4e2f-bc8b-39610f57f804	MCX9	\N	\N	daylily	2025-07-03 08:24:16.457098-07	INSERT	\N	\N	\N	f	f
d7ca409f-4d69-4362-b811-0ce8e606e368	generic_instance	json_addl	f69f3bb2-4931-451c-ba90-a2f283ef8618	CX21	{"cogs": {"cost": "1.70", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "1.5ml eppi tube", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "1.5ml eppi tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "1.70", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "NTC - water", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "1.5ml eppi tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.459479-07	UPDATE	\N	\N	\N	f	f
fd339f0a-6c28-414f-9737-f7c03c96f311	generic_instance	created_dt	f69f3bb2-4931-451c-ba90-a2f283ef8618	CX21	2025-07-03 08:24:16.454793-07	2025-07-03T08:24:16.454793-07:00	daylily	2025-07-03 08:24:16.459479-07	UPDATE	\N	\N	\N	f	f
4102e4bb-f3fc-43be-9045-1a054139c186	generic_instance	modified_dt	f69f3bb2-4931-451c-ba90-a2f283ef8618	CX21	2025-07-03 08:24:16.454793-07	2025-07-03T08:24:16.459479-07:00	daylily	2025-07-03 08:24:16.459479-07	UPDATE	\N	\N	\N	f	f
2467ce8d-8ad3-4afd-8414-1e3e8b8636f5	generic_instance_lineage	\N	45a26d48-60fb-4705-acc9-432c8b0fc38f	GL703	\N	\N	daylily	2025-07-03 08:24:16.459479-07	INSERT	\N	\N	\N	f	f
b3ef164c-881d-478b-a988-c0ea4e8f0d5b	generic_instance	\N	7a653734-9941-44ab-9446-6ef7acee9702	CX22	\N	\N	daylily	2025-07-03 08:24:16.463243-07	INSERT	\N	\N	\N	f	f
5617c3b3-2929-417b-8b05-46e16ae842de	generic_instance	\N	70493c68-0506-4017-a8f9-5d14bdef4cf7	CWX281	\N	\N	daylily	2025-07-03 08:24:16.468168-07	INSERT	\N	\N	\N	f	f
9c4f31cd-8ac6-4861-abc6-5c49d7daa954	generic_instance	json_addl	70493c68-0506-4017-a8f9-5d14bdef4cf7	CWX281	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.471057-07	UPDATE	\N	\N	\N	f	f
f0676e04-c02a-4003-9d1e-4378c683a884	generic_instance	created_dt	70493c68-0506-4017-a8f9-5d14bdef4cf7	CWX281	2025-07-03 08:24:16.468168-07	2025-07-03T08:24:16.468168-07:00	daylily	2025-07-03 08:24:16.471057-07	UPDATE	\N	\N	\N	f	f
7b8ae359-ec38-40b6-abbf-707d19feee82	generic_instance	modified_dt	70493c68-0506-4017-a8f9-5d14bdef4cf7	CWX281	2025-07-03 08:24:16.468168-07	2025-07-03T08:24:16.471057-07:00	daylily	2025-07-03 08:24:16.471057-07	UPDATE	\N	\N	\N	f	f
f1bfca6a-ca6e-47c5-8567-67e92d555cf9	generic_instance_lineage	\N	831e7ece-a335-46e1-bda0-65c973485d87	GL704	\N	\N	daylily	2025-07-03 08:24:16.472381-07	INSERT	\N	\N	\N	f	f
0febfa52-9c7f-4c4c-a684-735f84e17cfb	generic_instance	\N	ee0ded45-6595-40e7-8c32-07e783b35d22	CWX282	\N	\N	daylily	2025-07-03 08:24:16.472381-07	INSERT	\N	\N	\N	f	f
de1abb47-9819-44a7-8740-c16a094d9711	generic_instance	json_addl	ee0ded45-6595-40e7-8c32-07e783b35d22	CWX282	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.476382-07	UPDATE	\N	\N	\N	f	f
130f2572-614c-4e11-8d76-6e1d1ef841cd	generic_instance	created_dt	ee0ded45-6595-40e7-8c32-07e783b35d22	CWX282	2025-07-03 08:24:16.472381-07	2025-07-03T08:24:16.472381-07:00	daylily	2025-07-03 08:24:16.476382-07	UPDATE	\N	\N	\N	f	f
f43ac024-2ff7-45f8-ad7b-1b1e28321ac1	generic_instance	modified_dt	ee0ded45-6595-40e7-8c32-07e783b35d22	CWX282	2025-07-03 08:24:16.472381-07	2025-07-03T08:24:16.476382-07:00	daylily	2025-07-03 08:24:16.476382-07	UPDATE	\N	\N	\N	f	f
fa62d4b5-1cff-45ae-ac85-30da1e1c6658	generic_instance_lineage	\N	37ee6bc8-6ce6-4894-beee-511f523e017b	GL705	\N	\N	daylily	2025-07-03 08:24:16.480503-07	INSERT	\N	\N	\N	f	f
cfdedbf3-ad4c-4c8a-8918-3fe9adb13b01	generic_instance	\N	367a6a1a-9157-4b01-b2ec-c2dc81af3154	CWX283	\N	\N	daylily	2025-07-03 08:24:16.480503-07	INSERT	\N	\N	\N	f	f
6dd1e220-8504-4261-9da0-e0e72e13ec5e	generic_instance	json_addl	367a6a1a-9157-4b01-b2ec-c2dc81af3154	CWX283	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.486322-07	UPDATE	\N	\N	\N	f	f
d87497fa-c42b-40da-b98e-23fa371ca6fa	generic_instance	created_dt	367a6a1a-9157-4b01-b2ec-c2dc81af3154	CWX283	2025-07-03 08:24:16.480503-07	2025-07-03T08:24:16.480503-07:00	daylily	2025-07-03 08:24:16.486322-07	UPDATE	\N	\N	\N	f	f
6e0c2ac2-c83b-41a7-8b66-f12adc3d0bb7	generic_instance	modified_dt	367a6a1a-9157-4b01-b2ec-c2dc81af3154	CWX283	2025-07-03 08:24:16.480503-07	2025-07-03T08:24:16.486322-07:00	daylily	2025-07-03 08:24:16.486322-07	UPDATE	\N	\N	\N	f	f
97565cf7-4f94-475d-9e90-2f994b3c70e9	generic_instance_lineage	\N	3729b464-b20a-4709-821a-cbe13b701d3c	GL706	\N	\N	daylily	2025-07-03 08:24:16.487499-07	INSERT	\N	\N	\N	f	f
dcff5c89-0e04-4720-9bbd-4bddfd637a85	generic_instance	\N	ca914991-5199-43a4-a369-2075a264f28d	CWX284	\N	\N	daylily	2025-07-03 08:24:16.487499-07	INSERT	\N	\N	\N	f	f
394f2bff-ab39-409c-b163-5b8f47cec15d	generic_instance	json_addl	ca914991-5199-43a4-a369-2075a264f28d	CWX284	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.491628-07	UPDATE	\N	\N	\N	f	f
dfa454bd-fdac-46ff-b544-7388924a5122	generic_instance	created_dt	ca914991-5199-43a4-a369-2075a264f28d	CWX284	2025-07-03 08:24:16.487499-07	2025-07-03T08:24:16.487499-07:00	daylily	2025-07-03 08:24:16.491628-07	UPDATE	\N	\N	\N	f	f
523f15ca-58c5-42b9-b083-b03670549a9b	generic_instance	modified_dt	ca914991-5199-43a4-a369-2075a264f28d	CWX284	2025-07-03 08:24:16.487499-07	2025-07-03T08:24:16.491628-07:00	daylily	2025-07-03 08:24:16.491628-07	UPDATE	\N	\N	\N	f	f
f8a85ae5-eaa4-4cb4-a32e-a347228b27fd	generic_instance_lineage	\N	00e60851-7b73-4002-b00f-417ab5c984c6	GL707	\N	\N	daylily	2025-07-03 08:24:16.492824-07	INSERT	\N	\N	\N	f	f
e68dabef-740c-475e-a48e-6fa2e882e55c	generic_instance	\N	9ab7fce2-54c8-491d-a3ec-17f3bc41ceba	CWX285	\N	\N	daylily	2025-07-03 08:24:16.492824-07	INSERT	\N	\N	\N	f	f
b1b4a1d6-1400-4edf-9025-1b2d1954f6a9	generic_instance	json_addl	9ab7fce2-54c8-491d-a3ec-17f3bc41ceba	CWX285	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.496667-07	UPDATE	\N	\N	\N	f	f
769c6f8b-ded0-4330-a86e-da2fd28f1d70	generic_instance	created_dt	9ab7fce2-54c8-491d-a3ec-17f3bc41ceba	CWX285	2025-07-03 08:24:16.492824-07	2025-07-03T08:24:16.492824-07:00	daylily	2025-07-03 08:24:16.496667-07	UPDATE	\N	\N	\N	f	f
34cd00f8-4634-43b0-ae44-025bf44642ae	generic_instance	modified_dt	9ab7fce2-54c8-491d-a3ec-17f3bc41ceba	CWX285	2025-07-03 08:24:16.492824-07	2025-07-03T08:24:16.496667-07:00	daylily	2025-07-03 08:24:16.496667-07	UPDATE	\N	\N	\N	f	f
ac70cd97-50ad-4f38-86a6-c77c2d440cef	generic_instance_lineage	\N	3a97dd5a-03c7-4b3b-8c66-8aa8437289a9	GL708	\N	\N	daylily	2025-07-03 08:24:16.497839-07	INSERT	\N	\N	\N	f	f
c838016d-9121-407d-a51e-e76f9e245ebd	generic_instance	\N	851ce518-5a99-4a87-ba96-8c4e2b8fde5c	CWX286	\N	\N	daylily	2025-07-03 08:24:16.497839-07	INSERT	\N	\N	\N	f	f
ef9665f9-7f31-45bb-8863-3c687a713f0e	generic_instance	json_addl	851ce518-5a99-4a87-ba96-8c4e2b8fde5c	CWX286	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.501632-07	UPDATE	\N	\N	\N	f	f
9b8a6717-76ea-41e8-adf1-5c6681b6ad02	generic_instance	created_dt	851ce518-5a99-4a87-ba96-8c4e2b8fde5c	CWX286	2025-07-03 08:24:16.497839-07	2025-07-03T08:24:16.497839-07:00	daylily	2025-07-03 08:24:16.501632-07	UPDATE	\N	\N	\N	f	f
fc21e4f4-917d-4531-92f5-9d511b18d21c	generic_instance	modified_dt	851ce518-5a99-4a87-ba96-8c4e2b8fde5c	CWX286	2025-07-03 08:24:16.497839-07	2025-07-03T08:24:16.501632-07:00	daylily	2025-07-03 08:24:16.501632-07	UPDATE	\N	\N	\N	f	f
638ec912-6375-43dc-a787-f1a2d03326a2	generic_instance_lineage	\N	906ce14e-6d9e-4ae1-aabc-37d5f4a0d903	GL709	\N	\N	daylily	2025-07-03 08:24:16.502909-07	INSERT	\N	\N	\N	f	f
1a6d5539-ae7c-4bb2-b8ab-e19852c85306	generic_instance	\N	b22b7269-47b8-44e6-a218-8c486c50cc0e	CWX287	\N	\N	daylily	2025-07-03 08:24:16.502909-07	INSERT	\N	\N	\N	f	f
88870fa0-993c-4168-803e-4b12eb6bdd69	generic_instance	json_addl	b22b7269-47b8-44e6-a218-8c486c50cc0e	CWX287	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.506763-07	UPDATE	\N	\N	\N	f	f
29c1de7f-5e58-47ab-8e32-03d659e2af5f	generic_instance	created_dt	b22b7269-47b8-44e6-a218-8c486c50cc0e	CWX287	2025-07-03 08:24:16.502909-07	2025-07-03T08:24:16.502909-07:00	daylily	2025-07-03 08:24:16.506763-07	UPDATE	\N	\N	\N	f	f
baf6aedf-63ec-4c0c-b9ed-a166c50fa91e	generic_instance	modified_dt	b22b7269-47b8-44e6-a218-8c486c50cc0e	CWX287	2025-07-03 08:24:16.502909-07	2025-07-03T08:24:16.506763-07:00	daylily	2025-07-03 08:24:16.506763-07	UPDATE	\N	\N	\N	f	f
d1a98807-6c69-40f5-ad9e-1a56208452ee	generic_instance_lineage	\N	d04e6d0d-7093-4e46-af1d-da2141e98e60	GL710	\N	\N	daylily	2025-07-03 08:24:16.507977-07	INSERT	\N	\N	\N	f	f
0385d1ef-fec8-4232-ba79-5f492a9c6490	generic_instance	\N	6ccd1298-f01c-4572-a51a-1defbf10da94	CWX288	\N	\N	daylily	2025-07-03 08:24:16.507977-07	INSERT	\N	\N	\N	f	f
34d81b91-6f95-4c9f-ac9d-bc721bf921c0	generic_instance	json_addl	6ccd1298-f01c-4572-a51a-1defbf10da94	CWX288	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.511867-07	UPDATE	\N	\N	\N	f	f
b4c56c3e-5f05-4143-807e-a29c06fc91e6	generic_instance	created_dt	6ccd1298-f01c-4572-a51a-1defbf10da94	CWX288	2025-07-03 08:24:16.507977-07	2025-07-03T08:24:16.507977-07:00	daylily	2025-07-03 08:24:16.511867-07	UPDATE	\N	\N	\N	f	f
29d40f71-f854-4103-9adc-cf5dba08f64d	generic_instance	modified_dt	6ccd1298-f01c-4572-a51a-1defbf10da94	CWX288	2025-07-03 08:24:16.507977-07	2025-07-03T08:24:16.511867-07:00	daylily	2025-07-03 08:24:16.511867-07	UPDATE	\N	\N	\N	f	f
c791e7bc-a928-4238-a331-6aa5affd1e21	generic_instance_lineage	\N	5474c331-bd88-414c-bb91-a3f7620b87f1	GL711	\N	\N	daylily	2025-07-03 08:24:16.513313-07	INSERT	\N	\N	\N	f	f
42ee4545-381f-4c66-9f00-8fbfda08a102	generic_instance	\N	0cfbb220-c556-449b-9a17-8f8b74170237	CWX289	\N	\N	daylily	2025-07-03 08:24:16.513313-07	INSERT	\N	\N	\N	f	f
4809a6dc-62b5-4e19-a210-702f55c9ef53	generic_instance	json_addl	0cfbb220-c556-449b-9a17-8f8b74170237	CWX289	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.517405-07	UPDATE	\N	\N	\N	f	f
334f7b7f-d5df-45cc-a72e-f3cccd7bbd80	generic_instance	created_dt	0cfbb220-c556-449b-9a17-8f8b74170237	CWX289	2025-07-03 08:24:16.513313-07	2025-07-03T08:24:16.513313-07:00	daylily	2025-07-03 08:24:16.517405-07	UPDATE	\N	\N	\N	f	f
9026f130-f7fa-4166-b13f-973c6f8efcee	generic_instance	modified_dt	0cfbb220-c556-449b-9a17-8f8b74170237	CWX289	2025-07-03 08:24:16.513313-07	2025-07-03T08:24:16.517405-07:00	daylily	2025-07-03 08:24:16.517405-07	UPDATE	\N	\N	\N	f	f
cacffdd2-ede7-4802-be8c-23038c381d38	generic_instance_lineage	\N	97981b53-a786-4f91-b525-94526300fea8	GL712	\N	\N	daylily	2025-07-03 08:24:16.518506-07	INSERT	\N	\N	\N	f	f
cc1e682b-4cea-4015-b11c-4deb1ea89604	generic_instance	\N	a20882a3-20fb-4e2a-b68c-ed393bb816d3	CWX290	\N	\N	daylily	2025-07-03 08:24:16.518506-07	INSERT	\N	\N	\N	f	f
a2ce55d9-ed24-4fe0-b220-495d3a10ef82	generic_instance	json_addl	a20882a3-20fb-4e2a-b68c-ed393bb816d3	CWX290	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.522203-07	UPDATE	\N	\N	\N	f	f
a71e534d-8efa-4287-9b77-0a28df39841f	generic_instance	created_dt	a20882a3-20fb-4e2a-b68c-ed393bb816d3	CWX290	2025-07-03 08:24:16.518506-07	2025-07-03T08:24:16.518506-07:00	daylily	2025-07-03 08:24:16.522203-07	UPDATE	\N	\N	\N	f	f
1db53476-51f2-4d16-adeb-87f4297f81e4	generic_instance	modified_dt	a20882a3-20fb-4e2a-b68c-ed393bb816d3	CWX290	2025-07-03 08:24:16.518506-07	2025-07-03T08:24:16.522203-07:00	daylily	2025-07-03 08:24:16.522203-07	UPDATE	\N	\N	\N	f	f
b767ce44-ee38-4980-98d8-4929e9e2757d	generic_instance_lineage	\N	d8cad5ee-1a75-4ad0-a6a5-9235af5f7253	GL713	\N	\N	daylily	2025-07-03 08:24:16.523496-07	INSERT	\N	\N	\N	f	f
d20a70e2-3252-4203-8185-98c6a6f383a1	generic_instance	\N	9bfaafce-9d17-4839-a5ea-03bbdfcc4b2e	CWX291	\N	\N	daylily	2025-07-03 08:24:16.523496-07	INSERT	\N	\N	\N	f	f
9bb07d2f-91f0-4811-96e9-e0a93305760f	generic_instance	json_addl	9bfaafce-9d17-4839-a5ea-03bbdfcc4b2e	CWX291	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.527786-07	UPDATE	\N	\N	\N	f	f
37cb3f5e-0bce-4861-94a7-61570a720403	generic_instance	created_dt	9bfaafce-9d17-4839-a5ea-03bbdfcc4b2e	CWX291	2025-07-03 08:24:16.523496-07	2025-07-03T08:24:16.523496-07:00	daylily	2025-07-03 08:24:16.527786-07	UPDATE	\N	\N	\N	f	f
650937cd-e9b8-446e-9b33-110c54a3e7ba	generic_instance	modified_dt	9bfaafce-9d17-4839-a5ea-03bbdfcc4b2e	CWX291	2025-07-03 08:24:16.523496-07	2025-07-03T08:24:16.527786-07:00	daylily	2025-07-03 08:24:16.527786-07	UPDATE	\N	\N	\N	f	f
1f6035d4-107f-470b-b27a-af30ae31f75d	generic_instance_lineage	\N	005fec53-7606-4f3c-8446-b16d33d11d2a	GL714	\N	\N	daylily	2025-07-03 08:24:16.528932-07	INSERT	\N	\N	\N	f	f
d10cc23f-7fb0-448b-81dd-9e26a9c4ac57	generic_instance	\N	2401c6c6-d364-44ec-94ae-fbe7825e8654	CWX292	\N	\N	daylily	2025-07-03 08:24:16.528932-07	INSERT	\N	\N	\N	f	f
1916ea4b-0818-4679-afef-0f7326ca66bb	generic_instance	json_addl	2401c6c6-d364-44ec-94ae-fbe7825e8654	CWX292	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.533495-07	UPDATE	\N	\N	\N	f	f
66ece6ad-63fe-445e-8469-ff05d2725413	generic_instance	created_dt	2401c6c6-d364-44ec-94ae-fbe7825e8654	CWX292	2025-07-03 08:24:16.528932-07	2025-07-03T08:24:16.528932-07:00	daylily	2025-07-03 08:24:16.533495-07	UPDATE	\N	\N	\N	f	f
5ff45001-c33a-4507-ad01-a45db17672e8	generic_instance	modified_dt	2401c6c6-d364-44ec-94ae-fbe7825e8654	CWX292	2025-07-03 08:24:16.528932-07	2025-07-03T08:24:16.533495-07:00	daylily	2025-07-03 08:24:16.533495-07	UPDATE	\N	\N	\N	f	f
e6641277-6d14-437a-9490-911fb325bd41	generic_instance_lineage	\N	302111cc-b4e3-4e62-a328-812238628850	GL715	\N	\N	daylily	2025-07-03 08:24:16.53635-07	INSERT	\N	\N	\N	f	f
9ff90990-4ea2-4bbe-bf19-a3be33c05d62	generic_instance	\N	2259a89d-1a1e-4541-a8f0-1316da02a81d	CWX293	\N	\N	daylily	2025-07-03 08:24:16.53635-07	INSERT	\N	\N	\N	f	f
190f942b-a0c6-4866-b250-d1afce7d3ab8	generic_instance	json_addl	2259a89d-1a1e-4541-a8f0-1316da02a81d	CWX293	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.541064-07	UPDATE	\N	\N	\N	f	f
86af0e74-ab28-42e0-b8c3-45bc4ec97b90	generic_instance	created_dt	2259a89d-1a1e-4541-a8f0-1316da02a81d	CWX293	2025-07-03 08:24:16.53635-07	2025-07-03T08:24:16.53635-07:00	daylily	2025-07-03 08:24:16.541064-07	UPDATE	\N	\N	\N	f	f
b4f3dac5-41f1-4470-a66e-504bf703b3b2	generic_instance	modified_dt	2259a89d-1a1e-4541-a8f0-1316da02a81d	CWX293	2025-07-03 08:24:16.53635-07	2025-07-03T08:24:16.541064-07:00	daylily	2025-07-03 08:24:16.541064-07	UPDATE	\N	\N	\N	f	f
59605ace-3a93-489a-9e69-9ff1d5d4d6dd	generic_instance_lineage	\N	ca04384d-bcd7-446d-99ca-c967f8c1f915	GL716	\N	\N	daylily	2025-07-03 08:24:16.542302-07	INSERT	\N	\N	\N	f	f
b285b874-0cd0-49e9-af35-2b938e5ea838	generic_instance	\N	5103010b-3a86-45d4-b1cd-0a9e77622b35	CWX294	\N	\N	daylily	2025-07-03 08:24:16.542302-07	INSERT	\N	\N	\N	f	f
18ea1317-a123-4546-a1c7-48dad6b128f3	generic_instance	json_addl	5103010b-3a86-45d4-b1cd-0a9e77622b35	CWX294	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.54621-07	UPDATE	\N	\N	\N	f	f
24716be9-bd5d-4bc5-8d07-37a220fc71af	generic_instance	created_dt	5103010b-3a86-45d4-b1cd-0a9e77622b35	CWX294	2025-07-03 08:24:16.542302-07	2025-07-03T08:24:16.542302-07:00	daylily	2025-07-03 08:24:16.54621-07	UPDATE	\N	\N	\N	f	f
0fda1ac4-d215-4bbc-af1e-7bfcad171b20	generic_instance	modified_dt	5103010b-3a86-45d4-b1cd-0a9e77622b35	CWX294	2025-07-03 08:24:16.542302-07	2025-07-03T08:24:16.54621-07:00	daylily	2025-07-03 08:24:16.54621-07	UPDATE	\N	\N	\N	f	f
ad4cc9b0-c944-43b1-a77f-adfe48be8bc7	generic_instance_lineage	\N	df5b0784-62cf-44ba-9ce0-885ce263408a	GL717	\N	\N	daylily	2025-07-03 08:24:16.547555-07	INSERT	\N	\N	\N	f	f
632869c9-6019-4fc4-a5ea-43b8caf1a52d	generic_instance	\N	b1be5bf3-c765-4f3c-a272-e128495d6a30	CWX295	\N	\N	daylily	2025-07-03 08:24:16.547555-07	INSERT	\N	\N	\N	f	f
82b5c082-a0e8-4cae-9fad-aec56f3fbb1c	generic_instance	json_addl	b1be5bf3-c765-4f3c-a272-e128495d6a30	CWX295	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.551494-07	UPDATE	\N	\N	\N	f	f
ae7e547e-3194-4683-97d3-b8653b17c4aa	generic_instance	created_dt	b1be5bf3-c765-4f3c-a272-e128495d6a30	CWX295	2025-07-03 08:24:16.547555-07	2025-07-03T08:24:16.547555-07:00	daylily	2025-07-03 08:24:16.551494-07	UPDATE	\N	\N	\N	f	f
2c2a4d85-472c-4695-8bd1-feffec413854	generic_instance	modified_dt	b1be5bf3-c765-4f3c-a272-e128495d6a30	CWX295	2025-07-03 08:24:16.547555-07	2025-07-03T08:24:16.551494-07:00	daylily	2025-07-03 08:24:16.551494-07	UPDATE	\N	\N	\N	f	f
2326cc3e-04a7-4176-8ef7-113e84750b37	generic_instance_lineage	\N	ae50d1bd-f281-4928-a723-b9889c8465e8	GL718	\N	\N	daylily	2025-07-03 08:24:16.55288-07	INSERT	\N	\N	\N	f	f
41bef891-9971-43df-ab1b-de2db3dbcdbf	generic_instance	\N	28a66e77-5293-4587-8aaa-2a11cd102349	CWX296	\N	\N	daylily	2025-07-03 08:24:16.55288-07	INSERT	\N	\N	\N	f	f
d115bca2-e5bc-4f25-9212-7793ddaf0c68	generic_instance	json_addl	28a66e77-5293-4587-8aaa-2a11cd102349	CWX296	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.556945-07	UPDATE	\N	\N	\N	f	f
24b19c16-53b4-44cc-a00c-56c5f807439e	generic_instance	created_dt	28a66e77-5293-4587-8aaa-2a11cd102349	CWX296	2025-07-03 08:24:16.55288-07	2025-07-03T08:24:16.55288-07:00	daylily	2025-07-03 08:24:16.556945-07	UPDATE	\N	\N	\N	f	f
a5f346ba-0921-4e74-af6d-43559d9573a1	generic_instance	modified_dt	28a66e77-5293-4587-8aaa-2a11cd102349	CWX296	2025-07-03 08:24:16.55288-07	2025-07-03T08:24:16.556945-07:00	daylily	2025-07-03 08:24:16.556945-07	UPDATE	\N	\N	\N	f	f
29773622-45e7-4ab9-8667-f5adf31c62e9	generic_instance_lineage	\N	efc3906c-c305-45f3-a4eb-ff93349191af	GL719	\N	\N	daylily	2025-07-03 08:24:16.558048-07	INSERT	\N	\N	\N	f	f
bc30af41-5dd8-4402-8677-995ae5cc0c6a	generic_instance	\N	344aff1a-652b-4802-8b45-87e345a5dfaf	CWX297	\N	\N	daylily	2025-07-03 08:24:16.558048-07	INSERT	\N	\N	\N	f	f
a7436962-1054-4b83-b28f-1745c686d5cf	generic_instance	json_addl	344aff1a-652b-4802-8b45-87e345a5dfaf	CWX297	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.561707-07	UPDATE	\N	\N	\N	f	f
de4cb1c1-52e7-4c06-bc8b-ffe4bc82d0e0	generic_instance	created_dt	344aff1a-652b-4802-8b45-87e345a5dfaf	CWX297	2025-07-03 08:24:16.558048-07	2025-07-03T08:24:16.558048-07:00	daylily	2025-07-03 08:24:16.561707-07	UPDATE	\N	\N	\N	f	f
751ca76c-b000-4671-9762-df2d3ed8dfa6	generic_instance	modified_dt	344aff1a-652b-4802-8b45-87e345a5dfaf	CWX297	2025-07-03 08:24:16.558048-07	2025-07-03T08:24:16.561707-07:00	daylily	2025-07-03 08:24:16.561707-07	UPDATE	\N	\N	\N	f	f
a3e1b0ef-78c6-4601-b35c-b753411c8088	generic_instance_lineage	\N	bd6f0aed-b3a3-46be-8467-caf81efef5ef	GL720	\N	\N	daylily	2025-07-03 08:24:16.562895-07	INSERT	\N	\N	\N	f	f
b6678747-dcb0-465e-ae8e-09178e3e80ec	generic_instance	\N	e41faaaf-ce2a-43d6-87e5-e3870abe45d6	CWX298	\N	\N	daylily	2025-07-03 08:24:16.562895-07	INSERT	\N	\N	\N	f	f
0dc6ab6b-7c2d-4e1c-ba6f-239ff07daf8d	generic_instance	json_addl	e41faaaf-ce2a-43d6-87e5-e3870abe45d6	CWX298	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.568162-07	UPDATE	\N	\N	\N	f	f
7c7c5c14-74aa-4e99-9d8f-3fd3b312ce53	generic_instance	created_dt	e41faaaf-ce2a-43d6-87e5-e3870abe45d6	CWX298	2025-07-03 08:24:16.562895-07	2025-07-03T08:24:16.562895-07:00	daylily	2025-07-03 08:24:16.568162-07	UPDATE	\N	\N	\N	f	f
bee83ef8-35a6-490b-9fbf-8ec4546d7bf5	generic_instance	modified_dt	e41faaaf-ce2a-43d6-87e5-e3870abe45d6	CWX298	2025-07-03 08:24:16.562895-07	2025-07-03T08:24:16.568162-07:00	daylily	2025-07-03 08:24:16.568162-07	UPDATE	\N	\N	\N	f	f
64fbbcd5-e9d9-445c-9794-f87d0dc1e270	generic_instance_lineage	\N	55c65ed9-ca79-42ea-a36b-77ee8ca2f285	GL721	\N	\N	daylily	2025-07-03 08:24:16.569326-07	INSERT	\N	\N	\N	f	f
98e9066b-f29a-48ae-a445-ef36cf852109	generic_instance	\N	006ca986-054c-4e9f-8e46-dac3dc562a98	CWX299	\N	\N	daylily	2025-07-03 08:24:16.569326-07	INSERT	\N	\N	\N	f	f
fdaab2f4-78f2-448d-b8c6-7fda5829f5cb	generic_instance	json_addl	006ca986-054c-4e9f-8e46-dac3dc562a98	CWX299	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.573033-07	UPDATE	\N	\N	\N	f	f
4b150c11-e1c1-4a20-aadf-82bda5e8eca1	generic_instance	created_dt	006ca986-054c-4e9f-8e46-dac3dc562a98	CWX299	2025-07-03 08:24:16.569326-07	2025-07-03T08:24:16.569326-07:00	daylily	2025-07-03 08:24:16.573033-07	UPDATE	\N	\N	\N	f	f
914c494a-f3c5-4f1c-82c8-e7f7ae5c3259	generic_instance	modified_dt	006ca986-054c-4e9f-8e46-dac3dc562a98	CWX299	2025-07-03 08:24:16.569326-07	2025-07-03T08:24:16.573033-07:00	daylily	2025-07-03 08:24:16.573033-07	UPDATE	\N	\N	\N	f	f
910fbd80-9823-4dda-9510-7c40076df47e	generic_instance_lineage	\N	8cff34c6-56b9-4877-b32e-62a59c26898c	GL722	\N	\N	daylily	2025-07-03 08:24:16.574559-07	INSERT	\N	\N	\N	f	f
cadc4c2e-9536-4ad4-abc0-8a95f141103d	generic_instance	\N	233d4349-e767-46bf-a7ed-035e092867e8	CWX300	\N	\N	daylily	2025-07-03 08:24:16.574559-07	INSERT	\N	\N	\N	f	f
898db681-bfc8-49cc-94fb-bd186a9a89e8	generic_instance	json_addl	233d4349-e767-46bf-a7ed-035e092867e8	CWX300	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.579865-07	UPDATE	\N	\N	\N	f	f
29f37569-a0a9-4f53-993d-9140bd5d22f8	generic_instance	created_dt	233d4349-e767-46bf-a7ed-035e092867e8	CWX300	2025-07-03 08:24:16.574559-07	2025-07-03T08:24:16.574559-07:00	daylily	2025-07-03 08:24:16.579865-07	UPDATE	\N	\N	\N	f	f
5d16edb1-ea56-41f5-8f51-16ecbbf37f13	generic_instance	modified_dt	233d4349-e767-46bf-a7ed-035e092867e8	CWX300	2025-07-03 08:24:16.574559-07	2025-07-03T08:24:16.579865-07:00	daylily	2025-07-03 08:24:16.579865-07	UPDATE	\N	\N	\N	f	f
d2d6ae4e-7d6f-4c8c-99c0-401bc3781b58	generic_instance_lineage	\N	f79e64eb-f5b1-4911-95ab-d30eeed29d5c	GL723	\N	\N	daylily	2025-07-03 08:24:16.581962-07	INSERT	\N	\N	\N	f	f
8746106c-6a9a-4a95-a8d8-e26f397fa78f	generic_instance	\N	f20b5ca4-d90b-4558-8f3a-088e0959cb77	CWX301	\N	\N	daylily	2025-07-03 08:24:16.581962-07	INSERT	\N	\N	\N	f	f
7630abe2-1a82-4401-a5c9-92bf0f0af853	generic_instance	json_addl	f20b5ca4-d90b-4558-8f3a-088e0959cb77	CWX301	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.588873-07	UPDATE	\N	\N	\N	f	f
e27be677-0a82-47f9-9447-acebad90748c	generic_instance	created_dt	f20b5ca4-d90b-4558-8f3a-088e0959cb77	CWX301	2025-07-03 08:24:16.581962-07	2025-07-03T08:24:16.581962-07:00	daylily	2025-07-03 08:24:16.588873-07	UPDATE	\N	\N	\N	f	f
25a6dc77-f6eb-49ac-be83-ef32b460ac5d	generic_instance	modified_dt	f20b5ca4-d90b-4558-8f3a-088e0959cb77	CWX301	2025-07-03 08:24:16.581962-07	2025-07-03T08:24:16.588873-07:00	daylily	2025-07-03 08:24:16.588873-07	UPDATE	\N	\N	\N	f	f
dde0e244-2140-4177-b04e-1c96fb4b0c4d	generic_instance_lineage	\N	12e3d02c-641c-41c3-90a9-f74a92a0a132	GL724	\N	\N	daylily	2025-07-03 08:24:16.590183-07	INSERT	\N	\N	\N	f	f
f8c2a10e-e301-4c5f-8a2c-dd2f1f14fadc	generic_instance	\N	35620e62-853f-41df-b226-7036aae429e1	CWX302	\N	\N	daylily	2025-07-03 08:24:16.590183-07	INSERT	\N	\N	\N	f	f
8233a10c-b786-4b97-8495-b2522e4dd471	generic_instance	json_addl	35620e62-853f-41df-b226-7036aae429e1	CWX302	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.594043-07	UPDATE	\N	\N	\N	f	f
86064e7f-55ac-4407-b6de-66547118c1f2	generic_instance	created_dt	35620e62-853f-41df-b226-7036aae429e1	CWX302	2025-07-03 08:24:16.590183-07	2025-07-03T08:24:16.590183-07:00	daylily	2025-07-03 08:24:16.594043-07	UPDATE	\N	\N	\N	f	f
c7761f5c-2a0c-428f-b7e1-692445ee1e4b	generic_instance	modified_dt	35620e62-853f-41df-b226-7036aae429e1	CWX302	2025-07-03 08:24:16.590183-07	2025-07-03T08:24:16.594043-07:00	daylily	2025-07-03 08:24:16.594043-07	UPDATE	\N	\N	\N	f	f
e07ba7e7-dfaa-4a72-aca7-d0371d5fbb59	generic_instance_lineage	\N	a4faa36f-4c23-4978-8c13-7289003c95b4	GL725	\N	\N	daylily	2025-07-03 08:24:16.595467-07	INSERT	\N	\N	\N	f	f
4aa3ea99-93ba-4e8e-b7d3-c9301bb3445d	generic_instance	\N	156a6879-7978-4ba2-9f2d-be6064697797	CWX303	\N	\N	daylily	2025-07-03 08:24:16.595467-07	INSERT	\N	\N	\N	f	f
3e074172-d8b7-4431-84c3-f3df046ef25e	generic_instance	json_addl	156a6879-7978-4ba2-9f2d-be6064697797	CWX303	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.600109-07	UPDATE	\N	\N	\N	f	f
4451656f-ade2-4c78-99d6-313d6a7225c9	generic_instance	created_dt	156a6879-7978-4ba2-9f2d-be6064697797	CWX303	2025-07-03 08:24:16.595467-07	2025-07-03T08:24:16.595467-07:00	daylily	2025-07-03 08:24:16.600109-07	UPDATE	\N	\N	\N	f	f
8d7e171b-ed50-41cc-909b-18f0c34b6537	generic_instance	modified_dt	156a6879-7978-4ba2-9f2d-be6064697797	CWX303	2025-07-03 08:24:16.595467-07	2025-07-03T08:24:16.600109-07:00	daylily	2025-07-03 08:24:16.600109-07	UPDATE	\N	\N	\N	f	f
b022c003-1817-4dd5-a4b2-e781774833e7	generic_instance_lineage	\N	ed81f171-7bbd-4ceb-a455-c7e4afab5e8c	GL726	\N	\N	daylily	2025-07-03 08:24:16.601469-07	INSERT	\N	\N	\N	f	f
c4b4cd4b-2fb2-4e4e-9f6e-fb3d7f578eba	generic_instance	\N	121e9047-ad3e-4dfd-823b-83e1fc717cbc	CWX304	\N	\N	daylily	2025-07-03 08:24:16.601469-07	INSERT	\N	\N	\N	f	f
5aafd2b0-6428-446a-87ec-12a2258db836	generic_instance	json_addl	121e9047-ad3e-4dfd-823b-83e1fc717cbc	CWX304	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.605983-07	UPDATE	\N	\N	\N	f	f
3f316bf0-5842-4582-99c5-961c8aafb2c4	generic_instance	created_dt	121e9047-ad3e-4dfd-823b-83e1fc717cbc	CWX304	2025-07-03 08:24:16.601469-07	2025-07-03T08:24:16.601469-07:00	daylily	2025-07-03 08:24:16.605983-07	UPDATE	\N	\N	\N	f	f
3cf7b01f-5432-4621-884a-15469ea8b0bf	generic_instance	modified_dt	121e9047-ad3e-4dfd-823b-83e1fc717cbc	CWX304	2025-07-03 08:24:16.601469-07	2025-07-03T08:24:16.605983-07:00	daylily	2025-07-03 08:24:16.605983-07	UPDATE	\N	\N	\N	f	f
ae3d3e56-e2ea-4344-aebd-cd90f779a624	generic_instance_lineage	\N	c702cef1-171b-459e-a744-0950c11c0716	GL727	\N	\N	daylily	2025-07-03 08:24:16.608243-07	INSERT	\N	\N	\N	f	f
820ea349-48b3-46d9-8013-8ff07f46e8b1	generic_instance	\N	fda6b968-9927-4cf6-85f9-c44bb39014a7	CWX305	\N	\N	daylily	2025-07-03 08:24:16.608243-07	INSERT	\N	\N	\N	f	f
d94651ba-a186-430f-a2dc-8140e6c7e23a	generic_instance	json_addl	fda6b968-9927-4cf6-85f9-c44bb39014a7	CWX305	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.61229-07	UPDATE	\N	\N	\N	f	f
849aeada-5d76-4164-99f7-ffde0c42c17b	generic_instance	created_dt	fda6b968-9927-4cf6-85f9-c44bb39014a7	CWX305	2025-07-03 08:24:16.608243-07	2025-07-03T08:24:16.608243-07:00	daylily	2025-07-03 08:24:16.61229-07	UPDATE	\N	\N	\N	f	f
ec543718-5e4f-4d8c-b68b-8a1ed72262cd	generic_instance	modified_dt	fda6b968-9927-4cf6-85f9-c44bb39014a7	CWX305	2025-07-03 08:24:16.608243-07	2025-07-03T08:24:16.61229-07:00	daylily	2025-07-03 08:24:16.61229-07	UPDATE	\N	\N	\N	f	f
810d139b-0944-49f1-860f-d8b8fb8a0b81	generic_instance_lineage	\N	6887fb1a-1e4b-4eab-aa5a-2ef67ad1380c	GL728	\N	\N	daylily	2025-07-03 08:24:16.614298-07	INSERT	\N	\N	\N	f	f
471487c0-8c64-4b99-bd6a-8ebf56151cfe	generic_instance	\N	c7124a90-7b46-478e-b74a-dacd7a07dc62	CWX306	\N	\N	daylily	2025-07-03 08:24:16.614298-07	INSERT	\N	\N	\N	f	f
e5777631-350f-4af1-b137-b28d44e40635	generic_instance	json_addl	c7124a90-7b46-478e-b74a-dacd7a07dc62	CWX306	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.618636-07	UPDATE	\N	\N	\N	f	f
f2f640be-e4da-4b1f-8bc6-e8257d423bc9	generic_instance	created_dt	c7124a90-7b46-478e-b74a-dacd7a07dc62	CWX306	2025-07-03 08:24:16.614298-07	2025-07-03T08:24:16.614298-07:00	daylily	2025-07-03 08:24:16.618636-07	UPDATE	\N	\N	\N	f	f
96130081-bba6-4be5-a6ab-2586a1a76a5c	generic_instance	modified_dt	c7124a90-7b46-478e-b74a-dacd7a07dc62	CWX306	2025-07-03 08:24:16.614298-07	2025-07-03T08:24:16.618636-07:00	daylily	2025-07-03 08:24:16.618636-07	UPDATE	\N	\N	\N	f	f
e7839444-2a4e-4168-9b17-bf72ac1ed7e9	generic_instance_lineage	\N	4e0c9980-2ea4-4add-90ef-ec8da33fb859	GL729	\N	\N	daylily	2025-07-03 08:24:16.619981-07	INSERT	\N	\N	\N	f	f
1a957c3e-bab3-4149-953c-b64f2ee620ae	generic_instance	\N	cb5b12ec-e9e7-4707-bffc-0e453d1fc960	CWX307	\N	\N	daylily	2025-07-03 08:24:16.619981-07	INSERT	\N	\N	\N	f	f
63bf0654-7c80-42c9-868e-328199f51849	generic_instance	json_addl	cb5b12ec-e9e7-4707-bffc-0e453d1fc960	CWX307	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.62407-07	UPDATE	\N	\N	\N	f	f
e812875f-f1ef-4a0b-98f0-57881632a243	generic_instance	created_dt	cb5b12ec-e9e7-4707-bffc-0e453d1fc960	CWX307	2025-07-03 08:24:16.619981-07	2025-07-03T08:24:16.619981-07:00	daylily	2025-07-03 08:24:16.62407-07	UPDATE	\N	\N	\N	f	f
6985cb4e-477c-4949-b68a-5e45bc6dfd5a	generic_instance	modified_dt	cb5b12ec-e9e7-4707-bffc-0e453d1fc960	CWX307	2025-07-03 08:24:16.619981-07	2025-07-03T08:24:16.62407-07:00	daylily	2025-07-03 08:24:16.62407-07	UPDATE	\N	\N	\N	f	f
904ab28f-b2b1-4cf5-82c5-0f6d50e0e0c9	generic_instance_lineage	\N	57705daa-7d7d-466d-a46d-4c773f2ae7f4	GL730	\N	\N	daylily	2025-07-03 08:24:16.62529-07	INSERT	\N	\N	\N	f	f
f4dd3ed5-94dc-4440-8ed0-8024f39fdf69	generic_instance	\N	ffefe78a-8ac0-4e3e-9fa9-772ee8762f39	CWX308	\N	\N	daylily	2025-07-03 08:24:16.62529-07	INSERT	\N	\N	\N	f	f
7540d47e-3fdf-49d3-abc7-0a01ee25d9f3	generic_instance	json_addl	ffefe78a-8ac0-4e3e-9fa9-772ee8762f39	CWX308	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	daylily	2025-07-03 08:24:16.629386-07	UPDATE	\N	\N	\N	f	f
8e90b4bf-5651-4497-b66b-b8eae572d57d	generic_instance	created_dt	ffefe78a-8ac0-4e3e-9fa9-772ee8762f39	CWX308	2025-07-03 08:24:16.62529-07	2025-07-03T08:24:16.62529-07:00	daylily	2025-07-03 08:24:16.629386-07	UPDATE	\N	\N	\N	f	f
b6f72d44-0678-4c52-acfa-38dfde6cd0d8	generic_instance	modified_dt	ffefe78a-8ac0-4e3e-9fa9-772ee8762f39	CWX308	2025-07-03 08:24:16.62529-07	2025-07-03T08:24:16.629386-07:00	daylily	2025-07-03 08:24:16.629386-07	UPDATE	\N	\N	\N	f	f
2ad3a9f2-af46-4122-96f6-b453f07d6110	generic_instance_lineage	\N	1ef8a199-a281-46ef-80b9-030ab303de45	GL731	\N	\N	daylily	2025-07-03 08:24:16.630723-07	INSERT	\N	\N	\N	f	f
63fbe9a5-d0c5-489d-af6a-7699f201133c	generic_instance	\N	3ebb20a3-6d70-4ebb-9a8d-f070a39a60f0	MRX281	\N	\N	daylily	2025-07-03 08:24:16.633054-07	INSERT	\N	\N	\N	f	f
42731a40-d192-4fe6-b347-e8d43f7dd822	generic_instance_lineage	\N	e57f97ae-5c0b-4b4d-9ea5-587adbdedea1	GL732	\N	\N	daylily	2025-07-03 08:24:16.637466-07	INSERT	\N	\N	\N	f	f
2d292a6e-f0e0-4e7d-a224-848ccf3eb2e5	generic_instance	\N	5d6d4e76-14b5-4009-9f05-a8e3062bb797	MRX282	\N	\N	daylily	2025-07-03 08:24:16.637466-07	INSERT	\N	\N	\N	f	f
11696b65-6ebf-4133-81e9-56873887f306	generic_instance_lineage	\N	1ba7c0f0-996a-42c9-acc0-88db181a375f	GL733	\N	\N	daylily	2025-07-03 08:24:16.643287-07	INSERT	\N	\N	\N	f	f
9a6d6d64-4509-4293-85ca-6eaff492beb9	generic_instance	\N	a9013d2d-2a96-4508-b74a-6142dc7eaacd	MRX283	\N	\N	daylily	2025-07-03 08:24:16.643287-07	INSERT	\N	\N	\N	f	f
a540a943-a150-441f-aa7e-2a2146f3b38d	generic_instance_lineage	\N	478bd241-2058-4b89-a722-537011794ff8	GL734	\N	\N	daylily	2025-07-03 08:24:16.649412-07	INSERT	\N	\N	\N	f	f
70cb504f-b065-4575-9ab2-45a5164704c0	generic_instance	\N	2355589e-b1e9-469d-9593-602b77227c24	MRX284	\N	\N	daylily	2025-07-03 08:24:16.649412-07	INSERT	\N	\N	\N	f	f
7cfc7b2f-8016-4094-b832-e3362fe13a46	generic_instance_lineage	\N	9faad57c-286a-477c-bfef-71bed6709f1c	GL735	\N	\N	daylily	2025-07-03 08:24:16.655929-07	INSERT	\N	\N	\N	f	f
\.


--
-- Data for Name: generic_instance; Type: TABLE DATA; Schema: public; Owner: daylily
--

COPY public.generic_instance (uuid, euid, name, json_addl, btype, polymorphic_discriminator, super_type, b_sub_type, created_dt, version, bstatus, is_deleted, template_uuid, modified_dt, is_singleton) FROM stdin;
a95a3806-06d0-4ce3-bf5b-042fb400562b	FS1	file_set:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T125759", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	file_set	file_instance	file	generic	2025-07-03 05:57:59.406061-07	1.0	ready	f	1c57e0bf-8b44-426b-8f9e-e96aff46a341	2025-07-03 05:57:59.406061-07	f
c47bda8d-6c52-4cf8-83d6-213aaa276e3f	FS13	file_set:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T143038", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	file_set	file_instance	file	generic	2025-07-03 07:30:38.949974-07	1.0	ready	f	1c57e0bf-8b44-426b-8f9e-e96aff46a341	2025-07-03 07:30:38.949974-07	f
245c6374-146a-4bf6-898d-e3a0ce08f78d	FS25	file_set:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144901", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	file_set	file_instance	file	generic	2025-07-03 07:49:08.725504-07	1.0	ready	f	1c57e0bf-8b44-426b-8f9e-e96aff46a341	2025-07-03 07:49:08.725504-07	f
01045846-479f-43f7-8d20-e84e3bec3420	FI7003	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7003.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7003.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T125759", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 05:57:59.595243-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 05:57:59.944771-07	f
c76338b6-6ba8-4dc2-9b60-eb20a9e253a6	FI7004	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7004.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7004.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T125759", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 05:58:00.957798-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 05:58:01.279603-07	f
6dbd1b9c-394e-43ad-9935-dce8db433908	FI7005	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7005.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7005.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T125759", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 05:58:02.258533-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 05:58:02.539249-07	f
2daf2aa2-3633-4fd6-a8a4-784977febd55	FI7006	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7006.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7006.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T125759", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 05:58:03.395505-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 05:58:03.705654-07	f
c8552b03-f549-47c5-962f-bb123d63fe02	FI7007	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "backup_2025-07-03_00-06-33.sql", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "sql", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7007.sql", "current_s3_uri": "s3://daylily-dewey-0/0/FI7007.sql", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T130014", "original_file_md5": "", "original_file_name": "backup_2025-07-03_00-06-33.sql", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "sql", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1012876, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 06:00:25.954212-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 06:00:26.245305-07	f
c9f38309-e10a-4fae-a3d0-f1f8488f7eae	FI7008	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7008.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7008.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T130538", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 06:05:38.980249-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 06:05:39.314807-07	f
c98ef464-2308-47c8-aaa7-cd1cee22edd1	FI7009	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7009.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7009.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T130538", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 06:05:40.43767-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 06:05:40.788879-07	f
25cb2b88-6898-400b-9282-3285aa3cd2e6	FI7010	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7010.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7010.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T130538", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 06:05:41.701199-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 06:05:41.939746-07	f
6f5b9dcf-45e5-43cc-b340-ed644031f886	FI7011	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7011.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7011.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T130538", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 06:05:42.664112-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 06:05:43.012234-07	f
8c274106-80d4-4ca6-b317-2a1ad9cc1f7c	FX1	shared_ref:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "", "status": "active", "comments": "", "who_pays": "", "visibility": null, "end_datetime": "2025-07-04T13:13:20.881682+00:00", "presigned_url": "https://daylily-dewey-0.s3.amazonaws.com/0/FI7007.sql?AWSAccessKeyId=AKIAZYG7WWMZG2QOPT5Q&Signature=C4RJYCxHnMlvDZ7Dw0EXLXUofu8%3D&Expires=1751634800", "rclone_config": {}, "intended_users": [""], "reference_type": "presigned", "start_datetime": "2025-07-03T13:13:20.881682+00:00", "valid_duration": 86400.0}, "description": "Generic File Reference", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"form_label": "Reference Type", "property_key": "reference_type"}, {"form_label": "Visibility", "property_key": "visibility"}, {"form_label": "Reference Status", "property_key": "status"}, {"form_label": "Duration", "property_key": "valid_duration"}, {"form_label": "Start Datetime", "property_key": "start_datetime"}, {"form_label": "End Datetime", "property_key": "end_datetime"}, {"form_label": "Who Pays", "property_key": "who_pays"}], "controlled_properties": {}, "instantiation_layouts": [], "arbitrary_association_strings": []}	shared_ref	file_instance	file	generic	2025-07-03 06:13:20.882236-07	1.0	ready	f	64646531-f2fd-463a-9790-a4587d83f2a5	2025-07-03 06:13:20.882236-07	f
f7c652b1-51bc-483b-a485-8d70c3630310	FS2	file_set:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T130014", "tags": [], "comments": "", "duration": 1.0, "lab_code": "", "ref_type": "presigned_url", "description": "File set created by Dewey file manager", "creating_user": "john@daylilyinformatics.com", "rclone_config": {"host": "0.0.0.0", "port": 8080, "user": "user", "bucket": "daylily-dewey-0", "passwd": "passwd"}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	file_set	file_instance	file	generic	2025-07-03 06:00:25.841409-07	1.0	ready	f	1c57e0bf-8b44-426b-8f9e-e96aff46a341	2025-07-03 06:13:20.875767-07	f
44f71e98-963f-4ee5-90c1-d4e544a735da	FI7012	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "backup_2025-07-03_00-06-33.sql", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "sql", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7012.sql", "current_s3_uri": "s3://daylily-dewey-0/0/FI7012.sql", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134027", "original_file_md5": "", "original_file_name": "backup_2025-07-03_00-06-33.sql", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "sql", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1012876, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 06:40:41.075897-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 06:40:41.379167-07	f
c54a4608-c3e9-4453-83b3-302ca39467f3	FS5	file_set:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T134109", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	file_set	file_instance	file	generic	2025-07-03 06:41:43.824381-07	1.0	ready	f	1c57e0bf-8b44-426b-8f9e-e96aff46a341	2025-07-03 06:41:43.824381-07	f
5aa56483-8480-4d03-a3f2-36506ae89d6c	FI7013	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "NA12878/HG001_GRCh38_1_22_v4.2.1_benchmark_hifiasm_v11_phasetransfer.PS_FIX.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7013.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7013.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134109", "original_file_md5": "", "original_file_name": "HG001_GRCh38_1_22_v4.2.1_benchmark_hifiasm_v11_phasetransfer.PS_FIX.vcf.gz", "original_file_path": "NA12878", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 111830815, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 06:41:43.893844-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 06:41:44.167933-07	f
9312ed99-a2e8-45a4-9f4d-802f27c08ea5	FI7014	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "NA12878/Lee_washington_0250E_19948.pdf", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "pdf", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7014.pdf", "current_s3_uri": "s3://daylily-dewey-0/0/FI7014.pdf", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134109", "original_file_md5": "", "original_file_name": "Lee_washington_0250E_19948.pdf", "original_file_path": "NA12878", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "pdf", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 11203779, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 06:41:49.285358-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 06:41:49.405737-07	f
d77e413a-c96e-4048-b4a7-a5ca4c00828a	FS6	file_set:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T134212", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	file_set	file_instance	file	generic	2025-07-03 06:42:12.436379-07	1.0	ready	f	1c57e0bf-8b44-426b-8f9e-e96aff46a341	2025-07-03 06:42:12.436379-07	f
675337ff-ab88-4ba5-bce0-a80e5446111e	FI7015	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7015.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7015.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T134212", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 06:42:12.453434-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 06:42:12.710245-07	f
199bb1af-804b-411a-b0ba-f270dc958bb5	FI7016	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7016.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7016.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T134212", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 06:42:13.769262-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 06:42:14.076936-07	f
b5eca29d-be2c-4c68-8f21-61819fe90f20	FI7017	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7017.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7017.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T134212", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 06:42:14.980617-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 06:42:15.2375-07	f
7ee95105-e62b-4f4a-aadf-e76bc8bf12d1	FI7018	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7018.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7018.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T134212", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 06:42:15.917897-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 06:42:16.202901-07	f
d7eb218e-e19e-402c-b405-ba307911e0ff	FX2	shared_ref:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "", "status": "active", "comments": "", "who_pays": "", "visibility": null, "end_datetime": "2025-07-04T13:42:33.389713+00:00", "presigned_url": "https://daylily-dewey-0.s3.amazonaws.com/0/FI7008.png?AWSAccessKeyId=AKIAZYG7WWMZG2QOPT5Q&Signature=%2F%2BFmquTcNonZvjiNpR48aVbCdRY%3D&Expires=1751636553", "rclone_config": {}, "intended_users": [""], "reference_type": "presigned", "start_datetime": "2025-07-03T13:42:33.389713+00:00", "valid_duration": 86400.0}, "description": "Generic File Reference", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"form_label": "Reference Type", "property_key": "reference_type"}, {"form_label": "Visibility", "property_key": "visibility"}, {"form_label": "Reference Status", "property_key": "status"}, {"form_label": "Duration", "property_key": "valid_duration"}, {"form_label": "Start Datetime", "property_key": "start_datetime"}, {"form_label": "End Datetime", "property_key": "end_datetime"}, {"form_label": "Who Pays", "property_key": "who_pays"}], "controlled_properties": {}, "instantiation_layouts": [], "arbitrary_association_strings": []}	shared_ref	file_instance	file	generic	2025-07-03 06:42:33.390481-07	1.0	ready	f	64646531-f2fd-463a-9790-a4587d83f2a5	2025-07-03 06:42:33.390481-07	f
6531f3d5-1217-4d66-93f7-7e40c65d72bf	FX3	shared_ref:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "", "status": "active", "comments": "", "who_pays": "", "visibility": null, "end_datetime": "2025-07-04T13:42:33.419840+00:00", "presigned_url": "https://daylily-dewey-0.s3.amazonaws.com/0/FI7009.tsv?AWSAccessKeyId=AKIAZYG7WWMZG2QOPT5Q&Signature=eFiz0YY269I%2B2dYU0PuBf%2FxTtvs%3D&Expires=1751636553", "rclone_config": {}, "intended_users": [""], "reference_type": "presigned", "start_datetime": "2025-07-03T13:42:33.419840+00:00", "valid_duration": 86400.0}, "description": "Generic File Reference", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"form_label": "Reference Type", "property_key": "reference_type"}, {"form_label": "Visibility", "property_key": "visibility"}, {"form_label": "Reference Status", "property_key": "status"}, {"form_label": "Duration", "property_key": "valid_duration"}, {"form_label": "Start Datetime", "property_key": "start_datetime"}, {"form_label": "End Datetime", "property_key": "end_datetime"}, {"form_label": "Who Pays", "property_key": "who_pays"}], "controlled_properties": {}, "instantiation_layouts": [], "arbitrary_association_strings": []}	shared_ref	file_instance	file	generic	2025-07-03 06:42:33.420047-07	1.0	ready	f	64646531-f2fd-463a-9790-a4587d83f2a5	2025-07-03 06:42:33.420047-07	f
b0c04ae3-a3bd-4721-9834-9cb6ae4f3526	FX4	shared_ref:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "", "status": "active", "comments": "", "who_pays": "", "visibility": null, "end_datetime": "2025-07-04T13:42:33.468909+00:00", "presigned_url": "https://daylily-dewey-0.s3.amazonaws.com/0/FI7010.png?AWSAccessKeyId=AKIAZYG7WWMZG2QOPT5Q&Signature=e4I9K6RNCvecG17%2FyyMLPK0uutc%3D&Expires=1751636553", "rclone_config": {}, "intended_users": [""], "reference_type": "presigned", "start_datetime": "2025-07-03T13:42:33.468909+00:00", "valid_duration": 86400.0}, "description": "Generic File Reference", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"form_label": "Reference Type", "property_key": "reference_type"}, {"form_label": "Visibility", "property_key": "visibility"}, {"form_label": "Reference Status", "property_key": "status"}, {"form_label": "Duration", "property_key": "valid_duration"}, {"form_label": "Start Datetime", "property_key": "start_datetime"}, {"form_label": "End Datetime", "property_key": "end_datetime"}, {"form_label": "Who Pays", "property_key": "who_pays"}], "controlled_properties": {}, "instantiation_layouts": [], "arbitrary_association_strings": []}	shared_ref	file_instance	file	generic	2025-07-03 06:42:33.469089-07	1.0	ready	f	64646531-f2fd-463a-9790-a4587d83f2a5	2025-07-03 06:42:33.469089-07	f
9728d18d-696b-47f7-a548-33e93b36b0f0	FX5	shared_ref:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "", "status": "active", "comments": "", "who_pays": "", "visibility": null, "end_datetime": "2025-07-04T13:42:33.521039+00:00", "presigned_url": "https://daylily-dewey-0.s3.amazonaws.com/0/FI7011.tsv?AWSAccessKeyId=AKIAZYG7WWMZG2QOPT5Q&Signature=tq6WIRzolEpOZuBiZDdY4qGnNp0%3D&Expires=1751636553", "rclone_config": {}, "intended_users": [""], "reference_type": "presigned", "start_datetime": "2025-07-03T13:42:33.521039+00:00", "valid_duration": 86400.0}, "description": "Generic File Reference", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"form_label": "Reference Type", "property_key": "reference_type"}, {"form_label": "Visibility", "property_key": "visibility"}, {"form_label": "Reference Status", "property_key": "status"}, {"form_label": "Duration", "property_key": "valid_duration"}, {"form_label": "Start Datetime", "property_key": "start_datetime"}, {"form_label": "End Datetime", "property_key": "end_datetime"}, {"form_label": "Who Pays", "property_key": "who_pays"}], "controlled_properties": {}, "instantiation_layouts": [], "arbitrary_association_strings": []}	shared_ref	file_instance	file	generic	2025-07-03 06:42:33.521204-07	1.0	ready	f	64646531-f2fd-463a-9790-a4587d83f2a5	2025-07-03 06:42:33.521204-07	f
0d2b2a11-5fa7-407f-9053-c2933c838bdf	FS3	file_set:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T130538", "tags": [], "comments": "", "duration": 1.0, "lab_code": "", "ref_type": "presigned_url", "description": "Bulk upload", "end_operator": {"email": "john@daylilyinformatics.com", "dag_fnv2": ""}, "creating_user": "john@daylilyinformatics.com", "end_timestamp": "2025-07-03 09:06:37 EDT-0400", "rclone_config": {"host": "0.0.0.0", "port": 8080, "user": "user", "bucket": "daylily-dewey-0", "passwd": "passwd"}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [{"email": "john@daylilyinformatics.com", "dag_fnv2": ""}], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "0", "max_executions": "-1", "action_executed": "1", "executed_datetime": ["2025-07-03 09:06:37 EDT-0400"], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	file_set	file_instance	file	generic	2025-07-03 06:05:38.782513-07	1.0	complete	f	1c57e0bf-8b44-426b-8f9e-e96aff46a341	2025-07-03 06:42:33.38115-07	f
ce4a70c5-bcb6-4305-9736-b14390e6c4a2	FS7	file_set:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T134315", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	file_set	file_instance	file	generic	2025-07-03 06:43:21.789558-07	1.0	ready	f	1c57e0bf-8b44-426b-8f9e-e96aff46a341	2025-07-03 06:43:21.789558-07	f
0098cfb3-be50-4e6d-8633-fef7cc11846c	FI7019	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "Rare_Cancer_Research_Foundation_eval.lic", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "lic", "patient_id": "", "remote_url": "https://daylily-service-omics-analysis-us-west-2.s3.us-west-2.amazonaws.com/data/cached_envs/Rare_Cancer_Research_Foundation_eval.lic?response-content-disposition=inline&X-Amz-Content-Sha256=UNSIGNED-PAYLOAD&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEA4aCXVzLXdlc3QtMiJIMEYCIQCZXoyLH34gj8x7%2B3rf73TQmLO4DbXXuM8tknhpdj%2Bw3AIhAIrgR1QwhzUkSqhzS%2FlzNlkHH2LR0BeExhlYOOlC0E2aKrkDCBcQABoMNjcwNDg0MDUwNzM4IgwjfPjz6GvCEvqnDjEqlgPZc0VsvDt6%2BcJJMHJ23ouM15EJ1mTLlwiLz4J4nNIbTxJOOFRTu6qBJWmw2YGwoBZM4v5iMyBSMI4L4NYQy4RTEsCljp8S6EOzBhjfU604eTRuPT3qyzgyrD1PkBLCQl2KdRLOdz%2BFdWnDEvl9sqQoityKo%2BOJE7PTlmXcHTcQnPRh2Ns2b%2Br83xJrk%2BrAOBi4U4u9lNDFJztlr32cUCsbPE8GvnGZ%2FVcdJej%2B7E7LRoE82bzm3QGrkHBzls2OKOA1vqCQ59THyRi4Lyj7rRG%2B7z13moLLDUny1ivZSxjaNgAQUbE1lVEkpnUZEl7DLHlWlDPNVcW%2FPSd5kEc6VhYOWajuGwCX6HUYnixTu9VhbYsCk4jyTbAAkm2tBMq7znJeFv%2B4IZWG1Tqp5Tf%2B68YFX57h0u13rL7G1qSsWlU5s4k5fnmQHXnPm%2BNreE6mhBlgbrvlXUpm%2Bq1xL%2Bee3Jw6lx%2BCtjgude6iG%2FqiYGjWpNgsQH%2ByrvsJtPhfT7MthZQnBjbJnCmjY7viuMpF6btOSGsQ4RuUMIS6mMMGOt0Cit1Z8coQMzmtNe8V7viWPNQU1JTLgBAaIkHmCsqtFjS3VxpRsYTt%2FgM1GbCysKXYyOs5oVAfM9g%2F3BB2H5O824KPPEtU4NBybVXIvdAP%2B02ZVtn5aAaIReO%2BuJVhG0rhXp1ZWsbc6WP0rUhLKhZ3SN%2FnAGEcsTqH9H%2F69wQYhC95gFix%2BapN2H7Hc%2BQZjD2mHG%2FWa6de4I4b7S32pAFsshBg%2Bpaa9KnAZ3MydygxMU5MwO4kXsEWvaUNYnbLNlniQ9a1CT02KtRd6AkmI9ReLjANDsL665JzJkL5gI5YE%2Fr9menC%2FfL4Qegz1a7UNN3x22VhQxlycDVe3iJC%2Bz56ASG4Y%2FNp9P3hyFnZeQ9hvxnSmZEsU%2BHf%2BpUsBes0D8p2ltqLyQWOH02RfkA2ry73VjeYZaFaTHRVUQCAJipykllOXUbBTculx%2BuiP9tjDfTxnDaI5x4JMueXdAayZg%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAZYG7WWMZFG4S7ANK%2F20250703%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20250703T134309Z&X-Amz-Expires=18000&X-Amz-SignedHeaders=host&X-Amz-Signature=8a0bd2fd61a1c4f4c7519ba7ddcae103f73614e41c27e88c6615e23cbed7d3c0", "clinician_id": "", "external_ids": [], "original_url": "https://daylily-service-omics-analysis-us-west-2.s3.us-west-2.amazonaws.com/data/cached_envs/Rare_Cancer_Research_Foundation_eval.lic?response-content-disposition=inline&X-Amz-Content-Sha256=UNSIGNED-PAYLOAD&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEA4aCXVzLXdlc3QtMiJIMEYCIQCZXoyLH34gj8x7%2B3rf73TQmLO4DbXXuM8tknhpdj%2Bw3AIhAIrgR1QwhzUkSqhzS%2FlzNlkHH2LR0BeExhlYOOlC0E2aKrkDCBcQABoMNjcwNDg0MDUwNzM4IgwjfPjz6GvCEvqnDjEqlgPZc0VsvDt6%2BcJJMHJ23ouM15EJ1mTLlwiLz4J4nNIbTxJOOFRTu6qBJWmw2YGwoBZM4v5iMyBSMI4L4NYQy4RTEsCljp8S6EOzBhjfU604eTRuPT3qyzgyrD1PkBLCQl2KdRLOdz%2BFdWnDEvl9sqQoityKo%2BOJE7PTlmXcHTcQnPRh2Ns2b%2Br83xJrk%2BrAOBi4U4u9lNDFJztlr32cUCsbPE8GvnGZ%2FVcdJej%2B7E7LRoE82bzm3QGrkHBzls2OKOA1vqCQ59THyRi4Lyj7rRG%2B7z13moLLDUny1ivZSxjaNgAQUbE1lVEkpnUZEl7DLHlWlDPNVcW%2FPSd5kEc6VhYOWajuGwCX6HUYnixTu9VhbYsCk4jyTbAAkm2tBMq7znJeFv%2B4IZWG1Tqp5Tf%2B68YFX57h0u13rL7G1qSsWlU5s4k5fnmQHXnPm%2BNreE6mhBlgbrvlXUpm%2Bq1xL%2Bee3Jw6lx%2BCtjgude6iG%2FqiYGjWpNgsQH%2ByrvsJtPhfT7MthZQnBjbJnCmjY7viuMpF6btOSGsQ4RuUMIS6mMMGOt0Cit1Z8coQMzmtNe8V7viWPNQU1JTLgBAaIkHmCsqtFjS3VxpRsYTt%2FgM1GbCysKXYyOs5oVAfM9g%2F3BB2H5O824KPPEtU4NBybVXIvdAP%2B02ZVtn5aAaIReO%2BuJVhG0rhXp1ZWsbc6WP0rUhLKhZ3SN%2FnAGEcsTqH9H%2F69wQYhC95gFix%2BapN2H7Hc%2BQZjD2mHG%2FWa6de4I4b7S32pAFsshBg%2Bpaa9KnAZ3MydygxMU5MwO4kXsEWvaUNYnbLNlniQ9a1CT02KtRd6AkmI9ReLjANDsL665JzJkL5gI5YE%2Fr9menC%2FfL4Qegz1a7UNN3x22VhQxlycDVe3iJC%2Bz56ASG4Y%2FNp9P3hyFnZeQ9hvxnSmZEsU%2BHf%2BpUsBes0D8p2ltqLyQWOH02RfkA2ry73VjeYZaFaTHRVUQCAJipykllOXUbBTculx%2BuiP9tjDfTxnDaI5x4JMueXdAayZg%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAZYG7WWMZFG4S7ANK%2F20250703%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20250703T134309Z&X-Amz-Expires=18000&X-Amz-SignedHeaders=host&X-Amz-Signature=8a0bd2fd61a1c4f4c7519ba7ddcae103f73614e41c27e88c6615e23cbed7d3c0", "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "https://daylily-service-omics-analysis-us-west-2.s3.us-west-2.amazonaws.com/data/cached_envs/Rare_Cancer_Research_Foundation_eval.lic?response-content-disposition=inline&X-Amz-Content-Sha256=UNSIGNED-PAYLOAD&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEA4aCXVzLXdlc3QtMiJIMEYCIQCZXoyLH34gj8x7%2B3rf73TQmLO4DbXXuM8tknhpdj%2Bw3AIhAIrgR1QwhzUkSqhzS%2FlzNlkHH2LR0BeExhlYOOlC0E2aKrkDCBcQABoMNjcwNDg0MDUwNzM4IgwjfPjz6GvCEvqnDjEqlgPZc0VsvDt6%2BcJJMHJ23ouM15EJ1mTLlwiLz4J4nNIbTxJOOFRTu6qBJWmw2YGwoBZM4v5iMyBSMI4L4NYQy4RTEsCljp8S6EOzBhjfU604eTRuPT3qyzgyrD1PkBLCQl2KdRLOdz%2BFdWnDEvl9sqQoityKo%2BOJE7PTlmXcHTcQnPRh2Ns2b%2Br83xJrk%2BrAOBi4U4u9lNDFJztlr32cUCsbPE8GvnGZ%2FVcdJej%2B7E7LRoE82bzm3QGrkHBzls2OKOA1vqCQ59THyRi4Lyj7rRG%2B7z13moLLDUny1ivZSxjaNgAQUbE1lVEkpnUZEl7DLHlWlDPNVcW%2FPSd5kEc6VhYOWajuGwCX6HUYnixTu9VhbYsCk4jyTbAAkm2tBMq7znJeFv%2B4IZWG1Tqp5Tf%2B68YFX57h0u13rL7G1qSsWlU5s4k5fnmQHXnPm%2BNreE6mhBlgbrvlXUpm%2Bq1xL%2Bee3Jw6lx%2BCtjgude6iG%2FqiYGjWpNgsQH%2ByrvsJtPhfT7MthZQnBjbJnCmjY7viuMpF6btOSGsQ4RuUMIS6mMMGOt0Cit1Z8coQMzmtNe8V7viWPNQU1JTLgBAaIkHmCsqtFjS3VxpRsYTt%2FgM1GbCysKXYyOs5oVAfM9g%2F3BB2H5O824KPPEtU4NBybVXIvdAP%2B02ZVtn5aAaIReO%2BuJVhG0rhXp1ZWsbc6WP0rUhLKhZ3SN%2FnAGEcsTqH9H%2F69wQYhC95gFix%2BapN2H7Hc%2BQZjD2mHG%2FWa6de4I4b7S32pAFsshBg%2Bpaa9KnAZ3MydygxMU5MwO4kXsEWvaUNYnbLNlniQ9a1CT02KtRd6AkmI9ReLjANDsL665JzJkL5gI5YE%2Fr9menC%2FfL4Qegz1a7UNN3x22VhQxlycDVe3iJC%2Bz56ASG4Y%2FNp9P3hyFnZeQ9hvxnSmZEsU%2BHf%2BpUsBes0D8p2ltqLyQWOH02RfkA2ry73VjeYZaFaTHRVUQCAJipykllOXUbBTculx%2BuiP9tjDfTxnDaI5x4JMueXdAayZg%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAZYG7WWMZFG4S7ANK%2F20250703%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20250703T134309Z&X-Amz-Expires=18000&X-Amz-SignedHeaders=host&X-Amz-Signature=8a0bd2fd61a1c4f4c7519ba7ddcae103f73614e41c27e88c6615e23cbed7d3c0", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "remote", "upload_group_key": "20250703T134315", "original_file_md5": "", "original_file_name": "Rare_Cancer_Research_Foundation_eval.lic", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "lic", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 06:43:21.841829-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 06:43:22.101299-07	f
b444546e-7e18-4775-a0a5-6a48aaebd985	FS8	file_set:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T134326", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	file_set	file_instance	file	generic	2025-07-03 06:43:34.904833-07	1.0	ready	f	1c57e0bf-8b44-426b-8f9e-e96aff46a341	2025-07-03 06:43:34.904833-07	f
4f291506-a724-4288-b37b-8a748c6eb737	FI7020	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "Rare_Cancer_Research_Foundation_eval.lic", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "lic", "patient_id": "", "clinician_id": "", "external_ids": [], "original_url": "https://daylily-service-omics-analysis-us-west-2.s3.us-west-2.amazonaws.com/data/cached_envs/Rare_Cancer_Research_Foundation_eval.lic?response-content-disposition=inline&X-Amz-Content-Sha256=UNSIGNED-PAYLOAD&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEA4aCXVzLXdlc3QtMiJIMEYCIQCZXoyLH34gj8x7%2B3rf73TQmLO4DbXXuM8tknhpdj%2Bw3AIhAIrgR1QwhzUkSqhzS%2FlzNlkHH2LR0BeExhlYOOlC0E2aKrkDCBcQABoMNjcwNDg0MDUwNzM4IgwjfPjz6GvCEvqnDjEqlgPZc0VsvDt6%2BcJJMHJ23ouM15EJ1mTLlwiLz4J4nNIbTxJOOFRTu6qBJWmw2YGwoBZM4v5iMyBSMI4L4NYQy4RTEsCljp8S6EOzBhjfU604eTRuPT3qyzgyrD1PkBLCQl2KdRLOdz%2BFdWnDEvl9sqQoityKo%2BOJE7PTlmXcHTcQnPRh2Ns2b%2Br83xJrk%2BrAOBi4U4u9lNDFJztlr32cUCsbPE8GvnGZ%2FVcdJej%2B7E7LRoE82bzm3QGrkHBzls2OKOA1vqCQ59THyRi4Lyj7rRG%2B7z13moLLDUny1ivZSxjaNgAQUbE1lVEkpnUZEl7DLHlWlDPNVcW%2FPSd5kEc6VhYOWajuGwCX6HUYnixTu9VhbYsCk4jyTbAAkm2tBMq7znJeFv%2B4IZWG1Tqp5Tf%2B68YFX57h0u13rL7G1qSsWlU5s4k5fnmQHXnPm%2BNreE6mhBlgbrvlXUpm%2Bq1xL%2Bee3Jw6lx%2BCtjgude6iG%2FqiYGjWpNgsQH%2ByrvsJtPhfT7MthZQnBjbJnCmjY7viuMpF6btOSGsQ4RuUMIS6mMMGOt0Cit1Z8coQMzmtNe8V7viWPNQU1JTLgBAaIkHmCsqtFjS3VxpRsYTt%2FgM1GbCysKXYyOs5oVAfM9g%2F3BB2H5O824KPPEtU4NBybVXIvdAP%2B02ZVtn5aAaIReO%2BuJVhG0rhXp1ZWsbc6WP0rUhLKhZ3SN%2FnAGEcsTqH9H%2F69wQYhC95gFix%2BapN2H7Hc%2BQZjD2mHG%2FWa6de4I4b7S32pAFsshBg%2Bpaa9KnAZ3MydygxMU5MwO4kXsEWvaUNYnbLNlniQ9a1CT02KtRd6AkmI9ReLjANDsL665JzJkL5gI5YE%2Fr9menC%2FfL4Qegz1a7UNN3x22VhQxlycDVe3iJC%2Bz56ASG4Y%2FNp9P3hyFnZeQ9hvxnSmZEsU%2BHf%2BpUsBes0D8p2ltqLyQWOH02RfkA2ry73VjeYZaFaTHRVUQCAJipykllOXUbBTculx%2BuiP9tjDfTxnDaI5x4JMueXdAayZg%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAZYG7WWMZFG4S7ANK%2F20250703%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20250703T134309Z&X-Amz-Expires=18000&X-Amz-SignedHeaders=host&X-Amz-Signature=8a0bd2fd61a1c4f4c7519ba7ddcae103f73614e41c27e88c6615e23cbed7d3c0", "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7020.lic", "current_s3_uri": "s3://daylily-dewey-0/0/FI7020.lic", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134326", "original_file_md5": "", "original_file_name": "Rare_Cancer_Research_Foundation_eval.lic", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "lic", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 4732, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 06:43:34.954475-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 06:43:35.204833-07	f
2321b538-2fee-484b-8d18-7b6df27371ea	FS9	file_set:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T134442", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	file_set	file_instance	file	generic	2025-07-03 06:44:50.554495-07	1.0	ready	f	1c57e0bf-8b44-426b-8f9e-e96aff46a341	2025-07-03 06:44:50.554495-07	f
bee9e98b-3b9e-4fe1-8729-1b8fd38bac40	FI7021	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "benchmarks_summary.tsv.old", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "old", "patient_id": "", "clinician_id": "", "external_ids": [], "original_url": "https://daylily-service-omics-analysis-us-west-2.s3.us-west-2.amazonaws.com/FSxLustre20250123T214325Z/analysis_results/ubuntu/hg38/remaining/daylily/benchmarks_summary.tsv.old?response-content-disposition=inline&X-Amz-Content-Sha256=UNSIGNED-PAYLOAD&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEA4aCXVzLXdlc3QtMiJIMEYCIQCZXoyLH34gj8x7%2B3rf73TQmLO4DbXXuM8tknhpdj%2Bw3AIhAIrgR1QwhzUkSqhzS%2FlzNlkHH2LR0BeExhlYOOlC0E2aKrkDCBcQABoMNjcwNDg0MDUwNzM4IgwjfPjz6GvCEvqnDjEqlgPZc0VsvDt6%2BcJJMHJ23ouM15EJ1mTLlwiLz4J4nNIbTxJOOFRTu6qBJWmw2YGwoBZM4v5iMyBSMI4L4NYQy4RTEsCljp8S6EOzBhjfU604eTRuPT3qyzgyrD1PkBLCQl2KdRLOdz%2BFdWnDEvl9sqQoityKo%2BOJE7PTlmXcHTcQnPRh2Ns2b%2Br83xJrk%2BrAOBi4U4u9lNDFJztlr32cUCsbPE8GvnGZ%2FVcdJej%2B7E7LRoE82bzm3QGrkHBzls2OKOA1vqCQ59THyRi4Lyj7rRG%2B7z13moLLDUny1ivZSxjaNgAQUbE1lVEkpnUZEl7DLHlWlDPNVcW%2FPSd5kEc6VhYOWajuGwCX6HUYnixTu9VhbYsCk4jyTbAAkm2tBMq7znJeFv%2B4IZWG1Tqp5Tf%2B68YFX57h0u13rL7G1qSsWlU5s4k5fnmQHXnPm%2BNreE6mhBlgbrvlXUpm%2Bq1xL%2Bee3Jw6lx%2BCtjgude6iG%2FqiYGjWpNgsQH%2ByrvsJtPhfT7MthZQnBjbJnCmjY7viuMpF6btOSGsQ4RuUMIS6mMMGOt0Cit1Z8coQMzmtNe8V7viWPNQU1JTLgBAaIkHmCsqtFjS3VxpRsYTt%2FgM1GbCysKXYyOs5oVAfM9g%2F3BB2H5O824KPPEtU4NBybVXIvdAP%2B02ZVtn5aAaIReO%2BuJVhG0rhXp1ZWsbc6WP0rUhLKhZ3SN%2FnAGEcsTqH9H%2F69wQYhC95gFix%2BapN2H7Hc%2BQZjD2mHG%2FWa6de4I4b7S32pAFsshBg%2Bpaa9KnAZ3MydygxMU5MwO4kXsEWvaUNYnbLNlniQ9a1CT02KtRd6AkmI9ReLjANDsL665JzJkL5gI5YE%2Fr9menC%2FfL4Qegz1a7UNN3x22VhQxlycDVe3iJC%2Bz56ASG4Y%2FNp9P3hyFnZeQ9hvxnSmZEsU%2BHf%2BpUsBes0D8p2ltqLyQWOH02RfkA2ry73VjeYZaFaTHRVUQCAJipykllOXUbBTculx%2BuiP9tjDfTxnDaI5x4JMueXdAayZg%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAZYG7WWMZFG4S7ANK%2F20250703%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20250703T134434Z&X-Amz-Expires=18000&X-Amz-SignedHeaders=host&X-Amz-Signature=ae73c785dbdc400295532cbf30b18e254b86e4ae4d015a2c61fb2712337fb1b7", "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7021.old", "current_s3_uri": "s3://daylily-dewey-0/0/FI7021.old", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134442", "original_file_md5": "", "original_file_name": "benchmarks_summary.tsv.old", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "old", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 179, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [{"email": "john@daylilyinformatics.com", "dag_fnv2": ""}], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "1", "executed_datetime": ["2025-07-03 09:45:06 EDT-0400"], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 06:44:50.612932-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 06:45:06.36426-07	f
88d739e7-9eb0-4083-af3b-265a8d380792	FS10	file_set:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T134607", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	file_set	file_instance	file	generic	2025-07-03 06:46:19.410805-07	1.0	ready	f	1c57e0bf-8b44-426b-8f9e-e96aff46a341	2025-07-03 06:46:19.410805-07	f
d512ba32-ddef-4c26-acae-5dda9273bab1	FI7022	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "benchmarks_summary.tsv", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "tsv", "patient_id": "", "clinician_id": "", "external_ids": [], "original_url": "https://daylily-service-omics-analysis-us-west-2.s3.us-west-2.amazonaws.com/FSxLustre20250123T214325Z/analysis_results/ubuntu/hg38/remaining/daylily/benchmarks_summary.tsv?response-content-disposition=inline&X-Amz-Content-Sha256=UNSIGNED-PAYLOAD&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEA4aCXVzLXdlc3QtMiJIMEYCIQCZXoyLH34gj8x7%2B3rf73TQmLO4DbXXuM8tknhpdj%2Bw3AIhAIrgR1QwhzUkSqhzS%2FlzNlkHH2LR0BeExhlYOOlC0E2aKrkDCBcQABoMNjcwNDg0MDUwNzM4IgwjfPjz6GvCEvqnDjEqlgPZc0VsvDt6%2BcJJMHJ23ouM15EJ1mTLlwiLz4J4nNIbTxJOOFRTu6qBJWmw2YGwoBZM4v5iMyBSMI4L4NYQy4RTEsCljp8S6EOzBhjfU604eTRuPT3qyzgyrD1PkBLCQl2KdRLOdz%2BFdWnDEvl9sqQoityKo%2BOJE7PTlmXcHTcQnPRh2Ns2b%2Br83xJrk%2BrAOBi4U4u9lNDFJztlr32cUCsbPE8GvnGZ%2FVcdJej%2B7E7LRoE82bzm3QGrkHBzls2OKOA1vqCQ59THyRi4Lyj7rRG%2B7z13moLLDUny1ivZSxjaNgAQUbE1lVEkpnUZEl7DLHlWlDPNVcW%2FPSd5kEc6VhYOWajuGwCX6HUYnixTu9VhbYsCk4jyTbAAkm2tBMq7znJeFv%2B4IZWG1Tqp5Tf%2B68YFX57h0u13rL7G1qSsWlU5s4k5fnmQHXnPm%2BNreE6mhBlgbrvlXUpm%2Bq1xL%2Bee3Jw6lx%2BCtjgude6iG%2FqiYGjWpNgsQH%2ByrvsJtPhfT7MthZQnBjbJnCmjY7viuMpF6btOSGsQ4RuUMIS6mMMGOt0Cit1Z8coQMzmtNe8V7viWPNQU1JTLgBAaIkHmCsqtFjS3VxpRsYTt%2FgM1GbCysKXYyOs5oVAfM9g%2F3BB2H5O824KPPEtU4NBybVXIvdAP%2B02ZVtn5aAaIReO%2BuJVhG0rhXp1ZWsbc6WP0rUhLKhZ3SN%2FnAGEcsTqH9H%2F69wQYhC95gFix%2BapN2H7Hc%2BQZjD2mHG%2FWa6de4I4b7S32pAFsshBg%2Bpaa9KnAZ3MydygxMU5MwO4kXsEWvaUNYnbLNlniQ9a1CT02KtRd6AkmI9ReLjANDsL665JzJkL5gI5YE%2Fr9menC%2FfL4Qegz1a7UNN3x22VhQxlycDVe3iJC%2Bz56ASG4Y%2FNp9P3hyFnZeQ9hvxnSmZEsU%2BHf%2BpUsBes0D8p2ltqLyQWOH02RfkA2ry73VjeYZaFaTHRVUQCAJipykllOXUbBTculx%2BuiP9tjDfTxnDaI5x4JMueXdAayZg%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAZYG7WWMZFG4S7ANK%2F20250703%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20250703T134602Z&X-Amz-Expires=18000&X-Amz-SignedHeaders=host&X-Amz-Signature=e93a920e5c7fba5ed1f0b926fde5ba9498ec24e4ac83995dd935dad70a49f04e", "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7022.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7022.tsv", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134607", "original_file_md5": "", "original_file_name": "benchmarks_summary.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 742279, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 06:46:19.522793-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 06:46:19.801673-07	f
7e52fd3c-eedc-4e34-8caf-de2f5ad475fa	FS11	file_set:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T142630", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	file_set	file_instance	file	generic	2025-07-03 07:26:30.457615-07	1.0	ready	f	1c57e0bf-8b44-426b-8f9e-e96aff46a341	2025-07-03 07:26:30.457615-07	f
f835a3a0-a807-41d2-9f29-68d149957d8d	FI7023	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7023.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7023.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142630", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:26:30.810766-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:26:31.242334-07	f
6ad2a1ab-3b91-4c5b-b4e6-69d087c868f0	FI7024	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7024.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7024.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142630", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:26:34.225501-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:26:34.615521-07	f
7b446c7d-69b4-45f5-9320-dff366d18cce	FI7025	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7025.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7025.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142630", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:26:35.675084-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:26:35.966492-07	f
11568060-9f81-417e-baf9-9e7e56b3a759	FI7026	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7026.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7026.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142630", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:26:36.917375-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:26:37.251557-07	f
1aab709b-8c59-4117-9bcc-48c2bf549a79	FS12	file_set:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T142720", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	file_set	file_instance	file	generic	2025-07-03 07:27:20.521229-07	1.0	ready	f	1c57e0bf-8b44-426b-8f9e-e96aff46a341	2025-07-03 07:27:20.521229-07	f
0886f894-a724-4452-8007-54b3598c84d6	FI7027	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7027.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7027.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142720", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:27:20.555143-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:27:20.854339-07	f
051fd024-70cd-4874-ba13-088a4becc209	FI7028	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7028.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7028.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142720", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:27:21.787463-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:27:22.104097-07	f
96c5276b-4501-460e-80a1-c6506a98e020	FI7031	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7031.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7031.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T143038", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:30:39.072113-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:30:39.333909-07	f
5608248c-838c-458c-bbba-a6a0ba464775	FI7029	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7029.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7029.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142720", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:27:23.20872-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:27:23.50468-07	f
ec92091a-246e-472e-914a-66e628bd748b	FI7030	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7030.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7030.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T142720", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:27:24.475065-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:27:24.798792-07	f
c64b5702-5aea-4bf0-856b-ad9d6a0035c4	FI7032	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7032.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7032.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T143038", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:30:40.072812-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:30:40.362347-07	f
44188f89-cb29-4da6-90c3-592a3e96ce82	FI7033	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7033.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7033.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T143038", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:30:41.012048-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:30:41.242377-07	f
fd60bad1-ce79-4e33-a561-8db568b1514c	FI7034	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7034.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7034.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T143038", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:30:41.961138-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:30:42.258511-07	f
99921700-5afa-47c0-aa3b-f6a520dc526b	FS14	file_set:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T143216", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	file_set	file_instance	file	generic	2025-07-03 07:32:27.331864-07	1.0	ready	f	1c57e0bf-8b44-426b-8f9e-e96aff46a341	2025-07-03 07:32:27.331864-07	f
ef3752a0-fb6c-4294-ac4e-e0f65976b68a	FI7035	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "backup_2025-07-03_00-06-33.sql", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "sql", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7035.sql", "current_s3_uri": "s3://daylily-dewey-0/0/FI7035.sql", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143216", "original_file_md5": "", "original_file_name": "backup_2025-07-03_00-06-33.sql", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "sql", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1012876, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:32:27.440193-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:32:27.785525-07	f
def0e204-1561-4031-a477-de7549381c84	FS15	file_set:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "fastqs", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	file_set	file_instance	file	generic	2025-07-03 07:33:39.079499-07	1.0	ready	f	1c57e0bf-8b44-426b-8f9e-e96aff46a341	2025-07-03 07:33:39.079499-07	f
8c44a660-381a-4fd7-8bb3-6093c23991f0	FI7036	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "NA12878/HG001_GRCh38_1_22_v4.2.1_benchmark_hifiasm_v11_phasetransfer.PS_FIX.vcf.gz", "purpose": "Clinical", "category": "Diagnosis", "comments": "a comment", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "", "file_tags": "sometag", "file_type": "gz", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7036.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7036.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "fastqs", "original_file_md5": "", "original_file_name": "HG001_GRCh38_1_22_v4.2.1_benchmark_hifiasm_v11_phasetransfer.PS_FIX.vcf.gz", "original_file_path": "NA12878", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 111830815, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:33:39.200006-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:33:39.500051-07	f
76d325cb-204a-490a-817f-ba7917e86e20	FI7037	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "NA12878/Lee_washington_0250E_19948.pdf", "purpose": "Clinical", "category": "Diagnosis", "comments": "a comment", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "", "file_tags": "sometag", "file_type": "pdf", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7037.pdf", "current_s3_uri": "s3://daylily-dewey-0/0/FI7037.pdf", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "fastqs", "original_file_md5": "", "original_file_name": "Lee_washington_0250E_19948.pdf", "original_file_path": "NA12878", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "pdf", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 11203779, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:33:44.749755-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:33:44.943376-07	f
9ea2e910-4f31-4ed2-b6c4-449c753c3692	FS16	file_set:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T143413", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	file_set	file_instance	file	generic	2025-07-03 07:34:23.540595-07	1.0	ready	f	1c57e0bf-8b44-426b-8f9e-e96aff46a341	2025-07-03 07:34:23.540595-07	f
2158f0a3-f8e1-4bc9-84f7-4620f6fc8f28	FI7038	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "day_cmd.log", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "log", "patient_id": "", "remote_url": "https://daylily-service-omics-analysis-us-west-2.s3.us-west-2.amazonaws.com/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log?response-content-disposition=inline&X-Amz-Content-Sha256=UNSIGNED-PAYLOAD&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEA8aCXVzLXdlc3QtMiJHMEUCIEwkFC7typ%2B7k712by339aiZJx867GzZv1rwLjaMWjy9AiEAyFFAGSuNMLiCibMYT9IuQiYJjXD59ZexbYTehsL2hXoquQMIGBAAGgw2NzA0ODQwNTA3MzgiDOLLh4KTH%2Bja%2FjEo0CqWA32BzYyhHkRrNyY1g6iBHMD91fBPNWdcSCjPQ3hFJ9USdNJZJG9R7%2BuWbMT0EQgqOu%2FUTCKIX7JLD8tJJHb1bx90R7BfXWfXKbrg94tOrItTV11IAv5uFUeVa24iDH%2BLCoStSBxskYQJl7MosBPuA0BXnrGsSTi545yLxxZnoUfHNRqltwIQRNut3XxdoDTpLlkirFKMGkIeyY8wJLQZfGoXj9bcau%2BGPw3Ikb20%2FdA79erq9GcJaANTefB3JHU%2FK2c%2B8zZLVnSB1V3%2BtR%2FTWW6thVzY8IFeXGMd9jvKnt9bX4CciVWPyQeBvKVniiLYqwFAdBGbLYmTr72bCfP%2FxA6JAsHooKLkX6bNgGNYbbT38%2BNGS4PDV5tZ7Ud8H13vKDvwn7Bux6JRqbf9erjLK7UJ9KmsLFPGdCvGMWO0c3STI3sw1vJgRQ6DAwwM6kKq4ui3k3yBHMGfF%2FW8MSdLB93B4W3qeKTriP5f8UX8Kw2LmNzLQnat7mPAmCunY3NHOpwaKMiES5XVhXcHGlKFYTz9V6spE3cwhLqYwwY63gJO%2FcyKZeiz3yHTivM0FJNm2TMuIhLmUm9KHl%2BEDGDP9x4XfVnbWza1M%2Fhjhkp7OQVZK0WWMsRmgzSKN%2F%2B0AYGf%2ByswpK6fUKJbeFeDDIxOTh3d8TWJTXj3wDTKuQddBNqLVw6XUuOmVwFBB5%2FLpVvywLDb3iVcNqOmlFJcR2jBR53n0EpZbfIW9j28xVlI965f%2BhcXMhK0D6yvNody3PegC9lDNi7bRERsj72QGCC7Na3HNliMURha2TZoo24SHLa1YLcsyu5ojw%2FIN%2BggWTTMg%2FE%2BWsP9eeEIxNLDj5G7nXkCEdjbfuKPoWUdebP15z9Dknyl7W1sDN2hai5B%2Bk%2B6hTaZziZxr%2FI%2FU2EHgwWkj2ya5qArGlpfK%2BAhlH5%2BVA9I3P3pv7ckjp2J6f2TUErgI3maueoSiTlbQqpEeSVNGu%2BDODQx5799eMGf0f448b29I%2BXKUrDlH0Yoj1RmEA%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAZYG7WWMZCCDE37TC%2F20250703%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20250703T143406Z&X-Amz-Expires=21600&X-Amz-SignedHeaders=host&X-Amz-Signature=3d7e773fe4a25d0a6dd36aecd9176b13b1717033a69610cfb6815ed156a1b93c", "clinician_id": "", "external_ids": [], "original_url": "https://daylily-service-omics-analysis-us-west-2.s3.us-west-2.amazonaws.com/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log?response-content-disposition=inline&X-Amz-Content-Sha256=UNSIGNED-PAYLOAD&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEA8aCXVzLXdlc3QtMiJHMEUCIEwkFC7typ%2B7k712by339aiZJx867GzZv1rwLjaMWjy9AiEAyFFAGSuNMLiCibMYT9IuQiYJjXD59ZexbYTehsL2hXoquQMIGBAAGgw2NzA0ODQwNTA3MzgiDOLLh4KTH%2Bja%2FjEo0CqWA32BzYyhHkRrNyY1g6iBHMD91fBPNWdcSCjPQ3hFJ9USdNJZJG9R7%2BuWbMT0EQgqOu%2FUTCKIX7JLD8tJJHb1bx90R7BfXWfXKbrg94tOrItTV11IAv5uFUeVa24iDH%2BLCoStSBxskYQJl7MosBPuA0BXnrGsSTi545yLxxZnoUfHNRqltwIQRNut3XxdoDTpLlkirFKMGkIeyY8wJLQZfGoXj9bcau%2BGPw3Ikb20%2FdA79erq9GcJaANTefB3JHU%2FK2c%2B8zZLVnSB1V3%2BtR%2FTWW6thVzY8IFeXGMd9jvKnt9bX4CciVWPyQeBvKVniiLYqwFAdBGbLYmTr72bCfP%2FxA6JAsHooKLkX6bNgGNYbbT38%2BNGS4PDV5tZ7Ud8H13vKDvwn7Bux6JRqbf9erjLK7UJ9KmsLFPGdCvGMWO0c3STI3sw1vJgRQ6DAwwM6kKq4ui3k3yBHMGfF%2FW8MSdLB93B4W3qeKTriP5f8UX8Kw2LmNzLQnat7mPAmCunY3NHOpwaKMiES5XVhXcHGlKFYTz9V6spE3cwhLqYwwY63gJO%2FcyKZeiz3yHTivM0FJNm2TMuIhLmUm9KHl%2BEDGDP9x4XfVnbWza1M%2Fhjhkp7OQVZK0WWMsRmgzSKN%2F%2B0AYGf%2ByswpK6fUKJbeFeDDIxOTh3d8TWJTXj3wDTKuQddBNqLVw6XUuOmVwFBB5%2FLpVvywLDb3iVcNqOmlFJcR2jBR53n0EpZbfIW9j28xVlI965f%2BhcXMhK0D6yvNody3PegC9lDNi7bRERsj72QGCC7Na3HNliMURha2TZoo24SHLa1YLcsyu5ojw%2FIN%2BggWTTMg%2FE%2BWsP9eeEIxNLDj5G7nXkCEdjbfuKPoWUdebP15z9Dknyl7W1sDN2hai5B%2Bk%2B6hTaZziZxr%2FI%2FU2EHgwWkj2ya5qArGlpfK%2BAhlH5%2BVA9I3P3pv7ckjp2J6f2TUErgI3maueoSiTlbQqpEeSVNGu%2BDODQx5799eMGf0f448b29I%2BXKUrDlH0Yoj1RmEA%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAZYG7WWMZCCDE37TC%2F20250703%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20250703T143406Z&X-Amz-Expires=21600&X-Amz-SignedHeaders=host&X-Amz-Signature=3d7e773fe4a25d0a6dd36aecd9176b13b1717033a69610cfb6815ed156a1b93c", "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "https://daylily-service-omics-analysis-us-west-2.s3.us-west-2.amazonaws.com/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log?response-content-disposition=inline&X-Amz-Content-Sha256=UNSIGNED-PAYLOAD&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEA8aCXVzLXdlc3QtMiJHMEUCIEwkFC7typ%2B7k712by339aiZJx867GzZv1rwLjaMWjy9AiEAyFFAGSuNMLiCibMYT9IuQiYJjXD59ZexbYTehsL2hXoquQMIGBAAGgw2NzA0ODQwNTA3MzgiDOLLh4KTH%2Bja%2FjEo0CqWA32BzYyhHkRrNyY1g6iBHMD91fBPNWdcSCjPQ3hFJ9USdNJZJG9R7%2BuWbMT0EQgqOu%2FUTCKIX7JLD8tJJHb1bx90R7BfXWfXKbrg94tOrItTV11IAv5uFUeVa24iDH%2BLCoStSBxskYQJl7MosBPuA0BXnrGsSTi545yLxxZnoUfHNRqltwIQRNut3XxdoDTpLlkirFKMGkIeyY8wJLQZfGoXj9bcau%2BGPw3Ikb20%2FdA79erq9GcJaANTefB3JHU%2FK2c%2B8zZLVnSB1V3%2BtR%2FTWW6thVzY8IFeXGMd9jvKnt9bX4CciVWPyQeBvKVniiLYqwFAdBGbLYmTr72bCfP%2FxA6JAsHooKLkX6bNgGNYbbT38%2BNGS4PDV5tZ7Ud8H13vKDvwn7Bux6JRqbf9erjLK7UJ9KmsLFPGdCvGMWO0c3STI3sw1vJgRQ6DAwwM6kKq4ui3k3yBHMGfF%2FW8MSdLB93B4W3qeKTriP5f8UX8Kw2LmNzLQnat7mPAmCunY3NHOpwaKMiES5XVhXcHGlKFYTz9V6spE3cwhLqYwwY63gJO%2FcyKZeiz3yHTivM0FJNm2TMuIhLmUm9KHl%2BEDGDP9x4XfVnbWza1M%2Fhjhkp7OQVZK0WWMsRmgzSKN%2F%2B0AYGf%2ByswpK6fUKJbeFeDDIxOTh3d8TWJTXj3wDTKuQddBNqLVw6XUuOmVwFBB5%2FLpVvywLDb3iVcNqOmlFJcR2jBR53n0EpZbfIW9j28xVlI965f%2BhcXMhK0D6yvNody3PegC9lDNi7bRERsj72QGCC7Na3HNliMURha2TZoo24SHLa1YLcsyu5ojw%2FIN%2BggWTTMg%2FE%2BWsP9eeEIxNLDj5G7nXkCEdjbfuKPoWUdebP15z9Dknyl7W1sDN2hai5B%2Bk%2B6hTaZziZxr%2FI%2FU2EHgwWkj2ya5qArGlpfK%2BAhlH5%2BVA9I3P3pv7ckjp2J6f2TUErgI3maueoSiTlbQqpEeSVNGu%2BDODQx5799eMGf0f448b29I%2BXKUrDlH0Yoj1RmEA%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAZYG7WWMZCCDE37TC%2F20250703%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20250703T143406Z&X-Amz-Expires=21600&X-Amz-SignedHeaders=host&X-Amz-Signature=3d7e773fe4a25d0a6dd36aecd9176b13b1717033a69610cfb6815ed156a1b93c", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "remote", "upload_group_key": "20250703T143413", "original_file_md5": "", "original_file_name": "day_cmd.log", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "log", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:34:23.593818-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:34:23.863877-07	f
cf8a5e09-2051-44be-9121-e9a446b85e5d	FI7039	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "day_cmd.log", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "log", "patient_id": "", "remote_url": "https://daylily-service-omics-analysis-us-west-2.s3.us-west-2.amazonaws.com/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log?response-content-disposition=inline&X-Amz-Content-Sha256=UNSIGNED-PAYLOAD&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEA8aCXVzLXdlc3QtMiJHMEUCIEwkFC7typ%2B7k712by339aiZJx867GzZv1rwLjaMWjy9AiEAyFFAGSuNMLiCibMYT9IuQiYJjXD59ZexbYTehsL2hXoquQMIGBAAGgw2NzA0ODQwNTA3MzgiDOLLh4KTH%2Bja%2FjEo0CqWA32BzYyhHkRrNyY1g6iBHMD91fBPNWdcSCjPQ3hFJ9USdNJZJG9R7%2BuWbMT0EQgqOu%2FUTCKIX7JLD8tJJHb1bx90R7BfXWfXKbrg94tOrItTV11IAv5uFUeVa24iDH%2BLCoStSBxskYQJl7MosBPuA0BXnrGsSTi545yLxxZnoUfHNRqltwIQRNut3XxdoDTpLlkirFKMGkIeyY8wJLQZfGoXj9bcau%2BGPw3Ikb20%2FdA79erq9GcJaANTefB3JHU%2FK2c%2B8zZLVnSB1V3%2BtR%2FTWW6thVzY8IFeXGMd9jvKnt9bX4CciVWPyQeBvKVniiLYqwFAdBGbLYmTr72bCfP%2FxA6JAsHooKLkX6bNgGNYbbT38%2BNGS4PDV5tZ7Ud8H13vKDvwn7Bux6JRqbf9erjLK7UJ9KmsLFPGdCvGMWO0c3STI3sw1vJgRQ6DAwwM6kKq4ui3k3yBHMGfF%2FW8MSdLB93B4W3qeKTriP5f8UX8Kw2LmNzLQnat7mPAmCunY3NHOpwaKMiES5XVhXcHGlKFYTz9V6spE3cwhLqYwwY63gJO%2FcyKZeiz3yHTivM0FJNm2TMuIhLmUm9KHl%2BEDGDP9x4XfVnbWza1M%2Fhjhkp7OQVZK0WWMsRmgzSKN%2F%2B0AYGf%2ByswpK6fUKJbeFeDDIxOTh3d8TWJTXj3wDTKuQddBNqLVw6XUuOmVwFBB5%2FLpVvywLDb3iVcNqOmlFJcR2jBR53n0EpZbfIW9j28xVlI965f%2BhcXMhK0D6yvNody3PegC9lDNi7bRERsj72QGCC7Na3HNliMURha2TZoo24SHLa1YLcsyu5ojw%2FIN%2BggWTTMg%2FE%2BWsP9eeEIxNLDj5G7nXkCEdjbfuKPoWUdebP15z9Dknyl7W1sDN2hai5B%2Bk%2B6hTaZziZxr%2FI%2FU2EHgwWkj2ya5qArGlpfK%2BAhlH5%2BVA9I3P3pv7ckjp2J6f2TUErgI3maueoSiTlbQqpEeSVNGu%2BDODQx5799eMGf0f448b29I%2BXKUrDlH0Yoj1RmEA%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAZYG7WWMZCCDE37TC%2F20250703%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20250703T143406Z&X-Amz-Expires=21600&X-Amz-SignedHeaders=host&X-Amz-Signature=3d7e773fe4a25d0a6dd36aecd9176b13b1717033a69610cfb6815ed156a1b93c", "clinician_id": "", "external_ids": [], "original_url": "https://daylily-service-omics-analysis-us-west-2.s3.us-west-2.amazonaws.com/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log?response-content-disposition=inline&X-Amz-Content-Sha256=UNSIGNED-PAYLOAD&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEA8aCXVzLXdlc3QtMiJHMEUCIEwkFC7typ%2B7k712by339aiZJx867GzZv1rwLjaMWjy9AiEAyFFAGSuNMLiCibMYT9IuQiYJjXD59ZexbYTehsL2hXoquQMIGBAAGgw2NzA0ODQwNTA3MzgiDOLLh4KTH%2Bja%2FjEo0CqWA32BzYyhHkRrNyY1g6iBHMD91fBPNWdcSCjPQ3hFJ9USdNJZJG9R7%2BuWbMT0EQgqOu%2FUTCKIX7JLD8tJJHb1bx90R7BfXWfXKbrg94tOrItTV11IAv5uFUeVa24iDH%2BLCoStSBxskYQJl7MosBPuA0BXnrGsSTi545yLxxZnoUfHNRqltwIQRNut3XxdoDTpLlkirFKMGkIeyY8wJLQZfGoXj9bcau%2BGPw3Ikb20%2FdA79erq9GcJaANTefB3JHU%2FK2c%2B8zZLVnSB1V3%2BtR%2FTWW6thVzY8IFeXGMd9jvKnt9bX4CciVWPyQeBvKVniiLYqwFAdBGbLYmTr72bCfP%2FxA6JAsHooKLkX6bNgGNYbbT38%2BNGS4PDV5tZ7Ud8H13vKDvwn7Bux6JRqbf9erjLK7UJ9KmsLFPGdCvGMWO0c3STI3sw1vJgRQ6DAwwM6kKq4ui3k3yBHMGfF%2FW8MSdLB93B4W3qeKTriP5f8UX8Kw2LmNzLQnat7mPAmCunY3NHOpwaKMiES5XVhXcHGlKFYTz9V6spE3cwhLqYwwY63gJO%2FcyKZeiz3yHTivM0FJNm2TMuIhLmUm9KHl%2BEDGDP9x4XfVnbWza1M%2Fhjhkp7OQVZK0WWMsRmgzSKN%2F%2B0AYGf%2ByswpK6fUKJbeFeDDIxOTh3d8TWJTXj3wDTKuQddBNqLVw6XUuOmVwFBB5%2FLpVvywLDb3iVcNqOmlFJcR2jBR53n0EpZbfIW9j28xVlI965f%2BhcXMhK0D6yvNody3PegC9lDNi7bRERsj72QGCC7Na3HNliMURha2TZoo24SHLa1YLcsyu5ojw%2FIN%2BggWTTMg%2FE%2BWsP9eeEIxNLDj5G7nXkCEdjbfuKPoWUdebP15z9Dknyl7W1sDN2hai5B%2Bk%2B6hTaZziZxr%2FI%2FU2EHgwWkj2ya5qArGlpfK%2BAhlH5%2BVA9I3P3pv7ckjp2J6f2TUErgI3maueoSiTlbQqpEeSVNGu%2BDODQx5799eMGf0f448b29I%2BXKUrDlH0Yoj1RmEA%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAZYG7WWMZCCDE37TC%2F20250703%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20250703T143406Z&X-Amz-Expires=21600&X-Amz-SignedHeaders=host&X-Amz-Signature=3d7e773fe4a25d0a6dd36aecd9176b13b1717033a69610cfb6815ed156a1b93c", "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "", "current_s3_uri": "https://daylily-service-omics-analysis-us-west-2.s3.us-west-2.amazonaws.com/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log?response-content-disposition=inline&X-Amz-Content-Sha256=UNSIGNED-PAYLOAD&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEA8aCXVzLXdlc3QtMiJHMEUCIEwkFC7typ%2B7k712by339aiZJx867GzZv1rwLjaMWjy9AiEAyFFAGSuNMLiCibMYT9IuQiYJjXD59ZexbYTehsL2hXoquQMIGBAAGgw2NzA0ODQwNTA3MzgiDOLLh4KTH%2Bja%2FjEo0CqWA32BzYyhHkRrNyY1g6iBHMD91fBPNWdcSCjPQ3hFJ9USdNJZJG9R7%2BuWbMT0EQgqOu%2FUTCKIX7JLD8tJJHb1bx90R7BfXWfXKbrg94tOrItTV11IAv5uFUeVa24iDH%2BLCoStSBxskYQJl7MosBPuA0BXnrGsSTi545yLxxZnoUfHNRqltwIQRNut3XxdoDTpLlkirFKMGkIeyY8wJLQZfGoXj9bcau%2BGPw3Ikb20%2FdA79erq9GcJaANTefB3JHU%2FK2c%2B8zZLVnSB1V3%2BtR%2FTWW6thVzY8IFeXGMd9jvKnt9bX4CciVWPyQeBvKVniiLYqwFAdBGbLYmTr72bCfP%2FxA6JAsHooKLkX6bNgGNYbbT38%2BNGS4PDV5tZ7Ud8H13vKDvwn7Bux6JRqbf9erjLK7UJ9KmsLFPGdCvGMWO0c3STI3sw1vJgRQ6DAwwM6kKq4ui3k3yBHMGfF%2FW8MSdLB93B4W3qeKTriP5f8UX8Kw2LmNzLQnat7mPAmCunY3NHOpwaKMiES5XVhXcHGlKFYTz9V6spE3cwhLqYwwY63gJO%2FcyKZeiz3yHTivM0FJNm2TMuIhLmUm9KHl%2BEDGDP9x4XfVnbWza1M%2Fhjhkp7OQVZK0WWMsRmgzSKN%2F%2B0AYGf%2ByswpK6fUKJbeFeDDIxOTh3d8TWJTXj3wDTKuQddBNqLVw6XUuOmVwFBB5%2FLpVvywLDb3iVcNqOmlFJcR2jBR53n0EpZbfIW9j28xVlI965f%2BhcXMhK0D6yvNody3PegC9lDNi7bRERsj72QGCC7Na3HNliMURha2TZoo24SHLa1YLcsyu5ojw%2FIN%2BggWTTMg%2FE%2BWsP9eeEIxNLDj5G7nXkCEdjbfuKPoWUdebP15z9Dknyl7W1sDN2hai5B%2Bk%2B6hTaZziZxr%2FI%2FU2EHgwWkj2ya5qArGlpfK%2BAhlH5%2BVA9I3P3pv7ckjp2J6f2TUErgI3maueoSiTlbQqpEeSVNGu%2BDODQx5799eMGf0f448b29I%2BXKUrDlH0Yoj1RmEA%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAZYG7WWMZCCDE37TC%2F20250703%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20250703T143406Z&X-Amz-Expires=21600&X-Amz-SignedHeaders=host&X-Amz-Signature=3d7e773fe4a25d0a6dd36aecd9176b13b1717033a69610cfb6815ed156a1b93c", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "remote", "upload_group_key": "20250703T143413", "original_file_md5": "", "original_file_name": "day_cmd.log", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "log", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:34:24.203098-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:34:24.279969-07	f
382785ea-95f6-45b9-b63a-71ba7c409c3d	FS17	file_set:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T143426", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	file_set	file_instance	file	generic	2025-07-03 07:34:37.863248-07	1.0	ready	f	1c57e0bf-8b44-426b-8f9e-e96aff46a341	2025-07-03 07:34:37.863248-07	f
47f023f3-25bd-4d38-bbf1-c9465815efe0	FI7040	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "day_cmd.log", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "log", "patient_id": "", "clinician_id": "", "external_ids": [], "original_url": "https://daylily-service-omics-analysis-us-west-2.s3.us-west-2.amazonaws.com/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log?response-content-disposition=inline&X-Amz-Content-Sha256=UNSIGNED-PAYLOAD&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEA8aCXVzLXdlc3QtMiJHMEUCIEwkFC7typ%2B7k712by339aiZJx867GzZv1rwLjaMWjy9AiEAyFFAGSuNMLiCibMYT9IuQiYJjXD59ZexbYTehsL2hXoquQMIGBAAGgw2NzA0ODQwNTA3MzgiDOLLh4KTH%2Bja%2FjEo0CqWA32BzYyhHkRrNyY1g6iBHMD91fBPNWdcSCjPQ3hFJ9USdNJZJG9R7%2BuWbMT0EQgqOu%2FUTCKIX7JLD8tJJHb1bx90R7BfXWfXKbrg94tOrItTV11IAv5uFUeVa24iDH%2BLCoStSBxskYQJl7MosBPuA0BXnrGsSTi545yLxxZnoUfHNRqltwIQRNut3XxdoDTpLlkirFKMGkIeyY8wJLQZfGoXj9bcau%2BGPw3Ikb20%2FdA79erq9GcJaANTefB3JHU%2FK2c%2B8zZLVnSB1V3%2BtR%2FTWW6thVzY8IFeXGMd9jvKnt9bX4CciVWPyQeBvKVniiLYqwFAdBGbLYmTr72bCfP%2FxA6JAsHooKLkX6bNgGNYbbT38%2BNGS4PDV5tZ7Ud8H13vKDvwn7Bux6JRqbf9erjLK7UJ9KmsLFPGdCvGMWO0c3STI3sw1vJgRQ6DAwwM6kKq4ui3k3yBHMGfF%2FW8MSdLB93B4W3qeKTriP5f8UX8Kw2LmNzLQnat7mPAmCunY3NHOpwaKMiES5XVhXcHGlKFYTz9V6spE3cwhLqYwwY63gJO%2FcyKZeiz3yHTivM0FJNm2TMuIhLmUm9KHl%2BEDGDP9x4XfVnbWza1M%2Fhjhkp7OQVZK0WWMsRmgzSKN%2F%2B0AYGf%2ByswpK6fUKJbeFeDDIxOTh3d8TWJTXj3wDTKuQddBNqLVw6XUuOmVwFBB5%2FLpVvywLDb3iVcNqOmlFJcR2jBR53n0EpZbfIW9j28xVlI965f%2BhcXMhK0D6yvNody3PegC9lDNi7bRERsj72QGCC7Na3HNliMURha2TZoo24SHLa1YLcsyu5ojw%2FIN%2BggWTTMg%2FE%2BWsP9eeEIxNLDj5G7nXkCEdjbfuKPoWUdebP15z9Dknyl7W1sDN2hai5B%2Bk%2B6hTaZziZxr%2FI%2FU2EHgwWkj2ya5qArGlpfK%2BAhlH5%2BVA9I3P3pv7ckjp2J6f2TUErgI3maueoSiTlbQqpEeSVNGu%2BDODQx5799eMGf0f448b29I%2BXKUrDlH0Yoj1RmEA%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAZYG7WWMZCCDE37TC%2F20250703%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20250703T143406Z&X-Amz-Expires=21600&X-Amz-SignedHeaders=host&X-Amz-Signature=3d7e773fe4a25d0a6dd36aecd9176b13b1717033a69610cfb6815ed156a1b93c", "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7040.log", "current_s3_uri": "s3://daylily-dewey-0/0/FI7040.log", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143426", "original_file_md5": "", "original_file_name": "day_cmd.log", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "log", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 3567, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:34:37.902654-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:34:38.160581-07	f
471467e2-7d73-4ec1-957c-980027da7ccc	FI7041	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "day_cmd.log", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "log", "patient_id": "", "clinician_id": "", "external_ids": [], "original_url": "https://daylily-service-omics-analysis-us-west-2.s3.us-west-2.amazonaws.com/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log?response-content-disposition=inline&X-Amz-Content-Sha256=UNSIGNED-PAYLOAD&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEA8aCXVzLXdlc3QtMiJHMEUCIEwkFC7typ%2B7k712by339aiZJx867GzZv1rwLjaMWjy9AiEAyFFAGSuNMLiCibMYT9IuQiYJjXD59ZexbYTehsL2hXoquQMIGBAAGgw2NzA0ODQwNTA3MzgiDOLLh4KTH%2Bja%2FjEo0CqWA32BzYyhHkRrNyY1g6iBHMD91fBPNWdcSCjPQ3hFJ9USdNJZJG9R7%2BuWbMT0EQgqOu%2FUTCKIX7JLD8tJJHb1bx90R7BfXWfXKbrg94tOrItTV11IAv5uFUeVa24iDH%2BLCoStSBxskYQJl7MosBPuA0BXnrGsSTi545yLxxZnoUfHNRqltwIQRNut3XxdoDTpLlkirFKMGkIeyY8wJLQZfGoXj9bcau%2BGPw3Ikb20%2FdA79erq9GcJaANTefB3JHU%2FK2c%2B8zZLVnSB1V3%2BtR%2FTWW6thVzY8IFeXGMd9jvKnt9bX4CciVWPyQeBvKVniiLYqwFAdBGbLYmTr72bCfP%2FxA6JAsHooKLkX6bNgGNYbbT38%2BNGS4PDV5tZ7Ud8H13vKDvwn7Bux6JRqbf9erjLK7UJ9KmsLFPGdCvGMWO0c3STI3sw1vJgRQ6DAwwM6kKq4ui3k3yBHMGfF%2FW8MSdLB93B4W3qeKTriP5f8UX8Kw2LmNzLQnat7mPAmCunY3NHOpwaKMiES5XVhXcHGlKFYTz9V6spE3cwhLqYwwY63gJO%2FcyKZeiz3yHTivM0FJNm2TMuIhLmUm9KHl%2BEDGDP9x4XfVnbWza1M%2Fhjhkp7OQVZK0WWMsRmgzSKN%2F%2B0AYGf%2ByswpK6fUKJbeFeDDIxOTh3d8TWJTXj3wDTKuQddBNqLVw6XUuOmVwFBB5%2FLpVvywLDb3iVcNqOmlFJcR2jBR53n0EpZbfIW9j28xVlI965f%2BhcXMhK0D6yvNody3PegC9lDNi7bRERsj72QGCC7Na3HNliMURha2TZoo24SHLa1YLcsyu5ojw%2FIN%2BggWTTMg%2FE%2BWsP9eeEIxNLDj5G7nXkCEdjbfuKPoWUdebP15z9Dknyl7W1sDN2hai5B%2Bk%2B6hTaZziZxr%2FI%2FU2EHgwWkj2ya5qArGlpfK%2BAhlH5%2BVA9I3P3pv7ckjp2J6f2TUErgI3maueoSiTlbQqpEeSVNGu%2BDODQx5799eMGf0f448b29I%2BXKUrDlH0Yoj1RmEA%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAZYG7WWMZCCDE37TC%2F20250703%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20250703T143406Z&X-Amz-Expires=21600&X-Amz-SignedHeaders=host&X-Amz-Signature=3d7e773fe4a25d0a6dd36aecd9176b13b1717033a69610cfb6815ed156a1b93c", "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7041.log", "current_s3_uri": "s3://daylily-dewey-0/0/FI7041.log", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143426", "original_file_md5": "", "original_file_name": "day_cmd.log", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "log", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 3567, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:34:38.765076-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:34:38.844921-07	f
42090aeb-3afd-498a-b55f-4af3689f2379	FI7042	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "day_cmd.log", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "log", "patient_id": "", "clinician_id": "", "external_ids": [], "original_url": "https://daylily-service-omics-analysis-us-west-2.s3.us-west-2.amazonaws.com/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log?response-content-disposition=inline&X-Amz-Content-Sha256=UNSIGNED-PAYLOAD&X-Amz-Security-Token=IQoJb3JpZ2luX2VjEA8aCXVzLXdlc3QtMiJHMEUCIEwkFC7typ%2B7k712by339aiZJx867GzZv1rwLjaMWjy9AiEAyFFAGSuNMLiCibMYT9IuQiYJjXD59ZexbYTehsL2hXoquQMIGBAAGgw2NzA0ODQwNTA3MzgiDOLLh4KTH%2Bja%2FjEo0CqWA32BzYyhHkRrNyY1g6iBHMD91fBPNWdcSCjPQ3hFJ9USdNJZJG9R7%2BuWbMT0EQgqOu%2FUTCKIX7JLD8tJJHb1bx90R7BfXWfXKbrg94tOrItTV11IAv5uFUeVa24iDH%2BLCoStSBxskYQJl7MosBPuA0BXnrGsSTi545yLxxZnoUfHNRqltwIQRNut3XxdoDTpLlkirFKMGkIeyY8wJLQZfGoXj9bcau%2BGPw3Ikb20%2FdA79erq9GcJaANTefB3JHU%2FK2c%2B8zZLVnSB1V3%2BtR%2FTWW6thVzY8IFeXGMd9jvKnt9bX4CciVWPyQeBvKVniiLYqwFAdBGbLYmTr72bCfP%2FxA6JAsHooKLkX6bNgGNYbbT38%2BNGS4PDV5tZ7Ud8H13vKDvwn7Bux6JRqbf9erjLK7UJ9KmsLFPGdCvGMWO0c3STI3sw1vJgRQ6DAwwM6kKq4ui3k3yBHMGfF%2FW8MSdLB93B4W3qeKTriP5f8UX8Kw2LmNzLQnat7mPAmCunY3NHOpwaKMiES5XVhXcHGlKFYTz9V6spE3cwhLqYwwY63gJO%2FcyKZeiz3yHTivM0FJNm2TMuIhLmUm9KHl%2BEDGDP9x4XfVnbWza1M%2Fhjhkp7OQVZK0WWMsRmgzSKN%2F%2B0AYGf%2ByswpK6fUKJbeFeDDIxOTh3d8TWJTXj3wDTKuQddBNqLVw6XUuOmVwFBB5%2FLpVvywLDb3iVcNqOmlFJcR2jBR53n0EpZbfIW9j28xVlI965f%2BhcXMhK0D6yvNody3PegC9lDNi7bRERsj72QGCC7Na3HNliMURha2TZoo24SHLa1YLcsyu5ojw%2FIN%2BggWTTMg%2FE%2BWsP9eeEIxNLDj5G7nXkCEdjbfuKPoWUdebP15z9Dknyl7W1sDN2hai5B%2Bk%2B6hTaZziZxr%2FI%2FU2EHgwWkj2ya5qArGlpfK%2BAhlH5%2BVA9I3P3pv7ckjp2J6f2TUErgI3maueoSiTlbQqpEeSVNGu%2BDODQx5799eMGf0f448b29I%2BXKUrDlH0Yoj1RmEA%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIAZYG7WWMZCCDE37TC%2F20250703%2Fus-west-2%2Fs3%2Faws4_request&X-Amz-Date=20250703T143406Z&X-Amz-Expires=21600&X-Amz-SignedHeaders=host&X-Amz-Signature=3d7e773fe4a25d0a6dd36aecd9176b13b1717033a69610cfb6815ed156a1b93c", "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7042.log", "current_s3_uri": "s3://daylily-dewey-0/0/FI7042.log", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143426", "original_file_md5": "", "original_file_name": "day_cmd.log", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "log", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 3567, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:34:39.314264-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:34:39.393563-07	f
7ddba52c-036c-420d-b112-ba9419d50c1d	FS18	file_set:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T143457", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	file_set	file_instance	file	generic	2025-07-03 07:35:07.907667-07	1.0	ready	f	1c57e0bf-8b44-426b-8f9e-e96aff46a341	2025-07-03 07:35:07.907667-07	f
5f8a84d3-a799-47fa-8e6e-e95496c7d320	FI7043	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "day_cmd.log", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "log", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "remote_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "current_s3_key": "FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "current_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "remote", "upload_group_key": "20250703T143457", "original_file_md5": "", "original_file_name": "day_cmd.log", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "log", "current_s3_bucket_name": "daylily-service-omics-analysis-us-west-2", "original_file_data_type": "s3uri", "original_file_size_bytes": null, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:35:08.22041-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:35:08.51139-07	f
afd98d90-d88b-4d7f-8384-8af9b64f4550	FI7044	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "day_cmd.log", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "log", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "remote_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "current_s3_key": "FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "current_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "remote", "upload_group_key": "20250703T143457", "original_file_md5": "", "original_file_name": "day_cmd.log", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "log", "current_s3_bucket_name": "daylily-service-omics-analysis-us-west-2", "original_file_data_type": "s3uri", "original_file_size_bytes": null, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:35:09.124723-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:35:09.391963-07	f
cd291654-0794-44d6-ad84-b147d8b5f61b	FI7045	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "day_cmd.log", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "log", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "remote_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "current_s3_key": "FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "current_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "remote", "upload_group_key": "20250703T143457", "original_file_md5": "", "original_file_name": "day_cmd.log", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "log", "current_s3_bucket_name": "daylily-service-omics-analysis-us-west-2", "original_file_data_type": "s3uri", "original_file_size_bytes": null, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:35:09.824113-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:35:10.060044-07	f
d8f83592-c673-4ac6-aef5-b79834139010	FS19	file_set:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T143512", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	file_set	file_instance	file	generic	2025-07-03 07:35:23.676353-07	1.0	ready	f	1c57e0bf-8b44-426b-8f9e-e96aff46a341	2025-07-03 07:35:23.676353-07	f
2975d2a7-516a-4e2d-8149-1835e51641fa	FI7046	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "day_cmd.log", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "log", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7046.log", "current_s3_uri": "s3://daylily-dewey-0/0/FI7046.log", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143512", "original_file_md5": "", "original_file_name": "day_cmd.log", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "log", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 3567, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:35:24.127515-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:35:24.418254-07	f
6c35ea5e-3cf4-45dd-9cc7-b1e1dc8f78a1	FI7047	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "day_cmd.log", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "log", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7047.log", "current_s3_uri": "s3://daylily-dewey-0/0/FI7047.log", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143512", "original_file_md5": "", "original_file_name": "day_cmd.log", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "log", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 3567, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:35:25.377865-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:35:25.612131-07	f
f8e647ac-e9b7-4af5-ba52-ddac34812049	FI7048	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "day_cmd.log", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "log", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7048.log", "current_s3_uri": "s3://daylily-dewey-0/0/FI7048.log", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T143512", "original_file_md5": "", "original_file_name": "day_cmd.log", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "log", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 3567, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:35:26.392421-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:35:26.629641-07	f
d3138620-73a5-40a2-b862-ca9cff5395a9	FX6	shared_ref:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "", "status": "active", "comments": "", "who_pays": "", "visibility": null, "end_datetime": "2025-07-04T14:35:57.325615+00:00", "presigned_url": "https://daylily-dewey-0.s3.amazonaws.com/0/FI7012.sql?AWSAccessKeyId=AKIAZYG7WWMZG2QOPT5Q&Signature=fiiII5EZloIrVzXDX%2BssA8DqDz4%3D&Expires=1751639757", "rclone_config": {}, "intended_users": [""], "reference_type": "presigned", "start_datetime": "2025-07-03T14:35:57.325615+00:00", "valid_duration": 86400.0}, "description": "Generic File Reference", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"form_label": "Reference Type", "property_key": "reference_type"}, {"form_label": "Visibility", "property_key": "visibility"}, {"form_label": "Reference Status", "property_key": "status"}, {"form_label": "Duration", "property_key": "valid_duration"}, {"form_label": "Start Datetime", "property_key": "start_datetime"}, {"form_label": "End Datetime", "property_key": "end_datetime"}, {"form_label": "Who Pays", "property_key": "who_pays"}], "controlled_properties": {}, "instantiation_layouts": [], "arbitrary_association_strings": []}	shared_ref	file_instance	file	generic	2025-07-03 07:35:57.329457-07	1.0	ready	f	64646531-f2fd-463a-9790-a4587d83f2a5	2025-07-03 07:35:57.329457-07	f
d7e5884d-3ffe-451a-8fbd-99716aff9bb9	FS4	file_set:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T134027", "tags": [], "comments": "", "duration": 1.0, "lab_code": "", "ref_type": "presigned_url", "description": "File set created by Dewey file manager", "creating_user": "john@daylilyinformatics.com", "rclone_config": {"host": "0.0.0.0", "port": 8080, "user": "user", "bucket": "daylily-dewey-0", "passwd": "passwd"}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	file_set	file_instance	file	generic	2025-07-03 06:40:40.947157-07	1.0	ready	f	1c57e0bf-8b44-426b-8f9e-e96aff46a341	2025-07-03 07:35:57.312669-07	f
82e5c3f3-22bc-4207-b3a5-7b29bd0ae741	FI7049	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "day_cmd.log", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "log", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "remote_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "current_s3_key": "FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "current_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "remote", "upload_group_key": "20250703T143658", "original_file_md5": "", "original_file_name": "day_cmd.log", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "log", "current_s3_bucket_name": "daylily-service-omics-analysis-us-west-2", "original_file_data_type": "s3uri", "original_file_size_bytes": null, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:37:07.256559-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:37:07.618985-07	f
11d26eb0-da43-4f29-b0c9-c234ca621a60	FX7	shared_ref:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "", "status": "active", "comments": "", "who_pays": "", "visibility": null, "end_datetime": "2025-07-04T14:37:32.780986+00:00", "presigned_url": "https://daylily-service-omics-analysis-us-west-2.s3.amazonaws.com/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log?AWSAccessKeyId=AKIAZYG7WWMZG2QOPT5Q&Signature=IEn226xamRr329fJ4gYg6Yl%2B3Dk%3D&Expires=1751639852", "rclone_config": {}, "intended_users": [""], "reference_type": "presigned", "start_datetime": "2025-07-03T14:37:32.780986+00:00", "valid_duration": 86400.0}, "description": "Generic File Reference", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"form_label": "Reference Type", "property_key": "reference_type"}, {"form_label": "Visibility", "property_key": "visibility"}, {"form_label": "Reference Status", "property_key": "status"}, {"form_label": "Duration", "property_key": "valid_duration"}, {"form_label": "Start Datetime", "property_key": "start_datetime"}, {"form_label": "End Datetime", "property_key": "end_datetime"}, {"form_label": "Who Pays", "property_key": "who_pays"}], "controlled_properties": {}, "instantiation_layouts": [], "arbitrary_association_strings": []}	shared_ref	file_instance	file	generic	2025-07-03 07:37:32.782376-07	1.0	ready	f	64646531-f2fd-463a-9790-a4587d83f2a5	2025-07-03 07:37:32.782376-07	f
d5b8c535-499d-4586-ae3b-4bf5a9871387	FS20	file_set:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T143658", "tags": [], "comments": "", "duration": 1.0, "lab_code": "", "ref_type": "presigned_url", "description": "File set created by Dewey file manager", "creating_user": "john@daylilyinformatics.com", "rclone_config": {"host": "0.0.0.0", "port": 8080, "user": "user", "bucket": "daylily-dewey-0", "passwd": "passwd"}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	file_set	file_instance	file	generic	2025-07-03 07:37:06.886121-07	1.0	ready	f	1c57e0bf-8b44-426b-8f9e-e96aff46a341	2025-07-03 07:37:32.770575-07	f
dde25f19-7032-40ff-a0e6-8ae39183e84b	FI7050	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "main.py", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "py", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "remote_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/main.py", "current_s3_key": "FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/main.py", "current_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/main.py", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/main.py", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "remote", "upload_group_key": "20250703T143915", "original_file_md5": "", "original_file_name": "main.py", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "py", "current_s3_bucket_name": "daylily-service-omics-analysis-us-west-2", "original_file_data_type": "s3uri", "original_file_size_bytes": null, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:39:23.277961-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:39:23.525991-07	f
21aad253-3d80-4b11-9220-290cf6ff8a34	FX8	shared_ref:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "", "status": "active", "comments": "", "who_pays": "", "visibility": null, "end_datetime": "2025-07-04T14:39:39.331627+00:00", "presigned_url": "https://daylily-service-omics-analysis-us-west-2.s3.amazonaws.com/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/main.py?AWSAccessKeyId=AKIAZYG7WWMZG2QOPT5Q&Signature=gu7M7V9hQeyb%2FY7Eew7xEDdZygk%3D&Expires=1751639979", "rclone_config": {}, "intended_users": [""], "reference_type": "presigned", "start_datetime": "2025-07-03T14:39:39.331627+00:00", "valid_duration": 86400.0}, "description": "Generic File Reference", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"form_label": "Reference Type", "property_key": "reference_type"}, {"form_label": "Visibility", "property_key": "visibility"}, {"form_label": "Reference Status", "property_key": "status"}, {"form_label": "Duration", "property_key": "valid_duration"}, {"form_label": "Start Datetime", "property_key": "start_datetime"}, {"form_label": "End Datetime", "property_key": "end_datetime"}, {"form_label": "Who Pays", "property_key": "who_pays"}], "controlled_properties": {}, "instantiation_layouts": [], "arbitrary_association_strings": []}	shared_ref	file_instance	file	generic	2025-07-03 07:39:39.332504-07	1.0	ready	f	64646531-f2fd-463a-9790-a4587d83f2a5	2025-07-03 07:39:39.332504-07	f
5834526f-b5d3-4f42-8384-9c26e5c74bc8	FS21	file_set:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T143915", "tags": [], "comments": "", "duration": 1.0, "lab_code": "", "ref_type": "presigned_url", "description": "File set created by Dewey file manager", "creating_user": "john@daylilyinformatics.com", "rclone_config": {"host": "0.0.0.0", "port": 8080, "user": "user", "bucket": "daylily-dewey-0", "passwd": "passwd"}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	file_set	file_instance	file	generic	2025-07-03 07:39:23.027872-07	1.0	ready	f	1c57e0bf-8b44-426b-8f9e-e96aff46a341	2025-07-03 07:39:39.32352-07	f
1bccdecd-c0cf-41d5-a0b5-13be26cdb88c	FI7051	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "rule_dag.png", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "png", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "remote_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/rule_dag.png", "current_s3_key": "FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/rule_dag.png", "current_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/rule_dag.png", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/rule_dag.png", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "remote", "upload_group_key": "20250703T144046", "original_file_md5": "", "original_file_name": "rule_dag.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-service-omics-analysis-us-west-2", "original_file_data_type": "s3uri", "original_file_size_bytes": null, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:40:54.504759-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:40:54.771021-07	f
6182f00d-f0e3-4b74-888f-54f693a6f8fe	FX9	shared_ref:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "", "status": "active", "comments": "", "who_pays": "", "visibility": null, "end_datetime": "2025-07-04T14:41:07.377745+00:00", "presigned_url": "https://daylily-service-omics-analysis-us-west-2.s3.amazonaws.com/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/rule_dag.png?AWSAccessKeyId=AKIAZYG7WWMZG2QOPT5Q&Signature=mtisEgmGQGGQ8%2FfbzoEXcVhFD%2FU%3D&Expires=1751640067", "rclone_config": {}, "intended_users": [""], "reference_type": "presigned", "start_datetime": "2025-07-03T14:41:07.377745+00:00", "valid_duration": 86400.0}, "description": "Generic File Reference", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"form_label": "Reference Type", "property_key": "reference_type"}, {"form_label": "Visibility", "property_key": "visibility"}, {"form_label": "Reference Status", "property_key": "status"}, {"form_label": "Duration", "property_key": "valid_duration"}, {"form_label": "Start Datetime", "property_key": "start_datetime"}, {"form_label": "End Datetime", "property_key": "end_datetime"}, {"form_label": "Who Pays", "property_key": "who_pays"}], "controlled_properties": {}, "instantiation_layouts": [], "arbitrary_association_strings": []}	shared_ref	file_instance	file	generic	2025-07-03 07:41:07.378265-07	1.0	ready	f	64646531-f2fd-463a-9790-a4587d83f2a5	2025-07-03 07:41:07.378265-07	f
2ff79f5f-1663-4dce-bbf3-13e0cc787c2d	FS22	file_set:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144046", "tags": [], "comments": "", "duration": 1.0, "lab_code": "", "ref_type": "presigned_url", "description": "File set created by Dewey file manager", "creating_user": "john@daylilyinformatics.com", "rclone_config": {"host": "0.0.0.0", "port": 8080, "user": "user", "bucket": "daylily-dewey-0", "passwd": "passwd"}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	file_set	file_instance	file	generic	2025-07-03 07:40:54.268794-07	1.0	ready	f	1c57e0bf-8b44-426b-8f9e-e96aff46a341	2025-07-03 07:41:07.372769-07	f
cd99a9ea-224d-4cda-b8a2-2027e4be1767	FS23	file_set:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T144308", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	file_set	file_instance	file	generic	2025-07-03 07:43:08.238971-07	1.0	ready	f	1c57e0bf-8b44-426b-8f9e-e96aff46a341	2025-07-03 07:43:08.238971-07	f
751326b8-0128-4f6d-8f82-05bf70cbd6c8	FI7052	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7052.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7052.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T144308", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:43:08.26847-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:43:08.517518-07	f
42c52d3e-cb20-4c08-87fe-c36891087f40	FI7053	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7053.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7053.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T144308", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:43:09.559286-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:43:09.844984-07	f
ade4c01a-0859-4437-b6b1-63a99088aaae	FI7054	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7054.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7054.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T144308", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:43:10.449991-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:43:10.679092-07	f
50d62952-ee77-48c2-adf1-a4fd30234dd6	FI7055	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7055.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7055.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T144308", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:43:11.46863-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:43:11.77149-07	f
81d233d2-dffb-4b82-9cea-61c8009b5f2b	FS24	file_set:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	file_set	file_instance	file	generic	2025-07-03 07:46:17.728972-07	1.0	ready	f	1c57e0bf-8b44-426b-8f9e-e96aff46a341	2025-07-03 07:46:17.728972-07	f
ec941c78-3c25-4399-bb69-f266930774bb	CX8	plate:fixed-plate-24:1.0	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	plate	container_instance	container	fixed-plate-24	2025-07-03 08:23:08.671329-07	1.0	ready	f	6bae2028-8873-4717-9e10-821a62bcd9f0	2025-07-03 08:23:08.671329-07	f
f3300273-7c20-42a6-aa2d-101f1f721a6d	FI7056	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp2s1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7056.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7056.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp2s1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 809418, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:17.796677-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:18.064288-07	f
5b761ebc-9320-4d44-9d95-c80a289eb8ae	FI7057	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp2f1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7057.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7057.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp2f1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 980260, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:18.804449-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:18.918822-07	f
c260d8ec-016f-445e-977a-fb012e00481b	FI7058	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/tbxas1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7058.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7058.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "tbxas1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1085371, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:19.561462-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:19.679578-07	f
04b20693-2770-4741-ab5e-af1f0f85f28e	FI7059	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/ugt2b17.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7059.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7059.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "ugt2b17.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 987018, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:20.290547-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:20.409103-07	f
dc3b6ed6-0ec6-4cde-98e4-5ae861b129fe	FI7060	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/por.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7060.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7060.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "por.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 837876, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:21.05117-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:21.168782-07	f
3b9591c8-f892-4b39-ba2f-95894da01bd1	FI7061	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp2c8.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7061.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7061.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp2c8.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 947576, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:21.79047-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:21.90558-07	f
2eb9d931-5a00-44ea-99d0-b50cc6d630f1	FI7062	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp2b6.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7062.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7062.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp2b6.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 992530, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:22.523954-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:22.641055-07	f
4d7880f2-f6ba-498d-9d36-a065b7c26a62	FI7063	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/ugt2b7.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7063.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7063.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "ugt2b7.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1288708, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:23.119651-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:23.201282-07	f
6033d3d2-b9bb-4ec7-8afd-a90254f310aa	FI7064	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp1a1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7064.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7064.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp1a1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 419961, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:23.856566-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:24.018626-07	f
7340a717-a09a-4c3c-9cd6-0a5e52a77978	FI7065	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp2r1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7065.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7065.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp2r1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 473765, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:24.63568-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:24.753214-07	f
611dcea7-1da6-46e9-b035-b9658cef5ab2	FI7066	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/ugt2b15.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7066.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7066.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "ugt2b15.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1082490, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:25.35827-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:25.476017-07	f
b007b95e-b1a6-42c6-bc19-24e1e5f779b9	FI7067	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp2j2.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7067.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7067.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp2j2.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 503177, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:26.080622-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:26.192291-07	f
2763560b-cabd-424f-a685-df92ca63080c	FI7068	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp2e1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7068.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7068.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp2e1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 899848, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:26.846633-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:26.958402-07	f
cd4284b9-8e2c-4ab9-94e5-2ed1630e323e	FI7069	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/gstp1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7069.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7069.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "gstp1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 685918, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:27.532219-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:27.683211-07	f
81948d72-dcd4-4211-819c-96e65e693d0e	FI7070	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/nat2.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7070.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7070.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "nat2.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1160607, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:28.295525-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:28.417356-07	f
77115b00-ac73-4144-80b9-a0bd9b125164	FI7071	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp2a13.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7071.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7071.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp2a13.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 987555, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:29.044606-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:29.159846-07	f
ee740bac-1bae-443e-8f54-c4d4d7bf371f	FI7072	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp2a6.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7072.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7072.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp2a6.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 882980, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:29.749324-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:29.870826-07	f
cd142cc8-8540-40d1-8b89-0320a209d5f1	FI7073	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/slc22a2.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7073.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7073.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "slc22a2.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1197684, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:30.503351-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:30.617386-07	f
9a54fbb8-c9e1-4d2e-8283-db8e5d7b84fc	FI7074	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/slco1b1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7074.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7074.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "slco1b1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1325694, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:31.173251-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:31.281848-07	f
16c52e6c-5995-449c-bd88-bd1500aab2c2	FI7075	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/ifnl3.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7075.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7075.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "ifnl3.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 620168, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:31.760213-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:31.843366-07	f
ac534fb1-c4b3-4163-a549-9d6c18fad16f	FI7076	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/slco1b3.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7076.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7076.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "slco1b3.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1703270, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:32.281743-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:32.362033-07	f
9ab24170-9de3-4cf6-929e-9389976f8a03	FI7077	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/ugt1a4.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7077.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7077.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "ugt1a4.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1012518, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:33.088279-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:33.226272-07	f
5e52c4b2-b976-433e-ba92-2aabefb9585e	FI7078	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cftr.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7078.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7078.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cftr.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 738060, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:33.790585-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:33.899393-07	f
e391ed08-8608-44ff-abc0-16fe6c7baf3d	FI7079	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp3a4.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7079.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7079.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp3a4.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 457772, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:34.386032-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:34.468559-07	f
bbf86074-7a49-44d5-a85a-36cd8bbf2da6	FI7080	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/vkorc1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7080.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7080.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "vkorc1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 398598, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:34.887222-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:34.971248-07	f
6968dcc3-4411-486e-a424-89dc7a8a9875	FI7081	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/gstm1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7081.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7081.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "gstm1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 644122, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:35.525532-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:35.615832-07	f
f0f17b46-d607-4117-b350-68dd1beb7c59	FI7082	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/dpyd.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7082.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7082.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "dpyd.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 2534931, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:36.091124-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:36.225606-07	f
628aaab0-bc64-42ae-9199-121c2cfae0be	FI7083	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp1b1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7083.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7083.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp1b1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 935399, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:36.926256-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:37.02709-07	f
d730ed0a-3439-475d-a55e-031e0f4fac5c	FI7084	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp1a2.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7084.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7084.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp1a2.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 445280, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:37.522065-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:37.679385-07	f
b1f51d3d-d2a3-48c0-9487-bc386b4a658d	FI7085	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp2c9.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7085.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7085.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp2c9.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 938026, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:38.14582-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:38.232699-07	f
652238e6-836f-4e53-ad34-062159021e26	FI7086	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/ugt1a1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7086.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7086.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "ugt1a1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 922059, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:38.730285-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:38.814929-07	f
a99ca08f-34f7-4f31-b30f-d765444bc221	FI7087	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/gstt1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7087.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7087.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "gstt1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 553675, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:39.294807-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:39.391185-07	f
6eeeffdd-a461-4bb2-ace0-e0f74637b47d	FI7088	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp2d6.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7088.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7088.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp2d6.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 682704, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:39.833297-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:39.915264-07	f
1f8fc77a-8554-4716-8f6f-bd177139f82f	FI7089	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp4b1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7089.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7089.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp4b1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 734217, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:40.525269-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:40.630664-07	f
6a8be340-a613-4e02-a89f-e1f09be2be4a	FI7090	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/ryr1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7090.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7090.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "ryr1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1094999, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:41.378868-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:41.512513-07	f
378516a2-6482-48f1-9e85-e33026c12b21	FI7091	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp2c19.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7091.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7091.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp2c19.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 925411, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:42.275354-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:42.418971-07	f
50f413f9-6000-4723-896a-926c67f39d59	FI7092	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/sult1a1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7092.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7092.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "sult1a1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 468288, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:43.172535-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:43.309086-07	f
5dae0e31-2c76-4322-bdfa-2f2d2f239bb2	FI7093	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/tpmt.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7093.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7093.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "tpmt.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 824178, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:44.0301-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:44.229127-07	f
7828fb01-12d3-4b62-8452-7c919cea3ab9	FI7094	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp3a7.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7094.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7094.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp3a7.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 417287, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:44.854199-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:44.985226-07	f
e010952a-8da4-4573-9b04-112e6178d9f8	FI7095	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp2w1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7095.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7095.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp2w1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1084808, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:45.724499-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:45.883195-07	f
e2bd006f-4d32-4a15-b87c-68a3eb36b37e	FI7096	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp4f2.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7096.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7096.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp4f2.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1339117, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:46.597019-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:46.741659-07	f
ddbd7604-9140-472d-87fe-44279e0463b4	FI7097	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp26a1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7097.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7097.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp26a1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 611814, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:47.514419-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:47.670338-07	f
a6ab103e-ac0d-4bda-b989-6beb34419280	CX17	plate:fixed-plate-24:1.0	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	plate	container_instance	container	fixed-plate-24	2025-07-03 08:24:15.648493-07	1.0	ready	f	6bae2028-8873-4717-9e10-821a62bcd9f0	2025-07-03 08:24:15.648493-07	f
16a63367-c957-46fb-aee4-f026757bd059	FI7098	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/nudt15.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7098.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7098.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "nudt15.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 851608, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:48.337838-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:48.476434-07	f
82961788-e3f9-441f-96f3-de01e051f122	FI7108	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "tbxas1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7108.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7108.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T144901", "original_file_md5": "", "original_file_name": "tbxas1.vcf.gz", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1085371, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:49:08.773823-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:49:09.033598-07	f
eeee8645-360f-4290-a6f7-a7275d34db61	FS26	file_set:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T145008", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	file_set	file_instance	file	generic	2025-07-03 07:50:08.537193-07	1.0	ready	f	1c57e0bf-8b44-426b-8f9e-e96aff46a341	2025-07-03 07:50:08.537193-07	f
205773c7-233a-423b-9b6f-704a210048fc	FI7099	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp19a1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7099.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7099.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp19a1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1103392, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:49.184805-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:49.351009-07	f
8b323316-e042-4abb-9e7c-948747452c5e	FI7100	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/g6pd.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7100.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7100.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "g6pd.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 253692, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:50.043178-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:50.195309-07	f
a2bc4b6d-42e0-4993-91ea-cb3b84c406ba	FI7109	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7109.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7109.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T145008", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:50:08.569214-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:50:08.837143-07	f
51f409ba-37e0-46f3-954f-eeb8b60c8886	FI7101	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp3a43.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7101.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7101.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp3a43.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 534281, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:50.867942-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:51.025392-07	f
9379619e-93d1-4e5e-a313-9576af3ffaab	FI7110	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7110.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7110.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T145008", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:50:09.773183-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:50:10.082248-07	f
6be7b5f3-b9a6-436a-8c5a-d011d47eee2d	FI7102	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/slco2b1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7102.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7102.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "slco2b1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 696892, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:51.714891-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:51.879088-07	f
9a2d33f4-0c38-4b92-b85c-e3650dbf67d5	FI7103	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cacna1s.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7103.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7103.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cacna1s.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1004301, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:52.475961-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:52.59389-07	f
c01298e9-f706-48dd-b71e-7666927fccf3	FI7111	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7111.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7111.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T145008", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:50:10.713804-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:50:10.941052-07	f
7d5dcd5e-7d18-4094-aab9-64c5da8698ec	FI7104	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/slc15a2.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7104.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7104.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "slc15a2.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 793207, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:53.082013-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:53.194867-07	f
9e89f061-44e0-4163-95f7-03de5e7d595b	FI7112	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7112.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7112.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T145008", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:50:11.650525-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:50:11.952358-07	f
bb3c68a8-5a54-4a62-b9e3-d83cfeacf4f6	FI7105	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/tpmt.vcf", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "vcf", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7105.vcf", "current_s3_uri": "s3://daylily-dewey-0/0/FI7105.vcf", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "tpmt.vcf", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "vcf", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 28012779, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:53.724439-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:53.840522-07	f
3e50c2f7-5e0b-4ae9-b721-cc81d94c22de	CX1	plate:fixed-plate-24:1.0	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	plate	container_instance	container	fixed-plate-24	2025-07-03 08:21:02.222701-07	1.0	ready	f	6bae2028-8873-4717-9e10-821a62bcd9f0	2025-07-03 08:21:02.222701-07	f
b32d8a96-efe1-42c3-abe3-7a3241e0c102	FI7106	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/cyp3a5.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7106.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7106.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "cyp3a5.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 441422, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:55.434585-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:55.698071-07	f
ee6c92d8-88fe-4743-af74-1ea3834c8407	MRX7	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_7", "probe_seq_1": "TACATCATGTGCTCCAGT", "probe_seq_2": "CATAAGTACCGGCTATCA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:21:02.579544-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:21:02.579544-07	f
20150c7c-2e35-4810-8f6c-0e7779530fac	MRX8	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_8", "probe_seq_1": "TGTCGCGCCTTACCAGTC", "probe_seq_2": "GGGGTCCCTGTGTACCTT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:21:02.584966-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:21:02.584966-07	f
051aa336-0ce7-4f70-bc74-9efbc624fce9	FI7107	file:generic:1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "1kgp_vcf/nat1.vcf.gz", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "gz", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7107.gz", "current_s3_uri": "s3://daylily-dewey-0/0/FI7107.gz", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2025-07-02", "import_or_remote": "Import", "upload_group_key": "20250703T144506", "original_file_md5": "", "original_file_name": "nat1.vcf.gz", "original_file_path": "1kgp_vcf", "original_server_ip": "", "record_datetime_end": "2025-07-02", "original_file_suffix": "gz", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1078162, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	file	file_instance	file	generic	2025-07-03 07:46:56.362055-07	1.0	ready	f	9a500c83-8b62-48e1-a0b0-cddd7c498e29	2025-07-03 07:46:56.518702-07	f
25cad9ef-1267-48b7-9a19-239743a8caea	MRX9	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_9", "probe_seq_1": "TTTCGCGTCTCCCGGCGC", "probe_seq_2": "GAGCACGAATAGTAGCCA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:21:02.591786-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:21:02.591786-07	f
e04744fc-bc25-486c-a066-89db9af94e92	MRX10	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_10", "probe_seq_1": "CATAGAAAAAATCTGTAT", "probe_seq_2": "GCCCGTAAGCTATCTGCC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:21:02.596373-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:21:02.596373-07	f
b9dbd159-998b-4db7-98ba-fb4f220cd04b	MRX11	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_11", "probe_seq_1": "TAAATTAGCCTGTAACCC", "probe_seq_2": "GTCGTGGCAGGGCACAGG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:21:02.600841-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:21:02.600841-07	f
dacc9c78-e115-4376-8a8c-9869504d64a7	CWX1	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.273663-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.279459-07	f
438c11ad-e609-4041-aa3e-75e8f719cef9	CWX2	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.293073-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.305077-07	f
f4b245da-a7f5-4a87-b23f-e24f173695d1	CWX3	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.30858-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.316622-07	f
9abd0e47-97f9-4d8e-ac09-cec264f3ee0d	CWX4	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.318139-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.324744-07	f
a778a7e1-7154-40f8-90c2-8c264b021681	CWX85	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:08.707623-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:08.7125-07	f
a339e966-d124-4730-bf7b-f27ca87c4e8d	CWX5	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.326916-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.333175-07	f
486a95ec-28d0-4b5a-b33e-0de10f65806c	CWX6	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.335164-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.341516-07	f
dee275bb-a13a-455a-879c-29bcddda4116	CWX7	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.343198-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.349401-07	f
843a9ccc-36ca-4d8b-80c8-a38487e0a12d	CWX225	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:15.677773-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:15.683175-07	f
b79a21a8-b8d0-4b24-8b54-01005554bf43	CWX8	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.351211-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.356925-07	f
ad8e242c-7636-42b1-9284-401ea589e648	CWX9	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.359286-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.36539-07	f
84bbdac6-5ab1-480c-8715-8dd7271b19bf	CWX10	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.366841-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.372549-07	f
ccf9f662-857d-49b6-8a43-8f9097e466f6	CWX86	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:08.715457-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:08.721885-07	f
a1e066c9-b6ad-4644-a288-029b6337c599	CWX11	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.374692-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.379966-07	f
4b07afc5-df46-4a85-8d24-2e74bb943460	CWX12	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.381424-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.386963-07	f
01c9d691-4dd1-4a31-88b8-328c9e47a8ec	CWX13	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.388919-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.394171-07	f
e3a263a5-ad0e-42e5-90f7-3eea9eb68be0	CWX14	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.398253-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.405579-07	f
005041ba-80b0-494a-9982-4f02023c76b9	CWX15	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.407525-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.412588-07	f
d0e326c1-0ca3-40e7-a7a3-58add70136f4	CWX16	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.413998-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.419312-07	f
7bffe3df-228e-49d0-b3d3-ea7566c97b58	CWX87	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:08.723816-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:08.729768-07	f
6413aa15-e457-4196-a660-67d30c22a768	CWX17	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.421024-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.426035-07	f
aabb851f-9f1b-44e8-86e2-2239eb90be7b	CWX18	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.427792-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.433789-07	f
47df2675-749e-4933-b64a-98e9f92f9891	CWX19	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.435195-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.447507-07	f
37a7fae4-216d-4842-ab8f-5c048d4a2809	CWX226	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:15.690826-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:15.702113-07	f
1d487141-743b-4f01-ac33-bbdd81756836	CWX20	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.465875-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.50052-07	f
d2e4093e-86f0-4c14-b4de-2e90a1f635a5	CWX21	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.50501-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.513494-07	f
2ec41575-3c9e-41db-a210-0aa45aa7c05e	CWX22	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.514912-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.519435-07	f
29ca4ea9-a532-40b8-90a2-bf62f212272c	CWX88	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:08.735751-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:08.741924-07	f
b5b21e36-376a-4491-b698-452597c517e8	CWX23	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.520696-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.524632-07	f
23f00e77-1469-4ea9-a978-11a2da61dd29	CWX24	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.525924-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.530253-07	f
e4e0fd64-c65d-4cc7-8fa8-1a052ae78bcf	CWX25	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.531634-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.535926-07	f
9ad9dea4-053c-4429-83e2-de60c13f4f67	CWX26	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.537114-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.541082-07	f
d13c47a9-1614-4d77-a15d-1872be28308e	CWX27	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.542464-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.546637-07	f
248543ec-5f5f-4421-879c-11e5899229fa	CWX28	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.547786-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.55174-07	f
73f06c89-2ce5-4e44-a22e-9441cd06dc0d	MRX1	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_1", "probe_seq_1": "AGTACATTGATCACTAAT", "probe_seq_2": "CAGACCCTCTTTAATACC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:21:02.554872-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:21:02.554872-07	f
fa310d7e-c727-402a-9c6c-47eafd01096c	MRX2	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_2", "probe_seq_1": "TGACCTTTGTAGTATTCT", "probe_seq_2": "GTTTAACACTCGTACCTA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:21:02.557162-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:21:02.557162-07	f
155d76ca-0ef6-4610-8991-3caf5eebcaf9	MRX3	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_3", "probe_seq_1": "CTACATAAAGGTCGTGCG", "probe_seq_2": "TTATAAGTAGAAGACTTT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:21:02.562107-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:21:02.562107-07	f
7f61478d-3e5c-47b1-a942-508cfe92ce65	MRX4	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_4", "probe_seq_1": "TCAGTTATAAAACCCCCG", "probe_seq_2": "CCAGAGGTCTGCCCCGCG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:21:02.566978-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:21:02.566978-07	f
d7eb32c1-6513-4a8e-b241-0bba13be7302	MRX5	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_5", "probe_seq_1": "ATCACCTCAGACTAGCCA", "probe_seq_2": "TTTACGGCTACACCAGAC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:21:02.571207-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:21:02.571207-07	f
1122f95f-07f3-4153-b7b9-eb3710b3370e	MRX6	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_6", "probe_seq_1": "CTCCCACCTACGAAACAT", "probe_seq_2": "CTCGTGGGTTATCAGGTA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:21:02.575611-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:21:02.575611-07	f
eea51634-39be-405d-b1e1-f2b229f006d2	MRX12	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_12", "probe_seq_1": "AGGACTCTTCTAGAGCTG", "probe_seq_2": "CATCACGACTCCTTGTAT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:21:02.605545-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:21:02.605545-07	f
0e1fd8ab-89b9-4237-8f0d-86d844c9c3d9	MRX13	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_13", "probe_seq_1": "GTATGCGCCTATGGTCTT", "probe_seq_2": "ATCTTTATTGCTTTTATA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:21:02.610022-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:21:02.610022-07	f
6e9a2e5a-f406-43d6-b310-2357457c94db	MRX14	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_14", "probe_seq_1": "AGTGCTTTAGAACCATAC", "probe_seq_2": "AGCCAGCGATATCAAAGT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:21:02.614709-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:21:02.614709-07	f
3c3c3a45-565a-4e97-9854-2702e0725454	MRX15	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_15", "probe_seq_1": "GAGCACCACACAACGTTC", "probe_seq_2": "GGGTATTCTCACTCCACA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:21:02.619075-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:21:02.619075-07	f
3fddfa66-2e48-47d6-8b04-a54d3c08aa2f	MRX16	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_16", "probe_seq_1": "AAAACCCCAAGAAGCCCA", "probe_seq_2": "CTAATGAGTTCTCAATCG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:21:02.623871-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:21:02.623871-07	f
94af7ed1-6a16-463c-9cce-442d3344703f	MRX17	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_17", "probe_seq_1": "CCGTGCGCCGGCCTAAGA", "probe_seq_2": "CGAATAGGGTTACTCACA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:21:02.628909-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:21:02.628909-07	f
d9adf020-6a54-47b1-ae11-60769eb81e1d	MRX18	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_18", "probe_seq_1": "AACGCAACCGTCCACAGC", "probe_seq_2": "GTAGTACGCATGACTGAA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:21:02.633227-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:21:02.633227-07	f
f3dfc3a9-b905-4286-9e05-ba68f5762d52	MRX19	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_19", "probe_seq_1": "AGCGCTATCCGGGGCGTA", "probe_seq_2": "AGTTTACGTGTAGATACC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:21:02.637441-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:21:02.637441-07	f
130e9a38-6071-44ea-b12c-0ac555011c7d	MRX20	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_20", "probe_seq_1": "TTTAAAGCATCATTAAAG", "probe_seq_2": "GTCATCGGACTCGGCTCC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:21:02.642032-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:21:02.642032-07	f
add01ba4-438a-48a5-aa94-62b29f76862b	MRX21	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_21", "probe_seq_1": "CTAATGAACCCAACGAGT", "probe_seq_2": "AAAGGTATGGACCTGGTC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:21:02.646617-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:21:02.646617-07	f
ce5401cf-dae7-4b77-988d-a7200152e343	MRX22	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_22", "probe_seq_1": "ATGAACACACATTATTAG", "probe_seq_2": "CCATAAAACGAGGTCAGA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:21:02.650909-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:21:02.650909-07	f
7d46d938-2e29-4a0e-b9ce-c132e7db46e6	MRX23	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_23", "probe_seq_1": "ACACGTGGCCTTCTCGTT", "probe_seq_2": "GGGGGCCGCGATGGGAGT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:21:02.655356-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:21:02.655356-07	f
a0017668-1487-4923-bf46-02ad0e4adae5	MRX24	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_24", "probe_seq_1": "AAGCCGCGGAGGTCTCGA", "probe_seq_2": "CGCATACGACTTGAGGAT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:21:02.660099-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:21:02.660099-07	f
0e8b6bac-2a70-473a-a382-8230c530bc86	MRX25	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_25", "probe_seq_1": "AACCTTGGGTCCGACTTT", "probe_seq_2": "CGCCGTTATGAGTGTACC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:21:02.664624-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:21:02.664624-07	f
b3c512e2-3442-47bf-a7ae-228de3b99ea8	MRX26	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_26", "probe_seq_1": "CTAACGGAATCTCTAACT", "probe_seq_2": "ATGGCACGAAGAGGACTA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:21:02.668815-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:21:02.668815-07	f
bf8c8f2e-8303-4213-a40f-25a7cce691d0	MRX27	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_27", "probe_seq_1": "TTTCAACTCACAGTCAGT", "probe_seq_2": "GAGGGCTGACAAGGAATT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:21:02.672962-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:21:02.672962-07	f
848b30bc-7457-4760-8ed6-4733e1c9bceb	MRX28	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_28", "probe_seq_1": "AAAGTGGATAGTAAGGTA", "probe_seq_2": "CTACAGTTAGATCAGTGG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:21:02.679104-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:21:02.679104-07	f
2fc73838-fb40-4247-9446-971fd96cff46	CX2	plate:fixed-plate-24:1.0	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	plate	container_instance	container	fixed-plate-24	2025-07-03 08:21:02.688189-07	1.0	ready	f	6bae2028-8873-4717-9e10-821a62bcd9f0	2025-07-03 08:21:02.688189-07	f
b98e9974-de3d-498d-8371-53ec167ca66f	CWX29	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.692069-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.694601-07	f
d97fc469-cfcf-42f6-b9f4-c694040b88c4	MRX32	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_4", "probe_seq_1": "CGAGAGTACTAACGCATA", "probe_seq_2": "AACGGAGCAAGATTGCAG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:02.861318-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:02.861318-07	f
f25b65aa-479f-4c10-b039-ecf50ea2340a	MRX33	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_5", "probe_seq_1": "TTGAGTCTGGCGTACCCG", "probe_seq_2": "TGTCCTCGAGCTTGTTCC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:02.866124-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:02.866124-07	f
9fc34612-8c1e-42b4-a30b-c95060c031ab	MRX34	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_6", "probe_seq_1": "ATACACGTTTCCTCTGAA", "probe_seq_2": "TGGATTTCGAGACTTGCG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:02.870652-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:02.870652-07	f
c5783a54-74bf-4417-b8d6-cfa029a644ec	MRX35	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_7", "probe_seq_1": "AATCGTACTACCTACTCA", "probe_seq_2": "TGTTGAAACGAGAACGTG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:02.877972-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:02.877972-07	f
7af7bee4-1b4e-4357-84a6-13e465094334	MRX36	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_8", "probe_seq_1": "CCCATATCCAGACCCCTC", "probe_seq_2": "TGGAACGTGAGCACTACT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:02.882499-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:02.882499-07	f
5932edad-b324-4826-b79a-4ff76745d991	MRX37	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_9", "probe_seq_1": "TCTCGGTACAGTATACGG", "probe_seq_2": "GCGGGTCAGCTATGTGCA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:02.886662-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:02.886662-07	f
a68a2de5-45b2-4db5-873f-eb0708d9b999	CWX30	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.69569-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.700045-07	f
3b804416-77a1-4840-a1cb-ddce018c129b	CWX89	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:08.743657-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:08.749126-07	f
5653c452-7c65-4fbf-a63a-555795d12dfd	CWX31	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.701299-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.705287-07	f
6ebf400e-84b9-48d3-9d87-05463e046efb	CWX32	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.706361-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.710076-07	f
490088ed-6d23-4654-9f8e-032e607177cf	CWX33	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.711446-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.715244-07	f
109675e8-0498-4c1c-bad5-788a996e69a2	CWX227	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:15.704863-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:15.710942-07	f
f8b71f68-c515-4c50-9d0e-c4dff9409c13	CWX34	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.716431-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.720483-07	f
a0eef834-450f-4780-8b5c-7c1db8b7ea67	CWX35	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.721585-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.725532-07	f
64d6d4da-233a-483b-b150-f76b15678000	CWX36	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.727093-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.731263-07	f
1bc05522-5e1d-4faa-8aaa-5b4a38ebc55a	CWX90	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:08.750583-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:08.755784-07	f
02114dfa-d423-42f2-adf7-65becb5f1255	CWX37	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.732411-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.73656-07	f
ee64ced1-19ac-4f65-844c-7b996b0eda0f	CWX38	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.738142-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.742113-07	f
7842b714-574f-4cb9-a83a-3eb1e194767a	CWX39	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.743496-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.747516-07	f
d4573a76-c4f1-4ed7-9424-90fdadead6a9	CWX40	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.749097-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.753202-07	f
f37b4964-1b4a-44e9-b71e-faa7ff7a0bb8	CWX41	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.754263-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.758213-07	f
db55beaf-f629-47b6-bf32-d6585fc2f1ca	CWX42	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.759768-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.764508-07	f
8ce6ad46-136d-48a8-ae5a-0d860b66c006	CWX91	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:08.757206-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:08.762245-07	f
26349cf2-a95d-4895-a58b-3226f432bfa7	CWX43	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.767064-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.775639-07	f
a237b8a6-236f-4af0-94cc-2ccac3708511	CWX44	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.776932-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.782574-07	f
247ddbff-98e3-4459-b22f-b4e7df6ea651	CWX45	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.784106-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.788755-07	f
20febaa0-790b-4079-8124-3d2d9586d49e	CWX228	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:15.712473-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:15.721173-07	f
e37cde14-ce6c-4cb8-a239-67ebf9db6a5a	CWX46	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.790176-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.794017-07	f
0eaa618c-4ac4-4704-b7b0-95b542240475	CWX47	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.795569-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.800059-07	f
3bd67b36-e968-4875-95ce-1eb8a711d47a	CWX48	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.801109-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.804906-07	f
cb6e1bb2-13ab-4245-b0bb-398bfdae6ac6	CWX92	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:08.763819-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:08.769863-07	f
9104381b-76f1-478f-bf41-baaa240f776c	CWX49	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.806355-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.810561-07	f
b80d221a-eb76-431b-bc14-78ef538ad39e	CWX50	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.811933-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.816008-07	f
62874067-fe2e-4ba5-a527-2c53ef5166c8	CWX51	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.817355-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.821698-07	f
ff19a148-c912-4b79-a416-d77eb17f4c8c	CWX52	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.822872-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.826758-07	f
981a109f-2b0d-464c-ba8c-3ab530503986	CWX53	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.828239-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.832557-07	f
371656ae-8d28-434f-86a7-21cc03c87d92	CWX54	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.833692-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.837588-07	f
7554c1d0-df29-4789-b007-c08a1d9f22e1	CWX55	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.838828-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.842641-07	f
42ed148e-fc5f-4bfc-a277-2bfdc42d1941	CWX56	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:02.843892-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:02.847692-07	f
de59347a-8518-43d1-8141-7314cb217b54	MRX29	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_1", "probe_seq_1": "TCACCTAGAGCCTAATCT", "probe_seq_2": "AGCGGCACGTGGCGCAAG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:02.851007-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:02.851007-07	f
70e6974a-4889-42b3-aa50-1dc5162a753a	MRX30	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_2", "probe_seq_1": "CGATTCTCCTCTTGAACA", "probe_seq_2": "GTGTTCAGCACTAAGTAA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:02.852684-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:02.852684-07	f
6bb5fc9e-205d-4cf1-aaa7-b82e122bb446	MRX31	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_3", "probe_seq_1": "CAGCGGCGTCGGGATAAA", "probe_seq_2": "CGCAACGACTATCGCACG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:02.857377-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:02.857377-07	f
da26ed31-1065-4604-8b23-091ce21c1a2f	MRX38	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_10", "probe_seq_1": "ACCAGGACCGAGGGCGAG", "probe_seq_2": "GGTGACGACCCGGCCCCC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:02.891032-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:02.891032-07	f
d6222668-162e-4c2d-b085-2ebbc2e71828	MRX39	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_11", "probe_seq_1": "GCAACAAATCCCAATACC", "probe_seq_2": "CTACGCTCGTACCTCAGT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:02.895251-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:02.895251-07	f
67777ee6-6878-4749-b64d-57f9eaa72595	MRX40	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_12", "probe_seq_1": "AGGCACGGGCCGTGCGGG", "probe_seq_2": "ATTGCATTTACATTGCAC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:02.900199-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:02.900199-07	f
d57dd303-b4c6-4710-b3ef-6de6113f5106	MRX41	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_13", "probe_seq_1": "TTGGGGCAATTCGCTTGT", "probe_seq_2": "TGCTACTAAAATATTATT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:02.904584-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:02.904584-07	f
d9c14341-79eb-4fc7-b30f-5c19d96ca9c2	MRX42	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_14", "probe_seq_1": "GAATAGCAGCTCATTCGT", "probe_seq_2": "GCATTGATCTTGAAATGA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:02.909067-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:02.909067-07	f
7352677a-92c8-4835-901a-525861538e8e	MRX43	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_15", "probe_seq_1": "ACGTCCGCAGGGGTTCAA", "probe_seq_2": "ACAAGCCGCCACGCACCG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:02.913536-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:02.913536-07	f
a330becf-ac06-416e-957d-ac2852be9ea0	MRX44	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_16", "probe_seq_1": "GGGAAGGAGTTCAATTTC", "probe_seq_2": "TAACATGTCAGAGGCTTA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:02.918242-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:02.918242-07	f
a887fcca-84b8-4a78-8bfb-c400a989c31a	MRX45	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_17", "probe_seq_1": "TGATGTCAGCACTTAAGC", "probe_seq_2": "AGGCGTGGCGCCGGGTGG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:02.922626-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:02.922626-07	f
5d5e68be-5a3d-4089-a30b-5d7bbbe4a2a7	MRX46	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_18", "probe_seq_1": "GGAAGAGAAAGCGCAATC", "probe_seq_2": "CTCCGATGCCTTGTAAGG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:02.926982-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:02.926982-07	f
ccc52ad7-c722-4a84-b999-5f97f4d53ffb	MRX47	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_19", "probe_seq_1": "CATGGCCGATATTGGGCG", "probe_seq_2": "CCGTGGTCGGGGTGACGT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:02.931525-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:02.931525-07	f
1e7602c7-c553-4988-8eb4-014fe5e4c3ef	MRX48	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_20", "probe_seq_1": "TTCCGTTACGTATCCCGG", "probe_seq_2": "CGAAGTGCATCAAGTCGT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:02.935814-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:02.935814-07	f
db082c31-af48-4101-96af-118560adac8e	MRX49	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_21", "probe_seq_1": "CAACTTGGGCCCTCCAGC", "probe_seq_2": "ATGCATCTGTGGCGCCTC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:02.940253-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:02.940253-07	f
77754c02-be5c-43cb-b291-8ebdeefc4c2b	MRX50	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_22", "probe_seq_1": "TAGAAAGAGGTCCGAGGC", "probe_seq_2": "GATGCTTCGTAACGTGCT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:02.944649-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:02.944649-07	f
8a472e60-5250-43b5-914b-d1bfe5a47927	MRX51	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_23", "probe_seq_1": "GTCTTTATAGTCTCTTAT", "probe_seq_2": "GTTTCAATTAGCGGTATA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:02.94904-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:02.94904-07	f
d5438a6a-5b90-40be-a73d-c418e343a0d5	MRX52	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_24", "probe_seq_1": "GGCTTGATGTTGATGCCA", "probe_seq_2": "GGAAGCCGAGTGTGTTCA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:02.953494-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:02.953494-07	f
f1bd3edd-8879-473c-a1ad-d70687254885	MRX53	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_25", "probe_seq_1": "GGGGTAGGGCGTTTCAGA", "probe_seq_2": "GGGAATGTGTATACACGC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:02.957776-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:02.957776-07	f
a067d8df-fe65-4937-a679-ce28eac8f132	MRX54	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_26", "probe_seq_1": "TTTGATAGTGACTTGGGG", "probe_seq_2": "TCTCCAGCCAGGAAGTAG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:02.962054-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:02.962054-07	f
8070ee33-11a2-441b-8f99-11df1436e5af	MRX55	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_27", "probe_seq_1": "AGTGTACGCTAAGCAAAT", "probe_seq_2": "ATCGTTGATAGCCTCGGT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:02.966483-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:02.966483-07	f
a2e7703b-a781-4d45-9775-6e3812f6f5c7	MRX56	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_28", "probe_seq_1": "GGAGTCCTTCTGGCACAC", "probe_seq_2": "TGTTCACATGGTAAGTGC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:02.978671-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:02.978671-07	f
edd473cb-2593-4ef4-a328-bdbcbae49d87	MCX1	control:giab-HG002:1.0	{"cogs": {"cost": "5.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "giab NA24385", "volume": "", "comments": "", "lab_code": "", "volume_units": ""}, "description": "giab NA24385", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	control	content_instance	content	giab-HG002	2025-07-03 08:21:03.013724-07	1.0	ready	f	50d08295-a50b-422e-9c2a-6c1791c08e8f	2025-07-03 08:21:03.013724-07	f
d4b6618e-692a-4952-b7f5-164c2c7e45a2	CX3	tube:tube-generic-10ml:1.0	{"cogs": {"cost": "1.50", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "giab NA24385", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "Generic 10ml tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	tube	container_instance	container	tube-generic-10ml	2025-07-03 08:21:02.98976-07	1.0	ready	f	d1533fbe-9bf0-46f2-abff-3c74cc80c744	2025-07-03 08:21:03.017866-07	f
800a223f-1a21-493f-a97c-5462854e77dd	MRX65	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:03.277032-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:03.277032-07	f
bb6bc18e-109f-4955-a46a-b909b4878979	MRX66	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:03.282022-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:03.282022-07	f
92d63b59-0d46-4dcc-aadf-ee1ed83a3979	MRX67	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:03.287395-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:03.287395-07	f
ed22c26f-07be-4ee0-8510-9575fec8e997	MRX68	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:03.293428-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:03.293428-07	f
2e08566f-8372-48e0-a356-c2aef6cd92d1	MRX69	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:03.299631-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:03.299631-07	f
26562587-4869-4ae7-82f8-06ce603df7f3	MCX2	control:synthetic-cfdna:1.0	{"cogs": {"cost": "3.1", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "s ynthetic-cfdna", "volume": "", "comments": "", "lab_code": "", "volume_units": ""}, "description": "synthetic-cfdna", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	control	content_instance	content	synthetic-cfdna	2025-07-03 08:21:03.027568-07	1.0	ready	f	4765a344-581e-4bc6-9c70-cca99e85637d	2025-07-03 08:21:03.027568-07	f
14f6d115-81d1-4a24-b8d9-a675f97a5251	CX4	tube:tube-10ml-glass:1.0	{"cogs": {"cost": "4.70", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "s ynthetic-cfdna", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "10ml glass tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	tube	container_instance	container	tube-10ml-glass	2025-07-03 08:21:03.024427-07	1.0	ready	f	7651bf71-c917-4c36-bd28-ded3fa87e1c2	2025-07-03 08:21:03.030939-07	f
06518ad4-6d64-4cd9-ad67-13c330248de9	MRX87	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_3", "probe_seq_1": "TAACACTCAATCCCTTTC", "probe_seq_2": "CCAGGTGATATGTACCCT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:08.907676-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:08.907676-07	f
8571eed1-4c25-4b2d-9f40-2e3bf265e6b3	MCX3	control:water-ntc:1.0	{"cogs": {"cost": "0.10", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "NTC - water", "volume": "", "comments": "", "lab_code": "", "volume_units": ""}, "description": "NTC - water", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	control	content_instance	content	water-ntc	2025-07-03 08:21:03.038-07	1.0	ready	f	70d42dd6-b1f4-4293-ba62-97c902f6ffb0	2025-07-03 08:21:03.038-07	f
ed15ff68-5a5a-4bf8-99aa-b4619eb64fd1	CX5	tube:tube-eppi-1.5ml:1.0	{"cogs": {"cost": "1.70", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "NTC - water", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "1.5ml eppi tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	tube	container_instance	container	tube-eppi-1.5ml	2025-07-03 08:21:03.035534-07	1.0	ready	f	ceca7fef-9b77-433d-9420-8d8a15cf1147	2025-07-03 08:21:03.040775-07	f
c5f7422b-9ae1-4496-a645-28e6decfdcd9	CX6	plate:fixed-plate-24:1.0	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	plate	container_instance	container	fixed-plate-24	2025-07-03 08:21:03.044814-07	1.0	ready	f	6bae2028-8873-4717-9e10-821a62bcd9f0	2025-07-03 08:21:03.044814-07	f
95a96de5-0287-4548-a6bc-5427e9da5086	MRX208	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.829526-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.829526-07	f
da03904f-c644-4f7b-a603-11082db6655d	CWX93	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:08.771673-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:08.777739-07	f
a20f9cb2-4002-4a69-bff6-d52924a67a7b	CWX57	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:03.049675-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:03.053177-07	f
1a35323d-f111-442b-958f-716b90b5d1f2	CWX58	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:03.054349-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:03.063623-07	f
9b76a542-1cdd-4c04-bb17-9fd9766d78cd	CWX59	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:03.068927-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:03.076386-07	f
e025d800-f0c0-4537-b1b1-cf50c0b976af	CWX60	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:03.0776-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:03.081513-07	f
a0398a18-4ab8-469e-9010-55fda1de05b0	CWX229	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:15.723773-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:15.732297-07	f
9fb53135-baa2-4a38-895f-c36ea22b3b11	CWX61	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:03.082741-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:03.086591-07	f
4b2787c8-670e-45a3-93e9-7c32160e7800	CWX62	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:03.087697-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:03.091585-07	f
c188bf5e-cf34-431e-9949-ce0743474003	CWX63	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:03.093038-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:03.097542-07	f
bc77f3ee-c50b-43eb-8f41-d9fec67a5da5	CWX94	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:08.779117-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:08.783511-07	f
0c8feadc-bdae-4f20-b475-839e34f912c2	CWX64	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:03.09916-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:03.103438-07	f
96d3bf7e-a12d-4167-b3eb-6f30a69c3b24	CWX65	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:03.104822-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:03.108766-07	f
6e0aa5d8-9855-45b9-8da1-1ae98a7dc57c	CWX66	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:03.110199-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:03.114005-07	f
42c4b4ad-e85c-4b20-9276-a9f98084fd5c	CWX67	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:03.115153-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:03.119129-07	f
723fff3b-fa28-4af0-82a4-c5228c05a429	CWX68	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:03.120564-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:03.124341-07	f
fd2c7148-44ec-48bf-abe8-38d5fb4c9e93	CWX69	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:03.125565-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:03.130106-07	f
780f794b-7f49-4805-801b-912cabfd265e	CWX95	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:08.78484-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:08.789232-07	f
2b67f9bd-b749-40d0-8189-51ed2c0da68e	CWX70	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:03.138604-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:03.14646-07	f
467f92c0-58ff-47fa-85a6-e9268901935e	CWX71	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:03.147822-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:03.152282-07	f
ecc31348-3e43-46dc-b0d9-c81f9ba47938	CWX72	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:03.153581-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:03.157767-07	f
4ba3e341-5121-4322-bc64-7b192e496768	CWX230	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:15.734602-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:15.743868-07	f
f8ac8009-49e0-4a5d-a1d2-6dd107e496e7	CWX73	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:03.15928-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:03.163663-07	f
c30fd0fb-b696-4632-895b-ac4e0b089522	CWX74	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:03.165817-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:03.170429-07	f
18266d0e-48c1-4a3c-a530-bd7d02a60043	CWX75	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:03.171921-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:03.175836-07	f
06fc3873-3a8c-404d-a4af-b8c7116d3213	CWX96	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:08.790553-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:08.796577-07	f
ffb4a762-b5dc-4494-b040-eb49d98a3e89	CWX76	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:03.177099-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:03.180898-07	f
edc810e0-165b-4d5d-8e0e-aa000572600d	CWX77	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:03.182487-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:03.186572-07	f
de6985c5-652b-4542-b1e5-a5ff86fe3d52	CWX78	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:03.187621-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:03.191325-07	f
3f1c2a4d-7892-4b46-9059-f884aad34e80	CWX79	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:03.192956-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:03.197291-07	f
2f43dd3a-7c4c-4300-a06a-1af3bbf880be	CWX80	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:03.198703-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:03.202601-07	f
88dc0886-2c10-464c-9b20-0fe485c4223a	CWX81	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:03.203859-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:03.208327-07	f
85fa4716-fb10-4279-9999-ebd9c08f6b20	CWX97	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:08.799621-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:08.806605-07	f
479bb3b4-e408-4cc2-ab2f-1d56b69ace6e	CWX82	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:03.209531-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:03.213821-07	f
9039cfde-66bb-4188-b1ea-e6075bb37449	CWX83	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:03.220289-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:03.227158-07	f
9409b6fd-e01b-43b1-ab67-77fe15748f54	CWX84	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:21:03.228423-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:21:03.232536-07	f
42207d28-a267-4548-b1af-b7e93f4982d0	MRX57	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:03.235399-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:03.235399-07	f
9d3d6ea3-0299-4b7b-aa33-f361910a58e3	MRX58	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:03.23909-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:03.23909-07	f
e70c4c91-bcf8-487a-8cc8-45817cd72396	MRX59	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:03.244636-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:03.244636-07	f
ed1d3d5a-e234-4ee2-8c5f-b88300456d8b	MRX60	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:03.250346-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:03.250346-07	f
6fd97ea6-93e2-4184-a45d-46f3431ca109	MRX61	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:03.25557-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:03.25557-07	f
44ef5e8f-8ff4-402e-9ad6-300221cc3de9	MRX62	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:03.260864-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:03.260864-07	f
70dc8c52-c634-4743-95a3-f7bc39a9cc12	MRX63	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:03.266682-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:03.266682-07	f
6d743c90-ecc6-4b3b-bcab-30278ece70c4	MRX64	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:03.271948-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:03.271948-07	f
e1841886-6ac7-4de7-8c4e-1bb17dd37908	MRX70	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:03.305291-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:03.305291-07	f
071b553d-fd1a-432b-a258-99de30fb97d9	MRX71	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:03.311046-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:03.311046-07	f
6be732d5-ba40-4cd2-95a3-ebbb71ea380b	MRX72	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:03.316643-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:03.316643-07	f
02d4fd7d-df49-48a7-8b80-873d463d7037	MRX73	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:03.3223-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:03.3223-07	f
3f2989ce-c0d1-4b52-9140-b9b19d170768	MRX74	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:03.328263-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:03.328263-07	f
8bff6ef8-6bc4-4cc5-bf48-8acc081d6f99	MRX75	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:03.334026-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:03.334026-07	f
86c7fbcf-32e3-4270-8884-b1617558a29e	MRX76	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:03.341416-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:03.341416-07	f
e258c3a2-c387-44c9-93c7-60868c8046c8	MRX77	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:03.34811-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:03.34811-07	f
82dac929-c652-47f9-ad44-7a7ac8fade75	MRX78	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:03.353852-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:03.353852-07	f
ba650613-ec31-45d0-8935-5b94a1dc3079	MRX79	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:03.359328-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:03.359328-07	f
4d800139-262b-4779-ba70-ab7a011c53cb	MRX80	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:03.365447-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:03.365447-07	f
ca7c8e06-c44f-414b-a3f4-7f3abdc7d153	MRX81	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:03.371137-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:03.371137-07	f
c7de2b92-a075-4026-86ca-1f491615948c	MRX82	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:03.376844-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:03.376844-07	f
f3ca1576-6ec8-4e97-835f-dbc25be9d779	MRX83	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:03.382514-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:03.382514-07	f
e526049b-2502-40d3-92d6-00841b98b660	MRX84	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:21:03.388193-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:21:03.388193-07	f
6bb43301-6e9a-4fad-9cd4-8c43db0a1159	CX7	rack:tube-rack-4-empty:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "4 position Rack", "comments": "", "lab_code": ""}, "description": "4 position Rack", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	rack	container_instance	container	tube-rack-4-empty	2025-07-03 08:21:03.393933-07	1.0	ready	f	d5ee3d31-ad0b-4780-a97d-b0c78bc74ac3	2025-07-03 08:21:03.393933-07	f
cb8e67ed-339b-4c03-b725-1a60d7ad9e54	MRX88	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_4", "probe_seq_1": "TCGGCTCATTAGAGGGGT", "probe_seq_2": "AGGGTTACCCGGGCGTTT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:08.912175-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:08.912175-07	f
b2421416-938d-46ed-bcc4-1aca83a2e2ab	MRX89	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_5", "probe_seq_1": "CCCAGTAGTATAGGTCAG", "probe_seq_2": "TCTGGAGCCGCAGCAGTA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:08.916028-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:08.916028-07	f
7ea54466-0a39-4279-becc-c379d09d80cc	MRX90	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_6", "probe_seq_1": "AACTAGACATTTCGGCGC", "probe_seq_2": "GTATCGGCGCCACTCACA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:08.92017-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:08.92017-07	f
12f78a64-40a3-4753-bef0-11ad97568fa6	MRX120	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_8", "probe_seq_1": "AACTCCATAAGTACGCGG", "probe_seq_2": "GTAGCTCTACCCAACTGA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:09.20388-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:09.20388-07	f
b4b7ac21-3798-4e5d-8ed4-735cec860516	CWX98	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:08.808977-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:08.815741-07	f
c478dfca-9c72-4f15-93f1-fe3efd6274d1	CWX231	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:15.746269-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:15.755799-07	f
50a01be1-89ed-4814-93c6-de250293b4e6	CWX99	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:08.818645-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:08.823329-07	f
1fd00059-c3de-43e5-9cb4-7d4c273343ad	CWX100	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:08.824708-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:08.82929-07	f
39ac2d80-8f21-4823-8941-abfe02e40867	CX10	plate:fixed-plate-24:1.0	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	plate	container_instance	container	fixed-plate-24	2025-07-03 08:23:16.838138-07	1.0	ready	f	6bae2028-8873-4717-9e10-821a62bcd9f0	2025-07-03 08:23:16.838138-07	f
53c38748-0c95-4d1d-bb0c-83201d2d54db	CWX101	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:08.831455-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:08.836095-07	f
56bc2733-1be4-4f33-8c35-2b522209ae67	CWX102	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:08.837365-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:08.841445-07	f
72061fa5-3321-4a96-b64c-275d915fc9ee	CWX103	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:08.842566-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:08.84644-07	f
e6d570a4-5fcf-4887-b23c-31b36a64854c	CWX104	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:08.847831-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:08.851964-07	f
d4fbc620-8b0d-4f35-b5f0-0b2fe2f00571	CWX232	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:15.759711-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:15.768378-07	f
92d78938-91b6-4aee-a8b4-a74e51ad8b95	CWX105	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:08.853275-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:08.857328-07	f
5dff72b8-b5c4-4662-b097-b09c106854c6	CWX106	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:08.858591-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:08.862637-07	f
974b84e5-0ee2-4545-8ff0-2de08d139c6f	CWX107	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:08.863842-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:08.868073-07	f
32986026-4b6a-4a23-a81c-ae11fd925582	CWX108	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:08.869272-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:08.872983-07	f
8fd06f8b-3195-4a98-882f-49d1c9a4834b	CWX109	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:08.874087-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:08.877679-07	f
b7122674-3965-4b7f-b5be-c6b1c869e1c5	CWX110	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:08.878751-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:08.882174-07	f
de7745e0-b7ec-49e7-8436-e6f0152453b6	CWX111	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:08.883518-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:08.887717-07	f
78c3c9b5-ea50-4c51-8560-9929af6a202d	CWX112	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:08.8891-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:08.893199-07	f
4366985c-1c27-4e8e-8f17-5f2bb448f0d2	MRX85	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_1", "probe_seq_1": "AACGCCTAGTACGAAGTT", "probe_seq_2": "ACCCGCACACTATATAAC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:08.897574-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:08.897574-07	f
5753d5e2-98b7-44e0-88ce-6ee0096cf3ee	MRX86	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_2", "probe_seq_1": "CTGTTTTTTTGGTCTCTC", "probe_seq_2": "TTGGCTAAGTCCCGGGGA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:08.901124-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:08.901124-07	f
878e54ab-5d67-4d62-8602-5d9700906f19	MRX91	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_7", "probe_seq_1": "CGAGAGCCTATACAATTC", "probe_seq_2": "GCAGGCAGACCAGGCAAA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:08.924004-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:08.924004-07	f
ee71e870-facd-47db-a77a-0005c2199964	MRX92	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_8", "probe_seq_1": "GGGGGGGAAGGGGGACAG", "probe_seq_2": "ACACATCCGCAATACACT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:08.927791-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:08.927791-07	f
64e1cb24-3e88-4abb-807e-4717e3f32b73	MRX93	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_9", "probe_seq_1": "TTCATGGCCAACGTTGTC", "probe_seq_2": "TGCAAGCAGCCATTGATG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:08.93191-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:08.93191-07	f
968e67a1-9b12-4960-83a7-912ca28870e3	MRX94	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_10", "probe_seq_1": "TTTACAAAACCCTAGTTC", "probe_seq_2": "CGAGCGGTTTTGACAATA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:08.935993-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:08.935993-07	f
109cb0f0-2f2b-4a27-992d-38184b13d248	MRX95	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_11", "probe_seq_1": "CGTGATTGCAAGCGACTG", "probe_seq_2": "CAAGACGCATGGAGGGAT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:08.940342-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:08.940342-07	f
df54fb98-5e91-448f-81c4-c19586df73ee	MRX96	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_12", "probe_seq_1": "TCCCCACGTATAATGAAC", "probe_seq_2": "ACGTCTTCAAACATCCCT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:08.944735-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:08.944735-07	f
5d354a84-be84-403e-8777-7b9a2520761f	MRX97	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_13", "probe_seq_1": "CGTGTTAACGTAGCAAAG", "probe_seq_2": "CTCGAGACCTACCAGGAT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:08.94886-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:08.94886-07	f
94807e9b-79b5-43ee-928c-b1977a3f96f5	MRX98	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_14", "probe_seq_1": "TTATGACAAGGCCGGCCG", "probe_seq_2": "GCTCGTCGAGTACAAGCA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:08.95349-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:08.95349-07	f
7de1ceef-160a-464e-92ca-2b68522b3eba	MRX99	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_15", "probe_seq_1": "AGTAATGGTCCGGTCCCA", "probe_seq_2": "TGAGAGCTGAAGCTCAAG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:08.957729-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:08.957729-07	f
733386da-f0db-4671-9ea4-5dda616d2631	MRX100	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_16", "probe_seq_1": "TAGGACGCATAAAAACGT", "probe_seq_2": "GAGGCAGAGCGAGGTCCT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:08.962449-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:08.962449-07	f
64c5fd73-d9fc-4c22-9f2e-1f91998a02c1	MRX101	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_17", "probe_seq_1": "AAGACGTGCCAACACTGG", "probe_seq_2": "CTAGGTAAGCTCCGGAAT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:08.967393-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:08.967393-07	f
148a355a-e554-4686-85f1-484bb92e12eb	MRX102	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_18", "probe_seq_1": "AGATTTGAACTCTTCGGG", "probe_seq_2": "CTCCGGAACAGTTACCCC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:08.971744-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:08.971744-07	f
8259854f-3495-44e0-8850-57466281e5a5	MRX103	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_19", "probe_seq_1": "ATGAAAGGGGAGTAAGGG", "probe_seq_2": "CAGGGTGCCAGCATGGGA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:08.976128-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:08.976128-07	f
a9098742-f68e-4d41-8ad5-58b225c762ec	MRX104	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_20", "probe_seq_1": "TGCCGCAGGGACACAGCC", "probe_seq_2": "CGCCATCGCGCGTGGTTA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:08.980265-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:08.980265-07	f
4da41248-557d-40ac-9a52-c0deaea86440	MRX105	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_21", "probe_seq_1": "GTGCGCACCTATTTTGAG", "probe_seq_2": "TATATGAAACTTGTAACG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:08.984738-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:08.984738-07	f
c02979bf-e445-4ee7-8505-67bf010865fd	MRX106	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_22", "probe_seq_1": "CGTCTCGCAGAACCCTGT", "probe_seq_2": "ACTGTCCCTTTCTCTACT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:08.989051-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:08.989051-07	f
431e9885-70e6-4269-9e55-5a6b6de2a8ea	MRX107	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_23", "probe_seq_1": "CCGAAGCGTATCAGTGTA", "probe_seq_2": "CTTATTTTTAGCTGTGAA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:08.993641-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:08.993641-07	f
859f2a12-8212-4697-893b-19aee54a3c15	MRX108	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_24", "probe_seq_1": "ACGAATTCTTAACGTTAT", "probe_seq_2": "ACAACAGGGTGGGGTCCA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:08.998181-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:08.998181-07	f
9fe5d8ab-c3a9-4850-b662-07d1349c1d97	MRX109	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_25", "probe_seq_1": "CTTTACTAGGATCGCAAA", "probe_seq_2": "CGCTAGCATGGCACGCGA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:09.002359-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:09.002359-07	f
901f56f9-59e7-4b78-80b9-6d6f62d6aed7	MRX110	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_26", "probe_seq_1": "GATGGCAATTGCAGTGTT", "probe_seq_2": "AACGGCGATCCGAAAGAC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:09.006676-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:09.006676-07	f
c64e172a-31c7-41d4-83e5-a592747c1107	MRX111	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_27", "probe_seq_1": "ACGCAGTATGGCGTCGTT", "probe_seq_2": "GACACTATTCACATGCAA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:09.011055-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:09.011055-07	f
173f906f-f360-4aa8-bf05-20fed68593d5	MRX112	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_28", "probe_seq_1": "AAAATCTAGGGTACCCTG", "probe_seq_2": "GTTCCGAGCACAAAAACA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:09.01544-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:09.01544-07	f
36a78f96-96bc-4162-9372-16d115b2e6c3	CX9	plate:fixed-plate-24:1.0	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	plate	container_instance	container	fixed-plate-24	2025-07-03 08:23:09.022787-07	1.0	ready	f	6bae2028-8873-4717-9e10-821a62bcd9f0	2025-07-03 08:23:09.022787-07	f
5c89cd1b-c0d0-4846-896a-ceca2f29fcb6	CWX233	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:15.771546-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:15.777077-07	f
b806e72c-8000-4615-b23f-b63a79d2da4d	CWX113	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:09.027734-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:09.030328-07	f
701b1431-fa54-4f76-9076-c445285f8ff5	CWX114	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:09.031758-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:09.036299-07	f
6a2dc8d6-7470-4baa-b4d8-cd0e20761c7d	CWX115	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:09.037395-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:09.041179-07	f
7f1ef106-8285-4641-bc6c-b2cb1e199946	CWX116	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:09.042644-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:09.04668-07	f
5fe3fa73-82ec-4a72-a25d-f4ca3f7575d8	CWX234	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:15.778732-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:15.784492-07	f
00b1f951-e684-487b-91fc-803626c26deb	CWX117	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:09.048349-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:09.052106-07	f
a7ce7f5a-f641-42ae-bb64-e32665f3383f	CWX118	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:09.053251-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:09.05703-07	f
04078517-815f-4032-a223-10dc61971326	CWX119	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:09.058174-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:09.062041-07	f
fb489163-90aa-4b36-83d6-71b946fb43f5	CWX120	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:09.063341-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:09.067566-07	f
0e06a2a9-31a7-4ab9-9074-9cc9ce191775	CWX121	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:09.068635-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:09.072453-07	f
f893cb99-8261-4d16-82b7-195a2985abb8	CWX122	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:09.073646-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:09.07746-07	f
015fbb50-97f2-409e-9b39-e5f548a21402	CWX235	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:15.786514-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:15.79204-07	f
842e7d48-9605-4321-956e-17b670298506	CWX123	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:09.078849-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:09.08265-07	f
b1519897-fb8e-4607-898f-3f0575c8a04a	CWX124	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:09.08386-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:09.087607-07	f
550d0d97-6fda-47b0-8e07-d820e912a145	CWX125	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:09.088818-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:09.092469-07	f
bd871a52-972e-4179-a5f8-2aacf0466302	CWX126	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:09.093617-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:09.09752-07	f
4ba333d5-a753-4026-9358-75385fec301a	CWX127	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:09.099273-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:09.1057-07	f
60b56382-b9de-4a9d-892b-a6e1bcae9e89	CWX128	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:09.10718-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:09.110903-07	f
427711ca-ab90-45b5-b311-8733db30d9ad	CWX236	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:15.793739-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:15.799421-07	f
a9314a45-c45f-4522-a0e9-9a99459a6924	CWX129	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:09.112412-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:09.11608-07	f
43f0aac8-c4da-48d5-ade3-02f63fdd00a5	CWX130	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:09.117223-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:09.121175-07	f
2d325520-3bd4-44b8-93e0-874881087ae6	CWX131	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:09.122373-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:09.126199-07	f
57556f94-c37c-41c8-88db-53e8cee975b6	CWX132	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:09.127525-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:09.1318-07	f
820ef252-cff7-4a46-a803-2dc8eee77103	CWX133	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:09.132921-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:09.136654-07	f
3b447fb4-8438-4485-bac4-1553e554d7d2	CWX134	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:09.137843-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:09.141751-07	f
95d4d6ba-7610-486b-a74c-d1ad8255621a	CWX237	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:15.801168-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:15.814182-07	f
c6928136-7211-428a-aa93-424927c53157	CWX135	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:09.142959-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:09.14677-07	f
5633367b-29d0-4a2c-bf49-950a922e81bb	CWX136	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:09.14792-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:09.151572-07	f
4b42200e-1ca5-4eb8-8947-8cf89ae11946	CWX137	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:09.152702-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:09.156632-07	f
fc4f5d57-c0c6-4d38-bfd6-d50850a8f148	CWX138	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:09.15779-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:09.161736-07	f
cff309ac-096c-4bc0-93dc-5adc7c8fe0fb	CWX139	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:09.163105-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:09.167363-07	f
ca5acce6-59f8-4465-83e5-3bb593d9bec3	CWX140	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:09.168511-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:09.172406-07	f
cb05244f-66db-44ed-9642-77f2976dd0dc	MRX113	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_1", "probe_seq_1": "CCATCTAGTTGCAGCGGA", "probe_seq_2": "CGAGTGTTGAGCTTGCCT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:09.175657-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:09.175657-07	f
a3c5b2a8-b081-4788-ac6a-f73a3e942278	MRX114	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_2", "probe_seq_1": "TCCGGGCACCAAGCCCTC", "probe_seq_2": "CTAGACAGGGTACCCCTG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:09.177524-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:09.177524-07	f
ed1337dc-5377-45a7-a724-e41e29c46a91	MRX115	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_3", "probe_seq_1": "ATGGCCTCTGCCCCGGAC", "probe_seq_2": "TAGTGCAGCTACTCAGGC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:09.181932-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:09.181932-07	f
dd1014f2-4cd8-4e15-b4ce-5095401e8759	MRX116	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_4", "probe_seq_1": "ACAGCTATGGAGAAGATT", "probe_seq_2": "CCTAGGAAGTTTCGCCAG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:09.186355-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:09.186355-07	f
170ab794-eb07-48e0-b1c0-f87d77d20590	MRX117	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_5", "probe_seq_1": "AGTTTTCAAAGCTCAAAC", "probe_seq_2": "ATACGGCCCTCGGGGTTC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:09.190707-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:09.190707-07	f
7316e0cf-c132-4afb-aef5-8bcf84ab93df	MRX118	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_6", "probe_seq_1": "ACCTGCTATAATGAGACG", "probe_seq_2": "CGCGGGGGCATAAGCAAC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:09.195105-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:09.195105-07	f
8a6828b1-67d5-4a72-a00e-70795c41dca6	MRX119	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_7", "probe_seq_1": "GATTATCCATACTAGTGT", "probe_seq_2": "AAGATCCAGTGTACTACA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:09.199605-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:09.199605-07	f
239cf4e6-27d8-4b54-ad21-a528a3c52252	MRX121	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_9", "probe_seq_1": "GAGGAGGCGGCGGTATTT", "probe_seq_2": "CCTCCTGGGAAGGGATCA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:09.208299-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:09.208299-07	f
512ab731-ec11-4e70-ace2-225ccd1c827c	MRX122	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_10", "probe_seq_1": "AGATGGGTGCACAGTCCC", "probe_seq_2": "GCCCTTCTACTAGGGGGG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:09.212744-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:09.212744-07	f
60bb2ea0-c3cb-442c-8ced-8495f8693f6f	MRX123	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_11", "probe_seq_1": "TATTCGAGTTTCCTTACA", "probe_seq_2": "ATTTTGCTTTGGGGATAC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:09.217046-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:09.217046-07	f
cb072692-f302-4e41-9830-dc8d5c7d4289	MRX124	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_12", "probe_seq_1": "ATCTTTCAATATCCACAT", "probe_seq_2": "CGATATCTGTCCATGAGT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:09.221322-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:09.221322-07	f
db516863-3a58-4e48-bdea-253df93b7eba	MRX125	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_13", "probe_seq_1": "TCCGTCCCCGTTACCACG", "probe_seq_2": "GGCGTTCCTGTAGCGTTA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:09.225846-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:09.225846-07	f
a6f37e7d-af69-46c4-a585-db6bfa69c506	MRX126	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_14", "probe_seq_1": "ACCGGACCTATGCGACAT", "probe_seq_2": "GGGCCTTTCCACGGCGTT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:09.23021-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:09.23021-07	f
c75f8621-9c89-4d84-9938-ed98b1430c39	MRX127	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_15", "probe_seq_1": "GATTGACCACTACTCGGC", "probe_seq_2": "GGATTTCAGTGACAGTGG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:09.234677-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:09.234677-07	f
ad6951f6-0d8e-43e5-a965-d1ac3e2d0035	MRX128	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_16", "probe_seq_1": "ATAAGTCGTTGCTTCGAC", "probe_seq_2": "GAGAGACGATTCACACGC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:09.239284-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:09.239284-07	f
558d4ccc-51af-41e5-a221-6a4d6de4e87e	MRX129	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_17", "probe_seq_1": "TAACGTTTCCCACCGGCA", "probe_seq_2": "CGCCGGACCCTCTGTTAA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:09.243663-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:09.243663-07	f
3e0d0af1-1210-4cab-84af-c3f9e045953f	MRX130	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_18", "probe_seq_1": "GTTGATTTTGACACGTTG", "probe_seq_2": "TCCCAACTTAGCGTCAGC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:09.248075-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:09.248075-07	f
eec83a36-273d-4e8d-a349-5c5abc492832	MRX131	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_19", "probe_seq_1": "TAAAGGGTACCATCCATG", "probe_seq_2": "GGATTCTGCGATATTCGA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:09.252411-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:09.252411-07	f
502f963c-82a7-45c1-be2b-d96b35f3bdf2	MRX132	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_20", "probe_seq_1": "GCCGGTAGCTGAGAACGG", "probe_seq_2": "AGCTGCGGACGAGACTCT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:09.256823-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:09.256823-07	f
00b8f24b-52a1-4993-b78a-1f8de01c1df5	MRX133	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_21", "probe_seq_1": "CCGTCTAACGAGCTCTTA", "probe_seq_2": "CCAAGCCGCGACTCGCCT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:09.261204-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:09.261204-07	f
5f74720c-3a0b-426f-981d-9c1db1d9f0cd	MRX134	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_22", "probe_seq_1": "TGCCGAGACTCCAGTAAG", "probe_seq_2": "TGACTGGGGTCCCAGAGC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:09.265782-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:09.265782-07	f
1f7b5d3a-40fc-4878-bcd6-78a9cef9ae5f	MRX135	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_23", "probe_seq_1": "CCCTTCATCCTTCGAAAA", "probe_seq_2": "TGCCGCTCCTTCAACGGG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:09.269957-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:09.269957-07	f
33903f7e-9f8c-4b25-aae5-d2273f155189	MRX136	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_24", "probe_seq_1": "CCGATAAGCCTGACGGCA", "probe_seq_2": "AAGGGGCAGCCTGTGACG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:09.274535-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:09.274535-07	f
b94ca2e5-3160-4c4d-ba63-e9b1fd3395e0	MRX137	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_25", "probe_seq_1": "CGTAAGAAAGAATAAGGC", "probe_seq_2": "AAATGCCCGTTTTTATCG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:09.278964-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:09.278964-07	f
88692020-71a2-4eda-89bc-ebb4cf452fed	MRX138	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_26", "probe_seq_1": "ATTGGATCACCAGCGAGA", "probe_seq_2": "TCTTTCTCTTCGAATTTA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:09.283551-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:09.283551-07	f
b26f5000-d3ef-4d72-be05-67228321dee4	MRX139	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_27", "probe_seq_1": "CACTGAGCGGGCAACCAC", "probe_seq_2": "GGGTTACTAAAGGGGATC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:09.287948-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:09.287948-07	f
46df43ab-6fe5-4e67-8f96-0ea193e774c6	MRX140	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_28", "probe_seq_1": "GCGAATGAGAGGCCAGAG", "probe_seq_2": "GAATATAGTGCTGCCTTC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:09.292413-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:09.292413-07	f
61bfac74-3f5c-474c-8b7a-6fe48898e386	CWX238	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:15.816498-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:15.82291-07	f
a1f6ccdc-9ee5-402b-a93d-533b39212bee	CWX239	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:15.824383-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:15.833736-07	f
f576931a-14bf-4549-bdad-cfbb1c5ad826	MRX256	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_4", "probe_seq_1": "GCATGCTTTCGGTCTCAT", "probe_seq_2": "TCAGCAACTCCCCCGACA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.289351-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.289351-07	f
f5c9a461-3037-48a9-ab53-74984ace3aec	MRX257	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_5", "probe_seq_1": "CCTACTGGGCAAATCTAT", "probe_seq_2": "GCTATTGTATCTAAGTGA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.29568-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.29568-07	f
103f84a1-a056-414a-b9a1-07ead9a853e6	MRX258	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_6", "probe_seq_1": "GTTACCGCTCACGTAGTT", "probe_seq_2": "CAGTAACTAGTGGCAGAG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.301211-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.301211-07	f
db19c6e5-98cc-4eba-9311-7d9ac31c3cf2	CWX141	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:16.850501-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:16.855209-07	f
b00c40c2-0afb-46fa-b628-daffa4dcd1c7	CWX142	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:16.858263-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:16.863962-07	f
662cdf0b-bf33-490a-997c-2ac9121bb7b9	CWX143	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:16.865432-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:16.871192-07	f
b3bb08ee-3efd-406b-ac7f-0959a69c2ff5	CWX144	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:16.872345-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:16.876773-07	f
4ff5edda-1503-4a1b-aa81-44f0d001dede	CWX240	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:15.83564-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:15.841215-07	f
c0ab8a50-6008-48dc-b1ed-35d21741141b	CWX145	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:16.878044-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:16.882158-07	f
44535f96-b261-441e-8059-2f7c5632858d	CWX146	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:16.88381-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:16.888274-07	f
4d5db19c-2950-4351-8d87-fe365be8e7aa	CWX147	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:16.891679-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:16.897352-07	f
3aa39100-5672-4c10-a9f9-d0f60b0f437a	CWX148	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:16.898679-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:16.903472-07	f
382bee5c-5a83-4842-8385-eec79639cc66	CWX149	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:16.904772-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:16.909154-07	f
36808dc6-c5b6-423e-848d-ed03c8a4e4d4	CWX150	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:16.910378-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:16.914525-07	f
c60228c8-099b-47f1-a890-387453eec26b	CWX241	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:15.84285-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:15.848337-07	f
304ab3f3-404d-4038-b781-6e028a3aac54	CWX151	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:16.91574-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:16.919569-07	f
bad6ad8a-77c4-4d48-81a1-858893e9a864	CWX152	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:16.9209-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:16.925115-07	f
6a5c1eac-1cbf-4e8f-87dd-63ceaee4a906	CWX153	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:16.92652-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:16.931112-07	f
c45f4209-c43e-4e0b-a0c6-4d8a5ea4d202	CWX154	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:16.932457-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:16.937418-07	f
cfba7c0a-1491-40ab-9d5c-27d8c5ce367e	CWX155	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:16.938673-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:16.942878-07	f
92af61ad-da5e-4378-bf20-1d69de145551	CWX156	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:16.944108-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:16.948262-07	f
7833d79c-9c5d-4177-b4ed-b29293438cc4	CWX242	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:15.850004-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:15.855556-07	f
8119664c-7b20-4d9d-b788-7ddefbc086e4	CWX157	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:16.94952-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:16.953604-07	f
be75f19a-9204-45e4-ac52-c07157044eb9	CWX158	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:16.955038-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:16.959483-07	f
03fd4454-0700-4766-adf4-40cd1f0847a2	CWX159	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:16.96069-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:16.966525-07	f
f653b34e-100a-46a3-8ab7-9afaa9cd74d0	CWX160	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:16.968505-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:16.972804-07	f
db0c762c-c36b-42de-bc97-6976ba8c53be	CWX161	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:16.974229-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:16.980754-07	f
01b72489-6786-4b1f-a8b7-25b56770d48f	CWX162	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:16.982042-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:16.986258-07	f
85859176-cfcb-4ba9-8078-981f2c92c5f7	CWX243	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:15.857224-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:15.862723-07	f
70bfdd8e-9926-4052-a785-1555e295832a	CWX163	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:16.987527-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:16.991348-07	f
ed6bd8d5-9e73-419c-afd6-6399c05e75a3	CWX164	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:16.992526-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:16.996405-07	f
bbb4417b-7369-41c9-91b0-8fb0773fdab6	CWX165	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:16.997679-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.00214-07	f
cf2a5e5f-bd04-40fa-a03e-f3efecc404b3	CWX166	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.003419-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.007562-07	f
6077eb5c-7bd2-436b-85a4-f1ec9e26cc5c	CWX167	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.008874-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.012753-07	f
4218d37a-128b-45cf-b2ef-1a111112b524	CWX168	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.013944-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.017956-07	f
82f7d2fc-6a90-4574-a482-546dc5be34c5	MRX141	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_1", "probe_seq_1": "CGAGGGTTATCGCTCGAG", "probe_seq_2": "TGTAAGCTCCCCTATCAG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:17.021151-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:17.021151-07	f
c5d41a18-1176-4b9d-bfd7-e222b66b1dfc	MRX142	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_2", "probe_seq_1": "TAATAGGTGCTCAGCTAC", "probe_seq_2": "ACAAAACTGACATACTAG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:17.023102-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:17.023102-07	f
ba952458-8cff-4ff5-a9d2-0a6ab5405602	MRX143	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_3", "probe_seq_1": "ACGACCTAAACGAGAATA", "probe_seq_2": "CACCCCTTAGGCGATACC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:17.029585-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:17.029585-07	f
525493a1-c68d-4d24-a2ee-d51be1599878	MRX144	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_4", "probe_seq_1": "TACCCGAAATTGCTACTA", "probe_seq_2": "TGGAGAAGTTATGGATCC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:17.034913-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:17.034913-07	f
ef23526a-8c13-4561-b742-e38af3f43e6d	MRX145	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_5", "probe_seq_1": "ACGAAAACAAACCAGTGA", "probe_seq_2": "TCTTACTGAGGCCCCGCA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:17.039681-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:17.039681-07	f
a6a0f6f1-6a24-475e-919b-c0fbe85a297c	MRX146	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_6", "probe_seq_1": "TAAGATCTTCAGCTAAGC", "probe_seq_2": "CAAAACATGCGTTAGCAA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:17.044588-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:17.044588-07	f
7ad6a870-8313-4efb-ba81-5f4fb0ab06c3	MRX147	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_7", "probe_seq_1": "ATGATTAACAGGCCCCAA", "probe_seq_2": "CGGACCCAGACAAGGTTT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:17.049301-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:17.049301-07	f
2fd021db-8cba-4e2b-9550-c0f078ed5ca5	MRX148	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_8", "probe_seq_1": "TCCGTACCTCCCTATAAG", "probe_seq_2": "GAAGCATTTGCATCTATG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:17.053597-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:17.053597-07	f
7c504afb-63e5-4f3e-9684-d7a685a74f33	MRX149	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_9", "probe_seq_1": "CTAGTGCCTGGTCAATCG", "probe_seq_2": "CCTGAGCGGCGACAGTCG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:17.059703-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:17.059703-07	f
3c9ca22e-fa5e-4304-916b-d9b14224b0da	MRX150	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_10", "probe_seq_1": "TATCTTGTTTACACCCAT", "probe_seq_2": "CTGCATGGACTCAACAGC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:17.067761-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:17.067761-07	f
47982770-15aa-46c6-bcb2-2dcfe0cf9fde	MRX151	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_11", "probe_seq_1": "GACCGGTGACCAGCGCAT", "probe_seq_2": "AGAAAATTAATTGTCTAA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:17.072967-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:17.072967-07	f
864663d1-dd5c-49ea-80c5-bcc1afaf6049	MRX152	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_12", "probe_seq_1": "CCCTCCCTAAGGCACCCT", "probe_seq_2": "GACTGCCACGCCCATTCA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:17.077892-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:17.077892-07	f
da2ddfd1-f611-4c8f-b85b-1c5e8f4d1a3f	MRX153	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_13", "probe_seq_1": "AGGGGATTTAGTCGAACA", "probe_seq_2": "TAATTTAGCGAACTTATA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:17.082628-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:17.082628-07	f
7aff050c-94f4-4a86-9bd3-e5ce0b2bcb7b	MRX154	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_14", "probe_seq_1": "GAGCCCCGTCTACTCGAT", "probe_seq_2": "GTACTAGGCAGGACCCGC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:17.087322-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:17.087322-07	f
e0a2ec5b-c727-46ff-8cfd-541ded1a37da	MRX155	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_15", "probe_seq_1": "GACTCGAACACGGCCCAA", "probe_seq_2": "TTTGAACGAGTTGATTTC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:17.091748-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:17.091748-07	f
461bb485-e99a-4c28-8499-92794e89166b	MRX156	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_16", "probe_seq_1": "TCTTCCAGATTAGCCTCC", "probe_seq_2": "GGTCCTACGGCCATGCCC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:17.096479-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:17.096479-07	f
952fb2a0-50f8-426f-baaa-95d86e9af105	MRX157	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_17", "probe_seq_1": "ACCGCATTCCCATTCGAA", "probe_seq_2": "CGTATAGCTCCCTCTAGT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:17.101187-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:17.101187-07	f
cb217410-9c27-445b-82c2-03e605c21be5	MRX158	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_18", "probe_seq_1": "TCCAGACGCCAGGTCGCC", "probe_seq_2": "AGTTACGATGCATGGCAC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:17.105736-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:17.105736-07	f
c7c571ec-1754-432e-b4e7-a8b18a0dc670	MRX159	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_19", "probe_seq_1": "GTTTGACTAATACAAATG", "probe_seq_2": "GCCTTAAGCCATTAGAAC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:17.110472-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:17.110472-07	f
e6c30471-04b1-41ce-a0f9-dea3962bb40f	MRX160	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_20", "probe_seq_1": "GAATGTGGTCGACATAAT", "probe_seq_2": "TCGTGAAATCGGCAGTGA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:17.115343-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:17.115343-07	f
61bd9a3a-9f50-40fa-bcb0-51a811d28a48	MRX161	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_21", "probe_seq_1": "CGACCAGTCCCGTCGGTA", "probe_seq_2": "ATTGCCGACGTCACACAC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:17.120297-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:17.120297-07	f
5b503e9f-0587-4d68-a122-eee644adffb3	MRX162	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_22", "probe_seq_1": "CCGCGTAGCGCTTTTGCG", "probe_seq_2": "CACCTGGACGAAGACAGA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:17.124922-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:17.124922-07	f
11bf48ed-61ca-4a6f-ad36-76f4f2eb9f55	MRX163	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_23", "probe_seq_1": "TCCACTACTGCATTGTAG", "probe_seq_2": "AATCCGCTTCTAATGGAC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:17.129883-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:17.129883-07	f
070fcb47-fb22-499a-afa0-e379811d58c6	MRX164	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_24", "probe_seq_1": "CCGCGTGCACTTAACTAG", "probe_seq_2": "CCTTCAAGCAATGGTTAC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:17.134831-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:17.134831-07	f
1ad3b52b-32cc-4ab5-a2b4-deb1c2a81bf2	MRX165	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_25", "probe_seq_1": "ATTTCGTTAACTTCCTTG", "probe_seq_2": "CTGAAATGTGCCGAAATA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:17.141995-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:17.141995-07	f
44cbde60-02c7-44d2-9032-fe68359e4f02	MRX166	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_26", "probe_seq_1": "CTGGATCACGACCGAGAC", "probe_seq_2": "TCACTTGGATAGCCCGGG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:17.147896-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:17.147896-07	f
0c7c36b7-7848-4205-af9c-ce41e543fd35	MRX167	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_27", "probe_seq_1": "TAACGGCGCGCGTGATGT", "probe_seq_2": "AACGAGCGAGCAACCATG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:17.152976-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:17.152976-07	f
35697e67-053e-4fbb-ac45-2223813e0a10	MRX168	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_28", "probe_seq_1": "CGAGTGGTAACGGCAGTT", "probe_seq_2": "GTAAGGAGGTGTGGTCAG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:23:17.157502-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:23:17.157502-07	f
1852629f-beb7-4f69-98e1-971a34af9928	CX11	plate:fixed-plate-24:1.0	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	plate	container_instance	container	fixed-plate-24	2025-07-03 08:23:17.165033-07	1.0	ready	f	6bae2028-8873-4717-9e10-821a62bcd9f0	2025-07-03 08:23:17.165033-07	f
fca6c1f5-1e60-4b0e-b8dc-de2108eb584c	CWX244	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:15.864295-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:15.870121-07	f
4b6dd84b-e2c0-413d-aa2a-2a520af200af	CWX169	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.169419-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.172522-07	f
e2728bf3-67bc-4811-bbe4-f0867231f6e0	CWX170	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.174531-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.179427-07	f
521b22bc-7d76-41ca-b735-17767c6be7bb	CWX171	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.180834-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.185151-07	f
7de76476-6e33-4d96-9a7b-333393b79eb4	CWX172	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.187614-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.191641-07	f
11d71ea0-89ac-4aad-89f4-68ba3f60525d	CWX245	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:15.872125-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:15.883029-07	f
c938574e-08aa-44ce-9f09-460240d55a12	CWX173	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.192961-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.197314-07	f
c43f8239-2b30-4153-8e0e-1ca8109d4847	CWX174	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.198868-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.203494-07	f
38d72d79-0c20-43bf-a047-265347aaf013	CWX175	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.20482-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.209266-07	f
ad4b62e4-5a8f-4bb6-963b-923da6668cf8	CWX176	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.210421-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.214462-07	f
45fef90b-f603-46b7-a07f-206b191d251e	CWX177	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.21569-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.219404-07	f
2d81e507-37d2-4131-b774-958d19466c06	CWX178	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.220545-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.224535-07	f
b5c6c063-303c-4bf1-a0b4-d952763bd8fc	CWX246	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:15.884775-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:15.891088-07	f
40c85e8c-792d-49af-930a-91f33e352dab	CWX179	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.225911-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.230668-07	f
7ed48914-f3a4-462b-baaa-845a9b3b31e6	CWX180	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.23526-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.241973-07	f
3802ef66-effb-4e9a-bcf9-d1ed8c84beba	CWX181	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.243204-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.247337-07	f
a9cc088c-0aec-4d5f-a5ab-fae4fe7f030f	CWX182	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.249395-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.254516-07	f
fc480673-f73e-44cb-b6d4-64df95e6b3d0	CWX183	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.255731-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.260046-07	f
8744e5ca-a013-46fc-99ac-b67f436d0827	CWX184	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.261154-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.265857-07	f
d7616aa4-1790-453e-97cd-801233b1bc1e	CWX247	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:15.894459-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:15.900439-07	f
1cf84a4a-f071-46fa-ae06-d39f5f8aa6df	CWX185	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.267378-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.271569-07	f
fb64119b-a9bf-4920-ac9f-42025d6486cd	CWX186	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.272936-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.277361-07	f
647cfc65-654d-4ff4-98b5-737645b3e930	CWX187	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.278636-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.28277-07	f
73d57912-9cb8-44c9-88a1-34e13b8b65fd	CWX188	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.28404-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.288198-07	f
8a96e5af-aa96-4369-9f19-a004222d92bf	CWX189	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.289374-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.29369-07	f
dd25a9e1-d8dc-4eea-be62-2d3d6a4a7829	CWX190	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.29493-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.2998-07	f
b9068f68-df90-4cc8-be11-f651be061b2a	CWX248	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:15.901925-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:15.909075-07	f
23e5c523-46c3-4f9b-a0ca-e30781d73d3c	CWX191	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.301168-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.305344-07	f
6d930317-d930-41e8-95c6-dccd9add54ff	CWX192	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.306639-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.310939-07	f
4ff0926c-88e1-40fd-90ac-dafe861eb746	CWX193	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.312181-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.316077-07	f
a72a0a3c-2e05-4aa3-bb8f-1d24163a15fa	CWX194	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.317516-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.322009-07	f
ebd9b442-c35d-4c7f-a9d3-514321045d51	CWX195	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.323224-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.327652-07	f
116a5243-bb5e-4e4c-b293-c07586d1d79e	CWX196	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.328821-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.334892-07	f
a7fdfb14-fc01-4349-a8ad-8de611a15b1c	MRX169	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_1", "probe_seq_1": "GCTACCCTACGTGCATAG", "probe_seq_2": "AGCTTGCAAAAAAGCAGA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.34078-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.34078-07	f
8205e9ec-903c-4f91-bc3b-bbdb81479279	MRX170	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_2", "probe_seq_1": "GAGAAGGCCACTATAACG", "probe_seq_2": "GGAACGGATTTAAAACCA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.342965-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.342965-07	f
4e28eeb7-6056-407f-92d0-53a349a90407	MRX171	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_3", "probe_seq_1": "GTATGTGTTAGAACCACC", "probe_seq_2": "GCACAGTCTCCTGAAGCT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.347665-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.347665-07	f
7a771f0c-8c4e-4184-91f0-be5b15eb4a1d	MRX172	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_4", "probe_seq_1": "GGTGTGTGGTATTCTTAA", "probe_seq_2": "CCTCCCAGAGCTCTTGGA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.354348-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.354348-07	f
b348c7b2-83c2-4c6e-ac63-25d711fed70a	MRX173	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_5", "probe_seq_1": "TGTCCGCCAGGATTTTTT", "probe_seq_2": "GGAAACCGACTAGTGCAT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.359712-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.359712-07	f
ad0b6013-0a58-4c0d-a2a2-acf5cf2fcf0b	MRX174	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_6", "probe_seq_1": "AATCAGGGAGCCAGACAT", "probe_seq_2": "AAGCCATCCGCTTAGTTA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.364937-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.364937-07	f
269038aa-810c-4401-9ae5-b2474d3afe38	MRX175	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_7", "probe_seq_1": "GAGCCATCGCGTAAGTCA", "probe_seq_2": "TCTGCCTAGGGTACCGTT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.36972-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.36972-07	f
538a6afc-a419-46c2-b3b1-79ac907c1b6f	MRX176	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_8", "probe_seq_1": "CGTTAACCGCACTGATGC", "probe_seq_2": "GCCACCTGTGACTGGACT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.374602-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.374602-07	f
376ecd05-c2eb-4b37-b19b-948ef80045b4	MRX177	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_9", "probe_seq_1": "GATTGGTGCGCTATTGTG", "probe_seq_2": "ACCCCTTTTGCCCATGAA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.379232-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.379232-07	f
6f2d6e6f-012e-46fc-82d9-8323bc889a37	MRX178	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_10", "probe_seq_1": "CCCGATCGCGAAGCACGA", "probe_seq_2": "GGAATCCCATTTAGGGGG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.383898-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.383898-07	f
009eef87-568f-4b42-8aff-4d78a5c238af	MRX179	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_11", "probe_seq_1": "CCGGTCTCCGTTCCTTGG", "probe_seq_2": "TAGCAACAGCAGGTATGA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.388733-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.388733-07	f
b2c3158a-af0b-43d2-a6fe-6b005335a909	MRX180	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_12", "probe_seq_1": "ATCGCTGAACGGAGTGAA", "probe_seq_2": "CGCCGCCGAGGCTAAAAG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.393467-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.393467-07	f
7ec4cacc-52bd-4a88-bfcd-3167baeab3a2	MRX181	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_13", "probe_seq_1": "AACGTGCCTACGGCGCCC", "probe_seq_2": "TCTGGCTCATATGGCTTT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.398828-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.398828-07	f
bda1c743-e436-4a9c-a9fc-71f22cda28c1	MRX182	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_14", "probe_seq_1": "CCCTATTGCTCCTACAAG", "probe_seq_2": "ACAATTATTGTAGAAGGC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.404929-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.404929-07	f
c17a9260-b36f-4a62-b076-e13de08d64cb	MRX183	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_15", "probe_seq_1": "AGGCAATCTAGAAGTCTG", "probe_seq_2": "GGCTGTCTTCGAACGTAA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.409505-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.409505-07	f
498ad847-9c74-4b5a-8725-0a56a3c84f94	MRX184	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_16", "probe_seq_1": "TCGCTCGAGGCGTCGTGA", "probe_seq_2": "TACTTATCTCTCAAGGTC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.416652-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.416652-07	f
9abbbeeb-2a9f-4a05-95f5-070fdc8c3483	MRX185	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_17", "probe_seq_1": "AATCAAATTTTTACGGCT", "probe_seq_2": "TCTCTTAGCTTAGGTGAC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.422874-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.422874-07	f
58b3a670-dedc-4045-b70d-d90479915351	MRX186	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_18", "probe_seq_1": "ATTGTTCTTTCATCCTAG", "probe_seq_2": "CCATCGTAGTACCTCCGG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.427323-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.427323-07	f
ef8398cc-3507-42a0-b8f5-99c0a9d033da	MRX187	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_19", "probe_seq_1": "TTGGATGTCTCAAATGGC", "probe_seq_2": "GGTGCTGATTGACAGTTC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.431746-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.431746-07	f
04853cf5-aee7-4abf-9240-600265068cb2	MRX188	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_20", "probe_seq_1": "TTGAGAGGAGGCATCGGG", "probe_seq_2": "TCACACACGAAATCTACG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.436354-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.436354-07	f
9401369a-7357-4e36-b527-2d208017597d	MRX189	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_21", "probe_seq_1": "AGTGGAACGTTATGCGCT", "probe_seq_2": "CCGAGTGCGTATCCATGA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.440646-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.440646-07	f
e6868a3f-63f7-4302-be5e-70a1bc5ec46e	MRX190	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_22", "probe_seq_1": "TCGGTCCGTTCTCACCCG", "probe_seq_2": "TCATGGGCTTTTGCGGTA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.445219-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.445219-07	f
8269b692-aa3c-45f6-9b49-415646e12639	MRX191	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_23", "probe_seq_1": "AGCAAGGGCAGGAATCAT", "probe_seq_2": "CCTTGGGAAGGGGACATT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.449624-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.449624-07	f
dfcce43b-c1c2-4169-aca9-4e84cb8f9eb9	MRX192	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_24", "probe_seq_1": "TTCTGATAACGGGAACCT", "probe_seq_2": "TTTGTAGGTATCAGGGTC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.454094-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.454094-07	f
4380dfd5-bc54-4a8f-9866-4012f2b64ab4	MRX193	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_25", "probe_seq_1": "GGCCCTGTAATCGGTCTC", "probe_seq_2": "CTAAGAAATGACATCGCC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.459277-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.459277-07	f
7191f250-f552-4d76-9ecd-78256934407d	MRX194	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_26", "probe_seq_1": "CTAGGATAAAGTCATCTA", "probe_seq_2": "GGGTAATCAAATGGGAAA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.464105-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.464105-07	f
5e7485c4-94f0-4ff4-b03a-fd6f67a262d0	MRX195	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_27", "probe_seq_1": "AGGAGTTTCGCGCGTCCT", "probe_seq_2": "GGGTGAGCGCGACGTTTC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.471022-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.471022-07	f
275089ba-354d-4b77-a90b-c9eb2be85073	MRX196	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_28", "probe_seq_1": "GGTCGTGTGATGCTCTTC", "probe_seq_2": "GTGCAAGGCCGCCCTTTC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.475801-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.475801-07	f
f9786af0-c6a2-470c-b39b-8bc18c43c409	MCX4	control:giab-HG002:1.0	{"cogs": {"cost": "5.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "giab NA24385", "volume": "", "comments": "", "lab_code": "", "volume_units": ""}, "description": "giab NA24385", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	control	content_instance	content	giab-HG002	2025-07-03 08:23:17.491289-07	1.0	ready	f	50d08295-a50b-422e-9c2a-6c1791c08e8f	2025-07-03 08:23:17.491289-07	f
d6dc35b7-cf4c-4e70-a692-c23346f61bcd	CX12	tube:tube-generic-10ml:1.0	{"cogs": {"cost": "1.50", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "giab NA24385", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "Generic 10ml tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	tube	container_instance	container	tube-generic-10ml	2025-07-03 08:23:17.484168-07	1.0	ready	f	d1533fbe-9bf0-46f2-abff-3c74cc80c744	2025-07-03 08:23:17.494103-07	f
4396a723-1c2c-490b-b006-94ab6438e2d8	MRX205	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.811179-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.811179-07	f
5bb46d8b-d389-4d1f-a86f-2bd41aef6aa1	MRX206	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.817124-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.817124-07	f
3813f28b-8d60-4dee-9752-1dabc8ba2ee2	MRX207	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.823295-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.823295-07	f
91cccd1e-a078-41ca-bbb3-67985eaaf192	MRX232	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_8", "probe_seq_1": "CATTAGGGATAGGGGTGA", "probe_seq_2": "CCAAACCTCGCCCATTGT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:24:15.993143-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:24:15.993143-07	f
cab1f6d5-eb37-459c-8fc4-4e3a8e81855f	MCX5	control:synthetic-cfdna:1.0	{"cogs": {"cost": "3.1", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "s ynthetic-cfdna", "volume": "", "comments": "", "lab_code": "", "volume_units": ""}, "description": "synthetic-cfdna", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	control	content_instance	content	synthetic-cfdna	2025-07-03 08:23:17.502951-07	1.0	ready	f	4765a344-581e-4bc6-9c70-cca99e85637d	2025-07-03 08:23:17.502951-07	f
9c6dc920-ebc9-44ca-bf88-13f9ca571726	CX13	tube:tube-10ml-glass:1.0	{"cogs": {"cost": "4.70", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "s ynthetic-cfdna", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "10ml glass tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	tube	container_instance	container	tube-10ml-glass	2025-07-03 08:23:17.500261-07	1.0	ready	f	7651bf71-c917-4c36-bd28-ded3fa87e1c2	2025-07-03 08:23:17.505305-07	f
e47de7d0-f3d5-4fd5-ae77-6f3d053caf84	MRX289	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.679546-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.679546-07	f
7fcfa70d-2016-45c2-a092-4f42ebed1393	MCX6	control:water-ntc:1.0	{"cogs": {"cost": "0.10", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "NTC - water", "volume": "", "comments": "", "lab_code": "", "volume_units": ""}, "description": "NTC - water", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	control	content_instance	content	water-ntc	2025-07-03 08:23:17.511787-07	1.0	ready	f	70d42dd6-b1f4-4293-ba62-97c902f6ffb0	2025-07-03 08:23:17.511787-07	f
fe1b21ed-3957-4df1-b13c-fcaccc035a83	CX14	tube:tube-eppi-1.5ml:1.0	{"cogs": {"cost": "1.70", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "NTC - water", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "1.5ml eppi tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	tube	container_instance	container	tube-eppi-1.5ml	2025-07-03 08:23:17.509268-07	1.0	ready	f	ceca7fef-9b77-433d-9420-8d8a15cf1147	2025-07-03 08:23:17.514422-07	f
43b72609-8eab-47a3-9cd2-fc54e4092892	CX15	plate:fixed-plate-24:1.0	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	plate	container_instance	container	fixed-plate-24	2025-07-03 08:23:17.5205-07	1.0	ready	f	6bae2028-8873-4717-9e10-821a62bcd9f0	2025-07-03 08:23:17.5205-07	f
35280b86-54a8-4337-9df0-abcb0cd2c340	CWX249	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:15.911667-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:15.91939-07	f
67da16e4-6e1d-451f-9319-d3b627ef5a3e	CWX197	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.525364-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.530954-07	f
3ab97a64-e71c-41b1-b799-60b928503b2e	CWX198	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.532606-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.536954-07	f
1773aa0b-fcb9-440c-99ba-e97d08c897d2	CWX199	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.538143-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.54221-07	f
245e0cf5-ffc9-4b32-bd05-857c40d77e4d	CWX200	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.543423-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.547902-07	f
3f3515b1-157e-4537-9270-70187b0b55b2	CWX201	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.549119-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.553401-07	f
6282522a-6923-4b31-8c11-9945e5d04e9c	CWX202	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.554699-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.558745-07	f
743e4c9a-0b92-48f1-b15e-aa89d9ef6f17	CWX203	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.560004-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.564347-07	f
3f257485-e9ed-43ba-9245-63fe12620c6e	CWX250	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:15.920899-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:15.92644-07	f
3fb7f5a8-6c7b-44c6-af3b-875bb8f8fcc9	CWX204	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.566042-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.570514-07	f
24c55350-7ef4-4463-90ab-46252ab1c1c5	CWX205	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.571878-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.576571-07	f
8ad330a0-fad3-4b94-be1d-6c2aade5b972	CWX206	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.577992-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.582129-07	f
801a01d3-8107-4911-9822-69f73fef2dee	CWX207	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.583496-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.588081-07	f
f6e70812-dce0-4359-848a-7c86a15b25eb	CWX208	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.589387-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.593702-07	f
226cfe6d-f38c-4275-b630-54bc1a181c4f	CWX209	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.594959-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.602179-07	f
c9afb65c-3bce-41aa-94ca-ed1c356a938c	CWX251	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:15.929132-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:15.935348-07	f
6d5743cc-81b2-46c0-bf95-ec24bf59c944	CWX210	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.605463-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.610861-07	f
e63af02f-2942-4849-a657-ab339e79a86c	CWX211	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.612317-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.616695-07	f
ab47b771-3bc5-4ec6-88ea-14a05ab36fe7	CWX212	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.618607-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.624378-07	f
32a38823-7229-4403-b7c5-4564b57105f6	CWX213	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.625699-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.629956-07	f
ed9a04fc-0e00-4fd1-8813-54c833e62db0	CWX214	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.631317-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.657687-07	f
39bff9ce-0ded-465c-b704-dda606530e7a	CWX215	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.685254-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.700558-07	f
cca985cb-db90-47cb-880a-3749d3611984	CWX252	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:15.938076-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:15.943407-07	f
fa6b2d22-7543-4429-b740-abd29980cabb	CWX216	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.703574-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.708097-07	f
2fe1c959-22e0-457c-b5f1-2e043fe9c6e4	CWX217	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.709524-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.713615-07	f
c42216f6-5feb-418b-af3c-792c2e43b666	CWX218	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.714876-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.719105-07	f
3ed95fad-05dc-4da6-a86e-f9cf24a0215a	MRX225	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_1", "probe_seq_1": "TAATATGCCCGCATAGAC", "probe_seq_2": "TGCCTGTCTGATCTTTAA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:24:15.952949-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:24:15.952949-07	f
7207e508-99fc-41ef-91ab-fd60d5306f4e	CWX219	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.720404-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.72516-07	f
6b65a1a8-ba49-4cb5-b381-72a0c2ab0f7d	MRX226	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_2", "probe_seq_1": "ACCCGTCATTTCAGGTTC", "probe_seq_2": "CAATACCCCTGTGGCGCC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:24:15.956035-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:24:15.956035-07	f
13a17e5f-6d85-417c-9d4a-956b5282800a	CWX220	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.72663-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.732601-07	f
f4382eed-95c5-4d1b-a046-f2b9b277ec34	CWX221	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.733993-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.738774-07	f
467ee04d-f443-49cf-b38d-090e047d1306	MRX227	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_3", "probe_seq_1": "TATGTAAGGACACTGCAC", "probe_seq_2": "GAGAGGCCGAGACCCGCC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:24:15.963002-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:24:15.963002-07	f
741b104f-826e-443e-8efd-9d6aaf9ec330	CWX222	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.740122-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.744649-07	f
3c31b867-fb64-4aaa-b939-f7e98b8abb92	MRX228	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_4", "probe_seq_1": "AACATGGTGGGCCCGAGT", "probe_seq_2": "TGGTGCTCTGGACCGTTT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:24:15.968746-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:24:15.968746-07	f
340f2341-eb12-4958-9c03-224ed6cc0747	CWX223	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.746439-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.751906-07	f
c6eb89d6-4a49-4da3-a615-6be4294b9235	CWX224	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:23:17.753422-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:23:17.757567-07	f
0d22f68d-dd27-4e92-ba89-b39005cfca81	MRX197	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.761172-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.761172-07	f
f186463f-a0de-4e66-9a85-60d861f29c20	MRX198	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.766127-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.766127-07	f
78f2fa4b-bfcb-46fc-a9ef-2f0bb27926f8	MRX199	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.772121-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.772121-07	f
906cdad8-145e-4040-b4f2-5ef1f31c907b	MRX200	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.781168-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.781168-07	f
dbc67543-2fb8-4e02-949f-1c2146b9f1c3	MRX201	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.786967-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.786967-07	f
d9d3755e-bd33-44fe-a4de-a68f8f0e4275	MRX202	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.792875-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.792875-07	f
9c4cca2b-65bf-4957-9a49-fefe346bdca8	MRX203	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.799241-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.799241-07	f
a6902e58-6edc-4679-a335-fffd804cd1b3	MRX204	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.805147-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.805147-07	f
9aea1a13-4383-4c3d-b24b-aad55d4e8111	MRX209	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.836604-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.836604-07	f
604c069f-71dd-4ee8-b4e8-ef6e7bb46e5c	MRX210	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.842567-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.842567-07	f
53dfbf91-c236-4d6e-b31c-a9bf4bc369a2	MRX211	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.848523-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.848523-07	f
3666e200-4c24-4efc-a76b-d72a6faa5e65	MRX212	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.855048-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.855048-07	f
c282db24-d774-48c1-8634-eb60d7b861fc	MRX213	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.867487-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.867487-07	f
5a55869e-cf2a-470b-b132-b2e375b2872c	MRX214	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.873982-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.873982-07	f
2d6a5e4e-5f46-48bc-ab25-2f3f5ddb6412	MRX215	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.879853-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.879853-07	f
e3be72b1-8052-4d5b-a017-c9db741c7bff	MRX216	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.885713-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.885713-07	f
c69af3be-aa7a-446d-a0c9-8a3c9d5b06f7	MRX217	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.891898-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.891898-07	f
264f523e-dc9b-4fb7-b188-68d48691adab	MRX218	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.900232-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.900232-07	f
b1f40745-93ea-4e41-8d30-5a900f02aab9	MRX219	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.906482-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.906482-07	f
1452bef9-03fa-4bf2-9908-b85eb998617a	MRX220	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.912796-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.912796-07	f
5551791d-203e-4584-b45d-4cce837d92a7	MRX221	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.919066-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.919066-07	f
a0110d14-da41-463b-86a6-81b15bdd0775	MRX222	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.925122-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.925122-07	f
c9e2dab1-38f2-4ade-9c19-814d76031f2d	MRX223	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.931001-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.931001-07	f
9999a8ff-1161-4ba5-a8c3-347e452a5f92	MRX224	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:23:17.936775-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:23:17.936775-07	f
8be1afd7-55da-4457-90b1-f02f5cc77e5c	CX16	rack:tube-rack-4-empty:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "4 position Rack", "comments": "", "lab_code": ""}, "description": "4 position Rack", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	rack	container_instance	container	tube-rack-4-empty	2025-07-03 08:23:17.942647-07	1.0	ready	f	d5ee3d31-ad0b-4780-a97d-b0c78bc74ac3	2025-07-03 08:23:17.942647-07	f
94ba58bf-d6b4-4163-bd4d-a69c9e105a68	MRX229	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_5", "probe_seq_1": "TGGTATCTCACCACTTTA", "probe_seq_2": "AGATTCCACGGGGAATAG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:24:15.9742-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:24:15.9742-07	f
fb4fba5f-dbd7-42d9-a5a3-7fe0cfca759d	MRX230	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_6", "probe_seq_1": "GTACGCGACACGAGCGGT", "probe_seq_2": "TATGGAAGAATAATGAAT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:24:15.979872-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:24:15.979872-07	f
c41f43b2-3170-48fd-97c3-304c73f09ad7	MRX231	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_7", "probe_seq_1": "GCCCAATTACGTTTGCGG", "probe_seq_2": "CGTTGATGGTGAGCGTAC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:24:15.985312-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:24:15.985312-07	f
43079fcb-f024-46aa-9320-7c51a5bd6494	MRX233	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_9", "probe_seq_1": "TGCCGAGAATATTACTTC", "probe_seq_2": "CGAAGGAAGTTAGCTGGC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:24:15.998517-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:24:15.998517-07	f
c76c7d77-7a2c-4aaf-8a66-637b61c1f62e	MRX234	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_10", "probe_seq_1": "TGCAACCTAAGTGTAAGG", "probe_seq_2": "GTACAGTTTATAAAGGTG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:24:16.005492-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:24:16.005492-07	f
d20fb8e0-c8c4-4754-a5b3-4937f66a496c	MRX235	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_11", "probe_seq_1": "AGGAACAGGCTGGTAGCT", "probe_seq_2": "ATGAGTTGTCTATACAAC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:24:16.010893-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:24:16.010893-07	f
5fff5b28-eb3e-4bbd-b79f-754f8f03eea3	MRX236	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_12", "probe_seq_1": "GGCGCCGGTCGAACCTAT", "probe_seq_2": "TGTTAGTCTTTGTGACAG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:24:16.016379-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:24:16.016379-07	f
1ddb556f-e2da-4b49-b71a-3201257874a6	MRX237	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_13", "probe_seq_1": "GCACAACTTCATACTCGC", "probe_seq_2": "AGGGATCACATGTGGTTA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:24:16.021218-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:24:16.021218-07	f
cbd6c137-3d08-43da-84e9-4f918b6946ce	MRX238	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_14", "probe_seq_1": "GTGTTCGGGGCAATAACG", "probe_seq_2": "CATATGATGTCAGGAGGC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:24:16.025934-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:24:16.025934-07	f
ba9a32dd-9ff0-4bd5-bec2-2be8156a1c60	MRX239	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_15", "probe_seq_1": "CAGTAGGAACCGACTGCT", "probe_seq_2": "GCTGTTCGTTCCGTCAAA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:24:16.030921-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:24:16.030921-07	f
653a1b4c-97ca-4469-88ee-08bdeeb09309	MRX240	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_16", "probe_seq_1": "TAACCCAGGACCAGACGC", "probe_seq_2": "TCAAAGTCTGTTATGGGG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:24:16.03599-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:24:16.03599-07	f
045dfa06-cd02-4830-815a-64a0eda2b595	MRX241	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_17", "probe_seq_1": "GGCTAGAGACACCTAGGG", "probe_seq_2": "CTAAGTTTCATCGCGTGC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:24:16.040974-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:24:16.040974-07	f
e5c9c69f-4bf8-4583-83c3-ab32883c6fd8	MRX242	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_18", "probe_seq_1": "GCGCATGCTAGGGTCAAG", "probe_seq_2": "AATCTCTACGAATCAAGT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:24:16.045785-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:24:16.045785-07	f
aa3dceeb-693f-4b52-96fd-fcdacec74279	MRX243	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_19", "probe_seq_1": "TCACGAAATTCCGATTGG", "probe_seq_2": "GCTGCTAAAAGGCGTTAA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:24:16.050413-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:24:16.050413-07	f
1806a41a-5c44-4be2-9d25-69589865a6fe	MRX244	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_20", "probe_seq_1": "CAGCACATCATTAACAGA", "probe_seq_2": "CACGGGTCAATCCTCAGG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:24:16.057149-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:24:16.057149-07	f
81670de8-fbad-4c5b-8159-65d031907aa0	MRX245	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_21", "probe_seq_1": "TTACGCCATACCGGTACC", "probe_seq_2": "CGGTACGGACGTGCAGAA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:24:16.061637-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:24:16.061637-07	f
8ca2ddc1-bf48-4c18-ae6f-54d1d4bef1e4	MRX246	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_22", "probe_seq_1": "CTCCGCCGTTGACGAACG", "probe_seq_2": "CACCCCGCTCAAGAAAGC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:24:16.066932-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:24:16.066932-07	f
e1feb2f2-726f-4aa6-8334-7875ca8a9d76	MRX247	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_23", "probe_seq_1": "GAGCCAGTCTTAGTTACG", "probe_seq_2": "GGTGCCTGTACATATGGG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:24:16.071479-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:24:16.071479-07	f
4ef01449-ede9-4847-9074-67d56aa0e462	MRX248	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_24", "probe_seq_1": "AGCGTGTGAAGGTCGCAT", "probe_seq_2": "TTCCACGTAACTACTAAC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:24:16.075965-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:24:16.075965-07	f
955793c4-ddbf-4bdf-a4e2-51f541d2d828	MRX249	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_25", "probe_seq_1": "TGTGACAGATAGTCGCGA", "probe_seq_2": "ACTAGTATCTCCACGTAT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:24:16.080477-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:24:16.080477-07	f
275fe9ac-a3f8-4182-bd15-d5982dab919d	MRX250	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_26", "probe_seq_1": "GTGCCCTTGTTCTATTAC", "probe_seq_2": "GATCCCCTCCGCCATGCT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:24:16.08627-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:24:16.08627-07	f
890ef728-0ffb-4f4d-90bd-29924f4bb340	MRX251	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_27", "probe_seq_1": "CTATCTCCTTTATATAAG", "probe_seq_2": "AGCAAAATCTTCCCCGTA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:24:16.090668-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:24:16.090668-07	f
5b8e86dd-588a-4518-9713-3a48aef550a6	MRX252	reagent:idt-probes-rare-mendelian:1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "probe_name": "id_probe_28", "probe_seq_1": "AGCTTAGTCGCCGACAAT", "probe_seq_2": "TGAGAGAGCCTATCCACC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	idt-probes-rare-mendelian	2025-07-03 08:24:16.095989-07	1.0	ready	f	95078f21-b1f0-46c6-99cf-e198717cca5e	2025-07-03 08:24:16.095989-07	f
f0bbdd4d-da05-445d-8c05-a8ed4769fe79	CX18	plate:fixed-plate-24:1.0	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	plate	container_instance	container	fixed-plate-24	2025-07-03 08:24:16.103179-07	1.0	ready	f	6bae2028-8873-4717-9e10-821a62bcd9f0	2025-07-03 08:24:16.103179-07	f
7a7a05e0-d926-4739-8f3c-c9a41c715f1a	CWX253	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.108836-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.111386-07	f
78ea25c1-5b2c-4e77-8781-5fffd25ab2df	CWX254	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.115443-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.11985-07	f
6a69eda7-1e5a-4d4a-8755-94aefc3c74e8	CWX255	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.121029-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.124914-07	f
217b0af9-7c61-4971-8098-d67ce949fc38	CWX256	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.126215-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.129956-07	f
4484b524-4a4b-408d-8153-f19da5a148c1	CWX257	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.131433-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.13607-07	f
220d86a6-088f-47a7-a11b-f64519da6a46	CWX258	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.137239-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.141143-07	f
c3897a38-49ee-4f23-9461-a55dff9a4413	CWX259	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.142482-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.146746-07	f
2588b288-14ab-4eab-b9ff-41c51104f8bc	CWX260	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.148035-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.152392-07	f
13fb23bc-4066-4e86-bde9-9a56cdc08e57	CWX261	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.153972-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.159719-07	f
4a40cc94-4c0d-4f56-a5be-7b17905d478f	CWX262	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.160832-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.165623-07	f
a34c5ab3-1540-4e32-a08c-9f90c1f7894e	CWX263	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.16817-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.172411-07	f
0007e7e3-47ad-48fa-8228-7f0dbe1f4a60	CWX264	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.173538-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.177886-07	f
e6a96820-ce9d-4ec4-8f59-fd93f0e11b76	CWX265	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.179232-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.183037-07	f
5eab8dce-d0dd-449f-966d-47154aef3cf0	CWX266	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.184262-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.188374-07	f
4e71540e-b707-4978-a4da-25a8ac6dbb48	CWX267	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.189469-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.193381-07	f
7f0ff5ac-55fd-40b8-993c-37fd45de7e70	CWX268	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.194592-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.199917-07	f
5914dd40-bbb0-451f-aabf-17c7970df831	CWX269	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.201071-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.204924-07	f
d53b5c53-dff3-4c58-be12-ef4497a67766	CWX270	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.20631-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.210561-07	f
561fbd02-0e42-4c28-b549-540297a81c6a	CWX271	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.212024-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.216014-07	f
ddc6efd5-96cb-476a-a565-bb7248a7b9a2	CWX272	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.217462-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.221429-07	f
f5a92643-727e-42cd-a337-e2fd476dc69d	CWX273	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.222941-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.228929-07	f
2754a4ac-782d-4991-9e07-ed7cad5dfc2a	CWX274	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.230475-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.234608-07	f
74137fae-e9d7-4e0b-b88c-f6e1ef1aeb96	CWX275	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.235951-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.240465-07	f
7b8a94d7-37af-4f1b-a3fc-1cd5399cb3f9	CWX276	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.241895-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.246288-07	f
165a43a4-b366-4fc2-a6c1-97f8c9b236dc	CWX277	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.247453-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.251863-07	f
b2fc4da9-2a05-49ca-8d5d-b9558cc97c68	CWX278	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.257121-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.261195-07	f
774fcd11-a6a1-4bdc-8b98-7fbc239ad4d7	CWX279	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.262563-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.267222-07	f
91999ed4-21f0-408d-8f7c-a00ad024b9a8	CWX280	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.268554-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.272866-07	f
e3b20962-4aaa-4e81-850f-90b4ff6bb72b	MRX253	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_1", "probe_seq_1": "AGATATAACAAATAGTCC", "probe_seq_2": "TTCTGCGAAAGGCTTGGC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.276217-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.276217-07	f
74ad366b-d25f-4da4-8084-53ea5fa20220	MRX254	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_2", "probe_seq_1": "CATGACGTTAATGCGGGC", "probe_seq_2": "TAGAGCGTGATTTAACGG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.278242-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.278242-07	f
0c89558c-3040-41f3-b14f-54281d89a861	MRX255	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_3", "probe_seq_1": "GAATCGTAGCTGCAATCC", "probe_seq_2": "CGATTCTATATGTAGCTT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.283122-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.283122-07	f
0c26d313-4b07-4a0f-a826-ea2b0242df8b	MRX259	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_7", "probe_seq_1": "AGTCCCGTAGGAGATCTT", "probe_seq_2": "GTGCAGACTTACATATTC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.307149-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.307149-07	f
5ef0757b-7e03-409b-87ea-70557be4dac0	MRX260	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_8", "probe_seq_1": "TACACTGTTGTGTTAAAG", "probe_seq_2": "TCGGTTTACTGTGTGAAG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.311778-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.311778-07	f
a5bb73df-6539-4173-9687-1763600076bc	MRX261	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_9", "probe_seq_1": "CAGTGTCCCGCGAGACTG", "probe_seq_2": "GGGATTCTATATGTTGTG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.316737-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.316737-07	f
4830667e-3fe8-4a03-beeb-80786dc41757	MRX262	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_10", "probe_seq_1": "AAGAGACACCAGTAGCCG", "probe_seq_2": "TAGATATTAAGCCCTGAA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.321763-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.321763-07	f
7512b9d2-980e-4174-952b-78facb42ec6d	MRX263	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_11", "probe_seq_1": "ACAGGTGCGCTGTTAGGC", "probe_seq_2": "TCCCTTTACCAGGTTAGC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.326841-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.326841-07	f
b0b03902-4397-4f5c-99cf-9bc996f64e45	MRX264	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_12", "probe_seq_1": "CCCCCCGAACTGAAGTAA", "probe_seq_2": "TGAAATTAGCGCAACTAT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.331496-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.331496-07	f
90a7b8ca-0912-4f4f-9e1b-35ee43101fe8	MRX265	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_13", "probe_seq_1": "TTTGCTTAATCAACGTCC", "probe_seq_2": "TCGATACCCATGAGCAAA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.336673-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.336673-07	f
ffae08fc-ee93-4871-9d7a-5c45332c4d86	MRX266	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_14", "probe_seq_1": "ATCTGAGCGGGATCATCT", "probe_seq_2": "AATTAAATCAACGCTTAT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.342636-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.342636-07	f
7823f81f-b135-4cb1-9937-96ad578a673d	MRX267	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_15", "probe_seq_1": "AGACAAACCCAGTGCAAA", "probe_seq_2": "GGACTCCGGTAGAATACG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.34712-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.34712-07	f
4a67fb94-8ae8-4ca3-bc46-81ed67993cf1	MRX268	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_16", "probe_seq_1": "ACGGTGCCAGGGAGCTGC", "probe_seq_2": "GAGTCTGACTCGCGTCGC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.35246-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.35246-07	f
c5f72729-d226-4ab5-b263-52ecffcea927	MRX269	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_17", "probe_seq_1": "TACCGCGACGGTATCCGT", "probe_seq_2": "AGAACCGCGTGGTTCCAT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.357321-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.357321-07	f
debc0fde-6d6a-4e56-899b-cd47f75025bb	MRX270	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_18", "probe_seq_1": "CAGTCAAGTCACGTCGAT", "probe_seq_2": "TGGTTCGATAGCTATGTA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.361823-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.361823-07	f
ebaff42d-7c7c-4d36-b7ce-02db9be33504	MRX271	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_19", "probe_seq_1": "TTTGCCATTGGGATAGTG", "probe_seq_2": "TTTAACCGTCAGTGATCT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.369633-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.369633-07	f
19af9824-019d-472e-a8b7-e5b098186cd7	MRX272	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_20", "probe_seq_1": "CAACTAACATTAACACTG", "probe_seq_2": "TATAGATATGCTCGCTCC", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.376565-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.376565-07	f
c2af1c10-826b-403b-945b-5068a0fe4998	MRX273	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_21", "probe_seq_1": "CGTAAGAGACCAAGCGCG", "probe_seq_2": "CCTACGGCTTTCTCGATA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.381663-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.381663-07	f
b9299cec-5b81-48be-b4e8-bcce06bf4d26	MRX274	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_22", "probe_seq_1": "CATAGTAGTGCGCACAGC", "probe_seq_2": "GTCGTTCCCGAACTAACT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.38591-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.38591-07	f
ac122017-f1e1-449f-bdf3-39016e27fad8	MRX275	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_23", "probe_seq_1": "GCAAGCGACGGCCTTCGA", "probe_seq_2": "GGGAACGCAAGGCAACGT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.391185-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.391185-07	f
e23c847f-8774-4450-9dc9-ec4f7233a6d8	MRX276	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_24", "probe_seq_1": "GCATTCGAGTCCAGAAGT", "probe_seq_2": "CCTCGAGTTCTAATGGGT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.395568-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.395568-07	f
f5228cdc-84b8-4f13-8171-543a997417ec	MRX277	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_25", "probe_seq_1": "TCAGGAGAGTGAACCCCA", "probe_seq_2": "TTGCTAGACGGAGCAATG", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.400057-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.400057-07	f
7d0c6c07-ac47-47df-b229-c005d7d1685d	MRX278	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_26", "probe_seq_1": "TAGTCTCTGAGACACAGT", "probe_seq_2": "ACCCCCGAATCTCACGAT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.404841-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.404841-07	f
c5a56bf4-3463-40b3-8b68-5e47b32a5cff	MRX279	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_27", "probe_seq_1": "TGGGTTTCAGTTAGAACT", "probe_seq_2": "ATCTGGTCCGAGTGGGGT", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.410484-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.410484-07	f
1b97a184-a3da-4d82-a043-a02a95207c0e	MRX280	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "probe_name": "id_probe_28", "probe_seq_1": "CGTGCTCCACTGGTATGG", "probe_seq_2": "GGAAGAACCCTTTGCTAA", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.415536-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.415536-07	f
845ebde8-23d0-447c-a05e-538a37a81f97	MCX7	control:giab-HG002:1.0	{"cogs": {"cost": "5.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "giab NA24385", "volume": "", "comments": "", "lab_code": "", "volume_units": ""}, "description": "giab NA24385", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	control	content_instance	content	giab-HG002	2025-07-03 08:24:16.432998-07	1.0	ready	f	50d08295-a50b-422e-9c2a-6c1791c08e8f	2025-07-03 08:24:16.432998-07	f
b9d2526e-ab8a-4ff8-a1b0-947071f5d2d7	CX19	tube:tube-generic-10ml:1.0	{"cogs": {"cost": "1.50", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "giab NA24385", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "Generic 10ml tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	tube	container_instance	container	tube-generic-10ml	2025-07-03 08:24:16.425128-07	1.0	ready	f	d1533fbe-9bf0-46f2-abff-3c74cc80c744	2025-07-03 08:24:16.436412-07	f
4f0b197f-9f39-43fd-90bd-3ee2cf2ca0a6	MCX8	control:synthetic-cfdna:1.0	{"cogs": {"cost": "3.1", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "s ynthetic-cfdna", "volume": "", "comments": "", "lab_code": "", "volume_units": ""}, "description": "synthetic-cfdna", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	control	content_instance	content	synthetic-cfdna	2025-07-03 08:24:16.448073-07	1.0	ready	f	4765a344-581e-4bc6-9c70-cca99e85637d	2025-07-03 08:24:16.448073-07	f
0bbf1234-7c4b-4afc-9352-56a600deb076	CX20	tube:tube-10ml-glass:1.0	{"cogs": {"cost": "4.70", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "s ynthetic-cfdna", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "10ml glass tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	tube	container_instance	container	tube-10ml-glass	2025-07-03 08:24:16.445109-07	1.0	ready	f	7651bf71-c917-4c36-bd28-ded3fa87e1c2	2025-07-03 08:24:16.450581-07	f
da2ba524-f58e-4e2f-bc8b-39610f57f804	MCX9	control:water-ntc:1.0	{"cogs": {"cost": "0.10", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "NTC - water", "volume": "", "comments": "", "lab_code": "", "volume_units": ""}, "description": "NTC - water", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	control	content_instance	content	water-ntc	2025-07-03 08:24:16.457098-07	1.0	ready	f	70d42dd6-b1f4-4293-ba62-97c902f6ffb0	2025-07-03 08:24:16.457098-07	f
f69f3bb2-4931-451c-ba90-a2f283ef8618	CX21	tube:tube-eppi-1.5ml:1.0	{"cogs": {"cost": "1.70", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "NTC - water", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "1.5ml eppi tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	tube	container_instance	container	tube-eppi-1.5ml	2025-07-03 08:24:16.454793-07	1.0	ready	f	ceca7fef-9b77-433d-9420-8d8a15cf1147	2025-07-03 08:24:16.459479-07	f
7a653734-9941-44ab-9446-6ef7acee9702	CX22	plate:fixed-plate-24:1.0	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	plate	container_instance	container	fixed-plate-24	2025-07-03 08:24:16.463243-07	1.0	ready	f	6bae2028-8873-4717-9e10-821a62bcd9f0	2025-07-03 08:24:16.463243-07	f
70493c68-0506-4017-a8f9-5d14bdef4cf7	CWX281	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.468168-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.471057-07	f
ee0ded45-6595-40e7-8c32-07e783b35d22	CWX282	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.472381-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.476382-07	f
367a6a1a-9157-4b01-b2ec-c2dc81af3154	CWX283	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.480503-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.486322-07	f
ca914991-5199-43a4-a369-2075a264f28d	CWX284	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.487499-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.491628-07	f
9ab7fce2-54c8-491d-a3ec-17f3bc41ceba	CWX285	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.492824-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.496667-07	f
851ce518-5a99-4a87-ba96-8c4e2b8fde5c	CWX286	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.497839-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.501632-07	f
b22b7269-47b8-44e6-a218-8c486c50cc0e	CWX287	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.502909-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.506763-07	f
6ccd1298-f01c-4572-a51a-1defbf10da94	CWX288	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.507977-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.511867-07	f
0cfbb220-c556-449b-9a17-8f8b74170237	CWX289	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.513313-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.517405-07	f
a20882a3-20fb-4e2a-b68c-ed393bb816d3	CWX290	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.518506-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.522203-07	f
9bfaafce-9d17-4839-a5ea-03bbdfcc4b2e	CWX291	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.523496-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.527786-07	f
2401c6c6-d364-44ec-94ae-fbe7825e8654	CWX292	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.528932-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.533495-07	f
2259a89d-1a1e-4541-a8f0-1316da02a81d	CWX293	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.53635-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.541064-07	f
5103010b-3a86-45d4-b1cd-0a9e77622b35	CWX294	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.542302-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.54621-07	f
b1be5bf3-c765-4f3c-a272-e128495d6a30	CWX295	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.547555-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.551494-07	f
28a66e77-5293-4587-8aaa-2a11cd102349	CWX296	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.55288-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.556945-07	f
344aff1a-652b-4802-8b45-87e345a5dfaf	CWX297	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.558048-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.561707-07	f
e41faaaf-ce2a-43d6-87e5-e3870abe45d6	CWX298	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.562895-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.568162-07	f
006ca986-054c-4e9f-8e46-dac3dc562a98	CWX299	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.569326-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.573033-07	f
233d4349-e767-46bf-a7ed-035e092867e8	CWX300	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.574559-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.579865-07	f
f20b5ca4-d90b-4558-8f3a-088e0959cb77	CWX301	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.581962-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.588873-07	f
35620e62-853f-41df-b226-7036aae429e1	CWX302	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.590183-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.594043-07	f
156a6879-7978-4ba2-9f2d-be6064697797	CWX303	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.595467-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.600109-07	f
121e9047-ad3e-4dfd-823b-83e1fc717cbc	CWX304	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.601469-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.605983-07	f
fda6b968-9927-4cf6-85f9-c44bb39014a7	CWX305	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.608243-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.61229-07	f
c7124a90-7b46-478e-b74a-dacd7a07dc62	CWX306	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.614298-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.618636-07	f
cb5b12ec-e9e7-4707-bffc-0e453d1fc960	CWX307	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.619981-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.62407-07	f
ffefe78a-8ac0-4e3e-9fa9-772ee8762f39	CWX308	well:fixed-plate-well:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	well	container_instance	container	fixed-plate-well	2025-07-03 08:24:16.62529-07	1.0	ready	f	507a3395-14f5-4e21-bf68-66df5d242281	2025-07-03 08:24:16.629386-07	f
3ebb20a3-6d70-4ebb-9a8d-f070a39a60f0	MRX281	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.633054-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.633054-07	f
5d6d4e76-14b5-4009-9f05-a8e3062bb797	MRX282	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.637466-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.637466-07	f
a9013d2d-2a96-4508-b74a-6142dc7eaacd	MRX283	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.643287-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.643287-07	f
2355589e-b1e9-469d-9593-602b77227c24	MRX284	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.649412-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.649412-07	f
ca939326-d2b1-4b88-80a3-15cb849ba2af	MRX285	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.655929-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.655929-07	f
88694a6f-07f1-4ef8-a9b8-97467620e235	MRX286	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.662326-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.662326-07	f
86c0d971-5f10-4cb5-ae88-c21624503ba6	MRX287	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.668647-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.668647-07	f
9f1fb3f4-0611-4f6e-a39a-71b3ae2dba25	MRX288	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.67413-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.67413-07	f
e72c6cc1-db0a-41bd-b271-a7d5b645df0c	MRX290	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.684741-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.684741-07	f
49a12695-1781-4daa-80c6-fbb71f9ba0ba	MRX291	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.690129-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.690129-07	f
5117b15a-474f-43b7-b9eb-e367af711f2a	MRX292	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.697069-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.697069-07	f
cb3a3bb1-aefc-4355-aae0-fcf5a8a94013	MRX293	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.705224-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.705224-07	f
51e3ccc5-9470-49f4-987d-b96a897a3fd2	MRX294	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.710408-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.710408-07	f
dea04455-e99a-4424-ac3b-fc7124de43f8	MRX295	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.715575-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.715575-07	f
0559ee06-9d4f-4f58-a12e-1f32bdff4633	MRX296	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.720517-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.720517-07	f
d20de8f6-379c-429c-9a0b-916b7833b1a4	MRX297	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.725843-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.725843-07	f
bb287c24-62a2-4531-8ca1-909803f845fc	MRX298	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.730975-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.730975-07	f
eb4ac051-5a75-4ba3-9725-081583664880	MRX299	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.736556-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.736556-07	f
89e14590-5f15-40df-8f99-d3784fb556be	MRX300	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.741873-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.741873-07	f
7fc49341-3ee8-463f-b10a-5919cd6f2f0a	MRX301	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.746901-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.746901-07	f
0262fa19-4669-4ca2-94d9-6fd44dfebbcf	MRX302	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.752108-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.752108-07	f
1e661782-0d7e-478b-834a-cf1b8e5619ad	MRX303	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.757112-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.757112-07	f
1d281d59-1305-4a9a-b542-7a874fea7e37	MRX304	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.762394-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.762394-07	f
97b8ade0-e5da-4ab0-97a5-a03f9ce9b362	MRX305	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.767513-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.767513-07	f
fa1ac264-1427-48af-9781-70e9f462b347	MRX306	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.772618-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.772618-07	f
ed96bfa1-f338-408f-86b5-48328def257e	MRX307	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.777579-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.777579-07	f
9cd2fa8f-9130-42fa-b43d-5e9110ceaec6	MRX308	reagent:naoh:1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "action_groups": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	reagent	content_instance	content	naoh	2025-07-03 08:24:16.782698-07	1.0	ready	f	37a16678-0fa3-401e-8467-f5842fa37866	2025-07-03 08:24:16.782698-07	f
88e5e9bc-0bf2-40e7-98fb-cc2576cf330d	CX23	rack:tube-rack-4-empty:1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "4 position Rack", "comments": "", "lab_code": ""}, "description": "4 position Rack", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	rack	container_instance	container	tube-rack-4-empty	2025-07-03 08:24:16.787731-07	1.0	ready	f	d5ee3d31-ad0b-4780-a97d-b0c78bc74ac3	2025-07-03 08:24:16.787731-07	f
\.


--
-- Data for Name: generic_instance_lineage; Type: TABLE DATA; Schema: public; Owner: daylily
--

COPY public.generic_instance_lineage (uuid, euid, parent_instance_uuid, parent_type, child_type, polymorphic_discriminator, child_instance_uuid, created_dt, name, super_type, btype, b_sub_type, json_addl, bstatus, version, is_deleted, modified_dt, is_singleton, relationship_type) FROM stdin;
54f3e9c7-4f9f-48c7-8b3a-67ecfc60d506	GL1	a95a3806-06d0-4ce3-bf5b-042fb400562b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	01045846-479f-43f7-8d20-e84e3bec3420	2025-07-03 05:58:00.946314-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T125759", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 05:58:00.946314-07	f	generic
96815b0b-1120-4db7-a199-f709a5961913	GL2	a95a3806-06d0-4ce3-bf5b-042fb400562b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	c76338b6-6ba8-4dc2-9b60-eb20a9e253a6	2025-07-03 05:58:02.248251-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T125759", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 05:58:02.248251-07	f	generic
8796fe96-65e1-4b68-aa6a-fc0a978ecbe8	GL3	a95a3806-06d0-4ce3-bf5b-042fb400562b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	6dbd1b9c-394e-43ad-9935-dce8db433908	2025-07-03 05:58:03.384589-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T125759", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 05:58:03.384589-07	f	generic
24ef7ef4-c0e6-485f-a117-52b457642f9f	GL4	a95a3806-06d0-4ce3-bf5b-042fb400562b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	2daf2aa2-3633-4fd6-a8a4-784977febd55	2025-07-03 05:58:04.619145-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T125759", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 05:58:04.619145-07	f	generic
54ee535d-5c73-49b5-a784-5365b0f0e7fa	GL5	f7c652b1-51bc-483b-a485-8d70c3630310	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	c8552b03-f549-47c5-962f-bb123d63fe02	2025-07-03 06:00:27.011789-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T130014", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 06:00:27.011789-07	f	generic
056be3d1-ec54-4164-8389-037e32201f10	GL6	0d2b2a11-5fa7-407f-9053-c2933c838bdf	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	c9f38309-e10a-4fae-a3d0-f1f8488f7eae	2025-07-03 06:05:40.421955-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T130538", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 06:05:40.421955-07	f	generic
db96380a-d111-4a49-b605-fd48025f640f	GL7	0d2b2a11-5fa7-407f-9053-c2933c838bdf	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	c98ef464-2308-47c8-aaa7-cd1cee22edd1	2025-07-03 06:05:41.67887-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T130538", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 06:05:41.67887-07	f	generic
0c13ca5b-8813-42af-b226-7b3d2c6ada7c	GL8	0d2b2a11-5fa7-407f-9053-c2933c838bdf	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	25cb2b88-6898-400b-9282-3285aa3cd2e6	2025-07-03 06:05:42.641494-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T130538", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 06:05:42.641494-07	f	generic
21d57f42-7ff7-45ef-b75d-182191c6a5d8	GL9	0d2b2a11-5fa7-407f-9053-c2933c838bdf	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	6f5b9dcf-45e5-43cc-b340-ed644031f886	2025-07-03 06:05:43.745629-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T130538", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 06:05:43.745629-07	f	generic
40d8c3fb-7df1-484a-872a-7d6847061560	GL10	c8552b03-f549-47c5-962f-bb123d63fe02	file:file:generic:1.0	file:shared_ref:generic:1.0	generic_instance_lineage	8c274106-80d4-4ca6-b317-2a1ad9cc1f7c	2025-07-03 06:13:20.915846-07	file:generic:1.0 :: shared_ref:generic:1.0	generic	file	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "backup_2025-07-03_00-06-33.sql", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "sql", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7007.sql", "current_s3_uri": "s3://daylily-dewey-0/0/FI7007.sql", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T130014", "original_file_md5": "", "original_file_name": "backup_2025-07-03_00-06-33.sql", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "sql", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1012876, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	ready	1.0	f	2025-07-03 06:13:20.915846-07	f	presigned
afaa336c-a28e-41ee-a3db-820af546ebcf	GL11	f7c652b1-51bc-483b-a485-8d70c3630310	file:file_set:generic:1.0	file:shared_ref:generic:1.0	generic_instance_lineage	8c274106-80d4-4ca6-b317-2a1ad9cc1f7c	2025-07-03 06:13:20.915846-07	file_set:generic:1.0 :: shared_ref:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T130014", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 06:13:20.915846-07	f	from_set
ffe8488a-73cd-44bb-a54c-1a32202d24c7	GL12	d7e5884d-3ffe-451a-8fbd-99716aff9bb9	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	44f71e98-963f-4ee5-90c1-d4e544a735da	2025-07-03 06:40:42.323173-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T134027", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 06:40:42.323173-07	f	generic
e087e2fe-5053-4cc2-9058-62ff7149a22e	GL13	c54a4608-c3e9-4453-83b3-302ca39467f3	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	5aa56483-8480-4d03-a3f2-36506ae89d6c	2025-07-03 06:41:49.286757-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T134109", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 06:41:49.286757-07	f	generic
ad18c92a-a1c6-47dd-982e-0e5f5cb9ece8	GL14	c54a4608-c3e9-4453-83b3-302ca39467f3	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	9312ed99-a2e8-45a4-9f4d-802f27c08ea5	2025-07-03 06:41:50.40051-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T134109", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 06:41:50.40051-07	f	generic
cf6b66d1-4204-48e0-89b8-a1b3d93e056f	GL15	d77e413a-c96e-4048-b4a7-a5ca4c00828a	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	675337ff-ab88-4ba5-bce0-a80e5446111e	2025-07-03 06:42:13.754991-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T134212", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 06:42:13.754991-07	f	generic
ca812806-312b-4c91-9783-4f3b844afe27	GL16	d77e413a-c96e-4048-b4a7-a5ca4c00828a	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	199bb1af-804b-411a-b0ba-f270dc958bb5	2025-07-03 06:42:14.974898-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T134212", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 06:42:14.974898-07	f	generic
321da091-bd88-482d-81be-9b7b69c4a70d	GL17	d77e413a-c96e-4048-b4a7-a5ca4c00828a	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	b5eca29d-be2c-4c68-8f21-61819fe90f20	2025-07-03 06:42:15.911874-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T134212", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 06:42:15.911874-07	f	generic
6c6fd1b8-10b5-4ffb-820e-e26fe831b976	GL18	d77e413a-c96e-4048-b4a7-a5ca4c00828a	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	7ee95105-e62b-4f4a-aadf-e76bc8bf12d1	2025-07-03 06:42:16.846283-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T134212", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 06:42:16.846283-07	f	generic
540527c1-2db0-4e63-8243-f4df42e12b65	GL19	c9f38309-e10a-4fae-a3d0-f1f8488f7eae	file:file:generic:1.0	file:shared_ref:generic:1.0	generic_instance_lineage	d7eb218e-e19e-402c-b405-ba307911e0ff	2025-07-03 06:42:33.412063-07	file:generic:1.0 :: shared_ref:generic:1.0	generic	file	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7008.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7008.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T130538", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	ready	1.0	f	2025-07-03 06:42:33.412063-07	f	presigned
e7bf76f4-992f-4dca-88b2-dc064be5140f	GL20	0d2b2a11-5fa7-407f-9053-c2933c838bdf	file:file_set:generic:1.0	file:shared_ref:generic:1.0	generic_instance_lineage	d7eb218e-e19e-402c-b405-ba307911e0ff	2025-07-03 06:42:33.412063-07	file_set:generic:1.0 :: shared_ref:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T130538", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "end_operator": {"email": "john@daylilyinformatics.com", "dag_fnv2": ""}, "creating_user": "", "end_timestamp": "2025-07-03 09:06:37 EDT-0400", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [{"email": "john@daylilyinformatics.com", "dag_fnv2": ""}], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "0", "max_executions": "-1", "action_executed": "1", "executed_datetime": ["2025-07-03 09:06:37 EDT-0400"], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	complete	1.0	f	2025-07-03 06:42:33.412063-07	f	from_set
2d8a3441-2f76-4196-b853-d102dded27ec	GL21	c98ef464-2308-47c8-aaa7-cd1cee22edd1	file:file:generic:1.0	file:shared_ref:generic:1.0	generic_instance_lineage	6531f3d5-1217-4d66-93f7-7e40c65d72bf	2025-07-03 06:42:33.460569-07	file:generic:1.0 :: shared_ref:generic:1.0	generic	file	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7009.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7009.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T130538", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	ready	1.0	f	2025-07-03 06:42:33.460569-07	f	presigned
f5e8b463-601f-46eb-8004-1956858c2087	GL22	0d2b2a11-5fa7-407f-9053-c2933c838bdf	file:file_set:generic:1.0	file:shared_ref:generic:1.0	generic_instance_lineage	6531f3d5-1217-4d66-93f7-7e40c65d72bf	2025-07-03 06:42:33.460569-07	file_set:generic:1.0 :: shared_ref:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T130538", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "end_operator": {"email": "john@daylilyinformatics.com", "dag_fnv2": ""}, "creating_user": "", "end_timestamp": "2025-07-03 09:06:37 EDT-0400", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [{"email": "john@daylilyinformatics.com", "dag_fnv2": ""}], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "0", "max_executions": "-1", "action_executed": "1", "executed_datetime": ["2025-07-03 09:06:37 EDT-0400"], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	complete	1.0	f	2025-07-03 06:42:33.460569-07	f	from_set
a8a79f41-7e7d-4177-842f-12b328b26065	GL23	25cb2b88-6898-400b-9282-3285aa3cd2e6	file:file:generic:1.0	file:shared_ref:generic:1.0	generic_instance_lineage	b0c04ae3-a3bd-4721-9834-9cb6ae4f3526	2025-07-03 06:42:33.514661-07	file:generic:1.0 :: shared_ref:generic:1.0	generic	file	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "test_png.png", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "png", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "original_url": "https://raw.githubusercontent.com/Daylily-Informatics/bloom/main/tests/test_png.png", "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7010.png", "current_s3_uri": "s3://daylily-dewey-0/0/FI7010.png", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T130538", "original_file_md5": "", "original_file_name": "test_png.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "url", "original_file_size_bytes": 364817, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	ready	1.0	f	2025-07-03 06:42:33.514661-07	f	presigned
d24ce294-8388-4b74-83ed-7fc017912513	GL24	0d2b2a11-5fa7-407f-9053-c2933c838bdf	file:file_set:generic:1.0	file:shared_ref:generic:1.0	generic_instance_lineage	b0c04ae3-a3bd-4721-9834-9cb6ae4f3526	2025-07-03 06:42:33.514661-07	file_set:generic:1.0 :: shared_ref:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T130538", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "end_operator": {"email": "john@daylilyinformatics.com", "dag_fnv2": ""}, "creating_user": "", "end_timestamp": "2025-07-03 09:06:37 EDT-0400", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [{"email": "john@daylilyinformatics.com", "dag_fnv2": ""}], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "0", "max_executions": "-1", "action_executed": "1", "executed_datetime": ["2025-07-03 09:06:37 EDT-0400"], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	complete	1.0	f	2025-07-03 06:42:33.514661-07	f	from_set
1643f84d-a93c-43a6-bfea-90a8578428e3	GL25	6f5b9dcf-45e5-43cc-b340-ed644031f886	file:file:generic:1.0	file:shared_ref:generic:1.0	generic_instance_lineage	9728d18d-696b-47f7-a548-33e93b36b0f0	2025-07-03 06:42:33.566659-07	file:generic:1.0 :: shared_ref:generic:1.0	generic	file	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"key": "value", "name": "pcluster-project-budget-tags.tsv", "purpose": "Research", "category": "Genomics", "comments": "Sample comments", "lab_code": "LAB123", "study_id": "STUDY001", "variable": "Age", "file_tags": "tag1,tag2", "file_type": "tsv", "patient_id": "PATIENT001", "clinician_id": "CLIN001", "external_ids": [], "sub_category": "DNA", "sub_variable": "Gender", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7011.tsv", "current_s3_uri": "s3://daylily-dewey-0/0/FI7011.tsv", "sub_category_2": "WGS", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/data/budget_tags/pcluster-project-budget-tags.tsv", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "2024-01-01T12:00:00", "import_or_remote": "Import", "upload_group_key": "20250703T130538", "original_file_md5": "", "original_file_name": "pcluster-project-budget-tags.tsv", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "2024-01-01T13:00:00", "original_file_suffix": "tsv", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "s3_uri", "original_file_size_bytes": 232, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	ready	1.0	f	2025-07-03 06:42:33.566659-07	f	presigned
8deb54e7-2959-42b4-80d6-feb2ec638056	GL26	0d2b2a11-5fa7-407f-9053-c2933c838bdf	file:file_set:generic:1.0	file:shared_ref:generic:1.0	generic_instance_lineage	9728d18d-696b-47f7-a548-33e93b36b0f0	2025-07-03 06:42:33.566659-07	file_set:generic:1.0 :: shared_ref:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T130538", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "end_operator": {"email": "john@daylilyinformatics.com", "dag_fnv2": ""}, "creating_user": "", "end_timestamp": "2025-07-03 09:06:37 EDT-0400", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [{"email": "john@daylilyinformatics.com", "dag_fnv2": ""}], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "0", "max_executions": "-1", "action_executed": "1", "executed_datetime": ["2025-07-03 09:06:37 EDT-0400"], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	complete	1.0	f	2025-07-03 06:42:33.566659-07	f	from_set
0e5eb005-a07a-46ec-a552-71c5ecce7dbf	GL27	ce4a70c5-bcb6-4305-9736-b14390e6c4a2	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	0098cfb3-be50-4e6d-8633-fef7cc11846c	2025-07-03 06:43:22.491105-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T134315", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 06:43:22.491105-07	f	generic
de008fbc-7243-4bdb-a8fa-f6761bb05959	GL28	b444546e-7e18-4775-a0a5-6a48aaebd985	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	4f291506-a724-4288-b37b-8a748c6eb737	2025-07-03 06:43:35.9415-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T134326", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 06:43:35.9415-07	f	generic
8bf7dc19-aa7f-420c-a2a6-2348ce9a21fa	GL29	2321b538-2fee-484b-8d18-7b6df27371ea	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	bee9e98b-3b9e-4fe1-8729-1b8fd38bac40	2025-07-03 06:44:51.517314-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T134442", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 06:44:51.517314-07	f	generic
c5c3501a-a482-4872-8cd5-80135dc2a955	GL30	88d739e7-9eb0-4083-af3b-265a8d380792	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	d512ba32-ddef-4c26-acae-5dda9273bab1	2025-07-03 06:46:20.996889-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T134607", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 06:46:20.996889-07	f	generic
00cf7c28-8134-4786-a02f-e5acd0185660	GL31	7e52fd3c-eedc-4e34-8caf-de2f5ad475fa	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	f835a3a0-a807-41d2-9f29-68d149957d8d	2025-07-03 07:26:34.211728-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T142630", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:26:34.211728-07	f	generic
4ef08fac-4436-4aa3-9afa-aba2b643d417	GL32	7e52fd3c-eedc-4e34-8caf-de2f5ad475fa	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	6ad2a1ab-3b91-4c5b-b4e6-69d087c868f0	2025-07-03 07:26:35.660855-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T142630", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:26:35.660855-07	f	generic
e2c2ce43-cead-4515-b697-5be784cfcf97	GL33	7e52fd3c-eedc-4e34-8caf-de2f5ad475fa	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	7b446c7d-69b4-45f5-9320-dff366d18cce	2025-07-03 07:26:36.904618-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T142630", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:26:36.904618-07	f	generic
8e3d71b8-d06e-40bd-a992-15f32dc14778	GL34	7e52fd3c-eedc-4e34-8caf-de2f5ad475fa	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	11568060-9f81-417e-baf9-9e7e56b3a759	2025-07-03 07:26:38.31003-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T142630", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:26:38.31003-07	f	generic
64f3ecb8-9358-43fd-9550-e13eeafb7d63	GL35	1aab709b-8c59-4117-9bcc-48c2bf549a79	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	0886f894-a724-4452-8007-54b3598c84d6	2025-07-03 07:27:21.772941-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T142720", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:27:21.772941-07	f	generic
8bc65425-a974-4a61-9bf8-639aaa6d8fcd	GL36	1aab709b-8c59-4117-9bcc-48c2bf549a79	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	051fd024-70cd-4874-ba13-088a4becc209	2025-07-03 07:27:23.158652-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T142720", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:27:23.158652-07	f	generic
47dde6fa-2e0b-43a5-bc61-0e8b2ecb48b7	GL37	1aab709b-8c59-4117-9bcc-48c2bf549a79	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	5608248c-838c-458c-bbba-a6a0ba464775	2025-07-03 07:27:24.445577-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T142720", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:27:24.445577-07	f	generic
c33c0266-22e6-4a82-a537-78a94d4d274f	GL38	1aab709b-8c59-4117-9bcc-48c2bf549a79	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	ec92091a-246e-472e-914a-66e628bd748b	2025-07-03 07:27:25.851199-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T142720", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:27:25.851199-07	f	generic
bda4f9c3-1d9d-4481-b0dc-f3e61b5377aa	GL39	c47bda8d-6c52-4cf8-83d6-213aaa276e3f	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	96c5276b-4501-460e-80a1-c6506a98e020	2025-07-03 07:30:40.065119-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T143038", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:30:40.065119-07	f	generic
b82352e9-606f-4a0e-bc02-6b29dd67745d	GL40	c47bda8d-6c52-4cf8-83d6-213aaa276e3f	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	c64b5702-5aea-4bf0-856b-ad9d6a0035c4	2025-07-03 07:30:41.005943-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T143038", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:30:41.005943-07	f	generic
105321bc-e952-45e3-b01d-490290c7ee0c	GL41	c47bda8d-6c52-4cf8-83d6-213aaa276e3f	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	44188f89-cb29-4da6-90c3-592a3e96ce82	2025-07-03 07:30:41.948242-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T143038", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:30:41.948242-07	f	generic
2c4895d4-8647-488d-82e2-2714097867d3	GL42	c47bda8d-6c52-4cf8-83d6-213aaa276e3f	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	fd60bad1-ce79-4e33-a561-8db568b1514c	2025-07-03 07:30:42.982424-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T143038", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:30:42.982424-07	f	generic
735fad8d-eac4-4bdd-8552-34ed0007dd55	GL43	99921700-5afa-47c0-aa3b-f6a520dc526b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	ef3752a0-fb6c-4294-ac4e-e0f65976b68a	2025-07-03 07:32:28.834993-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T143216", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:32:28.834993-07	f	generic
b6ee4da9-b42c-435f-95d7-0ef084d6888c	GL44	def0e204-1561-4031-a477-de7549381c84	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	8c44a660-381a-4fd7-8bb3-6093c23991f0	2025-07-03 07:33:44.754931-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "fastqs", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:33:44.754931-07	f	generic
5e1c9097-7b4d-4b95-bff7-048394f6b772	GL45	def0e204-1561-4031-a477-de7549381c84	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	76d325cb-204a-490a-817f-ba7917e86e20	2025-07-03 07:33:46.124552-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "fastqs", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:33:46.124552-07	f	generic
a37f01e7-1f5e-41b2-82ce-99c6ab0891f9	GL46	9ea2e910-4f31-4ed2-b6c4-449c753c3692	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	2158f0a3-f8e1-4bc9-84f7-4620f6fc8f28	2025-07-03 07:34:24.204351-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T143413", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:34:24.204351-07	f	generic
f0e1b391-4ccf-4109-953b-41720336197d	GL47	9ea2e910-4f31-4ed2-b6c4-449c753c3692	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	cf8a5e09-2051-44be-9121-e9a446b85e5d	2025-07-03 07:34:24.520501-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T143413", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:34:24.520501-07	f	generic
5e646d28-0b01-4fc8-95d4-837b4887af5c	GL48	382785ea-95f6-45b9-b63a-71ba7c409c3d	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	47f023f3-25bd-4d38-bbf1-c9465815efe0	2025-07-03 07:34:38.766287-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T143426", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:34:38.766287-07	f	generic
82fa7bdb-2695-41d2-b8e5-46378f282371	GL49	382785ea-95f6-45b9-b63a-71ba7c409c3d	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	471467e2-7d73-4ec1-957c-980027da7ccc	2025-07-03 07:34:39.315097-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T143426", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:34:39.315097-07	f	generic
17fe66cc-a1bb-4cad-846c-f3d85b03cdfb	GL50	382785ea-95f6-45b9-b63a-71ba7c409c3d	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	42090aeb-3afd-498a-b55f-4af3689f2379	2025-07-03 07:34:39.874607-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T143426", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:34:39.874607-07	f	generic
daeb0b62-5fab-4e71-a698-f4460e9ee66f	GL51	7ddba52c-036c-420d-b112-ba9419d50c1d	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	5f8a84d3-a799-47fa-8e6e-e95496c7d320	2025-07-03 07:35:09.126574-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T143457", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:35:09.126574-07	f	generic
8c70110d-9be7-44f3-b54c-7ee15ea3b0ef	GL52	7ddba52c-036c-420d-b112-ba9419d50c1d	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	afd98d90-d88b-4d7f-8384-8af9b64f4550	2025-07-03 07:35:09.825727-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T143457", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:35:09.825727-07	f	generic
7d6bdab9-e67c-4c0c-954e-29e82c47f92b	GL53	7ddba52c-036c-420d-b112-ba9419d50c1d	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	cd291654-0794-44d6-ad84-b147d8b5f61b	2025-07-03 07:35:10.436307-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T143457", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:35:10.436307-07	f	generic
3ce6bc0d-f5fa-4db4-b5f9-e3c2372ad3fc	GL54	d8f83592-c673-4ac6-aef5-b79834139010	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	2975d2a7-516a-4e2d-8149-1835e51641fa	2025-07-03 07:35:25.385411-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T143512", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:35:25.385411-07	f	generic
70dfbfc9-7337-4786-9fcc-908811003123	GL55	d8f83592-c673-4ac6-aef5-b79834139010	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	6c35ea5e-3cf4-45dd-9cc7-b1e1dc8f78a1	2025-07-03 07:35:26.394088-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T143512", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:35:26.394088-07	f	generic
41268426-8f93-40b8-a880-44396be641ad	GL56	d8f83592-c673-4ac6-aef5-b79834139010	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	f8e647ac-e9b7-4af5-ba52-ddac34812049	2025-07-03 07:35:27.2801-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T143512", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:35:27.2801-07	f	generic
af9d13bc-f3c6-4942-b097-afca275ee0dd	GL57	44f71e98-963f-4ee5-90c1-d4e544a735da	file:file:generic:1.0	file:shared_ref:generic:1.0	generic_instance_lineage	d3138620-73a5-40a2-b862-ca9cff5395a9	2025-07-03 07:35:57.37592-07	file:generic:1.0 :: shared_ref:generic:1.0	generic	file	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "backup_2025-07-03_00-06-33.sql", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "sql", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "current_s3_key": "0/FI7012.sql", "current_s3_uri": "s3://daylily-dewey-0/0/FI7012.sql", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "Import", "upload_group_key": "20250703T134027", "original_file_md5": "", "original_file_name": "backup_2025-07-03_00-06-33.sql", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "sql", "current_s3_bucket_name": "daylily-dewey-0", "original_file_data_type": "raw data", "original_file_size_bytes": 1012876, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	ready	1.0	f	2025-07-03 07:35:57.37592-07	f	presigned
ff55ed19-d77e-42e0-92fb-b1367a3f3e83	GL58	d7e5884d-3ffe-451a-8fbd-99716aff9bb9	file:file_set:generic:1.0	file:shared_ref:generic:1.0	generic_instance_lineage	d3138620-73a5-40a2-b862-ca9cff5395a9	2025-07-03 07:35:57.37592-07	file_set:generic:1.0 :: shared_ref:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T134027", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:35:57.37592-07	f	from_set
a35749a3-0485-4a30-80b8-6676da5d722e	GL59	d5b8c535-499d-4586-ae3b-4bf5a9871387	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	82e5c3f3-22bc-4207-b3a5-7b29bd0ae741	2025-07-03 07:37:08.404927-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T143658", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:37:08.404927-07	f	generic
08be9c2a-f237-4a44-b4b7-f630ecf3f8b8	GL60	82e5c3f3-22bc-4207-b3a5-7b29bd0ae741	file:file:generic:1.0	file:shared_ref:generic:1.0	generic_instance_lineage	11d26eb0-da43-4f29-b0c9-c234ca621a60	2025-07-03 07:37:32.832686-07	file:generic:1.0 :: shared_ref:generic:1.0	generic	file	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "day_cmd.log", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "log", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "remote_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "current_s3_key": "FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "current_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/day_cmd.log", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "remote", "upload_group_key": "20250703T143658", "original_file_md5": "", "original_file_name": "day_cmd.log", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "log", "current_s3_bucket_name": "daylily-service-omics-analysis-us-west-2", "original_file_data_type": "s3uri", "original_file_size_bytes": null, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	ready	1.0	f	2025-07-03 07:37:32.832686-07	f	presigned
fbb6d5e9-1e89-412f-b073-ad4485f40718	GL61	d5b8c535-499d-4586-ae3b-4bf5a9871387	file:file_set:generic:1.0	file:shared_ref:generic:1.0	generic_instance_lineage	11d26eb0-da43-4f29-b0c9-c234ca621a60	2025-07-03 07:37:32.832686-07	file_set:generic:1.0 :: shared_ref:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T143658", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:37:32.832686-07	f	from_set
d82bcd53-e48a-4fbc-add8-11809e0c36a0	GL62	5834526f-b5d3-4f42-8384-9c26e5c74bc8	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	dde25f19-7032-40ff-a0e6-8ae39183e84b	2025-07-03 07:39:24.078427-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T143915", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:39:24.078427-07	f	generic
ff98a429-d339-47c9-815a-cab619e20fcc	GL63	dde25f19-7032-40ff-a0e6-8ae39183e84b	file:file:generic:1.0	file:shared_ref:generic:1.0	generic_instance_lineage	21aad253-3d80-4b11-9220-290cf6ff8a34	2025-07-03 07:39:39.360451-07	file:generic:1.0 :: shared_ref:generic:1.0	generic	file	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "main.py", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "py", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "remote_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/main.py", "current_s3_key": "FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/main.py", "current_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/main.py", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/main.py", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "remote", "upload_group_key": "20250703T143915", "original_file_md5": "", "original_file_name": "main.py", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "py", "current_s3_bucket_name": "daylily-service-omics-analysis-us-west-2", "original_file_data_type": "s3uri", "original_file_size_bytes": null, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	ready	1.0	f	2025-07-03 07:39:39.360451-07	f	presigned
3177f944-91c2-4db1-a3e4-91f333ce65fd	GL64	5834526f-b5d3-4f42-8384-9c26e5c74bc8	file:file_set:generic:1.0	file:shared_ref:generic:1.0	generic_instance_lineage	21aad253-3d80-4b11-9220-290cf6ff8a34	2025-07-03 07:39:39.360451-07	file_set:generic:1.0 :: shared_ref:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T143915", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:39:39.360451-07	f	from_set
cfbf10c4-6988-40b1-8360-c61d9fc764b2	GL65	2ff79f5f-1663-4dce-bbf3-13e0cc787c2d	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	1bccdecd-c0cf-41d5-a0b5-13be26cdb88c	2025-07-03 07:40:55.423969-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144046", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:40:55.423969-07	f	generic
808f1552-93c1-40ee-bd40-b31602d8edae	GL66	1bccdecd-c0cf-41d5-a0b5-13be26cdb88c	file:file:generic:1.0	file:shared_ref:generic:1.0	generic_instance_lineage	6182f00d-f0e3-4b74-888f-54f693a6f8fe	2025-07-03 07:41:07.398316-07	file:generic:1.0 :: shared_ref:generic:1.0	generic	file	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "rule_dag.png", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": "", "file_type": "png", "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "john@daylilyinformatics.com", "remote_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/rule_dag.png", "current_s3_key": "FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/rule_dag.png", "current_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/rule_dag.png", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "s3://daylily-service-omics-analysis-us-west-2/FSxLustre20250128T054156Z/analysis_results/ubuntu/giab/hg38_full/daylily/rule_dag.png", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "remote", "upload_group_key": "20250703T144046", "original_file_md5": "", "original_file_name": "rule_dag.png", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "png", "current_s3_bucket_name": "daylily-service-omics-analysis-us-west-2", "original_file_data_type": "s3uri", "original_file_size_bytes": null, "original_local_server_name": ""}, "description": "Generic File", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/download_file/1.0": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Actions", "group_order": "2"}}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	ready	1.0	f	2025-07-03 07:41:07.398316-07	f	presigned
390be49f-6b09-4027-8804-ca78852fd601	GL67	2ff79f5f-1663-4dce-bbf3-13e0cc787c2d	file:file_set:generic:1.0	file:shared_ref:generic:1.0	generic_instance_lineage	6182f00d-f0e3-4b74-888f-54f693a6f8fe	2025-07-03 07:41:07.398316-07	file_set:generic:1.0 :: shared_ref:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144046", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:41:07.398316-07	f	from_set
7c6aa1df-232f-48f8-b5c5-a58cce4b1b6e	GL68	cd99a9ea-224d-4cda-b8a2-2027e4be1767	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	751326b8-0128-4f6d-8f82-05bf70cbd6c8	2025-07-03 07:43:09.55241-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T144308", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:43:09.55241-07	f	generic
e3e7fe3d-2472-4cbe-8837-3abbc8c5124c	GL69	cd99a9ea-224d-4cda-b8a2-2027e4be1767	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	42c52d3e-cb20-4c08-87fe-c36891087f40	2025-07-03 07:43:10.443934-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T144308", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:43:10.443934-07	f	generic
1b86b600-46b4-4f13-906b-7c4f124ee281	GL70	cd99a9ea-224d-4cda-b8a2-2027e4be1767	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	ade4c01a-0859-4437-b6b1-63a99088aaae	2025-07-03 07:43:11.454347-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T144308", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:43:11.454347-07	f	generic
ae6cb85a-3060-4047-b619-78f7a7e101ce	GL71	cd99a9ea-224d-4cda-b8a2-2027e4be1767	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	50d62952-ee77-48c2-adf1-a4fd30234dd6	2025-07-03 07:43:12.588778-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T144308", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:43:12.588778-07	f	generic
af276406-fc93-4896-b3e0-8227c6058a29	GL72	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	f3300273-7c20-42a6-aa2d-101f1f721a6d	2025-07-03 07:46:18.805557-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:18.805557-07	f	generic
72874abc-291e-4e36-89d2-13aa068cd7fb	GL73	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	5b761ebc-9320-4d44-9d95-c80a289eb8ae	2025-07-03 07:46:19.562247-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:19.562247-07	f	generic
f1409a26-5152-433b-a0dd-7a664bd9be6e	GL74	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	c260d8ec-016f-445e-977a-fb012e00481b	2025-07-03 07:46:20.291238-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:20.291238-07	f	generic
c2c9e2e4-8ed4-4737-85d4-fd75871c7f9c	GL75	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	04b20693-2770-4741-ab5e-af1f0f85f28e	2025-07-03 07:46:21.051861-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:21.051861-07	f	generic
efebb8ad-8a01-4e0e-afff-447eb897daed	GL76	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	dc3b6ed6-0ec6-4cde-98e4-5ae861b129fe	2025-07-03 07:46:21.79117-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:21.79117-07	f	generic
bb0894f4-e56c-4a77-a9cd-9881a32ad0c6	GL77	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	3b9591c8-f892-4b39-ba2f-95894da01bd1	2025-07-03 07:46:22.524617-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:22.524617-07	f	generic
87253686-7e6c-4c31-97af-b76cc05839e9	GL78	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	2eb9d931-5a00-44ea-99d0-b50cc6d630f1	2025-07-03 07:46:23.120485-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:23.120485-07	f	generic
fea94b55-19ff-41ec-8dfc-c3c4fe2b28e1	GL79	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	4d7880f2-f6ba-498d-9d36-a065b7c26a62	2025-07-03 07:46:23.857466-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:23.857466-07	f	generic
89861523-d00e-450e-9812-dce1eac3a79e	GL80	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	6033d3d2-b9bb-4ec7-8afd-a90254f310aa	2025-07-03 07:46:24.636402-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:24.636402-07	f	generic
0f10a224-e406-4a00-a3c7-bc402d35fc65	GL81	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	7340a717-a09a-4c3c-9cd6-0a5e52a77978	2025-07-03 07:46:25.359055-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:25.359055-07	f	generic
8bd2a248-5b53-47b5-ba43-d357e6bc7912	GL82	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	611dcea7-1da6-46e9-b035-b9658cef5ab2	2025-07-03 07:46:26.081292-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:26.081292-07	f	generic
c7533e6c-8ade-493f-b39f-af4f02fb5023	GL83	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	b007b95e-b1a6-42c6-bc19-24e1e5f779b9	2025-07-03 07:46:26.847362-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:26.847362-07	f	generic
444ce4d6-7aa1-479c-8f4d-58c0458182fd	GL84	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	2763560b-cabd-424f-a685-df92ca63080c	2025-07-03 07:46:27.532904-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:27.532904-07	f	generic
70d1e8e6-6b6c-4a63-9f2e-4dc40144a27a	GL85	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	cd4284b9-8e2c-4ab9-94e5-2ed1630e323e	2025-07-03 07:46:28.296882-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:28.296882-07	f	generic
1b99eca0-0f78-4139-80a1-9eb7ed70c6ad	GL86	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	81948d72-dcd4-4211-819c-96e65e693d0e	2025-07-03 07:46:29.045366-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:29.045366-07	f	generic
acfe0859-e3b1-4ab1-8d62-76e44bdbf393	GL87	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	77115b00-ac73-4144-80b9-a0bd9b125164	2025-07-03 07:46:29.750202-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:29.750202-07	f	generic
be8debd8-6d59-4539-a841-b09fcfce1cee	GL88	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	ee740bac-1bae-443e-8f54-c4d4d7bf371f	2025-07-03 07:46:30.504104-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:30.504104-07	f	generic
17757fe8-5d54-4261-83b0-2e79cc37403d	GL89	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	cd142cc8-8540-40d1-8b89-0320a209d5f1	2025-07-03 07:46:31.173946-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:31.173946-07	f	generic
4b9a07d8-6a0c-40f0-9018-5a56f314372a	GL90	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	9a54fbb8-c9e1-4d2e-8283-db8e5d7b84fc	2025-07-03 07:46:31.760985-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:31.760985-07	f	generic
b3c10c39-4c5b-4bca-a819-9aa56d472845	GL91	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	16c52e6c-5995-449c-bd88-bd1500aab2c2	2025-07-03 07:46:32.282593-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:32.282593-07	f	generic
9040ac20-4574-440d-aad4-75c340141fed	GL92	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	ac534fb1-c4b3-4163-a549-9d6c18fad16f	2025-07-03 07:46:33.089104-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:33.089104-07	f	generic
4732949c-99fb-4e5d-ae96-8ed0e3da8bd8	GL93	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	9ab24170-9de3-4cf6-929e-9389976f8a03	2025-07-03 07:46:33.791329-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:33.791329-07	f	generic
f2a2e709-b723-4798-b2c1-80e4e7aea581	GL94	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	5e52c4b2-b976-433e-ba92-2aabefb9585e	2025-07-03 07:46:34.387019-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:34.387019-07	f	generic
89caf14b-7c28-4103-a377-83fd4143a672	GL95	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	e391ed08-8608-44ff-abc0-16fe6c7baf3d	2025-07-03 07:46:34.887954-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:34.887954-07	f	generic
66ec8afd-6983-493f-b354-fad5336a0812	GL96	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	bbf86074-7a49-44d5-a85a-36cd8bbf2da6	2025-07-03 07:46:35.526478-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:35.526478-07	f	generic
82122e94-57f1-4c55-b3fa-0040228c1ef0	GL97	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	6968dcc3-4411-486e-a424-89dc7a8a9875	2025-07-03 07:46:36.091874-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:36.091874-07	f	generic
ded88850-1761-431c-80b7-bec5d2bd0a84	GL98	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	f0f17b46-d607-4117-b350-68dd1beb7c59	2025-07-03 07:46:36.927029-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:36.927029-07	f	generic
3f2a57be-c5c3-4cda-b0b8-ebb3a0e78e0a	GL99	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	628aaab0-bc64-42ae-9199-121c2cfae0be	2025-07-03 07:46:37.522901-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:37.522901-07	f	generic
b892a9eb-82cb-4adc-bfc7-ba08c1b4b394	GL100	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	d730ed0a-3439-475d-a55e-031e0f4fac5c	2025-07-03 07:46:38.146625-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:38.146625-07	f	generic
60162a63-a1bd-41e9-be90-3ac38f7dfda2	GL101	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	b1f51d3d-d2a3-48c0-9487-bc386b4a658d	2025-07-03 07:46:38.730951-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:38.730951-07	f	generic
ae6f130a-4d48-4508-a244-b89d0d6b6598	GL102	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	652238e6-836f-4e53-ad34-062159021e26	2025-07-03 07:46:39.29554-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:39.29554-07	f	generic
f64146e4-2eb2-4471-9b13-f2eae183b4f9	GL103	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	a99ca08f-34f7-4f31-b30f-d765444bc221	2025-07-03 07:46:39.834112-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:39.834112-07	f	generic
42f05e96-4756-4763-8b6f-4269b34ac2bf	GL104	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	6eeeffdd-a461-4bb2-ace0-e0f74637b47d	2025-07-03 07:46:40.527301-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:40.527301-07	f	generic
f0b37d08-2344-4cb4-8726-d018de18327d	GL105	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	1f8fc77a-8554-4716-8f6f-bd177139f82f	2025-07-03 07:46:41.380446-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:41.380446-07	f	generic
ab7edb2b-4f52-475e-b151-a597117af5ea	GL106	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	6a8be340-a613-4e02-a89f-e1f09be2be4a	2025-07-03 07:46:42.277334-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:42.277334-07	f	generic
195eee4f-bda2-4dfb-9830-adbad8b15f4a	GL107	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	378516a2-6482-48f1-9e85-e33026c12b21	2025-07-03 07:46:43.173901-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:43.173901-07	f	generic
af171de5-ad67-44f6-aa97-0f5ef910079a	GL108	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	50f413f9-6000-4723-896a-926c67f39d59	2025-07-03 07:46:44.031159-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:44.031159-07	f	generic
f550ad34-3355-4709-8674-2bf3bdc551a2	GL109	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	5dae0e31-2c76-4322-bdfa-2f2d2f239bb2	2025-07-03 07:46:44.855903-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:44.855903-07	f	generic
7bf69843-e848-4997-a763-9a10121c84b2	GL110	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	7828fb01-12d3-4b62-8452-7c919cea3ab9	2025-07-03 07:46:45.725978-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:45.725978-07	f	generic
3250fc4a-f90c-40b2-b636-33b7ca342d0d	GL111	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	e010952a-8da4-4573-9b04-112e6178d9f8	2025-07-03 07:46:46.598144-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:46.598144-07	f	generic
1ab9b618-bc2b-4c0f-a4f0-e027e7177231	GL112	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	e2bd006f-4d32-4a15-b87c-68a3eb36b37e	2025-07-03 07:46:47.516113-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:47.516113-07	f	generic
8c920f6d-9a0a-47d5-9cc4-b7d8a71662ea	GL113	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	ddbd7604-9140-472d-87fe-44279e0463b4	2025-07-03 07:46:48.338904-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:48.338904-07	f	generic
fc3e8ad2-022c-4996-9404-56452b372fe5	GL114	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	16a63367-c957-46fb-aee4-f026757bd059	2025-07-03 07:46:49.187999-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:49.187999-07	f	generic
a6f1a6b1-3a2d-4aa8-9385-193fb4f1b726	GL115	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	205773c7-233a-423b-9b6f-704a210048fc	2025-07-03 07:46:50.044671-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:50.044671-07	f	generic
c862aeba-ff72-4880-bb87-1f8f7df2721a	GL116	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	8b323316-e042-4abb-9e7c-948747452c5e	2025-07-03 07:46:50.870117-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:50.870117-07	f	generic
5513d62f-f16f-446d-9b48-0e09ffc2761c	GL117	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	51f409ba-37e0-46f3-954f-eeb8b60c8886	2025-07-03 07:46:51.716479-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:51.716479-07	f	generic
8fa72706-c3b9-4c81-a9a7-c4764ba8d24f	GL118	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	6be7b5f3-b9a6-436a-8c5a-d011d47eee2d	2025-07-03 07:46:52.478357-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:52.478357-07	f	generic
480f5b69-6f16-41c5-b6c7-e59d3e71b7f4	GL119	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	9a2d33f4-0c38-4b92-b85c-e3650dbf67d5	2025-07-03 07:46:53.083873-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:53.083873-07	f	generic
a5ea33ef-0dba-4dab-a402-65c29efe4d4a	GL120	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	7d5dcd5e-7d18-4094-aab9-64c5da8698ec	2025-07-03 07:46:53.726108-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:53.726108-07	f	generic
98bbaad3-4087-4909-aac4-888bd138c712	GL121	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	bb3c68a8-5a54-4a62-b9e3-d83cfeacf4f6	2025-07-03 07:46:55.436303-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:55.436303-07	f	generic
1088fa50-c5bd-4e91-b106-977c13134093	GL122	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	b32d8a96-efe1-42c3-abe3-7a3241e0c102	2025-07-03 07:46:56.365609-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:56.365609-07	f	generic
74beaaff-cde2-41ba-8f07-4ab53a6ca1a3	GL123	81d233d2-dffb-4b82-9cea-61c8009b5f2b	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	051aa336-0ce7-4f70-bc74-9efbc624fce9	2025-07-03 07:46:57.21458-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144506", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:46:57.21458-07	f	generic
e5f9d28a-edb7-4156-abc9-43bbf215f691	GL124	245c6374-146a-4bf6-898d-e3a0ce08f78d	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	82961788-e3f9-441f-96f3-de01e051f122	2025-07-03 07:49:09.90103-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "on-create", "name": "20250703T144901", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "File set created by Dewey file manager", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:49:09.90103-07	f	generic
99745052-4f67-459f-a147-92668662db3b	GL125	eeee8645-360f-4290-a6f7-a7275d34db61	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	a2bc4b6d-42e0-4993-91ea-cb3b84c406ba	2025-07-03 07:50:09.765023-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T145008", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:50:09.765023-07	f	generic
8bf2431f-d5cf-491f-bf01-d728c249643e	GL126	eeee8645-360f-4290-a6f7-a7275d34db61	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	9379619e-93d1-4e5e-a313-9576af3ffaab	2025-07-03 07:50:10.707766-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T145008", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:50:10.707766-07	f	generic
f8fc87a0-3ca8-4988-93d6-3b3611c8376c	GL127	eeee8645-360f-4290-a6f7-a7275d34db61	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	c01298e9-f706-48dd-b71e-7666927fccf3	2025-07-03 07:50:11.636939-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T145008", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:50:11.636939-07	f	generic
e7149242-57d5-4d9f-9211-d25bccb90e35	GL128	eeee8645-360f-4290-a6f7-a7275d34db61	file:file_set:generic:1.0	file:file:generic:1.0	generic_instance_lineage	9e89f061-44e0-4163-95f7-03de5e7d595b	2025-07-03 07:50:12.86093-07	file_set:generic:1.0 :: file:generic:1.0	generic	file_set	generic	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "20250703T145008", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "Bulk upload", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/add-file/1.0": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/remove-file/1.0": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/file_set/share-presigned/1.0": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "File Set Actions", "group_order": "2"}}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	ready	1.0	f	2025-07-03 07:50:12.86093-07	f	generic
9dc4ad96-63ce-4ec1-942f-be5d6df671e7	GL129	3e50c2f7-5e0b-4ae9-b721-cc81d94c22de	container_instance	container_instance	container_instance_lineage	dacc9c78-e115-4376-8a8c-9869504d64a7	2025-07-03 08:21:02.293073-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.293073-07	f	\N
e20e6d8a-b365-48ba-bac5-69eea0985427	GL130	3e50c2f7-5e0b-4ae9-b721-cc81d94c22de	container_instance	container_instance	container_instance_lineage	438c11ad-e609-4041-aa3e-75e8f719cef9	2025-07-03 08:21:02.30858-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.30858-07	f	\N
15cfe66c-4119-467b-8176-23876dc57dcf	GL131	3e50c2f7-5e0b-4ae9-b721-cc81d94c22de	container_instance	container_instance	container_instance_lineage	f4b245da-a7f5-4a87-b23f-e24f173695d1	2025-07-03 08:21:02.318139-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.318139-07	f	\N
92e7fcc0-f025-4518-a064-87e927e507d5	GL132	3e50c2f7-5e0b-4ae9-b721-cc81d94c22de	container_instance	container_instance	container_instance_lineage	9abd0e47-97f9-4d8e-ac09-cec264f3ee0d	2025-07-03 08:21:02.326916-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.326916-07	f	\N
4cf3557d-3cb8-4572-9720-a3a02f6ff8e2	GL133	3e50c2f7-5e0b-4ae9-b721-cc81d94c22de	container_instance	container_instance	container_instance_lineage	a339e966-d124-4730-bf7b-f27ca87c4e8d	2025-07-03 08:21:02.335164-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.335164-07	f	\N
33f91995-d65d-4c81-bff9-9b37dc831d72	GL134	3e50c2f7-5e0b-4ae9-b721-cc81d94c22de	container_instance	container_instance	container_instance_lineage	486a95ec-28d0-4b5a-b33e-0de10f65806c	2025-07-03 08:21:02.343198-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.343198-07	f	\N
952499b9-8efd-4df2-a63c-96321b52cfbd	GL135	3e50c2f7-5e0b-4ae9-b721-cc81d94c22de	container_instance	container_instance	container_instance_lineage	dee275bb-a13a-455a-879c-29bcddda4116	2025-07-03 08:21:02.351211-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.351211-07	f	\N
0a698f74-6ffd-470a-a995-96c5fbcbce67	GL136	3e50c2f7-5e0b-4ae9-b721-cc81d94c22de	container_instance	container_instance	container_instance_lineage	b79a21a8-b8d0-4b24-8b54-01005554bf43	2025-07-03 08:21:02.359286-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.359286-07	f	\N
3e26d0cc-9cf4-41a0-ae0e-2035e6f13df8	GL137	3e50c2f7-5e0b-4ae9-b721-cc81d94c22de	container_instance	container_instance	container_instance_lineage	ad8e242c-7636-42b1-9284-401ea589e648	2025-07-03 08:21:02.366841-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.366841-07	f	\N
62156688-56dd-4c18-996a-1cf187c3f186	GL138	3e50c2f7-5e0b-4ae9-b721-cc81d94c22de	container_instance	container_instance	container_instance_lineage	84bbdac6-5ab1-480c-8715-8dd7271b19bf	2025-07-03 08:21:02.374692-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.374692-07	f	\N
7c4c38ac-e94b-4152-9f9d-b55dec1291a1	GL139	3e50c2f7-5e0b-4ae9-b721-cc81d94c22de	container_instance	container_instance	container_instance_lineage	a1e066c9-b6ad-4644-a288-029b6337c599	2025-07-03 08:21:02.381424-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.381424-07	f	\N
b496e277-846f-48d0-9473-a24c303cf372	GL140	3e50c2f7-5e0b-4ae9-b721-cc81d94c22de	container_instance	container_instance	container_instance_lineage	4b07afc5-df46-4a85-8d24-2e74bb943460	2025-07-03 08:21:02.388919-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.388919-07	f	\N
8ee8f584-a084-4e05-9d27-c472da7df807	GL141	3e50c2f7-5e0b-4ae9-b721-cc81d94c22de	container_instance	container_instance	container_instance_lineage	01c9d691-4dd1-4a31-88b8-328c9e47a8ec	2025-07-03 08:21:02.398253-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.398253-07	f	\N
3735d12b-7137-4e39-b871-89926008d49f	GL142	3e50c2f7-5e0b-4ae9-b721-cc81d94c22de	container_instance	container_instance	container_instance_lineage	e3a263a5-ad0e-42e5-90f7-3eea9eb68be0	2025-07-03 08:21:02.407525-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.407525-07	f	\N
f1438358-2b8f-4b56-9f5b-88098a9cd8dd	GL143	3e50c2f7-5e0b-4ae9-b721-cc81d94c22de	container_instance	container_instance	container_instance_lineage	005041ba-80b0-494a-9982-4f02023c76b9	2025-07-03 08:21:02.413998-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.413998-07	f	\N
203d92b9-a022-4355-8833-7446c5c90db4	GL144	3e50c2f7-5e0b-4ae9-b721-cc81d94c22de	container_instance	container_instance	container_instance_lineage	d0e326c1-0ca3-40e7-a7a3-58add70136f4	2025-07-03 08:21:02.421024-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.421024-07	f	\N
0ec8e7dd-8845-47f7-9cd9-efaf356cad19	GL145	3e50c2f7-5e0b-4ae9-b721-cc81d94c22de	container_instance	container_instance	container_instance_lineage	6413aa15-e457-4196-a660-67d30c22a768	2025-07-03 08:21:02.427792-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.427792-07	f	\N
b3b2f32a-fdbf-46d1-bd83-a426b3b8498b	GL146	3e50c2f7-5e0b-4ae9-b721-cc81d94c22de	container_instance	container_instance	container_instance_lineage	aabb851f-9f1b-44e8-86e2-2239eb90be7b	2025-07-03 08:21:02.435195-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.435195-07	f	\N
38e5e199-6846-415d-97f6-85e795b8daf5	GL147	3e50c2f7-5e0b-4ae9-b721-cc81d94c22de	container_instance	container_instance	container_instance_lineage	47df2675-749e-4933-b64a-98e9f92f9891	2025-07-03 08:21:02.465875-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.465875-07	f	\N
0f7d966a-51dd-4977-b50f-11513f15cdbb	GL148	3e50c2f7-5e0b-4ae9-b721-cc81d94c22de	container_instance	container_instance	container_instance_lineage	1d487141-743b-4f01-ac33-bbdd81756836	2025-07-03 08:21:02.50501-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.50501-07	f	\N
226492e2-9fbe-4157-829c-b326a3dcfff9	GL149	3e50c2f7-5e0b-4ae9-b721-cc81d94c22de	container_instance	container_instance	container_instance_lineage	d2e4093e-86f0-4c14-b4de-2e90a1f635a5	2025-07-03 08:21:02.514912-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.514912-07	f	\N
1f2797a2-3440-4e7a-85fa-82a15f6cca11	GL150	3e50c2f7-5e0b-4ae9-b721-cc81d94c22de	container_instance	container_instance	container_instance_lineage	2ec41575-3c9e-41db-a210-0aa45aa7c05e	2025-07-03 08:21:02.520696-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.520696-07	f	\N
082a80b0-d445-4f92-af95-269f521cf314	GL151	3e50c2f7-5e0b-4ae9-b721-cc81d94c22de	container_instance	container_instance	container_instance_lineage	b5b21e36-376a-4491-b698-452597c517e8	2025-07-03 08:21:02.525924-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.525924-07	f	\N
4e5e2d5d-b9a4-4093-a46e-dd0a811cd911	GL152	3e50c2f7-5e0b-4ae9-b721-cc81d94c22de	container_instance	container_instance	container_instance_lineage	23f00e77-1469-4ea9-a978-11a2da61dd29	2025-07-03 08:21:02.531634-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.531634-07	f	\N
800ca384-0363-473f-84b6-bec48c07906d	GL153	3e50c2f7-5e0b-4ae9-b721-cc81d94c22de	container_instance	container_instance	container_instance_lineage	e4e0fd64-c65d-4cc7-8fa8-1a052ae78bcf	2025-07-03 08:21:02.537114-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.537114-07	f	\N
57a46487-2d86-4390-9ec1-e291ca334663	GL154	3e50c2f7-5e0b-4ae9-b721-cc81d94c22de	container_instance	container_instance	container_instance_lineage	9ad9dea4-053c-4429-83e2-de60c13f4f67	2025-07-03 08:21:02.542464-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.542464-07	f	\N
2b8f67e1-00ee-437f-be89-571705f8eabc	GL155	3e50c2f7-5e0b-4ae9-b721-cc81d94c22de	container_instance	container_instance	container_instance_lineage	d13c47a9-1614-4d77-a15d-1872be28308e	2025-07-03 08:21:02.547786-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.547786-07	f	\N
2bac7721-51a7-415d-95e6-f1fb15adef8e	GL156	3e50c2f7-5e0b-4ae9-b721-cc81d94c22de	container_instance	container_instance	container_instance_lineage	248543ec-5f5f-4421-879c-11e5899229fa	2025-07-03 08:21:02.553092-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.553092-07	f	\N
492d980e-25a8-46c3-9b87-75d557ba1cac	GL157	dacc9c78-e115-4376-8a8c-9869504d64a7	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	73f06c89-2ce5-4e44-a22e-9441cd06dc0d	2025-07-03 08:21:02.557162-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.557162-07	f	generic
47c42ef6-61a9-4a61-a897-95eb26e74770	GL191	2fc73838-fb40-4247-9446-971fd96cff46	container_instance	container_instance	container_instance_lineage	a0eef834-450f-4780-8b5c-7c1db8b7ea67	2025-07-03 08:21:02.727093-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.727093-07	f	\N
b1b45f70-b597-4133-9046-681cc3b0dd8d	GL192	2fc73838-fb40-4247-9446-971fd96cff46	container_instance	container_instance	container_instance_lineage	64d6d4da-233a-483b-b150-f76b15678000	2025-07-03 08:21:02.732411-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.732411-07	f	\N
5d40c347-7c01-4135-8d6e-a07001f5b902	GL193	2fc73838-fb40-4247-9446-971fd96cff46	container_instance	container_instance	container_instance_lineage	02114dfa-d423-42f2-adf7-65becb5f1255	2025-07-03 08:21:02.738142-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.738142-07	f	\N
07c15c91-54ca-4638-8729-15abe16442d9	GL194	2fc73838-fb40-4247-9446-971fd96cff46	container_instance	container_instance	container_instance_lineage	ee64ced1-19ac-4f65-844c-7b996b0eda0f	2025-07-03 08:21:02.743496-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.743496-07	f	\N
d09f5f2e-1cce-425a-b20d-3e971b0575fd	GL158	438c11ad-e609-4041-aa3e-75e8f719cef9	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	fa310d7e-c727-402a-9c6c-47eafd01096c	2025-07-03 08:21:02.562107-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.562107-07	f	generic
33d3182a-a8d7-439a-a752-6ba1d8219778	GL159	f4b245da-a7f5-4a87-b23f-e24f173695d1	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	155d76ca-0ef6-4610-8991-3caf5eebcaf9	2025-07-03 08:21:02.566978-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.566978-07	f	generic
926a9448-460d-4f6b-a4f9-e26815314305	GL160	9abd0e47-97f9-4d8e-ac09-cec264f3ee0d	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	7f61478d-3e5c-47b1-a942-508cfe92ce65	2025-07-03 08:21:02.571207-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.571207-07	f	generic
09b81e88-6598-4ce5-81a2-206f8e0a9a46	GL161	a339e966-d124-4730-bf7b-f27ca87c4e8d	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	d7eb32c1-6513-4a8e-b241-0bba13be7302	2025-07-03 08:21:02.575611-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.575611-07	f	generic
94fa20cd-f1a2-441b-84b1-f8e71b11144f	GL162	486a95ec-28d0-4b5a-b33e-0de10f65806c	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	1122f95f-07f3-4153-b7b9-eb3710b3370e	2025-07-03 08:21:02.579544-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.579544-07	f	generic
b76e6d69-7def-42ad-98a1-1c1500039235	GL163	dee275bb-a13a-455a-879c-29bcddda4116	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	ee6c92d8-88fe-4743-af74-1ea3834c8407	2025-07-03 08:21:02.584966-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.584966-07	f	generic
4e88924c-588b-47d8-9221-72432cbbec27	GL164	b79a21a8-b8d0-4b24-8b54-01005554bf43	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	20150c7c-2e35-4810-8f6c-0e7779530fac	2025-07-03 08:21:02.591786-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.591786-07	f	generic
abbdef57-8708-42b4-9dcc-ee3591a5ce2e	GL165	ad8e242c-7636-42b1-9284-401ea589e648	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	25cad9ef-1267-48b7-9a19-239743a8caea	2025-07-03 08:21:02.596373-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.596373-07	f	generic
815f0362-3cc3-4d9d-a0e8-a3e79c963c64	GL166	84bbdac6-5ab1-480c-8715-8dd7271b19bf	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	e04744fc-bc25-486c-a066-89db9af94e92	2025-07-03 08:21:02.600841-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.600841-07	f	generic
773ed0fe-4337-4821-b638-76ce2a6d994e	GL167	a1e066c9-b6ad-4644-a288-029b6337c599	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	b9dbd159-998b-4db7-98ba-fb4f220cd04b	2025-07-03 08:21:02.605545-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.605545-07	f	generic
cf9b3d21-6aee-456b-b30b-daa1c8c0ab84	GL168	4b07afc5-df46-4a85-8d24-2e74bb943460	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	eea51634-39be-405d-b1e1-f2b229f006d2	2025-07-03 08:21:02.610022-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.610022-07	f	generic
947c868a-17cf-433a-8327-a29f7629f8d3	GL169	01c9d691-4dd1-4a31-88b8-328c9e47a8ec	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	0e1fd8ab-89b9-4237-8f0d-86d844c9c3d9	2025-07-03 08:21:02.614709-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.614709-07	f	generic
5052c3dc-b03b-463c-bec8-6e5bcdacc2f4	GL170	e3a263a5-ad0e-42e5-90f7-3eea9eb68be0	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	6e9a2e5a-f406-43d6-b310-2357457c94db	2025-07-03 08:21:02.619075-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.619075-07	f	generic
ff72f753-ba2a-480e-ba40-825a5b2313b1	GL171	005041ba-80b0-494a-9982-4f02023c76b9	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	3c3c3a45-565a-4e97-9854-2702e0725454	2025-07-03 08:21:02.623871-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.623871-07	f	generic
b23eed0b-8300-422c-a96e-37e46b7ff159	GL172	d0e326c1-0ca3-40e7-a7a3-58add70136f4	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	3fddfa66-2e48-47d6-8b04-a54d3c08aa2f	2025-07-03 08:21:02.628909-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.628909-07	f	generic
8ff2b226-0fb7-4f02-8478-a0f269d8a809	GL173	6413aa15-e457-4196-a660-67d30c22a768	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	94af7ed1-6a16-463c-9cce-442d3344703f	2025-07-03 08:21:02.633227-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.633227-07	f	generic
8e1bc899-9b88-4ba3-a783-e243843f01f4	GL174	aabb851f-9f1b-44e8-86e2-2239eb90be7b	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	d9adf020-6a54-47b1-ae11-60769eb81e1d	2025-07-03 08:21:02.637441-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.637441-07	f	generic
f41f629e-60f5-467d-8af0-5beffba4a517	GL175	47df2675-749e-4933-b64a-98e9f92f9891	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	f3dfc3a9-b905-4286-9e05-ba68f5762d52	2025-07-03 08:21:02.642032-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.642032-07	f	generic
b723df62-3595-4170-965e-801bf4ff97a9	GL176	1d487141-743b-4f01-ac33-bbdd81756836	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	130e9a38-6071-44ea-b12c-0ac555011c7d	2025-07-03 08:21:02.646617-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.646617-07	f	generic
ec9e080d-ef8a-4fe8-af23-2805090191bb	GL177	d2e4093e-86f0-4c14-b4de-2e90a1f635a5	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	add01ba4-438a-48a5-aa94-62b29f76862b	2025-07-03 08:21:02.650909-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.650909-07	f	generic
8dd330c3-9d9a-4849-a349-a69079627dba	GL178	2ec41575-3c9e-41db-a210-0aa45aa7c05e	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	ce5401cf-dae7-4b77-988d-a7200152e343	2025-07-03 08:21:02.655356-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.655356-07	f	generic
17d06cc1-5d17-477f-99e1-b734663d9520	GL179	b5b21e36-376a-4491-b698-452597c517e8	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	7d46d938-2e29-4a0e-b9ce-c132e7db46e6	2025-07-03 08:21:02.660099-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.660099-07	f	generic
94c423d6-8770-44ec-b075-685b53f7ee45	GL180	23f00e77-1469-4ea9-a978-11a2da61dd29	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	a0017668-1487-4923-bf46-02ad0e4adae5	2025-07-03 08:21:02.664624-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.664624-07	f	generic
7f6140fa-6011-44e2-ba73-95d1654bad7e	GL181	e4e0fd64-c65d-4cc7-8fa8-1a052ae78bcf	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	0e8b6bac-2a70-473a-a382-8230c530bc86	2025-07-03 08:21:02.668815-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.668815-07	f	generic
69b4f8ff-fbbe-4a23-938e-27a8f837c23d	GL182	9ad9dea4-053c-4429-83e2-de60c13f4f67	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	b3c512e2-3442-47bf-a7ae-228de3b99ea8	2025-07-03 08:21:02.672962-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.672962-07	f	generic
a3f7dc99-e0bb-458d-a8d6-a1fb2260d74e	GL183	d13c47a9-1614-4d77-a15d-1872be28308e	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	bf8c8f2e-8303-4213-a40f-25a7cce691d0	2025-07-03 08:21:02.679104-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.679104-07	f	generic
5c1b0cf0-f7e1-4556-885d-c258129f2d93	GL184	248543ec-5f5f-4421-879c-11e5899229fa	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	848b30bc-7457-4760-8ed6-4733e1c9bceb	2025-07-03 08:21:02.685324-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.685324-07	f	generic
44ad6c97-bef2-47da-8ed3-7319f5e5a742	GL185	2fc73838-fb40-4247-9446-971fd96cff46	container_instance	container_instance	container_instance_lineage	b98e9974-de3d-498d-8371-53ec167ca66f	2025-07-03 08:21:02.69569-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.69569-07	f	\N
8a363d88-87bf-411b-9738-90b7e3aac55b	GL186	2fc73838-fb40-4247-9446-971fd96cff46	container_instance	container_instance	container_instance_lineage	a68a2de5-45b2-4db5-873f-eb0708d9b999	2025-07-03 08:21:02.701299-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.701299-07	f	\N
6154ccf3-52a5-4582-a1df-5d877ae9ac48	GL187	2fc73838-fb40-4247-9446-971fd96cff46	container_instance	container_instance	container_instance_lineage	5653c452-7c65-4fbf-a63a-555795d12dfd	2025-07-03 08:21:02.706361-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.706361-07	f	\N
e27c1763-c034-47a7-823e-c177b1630ceb	GL188	2fc73838-fb40-4247-9446-971fd96cff46	container_instance	container_instance	container_instance_lineage	6ebf400e-84b9-48d3-9d87-05463e046efb	2025-07-03 08:21:02.711446-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.711446-07	f	\N
a9b23622-9e25-4fde-997c-6e5ae8605847	GL189	2fc73838-fb40-4247-9446-971fd96cff46	container_instance	container_instance	container_instance_lineage	490088ed-6d23-4654-9f8e-032e607177cf	2025-07-03 08:21:02.716431-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.716431-07	f	\N
2d05e9db-7e43-48cc-a065-08054f89147e	GL190	2fc73838-fb40-4247-9446-971fd96cff46	container_instance	container_instance	container_instance_lineage	f8b71f68-c515-4c50-9d0e-c4dff9409c13	2025-07-03 08:21:02.721585-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.721585-07	f	\N
2852f056-f0d6-421c-b344-fcc6f0bd0bff	GL195	2fc73838-fb40-4247-9446-971fd96cff46	container_instance	container_instance	container_instance_lineage	7842b714-574f-4cb9-a83a-3eb1e194767a	2025-07-03 08:21:02.749097-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.749097-07	f	\N
2a18e5a8-fe8f-4144-a6bd-ad89f5db9f53	GL196	2fc73838-fb40-4247-9446-971fd96cff46	container_instance	container_instance	container_instance_lineage	d4573a76-c4f1-4ed7-9424-90fdadead6a9	2025-07-03 08:21:02.754263-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.754263-07	f	\N
6babb41b-8acb-48d1-b062-c9551c09734f	GL197	2fc73838-fb40-4247-9446-971fd96cff46	container_instance	container_instance	container_instance_lineage	f37b4964-1b4a-44e9-b71e-faa7ff7a0bb8	2025-07-03 08:21:02.759768-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.759768-07	f	\N
a4e2f96e-ba92-40c2-bf66-5b5f1605c8e1	GL198	2fc73838-fb40-4247-9446-971fd96cff46	container_instance	container_instance	container_instance_lineage	db55beaf-f629-47b6-bf32-d6585fc2f1ca	2025-07-03 08:21:02.767064-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.767064-07	f	\N
34749d2e-9c24-49a3-bcde-0cd627320294	GL199	2fc73838-fb40-4247-9446-971fd96cff46	container_instance	container_instance	container_instance_lineage	26349cf2-a95d-4895-a58b-3226f432bfa7	2025-07-03 08:21:02.776932-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.776932-07	f	\N
673f53a6-8f9a-458f-9695-698756484132	GL200	2fc73838-fb40-4247-9446-971fd96cff46	container_instance	container_instance	container_instance_lineage	a237b8a6-236f-4af0-94cc-2ccac3708511	2025-07-03 08:21:02.784106-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.784106-07	f	\N
52853f6b-81be-414c-aa56-4fff379d8cf0	GL201	2fc73838-fb40-4247-9446-971fd96cff46	container_instance	container_instance	container_instance_lineage	247ddbff-98e3-4459-b22f-b4e7df6ea651	2025-07-03 08:21:02.790176-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.790176-07	f	\N
76023953-9a1e-4063-9425-5f905ba3f229	GL202	2fc73838-fb40-4247-9446-971fd96cff46	container_instance	container_instance	container_instance_lineage	e37cde14-ce6c-4cb8-a239-67ebf9db6a5a	2025-07-03 08:21:02.795569-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.795569-07	f	\N
de84edea-b8f7-4b00-825b-053219c05727	GL203	2fc73838-fb40-4247-9446-971fd96cff46	container_instance	container_instance	container_instance_lineage	0eaa618c-4ac4-4704-b7b0-95b542240475	2025-07-03 08:21:02.801109-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.801109-07	f	\N
26c05e2c-0945-4aad-b974-992e655c6fde	GL204	2fc73838-fb40-4247-9446-971fd96cff46	container_instance	container_instance	container_instance_lineage	3bd67b36-e968-4875-95ce-1eb8a711d47a	2025-07-03 08:21:02.806355-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.806355-07	f	\N
b8d55808-a282-47b1-9f9c-75d2946bf11d	GL205	2fc73838-fb40-4247-9446-971fd96cff46	container_instance	container_instance	container_instance_lineage	9104381b-76f1-478f-bf41-baaa240f776c	2025-07-03 08:21:02.811933-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.811933-07	f	\N
50f06a3a-15ad-486b-8401-2ba99d5683a1	GL206	2fc73838-fb40-4247-9446-971fd96cff46	container_instance	container_instance	container_instance_lineage	b80d221a-eb76-431b-bc14-78ef538ad39e	2025-07-03 08:21:02.817355-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.817355-07	f	\N
1ab8f294-117c-4879-a40e-924c07298d53	GL207	2fc73838-fb40-4247-9446-971fd96cff46	container_instance	container_instance	container_instance_lineage	62874067-fe2e-4ba5-a527-2c53ef5166c8	2025-07-03 08:21:02.822872-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.822872-07	f	\N
73eb904b-2928-4580-8568-9ce82906705b	GL208	2fc73838-fb40-4247-9446-971fd96cff46	container_instance	container_instance	container_instance_lineage	ff19a148-c912-4b79-a416-d77eb17f4c8c	2025-07-03 08:21:02.828239-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.828239-07	f	\N
a833636f-76b0-4ea8-81f2-bcea7cc2c27f	GL209	2fc73838-fb40-4247-9446-971fd96cff46	container_instance	container_instance	container_instance_lineage	981a109f-2b0d-464c-ba8c-3ab530503986	2025-07-03 08:21:02.833692-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.833692-07	f	\N
2ab06d6c-2d35-45bc-947f-8375fabc287f	GL210	2fc73838-fb40-4247-9446-971fd96cff46	container_instance	container_instance	container_instance_lineage	371656ae-8d28-434f-86a7-21cc03c87d92	2025-07-03 08:21:02.838828-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.838828-07	f	\N
d71443dc-a34a-48f5-9c2e-0b65da03c707	GL211	2fc73838-fb40-4247-9446-971fd96cff46	container_instance	container_instance	container_instance_lineage	7554c1d0-df29-4789-b007-c08a1d9f22e1	2025-07-03 08:21:02.843892-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.843892-07	f	\N
20d98b3c-6fb6-49f0-929c-621a3a3e0448	GL212	2fc73838-fb40-4247-9446-971fd96cff46	container_instance	container_instance	container_instance_lineage	42ed148e-fc5f-4bfc-a277-2bfdc42d1941	2025-07-03 08:21:02.848923-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:02.848923-07	f	\N
52ec52f5-0695-4df6-b0c7-e3f502c343fb	GL213	b98e9974-de3d-498d-8371-53ec167ca66f	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	de59347a-8518-43d1-8141-7314cb217b54	2025-07-03 08:21:02.852684-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.852684-07	f	generic
923970ef-2781-4273-a175-761c91dabee8	GL259	c5f7422b-9ae1-4496-a645-28e6decfdcd9	container_instance	container_instance	container_instance_lineage	ecc31348-3e43-46dc-b0d9-c81f9ba47938	2025-07-03 08:21:03.15928-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:03.15928-07	f	\N
4e09b61d-bcf7-4dd5-9abc-a2663c96e8c7	GL260	c5f7422b-9ae1-4496-a645-28e6decfdcd9	container_instance	container_instance	container_instance_lineage	f8ac8009-49e0-4a5d-a1d2-6dd107e496e7	2025-07-03 08:21:03.165817-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:03.165817-07	f	\N
c60bfba3-c498-4612-af14-0dd994a6470a	GL214	a68a2de5-45b2-4db5-873f-eb0708d9b999	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	70e6974a-4889-42b3-aa50-1dc5162a753a	2025-07-03 08:21:02.857377-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.857377-07	f	generic
0b161150-52f2-4580-9feb-cb6af2f34acf	GL215	5653c452-7c65-4fbf-a63a-555795d12dfd	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	6bb5fc9e-205d-4cf1-aaa7-b82e122bb446	2025-07-03 08:21:02.861318-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.861318-07	f	generic
811021ad-2a1f-45af-8b66-9c09c7e1d0d8	GL216	6ebf400e-84b9-48d3-9d87-05463e046efb	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	d97fc469-cfcf-42f6-b9f4-c694040b88c4	2025-07-03 08:21:02.866124-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.866124-07	f	generic
fa3702d0-b8ea-4bd5-ae46-8dfd9d498456	GL217	490088ed-6d23-4654-9f8e-032e607177cf	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	f25b65aa-479f-4c10-b039-ecf50ea2340a	2025-07-03 08:21:02.870652-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.870652-07	f	generic
582ac088-6f0f-4c81-a954-107b621eb12e	GL218	f8b71f68-c515-4c50-9d0e-c4dff9409c13	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	9fc34612-8c1e-42b4-a30b-c95060c031ab	2025-07-03 08:21:02.877972-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.877972-07	f	generic
ffe2f273-ef35-4502-bd7b-1c8efd609f16	GL219	a0eef834-450f-4780-8b5c-7c1db8b7ea67	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	c5783a54-74bf-4417-b8d6-cfa029a644ec	2025-07-03 08:21:02.882499-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.882499-07	f	generic
fe5dea71-c542-452a-9eeb-c07775028458	GL220	64d6d4da-233a-483b-b150-f76b15678000	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	7af7bee4-1b4e-4357-84a6-13e465094334	2025-07-03 08:21:02.886662-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.886662-07	f	generic
d87aaabd-2480-4f6a-b772-a019a906503f	GL221	02114dfa-d423-42f2-adf7-65becb5f1255	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	5932edad-b324-4826-b79a-4ff76745d991	2025-07-03 08:21:02.891032-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.891032-07	f	generic
7e9a729c-47ca-484e-8e76-eab9adef2fdc	GL222	ee64ced1-19ac-4f65-844c-7b996b0eda0f	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	da26ed31-1065-4604-8b23-091ce21c1a2f	2025-07-03 08:21:02.895251-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.895251-07	f	generic
46382e63-9abd-4b81-9ef9-27d6a97a295a	GL223	7842b714-574f-4cb9-a83a-3eb1e194767a	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	d6222668-162e-4c2d-b085-2ebbc2e71828	2025-07-03 08:21:02.900199-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.900199-07	f	generic
93f9b7b4-f8ce-4cee-9283-732312d1169e	GL224	d4573a76-c4f1-4ed7-9424-90fdadead6a9	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	67777ee6-6878-4749-b64d-57f9eaa72595	2025-07-03 08:21:02.904584-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.904584-07	f	generic
3eaac0cb-57fd-4f7a-bd2f-3b5b0b694152	GL225	f37b4964-1b4a-44e9-b71e-faa7ff7a0bb8	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	d57dd303-b4c6-4710-b3ef-6de6113f5106	2025-07-03 08:21:02.909067-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.909067-07	f	generic
e512746a-c20f-4fac-b008-c4f72f3676d8	GL226	db55beaf-f629-47b6-bf32-d6585fc2f1ca	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	d9c14341-79eb-4fc7-b30f-5c19d96ca9c2	2025-07-03 08:21:02.913536-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.913536-07	f	generic
ef2e2bb9-d7be-4366-9228-bcd8cdbd1f30	GL227	26349cf2-a95d-4895-a58b-3226f432bfa7	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	7352677a-92c8-4835-901a-525861538e8e	2025-07-03 08:21:02.918242-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.918242-07	f	generic
3c82adc9-90f3-4c77-a4c1-859d21e934db	GL228	a237b8a6-236f-4af0-94cc-2ccac3708511	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	a330becf-ac06-416e-957d-ac2852be9ea0	2025-07-03 08:21:02.922626-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.922626-07	f	generic
41c488f6-b789-4431-ad59-3e865965fa45	GL229	247ddbff-98e3-4459-b22f-b4e7df6ea651	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	a887fcca-84b8-4a78-8bfb-c400a989c31a	2025-07-03 08:21:02.926982-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.926982-07	f	generic
c1aed6a3-140e-4484-a094-17f4e342418d	GL230	e37cde14-ce6c-4cb8-a239-67ebf9db6a5a	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	5d5e68be-5a3d-4089-a30b-5d7bbbe4a2a7	2025-07-03 08:21:02.931525-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.931525-07	f	generic
16d98e3a-cb00-4c8a-9457-d2c8b473cf5d	GL231	0eaa618c-4ac4-4704-b7b0-95b542240475	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	ccc52ad7-c722-4a84-b999-5f97f4d53ffb	2025-07-03 08:21:02.935814-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.935814-07	f	generic
05975054-b74d-4f4d-8b2b-46ee393fa5f7	GL232	3bd67b36-e968-4875-95ce-1eb8a711d47a	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	1e7602c7-c553-4988-8eb4-014fe5e4c3ef	2025-07-03 08:21:02.940253-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.940253-07	f	generic
9bd72e8b-3b17-4340-b9b5-94acef1223b7	GL233	9104381b-76f1-478f-bf41-baaa240f776c	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	db082c31-af48-4101-96af-118560adac8e	2025-07-03 08:21:02.944649-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.944649-07	f	generic
f1004776-858c-4abb-beae-0ca718aff719	GL234	b80d221a-eb76-431b-bc14-78ef538ad39e	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	77754c02-be5c-43cb-b291-8ebdeefc4c2b	2025-07-03 08:21:02.94904-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.94904-07	f	generic
490193a3-525a-4aa2-9895-913a08e17ff6	GL235	62874067-fe2e-4ba5-a527-2c53ef5166c8	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	8a472e60-5250-43b5-914b-d1bfe5a47927	2025-07-03 08:21:02.953494-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.953494-07	f	generic
a6626ad6-cbe4-4085-9a33-bc601df27938	GL236	ff19a148-c912-4b79-a416-d77eb17f4c8c	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	d5438a6a-5b90-40be-a73d-c418e343a0d5	2025-07-03 08:21:02.957776-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.957776-07	f	generic
cfe1b78e-ffad-4337-be90-af1c9d06c9e7	GL237	981a109f-2b0d-464c-ba8c-3ab530503986	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	f1bd3edd-8879-473c-a1ad-d70687254885	2025-07-03 08:21:02.962054-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.962054-07	f	generic
9357d737-db54-47c5-baeb-8e0ab20dacf5	GL238	371656ae-8d28-434f-86a7-21cc03c87d92	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	a067d8df-fe65-4937-a679-ce28eac8f132	2025-07-03 08:21:02.966483-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.966483-07	f	generic
2d826236-66eb-4747-b86e-4269cef09d35	GL239	7554c1d0-df29-4789-b007-c08a1d9f22e1	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	8070ee33-11a2-441b-8f99-11df1436e5af	2025-07-03 08:21:02.978671-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.978671-07	f	generic
bc3fe620-0fa6-42f9-b3d7-689d16400cfe	GL240	42ed148e-fc5f-4bfc-a277-2bfdc42d1941	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	a2e7703b-a781-4d45-9775-6e3812f6f5c7	2025-07-03 08:21:02.98473-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:02.98473-07	f	generic
65dda973-c896-4996-b9a2-395234b91dc5	GL241	d4b6618e-692a-4952-b7f5-164c2c7e45a2	container:tube:tube-generic-10ml:1.0	content:control:giab-HG002:1.0	generic_instance_lineage	edd473cb-2593-4ef4-a328-bdbcbae49d87	2025-07-03 08:21:03.017866-07	tube:tube-generic-10ml:1.0 :: control:giab-HG002:1.0	generic	tube	tube-generic-10ml	{"cogs": {"cost": "1.50", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "giab NA24385", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "Generic 10ml tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:03.017866-07	f	generic
4e428bc7-aa44-455f-82e9-1b0f48b3bfd5	GL261	c5f7422b-9ae1-4496-a645-28e6decfdcd9	container_instance	container_instance	container_instance_lineage	c30fd0fb-b696-4632-895b-ac4e0b089522	2025-07-03 08:21:03.171921-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:03.171921-07	f	\N
9a121f62-bcd3-48be-a7e6-8f9f962b21d2	GL242	14f6d115-81d1-4a24-b8d9-a675f97a5251	container:tube:tube-10ml-glass:1.0	content:control:synthetic-cfdna:1.0	generic_instance_lineage	26562587-4869-4ae7-82f8-06ce603df7f3	2025-07-03 08:21:03.030939-07	tube:tube-10ml-glass:1.0 :: control:synthetic-cfdna:1.0	generic	tube	tube-10ml-glass	{"cogs": {"cost": "4.70", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "s ynthetic-cfdna", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "10ml glass tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:03.030939-07	f	generic
77853896-887c-4d79-bee7-2cd9078883c5	GL243	ed15ff68-5a5a-4bf8-99aa-b4619eb64fd1	container:tube:tube-eppi-1.5ml:1.0	content:control:water-ntc:1.0	generic_instance_lineage	8571eed1-4c25-4b2d-9f40-2e3bf265e6b3	2025-07-03 08:21:03.040775-07	tube:tube-eppi-1.5ml:1.0 :: control:water-ntc:1.0	generic	tube	tube-eppi-1.5ml	{"cogs": {"cost": "1.70", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "NTC - water", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "1.5ml eppi tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:03.040775-07	f	generic
6c40ddce-4044-4bf5-90fe-a7f8927f532c	GL244	c5f7422b-9ae1-4496-a645-28e6decfdcd9	container_instance	container_instance	container_instance_lineage	a20f9cb2-4002-4a69-bff6-d52924a67a7b	2025-07-03 08:21:03.054349-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:03.054349-07	f	\N
9e36ce27-9cda-4b79-9d10-582ff86a73c5	GL245	c5f7422b-9ae1-4496-a645-28e6decfdcd9	container_instance	container_instance	container_instance_lineage	1a35323d-f111-442b-958f-716b90b5d1f2	2025-07-03 08:21:03.068927-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:03.068927-07	f	\N
6b687a84-7642-4f88-b4d6-78e5c6091d39	GL246	c5f7422b-9ae1-4496-a645-28e6decfdcd9	container_instance	container_instance	container_instance_lineage	9b76a542-1cdd-4c04-bb17-9fd9766d78cd	2025-07-03 08:21:03.0776-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:03.0776-07	f	\N
2790e507-5b50-4002-8729-7ca14689e420	GL247	c5f7422b-9ae1-4496-a645-28e6decfdcd9	container_instance	container_instance	container_instance_lineage	e025d800-f0c0-4537-b1b1-cf50c0b976af	2025-07-03 08:21:03.082741-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:03.082741-07	f	\N
5d59836e-6817-4a46-9b6a-cbe41017d7c2	GL248	c5f7422b-9ae1-4496-a645-28e6decfdcd9	container_instance	container_instance	container_instance_lineage	9fb53135-baa2-4a38-895f-c36ea22b3b11	2025-07-03 08:21:03.087697-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:03.087697-07	f	\N
c488e6e8-5fb6-4cf1-9c6b-df9f32f49895	GL249	c5f7422b-9ae1-4496-a645-28e6decfdcd9	container_instance	container_instance	container_instance_lineage	4b2787c8-670e-45a3-93e9-7c32160e7800	2025-07-03 08:21:03.093038-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:03.093038-07	f	\N
fef9e938-8114-4b49-9324-c0c65b6decb3	GL250	c5f7422b-9ae1-4496-a645-28e6decfdcd9	container_instance	container_instance	container_instance_lineage	c188bf5e-cf34-431e-9949-ce0743474003	2025-07-03 08:21:03.09916-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:03.09916-07	f	\N
131c2d42-303a-4ab7-928b-876a6d1fecf8	GL251	c5f7422b-9ae1-4496-a645-28e6decfdcd9	container_instance	container_instance	container_instance_lineage	0c8feadc-bdae-4f20-b475-839e34f912c2	2025-07-03 08:21:03.104822-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:03.104822-07	f	\N
84d114ba-5d11-408f-bc8d-6c7a702b8d2d	GL252	c5f7422b-9ae1-4496-a645-28e6decfdcd9	container_instance	container_instance	container_instance_lineage	96d3bf7e-a12d-4167-b3eb-6f30a69c3b24	2025-07-03 08:21:03.110199-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:03.110199-07	f	\N
14bdc49f-93f2-4d8b-adbb-dd480f1d4f5d	GL253	c5f7422b-9ae1-4496-a645-28e6decfdcd9	container_instance	container_instance	container_instance_lineage	6e0aa5d8-9855-45b9-8da1-1ae98a7dc57c	2025-07-03 08:21:03.115153-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:03.115153-07	f	\N
fd8a8aec-ee5b-4b81-88da-1f736db065f1	GL254	c5f7422b-9ae1-4496-a645-28e6decfdcd9	container_instance	container_instance	container_instance_lineage	42c4b4ad-e85c-4b20-9276-a9f98084fd5c	2025-07-03 08:21:03.120564-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:03.120564-07	f	\N
cbbd217c-0824-417f-a034-d06d7d3f50f3	GL255	c5f7422b-9ae1-4496-a645-28e6decfdcd9	container_instance	container_instance	container_instance_lineage	723fff3b-fa28-4af0-82a4-c5228c05a429	2025-07-03 08:21:03.125565-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:03.125565-07	f	\N
de385663-a99e-4d19-bd08-64567298633a	GL256	c5f7422b-9ae1-4496-a645-28e6decfdcd9	container_instance	container_instance	container_instance_lineage	fd2c7148-44ec-48bf-abe8-38d5fb4c9e93	2025-07-03 08:21:03.138604-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:03.138604-07	f	\N
8a9c03a2-5b4c-4f87-a61a-f7d6d32bf1d1	GL257	c5f7422b-9ae1-4496-a645-28e6decfdcd9	container_instance	container_instance	container_instance_lineage	2b67f9bd-b749-40d0-8189-51ed2c0da68e	2025-07-03 08:21:03.147822-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:03.147822-07	f	\N
4b8562a1-3f1b-48e1-aff6-3425a41b0b8a	GL258	c5f7422b-9ae1-4496-a645-28e6decfdcd9	container_instance	container_instance	container_instance_lineage	467f92c0-58ff-47fa-85a6-e9268901935e	2025-07-03 08:21:03.153581-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:03.153581-07	f	\N
9db54eed-808f-4ce6-8e2c-03ddab45bce1	GL262	c5f7422b-9ae1-4496-a645-28e6decfdcd9	container_instance	container_instance	container_instance_lineage	18266d0e-48c1-4a3c-a530-bd7d02a60043	2025-07-03 08:21:03.177099-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:03.177099-07	f	\N
e720145e-8133-4106-9e2e-ae0737e22666	GL263	c5f7422b-9ae1-4496-a645-28e6decfdcd9	container_instance	container_instance	container_instance_lineage	ffb4a762-b5dc-4494-b040-eb49d98a3e89	2025-07-03 08:21:03.182487-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:03.182487-07	f	\N
804320c8-a0c9-4c29-89a0-fcb543c1c226	GL264	c5f7422b-9ae1-4496-a645-28e6decfdcd9	container_instance	container_instance	container_instance_lineage	edc810e0-165b-4d5d-8e0e-aa000572600d	2025-07-03 08:21:03.187621-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:03.187621-07	f	\N
8b551b71-f40d-4f21-bf52-8bf726d2fd31	GL265	c5f7422b-9ae1-4496-a645-28e6decfdcd9	container_instance	container_instance	container_instance_lineage	de6985c5-652b-4542-b1e5-a5ff86fe3d52	2025-07-03 08:21:03.192956-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:03.192956-07	f	\N
90c0564c-0c11-4afc-873c-d31e5aa13fc5	GL266	c5f7422b-9ae1-4496-a645-28e6decfdcd9	container_instance	container_instance	container_instance_lineage	3f1c2a4d-7892-4b46-9059-f884aad34e80	2025-07-03 08:21:03.198703-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:03.198703-07	f	\N
0e183d54-ca39-4fab-9cf4-2a43c7f15220	GL267	c5f7422b-9ae1-4496-a645-28e6decfdcd9	container_instance	container_instance	container_instance_lineage	2f43dd3a-7c4c-4300-a06a-1af3bbf880be	2025-07-03 08:21:03.203859-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:03.203859-07	f	\N
881652e7-d59f-447c-9b11-da331084f35a	GL268	c5f7422b-9ae1-4496-a645-28e6decfdcd9	container_instance	container_instance	container_instance_lineage	88dc0886-2c10-464c-9b20-0fe485c4223a	2025-07-03 08:21:03.209531-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:03.209531-07	f	\N
ebf4f643-f9f2-4ff6-b993-ee5cee172801	GL269	c5f7422b-9ae1-4496-a645-28e6decfdcd9	container_instance	container_instance	container_instance_lineage	479bb3b4-e408-4cc2-ab2f-1d56b69ace6e	2025-07-03 08:21:03.220289-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:03.220289-07	f	\N
c367a08e-49cb-41a6-b5be-8c2032a41839	GL270	c5f7422b-9ae1-4496-a645-28e6decfdcd9	container_instance	container_instance	container_instance_lineage	9039cfde-66bb-4188-b1ea-e6075bb37449	2025-07-03 08:21:03.228423-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:03.228423-07	f	\N
4e977570-3baa-4586-92d2-e0c6c6f178e1	GL271	c5f7422b-9ae1-4496-a645-28e6decfdcd9	container_instance	container_instance	container_instance_lineage	9409b6fd-e01b-43b1-ab67-77fe15748f54	2025-07-03 08:21:03.233611-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:21:03.233611-07	f	\N
8838f3c9-c234-4079-a0ab-ff182c5c4a27	GL272	a20f9cb2-4002-4a69-bff6-d52924a67a7b	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	42207d28-a267-4548-b1af-b7e93f4982d0	2025-07-03 08:21:03.23909-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:03.23909-07	f	generic
841c638a-f684-4594-a793-c513c1f6e97b	GL273	1a35323d-f111-442b-958f-716b90b5d1f2	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	9d3d6ea3-0299-4b7b-aa33-f361910a58e3	2025-07-03 08:21:03.244636-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:03.244636-07	f	generic
46179fbe-cf4b-4434-a326-59fb46c8ebe8	GL324	ec941c78-3c25-4399-bb69-f266930774bb	container_instance	container_instance	container_instance_lineage	5dff72b8-b5c4-4662-b097-b09c106854c6	2025-07-03 08:23:08.863842-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:08.863842-07	f	\N
8179c198-0681-42c1-9b15-6234dfe69da3	GL325	ec941c78-3c25-4399-bb69-f266930774bb	container_instance	container_instance	container_instance_lineage	974b84e5-0ee2-4545-8ff0-2de08d139c6f	2025-07-03 08:23:08.869272-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:08.869272-07	f	\N
85d8ef1d-1e34-4970-b863-0119887cb537	GL326	ec941c78-3c25-4399-bb69-f266930774bb	container_instance	container_instance	container_instance_lineage	32986026-4b6a-4a23-a81c-ae11fd925582	2025-07-03 08:23:08.874087-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:08.874087-07	f	\N
1ab2da74-9e5d-417d-8628-f23b553a8990	GL327	ec941c78-3c25-4399-bb69-f266930774bb	container_instance	container_instance	container_instance_lineage	8fd06f8b-3195-4a98-882f-49d1c9a4834b	2025-07-03 08:23:08.878751-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:08.878751-07	f	\N
dcdabb27-0af0-418c-9d52-511677e47cde	GL274	9b76a542-1cdd-4c04-bb17-9fd9766d78cd	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	e70c4c91-bcf8-487a-8cc8-45817cd72396	2025-07-03 08:21:03.250346-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:03.250346-07	f	generic
3162a194-c8ff-424b-90b6-f49e372d6576	GL275	e025d800-f0c0-4537-b1b1-cf50c0b976af	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	ed1d3d5a-e234-4ee2-8c5f-b88300456d8b	2025-07-03 08:21:03.25557-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:03.25557-07	f	generic
2c105010-4a1a-473d-85a2-3e37982e6505	GL276	9fb53135-baa2-4a38-895f-c36ea22b3b11	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	6fd97ea6-93e2-4184-a45d-46f3431ca109	2025-07-03 08:21:03.260864-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:03.260864-07	f	generic
22ec3501-74c9-4013-a586-e6a6b5abe88b	GL277	4b2787c8-670e-45a3-93e9-7c32160e7800	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	44ef5e8f-8ff4-402e-9ad6-300221cc3de9	2025-07-03 08:21:03.266682-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:03.266682-07	f	generic
c6419250-2374-421e-bd13-c13f08c20792	GL278	c188bf5e-cf34-431e-9949-ce0743474003	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	70dc8c52-c634-4743-95a3-f7bc39a9cc12	2025-07-03 08:21:03.271948-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:03.271948-07	f	generic
ca8920ae-76fa-4082-9d4f-b4f84d940480	GL279	0c8feadc-bdae-4f20-b475-839e34f912c2	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	6d743c90-ecc6-4b3b-bcab-30278ece70c4	2025-07-03 08:21:03.277032-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:03.277032-07	f	generic
37f67b45-c251-4ca9-bca5-df5c948b6b71	GL280	96d3bf7e-a12d-4167-b3eb-6f30a69c3b24	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	800a223f-1a21-493f-a97c-5462854e77dd	2025-07-03 08:21:03.282022-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:03.282022-07	f	generic
ea70c007-16b1-4589-bbf7-5908e16e4046	GL281	6e0aa5d8-9855-45b9-8da1-1ae98a7dc57c	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	bb6bc18e-109f-4955-a46a-b909b4878979	2025-07-03 08:21:03.287395-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:03.287395-07	f	generic
32fa3742-7427-41e3-aa23-a27a43e065d6	GL282	42c4b4ad-e85c-4b20-9276-a9f98084fd5c	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	92d63b59-0d46-4dcc-aadf-ee1ed83a3979	2025-07-03 08:21:03.293428-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:03.293428-07	f	generic
25c57886-f486-4aa4-9851-ef4dc99b7d5c	GL283	723fff3b-fa28-4af0-82a4-c5228c05a429	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	ed22c26f-07be-4ee0-8510-9575fec8e997	2025-07-03 08:21:03.299631-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:03.299631-07	f	generic
234f1d94-1d08-431b-b9c7-397b8728fe41	GL284	fd2c7148-44ec-48bf-abe8-38d5fb4c9e93	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	2e08566f-8372-48e0-a356-c2aef6cd92d1	2025-07-03 08:21:03.305291-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:03.305291-07	f	generic
3fc3476c-c2e6-49b5-8b56-b8979e449b7b	GL285	2b67f9bd-b749-40d0-8189-51ed2c0da68e	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	e1841886-6ac7-4de7-8c4e-1bb17dd37908	2025-07-03 08:21:03.311046-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:03.311046-07	f	generic
001008c3-0745-41c9-a20d-633615ed3a67	GL286	467f92c0-58ff-47fa-85a6-e9268901935e	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	071b553d-fd1a-432b-a258-99de30fb97d9	2025-07-03 08:21:03.316643-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:03.316643-07	f	generic
9b260b00-055f-43e1-aded-a45bbc02af39	GL287	ecc31348-3e43-46dc-b0d9-c81f9ba47938	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	6be732d5-ba40-4cd2-95a3-ebbb71ea380b	2025-07-03 08:21:03.3223-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:03.3223-07	f	generic
d989078d-4d91-407c-bd1a-cabafb5936f3	GL288	f8ac8009-49e0-4a5d-a1d2-6dd107e496e7	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	02d4fd7d-df49-48a7-8b80-873d463d7037	2025-07-03 08:21:03.328263-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:03.328263-07	f	generic
0922e1ea-73cd-43ae-95bf-66960355d3ba	GL289	c30fd0fb-b696-4632-895b-ac4e0b089522	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	3f2989ce-c0d1-4b52-9140-b9b19d170768	2025-07-03 08:21:03.334026-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:03.334026-07	f	generic
0fc74178-ff14-4450-88c2-560425d22852	GL290	18266d0e-48c1-4a3c-a530-bd7d02a60043	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	8bff6ef8-6bc4-4cc5-bf48-8acc081d6f99	2025-07-03 08:21:03.341416-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:03.341416-07	f	generic
da673b37-23d5-4b53-9774-c2a00fd4f313	GL291	ffb4a762-b5dc-4494-b040-eb49d98a3e89	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	86c7fbcf-32e3-4270-8884-b1617558a29e	2025-07-03 08:21:03.34811-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:03.34811-07	f	generic
102d4f4b-3df3-43f4-8be9-a208b12463b8	GL292	edc810e0-165b-4d5d-8e0e-aa000572600d	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	e258c3a2-c387-44c9-93c7-60868c8046c8	2025-07-03 08:21:03.353852-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:03.353852-07	f	generic
a09a0eb6-3eba-414c-9c5a-afb50eebd4cc	GL293	de6985c5-652b-4542-b1e5-a5ff86fe3d52	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	82dac929-c652-47f9-ad44-7a7ac8fade75	2025-07-03 08:21:03.359328-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:03.359328-07	f	generic
6fff4d6b-2843-4149-ad1b-6fde2cf5ee2b	GL294	3f1c2a4d-7892-4b46-9059-f884aad34e80	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	ba650613-ec31-45d0-8935-5b94a1dc3079	2025-07-03 08:21:03.365447-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:03.365447-07	f	generic
79a4a65e-34f6-45e2-8c2a-cec28ca4259c	GL295	2f43dd3a-7c4c-4300-a06a-1af3bbf880be	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	4d800139-262b-4779-ba70-ab7a011c53cb	2025-07-03 08:21:03.371137-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:03.371137-07	f	generic
3e137116-98fc-4c3a-9b8a-e44d70090824	GL296	88dc0886-2c10-464c-9b20-0fe485c4223a	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	ca7c8e06-c44f-414b-a3f4-7f3abdc7d153	2025-07-03 08:21:03.376844-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:03.376844-07	f	generic
39921349-1c43-4fa3-ac6a-92fdbe295072	GL297	479bb3b4-e408-4cc2-ab2f-1d56b69ace6e	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	c7de2b92-a075-4026-86ca-1f491615948c	2025-07-03 08:21:03.382514-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:03.382514-07	f	generic
51610499-5fa3-46a2-8e63-d240c0fce935	GL298	9039cfde-66bb-4188-b1ea-e6075bb37449	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	f3ca1576-6ec8-4e97-835f-dbc25be9d779	2025-07-03 08:21:03.388193-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:03.388193-07	f	generic
62462a73-6790-4ca9-85bd-371aafe6c537	GL299	9409b6fd-e01b-43b1-ab67-77fe15748f54	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	e526049b-2502-40d3-92d6-00841b98b660	2025-07-03 08:21:03.393933-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:21:03.393933-07	f	generic
04464188-a83d-4e7e-9c68-2c94bd16a0ce	GL328	ec941c78-3c25-4399-bb69-f266930774bb	container_instance	container_instance	container_instance_lineage	b7122674-3965-4b7f-b5be-c6b1c869e1c5	2025-07-03 08:23:08.883518-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:08.883518-07	f	\N
f4b4bebf-5690-4124-ae96-e757f67ab71d	GL329	ec941c78-3c25-4399-bb69-f266930774bb	container_instance	container_instance	container_instance_lineage	de7745e0-b7ec-49e7-8436-e6f0152453b6	2025-07-03 08:23:08.8891-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:08.8891-07	f	\N
bba10ec4-9246-4469-998b-88a4bd0a2163	GL303	ec941c78-3c25-4399-bb69-f266930774bb	container_instance	container_instance	container_instance_lineage	a778a7e1-7154-40f8-90c2-8c264b021681	2025-07-03 08:23:08.715457-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:08.715457-07	f	\N
7048bc4b-058d-4fe4-822f-f2b6d5368a5c	GL304	ec941c78-3c25-4399-bb69-f266930774bb	container_instance	container_instance	container_instance_lineage	ccf9f662-857d-49b6-8a43-8f9097e466f6	2025-07-03 08:23:08.723816-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:08.723816-07	f	\N
c3a38f19-4c21-4950-a3b2-008298f34e12	GL305	ec941c78-3c25-4399-bb69-f266930774bb	container_instance	container_instance	container_instance_lineage	7bffe3df-228e-49d0-b3d3-ea7566c97b58	2025-07-03 08:23:08.735751-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:08.735751-07	f	\N
c6a27932-f97c-41d6-8bce-8a647a177d63	GL306	ec941c78-3c25-4399-bb69-f266930774bb	container_instance	container_instance	container_instance_lineage	29ca4ea9-a532-40b8-90a2-bf62f212272c	2025-07-03 08:23:08.743657-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:08.743657-07	f	\N
b72dec32-16c0-4ddb-ae94-2743c0db5d82	GL307	ec941c78-3c25-4399-bb69-f266930774bb	container_instance	container_instance	container_instance_lineage	3b804416-77a1-4840-a1cb-ddce018c129b	2025-07-03 08:23:08.750583-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:08.750583-07	f	\N
ce9cd592-fd05-4d59-ab69-cb1fdbf569ff	GL308	ec941c78-3c25-4399-bb69-f266930774bb	container_instance	container_instance	container_instance_lineage	1bc05522-5e1d-4faa-8aaa-5b4a38ebc55a	2025-07-03 08:23:08.757206-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:08.757206-07	f	\N
ca6d1ddb-c0ce-42b6-9f23-eac0863bfc7b	GL309	ec941c78-3c25-4399-bb69-f266930774bb	container_instance	container_instance	container_instance_lineage	8ce6ad46-136d-48a8-ae5a-0d860b66c006	2025-07-03 08:23:08.763819-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:08.763819-07	f	\N
64132bfa-d1cb-4d53-8820-2a3d526b34ce	GL310	ec941c78-3c25-4399-bb69-f266930774bb	container_instance	container_instance	container_instance_lineage	cb6e1bb2-13ab-4245-b0bb-398bfdae6ac6	2025-07-03 08:23:08.771673-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:08.771673-07	f	\N
074be2e1-7e8f-45c2-aac1-b0ecadfb866a	GL311	ec941c78-3c25-4399-bb69-f266930774bb	container_instance	container_instance	container_instance_lineage	da03904f-c644-4f7b-a603-11082db6655d	2025-07-03 08:23:08.779117-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:08.779117-07	f	\N
c5c8cfa0-fe0b-480b-ab03-e5921141b8e1	GL312	ec941c78-3c25-4399-bb69-f266930774bb	container_instance	container_instance	container_instance_lineage	bc77f3ee-c50b-43eb-8f41-d9fec67a5da5	2025-07-03 08:23:08.78484-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:08.78484-07	f	\N
62afa9ab-7bbd-451f-b344-504dc4fb7311	GL313	ec941c78-3c25-4399-bb69-f266930774bb	container_instance	container_instance	container_instance_lineage	780f794b-7f49-4805-801b-912cabfd265e	2025-07-03 08:23:08.790553-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:08.790553-07	f	\N
7aaf3aba-b740-494f-ae80-8d4d6a0539db	GL314	ec941c78-3c25-4399-bb69-f266930774bb	container_instance	container_instance	container_instance_lineage	06fc3873-3a8c-404d-a4af-b8c7116d3213	2025-07-03 08:23:08.799621-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:08.799621-07	f	\N
a2a85981-bdb9-412b-b9b4-5878c7969e70	GL315	ec941c78-3c25-4399-bb69-f266930774bb	container_instance	container_instance	container_instance_lineage	85fa4716-fb10-4279-9999-ebd9c08f6b20	2025-07-03 08:23:08.808977-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:08.808977-07	f	\N
fcba1c1e-024c-4be0-941a-5937852dc425	GL316	ec941c78-3c25-4399-bb69-f266930774bb	container_instance	container_instance	container_instance_lineage	b4b7ac21-3798-4e5d-8ed4-735cec860516	2025-07-03 08:23:08.818645-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:08.818645-07	f	\N
222726b1-4659-44c6-b6ac-704278ca2587	GL317	ec941c78-3c25-4399-bb69-f266930774bb	container_instance	container_instance	container_instance_lineage	50a01be1-89ed-4814-93c6-de250293b4e6	2025-07-03 08:23:08.824708-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:08.824708-07	f	\N
a8736dda-29e9-4b62-b8f3-af9869d44f6b	GL318	ec941c78-3c25-4399-bb69-f266930774bb	container_instance	container_instance	container_instance_lineage	1fd00059-c3de-43e5-9cb4-7d4c273343ad	2025-07-03 08:23:08.831455-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:08.831455-07	f	\N
d70377da-64d3-4af3-b6c9-080adf9e5495	GL319	ec941c78-3c25-4399-bb69-f266930774bb	container_instance	container_instance	container_instance_lineage	53c38748-0c95-4d1d-bb0c-83201d2d54db	2025-07-03 08:23:08.837365-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:08.837365-07	f	\N
beb97add-7667-4b29-9254-37ffc0604634	GL320	ec941c78-3c25-4399-bb69-f266930774bb	container_instance	container_instance	container_instance_lineage	56bc2733-1be4-4f33-8c35-2b522209ae67	2025-07-03 08:23:08.842566-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:08.842566-07	f	\N
beac40fc-cbef-451f-8ab7-863bd606db77	GL321	ec941c78-3c25-4399-bb69-f266930774bb	container_instance	container_instance	container_instance_lineage	72061fa5-3321-4a96-b64c-275d915fc9ee	2025-07-03 08:23:08.847831-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:08.847831-07	f	\N
10b53fe6-003b-4068-82a5-ca3a0e5de42c	GL322	ec941c78-3c25-4399-bb69-f266930774bb	container_instance	container_instance	container_instance_lineage	e6d570a4-5fcf-4887-b23c-31b36a64854c	2025-07-03 08:23:08.853275-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:08.853275-07	f	\N
4f9a6005-3c45-4eee-9a6b-44808a8402bd	GL323	ec941c78-3c25-4399-bb69-f266930774bb	container_instance	container_instance	container_instance_lineage	92d78938-91b6-4aee-a8b4-a74e51ad8b95	2025-07-03 08:23:08.858591-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:08.858591-07	f	\N
2b96af66-a954-43b7-9879-ad233e56b391	GL330	ec941c78-3c25-4399-bb69-f266930774bb	container_instance	container_instance	container_instance_lineage	78c3c9b5-ea50-4c51-8560-9929af6a202d	2025-07-03 08:23:08.894521-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:08.894521-07	f	\N
edeabf86-33b6-44a3-a590-5b2fd61df19f	GL331	a778a7e1-7154-40f8-90c2-8c264b021681	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	4366985c-1c27-4e8e-8f17-5f2bb448f0d2	2025-07-03 08:23:08.901124-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:08.901124-07	f	generic
7d6f7ec7-5baf-4825-bda5-04224bb9460c	GL332	ccf9f662-857d-49b6-8a43-8f9097e466f6	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	5753d5e2-98b7-44e0-88ce-6ee0096cf3ee	2025-07-03 08:23:08.907676-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:08.907676-07	f	generic
ee3b566e-9c9a-4310-ab79-44f89489453c	GL333	7bffe3df-228e-49d0-b3d3-ea7566c97b58	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	06518ad4-6d64-4cd9-ad67-13c330248de9	2025-07-03 08:23:08.912175-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:08.912175-07	f	generic
965312e8-cc48-4124-a96e-76a6a9e775a3	GL379	36a78f96-96bc-4162-9372-16d115b2e6c3	container_instance	container_instance	container_instance_lineage	820ef252-cff7-4a46-a803-2dc8eee77103	2025-07-03 08:23:09.137843-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:09.137843-07	f	\N
e0868793-d127-4bfb-a19d-e2fb2741fa3e	GL380	36a78f96-96bc-4162-9372-16d115b2e6c3	container_instance	container_instance	container_instance_lineage	3b447fb4-8438-4485-bac4-1553e554d7d2	2025-07-03 08:23:09.142959-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:09.142959-07	f	\N
87df48f5-fead-4a10-9166-10a081b27e42	GL381	36a78f96-96bc-4162-9372-16d115b2e6c3	container_instance	container_instance	container_instance_lineage	c6928136-7211-428a-aa93-424927c53157	2025-07-03 08:23:09.14792-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:09.14792-07	f	\N
ae944efe-2835-418f-93fc-fec9a20e20f7	GL382	36a78f96-96bc-4162-9372-16d115b2e6c3	container_instance	container_instance	container_instance_lineage	5633367b-29d0-4a2c-bf49-950a922e81bb	2025-07-03 08:23:09.152702-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:09.152702-07	f	\N
39fc1d4e-f58d-4fe7-b0f7-8fdded5395b0	GL383	36a78f96-96bc-4162-9372-16d115b2e6c3	container_instance	container_instance	container_instance_lineage	4b42200e-1ca5-4eb8-8947-8cf89ae11946	2025-07-03 08:23:09.15779-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:09.15779-07	f	\N
8a450bd4-2d67-4314-84fa-9e096cc6149f	GL334	29ca4ea9-a532-40b8-90a2-bf62f212272c	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	cb8e67ed-339b-4c03-b725-1a60d7ad9e54	2025-07-03 08:23:08.916028-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:08.916028-07	f	generic
a91c15e0-0d49-4e69-91d7-7f386b35eb0a	GL335	3b804416-77a1-4840-a1cb-ddce018c129b	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	b2421416-938d-46ed-bcc4-1aca83a2e2ab	2025-07-03 08:23:08.92017-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:08.92017-07	f	generic
2db7a90c-1a11-4989-95e9-35419eeb39f9	GL336	1bc05522-5e1d-4faa-8aaa-5b4a38ebc55a	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	7ea54466-0a39-4279-becc-c379d09d80cc	2025-07-03 08:23:08.924004-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:08.924004-07	f	generic
295a6f67-0a29-4b88-96e2-c6362e7c325f	GL337	8ce6ad46-136d-48a8-ae5a-0d860b66c006	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	878e54ab-5d67-4d62-8602-5d9700906f19	2025-07-03 08:23:08.927791-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:08.927791-07	f	generic
689af8bc-b8ac-46b8-8a96-7a87f191a7e8	GL338	cb6e1bb2-13ab-4245-b0bb-398bfdae6ac6	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	ee71e870-facd-47db-a77a-0005c2199964	2025-07-03 08:23:08.93191-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:08.93191-07	f	generic
fe7a8b20-63eb-4386-a27d-c83949a04533	GL339	da03904f-c644-4f7b-a603-11082db6655d	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	64e1cb24-3e88-4abb-807e-4717e3f32b73	2025-07-03 08:23:08.935993-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:08.935993-07	f	generic
88a734ce-3de2-4eb7-9806-bab64116c701	GL340	bc77f3ee-c50b-43eb-8f41-d9fec67a5da5	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	968e67a1-9b12-4960-83a7-912ca28870e3	2025-07-03 08:23:08.940342-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:08.940342-07	f	generic
eaf76fd7-118e-4dbf-897c-f4a1b4930935	GL341	780f794b-7f49-4805-801b-912cabfd265e	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	109cb0f0-2f2b-4a27-992d-38184b13d248	2025-07-03 08:23:08.944735-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:08.944735-07	f	generic
c20c35d3-4f9d-46bf-8317-94bd9413f0d6	GL342	06fc3873-3a8c-404d-a4af-b8c7116d3213	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	df54fb98-5e91-448f-81c4-c19586df73ee	2025-07-03 08:23:08.94886-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:08.94886-07	f	generic
5fa030fd-745e-4694-a871-66342e421b9d	GL343	85fa4716-fb10-4279-9999-ebd9c08f6b20	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	5d354a84-be84-403e-8777-7b9a2520761f	2025-07-03 08:23:08.95349-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:08.95349-07	f	generic
9023c24c-9905-4973-beee-b40b7271f8b2	GL344	b4b7ac21-3798-4e5d-8ed4-735cec860516	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	94807e9b-79b5-43ee-928c-b1977a3f96f5	2025-07-03 08:23:08.957729-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:08.957729-07	f	generic
e36d8a9f-f261-4f80-9459-4e21d1ac8284	GL345	50a01be1-89ed-4814-93c6-de250293b4e6	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	7de1ceef-160a-464e-92ca-2b68522b3eba	2025-07-03 08:23:08.962449-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:08.962449-07	f	generic
d7373e5b-8f66-444e-a2ce-c353641f3149	GL346	1fd00059-c3de-43e5-9cb4-7d4c273343ad	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	733386da-f0db-4671-9ea4-5dda616d2631	2025-07-03 08:23:08.967393-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:08.967393-07	f	generic
f98baef5-46bd-44a9-85d5-b6a98e09c842	GL347	53c38748-0c95-4d1d-bb0c-83201d2d54db	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	64c5fd73-d9fc-4c22-9f2e-1f91998a02c1	2025-07-03 08:23:08.971744-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:08.971744-07	f	generic
5127cd1d-6907-499e-b7e1-30808e04528b	GL348	56bc2733-1be4-4f33-8c35-2b522209ae67	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	148a355a-e554-4686-85f1-484bb92e12eb	2025-07-03 08:23:08.976128-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:08.976128-07	f	generic
d649b2f7-f439-4b04-b383-d8927030277c	GL349	72061fa5-3321-4a96-b64c-275d915fc9ee	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	8259854f-3495-44e0-8850-57466281e5a5	2025-07-03 08:23:08.980265-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:08.980265-07	f	generic
2fa16d80-d276-4209-90fd-a7a87771ee22	GL350	e6d570a4-5fcf-4887-b23c-31b36a64854c	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	a9098742-f68e-4d41-8ad5-58b225c762ec	2025-07-03 08:23:08.984738-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:08.984738-07	f	generic
05b798ea-0c44-4bd8-8524-25ba31e10209	GL351	92d78938-91b6-4aee-a8b4-a74e51ad8b95	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	4da41248-557d-40ac-9a52-c0deaea86440	2025-07-03 08:23:08.989051-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:08.989051-07	f	generic
354556e6-e1b3-447d-bc8e-073ca8c80e08	GL352	5dff72b8-b5c4-4662-b097-b09c106854c6	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	c02979bf-e445-4ee7-8505-67bf010865fd	2025-07-03 08:23:08.993641-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:08.993641-07	f	generic
e3acbae5-b7cd-41ac-aebf-25d0f233795a	GL353	974b84e5-0ee2-4545-8ff0-2de08d139c6f	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	431e9885-70e6-4269-9e55-5a6b6de2a8ea	2025-07-03 08:23:08.998181-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:08.998181-07	f	generic
faf2bc63-372f-4d73-ae5b-b7cb93fe8190	GL354	32986026-4b6a-4a23-a81c-ae11fd925582	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	859f2a12-8212-4697-893b-19aee54a3c15	2025-07-03 08:23:09.002359-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.002359-07	f	generic
8813022a-cc82-4d68-b5da-a61f008d586a	GL355	8fd06f8b-3195-4a98-882f-49d1c9a4834b	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	9fe5d8ab-c3a9-4850-b662-07d1349c1d97	2025-07-03 08:23:09.006676-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.006676-07	f	generic
a2a67686-4243-495c-ac14-47921d1125bb	GL356	b7122674-3965-4b7f-b5be-c6b1c869e1c5	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	901f56f9-59e7-4b78-80b9-6d6f62d6aed7	2025-07-03 08:23:09.011055-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.011055-07	f	generic
bd06de39-b100-4b10-8389-caf6c681f7a2	GL357	de7745e0-b7ec-49e7-8436-e6f0152453b6	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	c64e172a-31c7-41d4-83e5-a592747c1107	2025-07-03 08:23:09.01544-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.01544-07	f	generic
e6161a8c-e2b8-4c00-b382-3dac9c037ec1	GL358	78c3c9b5-ea50-4c51-8560-9929af6a202d	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	173f906f-f360-4aa8-bf05-20fed68593d5	2025-07-03 08:23:09.019921-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.019921-07	f	generic
cdb5cec8-fdbd-4b2c-8f26-c0621c14b9a3	GL359	36a78f96-96bc-4162-9372-16d115b2e6c3	container_instance	container_instance	container_instance_lineage	b806e72c-8000-4615-b23f-b63a79d2da4d	2025-07-03 08:23:09.031758-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:09.031758-07	f	\N
fec6528b-07bf-4b1e-ac36-168b46d1e7e6	GL360	36a78f96-96bc-4162-9372-16d115b2e6c3	container_instance	container_instance	container_instance_lineage	701b1431-fa54-4f76-9076-c445285f8ff5	2025-07-03 08:23:09.037395-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:09.037395-07	f	\N
bd4f4fe4-b242-41f2-b297-7c4e25c79a43	GL361	36a78f96-96bc-4162-9372-16d115b2e6c3	container_instance	container_instance	container_instance_lineage	6a2dc8d6-7470-4baa-b4d8-cd0e20761c7d	2025-07-03 08:23:09.042644-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:09.042644-07	f	\N
5967cb29-d6df-4e6e-a318-d6279ff16249	GL362	36a78f96-96bc-4162-9372-16d115b2e6c3	container_instance	container_instance	container_instance_lineage	7f1ef106-8285-4641-bc6c-b2cb1e199946	2025-07-03 08:23:09.048349-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:09.048349-07	f	\N
377a531c-2df4-483d-bff5-e5251a9caa86	GL363	36a78f96-96bc-4162-9372-16d115b2e6c3	container_instance	container_instance	container_instance_lineage	00b1f951-e684-487b-91fc-803626c26deb	2025-07-03 08:23:09.053251-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:09.053251-07	f	\N
b092a54b-5cbf-4063-a27e-9b90a7fdb23d	GL364	36a78f96-96bc-4162-9372-16d115b2e6c3	container_instance	container_instance	container_instance_lineage	a7ce7f5a-f641-42ae-bb64-e32665f3383f	2025-07-03 08:23:09.058174-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:09.058174-07	f	\N
90baf748-56e6-4cb8-8fe4-05283ab8b489	GL365	36a78f96-96bc-4162-9372-16d115b2e6c3	container_instance	container_instance	container_instance_lineage	04078517-815f-4032-a223-10dc61971326	2025-07-03 08:23:09.063341-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:09.063341-07	f	\N
5ab67372-ca46-47ce-a91f-b01a6564a685	GL366	36a78f96-96bc-4162-9372-16d115b2e6c3	container_instance	container_instance	container_instance_lineage	fb489163-90aa-4b36-83d6-71b946fb43f5	2025-07-03 08:23:09.068635-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:09.068635-07	f	\N
6c8544e3-8e57-44c8-9d6d-649071cc1eae	GL367	36a78f96-96bc-4162-9372-16d115b2e6c3	container_instance	container_instance	container_instance_lineage	0e06a2a9-31a7-4ab9-9074-9cc9ce191775	2025-07-03 08:23:09.073646-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:09.073646-07	f	\N
47d078b3-fbfd-458f-b210-4264f46ba19e	GL368	36a78f96-96bc-4162-9372-16d115b2e6c3	container_instance	container_instance	container_instance_lineage	f893cb99-8261-4d16-82b7-195a2985abb8	2025-07-03 08:23:09.078849-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:09.078849-07	f	\N
49b6322f-e51e-4130-bb66-bdde2a7c40d9	GL369	36a78f96-96bc-4162-9372-16d115b2e6c3	container_instance	container_instance	container_instance_lineage	842e7d48-9605-4321-956e-17b670298506	2025-07-03 08:23:09.08386-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:09.08386-07	f	\N
4137bc31-053d-4639-97ea-f15d689dc56b	GL370	36a78f96-96bc-4162-9372-16d115b2e6c3	container_instance	container_instance	container_instance_lineage	b1519897-fb8e-4607-898f-3f0575c8a04a	2025-07-03 08:23:09.088818-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:09.088818-07	f	\N
71348741-eb8b-4c9c-bfcf-4af826116a13	GL371	36a78f96-96bc-4162-9372-16d115b2e6c3	container_instance	container_instance	container_instance_lineage	550d0d97-6fda-47b0-8e07-d820e912a145	2025-07-03 08:23:09.093617-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:09.093617-07	f	\N
2c63e770-24e0-4053-8e3f-8a97aa1d7a79	GL372	36a78f96-96bc-4162-9372-16d115b2e6c3	container_instance	container_instance	container_instance_lineage	bd871a52-972e-4179-a5f8-2aacf0466302	2025-07-03 08:23:09.099273-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:09.099273-07	f	\N
656a6e1c-4bf3-40d4-9a3d-f03ef843d3b0	GL373	36a78f96-96bc-4162-9372-16d115b2e6c3	container_instance	container_instance	container_instance_lineage	4ba333d5-a753-4026-9358-75385fec301a	2025-07-03 08:23:09.10718-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:09.10718-07	f	\N
cd1cf9f9-3297-4276-895f-f1a678b9f1dc	GL374	36a78f96-96bc-4162-9372-16d115b2e6c3	container_instance	container_instance	container_instance_lineage	60b56382-b9de-4a9d-892b-a6e1bcae9e89	2025-07-03 08:23:09.112412-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:09.112412-07	f	\N
f2c44965-a71e-44b9-b5ae-1aabb95e4d1e	GL375	36a78f96-96bc-4162-9372-16d115b2e6c3	container_instance	container_instance	container_instance_lineage	a9314a45-c45f-4522-a0e9-9a99459a6924	2025-07-03 08:23:09.117223-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:09.117223-07	f	\N
9cc96fc0-394e-4a0b-9dc0-7287b39afbaf	GL376	36a78f96-96bc-4162-9372-16d115b2e6c3	container_instance	container_instance	container_instance_lineage	43f0aac8-c4da-48d5-ade3-02f63fdd00a5	2025-07-03 08:23:09.122373-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:09.122373-07	f	\N
0f9233ce-f7d6-4a54-8a34-70a67560b75f	GL377	36a78f96-96bc-4162-9372-16d115b2e6c3	container_instance	container_instance	container_instance_lineage	2d325520-3bd4-44b8-93e0-874881087ae6	2025-07-03 08:23:09.127525-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:09.127525-07	f	\N
c89fc23d-30a7-4027-8168-f25bc8e1de41	GL378	36a78f96-96bc-4162-9372-16d115b2e6c3	container_instance	container_instance	container_instance_lineage	57556f94-c37c-41c8-88db-53e8cee975b6	2025-07-03 08:23:09.132921-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:09.132921-07	f	\N
9a66eec3-08fc-44ae-83ae-fc680437decf	GL384	36a78f96-96bc-4162-9372-16d115b2e6c3	container_instance	container_instance	container_instance_lineage	fc4f5d57-c0c6-4d38-bfd6-d50850a8f148	2025-07-03 08:23:09.163105-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:09.163105-07	f	\N
9cc4f63d-0981-4b9c-b12c-d0d0c7f54e33	GL385	36a78f96-96bc-4162-9372-16d115b2e6c3	container_instance	container_instance	container_instance_lineage	cff309ac-096c-4bc0-93dc-5adc7c8fe0fb	2025-07-03 08:23:09.168511-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:09.168511-07	f	\N
b95e04d7-8737-4824-a2a0-eb7896990a5a	GL386	36a78f96-96bc-4162-9372-16d115b2e6c3	container_instance	container_instance	container_instance_lineage	ca5acce6-59f8-4465-83e5-3bb593d9bec3	2025-07-03 08:23:09.173761-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:09.173761-07	f	\N
703a5aaf-2a74-4cf9-aaff-eeaf6fee40cd	GL387	b806e72c-8000-4615-b23f-b63a79d2da4d	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	cb05244f-66db-44ed-9642-77f2976dd0dc	2025-07-03 08:23:09.177524-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.177524-07	f	generic
b85974ac-48d4-4f95-8d62-4ed766ae5ce7	GL388	701b1431-fa54-4f76-9076-c445285f8ff5	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	a3c5b2a8-b081-4788-ac6a-f73a3e942278	2025-07-03 08:23:09.181932-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.181932-07	f	generic
5d0af38f-b391-4396-a686-0025a58a3f8e	GL389	6a2dc8d6-7470-4baa-b4d8-cd0e20761c7d	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	ed1337dc-5377-45a7-a724-e41e29c46a91	2025-07-03 08:23:09.186355-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.186355-07	f	generic
178aac1e-4d01-4c01-9f6c-c44d25d1fa1f	GL435	39ac2d80-8f21-4823-8941-abfe02e40867	container_instance	container_instance	container_instance_lineage	db0c762c-c36b-42de-bc97-6976ba8c53be	2025-07-03 08:23:16.982042-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:16.982042-07	f	\N
d5ce552e-a6d9-4d70-b1ce-305f17f7bc13	GL436	39ac2d80-8f21-4823-8941-abfe02e40867	container_instance	container_instance	container_instance_lineage	01b72489-6786-4b1f-a8b7-25b56770d48f	2025-07-03 08:23:16.987527-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:16.987527-07	f	\N
ef7b9389-0bc2-437d-81ed-dcb394ffbc7f	GL437	39ac2d80-8f21-4823-8941-abfe02e40867	container_instance	container_instance	container_instance_lineage	70bfdd8e-9926-4052-a785-1555e295832a	2025-07-03 08:23:16.992526-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:16.992526-07	f	\N
25a39b3d-48be-4584-b83a-43e2985ce506	GL438	39ac2d80-8f21-4823-8941-abfe02e40867	container_instance	container_instance	container_instance_lineage	ed6bd8d5-9e73-419c-afd6-6399c05e75a3	2025-07-03 08:23:16.997679-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:16.997679-07	f	\N
5861cc21-e93e-46cf-bcdb-afc4afc7225b	GL390	7f1ef106-8285-4641-bc6c-b2cb1e199946	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	dd1014f2-4cd8-4e15-b4ce-5095401e8759	2025-07-03 08:23:09.190707-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.190707-07	f	generic
35e7d335-59f2-47ef-a757-8bd04a00a2f8	GL391	00b1f951-e684-487b-91fc-803626c26deb	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	170ab794-eb07-48e0-b1c0-f87d77d20590	2025-07-03 08:23:09.195105-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.195105-07	f	generic
8576a290-e0f4-4301-a63b-df8c53ff07bd	GL392	a7ce7f5a-f641-42ae-bb64-e32665f3383f	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	7316e0cf-c132-4afb-aef5-8bcf84ab93df	2025-07-03 08:23:09.199605-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.199605-07	f	generic
319d79f6-0a91-4668-b6e6-6e1676cf9bf2	GL393	04078517-815f-4032-a223-10dc61971326	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	8a6828b1-67d5-4a72-a00e-70795c41dca6	2025-07-03 08:23:09.20388-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.20388-07	f	generic
60fb78d3-e9c7-4553-af86-bcd43b4c3e39	GL394	fb489163-90aa-4b36-83d6-71b946fb43f5	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	12f78a64-40a3-4753-bef0-11ad97568fa6	2025-07-03 08:23:09.208299-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.208299-07	f	generic
ea0421db-1589-4190-9e5b-4191495ba9c3	GL395	0e06a2a9-31a7-4ab9-9074-9cc9ce191775	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	239cf4e6-27d8-4b54-ad21-a528a3c52252	2025-07-03 08:23:09.212744-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.212744-07	f	generic
52b216e5-610e-402d-b90d-b587a6dd1e20	GL396	f893cb99-8261-4d16-82b7-195a2985abb8	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	512ab731-ec11-4e70-ace2-225ccd1c827c	2025-07-03 08:23:09.217046-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.217046-07	f	generic
b6fd6f13-169e-436e-91ef-ba8fc1fd6b3a	GL397	842e7d48-9605-4321-956e-17b670298506	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	60bb2ea0-c3cb-442c-8ced-8495f8693f6f	2025-07-03 08:23:09.221322-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.221322-07	f	generic
b9dc311a-c011-463c-a9f6-e94d8b415f37	GL398	b1519897-fb8e-4607-898f-3f0575c8a04a	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	cb072692-f302-4e41-9830-dc8d5c7d4289	2025-07-03 08:23:09.225846-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.225846-07	f	generic
1d89e2d6-5241-46ab-a357-1d0c0e59382b	GL399	550d0d97-6fda-47b0-8e07-d820e912a145	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	db516863-3a58-4e48-bdea-253df93b7eba	2025-07-03 08:23:09.23021-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.23021-07	f	generic
420cf78c-cf53-4208-9ca7-b8d1f3e0d774	GL400	bd871a52-972e-4179-a5f8-2aacf0466302	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	a6f37e7d-af69-46c4-a585-db6bfa69c506	2025-07-03 08:23:09.234677-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.234677-07	f	generic
fcd6944d-0c86-4a7a-9aa2-db3490ea589b	GL401	4ba333d5-a753-4026-9358-75385fec301a	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	c75f8621-9c89-4d84-9938-ed98b1430c39	2025-07-03 08:23:09.239284-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.239284-07	f	generic
7899ebf9-9a37-4935-9f46-c8fdacde5a1f	GL402	60b56382-b9de-4a9d-892b-a6e1bcae9e89	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	ad6951f6-0d8e-43e5-a965-d1ac3e2d0035	2025-07-03 08:23:09.243663-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.243663-07	f	generic
598f3faa-a496-4f66-8e75-e1882d5a2d65	GL403	a9314a45-c45f-4522-a0e9-9a99459a6924	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	558d4ccc-51af-41e5-a221-6a4d6de4e87e	2025-07-03 08:23:09.248075-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.248075-07	f	generic
a701bf49-691f-4c59-8b85-2127da8e5c06	GL404	43f0aac8-c4da-48d5-ade3-02f63fdd00a5	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	3e0d0af1-1210-4cab-84af-c3f9e045953f	2025-07-03 08:23:09.252411-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.252411-07	f	generic
3ed88f84-e80e-4785-9d78-b9a314c4097e	GL405	2d325520-3bd4-44b8-93e0-874881087ae6	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	eec83a36-273d-4e8d-a349-5c5abc492832	2025-07-03 08:23:09.256823-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.256823-07	f	generic
bb9ee74f-1635-4228-9ffb-7af827d417ed	GL406	57556f94-c37c-41c8-88db-53e8cee975b6	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	502f963c-82a7-45c1-be2b-d96b35f3bdf2	2025-07-03 08:23:09.261204-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.261204-07	f	generic
077f07fa-ffe9-41ed-81f8-46e6f1d87674	GL407	820ef252-cff7-4a46-a803-2dc8eee77103	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	00b8f24b-52a1-4993-b78a-1f8de01c1df5	2025-07-03 08:23:09.265782-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.265782-07	f	generic
fb64de69-cc24-44f9-a83b-307030ff461f	GL408	3b447fb4-8438-4485-bac4-1553e554d7d2	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	5f74720c-3a0b-426f-981d-9c1db1d9f0cd	2025-07-03 08:23:09.269957-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.269957-07	f	generic
355e4638-e17a-4fe7-87df-ba57989a4c0f	GL409	c6928136-7211-428a-aa93-424927c53157	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	1f7b5d3a-40fc-4878-bcd6-78a9cef9ae5f	2025-07-03 08:23:09.274535-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.274535-07	f	generic
7511643a-c11c-4603-9aaa-51dfacd96239	GL410	5633367b-29d0-4a2c-bf49-950a922e81bb	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	33903f7e-9f8c-4b25-aae5-d2273f155189	2025-07-03 08:23:09.278964-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.278964-07	f	generic
9253402c-c2be-41a1-aaa2-5c97ff05605e	GL411	4b42200e-1ca5-4eb8-8947-8cf89ae11946	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	b94ca2e5-3160-4c4d-ba63-e9b1fd3395e0	2025-07-03 08:23:09.283551-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.283551-07	f	generic
94d325b7-2307-4805-87a4-d6feeef522f7	GL412	fc4f5d57-c0c6-4d38-bfd6-d50850a8f148	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	88692020-71a2-4eda-89bc-ebb4cf452fed	2025-07-03 08:23:09.287948-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.287948-07	f	generic
9513e3a5-1064-45d6-9499-4df4bc43ef72	GL413	cff309ac-096c-4bc0-93dc-5adc7c8fe0fb	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	b26f5000-d3ef-4d72-be05-67228321dee4	2025-07-03 08:23:09.292413-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.292413-07	f	generic
1aff911b-21b8-4861-9006-ab254d38e0e9	GL414	ca5acce6-59f8-4465-83e5-3bb593d9bec3	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	46df43ab-6fe5-4e67-8f96-0ea193e774c6	2025-07-03 08:23:09.296939-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:09.296939-07	f	generic
74601e8a-d8b3-4fd8-ba7b-437d201f9754	GL415	39ac2d80-8f21-4823-8941-abfe02e40867	container_instance	container_instance	container_instance_lineage	db19c6e5-98cc-4eba-9311-7d9ac31c3cf2	2025-07-03 08:23:16.858263-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:16.858263-07	f	\N
85e42541-418b-4d4a-a4d1-49d0df5661bf	GL416	39ac2d80-8f21-4823-8941-abfe02e40867	container_instance	container_instance	container_instance_lineage	b00c40c2-0afb-46fa-b628-daffa4dcd1c7	2025-07-03 08:23:16.865432-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:16.865432-07	f	\N
826b746c-3175-4835-a1dd-1fa25531c5ee	GL417	39ac2d80-8f21-4823-8941-abfe02e40867	container_instance	container_instance	container_instance_lineage	662cdf0b-bf33-490a-997c-2ac9121bb7b9	2025-07-03 08:23:16.872345-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:16.872345-07	f	\N
dd9a2971-6a8d-45e8-b8ee-83c101c8bca7	GL418	39ac2d80-8f21-4823-8941-abfe02e40867	container_instance	container_instance	container_instance_lineage	b3bb08ee-3efd-406b-ac7f-0959a69c2ff5	2025-07-03 08:23:16.878044-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:16.878044-07	f	\N
14ce0dba-39dd-41b0-bc3f-72f6ce423a9d	GL419	39ac2d80-8f21-4823-8941-abfe02e40867	container_instance	container_instance	container_instance_lineage	c0ab8a50-6008-48dc-b1ed-35d21741141b	2025-07-03 08:23:16.88381-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:16.88381-07	f	\N
1e49a2c7-894f-4685-90d3-72c574cb8b29	GL420	39ac2d80-8f21-4823-8941-abfe02e40867	container_instance	container_instance	container_instance_lineage	44535f96-b261-441e-8059-2f7c5632858d	2025-07-03 08:23:16.891679-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:16.891679-07	f	\N
b5bf008b-a217-44f2-816f-63881ecf045f	GL421	39ac2d80-8f21-4823-8941-abfe02e40867	container_instance	container_instance	container_instance_lineage	4d5db19c-2950-4351-8d87-fe365be8e7aa	2025-07-03 08:23:16.898679-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:16.898679-07	f	\N
70fbbb5a-cb5f-49d1-bdc9-38445865e557	GL422	39ac2d80-8f21-4823-8941-abfe02e40867	container_instance	container_instance	container_instance_lineage	3aa39100-5672-4c10-a9f9-d0f60b0f437a	2025-07-03 08:23:16.904772-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:16.904772-07	f	\N
1f479497-6bff-486c-b662-149739c24c70	GL423	39ac2d80-8f21-4823-8941-abfe02e40867	container_instance	container_instance	container_instance_lineage	382bee5c-5a83-4842-8385-eec79639cc66	2025-07-03 08:23:16.910378-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:16.910378-07	f	\N
f82d692f-aa1e-4bea-9cd1-cccf330bc8e6	GL424	39ac2d80-8f21-4823-8941-abfe02e40867	container_instance	container_instance	container_instance_lineage	36808dc6-c5b6-423e-848d-ed03c8a4e4d4	2025-07-03 08:23:16.91574-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:16.91574-07	f	\N
50d1361b-7633-44a4-a0a6-2ab9a8040d94	GL425	39ac2d80-8f21-4823-8941-abfe02e40867	container_instance	container_instance	container_instance_lineage	304ab3f3-404d-4038-b781-6e028a3aac54	2025-07-03 08:23:16.9209-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:16.9209-07	f	\N
ee2fbff9-7804-4a34-9ef2-40c364100fda	GL426	39ac2d80-8f21-4823-8941-abfe02e40867	container_instance	container_instance	container_instance_lineage	bad6ad8a-77c4-4d48-81a1-858893e9a864	2025-07-03 08:23:16.92652-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:16.92652-07	f	\N
bbb4a931-9249-4bf4-a71f-fa7e511dc1c4	GL427	39ac2d80-8f21-4823-8941-abfe02e40867	container_instance	container_instance	container_instance_lineage	6a5c1eac-1cbf-4e8f-87dd-63ceaee4a906	2025-07-03 08:23:16.932457-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:16.932457-07	f	\N
a24fd27a-cd4e-4640-8230-bd6cdc0238b7	GL428	39ac2d80-8f21-4823-8941-abfe02e40867	container_instance	container_instance	container_instance_lineage	c45f4209-c43e-4e0b-a0c6-4d8a5ea4d202	2025-07-03 08:23:16.938673-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:16.938673-07	f	\N
e6006c9a-7cf0-4e4e-bc4b-97570b66418c	GL429	39ac2d80-8f21-4823-8941-abfe02e40867	container_instance	container_instance	container_instance_lineage	cfba7c0a-1491-40ab-9d5c-27d8c5ce367e	2025-07-03 08:23:16.944108-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:16.944108-07	f	\N
33fb8cab-f33a-45f3-915f-2fab93aa6545	GL430	39ac2d80-8f21-4823-8941-abfe02e40867	container_instance	container_instance	container_instance_lineage	92af61ad-da5e-4378-bf20-1d69de145551	2025-07-03 08:23:16.94952-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:16.94952-07	f	\N
64b043b9-29e4-48fb-b4af-03486b298b9e	GL431	39ac2d80-8f21-4823-8941-abfe02e40867	container_instance	container_instance	container_instance_lineage	8119664c-7b20-4d9d-b788-7ddefbc086e4	2025-07-03 08:23:16.955038-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:16.955038-07	f	\N
73735461-d4d8-4f6a-86bd-8b566d592fd7	GL432	39ac2d80-8f21-4823-8941-abfe02e40867	container_instance	container_instance	container_instance_lineage	be75f19a-9204-45e4-ac52-c07157044eb9	2025-07-03 08:23:16.96069-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:16.96069-07	f	\N
27560227-45fc-4f81-a66e-ac717bf72320	GL433	39ac2d80-8f21-4823-8941-abfe02e40867	container_instance	container_instance	container_instance_lineage	03fd4454-0700-4766-adf4-40cd1f0847a2	2025-07-03 08:23:16.968505-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:16.968505-07	f	\N
e5518eff-c507-4a67-8647-582810906e88	GL434	39ac2d80-8f21-4823-8941-abfe02e40867	container_instance	container_instance	container_instance_lineage	f653b34e-100a-46a3-8ab7-9afaa9cd74d0	2025-07-03 08:23:16.974229-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:16.974229-07	f	\N
3da3c520-44d8-408d-bce2-574814903a11	GL439	39ac2d80-8f21-4823-8941-abfe02e40867	container_instance	container_instance	container_instance_lineage	bbb4417b-7369-41c9-91b0-8fb0773fdab6	2025-07-03 08:23:17.003419-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.003419-07	f	\N
dc00b76c-2219-4851-a0bf-15ca81b49e4e	GL440	39ac2d80-8f21-4823-8941-abfe02e40867	container_instance	container_instance	container_instance_lineage	cf2a5e5f-bd04-40fa-a03e-f3efecc404b3	2025-07-03 08:23:17.008874-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.008874-07	f	\N
d89a3d71-6866-47df-81ac-4e80f39eeef5	GL441	39ac2d80-8f21-4823-8941-abfe02e40867	container_instance	container_instance	container_instance_lineage	6077eb5c-7bd2-436b-85a4-f1ec9e26cc5c	2025-07-03 08:23:17.013944-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.013944-07	f	\N
871df990-bfd8-4d14-a5e5-2f40cc17e76a	GL442	39ac2d80-8f21-4823-8941-abfe02e40867	container_instance	container_instance	container_instance_lineage	4218d37a-128b-45cf-b2ef-1a111112b524	2025-07-03 08:23:17.01932-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.01932-07	f	\N
ff8de42c-43e0-4b9e-91d5-3dd56d1bef14	GL443	db19c6e5-98cc-4eba-9311-7d9ac31c3cf2	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	82f7d2fc-6a90-4574-a482-546dc5be34c5	2025-07-03 08:23:17.023102-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.023102-07	f	generic
e09f0196-336a-4f70-97c5-428580379204	GL444	b00c40c2-0afb-46fa-b628-daffa4dcd1c7	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	c5d41a18-1176-4b9d-bfd7-e222b66b1dfc	2025-07-03 08:23:17.029585-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.029585-07	f	generic
c9f99b5f-6046-487a-8b40-55df8e34e46f	GL445	662cdf0b-bf33-490a-997c-2ac9121bb7b9	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	ba952458-8cff-4ff5-a9d2-0a6ab5405602	2025-07-03 08:23:17.034913-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.034913-07	f	generic
3c7a4747-1927-42f2-8229-775b1d5a68ff	GL491	1852629f-beb7-4f69-98e1-971a34af9928	container_instance	container_instance	container_instance_lineage	8a96e5af-aa96-4369-9f19-a004222d92bf	2025-07-03 08:23:17.29493-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.29493-07	f	\N
b7eb8ee6-4cba-41c8-885f-1470368a38b3	GL492	1852629f-beb7-4f69-98e1-971a34af9928	container_instance	container_instance	container_instance_lineage	dd25a9e1-d8dc-4eea-be62-2d3d6a4a7829	2025-07-03 08:23:17.301168-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.301168-07	f	\N
743efc0f-8bcd-435a-9f46-4f3dc62d85a6	GL446	b3bb08ee-3efd-406b-ac7f-0959a69c2ff5	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	525493a1-c68d-4d24-a2ee-d51be1599878	2025-07-03 08:23:17.039681-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.039681-07	f	generic
94ad3b04-a6fa-430a-acba-645ce02212fa	GL447	c0ab8a50-6008-48dc-b1ed-35d21741141b	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	ef23526a-8c13-4561-b742-e38af3f43e6d	2025-07-03 08:23:17.044588-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.044588-07	f	generic
93037a80-2592-47d1-9806-8f9b979149f5	GL448	44535f96-b261-441e-8059-2f7c5632858d	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	a6a0f6f1-6a24-475e-919b-c0fbe85a297c	2025-07-03 08:23:17.049301-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.049301-07	f	generic
37476cb1-99c8-423d-a35e-8a806250b1b9	GL449	4d5db19c-2950-4351-8d87-fe365be8e7aa	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	7ad6a870-8313-4efb-ba81-5f4fb0ab06c3	2025-07-03 08:23:17.053597-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.053597-07	f	generic
5ad35ea5-0160-4b55-b204-cfee16c12b02	GL450	3aa39100-5672-4c10-a9f9-d0f60b0f437a	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	2fd021db-8cba-4e2b-9550-c0f078ed5ca5	2025-07-03 08:23:17.059703-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.059703-07	f	generic
c169b2c3-5d29-4c89-9d27-0c73abf43119	GL451	382bee5c-5a83-4842-8385-eec79639cc66	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	7c504afb-63e5-4f3e-9684-d7a685a74f33	2025-07-03 08:23:17.067761-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.067761-07	f	generic
9f726eea-7186-421e-b3f4-d28c19b15fc4	GL452	36808dc6-c5b6-423e-848d-ed03c8a4e4d4	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	3c9ca22e-fa5e-4304-916b-d9b14224b0da	2025-07-03 08:23:17.072967-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.072967-07	f	generic
edb196ae-ea93-47f3-afe8-77dbb23a98c6	GL453	304ab3f3-404d-4038-b781-6e028a3aac54	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	47982770-15aa-46c6-bcb2-2dcfe0cf9fde	2025-07-03 08:23:17.077892-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.077892-07	f	generic
f2dc86cc-069e-43ba-8ee1-84b2417f8a22	GL454	bad6ad8a-77c4-4d48-81a1-858893e9a864	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	864663d1-dd5c-49ea-80c5-bcc1afaf6049	2025-07-03 08:23:17.082628-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.082628-07	f	generic
484a0633-3de3-4c63-ab69-983e87374dc0	GL455	6a5c1eac-1cbf-4e8f-87dd-63ceaee4a906	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	da2ddfd1-f611-4c8f-b85b-1c5e8f4d1a3f	2025-07-03 08:23:17.087322-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.087322-07	f	generic
9c1a6464-47ce-4a78-bc4a-2b812bfad394	GL456	c45f4209-c43e-4e0b-a0c6-4d8a5ea4d202	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	7aff050c-94f4-4a86-9bd3-e5ce0b2bcb7b	2025-07-03 08:23:17.091748-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.091748-07	f	generic
ddc86ea0-9241-4da0-bd9d-4147c24a40ef	GL457	cfba7c0a-1491-40ab-9d5c-27d8c5ce367e	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	e0a2ec5b-c727-46ff-8cfd-541ded1a37da	2025-07-03 08:23:17.096479-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.096479-07	f	generic
6179cc59-385a-4a4a-8cc3-3b1c3fd6dd10	GL458	92af61ad-da5e-4378-bf20-1d69de145551	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	461bb485-e99a-4c28-8499-92794e89166b	2025-07-03 08:23:17.101187-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.101187-07	f	generic
28a09cce-5e74-4bcf-9402-f2737dac2e7c	GL459	8119664c-7b20-4d9d-b788-7ddefbc086e4	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	952fb2a0-50f8-426f-baaa-95d86e9af105	2025-07-03 08:23:17.105736-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.105736-07	f	generic
a5d3b00d-160b-4960-baab-4b81a68d3210	GL460	be75f19a-9204-45e4-ac52-c07157044eb9	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	cb217410-9c27-445b-82c2-03e605c21be5	2025-07-03 08:23:17.110472-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.110472-07	f	generic
f01fc837-0b95-4eb1-ae0e-efb1bc4352b4	GL461	03fd4454-0700-4766-adf4-40cd1f0847a2	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	c7c571ec-1754-432e-b4e7-a8b18a0dc670	2025-07-03 08:23:17.115343-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.115343-07	f	generic
7c4c632e-22d5-45b2-bce0-1e207ef28dca	GL462	f653b34e-100a-46a3-8ab7-9afaa9cd74d0	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	e6c30471-04b1-41ce-a0f9-dea3962bb40f	2025-07-03 08:23:17.120297-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.120297-07	f	generic
11d20684-9b3c-4482-b247-2df05f3bb0ce	GL463	db0c762c-c36b-42de-bc97-6976ba8c53be	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	61bd9a3a-9f50-40fa-bcb0-51a811d28a48	2025-07-03 08:23:17.124922-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.124922-07	f	generic
e7d5efbd-d094-4870-afbe-75756790f551	GL464	01b72489-6786-4b1f-a8b7-25b56770d48f	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	5b503e9f-0587-4d68-a122-eee644adffb3	2025-07-03 08:23:17.129883-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.129883-07	f	generic
0cef7180-14f3-42cd-a30e-5b4afa6f9143	GL465	70bfdd8e-9926-4052-a785-1555e295832a	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	11bf48ed-61ca-4a6f-ad36-76f4f2eb9f55	2025-07-03 08:23:17.134831-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.134831-07	f	generic
c99bbeeb-3988-463d-ac06-71c63e29b0a5	GL466	ed6bd8d5-9e73-419c-afd6-6399c05e75a3	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	070fcb47-fb22-499a-afa0-e379811d58c6	2025-07-03 08:23:17.141995-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.141995-07	f	generic
aef6ce20-c08b-48ea-b26d-c951053fdb4f	GL467	bbb4417b-7369-41c9-91b0-8fb0773fdab6	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	1ad3b52b-32cc-4ab5-a2b4-deb1c2a81bf2	2025-07-03 08:23:17.147896-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.147896-07	f	generic
19cf5f23-9718-496a-a07a-016fe8ee09a6	GL468	cf2a5e5f-bd04-40fa-a03e-f3efecc404b3	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	44cbde60-02c7-44d2-9032-fe68359e4f02	2025-07-03 08:23:17.152976-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.152976-07	f	generic
b6822c5d-b88d-40dc-a9b2-1c6753b477b2	GL469	6077eb5c-7bd2-436b-85a4-f1ec9e26cc5c	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	0c7c36b7-7848-4205-af9c-ce41e543fd35	2025-07-03 08:23:17.157502-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.157502-07	f	generic
d7c796a9-e84f-47ac-a58a-2308a0b3a9fe	GL470	4218d37a-128b-45cf-b2ef-1a111112b524	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	35697e67-053e-4fbb-ac45-2223813e0a10	2025-07-03 08:23:17.162168-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.162168-07	f	generic
3dea1ad0-ff27-43bf-83fe-cb70f0bd154a	GL471	1852629f-beb7-4f69-98e1-971a34af9928	container_instance	container_instance	container_instance_lineage	4b6dd84b-e2c0-413d-aa2a-2a520af200af	2025-07-03 08:23:17.174531-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.174531-07	f	\N
f003b658-5069-4534-840c-5ed3ed5ea149	GL472	1852629f-beb7-4f69-98e1-971a34af9928	container_instance	container_instance	container_instance_lineage	e2728bf3-67bc-4811-bbe4-f0867231f6e0	2025-07-03 08:23:17.180834-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.180834-07	f	\N
fbcb4f04-2ebd-4692-9fa7-b493a77db7e8	GL473	1852629f-beb7-4f69-98e1-971a34af9928	container_instance	container_instance	container_instance_lineage	521b22bc-7d76-41ca-b735-17767c6be7bb	2025-07-03 08:23:17.187614-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.187614-07	f	\N
436965c6-433f-4bca-b2c7-22f0d14da31b	GL474	1852629f-beb7-4f69-98e1-971a34af9928	container_instance	container_instance	container_instance_lineage	7de76476-6e33-4d96-9a7b-333393b79eb4	2025-07-03 08:23:17.192961-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.192961-07	f	\N
101b2aa8-0038-41be-9569-78350b55685d	GL475	1852629f-beb7-4f69-98e1-971a34af9928	container_instance	container_instance	container_instance_lineage	c938574e-08aa-44ce-9f09-460240d55a12	2025-07-03 08:23:17.198868-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.198868-07	f	\N
8dcc3dcf-7e45-4b0d-9b77-ed993558750d	GL476	1852629f-beb7-4f69-98e1-971a34af9928	container_instance	container_instance	container_instance_lineage	c43f8239-2b30-4153-8e0e-1ca8109d4847	2025-07-03 08:23:17.20482-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.20482-07	f	\N
b1581797-e19f-4c79-b915-9a3b77b87fe7	GL477	1852629f-beb7-4f69-98e1-971a34af9928	container_instance	container_instance	container_instance_lineage	38d72d79-0c20-43bf-a047-265347aaf013	2025-07-03 08:23:17.210421-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.210421-07	f	\N
de715d8c-6933-4329-81cc-1f373d667034	GL478	1852629f-beb7-4f69-98e1-971a34af9928	container_instance	container_instance	container_instance_lineage	ad4b62e4-5a8f-4bb6-963b-923da6668cf8	2025-07-03 08:23:17.21569-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.21569-07	f	\N
90a878b6-8fcd-4d61-afa8-21e1768824ae	GL479	1852629f-beb7-4f69-98e1-971a34af9928	container_instance	container_instance	container_instance_lineage	45fef90b-f603-46b7-a07f-206b191d251e	2025-07-03 08:23:17.220545-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.220545-07	f	\N
92e8b630-74d0-461e-a534-fa5c88614b3f	GL480	1852629f-beb7-4f69-98e1-971a34af9928	container_instance	container_instance	container_instance_lineage	2d81e507-37d2-4131-b774-958d19466c06	2025-07-03 08:23:17.225911-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.225911-07	f	\N
0832cc59-ca71-4460-b826-828898f42e1e	GL481	1852629f-beb7-4f69-98e1-971a34af9928	container_instance	container_instance	container_instance_lineage	40c85e8c-792d-49af-930a-91f33e352dab	2025-07-03 08:23:17.23526-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.23526-07	f	\N
56d1a636-c82f-41a3-92b3-afba14b16911	GL482	1852629f-beb7-4f69-98e1-971a34af9928	container_instance	container_instance	container_instance_lineage	7ed48914-f3a4-462b-baaa-845a9b3b31e6	2025-07-03 08:23:17.243204-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.243204-07	f	\N
b446d6cb-f7f3-49e1-9650-7b6b0519af6d	GL483	1852629f-beb7-4f69-98e1-971a34af9928	container_instance	container_instance	container_instance_lineage	3802ef66-effb-4e9a-bcf9-d1ed8c84beba	2025-07-03 08:23:17.249395-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.249395-07	f	\N
711ed583-eefa-4f39-8784-645d104d9648	GL484	1852629f-beb7-4f69-98e1-971a34af9928	container_instance	container_instance	container_instance_lineage	a9cc088c-0aec-4d5f-a5ab-fae4fe7f030f	2025-07-03 08:23:17.255731-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.255731-07	f	\N
e86af435-51d6-46ff-aeb2-36e10bbee947	GL485	1852629f-beb7-4f69-98e1-971a34af9928	container_instance	container_instance	container_instance_lineage	fc480673-f73e-44cb-b6d4-64df95e6b3d0	2025-07-03 08:23:17.261154-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.261154-07	f	\N
9b86fafd-0e91-41b3-9958-2e341240e13d	GL486	1852629f-beb7-4f69-98e1-971a34af9928	container_instance	container_instance	container_instance_lineage	8744e5ca-a013-46fc-99ac-b67f436d0827	2025-07-03 08:23:17.267378-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.267378-07	f	\N
ce43688c-c273-4f42-ab5a-38e2f0be6274	GL487	1852629f-beb7-4f69-98e1-971a34af9928	container_instance	container_instance	container_instance_lineage	1cf84a4a-f071-46fa-ae06-d39f5f8aa6df	2025-07-03 08:23:17.272936-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.272936-07	f	\N
eb343497-2a96-4f4e-82dd-f15e3c19ed8a	GL488	1852629f-beb7-4f69-98e1-971a34af9928	container_instance	container_instance	container_instance_lineage	fb64119b-a9bf-4920-ac9f-42025d6486cd	2025-07-03 08:23:17.278636-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.278636-07	f	\N
754681d9-0ece-4b7f-bf4d-e8dc2d256a6d	GL489	1852629f-beb7-4f69-98e1-971a34af9928	container_instance	container_instance	container_instance_lineage	647cfc65-654d-4ff4-98b5-737645b3e930	2025-07-03 08:23:17.28404-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.28404-07	f	\N
1a714a26-b5da-442a-94bb-a4a2bb5dc5ea	GL490	1852629f-beb7-4f69-98e1-971a34af9928	container_instance	container_instance	container_instance_lineage	73d57912-9cb8-44c9-88a1-34e13b8b65fd	2025-07-03 08:23:17.289374-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.289374-07	f	\N
fc077a22-296f-4e69-a3a4-6d40439d18ea	GL493	1852629f-beb7-4f69-98e1-971a34af9928	container_instance	container_instance	container_instance_lineage	23e5c523-46c3-4f9b-a0ca-e30781d73d3c	2025-07-03 08:23:17.306639-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.306639-07	f	\N
d70e0389-0c1b-4a2a-92c4-1a5dc0511299	GL494	1852629f-beb7-4f69-98e1-971a34af9928	container_instance	container_instance	container_instance_lineage	6d930317-d930-41e8-95c6-dccd9add54ff	2025-07-03 08:23:17.312181-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.312181-07	f	\N
f1c9dfb8-17d8-43a0-9d0f-7cafa83b3f4c	GL495	1852629f-beb7-4f69-98e1-971a34af9928	container_instance	container_instance	container_instance_lineage	4ff0926c-88e1-40fd-90ac-dafe861eb746	2025-07-03 08:23:17.317516-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.317516-07	f	\N
478f219f-c250-4739-aa4f-89be44446078	GL496	1852629f-beb7-4f69-98e1-971a34af9928	container_instance	container_instance	container_instance_lineage	a72a0a3c-2e05-4aa3-bb8f-1d24163a15fa	2025-07-03 08:23:17.323224-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.323224-07	f	\N
2adb0e7f-f3b2-4acc-a2d8-f99eea0cf1f9	GL497	1852629f-beb7-4f69-98e1-971a34af9928	container_instance	container_instance	container_instance_lineage	ebd9b442-c35d-4c7f-a9d3-514321045d51	2025-07-03 08:23:17.328821-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.328821-07	f	\N
57a29989-1636-4134-8a6d-0dddf3c0745a	GL498	1852629f-beb7-4f69-98e1-971a34af9928	container_instance	container_instance	container_instance_lineage	116a5243-bb5e-4e4c-b293-c07586d1d79e	2025-07-03 08:23:17.336917-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.336917-07	f	\N
d9c50ec5-f17c-4f1d-9bd5-b2598006394b	GL499	4b6dd84b-e2c0-413d-aa2a-2a520af200af	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	a7fdfb14-fc01-4349-a8ad-8de611a15b1c	2025-07-03 08:23:17.342965-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.342965-07	f	generic
208fabab-e2e4-4232-b8cb-99afe8c25b9f	GL500	e2728bf3-67bc-4811-bbe4-f0867231f6e0	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	8205e9ec-903c-4f91-bc3b-bbdb81479279	2025-07-03 08:23:17.347665-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.347665-07	f	generic
91603704-a959-40ac-9fb5-afe7e62a93e0	GL501	521b22bc-7d76-41ca-b735-17767c6be7bb	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	4e28eeb7-6056-407f-92d0-53a349a90407	2025-07-03 08:23:17.354348-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.354348-07	f	generic
ad6515f0-ae8c-4813-98af-f1ad0a80bd22	GL533	43b72609-8eab-47a3-9cd2-fc54e4092892	container_instance	container_instance	container_instance_lineage	245e0cf5-ffc9-4b32-bd05-857c40d77e4d	2025-07-03 08:23:17.549119-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.549119-07	f	\N
225dfd3a-ab52-453e-b228-f639f4607292	GL502	7de76476-6e33-4d96-9a7b-333393b79eb4	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	7a771f0c-8c4e-4184-91f0-be5b15eb4a1d	2025-07-03 08:23:17.359712-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.359712-07	f	generic
e7d98d83-461d-46e1-b18b-860a4fa58d1a	GL503	c938574e-08aa-44ce-9f09-460240d55a12	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	b348c7b2-83c2-4c6e-ac63-25d711fed70a	2025-07-03 08:23:17.364937-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.364937-07	f	generic
dd3a1113-92e0-4264-a566-0ac6be54223d	GL504	c43f8239-2b30-4153-8e0e-1ca8109d4847	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	ad0b6013-0a58-4c0d-a2a2-acf5cf2fcf0b	2025-07-03 08:23:17.36972-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.36972-07	f	generic
748d5cb4-3405-421e-9e56-fdafcb8228f6	GL505	38d72d79-0c20-43bf-a047-265347aaf013	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	269038aa-810c-4401-9ae5-b2474d3afe38	2025-07-03 08:23:17.374602-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.374602-07	f	generic
d26d60f3-7c03-424b-ad7a-d39921a63425	GL506	ad4b62e4-5a8f-4bb6-963b-923da6668cf8	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	538a6afc-a419-46c2-b3b1-79ac907c1b6f	2025-07-03 08:23:17.379232-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.379232-07	f	generic
c449c4f7-937f-4534-9a18-98ad969e82cf	GL507	45fef90b-f603-46b7-a07f-206b191d251e	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	376ecd05-c2eb-4b37-b19b-948ef80045b4	2025-07-03 08:23:17.383898-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.383898-07	f	generic
7a7224e8-222a-42a7-83cf-c40ee155501c	GL508	2d81e507-37d2-4131-b774-958d19466c06	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	6f2d6e6f-012e-46fc-82d9-8323bc889a37	2025-07-03 08:23:17.388733-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.388733-07	f	generic
84e4b868-c7d5-46fa-ac1c-9948f2b25016	GL509	40c85e8c-792d-49af-930a-91f33e352dab	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	009eef87-568f-4b42-8aff-4d78a5c238af	2025-07-03 08:23:17.393467-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.393467-07	f	generic
c9f43a9a-337e-4d13-ac73-8706fcb15a20	GL510	7ed48914-f3a4-462b-baaa-845a9b3b31e6	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	b2c3158a-af0b-43d2-a6fe-6b005335a909	2025-07-03 08:23:17.398828-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.398828-07	f	generic
d93f0d46-960f-42f5-874c-dcd85cec0337	GL511	3802ef66-effb-4e9a-bcf9-d1ed8c84beba	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	7ec4cacc-52bd-4a88-bfcd-3167baeab3a2	2025-07-03 08:23:17.404929-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.404929-07	f	generic
fae8a86e-5587-403c-9648-8f6826596e9b	GL512	a9cc088c-0aec-4d5f-a5ab-fae4fe7f030f	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	bda1c743-e436-4a9c-a9fc-71f22cda28c1	2025-07-03 08:23:17.409505-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.409505-07	f	generic
d005a268-21e0-47cf-bbc4-8694d07be3b7	GL513	fc480673-f73e-44cb-b6d4-64df95e6b3d0	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	c17a9260-b36f-4a62-b076-e13de08d64cb	2025-07-03 08:23:17.416652-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.416652-07	f	generic
761d0c36-8f03-491e-8f44-0bd503b7cc0e	GL514	8744e5ca-a013-46fc-99ac-b67f436d0827	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	498ad847-9c74-4b5a-8725-0a56a3c84f94	2025-07-03 08:23:17.422874-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.422874-07	f	generic
2c081129-7e27-4ab7-8d41-2d6e0227837e	GL515	1cf84a4a-f071-46fa-ae06-d39f5f8aa6df	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	9abbbeeb-2a9f-4a05-95f5-070fdc8c3483	2025-07-03 08:23:17.427323-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.427323-07	f	generic
400b16bf-edce-406b-bc9b-7ffab4cfe2c2	GL516	fb64119b-a9bf-4920-ac9f-42025d6486cd	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	58b3a670-dedc-4045-b70d-d90479915351	2025-07-03 08:23:17.431746-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.431746-07	f	generic
ebfaecf1-8a8c-4b7e-b55f-c680c68c5949	GL517	647cfc65-654d-4ff4-98b5-737645b3e930	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	ef8398cc-3507-42a0-b8f5-99c0a9d033da	2025-07-03 08:23:17.436354-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.436354-07	f	generic
a655b533-585a-4ddf-9581-6512e81ae166	GL518	73d57912-9cb8-44c9-88a1-34e13b8b65fd	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	04853cf5-aee7-4abf-9240-600265068cb2	2025-07-03 08:23:17.440646-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.440646-07	f	generic
0b8f0035-8c7a-4c51-8eb4-08d77e18103b	GL519	8a96e5af-aa96-4369-9f19-a004222d92bf	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	9401369a-7357-4e36-b527-2d208017597d	2025-07-03 08:23:17.445219-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.445219-07	f	generic
09a206d5-d266-4242-8f79-c83cc5366fe7	GL520	dd25a9e1-d8dc-4eea-be62-2d3d6a4a7829	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	e6868a3f-63f7-4302-be5e-70a1bc5ec46e	2025-07-03 08:23:17.449624-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.449624-07	f	generic
97d1bdcb-bcf8-4f31-88de-5e0ba783e762	GL521	23e5c523-46c3-4f9b-a0ca-e30781d73d3c	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	8269b692-aa3c-45f6-9b49-415646e12639	2025-07-03 08:23:17.454094-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.454094-07	f	generic
70b892ad-d58c-42c2-8152-e0b156e50d96	GL522	6d930317-d930-41e8-95c6-dccd9add54ff	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	dfcce43b-c1c2-4169-aca9-4e84cb8f9eb9	2025-07-03 08:23:17.459277-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.459277-07	f	generic
ddf496e1-7004-4cd9-873d-ca9ac78c5c8e	GL523	4ff0926c-88e1-40fd-90ac-dafe861eb746	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	4380dfd5-bc54-4a8f-9866-4012f2b64ab4	2025-07-03 08:23:17.464105-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.464105-07	f	generic
ecd66eb6-753d-45ae-ba80-6ba2e90bd86c	GL524	a72a0a3c-2e05-4aa3-bb8f-1d24163a15fa	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	7191f250-f552-4d76-9ecd-78256934407d	2025-07-03 08:23:17.471022-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.471022-07	f	generic
ed977d1c-245a-4c41-8a4d-2db1f1485b13	GL525	ebd9b442-c35d-4c7f-a9d3-514321045d51	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	5e7485c4-94f0-4ff4-b03a-fd6f67a262d0	2025-07-03 08:23:17.475801-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.475801-07	f	generic
ae9018b8-f4d5-4e32-a17c-1d40eb514619	GL526	116a5243-bb5e-4e4c-b293-c07586d1d79e	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	275089ba-354d-4b77-a90b-c9eb2be85073	2025-07-03 08:23:17.480331-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.480331-07	f	generic
c4bb2d2b-0eb9-4e7a-b314-4a4ca8af24c1	GL527	d6dc35b7-cf4c-4e70-a692-c23346f61bcd	container:tube:tube-generic-10ml:1.0	content:control:giab-HG002:1.0	generic_instance_lineage	f9786af0-c6a2-470c-b39b-8bc18c43c409	2025-07-03 08:23:17.494103-07	tube:tube-generic-10ml:1.0 :: control:giab-HG002:1.0	generic	tube	tube-generic-10ml	{"cogs": {"cost": "1.50", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "giab NA24385", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "Generic 10ml tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.494103-07	f	generic
3a410823-ee1b-4542-bd9a-d41c070c02a2	GL528	9c6dc920-ebc9-44ca-bf88-13f9ca571726	container:tube:tube-10ml-glass:1.0	content:control:synthetic-cfdna:1.0	generic_instance_lineage	cab1f6d5-eb37-459c-8fc4-4e3a8e81855f	2025-07-03 08:23:17.505305-07	tube:tube-10ml-glass:1.0 :: control:synthetic-cfdna:1.0	generic	tube	tube-10ml-glass	{"cogs": {"cost": "4.70", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "s ynthetic-cfdna", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "10ml glass tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.505305-07	f	generic
ef0f706f-574e-4646-8556-90f289c2653e	GL529	fe1b21ed-3957-4df1-b13c-fcaccc035a83	container:tube:tube-eppi-1.5ml:1.0	content:control:water-ntc:1.0	generic_instance_lineage	7fcfa70d-2016-45c2-a092-4f42ebed1393	2025-07-03 08:23:17.514422-07	tube:tube-eppi-1.5ml:1.0 :: control:water-ntc:1.0	generic	tube	tube-eppi-1.5ml	{"cogs": {"cost": "1.70", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "NTC - water", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "1.5ml eppi tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.514422-07	f	generic
91f5a54b-5d03-4ffe-a643-76c4370b80e9	GL530	43b72609-8eab-47a3-9cd2-fc54e4092892	container_instance	container_instance	container_instance_lineage	67da16e4-6e1d-451f-9319-d3b627ef5a3e	2025-07-03 08:23:17.532606-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.532606-07	f	\N
e07a7bdc-4437-469c-88d1-ab4e8f2e0b7b	GL531	43b72609-8eab-47a3-9cd2-fc54e4092892	container_instance	container_instance	container_instance_lineage	3ab97a64-e71c-41b1-b799-60b928503b2e	2025-07-03 08:23:17.538143-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.538143-07	f	\N
00c14f88-0d17-46a0-b361-30b1db2e82be	GL532	43b72609-8eab-47a3-9cd2-fc54e4092892	container_instance	container_instance	container_instance_lineage	1773aa0b-fcb9-440c-99ba-e97d08c897d2	2025-07-03 08:23:17.543423-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.543423-07	f	\N
a66eb90a-d406-4b1b-8a15-e8e68faa0a5a	GL534	43b72609-8eab-47a3-9cd2-fc54e4092892	container_instance	container_instance	container_instance_lineage	3f3515b1-157e-4537-9270-70187b0b55b2	2025-07-03 08:23:17.554699-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.554699-07	f	\N
cbabf5f8-b703-43ee-8dd5-3bc08bbe8cbf	GL535	43b72609-8eab-47a3-9cd2-fc54e4092892	container_instance	container_instance	container_instance_lineage	6282522a-6923-4b31-8c11-9945e5d04e9c	2025-07-03 08:23:17.560004-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.560004-07	f	\N
206bdf72-c5b7-4fd1-b83a-523d1d96ac15	GL536	43b72609-8eab-47a3-9cd2-fc54e4092892	container_instance	container_instance	container_instance_lineage	743e4c9a-0b92-48f1-b15e-aa89d9ef6f17	2025-07-03 08:23:17.566042-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.566042-07	f	\N
abcd409c-29c1-4b65-98e5-11f2f9263ddd	GL537	43b72609-8eab-47a3-9cd2-fc54e4092892	container_instance	container_instance	container_instance_lineage	3fb7f5a8-6c7b-44c6-af3b-875bb8f8fcc9	2025-07-03 08:23:17.571878-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.571878-07	f	\N
813107eb-d656-4fb7-aab3-8d12a0e8711d	GL538	43b72609-8eab-47a3-9cd2-fc54e4092892	container_instance	container_instance	container_instance_lineage	24c55350-7ef4-4463-90ab-46252ab1c1c5	2025-07-03 08:23:17.577992-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.577992-07	f	\N
40c0072f-2304-438d-ac92-c751dc363d54	GL539	43b72609-8eab-47a3-9cd2-fc54e4092892	container_instance	container_instance	container_instance_lineage	8ad330a0-fad3-4b94-be1d-6c2aade5b972	2025-07-03 08:23:17.583496-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.583496-07	f	\N
95530e78-12ef-44c1-80f4-d1edb0c6abf1	GL540	43b72609-8eab-47a3-9cd2-fc54e4092892	container_instance	container_instance	container_instance_lineage	801a01d3-8107-4911-9822-69f73fef2dee	2025-07-03 08:23:17.589387-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.589387-07	f	\N
05f8ddab-694b-462d-92e7-3e4d8f1c88ca	GL541	43b72609-8eab-47a3-9cd2-fc54e4092892	container_instance	container_instance	container_instance_lineage	f6e70812-dce0-4359-848a-7c86a15b25eb	2025-07-03 08:23:17.594959-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.594959-07	f	\N
fdfe2b7f-84f2-4096-8e9a-1223df323f2d	GL542	43b72609-8eab-47a3-9cd2-fc54e4092892	container_instance	container_instance	container_instance_lineage	226cfe6d-f38c-4275-b630-54bc1a181c4f	2025-07-03 08:23:17.605463-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.605463-07	f	\N
24a4da19-904a-4b6b-9215-1c3bf022926c	GL543	43b72609-8eab-47a3-9cd2-fc54e4092892	container_instance	container_instance	container_instance_lineage	6d5743cc-81b2-46c0-bf95-ec24bf59c944	2025-07-03 08:23:17.612317-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.612317-07	f	\N
04b0a9cf-6877-41d3-b1bc-b06cbdc6c386	GL544	43b72609-8eab-47a3-9cd2-fc54e4092892	container_instance	container_instance	container_instance_lineage	e63af02f-2942-4849-a657-ab339e79a86c	2025-07-03 08:23:17.618607-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.618607-07	f	\N
7626f0e3-d4d8-4151-9e78-5f37f2bd2533	GL545	43b72609-8eab-47a3-9cd2-fc54e4092892	container_instance	container_instance	container_instance_lineage	ab47b771-3bc5-4ec6-88ea-14a05ab36fe7	2025-07-03 08:23:17.625699-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.625699-07	f	\N
fcb84219-4735-4483-bf9a-1d6fb3746398	GL546	43b72609-8eab-47a3-9cd2-fc54e4092892	container_instance	container_instance	container_instance_lineage	32a38823-7229-4403-b7c5-4564b57105f6	2025-07-03 08:23:17.631317-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.631317-07	f	\N
ae2cc2a9-2643-4b26-bb1b-048663237d5e	GL547	43b72609-8eab-47a3-9cd2-fc54e4092892	container_instance	container_instance	container_instance_lineage	ed9a04fc-0e00-4fd1-8813-54c833e62db0	2025-07-03 08:23:17.685254-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.685254-07	f	\N
f78e5840-de52-49a1-9f70-67ed6f26f822	GL548	43b72609-8eab-47a3-9cd2-fc54e4092892	container_instance	container_instance	container_instance_lineage	39bff9ce-0ded-465c-b704-dda606530e7a	2025-07-03 08:23:17.703574-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.703574-07	f	\N
86c40630-ff92-4e95-a713-dd50a3d7a8a8	GL549	43b72609-8eab-47a3-9cd2-fc54e4092892	container_instance	container_instance	container_instance_lineage	fa6b2d22-7543-4429-b740-abd29980cabb	2025-07-03 08:23:17.709524-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.709524-07	f	\N
2bda3132-9788-40e7-ac3a-58fac4c774b7	GL550	43b72609-8eab-47a3-9cd2-fc54e4092892	container_instance	container_instance	container_instance_lineage	2fe1c959-22e0-457c-b5f1-2e043fe9c6e4	2025-07-03 08:23:17.714876-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.714876-07	f	\N
c1d9a385-213c-45a7-8ef9-0e4d0054f47a	GL551	43b72609-8eab-47a3-9cd2-fc54e4092892	container_instance	container_instance	container_instance_lineage	c42216f6-5feb-418b-af3c-792c2e43b666	2025-07-03 08:23:17.720404-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.720404-07	f	\N
f8f1ced5-70e6-4a61-a9d9-d06d1f4ec0a5	GL552	43b72609-8eab-47a3-9cd2-fc54e4092892	container_instance	container_instance	container_instance_lineage	7207e508-99fc-41ef-91ab-fd60d5306f4e	2025-07-03 08:23:17.72663-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.72663-07	f	\N
dd69217d-fc02-4b82-bd19-6e38124d5c8f	GL553	43b72609-8eab-47a3-9cd2-fc54e4092892	container_instance	container_instance	container_instance_lineage	13a17e5f-6d85-417c-9d4a-956b5282800a	2025-07-03 08:23:17.733993-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.733993-07	f	\N
fd45b679-7d8e-4072-ae36-602a54610c84	GL554	43b72609-8eab-47a3-9cd2-fc54e4092892	container_instance	container_instance	container_instance_lineage	f4382eed-95c5-4d1b-a046-f2b9b277ec34	2025-07-03 08:23:17.740122-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.740122-07	f	\N
ef95e543-129f-4d06-8bb3-b580c983cc3c	GL555	43b72609-8eab-47a3-9cd2-fc54e4092892	container_instance	container_instance	container_instance_lineage	741b104f-826e-443e-8efd-9d6aaf9ec330	2025-07-03 08:23:17.746439-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.746439-07	f	\N
0e0d51ed-217a-42e4-a32c-0ac91f8375e1	GL556	43b72609-8eab-47a3-9cd2-fc54e4092892	container_instance	container_instance	container_instance_lineage	340f2341-eb12-4958-9c03-224ed6cc0747	2025-07-03 08:23:17.753422-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.753422-07	f	\N
c9960503-0eeb-44bc-9fc8-7aa89a55623b	GL557	43b72609-8eab-47a3-9cd2-fc54e4092892	container_instance	container_instance	container_instance_lineage	c6eb89d6-4a49-4da3-a615-6be4294b9235	2025-07-03 08:23:17.759189-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:23:17.759189-07	f	\N
ace327c9-a487-436d-859b-ccc19cbb11f2	GL599	a6ab103e-ac0d-4bda-b989-6beb34419280	container_instance	container_instance	container_instance_lineage	015fbb50-97f2-409e-9b39-e5f548a21402	2025-07-03 08:24:15.793739-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:15.793739-07	f	\N
7eda35a5-c22f-4eea-a5d5-c9ca6422577e	GL600	a6ab103e-ac0d-4bda-b989-6beb34419280	container_instance	container_instance	container_instance_lineage	427711ca-ab90-45b5-b311-8733db30d9ad	2025-07-03 08:24:15.801168-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:15.801168-07	f	\N
88920470-3dea-40a5-95cd-e345224eecf6	GL601	a6ab103e-ac0d-4bda-b989-6beb34419280	container_instance	container_instance	container_instance_lineage	95d4d6ba-7610-486b-a74c-d1ad8255621a	2025-07-03 08:24:15.816498-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:15.816498-07	f	\N
991a8ecc-ff6c-43fa-bc74-8aaa5af0816b	GL558	67da16e4-6e1d-451f-9319-d3b627ef5a3e	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	0d22f68d-dd27-4e92-ba89-b39005cfca81	2025-07-03 08:23:17.766127-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.766127-07	f	generic
204e9535-3393-4a36-83ee-179de1d3e671	GL559	3ab97a64-e71c-41b1-b799-60b928503b2e	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	f186463f-a0de-4e66-9a85-60d861f29c20	2025-07-03 08:23:17.772121-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.772121-07	f	generic
a73eaa6e-cb5c-4dab-828a-0c27cdc81801	GL560	1773aa0b-fcb9-440c-99ba-e97d08c897d2	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	78f2fa4b-bfcb-46fc-a9ef-2f0bb27926f8	2025-07-03 08:23:17.781168-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.781168-07	f	generic
37a75b76-cfaf-4790-8424-090bc57d284a	GL561	245e0cf5-ffc9-4b32-bd05-857c40d77e4d	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	906cdad8-145e-4040-b4f2-5ef1f31c907b	2025-07-03 08:23:17.786967-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.786967-07	f	generic
36f5a79b-9d80-4c2a-a1f4-aaca37b2bd6a	GL562	3f3515b1-157e-4537-9270-70187b0b55b2	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	dbc67543-2fb8-4e02-949f-1c2146b9f1c3	2025-07-03 08:23:17.792875-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.792875-07	f	generic
5e3cfb28-fc2c-4fc9-aa77-fdc989624052	GL563	6282522a-6923-4b31-8c11-9945e5d04e9c	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	d9d3755e-bd33-44fe-a4de-a68f8f0e4275	2025-07-03 08:23:17.799241-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.799241-07	f	generic
c485273e-9dcf-422e-b10c-405cd3958b6f	GL564	743e4c9a-0b92-48f1-b15e-aa89d9ef6f17	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	9c4cca2b-65bf-4957-9a49-fefe346bdca8	2025-07-03 08:23:17.805147-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.805147-07	f	generic
fbd21f8d-51f8-4a37-a07c-deed87faa0e2	GL565	3fb7f5a8-6c7b-44c6-af3b-875bb8f8fcc9	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	a6902e58-6edc-4679-a335-fffd804cd1b3	2025-07-03 08:23:17.811179-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.811179-07	f	generic
b51ec6b2-0b5e-48f3-8755-de7aabf8f9a0	GL566	24c55350-7ef4-4463-90ab-46252ab1c1c5	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	4396a723-1c2c-490b-b006-94ab6438e2d8	2025-07-03 08:23:17.817124-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.817124-07	f	generic
07ef2f1c-009c-492d-9d77-72862b15a277	GL567	8ad330a0-fad3-4b94-be1d-6c2aade5b972	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	5bb46d8b-d389-4d1f-a86f-2bd41aef6aa1	2025-07-03 08:23:17.823295-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.823295-07	f	generic
dcf4e965-78bd-4eb8-a1f4-547fdc7c6d17	GL568	801a01d3-8107-4911-9822-69f73fef2dee	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	3813f28b-8d60-4dee-9752-1dabc8ba2ee2	2025-07-03 08:23:17.829526-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.829526-07	f	generic
eec2f323-46ea-4380-8c58-4b1065f85ad5	GL569	f6e70812-dce0-4359-848a-7c86a15b25eb	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	95a96de5-0287-4548-a6bc-5427e9da5086	2025-07-03 08:23:17.836604-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.836604-07	f	generic
554e32b1-9710-41fe-b057-84dfe55c8ab0	GL570	226cfe6d-f38c-4275-b630-54bc1a181c4f	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	9aea1a13-4383-4c3d-b24b-aad55d4e8111	2025-07-03 08:23:17.842567-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.842567-07	f	generic
eb262066-36a2-47f0-8401-dfcc895d3057	GL571	6d5743cc-81b2-46c0-bf95-ec24bf59c944	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	604c069f-71dd-4ee8-b4e8-ef6e7bb46e5c	2025-07-03 08:23:17.848523-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.848523-07	f	generic
b8044b01-4e2a-4b8c-953e-95d9546307a4	GL572	e63af02f-2942-4849-a657-ab339e79a86c	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	53dfbf91-c236-4d6e-b31c-a9bf4bc369a2	2025-07-03 08:23:17.855048-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.855048-07	f	generic
9560ac70-bc41-4df7-8137-2c05a37394ae	GL573	ab47b771-3bc5-4ec6-88ea-14a05ab36fe7	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	3666e200-4c24-4efc-a76b-d72a6faa5e65	2025-07-03 08:23:17.867487-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.867487-07	f	generic
5af314ed-67f6-4618-9ffa-68a2f25fd586	GL574	32a38823-7229-4403-b7c5-4564b57105f6	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	c282db24-d774-48c1-8634-eb60d7b861fc	2025-07-03 08:23:17.873982-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.873982-07	f	generic
f77365a9-4599-4ae3-9a9d-b2d387eed316	GL575	ed9a04fc-0e00-4fd1-8813-54c833e62db0	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	5a55869e-cf2a-470b-b132-b2e375b2872c	2025-07-03 08:23:17.879853-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.879853-07	f	generic
70b82c8f-8355-4f5b-9725-e4584bc36ab2	GL576	39bff9ce-0ded-465c-b704-dda606530e7a	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	2d6a5e4e-5f46-48bc-ab25-2f3f5ddb6412	2025-07-03 08:23:17.885713-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.885713-07	f	generic
8632faff-f351-441f-9874-060b94ea3d39	GL577	fa6b2d22-7543-4429-b740-abd29980cabb	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	e3be72b1-8052-4d5b-a017-c9db741c7bff	2025-07-03 08:23:17.891898-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.891898-07	f	generic
21ecb073-d644-4add-bd5e-5ca8503a60ec	GL578	2fe1c959-22e0-457c-b5f1-2e043fe9c6e4	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	c69af3be-aa7a-446d-a0c9-8a3c9d5b06f7	2025-07-03 08:23:17.900232-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.900232-07	f	generic
4109ce5e-ebe2-45ea-8d4c-cbad57f41159	GL579	c42216f6-5feb-418b-af3c-792c2e43b666	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	264f523e-dc9b-4fb7-b188-68d48691adab	2025-07-03 08:23:17.906482-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.906482-07	f	generic
61d1d09f-4fb2-4dde-9156-702072657d31	GL580	7207e508-99fc-41ef-91ab-fd60d5306f4e	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	b1f40745-93ea-4e41-8d30-5a900f02aab9	2025-07-03 08:23:17.912796-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.912796-07	f	generic
ff05c678-a518-47e4-a180-36ee383a75c3	GL581	13a17e5f-6d85-417c-9d4a-956b5282800a	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	1452bef9-03fa-4bf2-9908-b85eb998617a	2025-07-03 08:23:17.919066-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.919066-07	f	generic
9a04149d-0f0b-49a4-aea9-732058930817	GL582	f4382eed-95c5-4d1b-a046-f2b9b277ec34	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	5551791d-203e-4584-b45d-4cce837d92a7	2025-07-03 08:23:17.925122-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.925122-07	f	generic
ba879a78-3dee-475f-8ec0-53e22579cfb2	GL583	741b104f-826e-443e-8efd-9d6aaf9ec330	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	a0110d14-da41-463b-86a6-81b15bdd0775	2025-07-03 08:23:17.931001-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.931001-07	f	generic
a511a619-24ee-41a3-a8e6-ee7a7f1bbaad	GL584	340f2341-eb12-4958-9c03-224ed6cc0747	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	c9e2dab1-38f2-4ade-9c19-814d76031f2d	2025-07-03 08:23:17.936775-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.936775-07	f	generic
0401fdfb-3f0e-47f9-8bab-da87d20816bf	GL585	c6eb89d6-4a49-4da3-a615-6be4294b9235	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	9999a8ff-1161-4ba5-a8c3-347e452a5f92	2025-07-03 08:23:17.942647-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:23:17.942647-07	f	generic
5a3319de-98fc-4251-acae-a6406be2d8db	GL589	a6ab103e-ac0d-4bda-b989-6beb34419280	container_instance	container_instance	container_instance_lineage	843a9ccc-36ca-4d8b-80c8-a38487e0a12d	2025-07-03 08:24:15.690826-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:15.690826-07	f	\N
f1be33a3-0cc2-4b05-90fc-4d8a05d08c16	GL590	a6ab103e-ac0d-4bda-b989-6beb34419280	container_instance	container_instance	container_instance_lineage	37a7fae4-216d-4842-ab8f-5c048d4a2809	2025-07-03 08:24:15.704863-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:15.704863-07	f	\N
7a924f25-8fc4-499a-98f6-05a8f45a7c52	GL591	a6ab103e-ac0d-4bda-b989-6beb34419280	container_instance	container_instance	container_instance_lineage	109675e8-0498-4c1c-bad5-788a996e69a2	2025-07-03 08:24:15.712473-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:15.712473-07	f	\N
2c9f09cb-5300-4d3c-b98d-662a63a2eb52	GL592	a6ab103e-ac0d-4bda-b989-6beb34419280	container_instance	container_instance	container_instance_lineage	20febaa0-790b-4079-8124-3d2d9586d49e	2025-07-03 08:24:15.723773-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:15.723773-07	f	\N
66664f7c-b45c-4109-a12b-442de26397ce	GL593	a6ab103e-ac0d-4bda-b989-6beb34419280	container_instance	container_instance	container_instance_lineage	a0398a18-4ab8-469e-9010-55fda1de05b0	2025-07-03 08:24:15.734602-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:15.734602-07	f	\N
136d6f57-3093-4fd7-af0a-a7bc175b96f6	GL594	a6ab103e-ac0d-4bda-b989-6beb34419280	container_instance	container_instance	container_instance_lineage	4ba3e341-5121-4322-bc64-7b192e496768	2025-07-03 08:24:15.746269-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:15.746269-07	f	\N
442c993b-ad25-4467-acc7-acd96b0e797f	GL595	a6ab103e-ac0d-4bda-b989-6beb34419280	container_instance	container_instance	container_instance_lineage	c478dfca-9c72-4f15-93f1-fe3efd6274d1	2025-07-03 08:24:15.759711-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:15.759711-07	f	\N
64995098-fc5c-4cd7-bf45-d693f9e3b958	GL596	a6ab103e-ac0d-4bda-b989-6beb34419280	container_instance	container_instance	container_instance_lineage	d4fbc620-8b0d-4f35-b5f0-0b2fe2f00571	2025-07-03 08:24:15.771546-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:15.771546-07	f	\N
6a836416-bb10-49b4-b5a5-496236afd94d	GL597	a6ab103e-ac0d-4bda-b989-6beb34419280	container_instance	container_instance	container_instance_lineage	5c89cd1b-c0d0-4846-896a-ceca2f29fcb6	2025-07-03 08:24:15.778732-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:15.778732-07	f	\N
81fdfe0f-4a70-4c28-ac45-885487399eca	GL598	a6ab103e-ac0d-4bda-b989-6beb34419280	container_instance	container_instance	container_instance_lineage	5fe3fa73-82ec-4a72-a25d-f4ca3f7575d8	2025-07-03 08:24:15.786514-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:15.786514-07	f	\N
c1a191a2-5708-4838-8c42-518388d88bb4	GL602	a6ab103e-ac0d-4bda-b989-6beb34419280	container_instance	container_instance	container_instance_lineage	61bfac74-3f5c-474c-8b7a-6fe48898e386	2025-07-03 08:24:15.824383-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:15.824383-07	f	\N
f08fc12e-74c2-4b22-ae8b-73110944312f	GL603	a6ab103e-ac0d-4bda-b989-6beb34419280	container_instance	container_instance	container_instance_lineage	a1f6ccdc-9ee5-402b-a93d-533b39212bee	2025-07-03 08:24:15.83564-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:15.83564-07	f	\N
b7fb325c-b99c-4d31-91df-32d531de33ae	GL604	a6ab103e-ac0d-4bda-b989-6beb34419280	container_instance	container_instance	container_instance_lineage	4ff5edda-1503-4a1b-aa81-44f0d001dede	2025-07-03 08:24:15.84285-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:15.84285-07	f	\N
dc72e345-5707-47fb-9bee-5003ef2dd689	GL605	a6ab103e-ac0d-4bda-b989-6beb34419280	container_instance	container_instance	container_instance_lineage	c60228c8-099b-47f1-a890-387453eec26b	2025-07-03 08:24:15.850004-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:15.850004-07	f	\N
8528102b-a975-4278-bb42-db6feccbfc84	GL606	a6ab103e-ac0d-4bda-b989-6beb34419280	container_instance	container_instance	container_instance_lineage	7833d79c-9c5d-4177-b4ed-b29293438cc4	2025-07-03 08:24:15.857224-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:15.857224-07	f	\N
4ba073c0-9fda-4718-ac23-fd5e4bfbc5c4	GL607	a6ab103e-ac0d-4bda-b989-6beb34419280	container_instance	container_instance	container_instance_lineage	85859176-cfcb-4ba9-8078-981f2c92c5f7	2025-07-03 08:24:15.864295-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:15.864295-07	f	\N
85903bd7-3177-4463-885e-78e4db0c1929	GL608	a6ab103e-ac0d-4bda-b989-6beb34419280	container_instance	container_instance	container_instance_lineage	fca6c1f5-1e60-4b0e-b8dc-de2108eb584c	2025-07-03 08:24:15.872125-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:15.872125-07	f	\N
3adb2680-c576-4c53-a5b9-0ca84a315b80	GL609	a6ab103e-ac0d-4bda-b989-6beb34419280	container_instance	container_instance	container_instance_lineage	11d71ea0-89ac-4aad-89f4-68ba3f60525d	2025-07-03 08:24:15.884775-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:15.884775-07	f	\N
acf08066-ab9c-4a70-8d4c-9ecc63390409	GL610	a6ab103e-ac0d-4bda-b989-6beb34419280	container_instance	container_instance	container_instance_lineage	b5c6c063-303c-4bf1-a0b4-d952763bd8fc	2025-07-03 08:24:15.894459-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:15.894459-07	f	\N
a8535f8f-fa7e-473c-a2b4-30b0a1612da5	GL611	a6ab103e-ac0d-4bda-b989-6beb34419280	container_instance	container_instance	container_instance_lineage	d7616aa4-1790-453e-97cd-801233b1bc1e	2025-07-03 08:24:15.901925-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:15.901925-07	f	\N
6b1669e9-21a0-49ba-a9cc-0375fbb1994d	GL612	a6ab103e-ac0d-4bda-b989-6beb34419280	container_instance	container_instance	container_instance_lineage	b9068f68-df90-4cc8-be11-f651be061b2a	2025-07-03 08:24:15.911667-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:15.911667-07	f	\N
d7786da3-1950-47d6-87aa-96f83aa833db	GL613	a6ab103e-ac0d-4bda-b989-6beb34419280	container_instance	container_instance	container_instance_lineage	35280b86-54a8-4337-9df0-abcb0cd2c340	2025-07-03 08:24:15.920899-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:15.920899-07	f	\N
96afcc22-b8db-4421-9597-d4dd2a8892fe	GL614	a6ab103e-ac0d-4bda-b989-6beb34419280	container_instance	container_instance	container_instance_lineage	3f257485-e9ed-43ba-9245-63fe12620c6e	2025-07-03 08:24:15.929132-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:15.929132-07	f	\N
870719ee-b6b5-4dbb-b2f4-bd30e3e96995	GL615	a6ab103e-ac0d-4bda-b989-6beb34419280	container_instance	container_instance	container_instance_lineage	c9afb65c-3bce-41aa-94ca-ed1c356a938c	2025-07-03 08:24:15.938076-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:15.938076-07	f	\N
4342c3b1-5931-4411-a5ae-cca89b42c12b	GL616	a6ab103e-ac0d-4bda-b989-6beb34419280	container_instance	container_instance	container_instance_lineage	cca985cb-db90-47cb-880a-3749d3611984	2025-07-03 08:24:15.94856-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:15.94856-07	f	\N
e9e3c63e-e74b-4260-8cdd-e5f89b3027c8	GL617	843a9ccc-36ca-4d8b-80c8-a38487e0a12d	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	3ed95fad-05dc-4da6-a86e-f9cf24a0215a	2025-07-03 08:24:15.956035-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:15.956035-07	f	generic
82b804fa-0b12-4c36-a305-5f3e3cda81d8	GL651	f0bbdd4d-da05-445d-8c05-a8ed4769fe79	container_instance	container_instance	container_instance_lineage	c3897a38-49ee-4f23-9461-a55dff9a4413	2025-07-03 08:24:16.148035-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.148035-07	f	\N
8d3f12ab-72a4-4eed-9bb2-e66f2c4e6e02	GL652	f0bbdd4d-da05-445d-8c05-a8ed4769fe79	container_instance	container_instance	container_instance_lineage	2588b288-14ab-4eab-b9ff-41c51104f8bc	2025-07-03 08:24:16.153972-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.153972-07	f	\N
6002beae-1ea4-4071-a5c5-cadcbdec0af0	GL653	f0bbdd4d-da05-445d-8c05-a8ed4769fe79	container_instance	container_instance	container_instance_lineage	13fb23bc-4066-4e86-bde9-9a56cdc08e57	2025-07-03 08:24:16.160832-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.160832-07	f	\N
738644c7-f56a-4454-8db8-7f06be1a0213	GL654	f0bbdd4d-da05-445d-8c05-a8ed4769fe79	container_instance	container_instance	container_instance_lineage	4a40cc94-4c0d-4f56-a5be-7b17905d478f	2025-07-03 08:24:16.16817-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.16817-07	f	\N
a078c524-0cac-44f8-89ae-77f51e0f2ed6	GL655	f0bbdd4d-da05-445d-8c05-a8ed4769fe79	container_instance	container_instance	container_instance_lineage	a34c5ab3-1540-4e32-a08c-9f90c1f7894e	2025-07-03 08:24:16.173538-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.173538-07	f	\N
703807e9-20eb-4909-9b0f-529685666294	GL618	37a7fae4-216d-4842-ab8f-5c048d4a2809	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	6b65a1a8-ba49-4cb5-b381-72a0c2ab0f7d	2025-07-03 08:24:15.963002-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:15.963002-07	f	generic
fd55de73-a8a6-4e98-b1ad-bcf025eb76a6	GL619	109675e8-0498-4c1c-bad5-788a996e69a2	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	467ee04d-f443-49cf-b38d-090e047d1306	2025-07-03 08:24:15.968746-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:15.968746-07	f	generic
f92f9f48-c82d-4eba-b862-92b4d7c943fb	GL620	20febaa0-790b-4079-8124-3d2d9586d49e	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	3c31b867-fb64-4aaa-b939-f7e98b8abb92	2025-07-03 08:24:15.9742-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:15.9742-07	f	generic
265f4ea7-d290-49fb-a84c-2b862840fbb9	GL621	a0398a18-4ab8-469e-9010-55fda1de05b0	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	94ba58bf-d6b4-4163-bd4d-a69c9e105a68	2025-07-03 08:24:15.979872-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:15.979872-07	f	generic
f804dfcd-db6b-4d6a-b8c9-3d619404ed21	GL622	4ba3e341-5121-4322-bc64-7b192e496768	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	fb4fba5f-dbd7-42d9-a5a3-7fe0cfca759d	2025-07-03 08:24:15.985312-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:15.985312-07	f	generic
adbe5a56-c700-42a9-b860-03dddfa86bb2	GL623	c478dfca-9c72-4f15-93f1-fe3efd6274d1	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	c41f43b2-3170-48fd-97c3-304c73f09ad7	2025-07-03 08:24:15.993143-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:15.993143-07	f	generic
63ddbe12-b72a-4a5b-9f84-dfa561b0777c	GL624	d4fbc620-8b0d-4f35-b5f0-0b2fe2f00571	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	91cccd1e-a078-41ca-bbb3-67985eaaf192	2025-07-03 08:24:15.998517-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:15.998517-07	f	generic
731da033-4f83-42a5-9c13-160e3fd0234b	GL625	5c89cd1b-c0d0-4846-896a-ceca2f29fcb6	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	43079fcb-f024-46aa-9320-7c51a5bd6494	2025-07-03 08:24:16.005492-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.005492-07	f	generic
dd716bc7-3f79-459d-a219-cccd102d80eb	GL626	5fe3fa73-82ec-4a72-a25d-f4ca3f7575d8	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	c76c7d77-7a2c-4aaf-8a66-637b61c1f62e	2025-07-03 08:24:16.010893-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.010893-07	f	generic
07d31f16-2540-4795-89ab-83ba3d4316ec	GL627	015fbb50-97f2-409e-9b39-e5f548a21402	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	d20fb8e0-c8c4-4754-a5b3-4937f66a496c	2025-07-03 08:24:16.016379-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.016379-07	f	generic
5efb0bdd-c4b9-4a92-9652-f35ee49f40a4	GL628	427711ca-ab90-45b5-b311-8733db30d9ad	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	5fff5b28-eb3e-4bbd-b79f-754f8f03eea3	2025-07-03 08:24:16.021218-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.021218-07	f	generic
0ee22f53-49d0-4e74-ad14-ae16da4384df	GL629	95d4d6ba-7610-486b-a74c-d1ad8255621a	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	1ddb556f-e2da-4b49-b71a-3201257874a6	2025-07-03 08:24:16.025934-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.025934-07	f	generic
8bec5b3b-e1a8-4ce1-a3ee-3f2488d98676	GL630	61bfac74-3f5c-474c-8b7a-6fe48898e386	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	cbd6c137-3d08-43da-84e9-4f918b6946ce	2025-07-03 08:24:16.030921-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.030921-07	f	generic
ee2ef3b9-e776-4631-ba4c-9d95d9ab5e2e	GL631	a1f6ccdc-9ee5-402b-a93d-533b39212bee	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	ba9a32dd-9ff0-4bd5-bec2-2be8156a1c60	2025-07-03 08:24:16.03599-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.03599-07	f	generic
08c934fa-b964-4afb-9bbe-a201e7014d94	GL632	4ff5edda-1503-4a1b-aa81-44f0d001dede	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	653a1b4c-97ca-4469-88ee-08bdeeb09309	2025-07-03 08:24:16.040974-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.040974-07	f	generic
f6cd6d86-310f-4c84-b660-b47c5427993a	GL633	c60228c8-099b-47f1-a890-387453eec26b	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	045dfa06-cd02-4830-815a-64a0eda2b595	2025-07-03 08:24:16.045785-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.045785-07	f	generic
32ce890f-5851-4915-a0fd-d6c5ea3925a9	GL634	7833d79c-9c5d-4177-b4ed-b29293438cc4	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	e5c9c69f-4bf8-4583-83c3-ab32883c6fd8	2025-07-03 08:24:16.050413-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.050413-07	f	generic
e954ff06-ed30-415d-bd18-e3e6e46ae774	GL635	85859176-cfcb-4ba9-8078-981f2c92c5f7	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	aa3dceeb-693f-4b52-96fd-fcdacec74279	2025-07-03 08:24:16.057149-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.057149-07	f	generic
bb35d05f-6d49-49cd-b3b0-b14da7148cd7	GL636	fca6c1f5-1e60-4b0e-b8dc-de2108eb584c	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	1806a41a-5c44-4be2-9d25-69589865a6fe	2025-07-03 08:24:16.061637-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.061637-07	f	generic
5aed10fb-b550-42c8-a69f-48f87bbe0dda	GL637	11d71ea0-89ac-4aad-89f4-68ba3f60525d	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	81670de8-fbad-4c5b-8159-65d031907aa0	2025-07-03 08:24:16.066932-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.066932-07	f	generic
07d42fd6-e6fa-4636-be25-6ea6d0465c8f	GL638	b5c6c063-303c-4bf1-a0b4-d952763bd8fc	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	8ca2ddc1-bf48-4c18-ae6f-54d1d4bef1e4	2025-07-03 08:24:16.071479-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.071479-07	f	generic
f5cbe271-a522-4ca7-b600-a40d038b5efc	GL639	d7616aa4-1790-453e-97cd-801233b1bc1e	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	e1feb2f2-726f-4aa6-8334-7875ca8a9d76	2025-07-03 08:24:16.075965-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.075965-07	f	generic
66e1fbee-48f4-402e-9a7e-267de546c2c1	GL640	b9068f68-df90-4cc8-be11-f651be061b2a	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	4ef01449-ede9-4847-9074-67d56aa0e462	2025-07-03 08:24:16.080477-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.080477-07	f	generic
61a5e379-dc70-40d1-8540-a84cff8b5596	GL641	35280b86-54a8-4337-9df0-abcb0cd2c340	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	955793c4-ddbf-4bdf-a4e2-51f541d2d828	2025-07-03 08:24:16.08627-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.08627-07	f	generic
f042aa35-b703-49ce-a177-5dec670b35f8	GL642	3f257485-e9ed-43ba-9245-63fe12620c6e	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	275fe9ac-a3f8-4182-bd15-d5982dab919d	2025-07-03 08:24:16.090668-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.090668-07	f	generic
1bfb4d22-5674-48f0-a82c-7e3d344d6ecf	GL643	c9afb65c-3bce-41aa-94ca-ed1c356a938c	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	890ef728-0ffb-4f4d-90bd-29924f4bb340	2025-07-03 08:24:16.095989-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.095989-07	f	generic
81eb69f1-60cc-4b3d-ab64-480dc08cc57a	GL644	cca985cb-db90-47cb-880a-3749d3611984	container:well:fixed-plate-well:1.0	content:reagent:idt-probes-rare-mendelian:1.0	generic_instance_lineage	5b8e86dd-588a-4518-9713-3a48aef550a6	2025-07-03 08:24:16.10054-07	well:fixed-plate-well:1.0 :: reagent:idt-probes-rare-mendelian:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.10054-07	f	generic
43cd7bd4-c097-406a-be9b-ada4b9091cf3	GL645	f0bbdd4d-da05-445d-8c05-a8ed4769fe79	container_instance	container_instance	container_instance_lineage	7a7a05e0-d926-4739-8f3c-c9a41c715f1a	2025-07-03 08:24:16.115443-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.115443-07	f	\N
d67ff291-c680-4499-b6b9-b4de08922b84	GL646	f0bbdd4d-da05-445d-8c05-a8ed4769fe79	container_instance	container_instance	container_instance_lineage	78ea25c1-5b2c-4e77-8781-5fffd25ab2df	2025-07-03 08:24:16.121029-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.121029-07	f	\N
008a06f2-0160-48c3-a881-b7657e127b0c	GL647	f0bbdd4d-da05-445d-8c05-a8ed4769fe79	container_instance	container_instance	container_instance_lineage	6a69eda7-1e5a-4d4a-8755-94aefc3c74e8	2025-07-03 08:24:16.126215-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.126215-07	f	\N
1d5cc72e-a788-444c-be4a-850c50af8eb7	GL648	f0bbdd4d-da05-445d-8c05-a8ed4769fe79	container_instance	container_instance	container_instance_lineage	217b0af9-7c61-4971-8098-d67ce949fc38	2025-07-03 08:24:16.131433-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.131433-07	f	\N
6de10070-8c76-49bd-8956-1a0634fc1cca	GL649	f0bbdd4d-da05-445d-8c05-a8ed4769fe79	container_instance	container_instance	container_instance_lineage	4484b524-4a4b-408d-8153-f19da5a148c1	2025-07-03 08:24:16.137239-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.137239-07	f	\N
4052a3fd-3659-44d6-9b95-8b211f5300ef	GL650	f0bbdd4d-da05-445d-8c05-a8ed4769fe79	container_instance	container_instance	container_instance_lineage	220d86a6-088f-47a7-a11b-f64519da6a46	2025-07-03 08:24:16.142482-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.142482-07	f	\N
72a6dd12-e29b-4456-893d-207104cbad11	GL656	f0bbdd4d-da05-445d-8c05-a8ed4769fe79	container_instance	container_instance	container_instance_lineage	0007e7e3-47ad-48fa-8228-7f0dbe1f4a60	2025-07-03 08:24:16.179232-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.179232-07	f	\N
7ff212fc-f2a2-4b75-b36e-5dc8ee450acb	GL657	f0bbdd4d-da05-445d-8c05-a8ed4769fe79	container_instance	container_instance	container_instance_lineage	e6a96820-ce9d-4ec4-8f59-fd93f0e11b76	2025-07-03 08:24:16.184262-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.184262-07	f	\N
989b1900-2d75-4096-a4c6-2ccf62e9d9e2	GL658	f0bbdd4d-da05-445d-8c05-a8ed4769fe79	container_instance	container_instance	container_instance_lineage	5eab8dce-d0dd-449f-966d-47154aef3cf0	2025-07-03 08:24:16.189469-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.189469-07	f	\N
86dd4863-f3ee-40ab-b399-c8e65d4d65a0	GL659	f0bbdd4d-da05-445d-8c05-a8ed4769fe79	container_instance	container_instance	container_instance_lineage	4e71540e-b707-4978-a4da-25a8ac6dbb48	2025-07-03 08:24:16.194592-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.194592-07	f	\N
8a7af7d7-ec12-46f0-8775-aa037d309fd8	GL660	f0bbdd4d-da05-445d-8c05-a8ed4769fe79	container_instance	container_instance	container_instance_lineage	7f0ff5ac-55fd-40b8-993c-37fd45de7e70	2025-07-03 08:24:16.201071-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.201071-07	f	\N
2c506eda-c6e3-419a-acbe-eb00c2014cc4	GL661	f0bbdd4d-da05-445d-8c05-a8ed4769fe79	container_instance	container_instance	container_instance_lineage	5914dd40-bbb0-451f-aabf-17c7970df831	2025-07-03 08:24:16.20631-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.20631-07	f	\N
938c645b-9d7b-40ec-a03f-632a15414104	GL662	f0bbdd4d-da05-445d-8c05-a8ed4769fe79	container_instance	container_instance	container_instance_lineage	d53b5c53-dff3-4c58-be12-ef4497a67766	2025-07-03 08:24:16.212024-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.212024-07	f	\N
7f015c91-b548-4b5d-b668-eeee4ea78342	GL663	f0bbdd4d-da05-445d-8c05-a8ed4769fe79	container_instance	container_instance	container_instance_lineage	561fbd02-0e42-4c28-b549-540297a81c6a	2025-07-03 08:24:16.217462-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.217462-07	f	\N
cb7f1fb2-ba5e-4eb8-8ada-c266202ad206	GL664	f0bbdd4d-da05-445d-8c05-a8ed4769fe79	container_instance	container_instance	container_instance_lineage	ddc6efd5-96cb-476a-a565-bb7248a7b9a2	2025-07-03 08:24:16.222941-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.222941-07	f	\N
68b9b27e-f99d-4291-8e6a-754536172b0a	GL665	f0bbdd4d-da05-445d-8c05-a8ed4769fe79	container_instance	container_instance	container_instance_lineage	f5a92643-727e-42cd-a337-e2fd476dc69d	2025-07-03 08:24:16.230475-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.230475-07	f	\N
a65743ed-2090-4b59-b0ce-f99809235d67	GL666	f0bbdd4d-da05-445d-8c05-a8ed4769fe79	container_instance	container_instance	container_instance_lineage	2754a4ac-782d-4991-9e07-ed7cad5dfc2a	2025-07-03 08:24:16.235951-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.235951-07	f	\N
ae2afa26-cc66-4df9-9354-d25769ef3181	GL667	f0bbdd4d-da05-445d-8c05-a8ed4769fe79	container_instance	container_instance	container_instance_lineage	74137fae-e9d7-4e0b-b88c-f6e1ef1aeb96	2025-07-03 08:24:16.241895-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.241895-07	f	\N
ce18cd66-32c6-4ce9-a103-dc52d3425d65	GL668	f0bbdd4d-da05-445d-8c05-a8ed4769fe79	container_instance	container_instance	container_instance_lineage	7b8a94d7-37af-4f1b-a3fc-1cd5399cb3f9	2025-07-03 08:24:16.247453-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.247453-07	f	\N
2f763498-ce03-48fc-bb96-306795f1bf84	GL669	f0bbdd4d-da05-445d-8c05-a8ed4769fe79	container_instance	container_instance	container_instance_lineage	165a43a4-b366-4fc2-a6c1-97f8c9b236dc	2025-07-03 08:24:16.257121-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.257121-07	f	\N
52f6dde3-ae7c-4031-8235-52e3b1aea117	GL670	f0bbdd4d-da05-445d-8c05-a8ed4769fe79	container_instance	container_instance	container_instance_lineage	b2fc4da9-2a05-49ca-8d5d-b9558cc97c68	2025-07-03 08:24:16.262563-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.262563-07	f	\N
fc77ce96-036c-4f03-a473-a11c8067b14e	GL671	f0bbdd4d-da05-445d-8c05-a8ed4769fe79	container_instance	container_instance	container_instance_lineage	774fcd11-a6a1-4bdc-8b98-7fbc239ad4d7	2025-07-03 08:24:16.268554-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.268554-07	f	\N
4afbaba8-6480-45fb-aeed-9c06de5409f2	GL672	f0bbdd4d-da05-445d-8c05-a8ed4769fe79	container_instance	container_instance	container_instance_lineage	91999ed4-21f0-408d-8f7c-a00ad024b9a8	2025-07-03 08:24:16.274066-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.274066-07	f	\N
87c3ad43-cdba-4c8f-a3e6-0c5dcffa7812	GL673	7a7a05e0-d926-4739-8f3c-c9a41c715f1a	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	e3b20962-4aaa-4e81-850f-90b4ff6bb72b	2025-07-03 08:24:16.278242-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.278242-07	f	generic
efc3906c-c305-45f3-a4eb-ff93349191af	GL719	7a653734-9941-44ab-9446-6ef7acee9702	container_instance	container_instance	container_instance_lineage	28a66e77-5293-4587-8aaa-2a11cd102349	2025-07-03 08:24:16.558048-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.558048-07	f	\N
bd6f0aed-b3a3-46be-8467-caf81efef5ef	GL720	7a653734-9941-44ab-9446-6ef7acee9702	container_instance	container_instance	container_instance_lineage	344aff1a-652b-4802-8b45-87e345a5dfaf	2025-07-03 08:24:16.562895-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.562895-07	f	\N
55c65ed9-ca79-42ea-a36b-77ee8ca2f285	GL721	7a653734-9941-44ab-9446-6ef7acee9702	container_instance	container_instance	container_instance_lineage	e41faaaf-ce2a-43d6-87e5-e3870abe45d6	2025-07-03 08:24:16.569326-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.569326-07	f	\N
b2659560-f732-4da9-ace5-539448f993f5	GL674	78ea25c1-5b2c-4e77-8781-5fffd25ab2df	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	74ad366b-d25f-4da4-8084-53ea5fa20220	2025-07-03 08:24:16.283122-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.283122-07	f	generic
74e2cf41-d124-44c0-998c-d4670f1df903	GL675	6a69eda7-1e5a-4d4a-8755-94aefc3c74e8	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	0c89558c-3040-41f3-b14f-54281d89a861	2025-07-03 08:24:16.289351-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.289351-07	f	generic
3f1e9150-2a19-442f-bcd7-14f218946eee	GL676	217b0af9-7c61-4971-8098-d67ce949fc38	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	f576931a-14bf-4549-bdad-cfbb1c5ad826	2025-07-03 08:24:16.29568-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.29568-07	f	generic
c8a926dc-bfbb-489a-8992-40c4133997ad	GL677	4484b524-4a4b-408d-8153-f19da5a148c1	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	f5c9a461-3037-48a9-ab53-74984ace3aec	2025-07-03 08:24:16.301211-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.301211-07	f	generic
aeacbbe7-6ba7-4e70-939a-f86a95f641be	GL678	220d86a6-088f-47a7-a11b-f64519da6a46	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	103f84a1-a056-414a-b9a1-07ead9a853e6	2025-07-03 08:24:16.307149-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.307149-07	f	generic
5e087394-2c8b-4e81-ad07-da72b87b5915	GL679	c3897a38-49ee-4f23-9461-a55dff9a4413	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	0c26d313-4b07-4a0f-a826-ea2b0242df8b	2025-07-03 08:24:16.311778-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.311778-07	f	generic
3a64133b-eb89-4bde-8e3a-9fc9a23f4b05	GL680	2588b288-14ab-4eab-b9ff-41c51104f8bc	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	5ef0757b-7e03-409b-87ea-70557be4dac0	2025-07-03 08:24:16.316737-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.316737-07	f	generic
775e4d55-c66f-423d-a000-8cd9eed7099b	GL681	13fb23bc-4066-4e86-bde9-9a56cdc08e57	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	a5bb73df-6539-4173-9687-1763600076bc	2025-07-03 08:24:16.321763-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.321763-07	f	generic
d4157848-b040-435a-b13a-92c5006c7d2b	GL682	4a40cc94-4c0d-4f56-a5be-7b17905d478f	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	4830667e-3fe8-4a03-beeb-80786dc41757	2025-07-03 08:24:16.326841-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.326841-07	f	generic
ada1fbec-c3af-44bf-ae9f-3f6e6eecafa1	GL683	a34c5ab3-1540-4e32-a08c-9f90c1f7894e	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	7512b9d2-980e-4174-952b-78facb42ec6d	2025-07-03 08:24:16.331496-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.331496-07	f	generic
460b6197-2e73-4e6b-bba4-e6d8c8231d5c	GL684	0007e7e3-47ad-48fa-8228-7f0dbe1f4a60	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	b0b03902-4397-4f5c-99cf-9bc996f64e45	2025-07-03 08:24:16.336673-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.336673-07	f	generic
ddb35748-5461-4510-9441-6dd66ce69f9c	GL685	e6a96820-ce9d-4ec4-8f59-fd93f0e11b76	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	90a7b8ca-0912-4f4f-9e1b-35ee43101fe8	2025-07-03 08:24:16.342636-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.342636-07	f	generic
a4d06364-7a57-41bd-ae5f-cbe0989e7ac6	GL686	5eab8dce-d0dd-449f-966d-47154aef3cf0	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	ffae08fc-ee93-4871-9d7a-5c45332c4d86	2025-07-03 08:24:16.34712-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.34712-07	f	generic
f25eb45f-f21a-471d-ab99-802a4c35dc7d	GL687	4e71540e-b707-4978-a4da-25a8ac6dbb48	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	7823f81f-b135-4cb1-9937-96ad578a673d	2025-07-03 08:24:16.35246-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.35246-07	f	generic
d2b3c985-dbb8-46f7-adb8-22b0ee6bc0b7	GL688	7f0ff5ac-55fd-40b8-993c-37fd45de7e70	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	4a67fb94-8ae8-4ca3-bc46-81ed67993cf1	2025-07-03 08:24:16.357321-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.357321-07	f	generic
de96d141-5742-4e82-a795-eda650bf3640	GL689	5914dd40-bbb0-451f-aabf-17c7970df831	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	c5f72729-d226-4ab5-b263-52ecffcea927	2025-07-03 08:24:16.361823-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.361823-07	f	generic
9c5ff9e1-f03d-49e1-b457-255c47315246	GL690	d53b5c53-dff3-4c58-be12-ef4497a67766	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	debc0fde-6d6a-4e56-899b-cd47f75025bb	2025-07-03 08:24:16.369633-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.369633-07	f	generic
ff8cbce0-c5a2-48aa-9185-f755a67ef106	GL691	561fbd02-0e42-4c28-b549-540297a81c6a	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	ebaff42d-7c7c-4d36-b7ce-02db9be33504	2025-07-03 08:24:16.376565-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.376565-07	f	generic
db514589-ad0b-43a6-9db4-9235d08dcebf	GL692	ddc6efd5-96cb-476a-a565-bb7248a7b9a2	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	19af9824-019d-472e-a8b7-e5b098186cd7	2025-07-03 08:24:16.381663-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.381663-07	f	generic
6c29cd28-4078-4524-b84c-3715029e5e39	GL693	f5a92643-727e-42cd-a337-e2fd476dc69d	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	c2af1c10-826b-403b-945b-5068a0fe4998	2025-07-03 08:24:16.38591-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.38591-07	f	generic
87a949a8-d919-4371-8130-186f927a470c	GL694	2754a4ac-782d-4991-9e07-ed7cad5dfc2a	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	b9299cec-5b81-48be-b4e8-bcce06bf4d26	2025-07-03 08:24:16.391185-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.391185-07	f	generic
2e61ac89-4ec1-454a-9f36-d8f1e7fa6e6d	GL695	74137fae-e9d7-4e0b-b88c-f6e1ef1aeb96	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	ac122017-f1e1-449f-bdf3-39016e27fad8	2025-07-03 08:24:16.395568-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.395568-07	f	generic
7cc98bd0-c920-481f-9d19-2bb5e1afde70	GL696	7b8a94d7-37af-4f1b-a3fc-1cd5399cb3f9	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	e23c847f-8774-4450-9dc9-ec4f7233a6d8	2025-07-03 08:24:16.400057-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.400057-07	f	generic
7981eae4-84ba-4695-84c8-31e3abd962b3	GL697	165a43a4-b366-4fc2-a6c1-97f8c9b236dc	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	f5228cdc-84b8-4f13-8171-543a997417ec	2025-07-03 08:24:16.404841-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.404841-07	f	generic
6c427ffb-e96a-464d-a307-d5bb52318f11	GL698	b2fc4da9-2a05-49ca-8d5d-b9558cc97c68	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	7d0c6c07-ac47-47df-b229-c005d7d1685d	2025-07-03 08:24:16.410484-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.410484-07	f	generic
662e0651-f894-4aad-a4ad-59700fc4ec95	GL699	774fcd11-a6a1-4bdc-8b98-7fbc239ad4d7	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	c5a56bf4-3463-40b3-8b68-5e47b32a5cff	2025-07-03 08:24:16.415536-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.415536-07	f	generic
417511eb-04d3-4027-ba31-6ba3da9e772d	GL700	91999ed4-21f0-408d-8f7c-a00ad024b9a8	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	1b97a184-a3da-4d82-a043-a02a95207c0e	2025-07-03 08:24:16.421056-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.421056-07	f	generic
ed5b3cd1-d211-4309-ac26-90ed4d8816b6	GL701	b9d2526e-ab8a-4ff8-a1b0-947071f5d2d7	container:tube:tube-generic-10ml:1.0	content:control:giab-HG002:1.0	generic_instance_lineage	845ebde8-23d0-447c-a05e-538a37a81f97	2025-07-03 08:24:16.436412-07	tube:tube-generic-10ml:1.0 :: control:giab-HG002:1.0	generic	tube	tube-generic-10ml	{"cogs": {"cost": "1.50", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "giab NA24385", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "Generic 10ml tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.436412-07	f	generic
8cff34c6-56b9-4877-b32e-62a59c26898c	GL722	7a653734-9941-44ab-9446-6ef7acee9702	container_instance	container_instance	container_instance_lineage	006ca986-054c-4e9f-8e46-dac3dc562a98	2025-07-03 08:24:16.574559-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.574559-07	f	\N
ad8a63b8-3a0f-4014-8136-dfae38a1650c	GL702	0bbf1234-7c4b-4afc-9352-56a600deb076	container:tube:tube-10ml-glass:1.0	content:control:synthetic-cfdna:1.0	generic_instance_lineage	4f0b197f-9f39-43fd-90bd-3ee2cf2ca0a6	2025-07-03 08:24:16.450581-07	tube:tube-10ml-glass:1.0 :: control:synthetic-cfdna:1.0	generic	tube	tube-10ml-glass	{"cogs": {"cost": "4.70", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "s ynthetic-cfdna", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "10ml glass tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.450581-07	f	generic
45a26d48-60fb-4705-acc9-432c8b0fc38f	GL703	f69f3bb2-4931-451c-ba90-a2f283ef8618	container:tube:tube-eppi-1.5ml:1.0	content:control:water-ntc:1.0	generic_instance_lineage	da2ba524-f58e-4e2f-bc8b-39610f57f804	2025-07-03 08:24:16.459479-07	tube:tube-eppi-1.5ml:1.0 :: control:water-ntc:1.0	generic	tube	tube-eppi-1.5ml	{"cogs": {"cost": "1.70", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "NTC - water", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "1.5ml eppi tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.459479-07	f	generic
831e7ece-a335-46e1-bda0-65c973485d87	GL704	7a653734-9941-44ab-9446-6ef7acee9702	container_instance	container_instance	container_instance_lineage	70493c68-0506-4017-a8f9-5d14bdef4cf7	2025-07-03 08:24:16.472381-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.472381-07	f	\N
37ee6bc8-6ce6-4894-beee-511f523e017b	GL705	7a653734-9941-44ab-9446-6ef7acee9702	container_instance	container_instance	container_instance_lineage	ee0ded45-6595-40e7-8c32-07e783b35d22	2025-07-03 08:24:16.480503-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.480503-07	f	\N
3729b464-b20a-4709-821a-cbe13b701d3c	GL706	7a653734-9941-44ab-9446-6ef7acee9702	container_instance	container_instance	container_instance_lineage	367a6a1a-9157-4b01-b2ec-c2dc81af3154	2025-07-03 08:24:16.487499-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.487499-07	f	\N
00e60851-7b73-4002-b00f-417ab5c984c6	GL707	7a653734-9941-44ab-9446-6ef7acee9702	container_instance	container_instance	container_instance_lineage	ca914991-5199-43a4-a369-2075a264f28d	2025-07-03 08:24:16.492824-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.492824-07	f	\N
3a97dd5a-03c7-4b3b-8c66-8aa8437289a9	GL708	7a653734-9941-44ab-9446-6ef7acee9702	container_instance	container_instance	container_instance_lineage	9ab7fce2-54c8-491d-a3ec-17f3bc41ceba	2025-07-03 08:24:16.497839-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.497839-07	f	\N
906ce14e-6d9e-4ae1-aabc-37d5f4a0d903	GL709	7a653734-9941-44ab-9446-6ef7acee9702	container_instance	container_instance	container_instance_lineage	851ce518-5a99-4a87-ba96-8c4e2b8fde5c	2025-07-03 08:24:16.502909-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.502909-07	f	\N
d04e6d0d-7093-4e46-af1d-da2141e98e60	GL710	7a653734-9941-44ab-9446-6ef7acee9702	container_instance	container_instance	container_instance_lineage	b22b7269-47b8-44e6-a218-8c486c50cc0e	2025-07-03 08:24:16.507977-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.507977-07	f	\N
5474c331-bd88-414c-bb91-a3f7620b87f1	GL711	7a653734-9941-44ab-9446-6ef7acee9702	container_instance	container_instance	container_instance_lineage	6ccd1298-f01c-4572-a51a-1defbf10da94	2025-07-03 08:24:16.513313-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.513313-07	f	\N
97981b53-a786-4f91-b525-94526300fea8	GL712	7a653734-9941-44ab-9446-6ef7acee9702	container_instance	container_instance	container_instance_lineage	0cfbb220-c556-449b-9a17-8f8b74170237	2025-07-03 08:24:16.518506-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.518506-07	f	\N
d8cad5ee-1a75-4ad0-a6a5-9235af5f7253	GL713	7a653734-9941-44ab-9446-6ef7acee9702	container_instance	container_instance	container_instance_lineage	a20882a3-20fb-4e2a-b68c-ed393bb816d3	2025-07-03 08:24:16.523496-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.523496-07	f	\N
005fec53-7606-4f3c-8446-b16d33d11d2a	GL714	7a653734-9941-44ab-9446-6ef7acee9702	container_instance	container_instance	container_instance_lineage	9bfaafce-9d17-4839-a5ea-03bbdfcc4b2e	2025-07-03 08:24:16.528932-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.528932-07	f	\N
302111cc-b4e3-4e62-a328-812238628850	GL715	7a653734-9941-44ab-9446-6ef7acee9702	container_instance	container_instance	container_instance_lineage	2401c6c6-d364-44ec-94ae-fbe7825e8654	2025-07-03 08:24:16.53635-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.53635-07	f	\N
ca04384d-bcd7-446d-99ca-c967f8c1f915	GL716	7a653734-9941-44ab-9446-6ef7acee9702	container_instance	container_instance	container_instance_lineage	2259a89d-1a1e-4541-a8f0-1316da02a81d	2025-07-03 08:24:16.542302-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.542302-07	f	\N
df5b0784-62cf-44ba-9ce0-885ce263408a	GL717	7a653734-9941-44ab-9446-6ef7acee9702	container_instance	container_instance	container_instance_lineage	5103010b-3a86-45d4-b1cd-0a9e77622b35	2025-07-03 08:24:16.547555-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.547555-07	f	\N
ae50d1bd-f281-4928-a723-b9889c8465e8	GL718	7a653734-9941-44ab-9446-6ef7acee9702	container_instance	container_instance	container_instance_lineage	b1be5bf3-c765-4f3c-a272-e128495d6a30	2025-07-03 08:24:16.55288-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.55288-07	f	\N
f79e64eb-f5b1-4911-95ab-d30eeed29d5c	GL723	7a653734-9941-44ab-9446-6ef7acee9702	container_instance	container_instance	container_instance_lineage	233d4349-e767-46bf-a7ed-035e092867e8	2025-07-03 08:24:16.581962-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.581962-07	f	\N
12e3d02c-641c-41c3-90a9-f74a92a0a132	GL724	7a653734-9941-44ab-9446-6ef7acee9702	container_instance	container_instance	container_instance_lineage	f20b5ca4-d90b-4558-8f3a-088e0959cb77	2025-07-03 08:24:16.590183-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.590183-07	f	\N
a4faa36f-4c23-4978-8c13-7289003c95b4	GL725	7a653734-9941-44ab-9446-6ef7acee9702	container_instance	container_instance	container_instance_lineage	35620e62-853f-41df-b226-7036aae429e1	2025-07-03 08:24:16.595467-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.595467-07	f	\N
ed81f171-7bbd-4ceb-a455-c7e4afab5e8c	GL726	7a653734-9941-44ab-9446-6ef7acee9702	container_instance	container_instance	container_instance_lineage	156a6879-7978-4ba2-9f2d-be6064697797	2025-07-03 08:24:16.601469-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.601469-07	f	\N
c702cef1-171b-459e-a744-0950c11c0716	GL727	7a653734-9941-44ab-9446-6ef7acee9702	container_instance	container_instance	container_instance_lineage	121e9047-ad3e-4dfd-823b-83e1fc717cbc	2025-07-03 08:24:16.608243-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.608243-07	f	\N
6887fb1a-1e4b-4eab-aa5a-2ef67ad1380c	GL728	7a653734-9941-44ab-9446-6ef7acee9702	container_instance	container_instance	container_instance_lineage	fda6b968-9927-4cf6-85f9-c44bb39014a7	2025-07-03 08:24:16.614298-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.614298-07	f	\N
4e0c9980-2ea4-4add-90ef-ec8da33fb859	GL729	7a653734-9941-44ab-9446-6ef7acee9702	container_instance	container_instance	container_instance_lineage	c7124a90-7b46-478e-b74a-dacd7a07dc62	2025-07-03 08:24:16.619981-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.619981-07	f	\N
57705daa-7d7d-466d-a46d-4c773f2ae7f4	GL730	7a653734-9941-44ab-9446-6ef7acee9702	container_instance	container_instance	container_instance_lineage	cb5b12ec-e9e7-4707-bffc-0e453d1fc960	2025-07-03 08:24:16.62529-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.62529-07	f	\N
1ef8a199-a281-46ef-80b9-030ab303de45	GL731	7a653734-9941-44ab-9446-6ef7acee9702	container_instance	container_instance	container_instance_lineage	ffefe78a-8ac0-4e3e-9fa9-772ee8762f39	2025-07-03 08:24:16.630723-07	plate:fixed-plate-24:1.0 :: well:fixed-plate-well:1.0	container	plate	fixed-plate-24	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {"core": {"actions": {"action/core/add-relationships/1.0": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}, "action/core/print_barcode_label/1.0": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/download_plate_map/1.0": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}, "action/plate/stamp_plates_into_plate/1.0": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}, "group_name": "Plate Actions", "group_order": "2"}}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	ready	1.0	f	2025-07-03 08:24:16.630723-07	f	\N
e57f97ae-5c0b-4b4d-9ea5-587adbdedea1	GL732	70493c68-0506-4017-a8f9-5d14bdef4cf7	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	3ebb20a3-6d70-4ebb-9a8d-f070a39a60f0	2025-07-03 08:24:16.637466-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.637466-07	f	generic
1ba7c0f0-996a-42c9-acc0-88db181a375f	GL733	ee0ded45-6595-40e7-8c32-07e783b35d22	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	5d6d4e76-14b5-4009-9f05-a8e3062bb797	2025-07-03 08:24:16.643287-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.643287-07	f	generic
478bd241-2058-4b89-a722-537011794ff8	GL734	367a6a1a-9157-4b01-b2ec-c2dc81af3154	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	a9013d2d-2a96-4508-b74a-6142dc7eaacd	2025-07-03 08:24:16.649412-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.649412-07	f	generic
9faad57c-286a-477c-bfef-71bed6709f1c	GL735	ca914991-5199-43a4-a369-2075a264f28d	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	2355589e-b1e9-469d-9593-602b77227c24	2025-07-03 08:24:16.655929-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "A4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.655929-07	f	generic
7916e637-e90f-4e7e-858e-e2c038f113e3	GL736	9ab7fce2-54c8-491d-a3ec-17f3bc41ceba	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	ca939326-d2b1-4b88-80a3-15cb849ba2af	2025-07-03 08:24:16.662326-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.662326-07	f	generic
2b1f02f9-59c7-4526-a897-b4ba3cf6d12f	GL737	851ce518-5a99-4a87-ba96-8c4e2b8fde5c	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	88694a6f-07f1-4ef8-a9b8-97467620e235	2025-07-03 08:24:16.668647-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.668647-07	f	generic
c0364bde-d980-4756-b7da-a924defc131a	GL738	b22b7269-47b8-44e6-a218-8c486c50cc0e	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	86c0d971-5f10-4cb5-ae88-c21624503ba6	2025-07-03 08:24:16.67413-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.67413-07	f	generic
157ddf80-29b8-445d-93d1-b2088c505dc8	GL739	6ccd1298-f01c-4572-a51a-1defbf10da94	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	9f1fb3f4-0611-4f6e-a39a-71b3ae2dba25	2025-07-03 08:24:16.679546-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "B4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.679546-07	f	generic
b931f168-c2da-4392-a535-653e5daa6ffa	GL740	0cfbb220-c556-449b-9a17-8f8b74170237	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	e47de7d0-f3d5-4fd5-ae77-6f3d053caf84	2025-07-03 08:24:16.684741-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.684741-07	f	generic
1d967a38-f06d-47f0-a544-95629e27e900	GL741	a20882a3-20fb-4e2a-b68c-ed393bb816d3	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	e72c6cc1-db0a-41bd-b271-a7d5b645df0c	2025-07-03 08:24:16.690129-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.690129-07	f	generic
b993f737-d611-4508-a4a0-fe83b78dc322	GL742	9bfaafce-9d17-4839-a5ea-03bbdfcc4b2e	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	49a12695-1781-4daa-80c6-fbb71f9ba0ba	2025-07-03 08:24:16.697069-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.697069-07	f	generic
80c92374-35f4-4ed1-82ad-ed50eda7a153	GL743	2401c6c6-d364-44ec-94ae-fbe7825e8654	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	5117b15a-474f-43b7-b9eb-e367af711f2a	2025-07-03 08:24:16.705224-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "C4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.705224-07	f	generic
a2eda545-dd1a-49f3-aa99-a19375707042	GL744	2259a89d-1a1e-4541-a8f0-1316da02a81d	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	cb3a3bb1-aefc-4355-aae0-fcf5a8a94013	2025-07-03 08:24:16.710408-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.710408-07	f	generic
74a5f05a-5eb8-42c3-aaa2-7bcc9fb6aafd	GL745	5103010b-3a86-45d4-b1cd-0a9e77622b35	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	51e3ccc5-9470-49f4-987d-b96a897a3fd2	2025-07-03 08:24:16.715575-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.715575-07	f	generic
87175ff8-4efe-4080-b847-74d4fbbe650d	GL746	b1be5bf3-c765-4f3c-a272-e128495d6a30	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	dea04455-e99a-4424-ac3b-fc7124de43f8	2025-07-03 08:24:16.720517-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.720517-07	f	generic
ac1ae2ef-839c-4f9c-a3e1-6ee4a006106b	GL747	28a66e77-5293-4587-8aaa-2a11cd102349	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	0559ee06-9d4f-4f58-a12e-1f32bdff4633	2025-07-03 08:24:16.725843-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "D4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.725843-07	f	generic
18f6a81b-71be-42e0-8cb5-7050c49b376c	GL748	344aff1a-652b-4802-8b45-87e345a5dfaf	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	d20de8f6-379c-429c-9a0b-916b7833b1a4	2025-07-03 08:24:16.730975-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.730975-07	f	generic
63dc7384-6002-45ac-9818-63728e7cce47	GL749	e41faaaf-ce2a-43d6-87e5-e3870abe45d6	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	bb287c24-62a2-4531-8ca1-909803f845fc	2025-07-03 08:24:16.736556-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.736556-07	f	generic
61b19c14-f268-41dc-ba4f-d56d1793e5c7	GL750	006ca986-054c-4e9f-8e46-dac3dc562a98	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	eb4ac051-5a75-4ba3-9725-081583664880	2025-07-03 08:24:16.741873-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.741873-07	f	generic
68c70c14-45bf-4eba-8275-6b146d767096	GL751	233d4349-e767-46bf-a7ed-035e092867e8	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	89e14590-5f15-40df-8f99-d3784fb556be	2025-07-03 08:24:16.746901-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "E4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.746901-07	f	generic
e64808e3-e2bd-4528-9fe1-c2f41c80b1d6	GL752	f20b5ca4-d90b-4558-8f3a-088e0959cb77	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	7fc49341-3ee8-463f-b10a-5919cd6f2f0a	2025-07-03 08:24:16.752108-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.752108-07	f	generic
81ebce8b-fd0c-4e36-880a-544ddd442ab0	GL753	35620e62-853f-41df-b226-7036aae429e1	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	0262fa19-4669-4ca2-94d9-6fd44dfebbcf	2025-07-03 08:24:16.757112-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.757112-07	f	generic
446912ce-94a9-491f-9925-00e5e85aff7b	GL754	156a6879-7978-4ba2-9f2d-be6064697797	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	1e661782-0d7e-478b-834a-cf1b8e5619ad	2025-07-03 08:24:16.762394-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.762394-07	f	generic
0b6fa778-a9ff-4848-a072-6f37f098aa01	GL755	121e9047-ad3e-4dfd-823b-83e1fc717cbc	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	1d281d59-1305-4a9a-b542-7a874fea7e37	2025-07-03 08:24:16.767513-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "F4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.767513-07	f	generic
03021265-7b72-4ab7-b3f3-06c3e8bb2432	GL756	fda6b968-9927-4cf6-85f9-c44bb39014a7	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	97b8ade0-e5da-4ab0-97a5-a03f9ce9b362	2025-07-03 08:24:16.772618-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G1", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.772618-07	f	generic
852c51fb-14e4-4bef-a935-2919b6048be1	GL757	c7124a90-7b46-478e-b74a-dacd7a07dc62	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	fa1ac264-1427-48af-9781-70e9f462b347	2025-07-03 08:24:16.777579-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G2", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.777579-07	f	generic
d649b760-159d-41fb-9970-82ab7455d1c7	GL758	cb5b12ec-e9e7-4707-bffc-0e453d1fc960	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	ed96bfa1-f338-408f-86b5-48328def257e	2025-07-03 08:24:16.782698-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G3", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.782698-07	f	generic
ef5a37e6-2b00-4820-a71c-03d6039f4358	GL759	ffefe78a-8ac0-4e3e-9fa9-772ee8762f39	container:well:fixed-plate-well:1.0	content:reagent:naoh:1.0	generic_instance_lineage	9cd2fa8f-9130-42fa-b43d-5e9110ceaec6	2025-07-03 08:24:16.787731-07	well:fixed-plate-well:1.0 :: reagent:naoh:1.0	generic	well	fixed-plate-well	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "G4", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}, "action_groups": {"core": {"actions": {"action/core/set_object_status/1.0": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}, "group_name": "Core Actions", "group_order": "1"}}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	ready	1.0	f	2025-07-03 08:24:16.787731-07	f	generic
\.


--
-- Data for Name: generic_template; Type: TABLE DATA; Schema: public; Owner: daylily
--

COPY public.generic_template (uuid, euid, name, instance_prefix, polymorphic_discriminator, super_type, btype, b_sub_type, created_dt, version, json_addl, json_addl_schema, bstatus, is_deleted, modified_dt, is_singleton) FROM stdin;
4a35bc93-97d4-43e5-b228-085f24c16807	GT1	accessioning-ay:create_package_and_first_workflow_step_assay_root:1.0	XX	action_template	action	accessioning-ay	create_package_and_first_workflow_step_assay_root	2025-07-03 05:56:45.597163-07	1.0	{"properties": {"name": "", "comments": ""}, "action_template": {"curr_user": "", "action_name": "Register Package", "action_user": [], "description": "Create Package and First Workflow Step", "method_name": "do_action_create_package_and_first_workflow_step_assay", "action_order": "0", "capture_data": "yes", "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {"container/package/generic/*/": {"json_addl": {"properties": {"comments": ""}}}}, "action_simple_value": "", "child_workflow_object": {"workflow/accessioning/accession-package-kit-tubes-testreq/1.0": {"json_addl": {"properties": {"step_number": "1"}}}}, "child_workflow_step_obj": {"workflow_step/accessioning-steps/package-generated/1.0": {"json_addl": {"properties": {"step_number": "1"}}}}, "deactivate_actions_when_executed": [], "workflow_step_to_attach_as_child": {"workflow_step/queue/all-purpose/1.0/": {"json_addl": {"properties": {"name": "", "comments": "", "step_number": "0"}}}}}}	\N	ready	f	2025-07-03 05:56:45.597163-07	t
f231825e-55c9-4b14-9ea1-98df6b41172b	GT2	assay-queue-accessioning:create_package_and_first_workflow_step:1.0	XX	action_template	action	assay-queue-accessioning	create_package_and_first_workflow_step	2025-07-03 05:56:46.067501-07	1.0	{"properties": {"name": "", "comments": ""}, "action_template": {"curr_user": "", "action_name": "Register Packagexx", "action_user": [], "description": "Create Package and First Workflow Stepxx", "method_name": "do_action_create_package_and_first_workflow_stepxx", "action_order": "0", "capture_data": "yes", "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {"container/package/generic/*/": {"json_addl": {"properties": {"comments": ""}}}}, "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/accessioning-steps/package-generated/1.0": {"json_addl": {"properties": {"step_number": "1"}}}}, "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:46.067501-07	t
07d797c4-599d-4b25-af0b-36ef32b875d7	GT3	core:set_object_status:1.0	XX	action_template	action	core	set_object_status	2025-07-03 05:56:46.532999-07	1.0	{"action_template": {"curr_user": "", "action_name": "Set Status", "action_user": [], "description": "Set Status", "method_name": "do_action_set_object_status", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;in_progress&quot;>In Progress</option><option value=&quot;complete&quot;>Complete</option><option value=&quot;failed&quot;>Fail</option><option value=&quot;abandoned&quot;>Abandon</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:46.532999-07	t
08fac2ba-c5e1-4393-9926-e4c3d67328c5	GT4	core:print_barcode_label:1.0	XX	action_template	action	core	print_barcode_label	2025-07-03 05:56:46.542929-07	1.0	{"properties": {"name": "", "comments": ""}, "action_template": {"curr_user": "", "action_name": "Print Label", "action_user": [], "description": "Print Label2", "method_name": "do_action_print_barcode_label", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:46.542929-07	t
a523522b-428e-4175-9e8f-00aa181fb7f5	GT5	core:add-relationships:1.0	XX	action_template	action	core	add-relationships	2025-07-03 05:56:46.552081-07	1.0	{"action_template": {"curr_user": "", "action_name": "Add Parent or Child Relationship(s)", "action_user": [], "description": "Add Parent or Child Relationships (aka lineages).", "method_name": "do_action_add_relationships", "action_order": "0", "capture_data": "yes", "captured_data": {"_parent_euids": "EUIDS (one per line):<textarea required name=&quot;euids&quot;/></textarea>", "_relationship_type": "Relationship Type (not yet controlled): <input required type=&quot;text&quot; name=&quot;relationship_type&quot; value=&quot;generic&quot;/>", "_lineage_type_to_create": "Create Relationship(s): <select required name=&quot;lineage_type_to_create&quot;><option value=&quot;parent&quot;>Parents of this record</option><option value=&quot;child&quot;>Children of this record</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:46.552081-07	t
58348944-3c6d-418a-844a-1fd146c8f2e1	GT6	data:record_cfdna_quant_outcome:1.0	XX	action_template	action	data	record_cfdna_quant_outcome	2025-07-03 05:56:47.012849-07	1.0	{"properties": {"name": "", "comments": ""}, "action_template": {"curr_user": "", "action_name": "Record Quantification Outcome", "action_user": [], "method_name": "do_action_record_cfdna_quant_outcome", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_object_status": "<select name=&quot;object_status&quot;><option value=&quot;pass&quot;>PASS</option><option value=&quot;fail&quot;>FAIL</option><option value=&quot;repeat&quot;>REPEAT</option></select>", "Interpretation Comments": ""}, "action_enabled": "0", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:47.012849-07	t
48960c63-2668-40d4-b442-7121e3ba9054	GT7	data:log-temperature:1.0	XX	action_template	action	data	log-temperature	2025-07-03 05:56:47.022972-07	1.0	{"properties": {"name": "", "comments": ""}, "action_template": {"curr_user": "", "action_name": "Log Temperature", "action_user": [], "method_name": "do_action_log_temperature", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Log Comment": "", "Temperature (celcuius)": ""}, "action_enabled": "0", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:47.022972-07	t
37a4e666-4b84-466d-8498-0baefd7a3edd	GT8	equipment:add-container-to:1.0	XX	action_template	action	equipment	add-container-to	2025-07-03 05:56:47.483943-07	1.0	{"properties": {"name": "", "comments": ""}, "action_template": {"curr_user": "", "action_name": "Add Container To", "action_user": [], "method_name": "do_action_add_container_to", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "simple_value": "", "captured_data": {"_capture_barcodes": "<textarea name=&quot;container_euids&quot; rows=&quot;12&quot; cols=&quot;50&quot;></textarea>"}, "action_enabled": "1", "child_data_obj": {}, "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "workflow_to_attach": {}, "child_container_obj": {}, "child_workflow_step_obj": {}, "attach_under_root_workflow": {}, "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:47.483943-07	t
09f033cd-f0df-462b-a151-43d92049b8b2	GT9	equipment:remove-container-from:1.0	XX	action_template	action	equipment	remove-container-from	2025-07-03 05:56:47.492776-07	1.0	{"properties": {"name": "", "comments": ""}, "action_template": {"curr_user": "", "action_name": "Remove Container", "action_user": [], "method_name": "do_action_remove_container_from", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_capture_barcodes": "<textarea name=&quot;container_euids&quot; rows=&quot;12&quot; cols=&quot;50&quot;></textarea>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "child_container_obj": {}, "test_requisition_obj": {}, "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:47.492776-07	t
3412d0e5-3fd2-40c1-a84b-ceb1ddb90e3e	GT10	equipment:download-inventory-tsv:1.0	XX	action_template	action	equipment	download-inventory-tsv	2025-07-03 05:56:47.501987-07	1.0	{"properties": {"name": "", "comments": ""}, "action_template": {"curr_user": "", "action_name": "Download Inventory TSV", "action_user": [], "method_name": "do_action_generate_inventory_tsv", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "child_container_obj": {}, "test_requisition_obj": {}, "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:47.501987-07	t
bfab0270-6453-43cb-a2e8-9d3abd12eca8	GT11	equipment:log-temperature:1.0	XX	action_template	action	equipment	log-temperature	2025-07-03 05:56:47.510056-07	1.0	{"properties": {"name": "", "comments": ""}, "action_template": {"curr_user": "", "action_name": "Log Temperature", "action_user": [], "method_name": "do_action_log_temperature", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Temperature (celcius)": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "child_container_obj": {"data/generic/temperature-log/*/": {"json_addl": {"properties": {"comments": ""}}}}, "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:47.510056-07	t
574c583b-b1e3-4142-9922-00529d7126e8	GT12	file_set:add-file:1.0	XX	action_template	action	file_set	add-file	2025-07-03 05:56:47.963962-07	1.0	{"action_template": {"curr_user": "", "action_name": "Add File To Set", "action_user": [], "description": "Add a file to a file set.", "method_name": "do_action_add_file_to_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:47.963962-07	t
1a25617a-4bc5-471c-ad68-08087dfe2488	GT13	file_set:remove-file:1.0	XX	action_template	action	file_set	remove-file	2025-07-03 05:56:47.975234-07	1.0	{"action_template": {"curr_user": "", "action_name": "Remove File From File Set", "action_user": [], "description": "Remove a file from a file set.", "method_name": "do_action_remove_file_from_file_set", "action_order": "0", "capture_data": "yes", "captured_data": {"file_euid": "File EUID: <input name=file_euid />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:47.975234-07	t
fe026be9-ef1e-4db0-8f53-4f254e4c212f	GT14	file_set:share-presigned:1.0	XX	action_template	action	file_set	share-presigned	2025-07-03 05:56:47.982632-07	1.0	{"action_template": {"curr_user": "", "action_name": "Share File Set", "action_user": [], "description": "Create presigned URLs for each file in the set.", "method_name": "do_action_share_file_set_presigned_urls", "action_order": "0", "capture_data": "yes", "captured_data": {"duration": "Duration (days): <input name=duration value=\\"1\\" />"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:47.982632-07	t
bad547f4-644b-4456-9c0b-1d26d160b9a7	GT15	file:download_file:1.0	XX	action_template	action	file	download_file	2025-07-03 05:56:48.43328-07	1.0	{"action_template": {"curr_user": "", "action_name": "Download A File", "action_user": [], "description": "Download File From S3", "method_name": "do_action_download_file", "action_order": "0", "capture_data": "yes", "captured_data": {"_dowload_type": "File Creation Mode: <select name=&quot;download_type&quot;><option value=&quot;dewey&quot; selected >dewey</option><option value=&quot;hybrid&quot;>hybrid</option><option value=&quot;orig&quot;>orig</option></select>", "_create_metadata_file": "Generate Metadata yaml:<select name=&quot;create_metadata_file&quot;><option selected value=&quot;yes&quot;>yes</option><option value=&quot;no&quot;>no</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:48.43328-07	t
4439e2c0-0ccf-4acd-8454-1cf2dc9245f3	GT16	move-queues:move-among-ay-top-queues:1.0	XX	action_template	action	move-queues	move-among-ay-top-queues	2025-07-03 05:56:48.901482-07	1.0	{"properties": {"name": "", "comments": ""}, "action_template": {"curr_user": "", "action_name": "Move Workset To Another Queue", "action_user": [], "description": "Move Workset To Another Queue", "method_name": "do_action_move_workset_to_another_queue", "action_order": "0", "capture_data": "yes", "captured_data": {"_move_to_q": "<select name=&quot;q_selection&quot;><option value=&quot;workflow_step/queue/plasma-isolation-queue-removed/1.0&quot;>Move To Completed Q</option>  <option value=&quot;workflow_step/queue/plasma-isolation-queue-exception/1.0&quot;>Move To Exception Q</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_object": {}, "child_workflow_step_obj": {}, "deactivate_actions_when_executed": [], "workflow_step_to_attach_as_child": {}}}	\N	ready	f	2025-07-03 05:56:48.901482-07	t
93e3736d-0342-4b13-bcb1-43f6f0ad046d	GT17	object:set-child-object:1.0	XX	action_template	action	object	set-child-object	2025-07-03 05:56:49.356324-07	1.0	{"action_template": {"verify": {"instance_type": "*/*/*/*"}, "curr_user": "", "action_name": "Set Object As Child", "action_user": [], "description": "Set Child Obj", "method_name": "do_action_set_child_object", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:49.356324-07	t
f204a8a7-c93f-4647-8a83-60cd13730368	GT18	packages:create_package_and_first_workflow_step:1.0	XX	action_template	action	packages	create_package_and_first_workflow_step	2025-07-03 05:56:49.845446-07	1.0	{"properties": {"name": "", "comments": ""}, "action_template": {"curr_user": "", "action_name": "Register Package", "action_user": [], "description": "Create Package and First Workflow Step", "method_name": "do_action_create_package_and_first_workflow_step", "action_order": "0", "capture_data": "yes", "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {"container/package/generic/*/": {"json_addl": {"properties": {"comments": "x"}}}}, "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/accessioning-steps/package-generated/1.0": {"json_addl": {"properties": {"step_number": "1"}}}}, "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:49.845446-07	t
049706cb-9b8f-467d-954f-9642c7abeb77	GT19	plate:download_plate_map:1.0	XX	action_template	action	plate	download_plate_map	2025-07-03 05:56:50.294806-07	1.0	{"properties": {"name": "", "comments": ""}, "action_template": {"curr_user": "", "action_name": "Download Plate Map TSV", "action_user": [], "method_name": "do_action_download_plate_map", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:50.294806-07	t
5b308530-e7e5-458e-8cf2-1dd82bbdf73b	GT20	plate:stamp_plates_into_plate:1.0	XX	action_template	action	plate	stamp_plates_into_plate	2025-07-03 05:56:50.304285-07	1.0	{"action_template": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Stamp plates into plate", "action_user": [], "description": "Stamp 1-* plates into an existing plate. Link the plate and well relationships, and link contents. Assume direct mapping of wells.", "method_name": "do_stamp_plates_into_plate", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "Source Plate EUIDs:<textarea name=&quot;source_barcodes&quot; rows=&quot;6&quot; cols=&quot;30&quot;></textarea><br>", "Destination Plate EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/plate-operations/plate-added-to/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:50.304285-07	t
589df01e-8e35-4949-8af2-873ba2772644	GT21	pool:create-pool-from-containers:1.0	XX	action_template	action	pool	create-pool-from-containers	2025-07-03 05:56:50.771645-07	1.0	{"action_template": {"verify": {"instance_type": "container/*/*/*"}, "curr_user": "", "action_name": "Create A Pool", "action_user": [], "description": "Create Pool", "method_name": "do_action_create_pool", "verify_data": "yes", "action_order": "0", "capture_data": "yes", "captured_data": {"_capture_barcodes": "<textarea name=&quot;discard_barcodes&quot; rows=&quot;24&quot; cols=&quot;50&quot;></textarea>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "executed_datetime": [], "action_simple_value": "", "child_container_obj": {"container/tube/tube-eppi-1.5ml/1.0/": {"json_addl": {"properties": {"comments": ""}}}}, "child_workflow_step_obj": {"workflow_step/pool/generic/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:50.771645-07	t
ef501631-cae6-4e0e-b901-c6fc61245b56	GT22	test_requisitions:set_verification_state:1.0	XX	action_template	action	test_requisitions	set_verification_state	2025-07-03 05:56:51.235388-07	1.0	{"properties": {"name": "", "comments": ""}, "action_template": {"curr_user": "", "action_name": "Verify Test Req", "action_user": [], "method_name": "do_action_set_verification_state", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"comment": "", "Verification Status": ""}, "action_enabled": "1", "max_executions": "1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:51.235388-07	t
b08268ac-cb7a-4e88-baae-ed61c441ea37	GT23	test_requisitions:add_container_to_assay_q:1.0	XX	action_template	action	test_requisitions	add_container_to_assay_q	2025-07-03 05:56:51.245386-07	1.0	{"properties": {"name": "", "comments": ""}, "action_template": {"curr_user": "", "action_name": "Add Specimen to Assay Queue", "action_user": [], "method_name": "do_action_add_container_to_assay_q", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "workflow_obj": {"workflow/assay/generic/1.0": {"json_addl": {"properties": {"step_number": "1"}}}}, "captured_data": {"Container EUID": "", "___workflow/assay/": "<select name=&quot;assay_selection&quot;><option value=&quot;/add/an/assay/wf&quot;>AddAnAssayInTheObjectActionDef</option></select>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:51.245386-07	t
98df25d3-bf76-4ae0-a3cf-295bb667652f	GT24	workflow_step_accessioning:create_child_container_and_link_child_workflow_step:1.0	XX	action_template	action	workflow_step_accessioning	create_child_container_and_link_child_workflow_step	2025-07-03 05:56:51.702162-07	1.0	{"properties": {"name": "", "comments": ""}, "action_template": {"curr_user": "", "action_name": "Register Kit", "action_user": [], "method_name": "do_action_create_child_container_and_link_child_workflow_step", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "simple_value": "", "captured_data": {"Comments": "", "Kit Barcode": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "child_container_obj": {"container/kit/generic/*/": {"json_addl": {"properties": {"comments": ""}}}}, "child_workflow_step_obj": {"workflow_step/accessioning-steps/kit-generated/*/": {"json_addl": {"properties": {"comments": ""}}}}, "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:51.702162-07	t
6b0e3b65-2244-4963-b36c-e2a9582344b2	GT65	sample:blood-plasma:1.0	MX	content_template	content	sample	blood-plasma	2025-07-03 05:57:00.098763-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "plasma (whole blood)", "volume": "", "comments": "", "lab_code": "", "volume_units": ""}, "description": "plasma (whole blood)", "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:00.098763-07	t
cae63b72-f330-441c-8200-30509666adc2	GT25	workflow_step_accessioning:create_child_container_and_link_child_workflow_step_specimen:1.0	XX	action_template	action	workflow_step_accessioning	create_child_container_and_link_child_workflow_step_specimen	2025-07-03 05:56:51.71187-07	1.0	{"properties": {"name": "", "comments": ""}, "action_template": {"curr_user": "", "action_name": "Register Specimen Container", "action_user": [], "method_name": "do_action_create_child_container_and_link_child_workflow_step", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "simple_value": "", "captured_data": {"DOB": "", "MRN": "", "Comment": "", "QC Pass": "", "Collection Date": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "child_container_obj": {"container/tube/tube-strek-10ml/*/": {"json_addl": {"properties": {"comments": ""}}, "instantiation_layouts": [[{"content/specimen/blood-whole/*/": {"json_addl": {"properties": {"comment": ""}}}}]]}}, "child_workflow_step_obj": {"workflow_step/accessioning-steps/container-generated/*/": {"json_addl": {"properties": {"comments": ""}}}}, "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:51.71187-07	t
65bccbd3-92a2-4f05-a6f9-9cbf95af621d	GT26	workflow_step_accessioning:create_test_req_and_link_child_workflow_step_dup:1.0	XX	action_template	action	workflow_step_accessioning	create_test_req_and_link_child_workflow_step_dup	2025-07-03 05:56:51.720854-07	1.0	{"action_template": {"curr_user": "", "action_name": "Create New Test Req & Assoc To Specimen Tube", "action_user": [], "method_name": "do_action_create_test_req_and_link_child_workflow_step", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Physician": "", "Order Type": "", "Patient Name": ""}, "action_enabled": "1", "max_executions": "1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "child_container_obj": {}, "test_requisition_obj": {"test_requisition/clinical/pan-cancer-panel/1.0/": {"json_addl": {"properties": {"comments": ""}}}}, "child_workflow_step_obj": {"workflow_step/accessioning-steps/test-requisition-generated2/*/": {"json_addl": {"properties": {"comments": ""}}}}, "deactivate_actions_when_executed": ["action/workflow_step_accessioning/create_test_req_and_link_child_workflow_step_dup/1.0"]}}	\N	ready	f	2025-07-03 05:56:51.720854-07	t
7463cce0-e0a2-401c-a36a-50823db2f8cf	GT27	workflow_step_accessioning:ycreate_test_req_and_link_child_workflow_step:1.0	XX	action_template	action	workflow_step_accessioning	ycreate_test_req_and_link_child_workflow_step	2025-07-03 05:56:51.729001-07	1.0	{"properties": {"name": "", "comments": ""}, "action_template": {"curr_user": "", "action_name": "Link Tube To Existing  Req.b", "action_user": [], "method_name": "do_action_ycreate_test_req_and_link_child_workflow_step", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Comment": "", "Tube EUID": "", "Test Requisition EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "child_container_obj": {}, "test_requisition_obj": {"test_requisition/clinical/pan-cancer-panel/1.0/": {"json_addl": {"properties": {"comments": ""}}}}, "child_workflow_step_obj": {"workflow_step/accessioning-steps/test-requisition-generated2/*/": {"json_addl": {"properties": {"comments": ""}}}}, "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:51.729001-07	t
48bf27c7-068c-4625-bce5-1826388f36fb	GT28	workflow_step_accessioning:ycreate_test_req_and_link_child_workflow_step2:1.0	XX	action_template	action	workflow_step_accessioning	ycreate_test_req_and_link_child_workflow_step2	2025-07-03 05:56:51.736163-07	1.0	{"properties": {"name": "", "comments": ""}, "action_template": {"curr_user": "", "action_name": "Link Tube To Existing  Req.b", "action_user": [], "method_name": "do_action_ycreate_test_req_and_link_child_workflow_step", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Comment": "", "Tube EUID": "", "Test Requisition EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "child_container_obj": {}, "test_requisition_obj": {"test_requisition/clinical/pan-cancer-panel/1.0/": {"json_addl": {"properties": {"comments": ""}}}}, "child_workflow_step_obj": {"workflow_step/accessioning-steps/test-requisition-generated2/*/": {"json_addl": {"properties": {"comments": ""}}}}, "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:51.736163-07	t
77d0a516-d631-4452-9664-b46b1fb760f9	GT29	workflow_step_analysis:download_quant_data:1.0	XX	action_template	action	workflow_step_analysis	download_quant_data	2025-07-03 05:56:52.188322-07	1.0	{"properties": {"name": "", "comments": ""}, "action_template": {"curr_user": "", "action_name": "Download Quant Data TSV", "action_user": [], "method_name": "do_action_download_quant_data", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "0", "child_data_obj": {}, "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "child_container_obj": {}, "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:52.188322-07	t
a963a0b5-9398-493c-8a4f-7a0f6cb58a0a	GT30	workflow_step_destroy:complete_specimen_destroy:1.0	XX	action_template	action	workflow_step_destroy	complete_specimen_destroy	2025-07-03 05:56:52.641013-07	1.0	{"properties": {"name": "", "comments": ""}, "action_template": {"curr_user": "", "action_name": "Reverse Destroy", "action_user": [], "method_name": "do_action_complete_specimen_destroy", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "simple_value": "", "captured_data": {"Comments": ""}, "action_enabled": "0", "max_executions": "1", "action_executed": "0", "executed_datetime": [], "child_container_obj": {}, "child_workflow_step_obj": {}, "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:52.641013-07	t
a9a238f2-2fb8-41c8-b64e-5f0f2f51e26d	GT31	workflow_step_extraction:cfdna_post_extraction_quant:1.0	XX	action_template	action	workflow_step_extraction	cfdna_post_extraction_quant	2025-07-03 05:56:53.087159-07	1.0	{"properties": {"name": "", "comments": ""}, "action_template": {"curr_user": "", "action_name": "Quant Plate", "action_user": [], "method_name": "do_action_cfdna_post_extraction_quant", "action_order": "0", "capture_data": "no", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "child_container_obj": {"data/generic/cfdna-quant/*/": {"json_addl": {"properties": {"comments": ""}}}}, "child_workflow_step_obj": {"workflow_step/extraction-steps/cfdna-quantified/*/": {"json_addl": {"properties": {"comments": "", "step_number": "1"}}}}, "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:53.087159-07	t
595b2ed6-7759-4dd9-bce7-e755190406e9	GT32	workflow_step_plate_operations:stamp_copy_plate:1.0	XX	action_template	action	workflow_step_plate_operations	stamp_copy_plate	2025-07-03 05:56:53.556901-07	1.0	{"properties": {"name": "", "comments": ""}, "action_template": {"curr_user": "", "action_name": "Stamp Plate (no well address changes)", "action_user": [], "method_name": "do_action_stamp_copy_plate", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "simple_value": "", "captured_data": {"_capture_barcodes": "Plate EUID:<textarea name=&quot;plate_euid&quot; rows=&quot;1&quot; cols=&quot;16&quot;></textarea>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "child_container_obj": {}, "child_workflow_step_obj": {"workflow_step/plate-operations/stamp-copy-plate/*/": {"json_addl": {"properties": {"comments": "", "step_number": ""}}}}, "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:53.556901-07	t
f9d50b10-de39-4ad4-ade4-767adbd86ae8	GT33	workflow_step_plate_operations:cfdna_quant:1.0	XX	action_template	action	workflow_step_plate_operations	cfdna_quant	2025-07-03 05:56:53.567628-07	1.0	{"properties": {"name": "", "comments": ""}, "action_template": {"curr_user": "", "action_name": "Upload cfDNA Quant Data", "action_user": [], "method_name": "do_action_cfdna_quant", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_uploaded_file": "<input type=file name=gdna_quant ></input>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "child_container_obj": {"data/generic/gdna-quantification/*/": {"json_addl": {"properties": {"comments": ""}}}}, "child_workflow_step_obj": {"workflow_step/analysis/gdna-quantified/*/": {"json_addl": {"properties": {"comments": "", "step_number": "1"}}}}, "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:53.567628-07	t
72f7584f-a95f-47ca-b02d-10e9f734bfef	GT34	workflow_step_queue:link_tubes_auto:1.0	XX	action_template	action	workflow_step_queue	link_tubes_auto	2025-07-03 05:56:54.021859-07	1.0	{"properties": {"name": "", "comments": ""}, "action_template": {"curr_user": "", "action_name": "Plasma Isolation Tube to Tube", "action_user": [], "method_name": "do_action_link_tubes_auto", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "simple_value": "", "captured_data": {"_capture_barcodes": "<textarea name=&quot;discard_barcodes&quot; rows=&quot;24&quot; cols=&quot;50&quot;></textarea>"}, "action_enabled": "1", "child_data_obj": {}, "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "child_container_obj": {}, "child_workflow_step_obj": {"workflow_step/queue/extraction-queue-available/*/": {"json_addl": {"properties": {"comments": "", "step_number": "0"}}}}, "attach_under_root_workflow": {"workflow/assay/hla-typing/1.2/": {}}, "attach_under_root_workflow_queue": {"workflow_step/queue/all-purpose/1.0/": {}}, "deactivate_actions_when_executed": [], "workflow_step_to_attach_as_child": {"workflow/assay-workset/hla-typing/1.2/": {"json_addl": {"properties": {"name": "", "comments": "", "step_number": "0"}}}}}}	\N	ready	f	2025-07-03 05:56:54.021859-07	t
2810a6e6-bf77-40f3-aa57-feeecaf7d589	GT35	workflow_step_queue:create_test_req_and_link_child_workflow_step:1.0	XX	action_template	action	workflow_step_queue	create_test_req_and_link_child_workflow_step	2025-07-03 05:56:54.03264-07	1.0	{"properties": {"name": "", "comments": ""}, "action_template": {"curr_user": "", "action_name": "Create New Test Req & Assoc To Specimen Tube", "action_user": [], "method_name": "do_action_create_test_req_and_link_child_workflow_step", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Physician": "", "Order Type": "", "Patient Name": ""}, "action_enabled": "1", "max_executions": "1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "child_container_obj": {}, "test_requisition_obj": {"test_requisition/clinical/pan-cancer-panel/1.0/": {"json_addl": {"properties": {"comments": ""}}}}, "child_workflow_step_obj": {"workflow_step/accessioning-steps/test-requisition-generated2/*/": {"json_addl": {"properties": {"comments": ""}}}}, "deactivate_actions_when_executed": ["ycreate_test_req_and_link_child_workflow_step"]}}	\N	ready	f	2025-07-03 05:56:54.03264-07	t
c28e1cd1-224c-429c-8397-5813b2b9d50e	GT36	workflow_step_queue:ycreate_test_req_and_link_child_workflow_step:1.0	XX	action_template	action	workflow_step_queue	ycreate_test_req_and_link_child_workflow_step	2025-07-03 05:56:54.040451-07	1.0	{"properties": {"name": "", "comments": ""}, "action_template": {"curr_user": "", "action_name": "Link Tube To Existing  Req.b", "action_user": [], "method_name": "do_action_ycreate_test_req_and_link_child_workflow_step", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Comment": "", "Tube EUID": "", "Test Requisition EUID": ""}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "action_simple_value": "", "child_container_obj": {}, "test_requisition_obj": {"test_requisition/clinical/pan-cancer-panel/1.0/": {"json_addl": {"properties": {"comments": ""}}}}, "child_workflow_step_obj": {"workflow_step/accessioning-steps/test-requisition-generated2/*/": {"json_addl": {"properties": {"comments": ""}}}}, "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:54.040451-07	t
2e9cce06-c2e4-4bc6-ab28-9f1fb91072cd	GT37	workflow_step_queue:fill_plate_undirected:1.0	XX	action_template	action	workflow_step_queue	fill_plate_undirected	2025-07-03 05:56:54.048666-07	1.0	{"properties": {"name": "", "comments": ""}, "action_template": {"curr_user": "", "action_name": "Map Tubes To Extraction Plate Wells", "action_user": [], "method_name": "do_action_fill_plate_undirected", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "simple_value": "", "captured_data": {"_capture_barcodes": "<textarea name=&quot;discard_barcodes&quot; rows=&quot;24&quot; cols=&quot;50&quot;></textarea>"}, "action_enabled": "1", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "child_container_obj": {}, "child_workflow_step_obj": {"workflow_step/plate-operations/extraction-plate-filled/*/": {"json_addl": {"properties": {"comments": "", "step_number": ""}}}}, "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:54.048666-07	t
1eded08b-0b24-4da3-9259-f4cbcca1b4a4	GT62	reagent:ILMN-nova-seq-kit:1.0	MRX	content_template	content	reagent	ILMN-nova-seq-kit	2025-07-03 05:56:59.623449-07	1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "Novaseq sequencing kit", "volume": "", "comments": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "Novaseq sequencing kit", "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:56:59.623449-07	t
2855fdeb-00d2-46d8-823d-8dd13fb1d82f	GT38	workflow_step_queue:fill_plate_directed:1.0	XX	action_template	action	workflow_step_queue	fill_plate_directed	2025-07-03 05:56:54.056092-07	1.0	{"action_template": {"curr_user": "", "action_name": "Fill Plate & Specify Map", "action_user": [], "method_name": "do_action_fill_plate_directed", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "simple_value": "", "captured_data": {"_capture_barcodes": "<textarea name=&quot;discard_barcodes&quot; rows=&quot;24&quot; cols=&quot;50&quot;></textarea>"}, "action_enabled": "0", "max_executions": "-1", "action_executed": "0", "executed_datetime": [], "child_container_obj": {}, "child_workflow_step_obj": {"workflow_step/plate-operations/plate-filled/*/": {"json_addl": {"properties": {"comments": "", "step_number": "auto(int())"}}}}, "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:54.056092-07	t
094d3f62-c3d3-4501-a801-9ffab36ec973	GT39	workflow:create_package_and_first_workflow_step:1.0	XX	action_template	action	workflow	create_package_and_first_workflow_step	2025-07-03 05:56:54.51131-07	1.0	{"properties": {"name": "", "comments": ""}, "action_template": {"curr_user": "", "action_name": "Register Package", "action_user": [], "method_name": "do_action_create_package_and_first_workflow_step", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"Carrier Name": "", "Tracking Number": ""}, "action_enabled": "1", "max_executions": "1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {"container/package/generic/*/": {"json_addl": {"properties": {"comments": "x"}}}}, "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/accessioning-steps/package-generated/1.0": {"json_addl": {"properties": {"step_number": "1"}}}}, "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:54.51131-07	t
558b3f42-d3db-46dc-bc44-dba056406fe5	GT40	workflow:destroy_specimen_containers:1.0	XX	action_template	action	workflow	destroy_specimen_containers	2025-07-03 05:56:54.522819-07	1.0	{"properties": {"name": "", "comments": ""}, "action_template": {"curr_user": "", "action_name": "Scan Conatiner Barcodes", "action_user": [], "method_name": "do_action_destroy_specimen_containers", "action_order": "0", "capture_data": "yes", "printer_opts": {"label_style": "", "printer_name": ""}, "captured_data": {"_capture_barcodes": "<textarea name=&quot;discard_barcodes&quot; rows=&quot;24&quot; cols=&quot;50&quot;></textarea>"}, "action_enabled": "1", "max_executions": "1", "action_executed": "0", "executed_datetime": [], "new_container_obj": {}, "action_simple_value": "", "child_workflow_step_obj": {"workflow_step/destroy-steps/specimen-containers-destroyed/1.0": {"json_addl": {"properties": {"comment": "", "step_number": "1"}}}}, "deactivate_actions_when_executed": []}}	\N	ready	f	2025-07-03 05:56:54.522819-07	t
a54b1949-0198-401f-81ef-6071793ccfe9	GT41	generic:patient:1.0	AX	actor_template	actor	generic	patient	2025-07-03 05:56:54.982533-07	1.0	{"singleton": "0", "properties": {"fax": "", "name": "A Patient", "phone": "", "address": "", "comments": "", "lab_code": "", "other_uids": [], "patient_id": ""}, "description": "Patient", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:56:54.982533-07	t
e76cf793-06c2-4109-8df8-6e0e3517ddd6	GT42	generic:physician:1.0	AX	actor_template	actor	generic	physician	2025-07-03 05:56:54.992571-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"fax": "", "name": "Clinician", "phone": "", "address": "", "comments": "", "lab_code": "", "other_uids": [], "clinician_id": ""}, "description": "Clinician", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:56:54.992571-07	t
2959b93a-fd1b-4d2c-a5ca-b6fd917b4cbc	GT43	kit:generic:1.0	CX	container_template	container	kit	generic	2025-07-03 05:56:55.446608-07	1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "Collection Kit", "comments": "", "lab_code": ""}, "description": "Collection Kit", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:56:55.446608-07	t
4bb99ae7-dc05-460e-95c5-6fec42076080	GT44	package:generic:1.0	CX	container_template	container	package	generic	2025-07-03 05:56:55.898352-07	1.0	{"cogs": {"cost": "30.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "Package", "comments": "", "lab_code": ""}, "description": "Package", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:56:55.898352-07	t
6bae2028-8873-4717-9e10-821a62bcd9f0	GT45	plate:fixed-plate-24:1.0	CX	container_template	container	plate	fixed-plate-24	2025-07-03 05:56:56.347558-07	1.0	{"cogs": {"cost": "2.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "24 fixed well plate", "cont_address": {"col_idx": "", "row_idx": "", "col_name": "", "row_name": ""}, "action_groups": {}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}]]}	\N	ready	f	2025-07-03 05:56:56.347558-07	t
ab7ccba5-c577-4dad-b713-59623d41b45a	GT63	sample:gdna:1.0	MX	content_template	content	sample	gdna	2025-07-03 05:57:00.08252-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "genomic DNA", "volume": "", "comments": "", "lab_code": "", "volume_units": ""}, "description": "genomic DNA", "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:00.08252-07	t
36f86ab4-1ac5-49a6-8dc3-0285b1185fe3	GT46	plate:fixed-plate-96:1.0	CX	container_template	container	plate	fixed-plate-96	2025-07-03 05:56:56.357901-07	1.0	{"cogs": {"cost": "4.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "96 well plate fixed", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "lot_expiration_date": ""}, "description": "96 well fixed plate", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate": {"actions": {"action/plate/*/1.0/": {}}, "group_name": "Plate Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A1"}, "cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A2"}, "cont_address": {"name": "A2", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A3"}, "cont_address": {"name": "A3", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A4"}, "cont_address": {"name": "A4", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A5"}, "cont_address": {"name": "A5", "col_idx": "4", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A6"}, "cont_address": {"name": "A6", "col_idx": "5", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A7"}, "cont_address": {"name": "A7", "col_idx": "6", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A8"}, "cont_address": {"name": "A8", "col_idx": "7", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A9"}, "cont_address": {"name": "A9", "col_idx": "8", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A10"}, "cont_address": {"name": "A10", "col_idx": "9", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A11"}, "cont_address": {"name": "A11", "col_idx": "10", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "A12"}, "cont_address": {"name": "A12", "col_idx": "11", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B1"}, "cont_address": {"name": "B1", "col_idx": "0", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B2"}, "cont_address": {"name": "B2", "col_idx": "1", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B3"}, "cont_address": {"name": "B3", "col_idx": "2", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B4"}, "cont_address": {"name": "B4", "col_idx": "3", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B5"}, "cont_address": {"name": "B5", "col_idx": "4", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B6"}, "cont_address": {"name": "B6", "col_idx": "5", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B7"}, "cont_address": {"name": "B7", "col_idx": "6", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B8"}, "cont_address": {"name": "B8", "col_idx": "7", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B9"}, "cont_address": {"name": "B9", "col_idx": "8", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B10"}, "cont_address": {"name": "B10", "col_idx": "9", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B11"}, "cont_address": {"name": "B11", "col_idx": "10", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "B12"}, "cont_address": {"name": "B12", "col_idx": "11", "row_idx": "1", "col_name": "2", "row_name": "B"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C1"}, "cont_address": {"name": "C1", "col_idx": "0", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C2"}, "cont_address": {"name": "C2", "col_idx": "1", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C3"}, "cont_address": {"name": "C3", "col_idx": "2", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C4"}, "cont_address": {"name": "C4", "col_idx": "3", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C5"}, "cont_address": {"name": "C5", "col_idx": "4", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C6"}, "cont_address": {"name": "C6", "col_idx": "5", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C7"}, "cont_address": {"name": "C7", "col_idx": "6", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C8"}, "cont_address": {"name": "C8", "col_idx": "7", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C9"}, "cont_address": {"name": "C9", "col_idx": "8", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C10"}, "cont_address": {"name": "C10", "col_idx": "9", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C11"}, "cont_address": {"name": "C11", "col_idx": "10", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "C12"}, "cont_address": {"name": "C12", "col_idx": "11", "row_idx": "2", "col_name": "3", "row_name": "C"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D1"}, "cont_address": {"name": "D1", "col_idx": "0", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D2"}, "cont_address": {"name": "D2", "col_idx": "1", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D3"}, "cont_address": {"name": "D3", "col_idx": "2", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D4"}, "cont_address": {"name": "D4", "col_idx": "3", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D5"}, "cont_address": {"name": "D5", "col_idx": "4", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D6"}, "cont_address": {"name": "D6", "col_idx": "5", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D7"}, "cont_address": {"name": "D7", "col_idx": "6", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D8"}, "cont_address": {"name": "D8", "col_idx": "7", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D9"}, "cont_address": {"name": "D9", "col_idx": "8", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D10"}, "cont_address": {"name": "D10", "col_idx": "9", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D11"}, "cont_address": {"name": "D11", "col_idx": "10", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "D12"}, "cont_address": {"name": "D12", "col_idx": "11", "row_idx": "3", "col_name": "4", "row_name": "D"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E1"}, "cont_address": {"name": "E1", "col_idx": "0", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E2"}, "cont_address": {"name": "E2", "col_idx": "1", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E3"}, "cont_address": {"name": "E3", "col_idx": "2", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E4"}, "cont_address": {"name": "E4", "col_idx": "3", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E5"}, "cont_address": {"name": "E5", "col_idx": "4", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E6"}, "cont_address": {"name": "E6", "col_idx": "5", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E7"}, "cont_address": {"name": "E7", "col_idx": "6", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E8"}, "cont_address": {"name": "E8", "col_idx": "7", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E9"}, "cont_address": {"name": "E9", "col_idx": "8", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E10"}, "cont_address": {"name": "E10", "col_idx": "9", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E11"}, "cont_address": {"name": "E11", "col_idx": "10", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "E12"}, "cont_address": {"name": "E12", "col_idx": "11", "row_idx": "4", "col_name": "5", "row_name": "E"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F1"}, "cont_address": {"name": "F1", "col_idx": "0", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F2"}, "cont_address": {"name": "F2", "col_idx": "1", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F3"}, "cont_address": {"name": "F3", "col_idx": "2", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F4"}, "cont_address": {"name": "F4", "col_idx": "3", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F5"}, "cont_address": {"name": "F5", "col_idx": "4", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F6"}, "cont_address": {"name": "F6", "col_idx": "5", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F7"}, "cont_address": {"name": "F7", "col_idx": "6", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F8"}, "cont_address": {"name": "F8", "col_idx": "7", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F9"}, "cont_address": {"name": "F9", "col_idx": "8", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F10"}, "cont_address": {"name": "F10", "col_idx": "9", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F11"}, "cont_address": {"name": "F11", "col_idx": "10", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "F12"}, "cont_address": {"name": "F12", "col_idx": "11", "row_idx": "5", "col_name": "6", "row_name": "F"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G1"}, "cont_address": {"name": "G1", "col_idx": "0", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G2"}, "cont_address": {"name": "G2", "col_idx": "1", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G3"}, "cont_address": {"name": "G3", "col_idx": "2", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G4"}, "cont_address": {"name": "G4", "col_idx": "3", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G5"}, "cont_address": {"name": "G5", "col_idx": "4", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G6"}, "cont_address": {"name": "G6", "col_idx": "5", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G7"}, "cont_address": {"name": "G7", "col_idx": "6", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G8"}, "cont_address": {"name": "G8", "col_idx": "7", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G9"}, "cont_address": {"name": "G9", "col_idx": "8", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G10"}, "cont_address": {"name": "G10", "col_idx": "9", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G11"}, "cont_address": {"name": "G11", "col_idx": "10", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "G12"}, "cont_address": {"name": "G12", "col_idx": "11", "row_idx": "6", "col_name": "7", "row_name": "G"}}}}], [{"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "H1"}, "cont_address": {"name": "H1", "col_idx": "0", "row_idx": "7", "col_name": "8", "row_name": "H"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "H2"}, "cont_address": {"name": "H2", "col_idx": "1", "row_idx": "7", "col_name": "8", "row_name": "H"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "H3"}, "cont_address": {"name": "H3", "col_idx": "2", "row_idx": "7", "col_name": "8", "row_name": "H"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "H4"}, "cont_address": {"name": "H4", "col_idx": "3", "row_idx": "7", "col_name": "8", "row_name": "H"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "H5"}, "cont_address": {"name": "H5", "col_idx": "4", "row_idx": "7", "col_name": "8", "row_name": "H"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "H6"}, "cont_address": {"name": "H6", "col_idx": "5", "row_idx": "7", "col_name": "8", "row_name": "H"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "H7"}, "cont_address": {"name": "H7", "col_idx": "6", "row_idx": "7", "col_name": "8", "row_name": "H"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "H8"}, "cont_address": {"name": "H8", "col_idx": "7", "row_idx": "7", "col_name": "8", "row_name": "H"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "H9"}, "cont_address": {"name": "H9", "col_idx": "8", "row_idx": "7", "col_name": "8", "row_name": "H"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "H10"}, "cont_address": {"name": "H10", "col_idx": "9", "row_idx": "7", "col_name": "8", "row_name": "H"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "H11"}, "cont_address": {"name": "H11", "col_idx": "10", "row_idx": "7", "col_name": "8", "row_name": "H"}}}}, {"container/well/fixed-plate-well/*/": {"json_addl": {"properties": {"name": "H12"}, "cont_address": {"name": "H12", "col_idx": "11", "row_idx": "7", "col_name": "8", "row_name": "H"}}}}]]}	\N	ready	f	2025-07-03 05:56:56.357901-07	t
88594a47-129c-4db5-886a-aca0f62ba7d2	GT47	rack:tube-rack-generic:1.0	CX	container_template	container	rack	tube-rack-generic	2025-07-03 05:56:56.812244-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "4 position Rack", "comments": "", "lab_code": ""}, "description": "4 position Rack", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/tube/tube-strek-10ml/*/": {"json_addl": {"cont_address": {"name": "A1", "col_idx": "0", "row_idx": "0", "col_name": "1", "row_name": "A"}}}}, {"container/tube/tube-strek-10ml/*/": {"json_addl": {"cont_address": {"name": "B1", "col_idx": "1", "row_idx": "0", "col_name": "1", "row_name": "B"}}}}, {"container/tube/tube-strek-10ml/*/": {"json_addl": {"cont_address": {"name": "C1", "col_idx": "2", "row_idx": "0", "col_name": "1", "row_name": "C"}}}}, {"container/tube/tube-eppi-1.5ml/*/": {"json_addl": {"cont_address": {"name": "D1", "col_idx": "3", "row_idx": "0", "col_name": "1", "row_name": "D"}}}}]]}	\N	ready	f	2025-07-03 05:56:56.812244-07	t
d5ee3d31-ad0b-4780-a97d-b0c78bc74ac3	GT48	rack:tube-rack-4-empty:1.0	CX	container_template	container	rack	tube-rack-4-empty	2025-07-03 05:56:56.82144-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "4 position Rack", "comments": "", "lab_code": ""}, "description": "4 position Rack", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {}, "valid_content": ["none"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:56:56.82144-07	t
4c7ddd40-8031-4060-87e4-1755fea7ba18	GT49	shelf:generic:1.0	CX	container_template	container	shelf	generic	2025-07-03 05:56:57.280881-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "Shelf", "comments": "", "lab_code": ""}, "description": "Shelf", "cont_address": {"col_idx": "0", "row_idx": "0", "capacity": "", "col_name": "", "row_name": ""}, "action_groups": {}, "valid_content": [""], "valid_parents": [""], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "fridge": {"actions": {"action/equipment/*/1.0/": {}}, "group_name": "Refrigerator Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:56:57.280881-07	t
c5ef9747-3879-4e3e-bab9-56164b1dda95	GT50	tube:tube-strek-10ml:1.0	CX	container_template	container	tube	tube-strek-10ml	2025-07-03 05:56:57.734023-07	1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "strek 10ml", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "strek 10ml", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:56:57.734023-07	t
d1533fbe-9bf0-46f2-abff-3c74cc80c744	GT51	tube:tube-generic-10ml:1.0	CX	container_template	container	tube	tube-generic-10ml	2025-07-03 05:56:57.744961-07	1.0	{"cogs": {"cost": "1.50", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "10ml tube generic", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "Generic 10ml tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:56:57.744961-07	t
0b2a2111-b3a7-415a-b61b-e9fc0051c56b	GT64	sample:cfdna:1.0	MX	content_template	content	sample	cfdna	2025-07-03 05:57:00.091674-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "cfDNA", "volume": "", "comments": "", "lab_code": "", "volume_units": ""}, "description": "cell free DNA ", "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:00.091674-07	t
ceca7fef-9b77-433d-9420-8d8a15cf1147	GT52	tube:tube-eppi-1.5ml:1.0	CX	container_template	container	tube	tube-eppi-1.5ml	2025-07-03 05:56:57.752706-07	1.0	{"cogs": {"cost": "1.70", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "1.5ml eppi tube", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "1.5ml eppi tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:56:57.752706-07	t
7651bf71-c917-4c36-bd28-ded3fa87e1c2	GT53	tube:tube-10ml-glass:1.0	CX	container_template	container	tube	tube-10ml-glass	2025-07-03 05:56:57.760583-07	1.0	{"cogs": {"cost": "4.70", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "10ml glass tube", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "manufacturer_other": "", "lot_expiration_date": ""}, "description": "10ml glass tube", "cont_address": {"name": "", "col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {}, "valid_content": ["content/.*"], "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:56:57.760583-07	t
507a3395-14f5-4e21-bf68-66df5d242281	GT54	well:fixed-plate-well:1.0	CWX	container_template	container	well	fixed-plate-well	2025-07-03 05:56:58.215895-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "", "height": "", "volume": "", "comments": "", "diameter": "", "lab_code": "", "bottom_type": ""}, "description": "plate well - fixed", "cont_address": {"col_idx": "0", "row_idx": "0", "col_name": "", "row_name": ""}, "action_groups": {}, "valid_content": ["content/.*"], "valid_parents": ["container/plate/fixed-plate.*?/.*?"], "action_imports": {"core": {"actions": {"action/core/set_object_status/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:56:58.215895-07	t
f974507f-5cd4-48ce-82c2-696aa8770434	GT55	control:giab-HG001:1.0	MCX	content_template	content	control	giab-HG001	2025-07-03 05:56:58.667599-07	1.0	{"cogs": {"cost": "7.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "giab NA12878", "volume": "", "comments": "", "lab_code": "", "volume_units": ""}, "description": "giab NA12878", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:56:58.667599-07	t
50d08295-a50b-422e-9c2a-6c1791c08e8f	GT56	control:giab-HG002:1.0	MCX	content_template	content	control	giab-HG002	2025-07-03 05:56:58.67764-07	1.0	{"cogs": {"cost": "5.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "giab NA24385", "volume": "", "comments": "", "lab_code": "", "volume_units": ""}, "description": "giab NA24385", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:56:58.67764-07	t
70d42dd6-b1f4-4293-ba62-97c902f6ffb0	GT57	control:water-ntc:1.0	MCX	content_template	content	control	water-ntc	2025-07-03 05:56:58.684637-07	1.0	{"cogs": {"cost": "0.10", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "NTC - water", "volume": "", "comments": "", "lab_code": "", "volume_units": ""}, "description": "NTC - water", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:56:58.684637-07	t
4765a344-581e-4bc6-9c70-cca99e85637d	GT58	control:synthetic-cfdna:1.0	MCX	content_template	content	control	synthetic-cfdna	2025-07-03 05:56:58.691421-07	1.0	{"cogs": {"cost": "3.1", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "s ynthetic-cfdna", "volume": "", "comments": "", "lab_code": "", "volume_units": ""}, "description": "synthetic-cfdna", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:56:58.691421-07	t
ba1686fd-a5c7-44d1-82c4-4fae8ebcb888	GT59	pool:generic:1.0	MX	content_template	content	pool	generic	2025-07-03 05:56:59.14715-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "Pool of Contents", "comments": "", "lab_code": ""}, "description": "Pool of Contents", "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:56:59.14715-07	t
37a16678-0fa3-401e-8467-f5842fa37866	GT60	reagent:naoh:1.0	MRX	content_template	content	reagent	naoh	2025-07-03 05:56:59.607026-07	1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "NaOH", "volume": "", "comments": "", "lab_code": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "NaOH", "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:56:59.607026-07	t
95078f21-b1f0-46c6-99cf-e198717cca5e	GT61	reagent:idt-probes-rare-mendelian:1.0	MRX	content_template	content	reagent	idt-probes-rare-mendelian	2025-07-03 05:56:59.616012-07	1.0	{"cogs": {"cost": "10.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "idt probes for rare mendelian disease, v1.0", "volume": "", "comments": "", "lot_number": "", "manufacturer": "", "volume_units": "", "received_date": "", "expiration_date": "", "storage_conditions": "", "qualification_status": ""}, "description": "idt probes for rare mendelian disease, v1.0", "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:56:59.616012-07	t
f43b7ac4-b6c7-4159-9ede-58f9dabc5414	GT66	specimen:blood-whole:1.0	MX	content_template	content	specimen	blood-whole	2025-07-03 05:57:00.560583-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "Whole Blood", "volume": "", "comments": "", "lab_code": "", "volume_units": ""}, "description": "Whole Blood", "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:00.560583-07	t
bb3a55ef-c1d8-4b21-8126-e6d2f8b71cc7	GT67	specimen:ffpe-block:1.0	MX	content_template	content	specimen	ffpe-block	2025-07-03 05:57:00.571084-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "FFPE Block", "volume": "", "comments": "", "lab_code": "", "volume_units": ""}, "description": "FFPE Block", "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:00.571084-07	t
f12aaba1-dba0-4228-8f42-bc8404095117	GT68	specimen:ffpe-slice:1.0	MX	content_template	content	specimen	ffpe-slice	2025-07-03 05:57:00.578556-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "actions": {}, "singleton": "0", "properties": {"name": "FFPE Slides", "volume": "", "comments": "", "lab_code": "", "volume_units": ""}, "description": "FFPE Slides", "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:00.578556-07	t
2b1a67ed-2156-484d-aeb7-04566a9307db	GT69	generic:gdna-quantification:1.0	DX	data_template	data	generic	gdna-quantification	2025-07-03 05:57:01.034457-07	1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "DNA Quant Data", "comments": "", "lab_code": ""}, "description": "DNA Quant Data", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "data": {"actions": {"action/data/record_cfdna_quant_outcome/1.0/": {}}, "group_name": "Data Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:01.034457-07	t
e14dcbc9-15a4-4df5-a246-545a24e0fdec	GT70	generic:temperature-log:1.0	DX	data_template	data	generic	temperature-log	2025-07-03 05:57:01.044305-07	1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "Temperate Log Data", "comments": "", "lab_code": ""}, "description": "Temperature Log Data", "action_groups": {}, "action_imports": {}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:01.044305-07	t
1167bc43-6f91-475b-bd87-810fe80ff930	GT71	freezer:freezer-m80c:1.0	EX	equipment_template	equipment	freezer	freezer-m80c	2025-07-03 05:57:01.492497-07	1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "-80C Freezer", "model": "", "serial": "", "comments": "", "lab_code": "", "manufacturer": "", "operataional_status": ""}, "description": "-80C Freezer.", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "fridge": {"actions": {"action/equipment/*/1.0/": {}}, "group_name": "Refrigerator Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/shelf/generic/1.0": {"json_addl": {"cont_address": {"name": "Shelf-1", "col_idx": "0", "row_idx": "0", "capacity": "10", "col_name": "", "row_name": "Shelf-1"}}}}, {"container/shelf/generic/1.0": {"json_addl": {"cont_address": {"name": "Shelf-2", "col_idx": "1", "row_idx": "0", "capacity": "10", "col_name": "", "row_name": "Shelf-2"}}}}, {"container/shelf/generic/1.0": {"json_addl": {"cont_address": {"name": "Shelf-3", "col_idx": "2", "row_idx": "0", "capacity": "10", "col_name": "", "row_name": "Shelf-3"}}}}, {"container/shelf/generic/1.0": {"json_addl": {"cont_address": {"name": "Shelf-4", "col_idx": "3", "row_idx": "0", "capacity": "10", "col_name": "", "row_name": "Shelf-4"}}}}]]}	\N	ready	f	2025-07-03 05:57:01.492497-07	t
85684d6e-da67-48c5-b565-3891c44d2859	GT72	freezer:freezer-m20c:1.0	EX	equipment_template	equipment	freezer	freezer-m20c	2025-07-03 05:57:01.502472-07	1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "-20C Freezer", "model": "", "serial": "", "comments": "", "lab_code": "", "manufacturer": "", "operataional_status": ""}, "description": "-20C Freezer", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "fridge": {"actions": {"action/equipment/*/1.0/": {}}, "group_name": "Refrigerator Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"container/shelf/generic/1.0": {"json_addl": {"cont_address": {"name": "Shelf-1", "col_idx": "0", "row_idx": "0", "capacity": "10", "col_name": "", "row_name": "Shelf-1"}}}}, {"container/shelf/generic/1.0": {"json_addl": {"cont_address": {"name": "Shelf-2", "col_idx": "1", "row_idx": "0", "capacity": "10", "col_name": "", "row_name": "Shelf-2"}}}}]]}	\N	ready	f	2025-07-03 05:57:01.502472-07	t
42b5243c-25db-4bdb-862b-82973a4a162d	GT73	liquid_handler:bravo:1.0	EX	equipment_template	equipment	liquid_handler	bravo	2025-07-03 05:57:01.955911-07	1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "Bravo Liquid Handler", "model": "", "serial": "", "comments": "", "lab_code": "", "manufacturer": "", "operataional_status": ""}, "description": "Bravo Liquid Handler", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:01.955911-07	t
e099d6f0-db3b-4613-b7a6-496537eef95a	GT74	sequencers:novaseq-6000:1.0	EX	equipment_template	equipment	sequencers	novaseq-6000	2025-07-03 05:57:02.395505-07	1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "Illumnina Novaseq 6000", "model": "", "serial": "", "comments": "", "lab_code": "", "manufacturer": "Illumina", "operataional_status": ""}, "description": "Illumnina Novaseq 6000", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:02.395505-07	t
7514dcbb-cca8-4793-8b61-6aa4ce3312b9	GT75	sequencers:miseq-y:1.0	EX	equipment_template	equipment	sequencers	miseq-y	2025-07-03 05:57:02.403676-07	1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "Illumnina MiSeq y", "model": "", "serial": "", "comments": "", "lab_code": "", "manufacturer": "Illumina", "operataional_status": ""}, "description": "Illumnina MiSeq y", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:02.403676-07	t
c91acde4-57c5-4dcf-9704-cbb0fb297475	GT76	sequencers:ont-minion:1.0	EX	equipment_template	equipment	sequencers	ont-minion	2025-07-03 05:57:02.41276-07	1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "ONT MinION", "model": "", "serial": "", "comments": "", "lab_code": "", "manufacturer": "Oxford Nanopore Technologies", "operataional_status": ""}, "description": "ONT MinION", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:02.41276-07	t
1c57e0bf-8b44-426b-8f9e-e96aff46a341	GT77	file_set:generic:1.0	FS	file_template	file	file_set	generic	2025-07-03 05:57:02.835006-07	1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"tag": "", "name": "", "tags": [], "comments": "", "duration": "", "lab_code": "", "ref_type": "", "description": "", "creating_user": "", "rclone_config": {}}, "description": "Generic File Set", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "file_set": {"actions": {"action/file_set/*/1.0/": {}}, "group_name": "File Set Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Set Name", "value_type": "string", "property_key": "name"}, {"required": false, "form_label": "Set Description", "value_type": "string", "property_key": "description"}, {"required": false, "form_label": "Tag", "value_type": "uid-interactive", "property_key": "tag"}, {"required": false, "form_label": "Tags", "value_type": "list", "property_key": "tags"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "Ref Type", "value_type": "uid-static", "property_key": "ref_type"}, {"required": false, "form_label": "Comments", "value_type": "hidden", "property_key": "comments"}], "controlled_properties": {}, "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:02.835006-07	t
9a500c83-8b62-48e1-a0b0-cddd7c498e29	GT78	file:generic:1.0	FI	file_template	file	file	generic	2025-07-03 05:57:03.249325-07	1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "", "purpose": "", "category": "", "comments": "", "lab_code": "", "study_id": "", "variable": "", "file_tags": [], "patient_id": "", "clinician_id": "", "external_ids": [], "sub_category": "", "sub_variable": "", "creating_user": "", "current_s3_key": "", "current_s3_uri": "", "sub_category_2": "", "health_event_id": "", "original_s3_uri": "", "outcome_comment": "", "purpose_subtype": "", "record_datetime": "", "import_or_remote": "", "upload_group_key": "", "original_file_md5": "", "original_file_name": "", "original_file_path": "", "original_server_ip": "", "record_datetime_end": "", "original_file_suffix": "", "current_s3_bucket_name": "", "original_file_size_bytes": "", "original_local_server_name": ""}, "description": "Generic File", "action_groups": {}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "files": {"actions": {"action/file/*/1.0/": {}}, "group_name": "File Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"required": false, "form_label": "Record Start (datetime)", "value_type": "datetime", "property_key": "record_datetime"}, {"required": false, "form_label": "Record End (datetime)", "value_type": "datetime", "property_key": "record_datetime_end"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "1_break"}, {"required": false, "form_label": "Patient ID", "value_type": "uid-interactive", "property_key": "patient_id"}, {"required": false, "form_label": "Clinician ID", "value_type": "uid-static", "property_key": "clinician_id"}, {"required": false, "form_label": "Lab Code", "value_type": "uid-static", "property_key": "lab_code"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "2_break"}, {"required": false, "form_label": "Study ID", "value_type": "uid-static", "property_key": "study_id"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "3_break"}, {"required": false, "form_label": "Purpose", "value_type": "controlled", "property_key": "purpose"}, {"required": false, "form_label": "Purpose Subtype", "value_type": "controlled", "property_key": "purpose_subtype"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "4_break"}, {"required": false, "form_label": "Category", "value_type": "controlled", "property_key": "category"}, {"required": false, "form_label": "Sub Category", "value_type": "controlled", "property_key": "sub_category"}, {"required": false, "form_label": "Sub Category 2", "value_type": "controlled", "property_key": "sub_category_2"}, {"required": false, "form_label": "", "value_type": "break", "property_key": "5_break"}, {"required": false, "form_label": "Creating User", "value_type": "uid-static", "property_key": "creating_user"}, {"required": false, "form_label": "File Tags", "value_type": "list", "property_key": "file_tags"}, {"required": false, "form_label": "Comments", "value_type": "string", "property_key": "comments"}, {"required": false, "form_label": "Public URLs & S3 URIs: Leave Remote / Import", "value_type": "controlled", "property_key": "import_or_remote"}], "controlled_properties": {"purpose": {"enum": ["", "Clinical", "Research", "Other"], "type": "string"}, "category": {"on": "purpose", "enum": {"": [""], "Other": [""], "Clinical": ["", "Diagnosis", "Staging", "Procedure", "Treatments", "Testing", "Imaging", "Hospitalization", "Biospecimen collection", "Disease status", "Other"], "Research": ["", "Test results summary", "Sequencing file", "Other"]}, "type": "dependent string"}, "sub_category": {"on": "category", "enum": {"": [""], "Other": [""], "Imaging": ["", "Abdomen", "Abdomen, pelvis", "Brain", "Chest", "Chest, abdomen", "Chest, abdomen, pelvis", "Other"], "Staging": [""], "Testing": ["", "Blood", "Cancer marker", "ctDNA monitoring", "Genetics", "Pathology"], "Diagnosis": [""], "Procedure": [""], "Treatments": [""], "Disease status": [""], "Hospitalization": [""], "Sequencing file": [""], "Test results summary": [""], "Biospecimen collection": [""]}, "type": "dependent string"}, "sub_category_2": {"on": "category", "enum": {"": [""], "Imaging": ["", "CT", "PET CT", "MRI", "X-Ray", "Ultrasound"]}, "type": "dependent string"}, "purpose_subtype": {"enum": ["", "Institution record", "xCures record", "Summary document", "Data file", "Literature", "Other"], "type": "string"}, "import_or_remote": {"enum": ["Remote", "Import"], "type": "string"}}, "instantiation_layouts": [], "arbitrary_association_strings": []}	\N	ready	f	2025-07-03 05:57:03.249325-07	t
64646531-f2fd-463a-9790-a4587d83f2a5	GT79	shared_ref:generic:1.0	FX	file_template	file	shared_ref	generic	2025-07-03 05:57:03.668781-07	1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "file_type": "", "singleton": "0", "properties": {"name": "", "status": "", "comments": "", "who_pays": "", "visibility": "", "end_datetime": "", "presigned_url": "", "intended_users": [""], "reference_type": "", "start_datetime": "", "valid_duration": ""}, "description": "Generic File Reference", "action_groups": {}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "ui_form_properties": [{"form_label": "Reference Type", "property_key": "reference_type"}, {"form_label": "Visibility", "property_key": "visibility"}, {"form_label": "Reference Status", "property_key": "status"}, {"form_label": "Duration", "property_key": "valid_duration"}, {"form_label": "Start Datetime", "property_key": "start_datetime"}, {"form_label": "End Datetime", "property_key": "end_datetime"}, {"form_label": "Who Pays", "property_key": "who_pays"}], "controlled_properties": {}, "instantiation_layouts": [], "arbitrary_association_strings": []}	\N	ready	f	2025-07-03 05:57:03.668781-07	t
b1dfed48-902d-4761-8e91-3588388a3247	GT80	generic:generic:1.0	GX	generic_template	generic	generic	generic	2025-07-03 05:57:04.1234-07	1.0	{"cogs": {"cost": "1.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "Generic Object Name (displayed in UI, can be easily overidded in child template references)", "comments": "", "lab_code": ""}, "description": "Generic Object Description", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "packages": {"actions": {"action/packages/create_package_and_first_workflow_step/1.0/": {"action_name": "Override Action Name", "description": "Override Default Description"}}, "group_name": "Package Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:04.1234-07	t
b7102e95-8645-4f7f-ad5c-269b41b356ac	GT81	generic:health-event:1.0	EV	health_event_template	health_event	generic	health-event	2025-07-03 05:57:04.535922-07	1.0	{"cogs": {"cost": "0.00", "state": "inactive", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "Generic Object Name (displayed in UI, can be easily overidded in child template references)", "comments": "", "lab_code": ""}, "description": "Generic Health Event", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:04.535922-07	t
285d1278-67df-4d62-b2de-67d49bca1572	GT88	analysis:plasma-quantification:1.0	WSX	workflow_step_template	workflow_step	analysis	plasma-quantification	2025-07-03 05:57:06.256703-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "Plasma QUANT", "comments": "", "lab_code": "", "step_number": ""}, "description": "Plasma QUANT", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "data_action": {"actions": {"action/workflow_step_analysis/download_quant_data/1.0/": {}}, "group_name": "Data Actions", "group_order": "2"}}, "expected_inputs": [], "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:06.256703-07	t
788e3a58-f719-4bfe-a013-2d3e7d33fc1d	GT82	clinical:pan-cancer-panel:1.0	TRX	test_requisition_template	test_requisition	clinical	pan-cancer-panel	2025-07-03 05:57:04.952392-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "layouts": [], "singleton": "0", "properties": {"name": "pan-cancer 239 Gene Panel Requisition", "comments": "", "lab_code": ""}, "description": "pan-cancer 239 Gene Panel Requisition", "action_groups": {}, "valid_parents": [{"workflow/assay/hla-typing/1.2/": {}}, {"workflow/assay/carrier-screen/3.9/": {}}], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "test_requisitions": {"actions": {"action/test_requisitions/set_verification_state/1.0/": {}, "action/test_requisitions/add_container_to_assay_q/1.0/": {"captured_data": {"___workflow/assay/": "<select name=&quot;assay_selection&quot;><option value=&quot;workflow/assay/hla-typing/1.2&quot;>HLA Typing v1.2</option><option value=&quot;workflow/assay/carrier-screen/3.9&quot;>Carrier Screen v3.9</option></select>"}}}, "group_name": "Test Req Actions", "group_order": "2"}}}	\N	ready	f	2025-07-03 05:57:04.952392-07	t
544eb81d-8a02-4606-87b5-4eff70d7d1e1	GT83	research:noamp-wgs:1.0	TRX	test_requisition_template	test_requisition	research	noamp-wgs	2025-07-03 05:57:05.368267-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "layouts": [], "singleton": "0", "properties": {"name": "noamp-wgs 3x", "comments": "", "lab_code": "R&D"}, "description": "noamp-wgs 3x", "action_groups": {}, "valid_parents": [], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "test_requisitions": {"actions": {"action/test_requisitions/*/1.0/": {}}, "group_name": "Test Req Actions", "group_order": "2"}}}	\N	ready	f	2025-07-03 05:57:05.368267-07	t
4f62b8c2-2d9c-49fe-a609-086f412026be	GT84	accessioning-steps:package-generated:1.0	WSX	workflow_step_template	workflow_step	accessioning-steps	package-generated	2025-07-03 05:57:05.787722-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "Register Kits To Package", "comments": "", "lab_code": "", "step_number": ""}, "description": "Register Kits To Package", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "create_child": {"actions": {"action/workflow_step_accessioning/create_child_container_and_link_child_workflow_step/1.0/": {"deactivate_when_executed": ["action/core/print_barcode_label/1.0"]}}, "group_name": "Acc Actions", "group_order": "2"}, "acc-queue-move": {"actions": {"action/move-queues/*/1.0/": {}}, "group_name": "Move Queues", "group_order": "3"}}, "expected_inputs": [], "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:05.787722-07	t
19e90d71-6ca5-4337-9611-dcaefae94281	GT85	accessioning-steps:kit-generated:1.0	WSX	workflow_step_template	workflow_step	accessioning-steps	kit-generated	2025-07-03 05:57:05.795953-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "Register Specimen Containers To Kit", "comments": "", "lab_code": "", "step_number": "", "carrier_name": "", "equipment_euid": [], "carrier_tracking_number": ""}, "description": "Register Specimen Containers To Kit", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "specimen_actions": {"actions": {"action/workflow_step_accessioning/create_child_container_and_link_child_workflow_step_specimen/1.0/": {}}, "group_name": "Specimen Actions", "group_order": "2"}}, "expected_inputs": [], "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "expected_outputs": [{"container/package/generic/*/": {"json_addl": {"properties": {"carrier_name": "", "carrier_tracking_number": ""}}}}], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:05.795953-07	t
990367b0-c995-4ac3-9d8a-527eb85151b8	GT86	accessioning-steps:container-generated:1.0	WSX	workflow_step_template	workflow_step	accessioning-steps	container-generated	2025-07-03 05:57:05.802888-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "Register Specimen Containers To Test Req // Assay", "comments": "", "lab_code": "", "step_number": ""}, "description": "Register Specimen Container To Test Req // Assay", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "test_req": {"actions": {"action/workflow_step_accessioning/create_test_req_and_link_child_workflow_step_dup/1.0/": {}}, "group_name": "Acc Actions", "group_order": "2"}}, "expected_inputs": [{"container/kit/generic/*/": {}}], "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "expected_outputs": [{"container/tube/tube-strek-10ml/*/": {"json_addl": {"properties": {"comments": ""}}}}], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:05.802888-07	t
dd00cc24-18ce-4b4c-a6a1-c7ac503f2ec1	GT87	accessioning-steps:test-requisition-generated2:1.0	WSX	workflow_step_template	workflow_step	accessioning-steps	test-requisition-generated2	2025-07-03 05:57:05.809117-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "Associate Other Specimen Containers To Test Requisition", "comments": "", "lab_code": "", "step_number": ""}, "description": "Associate Other Specimen Containers To Test Requisition", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "test_req": {"actions": {"action/workflow_step_accessioning/ycreate_test_req_and_link_child_workflow_step2/1.0/": {}}, "group_name": "Test Requisition Actions", "group_order": "2"}}, "expected_inputs": [{"container/kit/generic/*/": {}}], "step_properties": {"name": "", "comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "expected_outputs": [{"container/tube/tube-strek-10ml/*/": {"json_addl": {"properties": {"comments": ""}}}}], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:05.809117-07	t
9a550109-ad1e-4afd-b127-3ee9378a542f	GT89	analysis:gdna-quantified:1.0	WSX	workflow_step_template	workflow_step	analysis	gdna-quantified	2025-07-03 05:57:06.265499-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "DNA QUANTIFIED", "comments": "", "lab_code": "", "step_number": ""}, "description": "DNA QUANTIFIED", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [{"container/tube/tube-strek-10ml/*/": {"holds_content": "content/specimen/blood-whole/*"}}, {"container/tube/tube-strek-10ml/*/": {"holds_content": "content/specimen/blood-whole/*"}}, {"container/tube/tube-strek-10ml/*/": {"holds_content": "content/specimen/blood-whole/*"}}, {"container/tube/tube-strek-10ml/*/": {"holds_content": "content/specimen/blood-whole/*"}}], "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "expected_outputs": [{"container/tube/tube-generic-10ml/*/": {"holds_content": "content/sample/blood-plasma/*"}}, {"container/tube/tube-generic-10ml/*/": {"holds_content": "content/sample/blood-plasma/*"}}, {"container/tube/tube-generic-10ml/*/": {"holds_content": "content/sample/blood-plasma/*"}}, {"container/tube/tube-generic-10ml/*/": {"holds_content": "content/sample/blood-plasma/*"}}], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:06.265499-07	t
b7eaddd6-0201-4579-b422-95af5dbb0805	GT90	destroy-steps:specimen-containers-destroyed:1.0	WSX	workflow_step_template	workflow_step	destroy-steps	specimen-containers-destroyed	2025-07-03 05:57:06.697285-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "Specimens Destroyed", "comments": "", "lab_code": "", "step_number": ""}, "description": "Specimens Destroyed", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "destroyed": {"actions": {"action/workflow_step_destroy/complete_specimen_destroy/*/1.0/": {}}, "group_name": "Destroy Actions", "group_order": "2"}}, "expected_inputs": [], "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:06.697285-07	t
3862d63d-1127-4727-af80-064ff9c105e2	GT91	exception:specimen-containers-destroyed:1.0	WSX	workflow_step_template	workflow_step	exception	specimen-containers-destroyed	2025-07-03 05:57:07.119389-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "Desctroyed Containers Exceptions", "comments": "", "lab_code": "", "step_number": ""}, "description": "Destroyed Containers Exceptions", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:07.119389-07	t
34966c45-2cd1-485a-b2db-cfb932024bed	GT92	extraction-steps:loading:1.0	WSX	workflow_step_template	workflow_step	extraction-steps	loading	2025-07-03 05:57:07.541107-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "Loading blood-whole in tube-strek-10ml tubes.", "comments": "", "lab_code": "", "step_number": ""}, "description": "Loading blood-whole in tube-strek-10ml tubes.", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [{"container/tube/tube-strek-10ml/*/": {"holds_content": "content/specimen/blood-whole/*"}}, {"container/tube/tube-strek-10ml/*/": {"holds_content": "content/specimen/blood-whole/*"}}, {"container/tube/tube-strek-10ml/*/": {"holds_content": "content/specimen/blood-whole/*"}}, {"container/tube/tube-strek-10ml/*/": {"holds_content": "content/specimen/blood-whole/*"}}], "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:07.541107-07	t
4f614c77-6499-4b50-bb8b-3bde0c5cd5be	GT93	extraction-steps:gdna-extraction:1.0	WSX	workflow_step_template	workflow_step	extraction-steps	gdna-extraction	2025-07-03 05:57:07.550559-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "Extracting gdna from blood-whole in tubes to wells in a fixed-plate-96.", "comments": "", "lab_code": "", "step_number": ""}, "description": "Extracting gdna from blood-whole in tubes to wells in a fixed-plate-96.", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [{"container/tube/tube-strek-10ml/*/": {"holds_content": "content/specimen/blood-whole/*"}}, {"container/tube/tube-strek-10ml/*/": {"holds_content": "content/specimen/blood-whole/*"}}, {"container/tube/tube-strek-10ml/*/": {"holds_content": "content/specimen/blood-whole/*"}}, {"container/tube/tube-strek-10ml/*/": {"holds_content": "content/specimen/blood-whole/*"}}], "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "expected_outputs": [[{"container/plate/fixed-plate-96/*/": {"child_of_lineages": [{"container/well/fixed-plate-well/*/": {"holds_content": "content/sample/gdna/*"}}]}}, {"container/plate/fixed-plate-96/*/": {"child_of_lineages": [{"container/well/fixed-plate-well/*/": {"holds_content": "content/sample/gdna/*"}}]}}, {"container/plate/fixed-plate-96/*/": {"child_of_lineages": [{"container/well/fixed-plate-well/*/": {"holds_content": "content/sample/gdna/*"}}]}}, {"container/plate/fixed-plate-96/*/": {"child_of_lineages": [{"container/well/fixed-plate-well/*/": {"holds_content": "content/sample/gdna/*"}}]}}]], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:07.550559-07	t
c96906f6-fa81-4c5f-90dc-25a5a099b745	GT94	extraction-steps:cfdna-extracted:1.0	WSX	workflow_step_template	workflow_step	extraction-steps	cfdna-extracted	2025-07-03 05:57:07.557425-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "cfDNA Extraction Plate", "comments": "", "lab_code": "", "step_number": ""}, "description": "cfDNA Extraction Plate ", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "cfdna": {"actions": {"action/workflow_step_extraction/cfdna_post_extraction_quant/1.0/": {}}, "group_name": "cfDNA Actions", "group_order": "2"}}, "expected_inputs": [[{"container/tube/tube-generic-10ml/*/": {"holds_content": "content/sample/blood-plasma/*"}}, {"container/tube/tube-generic-10ml/*/": {"holds_content": "content/sample/blood-plasma/*"}}, {"container/tube/tube-generic-10ml/*/": {"holds_content": "content/sample/blood-plasma/*"}}, {"container/tube/tube-generic-10ml/*/": {"holds_content": "content/sample/blood-plasma/*"}}]], "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "expected_outputs": [[{"container/plate/fixed-plate-96/*/": {"child_of_lineages": [{"container/well/fixed-plate-well/*/": {"holds_content": "content/sample/cfdna/*"}}]}}, {"container/plate/fixed-plate-96/*/": {"child_of_lineages": [{"container/well/fixed-plate-well/*/": {"holds_content": "content/sample/cfdna/*"}}]}}, {"container/plate/fixed-plate-96/*/": {"child_of_lineages": [{"container/well/fixed-plate-well/*/": {"holds_content": "content/sample/cfdna/*"}}]}}, {"container/plate/fixed-plate-96/*/": {"child_of_lineages": [{"container/well/fixed-plate-well/*/": {"holds_content": "content/sample/cfdna/*"}}]}}]], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:07.557425-07	t
13028ffb-87ef-4e09-ba4e-5941405f90b9	GT113	destroy:specimen-containers:1.0	WX	workflow_template	workflow	destroy	specimen-containers	2025-07-03 05:57:10.603883-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"comments": "", "lab_code": ""}, "description": "Destroy Specimen Containers", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "destroy": {"actions": {"action/workflow/destroy_specimen_containers/1.0/": {}}, "group_name": "Destroy Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:10.603883-07	t
6c7c2ad6-3774-42c5-a226-c09b52e011b0	GT95	extraction-steps:plasma-isolated:1.0	WSX	workflow_step_template	workflow_step	extraction-steps	plasma-isolated	2025-07-03 05:57:07.563973-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "plasma isolation", "comments": "", "lab_code": "", "step_number": ""}, "description": "plasma isolation", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [{"container/tube/tube-strek-10ml/*/": {"holds_content": "content/specimen/blood-whole/*"}}, {"container/tube/tube-strek-10ml/*/": {"holds_content": "content/specimen/blood-whole/*"}}, {"container/tube/tube-strek-10ml/*/": {"holds_content": "content/specimen/blood-whole/*"}}, {"container/tube/tube-strek-10ml/*/": {"holds_content": "content/specimen/blood-whole/*"}}], "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "expected_outputs": [{"container/tube/tube-generic-10ml/*/": {"holds_content": "content/sample/blood-plasma/*"}}, {"container/tube/tube-generic-10ml/*/": {"holds_content": "content/sample/blood-plasma/*"}}, {"container/tube/tube-generic-10ml/*/": {"holds_content": "content/sample/blood-plasma/*"}}, {"container/tube/tube-generic-10ml/*/": {"holds_content": "content/sample/blood-plasma/*"}}], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:07.563973-07	t
3c8306f0-1d56-4dae-b063-76cf78fd96dd	GT96	plate-operations:heat-seal:1.0	WSX	workflow_step_template	workflow_step	plate-operations	heat-seal	2025-07-03 05:57:07.985297-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "Heatseal Plate", "comments": "", "lab_code": "", "step_number": ""}, "description": "Heatseal Plate", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [{"container/plate/*/*/": {}}], "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "expected_outputs": [{"container/plate/*/*/": {}}], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:07.985297-07	t
6f7b03b6-fc05-48be-9372-a2480ce99349	GT97	plate-operations:extraction-plate-filled:1.0	WSX	workflow_step_template	workflow_step	plate-operations	extraction-plate-filled	2025-07-03 05:57:07.993607-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "properties": {"name": "Filled Extraction Plate", "comments": "", "lab_code": "", "step_number": ""}, "description": "Filled Extraction Plate", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "eqpt": {"actions": {"action/object/set-child-object/1.0/": {"verify": {"instance_type": "equipment/*/*/*"}, "action_name": "Link Equipment", "verify_data": "yes"}}, "group_name": "Equipment", "group_order": "4"}, "pool": {"actions": {"action/pool/create-pool-from-containers/1.0/": {"verify": {"instance_type": "container/*/*/*"}, "action_name": "Create Pool From Containers", "verify_data": "yes"}}, "group_name": "Pool", "group_order": "5"}, "rgnts": {"actions": {"action/object/set-child-object/1.0/": {"verify": {"instance_type": "container/*/*/*"}, "action_name": "Link Reagent Container", "verify_data": "yes"}}, "group_name": "Add Reagent Container", "group_order": "3"}, "plate_operations": {"actions": {"action/workflow_step_plate_operations/cfdna_quant/1.0/": {}, "action/workflow_step_plate_operations/stamp_copy_plate/1.0/": {}}, "group_name": "Plate Operations", "group_order": "2"}, "stamp-plates-into-plate": {"actions": {"action/plate/stamp_plates_into_plate/1.0/": {"verify": {"instance_type": "container/*/*/*"}, "action_name": "Add Plates to Plate", "verify_data": "yes"}}, "group_name": "Stamp Plates Into Plate", "group_order": "6"}}, "expected_inputs": [{"container/plate/*/*/": {}}], "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "expected_outputs": [{"container/plate/*/*/": {}}], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:07.993607-07	t
806a641e-fc42-4c92-ab2f-25b248c82db4	GT98	plate-operations:stamp-copy-plate:1.0	WSX	workflow_step_template	workflow_step	plate-operations	stamp-copy-plate	2025-07-03 05:57:08.000826-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "Stamp Copy Plate", "comments": "", "lab_code": "", "step_number": ""}, "description": "Stamp Copy Plate", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate_operations": {"actions": {"action/workflow_step_plate_operations/cfdna_quant/1.0/": {}, "action/workflow_step_plate_operations/stamp_copy_plate/1.0/": {}}, "group_name": "Plate Operations", "group_order": "2"}}, "expected_inputs": [], "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:08.000826-07	t
575cfbc5-68d3-40a0-a752-15638c82b871	GT99	plate-operations:plate-added-to:1.0	WSX	workflow_step_template	workflow_step	plate-operations	plate-added-to	2025-07-03 05:57:08.007888-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "Plate Added To", "comments": "", "lab_code": "", "step_number": ""}, "description": "Existing plate, had 1+ plates merged into it.", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "plate_operations": {"actions": {"action/workflow_step_plate_operations/cfdna_quant/1.0/": {}, "action/workflow_step_plate_operations/stamp_copy_plate/1.0/": {}}, "group_name": "Plate Operations", "group_order": "2"}}, "expected_inputs": [], "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:08.007888-07	t
34f5d0b7-534e-45dc-9e95-89b082c9f496	GT100	pool:generic:1.0	WSX	workflow_step_template	workflow_step	pool	generic	2025-07-03 05:57:08.426236-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "generic pool", "comments": "", "lab_code": "", "step_number": ""}, "description": "pool", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:08.426236-07	t
ba6f37f3-857e-430f-b4ea-46be96b43945	GT101	queue:all-purpose:1.0	QX	workflow_step_template	workflow_step	queue	all-purpose	2025-07-03 05:57:08.844981-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "OVERRIDEME", "comments": "", "lab_code": "", "step_number": "OVERRIDEME"}, "description": "OVERRIDEME", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:08.844981-07	t
ce71f5cc-14fb-4bdc-8798-b726f5c74966	GT102	queue:plasma-isolation-queue-available:1.0	QX	workflow_step_template	workflow_step	queue	plasma-isolation-queue-available	2025-07-03 05:57:08.854493-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "Containers available for plasma isolation in this queue.", "comments": "", "lab_code": "", "step_number": ""}, "description": "Containers available for plasma isolation in this queue.", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "tube_xfer": {"actions": {"action/workflow_step_queue/link_tubes_auto/1.0/": {}}, "group_name": "Tube Transfer Actions", "group_order": "2"}}, "expected_inputs": [], "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:08.854493-07	t
c5674a0e-552c-4310-aa7b-92d147f1e22c	GT103	queue:plasma-isolation-queue-removed:1.0	QX	workflow_step_template	workflow_step	queue	plasma-isolation-queue-removed	2025-07-03 05:57:08.861101-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "Containers Removed From Queue", "comments": "", "lab_code": "", "step_number": ""}, "description": "Containers Removed From Queue", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "expected_outputs": [{"container/package/generic/*/": {"json_addl": {"properties": {"carrier_name": "str()", "carrier_tracking_number": "str()"}}}}], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:08.861101-07	t
c719e675-9b0f-4225-b133-86ad974014d2	GT104	queue:plasma-isolation-queue-exception:1.0	QX	workflow_step_template	workflow_step	queue	plasma-isolation-queue-exception	2025-07-03 05:57:08.867425-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "Plasma Isolation Exception Queue", "comments": "", "lab_code": "", "step_number": ""}, "description": "Plasma Isolation Exception Queue", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "test_req_2": {"actions": {"action/workflow_step_queue/link_tubes_auto/1.0/": {}, "action/workflow_step_queue/ycreate_test_req_and_link_child_workflow_step/1.0/": {}}, "group_name": "Test Requisition Actions", "group_order": "2"}}, "expected_inputs": [{"container/kit/generic/*/": {}}], "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "expected_outputs": [{"container/tube/tube-strek-10ml/*/": {"json_addl": {"properties": {"comments": ""}}}}], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:08.867425-07	t
e498236a-d679-4005-8c6f-dcef0cdc44af	GT105	queue:extraction-queue-available:1.0	QX	workflow_step_template	workflow_step	queue	extraction-queue-available	2025-07-03 05:57:08.874083-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "Plasma To Plates For Extraction", "comments": "", "lab_code": "", "step_number": ""}, "description": "Plasma To Plates For Extraction", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "fill_plate": {"actions": {"action/workflow_step_queue/fill_plate_directed/1.0/": {}, "action/workflow_step_queue/fill_plate_undirected/1.0/": {}}, "group_name": "Fill Plate Actions", "group_order": "2"}}, "expected_inputs": [], "step_properties": {"comments": "", "step_number": "", "end_operator": "", "end_timestamp": "", "start_operator": "", "start_timestamp": ""}, "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:08.874083-07	t
6e815277-8197-46dc-8cc3-c0741db7cb31	GT106	accessioning:accession-package-kit-tubes-testreq:1.0	WX	workflow_template	workflow	accessioning	accession-package-kit-tubes-testreq	2025-07-03 05:57:09.302089-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "Carrier Package, Kit, Specimen Container, Test Requisition Accessioning & Specimen Assay Queue Registration", "comments": "", "lab_code": ""}, "description": "Carrier Package, Kit, Specimen Container, Test Requisition Accessioning & Specimen Assay Queue Registration", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "accessioning": {"actions": {"action/workflow/create_package_and_first_workflow_step/1.0/": {}}, "group_name": "Accessioning Actions", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": []}	\N	ready	f	2025-07-03 05:57:09.302089-07	t
ed324835-92b7-4d82-b255-b5909b887a57	GT107	assay-workset:hla-typing:1.2	WX	workflow_template	workflow	assay-workset	hla-typing	2025-07-03 05:57:09.725266-07	1.2	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "Workset for hla-typing 1.2", "comments": "", "lab_code": ""}, "description": "Workset for hla-typing 1.2", "action_groups": {}, "valid_parents": [{"workflow/assay/hla-typing/1.2/": {}}], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "queue-move": {"actions": {"action/move-queues/*/1.0/": {}}, "group_name": "Move Queues", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[]]}	\N	ready	f	2025-07-03 05:57:09.725266-07	t
02ca20b8-6864-4bf7-bae8-cbf445b3324e	GT112	assay:accessioning-RnD:1.0	AY	workflow_template	workflow	assay	accessioning-RnD	2025-07-03 05:57:10.17588-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "1", "assay_type": "accessioning", "properties": {"name": "Accessioning For R&D", "ay_addl": "fedex", "comments": "", "lab_code": ""}, "description": "Accessioning For R&D", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "accessioning": {"actions": {"action/accessioning-ay/*/1.0/": {}}, "group_name": "Package Receipt", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"workflow_step/queue/all-purpose/1.0/": {"json_addl": {"properties": {"name": "In Propgress RnD Accessioning WorkSets", "comments": "", "lab_code": "", "step_number": "1"}, "description": "In Progress RnD Accessioning  WorkSets"}, "description": "In Progress RnD Accessioning WorkSets", "action_imports": {"corei": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core ActionsIMP", "group_order": "1"}}, "assay-queue-accessioning": {"actions": {"action/assay-queue-accessioning/*/1.0/": {}}, "group_name": "Assay Queue Actions", "group_order": "2"}}}, {"workflow_step/queue/plasma-isolation-queue-removed/1.0/": {"json_addl": {"properties": {"name": "Completed RnD Accessioning Worksets", "comments": "", "lab_code": "", "step_number": "2"}, "description": "Completed RnD Accessioning Worksets"}, "description": "Completed RnD Accessioning Worksets"}}, {"workflow_step/queue/plasma-isolation-queue-exception/1.0/": {"json_addl": {"properties": {"name": "Exception RnD Accessioning Queue", "comments": "", "lab_code": "", "step_number": "3"}, "description": "Exception RnD Accessioning Queue"}, "description": "Exception RnD Accessioning Queue"}}]]}	\N	ready	f	2025-07-03 05:57:10.17588-07	t
538ca2de-0f1c-4e0b-ac0f-3aafbd176d9e	GT108	assay-workset:carrier-screen:3.9	WX	workflow_template	workflow	assay-workset	carrier-screen	2025-07-03 05:57:09.735431-07	3.9	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"name": "Workset for carrier-screen 3.9", "comments": "", "lab_code": ""}, "description": "Workset for carrier-screen 3.9", "action_groups": {}, "valid_parents": [{"workflow/assay/carrier-screen/3.9/": {}}], "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "queue-move": {"actions": {"action/move-queues/*/1.0/": {}}, "group_name": "Move Queues", "group_order": "3"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"workflow_step/queue/plasma-isolation-queue-available/1.0/": {"json_addl": {"properties": {"comments": "", "step_number": "1"}}}}, {"workflow_step/queue/plasma-isolation-queue-removed/1.0/": {"json_addl": {"properties": {"comments": "", "step_number": "2"}}}}, {"workflow_step/queue/plasma-isolation-queue-exception/1.0/": {"json_addl": {"properties": {"comments": "", "step_number": "3"}}}}]]}	\N	ready	f	2025-07-03 05:57:09.735431-07	t
d9a37828-c1d8-44f0-979b-e293cb521c30	GT109	assay:hla-typing:1.2	AY	workflow_template	workflow	assay	hla-typing	2025-07-03 05:57:10.154217-07	1.2	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "1", "assay_type": "assay", "properties": {"name": "HLA Typing v1.2", "comments": "", "lab_code": ""}, "description": "HLA Typing v1.2", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"workflow_step/queue/plasma-isolation-queue-available/1.0/": {"json_addl": {"properties": {"name": "Ready For Plasma Isolation Queue", "comments": "", "lab_code": "", "step_number": "1"}, "description": "Ready For Plasma Isolation Queue"}, "description": "Ready For Plasma Isolation Queue"}}, {"workflow_step/queue/all-purpose/1.0/": {"json_addl": {"properties": {"name": "In Propgress / Active WorkSets", "comments": "", "lab_code": "", "step_number": "2"}, "description": "In Progress / Active  WorkSets"}, "description": "In Progress / Active WorkSets"}}, {"workflow_step/queue/plasma-isolation-queue-removed/1.0/": {"json_addl": {"properties": {"name": "Completed WorkSets", "comments": "", "lab_code": "", "step_number": "3"}, "description": "Completed WorkSets"}, "description": "Completed WorkSets"}}, {"workflow_step/queue/plasma-isolation-queue-exception/1.0/": {"json_addl": {"properties": {"name": "Exception Queue of WorkSets", "comments": "", "lab_code": "", "step_number": "4"}, "description": "Exception Queue of WorkSets"}, "description": "Exception Queue of WorkSets"}}]]}	\N	ready	f	2025-07-03 05:57:10.154217-07	t
1251878d-5c20-490d-b895-0349b9ae827b	GT110	assay:carrier-screen:3.9	AY	workflow_template	workflow	assay	carrier-screen	2025-07-03 05:57:10.162593-07	3.9	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "1", "assay_type": "assay", "properties": {"name": "Carrier Screen 3.9", "comments": "", "lab_code": ""}, "description": "Carrier Screen 3.9", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"workflow_step/queue/plasma-isolation-queue-available/1.0/": {"json_addl": {"properties": {"name": "Ready For Plasma Isolation Queue", "comments": "", "lab_code": "", "step_number": "1"}, "description": "Ready For Plasma Isolation Queue"}, "description": "Ready For Plasma Isolation Queue"}}, {"workflow_step/queue/all-purpose/1.0/": {"json_addl": {"properties": {"name": "In Propgress / Active WorkSets", "comments": "", "lab_code": "", "step_number": "2"}, "description": "In Progress / Active  WorkSets"}, "description": "In Progress / Active WorkSets"}}, {"workflow_step/queue/plasma-isolation-queue-removed/1.0/": {"json_addl": {"properties": {"name": "Completed WorkSets", "comments": "", "lab_code": "", "step_number": "3"}, "description": "Completed WorkSets"}, "description": "Completed WorkSets"}}, {"workflow_step/queue/plasma-isolation-queue-exception/1.0/": {"json_addl": {"properties": {"name": "Exception Queue of WorkSets", "comments": "", "lab_code": "", "step_number": "4"}, "description": "Exception Queue of WorkSets"}, "description": "Exception Queue of WorkSets"}}]]}	\N	ready	f	2025-07-03 05:57:10.162593-07	t
a3e7194e-e1f3-47ab-9c88-8ffd6d8cbfd0	GT111	assay:accessioning-clinical:1.0	AY	workflow_template	workflow	assay	accessioning-clinical	2025-07-03 05:57:10.16952-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "1", "assay_type": "accessioning", "properties": {"name": "Accessioning - Clinical", "ay_addl": "fedex", "comments": "", "lab_code": ""}, "description": "Accessioning - Clinical", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}, "accessioning": {"actions": {"action/accessioning-ay/*/1.0/": {}}, "group_name": "Package Receipt", "group_order": "2"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[{"workflow_step/queue/all-purpose/1.0/": {"json_addl": {"properties": {"name": "In Propgress Prod Accessioning WorkSets", "comments": "", "lab_code": "", "step_number": "1"}, "description": "In Progress Prod Accessioning  WorkSets"}, "description": "In Progress Prod Accessioning WorkSets"}}, {"workflow_step/queue/plasma-isolation-queue-removed/1.0/": {"json_addl": {"properties": {"name": "Completed Prod Accessioning Worksets", "comments": "", "lab_code": "", "step_number": "2"}, "description": "Completed Prod Accessioning Worksets"}, "description": "Completed Prod Accessioning Worksets"}}, {"workflow_step/queue/plasma-isolation-queue-exception/1.0/": {"json_addl": {"properties": {"name": "Exception Prod Accessioning Queue", "comments": "", "lab_code": "", "step_number": "3"}, "description": "Exception Prod Accessioning Queue"}, "description": "Exception Prod Accessioning Queue"}}]]}	\N	ready	f	2025-07-03 05:57:10.16952-07	t
ce435b29-2c59-4e13-96c6-00e1b8c50b37	GT114	extraction:blood-whole-to-gdna:1.0	WX	workflow_template	workflow	extraction	blood-whole-to-gdna	2025-07-03 05:57:11.0287-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"comments": "", "lab_code": ""}, "description": "DNA extraction from blood-whole in tube-strek-10ml tubes to a fixed-plate-96 with gdna in each well.", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[]]}	\N	ready	f	2025-07-03 05:57:11.0287-07	t
7f795301-5b20-47bf-9656-9073ca292af6	GT115	extraction:blood-whole-to-plasma-to-cfdna:1.0	WX	workflow_template	workflow	extraction	blood-whole-to-plasma-to-cfdna	2025-07-03 05:57:11.03695-07	1.0	{"cogs": {"cost": "0.00", "state": "active", "allocation_type": "", "cost_split_by_children": [{"*/*/*/*": {}}]}, "singleton": "0", "properties": {"comments": "", "lab_code": ""}, "description": "DNA extraction from blood-whole in tube-strek-10ml tubes to a fixed-plate-96 with gdna in each well.", "action_groups": {}, "action_imports": {"core": {"actions": {"action/core/*/1.0/": {}}, "group_name": "Core Actions", "group_order": "1"}}, "expected_inputs": [], "expected_outputs": [], "instantiation_layouts": [[]]}	\N	ready	f	2025-07-03 05:57:11.03695-07	t
\.


--
-- Name: audit_log_seq; Type: SEQUENCE SET; Schema: public; Owner: daylily
--

SELECT pg_catalog.setval('public.audit_log_seq', 1, false);


--
-- Name: ax_instance_seq; Type: SEQUENCE SET; Schema: public; Owner: daylily
--

SELECT pg_catalog.setval('public.ax_instance_seq', 1, false);


--
-- Name: ay_instance_seq; Type: SEQUENCE SET; Schema: public; Owner: daylily
--

SELECT pg_catalog.setval('public.ay_instance_seq', 1, false);


--
-- Name: cwx_instance_seq; Type: SEQUENCE SET; Schema: public; Owner: daylily
--

SELECT pg_catalog.setval('public.cwx_instance_seq', 308, true);


--
-- Name: cx_instance_seq; Type: SEQUENCE SET; Schema: public; Owner: daylily
--

SELECT pg_catalog.setval('public.cx_instance_seq', 23, true);


--
-- Name: dx_instance_seq; Type: SEQUENCE SET; Schema: public; Owner: daylily
--

SELECT pg_catalog.setval('public.dx_instance_seq', 1, false);


--
-- Name: ev_instance_seq; Type: SEQUENCE SET; Schema: public; Owner: daylily
--

SELECT pg_catalog.setval('public.ev_instance_seq', 1, false);


--
-- Name: ex_instance_seq; Type: SEQUENCE SET; Schema: public; Owner: daylily
--

SELECT pg_catalog.setval('public.ex_instance_seq', 1, false);


--
-- Name: fi_instance_seq; Type: SEQUENCE SET; Schema: public; Owner: daylily
--

SELECT pg_catalog.setval('public.fi_instance_seq', 7112, true);


--
-- Name: fs_instance_seq; Type: SEQUENCE SET; Schema: public; Owner: daylily
--

SELECT pg_catalog.setval('public.fs_instance_seq', 26, true);


--
-- Name: fx_instance_seq; Type: SEQUENCE SET; Schema: public; Owner: daylily
--

SELECT pg_catalog.setval('public.fx_instance_seq', 9, true);


--
-- Name: generic_instance_lineage_seq; Type: SEQUENCE SET; Schema: public; Owner: daylily
--

SELECT pg_catalog.setval('public.generic_instance_lineage_seq', 762, true);


--
-- Name: generic_instance_seq; Type: SEQUENCE SET; Schema: public; Owner: daylily
--

SELECT pg_catalog.setval('public.generic_instance_seq', 1, false);


--
-- Name: generic_template_seq; Type: SEQUENCE SET; Schema: public; Owner: daylily
--

SELECT pg_catalog.setval('public.generic_template_seq', 115, true);


--
-- Name: gx_instance_seq; Type: SEQUENCE SET; Schema: public; Owner: daylily
--

SELECT pg_catalog.setval('public.gx_instance_seq', 1, false);


--
-- Name: mcxq_instance_seq; Type: SEQUENCE SET; Schema: public; Owner: daylily
--

SELECT pg_catalog.setval('public.mcxq_instance_seq', 9, true);


--
-- Name: mrxq_instance_seq; Type: SEQUENCE SET; Schema: public; Owner: daylily
--

SELECT pg_catalog.setval('public.mrxq_instance_seq', 308, true);


--
-- Name: mx_instance_seq; Type: SEQUENCE SET; Schema: public; Owner: daylily
--

SELECT pg_catalog.setval('public.mx_instance_seq', 1, false);


--
-- Name: trx_instance_seq; Type: SEQUENCE SET; Schema: public; Owner: daylily
--

SELECT pg_catalog.setval('public.trx_instance_seq', 1, false);


--
-- Name: wsq_instance_seq; Type: SEQUENCE SET; Schema: public; Owner: daylily
--

SELECT pg_catalog.setval('public.wsq_instance_seq', 1, false);


--
-- Name: wsx_instance_seq; Type: SEQUENCE SET; Schema: public; Owner: daylily
--

SELECT pg_catalog.setval('public.wsx_instance_seq', 1, false);


--
-- Name: wx_instance_seq; Type: SEQUENCE SET; Schema: public; Owner: daylily
--

SELECT pg_catalog.setval('public.wx_instance_seq', 1, false);


--
-- Name: xx_instance_seq; Type: SEQUENCE SET; Schema: public; Owner: daylily
--

SELECT pg_catalog.setval('public.xx_instance_seq', 1, false);


--
-- Name: audit_log audit_log_pkey; Type: CONSTRAINT; Schema: public; Owner: daylily
--

ALTER TABLE ONLY public.audit_log
    ADD CONSTRAINT audit_log_pkey PRIMARY KEY (uuid);


--
-- Name: generic_instance generic_instance_euid_key; Type: CONSTRAINT; Schema: public; Owner: daylily
--

ALTER TABLE ONLY public.generic_instance
    ADD CONSTRAINT generic_instance_euid_key UNIQUE (euid);


--
-- Name: generic_instance_lineage generic_instance_lineage_euid_key; Type: CONSTRAINT; Schema: public; Owner: daylily
--

ALTER TABLE ONLY public.generic_instance_lineage
    ADD CONSTRAINT generic_instance_lineage_euid_key UNIQUE (euid);


--
-- Name: generic_instance_lineage generic_instance_lineage_pkey; Type: CONSTRAINT; Schema: public; Owner: daylily
--

ALTER TABLE ONLY public.generic_instance_lineage
    ADD CONSTRAINT generic_instance_lineage_pkey PRIMARY KEY (uuid);


--
-- Name: generic_instance generic_instance_pkey; Type: CONSTRAINT; Schema: public; Owner: daylily
--

ALTER TABLE ONLY public.generic_instance
    ADD CONSTRAINT generic_instance_pkey PRIMARY KEY (uuid);


--
-- Name: generic_template generic_template_euid_key; Type: CONSTRAINT; Schema: public; Owner: daylily
--

ALTER TABLE ONLY public.generic_template
    ADD CONSTRAINT generic_template_euid_key UNIQUE (euid);


--
-- Name: generic_template generic_template_pkey; Type: CONSTRAINT; Schema: public; Owner: daylily
--

ALTER TABLE ONLY public.generic_template
    ADD CONSTRAINT generic_template_pkey PRIMARY KEY (uuid);


--
-- Name: generic_template unique_super_type_btype_b_sub_type_version; Type: CONSTRAINT; Schema: public; Owner: daylily
--

ALTER TABLE ONLY public.generic_template
    ADD CONSTRAINT unique_super_type_btype_b_sub_type_version UNIQUE (super_type, btype, b_sub_type, version);


--
-- Name: idx_audit_log_changed_at; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_audit_log_changed_at ON public.audit_log USING btree (changed_at);


--
-- Name: idx_audit_log_changed_by; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_audit_log_changed_by ON public.audit_log USING btree (changed_by);


--
-- Name: idx_audit_log_is_deleted; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_audit_log_is_deleted ON public.audit_log USING btree (is_deleted);


--
-- Name: idx_audit_log_json_addl_gin; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_audit_log_json_addl_gin ON public.audit_log USING gin (json_addl);


--
-- Name: idx_audit_log_operation_type; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_audit_log_operation_type ON public.audit_log USING btree (operation_type);


--
-- Name: idx_audit_log_rel_table_euid_fk; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_audit_log_rel_table_euid_fk ON public.audit_log USING btree (rel_table_euid_fk);


--
-- Name: idx_audit_log_rel_table_name; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_audit_log_rel_table_name ON public.audit_log USING btree (rel_table_name);


--
-- Name: idx_audit_log_rel_table_uuid_fk; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_audit_log_rel_table_uuid_fk ON public.audit_log USING btree (rel_table_uuid_fk);


--
-- Name: idx_generic_instance_b_sub_type; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_generic_instance_b_sub_type ON public.generic_instance USING btree (b_sub_type);


--
-- Name: idx_generic_instance_composite; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_generic_instance_composite ON public.generic_template USING btree (super_type, btype, b_sub_type, version, is_deleted);


--
-- Name: idx_generic_instance_euid; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_generic_instance_euid ON public.generic_instance USING btree (euid);


--
-- Name: idx_generic_instance_is_deleted; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_generic_instance_is_deleted ON public.generic_instance USING btree (is_deleted);


--
-- Name: idx_generic_instance_json_addl_gin; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_generic_instance_json_addl_gin ON public.generic_instance USING gin (json_addl);


--
-- Name: idx_generic_instance_lineage_child_uuid; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_generic_instance_lineage_child_uuid ON public.generic_instance_lineage USING btree (child_instance_uuid);


--
-- Name: idx_generic_instance_lineage_euid; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_generic_instance_lineage_euid ON public.generic_instance_lineage USING btree (euid);


--
-- Name: idx_generic_instance_lineage_is_deleted; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_generic_instance_lineage_is_deleted ON public.generic_instance_lineage USING btree (is_deleted);


--
-- Name: idx_generic_instance_lineage_json_addl_gin; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_generic_instance_lineage_json_addl_gin ON public.generic_instance_lineage USING gin (json_addl);


--
-- Name: idx_generic_instance_lineage_mod_dt; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_generic_instance_lineage_mod_dt ON public.generic_instance_lineage USING btree (modified_dt);


--
-- Name: idx_generic_instance_lineage_parent_uuid; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_generic_instance_lineage_parent_uuid ON public.generic_instance_lineage USING btree (parent_instance_uuid);


--
-- Name: idx_generic_instance_lineage_polymorphic_discriminator; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_generic_instance_lineage_polymorphic_discriminator ON public.generic_instance_lineage USING btree (polymorphic_discriminator);


--
-- Name: idx_generic_instance_lineage_relationship_type; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_generic_instance_lineage_relationship_type ON public.generic_instance_lineage USING btree (relationship_type);


--
-- Name: idx_generic_instance_mod_df; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_generic_instance_mod_df ON public.generic_instance USING btree (modified_dt);


--
-- Name: idx_generic_instance_polymorphic_discriminator; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_generic_instance_polymorphic_discriminator ON public.generic_instance USING btree (polymorphic_discriminator);


--
-- Name: idx_generic_instance_singleton; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_generic_instance_singleton ON public.generic_instance USING btree (is_singleton);


--
-- Name: idx_generic_instance_super_type; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_generic_instance_super_type ON public.generic_instance USING btree (super_type);


--
-- Name: idx_generic_instance_template_uuid; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_generic_instance_template_uuid ON public.generic_instance USING btree (template_uuid);


--
-- Name: idx_generic_instance_type; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_generic_instance_type ON public.generic_instance USING btree (btype);


--
-- Name: idx_generic_instance_version; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_generic_instance_version ON public.generic_instance USING btree (version);


--
-- Name: idx_generic_template_b_sub_type; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_generic_template_b_sub_type ON public.generic_template USING btree (b_sub_type);


--
-- Name: idx_generic_template_composite; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_generic_template_composite ON public.generic_template USING btree (super_type, btype, b_sub_type, version, is_deleted);


--
-- Name: idx_generic_template_euid; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_generic_template_euid ON public.generic_template USING btree (euid);


--
-- Name: idx_generic_template_instance_prefix; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_generic_template_instance_prefix ON public.generic_template USING btree (instance_prefix);


--
-- Name: idx_generic_template_is_deleted; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_generic_template_is_deleted ON public.generic_template USING btree (is_deleted);


--
-- Name: idx_generic_template_json_addl_gin; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_generic_template_json_addl_gin ON public.generic_template USING gin (json_addl);


--
-- Name: idx_generic_template_mod_df; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_generic_template_mod_df ON public.generic_template USING btree (modified_dt);


--
-- Name: idx_generic_template_polymorphic_discriminator; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_generic_template_polymorphic_discriminator ON public.generic_template USING btree (polymorphic_discriminator);


--
-- Name: idx_generic_template_singleton; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_generic_template_singleton ON public.generic_template USING btree (is_singleton);


--
-- Name: idx_generic_template_super_type; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_generic_template_super_type ON public.generic_template USING btree (super_type);


--
-- Name: idx_generic_template_type; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_generic_template_type ON public.generic_template USING btree (btype);


--
-- Name: idx_generic_template_version; Type: INDEX; Schema: public; Owner: daylily
--

CREATE INDEX idx_generic_template_version ON public.generic_template USING btree (version);


--
-- Name: idx_genric_instance_unique_singleton_key; Type: INDEX; Schema: public; Owner: daylily
--

CREATE UNIQUE INDEX idx_genric_instance_unique_singleton_key ON public.generic_instance USING btree (super_type, btype, b_sub_type, version) WHERE (is_singleton = true);


--
-- Name: idx_genric_template_unique_singleton_key; Type: INDEX; Schema: public; Owner: daylily
--

CREATE UNIQUE INDEX idx_genric_template_unique_singleton_key ON public.generic_template USING btree (super_type, btype, b_sub_type, version) WHERE (is_singleton = true);


--
-- Name: generic_instance generic_instance_audit_trigger_after_delete; Type: TRIGGER; Schema: public; Owner: daylily
--

CREATE TRIGGER generic_instance_audit_trigger_after_delete AFTER DELETE ON public.generic_instance FOR EACH ROW EXECUTE FUNCTION public.record_delete();


--
-- Name: generic_instance generic_instance_audit_trigger_after_insert; Type: TRIGGER; Schema: public; Owner: daylily
--

CREATE TRIGGER generic_instance_audit_trigger_after_insert AFTER INSERT ON public.generic_instance FOR EACH ROW EXECUTE FUNCTION public.record_insert();


--
-- Name: generic_instance generic_instance_audit_trigger_after_update; Type: TRIGGER; Schema: public; Owner: daylily
--

CREATE TRIGGER generic_instance_audit_trigger_after_update AFTER UPDATE ON public.generic_instance FOR EACH ROW EXECUTE FUNCTION public.record_update();


--
-- Name: generic_instance_lineage generic_instance_lineage_audit_trigger_after_delete; Type: TRIGGER; Schema: public; Owner: daylily
--

CREATE TRIGGER generic_instance_lineage_audit_trigger_after_delete AFTER DELETE ON public.generic_instance_lineage FOR EACH ROW EXECUTE FUNCTION public.record_delete();


--
-- Name: generic_instance_lineage generic_instance_lineage_audit_trigger_after_insert; Type: TRIGGER; Schema: public; Owner: daylily
--

CREATE TRIGGER generic_instance_lineage_audit_trigger_after_insert AFTER INSERT ON public.generic_instance_lineage FOR EACH ROW EXECUTE FUNCTION public.record_insert();


--
-- Name: generic_instance_lineage generic_instance_lineage_audit_trigger_after_update; Type: TRIGGER; Schema: public; Owner: daylily
--

CREATE TRIGGER generic_instance_lineage_audit_trigger_after_update AFTER UPDATE ON public.generic_instance_lineage FOR EACH ROW EXECUTE FUNCTION public.record_update();


--
-- Name: generic_instance_lineage generic_instance_lineage_soft_delete; Type: TRIGGER; Schema: public; Owner: daylily
--

CREATE TRIGGER generic_instance_lineage_soft_delete BEFORE DELETE ON public.generic_instance_lineage FOR EACH ROW EXECUTE FUNCTION public.soft_delete_row();


--
-- Name: generic_template generic_template_audit_trigger_after_delete; Type: TRIGGER; Schema: public; Owner: daylily
--

CREATE TRIGGER generic_template_audit_trigger_after_delete AFTER DELETE ON public.generic_template FOR EACH ROW EXECUTE FUNCTION public.record_delete();


--
-- Name: generic_template generic_template_audit_trigger_after_insert; Type: TRIGGER; Schema: public; Owner: daylily
--

CREATE TRIGGER generic_template_audit_trigger_after_insert AFTER INSERT ON public.generic_template FOR EACH ROW EXECUTE FUNCTION public.record_insert();


--
-- Name: generic_template generic_template_audit_trigger_after_update; Type: TRIGGER; Schema: public; Owner: daylily
--

CREATE TRIGGER generic_template_audit_trigger_after_update AFTER UPDATE ON public.generic_template FOR EACH ROW EXECUTE FUNCTION public.record_update();


--
-- Name: generic_instance trigger_generic_instance_soft_delete; Type: TRIGGER; Schema: public; Owner: daylily
--

CREATE TRIGGER trigger_generic_instance_soft_delete BEFORE DELETE ON public.generic_instance FOR EACH ROW EXECUTE FUNCTION public.soft_delete_row();


--
-- Name: generic_template trigger_generic_template_soft_delete; Type: TRIGGER; Schema: public; Owner: daylily
--

CREATE TRIGGER trigger_generic_template_soft_delete BEFORE DELETE ON public.generic_template FOR EACH ROW EXECUTE FUNCTION public.soft_delete_row();


--
-- Name: generic_instance trigger_set_generic_instance_euid; Type: TRIGGER; Schema: public; Owner: daylily
--

CREATE TRIGGER trigger_set_generic_instance_euid BEFORE INSERT ON public.generic_instance FOR EACH ROW EXECUTE FUNCTION public.set_generic_instance_euid();


--
-- Name: generic_instance_lineage update_generic_instance_lineage_mod_dt; Type: TRIGGER; Schema: public; Owner: daylily
--

CREATE TRIGGER update_generic_instance_lineage_mod_dt BEFORE UPDATE ON public.generic_instance_lineage FOR EACH ROW EXECUTE FUNCTION public.update_modified_dt();


--
-- Name: generic_instance update_generic_instance_mod_dt; Type: TRIGGER; Schema: public; Owner: daylily
--

CREATE TRIGGER update_generic_instance_mod_dt BEFORE UPDATE ON public.generic_instance FOR EACH ROW EXECUTE FUNCTION public.update_modified_dt();


--
-- Name: generic_template update_generic_template_mod_dt; Type: TRIGGER; Schema: public; Owner: daylily
--

CREATE TRIGGER update_generic_template_mod_dt BEFORE UPDATE ON public.generic_template FOR EACH ROW EXECUTE FUNCTION public.update_modified_dt();


--
-- Name: generic_instance_lineage generic_instance_lineage_child_instance_uuid_fkey; Type: FK CONSTRAINT; Schema: public; Owner: daylily
--

ALTER TABLE ONLY public.generic_instance_lineage
    ADD CONSTRAINT generic_instance_lineage_child_instance_uuid_fkey FOREIGN KEY (child_instance_uuid) REFERENCES public.generic_instance(uuid);


--
-- Name: generic_instance_lineage generic_instance_lineage_parent_instance_uuid_fkey; Type: FK CONSTRAINT; Schema: public; Owner: daylily
--

ALTER TABLE ONLY public.generic_instance_lineage
    ADD CONSTRAINT generic_instance_lineage_parent_instance_uuid_fkey FOREIGN KEY (parent_instance_uuid) REFERENCES public.generic_instance(uuid);


--
-- Name: generic_instance generic_instance_template_uuid_fkey; Type: FK CONSTRAINT; Schema: public; Owner: daylily
--

ALTER TABLE ONLY public.generic_instance
    ADD CONSTRAINT generic_instance_template_uuid_fkey FOREIGN KEY (template_uuid) REFERENCES public.generic_template(uuid);


--
-- PostgreSQL database dump complete
--

